import { uuid, isString } from 'substance'
import { qualifiedId as _qualifiedId, applyCellTransformations } from './engineHelpers'
import Cell from './Cell'

/*
  Engine's internal model of a Document.
*/
export default class Document {
  constructor (engine, data) {
    this.engine = engine
    this.data = data
    if (!data.id) throw new Error("'id' is required")
    this.id = data.id
    this.name = data.name
    this.lang = data.lang || 'mini'
    if (data.hasOwnProperty('autorun')) {
      this.autorun = data.autorun
    } else {
      // TODO: using manual execution as a default for now
      this.autorun = true
    }
    this.cells = data.cells.map(cellData => this._createCell(cellData))
    // registration hook used for propagating initial cell state to the application
    if (data.onCellRegister) this.onCellRegister = data.onCellRegister
  }

  get type () { return 'document' }

  getCells () {
    return this.cells
  }

  setAutorun (val) {
    this.autorun = val
  }

  insertCellAt (pos, cellData) {
    let cell = this._createCell(cellData)
    this._registerCell(cell)
    this.cells.splice(pos, 0, cell)
    return cell
  }

  removeCell (id) {
    const qualifiedId = _qualifiedId(this.id, id)
    const cells = this.cells
    let pos = cells.findIndex(cell => cell.id === qualifiedId)
    if (pos >= 0) {
      let cell = cells[pos]
      this.cells.splice(pos, 1)
      this.engine._unregisterCell(cell)
    } else {
      console.error('Unknown cell', id)
    }
  }

  updateCell (id, cellData) {
    let qualifiedId = _qualifiedId(this.id, id)
    if (isString(cellData)) {
      cellData = { source: cellData }
    }
    this.engine._updateCell(qualifiedId, cellData)
  }

  rename (newName) {
    if (newName === this.name) return
    const docId = this.id
    let graph = this.engine._graph
    let cells = this.cells
    let affectedCells = new Set()
    for (let i = 0; i < cells.length; i++) {
      let cell = cells[i]
      if (graph._cellProvidesOutput(cell)) {
        let ids = graph._ins[cell.output]
        if (ids) {
          ids.forEach(cellId => {
            let cell = graph.getCell(cellId)
            cell.inputs.forEach(s => {
              if (s.docId === docId) {
                s._update = { type: 'rename', scope: newName }
              }
            })
            affectedCells.add(cell)
          })
        }
      }
    }
    affectedCells.forEach(cell => {
      applyCellTransformations(cell)
      this.engine._resetCell(cell)
    })
    this.name = newName
    this._sendSourceUpdate(affectedCells)
  }

  onCellRegister(cell) { // eslint-disable-line
  }

  dump () {
    let cells = this.cells.map(cell => cell.dump())
    return {
      type: 'document',
      id: this.id,
      name: this.name,
      lang: this.lang,
      autorun: this.autorun,
      cells
    }
  }

  _createCell (cellData) {
    if (isString(cellData)) {
      let source = cellData
      cellData = {
        id: uuid(),
        docId: this.id,
        source,
        lang: this.lang
      }
    }
    return new Cell(this, cellData)
  }

  _registerCell (cell) {
    const engine = this.engine
    engine._registerCell(cell)
    this.onCellRegister(cell)
  }

  _registerCells (block) {
    if (!block) block = this.cells
    block.forEach(cell => this._registerCell(cell))
  }

  _sendSourceUpdate (cells) {
    if (cells.size > 0) {
      this.engine._sendUpdate('source', cells)
    }
  }
}
