import { isArray, tableHelpers } from 'substance'
import { JavascriptContext } from 'stencila-js'
import { Engine, cellStateToString, MiniContext, parseSymbol } from '../index'
import { libtest } from './libtest'
import TestContext from './TestContext'

const { getRowCol, getIndexesFromRange, getRangeFromMatrix } = tableHelpers

export function setupEngine () {
  let context = new TestContext()
  context.configure({
    contexts: [
      { id: 'mickey', lang: 'mini', client: MiniContext },
      { id: 'goofy', lang: 'js', client: JavascriptContext }
    ],
    libraries: [
      { lang: 'js', lib: libtest }
    ]
  })
  let engine = new Engine(context)
  // EXPERIMENTAL: register all library content as globals
  let jsContext = context.getLanguageContext('js')
  let miniContext = context.getLanguageContext('mini')
  let names = Object.keys(libtest.funcs)
  names.forEach(name => {
    // TODO: do we want that extra level here?
    // need to discuss if the function type could
    // be simplified
    engine._addGlobal(name, {
      type: 'function',
      data: {
        name,
        library: libtest.name,
        context: jsContext.id
      }
    })
  })

  let graph = engine._graph
  return { engine, context, graph, miniContext, jsContext }
}

export function getValue (cell) {
  if (cell.value) {
    return cell.value.data
  }
}

export function getValues (cells) {
  return cells.map(rowOrCell => {
    if (isArray(rowOrCell)) {
      return rowOrCell.map(getValue)
    } else {
      return getValue(rowOrCell)
    }
  })
}

export function getSource (cell) {
  return cell.source
}

export function getSources (cells) {
  return cells.map(rowOrCell => {
    if (isArray(rowOrCell)) {
      return rowOrCell.map(getSource)
    } else {
      return getSource(rowOrCell)
    }
  })
}

export function getErrors (cells) {
  return cells.map(cell => {
    return cell.errors.map(err => {
      return err.name || 'unknown'
    })
  })
}

export function getStates (cells) {
  return cells.map(cell => {
    return cellStateToString(cell.status)
  })
}

export function queryValues (engine, expr) {
  let symbol = parseSymbol(expr)
  if (!symbol.scope) throw new Error('query must use fully qualified identifiers')
  let docId = engine._lookupDocumentId(symbol.scope)
  if (!docId) throw new Error('Unknown resource:', symbol.scope)
  switch (symbol.type) {
    case 'var': {
      return engine._graph.getValue(expr)
    }
    case 'cell': {
      let sheet = engine._docs[docId]
      let [row, col] = getRowCol(symbol.name)
      return getValue(sheet.cells[row][col])
    }
    case 'range': {
      let sheet = engine._docs[docId]
      const { startRow, startCol, endRow, endCol } = getIndexesFromRange(symbol.anchor, symbol.focus)
      let cells = getRangeFromMatrix(sheet.getCells(), startRow, startCol, endRow, endCol)
      return getValues(cells)
    }
    default:
      //
  }
}

/*
  Waits for all actions to be finished.
  This is the slowest kind of scheduling, as every cycle
  takes as long as the longest evaluation.
  In a real environment, the Engine should be triggered as often as possible,
  but still with a little delay, so that all 'simultanous' actions can be
  done at once.
*/
export function cycle (engine) {
  console.error('DEPRECATED: use engine.cycle() instead')
  return engine.cycle()
}

/*
  Triggers a cycle as long as next actions are coming in.
*/
export function play (engine) {
  console.error('DEPRECATED: use engine.runOnce() instead')
  return engine.runOnce()
}

export function setSheetSelection (sheetSession, expr) {
  let { anchorRow, anchorCol, focusRow, focusCol } = _getCoordinatesFromExpr(expr)
  let selData = {
    type: 'range',
    anchorRow,
    anchorCol,
    focusRow,
    focusCol
  }
  sheetSession.setSelection({
    type: 'custom',
    customType: 'sheet',
    data: selData
  })
}

export function checkSelection (t, sel, expr) {
  let expectedSelData = _getCoordinatesFromExpr(expr)
  expectedSelData.type = 'range'
  t.deepEqual(sel.data, expectedSelData, 'selection should be correct')
}

function _getCoordinatesFromExpr (expr) {
  let [start, end] = expr.split(':')
  let [anchorRow, anchorCol] = getRowCol(start)
  let focusRow, focusCol
  if (end) {
    ([focusRow, focusCol] = getRowCol(end))
  } else {
    ([focusRow, focusCol] = [anchorRow, anchorCol])
  }
  return { anchorRow, anchorCol, focusRow, focusCol }
}

// This is useful for writing tests, to use queries such as 'A1:A10'
export function queryCells (cells, query) {
  let symbol = parseSymbol(query)
  switch (symbol.type) {
    case 'cell': {
      const [row, col] = getRowCol(symbol.name)
      return cells[row][col]
    }
    case 'range': {
      const { startRow, startCol, endRow, endCol } = getIndexesFromRange(symbol.anchor, symbol.focus)
      return getRangeFromMatrix(cells, startRow, startCol, endRow, endCol)
    }
    default:
      throw new Error('Unsupported query')
  }
}
