(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (factory((global.stencilaEngine = {})));
}(this, (function (exports) { 'use strict';

  function forEach (iteratee, func) {
    if (!iteratee) return
    if (iteratee.constructor.prototype.forEach) {
      iteratee.forEach(func);
    } else {
      Object.keys(iteratee).forEach(function (key) {
        return func(iteratee[key], key)
      });
    }
  }

  function isObject (val) {
    const type = typeof val;
    return Boolean(val) && (type === 'object' || type === 'function')
  }

  const DEBUG = false;
  let count = 0;
  const COUNT_MSG = '%s listeners registered in the whole system.';


  class EventEmitter {
    
    emit (event) {
      if (event in this.__events__) {
        
        
        var bindings = this.__events__[event].slice();
        var args = Array.prototype.slice.call(arguments, 1);
        for (var i = 0, len = bindings.length; i < len; i++) {
          var binding = bindings[i];
          
          binding.method.apply(binding.context, args);
        }
        return true
      }
      return false
    }

    
    on (event, method, context) {
      
      _on.call(this, event, method, context);
    }

    
    off (event, method, context) { 
      if (arguments.length === 1 && isObject(arguments[0])) {
        _disconnect.call(this, arguments[0]);
      } else {
        _off.apply(this, arguments);
      }
    }

    _debugEvents () {
      
      console.log('### EventEmitter: ', this);
      forEach(this.__events__, (handlers, name) => {
        console.log('- %s listeners for %s: ', handlers.length, name, handlers);
      });
      
    }

    get __events__ () {
      if (!this.___events___) {
        this.___events___ = {};
      }
      return this.___events___
    }
  }


  function _on (event, method, context) {
    
    var bindings;
    validateMethod(method, context);
    if (this.__events__.hasOwnProperty(event)) {
      bindings = this.__events__[event];
    } else {
      
      bindings = this.__events__[event] = [];
    }
    
    bindings.push({
      method: method,
      context: context || null
    });
    if (DEBUG) {
      count++;
      console.info('_on()', event, method.name, context, this);
      console.info(COUNT_MSG, count);
    }
    return this
    
  }


  function _off (event, method, context) {
    
    if (arguments.length === 0) {
      if (DEBUG) {
        forEach(this.__events__, (bindings) => {
          bindings.forEach((b) => {
            console.info('_off()', b.method.name, b.context, this);
          });
          count -= bindings.length;
        });
        console.info(COUNT_MSG, count);
      }
      this.___events___ = {};
      return this
    }
    if (arguments.length === 1) {
      
      if (DEBUG) {
        count -= (this.__events__[event] || []).length;
        console.info(COUNT_MSG, count);
      }
      delete this.__events__[event];
      return this
    }
    validateMethod(method, context);
    if (!(event in this.__events__) || !this.__events__[event].length) {
      if (DEBUG) console.info('NO MATCHING BINDINGS');
      
      return this
    }
    
    if (arguments.length < 3) {
      context = null;
    }
    
    let bindings = this.__events__[event];
    for (let i = bindings.length - 1; i >= 0; i--) {
      const b = bindings[i];
      if (b.method === method && b.context === context) {
        bindings.splice(i, 1);
        if (DEBUG) count--;
      }
    }
    
    if (bindings.length === 0) {
      delete this.__events__[event];
    }
    if (DEBUG) console.info(COUNT_MSG, count);
    return this
    
  }


  function _disconnect (context) {
    
    
    forEach(this.__events__, (bindings, event) => {
      for (let i = bindings.length - 1; i >= 0; i--) {
        
        
        if (bindings[i] && bindings[i].context === context) {
          _off.call(this, event, bindings[i].method, context);
        }
      }
    });
    return this
    
  }

  function validateMethod (method, context) {
    
    if (typeof method === 'string') {
      
      if (context === undefined || context === null) {
        throw new Error('Method name "' + method + '" has no context.')
      }
      if (!(method in context)) {
        
        
        throw new Error('Method not found: "' + method + '"')
      }
      if (typeof context[method] !== 'function') {
        
        
        throw new Error('Property "' + method + '" is not a function')
      }
    } else if (typeof method !== 'function') {
      throw new Error('Invalid callback. Function or method name expected.')
    }
  }

  function uuid (prefix, len) {
    if (prefix && prefix[prefix.length - 1] !== '-') {
      prefix = prefix.concat('-');
    }
    var chars = '0123456789abcdefghijklmnopqrstuvwxyz'.split('');
    var uuid = [];
    var radix = 16;
    var idx;
    len = len || 32;
    if (len) {
      
      for (idx = 0; idx < len; idx++) uuid[idx] = chars[0 | Math.random() * radix];
    } else {
      
      var r;
      
      uuid[8] = uuid[13] = uuid[18] = uuid[23] = '-';
      uuid[14] = '4';
      
      
      for (idx = 0; idx < 36; idx++) {
        if (!uuid[idx]) {
          r = 0 | Math.random() * 16;
          uuid[idx] = chars[(idx === 19) ? (r & 0x3) | 0x8 : r];
        }
      }
    }
    return (prefix || '') + uuid.join('')
  }

  function map (iteratee, func) {
    if (!iteratee) return []
    if (!func) func = function (item) { return item };
    if (Array.isArray(iteratee)) {
      return iteratee.map(func)
    }
    if (iteratee instanceof Map) {
      let result = [];
      for (let [name, val] of iteratee) {
        result.push(func(val, name));
      }
      return result
    }
    if (iteratee instanceof Set) {
      let result = [];
      let idx = 0;
      iteratee.forEach(item => {
        result.push(func(item, idx++));
      });
      return result
    }
    return Object.keys(iteratee).map(function (key) {
      return func(iteratee[key], key)
    })
  }

  function isPlainObject (o) {
    return Boolean(o) && o.constructor === {}.constructor
  }

  function isArray (a) {
    return Array.isArray(a)
  }

  function clone (val) {
    if (isArray(val)) {
      return val.slice(0)
    }
    if (isObject(val)) {
      return Object.assign({}, val)
    }
    
    
    return val
  }

  const platform = {

    inBrowser: false,

    inNodeJS: false,

    inElectron: false,

    
    isIE: false,
    

    isFF: false,

    isWebkit: false,

    
    version: -1,

    

    isWindows: false,

    isMac: false,

    devtools: false,

    
    
    _reset: detect
  };

  function detect () {
    if (typeof window !== 'undefined') {
      platform.inBrowser = true;

      
      const ua = window.navigator.userAgent;
      const msie = ua.indexOf('MSIE ');
      const trident = ua.indexOf('Trident/');
      const edge = ua.indexOf('Edge/');

      if (msie > 0) {
        
        platform.isIE = true;
        platform.version = 10;
        
        
      } else if (trident > 0) {
        
        platform.isIE = true;
        platform.version = 11;
        platform.isTrident = true;
        
        
        
      } else if (edge > 0) {
        
        platform.isIE = true;
        platform.isEdge = true;
        platform.version = 12;
        
        parseInt(ua.substring(edge + 5, ua.indexOf('.', edge)), 10);
      }

      
      platform.isFF = window.navigator.userAgent.toLowerCase().indexOf('firefox') > -1;

      
      platform.isWebkit = !platform.isFF && !platform.isIE;
    } else {
      platform.inBrowser = false;
    }

    if (platform.inBrowser) {
      platform.isWindows = (window.navigator !== undefined && window.navigator.appVersion && window.navigator.appVersion.indexOf('Win') !== -1);
      platform.isMac = (window.navigator !== undefined && window.navigator.platform.indexOf('Mac') >= 0);
    }

    
    if (typeof process !== 'undefined') {
      if (platform.inBrowser) {
        platform.inElectron = true;
      } else {
        platform.inNodeJS = true;
      }
    }

    if (platform.inBrowser) {
      let widthThreshold = window.outerWidth - window.innerWidth > 160;
      let heightThreshold = window.outerHeight - window.innerHeight > 160;
      let orientation = widthThreshold ? 'vertical' : 'horizontal';
      if (!(heightThreshold && widthThreshold) &&
        ((window.Firebug && window.Firebug.chrome && window.Firebug.chrome.isInitialized) || widthThreshold || heightThreshold)) {
        platform.devtools = true;
        platform.devtoolsOrientation = orientation;
      }
    }
  }

  detect();

  function cloneDeep (val) {
    if (isArray(val)) {
      return _cloneArrayDeep(val)
    }
    
    
    
    if (platform.inBrowser && val instanceof window.File) {
      return val
    }
    if (isObject(val)) {
      return _cloneObjectDeep(val)
    }
    
    
    return val
  }

  function _cloneObjectDeep (obj) {
    let res = {};
    forEach(obj, (val, key) => {
      res[key] = cloneDeep(val);
    });
    return res
  }

  function _cloneArrayDeep (arr) {
    return arr.map(cloneDeep)
  }

  function isNumber (n) {
    return typeof n === 'number'
  }

  function isEqual (a, b) {
    if (a === b) return true
    if (isArray(a) && isArray(b)) {
      if (a.length !== b.length) return false
      for (let i = 0; i < a.length; i++) {
        if (!isEqual(a[i], b[i])) return false
      }
      return true
    }
    if (isPlainObject(a) && isPlainObject(b)) {
      let akeys = Object.keys(a).sort();
      let bkeys = Object.keys(b).sort();
      if (!isEqual(akeys, bkeys)) return false
      for (let i = 0; i < akeys.length; i++) {
        let key = akeys[i];
        if (!isEqual(a[key], b[key])) return false
      }
      return true
    }
    return false
  }

  function isNil (o) {
    return o === null || o === undefined
  }

  function isString (s) {
    return typeof s === 'string'
  }

  var freeGlobal = typeof global == 'object' && global && global.Object === Object && global;

  var freeSelf = typeof self == 'object' && self && self.Object === Object && self;


  var root = freeGlobal || freeSelf || Function('return this')();

  var Symbol$1 = root.Symbol;

  function arrayMap(array, iteratee) {
    var index = -1,
        length = array == null ? 0 : array.length,
        result = Array(length);

    while (++index < length) {
      result[index] = iteratee(array[index], index, array);
    }
    return result;
  }

  var isArray$1 = Array.isArray;

  var objectProto = Object.prototype;


  var hasOwnProperty = objectProto.hasOwnProperty;


  var nativeObjectToString = objectProto.toString;


  var symToStringTag = Symbol$1 ? Symbol$1.toStringTag : undefined;


  function getRawTag(value) {
    var isOwn = hasOwnProperty.call(value, symToStringTag),
        tag = value[symToStringTag];

    try {
      value[symToStringTag] = undefined;
      var unmasked = true;
    } catch (e) {}

    var result = nativeObjectToString.call(value);
    if (unmasked) {
      if (isOwn) {
        value[symToStringTag] = tag;
      } else {
        delete value[symToStringTag];
      }
    }
    return result;
  }

  var objectProto$1 = Object.prototype;


  var nativeObjectToString$1 = objectProto$1.toString;


  function objectToString(value) {
    return nativeObjectToString$1.call(value);
  }

  var nullTag = '[object Null]',
      undefinedTag = '[object Undefined]';


  var symToStringTag$1 = Symbol$1 ? Symbol$1.toStringTag : undefined;


  function baseGetTag(value) {
    if (value == null) {
      return value === undefined ? undefinedTag : nullTag;
    }
    return (symToStringTag$1 && symToStringTag$1 in Object(value))
      ? getRawTag(value)
      : objectToString(value);
  }

  function isObjectLike(value) {
    return value != null && typeof value == 'object';
  }

  var symbolTag = '[object Symbol]';


  function isSymbol(value) {
    return typeof value == 'symbol' ||
      (isObjectLike(value) && baseGetTag(value) == symbolTag);
  }

  var INFINITY = 1 / 0;


  var symbolProto = Symbol$1 ? Symbol$1.prototype : undefined,
      symbolToString = symbolProto ? symbolProto.toString : undefined;


  function baseToString(value) {
    
    if (typeof value == 'string') {
      return value;
    }
    if (isArray$1(value)) {
      
      return arrayMap(value, baseToString) + '';
    }
    if (isSymbol(value)) {
      return symbolToString ? symbolToString.call(value) : '';
    }
    var result = (value + '');
    return (result == '0' && (1 / value) == -INFINITY) ? '-0' : result;
  }

  function toString(value) {
    return value == null ? '' : baseToString(value);
  }

  function baseSlice(array, start, end) {
    var index = -1,
        length = array.length;

    if (start < 0) {
      start = -start > length ? 0 : (length + start);
    }
    end = end > length ? length : end;
    if (end < 0) {
      end += length;
    }
    length = start > end ? 0 : ((end - start) >>> 0);
    start >>>= 0;

    var result = Array(length);
    while (++index < length) {
      result[index] = array[index + start];
    }
    return result;
  }

  function isObject$1(value) {
    var type = typeof value;
    return value != null && (type == 'object' || type == 'function');
  }

  var reIsDeepProp = /\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/,
      reIsPlainProp = /^\w*$/;


  function isKey(value, object) {
    if (isArray$1(value)) {
      return false;
    }
    var type = typeof value;
    if (type == 'number' || type == 'symbol' || type == 'boolean' ||
        value == null || isSymbol(value)) {
      return true;
    }
    return reIsPlainProp.test(value) || !reIsDeepProp.test(value) ||
      (object != null && value in Object(object));
  }

  var asyncTag = '[object AsyncFunction]',
      funcTag = '[object Function]',
      genTag = '[object GeneratorFunction]',
      proxyTag = '[object Proxy]';


  function isFunction(value) {
    if (!isObject$1(value)) {
      return false;
    }
    
    
    var tag = baseGetTag(value);
    return tag == funcTag || tag == genTag || tag == asyncTag || tag == proxyTag;
  }

  var coreJsData = root['__core-js_shared__'];

  var maskSrcKey = (function() {
    var uid = /[^.]+$/.exec(coreJsData && coreJsData.keys && coreJsData.keys.IE_PROTO || '');
    return uid ? ('Symbol(src)_1.' + uid) : '';
  }());


  function isMasked(func) {
    return !!maskSrcKey && (maskSrcKey in func);
  }

  var funcProto = Function.prototype;


  var funcToString = funcProto.toString;


  function toSource(func) {
    if (func != null) {
      try {
        return funcToString.call(func);
      } catch (e) {}
      try {
        return (func + '');
      } catch (e) {}
    }
    return '';
  }

  var reRegExpChar = /[\\^$.*+?()[\]{}|]/g;


  var reIsHostCtor = /^\[object .+?Constructor\]$/;


  var funcProto$1 = Function.prototype,
      objectProto$2 = Object.prototype;


  var funcToString$1 = funcProto$1.toString;


  var hasOwnProperty$1 = objectProto$2.hasOwnProperty;


  var reIsNative = RegExp('^' +
    funcToString$1.call(hasOwnProperty$1).replace(reRegExpChar, '\\$&')
    .replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g, '$1.*?') + '$'
  );


  function baseIsNative(value) {
    if (!isObject$1(value) || isMasked(value)) {
      return false;
    }
    var pattern = isFunction(value) ? reIsNative : reIsHostCtor;
    return pattern.test(toSource(value));
  }

  function getValue(object, key) {
    return object == null ? undefined : object[key];
  }

  function getNative(object, key) {
    var value = getValue(object, key);
    return baseIsNative(value) ? value : undefined;
  }

  var nativeCreate = getNative(Object, 'create');

  function hashClear() {
    this.__data__ = nativeCreate ? nativeCreate(null) : {};
    this.size = 0;
  }

  function hashDelete(key) {
    var result = this.has(key) && delete this.__data__[key];
    this.size -= result ? 1 : 0;
    return result;
  }

  var HASH_UNDEFINED = '__lodash_hash_undefined__';


  var objectProto$3 = Object.prototype;


  var hasOwnProperty$2 = objectProto$3.hasOwnProperty;


  function hashGet(key) {
    var data = this.__data__;
    if (nativeCreate) {
      var result = data[key];
      return result === HASH_UNDEFINED ? undefined : result;
    }
    return hasOwnProperty$2.call(data, key) ? data[key] : undefined;
  }

  var objectProto$4 = Object.prototype;


  var hasOwnProperty$3 = objectProto$4.hasOwnProperty;


  function hashHas(key) {
    var data = this.__data__;
    return nativeCreate ? (data[key] !== undefined) : hasOwnProperty$3.call(data, key);
  }

  var HASH_UNDEFINED$1 = '__lodash_hash_undefined__';


  function hashSet(key, value) {
    var data = this.__data__;
    this.size += this.has(key) ? 0 : 1;
    data[key] = (nativeCreate && value === undefined) ? HASH_UNDEFINED$1 : value;
    return this;
  }

  function Hash$1(entries) {
    var index = -1,
        length = entries == null ? 0 : entries.length;

    this.clear();
    while (++index < length) {
      var entry = entries[index];
      this.set(entry[0], entry[1]);
    }
  }


  Hash$1.prototype.clear = hashClear;
  Hash$1.prototype['delete'] = hashDelete;
  Hash$1.prototype.get = hashGet;
  Hash$1.prototype.has = hashHas;
  Hash$1.prototype.set = hashSet;

  function listCacheClear() {
    this.__data__ = [];
    this.size = 0;
  }

  function eq(value, other) {
    return value === other || (value !== value && other !== other);
  }

  function assocIndexOf(array, key) {
    var length = array.length;
    while (length--) {
      if (eq(array[length][0], key)) {
        return length;
      }
    }
    return -1;
  }

  var arrayProto = Array.prototype;


  var splice = arrayProto.splice;


  function listCacheDelete(key) {
    var data = this.__data__,
        index = assocIndexOf(data, key);

    if (index < 0) {
      return false;
    }
    var lastIndex = data.length - 1;
    if (index == lastIndex) {
      data.pop();
    } else {
      splice.call(data, index, 1);
    }
    --this.size;
    return true;
  }

  function listCacheGet(key) {
    var data = this.__data__,
        index = assocIndexOf(data, key);

    return index < 0 ? undefined : data[index][1];
  }

  function listCacheHas(key) {
    return assocIndexOf(this.__data__, key) > -1;
  }

  function listCacheSet(key, value) {
    var data = this.__data__,
        index = assocIndexOf(data, key);

    if (index < 0) {
      ++this.size;
      data.push([key, value]);
    } else {
      data[index][1] = value;
    }
    return this;
  }

  function ListCache(entries) {
    var index = -1,
        length = entries == null ? 0 : entries.length;

    this.clear();
    while (++index < length) {
      var entry = entries[index];
      this.set(entry[0], entry[1]);
    }
  }


  ListCache.prototype.clear = listCacheClear;
  ListCache.prototype['delete'] = listCacheDelete;
  ListCache.prototype.get = listCacheGet;
  ListCache.prototype.has = listCacheHas;
  ListCache.prototype.set = listCacheSet;

  var Map$1 = getNative(root, 'Map');

  function mapCacheClear() {
    this.size = 0;
    this.__data__ = {
      'hash': new Hash$1,
      'map': new (Map$1 || ListCache),
      'string': new Hash$1
    };
  }

  function isKeyable(value) {
    var type = typeof value;
    return (type == 'string' || type == 'number' || type == 'symbol' || type == 'boolean')
      ? (value !== '__proto__')
      : (value === null);
  }

  function getMapData(map, key) {
    var data = map.__data__;
    return isKeyable(key)
      ? data[typeof key == 'string' ? 'string' : 'hash']
      : data.map;
  }

  function mapCacheDelete(key) {
    var result = getMapData(this, key)['delete'](key);
    this.size -= result ? 1 : 0;
    return result;
  }

  function mapCacheGet(key) {
    return getMapData(this, key).get(key);
  }

  function mapCacheHas(key) {
    return getMapData(this, key).has(key);
  }

  function mapCacheSet(key, value) {
    var data = getMapData(this, key),
        size = data.size;

    data.set(key, value);
    this.size += data.size == size ? 0 : 1;
    return this;
  }

  function MapCache(entries) {
    var index = -1,
        length = entries == null ? 0 : entries.length;

    this.clear();
    while (++index < length) {
      var entry = entries[index];
      this.set(entry[0], entry[1]);
    }
  }


  MapCache.prototype.clear = mapCacheClear;
  MapCache.prototype['delete'] = mapCacheDelete;
  MapCache.prototype.get = mapCacheGet;
  MapCache.prototype.has = mapCacheHas;
  MapCache.prototype.set = mapCacheSet;

  var FUNC_ERROR_TEXT$1 = 'Expected a function';


  function memoize(func, resolver) {
    if (typeof func != 'function' || (resolver != null && typeof resolver != 'function')) {
      throw new TypeError(FUNC_ERROR_TEXT$1);
    }
    var memoized = function() {
      var args = arguments,
          key = resolver ? resolver.apply(this, args) : args[0],
          cache = memoized.cache;

      if (cache.has(key)) {
        return cache.get(key);
      }
      var result = func.apply(this, args);
      memoized.cache = cache.set(key, result) || cache;
      return result;
    };
    memoized.cache = new (memoize.Cache || MapCache);
    return memoized;
  }


  memoize.Cache = MapCache;

  var MAX_MEMOIZE_SIZE = 500;


  function memoizeCapped(func) {
    var result = memoize(func, function(key) {
      if (cache.size === MAX_MEMOIZE_SIZE) {
        cache.clear();
      }
      return key;
    });

    var cache = result.cache;
    return result;
  }

  var rePropName = /[^.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(?:\.|\[\])(?:\.|\[\]|$))/g;


  var reEscapeChar = /\\(\\)?/g;


  var stringToPath = memoizeCapped(function(string) {
    var result = [];
    if (string.charCodeAt(0) === 46 ) {
      result.push('');
    }
    string.replace(rePropName, function(match, number, quote, subString) {
      result.push(quote ? subString.replace(reEscapeChar, '$1') : (number || match));
    });
    return result;
  });

  function castPath(value, object) {
    if (isArray$1(value)) {
      return value;
    }
    return isKey(value, object) ? [value] : stringToPath(toString(value));
  }

  var INFINITY$1 = 1 / 0;


  function toKey(value) {
    if (typeof value == 'string' || isSymbol(value)) {
      return value;
    }
    var result = (value + '');
    return (result == '0' && (1 / value) == -INFINITY$1) ? '-0' : result;
  }

  function baseGet(object, path) {
    path = castPath(path, object);

    var index = 0,
        length = path.length;

    while (object != null && index < length) {
      object = object[toKey(path[index++])];
    }
    return (index && index == length) ? object : undefined;
  }

  function get(object, path, defaultValue) {
    var result = object == null ? undefined : baseGet(object, path);
    return result === undefined ? defaultValue : result;
  }

  function stackClear() {
    this.__data__ = new ListCache;
    this.size = 0;
  }

  function stackDelete(key) {
    var data = this.__data__,
        result = data['delete'](key);

    this.size = data.size;
    return result;
  }

  function stackGet(key) {
    return this.__data__.get(key);
  }

  function stackHas(key) {
    return this.__data__.has(key);
  }

  var LARGE_ARRAY_SIZE = 200;


  function stackSet(key, value) {
    var data = this.__data__;
    if (data instanceof ListCache) {
      var pairs = data.__data__;
      if (!Map$1 || (pairs.length < LARGE_ARRAY_SIZE - 1)) {
        pairs.push([key, value]);
        this.size = ++data.size;
        return this;
      }
      data = this.__data__ = new MapCache(pairs);
    }
    data.set(key, value);
    this.size = data.size;
    return this;
  }

  function Stack(entries) {
    var data = this.__data__ = new ListCache(entries);
    this.size = data.size;
  }


  Stack.prototype.clear = stackClear;
  Stack.prototype['delete'] = stackDelete;
  Stack.prototype.get = stackGet;
  Stack.prototype.has = stackHas;
  Stack.prototype.set = stackSet;

  var HASH_UNDEFINED$2 = '__lodash_hash_undefined__';


  function setCacheAdd(value) {
    this.__data__.set(value, HASH_UNDEFINED$2);
    return this;
  }

  function setCacheHas(value) {
    return this.__data__.has(value);
  }

  function SetCache(values) {
    var index = -1,
        length = values == null ? 0 : values.length;

    this.__data__ = new MapCache;
    while (++index < length) {
      this.add(values[index]);
    }
  }


  SetCache.prototype.add = SetCache.prototype.push = setCacheAdd;
  SetCache.prototype.has = setCacheHas;

  function cacheHas(cache, key) {
    return cache.has(key);
  }

  var Uint8Array = root.Uint8Array;

  function setToArray(set) {
    var index = -1,
        result = Array(set.size);

    set.forEach(function(value) {
      result[++index] = value;
    });
    return result;
  }


  var symbolProto$1 = Symbol$1 ? Symbol$1.prototype : undefined,
      symbolValueOf = symbolProto$1 ? symbolProto$1.valueOf : undefined;

  function baseTimes(n, iteratee) {
    var index = -1,
        result = Array(n);

    while (++index < n) {
      result[index] = iteratee(index);
    }
    return result;
  }

  var argsTag = '[object Arguments]';


  function baseIsArguments(value) {
    return isObjectLike(value) && baseGetTag(value) == argsTag;
  }

  var objectProto$6 = Object.prototype;


  var hasOwnProperty$4 = objectProto$6.hasOwnProperty;


  var propertyIsEnumerable$1 = objectProto$6.propertyIsEnumerable;


  var isArguments = baseIsArguments(function() { return arguments; }()) ? baseIsArguments : function(value) {
    return isObjectLike(value) && hasOwnProperty$4.call(value, 'callee') &&
      !propertyIsEnumerable$1.call(value, 'callee');
  };

  function stubFalse() {
    return false;
  }

  var freeExports = typeof exports == 'object' && exports && !exports.nodeType && exports;


  var freeModule = freeExports && typeof module == 'object' && module && !module.nodeType && module;


  var moduleExports = freeModule && freeModule.exports === freeExports;


  var Buffer = moduleExports ? root.Buffer : undefined;


  var nativeIsBuffer = Buffer ? Buffer.isBuffer : undefined;


  var isBuffer = nativeIsBuffer || stubFalse;

  var MAX_SAFE_INTEGER = 9007199254740991;


  var reIsUint = /^(?:0|[1-9]\d*)$/;


  function isIndex(value, length) {
    var type = typeof value;
    length = length == null ? MAX_SAFE_INTEGER : length;

    return !!length &&
      (type == 'number' ||
        (type != 'symbol' && reIsUint.test(value))) &&
          (value > -1 && value % 1 == 0 && value < length);
  }

  var MAX_SAFE_INTEGER$1 = 9007199254740991;


  function isLength(value) {
    return typeof value == 'number' &&
      value > -1 && value % 1 == 0 && value <= MAX_SAFE_INTEGER$1;
  }

  var argsTag$1 = '[object Arguments]',
      arrayTag = '[object Array]',
      boolTag$1 = '[object Boolean]',
      dateTag$1 = '[object Date]',
      errorTag$1 = '[object Error]',
      funcTag$1 = '[object Function]',
      mapTag$1 = '[object Map]',
      numberTag$1 = '[object Number]',
      objectTag = '[object Object]',
      regexpTag$1 = '[object RegExp]',
      setTag$1 = '[object Set]',
      stringTag$1 = '[object String]',
      weakMapTag = '[object WeakMap]';

  var arrayBufferTag$1 = '[object ArrayBuffer]',
      dataViewTag$1 = '[object DataView]',
      float32Tag = '[object Float32Array]',
      float64Tag = '[object Float64Array]',
      int8Tag = '[object Int8Array]',
      int16Tag = '[object Int16Array]',
      int32Tag = '[object Int32Array]',
      uint8Tag = '[object Uint8Array]',
      uint8ClampedTag = '[object Uint8ClampedArray]',
      uint16Tag = '[object Uint16Array]',
      uint32Tag = '[object Uint32Array]';


  var typedArrayTags = {};
  typedArrayTags[float32Tag] = typedArrayTags[float64Tag] =
  typedArrayTags[int8Tag] = typedArrayTags[int16Tag] =
  typedArrayTags[int32Tag] = typedArrayTags[uint8Tag] =
  typedArrayTags[uint8ClampedTag] = typedArrayTags[uint16Tag] =
  typedArrayTags[uint32Tag] = true;
  typedArrayTags[argsTag$1] = typedArrayTags[arrayTag] =
  typedArrayTags[arrayBufferTag$1] = typedArrayTags[boolTag$1] =
  typedArrayTags[dataViewTag$1] = typedArrayTags[dateTag$1] =
  typedArrayTags[errorTag$1] = typedArrayTags[funcTag$1] =
  typedArrayTags[mapTag$1] = typedArrayTags[numberTag$1] =
  typedArrayTags[objectTag] = typedArrayTags[regexpTag$1] =
  typedArrayTags[setTag$1] = typedArrayTags[stringTag$1] =
  typedArrayTags[weakMapTag] = false;


  function baseIsTypedArray(value) {
    return isObjectLike(value) &&
      isLength(value.length) && !!typedArrayTags[baseGetTag(value)];
  }

  function baseUnary(func) {
    return function(value) {
      return func(value);
    };
  }

  var freeExports$1 = typeof exports == 'object' && exports && !exports.nodeType && exports;


  var freeModule$1 = freeExports$1 && typeof module == 'object' && module && !module.nodeType && module;


  var moduleExports$1 = freeModule$1 && freeModule$1.exports === freeExports$1;


  var freeProcess = moduleExports$1 && freeGlobal.process;


  var nodeUtil = (function() {
    try {
      
      var types = freeModule$1 && freeModule$1.require && freeModule$1.require('util').types;

      if (types) {
        return types;
      }

      
      return freeProcess && freeProcess.binding && freeProcess.binding('util');
    } catch (e) {}
  }());

  var nodeIsTypedArray = nodeUtil && nodeUtil.isTypedArray;


  var isTypedArray = nodeIsTypedArray ? baseUnary(nodeIsTypedArray) : baseIsTypedArray;

  var objectProto$7 = Object.prototype;


  var hasOwnProperty$5 = objectProto$7.hasOwnProperty;


  function arrayLikeKeys(value, inherited) {
    var isArr = isArray$1(value),
        isArg = !isArr && isArguments(value),
        isBuff = !isArr && !isArg && isBuffer(value),
        isType = !isArr && !isArg && !isBuff && isTypedArray(value),
        skipIndexes = isArr || isArg || isBuff || isType,
        result = skipIndexes ? baseTimes(value.length, String) : [],
        length = result.length;

    for (var key in value) {
      if ((inherited || hasOwnProperty$5.call(value, key)) &&
          !(skipIndexes && (
             
             key == 'length' ||
             
             (isBuff && (key == 'offset' || key == 'parent')) ||
             
             (isType && (key == 'buffer' || key == 'byteLength' || key == 'byteOffset')) ||
             
             isIndex(key, length)
          ))) {
        result.push(key);
      }
    }
    return result;
  }

  var objectProto$8 = Object.prototype;


  function isPrototype(value) {
    var Ctor = value && value.constructor,
        proto = (typeof Ctor == 'function' && Ctor.prototype) || objectProto$8;

    return value === proto;
  }

  function overArg(func, transform) {
    return function(arg) {
      return func(transform(arg));
    };
  }

  function isArrayLike(value) {
    return value != null && isLength(value.length) && !isFunction(value);
  }

  var DataView = getNative(root, 'DataView');

  var Promise$1 = getNative(root, 'Promise');

  var Set$1 = getNative(root, 'Set');

  var WeakMap$1 = getNative(root, 'WeakMap');

  var mapTag$2 = '[object Map]',
      objectTag$1 = '[object Object]',
      promiseTag = '[object Promise]',
      setTag$2 = '[object Set]',
      weakMapTag$1 = '[object WeakMap]';

  var dataViewTag$2 = '[object DataView]';


  var dataViewCtorString = toSource(DataView),
      mapCtorString = toSource(Map$1),
      promiseCtorString = toSource(Promise$1),
      setCtorString = toSource(Set$1),
      weakMapCtorString = toSource(WeakMap$1);


  var getTag = baseGetTag;


  if ((DataView && getTag(new DataView(new ArrayBuffer(1))) != dataViewTag$2) ||
      (Map$1 && getTag(new Map$1) != mapTag$2) ||
      (Promise$1 && getTag(Promise$1.resolve()) != promiseTag) ||
      (Set$1 && getTag(new Set$1) != setTag$2) ||
      (WeakMap$1 && getTag(new WeakMap$1) != weakMapTag$1)) {
    getTag = function(value) {
      var result = baseGetTag(value),
          Ctor = result == objectTag$1 ? value.constructor : undefined,
          ctorString = Ctor ? toSource(Ctor) : '';

      if (ctorString) {
        switch (ctorString) {
          case dataViewCtorString: return dataViewTag$2;
          case mapCtorString: return mapTag$2;
          case promiseCtorString: return promiseTag;
          case setCtorString: return setTag$2;
          case weakMapCtorString: return weakMapTag$1;
        }
      }
      return result;
    };
  }

  function identity(value) {
    return value;
  }

  function baseFindIndex(array, predicate, fromIndex, fromRight) {
    var length = array.length,
        index = fromIndex + (fromRight ? 1 : -1);

    while ((fromRight ? index-- : ++index < length)) {
      if (predicate(array[index], index, array)) {
        return index;
      }
    }
    return -1;
  }

  var defineProperty = (function() {
    try {
      var func = getNative(Object, 'defineProperty');
      func({}, '', {});
      return func;
    } catch (e) {}
  }());

  function baseAssignValue(object, key, value) {
    if (key == '__proto__' && defineProperty) {
      defineProperty(object, key, {
        'configurable': true,
        'enumerable': true,
        'value': value,
        'writable': true
      });
    } else {
      object[key] = value;
    }
  }

  function assignMergeValue(object, key, value) {
    if ((value !== undefined && !eq(object[key], value)) ||
        (value === undefined && !(key in object))) {
      baseAssignValue(object, key, value);
    }
  }

  function createBaseFor(fromRight) {
    return function(object, iteratee, keysFunc) {
      var index = -1,
          iterable = Object(object),
          props = keysFunc(object),
          length = props.length;

      while (length--) {
        var key = props[fromRight ? length : ++index];
        if (iteratee(iterable[key], key, iterable) === false) {
          break;
        }
      }
      return object;
    };
  }

  var baseFor = createBaseFor();

  var freeExports$2 = typeof exports == 'object' && exports && !exports.nodeType && exports;


  var freeModule$2 = freeExports$2 && typeof module == 'object' && module && !module.nodeType && module;


  var moduleExports$2 = freeModule$2 && freeModule$2.exports === freeExports$2;


  var Buffer$1 = moduleExports$2 ? root.Buffer : undefined,
      allocUnsafe = Buffer$1 ? Buffer$1.allocUnsafe : undefined;


  function cloneBuffer(buffer, isDeep) {
    if (isDeep) {
      return buffer.slice();
    }
    var length = buffer.length,
        result = allocUnsafe ? allocUnsafe(length) : new buffer.constructor(length);

    buffer.copy(result);
    return result;
  }

  function cloneArrayBuffer(arrayBuffer) {
    var result = new arrayBuffer.constructor(arrayBuffer.byteLength);
    new Uint8Array(result).set(new Uint8Array(arrayBuffer));
    return result;
  }

  function cloneTypedArray(typedArray, isDeep) {
    var buffer = isDeep ? cloneArrayBuffer(typedArray.buffer) : typedArray.buffer;
    return new typedArray.constructor(buffer, typedArray.byteOffset, typedArray.length);
  }

  function copyArray(source, array) {
    var index = -1,
        length = source.length;

    array || (array = Array(length));
    while (++index < length) {
      array[index] = source[index];
    }
    return array;
  }

  var objectCreate = Object.create;


  var baseCreate = (function() {
    function object() {}
    return function(proto) {
      if (!isObject$1(proto)) {
        return {};
      }
      if (objectCreate) {
        return objectCreate(proto);
      }
      object.prototype = proto;
      var result = new object;
      object.prototype = undefined;
      return result;
    };
  }());

  var getPrototype = overArg(Object.getPrototypeOf, Object);

  function initCloneObject(object) {
    return (typeof object.constructor == 'function' && !isPrototype(object))
      ? baseCreate(getPrototype(object))
      : {};
  }

  function isArrayLikeObject(value) {
    return isObjectLike(value) && isArrayLike(value);
  }

  var objectTag$3 = '[object Object]';


  var funcProto$2 = Function.prototype,
      objectProto$12 = Object.prototype;


  var funcToString$2 = funcProto$2.toString;


  var hasOwnProperty$9 = objectProto$12.hasOwnProperty;


  var objectCtorString = funcToString$2.call(Object);


  function isPlainObject$1(value) {
    if (!isObjectLike(value) || baseGetTag(value) != objectTag$3) {
      return false;
    }
    var proto = getPrototype(value);
    if (proto === null) {
      return true;
    }
    var Ctor = hasOwnProperty$9.call(proto, 'constructor') && proto.constructor;
    return typeof Ctor == 'function' && Ctor instanceof Ctor &&
      funcToString$2.call(Ctor) == objectCtorString;
  }

  function safeGet(object, key) {
    return key == '__proto__'
      ? undefined
      : object[key];
  }

  var objectProto$13 = Object.prototype;


  var hasOwnProperty$10 = objectProto$13.hasOwnProperty;


  function assignValue(object, key, value) {
    var objValue = object[key];
    if (!(hasOwnProperty$10.call(object, key) && eq(objValue, value)) ||
        (value === undefined && !(key in object))) {
      baseAssignValue(object, key, value);
    }
  }

  function copyObject(source, props, object, customizer) {
    var isNew = !object;
    object || (object = {});

    var index = -1,
        length = props.length;

    while (++index < length) {
      var key = props[index];

      var newValue = customizer
        ? customizer(object[key], source[key], key, object, source)
        : undefined;

      if (newValue === undefined) {
        newValue = source[key];
      }
      if (isNew) {
        baseAssignValue(object, key, newValue);
      } else {
        assignValue(object, key, newValue);
      }
    }
    return object;
  }

  function nativeKeysIn(object) {
    var result = [];
    if (object != null) {
      for (var key in Object(object)) {
        result.push(key);
      }
    }
    return result;
  }

  var objectProto$14 = Object.prototype;


  var hasOwnProperty$11 = objectProto$14.hasOwnProperty;


  function baseKeysIn(object) {
    if (!isObject$1(object)) {
      return nativeKeysIn(object);
    }
    var isProto = isPrototype(object),
        result = [];

    for (var key in object) {
      if (!(key == 'constructor' && (isProto || !hasOwnProperty$11.call(object, key)))) {
        result.push(key);
      }
    }
    return result;
  }

  function keysIn(object) {
    return isArrayLike(object) ? arrayLikeKeys(object, true) : baseKeysIn(object);
  }

  function toPlainObject(value) {
    return copyObject(value, keysIn(value));
  }

  function baseMergeDeep(object, source, key, srcIndex, mergeFunc, customizer, stack) {
    var objValue = safeGet(object, key),
        srcValue = safeGet(source, key),
        stacked = stack.get(srcValue);

    if (stacked) {
      assignMergeValue(object, key, stacked);
      return;
    }
    var newValue = customizer
      ? customizer(objValue, srcValue, (key + ''), object, source, stack)
      : undefined;

    var isCommon = newValue === undefined;

    if (isCommon) {
      var isArr = isArray$1(srcValue),
          isBuff = !isArr && isBuffer(srcValue),
          isTyped = !isArr && !isBuff && isTypedArray(srcValue);

      newValue = srcValue;
      if (isArr || isBuff || isTyped) {
        if (isArray$1(objValue)) {
          newValue = objValue;
        }
        else if (isArrayLikeObject(objValue)) {
          newValue = copyArray(objValue);
        }
        else if (isBuff) {
          isCommon = false;
          newValue = cloneBuffer(srcValue, true);
        }
        else if (isTyped) {
          isCommon = false;
          newValue = cloneTypedArray(srcValue, true);
        }
        else {
          newValue = [];
        }
      }
      else if (isPlainObject$1(srcValue) || isArguments(srcValue)) {
        newValue = objValue;
        if (isArguments(objValue)) {
          newValue = toPlainObject(objValue);
        }
        else if (!isObject$1(objValue) || (srcIndex && isFunction(objValue))) {
          newValue = initCloneObject(srcValue);
        }
      }
      else {
        isCommon = false;
      }
    }
    if (isCommon) {
      
      stack.set(srcValue, newValue);
      mergeFunc(newValue, srcValue, srcIndex, customizer, stack);
      stack['delete'](srcValue);
    }
    assignMergeValue(object, key, newValue);
  }

  function baseMerge(object, source, srcIndex, customizer, stack) {
    if (object === source) {
      return;
    }
    baseFor(source, function(srcValue, key) {
      if (isObject$1(srcValue)) {
        stack || (stack = new Stack);
        baseMergeDeep(object, source, key, srcIndex, baseMerge, customizer, stack);
      }
      else {
        var newValue = customizer
          ? customizer(safeGet(object, key), srcValue, (key + ''), object, source, stack)
          : undefined;

        if (newValue === undefined) {
          newValue = srcValue;
        }
        assignMergeValue(object, key, newValue);
      }
    }, keysIn);
  }

  function apply(func, thisArg, args) {
    switch (args.length) {
      case 0: return func.call(thisArg);
      case 1: return func.call(thisArg, args[0]);
      case 2: return func.call(thisArg, args[0], args[1]);
      case 3: return func.call(thisArg, args[0], args[1], args[2]);
    }
    return func.apply(thisArg, args);
  }

  var nativeMax$2 = Math.max;


  function overRest(func, start, transform) {
    start = nativeMax$2(start === undefined ? (func.length - 1) : start, 0);
    return function() {
      var args = arguments,
          index = -1,
          length = nativeMax$2(args.length - start, 0),
          array = Array(length);

      while (++index < length) {
        array[index] = args[start + index];
      }
      index = -1;
      var otherArgs = Array(start + 1);
      while (++index < start) {
        otherArgs[index] = args[index];
      }
      otherArgs[start] = transform(array);
      return apply(func, this, otherArgs);
    };
  }

  function constant(value) {
    return function() {
      return value;
    };
  }

  var baseSetToString = !defineProperty ? identity : function(func, string) {
    return defineProperty(func, 'toString', {
      'configurable': true,
      'enumerable': false,
      'value': constant(string),
      'writable': true
    });
  };

  var HOT_COUNT = 800,
      HOT_SPAN = 16;


  var nativeNow = Date.now;


  function shortOut(func) {
    var count = 0,
        lastCalled = 0;

    return function() {
      var stamp = nativeNow(),
          remaining = HOT_SPAN - (stamp - lastCalled);

      lastCalled = stamp;
      if (remaining > 0) {
        if (++count >= HOT_COUNT) {
          return arguments[0];
        }
      } else {
        count = 0;
      }
      return func.apply(undefined, arguments);
    };
  }

  var setToString = shortOut(baseSetToString);

  function baseRest(func, start) {
    return setToString(overRest(func, start, identity), func + '');
  }

  function isIterateeCall(value, index, object) {
    if (!isObject$1(object)) {
      return false;
    }
    var type = typeof index;
    if (type == 'number'
          ? (isArrayLike(object) && isIndex(index, object.length))
          : (type == 'string' && index in object)
        ) {
      return eq(object[index], value);
    }
    return false;
  }

  function createAssigner(assigner) {
    return baseRest(function(object, sources) {
      var index = -1,
          length = sources.length,
          customizer = length > 1 ? sources[length - 1] : undefined,
          guard = length > 2 ? sources[2] : undefined;

      customizer = (assigner.length > 3 && typeof customizer == 'function')
        ? (length--, customizer)
        : undefined;

      if (guard && isIterateeCall(sources[0], sources[1], guard)) {
        customizer = length < 3 ? undefined : customizer;
        length = 1;
      }
      object = Object(object);
      while (++index < length) {
        var source = sources[index];
        if (source) {
          assigner(object, source, index, customizer);
        }
      }
      return object;
    });
  }

  var merge = createAssigner(function(object, source, srcIndex) {
    baseMerge(object, source, srcIndex);
  });

  var mergeWith = createAssigner(function(object, source, srcIndex, customizer) {
    baseMerge(object, source, srcIndex, customizer);
  });

  function baseSet(object, path, value, customizer) {
    if (!isObject$1(object)) {
      return object;
    }
    path = castPath(path, object);

    var index = -1,
        length = path.length,
        lastIndex = length - 1,
        nested = object;

    while (nested != null && ++index < length) {
      var key = toKey(path[index]),
          newValue = value;

      if (index != lastIndex) {
        var objValue = nested[key];
        newValue = customizer ? customizer(objValue, key, nested) : undefined;
        if (newValue === undefined) {
          newValue = isObject$1(objValue)
            ? objValue
            : (isIndex(path[index + 1]) ? [] : {});
        }
      }
      assignValue(nested, key, newValue);
      nested = nested[key];
    }
    return object;
  }

  function setWith(object, path, value, customizer) {
    customizer = typeof customizer == 'function' ? customizer : undefined;
    return object == null ? object : baseSet(object, path, value, customizer);
  }

  function baseIsNaN(value) {
    return value !== value;
  }

  function strictIndexOf(array, value, fromIndex) {
    var index = fromIndex - 1,
        length = array.length;

    while (++index < length) {
      if (array[index] === value) {
        return index;
      }
    }
    return -1;
  }

  function baseIndexOf(array, value, fromIndex) {
    return value === value
      ? strictIndexOf(array, value, fromIndex)
      : baseFindIndex(array, baseIsNaN, fromIndex);
  }

  function arrayIncludes(array, value) {
    var length = array == null ? 0 : array.length;
    return !!length && baseIndexOf(array, value, 0) > -1;
  }

  function arrayIncludesWith(array, value, comparator) {
    var index = -1,
        length = array == null ? 0 : array.length;

    while (++index < length) {
      if (comparator(value, array[index])) {
        return true;
      }
    }
    return false;
  }

  function noop() {
    
  }

  var INFINITY$3 = 1 / 0;


  var createSet = !(Set$1 && (1 / setToArray(new Set$1([,-0]))[1]) == INFINITY$3) ? noop : function(values) {
    return new Set$1(values);
  };

  function last(array) {
    var length = array == null ? 0 : array.length;
    return length ? array[length - 1] : undefined;
  }

  function parent(object, path) {
    return path.length < 2 ? object : baseGet(object, baseSlice(path, 0, -1));
  }

  function baseUnset(object, path) {
    path = castPath(path, object);
    object = parent(object, path);
    return object == null || delete object[toKey(last(path))];
  }

  function unset(object, path) {
    return object == null ? true : baseUnset(object, path);
  }

  var LARGE_ARRAY_SIZE$2 = 200;


  function baseDifference(array, values, iteratee, comparator) {
    var index = -1,
        includes = arrayIncludes,
        isCommon = true,
        length = array.length,
        result = [],
        valuesLength = values.length;

    if (!length) {
      return result;
    }
    if (iteratee) {
      values = arrayMap(values, baseUnary(iteratee));
    }
    if (comparator) {
      includes = arrayIncludesWith;
      isCommon = false;
    }
    else if (values.length >= LARGE_ARRAY_SIZE$2) {
      includes = cacheHas;
      isCommon = false;
      values = new SetCache(values);
    }
    outer:
    while (++index < length) {
      var value = array[index],
          computed = iteratee == null ? value : iteratee(value);

      value = (comparator || value !== 0) ? value : 0;
      if (isCommon && computed === computed) {
        var valuesIndex = valuesLength;
        while (valuesIndex--) {
          if (values[valuesIndex] === computed) {
            continue outer;
          }
        }
        result.push(value);
      }
      else if (!includes(values, computed, comparator)) {
        result.push(value);
      }
    }
    return result;
  }

  var without = baseRest(function(array, values) {
    return isArrayLikeObject(array)
      ? baseDifference(array, values)
      : [];
  });

  class PathObject {
    
    constructor (root) {
      if (root) {
        this.__root__ = root;
      }
    }

    contains (id) {
      return Boolean(this.getRoot()[id])
    }

    getRoot () {
      if (this.__root__) {
        return this.__root__
      } else {
        return this
      }
    }

    
    get (path) {
      if (!path) {
        return undefined
      }
      if (isString(path)) {
        return this.getRoot()[path]
      }
      if (arguments.length > 1) {
        path = Array.prototype.slice(arguments, 0);
      }
      if (!isArray(path)) {
        throw new Error('Illegal argument for PathObject.get()')
      }
      return get(this.getRoot(), path)
    }

    set (path, value) {
      if (!path) {
        throw new Error('Illegal argument: PathObject.set(>path<, value) - path is mandatory.')
      }
      if (isString(path)) {
        this.getRoot()[path] = value;
      } else {
        setWith(this.getRoot(), path, value);
      }
    }

    delete (path) {
      if (isString(path)) {
        delete this.getRoot()[path];
      } else if (path.length === 1) {
        delete this.getRoot()[path[0]];
      } else {
        var success = unset(this.getRoot(), path);
        if (!success) {
          throw new Error('Could not delete property at path' + path)
        }
      }
    }

    clear () {
      var root = this.getRoot();
      for (var key in root) {
        if (root.hasOwnProperty(key)) {
          delete root[key];
        }
      }
    }
  }

  PathObject.prototype._isPathObject = true;

  class Conflict extends Error {
    constructor (a, b) {
      super('Conflict: ' + JSON.stringify(a) + ' vs ' + JSON.stringify(b));
      this.a = a;
      this.b = b;
    }
  }

  const INSERT = 'insert';
  const DELETE = 'delete';

  class TextOperation {
    constructor (data) {
      if (!data || data.type === undefined || data.pos === undefined || data.str === undefined) {
        throw new Error('Illegal argument: insufficient data.')
      }
      
      this.type = data.type;
      
      this.pos = data.pos;
      
      this.str = data.str;
      
      if (!this.isInsert() && !this.isDelete()) {
        throw new Error('Illegal type.')
      }
      if (!isString(this.str)) {
        throw new Error('Illegal argument: expecting string.')
      }
      if (!isNumber(this.pos) || this.pos < 0) {
        throw new Error('Illegal argument: expecting positive number as pos.')
      }
    }

    apply (str) {
      if (this.isEmpty()) return str
      if (this.type === INSERT) {
        if (str.length < this.pos) {
          throw new Error('Provided string is too short.')
        }
        if (str.splice) {
          return str.splice(this.pos, 0, this.str)
        } else {
          return str.slice(0, this.pos).concat(this.str).concat(str.slice(this.pos))
        }
      } else  {
        if (str.length < this.pos + this.str.length) {
          throw new Error('Provided string is too short.')
        }
        if (str.splice) {
          return str.splice(this.pos, this.str.length)
        } else {
          return str.slice(0, this.pos).concat(str.slice(this.pos + this.str.length))
        }
      }
    }

    clone () {
      return new TextOperation(this)
    }

    isNOP () {
      return this.type === 'NOP' || this.str.length === 0
    }

    isInsert () {
      return this.type === INSERT
    }

    isDelete () {
      return this.type === DELETE
    }

    getLength () {
      return this.str.length
    }

    invert () {
      var data = {
        type: this.isInsert() ? DELETE : INSERT,
        pos: this.pos,
        str: this.str
      };
      return new TextOperation(data)
    }

    hasConflict (other) {
      return _hasConflict(this, other)
    }

    isEmpty () {
      return this.str.length === 0
    }

    toJSON () {
      return {
        type: this.type,
        pos: this.pos,
        str: this.str
      }
    }

    toString () {
      return ['(', (this.isInsert() ? INSERT : DELETE), ',', this.pos, ",'", this.str, "')"].join('')
    }
  }

  TextOperation.prototype._isOperation = true;
  TextOperation.prototype._isTextOperation = true;

  function _hasConflict (a, b) {
    
    
    
    if (a.type === INSERT && b.type === INSERT) return (a.pos === b.pos)
    
    
    
    if (a.type === DELETE && b.type === DELETE) {
      
      return !(a.pos >= b.pos + b.str.length || b.pos >= a.pos + a.str.length)
    }
    
    
    
    var del, ins;
    if (a.type === DELETE) {
      del = a; ins = b;
    } else {
      del = b; ins = a;
    }
    return (ins.pos >= del.pos && ins.pos < del.pos + del.str.length)
  }




  function transformInsertInsert (a, b) {
    if (a.pos === b.pos) {
      b.pos += a.str.length;
    } else if (a.pos < b.pos) {
      b.pos += a.str.length;
    } else {
      a.pos += b.str.length;
    }
  }





  function transformDeleteDelete (a, b, first) {
    
    if (a.pos > b.pos) {
      return transformDeleteDelete(b, a, !first)
    }
    if (a.pos === b.pos && a.str.length > b.str.length) {
      return transformDeleteDelete(b, a, !first)
    }
    
    if (b.pos < a.pos + a.str.length) {
      var s = b.pos - a.pos;
      var s1 = a.str.length - s;
      var s2 = s + b.str.length;
      a.str = a.str.slice(0, s) + a.str.slice(s2);
      b.str = b.str.slice(s1);
      b.pos -= s;
    } else {
      b.pos -= a.str.length;
    }
  }





  function transformInsertDelete (a, b) {
    if (a.type === DELETE) {
      return transformInsertDelete(b, a)
    }
    
    
    if (a.pos <= b.pos) {
      b.pos += a.str.length;
    
    } else if (a.pos >= b.pos + b.str.length) {
      a.pos -= b.str.length;
    
    
    
    } else {
      var s = a.pos - b.pos;
      b.str = b.str.slice(0, s) + a.str + b.str.slice(s);
      a.str = '';
    }
  }

  function transform (a, b, options) {
    options = options || {};
    if (options['no-conflict'] && _hasConflict(a, b)) {
      throw new Conflict(a, b)
    }
    if (!options.inplace) {
      a = a.clone();
      b = b.clone();
    }
    if (a.type === INSERT && b.type === INSERT) {
      transformInsertInsert(a, b);
    } else if (a.type === DELETE && b.type === DELETE) {
      transformDeleteDelete(a, b, true);
    } else {
      transformInsertDelete(a, b);
    }
    return [a, b]
  }

  TextOperation.transform = function () {
    return transform.apply(null, arguments)
  };



  TextOperation.Insert = function (pos, str) {
    return new TextOperation({ type: INSERT, pos: pos, str: str })
  };

  TextOperation.Delete = function (pos, str) {
    return new TextOperation({ type: DELETE, pos: pos, str: str })
  };

  TextOperation.INSERT = INSERT;
  TextOperation.DELETE = DELETE;

  TextOperation.fromJSON = function (data) {
    return new TextOperation(data)
  };

  const NOP = 'NOP';
  const DELETE$1 = 'delete';
  const INSERT$1 = 'insert';

  class ArrayOperation {
    constructor (data) {
      if (!data || !data.type) {
        throw new Error('Illegal argument: insufficient data.')
      }
      this.type = data.type;
      if (this.type === NOP) return

      if (this.type !== INSERT$1 && this.type !== DELETE$1) {
        throw new Error('Illegal type.')
      }
      
      this.pos = data.pos;
      
      this.val = data.val;
      if (!isNumber(this.pos) || this.pos < 0) {
        throw new Error('Illegal argument: expecting positive number as pos.')
      }
    }

    apply (array) {
      if (this.type === NOP) {
        return array
      }
      if (this.type === INSERT$1) {
        if (array.length < this.pos) {
          throw new Error('Provided array is too small.')
        }
        array.splice(this.pos, 0, this.val);
        return array
      
      } else  {
        if (array.length < this.pos) {
          throw new Error('Provided array is too small.')
        }
        if (!isEqual(array[this.pos], this.val)) {
          throw Error('Unexpected value at position ' + this.pos + '. Expected ' + this.val + ', found ' + array[this.pos])
        }
        array.splice(this.pos, 1);
        return array
      }
    }

    clone () {
      var data = {
        type: this.type,
        pos: this.pos,
        val: cloneDeep(this.val)
      };
      return new ArrayOperation(data)
    }

    invert () {
      var data = this.toJSON();
      if (this.type === NOP) data.type = NOP;
      else if (this.type === INSERT$1) data.type = DELETE$1;
      else  data.type = INSERT$1;
      return new ArrayOperation(data)
    }

    hasConflict (other) {
      return ArrayOperation.hasConflict(this, other)
    }

    toJSON () {
      var result = {
        type: this.type
      };
      if (this.type === NOP) return result
      result.pos = this.pos;
      result.val = cloneDeep(this.val);
      return result
    }

    isInsert () {
      return this.type === INSERT$1
    }

    isDelete () {
      return this.type === DELETE$1
    }

    getOffset () {
      return this.pos
    }

    getValue () {
      return this.val
    }

    isNOP () {
      return this.type === NOP
    }

    toString () {
      return ['(', (this.isInsert() ? INSERT$1 : DELETE$1), ',', this.getOffset(), ",'", this.getValue(), "')"].join('')
    }
  }

  ArrayOperation.prototype._isOperation = true;
  ArrayOperation.prototype._isArrayOperation = true;

  function hasConflict (a, b) {
    if (a.type === NOP || b.type === NOP) return false
    if (a.type === INSERT$1 && b.type === INSERT$1) {
      return a.pos === b.pos
    } else {
      return false
    }
  }

  function transformInsertInsert$1 (a, b) {
    if (a.pos === b.pos) {
      b.pos += 1;
    
    } else if (a.pos < b.pos) {
      b.pos += 1;
    
    } else {
      a.pos += 1;
    }
  }

  function transformDeleteDelete$1 (a, b) {
    
    if (a.pos === b.pos) {
      b.type = NOP;
      a.type = NOP;
      return
    }
    if (a.pos < b.pos) {
      b.pos -= 1;
    } else {
      a.pos -= 1;
    }
  }

  function transformInsertDelete$1 (a, b) {
    
    if (a.type === DELETE$1) {
      var tmp = a;
      a = b;
      b = tmp;
    }
    if (a.pos <= b.pos) {
      b.pos += 1;
    } else {
      a.pos -= 1;
    }
  }

  var transform$1 = function (a, b, options) {
    options = options || {};
    
    
    if (options['no-conflict'] && hasConflict(a, b)) {
      throw new Conflict(a, b)
    }
    
    if (!options.inplace) {
      a = a.clone();
      b = b.clone();
    }
    if (a.type === NOP || b.type === NOP) ; else if (a.type === INSERT$1 && b.type === INSERT$1) {
      transformInsertInsert$1(a, b);
    } else if (a.type === DELETE$1 && b.type === DELETE$1) {
      transformDeleteDelete$1(a, b);
    } else {
      transformInsertDelete$1(a, b);
    }
    return [a, b]
  };

  ArrayOperation.transform = transform$1;
  ArrayOperation.hasConflict = hasConflict;



  ArrayOperation.Insert = function (pos, val) {
    return new ArrayOperation({type: INSERT$1, pos: pos, val: val})
  };

  ArrayOperation.Delete = function (pos, val) {
    return new ArrayOperation({ type: DELETE$1, pos: pos, val: val })
  };

  ArrayOperation.fromJSON = function (data) {
    return new ArrayOperation(data)
  };

  ArrayOperation.NOP = NOP;
  ArrayOperation.DELETE = DELETE$1;
  ArrayOperation.INSERT = INSERT$1;

  const SHIFT = 'shift';

  class CoordinateOperation {
    constructor (data) {
      if (!data || data.type === undefined) {
        throw new Error('Illegal argument: insufficient data.')
      }
      
      this.type = data.type;
      
      this.val = data.val;
      
      if (!this.isShift()) {
        throw new Error('Illegal type.')
      }
      if (!isNumber(this.val)) {
        throw new Error('Illegal argument: expecting number as shift value.')
      }
    }

    apply (coor) {
      coor.offset = coor.offset + this.val;
    }

    isShift () {
      return this.type === SHIFT
    }

    isNOP () {
      switch (this.type) {
        case SHIFT: {
          return this.val === 0
        }
        default:
          return false
      }
    }

    clone () {
      return new CoordinateOperation(this)
    }

    invert () {
      let data;
      switch (this.type) {
        case SHIFT:
          data = {
            type: SHIFT,
            val: -this.val
          };
          break
        default:
          throw new Error('Invalid type.')
      }
      return new CoordinateOperation(data)
    }

    hasConflict () {
      
      return false
    }

    toJSON () {
      return {
        type: this.type,
        val: this.val
      }
    }

    toString () {
      return ['(', (this.type), ',', this.val, "')"].join('')
    }
  }

  CoordinateOperation.prototype._isOperation = true;
  CoordinateOperation.prototype._isCoordinateOperation = true;

  function transformShiftShift (a, b) {
    a.val += b.val;
    b.val += a.val;
  }

  function transform$2 (a, b, options) {
    options = options || {};
    
    if (!options.inplace) {
      a = a.clone();
      b = b.clone();
    }
    if (a.type === SHIFT && b.type === SHIFT) {
      transformShiftShift(a, b);
    } else {
      throw new Error('Illegal type')
    }
    return [a, b]
  }

  CoordinateOperation.transform = function (...args) {
    return transform$2(...args)
  };

  CoordinateOperation.fromJSON = function (json) {
    return new CoordinateOperation(json)
  };

  CoordinateOperation.Shift = function (val) {
    return new CoordinateOperation({
      type: SHIFT,
      val: val
    })
  };

  const NOP$1 = 'NOP';
  const CREATE = 'create';
  const DELETE$2 = 'delete';
  const UPDATE = 'update';
  const SET = 'set';

  class ObjectOperation {
    constructor (data) {
      
      if (!data) {
        throw new Error('Data of ObjectOperation is missing.')
      }
      
      if (!data.type) {
        throw new Error('Invalid data: type is mandatory.')
      }
      this.type = data.type;
      if (data.type === NOP$1) {
        return
      }
      this.path = data.path;
      if (!data.path) {
        throw new Error('Invalid data: path is mandatory.')
      }
      if (this.type === CREATE || this.type === DELETE$2) {
        if (!data.val) {
          throw new Error('Invalid data: value is missing.')
        }
        this.val = data.val;
      } else if (this.type === UPDATE) {
        if (data.diff) {
          this.diff = data.diff;
          if (data.diff._isTextOperation) {
            this.propertyType = 'string';
          } else if (data.diff._isArrayOperation) {
            this.propertyType = 'array';
          } else if (data.diff._isCoordinateOperation) {
            this.propertyType = 'coordinate';
          } else {
            throw new Error('Invalid data: diff must be a TextOperation or an ArrayOperation.')
          }
        } else {
          throw new Error('Invalid data: diff is mandatory for update operation.')
        }
      } else if (this.type === SET) {
        this.val = data.val;
        this.original = data.original;
      } else {
        throw new Error('Invalid type: ' + data.type)
      }
    }

    apply (obj) {
      if (this.type === NOP$1) return obj
      var adapter;
      if (obj._isPathObject) {
        adapter = obj;
      } else {
        adapter = new PathObject(obj);
      }
      if (this.type === CREATE) {
        adapter.set(this.path, cloneDeep(this.val));
        return obj
      }
      if (this.type === DELETE$2) {
        adapter.delete(this.path, 'strict');
      } else if (this.type === UPDATE) {
        var diff = this.diff;
        switch (this.propertyType) {
          case 'array': {
            let arr = adapter.get(this.path);
            diff.apply(arr);
            break
          }
          case 'string': {
            let str = adapter.get(this.path);
            if (isNil(str)) str = '';
            str = diff.apply(str);
            adapter.set(this.path, str);
            break
          }
          case 'coordinate': {
            let coor = adapter.get(this.path);
            if (!coor) throw new Error('No coordinate with path ' + this.path)
            diff.apply(coor);
            break
          }
          default:
            throw new Error('Invalid state.')
        }
      } else if (this.type === SET) {
        
        adapter.set(this.path, cloneDeep(this.val));
      } else {
        throw new Error('Invalid type.')
      }
      return obj
    }

    clone () {
      var data = {
        type: this.type,
        path: this.path
      };
      if (this.val) {
        data.val = cloneDeep(this.val);
      }
      if (this.diff) {
        data.diff = this.diff.clone();
      }
      return new ObjectOperation(data)
    }

    isNOP () {
      if (this.type === NOP$1) return true
      else if (this.type === UPDATE) return this.diff.isNOP()
    }

    isCreate () {
      return this.type === CREATE
    }

    isDelete () {
      return this.type === DELETE$2
    }

    isUpdate (propertyType) {
      if (propertyType) {
        return (this.type === UPDATE && this.propertyType === propertyType)
      } else {
        return this.type === UPDATE
      }
    }

    isSet () {
      return this.type === SET
    }

    invert () {
      if (this.type === NOP$1) {
        return new ObjectOperation({ type: NOP$1 })
      }
      var result = new ObjectOperation(this);
      if (this.type === CREATE) {
        result.type = DELETE$2;
      } else if (this.type === DELETE$2) {
        result.type = CREATE;
      } else if (this.type === UPDATE) {
        var invertedDiff;
        if (this.diff._isTextOperation) {
          invertedDiff = TextOperation.fromJSON(this.diff.toJSON()).invert();
        } else if (this.diff._isArrayOperation) {
          invertedDiff = ArrayOperation.fromJSON(this.diff.toJSON()).invert();
        } else if (this.diff._isCoordinateOperation) {
          invertedDiff = CoordinateOperation.fromJSON(this.diff.toJSON()).invert();
        } else {
          throw new Error('Illegal type')
        }
        result.diff = invertedDiff;
      } else  {
        result.val = this.original;
        result.original = this.val;
      }
      return result
    }

    hasConflict (other) {
      return ObjectOperation.hasConflict(this, other)
    }

    toJSON () {
      if (this.type === NOP$1) {
        return { type: NOP$1 }
      }
      var data = {
        type: this.type,
        path: this.path
      };
      if (this.type === CREATE || this.type === DELETE$2) {
        data.val = this.val;
      } else if (this.type === UPDATE) {
        if (this.diff._isTextOperation) {
          data.propertyType = 'string';
        } else if (this.diff._isArrayOperation) {
          data.propertyType = 'array';
        } else if (this.diff._isCoordinateOperation) {
          data.propertyType = 'coordinate';
        } else {
          throw new Error('Invalid property type.')
        }
        data.diff = this.diff.toJSON();
      } else  {
        data.val = this.val;
        data.original = this.original;
      }
      return data
    }

    getType () {
      return this.type
    }

    getPath () {
      return this.path
    }

    getValue () {
      return this.val
    }

    getOldValue () {
      return this.original
    }

    getValueOp () {
      return this.diff
    }

    
    toString () {
      switch (this.type) {
        case CREATE:
          return ['(+,', JSON.stringify(this.path), JSON.stringify(this.val), ')'].join('')
        case DELETE$2:
          return ['(-,', JSON.stringify(this.path), JSON.stringify(this.val), ')'].join('')
        case UPDATE:
          return ['(>>,', JSON.stringify(this.path), this.propertyType, this.diff.toString(), ')'].join('')
        case SET:
          return ['(=,', JSON.stringify(this.path), this.val, this.original, ')'].join('')
        case NOP$1:
          return 'NOP'
        default:
          throw new Error('Invalid type')
      }
    }
  }

  ObjectOperation.prototype._isOperation = true;
  ObjectOperation.prototype._isObjectOperation = true;



  function hasConflict$1 (a, b) {
    if (a.type === NOP$1 || b.type === NOP$1) return false
    return isEqual(a.path, b.path)
  }

  function transformDeleteDelete$2 (a, b) {
    
    
    a.type = NOP$1;
    b.type = NOP$1;
  }

  function transformCreateCreate () {
    throw new Error('Can not transform two concurring creates of the same property')
  }

  function transformDeleteCreate () {
    throw new Error('Illegal state: can not create and delete a value at the same time.')
  }

  function transformDeleteUpdate (a, b, flipped) {
    if (a.type !== DELETE$2) {
      return transformDeleteUpdate(b, a, true)
    }
    var op;
    switch (b.propertyType) {
      case 'string':
        op = TextOperation.fromJSON(b.diff);
        break
      case 'array':
        op = ArrayOperation.fromJSON(b.diff);
        break
      case 'coordinate':
        op = CoordinateOperation.fromJSON(b.diff);
        break
      default:
        throw new Error('Illegal type')
    }
    
    if (!flipped) {
      a.type = NOP$1;
      b.type = CREATE;
      b.val = op.apply(a.val);
    
    } else {
      a.val = op.apply(a.val);
      b.type = NOP$1;
    }
  }

  function transformCreateUpdate () {
    
    throw new Error('Can not transform a concurring create and update of the same property')
  }

  function transformUpdateUpdate (a, b) {
    
    let opA, opB, t;
    switch (b.propertyType) {
      case 'string':
        opA = TextOperation.fromJSON(a.diff);
        opB = TextOperation.fromJSON(b.diff);
        t = TextOperation.transform(opA, opB, {inplace: true});
        break
      case 'array':
        opA = ArrayOperation.fromJSON(a.diff);
        opB = ArrayOperation.fromJSON(b.diff);
        t = ArrayOperation.transform(opA, opB, {inplace: true});
        break
      case 'coordinate':
        opA = CoordinateOperation.fromJSON(a.diff);
        opB = CoordinateOperation.fromJSON(b.diff);
        t = CoordinateOperation.transform(opA, opB, {inplace: true});
        break
      default:
        throw new Error('Illegal type')
    }
    a.diff = t[0];
    b.diff = t[1];
  }

  function transformCreateSet () {
    throw new Error('Illegal state: can not create and set a value at the same time.')
  }

  function transformDeleteSet (a, b, flipped) {
    if (a.type !== DELETE$2) return transformDeleteSet(b, a, true)
    if (!flipped) {
      a.type = NOP$1;
      b.type = CREATE;
      b.original = undefined;
    } else {
      a.val = b.val;
      b.type = NOP$1;
    }
  }

  function transformUpdateSet () {
    throw new Error('Unresolvable conflict: update + set.')
  }

  function transformSetSet (a, b) {
    a.type = NOP$1;
    b.original = a.val;
  }

  const _NOP = 0;
  const _CREATE = 1;
  const _DELETE = 2;
  const _UPDATE = 4;
  const _SET = 8;

  const CODE = (() => {
    const c = {};
    c[NOP$1] = _NOP;
    c[CREATE] = _CREATE;
    c[DELETE$2] = _DELETE;
    c[UPDATE] = _UPDATE;
    c[SET] = _SET;
    return c
  })();

  const __transform__ = (() => {
    
    const t = {};
    t[_DELETE | _DELETE] = transformDeleteDelete$2;
    t[_DELETE | _CREATE] = transformDeleteCreate;
    t[_DELETE | _UPDATE] = transformDeleteUpdate;
    t[_CREATE | _CREATE] = transformCreateCreate;
    t[_CREATE | _UPDATE] = transformCreateUpdate;
    t[_UPDATE | _UPDATE] = transformUpdateUpdate;
    t[_CREATE | _SET] = transformCreateSet;
    t[_DELETE | _SET] = transformDeleteSet;
    t[_UPDATE | _SET] = transformUpdateSet;
    t[_SET    | _SET] = transformSetSet;
    
    return t
  })();

  function transform$3 (a, b, options) {
    options = options || {};
    if (options['no-conflict'] && hasConflict$1(a, b)) {
      throw new Conflict(a, b)
    }
    if (!options.inplace) {
      a = a.clone();
      b = b.clone();
    }
    if (a.isNOP() || b.isNOP()) {
      return [a, b]
    }
    var sameProp = isEqual(a.path, b.path);
    
    if (sameProp) {
      __transform__[CODE[a.type] | CODE[b.type]](a, b);
    }
    return [a, b]
  }

  ObjectOperation.transform = transform$3;
  ObjectOperation.hasConflict = hasConflict$1;



  ObjectOperation.Create = function (idOrPath, val) {
    var path;
    if (isString(idOrPath)) {
      path = [idOrPath];
    } else {
      path = idOrPath;
    }
    return new ObjectOperation({type: CREATE, path: path, val: val})
  };

  ObjectOperation.Delete = function (idOrPath, val) {
    var path;
    if (isString(idOrPath)) {
      path = [idOrPath];
    } else {
      path = idOrPath;
    }
    return new ObjectOperation({type: DELETE$2, path: path, val: val})
  };

  ObjectOperation.Update = function (path, op) {
    return new ObjectOperation({
      type: UPDATE,
      path: path,
      diff: op
    })
  };

  ObjectOperation.Set = function (path, oldVal, newVal) {
    return new ObjectOperation({
      type: SET,
      path: path,
      val: cloneDeep(newVal),
      original: cloneDeep(oldVal)
    })
  };

  ObjectOperation.fromJSON = function (data) {
    data = cloneDeep(data);
    if (data.type === 'update') {
      switch (data.propertyType) {
        case 'string':
          data.diff = TextOperation.fromJSON(data.diff);
          break
        case 'array':
          data.diff = ArrayOperation.fromJSON(data.diff);
          break
        case 'coordinate':
          data.diff = CoordinateOperation.fromJSON(data.diff);
          break
        default:
          throw new Error('Unsupported update diff:' + JSON.stringify(data.diff))
      }
    }
    var op = new ObjectOperation(data);
    return op
  };

  ObjectOperation.NOP = NOP$1;
  ObjectOperation.CREATE = CREATE;
  ObjectOperation.DELETE = DELETE$2;
  ObjectOperation.UPDATE = UPDATE;
  ObjectOperation.SET = SET;

  class OperationSerializer {
    constructor () {
      this.SEPARATOR = '\t';
    }

    serialize (op) {
      var out = [];
      switch (op.type) {
        case 'create':
          out.push('c');
          out.push(op.val.id);
          out.push(op.val);
          break
        case 'delete':
          out.push('d');
          out.push(op.val.id);
          out.push(op.val);
          break
        case 'set':
          out.push('s');
          out.push(op.path.join('.'));
          out.push(op.val);
          out.push(op.original);
          break
        case 'update':
          out.push('u');
          out.push(op.path.join('.'));
          Array.prototype.push.apply(out, this.serializePrimitiveOp(op.diff));
          break
        default:
          throw new Error('Unsupported operation type.')
      }
      return out
    }

    serializePrimitiveOp (op) {
      var out = [];
      if (op._isTextOperation) {
        if (op.isInsert()) {
          out.push('t+');
        } else if (op.isDelete()) {
          out.push('t-');
        }
        out.push(op.pos);
        out.push(op.str);
      } else if (op._isArrayOperation) {
        if (op.isInsert()) {
          out.push('a+');
        } else if (op.isDelete()) {
          out.push('a-');
        }
        out.push(op.pos);
        out.push(op.val);
      } else if (op._isCoordinateOperation) {
        if (op.isShift()) {
          out.push('c>>');
        } else {
          throw new Error('Unsupported CoordinateOperation type.')
        }
        out.push(op.pos);
        out.push(op.val);
      } else {
        throw new Error('Unsupported operation type.')
      }
      return out
    }

    deserialize (str, tokenizer) {
      if (!tokenizer) {
        tokenizer = new Tokenizer(str, this.SEPARATOR);
      }
      var type = tokenizer.getString();
      var op, path, val, oldVal, diff;
      switch (type) {
        case 'c':
          path = tokenizer.getPath();
          val = tokenizer.getObject();
          op = ObjectOperation.Create(path, val);
          break
        case 'd':
          path = tokenizer.getPath();
          val = tokenizer.getObject();
          op = ObjectOperation.Delete(path, val);
          break
        case 's':
          path = tokenizer.getPath();
          val = tokenizer.getAny();
          oldVal = tokenizer.getAny();
          op = ObjectOperation.Set(path, oldVal, val);
          break
        case 'u':
          path = tokenizer.getPath();
          diff = this.deserializePrimitiveOp(str, tokenizer);
          op = ObjectOperation.Update(path, diff);
          break
        default:
          throw new Error('Illegal type for ObjectOperation: ' + type)
      }
      return op
    }

    deserializePrimitiveOp (str, tokenizer) {
      if (!tokenizer) {
        tokenizer = new Tokenizer(str, this.SEPARATOR);
      }
      var type = tokenizer.getString();
      var op, pos, val;
      switch (type) {
        case 't+':
          pos = tokenizer.getNumber();
          val = tokenizer.getString();
          op = TextOperation.Insert(pos, val);
          break
        case 't-':
          pos = tokenizer.getNumber();
          val = tokenizer.getString();
          op = TextOperation.Delete(pos, val);
          break
        case 'a+':
          pos = tokenizer.getNumber();
          val = tokenizer.getAny();
          op = ArrayOperation.Insert(pos, val);
          break
        case 'a-':
          pos = tokenizer.getNumber();
          val = tokenizer.getAny();
          op = ArrayOperation.Delete(pos, val);
          break
        case 'c>>':
          val = tokenizer.getNumber();
          op = CoordinateOperation.Shift(val);
          break
        default:
          throw new Error('Unsupported operation type: ' + type)
      }
      return op
    }
  }

  class Tokenizer {
    constructor (str, sep) {
      if (isArray(arguments[0])) {
        this.tokens = arguments[0];
      } else {
        this.tokens = str.split(sep);
      }
      this.pos = -1;
    }

    error (msg) {
      throw new Error('Parsing error: ' + msg + '\n' + this.tokens[this.pos])
    }

    getString () {
      this.pos++;
      var str = this.tokens[this.pos];
      if (str[0] === '"') {
        str = str.slice(1, -1);
      }
      return str
    }

    getNumber () {
      this.pos++;
      var number;
      var token = this.tokens[this.pos];
      try {
        if (isNumber(token)) {
          number = token;
        } else {
          number = parseInt(this.tokens[this.pos], 10);
        }
        return number
      } catch (err) {
        this.error('expected number');
      }
    }

    getObject () {
      this.pos++;
      var obj;
      var token = this.tokens[this.pos];
      try {
        if (isObject(token)) {
          obj = token;
        } else {
          obj = JSON.parse(this.tokens[this.pos]);
        }
        return obj
      } catch (err) {
        this.error('expected object');
      }
    }

    getAny () {
      this.pos++;
      var token = this.tokens[this.pos];
      return token
    }

    getPath () {
      var str = this.getString();
      return str.split('.')
    }
  }

  OperationSerializer.Tokenizer = Tokenizer;

  function last$1 (arr) {
    return arr[arr.length - 1]
  }

  class Selection {
    constructor () {
      
      var _internal = {};
      Object.defineProperty(this, '_internal', {
        enumerable: false,
        value: _internal
      });
      
      _internal.doc = null;
    }

    clone () {
      var newSel = this._clone();
      if (this._internal.doc) {
        newSel.attach(this._internal.doc);
      }
      return newSel
    }

    
    getDocument () {
      var doc = this._internal.doc;
      if (!doc) {
        throw new Error('Selection is not attached to a document.')
      }
      return doc
    }

    isAttached () {
      return Boolean(this._internal.doc)
    }

    
    attach (doc) {
      this._internal.doc = doc;
      return this
    }

    
    isNull () { return false }

    
    isPropertySelection () { return false }

    
    isContainerSelection () { return false }

    
    isNodeSelection () { return false }

    isCustomSelection () { return false }

    
    isCollapsed () { return true }

    
    isReverse () { return false }

    getType () {
      throw new Error('Selection.getType() is abstract.')
    }

    get type () {
      return this.getType()
    }

    
    equals (other) {
      if (this === other) {
        return true
      } else if (!other) {
        return false
      } else if (this.isNull() !== other.isNull()) {
        return false
      } else if (this.getType() !== other.getType()) {
        return false
      } else {
        
        
        return true
      }
    }

    
    toString () {
      return 'null'
    }

    
    toJSON () {
      throw new Error('This method is abstract.')
    }

    createWith (update) {
      let SelectionClass = this.constructor;
      let data = this.toJSON();
      Object.assign(data, update);
      return SelectionClass.fromJSON(data)
    }
  }


  Selection.prototype._isSelection = true;


  class NullSelection extends Selection {
    isNull () {
      return true
    }

    getType () {
      return 'null'
    }

    toJSON () {
      return null
    }

    clone () {
      return this
    }
  }



  Selection.nullSelection = Object.freeze(new NullSelection());

  function isArrayEqual (arr1, arr2) {
    if (arr1 === arr2) return true
    if (!isArray(arr1) || !isArray(arr2)) return false
    if (arr1.length !== arr2.length) return false
    let L = arr1.length;
    for (var i = 0; i < L; i++) {
      if (arr1[i] !== arr2[i]) return false
    }
    return true
  }

  class Coordinate {
    
    constructor (path, offset) {
      
      if (arguments[0] === 'SKIP') return
      if (arguments.length === 1) {
        let data = arguments[0];
        this.path = data.path;
        this.offset = data.offset;
      } else {
        this.path = path;
        this.offset = offset;
      }
      if (!isArray(this.path)) {
        throw new Error('Invalid arguments: path should be an array.')
      }
      if (!isNumber(this.offset) || this.offset < 0) {
        throw new Error('Invalid arguments: offset must be a positive number.')
      }
    }

    equals (other) {
      return (other === this ||
        (isArrayEqual(other.path, this.path) && other.offset === this.offset))
    }

    withCharPos (offset) {
      return new Coordinate(this.path, offset)
    }

    getNodeId () {
      return this.path[0]
    }

    getPath () {
      return this.path
    }

    getOffset () {
      return this.offset
    }

    toJSON () {
      return {
        path: this.path.slice(),
        offset: this.offset
      }
    }

    toString () {
      return '(' + this.path.join('.') + ', ' + this.offset + ')'
    }

    isPropertyCoordinate () {
      return this.path.length > 1
    }

    isNodeCoordinate () {
      return this.path.length === 1
    }

    hasSamePath (other) {
      return isArrayEqual(this.path, other.path)
    }
  }

  Coordinate.prototype._isCoordinate = true;

  class PropertySelection extends Selection {
    
    constructor (path, startOffset, endOffset, reverse, containerId, surfaceId) {
      super();

      if (arguments.length === 1) {
        let data = arguments[0];
        path = data.path;
        startOffset = data.startOffset;
        endOffset = data.endOffset;
        reverse = data.reverse;
        containerId = data.containerId;
        surfaceId = data.surfaceId;
      }

      if (!path || !isNumber(startOffset)) {
        throw new Error('Invalid arguments: `path` and `startOffset` are mandatory')
      }

      this.start = new Coordinate(path, startOffset);
      this.end = new Coordinate(path, isNumber(endOffset) ? endOffset : startOffset);

      
      this.reverse = Boolean(reverse);

      this.containerId = containerId;

      
      this.surfaceId = surfaceId;
    }

    get path () {
      return this.start.path
    }

    get startOffset () {
      console.warn('DEPRECATED: Use sel.start.offset instead');
      return this.start.offset
    }

    get endOffset () {
      console.warn('DEPRECATED: Use sel.end.offset instead');
      return this.end.offset
    }

    
    toJSON () {
      return {
        type: 'property',
        path: this.start.path,
        startOffset: this.start.offset,
        endOffset: this.end.offset,
        reverse: this.reverse,
        containerId: this.containerId,
        surfaceId: this.surfaceId
      }
    }

    isPropertySelection () {
      return true
    }

    getType () {
      return 'property'
    }

    isNull () {
      return false
    }

    isCollapsed () {
      return this.start.offset === this.end.offset
    }

    isReverse () {
      return this.reverse
    }

    equals (other) {
      return (
        Selection.prototype.equals.call(this, other) &&
        (this.start.equals(other.start) && this.end.equals(other.end))
      )
    }

    toString () {
      
      return [
        'PropertySelection(', JSON.stringify(this.path), ', ',
        this.start.offset, ' -> ', this.end.offset,
        (this.reverse ? ', reverse' : ''),
        (this.surfaceId ? (', ' + this.surfaceId) : ''),
        ')'
      ].join('')
    }

    
    collapse (direction) {
      var offset;
      if (direction === 'left') {
        offset = this.start.offset;
      } else {
        offset = this.end.offset;
      }
      return this.createWithNewRange(offset, offset)
    }

    
    

    
    getPath () {
      return this.start.path
    }

    getNodeId () {
      return this.start.path[0]
    }

    
    isInsideOf (other, strict) {
      if (other.isNull()) return false
      if (other.isContainerSelection()) {
        return other.contains(this, strict)
      }
      if (strict) {
        return (isArrayEqual(this.path, other.path) &&
          this.start.offset > other.start.offset &&
          this.end.offset < other.end.offset)
      } else {
        return (isArrayEqual(this.path, other.path) &&
          this.start.offset >= other.start.offset &&
          this.end.offset <= other.end.offset)
      }
    }

    
    contains (other, strict) {
      if (other.isNull()) return false
      return other.isInsideOf(this, strict)
    }

    
    overlaps (other, strict) {
      if (other.isNull()) return false
      if (other.isContainerSelection()) {
        
        return other.overlaps(this)
      }
      if (!isArrayEqual(this.path, other.path)) return false
      if (strict) {
        return (!(this.start.offset >= other.end.offset || this.end.offset <= other.start.offset))
      } else {
        return (!(this.start.offset > other.end.offset || this.end.offset < other.start.offset))
      }
    }

    
    isRightAlignedWith (other) {
      if (other.isNull()) return false
      if (other.isContainerSelection()) {
        
        return other.isRightAlignedWith(this)
      }
      return (isArrayEqual(this.path, other.path) &&
        this.end.offset === other.end.offset)
    }

    
    isLeftAlignedWith (other) {
      if (other.isNull()) return false
      if (other.isContainerSelection()) {
        
        return other.isLeftAlignedWith(this)
      }
      return (isArrayEqual(this.path, other.path) &&
        this.start.offset === other.start.offset)
    }

    
    expand (other) {
      if (other.isNull()) return this

      
      
      
      if (other.isContainerSelection()) {
        return other.expand(this)
      }
      if (!isArrayEqual(this.path, other.path)) {
        throw new Error('Can not expand PropertySelection to a different property.')
      }
      var newStartOffset = Math.min(this.start.offset, other.start.offset);
      var newEndOffset = Math.max(this.end.offset, other.end.offset);
      return this.createWithNewRange(newStartOffset, newEndOffset)
    }

    
    truncateWith (other) {
      if (other.isNull()) return this
      if (other.isInsideOf(this, 'strict')) {
        
        throw new Error('Can not truncate with a contained selections')
      }
      if (!this.overlaps(other)) {
        return this
      }
      var otherStartOffset, otherEndOffset;
      if (other.isPropertySelection()) {
        otherStartOffset = other.start.offset;
        otherEndOffset = other.end.offset;
      } else if (other.isContainerSelection()) {
        
        if (isArrayEqual(other.start.path, this.start.path)) {
          otherStartOffset = other.start.offset;
        } else {
          otherStartOffset = this.start.offset;
        }
        if (isArrayEqual(other.end.path, this.start.path)) {
          otherEndOffset = other.end.offset;
        } else {
          otherEndOffset = this.end.offset;
        }
      } else {
        return this
      }

      var newStartOffset;
      var newEndOffset;
      if (this.start.offset > otherStartOffset && this.end.offset > otherEndOffset) {
        newStartOffset = otherEndOffset;
        newEndOffset = this.end.offset;
      } else if (this.start.offset < otherStartOffset && this.end.offset < otherEndOffset) {
        newStartOffset = this.start.offset;
        newEndOffset = otherStartOffset;
      } else if (this.start.offset === otherStartOffset) {
        if (this.end.offset <= otherEndOffset) {
          return Selection.nullSelection
        } else {
          newStartOffset = otherEndOffset;
          newEndOffset = this.end.offset;
        }
      } else if (this.end.offset === otherEndOffset) {
        if (this.start.offset >= otherStartOffset) {
          return Selection.nullSelection
        } else {
          newStartOffset = this.start.offset;
          newEndOffset = otherStartOffset;
        }
      } else if (other.contains(this)) {
        return Selection.nullSelection
      } else {
        
        throw new Error('Illegal state.')
      }
      return this.createWithNewRange(newStartOffset, newEndOffset)
    }

    
    createWithNewRange (startOffset, endOffset) {
      var sel = new PropertySelection(this.path, startOffset, endOffset, false, this.containerId, this.surfaceId);
      var doc = this._internal.doc;
      if (doc) {
        sel.attach(doc);
      }
      return sel
    }

    _clone () {
      return new PropertySelection(this.start.path, this.start.offset, this.end.offset, this.reverse, this.containerId, this.surfaceId)
    }
  }

  PropertySelection.fromJSON = function (json) {
    return new PropertySelection(json)
  };

  class ContainerSelection extends Selection {
    constructor (containerId, startPath, startOffset, endPath, endOffset, reverse, surfaceId) {
      super();

      if (arguments.length === 1) {
        let data = arguments[0];
        containerId = data.containerId;
        startPath = data.startPath;
        startOffset = data.startOffset;
        endPath = data.endPath;
        endOffset = data.endOffset;
        reverse = data.reverse;
        surfaceId = data.surfaceId;
      }

      
      this.containerId = containerId;
      if (!this.containerId) throw new Error('Invalid arguments: `containerId` is mandatory')

      this.start = new Coordinate(startPath, startOffset);
      this.end = new Coordinate(isNil(endPath) ? startPath : endPath, isNil(endOffset) ? startOffset : endOffset);

      this.reverse = Boolean(reverse);

      this.surfaceId = surfaceId;
    }

    

    get startPath () {
      console.warn('DEPRECATED: use sel.start.path instead.');
      return this.start.path
    }

    get startOffset () {
      console.warn('DEPRECATED: use sel.start.offset instead.');
      return this.start.offset
    }

    get endPath () {
      console.warn('DEPRECATED: use sel.end.path instead.');
      return this.end.path
    }

    get endOffset () {
      console.warn('DEPRECATED: use sel.end.offset instead.');
      return this.end.offset
    }

    

    toJSON () {
      return {
        type: 'container',
        containerId: this.containerId,
        startPath: this.start.path,
        startOffset: this.start.offset,
        endPath: this.end.path,
        endOffset: this.end.offset,
        reverse: this.reverse,
        surfaceId: this.surfaceId
      }
    }

    isContainerSelection () {
      return true
    }

    getType () {
      return 'container'
    }

    isNull () {
      return false
    }

    isCollapsed () {
      return this.start.equals(this.end)
    }

    isReverse () {
      return this.reverse
    }

    equals (other) {
      return (
        Selection.prototype.equals.call(this, other) &&
        this.containerId === other.containerId &&
        (this.start.equals(other.start) && this.end.equals(other.end))
      )
    }

    toString () {
      
      return [
        'ContainerSelection(',
        this.containerId, ', ',
        JSON.stringify(this.start.path), ', ', this.start.offset,
        ' -> ',
        JSON.stringify(this.end.path), ', ', this.end.offset,
        (this.reverse ? ', reverse' : ''),
        (this.surfaceId ? (', ' + this.surfaceId) : ''),
        ')'
      ].join('')
    }

    
    getContainer () {
      if (!this._internal.container) {
        this._internal.container = this.getDocument().get(this.containerId);
      }
      return this._internal.container
    }

    isInsideOf (other, strict) {
      
      
      if (other.isNull()) return false
      strict = Boolean(strict);
      let r1 = this._range(this);
      let r2 = this._range(other);
      return (r2.start.isBefore(r1.start, strict) &&
        r1.end.isBefore(r2.end, strict))
    }

    contains (other, strict) {
      
      
      if (other.isNull()) return false
      strict = Boolean(strict);
      let r1 = this._range(this);
      let r2 = this._range(other);
      return (r1.start.isBefore(r2.start, strict) &&
        r2.end.isBefore(r1.end, strict))
    }

    containsNode (nodeId, strict) {
      const container = this.getContainer();
      if (!container.contains(nodeId)) return false
      const coor = new Coordinate([nodeId], 0);
      const address = container.getAddress(coor);
      const r = this._range(this);
      
      let contained = r.start.isBefore(address, strict);
      if (contained) {
        address.offset = 1;
        contained = r.end.isAfter(address, strict);
      }
      return contained
    }

    overlaps (other) {
      let r1 = this._range(this);
      let r2 = this._range(other);
      
      return !(r1.end.isBefore(r2.start, false) ||
        r2.end.isBefore(r1.start, false))
    }

    isLeftAlignedWith (other) {
      let r1 = this._range(this);
      let r2 = this._range(other);
      return r1.start.isEqual(r2.start)
    }

    isRightAlignedWith (other) {
      let r1 = this._range(this);
      let r2 = this._range(other);
      return r1.end.isEqual(r2.end)
    }

    
    collapse (direction) {
      let coor;
      if (direction === 'left') {
        coor = this.start;
      } else {
        coor = this.end;
      }
      return _createNewSelection(this, coor, coor)
    }

    expand (other) {
      let r1 = this._range(this);
      let r2 = this._range(other);
      let start;
      let end;

      if (r1.start.isEqual(r2.start)) {
        start = new Coordinate(this.start.path, Math.min(this.start.offset, other.start.offset));
      } else if (r1.start.isAfter(r2.start)) {
        start = new Coordinate(other.start.path, other.start.offset);
      } else {
        start = this.start;
      }
      if (r1.end.isEqual(r2.end)) {
        end = new Coordinate(this.end.path, Math.max(this.end.offset, other.end.offset));
      } else if (r1.end.isBefore(r2.end, false)) {
        end = new Coordinate(other.end.path, other.end.offset);
      } else {
        end = this.end;
      }

      return _createNewSelection(this, start, end)
    }

    truncateWith (other) {
      if (other.isInsideOf(this, 'strict')) {
        
        throw new Error('Can not truncate with a contained selections')
      }
      if (!this.overlaps(other)) {
        return this
      }
      let r1 = this._range(this);
      let r2 = this._range(other);
      let start, end;
      if (r2.start.isBefore(r1.start, 'strict') && r2.end.isBefore(r1.end, 'strict')) {
        start = other.end;
        end = this.end;
      } else if (r1.start.isBefore(r2.start, 'strict') && r1.end.isBefore(r2.end, 'strict')) {
        start = this.start;
        end = other.start;
      } else if (r1.start.isEqual(r2.start)) {
        if (r2.end.isBefore(r1.end, 'strict')) {
          start = other.end;
          end = this.end;
        } else {
          
          return Selection.nullSelection
        }
      } else if (r1.end.isEqual(r2.end)) {
        if (r1.start.isBefore(r2.start, 'strict')) {
          start = this.start;
          end = other.start;
        } else {
          
          return Selection.nullSelection
        }
      } else if (this.isInsideOf(other)) {
        return Selection.nullSelection
      } else {
        throw new Error('Could not determine coordinates for truncate. Check input')
      }
      return _createNewSelection(this, start, end)
    }

    
    getNodeIds () {
      const container = this.getContainer();
      const startPos = container.getPosition(this.start.path[0]);
      const endPos = container.getPosition(this.end.path[0]);
      return container.getContent().slice(startPos, endPos + 1)
    }

    
    splitIntoPropertySelections () {
      let sels = [];
      let fragments = this.getFragments();
      fragments.forEach(function (fragment) {
        if (fragment instanceof Selection.Fragment) {
          sels.push(
            new PropertySelection(fragment.path, fragment.startOffset,
              fragment.endOffset, false, this.containerId, this.surfaceId)
          );
        }
      }.bind(this));
      return sels
    }

    _clone () {
      return new ContainerSelection(this)
    }

    _range (sel) {
      
      
      
      if (sel._internal.addressRange) {
        return sel._internal.addressRange
      }

      let container = this.getContainer();
      let startAddress = container.getAddress(sel.start);
      let endAddress;
      if (sel.isCollapsed()) {
        endAddress = startAddress;
      } else {
        endAddress = container.getAddress(sel.end);
      }
      let addressRange = {
        start: startAddress,
        end: endAddress
      };
      if (sel._isContainerSelection) {
        sel._internal.addressRange = addressRange;
      }
      return addressRange
    }

    get path () {
      throw new Error('ContainerSelection has no path property. Use startPath and endPath instead')
    }
  }

  ContainerSelection.prototype._isContainerSelection = true;

  ContainerSelection.fromJSON = function (properties) {
    let sel = new ContainerSelection(properties);
    return sel
  };

  function _createNewSelection (containerSel, start, end) {
    let newSel;

    if (start === end) {
      newSel = new PropertySelection({
        path: start.path,
        startOffset: start.offset,
        endOffset: start.offset,
        containerId: containerSel.containerId,
        surfaceId: containerSel.surfaceId
      });
    } else {
      newSel = new ContainerSelection(
        containerSel.containerId,
        start.path, start.offset, end.path, end.offset,
        false, containerSel.surfaceId
      );
    }
    
    const doc = containerSel._internal.doc;
    if (doc) {
      newSel.attach(doc);
    }
    return newSel
  }

  class NodeSelection extends Selection {
    constructor (containerId, nodeId, mode, reverse, surfaceId) {
      super();

      if (arguments.length === 1) {
        let data = arguments[0];
        containerId = data.containerId;
        nodeId = data.nodeId;
        mode = data.mode;
        reverse = data.reverse;
        surfaceId = data.surfaceId;
      }

      if (!isString(containerId)) {
        throw new Error("'containerId' is mandatory.")
      }
      if (!isString(nodeId)) {
        throw new Error("'nodeId' is mandatory.")
      }
      mode = mode || 'full';

      this.containerId = containerId;
      this.nodeId = nodeId;
      this.mode = mode;
      this.reverse = Boolean(reverse);
      this.surfaceId = surfaceId;

      this.start = new Coordinate([nodeId], 0);
      this.end = new Coordinate([nodeId], 1);
    }

    equals (other) {
      return (
        super.equals(other) &&
        this.nodeId === other.nodeId &&
        this.mode === other.mode
      )
    }

    isNodeSelection () {
      return true
    }

    getType () {
      return 'node'
    }

    getNodeId () {
      return this.nodeId
    }

    isFull () {
      return this.mode === 'full'
    }

    isBefore () {
      return this.mode === 'before'
    }

    isAfter () {
      return this.mode === 'after'
    }

    isCollapsed () {
      return this.mode !== 'full'
    }

    toJSON () {
      return {
        type: 'node',
        nodeId: this.nodeId,
        mode: this.mode,
        reverse: this.reverse,
        containerId: this.containerId,
        surfaceId: this.surfaceId
      }
    }

    toString () {
      
      return [
        'NodeSelection(',
        this.containerId, '.', this.nodeId, ', ',
        this.mode, ', ',
        (this.reverse ? ', reverse' : ''),
        (this.surfaceId ? (', ' + this.surfaceId) : ''),
        ')'
      ].join('')
    }

    collapse (direction) {
      if (direction === 'left') {
        if (this.isBefore()) {
          return this
        } else {
          return new NodeSelection(this.containerId, this.nodeId, 'before', this.reverse, this.surfaceId)
        }
      } else if (direction === 'right') {
        if (this.isAfter()) {
          return this
        } else {
          return new NodeSelection(this.containerId, this.nodeId, 'after', this.reverse, this.surfaceId)
        }
      } else {
        throw new Error("'direction' must be either 'left' or 'right'")
      }
    }

    _getCoordinate () {
      if (this.mode === 'before') {
        return new Coordinate([this.nodeId], 0)
      } else if (this.mode === 'after') {
        return new Coordinate([this.nodeId], 1)
      }
    }

    _clone () {
      return new NodeSelection(this)
    }
  }

  NodeSelection.prototype._isNodeSelection = true;

  NodeSelection.fromJSON = function (json) {
    return new NodeSelection(json)
  };


  NodeSelection._createFromCoordinate = function (coor) {
    var containerId = coor.containerId;
    var nodeId = coor.getNodeId();
    var mode = coor.offset === 0 ? 'before' : 'after';
    return new NodeSelection(containerId, nodeId, mode, false)
  };

  class CustomSelection extends Selection {
    constructor (customType, data, surfaceId) {
      super();

      if (arguments.length === 1) {
        let _data = arguments[0];
        customType = _data.customType;
        data = _data.data;
        surfaceId = _data.surfaceId;
      }

      this.customType = customType;
      this.data = data || {};
      this.surfaceId = surfaceId;
    }

    isCustomSelection () {
      return true
    }

    getType () {
      return 'custom'
    }

    getCustomType () {
      return this.customType
    }

    toJSON () {
      return {
        type: 'custom',
        customType: this.customType,
        data: cloneDeep(this.data),
        surfaceId: this.surfaceId
      }
    }

    toString () {
      
      return [
        'CustomSelection(',
        this.customType, ', ',
        JSON.stringify(this.data),
        ')'
      ].join('')
    }

    equals (other) {
      return (
        Selection.prototype.equals.call(this, other) &&
        other.isCustomSelection() &&
        isEqual(this.data, other.data)
      )
    }

    _clone () {
      return new CustomSelection(this)
    }
  }

  CustomSelection.prototype._isCustomSelection = true;

  CustomSelection.fromJSON = function (json) {
    return new CustomSelection(json)
  };

  function fromJSON (json) {
    if (!json) return Selection.nullSelection
    var type = json.type;
    switch (type) {
      case 'property':
        return PropertySelection.fromJSON(json)
      case 'container':
        return ContainerSelection.fromJSON(json)
      case 'node':
        return NodeSelection.fromJSON(json)
      case 'custom':
        return CustomSelection.fromJSON(json)
      default:
        
        return Selection.nullSelection
    }
  }


  function isFirst (doc, coor) {
    if (coor.isNodeCoordinate() && coor.offset === 0) return true
    let node = doc.get(coor.path[0]).getContainerRoot();
    if (node.isText() && coor.offset === 0) return true
    if (node.isList()) {
      let itemId = coor.path[0];
      if (node.items[0] === itemId && coor.offset === 0) return true
    }
  }


  function isLast (doc, coor) {
    if (coor.isNodeCoordinate() && coor.offset > 0) return true
    let node = doc.get(coor.path[0]).getContainerRoot();
    if (node.isText() && coor.offset >= node.getLength()) return true
    if (node.isList()) {
      let itemId = coor.path[0];
      let item = doc.get(itemId);
      if (last$1(node.items) === itemId && coor.offset === item.getLength()) return true
    }
  }

  function isEntirelySelected (doc, node, start, end) {
    let { isEntirelySelected } = getRangeInfo(doc, node, start, end);
    return isEntirelySelected
  }

  function getRangeInfo (doc, node, start, end) {
    let isFirst = true;
    let isLast = true;
    if (node.isText()) {
      if (start && start.offset !== 0) isFirst = false;
      if (end && end.offset < node.getLength()) isLast = false;
    } else if (node.isList()) {
      if (start) {
        let itemId = start.path[0];
        let itemPos = node.getItemPosition(itemId);
        if (itemPos > 0 || start.offset !== 0) isFirst = false;
      }
      if (end) {
        let itemId = end.path[0];
        let itemPos = node.getItemPosition(itemId);
        let item = doc.get(itemId);
        if (itemPos < node.items.length - 1 || end.offset < item.getLength()) isLast = false;
      }
    }
    let isEntirelySelected = isFirst && isLast;
    return {isFirst, isLast, isEntirelySelected}
  }

  function setCursor (tx, node, containerId, mode) {
    if (node.isText()) {
      let offset = 0;
      if (mode === 'after') {
        let text = node.getText();
        offset = text.length;
      }
      tx.setSelection({
        type: 'property',
        path: node.getPath(),
        startOffset: offset,
        containerId: containerId
      });
    } else if (node.isList()) {
      let item, offset;
      if (mode === 'after') {
        item = node.getLastItem();
        offset = item.getLength();
      } else {
        item = node.getFirstItem();
        offset = 0;
      }
      tx.setSelection({
        type: 'property',
        path: item.getPath(),
        startOffset: offset,
        containerId: containerId
      });
    } else {
      tx.setSelection({
        type: 'node',
        containerId: containerId,
        nodeId: node.id
        
        
        
      });
    }
  }

  function selectNode (tx, nodeId, containerId) {
    tx.setSelection(createNodeSelection({ doc: tx, nodeId, containerId }));
  }

  function createNodeSelection ({ doc, nodeId, containerId, mode, reverse, surfaceId }) {
    let node = doc.get(nodeId);
    if (!node) return Selection.nullSelection
    node = node.getContainerRoot();
    if (node.isText()) {
      return new PropertySelection({
        path: node.getPath(),
        startOffset: mode === 'after' ? node.getLength() : 0,
        endOffset: mode === 'before' ? 0 : node.getLength(),
        reverse: reverse,
        containerId: containerId,
        surfaceId: surfaceId
      })
    } else if (node.isList() && node.getLength() > 0) {
      let first = node.getFirstItem();
      let last = node.getLastItem();
      let start = {
        path: first.getPath(),
        offset: 0
      };
      let end = {
        path: last.getPath(),
        offset: last.getLength()
      };
      if (mode === 'after') start = end;
      else if (mode === 'before') end = start;
      return new ContainerSelection({
        startPath: start.path,
        startOffset: start.offset,
        endPath: end.path,
        endOffset: end.offset,
        reverse: reverse,
        containerId: containerId,
        surfaceId: surfaceId
      })
    } else {
      return new NodeSelection({ nodeId, mode, reverse, containerId, surfaceId })
    }
  }

  function stepIntoIsolatedNode (editorSession, comp) {
    
    
    if (comp.grabFocus()) return true

    
    let surface = comp.find('.sc-surface');
    if (surface) {
      
      if (surface._isTextPropertyEditor) {
        const doc = editorSession.getDocument();
        const path = surface.getPath();
        const text = doc.get(path, 'strict');
        editorSession.setSelection({
          type: 'property',
          path: path,
          startOffset: text.length,
          surfaceId: surface.id
        });
        return true
      } else if (surface._isContainerEditor) {
        let container = surface.getContainer();
        if (container.length > 0) {
          let first = container.getChildAt(0);
          setCursor(editorSession, first, container.id, 'after');
        }
        return true
      }
    }
    return false
  }

  function augmentSelection (selData, oldSel) {
    
    if (selData && oldSel && !selData.surfaceId && !oldSel.isNull()) {
      selData.containerId = selData.containerId || oldSel.containerId;
      selData.surfaceId = selData.surfaceId || oldSel.surfaceId;
    }
    return selData
  }

  var selectionHelpers = /*#__PURE__*/Object.freeze({
    fromJSON: fromJSON,
    isFirst: isFirst,
    isLast: isLast,
    isEntirelySelected: isEntirelySelected,
    getRangeInfo: getRangeInfo,
    setCursor: setCursor,
    selectNode: selectNode,
    createNodeSelection: createNodeSelection,
    stepIntoIsolatedNode: stepIntoIsolatedNode,
    augmentSelection: augmentSelection
  });

  class DocumentChange {
    constructor (ops, before, after) {
      if (arguments.length === 1 && isPlainObject(arguments[0])) {
        let data = arguments[0];
        
        this.sha = data.sha;
        
        this.timestamp = data.timestamp;
        
        this.before = data.before || {};
        
        this.ops = data.ops;
        this.info = data.info; 
        
        this.after = data.after || {};
      } else if (arguments.length === 3) {
        this.sha = uuid();
        this.info = {};
        this.timestamp = Date.now();
        this.ops = ops.slice(0);
        this.before = before || {};
        this.after = after || {};
      } else {
        throw new Error('Illegal arguments.')
      }
      
      this.updated = null;
      
      this.created = null;
      
      this.deleted = null;
    }

    
    _extractInformation (doc) {
      
      
      if (this._extracted) return

      let ops = this.ops;
      let created = {};
      let deleted = {};
      let updated = {};
      let affectedContainerAnnos = [];

      
      function _checkAnnotation (op) {
        switch (op.type) {
          case 'create':
          case 'delete': {
            let node = op.val;
            if (node.hasOwnProperty('start')) {
              updated[node.start.path] = true;
            }
            if (node.hasOwnProperty('end')) {
              updated[node.end.path] = true;
            }
            break
          }
          case 'update':
          case 'set': {
            
            let node = doc.get(op.path[0]);
            if (node) {
              if (node.isPropertyAnnotation()) {
                updated[node.start.path] = true;
              } else if (node.isContainerAnnotation()) {
                affectedContainerAnnos.push(node);
              }
            }
            break
          }
          default:
            
            throw new Error('Illegal state')
        }
      }

      for (let i = 0; i < ops.length; i++) {
        let op = ops[i];
        if (op.type === 'create') {
          created[op.val.id] = op.val;
          delete deleted[op.val.id];
        }
        if (op.type === 'delete') {
          delete created[op.val.id];
          deleted[op.val.id] = op.val;
        }
        if (op.type === 'set' || op.type === 'update') {
          updated[op.path] = true;
          
          updated[op.path[0]] = true;
        }
        _checkAnnotation(op);
      }

      affectedContainerAnnos.forEach(function (anno) {
        let container = doc.get(anno.containerId, 'strict');
        let startPos = container.getPosition(anno.start.path[0]);
        let endPos = container.getPosition(anno.end.path[0]);
        for (let pos = startPos; pos <= endPos; pos++) {
          let node = container.getChildAt(pos);
          let path;
          if (node.isText()) {
            path = node.getPath();
          } else {
            path = [node.id];
          }
          if (!deleted[node.id]) {
            updated[path] = true;
          }
        }
      });

      
      if (Object.keys(deleted).length > 0) {
        forEach(updated, function (_, key) {
          let nodeId = key.split(',')[0];
          if (deleted[nodeId]) {
            delete updated[key];
          }
        });
      }

      this.created = created;
      this.deleted = deleted;
      this.updated = updated;

      this._extracted = true;
    }

    invert () {
      
      let copy = this.toJSON();
      copy.ops = [];
      
      let tmp = copy.before;
      copy.before = copy.after;
      copy.after = tmp;
      let inverted = DocumentChange.fromJSON(copy);
      let ops = [];
      for (let i = this.ops.length - 1; i >= 0; i--) {
        ops.push(this.ops[i].invert());
      }
      inverted.ops = ops;
      return inverted
    }

    
    isAffected (path) {
      console.error('DEPRECATED: use change.hasUpdated() instead');
      return this.hasUpdated(path)
    }

    isUpdated (path) {
      console.error('DEPRECATED: use change.hasUpdated() instead');
      return this.hasUpdated(path)
    }
    

    hasUpdated (path) {
      return this.updated[path]
    }

    hasDeleted (id) {
      return this.deleted[id]
    }

    serialize () {
      
      

      let opSerializer = new OperationSerializer();
      let data = this.toJSON();
      data.ops = this.ops.map(function (op) {
        return opSerializer.serialize(op)
      });
      return JSON.stringify(data)
    }

    clone () {
      return DocumentChange.fromJSON(this.toJSON())
    }

    toJSON () {
      let data = {
        
        sha: this.sha,
        
        before: clone(this.before),
        ops: map(this.ops, function (op) {
          return op.toJSON()
        }),
        info: this.info,
        
        after: clone(this.after)
      };

      
      
      data.after.selection = undefined;
      data.before.selection = undefined;

      let sel = this.before.selection;
      if (sel && sel._isSelection) {
        data.before.selection = sel.toJSON();
      }
      sel = this.after.selection;
      if (sel && sel._isSelection) {
        data.after.selection = sel.toJSON();
      }
      return data
    }
  }

  DocumentChange.deserialize = function (str) {
    let opSerializer = new OperationSerializer();
    let data = JSON.parse(str);
    data.ops = data.ops.map(function (opData) {
      return opSerializer.deserialize(opData)
    });
    if (data.before.selection) {
      data.before.selection = fromJSON(data.before.selection);
    }
    if (data.after.selection) {
      data.after.selection = fromJSON(data.after.selection);
    }
    return new DocumentChange(data)
  };

  DocumentChange.fromJSON = function (data) {
    
    let change = cloneDeep(data);
    change.ops = data.ops.map(function (opData) {
      return ObjectOperation.fromJSON(opData)
    });
    change.before.selection = fromJSON(data.before.selection);
    change.after.selection = fromJSON(data.after.selection);
    return new DocumentChange(change)
  };

  function transformDocumentChange (A, B) {
    _transformInplaceBatch(A, B);
  }

  function transformSelection (sel, a) {
    let newSel = sel.clone();
    let hasChanged = _transformSelectionInplace(newSel, a);
    if (hasChanged) {
      return newSel
    } else {
      return sel
    }
  }

  function _transformInplaceSingle (a, b) {
    for (let i = 0; i < a.ops.length; i++) {
      let opA = a.ops[i];
      for (let j = 0; j < b.ops.length; j++) {
        let opB = b.ops[j];
        
        
        ObjectOperation.transform(opA, opB, {inplace: true});
      }
    }
    if (a.before) {
      _transformSelectionInplace(a.before.selection, b);
    }
    if (a.after) {
      _transformSelectionInplace(a.after.selection, b);
    }
    if (b.before) {
      _transformSelectionInplace(b.before.selection, a);
    }
    if (b.after) {
      _transformSelectionInplace(b.after.selection, a);
    }
  }

  function _transformInplaceBatch (A, B) {
    if (!isArray(A)) {
      A = [A];
    }
    if (!isArray(B)) {
      B = [B];
    }
    for (let i = 0; i < A.length; i++) {
      let a = A[i];
      for (let j = 0; j < B.length; j++) {
        let b = B[j];
        _transformInplaceSingle(a, b);
      }
    }
  }

  function _transformSelectionInplace (sel, a) {
    if (!sel || (!sel.isPropertySelection() && !sel.isContainerSelection())) {
      return false
    }
    let ops = a.ops;
    let hasChanged = false;
    let isCollapsed = sel.isCollapsed();
    for (let i = 0; i < ops.length; i++) {
      let op = ops[i];
      hasChanged |= _transformCoordinateInplace(sel.start, op);
      if (!isCollapsed) {
        hasChanged |= _transformCoordinateInplace(sel.end, op);
      } else {
        if (sel.isContainerSelection()) {
          sel.end.path = sel.start.path;
        }
        sel.end.offset = sel.start.offset;
      }
    }
    return hasChanged
  }

  function _transformCoordinateInplace (coor, op) {
    if (!isEqual(op.path, coor.path)) return false
    let hasChanged = false;
    if (op.type === 'update' && op.propertyType === 'string') {
      let diff = op.diff;
      let newOffset;
      if (diff.isInsert() && diff.pos <= coor.offset) {
        newOffset = coor.offset + diff.str.length;
        
        coor.offset = newOffset;
        hasChanged = true;
      } else if (diff.isDelete() && diff.pos <= coor.offset) {
        newOffset = Math.max(diff.pos, coor.offset - diff.str.length);
        
        coor.offset = newOffset;
        hasChanged = true;
      }
    }
    return hasChanged
  }

  var operationHelpers = /*#__PURE__*/Object.freeze({
    transformDocumentChange: transformDocumentChange,
    transformSelection: transformSelection
  });

  function isFunction$1 (f) {
    return typeof f === 'function'
  }

  function deleteFromArray (array, value) {
    if (!array) return
    for (var i = 0; i < array.length; i++) {
      if (array[i] === value) {
        array.splice(i, 1);
        i--;
      }
    }
  }

  class TreeNode {}



  class TreeIndex {
    
    get (path) {
      if (arguments.length > 1) {
        path = Array.prototype.slice(arguments, 0);
      }
      if (isString(path)) {
        path = [path];
      }
      return get(this, path)
    }

    getAll (path) {
      if (arguments.length > 1) {
        path = Array.prototype.slice(arguments, 0);
      }
      if (isString(path)) {
        path = [path];
      }
      if (!isArray(path)) {
        throw new Error('Illegal argument for TreeIndex.get()')
      }
      var node = get(this, path);
      return this._collectValues(node)
    }

    set (path, value) {
      if (isString(path)) {
        path = [path];
      }
      setWith(this, path, value, function (val) {
        if (!val) return new TreeNode()
      });
    }

    delete (path) {
      if (isString(path)) {
        delete this[path];
      } else if (path.length === 1) {
        delete this[path[0]];
      } else {
        var key = path[path.length - 1];
        path = path.slice(0, -1);
        var parent = get(this, path);
        if (parent) {
          delete parent[key];
        }
      }
    }

    clear () {
      var root = this;
      for (var key in root) {
        if (root.hasOwnProperty(key)) {
          delete root[key];
        }
      }
    }

    traverse (fn) {
      this._traverse(this, [], fn);
    }

    forEach (...args) {
      this.traverse(...args);
    }

    _traverse (root, path, fn) {
      var id;
      for (id in root) {
        if (!root.hasOwnProperty(id)) continue
        var child = root[id];
        var childPath = path.concat([id]);
        if (child instanceof TreeNode) {
          this._traverse(child, childPath, fn);
        } else {
          fn(child, childPath);
        }
      }
    }

    _collectValues (root) {
      
      
      var vals = {};
      this._traverse(root, [], function (val, path) {
        var key = path[path.length - 1];
        vals[key] = val;
      });
      return vals
    }
  }

  class TreeIndexArrays extends TreeIndex {
    contains (path) {
      let val = super.get(path);
      return Boolean(val)
    }

    get (path) {
      let val = super.get(path);
      if (val instanceof TreeNode) {
        val = val.__values__ || [];
      }
      return val
    }

    set (path, arr) {
      let val = super.get(path);
      val.__values__ = arr;
    }

    add (path, value) {
      if (isString(path)) {
        path = [path];
      }
      if (!isArray(path)) {
        throw new Error('Illegal arguments.')
      }
      var arr;

      
      
      
      
      
      
      setWith(this, path.concat(['__values__', '__dummy__']), undefined, function (val, key) {
        if (key === '__values__') {
          if (!val) val = [];
          arr = val;
        } else if (!val) {
          val = new TreeNode();
        }
        return val
      });
      delete arr.__dummy__;
      arr.push(value);
    }

    remove (path, value) {
      var arr = get(this, path);
      if (arr instanceof TreeNode) {
        if (arguments.length === 1) {
          delete arr.__values__;
        } else {
          deleteFromArray(arr.__values__, value);
        }
      }
    }

    _collectValues (root) {
      var vals = [];
      this._traverse(root, [], function (val) {
        vals.push(val);
      });
      vals = Array.prototype.concat.apply([], vals);
      return vals
    }
  }

  TreeIndex.Arrays = TreeIndexArrays;

  function filter (iteratee, fn) {
    if (!iteratee) return []
    if (iteratee.constructor.prototype.filter && isFunction$1(iteratee.constructor.prototype.filter)) {
      return iteratee.filter(fn)
    }
    let result = [];
    forEach(iteratee, (val, key) => {
      if (fn(val, key)) {
        result.push(val);
      }
    });
    return result
  }

  function flatten (arr) {
    return Array.prototype.concat.apply([], arr)
  }

  function flattenOften (arr, max) {
    if (!(max > 0)) throw new Error("'max' must be a positive number")
    let l = arr.length;
    arr = flatten(arr);
    let round = 1;
    while (round < max && l < arr.length) {
      l = arr.length;
      arr = flatten(arr);
      round++;
    }
    return arr
  }

  class NodeIndex {
    
    select (node) { 
      throw new Error('This method is abstract.')
    }

    clear () {
      throw new Error('This method is abstract')
    }

    
    create (node) { 
      throw new Error('This method is abstract.')
    }

    
    delete (node) { 
      throw new Error('This method is abstract.')
    }

    set (node, path, newValue, oldValue) {
      this.update(node, path, newValue, oldValue);
    }

    
    update (node, path, newValue, oldValue) { 
      throw new Error('This method is abstract.')
    }

    
    reset (data) {
      this.clear();
      this._initialize(data);
    }

    
    clone () {
      var NodeIndexClass = this.constructor;
      var clone = new NodeIndexClass();
      return clone
    }

    _initialize (data) {
      forEach(data.getNodes(), function (node) {
        if (this.select(node)) {
          this.create(node);
        }
      }.bind(this));
    }
  }


  NodeIndex.create = function (prototype) {
    var index = Object.assign(new NodeIndex(), prototype);
    index.clone = function () {
      return NodeIndex.create(prototype)
    };
    return index
  };


  NodeIndex.filterByType = function (type) {
    return function (node) {
      return node.isInstanceOf(type)
    }
  };

  class DocumentIndex extends NodeIndex {}

  var annotationHelpers = {
    insertedText,
    deletedText,
    transferAnnotations,
    expandAnnotation,
    fuseAnnotation,
    truncateAnnotation
  };

  function insertedText (doc, coordinate, length) {
    if (!length) return
    var index = doc.getIndex('annotations');
    var annotations = index.get(coordinate.path);
    for (let i = 0; i < annotations.length; i++) {
      let anno = annotations[i];
      var pos = coordinate.offset;
      var start = anno.start.offset;
      var end = anno.end.offset;
      var newStart = start;
      var newEnd = end;
      if ((pos < start) ||
           (pos === start)) {
        newStart += length;
      }
      
      if ((pos < end) ||
           (pos === end && !anno.isInline())) {
        newEnd += length;
      }
      
      if (newStart !== start) {
        doc.set([anno.id, 'start', 'offset'], newStart);
      }
      if (newEnd !== end) {
        doc.set([anno.id, 'end', 'offset'], newEnd);
      }
    }

    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
  }

  function deletedText (doc, path, startOffset, endOffset) {
    if (startOffset === endOffset) return
    var index = doc.getIndex('annotations');
    var annotations = index.get(path);
    var length = endOffset - startOffset;
    for (let i = 0; i < annotations.length; i++) {
      let anno = annotations[i];
      var pos1 = startOffset;
      var pos2 = endOffset;
      var start = anno.start.offset;
      var end = anno.end.offset;
      var newStart = start;
      var newEnd = end;
      if (pos2 <= start) {
        newStart -= length;
        newEnd -= length;
        doc.set([anno.id, 'start', 'offset'], newStart);
        doc.set([anno.id, 'end', 'offset'], newEnd);
      } else {
        if (pos1 <= start) {
          newStart = start - Math.min(pos2 - pos1, start - pos1);
        }
        if (pos1 <= end) {
          newEnd = end - Math.min(pos2 - pos1, end - pos1);
        }
        
        if (start !== end && newStart === newEnd) {
          doc.delete(anno.id);
        } else {
          
          if (start !== newStart) {
            doc.set([anno.id, 'start', 'offset'], newStart);
          }
          if (end !== newEnd) {
            doc.set([anno.id, 'end', 'offset'], newEnd);
          }
        }
      }
    }
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
  }


  function transferAnnotations (doc, path, offset, newPath, newOffset) {
    var index = doc.getIndex('annotations');
    var annotations = index.get(path, offset);
    for (let i = 0; i < annotations.length; i++) {
      let a = annotations[i];
      var isInside = (offset > a.start.offset && offset < a.end.offset);
      var start = a.start.offset;
      var end = a.end.offset;
      
      if (isInside) {
        
        if (a.canSplit()) {
          let newAnno = a.toJSON();
          newAnno.id = uuid(a.type + '_');
          newAnno.start.path = newPath;
          newAnno.start.offset = newOffset;
          newAnno.end.path = newPath;
          newAnno.end.offset = newOffset + a.end.offset - offset;
          doc.create(newAnno);
        }
        
        let newStartOffset = a.start.offset;
        let newEndOffset = offset;
        
        if (newEndOffset === newStartOffset) {
          doc.delete(a.id);
        
        } else {
          
          if (newStartOffset !== start) {
            doc.set([a.id, 'start', 'offset'], newStartOffset);
          }
          if (newEndOffset !== end) {
            doc.set([a.id, 'end', 'offset'], newEndOffset);
          }
        }
      
      } else if (a.start.offset >= offset) {
        
        
        
        doc.set([a.id, 'start', 'path'], newPath);
        doc.set([a.id, 'start', 'offset'], newOffset + a.start.offset - offset);
        doc.set([a.id, 'end', 'path'], newPath);
        doc.set([a.id, 'end', 'offset'], newOffset + a.end.offset - offset);
      }
    }

    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
  }


  function truncateAnnotation (tx, anno, sel) {
    if (!sel || !sel._isSelection) throw new Error('Argument "selection" is required.')
    if (!anno || !anno.isAnnotation()) throw new Error('Argument "anno" is required and must be an annotation.')
    let annoSel = anno.getSelection();
    let newAnnoSel = annoSel.truncateWith(sel);
    anno._updateRange(tx, newAnnoSel);
    return anno
  }


  function expandAnnotation (tx, anno, sel) {
    if (!sel || !sel._isSelection) throw new Error('Argument "selection" is required.')
    if (!anno || !anno.isAnnotation()) throw new Error('Argument "anno" is required and must be an annotation.')
    let annoSel = anno.getSelection();
    let newAnnoSel = annoSel.expand(sel);
    anno._updateRange(tx, newAnnoSel);
    return anno
  }


  function fuseAnnotation (tx, annos) {
    if (!isArray(annos) || annos.length < 2) {
      throw new Error('fuseAnnotation(): at least two annotations are necessary.')
    }
    let sel, annoType;
    annos.forEach(function (anno, idx) {
      if (idx === 0) {
        sel = anno.getSelection();
        annoType = anno.type;
      } else {
        if (anno.type !== annoType) {
          throw new Error('fuseAnnotation(): all annotations must be of the same type.')
        }
        sel = sel.expand(anno.getSelection());
      }
    });
    
    for (var i = 1; i < annos.length; i++) {
      tx.delete(annos[i].id);
    }
    expandAnnotation(tx, annos[0], sel);
    tx.setSelection(sel);
  }

  function getPropertyAnnotationsForSelection (doc, sel, options) {
    options = options || {};
    if (!sel.isPropertySelection()) {
      return []
    }
    let path = sel.getPath();
    let annotations = doc.getIndex('annotations').get(path, sel.start.offset, sel.end.offset);
    if (options.type) {
      annotations = filter(annotations, DocumentIndex.filterByType(options.type));
    }
    return annotations
  }


  function getContainerAnnotationsForSelection (doc, sel, containerId, options) {
    
    
    
    
    
    if (!containerId) {
      throw new Error("'containerId' is required.")
    }
    options = options || {};
    let index = doc.getIndex('container-annotations');
    let annotations = [];
    if (index) {
      annotations = index.get(containerId, options.type);
      annotations = filter(annotations, function (anno) {
        return sel.overlaps(anno.getSelection())
      });
    }
    return annotations
  }


  function isContainerAnnotation (doc, type) {
    let schema = doc.getSchema();
    return schema.isInstanceOf(type, '@container-annotation')
  }


  function getTextForSelection (doc, sel) {
    if (!sel || sel.isNull()) {
      return ''
    } else if (sel.isPropertySelection()) {
      let text = doc.get(sel.start.path);
      return text.substring(sel.start.offset, sel.end.offset)
    } else if (sel.isContainerSelection()) {
      let result = [];
      let nodeIds = sel.getNodeIds();
      let L = nodeIds.length;
      for (let i = 0; i < L; i++) {
        let id = nodeIds[i];
        let node = doc.get(id);
        if (node.isText()) {
          let text = node.getText();
          if (i === L - 1) {
            text = text.slice(0, sel.end.offset);
          }
          if (i === 0) {
            text = text.slice(sel.start.offset);
          }
          result.push(text);
        }
      }
      return result.join('\n')
    }
  }

  function getMarkersForSelection (doc, sel) {
    
    if (!sel || !sel.isPropertySelection()) return []
    const path = sel.getPath();
    
    let markers = doc.getIndex('markers').get(path);
    const filtered = filter(markers, function (m) {
      return m.containsSelection(sel)
    });
    return filtered
  }


  function deleteNode (doc, node) {
    
    if (!node) {
      console.warn('Invalid arguments');
      return
    }
    
    if (node.isText()) {
      
      let annos = doc.getIndex('annotations').get(node.id);
      for (let i = 0; i < annos.length; i++) {
        doc.delete(annos[i].id);
      }
    }
    let nodeSchema = node.getSchema();
    
    
    
    doc.delete(node.id);
    
    let ownedProps = nodeSchema.getOwnedProperties();
    ownedProps.forEach(prop => {
      let value = node[prop.name];
      if (prop.isArray()) {
        let ids = value;
        if (ids.length > 0) {
          
          if (isArray(ids[0])) ids = flattenOften(ids, 2);
          ids.forEach((id) => {
            deleteNode(doc, doc.get(id));
          });
        }
      } else {
        deleteNode(doc, doc.get(value));
      }
    });
  }


  function copyNode (node) {
    let nodes = [];
    
    let doc = node.getDocument();
    let nodeSchema = node.getSchema();
    for (let prop of nodeSchema) {
      
      
      if ((prop.isReference() && prop.isOwned()) || (prop.type === 'file')) {
        let val = node[prop.name];
        nodes.push(_copyChildren(val));
      }
    }
    nodes.push(node.toJSON());
    let annotationIndex = node.getDocument().getIndex('annotations');
    let annotations = annotationIndex.get([node.id]);
    forEach(annotations, function (anno) {
      nodes.push(anno.toJSON());
    });
    let result = flatten(nodes).filter(Boolean);
    
    return result

    function _copyChildren (val) {
      if (!val) return null
      if (isArray(val)) {
        return flatten(val.map(_copyChildren))
      } else {
        let id = val;
        if (!id) return null
        let child = doc.get(id);
        if (!child) return
        return copyNode(child)
      }
    }
  }


  function deleteTextRange (doc, start, end) {
    if (!start) {
      start = {
        path: end.path,
        offset: 0
      };
    }
    let path = start.path;
    let text = doc.get(path);
    if (!end) {
      end = {
        path: start.path,
        offset: text.length
      };
    }
    
    if (!isArrayEqual(start.path, end.path)) {
      throw new Error('start and end must be on one property')
    }
    let startOffset = start.offset;
    if (startOffset < 0) throw new Error('start offset must be >= 0')
    let endOffset = end.offset;
    if (endOffset > text.length) throw new Error('end offset must be smaller than the text length')

    doc.update(path, { type: 'delete', start: startOffset, end: endOffset });
    
    let annos = doc.getAnnotations(path);
    annos.forEach(function (anno) {
      let annoStart = anno.start.offset;
      let annoEnd = anno.end.offset;
      
      if (annoEnd <= startOffset) ; else if (annoStart >= endOffset) {
        doc.update([anno.id, 'start'], { type: 'shift', value: startOffset - endOffset });
        doc.update([anno.id, 'end'], { type: 'shift', value: startOffset - endOffset });
      
      } else if (annoStart >= startOffset && annoEnd <= endOffset) {
        doc.delete(anno.id);
      
      } else if (annoStart >= startOffset && annoEnd >= endOffset) {
        if (annoStart > startOffset) {
          doc.update([anno.id, 'start'], { type: 'shift', value: startOffset - annoStart });
        }
        doc.update([anno.id, 'end'], { type: 'shift', value: startOffset - endOffset });
      
      } else if (annoStart <= startOffset && annoEnd <= endOffset) {
        doc.update([anno.id, 'end'], { type: 'shift', value: startOffset - annoEnd });
      
      } else if (annoStart < startOffset && annoEnd >= endOffset) {
        doc.update([anno.id, 'end'], { type: 'shift', value: startOffset - endOffset });
      } else {
        console.warn('TODO: handle annotation update case.');
      }
    });
  }

  function deleteListRange (doc, list, start, end) {
    if (doc !== list.getDocument()) {
      list = doc.get(list.id);
    }
    let startItem, endItem;
    if (!start) {
      startItem = list.getItemAt(0);
      start = {
        path: startItem.getPath(),
        offset: 0
      };
    } else {
      startItem = doc.get(start.path[0]);
    }
    if (!end) {
      endItem = list.getLastItem();
      end = {
        path: endItem.getPath(),
        offset: endItem.getLength()
      };
    } else {
      endItem = doc.get(end.path[0]);
    }
    let startPos = list.getItemPosition(startItem);
    let endPos = list.getItemPosition(endItem);
    
    if (startPos === endPos) {
      deleteTextRange(doc, start, end);
      return
    }
    
    if (startPos > endPos) {
      [start, end] = [end, start];
      [startPos, endPos] = [endPos, startPos];
      [startItem, endItem] = [endItem, startItem];
    }
    let firstEntirelySelected = isEntirelySelected(doc, startItem, start, null);
    let lastEntirelySelected = isEntirelySelected(doc, endItem, null, end);

    
    if (lastEntirelySelected) {
      list.removeItemAt(endPos);
      deleteNode(doc, endItem);
    } else {
      deleteTextRange(doc, null, end);
    }

    
    let items = list.getItems();
    for (let i = endPos - 1; i > startPos; i--) {
      let item = items[i];
      list.removeItemAt(i);
      deleteNode(doc, item);
    }

    
    if (firstEntirelySelected) {
      
      
      
      
      deleteTextRange(doc, start, null);
    } else {
      deleteTextRange(doc, start, null);
    }

    if (!firstEntirelySelected && !lastEntirelySelected) {
      mergeListItems(doc, list.id, startPos);
    }
  }

  function mergeListItems (doc, listId, itemPos) {
    
    let list = doc.get(listId);
    let targetItem = list.getItemAt(itemPos);
    let targetPath = targetItem.getPath();
    let targetLength = targetItem.getLength();
    let sourceItem = list.getItemAt(itemPos + 1);
    let sourcePath = sourceItem.getPath();
    
    list.removeItemAt(itemPos + 1);
    
    doc.update(targetPath, { type: 'insert', start: targetLength, text: sourceItem.getText() });
    
    annotationHelpers.transferAnnotations(doc, sourcePath, 0, targetPath, targetLength);
    deleteNode(doc, sourceItem);
  }

  function getNodes (doc, ids) {
    return ids.map((id) => {
      return doc.get(id, 'strict')
    })
  }



  const SNIPPET_ID = 'snippet';
  const TEXT_SNIPPET_ID = 'text-snippet';

  var documentHelpers = /*#__PURE__*/Object.freeze({
    getPropertyAnnotationsForSelection: getPropertyAnnotationsForSelection,
    getContainerAnnotationsForSelection: getContainerAnnotationsForSelection,
    isContainerAnnotation: isContainerAnnotation,
    getTextForSelection: getTextForSelection,
    getMarkersForSelection: getMarkersForSelection,
    deleteNode: deleteNode,
    copyNode: copyNode,
    deleteTextRange: deleteTextRange,
    deleteListRange: deleteListRange,
    mergeListItems: mergeListItems,
    getNodes: getNodes,
    SNIPPET_ID: SNIPPET_ID,
    TEXT_SNIPPET_ID: TEXT_SNIPPET_ID
  });

  const PLAINOBJ = {};


  class Registry {
    constructor (entries, validator) {
      this.entries = {};
      this.names = [];
      this.validator = validator;

      if (entries) {
        forEach(entries, function (entry, name) {
          this.add(name, entry);
        }.bind(this));
      }
    }

    
    contains (name) {
      return this.entries.hasOwnProperty(name)
    }

    
    add (name, entry) {
      if (this.validator) {
        this.validator(entry);
      }
      if (PLAINOBJ[name]) {
        throw new Error('Illegal key: "' + name + '" is a property of Object which is thus not allowed as a key.')
      }
      if (this.contains(name)) {
        this.remove(name);
      }
      this.entries[name] = entry;
      this.names.push(name);
    }

    
    remove (name) {
      let pos = this.names.indexOf(name);
      if (pos >= 0) {
        this.names.splice(pos, 1);
      }
      delete this.entries[name];
    }

    
    clear () {
      this.names = [];
      this.entries = {};
    }

    
    get (name, strict) {
      let result = this.entries[name];
      if (strict && !result) {
        throw new Error('No entry registered for name ' + name)
      }
      return result
    }

    
    forEach (callback) {
      for (let i = 0; i < this.names.length; i++) {
        let name = this.names[i];
        let _continue = callback(this.entries[name], name);
        if (_continue === false) {
          break
        }
      }
    }

    map (callback) {
      let result = [];
      this.forEach((entry, name) => {
        result.push(callback(entry, name));
      });
      return result
    }

    filter (callback) {
      let result = [];
      this.forEach(function (entry, name) {
        if (callback(entry, name)) {
          result.push(entry);
        }
      });
      return result
    }

    values () {
      return this.filter(() => { return true })
    }
  }

  Registry.prototype._isRegistry = true;

  function isBoolean (val) {
    return (val === true || val === false || (val && val.constructor === Boolean))
  }

  class NodeProperty {
    constructor (name, definition) {
      this.name = name;
      this.definition = definition;

      Object.freeze(this);
      Object.freeze(definition);
    }

    isArray () {
      return isArray(this.definition.type)
    }

    isReference () {
      if (this.isArray()) {
        return last$1(this.definition.type) === 'id'
      } else {
        return this.definition.type === 'id'
      }
    }

    isText () {
      return Boolean(this.definition._isText)
    }

    isOwned () {
      return Boolean(this.definition.owned)
    }

    isOptional () {
      return this.definition.optional || this.hasDefault()
    }

    isNotNull () {
      return Boolean(this.definition.notNull)
    }

    hasDefault () {
      return this.definition.hasOwnProperty('default')
    }

    getDefault () {
      return this.definition.default
    }

    createDefaultValue () {
      if (isArray(this.definition.type)) {
        return []
      }
      switch (this.definition.type) {
        case 'object':
          return {}
        case 'number':
          return -1
        case 'coordinate':
          return new Coordinate([], 0)
        case 'boolean':
          return false
        case 'id':
          return null
        case 'string':
          return ''
        default:
          return null
      }
    }

    get type () {
      return this.definition.type
    }

    get targetTypes () {
      return this.definition.targetTypes
    }
  }

  class NodeSchema {
    constructor (properties) {
      this.properties = properties;

      
      
      
      
      
      this._ownedPropNames = new Set();
      this._ownedProps = [];
      forEach(properties, (prop) => {
        if ((prop.isReference() && prop.isOwned()) || (prop.type === 'file')) {
          this._ownedPropNames.add(prop.name);
          this._ownedProps.push(prop);
        }
      });
    }

    getProperty (name) {
      return this.properties[name]
    }

    hasOwnedProperties () {
      return this._ownedPropNames.size > 0
    }

    getOwnedProperties () {
      return this._ownedProps.slice()
    }

    isOwned (name) {
      return this._ownedPropNames.has(name)
    }

    [Symbol.iterator] () {
      const properties = this.properties;
      let ids = Object.keys(properties);
      let idx = 0;
      return { 
        next () {
          let done = idx > ids.length - 1;
          if (done) {
            return { done }
          } else {
            return {
              value: properties[ids[idx++]]
            }
          }
        }
      }
    }
  }

  class Node extends EventEmitter {
    
    constructor () {
      super();

      
      
      this._initialize.apply(this, arguments);
    }

    _initialize (data) {
      const NodeClass = this.constructor;

      let schema = NodeClass.schema;
      for (let property of schema) {
        let name = property.name;
        
        
        const propIsGiven = (data[name] !== undefined);
        const isOptional = property.isOptional();
        const hasDefault = property.hasDefault();
        if (!isOptional && !propIsGiven) {
          throw new Error('Property ' + name + ' is mandatory for node type ' + this.type)
        }
        if (propIsGiven) {
          this[name] = _checked(property, data[name]);
        } else if (hasDefault) {
          this[name] = cloneDeep(_checked(property, property.getDefault()));
        }
      }
    }

    dispose () {
      this._disposed = true;
    }

    isDisposed () {
      return Boolean(this._disposed)
    }

    
    isInstanceOf (typeName) {
      return Node.isInstanceOf(this.constructor, typeName)
    }

    getSchema () {
      return this.constructor.schema
    }

    
    getTypeNames () {
      var typeNames = [];
      var NodeClass = this.constructor;
      while (NodeClass.type !== '@node') {
        typeNames.push(NodeClass.type);
        NodeClass = Object.getPrototypeOf(NodeClass);
      }
      return typeNames
    }

    
    getPropertyType (propertyName) {
      return this.constructor.schema.getProperty(propertyName).type
    }

    
    toJSON () {
      var data = {
        type: this.type
      };
      const schema = this.getSchema();
      for (let prop of schema) {
        let val = this[prop.name];
        if (prop.isOptional() && val === undefined) continue
        if (isArray(val) || isObject(val)) {
          val = cloneDeep(val);
        }
        data[prop.name] = val;
      }
      return data
    }

    get type () {
      return this.constructor.type
    }
  }

  Node.prototype._isNode = true;


  Object.defineProperty(Node, 'schema', {
    get () { return this._schema },
    set (schema) {
      let NodeClass = this;
      
      
      if (schema.type) {
        NodeClass.type = schema.type;
      }
      
      
      
      NodeClass._schema = compileSchema(NodeClass, schema);
    }
  });

  Node.define = Node.defineSchema = function define (schema) {
    this.schema = schema;
  };

  Node.schema = {
    type: '@node',
    id: 'string'
  };


  Node.isInstanceOf = function (NodeClass, typeName) {
    var type = NodeClass.type;
    while (type !== '@node') {
      if (type === typeName) return true
      var _super = Object.getPrototypeOf(NodeClass.prototype).constructor;
      if (_super && _super.type) {
        NodeClass = _super;
        type = NodeClass.type;
      } else {
        break
      }
    }
    return false
  };



  function compileSchema (NodeClass, schema) {
    let compiledSchema = _compileSchema(schema);
    let schemas = [compiledSchema];
    let clazz = NodeClass;
    while (clazz) {
      var parentProto = Object.getPrototypeOf(clazz.prototype);
      if (!parentProto) break
      clazz = parentProto.constructor;
      if (clazz && clazz._schema) {
        schemas.unshift(clazz._schema.properties);
      }
    }
    schemas.unshift({});
    return new NodeSchema(Object.assign.apply(null, schemas))
  }

  function _compileSchema (schema) {
    let compiledSchema = {};
    forEach(schema, function (definition, name) {
      
      if (name === 'type') {
        return
      }
      if (isString(definition) || isArray(definition)) {
        definition = { type: definition };
      }
      definition = _compileDefintion(definition);
      definition.name = name;
      compiledSchema[name] = new NodeProperty(name, definition);
    });
    return compiledSchema
  }

  const _valueTypes = new Set(['string', 'number', 'boolean', 'object', 'array']);

  function _compileDefintion (definition) {
    let result = definition;
    if (isArray(definition.type)) {
      
      
      
      
      let defs = definition.type;
      let lastIdx = defs.length - 1;
      let first = defs[0];
      let last = defs[lastIdx];
      let isCanonical = first === 'array';
      let hasTargetType = last !== 'id' && !_valueTypes.has(last);
      definition.targetTypes = definition.type;
      if (isCanonical) {
        definition.type = defs.slice();
        definition.type[lastIdx] = 'id';
        if (hasTargetType) definition.targetTypes = [last];
      } else {
        if (defs.length > 1) {
          defs.forEach(t => {
            if (_valueTypes.has(t)) {
              throw new Error('Multi-types must consist of node types.')
            }
          });
          definition.type = [ 'array', 'id' ];
          definition.targetTypes = defs;
        } else {
          if (_valueTypes.has(first)) {
            definition.type = [ 'array', first ];
          } else {
            definition.type = [ 'array', 'id' ];
            definition.targetTypes = defs;
          }
        }
      }
    } else if (definition.type === 'text') {
      result = {
        type: 'string',
        default: '',
        _isText: true
      };
    }
    return result
  }

  function _checked (prop, value) {
    let type;
    let name = prop.name;
    if (prop.isArray()) {
      type = 'array';
    } else {
      type = prop.type;
    }
    if (value === null) {
      if (prop.isNotNull()) {
        throw new Error('Value for property ' + name + ' is null.')
      } else {
        return value
      }
    }
    if (value === undefined) {
      throw new Error('Value for property ' + name + ' is undefined.')
    }
    if ((type === 'string' && !isString(value)) ||
        (type === 'boolean' && !isBoolean(value)) ||
        (type === 'number' && !isNumber(value)) ||
        (type === 'array' && !isArray(value)) ||
        (type === 'id' && !isString(value)) ||
        (type === 'object' && !isObject(value))) {
      throw new Error('Illegal value type for property ' + name + ': expected ' + type + ', was ' + (typeof value))
    }
    return value
  }

  class DocumentNode extends Node {
    _initialize (doc, props) {
      this.document = doc;
      super._initialize(props);
    }

    
    getDocument () {
      return this.document
    }

    
    hasParent () {
      return Boolean(this.parent)
    }

    
    getParent () {
      if (isString(this.parent)) return this.document.get(this.parent)
      return this.parent
    }

    setParent (parent) {
      if (isString(parent)) parent = this.document.get(parent);
      this.parent = parent;
    }

    
    getRoot () {
      let node = this;
      while (node.parent) {
        node = node.parent;
      }
      return node
    }

    getContainerRoot () {
      let node = this;
      while (node.parent) {
        
        if (node.parent.isContainer()) return node
        
        node = node.parent;
      }
      return node
    }

    
    hasChildren () {
      return false
    }

    
    getChildIndex(child) { 
      return -1
    }

    
    getChildAt(idx) { 
      return null
    }

    
    getChildCount () {
      return 0
    }

    
    

    

    
    isBlock () {
      return Boolean(this.constructor.isBlock)
    }

    
    isText () {
      return Boolean(this.constructor.isText)
    }

    isList () {
      return Boolean(this.constructor.isList)
    }

    isListItem () {
      return Boolean(this.constructor.isListItem)
    }

    isContainer () {
      return Boolean(this._isContainer)
    }

    

    isAnnotation () {
      return Boolean(this._isAnnotation)
    }

    isPropertyAnnotation () {
      return Boolean(this._isPropertyAnnotation)
    }

    isContainerAnnotation () {
      return Boolean(this._isContainerAnnotation)
    }

    
    isInline () {
      return Boolean(this.constructor.isInline)
    }
  }

  DocumentNode.prototype._isDocumentNode = true;


  DocumentNode.isBlock = false;


  DocumentNode.isText = false;


  DocumentNode.isPropertyAnnotation = false;


  DocumentNode.isContainerAnnotation = false;


  DocumentNode.isInline = false;

  class ContainerAddress {
    constructor (pos, offset) {
      this.pos = pos;
      this.offset = offset;
    }

    isBefore (other, strict) {
      strict = Boolean(strict);
      if (this.pos < other.pos) {
        return true
      } else if (this.pos > other.pos) {
        return false
      } else if (this.offset < other.offset) {
        return true
      } else if (this.offset > other.offset) {
        return false
      }
      if (strict) {
        return false
      } else {
        return true
      }
    }

    isAfter (other, strict) {
      return other.isBefore(this, strict)
    }

    isEqual (other) {
      return (this.pos === other.pos && this.offset === other.offset)
    }

    toString () {
      return [this.pos, '.', this.offset].join('')
    }
  }

  function ContainerMixin (DocumentNode) {
    class AbstractContainer extends DocumentNode {
      contains (nodeId) {
        return this.getPosition(nodeId) >= 0
      }

      getPosition (node, strict) {
        if (isString(node)) {
          node = this.document.get(node);
        }
        if (!node) return -1
        let pos = this._getPosition(node);
        if (strict && pos < 0) {
          throw new Error('Node is not within this container: ' + node.id)
        }
        return pos
      }

      getNodeAt (idx) {
        const nodeId = this.getNodeIdAt(idx);
        if (nodeId) {
          return this.getDocument().get(nodeId)
        }
      }

      getNodeIdAt (idx) {
        let content = this.getContent();
        if (idx < 0 || idx >= content.length) {
          
          return undefined
        } else {
          return content[idx]
        }
      }

      getNodes () {
        const doc = this.getDocument();
        return this.getContent().map(id => doc.get(id)).filter(Boolean)
      }

      show (nodeId, pos) {
        
        const arg1 = arguments[0];
        if (!isString(arg1)) {
          if (arg1._isNode) {
            nodeId = arg1.id;
          }
        }
        if (arguments.length > 1) {
          console.error('DEPRECATED: use container.showAt(pos, nodeId) instead');
        } else {
          pos = this.getLength();
        }
        return this.showAt(pos, nodeId)
      }

      showAt (pos, nodeId) {
        const doc = this.getDocument();
        const length = this.getLength();
        if (!isNumber(pos) || pos < 0 || pos > length) {
          throw new Error('Index out of bounds')
        }
        if (!isString(nodeId)) {
          if (nodeId._isNode) {
            nodeId = nodeId.id;
          } else {
            throw new Error('Invalid argument.')
          }
        }
        doc.update(this.getContentPath(), { type: 'insert', pos: pos, value: nodeId });
      }

      hide (nodeId) {
        const pos = this.getPosition(nodeId);
        this.hideAt(pos);
      }

      hideAt (pos) {
        const length = this.getLength();
        if (pos >= 0 && pos < length) {
          const doc = this.getDocument();
          doc.update(this.getContentPath(), { type: 'delete', pos: pos });
        } else {
          throw new Error('Index out of bounds.')
        }
      }

      getAddress (coor) {
        if (!coor._isCoordinate) {
          
          throw new Error('Illegal argument: Container.getAddress(coor) expects a Coordinate instance.')
        }
        var nodeId = coor.path[0];
        var nodePos = this.getPosition(nodeId);
        var offset;
        if (coor.isNodeCoordinate()) {
          if (coor.offset > 0) {
            offset = Number.MAX_VALUE;
          } else {
            offset = 0;
          }
        } else {
          offset = coor.offset;
        }
        return new ContainerAddress(nodePos, offset)
      }

      getLength () {
        return this.getContent().length
      }

      get length () {
        return this.getLength()
      }

      _getPosition (node) {
        if (this._isCaching) {
          return this._getCachedPosition(node)
        } else {
          return this._lookupPosition(node)
        }
      }

      _getCachedPosition (node) {
        let cache = this._cachedPositions || this._fillCache();
        let nodeId = node.id;
        let pos = -1;
        if (cache.hasOwnProperty(nodeId)) {
          pos = cache[nodeId];
        } else {
          pos = this._lookupPosition(node);
          cache[nodeId] = pos;
        }
        return pos
      }

      _fillCache () {
        let positions = {};
        this.nodes.forEach((id, pos) => {
          positions[id] = pos;
        });
        this._cachedPositions = positions;
        return positions
      }

      _invalidateCache () {
        this._cachedPositions = null;
      }

      _lookupPosition (node) {
        if (node.hasParent()) {
          node = node.getContainerRoot();
        }
        return this.getContent().indexOf(node.id)
      }

      _enableCaching () {
        
        if (this.document) {
          this.document.data.on('operation:applied', this._onOperationApplied, this);
          this._isCaching = true;
        }
      }

      _onOperationApplied (op) {
        if (op.type === 'set' || op.type === 'update') {
          if (op.path[0] === this.id) {
            this._invalidateCache();
          }
        }
      }

      _onDocumentChange (change) {
        if (change.hasUpdated(this.getContentPath())) {
          this._invalidateCache();
        }
      }

      
      

      hasChildren () {
        return this.getContent().length > 0
      }

      getChildIndex (child) {
        return this.getContent().indexOf(child.id)
      }

      getChildren () {
        return getNodes(this.getDocument(), this.getContent())
      }

      getChildAt (idx) {
        var childrenIds = this.getContent();
        if (idx < 0 || idx >= childrenIds.length) {
          throw new Error('Array index out of bounds: ' + idx + ', ' + childrenIds.length)
        }
        return this.getDocument().get(childrenIds[idx], 'strict')
      }

      getChildCount () {
        return this.getContent().length
      }
    }
    return AbstractContainer
  }

  class Container extends ContainerMixin(DocumentNode) {
    constructor (...args) {
      super(...args);

      
      
      
      this._enableCaching();
    }

    dispose () {
      this.document.off(this);
    }

    getContentPath () {
      return [this.id, 'nodes']
    }

    getContent () {
      return this.nodes
    }
  }

  Container.prototype._isContainer = true;

  Container.schema = {
    type: '@container',
    nodes: { type: ['array', 'id'], default: [], owned: true }
  };

  function AnnotationMixin (DocumentNode) {
    class AbstractAnnotation extends DocumentNode {
      constructor (doc, props) {
        super(doc, _normalizedProps(props));

        
        this.start = new Coordinate(this.start);
        this.end = new Coordinate(this.end);
      }

      

      get path () {
        console.warn('DEPRECATED: use annotation.start.path instead');
        return this.start.path
      }

      getPath () {
        return this.start.path
      }

      get startPath () {
        console.warn('DEPRECATED: use annotation.start.path instead.');
        return this.start.path
      }

      set startPath (path) {
        console.warn('DEPRECATED: use annotation.start.path instead.');
        this.start.path = path;
      }

      get startOffset () {
        console.warn('DEPRECATED: use annotation.start.offset instead.');
        return this.start.offset
      }

      set startOffset (offset) {
        console.warn('DEPRECATED: use annotation.start.offset instead.');
        this.start.offset = offset;
      }

      get endPath () {
        console.warn('DEPRECATED: use annotation.end.path instead.');
        return this.end.path
      }

      set endPath (path) {
        console.warn('DEPRECATED: use annotation.end.path instead.');
        this.end.path = path;
      }

      get endOffset () {
        console.warn('DEPRECATED: use annotation.end.offset instead.');
        return this.end.offset
      }

      set endOffset (offset) {
        console.warn('DEPRECATED: use annotation.end.offset instead.');
        this.end.offset = offset;
      }

      

      
      getText () {
        var doc = this.getDocument();
        if (!doc) {
          console.warn('Trying to use a Annotation which is not attached to the document.');
          return ''
        }
        return getTextForSelection(doc, this.getSelection())
      }

      isAnnotation () {
        return true
      }

      
      canSplit () {
        return true
      }

      
      isAnchor () {
        return false
      }

      
      getSelection () {
        const doc = this.getDocument();
        
        if (!doc) {
          console.warn('Trying to use a ContainerAnnotation which is not attached to the document.');
          return Selection.nullSelection()
        }
        if (this._isContainerAnnotation) {
          return doc.createSelection({
            type: 'container',
            containerId: this.containerId,
            startPath: this.start.path,
            startOffset: this.start.offset,
            endPath: this.end.path,
            endOffset: this.end.offset
          })
        } else {
          return this.getDocument().createSelection({
            type: 'property',
            path: this.start.path,
            startOffset: this.start.offset,
            endOffset: this.end.offset
          })
        }
      }

      _updateRange (tx, sel) {
        if (sel.isContainerSelection()) {
          
          if (!isEqual(this.start.path, sel.start.path)) {
            tx.set([this.id, 'start', 'path'], sel.start.path);
          }
          if (this.start.offset !== sel.start.offset) {
            tx.set([this.id, 'start', 'offset'], sel.start.offset);
          }
          if (!isEqual(this.end.path, sel.end.path)) {
            tx.set([this.id, 'end', 'path'], sel.end.path);
          }
          if (this.end.offset !== sel.end.offset) {
            tx.set([this.id, 'end', 'offset'], sel.end.offset);
          }
        } else if (sel.isPropertySelection()) {
          if (!isArrayEqual(this.start.path, sel.start.path)) {
            tx.set([this.id, 'path'], sel.start.path);
          }
          
          if (this.start.offset !== sel.start.offset) {
            tx.set([this.id, 'start', 'offset'], sel.start.offset);
          }
          if (this.end.offset !== sel.end.offset) {
            tx.set([this.id, 'end', 'offset'], sel.end.offset);
          }
        } else {
          throw new Error('Invalid selection.')
        }
      }
    }

    AbstractAnnotation.prototype._isAnnotation = true;

    AbstractAnnotation.schema = {
      start: { type: 'coordinate', default: { path: [], offset: 0 } },
      end: { type: 'coordinate', default: { path: [], offset: 0 } }
    };

    return AbstractAnnotation
  }

  function _normalizedProps (props) {
    
    
    if (!props.hasOwnProperty('start')) {
      
      
      
      let start, end;
      if (props.hasOwnProperty('startPath') || props.hasOwnProperty('path')) {
        start = {
          path: props.startPath || props.path,
          offset: props.startOffset
        };
      }
      if (props.hasOwnProperty('endPath') || props.hasOwnProperty('endOffset')) {
        end = {
          path: props.endPath || props.path,
          offset: props.endOffset
        };
      }
      if (start && !end) {
        end = cloneDeep(start);
      }
      if (start) {
        props = Object.assign({}, props);
        delete props.path;
        delete props.startPath;
        delete props.endPath;
        delete props.startOffset;
        delete props.endOffset;
        props.start = start;
        props.end = end;
      }
    } else if (props.hasOwnProperty('end') && !props.end.path) {
      props.end.path = props.start.path;
    }
    return props
  }

  class PropertyAnnotation extends AnnotationMixin(DocumentNode) {}

  PropertyAnnotation.prototype._isAnnotation = true;
  PropertyAnnotation.prototype._isPropertyAnnotation = true;

  PropertyAnnotation.isPropertyAnnotation = true;
  PropertyAnnotation.autoExpandRight = true;

  PropertyAnnotation.schema = {
    type: '@annotation',
    
    
    _content: { type: 'string', optional: true }
  };

  class ContainerAnnotation extends AnnotationMixin(DocumentNode) {
    setHighlighted (highlighted, scope) {
      if (this.highlighted !== highlighted) {
        this.highlighted = highlighted;
        this.highlightedScope = scope;
        this.emit('highlighted', highlighted, scope);
        forEach(this.fragments, function (frag) {
          frag.emit('highlighted', highlighted, scope);
        });
      }
    }
  }

  ContainerAnnotation.schema = {
    type: '@container-annotation',
    containerId: 'string',
    start: 'coordinate',
    end: 'coordinate'
  };

  ContainerAnnotation.prototype._isAnnotation = true;
  ContainerAnnotation.prototype._isContainerAnnotation = true;

  class EditingBehavior {
    constructor () {
      this._merge = {};
      this._mergeComponents = {};
      this._break = {};
    }

    defineMerge (firstType, secondType, impl) {
      if (!this._merge[firstType]) {
        this._merge[firstType] = {};
      }
      this._merge[firstType][secondType] = impl;
      return this
    }

    canMerge (firstType, secondType) {
      return (this._merge[firstType] && this._merge[firstType][secondType])
    }

    getMerger (firstType, secondType) {
      return this._merge[firstType][secondType]
    }

    defineComponentMerge (nodeType, impl) {
      this._mergeComponents[nodeType] = impl;
    }

    canMergeComponents (nodeType) {
      return this._mergeComponents[nodeType]
    }

    getComponentMerger (nodeType) {
      return this._mergeComponents[nodeType]
    }

    defineBreak (nodeType, impl) {
      this._break[nodeType] = impl;
      return this
    }

    canBreak (nodeType) {
      return this._break[nodeType]
    }

    getBreaker (nodeType) {
      return this._break[nodeType]
    }
  }

  function extend (...args) {
    return Object.assign(...args)
  }

  class ArrayIterator {
    constructor (arr) {
      this.arr = arr;
      this.pos = -1;
    }

    get _isArrayIterator () {
      return true
    }

    
    hasNext () {
      return this.pos < this.arr.length - 1
    }

    
    next () {
      this.pos += 1;
      var next = this.arr[this.pos];
      return next
    }

    
    back () {
      if (this.pos >= 0) {
        this.pos -= 1;
      }
      return this
    }

    peek () {
      return this.arr[this.pos + 1]
    }
  }

  function findIndex$1 (arr, predicate) {
    if (!isFunction$1(predicate)) return arr.indexOf(predicate)
    for (let i = 0; i < arr.length; i++) {
      if (predicate(arr[i])) return i
    }
    return -1
  }

  class DOMEventListener {
    constructor (eventName, handler, options) {
      
      if (!isString(eventName) || !isFunction$1(handler)) {
        throw new Error("Illegal arguments: 'eventName' must be a String, and 'handler' must be a Function.")
      }
      options = options || {};
      var origHandler = handler;
      var context = options.context;
      var capture = Boolean(options.capture);

      if (context) {
        handler = handler.bind(context);
      }
      if (options.once === true) {
        handler = _once(this, handler);
      }

      this.eventName = eventName;
      this.originalHandler = origHandler;
      this.handler = handler;
      this.capture = capture;
      this.context = context;
      this.options = options;
      
      this._el = null;
    }
  }

  DOMEventListener.prototype._isDOMEventListener = true;

  DOMEventListener.findIndex = function (eventListeners, eventName, handler) {
    var idx = -1;
    if (arguments[1]._isDOMEventListener) {
      idx = eventListeners.indexOf(arguments[1]);
    } else {
      idx = findIndex$1(eventListeners,
        _matches.bind(null, {
          eventName: eventName,
          originalHandler: handler
        })
      );
    }
    return idx
  };

  function _matches (l1, l2) {
    return l1.eventName === l2.eventName && l1.originalHandler === l2.originalHandler
  }

  function _once (listener, handler) {
    return function (event) {
      handler(event);
      listener._el.removeEventListener(listener);
    }
  }

  const NOT_IMPLEMENTED = 'This method is not implemented.';


  class DOMElement {
    

    

    

    

    

    

    

    

    

    
    getNativeElement () {
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    isTextNode () {
      
      return false
    }

    
    isElementNode () {
      
      return false
    }

    
    isCommentNode () {
      
      return false
    }

    
    isDocumentNode () {
      
      return false
    }

    
    getTagName () {
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    setTagName (tagName) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    getId () {
      return this.getAttribute('id')
    }

    
    setId (id) {
      this.setAttribute('id', id);
      return this
    }

    
    hasClass (className) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    addClass (classString) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    removeClass (classString) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    hasAttribute (name) {
      return Boolean(this.getAttribute(name))
    }

    
    attr () {
      if (arguments.length === 1) {
        if (isString(arguments[0])) {
          return this.getAttribute(arguments[0])
        } else if (isObject(arguments[0])) {
          forEach(arguments[0], function (value, name) {
            this.setAttribute(name, value);
          }.bind(this));
        }
      } else if (arguments.length === 2) {
        this.setAttribute(arguments[0], arguments[1]);
      }
      return this
    }

    
    removeAttr (name) {
      var names = name.split(/\s+/);
      if (names.length === 1) {
        this.removeAttribute(name);
      } else {
        names.forEach(function (name) {
          this.removeAttribute(name);
        }.bind(this));
      }
      return this
    }

    
    getAttribute (name) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    setAttribute (name, value) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    removeAttribute (name) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    getAttributes () {
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    htmlProp () {
      if (arguments.length === 1) {
        if (isString(arguments[0])) {
          return this.getProperty(arguments[0])
        } else if (isObject(arguments[0])) {
          forEach(arguments[0], function (value, name) {
            this.setProperty(name, value);
          }.bind(this));
        }
      } else if (arguments.length === 2) {
        this.setProperty(arguments[0], arguments[1]);
      }
      return this
    }

    getProperty (name) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    setProperty (name, value) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    val (value) {
      if (arguments.length === 0) {
        return this.getValue()
      } else {
        this.setValue(value);
        return this
      }
    }

    getValue () {
      return this.getProperty('value')
    }

    setValue (value) {
      this.setProperty('value', value);
      return this
    }

    
    css () {
      
      if (arguments.length === 1) {
        
        if (isString(arguments[0])) {
          return this.getStyle(arguments[0])
        } else if (isObject(arguments[0])) {
          forEach(arguments[0], function (value, name) {
            this.setStyle(name, value);
          }.bind(this));
        } else {
          throw new Error('Illegal arguments.')
        }
      } else if (arguments.length === 2) {
        this.setStyle(arguments[0], arguments[1]);
      } else {
        throw new Error('Illegal arguments.')
      }
      return this
    }

    getStyle (name) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    setStyle (name, value) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    text (text) {
      if (arguments.length === 0) {
        return this.getTextContent()
      } else {
        this.setTextContent(text);
      }
      return this
    }

    
    getTextContent () {
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    setTextContent (text) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    html (html) {
      if (arguments.length === 0) {
        return this.getInnerHTML()
      } else {
        this.setInnerHTML(html);
      }
      return this
    }

    
    getInnerHTML () {
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    getInnerXML () {
      return this.getInnerHTML()
    }

    
    setInnerHTML (html) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    setInnerXML (xml) {
      return this.setInnerHTML(xml)
    }

    
    getOuterHTML () {
      
      throw new Error(NOT_IMPLEMENTED)
    }

    getOuterXML () {
      return this.getOuterHTML()
    }

    
    on (eventName, handler, context, options) {
      
      if (!isString(eventName)) {
        throw new Error('Illegal argument: "event" must be a String.')
      }
      options = options || {};
      if (context) {
        options.context = context;
      }
      
      if (!handler || !isFunction$1(handler)) {
        throw new Error('Illegal argument: invalid handler function for event ' + eventName)
      }
      this.addEventListener(eventName, handler, options);
      return this
    }

    
    off (eventName, handler) {
      
      if (arguments.length === 1 && !isString(eventName)) {
        let context = arguments[0];
        this.getEventListeners().filter(function (l) {
          return l.context === context
        }).forEach(function (l) {
          this.removeEventListener(l);
        }.bind(this));
      } else {
        this.removeEventListener(eventName, handler);
      }
      return this
    }

    addEventListener (eventName, handler, options = {}) {
      let listener;
      if (arguments.length === 1 && arguments[0]) {
        listener = arguments[0];
      } else {
        listener = this._createEventListener(eventName, handler, options);
      }
      if (!this.eventListeners) {
        this.eventListeners = [];
      }
      listener._el = this;
      this.eventListeners.push(listener);
      this._addEventListenerNative(listener);
      return this
    }

    _createEventListener (eventName, handler, options) {
      return new DOMEventListener(eventName, handler, options)
    }

    _addEventListenerNative(listener) {} 

    removeEventListener (eventName, handler) {
      if (!this.eventListeners) return
      
      let listener = null;
      let idx = DOMEventListener.findIndex(this.eventListeners, eventName, handler);
      listener = this.eventListeners[idx];
      if (idx > -1) {
        this.eventListeners.splice(idx, 1);
        
        listener._el = null;
        this._removeEventListenerNative(listener);
      }
      return this
    }

    _removeEventListenerNative(listener) {} 

    removeAllEventListeners () {
      if (!this.eventListeners) return
      for (let i = 0; i < this.eventListeners.length; i++) {
        let listener = this.eventListeners[i];
        
        listener._el = null;
        this._removeEventListenerNative(listener);
      }
      delete this.eventListeners;
    }

    getEventListeners () {
      return this.eventListeners || []
    }

    
    getNodeType () {
      
      throw new Error(NOT_IMPLEMENTED)
    }

    getContentType () {
      
      throw new Error(NOT_IMPLEMENTED)
    }

    getChildCount () {
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    getChildNodes () {
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    getChildren () {
      
      throw new Error(NOT_IMPLEMENTED)
    }

    getChildAt (pos) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    getChildIndex (child) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    getChildNodeIterator () {
      return new ArrayIterator(this.getChildNodes())
    }

    getLastChild () {
      
      throw new Error(NOT_IMPLEMENTED)
    }

    getFirstChild () {
      
      throw new Error(NOT_IMPLEMENTED)
    }

    getNextSibling () {
      
      throw new Error(NOT_IMPLEMENTED)
    }

    getPreviousSibling () {
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    clone(deep) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    createElement (str) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    createTextNode (text) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    createComment (data) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    createProcessingInstruction (name, data) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    createCDATASection (data) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    is (cssSelector) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    getParent () {
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    getOwnerDocument () {
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    getDoctype () {
      
      throw new Error('NOT_IMPLEMENTED')
    }

    setDoctype(qualifiedNameStr, publicId, systemId) { 
      throw new Error('NOT_IMPLEMENTED')
    }

    setDocType (...args) {
      console.error('DEPRECATED: use setDoctype() instead');
      return this.setDoctype(...args)
    }

    
    find (cssSelector) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    findAll (cssSelector) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    append (child) {
      var children;
      if (arguments.length === 1) {
        if (isArray(child)) {
          children = child;
        } else {
          this.appendChild(child);
          return this
        }
      } else {
        children = arguments;
      }
      if (children) {
        Array.prototype.forEach.call(children, this.appendChild.bind(this));
      }
      return this
    }

    appendChild (child) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    insertAt (pos, child) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    insertBefore (newChild, before) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    removeAt (pos) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    removeChild (child) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    replaceChild (oldChild, newChild) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    
    remove () {
      var parent = this.getParent();
      if (parent) {
        parent.removeChild(this);
      }
    }

    
    empty () {
      
      throw new Error(NOT_IMPLEMENTED)
    }

    serialize () {
      return this.getOuterHTML()
    }

    isInDocument () {
      let el = this;
      while (el) {
        if (el.isDocumentNode()) {
          return true
        }
        el = el.getParent();
      }
    }

    
    focus () {
      
      return this
    }

    
    select () {
      
      return this
    }

    
    blur () {
      
      return this
    }

    
    click () {
      
      return this
    }

    

    getWidth () {
      
      return 0
    }

    getHeight () {
      
      return 0
    }

    
    getOuterHeight (withMargin) { 
      
      return 0
    }

    
    getOffset () {
      
      return { top: 0, left: 0 }
    }

    
    getPosition () {
      
      return { top: 0, left: 0 }
    }

    
    getElementFactory () {
      return this.createElement.bind(this)
    }

    
    emit(name, data) { 
      
      throw new Error(NOT_IMPLEMENTED)
    }

    

    get id () {
      return this.getId()
    }

    set id (id) {
      this.setId(id);
    }

    get tagName () {
      return this.getTagName()
    }

    set tagName (tagName) {
      this.setTagName(tagName);
    }

    get nodeName () {
      return this.getTagName()
    }

    get nodeType () {
      return this.getNodeType()
    }

    get className () {
      return this.getAttribute('class')
    }

    set className (className) {
      this.setAttribute('class', className);
    }

    get textContent () {
      return this.getTextContent()
    }

    set textContent (text) {
      this.setTextContent(text);
    }

    get innerHTML () {
      return this.getInnerHTML()
    }

    set innerHTML (html) {
      this.setInnerHTML(html);
    }

    get outerHTML () {
      return this.getOuterHTML()
    }

    get firstChild () {
      return this.getFirstChild()
    }

    get lastChild () {
      return this.getLastChild()
    }

    get nextSibling () {
      return this.getNextSibling()
    }

    get previousSibling () {
      return this.getPreviousSibling()
    }

    get parentNode () {
      return this.getParent()
    }

    get height () {
      return this.getHeight()
    }

    get width () {
      return this.getWidth()
    }

    get value () {
      return this.getValue()
    }

    set value (value) {
      return this.setValue(value)
    }
  }

  DOMElement.prototype._isDOMElement = true;

  DOMElement.pxStyles = {
    top: true,
    bottom: true,
    left: true,
    right: true,
    height: true,
    width: true
  };

  DOMElement.EMPTY_HTML = '<html><head></head><body></body></html>';

  const SIGNATURE = uuid('_BrowserDOMElement');

  function _attach (nativeEl, browserDOMElement) {
    nativeEl[SIGNATURE] = browserDOMElement;
  }

  function _detach (nativeEl) {
    delete nativeEl[SIGNATURE];
  }

  function _unwrap (nativeEl) {
    return nativeEl[SIGNATURE]
  }

  class BrowserDOMElement extends DOMElement {
    constructor (el) {
      super();
      console.assert(el instanceof window.Node, 'Expecting native DOM node.');
      this.el = el;
      
      
      _attach(el, this);
    }

    getNativeElement () {
      return this.el
    }

    getNodeType () {
      switch (this.el.nodeType) {
        case window.Node.TEXT_NODE:
          return 'text'
        case window.Node.ELEMENT_NODE:
          return 'element'
        case window.Node.DOCUMENT_NODE:
          return 'document'
        case window.Node.COMMENT_NODE:
          return 'comment'
        case window.Node.PROCESSING_INSTRUCTION_NODE:
          return 'directive'
        case window.Node.CDATA_SECTION_NODE:
          return 'cdata'
        default:
          
      }
    }

    getDoctype () {
      if (this.isDocumentNode()) {
        return this.el.doctype
      } else {
        return this.getOwnerDocument().getDoctype()
      }
    }

    setDoctype (qualifiedNameStr, publicId, systemId) {
      let ownerDocument = this._getNativeOwnerDocument();
      let oldDocType = ownerDocument.doctype;
      let newDocType = ownerDocument.implementation.createDocumentType(
        qualifiedNameStr, publicId, systemId
      );
      if (oldDocType) {
        oldDocType.parentNode.replaceChild(newDocType, oldDocType);
      } else {
        ownerDocument.insertBefore(newDocType, ownerDocument.firstChild);
      }
    }

    isTextNode () {
      return (this.el.nodeType === window.Node.TEXT_NODE)
    }

    isElementNode () {
      return (this.el.nodeType === window.Node.ELEMENT_NODE)
    }

    isCommentNode () {
      return (this.el.nodeType === window.Node.COMMENT_NODE)
    }

    isDocumentNode () {
      return (this.el.nodeType === window.Node.DOCUMENT_NODE)
    }

    hasClass (className) {
      return this.el.classList.contains(className)
    }

    addClass (className) {
      this.el.classList.add(className);
      return this
    }

    removeClass (className) {
      this.el.classList.remove(className);
      return this
    }

    getAttribute (name) {
      return this.el.getAttribute(name)
    }

    setAttribute (name, value) {
      this.el.setAttribute(name, String(value));
      return this
    }

    removeAttribute (name) {
      this.el.removeAttribute(name);
      return this
    }

    getAttributes () {
      if (!this.el.attributes._mapAdapter) {
        this.el.attributes._mapAdapter = new AttributesMapAdapter(this.el.attributes);
      }
      return this.el.attributes._mapAdapter
    }

    getProperty (name) {
      return this.el[name]
    }

    setProperty (name, value) {
      
      
      if (this._isXML()) throw new Error('setProperty() is only supported for HTML elements.')
      this.el[name] = value;
      return this
    }

    getTagName () {
      
      
      if (this._isXML()) {
        return this.el.tagName
      } else if (this.el.tagName) {
        return this.el.tagName.toLowerCase()
      }
    }

    setTagName (tagName) {
      let newEl = this.createElement(tagName);
      let attributes = this.el.attributes;
      let l = attributes.length;
      let i;
      for (i = 0; i < l; i++) {
        let attr = attributes.item(i);
        newEl.setAttribute(attr.name, attr.value);
      }
      if (this.eventListeners) {
        this.eventListeners.forEach(function (listener) {
          newEl.addEventListener(listener.eventName, listener.handler, listener.capture);
        });
      }
      newEl.append(this.getChildNodes());

      this._replaceNativeEl(newEl.getNativeElement());
      return this
    }

    getId () {
      return this.el.id
    }

    setId (id) {
      this.el.id = id;
      return this
    }

    getStyle (name) {
      let val = this.el.style[name];
      if (!val) {
        let computedStyle = this.getComputedStyle();
        val = computedStyle[name];
      }
      return val
    }

    getComputedStyle () {
      return window.getComputedStyle(this.el)
    }

    setStyle (name, value) {
      if (DOMElement.pxStyles[name] && isNumber(value)) value = value + 'px';
      this.el.style[name] = value;
      return this
    }

    getTextContent () {
      return this.el.textContent
    }

    setTextContent (text) {
      this.el.textContent = text;
      return this
    }

    getInnerHTML () {
      if (this._isXML()) {
        let xs = new window.XMLSerializer();
        let result = Array.prototype.map.call(this.el.childNodes, c => xs.serializeToString(c));
        return result.join('')
      } else {
        return this.el.innerHTML
      }
    }

    setInnerHTML (html) {
      
      
      this.el.innerHTML = html;
      return this
    }

    getOuterHTML () {
      
      
      if (this._isXML()) {
        let xs = new window.XMLSerializer();
        return xs.serializeToString(this.el)
      } else {
        return this.el.outerHTML
      }
    }

    _addEventListenerNative (listener) {
      this.el.addEventListener(listener.eventName, listener.handler, listener.capture);
    }

    _removeEventListenerNative (listener) {
      this.el.removeEventListener(listener.eventName, listener.handler);
    }

    getEventListeners () {
      return this.eventListeners || []
    }

    getChildCount () {
      return this.el.childNodes.length
    }

    getChildNodes () {
      let childNodes = [];
      for (let node = this.el.firstChild; node; node = node.nextSibling) {
        childNodes.push(BrowserDOMElement.wrap(node));
      }
      return childNodes
    }

    getChildNodeIterator () {
      return new BrowserChildNodeIterator(this.el)
    }

    get childNodes () {
      return this.getChildNodes()
    }

    getChildren () {
      
      
      let children = [];
      for (let node = this.el.firstChild; node; node = node.nextSibling) {
        if (node.nodeType === window.Node.ELEMENT_NODE) {
          children.push(BrowserDOMElement.wrap(node));
        }
      }
      return children
    }

    get children () {
      return this.getChildren()
    }

    getChildAt (pos) {
      return BrowserDOMElement.wrap(this.el.childNodes[pos])
    }

    getChildIndex (child) {
      
      if (!child._isBrowserDOMElement) {
        throw new Error('Expecting a BrowserDOMElement instance.')
      }
      return Array.prototype.indexOf.call(this.el.childNodes, child.el)
    }

    getFirstChild () {
      let firstChild = this.el.firstChild;
      
      if (firstChild) {
        return BrowserDOMElement.wrap(firstChild)
      } else {
        return null
      }
    }

    getLastChild () {
      var lastChild = this.el.lastChild;
      
      if (lastChild) {
        return BrowserDOMElement.wrap(lastChild)
      } else {
        return null
      }
    }

    getNextSibling () {
      let next = this.el.nextSibling;
      
      if (next) {
        return BrowserDOMElement.wrap(next)
      } else {
        return null
      }
    }

    getPreviousSibling () {
      let previous = this.el.previousSibling;
      
      if (previous) {
        return BrowserDOMElement.wrap(previous)
      } else {
        return null
      }
    }

    clone (deep) {
      let clone = this.el.cloneNode(deep);
      return BrowserDOMElement.wrap(clone)
    }

    createDocument (format) {
      return BrowserDOMElement.createDocument(format)
    }

    createElement (tagName) {
      let doc = this._getNativeOwnerDocument();
      let el = doc.createElement(tagName);
      return BrowserDOMElement.wrap(el)
    }

    createTextNode (text) {
      let doc = this._getNativeOwnerDocument();
      let el = doc.createTextNode(text);
      return BrowserDOMElement.wrap(el)
    }

    createComment (data) {
      let doc = this._getNativeOwnerDocument();
      let el = doc.createComment(data);
      return BrowserDOMElement.wrap(el)
    }

    createProcessingInstruction (name, data) {
      let doc = this._getNativeOwnerDocument();
      let el = doc.createProcessingInstruction(name, data);
      return BrowserDOMElement.wrap(el)
    }

    createCDATASection (data) {
      let doc = this._getNativeOwnerDocument();
      let el = doc.createCDATASection(data);
      return BrowserDOMElement.wrap(el)
    }

    is (cssSelector) {
      
      
      let el = this.el;
      
      if (this.isElementNode()) {
        return matches(el, cssSelector)
      } else {
        return false
      }
    }

    getParent () {
      let parent = this.el.parentNode;
      
      if (parent) {
        return BrowserDOMElement.wrap(parent)
      } else {
        return null
      }
    }

    getOwnerDocument () {
      return BrowserDOMElement.wrap(this._getNativeOwnerDocument())
    }

    get ownerDocument () {
      return this.getOwnerDocument()
    }

    _getNativeOwnerDocument () {
      return (this.isDocumentNode() ? this.el : this.el.ownerDocument)
    }

    find (cssSelector) {
      let result = null;
      if (this.el.querySelector) {
        result = this.el.querySelector(cssSelector);
      }
      if (result) {
        return BrowserDOMElement.wrap(result)
      } else {
        return null
      }
    }

    findAll (cssSelector) {
      let result = [];
      if (this.el.querySelectorAll) {
        result = this.el.querySelectorAll(cssSelector);
      }
      return Array.prototype.map.call(result, function (el) {
        return BrowserDOMElement.wrap(el)
      })
    }

    _normalizeChild (child) {
      if (isNil(child)) return child

      if (child instanceof window.Node) {
        child = BrowserDOMElement.wrap(child);
      
      
      
      } else if (child._isBrowserDOMElement && !(child instanceof BrowserDOMElement)) {
        child = BrowserDOMElement.wrap(child);
      } else if (isString(child) || isNumber(child)) {
        child = this.createTextNode(child);
      }
      
      if (!child || !child._isBrowserDOMElement) {
        throw new Error('Illegal child type.')
      }
      console.assert(_unwrap(child.el) === child, 'The backlink to the wrapper should be consistent');
      return child.getNativeElement()
    }

    appendChild (child) {
      let nativeChild = this._normalizeChild(child);
      if (nativeChild) {
        this.el.appendChild(nativeChild);
      }
      return this
    }

    insertAt (pos, child) {
      let nativeChild = this._normalizeChild(child);
      let childNodes = this.el.childNodes;
      if (pos >= childNodes.length) {
        this.el.appendChild(nativeChild);
      } else {
        this.el.insertBefore(nativeChild, childNodes[pos]);
      }
      return this
    }

    insertBefore (child, before) {
      
      if (isNil(before)) {
        return this.appendChild(child)
      }
      if (!before._isBrowserDOMElement) {
        throw new Error('insertBefore(): Illegal arguments. "before" must be a BrowserDOMElement instance.')
      }
      var nativeChild = this._normalizeChild(child);
      if (nativeChild) {
        this.el.insertBefore(nativeChild, before.el);
      }
      return this
    }

    removeAt (pos) {
      this.el.removeChild(this.el.childNodes[pos]);
      return this
    }

    removeChild (child) {
      
      if (!child || !child._isBrowserDOMElement) {
        throw new Error('removeChild(): Illegal arguments. Expecting a BrowserDOMElement instance.')
      }
      this.el.removeChild(child.el);
      return this
    }

    replaceChild (oldChild, newChild) {
      
      if (!newChild || !oldChild ||
          !newChild._isBrowserDOMElement || !oldChild._isBrowserDOMElement) {
        throw new Error('replaceChild(): Illegal arguments. Expecting BrowserDOMElement instances.')
      }
      
      this.el.replaceChild(newChild.el, oldChild.el);
      return this
    }

    empty () {
      let el = this.el;
      while (el.lastChild) {
        el.removeChild(el.lastChild);
      }
      return this
    }

    remove () {
      if (this.el.parentNode) {
        this.el.parentNode.removeChild(this.el);
      }
      return this
    }

    serialize () {
      let outerHTML = this.el.outerHTML;
      if (isString(outerHTML)) {
        return outerHTML
      } else {
        let xs = new window.XMLSerializer();
        return xs.serializeToString(this.el)
      }
    }

    isInDocument () {
      let el = this.el;
      while (el) {
        if (el.nodeType === window.Node.DOCUMENT_NODE) {
          return true
        }
        el = el.parentNode;
      }
    }

    _replaceNativeEl (newEl) {
      console.assert(newEl instanceof window.Node, 'Expecting a native element.');
      let oldEl = this.el;
      let parentNode = oldEl.parentNode;
      if (parentNode) {
        parentNode.replaceChild(newEl, oldEl);
      }
      this.el = newEl;
      _detach(oldEl);
      _attach(newEl, this);
    }

    _getChildNodeCount () {
      return this.el.childNodes.length
    }

    focus () {
      this.el.focus();
      return this
    }

    select () {
      this.el.select();
      return this
    }

    blur () {
      this.el.blur();
      return this
    }

    click () {
      this.el.click();
      return this
    }

    getWidth () {
      let rect = this.el.getClientRects()[0];
      if (rect) {
        return rect.width
      } else {
        return 0
      }
    }

    getHeight () {
      let rect = this.el.getClientRects()[0];
      if (rect) {
        return rect.height
      } else {
        return 0
      }
    }

    getOffset () {
      let rect = this.el.getBoundingClientRect();
      return {
        top: rect.top + document.body.scrollTop,
        left: rect.left + document.body.scrollLeft
      }
    }

    getPosition () {
      return {left: this.el.offsetLeft, top: this.el.offsetTop}
    }

    getOuterHeight (withMargin) {
      let outerHeight = this.el.offsetHeight;
      if (withMargin) {
        let style = this.getComputedStyle();
        outerHeight += parseInt(style.marginTop, 10) + parseInt(style.marginBottom, 10);
      }
      return outerHeight
    }

    getContentType () {
      return this._getNativeOwnerDocument().contentType
    }

    _isXML () {
      return this.getContentType() === 'application/xml'
    }

    emit (name, data) {
      let event;
      if (data) {
        event = new window.CustomEvent(name, {
          detail: data,
          bubbles: true,
          cancelable: true
        });
      } else {
        event = new window.Event(name, {
          bubbles: true,
          cancelable: true
        });
      }
      this.el.dispatchEvent(event);
    }
  }

  BrowserDOMElement.prototype._isBrowserDOMElement = true;


  BrowserDOMElement.createDocument = function (format) {
    let doc;
    if (format === 'xml') {
      
      doc = window.document.implementation.createDocument(null, 'dummy');
      
      doc.removeChild(doc.firstChild);
    } else {
      doc = (new window.DOMParser()).parseFromString(DOMElement.EMPTY_HTML, 'text/html');
    }
    return BrowserDOMElement.wrap(doc)
  };

  BrowserDOMElement.parseMarkup = function (str, format, options = {}) {
    if (!str) {
      return BrowserDOMElement.createDocument(format)
    }
    if (options.snippet) {
      str = `<div id='__snippet__'>${str}</div>`;
    }
    let doc;
    let parser = new window.DOMParser();
    if (format === 'html') {
      doc = BrowserDOMElement.wrap(
        _check(
          parser.parseFromString(str, 'text/html')
        )
      );
    } else if (format === 'xml') {
      doc = BrowserDOMElement.wrap(
        _check(
          parser.parseFromString(str, 'application/xml')
        )
      );
    }
    if (options.snippet) {
      let childNodes = doc.find('#__snippet__').childNodes;
      if (childNodes.length === 1) {
        return childNodes[0]
      } else {
        return childNodes
      }
    } else {
      return doc
    }

    function _check (doc) {
      if (doc) {
        let parserError = doc.querySelector('parsererror');
        if (parserError) {
          throw new Error('ParserError: ' + parserError)
        }
      }
      return doc
    }
  };

  BrowserDOMElement.wrap =
  BrowserDOMElement.wrapNativeElement = function (el) {
    if (el) {
      let _el = _unwrap(el);
      if (_el) {
        return _el
      } else if (el instanceof window.Node) {
        return new BrowserDOMElement(el)
      } else if (el._isBrowserDOMElement) {
        return new BrowserDOMElement(el.getNativeElement())
      } else if (el === window) {
        return BrowserDOMElement.getBrowserWindow()
      }
    } else {
      return null
    }
  };

  BrowserDOMElement.unwrap = function (nativeEl) {
    return _unwrap(nativeEl)
  };


  class BrowserWindow {
    constructor () {
      
      this.el = window;
      _attach(window, this);
    }
  }

  BrowserWindow.prototype.on = BrowserDOMElement.prototype.on;
  BrowserWindow.prototype.off = BrowserDOMElement.prototype.off;
  BrowserWindow.prototype.addEventListener = BrowserDOMElement.prototype.addEventListener;
  BrowserWindow.prototype.removeEventListener = BrowserDOMElement.prototype.removeEventListener;
  BrowserWindow.prototype._createEventListener = BrowserDOMElement.prototype._createEventListener;
  BrowserWindow.prototype._addEventListenerNative = BrowserDOMElement.prototype._addEventListenerNative;
  BrowserWindow.prototype._removeEventListenerNative = BrowserDOMElement.prototype._removeEventListenerNative;

  BrowserWindow.prototype.getEventListeners = BrowserDOMElement.prototype.getEventListeners;

  BrowserDOMElement.getBrowserWindow = function () {
    if (window[SIGNATURE]) return window[SIGNATURE]
    return new BrowserWindow(window)
  };

  BrowserDOMElement.isReverse = function (anchorNode, anchorOffset, focusNode, focusOffset) {
    
    
    if (focusNode && anchorNode) {
      if (!BrowserDOMElement.isReverse._r1) {
        BrowserDOMElement.isReverse._r1 = window.document.createRange();
        BrowserDOMElement.isReverse._r2 = window.document.createRange();
      }
      const _r1 = BrowserDOMElement.isReverse._r1;
      const _r2 = BrowserDOMElement.isReverse._r2;
      _r1.setStart(anchorNode.getNativeElement(), anchorOffset);
      _r2.setStart(focusNode.getNativeElement(), focusOffset);
      let cmp = _r1.compareBoundaryPoints(window.Range.START_TO_START, _r2);
      if (cmp === 1) {
        return true
      }
    }
    return false
  };

  BrowserDOMElement.getWindowSelection = function () {
    let nativeSel = window.getSelection();
    let result = {
      anchorNode: BrowserDOMElement.wrap(nativeSel.anchorNode),
      anchorOffset: nativeSel.anchorOffset,
      focusNode: BrowserDOMElement.wrap(nativeSel.focusNode),
      focusOffset: nativeSel.focusOffset
    };
    return result
  };

  function matches (el, selector) {
    let elProto = window.Element.prototype;
    let _matches = (
      elProto.matches || elProto.matchesSelector ||
      elProto.msMatchesSelector || elProto.webkitMatchesSelector
    );
    return _matches.call(el, selector)
  }

  class AttributesMapAdapter {
    constructor (attributes) {
      this.attributes = attributes;
    }

    get size () {
      return this.attributes.length
    }

    get (name) {
      let item = this.attributes.getNamedItem(name);
      if (item) {
        return item.value
      }
    }

    set (name, value) {
      this.attributes.setNamedItem(name, value);
    }

    forEach (fn) {
      const S = this.size;
      for (let i = 0; i < S; i++) {
        const item = this.attributes.item(i);
        fn(item.value, item.name);
      }
    }

    map (fn) {
      let result = [];
      this.forEach((val, key) => { result.push(fn(val, key)); });
      return result
    }

    keys () {
      return this.map((val, key) => { return key })
    }

    values () {
      return this.map((val) => { return val })
    }

    entries () {
      return this.map((val, key) => { return [key, val] })
    }
  }

  class BrowserChildNodeIterator {
    constructor (el) {
      this._next = el.firstChild;
      this._curr = null;
    }

    hasNext () {
      return Boolean(this._next)
    }

    next () {
      let next = this._next;
      this._curr = next;
      this._next = next.nextSibling;
      return BrowserDOMElement.wrap(next)
    }

    back () {
      this._next = this._curr;
      this._curr = this._curr.previousSibling;
    }

    peek () {
      return BrowserDOMElement.wrap(this._curr)
    }
  }

  var inBrowser = platform.inBrowser;

  var domelementtype = {
  	Text: "text", 
  	Directive: "directive", 
  	Comment: "comment", 
  	Script: "script", 
  	Style: "style", 
  	Tag: "tag", 
  	CDATA: "cdata", 
  	Doctype: "doctype",

  	isTag: function(elem){
  		return elem.type === "tag" || elem.type === "script" || elem.type === "style";
  	}
  };

  var boolbase = {
  	trueFunc: function trueFunc(){
  		return true;
  	},
  	falseFunc: function falseFunc(){
  		return false;
  	}
  };

  var parse_1 = parse;




  var re_nthElement = /^([+\-]?\d*n)?\s*(?:([+\-]?)\s*(\d+))?$/;


  function parse(formula){
  	formula = formula.trim().toLowerCase();

  	if(formula === "even"){
  		return [2, 0];
  	} else if(formula === "odd"){
  		return [2, 1];
  	} else {
  		var parsed = formula.match(re_nthElement);

  		if(!parsed){
  			throw new SyntaxError("n-th rule couldn't be parsed ('" + formula + "')");
  		}

  		var a;

  		if(parsed[1]){
  			a = parseInt(parsed[1], 10);
  			if(isNaN(a)){
  				if(parsed[1].charAt(0) === "-") a = -1;
  				else a = 1;
  			}
  		} else a = 0;

  		return [
  			a,
  			parsed[3] ? parseInt((parsed[2] || "") + parsed[3], 10) : 0
  		];
  	}
  }

  var compile_1 = compile;

  var trueFunc  = boolbase.trueFunc,
      falseFunc = boolbase.falseFunc;


  function compile(parsed){
  	var a = parsed[0],
  	    b = parsed[1] - 1;

  	
  	
  	if(b < 0 && a <= 0) return falseFunc;

  	
  	if(a ===-1) return function(pos){ return pos <= b; };
  	if(a === 0) return function(pos){ return pos === b; };
  	
  	if(a === 1) return b < 0 ? trueFunc : function(pos){ return pos >= b; };

  	
  	var bMod = b % a;
  	if(bMod < 0) bMod += a;

  	if(a > 1){
  		return function(pos){
  			return pos >= b && pos % a === bMod;
  		};
  	}

  	a *= -1; 

  	return function(pos){
  		return pos <= b && pos % a === bMod;
  	};
  }

  var nthCheck = function nthCheck(formula){
  	return compile_1(parse_1(formula));
  };

  var parse_1$1 = parse_1;
  var compile_1$1 = compile_1;
  nthCheck.parse = parse_1$1;
  nthCheck.compile = compile_1$1;

  var cssWhat = parse$1;

  var re_name = /^(?:\\.|[\w\-\u00c0-\uFFFF])+/,
      re_escape = /\\([\da-f]{1,6}\s?|(\s)|.)/ig,
      
      re_attr = /^\s*((?:\\.|[\w\u00c0-\uFFFF\-])+)\s*(?:(\S?)=\s*(?:(['"])(.*?)\3|(#?(?:\\.|[\w\u00c0-\uFFFF\-])*)|)|)\s*(i)?\]/;

  var actionTypes = {
  	__proto__: null,
  	"undefined": "exists",
  	"":  "equals",
  	"~": "element",
  	"^": "start",
  	"$": "end",
  	"*": "any",
  	"!": "not",
  	"|": "hyphen"
  };

  var simpleSelectors = {
  	__proto__: null,
  	">": "child",
  	"<": "parent",
  	"~": "sibling",
  	"+": "adjacent"
  };

  var attribSelectors = {
  	__proto__: null,
  	"#": ["id", "equals"],
  	".": ["class", "element"]
  };


  var unpackPseudos = {
  	__proto__: null,
  	"has": true,
  	"not": true,
  	"matches": true
  };

  var stripQuotesFromPseudos = {
  	__proto__: null,
  	"contains": true,
  	"icontains": true
  };

  var quotes = {
  	__proto__: null,
  	"\"": true,
  	"'": true
  };


  function funescape( _, escaped, escapedWhitespace ) {
  	var high = "0x" + escaped - 0x10000;
  	
  	
  	
  	return high !== high || escapedWhitespace ?
  		escaped :
  		
  		high < 0 ?
  			String.fromCharCode( high + 0x10000 ) :
  			
  			String.fromCharCode( high >> 10 | 0xD800, high & 0x3FF | 0xDC00 );
  }

  function unescapeCSS(str){
  	return str.replace(re_escape, funescape);
  }

  function isWhitespace(c){
  	return c === " " || c === "\n" || c === "\t" || c === "\f" || c === "\r";
  }

  function parse$1(selector, options){
  	var subselects = [];

  	selector = parseSelector(subselects, selector + "", options);

  	if(selector !== ""){
  		throw new SyntaxError("Unmatched selector: " + selector);
  	}

  	return subselects;
  }

  function parseSelector(subselects, selector, options){
  	var tokens = [],
  		sawWS = false,
  		data, firstChar, name, quot;

  	function getName(){
  		var sub = selector.match(re_name)[0];
  		selector = selector.substr(sub.length);
  		return unescapeCSS(sub);
  	}

  	function stripWhitespace(start){
  		while(isWhitespace(selector.charAt(start))) start++;
  		selector = selector.substr(start);
  	}

  	stripWhitespace(0);

  	while(selector !== ""){
  		firstChar = selector.charAt(0);

  		if(isWhitespace(firstChar)){
  			sawWS = true;
  			stripWhitespace(1);
  		} else if(firstChar in simpleSelectors){
  			tokens.push({type: simpleSelectors[firstChar]});
  			sawWS = false;

  			stripWhitespace(1);
  		} else if(firstChar === ","){
  			if(tokens.length === 0){
  				throw new SyntaxError("empty sub-selector");
  			}
  			subselects.push(tokens);
  			tokens = [];
  			sawWS = false;
  			stripWhitespace(1);
  		} else {
  			if(sawWS){
  				if(tokens.length > 0){
  					tokens.push({type: "descendant"});
  				}
  				sawWS = false;
  			}

  			if(firstChar === "*"){
  				selector = selector.substr(1);
  				tokens.push({type: "universal"});
  			} else if(firstChar in attribSelectors){
  				selector = selector.substr(1);
  				tokens.push({
  					type: "attribute",
  					name: attribSelectors[firstChar][0],
  					action: attribSelectors[firstChar][1],
  					value: getName(),
  					ignoreCase: false
  				});
  			} else if(firstChar === "["){
  				selector = selector.substr(1);
  				data = selector.match(re_attr);
  				if(!data){
  					throw new SyntaxError("Malformed attribute selector: " + selector);
  				}
  				selector = selector.substr(data[0].length);
  				name = unescapeCSS(data[1]);

  				if(
  					!options || (
  						"lowerCaseAttributeNames" in options ?
  							options.lowerCaseAttributeNames :
  							!options.xmlMode
  					)
  				){
  					name = name.toLowerCase();
  				}

  				tokens.push({
  					type: "attribute",
  					name: name,
  					action: actionTypes[data[2]],
  					value: unescapeCSS(data[4] || data[5] || ""),
  					ignoreCase: !!data[6]
  				});

  			} else if(firstChar === ":"){
  				if(selector.charAt(1) === ":"){
  					selector = selector.substr(2);
  					tokens.push({type: "pseudo-element", name: getName().toLowerCase()});
  					continue;
  				}

  				selector = selector.substr(1);

  				name = getName().toLowerCase();
  				data = null;

  				if(selector.charAt(0) === "("){
  					if(name in unpackPseudos){
  						quot = selector.charAt(1);
  						var quoted = quot in quotes;

  						selector = selector.substr(quoted + 1);

  						data = [];
  						selector = parseSelector(data, selector, options);

  						if(quoted){
  							if(selector.charAt(0) !== quot){
  								throw new SyntaxError("unmatched quotes in :" + name);
  							} else {
  								selector = selector.substr(1);
  							}
  						}

  						if(selector.charAt(0) !== ")"){
  							throw new SyntaxError("missing closing parenthesis in :" + name + " " + selector);
  						}

  						selector = selector.substr(1);
  					} else {
  						var pos = 1, counter = 1;

  						for(; counter > 0 && pos < selector.length; pos++){
  							if(selector.charAt(pos) === "(") counter++;
  							else if(selector.charAt(pos) === ")") counter--;
  						}

  						if(counter){
  							throw new SyntaxError("parenthesis not matched");
  						}

  						data = selector.substr(1, pos - 2);
  						selector = selector.substr(pos);

  						if(name in stripQuotesFromPseudos){
  							quot = data.charAt(0);

  							if(quot === data.slice(-1) && quot in quotes){
  								data = data.slice(1, -1);
  							}

  							data = unescapeCSS(data);
  						}
  					}
  				}

  				tokens.push({type: "pseudo", name: name, data: data});
  			} else if(re_name.test(selector)){
  				name = getName();

  				if(!options || ("lowerCaseTags" in options ? options.lowerCaseTags : !options.xmlMode)){
  					name = name.toLowerCase();
  				}

  				tokens.push({type: "tag", name: name});
  			} else {
  				if(tokens.length && tokens[tokens.length - 1].type === "descendant"){
  					tokens.pop();
  				}
  				addToken(subselects, tokens);
  				return selector;
  			}
  		}
  	}

  	addToken(subselects, tokens);

  	return selector;
  }

  function addToken(subselects, tokens){
  	if(subselects.length > 0 && tokens.length === 0){
  		throw new SyntaxError("empty sub-selector");
  	}

  	subselects.push(tokens);
  }

  const _encodeXMLContent = ((obj) => {
    let invObj = getInverseObj(obj);
    let replacer = getInverseReplacer(invObj);
    return getInverse(invObj, replacer)
  })({
    amp: '&',
    gt: '>',
    lt: '<'
  });

  const _encodeXMLAttr = ((obj) => {
    let invObj = getInverseObj(obj);
    let replacer = getInverseReplacer(invObj);
    return getInverse(invObj, replacer)
  })({
    quot: '"'
  });

  function getInverseObj (obj) {
    return Object.keys(obj).sort().reduce(function (inverse, name) {
      inverse[obj[name]] = '&' + name + ';';
      return inverse
    }, {})
  }

  function getInverseReplacer (inverse) {
    let single = [];
    let multiple = [];

    Object.keys(inverse).forEach(function (k) {
      if (k.length === 1) {
        single.push('\\' + k);
      } else {
        multiple.push(k);
      }
    });

    multiple.unshift('[' + single.join('') + ']');

    return new RegExp(multiple.join('|'), 'g')
  }

  const RE_NON_ASCII = /[^\0-\x7F]/g;
  const RE_ASTRAL_SYMBOLS = /[\uD800-\uDBFF][\uDC00-\uDFFF]/g;

  function singleCharReplacer (c) {
    return '&#x' + c.charCodeAt(0).toString(16).toUpperCase() + ';'
  }

  function astralReplacer (c) {
    var high = c.charCodeAt(0);
    var low = c.charCodeAt(1);
    var codePoint = (high - 0xD800) * 0x400 + low - 0xDC00 + 0x10000;
    return '&#x' + codePoint.toString(16).toUpperCase() + ';'
  }

  function getInverse (inverse, re) {
    function func (name) {
      return inverse[name]
    }
    return function (data) {
      return data
        .replace(re, func)
        .replace(RE_ASTRAL_SYMBOLS, astralReplacer)
        .replace(RE_NON_ASCII, singleCharReplacer)
    }
  }

  const booleanAttributes = {
    __proto__: null,
    allowfullscreen: true,
    async: true,
    autofocus: true,
    autoplay: true,
    checked: true,
    controls: true,
    default: true,
    defer: true,
    disabled: true,
    hidden: true,
    ismap: true,
    loop: true,
    multiple: true,
    muted: true,
    open: true,
    readonly: true,
    required: true,
    reversed: true,
    scoped: true,
    seamless: true,
    selected: true,
    typemustmatch: true
  };

  const unencodedElements = {
    __proto__: null,
    style: true,
    script: true,
    xmp: true,
    iframe: true,
    noembed: true,
    noframes: true,
    plaintext: true,
    noscript: true
  };

  const singleTag = {
    __proto__: null,
    area: true,
    base: true,
    basefont: true,
    br: true,
    col: true,
    command: true,
    embed: true,
    frame: true,
    hr: true,
    img: true,
    input: true,
    isindex: true,
    keygen: true,
    link: true,
    meta: true,
    param: true,
    source: true,
    track: true,
    wbr: true
  };

  class DomUtils {
    isTag (elem) {
      return domelementtype.isTag(elem)
    }

    removeElement (elem) {
      if (elem.prev) elem.prev.next = elem.next;
      if (elem.next) elem.next.prev = elem.prev;
      if (elem.parent) {
        var childs = elem.parent.childNodes;
        let pos = childs.lastIndexOf(elem);
        if (pos < 0) throw new Error('Invalid state')
        childs.splice(pos, 1);
        elem.parent = null;
      }
    }

    replaceElement (elem, replacement) {
      if (replacement.parent) this.removeElement(replacement);
      var prev = replacement.prev = elem.prev;
      if (prev) {
        prev.next = replacement;
      }

      var next = replacement.next = elem.next;
      if (next) {
        next.prev = replacement;
      }

      var parent = replacement.parent = elem.parent;
      if (parent) {
        var childs = parent.childNodes;
        let pos = childs.lastIndexOf(elem);
        if (pos < 0) throw new Error('Invalid state')
        childs[pos] = replacement;
      }
    }

    appendChild (elem, child) {
      if (child.parent) this.removeElement(child);
      child.parent = elem;

      if (elem.childNodes.push(child) !== 1) {
        var sibling = elem.childNodes[elem.childNodes.length - 2];
        sibling.next = child;
        child.prev = sibling;
        child.next = null;
      }
    }

    append (elem, next) {
      if (next.parent) this.removeElement(next);
      let parent = elem.parent;
      let currNext = elem.next;

      next.next = currNext;
      next.prev = elem;
      elem.next = next;
      next.parent = parent;

      if (currNext) {
        currNext.prev = next;
        if (parent) {
          var childs = parent.childNodes;
          let pos = childs.lastIndexOf(currNext);
          if (pos < 0) throw new Error('Invalid state')
          childs.splice(pos, 0, next);
        }
      } else if (parent) {
        parent.childNodes.push(next);
      }
    }

    prepend (elem, prev) {
      if (prev.parent) this.removeElement(prev);
      var parent = elem.parent;
      if (parent) {
        var childs = parent.childNodes;
        let pos = childs.lastIndexOf(elem);
        if (pos < 0) throw new Error('Invalid state')
        childs.splice(pos, 0, prev);
      }

      if (elem.prev) {
        elem.prev.next = prev;
      }

      prev.parent = parent;
      prev.prev = elem.prev;
      prev.next = elem;
      elem.prev = prev;
    }

    filter (test, element, recurse, limit) {
      if (!Array.isArray(element)) element = [element];

      if (typeof limit !== 'number' || !isFinite(limit)) {
        limit = Infinity;
      }
      return this.find(test, element, recurse !== false, limit)
    }

    find (test, elems, recurse, limit) {
      let result = [];
      let childs;

      for (var i = 0, j = elems.length; i < j; i++) {
        if (test(elems[i])) {
          result.push(elems[i]);
          if (--limit <= 0) break
        }

        childs = this.getChildren(elems[i]);
        if (recurse && childs && childs.length > 0) {
          childs = this.find(test, childs, recurse, limit);
          result = result.concat(childs);
          limit -= childs.length;
          if (limit <= 0) break
        }
      }

      return result
    }

    findOneChild (test, elems) {
      for (var i = 0, l = elems.length; i < l; i++) {
        if (test(elems[i])) return elems[i]
      }

      return null
    }

    findOne (test, elems) {
      var elem = null;

      for (var i = 0, l = elems.length; i < l && !elem; i++) {
        const child = elems[i];
        if (!this.isTag(child)) {
          continue
        } else if (test(child)) {
          elem = child;
        } else {
          const childNodes = this.getChildren(child);
          if (childNodes.length > 0) {
            elem = this.findOne(test, childNodes);
          }
        }
      }

      return elem
    }

    existsOne (test, elems) {
      for (var i = 0, l = elems.length; i < l; i++) {
        const elem = elems[i];
        
        if (!this.isTag(elem)) continue
        
        if (test(elem)) return true
        
        const childNodes = this.getChildren(elem);
        if (childNodes.length > 0 && this.existsOne(test, childNodes)) return true
      }
      return false
    }

    findAll (test, elems) {
      var result = [];
      for (var i = 0, j = elems.length; i < j; i++) {
        const elem = elems[i];
        if (!this.isTag(elem)) continue
        if (test(elem)) result.push(elem);
        const childNodes = this.getChildren(elem);
        if (childNodes.length > 0) {
          result = result.concat(this.findAll(test, childNodes));
        }
      }
      return result
    }

    getAttributes (el) {
      let attribs = el.getAttributes();
      
      
      if (attribs instanceof Map) {
        return Array.from(attribs)
      } else if (attribs && attribs.forEach) {
        let res = [];
        attribs.forEach((val, key) => {
          res.push([key, val]);
        });
        return res
      } else {
        return []
      }
    }

    formatAttribs (el, opts = {}) {
      let output = [];
      const attributes = this.getAttributes(el);
      attributes.forEach(([key, value]) => {
        if (!value && booleanAttributes[key]) {
          output.push(key);
        } else {
          output.push(key + '="' + (opts.decodeEntities ? _encodeXMLAttr(value) : value) + '"');
        }
      });
      return output.join(' ')
    }

    render (dom, opts) {
      if (!Array.isArray(dom)) dom = [dom];
      opts = opts || {};
      let output = [];
      for (var i = 0; i < dom.length; i++) {
        let elem = dom[i];
        switch (elem.type) {
          case 'root':
          case 'document': {
            output.push(this.render(this.getChildren(elem), opts));
            break
          }
          case domelementtype.Tag:
          case domelementtype.Script:
          case domelementtype.Style: {
            output.push(this.renderTag(elem, opts));
            break
          }
          case domelementtype.CDATA: {
            output.push(this.renderCdata(elem));
            break
          }
          case domelementtype.Comment: {
            output.push(this.renderComment(elem));
            break
          }
          case domelementtype.Directive: {
            output.push(this.renderDirective(elem));
            break
          }
          case domelementtype.Doctype: {
            output.push(this.renderDoctype(elem));
            break
          }
          case domelementtype.Text: {
            output.push(this.renderText(elem, opts));
            break
          }
          default:
            throw new Error('Not implement yet: render of element type ' + elem.type)
        }
      }
      return output.join('')
    }

    renderTag (elem, opts) {
      const name = this.getName(elem);
      if (name === 'svg') opts = {decodeEntities: opts.decodeEntities, xmlMode: true};
      let tag = '<' + name;
      let attribs = this.formatAttribs(elem, opts);
      if (attribs) {
        tag += ' ' + attribs;
      }
      const childNodes = this.getChildren(elem);
      if (opts.xmlMode && childNodes.length === 0) {
        tag += '/>';
      } else {
        tag += '>';
        if (childNodes.length > 0) {
          tag += this.render(childNodes, opts);
        }
        if (!singleTag[name] || opts.xmlMode) {
          tag += '</' + name + '>';
        }
      }
      return tag
    }

    renderDirective (elem) {
      return '<?' + this.getData(elem) + '?>'
    }

    renderDoctype (elem) {
      const { name, publicId, systemId } = this.getData(elem);
      let frags = ['DOCTYPE', name];
      if (publicId) {
        frags.push('PUBLIC');
        frags.push('"' + publicId + '"');
        if (systemId) frags.push('"' + systemId + '"');
      }
      return '<!' + frags.join(' ') + '>'
    }

    renderText (elem, opts) {
      let text = this.getText(elem);
      if (opts.decodeEntities) {
        const parent = this.getParent(elem);
        if (!(parent && this.getName(parent) in unencodedElements)) {
          text = _encodeXMLContent(text);
        }
      }
      return text
    }

    renderCdata (elem) {
      return '<![CDATA[' + this.getData(elem) + ']]>'
    }

    renderComment (elem) {
      return '<!--' + this.getData(elem) + '-->'
    }

    getInnerHTML (elem, opts) {
      const childNodes = this.getChildren(elem);
      return childNodes.map((child) => {
        return this.render(child, opts)
      }).join('')
    }

    getOuterHTML (elem, opts) {
      return this.render(elem, opts)
    }

    getData (elem) {
      return elem.data
    }

    getText (elem, sub) {
      if (Array.isArray(elem)) return elem.map(e => this.getText(e, sub)).join('')
      switch (elem.type) {
        case domelementtype.Tag:
        case domelementtype.Script:
        case domelementtype.Style:
          return this.getText(this.getChildren(elem), true)
        case domelementtype.Text:
        case domelementtype.CDATA:
          return elem.data
        case domelementtype.Comment:
          
          
          if (sub) {
            return ''
          }
          return elem.data
        default:
          return ''
      }
    }

    getChildren (elem) {
      return elem.childNodes
    }

    getParent (elem) {
      return elem.parent
    }

    getSiblings (elem) {
      var parent = this.getParent(elem);
      return parent ? this.getChildren(parent) : [elem]
    }

    getAttributeValue (elem, name) {
      return elem.getAttribute(name)
    }

    hasAttrib (elem, name) {
      return elem.hasAttribute(name)
    }

    getName (elem) {
      return elem.name
    }

    getNameWithoutNS (elem) {
      return elem.nameWithoutNS
    }
  }

  const domUtils = new DomUtils();
  domUtils.DomUtils = DomUtils;

  var universal = 50;
  var tag = 30;
  var attribute = 1;
  var pseudo = 0;
  var descendant = -1;
  var child = -1;
  var parent$1 = -1;
  var sibling = -1;
  var adjacent = -1;
  var procedure = {
  	universal: universal,
  	tag: tag,
  	attribute: attribute,
  	pseudo: pseudo,
  	descendant: descendant,
  	child: child,
  	parent: parent$1,
  	sibling: sibling,
  	adjacent: adjacent
  };

  var procedure$1 = Object.freeze({
    universal: universal,
    tag: tag,
    attribute: attribute,
    pseudo: pseudo,
    descendant: descendant,
    child: child,
    parent: parent$1,
    sibling: sibling,
    adjacent: adjacent,
    default: procedure
  });

  var procedure$2 = ( procedure$1 && procedure ) || procedure$1;

  var sort = sortByProcedure;





  var attributes = {
  	__proto__: null,
  	exists: 10,
  	equals: 8,
  	not: 7,
  	start: 6,
  	end: 6,
  	any: 5,
  	hyphen: 4,
  	element: 4
  };

  function sortByProcedure(arr){
  	var procs = arr.map(getProcedure);
  	for(var i = 1; i < arr.length; i++){
  		var procNew = procs[i];

  		if(procNew < 0) continue;

  		for(var j = i - 1; j >= 0 && procNew < procs[j]; j--){
  			var token = arr[j + 1];
  			arr[j + 1] = arr[j];
  			arr[j] = token;
  			procs[j + 1] = procs[j];
  			procs[j] = procNew;
  		}
  	}
  }

  function getProcedure(token){
  	var proc = procedure$2[token.type];

  	if(proc === procedure$2.attribute){
  		proc = attributes[token.action];

  		if(proc === attributes.equals && token.name === "id"){
  			
  			proc = 9;
  		}

  		if(token.ignoreCase){
  			
  			
  			proc >>= 1;
  		}
  	} else if(proc === procedure$2.pseudo){
  		if(!token.data){
  			proc = 3;
  		} else if(token.name === "has" || token.name === "contains"){
  			proc = 0; 
  		} else if(token.name === "matches" || token.name === "not"){
  			proc = 0;
  			for(var i = 0; i < token.data.length; i++){
  				
  				if(token.data[i].length !== 1) continue;
  				var cur = getProcedure(token.data[i][0]);
  				
  				if(cur === 0){
  					proc = 0;
  					break;
  				}
  				if(cur > proc) proc = cur;
  			}
  			if(token.data.length > 1 && proc > 0) proc -= 1;
  		} else {
  			proc = 1;
  		}
  	}
  	return proc;
  }

  var falseFunc$1 = boolbase.falseFunc;


  var reChars = /[-[\]{}()*+?.,\\^$|#\s]/g;

  function factory(adapter){
  	
  	var attributeRules = {
  		__proto__: null,
  		equals: function(next, data){
  			var name  = data.name,
  				value = data.value;

  			if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function equalsIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && attr.toLowerCase() === value && next(elem);
  				};
  			}

  			return function equals(elem){
  				return adapter.getAttributeValue(elem, name) === value && next(elem);
  			};
  		},
  		hyphen: function(next, data){
  			var name  = data.name,
  				value = data.value,
  				len = value.length;

  			if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function hyphenIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null &&
  							(attr.length === len || attr.charAt(len) === "-") &&
  							attr.substr(0, len).toLowerCase() === value &&
  							next(elem);
  				};
  			}

  			return function hyphen(elem){
  				var attr = adapter.getAttributeValue(elem, name);
  				return attr != null &&
  						attr.substr(0, len) === value &&
  						(attr.length === len || attr.charAt(len) === "-") &&
  						next(elem);
  			};
  		},
  		element: function(next, data){
  			var name = data.name,
  				value = data.value;
  			if (data.name === 'class') {
  				let value = data.value;
  				if (/\s/.test(value)) return function() { return false }
  				return function clazz(elem) {
  					let classes = elem.classes;
  					return classes && classes.has(value) && next(elem)
  				}
  			} else {
  				if(/\s/.test(value)){
  					return falseFunc$1;
  				}

  				value = value.replace(reChars, "\\$&");

  				var pattern = "(?:^|\\s)" + value + "(?:$|\\s)",
  					flags = data.ignoreCase ? "i" : "",
  					regex = new RegExp(pattern, flags);

  				return function element(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && regex.test(attr) && next(elem);
  				};
  			}
  		},
  		exists: function(next, data){
  			var name = data.name;
  			return function exists(elem){
  				return adapter.hasAttrib(elem, name) && next(elem);
  			};
  		},
  		start: function(next, data){
  			var name  = data.name,
  				value = data.value,
  				len = value.length;

  			if(len === 0){
  				return falseFunc$1;
  			}

  			if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function startIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && attr.substr(0, len).toLowerCase() === value && next(elem);
  				};
  			}

  			return function start(elem){
  				var attr = adapter.getAttributeValue(elem, name);
  				return attr != null && attr.substr(0, len) === value && next(elem);
  			};
  		},
  		end: function(next, data){
  			var name  = data.name,
  				value = data.value,
  				len   = -value.length;

  			if(len === 0){
  				return falseFunc$1;
  			}

  			if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function endIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && attr.substr(len).toLowerCase() === value && next(elem);
  				};
  			}

  			return function end(elem){
  				var attr = adapter.getAttributeValue(elem, name);
  				return attr != null && attr.substr(len) === value && next(elem);
  			};
  		},
  		any: function(next, data){
  			var name  = data.name,
  				value = data.value;

  			if(value === ""){
  				return falseFunc$1;
  			}

  			if(data.ignoreCase){
  				var regex = new RegExp(value.replace(reChars, "\\$&"), "i");

  				return function anyIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && regex.test(attr) && next(elem);
  				};
  			}

  			return function any(elem){
  				var attr = adapter.getAttributeValue(elem, name);
  				return attr != null && attr.indexOf(value) >= 0 && next(elem);
  			};
  		},
  		not: function(next, data){
  			var name  = data.name,
  				value = data.value;

  			if(value === ""){
  				return function notEmpty(elem){
  					return !!adapter.getAttributeValue(elem, name) && next(elem);
  				};
  			} else if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function notIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && attr.toLowerCase() !== value && next(elem);
  				};
  			}

  			return function not(elem){
  				return adapter.getAttributeValue(elem, name) !== value && next(elem);
  			};
  		}
  	};

  	return {
  		compile: function(next, data, options){
  			if(options && options.strict && (
  				data.ignoreCase || data.action === "not"
  			)) throw new Error("Unsupported attribute selector");
  			return attributeRules[data.action](next, data);
  		},
  		rules: attributeRules
  	};
  }

  var attributes$1 = factory;

  function generalFactory(adapter, Pseudos){
  	
  	return {
  		__proto__: null,

  		attribute: attributes$1(adapter).compile,
  		pseudo: Pseudos.compile,

  		
  		tag: function(next, data){
  			var name = data.name;
  			return function tag(elem){
  				return adapter.getNameWithoutNS(elem) === name && next(elem);
  			}
  		},

  		
  		descendant: function(next){
  			return function descendant(elem){

  				var found = false;

  				while(!found && (elem = adapter.getParent(elem))){
  					found = next(elem);
  				}

  				return found;
  			};
  		},
  		_flexibleDescendant: function(next){
  			
  			return function descendant(elem){

  				var found = next(elem);

  				while(!found && (elem = adapter.getParent(elem))){
  					found = next(elem);
  				}

  				return found;
  			};
  		},
  		parent: function(next, data, options){
  			if(options && options.strict) throw new Error("Parent selector isn't part of CSS3");

  			return function parent(elem){
  				return adapter.getChildren(elem).some(test);
  			};

  			function test(elem){
  				return adapter.isTag(elem) && next(elem);
  			}
  		},
  		child: function(next){
  			return function child(elem){
  				var parent = adapter.getParent(elem);
  				return !!parent && next(parent);
  			};
  		},
  		sibling: function(next){
  			return function sibling(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var i = 0; i < siblings.length; i++){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) break;
  						if(next(siblings[i])) return true;
  					}
  				}

  				return false;
  			};
  		},
  		adjacent: function(next){
  			return function adjacent(elem){
  				var siblings = adapter.getSiblings(elem),
  					lastElement;

  				for(var i = 0; i < siblings.length; i++){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) break;
  						lastElement = siblings[i];
  					}
  				}

  				return !!lastElement && next(lastElement);
  			};
  		},
  		universal: function(next){
  			return next;
  		}
  	};
  }

  var general = generalFactory;

  var trueFunc$1          = boolbase.trueFunc,
  	falseFunc$1$1         = boolbase.falseFunc;

  function filtersFactory(adapter){
  	var attributes  = attributes$1(adapter),
  		checkAttrib = attributes.rules.equals;

  	
  	function equals(a, b){
  		if(typeof adapter.equals === "function") return adapter.equals(a, b);

  		return a === b;
  	}

  	function getAttribFunc(name, value){
  		var data = {name: name, value: value};
  		return function attribFunc(next){
  			return checkAttrib(next, data);
  		};
  	}

  	function getChildFunc(next){
  		return function(elem){
  			return !!adapter.getParent(elem) && next(elem);
  		};
  	}

  	var filters = {
  		contains: function(next, text){
  			return function contains(elem){
  				return next(elem) && adapter.getText(elem).indexOf(text) >= 0;
  			};
  		},
  		icontains: function(next, text){
  			var itext = text.toLowerCase();
  			return function icontains(elem){
  				return next(elem) &&
  					adapter.getText(elem).toLowerCase().indexOf(itext) >= 0;
  			};
  		},

  		
  		"nth-child": function(next, rule){
  			var func = nthCheck(rule);

  			if(func === falseFunc$1$1) return func;
  			if(func === trueFunc$1)  return getChildFunc(next);

  			return function nthChild(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var i = 0, pos = 0; i < siblings.length; i++){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) break;
  						else pos++;
  					}
  				}

  				return func(pos) && next(elem);
  			};
  		},
  		"nth-last-child": function(next, rule){
  			var func = nthCheck(rule);

  			if(func === falseFunc$1$1) return func;
  			if(func === trueFunc$1)  return getChildFunc(next);

  			return function nthLastChild(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var pos = 0, i = siblings.length - 1; i >= 0; i--){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) break;
  						else pos++;
  					}
  				}

  				return func(pos) && next(elem);
  			};
  		},
  		"nth-of-type": function(next, rule){
  			var func = nthCheck(rule);

  			if(func === falseFunc$1$1) return func;
  			if(func === trueFunc$1)  return getChildFunc(next);

  			return function nthOfType(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var pos = 0, i = 0; i < siblings.length; i++){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) break;
  						if(adapter.getName(siblings[i]) === adapter.getName(elem)) pos++;
  					}
  				}

  				return func(pos) && next(elem);
  			};
  		},
  		"nth-last-of-type": function(next, rule){
  			var func = nthCheck(rule);

  			if(func === falseFunc$1$1) return func;
  			if(func === trueFunc$1)  return getChildFunc(next);

  			return function nthLastOfType(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var pos = 0, i = siblings.length - 1; i >= 0; i--){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) break;
  						if(adapter.getName(siblings[i]) === adapter.getName(elem)) pos++;
  					}
  				}

  				return func(pos) && next(elem);
  			};
  		},

  		
  		root: function(next){
  			return function(elem){
  				return !adapter.getParent(elem) && next(elem);
  			};
  		},

  		scope: function(next, rule, options, context){
  			if(!context || context.length === 0){
  				
  				return filters.root(next);
  			}

  			if(context.length === 1){
  				
  				return function(elem){
  					return equals(context[0], elem) && next(elem);
  				};
  			}

  			return function(elem){
  				return context.indexOf(elem) >= 0 && next(elem);
  			};
  		},

  		
  		checkbox: getAttribFunc("type", "checkbox"),
  		file: getAttribFunc("type", "file"),
  		password: getAttribFunc("type", "password"),
  		radio: getAttribFunc("type", "radio"),
  		reset: getAttribFunc("type", "reset"),
  		image: getAttribFunc("type", "image"),
  		submit: getAttribFunc("type", "submit")
  	};
  	return filters;
  }

  function pseudosFactory(adapter){
  	
  	function getFirstElement(elems){
  		for(var i = 0; elems && i < elems.length; i++){
  			if(adapter.isTag(elems[i])) return elems[i];
  		}
  	}

  	
  	var pseudos = {
  		empty: function(elem){
  			return !adapter.getChildren(elem).some(function(elem){
  				return adapter.isTag(elem) || elem.type === "text";
  			});
  		},

  		"first-child": function(elem){
  			return getFirstElement(adapter.getSiblings(elem)) === elem;
  		},
  		"last-child": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = siblings.length - 1; i >= 0; i--){
  				if(siblings[i] === elem) return true;
  				if(adapter.isTag(siblings[i])) break;
  			}

  			return false;
  		},
  		"first-of-type": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = 0; i < siblings.length; i++){
  				if(adapter.isTag(siblings[i])){
  					if(siblings[i] === elem) return true;
  					if(adapter.getName(siblings[i]) === adapter.getName(elem)) break;
  				}
  			}

  			return false;
  		},
  		"last-of-type": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = siblings.length - 1; i >= 0; i--){
  				if(adapter.isTag(siblings[i])){
  					if(siblings[i] === elem) return true;
  					if(adapter.getName(siblings[i]) === adapter.getName(elem)) break;
  				}
  			}

  			return false;
  		},
  		"only-of-type": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = 0, j = siblings.length; i < j; i++){
  				if(adapter.isTag(siblings[i])){
  					if(siblings[i] === elem) continue;
  					if(adapter.getName(siblings[i]) === adapter.getName(elem)) return false;
  				}
  			}

  			return true;
  		},
  		"only-child": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = 0; i < siblings.length; i++){
  				if(adapter.isTag(siblings[i]) && siblings[i] !== elem) return false;
  			}

  			return true;
  		},

  		
  		link: function(elem){
  			return adapter.hasAttrib(elem, "href");
  		},
  		visited: falseFunc$1$1, 
  		

  		
  		

  		
  		selected: function(elem){
  			if(adapter.hasAttrib(elem, "selected")) return true;
  			else if(adapter.getName(elem) !== "option") return false;

  			
  			var parent = adapter.getParent(elem);

  			if(
  				!parent ||
  				adapter.getName(parent) !== "select" ||
  				adapter.hasAttrib(parent, "multiple")
  			) return false;

  			var siblings = adapter.getChildren(parent),
  				sawElem  = false;

  			for(var i = 0; i < siblings.length; i++){
  				if(adapter.isTag(siblings[i])){
  					if(siblings[i] === elem){
  						sawElem = true;
  					} else if(!sawElem){
  						return false;
  					} else if(adapter.hasAttrib(siblings[i], "selected")){
  						return false;
  					}
  				}
  			}

  			return sawElem;
  		},
  		
  		
  		
  		
  		
  		
  		disabled: function(elem){
  			return adapter.hasAttrib(elem, "disabled");
  		},
  		enabled: function(elem){
  			return !adapter.hasAttrib(elem, "disabled");
  		},
  		
  		checked: function(elem){
  			return adapter.hasAttrib(elem, "checked") || pseudos.selected(elem);
  		},
  		
  		required: function(elem){
  			return adapter.hasAttrib(elem, "required");
  		},
  		
  		optional: function(elem){
  			return !adapter.hasAttrib(elem, "required");
  		},

  		

  		
  		parent: function(elem){
  			return !pseudos.empty(elem);
  		},
  		
  		header: function(elem){
  			var name = adapter.getName(elem);
  			return name === "h1" ||
  					name === "h2" ||
  					name === "h3" ||
  					name === "h4" ||
  					name === "h5" ||
  					name === "h6";
  		},

  		
  		button: function(elem){
  			var name = adapter.getName(elem);
  			return name === "button" ||
  					name === "input" &&
  					adapter.getAttributeValue(elem, "type") === "button";
  		},
  		
  		input: function(elem){
  			var name = adapter.getName(elem);
  			return name === "input" ||
  					name === "textarea" ||
  					name === "select" ||
  					name === "button";
  		},
  		
  		text: function(elem){
  			var attr;
  			return adapter.getName(elem) === "input" && (
  				!(attr = adapter.getAttributeValue(elem, "type")) ||
  				attr.toLowerCase() === "text"
  			);
  		}
  	};

  	return pseudos;
  }

  function verifyArgs(func, name, subselect){
  	if(subselect === null){
  		if(func.length > 1 && name !== "scope"){
  			throw new Error("pseudo-selector :" + name + " requires an argument");
  		}
  	} else {
  		if(func.length === 1){
  			throw new Error("pseudo-selector :" + name + " doesn't have any arguments");
  		}
  	}
  }


  var re_CSS3 = /^(?:(?:nth|last|first|only)-(?:child|of-type)|root|empty|(?:en|dis)abled|checked|not)$/;

  function factory$1(adapter){
  	var pseudos = pseudosFactory(adapter);
  	var filters = filtersFactory(adapter);

  	return {
  		compile: function(next, data, options, context){
  			var name = data.name,
  				subselect = data.data;

  			if(options && options.strict && !re_CSS3.test(name)){
  				throw new Error(":" + name + " isn't part of CSS3");
  			}

  			if(typeof filters[name] === "function"){
  				verifyArgs(filters[name], name,  subselect);
  				return filters[name](next, subselect, options, context);
  			} else if(typeof pseudos[name] === "function"){
  				var func = pseudos[name];
  				verifyArgs(func, name, subselect);

  				if(next === trueFunc$1) return func;

  				return function pseudoArgs(elem){
  					return func(elem, subselect) && next(elem);
  				};
  			} else {
  				throw new Error("unmatched pseudo-class :" + name);
  			}
  		},
  		filters: filters,
  		pseudos: pseudos
  	};
  }

  var pseudos = factory$1;

  var compile$1 = compileFactory;

  var trueFunc$1$1       = boolbase.trueFunc,
  	falseFunc$2      = boolbase.falseFunc;

  function compileFactory(adapter){
  	var Pseudos     = pseudos(adapter),
  		filters     = Pseudos.filters,
  		Rules 			= general(adapter, Pseudos);

  	function compile(selector, options, context){
  		var next = compileUnsafe(selector, options, context);
  		return wrap(next);
  	}

  	function wrap(next){
  		return function base(elem){
  			return adapter.isTag(elem) && next(elem);
  		};
  	}

  	function compileUnsafe(selector, options, context){
  		var token = cssWhat(selector, options);
  		return compileToken(token, options, context);
  	}

  	function includesScopePseudo(t){
  		return t.type === "pseudo" && (
  			t.name === "scope" || (
  				Array.isArray(t.data) &&
  				t.data.some(function(data){
  					return data.some(includesScopePseudo);
  				})
  			)
  		);
  	}

  	var DESCENDANT_TOKEN = {type: "descendant"},
  		FLEXIBLE_DESCENDANT_TOKEN = {type: "_flexibleDescendant"},
  		SCOPE_TOKEN = {type: "pseudo", name: "scope"},
  		PLACEHOLDER_ELEMENT = {};

  	
  	
  	function absolutize(token, context){
  		
  		var hasContext = !!context && !!context.length && context.every(function(e){
  			return e === PLACEHOLDER_ELEMENT || !!adapter.getParent(e);
  		});


  		token.forEach(function(t){
  			if(t.length > 0 && isTraversal(t[0]) && t[0].type !== "descendant"); else if(hasContext && !includesScopePseudo(t)){
  				t.unshift(DESCENDANT_TOKEN);
  			} else {
  				return;
  			}

  			t.unshift(SCOPE_TOKEN);
  		});
  	}

  	function compileToken(token, options, context){
  		token = token.filter(function(t){ return t.length > 0; });

  		token.forEach(sort);

  		var isArrayContext = Array.isArray(context);

  		context = (options && options.context) || context;

  		if(context && !isArrayContext) context = [context];

  		absolutize(token, context);

  		var shouldTestNextSiblings = false;

  		var query = token
  			.map(function(rules){
  				if(rules[0] && rules[1] && rules[0].name === "scope"){
  					var ruleType = rules[1].type;
  					if(isArrayContext && ruleType === "descendant") rules[1] = FLEXIBLE_DESCENDANT_TOKEN;
  					else if(ruleType === "adjacent" || ruleType === "sibling") shouldTestNextSiblings = true;
  				}
  				return compileRules(rules, options, context);
  			})
  			.reduce(reduceRules, falseFunc$2);

  		query.shouldTestNextSiblings = shouldTestNextSiblings;

  		return query;
  	}

  	function isTraversal(t){
  		return procedure$2[t.type] < 0;
  	}

  	function compileRules(rules, options, context){
  		return rules.reduce(function(func, rule){
  			if(func === falseFunc$2) return func;
  			return Rules[rule.type](func, rule, options, context);
  		}, options && options.rootFunc || trueFunc$1$1);
  	}

  	function reduceRules(a, b){
  		if(b === falseFunc$2 || a === trueFunc$1$1){
  			return a;
  		}
  		if(a === falseFunc$2 || b === trueFunc$1$1){
  			return b;
  		}

  		return function combine(elem){
  			return a(elem) || b(elem);
  		};
  	}

  	function containsTraversal(t){
  		return t.some(isTraversal);
  	}

  	
  	
  	
  	filters.not = function(next, token, options, context){
  		var opts = {
  			xmlMode: !!(options && options.xmlMode),
  			strict: !!(options && options.strict)
  		};

  		if(opts.strict){
  			if(token.length > 1 || token.some(containsTraversal)){
  				throw new Error("complex selectors in :not aren't allowed in strict mode");
  			}
  		}

  		var func = compileToken(token, opts, context);

  		if(func === falseFunc$2) return next;
  		if(func === trueFunc$1$1)  return falseFunc$2;

  		return function(elem){
  			return !func(elem) && next(elem);
  		};
  	};

  	filters.has = function(next, token, options){
  		var opts = {
  			xmlMode: !!(options && options.xmlMode),
  			strict: !!(options && options.strict)
  		};

  		
  		var context = token.some(containsTraversal) ? [PLACEHOLDER_ELEMENT] : null;

  		var func = compileToken(token, opts, context);

  		if(func === falseFunc$2) return falseFunc$2;
  		if(func === trueFunc$1$1){
  			return function(elem){
  				return adapter.getChildren(elem).some(adapter.isTag) && next(elem);
  			};
  		}

  		func = wrap(func);

  		if(context){
  			return function has(elem){
  				return next(elem) && (
  					(context[0] = elem), adapter.existsOne(func, adapter.getChildren(elem))
  				);
  			};
  		}

  		return function has(elem){
  			return next(elem) && adapter.existsOne(func, adapter.getChildren(elem));
  		};
  	};

  	filters.matches = function(next, token, options, context){
  		var opts = {
  			xmlMode: !!(options && options.xmlMode),
  			strict: !!(options && options.strict),
  			rootFunc: next
  		};

  		return compileToken(token, opts, context);
  	};

  	compile.compileToken = compileToken;
  	compile.compileUnsafe = compileUnsafe;
  	compile.Pseudos = Pseudos;

  	return compile;
  }

  var cssSelect = CSSselect;

  var falseFunc$3      = boolbase.falseFunc,
  	defaultCompile = compile$1(domUtils);

  function adapterCompile(adapter){
  	if(!adapter.__compile__){
  		adapter.__compile__ = compile$1(adapter);
  	}
  	return adapter.__compile__
  }

  function getSelectorFunc(searchFunc){
  	return function select(query, elems, options){
  		options = options || {};
  		options.adapter = options.adapter || domUtils;
  		var compile$$1 = adapterCompile(options.adapter);

  		if(typeof query !== "function") query = compile$$1.compileUnsafe(query, options, elems);
  		if(query.shouldTestNextSiblings) elems = appendNextSiblings((options && options.context) || elems, options.adapter);
  		if(!Array.isArray(elems)) elems = options.adapter.getChildren(elems);
  		else elems = options.adapter.removeSubsets(elems);
  		return searchFunc(query, elems, options);
  	};
  }

  function getNextSiblings(elem, adapter){
  	var siblings = adapter.getSiblings(elem);
  	if(!Array.isArray(siblings)) return [];
  	siblings = siblings.slice(0);
  	while(siblings.shift() !== elem);
  	return siblings;
  }

  function appendNextSiblings(elems, adapter){
  	
  	if(!Array.isArray(elems)) elems = [elems];
  	var newElems = elems.slice(0);

  	for(var i = 0, len = elems.length; i < len; i++){
  		var nextSiblings = getNextSiblings(newElems[i], adapter);
  		newElems.push.apply(newElems, nextSiblings);
  	}
  	return newElems;
  }

  var selectAll = getSelectorFunc(function selectAll(query, elems, options){
  	return (query === falseFunc$3 || !elems || elems.length === 0) ? [] : options.adapter.findAll(query, elems);
  });

  var selectOne = getSelectorFunc(function selectOne(query, elems, options){
  	return (query === falseFunc$3 || !elems || elems.length === 0) ? null : options.adapter.findOne(query, elems);
  });

  function is(elem, query, options){
  	options = options || {};
  	options.adapter = options.adapter || domUtils;
  	var compile$$1 = adapterCompile(options.adapter);
  	return (typeof query === "function" ? query : compile$$1(query, options))(elem);
  }


  function CSSselect(query, elems, options){
  	return selectAll(query, elems, options);
  }

  CSSselect.compile = defaultCompile;
  CSSselect.filters = defaultCompile.Pseudos.filters;
  CSSselect.pseudos = defaultCompile.Pseudos.pseudos;

  CSSselect.selectAll = selectAll;
  CSSselect.selectOne = selectOne;

  CSSselect.is = is;


  CSSselect.parse = defaultCompile;
  CSSselect.iterate = selectAll;


  CSSselect._compileUnsafe = defaultCompile.compileUnsafe;
  CSSselect._compileToken = defaultCompile.compileToken;

  var amp = "&";
  var apos = "'";
  var gt = ">";
  var lt = "<";
  var quot = "\"";
  var xmlJSON = {
  	amp: amp,
  	apos: apos,
  	gt: gt,
  	lt: lt,
  	quot: quot
  };

  var xml = Object.freeze({
  	amp: amp,
  	apos: apos,
  	gt: gt,
  	lt: lt,
  	quot: quot,
  	default: xmlJSON
  });

  var Aacute = "Á";
  var aacute = "á";
  var Abreve = "Ă";
  var abreve = "ă";
  var ac = "∾";
  var acd = "∿";
  var acE = "∾̳";
  var Acirc = "Â";
  var acirc = "â";
  var acute = "´";
  var Acy = "А";
  var acy = "а";
  var AElig = "Æ";
  var aelig = "æ";
  var af = "⁡";
  var Afr = "𝔄";
  var afr = "𝔞";
  var Agrave = "À";
  var agrave = "à";
  var alefsym = "ℵ";
  var aleph = "ℵ";
  var Alpha = "Α";
  var alpha = "α";
  var Amacr = "Ā";
  var amacr = "ā";
  var amalg = "⨿";
  var amp$1 = "&";
  var AMP = "&";
  var andand = "⩕";
  var And = "⩓";
  var and = "∧";
  var andd = "⩜";
  var andslope = "⩘";
  var andv = "⩚";
  var ang = "∠";
  var ange = "⦤";
  var angle = "∠";
  var angmsdaa = "⦨";
  var angmsdab = "⦩";
  var angmsdac = "⦪";
  var angmsdad = "⦫";
  var angmsdae = "⦬";
  var angmsdaf = "⦭";
  var angmsdag = "⦮";
  var angmsdah = "⦯";
  var angmsd = "∡";
  var angrt = "∟";
  var angrtvb = "⊾";
  var angrtvbd = "⦝";
  var angsph = "∢";
  var angst = "Å";
  var angzarr = "⍼";
  var Aogon = "Ą";
  var aogon = "ą";
  var Aopf = "𝔸";
  var aopf = "𝕒";
  var apacir = "⩯";
  var ap = "≈";
  var apE = "⩰";
  var ape = "≊";
  var apid = "≋";
  var apos$1 = "'";
  var ApplyFunction = "⁡";
  var approx = "≈";
  var approxeq = "≊";
  var Aring = "Å";
  var aring = "å";
  var Ascr = "𝒜";
  var ascr = "𝒶";
  var Assign = "≔";
  var ast = "*";
  var asymp = "≈";
  var asympeq = "≍";
  var Atilde = "Ã";
  var atilde = "ã";
  var Auml = "Ä";
  var auml = "ä";
  var awconint = "∳";
  var awint = "⨑";
  var backcong = "≌";
  var backepsilon = "϶";
  var backprime = "‵";
  var backsim = "∽";
  var backsimeq = "⋍";
  var Backslash = "∖";
  var Barv = "⫧";
  var barvee = "⊽";
  var barwed = "⌅";
  var Barwed = "⌆";
  var barwedge = "⌅";
  var bbrk = "⎵";
  var bbrktbrk = "⎶";
  var bcong = "≌";
  var Bcy = "Б";
  var bcy = "б";
  var bdquo = "„";
  var becaus = "∵";
  var because = "∵";
  var Because = "∵";
  var bemptyv = "⦰";
  var bepsi = "϶";
  var bernou = "ℬ";
  var Bernoullis = "ℬ";
  var Beta = "Β";
  var beta = "β";
  var beth = "ℶ";
  var between = "≬";
  var Bfr = "𝔅";
  var bfr = "𝔟";
  var bigcap = "⋂";
  var bigcirc = "◯";
  var bigcup = "⋃";
  var bigodot = "⨀";
  var bigoplus = "⨁";
  var bigotimes = "⨂";
  var bigsqcup = "⨆";
  var bigstar = "★";
  var bigtriangledown = "▽";
  var bigtriangleup = "△";
  var biguplus = "⨄";
  var bigvee = "⋁";
  var bigwedge = "⋀";
  var bkarow = "⤍";
  var blacklozenge = "⧫";
  var blacksquare = "▪";
  var blacktriangle = "▴";
  var blacktriangledown = "▾";
  var blacktriangleleft = "◂";
  var blacktriangleright = "▸";
  var blank = "␣";
  var blk12 = "▒";
  var blk14 = "░";
  var blk34 = "▓";
  var block = "█";
  var bne = "=⃥";
  var bnequiv = "≡⃥";
  var bNot = "⫭";
  var bnot = "⌐";
  var Bopf = "𝔹";
  var bopf = "𝕓";
  var bot = "⊥";
  var bottom = "⊥";
  var bowtie = "⋈";
  var boxbox = "⧉";
  var boxdl = "┐";
  var boxdL = "╕";
  var boxDl = "╖";
  var boxDL = "╗";
  var boxdr = "┌";
  var boxdR = "╒";
  var boxDr = "╓";
  var boxDR = "╔";
  var boxh = "─";
  var boxH = "═";
  var boxhd = "┬";
  var boxHd = "╤";
  var boxhD = "╥";
  var boxHD = "╦";
  var boxhu = "┴";
  var boxHu = "╧";
  var boxhU = "╨";
  var boxHU = "╩";
  var boxminus = "⊟";
  var boxplus = "⊞";
  var boxtimes = "⊠";
  var boxul = "┘";
  var boxuL = "╛";
  var boxUl = "╜";
  var boxUL = "╝";
  var boxur = "└";
  var boxuR = "╘";
  var boxUr = "╙";
  var boxUR = "╚";
  var boxv = "│";
  var boxV = "║";
  var boxvh = "┼";
  var boxvH = "╪";
  var boxVh = "╫";
  var boxVH = "╬";
  var boxvl = "┤";
  var boxvL = "╡";
  var boxVl = "╢";
  var boxVL = "╣";
  var boxvr = "├";
  var boxvR = "╞";
  var boxVr = "╟";
  var boxVR = "╠";
  var bprime = "‵";
  var breve = "˘";
  var Breve = "˘";
  var brvbar = "¦";
  var bscr = "𝒷";
  var Bscr = "ℬ";
  var bsemi = "⁏";
  var bsim = "∽";
  var bsime = "⋍";
  var bsolb = "⧅";
  var bsol = "\\";
  var bsolhsub = "⟈";
  var bull = "•";
  var bullet = "•";
  var bump = "≎";
  var bumpE = "⪮";
  var bumpe = "≏";
  var Bumpeq = "≎";
  var bumpeq = "≏";
  var Cacute = "Ć";
  var cacute = "ć";
  var capand = "⩄";
  var capbrcup = "⩉";
  var capcap = "⩋";
  var cap = "∩";
  var Cap = "⋒";
  var capcup = "⩇";
  var capdot = "⩀";
  var CapitalDifferentialD = "ⅅ";
  var caps = "∩︀";
  var caret = "⁁";
  var caron = "ˇ";
  var Cayleys = "ℭ";
  var ccaps = "⩍";
  var Ccaron = "Č";
  var ccaron = "č";
  var Ccedil = "Ç";
  var ccedil = "ç";
  var Ccirc = "Ĉ";
  var ccirc = "ĉ";
  var Cconint = "∰";
  var ccups = "⩌";
  var ccupssm = "⩐";
  var Cdot = "Ċ";
  var cdot = "ċ";
  var cedil = "¸";
  var Cedilla = "¸";
  var cemptyv = "⦲";
  var cent = "¢";
  var centerdot = "·";
  var CenterDot = "·";
  var cfr = "𝔠";
  var Cfr = "ℭ";
  var CHcy = "Ч";
  var chcy = "ч";
  var check = "✓";
  var checkmark = "✓";
  var Chi = "Χ";
  var chi = "χ";
  var circ = "ˆ";
  var circeq = "≗";
  var circlearrowleft = "↺";
  var circlearrowright = "↻";
  var circledast = "⊛";
  var circledcirc = "⊚";
  var circleddash = "⊝";
  var CircleDot = "⊙";
  var circledR = "®";
  var circledS = "Ⓢ";
  var CircleMinus = "⊖";
  var CirclePlus = "⊕";
  var CircleTimes = "⊗";
  var cir = "○";
  var cirE = "⧃";
  var cire = "≗";
  var cirfnint = "⨐";
  var cirmid = "⫯";
  var cirscir = "⧂";
  var ClockwiseContourIntegral = "∲";
  var CloseCurlyDoubleQuote = "”";
  var CloseCurlyQuote = "’";
  var clubs = "♣";
  var clubsuit = "♣";
  var colon = ":";
  var Colon = "∷";
  var Colone = "⩴";
  var colone = "≔";
  var coloneq = "≔";
  var comma = ",";
  var commat = "@";
  var comp = "∁";
  var compfn = "∘";
  var complement = "∁";
  var complexes = "ℂ";
  var cong = "≅";
  var congdot = "⩭";
  var Congruent = "≡";
  var conint = "∮";
  var Conint = "∯";
  var ContourIntegral = "∮";
  var copf = "𝕔";
  var Copf = "ℂ";
  var coprod = "∐";
  var Coproduct = "∐";
  var copy = "©";
  var COPY = "©";
  var copysr = "℗";
  var CounterClockwiseContourIntegral = "∳";
  var crarr = "↵";
  var cross = "✗";
  var Cross = "⨯";
  var Cscr = "𝒞";
  var cscr = "𝒸";
  var csub = "⫏";
  var csube = "⫑";
  var csup = "⫐";
  var csupe = "⫒";
  var ctdot = "⋯";
  var cudarrl = "⤸";
  var cudarrr = "⤵";
  var cuepr = "⋞";
  var cuesc = "⋟";
  var cularr = "↶";
  var cularrp = "⤽";
  var cupbrcap = "⩈";
  var cupcap = "⩆";
  var CupCap = "≍";
  var cup = "∪";
  var Cup = "⋓";
  var cupcup = "⩊";
  var cupdot = "⊍";
  var cupor = "⩅";
  var cups = "∪︀";
  var curarr = "↷";
  var curarrm = "⤼";
  var curlyeqprec = "⋞";
  var curlyeqsucc = "⋟";
  var curlyvee = "⋎";
  var curlywedge = "⋏";
  var curren = "¤";
  var curvearrowleft = "↶";
  var curvearrowright = "↷";
  var cuvee = "⋎";
  var cuwed = "⋏";
  var cwconint = "∲";
  var cwint = "∱";
  var cylcty = "⌭";
  var dagger = "†";
  var Dagger = "‡";
  var daleth = "ℸ";
  var darr = "↓";
  var Darr = "↡";
  var dArr = "⇓";
  var dash = "‐";
  var Dashv = "⫤";
  var dashv = "⊣";
  var dbkarow = "⤏";
  var dblac = "˝";
  var Dcaron = "Ď";
  var dcaron = "ď";
  var Dcy = "Д";
  var dcy = "д";
  var ddagger = "‡";
  var ddarr = "⇊";
  var DD = "ⅅ";
  var dd = "ⅆ";
  var DDotrahd = "⤑";
  var ddotseq = "⩷";
  var deg = "°";
  var Del = "∇";
  var Delta = "Δ";
  var delta = "δ";
  var demptyv = "⦱";
  var dfisht = "⥿";
  var Dfr = "𝔇";
  var dfr = "𝔡";
  var dHar = "⥥";
  var dharl = "⇃";
  var dharr = "⇂";
  var DiacriticalAcute = "´";
  var DiacriticalDot = "˙";
  var DiacriticalDoubleAcute = "˝";
  var DiacriticalGrave = "`";
  var DiacriticalTilde = "˜";
  var diam = "⋄";
  var diamond = "⋄";
  var Diamond = "⋄";
  var diamondsuit = "♦";
  var diams = "♦";
  var die = "¨";
  var DifferentialD = "ⅆ";
  var digamma = "ϝ";
  var disin = "⋲";
  var div = "÷";
  var divide = "÷";
  var divideontimes = "⋇";
  var divonx = "⋇";
  var DJcy = "Ђ";
  var djcy = "ђ";
  var dlcorn = "⌞";
  var dlcrop = "⌍";
  var dollar = "$";
  var Dopf = "𝔻";
  var dopf = "𝕕";
  var Dot = "¨";
  var dot = "˙";
  var DotDot = "⃜";
  var doteq = "≐";
  var doteqdot = "≑";
  var DotEqual = "≐";
  var dotminus = "∸";
  var dotplus = "∔";
  var dotsquare = "⊡";
  var doublebarwedge = "⌆";
  var DoubleContourIntegral = "∯";
  var DoubleDot = "¨";
  var DoubleDownArrow = "⇓";
  var DoubleLeftArrow = "⇐";
  var DoubleLeftRightArrow = "⇔";
  var DoubleLeftTee = "⫤";
  var DoubleLongLeftArrow = "⟸";
  var DoubleLongLeftRightArrow = "⟺";
  var DoubleLongRightArrow = "⟹";
  var DoubleRightArrow = "⇒";
  var DoubleRightTee = "⊨";
  var DoubleUpArrow = "⇑";
  var DoubleUpDownArrow = "⇕";
  var DoubleVerticalBar = "∥";
  var DownArrowBar = "⤓";
  var downarrow = "↓";
  var DownArrow = "↓";
  var Downarrow = "⇓";
  var DownArrowUpArrow = "⇵";
  var DownBreve = "̑";
  var downdownarrows = "⇊";
  var downharpoonleft = "⇃";
  var downharpoonright = "⇂";
  var DownLeftRightVector = "⥐";
  var DownLeftTeeVector = "⥞";
  var DownLeftVectorBar = "⥖";
  var DownLeftVector = "↽";
  var DownRightTeeVector = "⥟";
  var DownRightVectorBar = "⥗";
  var DownRightVector = "⇁";
  var DownTeeArrow = "↧";
  var DownTee = "⊤";
  var drbkarow = "⤐";
  var drcorn = "⌟";
  var drcrop = "⌌";
  var Dscr = "𝒟";
  var dscr = "𝒹";
  var DScy = "Ѕ";
  var dscy = "ѕ";
  var dsol = "⧶";
  var Dstrok = "Đ";
  var dstrok = "đ";
  var dtdot = "⋱";
  var dtri = "▿";
  var dtrif = "▾";
  var duarr = "⇵";
  var duhar = "⥯";
  var dwangle = "⦦";
  var DZcy = "Џ";
  var dzcy = "џ";
  var dzigrarr = "⟿";
  var Eacute = "É";
  var eacute = "é";
  var easter = "⩮";
  var Ecaron = "Ě";
  var ecaron = "ě";
  var Ecirc = "Ê";
  var ecirc = "ê";
  var ecir = "≖";
  var ecolon = "≕";
  var Ecy = "Э";
  var ecy = "э";
  var eDDot = "⩷";
  var Edot = "Ė";
  var edot = "ė";
  var eDot = "≑";
  var ee = "ⅇ";
  var efDot = "≒";
  var Efr = "𝔈";
  var efr = "𝔢";
  var eg = "⪚";
  var Egrave = "È";
  var egrave = "è";
  var egs = "⪖";
  var egsdot = "⪘";
  var el = "⪙";
  var Element = "∈";
  var elinters = "⏧";
  var ell = "ℓ";
  var els = "⪕";
  var elsdot = "⪗";
  var Emacr = "Ē";
  var emacr = "ē";
  var empty = "∅";
  var emptyset = "∅";
  var EmptySmallSquare = "◻";
  var emptyv = "∅";
  var EmptyVerySmallSquare = "▫";
  var emsp13 = " ";
  var emsp14 = " ";
  var emsp = " ";
  var ENG = "Ŋ";
  var eng = "ŋ";
  var ensp = " ";
  var Eogon = "Ę";
  var eogon = "ę";
  var Eopf = "𝔼";
  var eopf = "𝕖";
  var epar = "⋕";
  var eparsl = "⧣";
  var eplus = "⩱";
  var epsi = "ε";
  var Epsilon = "Ε";
  var epsilon = "ε";
  var epsiv = "ϵ";
  var eqcirc = "≖";
  var eqcolon = "≕";
  var eqsim = "≂";
  var eqslantgtr = "⪖";
  var eqslantless = "⪕";
  var Equal = "⩵";
  var equals = "=";
  var EqualTilde = "≂";
  var equest = "≟";
  var Equilibrium = "⇌";
  var equiv = "≡";
  var equivDD = "⩸";
  var eqvparsl = "⧥";
  var erarr = "⥱";
  var erDot = "≓";
  var escr = "ℯ";
  var Escr = "ℰ";
  var esdot = "≐";
  var Esim = "⩳";
  var esim = "≂";
  var Eta = "Η";
  var eta = "η";
  var ETH = "Ð";
  var eth = "ð";
  var Euml = "Ë";
  var euml = "ë";
  var euro = "€";
  var excl = "!";
  var exist = "∃";
  var Exists = "∃";
  var expectation = "ℰ";
  var exponentiale = "ⅇ";
  var ExponentialE = "ⅇ";
  var fallingdotseq = "≒";
  var Fcy = "Ф";
  var fcy = "ф";
  var female = "♀";
  var ffilig = "ﬃ";
  var fflig = "ﬀ";
  var ffllig = "ﬄ";
  var Ffr = "𝔉";
  var ffr = "𝔣";
  var filig = "ﬁ";
  var FilledSmallSquare = "◼";
  var FilledVerySmallSquare = "▪";
  var fjlig = "fj";
  var flat = "♭";
  var fllig = "ﬂ";
  var fltns = "▱";
  var fnof = "ƒ";
  var Fopf = "𝔽";
  var fopf = "𝕗";
  var forall = "∀";
  var ForAll = "∀";
  var fork = "⋔";
  var forkv = "⫙";
  var Fouriertrf = "ℱ";
  var fpartint = "⨍";
  var frac12 = "½";
  var frac13 = "⅓";
  var frac14 = "¼";
  var frac15 = "⅕";
  var frac16 = "⅙";
  var frac18 = "⅛";
  var frac23 = "⅔";
  var frac25 = "⅖";
  var frac34 = "¾";
  var frac35 = "⅗";
  var frac38 = "⅜";
  var frac45 = "⅘";
  var frac56 = "⅚";
  var frac58 = "⅝";
  var frac78 = "⅞";
  var frasl = "⁄";
  var frown = "⌢";
  var fscr = "𝒻";
  var Fscr = "ℱ";
  var gacute = "ǵ";
  var Gamma = "Γ";
  var gamma = "γ";
  var Gammad = "Ϝ";
  var gammad = "ϝ";
  var gap = "⪆";
  var Gbreve = "Ğ";
  var gbreve = "ğ";
  var Gcedil = "Ģ";
  var Gcirc = "Ĝ";
  var gcirc = "ĝ";
  var Gcy = "Г";
  var gcy = "г";
  var Gdot = "Ġ";
  var gdot = "ġ";
  var ge = "≥";
  var gE = "≧";
  var gEl = "⪌";
  var gel = "⋛";
  var geq = "≥";
  var geqq = "≧";
  var geqslant = "⩾";
  var gescc = "⪩";
  var ges = "⩾";
  var gesdot = "⪀";
  var gesdoto = "⪂";
  var gesdotol = "⪄";
  var gesl = "⋛︀";
  var gesles = "⪔";
  var Gfr = "𝔊";
  var gfr = "𝔤";
  var gg = "≫";
  var Gg = "⋙";
  var ggg = "⋙";
  var gimel = "ℷ";
  var GJcy = "Ѓ";
  var gjcy = "ѓ";
  var gla = "⪥";
  var gl = "≷";
  var glE = "⪒";
  var glj = "⪤";
  var gnap = "⪊";
  var gnapprox = "⪊";
  var gne = "⪈";
  var gnE = "≩";
  var gneq = "⪈";
  var gneqq = "≩";
  var gnsim = "⋧";
  var Gopf = "𝔾";
  var gopf = "𝕘";
  var grave = "`";
  var GreaterEqual = "≥";
  var GreaterEqualLess = "⋛";
  var GreaterFullEqual = "≧";
  var GreaterGreater = "⪢";
  var GreaterLess = "≷";
  var GreaterSlantEqual = "⩾";
  var GreaterTilde = "≳";
  var Gscr = "𝒢";
  var gscr = "ℊ";
  var gsim = "≳";
  var gsime = "⪎";
  var gsiml = "⪐";
  var gtcc = "⪧";
  var gtcir = "⩺";
  var gt$1 = ">";
  var GT = ">";
  var Gt = "≫";
  var gtdot = "⋗";
  var gtlPar = "⦕";
  var gtquest = "⩼";
  var gtrapprox = "⪆";
  var gtrarr = "⥸";
  var gtrdot = "⋗";
  var gtreqless = "⋛";
  var gtreqqless = "⪌";
  var gtrless = "≷";
  var gtrsim = "≳";
  var gvertneqq = "≩︀";
  var gvnE = "≩︀";
  var Hacek = "ˇ";
  var hairsp = " ";
  var half = "½";
  var hamilt = "ℋ";
  var HARDcy = "Ъ";
  var hardcy = "ъ";
  var harrcir = "⥈";
  var harr = "↔";
  var hArr = "⇔";
  var harrw = "↭";
  var Hat = "^";
  var hbar = "ℏ";
  var Hcirc = "Ĥ";
  var hcirc = "ĥ";
  var hearts = "♥";
  var heartsuit = "♥";
  var hellip = "…";
  var hercon = "⊹";
  var hfr = "𝔥";
  var Hfr = "ℌ";
  var HilbertSpace = "ℋ";
  var hksearow = "⤥";
  var hkswarow = "⤦";
  var hoarr = "⇿";
  var homtht = "∻";
  var hookleftarrow = "↩";
  var hookrightarrow = "↪";
  var hopf = "𝕙";
  var Hopf = "ℍ";
  var horbar = "―";
  var HorizontalLine = "─";
  var hscr = "𝒽";
  var Hscr = "ℋ";
  var hslash = "ℏ";
  var Hstrok = "Ħ";
  var hstrok = "ħ";
  var HumpDownHump = "≎";
  var HumpEqual = "≏";
  var hybull = "⁃";
  var hyphen = "‐";
  var Iacute = "Í";
  var iacute = "í";
  var ic = "⁣";
  var Icirc = "Î";
  var icirc = "î";
  var Icy = "И";
  var icy = "и";
  var Idot = "İ";
  var IEcy = "Е";
  var iecy = "е";
  var iexcl = "¡";
  var iff = "⇔";
  var ifr = "𝔦";
  var Ifr = "ℑ";
  var Igrave = "Ì";
  var igrave = "ì";
  var ii = "ⅈ";
  var iiiint = "⨌";
  var iiint = "∭";
  var iinfin = "⧜";
  var iiota = "℩";
  var IJlig = "Ĳ";
  var ijlig = "ĳ";
  var Imacr = "Ī";
  var imacr = "ī";
  var image = "ℑ";
  var ImaginaryI = "ⅈ";
  var imagline = "ℐ";
  var imagpart = "ℑ";
  var imath = "ı";
  var Im = "ℑ";
  var imof = "⊷";
  var imped = "Ƶ";
  var Implies = "⇒";
  var incare = "℅";
  var infin = "∞";
  var infintie = "⧝";
  var inodot = "ı";
  var intcal = "⊺";
  var int = "∫";
  var Int = "∬";
  var integers = "ℤ";
  var Integral = "∫";
  var intercal = "⊺";
  var Intersection = "⋂";
  var intlarhk = "⨗";
  var intprod = "⨼";
  var InvisibleComma = "⁣";
  var InvisibleTimes = "⁢";
  var IOcy = "Ё";
  var iocy = "ё";
  var Iogon = "Į";
  var iogon = "į";
  var Iopf = "𝕀";
  var iopf = "𝕚";
  var Iota = "Ι";
  var iota = "ι";
  var iprod = "⨼";
  var iquest = "¿";
  var iscr = "𝒾";
  var Iscr = "ℐ";
  var isin = "∈";
  var isindot = "⋵";
  var isinE = "⋹";
  var isins = "⋴";
  var isinsv = "⋳";
  var isinv = "∈";
  var it = "⁢";
  var Itilde = "Ĩ";
  var itilde = "ĩ";
  var Iukcy = "І";
  var iukcy = "і";
  var Iuml = "Ï";
  var iuml = "ï";
  var Jcirc = "Ĵ";
  var jcirc = "ĵ";
  var Jcy = "Й";
  var jcy = "й";
  var Jfr = "𝔍";
  var jfr = "𝔧";
  var jmath = "ȷ";
  var Jopf = "𝕁";
  var jopf = "𝕛";
  var Jscr = "𝒥";
  var jscr = "𝒿";
  var Jsercy = "Ј";
  var jsercy = "ј";
  var Jukcy = "Є";
  var jukcy = "є";
  var Kappa = "Κ";
  var kappa = "κ";
  var kappav = "ϰ";
  var Kcedil = "Ķ";
  var kcedil = "ķ";
  var Kcy = "К";
  var kcy = "к";
  var Kfr = "𝔎";
  var kfr = "𝔨";
  var kgreen = "ĸ";
  var KHcy = "Х";
  var khcy = "х";
  var KJcy = "Ќ";
  var kjcy = "ќ";
  var Kopf = "𝕂";
  var kopf = "𝕜";
  var Kscr = "𝒦";
  var kscr = "𝓀";
  var lAarr = "⇚";
  var Lacute = "Ĺ";
  var lacute = "ĺ";
  var laemptyv = "⦴";
  var lagran = "ℒ";
  var Lambda = "Λ";
  var lambda = "λ";
  var lang = "⟨";
  var Lang = "⟪";
  var langd = "⦑";
  var langle = "⟨";
  var lap = "⪅";
  var Laplacetrf = "ℒ";
  var laquo = "«";
  var larrb = "⇤";
  var larrbfs = "⤟";
  var larr = "←";
  var Larr = "↞";
  var lArr = "⇐";
  var larrfs = "⤝";
  var larrhk = "↩";
  var larrlp = "↫";
  var larrpl = "⤹";
  var larrsim = "⥳";
  var larrtl = "↢";
  var latail = "⤙";
  var lAtail = "⤛";
  var lat = "⪫";
  var late = "⪭";
  var lates = "⪭︀";
  var lbarr = "⤌";
  var lBarr = "⤎";
  var lbbrk = "❲";
  var lbrace = "{";
  var lbrack = "[";
  var lbrke = "⦋";
  var lbrksld = "⦏";
  var lbrkslu = "⦍";
  var Lcaron = "Ľ";
  var lcaron = "ľ";
  var Lcedil = "Ļ";
  var lcedil = "ļ";
  var lceil = "⌈";
  var lcub = "{";
  var Lcy = "Л";
  var lcy = "л";
  var ldca = "⤶";
  var ldquo = "“";
  var ldquor = "„";
  var ldrdhar = "⥧";
  var ldrushar = "⥋";
  var ldsh = "↲";
  var le = "≤";
  var lE = "≦";
  var LeftAngleBracket = "⟨";
  var LeftArrowBar = "⇤";
  var leftarrow = "←";
  var LeftArrow = "←";
  var Leftarrow = "⇐";
  var LeftArrowRightArrow = "⇆";
  var leftarrowtail = "↢";
  var LeftCeiling = "⌈";
  var LeftDoubleBracket = "⟦";
  var LeftDownTeeVector = "⥡";
  var LeftDownVectorBar = "⥙";
  var LeftDownVector = "⇃";
  var LeftFloor = "⌊";
  var leftharpoondown = "↽";
  var leftharpoonup = "↼";
  var leftleftarrows = "⇇";
  var leftrightarrow = "↔";
  var LeftRightArrow = "↔";
  var Leftrightarrow = "⇔";
  var leftrightarrows = "⇆";
  var leftrightharpoons = "⇋";
  var leftrightsquigarrow = "↭";
  var LeftRightVector = "⥎";
  var LeftTeeArrow = "↤";
  var LeftTee = "⊣";
  var LeftTeeVector = "⥚";
  var leftthreetimes = "⋋";
  var LeftTriangleBar = "⧏";
  var LeftTriangle = "⊲";
  var LeftTriangleEqual = "⊴";
  var LeftUpDownVector = "⥑";
  var LeftUpTeeVector = "⥠";
  var LeftUpVectorBar = "⥘";
  var LeftUpVector = "↿";
  var LeftVectorBar = "⥒";
  var LeftVector = "↼";
  var lEg = "⪋";
  var leg = "⋚";
  var leq = "≤";
  var leqq = "≦";
  var leqslant = "⩽";
  var lescc = "⪨";
  var les = "⩽";
  var lesdot = "⩿";
  var lesdoto = "⪁";
  var lesdotor = "⪃";
  var lesg = "⋚︀";
  var lesges = "⪓";
  var lessapprox = "⪅";
  var lessdot = "⋖";
  var lesseqgtr = "⋚";
  var lesseqqgtr = "⪋";
  var LessEqualGreater = "⋚";
  var LessFullEqual = "≦";
  var LessGreater = "≶";
  var lessgtr = "≶";
  var LessLess = "⪡";
  var lesssim = "≲";
  var LessSlantEqual = "⩽";
  var LessTilde = "≲";
  var lfisht = "⥼";
  var lfloor = "⌊";
  var Lfr = "𝔏";
  var lfr = "𝔩";
  var lg = "≶";
  var lgE = "⪑";
  var lHar = "⥢";
  var lhard = "↽";
  var lharu = "↼";
  var lharul = "⥪";
  var lhblk = "▄";
  var LJcy = "Љ";
  var ljcy = "љ";
  var llarr = "⇇";
  var ll = "≪";
  var Ll = "⋘";
  var llcorner = "⌞";
  var Lleftarrow = "⇚";
  var llhard = "⥫";
  var lltri = "◺";
  var Lmidot = "Ŀ";
  var lmidot = "ŀ";
  var lmoustache = "⎰";
  var lmoust = "⎰";
  var lnap = "⪉";
  var lnapprox = "⪉";
  var lne = "⪇";
  var lnE = "≨";
  var lneq = "⪇";
  var lneqq = "≨";
  var lnsim = "⋦";
  var loang = "⟬";
  var loarr = "⇽";
  var lobrk = "⟦";
  var longleftarrow = "⟵";
  var LongLeftArrow = "⟵";
  var Longleftarrow = "⟸";
  var longleftrightarrow = "⟷";
  var LongLeftRightArrow = "⟷";
  var Longleftrightarrow = "⟺";
  var longmapsto = "⟼";
  var longrightarrow = "⟶";
  var LongRightArrow = "⟶";
  var Longrightarrow = "⟹";
  var looparrowleft = "↫";
  var looparrowright = "↬";
  var lopar = "⦅";
  var Lopf = "𝕃";
  var lopf = "𝕝";
  var loplus = "⨭";
  var lotimes = "⨴";
  var lowast = "∗";
  var lowbar = "_";
  var LowerLeftArrow = "↙";
  var LowerRightArrow = "↘";
  var loz = "◊";
  var lozenge = "◊";
  var lozf = "⧫";
  var lpar = "(";
  var lparlt = "⦓";
  var lrarr = "⇆";
  var lrcorner = "⌟";
  var lrhar = "⇋";
  var lrhard = "⥭";
  var lrm = "‎";
  var lrtri = "⊿";
  var lsaquo = "‹";
  var lscr = "𝓁";
  var Lscr = "ℒ";
  var lsh = "↰";
  var Lsh = "↰";
  var lsim = "≲";
  var lsime = "⪍";
  var lsimg = "⪏";
  var lsqb = "[";
  var lsquo = "‘";
  var lsquor = "‚";
  var Lstrok = "Ł";
  var lstrok = "ł";
  var ltcc = "⪦";
  var ltcir = "⩹";
  var lt$1 = "<";
  var LT = "<";
  var Lt = "≪";
  var ltdot = "⋖";
  var lthree = "⋋";
  var ltimes = "⋉";
  var ltlarr = "⥶";
  var ltquest = "⩻";
  var ltri = "◃";
  var ltrie = "⊴";
  var ltrif = "◂";
  var ltrPar = "⦖";
  var lurdshar = "⥊";
  var luruhar = "⥦";
  var lvertneqq = "≨︀";
  var lvnE = "≨︀";
  var macr = "¯";
  var male = "♂";
  var malt = "✠";
  var maltese = "✠";
  var map$1 = "↦";
  var mapsto = "↦";
  var mapstodown = "↧";
  var mapstoleft = "↤";
  var mapstoup = "↥";
  var marker = "▮";
  var mcomma = "⨩";
  var Mcy = "М";
  var mcy = "м";
  var mdash = "—";
  var mDDot = "∺";
  var measuredangle = "∡";
  var MediumSpace = " ";
  var Mellintrf = "ℳ";
  var Mfr = "𝔐";
  var mfr = "𝔪";
  var mho = "℧";
  var micro = "µ";
  var midast = "*";
  var midcir = "⫰";
  var mid = "∣";
  var middot = "·";
  var minusb = "⊟";
  var minus = "−";
  var minusd = "∸";
  var minusdu = "⨪";
  var MinusPlus = "∓";
  var mlcp = "⫛";
  var mldr = "…";
  var mnplus = "∓";
  var models = "⊧";
  var Mopf = "𝕄";
  var mopf = "𝕞";
  var mp = "∓";
  var mscr = "𝓂";
  var Mscr = "ℳ";
  var mstpos = "∾";
  var Mu = "Μ";
  var mu = "μ";
  var multimap = "⊸";
  var mumap = "⊸";
  var nabla = "∇";
  var Nacute = "Ń";
  var nacute = "ń";
  var nang = "∠⃒";
  var nap = "≉";
  var napE = "⩰̸";
  var napid = "≋̸";
  var napos = "ŉ";
  var napprox = "≉";
  var natural = "♮";
  var naturals = "ℕ";
  var natur = "♮";
  var nbsp = " ";
  var nbump = "≎̸";
  var nbumpe = "≏̸";
  var ncap = "⩃";
  var Ncaron = "Ň";
  var ncaron = "ň";
  var Ncedil = "Ņ";
  var ncedil = "ņ";
  var ncong = "≇";
  var ncongdot = "⩭̸";
  var ncup = "⩂";
  var Ncy = "Н";
  var ncy = "н";
  var ndash = "–";
  var nearhk = "⤤";
  var nearr = "↗";
  var neArr = "⇗";
  var nearrow = "↗";
  var ne = "≠";
  var nedot = "≐̸";
  var NegativeMediumSpace = "​";
  var NegativeThickSpace = "​";
  var NegativeThinSpace = "​";
  var NegativeVeryThinSpace = "​";
  var nequiv = "≢";
  var nesear = "⤨";
  var nesim = "≂̸";
  var NestedGreaterGreater = "≫";
  var NestedLessLess = "≪";
  var NewLine = "\n";
  var nexist = "∄";
  var nexists = "∄";
  var Nfr = "𝔑";
  var nfr = "𝔫";
  var ngE = "≧̸";
  var nge = "≱";
  var ngeq = "≱";
  var ngeqq = "≧̸";
  var ngeqslant = "⩾̸";
  var nges = "⩾̸";
  var nGg = "⋙̸";
  var ngsim = "≵";
  var nGt = "≫⃒";
  var ngt = "≯";
  var ngtr = "≯";
  var nGtv = "≫̸";
  var nharr = "↮";
  var nhArr = "⇎";
  var nhpar = "⫲";
  var ni = "∋";
  var nis = "⋼";
  var nisd = "⋺";
  var niv = "∋";
  var NJcy = "Њ";
  var njcy = "њ";
  var nlarr = "↚";
  var nlArr = "⇍";
  var nldr = "‥";
  var nlE = "≦̸";
  var nle = "≰";
  var nleftarrow = "↚";
  var nLeftarrow = "⇍";
  var nleftrightarrow = "↮";
  var nLeftrightarrow = "⇎";
  var nleq = "≰";
  var nleqq = "≦̸";
  var nleqslant = "⩽̸";
  var nles = "⩽̸";
  var nless = "≮";
  var nLl = "⋘̸";
  var nlsim = "≴";
  var nLt = "≪⃒";
  var nlt = "≮";
  var nltri = "⋪";
  var nltrie = "⋬";
  var nLtv = "≪̸";
  var nmid = "∤";
  var NoBreak = "⁠";
  var NonBreakingSpace = " ";
  var nopf = "𝕟";
  var Nopf = "ℕ";
  var Not = "⫬";
  var not = "¬";
  var NotCongruent = "≢";
  var NotCupCap = "≭";
  var NotDoubleVerticalBar = "∦";
  var NotElement = "∉";
  var NotEqual = "≠";
  var NotEqualTilde = "≂̸";
  var NotExists = "∄";
  var NotGreater = "≯";
  var NotGreaterEqual = "≱";
  var NotGreaterFullEqual = "≧̸";
  var NotGreaterGreater = "≫̸";
  var NotGreaterLess = "≹";
  var NotGreaterSlantEqual = "⩾̸";
  var NotGreaterTilde = "≵";
  var NotHumpDownHump = "≎̸";
  var NotHumpEqual = "≏̸";
  var notin = "∉";
  var notindot = "⋵̸";
  var notinE = "⋹̸";
  var notinva = "∉";
  var notinvb = "⋷";
  var notinvc = "⋶";
  var NotLeftTriangleBar = "⧏̸";
  var NotLeftTriangle = "⋪";
  var NotLeftTriangleEqual = "⋬";
  var NotLess = "≮";
  var NotLessEqual = "≰";
  var NotLessGreater = "≸";
  var NotLessLess = "≪̸";
  var NotLessSlantEqual = "⩽̸";
  var NotLessTilde = "≴";
  var NotNestedGreaterGreater = "⪢̸";
  var NotNestedLessLess = "⪡̸";
  var notni = "∌";
  var notniva = "∌";
  var notnivb = "⋾";
  var notnivc = "⋽";
  var NotPrecedes = "⊀";
  var NotPrecedesEqual = "⪯̸";
  var NotPrecedesSlantEqual = "⋠";
  var NotReverseElement = "∌";
  var NotRightTriangleBar = "⧐̸";
  var NotRightTriangle = "⋫";
  var NotRightTriangleEqual = "⋭";
  var NotSquareSubset = "⊏̸";
  var NotSquareSubsetEqual = "⋢";
  var NotSquareSuperset = "⊐̸";
  var NotSquareSupersetEqual = "⋣";
  var NotSubset = "⊂⃒";
  var NotSubsetEqual = "⊈";
  var NotSucceeds = "⊁";
  var NotSucceedsEqual = "⪰̸";
  var NotSucceedsSlantEqual = "⋡";
  var NotSucceedsTilde = "≿̸";
  var NotSuperset = "⊃⃒";
  var NotSupersetEqual = "⊉";
  var NotTilde = "≁";
  var NotTildeEqual = "≄";
  var NotTildeFullEqual = "≇";
  var NotTildeTilde = "≉";
  var NotVerticalBar = "∤";
  var nparallel = "∦";
  var npar = "∦";
  var nparsl = "⫽⃥";
  var npart = "∂̸";
  var npolint = "⨔";
  var npr = "⊀";
  var nprcue = "⋠";
  var nprec = "⊀";
  var npreceq = "⪯̸";
  var npre = "⪯̸";
  var nrarrc = "⤳̸";
  var nrarr = "↛";
  var nrArr = "⇏";
  var nrarrw = "↝̸";
  var nrightarrow = "↛";
  var nRightarrow = "⇏";
  var nrtri = "⋫";
  var nrtrie = "⋭";
  var nsc = "⊁";
  var nsccue = "⋡";
  var nsce = "⪰̸";
  var Nscr = "𝒩";
  var nscr = "𝓃";
  var nshortmid = "∤";
  var nshortparallel = "∦";
  var nsim = "≁";
  var nsime = "≄";
  var nsimeq = "≄";
  var nsmid = "∤";
  var nspar = "∦";
  var nsqsube = "⋢";
  var nsqsupe = "⋣";
  var nsub = "⊄";
  var nsubE = "⫅̸";
  var nsube = "⊈";
  var nsubset = "⊂⃒";
  var nsubseteq = "⊈";
  var nsubseteqq = "⫅̸";
  var nsucc = "⊁";
  var nsucceq = "⪰̸";
  var nsup = "⊅";
  var nsupE = "⫆̸";
  var nsupe = "⊉";
  var nsupset = "⊃⃒";
  var nsupseteq = "⊉";
  var nsupseteqq = "⫆̸";
  var ntgl = "≹";
  var Ntilde = "Ñ";
  var ntilde = "ñ";
  var ntlg = "≸";
  var ntriangleleft = "⋪";
  var ntrianglelefteq = "⋬";
  var ntriangleright = "⋫";
  var ntrianglerighteq = "⋭";
  var Nu = "Ν";
  var nu = "ν";
  var num = "#";
  var numero = "№";
  var numsp = " ";
  var nvap = "≍⃒";
  var nvdash = "⊬";
  var nvDash = "⊭";
  var nVdash = "⊮";
  var nVDash = "⊯";
  var nvge = "≥⃒";
  var nvgt = ">⃒";
  var nvHarr = "⤄";
  var nvinfin = "⧞";
  var nvlArr = "⤂";
  var nvle = "≤⃒";
  var nvlt = "<⃒";
  var nvltrie = "⊴⃒";
  var nvrArr = "⤃";
  var nvrtrie = "⊵⃒";
  var nvsim = "∼⃒";
  var nwarhk = "⤣";
  var nwarr = "↖";
  var nwArr = "⇖";
  var nwarrow = "↖";
  var nwnear = "⤧";
  var Oacute = "Ó";
  var oacute = "ó";
  var oast = "⊛";
  var Ocirc = "Ô";
  var ocirc = "ô";
  var ocir = "⊚";
  var Ocy = "О";
  var ocy = "о";
  var odash = "⊝";
  var Odblac = "Ő";
  var odblac = "ő";
  var odiv = "⨸";
  var odot = "⊙";
  var odsold = "⦼";
  var OElig = "Œ";
  var oelig = "œ";
  var ofcir = "⦿";
  var Ofr = "𝔒";
  var ofr = "𝔬";
  var ogon = "˛";
  var Ograve = "Ò";
  var ograve = "ò";
  var ogt = "⧁";
  var ohbar = "⦵";
  var ohm = "Ω";
  var oint = "∮";
  var olarr = "↺";
  var olcir = "⦾";
  var olcross = "⦻";
  var oline = "‾";
  var olt = "⧀";
  var Omacr = "Ō";
  var omacr = "ō";
  var Omega = "Ω";
  var omega = "ω";
  var Omicron = "Ο";
  var omicron = "ο";
  var omid = "⦶";
  var ominus = "⊖";
  var Oopf = "𝕆";
  var oopf = "𝕠";
  var opar = "⦷";
  var OpenCurlyDoubleQuote = "“";
  var OpenCurlyQuote = "‘";
  var operp = "⦹";
  var oplus = "⊕";
  var orarr = "↻";
  var Or = "⩔";
  var or = "∨";
  var ord = "⩝";
  var order = "ℴ";
  var orderof = "ℴ";
  var ordf = "ª";
  var ordm = "º";
  var origof = "⊶";
  var oror = "⩖";
  var orslope = "⩗";
  var orv = "⩛";
  var oS = "Ⓢ";
  var Oscr = "𝒪";
  var oscr = "ℴ";
  var Oslash = "Ø";
  var oslash = "ø";
  var osol = "⊘";
  var Otilde = "Õ";
  var otilde = "õ";
  var otimesas = "⨶";
  var Otimes = "⨷";
  var otimes = "⊗";
  var Ouml = "Ö";
  var ouml = "ö";
  var ovbar = "⌽";
  var OverBar = "‾";
  var OverBrace = "⏞";
  var OverBracket = "⎴";
  var OverParenthesis = "⏜";
  var para = "¶";
  var parallel = "∥";
  var par = "∥";
  var parsim = "⫳";
  var parsl = "⫽";
  var part = "∂";
  var PartialD = "∂";
  var Pcy = "П";
  var pcy = "п";
  var percnt = "%";
  var period = ".";
  var permil = "‰";
  var perp = "⊥";
  var pertenk = "‱";
  var Pfr = "𝔓";
  var pfr = "𝔭";
  var Phi = "Φ";
  var phi = "φ";
  var phiv = "ϕ";
  var phmmat = "ℳ";
  var phone = "☎";
  var Pi = "Π";
  var pi = "π";
  var pitchfork = "⋔";
  var piv = "ϖ";
  var planck = "ℏ";
  var planckh = "ℎ";
  var plankv = "ℏ";
  var plusacir = "⨣";
  var plusb = "⊞";
  var pluscir = "⨢";
  var plus = "+";
  var plusdo = "∔";
  var plusdu = "⨥";
  var pluse = "⩲";
  var PlusMinus = "±";
  var plusmn = "±";
  var plussim = "⨦";
  var plustwo = "⨧";
  var pm = "±";
  var Poincareplane = "ℌ";
  var pointint = "⨕";
  var popf = "𝕡";
  var Popf = "ℙ";
  var pound = "£";
  var prap = "⪷";
  var Pr = "⪻";
  var pr = "≺";
  var prcue = "≼";
  var precapprox = "⪷";
  var prec = "≺";
  var preccurlyeq = "≼";
  var Precedes = "≺";
  var PrecedesEqual = "⪯";
  var PrecedesSlantEqual = "≼";
  var PrecedesTilde = "≾";
  var preceq = "⪯";
  var precnapprox = "⪹";
  var precneqq = "⪵";
  var precnsim = "⋨";
  var pre = "⪯";
  var prE = "⪳";
  var precsim = "≾";
  var prime = "′";
  var Prime = "″";
  var primes = "ℙ";
  var prnap = "⪹";
  var prnE = "⪵";
  var prnsim = "⋨";
  var prod = "∏";
  var Product = "∏";
  var profalar = "⌮";
  var profline = "⌒";
  var profsurf = "⌓";
  var prop = "∝";
  var Proportional = "∝";
  var Proportion = "∷";
  var propto = "∝";
  var prsim = "≾";
  var prurel = "⊰";
  var Pscr = "𝒫";
  var pscr = "𝓅";
  var Psi = "Ψ";
  var psi = "ψ";
  var puncsp = " ";
  var Qfr = "𝔔";
  var qfr = "𝔮";
  var qint = "⨌";
  var qopf = "𝕢";
  var Qopf = "ℚ";
  var qprime = "⁗";
  var Qscr = "𝒬";
  var qscr = "𝓆";
  var quaternions = "ℍ";
  var quatint = "⨖";
  var quest = "?";
  var questeq = "≟";
  var quot$1 = "\"";
  var QUOT = "\"";
  var rAarr = "⇛";
  var race = "∽̱";
  var Racute = "Ŕ";
  var racute = "ŕ";
  var radic = "√";
  var raemptyv = "⦳";
  var rang = "⟩";
  var Rang = "⟫";
  var rangd = "⦒";
  var range = "⦥";
  var rangle = "⟩";
  var raquo = "»";
  var rarrap = "⥵";
  var rarrb = "⇥";
  var rarrbfs = "⤠";
  var rarrc = "⤳";
  var rarr = "→";
  var Rarr = "↠";
  var rArr = "⇒";
  var rarrfs = "⤞";
  var rarrhk = "↪";
  var rarrlp = "↬";
  var rarrpl = "⥅";
  var rarrsim = "⥴";
  var Rarrtl = "⤖";
  var rarrtl = "↣";
  var rarrw = "↝";
  var ratail = "⤚";
  var rAtail = "⤜";
  var ratio = "∶";
  var rationals = "ℚ";
  var rbarr = "⤍";
  var rBarr = "⤏";
  var RBarr = "⤐";
  var rbbrk = "❳";
  var rbrace = "}";
  var rbrack = "]";
  var rbrke = "⦌";
  var rbrksld = "⦎";
  var rbrkslu = "⦐";
  var Rcaron = "Ř";
  var rcaron = "ř";
  var Rcedil = "Ŗ";
  var rcedil = "ŗ";
  var rceil = "⌉";
  var rcub = "}";
  var Rcy = "Р";
  var rcy = "р";
  var rdca = "⤷";
  var rdldhar = "⥩";
  var rdquo = "”";
  var rdquor = "”";
  var rdsh = "↳";
  var real = "ℜ";
  var realine = "ℛ";
  var realpart = "ℜ";
  var reals = "ℝ";
  var Re = "ℜ";
  var rect = "▭";
  var reg = "®";
  var REG = "®";
  var ReverseElement = "∋";
  var ReverseEquilibrium = "⇋";
  var ReverseUpEquilibrium = "⥯";
  var rfisht = "⥽";
  var rfloor = "⌋";
  var rfr = "𝔯";
  var Rfr = "ℜ";
  var rHar = "⥤";
  var rhard = "⇁";
  var rharu = "⇀";
  var rharul = "⥬";
  var Rho = "Ρ";
  var rho = "ρ";
  var rhov = "ϱ";
  var RightAngleBracket = "⟩";
  var RightArrowBar = "⇥";
  var rightarrow = "→";
  var RightArrow = "→";
  var Rightarrow = "⇒";
  var RightArrowLeftArrow = "⇄";
  var rightarrowtail = "↣";
  var RightCeiling = "⌉";
  var RightDoubleBracket = "⟧";
  var RightDownTeeVector = "⥝";
  var RightDownVectorBar = "⥕";
  var RightDownVector = "⇂";
  var RightFloor = "⌋";
  var rightharpoondown = "⇁";
  var rightharpoonup = "⇀";
  var rightleftarrows = "⇄";
  var rightleftharpoons = "⇌";
  var rightrightarrows = "⇉";
  var rightsquigarrow = "↝";
  var RightTeeArrow = "↦";
  var RightTee = "⊢";
  var RightTeeVector = "⥛";
  var rightthreetimes = "⋌";
  var RightTriangleBar = "⧐";
  var RightTriangle = "⊳";
  var RightTriangleEqual = "⊵";
  var RightUpDownVector = "⥏";
  var RightUpTeeVector = "⥜";
  var RightUpVectorBar = "⥔";
  var RightUpVector = "↾";
  var RightVectorBar = "⥓";
  var RightVector = "⇀";
  var ring = "˚";
  var risingdotseq = "≓";
  var rlarr = "⇄";
  var rlhar = "⇌";
  var rlm = "‏";
  var rmoustache = "⎱";
  var rmoust = "⎱";
  var rnmid = "⫮";
  var roang = "⟭";
  var roarr = "⇾";
  var robrk = "⟧";
  var ropar = "⦆";
  var ropf = "𝕣";
  var Ropf = "ℝ";
  var roplus = "⨮";
  var rotimes = "⨵";
  var RoundImplies = "⥰";
  var rpar = ")";
  var rpargt = "⦔";
  var rppolint = "⨒";
  var rrarr = "⇉";
  var Rrightarrow = "⇛";
  var rsaquo = "›";
  var rscr = "𝓇";
  var Rscr = "ℛ";
  var rsh = "↱";
  var Rsh = "↱";
  var rsqb = "]";
  var rsquo = "’";
  var rsquor = "’";
  var rthree = "⋌";
  var rtimes = "⋊";
  var rtri = "▹";
  var rtrie = "⊵";
  var rtrif = "▸";
  var rtriltri = "⧎";
  var RuleDelayed = "⧴";
  var ruluhar = "⥨";
  var rx = "℞";
  var Sacute = "Ś";
  var sacute = "ś";
  var sbquo = "‚";
  var scap = "⪸";
  var Scaron = "Š";
  var scaron = "š";
  var Sc = "⪼";
  var sc = "≻";
  var sccue = "≽";
  var sce = "⪰";
  var scE = "⪴";
  var Scedil = "Ş";
  var scedil = "ş";
  var Scirc = "Ŝ";
  var scirc = "ŝ";
  var scnap = "⪺";
  var scnE = "⪶";
  var scnsim = "⋩";
  var scpolint = "⨓";
  var scsim = "≿";
  var Scy = "С";
  var scy = "с";
  var sdotb = "⊡";
  var sdot = "⋅";
  var sdote = "⩦";
  var searhk = "⤥";
  var searr = "↘";
  var seArr = "⇘";
  var searrow = "↘";
  var sect = "§";
  var semi = ";";
  var seswar = "⤩";
  var setminus = "∖";
  var setmn = "∖";
  var sext = "✶";
  var Sfr = "𝔖";
  var sfr = "𝔰";
  var sfrown = "⌢";
  var sharp = "♯";
  var SHCHcy = "Щ";
  var shchcy = "щ";
  var SHcy = "Ш";
  var shcy = "ш";
  var ShortDownArrow = "↓";
  var ShortLeftArrow = "←";
  var shortmid = "∣";
  var shortparallel = "∥";
  var ShortRightArrow = "→";
  var ShortUpArrow = "↑";
  var shy = "­";
  var Sigma = "Σ";
  var sigma = "σ";
  var sigmaf = "ς";
  var sigmav = "ς";
  var sim = "∼";
  var simdot = "⩪";
  var sime = "≃";
  var simeq = "≃";
  var simg = "⪞";
  var simgE = "⪠";
  var siml = "⪝";
  var simlE = "⪟";
  var simne = "≆";
  var simplus = "⨤";
  var simrarr = "⥲";
  var slarr = "←";
  var SmallCircle = "∘";
  var smallsetminus = "∖";
  var smashp = "⨳";
  var smeparsl = "⧤";
  var smid = "∣";
  var smile = "⌣";
  var smt = "⪪";
  var smte = "⪬";
  var smtes = "⪬︀";
  var SOFTcy = "Ь";
  var softcy = "ь";
  var solbar = "⌿";
  var solb = "⧄";
  var sol = "/";
  var Sopf = "𝕊";
  var sopf = "𝕤";
  var spades = "♠";
  var spadesuit = "♠";
  var spar = "∥";
  var sqcap = "⊓";
  var sqcaps = "⊓︀";
  var sqcup = "⊔";
  var sqcups = "⊔︀";
  var Sqrt = "√";
  var sqsub = "⊏";
  var sqsube = "⊑";
  var sqsubset = "⊏";
  var sqsubseteq = "⊑";
  var sqsup = "⊐";
  var sqsupe = "⊒";
  var sqsupset = "⊐";
  var sqsupseteq = "⊒";
  var square = "□";
  var Square = "□";
  var SquareIntersection = "⊓";
  var SquareSubset = "⊏";
  var SquareSubsetEqual = "⊑";
  var SquareSuperset = "⊐";
  var SquareSupersetEqual = "⊒";
  var SquareUnion = "⊔";
  var squarf = "▪";
  var squ = "□";
  var squf = "▪";
  var srarr = "→";
  var Sscr = "𝒮";
  var sscr = "𝓈";
  var ssetmn = "∖";
  var ssmile = "⌣";
  var sstarf = "⋆";
  var Star = "⋆";
  var star = "☆";
  var starf = "★";
  var straightepsilon = "ϵ";
  var straightphi = "ϕ";
  var strns = "¯";
  var sub = "⊂";
  var Sub = "⋐";
  var subdot = "⪽";
  var subE = "⫅";
  var sube = "⊆";
  var subedot = "⫃";
  var submult = "⫁";
  var subnE = "⫋";
  var subne = "⊊";
  var subplus = "⪿";
  var subrarr = "⥹";
  var subset = "⊂";
  var Subset = "⋐";
  var subseteq = "⊆";
  var subseteqq = "⫅";
  var SubsetEqual = "⊆";
  var subsetneq = "⊊";
  var subsetneqq = "⫋";
  var subsim = "⫇";
  var subsub = "⫕";
  var subsup = "⫓";
  var succapprox = "⪸";
  var succ = "≻";
  var succcurlyeq = "≽";
  var Succeeds = "≻";
  var SucceedsEqual = "⪰";
  var SucceedsSlantEqual = "≽";
  var SucceedsTilde = "≿";
  var succeq = "⪰";
  var succnapprox = "⪺";
  var succneqq = "⪶";
  var succnsim = "⋩";
  var succsim = "≿";
  var SuchThat = "∋";
  var sum = "∑";
  var Sum = "∑";
  var sung = "♪";
  var sup1 = "¹";
  var sup2 = "²";
  var sup3 = "³";
  var sup = "⊃";
  var Sup = "⋑";
  var supdot = "⪾";
  var supdsub = "⫘";
  var supE = "⫆";
  var supe = "⊇";
  var supedot = "⫄";
  var Superset = "⊃";
  var SupersetEqual = "⊇";
  var suphsol = "⟉";
  var suphsub = "⫗";
  var suplarr = "⥻";
  var supmult = "⫂";
  var supnE = "⫌";
  var supne = "⊋";
  var supplus = "⫀";
  var supset = "⊃";
  var Supset = "⋑";
  var supseteq = "⊇";
  var supseteqq = "⫆";
  var supsetneq = "⊋";
  var supsetneqq = "⫌";
  var supsim = "⫈";
  var supsub = "⫔";
  var supsup = "⫖";
  var swarhk = "⤦";
  var swarr = "↙";
  var swArr = "⇙";
  var swarrow = "↙";
  var swnwar = "⤪";
  var szlig = "ß";
  var Tab = "\t";
  var target = "⌖";
  var Tau = "Τ";
  var tau = "τ";
  var tbrk = "⎴";
  var Tcaron = "Ť";
  var tcaron = "ť";
  var Tcedil = "Ţ";
  var tcedil = "ţ";
  var Tcy = "Т";
  var tcy = "т";
  var tdot = "⃛";
  var telrec = "⌕";
  var Tfr = "𝔗";
  var tfr = "𝔱";
  var there4 = "∴";
  var therefore = "∴";
  var Therefore = "∴";
  var Theta = "Θ";
  var theta = "θ";
  var thetasym = "ϑ";
  var thetav = "ϑ";
  var thickapprox = "≈";
  var thicksim = "∼";
  var ThickSpace = "  ";
  var ThinSpace = " ";
  var thinsp = " ";
  var thkap = "≈";
  var thksim = "∼";
  var THORN = "Þ";
  var thorn = "þ";
  var tilde = "˜";
  var Tilde = "∼";
  var TildeEqual = "≃";
  var TildeFullEqual = "≅";
  var TildeTilde = "≈";
  var timesbar = "⨱";
  var timesb = "⊠";
  var times = "×";
  var timesd = "⨰";
  var tint = "∭";
  var toea = "⤨";
  var topbot = "⌶";
  var topcir = "⫱";
  var top = "⊤";
  var Topf = "𝕋";
  var topf = "𝕥";
  var topfork = "⫚";
  var tosa = "⤩";
  var tprime = "‴";
  var trade = "™";
  var TRADE = "™";
  var triangle = "▵";
  var triangledown = "▿";
  var triangleleft = "◃";
  var trianglelefteq = "⊴";
  var triangleq = "≜";
  var triangleright = "▹";
  var trianglerighteq = "⊵";
  var tridot = "◬";
  var trie = "≜";
  var triminus = "⨺";
  var TripleDot = "⃛";
  var triplus = "⨹";
  var trisb = "⧍";
  var tritime = "⨻";
  var trpezium = "⏢";
  var Tscr = "𝒯";
  var tscr = "𝓉";
  var TScy = "Ц";
  var tscy = "ц";
  var TSHcy = "Ћ";
  var tshcy = "ћ";
  var Tstrok = "Ŧ";
  var tstrok = "ŧ";
  var twixt = "≬";
  var twoheadleftarrow = "↞";
  var twoheadrightarrow = "↠";
  var Uacute = "Ú";
  var uacute = "ú";
  var uarr = "↑";
  var Uarr = "↟";
  var uArr = "⇑";
  var Uarrocir = "⥉";
  var Ubrcy = "Ў";
  var ubrcy = "ў";
  var Ubreve = "Ŭ";
  var ubreve = "ŭ";
  var Ucirc = "Û";
  var ucirc = "û";
  var Ucy = "У";
  var ucy = "у";
  var udarr = "⇅";
  var Udblac = "Ű";
  var udblac = "ű";
  var udhar = "⥮";
  var ufisht = "⥾";
  var Ufr = "𝔘";
  var ufr = "𝔲";
  var Ugrave = "Ù";
  var ugrave = "ù";
  var uHar = "⥣";
  var uharl = "↿";
  var uharr = "↾";
  var uhblk = "▀";
  var ulcorn = "⌜";
  var ulcorner = "⌜";
  var ulcrop = "⌏";
  var ultri = "◸";
  var Umacr = "Ū";
  var umacr = "ū";
  var uml = "¨";
  var UnderBar = "_";
  var UnderBrace = "⏟";
  var UnderBracket = "⎵";
  var UnderParenthesis = "⏝";
  var Union = "⋃";
  var UnionPlus = "⊎";
  var Uogon = "Ų";
  var uogon = "ų";
  var Uopf = "𝕌";
  var uopf = "𝕦";
  var UpArrowBar = "⤒";
  var uparrow = "↑";
  var UpArrow = "↑";
  var Uparrow = "⇑";
  var UpArrowDownArrow = "⇅";
  var updownarrow = "↕";
  var UpDownArrow = "↕";
  var Updownarrow = "⇕";
  var UpEquilibrium = "⥮";
  var upharpoonleft = "↿";
  var upharpoonright = "↾";
  var uplus = "⊎";
  var UpperLeftArrow = "↖";
  var UpperRightArrow = "↗";
  var upsi = "υ";
  var Upsi = "ϒ";
  var upsih = "ϒ";
  var Upsilon = "Υ";
  var upsilon = "υ";
  var UpTeeArrow = "↥";
  var UpTee = "⊥";
  var upuparrows = "⇈";
  var urcorn = "⌝";
  var urcorner = "⌝";
  var urcrop = "⌎";
  var Uring = "Ů";
  var uring = "ů";
  var urtri = "◹";
  var Uscr = "𝒰";
  var uscr = "𝓊";
  var utdot = "⋰";
  var Utilde = "Ũ";
  var utilde = "ũ";
  var utri = "▵";
  var utrif = "▴";
  var uuarr = "⇈";
  var Uuml = "Ü";
  var uuml = "ü";
  var uwangle = "⦧";
  var vangrt = "⦜";
  var varepsilon = "ϵ";
  var varkappa = "ϰ";
  var varnothing = "∅";
  var varphi = "ϕ";
  var varpi = "ϖ";
  var varpropto = "∝";
  var varr = "↕";
  var vArr = "⇕";
  var varrho = "ϱ";
  var varsigma = "ς";
  var varsubsetneq = "⊊︀";
  var varsubsetneqq = "⫋︀";
  var varsupsetneq = "⊋︀";
  var varsupsetneqq = "⫌︀";
  var vartheta = "ϑ";
  var vartriangleleft = "⊲";
  var vartriangleright = "⊳";
  var vBar = "⫨";
  var Vbar = "⫫";
  var vBarv = "⫩";
  var Vcy = "В";
  var vcy = "в";
  var vdash = "⊢";
  var vDash = "⊨";
  var Vdash = "⊩";
  var VDash = "⊫";
  var Vdashl = "⫦";
  var veebar = "⊻";
  var vee = "∨";
  var Vee = "⋁";
  var veeeq = "≚";
  var vellip = "⋮";
  var verbar = "|";
  var Verbar = "‖";
  var vert = "|";
  var Vert = "‖";
  var VerticalBar = "∣";
  var VerticalLine = "|";
  var VerticalSeparator = "❘";
  var VerticalTilde = "≀";
  var VeryThinSpace = " ";
  var Vfr = "𝔙";
  var vfr = "𝔳";
  var vltri = "⊲";
  var vnsub = "⊂⃒";
  var vnsup = "⊃⃒";
  var Vopf = "𝕍";
  var vopf = "𝕧";
  var vprop = "∝";
  var vrtri = "⊳";
  var Vscr = "𝒱";
  var vscr = "𝓋";
  var vsubnE = "⫋︀";
  var vsubne = "⊊︀";
  var vsupnE = "⫌︀";
  var vsupne = "⊋︀";
  var Vvdash = "⊪";
  var vzigzag = "⦚";
  var Wcirc = "Ŵ";
  var wcirc = "ŵ";
  var wedbar = "⩟";
  var wedge = "∧";
  var Wedge = "⋀";
  var wedgeq = "≙";
  var weierp = "℘";
  var Wfr = "𝔚";
  var wfr = "𝔴";
  var Wopf = "𝕎";
  var wopf = "𝕨";
  var wp = "℘";
  var wr = "≀";
  var wreath = "≀";
  var Wscr = "𝒲";
  var wscr = "𝓌";
  var xcap = "⋂";
  var xcirc = "◯";
  var xcup = "⋃";
  var xdtri = "▽";
  var Xfr = "𝔛";
  var xfr = "𝔵";
  var xharr = "⟷";
  var xhArr = "⟺";
  var Xi = "Ξ";
  var xi = "ξ";
  var xlarr = "⟵";
  var xlArr = "⟸";
  var xmap = "⟼";
  var xnis = "⋻";
  var xodot = "⨀";
  var Xopf = "𝕏";
  var xopf = "𝕩";
  var xoplus = "⨁";
  var xotime = "⨂";
  var xrarr = "⟶";
  var xrArr = "⟹";
  var Xscr = "𝒳";
  var xscr = "𝓍";
  var xsqcup = "⨆";
  var xuplus = "⨄";
  var xutri = "△";
  var xvee = "⋁";
  var xwedge = "⋀";
  var Yacute = "Ý";
  var yacute = "ý";
  var YAcy = "Я";
  var yacy = "я";
  var Ycirc = "Ŷ";
  var ycirc = "ŷ";
  var Ycy = "Ы";
  var ycy = "ы";
  var yen = "¥";
  var Yfr = "𝔜";
  var yfr = "𝔶";
  var YIcy = "Ї";
  var yicy = "ї";
  var Yopf = "𝕐";
  var yopf = "𝕪";
  var Yscr = "𝒴";
  var yscr = "𝓎";
  var YUcy = "Ю";
  var yucy = "ю";
  var yuml = "ÿ";
  var Yuml = "Ÿ";
  var Zacute = "Ź";
  var zacute = "ź";
  var Zcaron = "Ž";
  var zcaron = "ž";
  var Zcy = "З";
  var zcy = "з";
  var Zdot = "Ż";
  var zdot = "ż";
  var zeetrf = "ℨ";
  var ZeroWidthSpace = "​";
  var Zeta = "Ζ";
  var zeta = "ζ";
  var zfr = "𝔷";
  var Zfr = "ℨ";
  var ZHcy = "Ж";
  var zhcy = "ж";
  var zigrarr = "⇝";
  var zopf = "𝕫";
  var Zopf = "ℤ";
  var Zscr = "𝒵";
  var zscr = "𝓏";
  var zwj = "‍";
  var zwnj = "‌";
  var entitiesJSON = {
  	Aacute: Aacute,
  	aacute: aacute,
  	Abreve: Abreve,
  	abreve: abreve,
  	ac: ac,
  	acd: acd,
  	acE: acE,
  	Acirc: Acirc,
  	acirc: acirc,
  	acute: acute,
  	Acy: Acy,
  	acy: acy,
  	AElig: AElig,
  	aelig: aelig,
  	af: af,
  	Afr: Afr,
  	afr: afr,
  	Agrave: Agrave,
  	agrave: agrave,
  	alefsym: alefsym,
  	aleph: aleph,
  	Alpha: Alpha,
  	alpha: alpha,
  	Amacr: Amacr,
  	amacr: amacr,
  	amalg: amalg,
  	amp: amp$1,
  	AMP: AMP,
  	andand: andand,
  	And: And,
  	and: and,
  	andd: andd,
  	andslope: andslope,
  	andv: andv,
  	ang: ang,
  	ange: ange,
  	angle: angle,
  	angmsdaa: angmsdaa,
  	angmsdab: angmsdab,
  	angmsdac: angmsdac,
  	angmsdad: angmsdad,
  	angmsdae: angmsdae,
  	angmsdaf: angmsdaf,
  	angmsdag: angmsdag,
  	angmsdah: angmsdah,
  	angmsd: angmsd,
  	angrt: angrt,
  	angrtvb: angrtvb,
  	angrtvbd: angrtvbd,
  	angsph: angsph,
  	angst: angst,
  	angzarr: angzarr,
  	Aogon: Aogon,
  	aogon: aogon,
  	Aopf: Aopf,
  	aopf: aopf,
  	apacir: apacir,
  	ap: ap,
  	apE: apE,
  	ape: ape,
  	apid: apid,
  	apos: apos$1,
  	ApplyFunction: ApplyFunction,
  	approx: approx,
  	approxeq: approxeq,
  	Aring: Aring,
  	aring: aring,
  	Ascr: Ascr,
  	ascr: ascr,
  	Assign: Assign,
  	ast: ast,
  	asymp: asymp,
  	asympeq: asympeq,
  	Atilde: Atilde,
  	atilde: atilde,
  	Auml: Auml,
  	auml: auml,
  	awconint: awconint,
  	awint: awint,
  	backcong: backcong,
  	backepsilon: backepsilon,
  	backprime: backprime,
  	backsim: backsim,
  	backsimeq: backsimeq,
  	Backslash: Backslash,
  	Barv: Barv,
  	barvee: barvee,
  	barwed: barwed,
  	Barwed: Barwed,
  	barwedge: barwedge,
  	bbrk: bbrk,
  	bbrktbrk: bbrktbrk,
  	bcong: bcong,
  	Bcy: Bcy,
  	bcy: bcy,
  	bdquo: bdquo,
  	becaus: becaus,
  	because: because,
  	Because: Because,
  	bemptyv: bemptyv,
  	bepsi: bepsi,
  	bernou: bernou,
  	Bernoullis: Bernoullis,
  	Beta: Beta,
  	beta: beta,
  	beth: beth,
  	between: between,
  	Bfr: Bfr,
  	bfr: bfr,
  	bigcap: bigcap,
  	bigcirc: bigcirc,
  	bigcup: bigcup,
  	bigodot: bigodot,
  	bigoplus: bigoplus,
  	bigotimes: bigotimes,
  	bigsqcup: bigsqcup,
  	bigstar: bigstar,
  	bigtriangledown: bigtriangledown,
  	bigtriangleup: bigtriangleup,
  	biguplus: biguplus,
  	bigvee: bigvee,
  	bigwedge: bigwedge,
  	bkarow: bkarow,
  	blacklozenge: blacklozenge,
  	blacksquare: blacksquare,
  	blacktriangle: blacktriangle,
  	blacktriangledown: blacktriangledown,
  	blacktriangleleft: blacktriangleleft,
  	blacktriangleright: blacktriangleright,
  	blank: blank,
  	blk12: blk12,
  	blk14: blk14,
  	blk34: blk34,
  	block: block,
  	bne: bne,
  	bnequiv: bnequiv,
  	bNot: bNot,
  	bnot: bnot,
  	Bopf: Bopf,
  	bopf: bopf,
  	bot: bot,
  	bottom: bottom,
  	bowtie: bowtie,
  	boxbox: boxbox,
  	boxdl: boxdl,
  	boxdL: boxdL,
  	boxDl: boxDl,
  	boxDL: boxDL,
  	boxdr: boxdr,
  	boxdR: boxdR,
  	boxDr: boxDr,
  	boxDR: boxDR,
  	boxh: boxh,
  	boxH: boxH,
  	boxhd: boxhd,
  	boxHd: boxHd,
  	boxhD: boxhD,
  	boxHD: boxHD,
  	boxhu: boxhu,
  	boxHu: boxHu,
  	boxhU: boxhU,
  	boxHU: boxHU,
  	boxminus: boxminus,
  	boxplus: boxplus,
  	boxtimes: boxtimes,
  	boxul: boxul,
  	boxuL: boxuL,
  	boxUl: boxUl,
  	boxUL: boxUL,
  	boxur: boxur,
  	boxuR: boxuR,
  	boxUr: boxUr,
  	boxUR: boxUR,
  	boxv: boxv,
  	boxV: boxV,
  	boxvh: boxvh,
  	boxvH: boxvH,
  	boxVh: boxVh,
  	boxVH: boxVH,
  	boxvl: boxvl,
  	boxvL: boxvL,
  	boxVl: boxVl,
  	boxVL: boxVL,
  	boxvr: boxvr,
  	boxvR: boxvR,
  	boxVr: boxVr,
  	boxVR: boxVR,
  	bprime: bprime,
  	breve: breve,
  	Breve: Breve,
  	brvbar: brvbar,
  	bscr: bscr,
  	Bscr: Bscr,
  	bsemi: bsemi,
  	bsim: bsim,
  	bsime: bsime,
  	bsolb: bsolb,
  	bsol: bsol,
  	bsolhsub: bsolhsub,
  	bull: bull,
  	bullet: bullet,
  	bump: bump,
  	bumpE: bumpE,
  	bumpe: bumpe,
  	Bumpeq: Bumpeq,
  	bumpeq: bumpeq,
  	Cacute: Cacute,
  	cacute: cacute,
  	capand: capand,
  	capbrcup: capbrcup,
  	capcap: capcap,
  	cap: cap,
  	Cap: Cap,
  	capcup: capcup,
  	capdot: capdot,
  	CapitalDifferentialD: CapitalDifferentialD,
  	caps: caps,
  	caret: caret,
  	caron: caron,
  	Cayleys: Cayleys,
  	ccaps: ccaps,
  	Ccaron: Ccaron,
  	ccaron: ccaron,
  	Ccedil: Ccedil,
  	ccedil: ccedil,
  	Ccirc: Ccirc,
  	ccirc: ccirc,
  	Cconint: Cconint,
  	ccups: ccups,
  	ccupssm: ccupssm,
  	Cdot: Cdot,
  	cdot: cdot,
  	cedil: cedil,
  	Cedilla: Cedilla,
  	cemptyv: cemptyv,
  	cent: cent,
  	centerdot: centerdot,
  	CenterDot: CenterDot,
  	cfr: cfr,
  	Cfr: Cfr,
  	CHcy: CHcy,
  	chcy: chcy,
  	check: check,
  	checkmark: checkmark,
  	Chi: Chi,
  	chi: chi,
  	circ: circ,
  	circeq: circeq,
  	circlearrowleft: circlearrowleft,
  	circlearrowright: circlearrowright,
  	circledast: circledast,
  	circledcirc: circledcirc,
  	circleddash: circleddash,
  	CircleDot: CircleDot,
  	circledR: circledR,
  	circledS: circledS,
  	CircleMinus: CircleMinus,
  	CirclePlus: CirclePlus,
  	CircleTimes: CircleTimes,
  	cir: cir,
  	cirE: cirE,
  	cire: cire,
  	cirfnint: cirfnint,
  	cirmid: cirmid,
  	cirscir: cirscir,
  	ClockwiseContourIntegral: ClockwiseContourIntegral,
  	CloseCurlyDoubleQuote: CloseCurlyDoubleQuote,
  	CloseCurlyQuote: CloseCurlyQuote,
  	clubs: clubs,
  	clubsuit: clubsuit,
  	colon: colon,
  	Colon: Colon,
  	Colone: Colone,
  	colone: colone,
  	coloneq: coloneq,
  	comma: comma,
  	commat: commat,
  	comp: comp,
  	compfn: compfn,
  	complement: complement,
  	complexes: complexes,
  	cong: cong,
  	congdot: congdot,
  	Congruent: Congruent,
  	conint: conint,
  	Conint: Conint,
  	ContourIntegral: ContourIntegral,
  	copf: copf,
  	Copf: Copf,
  	coprod: coprod,
  	Coproduct: Coproduct,
  	copy: copy,
  	COPY: COPY,
  	copysr: copysr,
  	CounterClockwiseContourIntegral: CounterClockwiseContourIntegral,
  	crarr: crarr,
  	cross: cross,
  	Cross: Cross,
  	Cscr: Cscr,
  	cscr: cscr,
  	csub: csub,
  	csube: csube,
  	csup: csup,
  	csupe: csupe,
  	ctdot: ctdot,
  	cudarrl: cudarrl,
  	cudarrr: cudarrr,
  	cuepr: cuepr,
  	cuesc: cuesc,
  	cularr: cularr,
  	cularrp: cularrp,
  	cupbrcap: cupbrcap,
  	cupcap: cupcap,
  	CupCap: CupCap,
  	cup: cup,
  	Cup: Cup,
  	cupcup: cupcup,
  	cupdot: cupdot,
  	cupor: cupor,
  	cups: cups,
  	curarr: curarr,
  	curarrm: curarrm,
  	curlyeqprec: curlyeqprec,
  	curlyeqsucc: curlyeqsucc,
  	curlyvee: curlyvee,
  	curlywedge: curlywedge,
  	curren: curren,
  	curvearrowleft: curvearrowleft,
  	curvearrowright: curvearrowright,
  	cuvee: cuvee,
  	cuwed: cuwed,
  	cwconint: cwconint,
  	cwint: cwint,
  	cylcty: cylcty,
  	dagger: dagger,
  	Dagger: Dagger,
  	daleth: daleth,
  	darr: darr,
  	Darr: Darr,
  	dArr: dArr,
  	dash: dash,
  	Dashv: Dashv,
  	dashv: dashv,
  	dbkarow: dbkarow,
  	dblac: dblac,
  	Dcaron: Dcaron,
  	dcaron: dcaron,
  	Dcy: Dcy,
  	dcy: dcy,
  	ddagger: ddagger,
  	ddarr: ddarr,
  	DD: DD,
  	dd: dd,
  	DDotrahd: DDotrahd,
  	ddotseq: ddotseq,
  	deg: deg,
  	Del: Del,
  	Delta: Delta,
  	delta: delta,
  	demptyv: demptyv,
  	dfisht: dfisht,
  	Dfr: Dfr,
  	dfr: dfr,
  	dHar: dHar,
  	dharl: dharl,
  	dharr: dharr,
  	DiacriticalAcute: DiacriticalAcute,
  	DiacriticalDot: DiacriticalDot,
  	DiacriticalDoubleAcute: DiacriticalDoubleAcute,
  	DiacriticalGrave: DiacriticalGrave,
  	DiacriticalTilde: DiacriticalTilde,
  	diam: diam,
  	diamond: diamond,
  	Diamond: Diamond,
  	diamondsuit: diamondsuit,
  	diams: diams,
  	die: die,
  	DifferentialD: DifferentialD,
  	digamma: digamma,
  	disin: disin,
  	div: div,
  	divide: divide,
  	divideontimes: divideontimes,
  	divonx: divonx,
  	DJcy: DJcy,
  	djcy: djcy,
  	dlcorn: dlcorn,
  	dlcrop: dlcrop,
  	dollar: dollar,
  	Dopf: Dopf,
  	dopf: dopf,
  	Dot: Dot,
  	dot: dot,
  	DotDot: DotDot,
  	doteq: doteq,
  	doteqdot: doteqdot,
  	DotEqual: DotEqual,
  	dotminus: dotminus,
  	dotplus: dotplus,
  	dotsquare: dotsquare,
  	doublebarwedge: doublebarwedge,
  	DoubleContourIntegral: DoubleContourIntegral,
  	DoubleDot: DoubleDot,
  	DoubleDownArrow: DoubleDownArrow,
  	DoubleLeftArrow: DoubleLeftArrow,
  	DoubleLeftRightArrow: DoubleLeftRightArrow,
  	DoubleLeftTee: DoubleLeftTee,
  	DoubleLongLeftArrow: DoubleLongLeftArrow,
  	DoubleLongLeftRightArrow: DoubleLongLeftRightArrow,
  	DoubleLongRightArrow: DoubleLongRightArrow,
  	DoubleRightArrow: DoubleRightArrow,
  	DoubleRightTee: DoubleRightTee,
  	DoubleUpArrow: DoubleUpArrow,
  	DoubleUpDownArrow: DoubleUpDownArrow,
  	DoubleVerticalBar: DoubleVerticalBar,
  	DownArrowBar: DownArrowBar,
  	downarrow: downarrow,
  	DownArrow: DownArrow,
  	Downarrow: Downarrow,
  	DownArrowUpArrow: DownArrowUpArrow,
  	DownBreve: DownBreve,
  	downdownarrows: downdownarrows,
  	downharpoonleft: downharpoonleft,
  	downharpoonright: downharpoonright,
  	DownLeftRightVector: DownLeftRightVector,
  	DownLeftTeeVector: DownLeftTeeVector,
  	DownLeftVectorBar: DownLeftVectorBar,
  	DownLeftVector: DownLeftVector,
  	DownRightTeeVector: DownRightTeeVector,
  	DownRightVectorBar: DownRightVectorBar,
  	DownRightVector: DownRightVector,
  	DownTeeArrow: DownTeeArrow,
  	DownTee: DownTee,
  	drbkarow: drbkarow,
  	drcorn: drcorn,
  	drcrop: drcrop,
  	Dscr: Dscr,
  	dscr: dscr,
  	DScy: DScy,
  	dscy: dscy,
  	dsol: dsol,
  	Dstrok: Dstrok,
  	dstrok: dstrok,
  	dtdot: dtdot,
  	dtri: dtri,
  	dtrif: dtrif,
  	duarr: duarr,
  	duhar: duhar,
  	dwangle: dwangle,
  	DZcy: DZcy,
  	dzcy: dzcy,
  	dzigrarr: dzigrarr,
  	Eacute: Eacute,
  	eacute: eacute,
  	easter: easter,
  	Ecaron: Ecaron,
  	ecaron: ecaron,
  	Ecirc: Ecirc,
  	ecirc: ecirc,
  	ecir: ecir,
  	ecolon: ecolon,
  	Ecy: Ecy,
  	ecy: ecy,
  	eDDot: eDDot,
  	Edot: Edot,
  	edot: edot,
  	eDot: eDot,
  	ee: ee,
  	efDot: efDot,
  	Efr: Efr,
  	efr: efr,
  	eg: eg,
  	Egrave: Egrave,
  	egrave: egrave,
  	egs: egs,
  	egsdot: egsdot,
  	el: el,
  	Element: Element,
  	elinters: elinters,
  	ell: ell,
  	els: els,
  	elsdot: elsdot,
  	Emacr: Emacr,
  	emacr: emacr,
  	empty: empty,
  	emptyset: emptyset,
  	EmptySmallSquare: EmptySmallSquare,
  	emptyv: emptyv,
  	EmptyVerySmallSquare: EmptyVerySmallSquare,
  	emsp13: emsp13,
  	emsp14: emsp14,
  	emsp: emsp,
  	ENG: ENG,
  	eng: eng,
  	ensp: ensp,
  	Eogon: Eogon,
  	eogon: eogon,
  	Eopf: Eopf,
  	eopf: eopf,
  	epar: epar,
  	eparsl: eparsl,
  	eplus: eplus,
  	epsi: epsi,
  	Epsilon: Epsilon,
  	epsilon: epsilon,
  	epsiv: epsiv,
  	eqcirc: eqcirc,
  	eqcolon: eqcolon,
  	eqsim: eqsim,
  	eqslantgtr: eqslantgtr,
  	eqslantless: eqslantless,
  	Equal: Equal,
  	equals: equals,
  	EqualTilde: EqualTilde,
  	equest: equest,
  	Equilibrium: Equilibrium,
  	equiv: equiv,
  	equivDD: equivDD,
  	eqvparsl: eqvparsl,
  	erarr: erarr,
  	erDot: erDot,
  	escr: escr,
  	Escr: Escr,
  	esdot: esdot,
  	Esim: Esim,
  	esim: esim,
  	Eta: Eta,
  	eta: eta,
  	ETH: ETH,
  	eth: eth,
  	Euml: Euml,
  	euml: euml,
  	euro: euro,
  	excl: excl,
  	exist: exist,
  	Exists: Exists,
  	expectation: expectation,
  	exponentiale: exponentiale,
  	ExponentialE: ExponentialE,
  	fallingdotseq: fallingdotseq,
  	Fcy: Fcy,
  	fcy: fcy,
  	female: female,
  	ffilig: ffilig,
  	fflig: fflig,
  	ffllig: ffllig,
  	Ffr: Ffr,
  	ffr: ffr,
  	filig: filig,
  	FilledSmallSquare: FilledSmallSquare,
  	FilledVerySmallSquare: FilledVerySmallSquare,
  	fjlig: fjlig,
  	flat: flat,
  	fllig: fllig,
  	fltns: fltns,
  	fnof: fnof,
  	Fopf: Fopf,
  	fopf: fopf,
  	forall: forall,
  	ForAll: ForAll,
  	fork: fork,
  	forkv: forkv,
  	Fouriertrf: Fouriertrf,
  	fpartint: fpartint,
  	frac12: frac12,
  	frac13: frac13,
  	frac14: frac14,
  	frac15: frac15,
  	frac16: frac16,
  	frac18: frac18,
  	frac23: frac23,
  	frac25: frac25,
  	frac34: frac34,
  	frac35: frac35,
  	frac38: frac38,
  	frac45: frac45,
  	frac56: frac56,
  	frac58: frac58,
  	frac78: frac78,
  	frasl: frasl,
  	frown: frown,
  	fscr: fscr,
  	Fscr: Fscr,
  	gacute: gacute,
  	Gamma: Gamma,
  	gamma: gamma,
  	Gammad: Gammad,
  	gammad: gammad,
  	gap: gap,
  	Gbreve: Gbreve,
  	gbreve: gbreve,
  	Gcedil: Gcedil,
  	Gcirc: Gcirc,
  	gcirc: gcirc,
  	Gcy: Gcy,
  	gcy: gcy,
  	Gdot: Gdot,
  	gdot: gdot,
  	ge: ge,
  	gE: gE,
  	gEl: gEl,
  	gel: gel,
  	geq: geq,
  	geqq: geqq,
  	geqslant: geqslant,
  	gescc: gescc,
  	ges: ges,
  	gesdot: gesdot,
  	gesdoto: gesdoto,
  	gesdotol: gesdotol,
  	gesl: gesl,
  	gesles: gesles,
  	Gfr: Gfr,
  	gfr: gfr,
  	gg: gg,
  	Gg: Gg,
  	ggg: ggg,
  	gimel: gimel,
  	GJcy: GJcy,
  	gjcy: gjcy,
  	gla: gla,
  	gl: gl,
  	glE: glE,
  	glj: glj,
  	gnap: gnap,
  	gnapprox: gnapprox,
  	gne: gne,
  	gnE: gnE,
  	gneq: gneq,
  	gneqq: gneqq,
  	gnsim: gnsim,
  	Gopf: Gopf,
  	gopf: gopf,
  	grave: grave,
  	GreaterEqual: GreaterEqual,
  	GreaterEqualLess: GreaterEqualLess,
  	GreaterFullEqual: GreaterFullEqual,
  	GreaterGreater: GreaterGreater,
  	GreaterLess: GreaterLess,
  	GreaterSlantEqual: GreaterSlantEqual,
  	GreaterTilde: GreaterTilde,
  	Gscr: Gscr,
  	gscr: gscr,
  	gsim: gsim,
  	gsime: gsime,
  	gsiml: gsiml,
  	gtcc: gtcc,
  	gtcir: gtcir,
  	gt: gt$1,
  	GT: GT,
  	Gt: Gt,
  	gtdot: gtdot,
  	gtlPar: gtlPar,
  	gtquest: gtquest,
  	gtrapprox: gtrapprox,
  	gtrarr: gtrarr,
  	gtrdot: gtrdot,
  	gtreqless: gtreqless,
  	gtreqqless: gtreqqless,
  	gtrless: gtrless,
  	gtrsim: gtrsim,
  	gvertneqq: gvertneqq,
  	gvnE: gvnE,
  	Hacek: Hacek,
  	hairsp: hairsp,
  	half: half,
  	hamilt: hamilt,
  	HARDcy: HARDcy,
  	hardcy: hardcy,
  	harrcir: harrcir,
  	harr: harr,
  	hArr: hArr,
  	harrw: harrw,
  	Hat: Hat,
  	hbar: hbar,
  	Hcirc: Hcirc,
  	hcirc: hcirc,
  	hearts: hearts,
  	heartsuit: heartsuit,
  	hellip: hellip,
  	hercon: hercon,
  	hfr: hfr,
  	Hfr: Hfr,
  	HilbertSpace: HilbertSpace,
  	hksearow: hksearow,
  	hkswarow: hkswarow,
  	hoarr: hoarr,
  	homtht: homtht,
  	hookleftarrow: hookleftarrow,
  	hookrightarrow: hookrightarrow,
  	hopf: hopf,
  	Hopf: Hopf,
  	horbar: horbar,
  	HorizontalLine: HorizontalLine,
  	hscr: hscr,
  	Hscr: Hscr,
  	hslash: hslash,
  	Hstrok: Hstrok,
  	hstrok: hstrok,
  	HumpDownHump: HumpDownHump,
  	HumpEqual: HumpEqual,
  	hybull: hybull,
  	hyphen: hyphen,
  	Iacute: Iacute,
  	iacute: iacute,
  	ic: ic,
  	Icirc: Icirc,
  	icirc: icirc,
  	Icy: Icy,
  	icy: icy,
  	Idot: Idot,
  	IEcy: IEcy,
  	iecy: iecy,
  	iexcl: iexcl,
  	iff: iff,
  	ifr: ifr,
  	Ifr: Ifr,
  	Igrave: Igrave,
  	igrave: igrave,
  	ii: ii,
  	iiiint: iiiint,
  	iiint: iiint,
  	iinfin: iinfin,
  	iiota: iiota,
  	IJlig: IJlig,
  	ijlig: ijlig,
  	Imacr: Imacr,
  	imacr: imacr,
  	image: image,
  	ImaginaryI: ImaginaryI,
  	imagline: imagline,
  	imagpart: imagpart,
  	imath: imath,
  	Im: Im,
  	imof: imof,
  	imped: imped,
  	Implies: Implies,
  	incare: incare,
  	infin: infin,
  	infintie: infintie,
  	inodot: inodot,
  	intcal: intcal,
  	int: int,
  	Int: Int,
  	integers: integers,
  	Integral: Integral,
  	intercal: intercal,
  	Intersection: Intersection,
  	intlarhk: intlarhk,
  	intprod: intprod,
  	InvisibleComma: InvisibleComma,
  	InvisibleTimes: InvisibleTimes,
  	IOcy: IOcy,
  	iocy: iocy,
  	Iogon: Iogon,
  	iogon: iogon,
  	Iopf: Iopf,
  	iopf: iopf,
  	Iota: Iota,
  	iota: iota,
  	iprod: iprod,
  	iquest: iquest,
  	iscr: iscr,
  	Iscr: Iscr,
  	isin: isin,
  	isindot: isindot,
  	isinE: isinE,
  	isins: isins,
  	isinsv: isinsv,
  	isinv: isinv,
  	it: it,
  	Itilde: Itilde,
  	itilde: itilde,
  	Iukcy: Iukcy,
  	iukcy: iukcy,
  	Iuml: Iuml,
  	iuml: iuml,
  	Jcirc: Jcirc,
  	jcirc: jcirc,
  	Jcy: Jcy,
  	jcy: jcy,
  	Jfr: Jfr,
  	jfr: jfr,
  	jmath: jmath,
  	Jopf: Jopf,
  	jopf: jopf,
  	Jscr: Jscr,
  	jscr: jscr,
  	Jsercy: Jsercy,
  	jsercy: jsercy,
  	Jukcy: Jukcy,
  	jukcy: jukcy,
  	Kappa: Kappa,
  	kappa: kappa,
  	kappav: kappav,
  	Kcedil: Kcedil,
  	kcedil: kcedil,
  	Kcy: Kcy,
  	kcy: kcy,
  	Kfr: Kfr,
  	kfr: kfr,
  	kgreen: kgreen,
  	KHcy: KHcy,
  	khcy: khcy,
  	KJcy: KJcy,
  	kjcy: kjcy,
  	Kopf: Kopf,
  	kopf: kopf,
  	Kscr: Kscr,
  	kscr: kscr,
  	lAarr: lAarr,
  	Lacute: Lacute,
  	lacute: lacute,
  	laemptyv: laemptyv,
  	lagran: lagran,
  	Lambda: Lambda,
  	lambda: lambda,
  	lang: lang,
  	Lang: Lang,
  	langd: langd,
  	langle: langle,
  	lap: lap,
  	Laplacetrf: Laplacetrf,
  	laquo: laquo,
  	larrb: larrb,
  	larrbfs: larrbfs,
  	larr: larr,
  	Larr: Larr,
  	lArr: lArr,
  	larrfs: larrfs,
  	larrhk: larrhk,
  	larrlp: larrlp,
  	larrpl: larrpl,
  	larrsim: larrsim,
  	larrtl: larrtl,
  	latail: latail,
  	lAtail: lAtail,
  	lat: lat,
  	late: late,
  	lates: lates,
  	lbarr: lbarr,
  	lBarr: lBarr,
  	lbbrk: lbbrk,
  	lbrace: lbrace,
  	lbrack: lbrack,
  	lbrke: lbrke,
  	lbrksld: lbrksld,
  	lbrkslu: lbrkslu,
  	Lcaron: Lcaron,
  	lcaron: lcaron,
  	Lcedil: Lcedil,
  	lcedil: lcedil,
  	lceil: lceil,
  	lcub: lcub,
  	Lcy: Lcy,
  	lcy: lcy,
  	ldca: ldca,
  	ldquo: ldquo,
  	ldquor: ldquor,
  	ldrdhar: ldrdhar,
  	ldrushar: ldrushar,
  	ldsh: ldsh,
  	le: le,
  	lE: lE,
  	LeftAngleBracket: LeftAngleBracket,
  	LeftArrowBar: LeftArrowBar,
  	leftarrow: leftarrow,
  	LeftArrow: LeftArrow,
  	Leftarrow: Leftarrow,
  	LeftArrowRightArrow: LeftArrowRightArrow,
  	leftarrowtail: leftarrowtail,
  	LeftCeiling: LeftCeiling,
  	LeftDoubleBracket: LeftDoubleBracket,
  	LeftDownTeeVector: LeftDownTeeVector,
  	LeftDownVectorBar: LeftDownVectorBar,
  	LeftDownVector: LeftDownVector,
  	LeftFloor: LeftFloor,
  	leftharpoondown: leftharpoondown,
  	leftharpoonup: leftharpoonup,
  	leftleftarrows: leftleftarrows,
  	leftrightarrow: leftrightarrow,
  	LeftRightArrow: LeftRightArrow,
  	Leftrightarrow: Leftrightarrow,
  	leftrightarrows: leftrightarrows,
  	leftrightharpoons: leftrightharpoons,
  	leftrightsquigarrow: leftrightsquigarrow,
  	LeftRightVector: LeftRightVector,
  	LeftTeeArrow: LeftTeeArrow,
  	LeftTee: LeftTee,
  	LeftTeeVector: LeftTeeVector,
  	leftthreetimes: leftthreetimes,
  	LeftTriangleBar: LeftTriangleBar,
  	LeftTriangle: LeftTriangle,
  	LeftTriangleEqual: LeftTriangleEqual,
  	LeftUpDownVector: LeftUpDownVector,
  	LeftUpTeeVector: LeftUpTeeVector,
  	LeftUpVectorBar: LeftUpVectorBar,
  	LeftUpVector: LeftUpVector,
  	LeftVectorBar: LeftVectorBar,
  	LeftVector: LeftVector,
  	lEg: lEg,
  	leg: leg,
  	leq: leq,
  	leqq: leqq,
  	leqslant: leqslant,
  	lescc: lescc,
  	les: les,
  	lesdot: lesdot,
  	lesdoto: lesdoto,
  	lesdotor: lesdotor,
  	lesg: lesg,
  	lesges: lesges,
  	lessapprox: lessapprox,
  	lessdot: lessdot,
  	lesseqgtr: lesseqgtr,
  	lesseqqgtr: lesseqqgtr,
  	LessEqualGreater: LessEqualGreater,
  	LessFullEqual: LessFullEqual,
  	LessGreater: LessGreater,
  	lessgtr: lessgtr,
  	LessLess: LessLess,
  	lesssim: lesssim,
  	LessSlantEqual: LessSlantEqual,
  	LessTilde: LessTilde,
  	lfisht: lfisht,
  	lfloor: lfloor,
  	Lfr: Lfr,
  	lfr: lfr,
  	lg: lg,
  	lgE: lgE,
  	lHar: lHar,
  	lhard: lhard,
  	lharu: lharu,
  	lharul: lharul,
  	lhblk: lhblk,
  	LJcy: LJcy,
  	ljcy: ljcy,
  	llarr: llarr,
  	ll: ll,
  	Ll: Ll,
  	llcorner: llcorner,
  	Lleftarrow: Lleftarrow,
  	llhard: llhard,
  	lltri: lltri,
  	Lmidot: Lmidot,
  	lmidot: lmidot,
  	lmoustache: lmoustache,
  	lmoust: lmoust,
  	lnap: lnap,
  	lnapprox: lnapprox,
  	lne: lne,
  	lnE: lnE,
  	lneq: lneq,
  	lneqq: lneqq,
  	lnsim: lnsim,
  	loang: loang,
  	loarr: loarr,
  	lobrk: lobrk,
  	longleftarrow: longleftarrow,
  	LongLeftArrow: LongLeftArrow,
  	Longleftarrow: Longleftarrow,
  	longleftrightarrow: longleftrightarrow,
  	LongLeftRightArrow: LongLeftRightArrow,
  	Longleftrightarrow: Longleftrightarrow,
  	longmapsto: longmapsto,
  	longrightarrow: longrightarrow,
  	LongRightArrow: LongRightArrow,
  	Longrightarrow: Longrightarrow,
  	looparrowleft: looparrowleft,
  	looparrowright: looparrowright,
  	lopar: lopar,
  	Lopf: Lopf,
  	lopf: lopf,
  	loplus: loplus,
  	lotimes: lotimes,
  	lowast: lowast,
  	lowbar: lowbar,
  	LowerLeftArrow: LowerLeftArrow,
  	LowerRightArrow: LowerRightArrow,
  	loz: loz,
  	lozenge: lozenge,
  	lozf: lozf,
  	lpar: lpar,
  	lparlt: lparlt,
  	lrarr: lrarr,
  	lrcorner: lrcorner,
  	lrhar: lrhar,
  	lrhard: lrhard,
  	lrm: lrm,
  	lrtri: lrtri,
  	lsaquo: lsaquo,
  	lscr: lscr,
  	Lscr: Lscr,
  	lsh: lsh,
  	Lsh: Lsh,
  	lsim: lsim,
  	lsime: lsime,
  	lsimg: lsimg,
  	lsqb: lsqb,
  	lsquo: lsquo,
  	lsquor: lsquor,
  	Lstrok: Lstrok,
  	lstrok: lstrok,
  	ltcc: ltcc,
  	ltcir: ltcir,
  	lt: lt$1,
  	LT: LT,
  	Lt: Lt,
  	ltdot: ltdot,
  	lthree: lthree,
  	ltimes: ltimes,
  	ltlarr: ltlarr,
  	ltquest: ltquest,
  	ltri: ltri,
  	ltrie: ltrie,
  	ltrif: ltrif,
  	ltrPar: ltrPar,
  	lurdshar: lurdshar,
  	luruhar: luruhar,
  	lvertneqq: lvertneqq,
  	lvnE: lvnE,
  	macr: macr,
  	male: male,
  	malt: malt,
  	maltese: maltese,
  	map: map$1,
  	mapsto: mapsto,
  	mapstodown: mapstodown,
  	mapstoleft: mapstoleft,
  	mapstoup: mapstoup,
  	marker: marker,
  	mcomma: mcomma,
  	Mcy: Mcy,
  	mcy: mcy,
  	mdash: mdash,
  	mDDot: mDDot,
  	measuredangle: measuredangle,
  	MediumSpace: MediumSpace,
  	Mellintrf: Mellintrf,
  	Mfr: Mfr,
  	mfr: mfr,
  	mho: mho,
  	micro: micro,
  	midast: midast,
  	midcir: midcir,
  	mid: mid,
  	middot: middot,
  	minusb: minusb,
  	minus: minus,
  	minusd: minusd,
  	minusdu: minusdu,
  	MinusPlus: MinusPlus,
  	mlcp: mlcp,
  	mldr: mldr,
  	mnplus: mnplus,
  	models: models,
  	Mopf: Mopf,
  	mopf: mopf,
  	mp: mp,
  	mscr: mscr,
  	Mscr: Mscr,
  	mstpos: mstpos,
  	Mu: Mu,
  	mu: mu,
  	multimap: multimap,
  	mumap: mumap,
  	nabla: nabla,
  	Nacute: Nacute,
  	nacute: nacute,
  	nang: nang,
  	nap: nap,
  	napE: napE,
  	napid: napid,
  	napos: napos,
  	napprox: napprox,
  	natural: natural,
  	naturals: naturals,
  	natur: natur,
  	nbsp: nbsp,
  	nbump: nbump,
  	nbumpe: nbumpe,
  	ncap: ncap,
  	Ncaron: Ncaron,
  	ncaron: ncaron,
  	Ncedil: Ncedil,
  	ncedil: ncedil,
  	ncong: ncong,
  	ncongdot: ncongdot,
  	ncup: ncup,
  	Ncy: Ncy,
  	ncy: ncy,
  	ndash: ndash,
  	nearhk: nearhk,
  	nearr: nearr,
  	neArr: neArr,
  	nearrow: nearrow,
  	ne: ne,
  	nedot: nedot,
  	NegativeMediumSpace: NegativeMediumSpace,
  	NegativeThickSpace: NegativeThickSpace,
  	NegativeThinSpace: NegativeThinSpace,
  	NegativeVeryThinSpace: NegativeVeryThinSpace,
  	nequiv: nequiv,
  	nesear: nesear,
  	nesim: nesim,
  	NestedGreaterGreater: NestedGreaterGreater,
  	NestedLessLess: NestedLessLess,
  	NewLine: NewLine,
  	nexist: nexist,
  	nexists: nexists,
  	Nfr: Nfr,
  	nfr: nfr,
  	ngE: ngE,
  	nge: nge,
  	ngeq: ngeq,
  	ngeqq: ngeqq,
  	ngeqslant: ngeqslant,
  	nges: nges,
  	nGg: nGg,
  	ngsim: ngsim,
  	nGt: nGt,
  	ngt: ngt,
  	ngtr: ngtr,
  	nGtv: nGtv,
  	nharr: nharr,
  	nhArr: nhArr,
  	nhpar: nhpar,
  	ni: ni,
  	nis: nis,
  	nisd: nisd,
  	niv: niv,
  	NJcy: NJcy,
  	njcy: njcy,
  	nlarr: nlarr,
  	nlArr: nlArr,
  	nldr: nldr,
  	nlE: nlE,
  	nle: nle,
  	nleftarrow: nleftarrow,
  	nLeftarrow: nLeftarrow,
  	nleftrightarrow: nleftrightarrow,
  	nLeftrightarrow: nLeftrightarrow,
  	nleq: nleq,
  	nleqq: nleqq,
  	nleqslant: nleqslant,
  	nles: nles,
  	nless: nless,
  	nLl: nLl,
  	nlsim: nlsim,
  	nLt: nLt,
  	nlt: nlt,
  	nltri: nltri,
  	nltrie: nltrie,
  	nLtv: nLtv,
  	nmid: nmid,
  	NoBreak: NoBreak,
  	NonBreakingSpace: NonBreakingSpace,
  	nopf: nopf,
  	Nopf: Nopf,
  	Not: Not,
  	not: not,
  	NotCongruent: NotCongruent,
  	NotCupCap: NotCupCap,
  	NotDoubleVerticalBar: NotDoubleVerticalBar,
  	NotElement: NotElement,
  	NotEqual: NotEqual,
  	NotEqualTilde: NotEqualTilde,
  	NotExists: NotExists,
  	NotGreater: NotGreater,
  	NotGreaterEqual: NotGreaterEqual,
  	NotGreaterFullEqual: NotGreaterFullEqual,
  	NotGreaterGreater: NotGreaterGreater,
  	NotGreaterLess: NotGreaterLess,
  	NotGreaterSlantEqual: NotGreaterSlantEqual,
  	NotGreaterTilde: NotGreaterTilde,
  	NotHumpDownHump: NotHumpDownHump,
  	NotHumpEqual: NotHumpEqual,
  	notin: notin,
  	notindot: notindot,
  	notinE: notinE,
  	notinva: notinva,
  	notinvb: notinvb,
  	notinvc: notinvc,
  	NotLeftTriangleBar: NotLeftTriangleBar,
  	NotLeftTriangle: NotLeftTriangle,
  	NotLeftTriangleEqual: NotLeftTriangleEqual,
  	NotLess: NotLess,
  	NotLessEqual: NotLessEqual,
  	NotLessGreater: NotLessGreater,
  	NotLessLess: NotLessLess,
  	NotLessSlantEqual: NotLessSlantEqual,
  	NotLessTilde: NotLessTilde,
  	NotNestedGreaterGreater: NotNestedGreaterGreater,
  	NotNestedLessLess: NotNestedLessLess,
  	notni: notni,
  	notniva: notniva,
  	notnivb: notnivb,
  	notnivc: notnivc,
  	NotPrecedes: NotPrecedes,
  	NotPrecedesEqual: NotPrecedesEqual,
  	NotPrecedesSlantEqual: NotPrecedesSlantEqual,
  	NotReverseElement: NotReverseElement,
  	NotRightTriangleBar: NotRightTriangleBar,
  	NotRightTriangle: NotRightTriangle,
  	NotRightTriangleEqual: NotRightTriangleEqual,
  	NotSquareSubset: NotSquareSubset,
  	NotSquareSubsetEqual: NotSquareSubsetEqual,
  	NotSquareSuperset: NotSquareSuperset,
  	NotSquareSupersetEqual: NotSquareSupersetEqual,
  	NotSubset: NotSubset,
  	NotSubsetEqual: NotSubsetEqual,
  	NotSucceeds: NotSucceeds,
  	NotSucceedsEqual: NotSucceedsEqual,
  	NotSucceedsSlantEqual: NotSucceedsSlantEqual,
  	NotSucceedsTilde: NotSucceedsTilde,
  	NotSuperset: NotSuperset,
  	NotSupersetEqual: NotSupersetEqual,
  	NotTilde: NotTilde,
  	NotTildeEqual: NotTildeEqual,
  	NotTildeFullEqual: NotTildeFullEqual,
  	NotTildeTilde: NotTildeTilde,
  	NotVerticalBar: NotVerticalBar,
  	nparallel: nparallel,
  	npar: npar,
  	nparsl: nparsl,
  	npart: npart,
  	npolint: npolint,
  	npr: npr,
  	nprcue: nprcue,
  	nprec: nprec,
  	npreceq: npreceq,
  	npre: npre,
  	nrarrc: nrarrc,
  	nrarr: nrarr,
  	nrArr: nrArr,
  	nrarrw: nrarrw,
  	nrightarrow: nrightarrow,
  	nRightarrow: nRightarrow,
  	nrtri: nrtri,
  	nrtrie: nrtrie,
  	nsc: nsc,
  	nsccue: nsccue,
  	nsce: nsce,
  	Nscr: Nscr,
  	nscr: nscr,
  	nshortmid: nshortmid,
  	nshortparallel: nshortparallel,
  	nsim: nsim,
  	nsime: nsime,
  	nsimeq: nsimeq,
  	nsmid: nsmid,
  	nspar: nspar,
  	nsqsube: nsqsube,
  	nsqsupe: nsqsupe,
  	nsub: nsub,
  	nsubE: nsubE,
  	nsube: nsube,
  	nsubset: nsubset,
  	nsubseteq: nsubseteq,
  	nsubseteqq: nsubseteqq,
  	nsucc: nsucc,
  	nsucceq: nsucceq,
  	nsup: nsup,
  	nsupE: nsupE,
  	nsupe: nsupe,
  	nsupset: nsupset,
  	nsupseteq: nsupseteq,
  	nsupseteqq: nsupseteqq,
  	ntgl: ntgl,
  	Ntilde: Ntilde,
  	ntilde: ntilde,
  	ntlg: ntlg,
  	ntriangleleft: ntriangleleft,
  	ntrianglelefteq: ntrianglelefteq,
  	ntriangleright: ntriangleright,
  	ntrianglerighteq: ntrianglerighteq,
  	Nu: Nu,
  	nu: nu,
  	num: num,
  	numero: numero,
  	numsp: numsp,
  	nvap: nvap,
  	nvdash: nvdash,
  	nvDash: nvDash,
  	nVdash: nVdash,
  	nVDash: nVDash,
  	nvge: nvge,
  	nvgt: nvgt,
  	nvHarr: nvHarr,
  	nvinfin: nvinfin,
  	nvlArr: nvlArr,
  	nvle: nvle,
  	nvlt: nvlt,
  	nvltrie: nvltrie,
  	nvrArr: nvrArr,
  	nvrtrie: nvrtrie,
  	nvsim: nvsim,
  	nwarhk: nwarhk,
  	nwarr: nwarr,
  	nwArr: nwArr,
  	nwarrow: nwarrow,
  	nwnear: nwnear,
  	Oacute: Oacute,
  	oacute: oacute,
  	oast: oast,
  	Ocirc: Ocirc,
  	ocirc: ocirc,
  	ocir: ocir,
  	Ocy: Ocy,
  	ocy: ocy,
  	odash: odash,
  	Odblac: Odblac,
  	odblac: odblac,
  	odiv: odiv,
  	odot: odot,
  	odsold: odsold,
  	OElig: OElig,
  	oelig: oelig,
  	ofcir: ofcir,
  	Ofr: Ofr,
  	ofr: ofr,
  	ogon: ogon,
  	Ograve: Ograve,
  	ograve: ograve,
  	ogt: ogt,
  	ohbar: ohbar,
  	ohm: ohm,
  	oint: oint,
  	olarr: olarr,
  	olcir: olcir,
  	olcross: olcross,
  	oline: oline,
  	olt: olt,
  	Omacr: Omacr,
  	omacr: omacr,
  	Omega: Omega,
  	omega: omega,
  	Omicron: Omicron,
  	omicron: omicron,
  	omid: omid,
  	ominus: ominus,
  	Oopf: Oopf,
  	oopf: oopf,
  	opar: opar,
  	OpenCurlyDoubleQuote: OpenCurlyDoubleQuote,
  	OpenCurlyQuote: OpenCurlyQuote,
  	operp: operp,
  	oplus: oplus,
  	orarr: orarr,
  	Or: Or,
  	or: or,
  	ord: ord,
  	order: order,
  	orderof: orderof,
  	ordf: ordf,
  	ordm: ordm,
  	origof: origof,
  	oror: oror,
  	orslope: orslope,
  	orv: orv,
  	oS: oS,
  	Oscr: Oscr,
  	oscr: oscr,
  	Oslash: Oslash,
  	oslash: oslash,
  	osol: osol,
  	Otilde: Otilde,
  	otilde: otilde,
  	otimesas: otimesas,
  	Otimes: Otimes,
  	otimes: otimes,
  	Ouml: Ouml,
  	ouml: ouml,
  	ovbar: ovbar,
  	OverBar: OverBar,
  	OverBrace: OverBrace,
  	OverBracket: OverBracket,
  	OverParenthesis: OverParenthesis,
  	para: para,
  	parallel: parallel,
  	par: par,
  	parsim: parsim,
  	parsl: parsl,
  	part: part,
  	PartialD: PartialD,
  	Pcy: Pcy,
  	pcy: pcy,
  	percnt: percnt,
  	period: period,
  	permil: permil,
  	perp: perp,
  	pertenk: pertenk,
  	Pfr: Pfr,
  	pfr: pfr,
  	Phi: Phi,
  	phi: phi,
  	phiv: phiv,
  	phmmat: phmmat,
  	phone: phone,
  	Pi: Pi,
  	pi: pi,
  	pitchfork: pitchfork,
  	piv: piv,
  	planck: planck,
  	planckh: planckh,
  	plankv: plankv,
  	plusacir: plusacir,
  	plusb: plusb,
  	pluscir: pluscir,
  	plus: plus,
  	plusdo: plusdo,
  	plusdu: plusdu,
  	pluse: pluse,
  	PlusMinus: PlusMinus,
  	plusmn: plusmn,
  	plussim: plussim,
  	plustwo: plustwo,
  	pm: pm,
  	Poincareplane: Poincareplane,
  	pointint: pointint,
  	popf: popf,
  	Popf: Popf,
  	pound: pound,
  	prap: prap,
  	Pr: Pr,
  	pr: pr,
  	prcue: prcue,
  	precapprox: precapprox,
  	prec: prec,
  	preccurlyeq: preccurlyeq,
  	Precedes: Precedes,
  	PrecedesEqual: PrecedesEqual,
  	PrecedesSlantEqual: PrecedesSlantEqual,
  	PrecedesTilde: PrecedesTilde,
  	preceq: preceq,
  	precnapprox: precnapprox,
  	precneqq: precneqq,
  	precnsim: precnsim,
  	pre: pre,
  	prE: prE,
  	precsim: precsim,
  	prime: prime,
  	Prime: Prime,
  	primes: primes,
  	prnap: prnap,
  	prnE: prnE,
  	prnsim: prnsim,
  	prod: prod,
  	Product: Product,
  	profalar: profalar,
  	profline: profline,
  	profsurf: profsurf,
  	prop: prop,
  	Proportional: Proportional,
  	Proportion: Proportion,
  	propto: propto,
  	prsim: prsim,
  	prurel: prurel,
  	Pscr: Pscr,
  	pscr: pscr,
  	Psi: Psi,
  	psi: psi,
  	puncsp: puncsp,
  	Qfr: Qfr,
  	qfr: qfr,
  	qint: qint,
  	qopf: qopf,
  	Qopf: Qopf,
  	qprime: qprime,
  	Qscr: Qscr,
  	qscr: qscr,
  	quaternions: quaternions,
  	quatint: quatint,
  	quest: quest,
  	questeq: questeq,
  	quot: quot$1,
  	QUOT: QUOT,
  	rAarr: rAarr,
  	race: race,
  	Racute: Racute,
  	racute: racute,
  	radic: radic,
  	raemptyv: raemptyv,
  	rang: rang,
  	Rang: Rang,
  	rangd: rangd,
  	range: range,
  	rangle: rangle,
  	raquo: raquo,
  	rarrap: rarrap,
  	rarrb: rarrb,
  	rarrbfs: rarrbfs,
  	rarrc: rarrc,
  	rarr: rarr,
  	Rarr: Rarr,
  	rArr: rArr,
  	rarrfs: rarrfs,
  	rarrhk: rarrhk,
  	rarrlp: rarrlp,
  	rarrpl: rarrpl,
  	rarrsim: rarrsim,
  	Rarrtl: Rarrtl,
  	rarrtl: rarrtl,
  	rarrw: rarrw,
  	ratail: ratail,
  	rAtail: rAtail,
  	ratio: ratio,
  	rationals: rationals,
  	rbarr: rbarr,
  	rBarr: rBarr,
  	RBarr: RBarr,
  	rbbrk: rbbrk,
  	rbrace: rbrace,
  	rbrack: rbrack,
  	rbrke: rbrke,
  	rbrksld: rbrksld,
  	rbrkslu: rbrkslu,
  	Rcaron: Rcaron,
  	rcaron: rcaron,
  	Rcedil: Rcedil,
  	rcedil: rcedil,
  	rceil: rceil,
  	rcub: rcub,
  	Rcy: Rcy,
  	rcy: rcy,
  	rdca: rdca,
  	rdldhar: rdldhar,
  	rdquo: rdquo,
  	rdquor: rdquor,
  	rdsh: rdsh,
  	real: real,
  	realine: realine,
  	realpart: realpart,
  	reals: reals,
  	Re: Re,
  	rect: rect,
  	reg: reg,
  	REG: REG,
  	ReverseElement: ReverseElement,
  	ReverseEquilibrium: ReverseEquilibrium,
  	ReverseUpEquilibrium: ReverseUpEquilibrium,
  	rfisht: rfisht,
  	rfloor: rfloor,
  	rfr: rfr,
  	Rfr: Rfr,
  	rHar: rHar,
  	rhard: rhard,
  	rharu: rharu,
  	rharul: rharul,
  	Rho: Rho,
  	rho: rho,
  	rhov: rhov,
  	RightAngleBracket: RightAngleBracket,
  	RightArrowBar: RightArrowBar,
  	rightarrow: rightarrow,
  	RightArrow: RightArrow,
  	Rightarrow: Rightarrow,
  	RightArrowLeftArrow: RightArrowLeftArrow,
  	rightarrowtail: rightarrowtail,
  	RightCeiling: RightCeiling,
  	RightDoubleBracket: RightDoubleBracket,
  	RightDownTeeVector: RightDownTeeVector,
  	RightDownVectorBar: RightDownVectorBar,
  	RightDownVector: RightDownVector,
  	RightFloor: RightFloor,
  	rightharpoondown: rightharpoondown,
  	rightharpoonup: rightharpoonup,
  	rightleftarrows: rightleftarrows,
  	rightleftharpoons: rightleftharpoons,
  	rightrightarrows: rightrightarrows,
  	rightsquigarrow: rightsquigarrow,
  	RightTeeArrow: RightTeeArrow,
  	RightTee: RightTee,
  	RightTeeVector: RightTeeVector,
  	rightthreetimes: rightthreetimes,
  	RightTriangleBar: RightTriangleBar,
  	RightTriangle: RightTriangle,
  	RightTriangleEqual: RightTriangleEqual,
  	RightUpDownVector: RightUpDownVector,
  	RightUpTeeVector: RightUpTeeVector,
  	RightUpVectorBar: RightUpVectorBar,
  	RightUpVector: RightUpVector,
  	RightVectorBar: RightVectorBar,
  	RightVector: RightVector,
  	ring: ring,
  	risingdotseq: risingdotseq,
  	rlarr: rlarr,
  	rlhar: rlhar,
  	rlm: rlm,
  	rmoustache: rmoustache,
  	rmoust: rmoust,
  	rnmid: rnmid,
  	roang: roang,
  	roarr: roarr,
  	robrk: robrk,
  	ropar: ropar,
  	ropf: ropf,
  	Ropf: Ropf,
  	roplus: roplus,
  	rotimes: rotimes,
  	RoundImplies: RoundImplies,
  	rpar: rpar,
  	rpargt: rpargt,
  	rppolint: rppolint,
  	rrarr: rrarr,
  	Rrightarrow: Rrightarrow,
  	rsaquo: rsaquo,
  	rscr: rscr,
  	Rscr: Rscr,
  	rsh: rsh,
  	Rsh: Rsh,
  	rsqb: rsqb,
  	rsquo: rsquo,
  	rsquor: rsquor,
  	rthree: rthree,
  	rtimes: rtimes,
  	rtri: rtri,
  	rtrie: rtrie,
  	rtrif: rtrif,
  	rtriltri: rtriltri,
  	RuleDelayed: RuleDelayed,
  	ruluhar: ruluhar,
  	rx: rx,
  	Sacute: Sacute,
  	sacute: sacute,
  	sbquo: sbquo,
  	scap: scap,
  	Scaron: Scaron,
  	scaron: scaron,
  	Sc: Sc,
  	sc: sc,
  	sccue: sccue,
  	sce: sce,
  	scE: scE,
  	Scedil: Scedil,
  	scedil: scedil,
  	Scirc: Scirc,
  	scirc: scirc,
  	scnap: scnap,
  	scnE: scnE,
  	scnsim: scnsim,
  	scpolint: scpolint,
  	scsim: scsim,
  	Scy: Scy,
  	scy: scy,
  	sdotb: sdotb,
  	sdot: sdot,
  	sdote: sdote,
  	searhk: searhk,
  	searr: searr,
  	seArr: seArr,
  	searrow: searrow,
  	sect: sect,
  	semi: semi,
  	seswar: seswar,
  	setminus: setminus,
  	setmn: setmn,
  	sext: sext,
  	Sfr: Sfr,
  	sfr: sfr,
  	sfrown: sfrown,
  	sharp: sharp,
  	SHCHcy: SHCHcy,
  	shchcy: shchcy,
  	SHcy: SHcy,
  	shcy: shcy,
  	ShortDownArrow: ShortDownArrow,
  	ShortLeftArrow: ShortLeftArrow,
  	shortmid: shortmid,
  	shortparallel: shortparallel,
  	ShortRightArrow: ShortRightArrow,
  	ShortUpArrow: ShortUpArrow,
  	shy: shy,
  	Sigma: Sigma,
  	sigma: sigma,
  	sigmaf: sigmaf,
  	sigmav: sigmav,
  	sim: sim,
  	simdot: simdot,
  	sime: sime,
  	simeq: simeq,
  	simg: simg,
  	simgE: simgE,
  	siml: siml,
  	simlE: simlE,
  	simne: simne,
  	simplus: simplus,
  	simrarr: simrarr,
  	slarr: slarr,
  	SmallCircle: SmallCircle,
  	smallsetminus: smallsetminus,
  	smashp: smashp,
  	smeparsl: smeparsl,
  	smid: smid,
  	smile: smile,
  	smt: smt,
  	smte: smte,
  	smtes: smtes,
  	SOFTcy: SOFTcy,
  	softcy: softcy,
  	solbar: solbar,
  	solb: solb,
  	sol: sol,
  	Sopf: Sopf,
  	sopf: sopf,
  	spades: spades,
  	spadesuit: spadesuit,
  	spar: spar,
  	sqcap: sqcap,
  	sqcaps: sqcaps,
  	sqcup: sqcup,
  	sqcups: sqcups,
  	Sqrt: Sqrt,
  	sqsub: sqsub,
  	sqsube: sqsube,
  	sqsubset: sqsubset,
  	sqsubseteq: sqsubseteq,
  	sqsup: sqsup,
  	sqsupe: sqsupe,
  	sqsupset: sqsupset,
  	sqsupseteq: sqsupseteq,
  	square: square,
  	Square: Square,
  	SquareIntersection: SquareIntersection,
  	SquareSubset: SquareSubset,
  	SquareSubsetEqual: SquareSubsetEqual,
  	SquareSuperset: SquareSuperset,
  	SquareSupersetEqual: SquareSupersetEqual,
  	SquareUnion: SquareUnion,
  	squarf: squarf,
  	squ: squ,
  	squf: squf,
  	srarr: srarr,
  	Sscr: Sscr,
  	sscr: sscr,
  	ssetmn: ssetmn,
  	ssmile: ssmile,
  	sstarf: sstarf,
  	Star: Star,
  	star: star,
  	starf: starf,
  	straightepsilon: straightepsilon,
  	straightphi: straightphi,
  	strns: strns,
  	sub: sub,
  	Sub: Sub,
  	subdot: subdot,
  	subE: subE,
  	sube: sube,
  	subedot: subedot,
  	submult: submult,
  	subnE: subnE,
  	subne: subne,
  	subplus: subplus,
  	subrarr: subrarr,
  	subset: subset,
  	Subset: Subset,
  	subseteq: subseteq,
  	subseteqq: subseteqq,
  	SubsetEqual: SubsetEqual,
  	subsetneq: subsetneq,
  	subsetneqq: subsetneqq,
  	subsim: subsim,
  	subsub: subsub,
  	subsup: subsup,
  	succapprox: succapprox,
  	succ: succ,
  	succcurlyeq: succcurlyeq,
  	Succeeds: Succeeds,
  	SucceedsEqual: SucceedsEqual,
  	SucceedsSlantEqual: SucceedsSlantEqual,
  	SucceedsTilde: SucceedsTilde,
  	succeq: succeq,
  	succnapprox: succnapprox,
  	succneqq: succneqq,
  	succnsim: succnsim,
  	succsim: succsim,
  	SuchThat: SuchThat,
  	sum: sum,
  	Sum: Sum,
  	sung: sung,
  	sup1: sup1,
  	sup2: sup2,
  	sup3: sup3,
  	sup: sup,
  	Sup: Sup,
  	supdot: supdot,
  	supdsub: supdsub,
  	supE: supE,
  	supe: supe,
  	supedot: supedot,
  	Superset: Superset,
  	SupersetEqual: SupersetEqual,
  	suphsol: suphsol,
  	suphsub: suphsub,
  	suplarr: suplarr,
  	supmult: supmult,
  	supnE: supnE,
  	supne: supne,
  	supplus: supplus,
  	supset: supset,
  	Supset: Supset,
  	supseteq: supseteq,
  	supseteqq: supseteqq,
  	supsetneq: supsetneq,
  	supsetneqq: supsetneqq,
  	supsim: supsim,
  	supsub: supsub,
  	supsup: supsup,
  	swarhk: swarhk,
  	swarr: swarr,
  	swArr: swArr,
  	swarrow: swarrow,
  	swnwar: swnwar,
  	szlig: szlig,
  	Tab: Tab,
  	target: target,
  	Tau: Tau,
  	tau: tau,
  	tbrk: tbrk,
  	Tcaron: Tcaron,
  	tcaron: tcaron,
  	Tcedil: Tcedil,
  	tcedil: tcedil,
  	Tcy: Tcy,
  	tcy: tcy,
  	tdot: tdot,
  	telrec: telrec,
  	Tfr: Tfr,
  	tfr: tfr,
  	there4: there4,
  	therefore: therefore,
  	Therefore: Therefore,
  	Theta: Theta,
  	theta: theta,
  	thetasym: thetasym,
  	thetav: thetav,
  	thickapprox: thickapprox,
  	thicksim: thicksim,
  	ThickSpace: ThickSpace,
  	ThinSpace: ThinSpace,
  	thinsp: thinsp,
  	thkap: thkap,
  	thksim: thksim,
  	THORN: THORN,
  	thorn: thorn,
  	tilde: tilde,
  	Tilde: Tilde,
  	TildeEqual: TildeEqual,
  	TildeFullEqual: TildeFullEqual,
  	TildeTilde: TildeTilde,
  	timesbar: timesbar,
  	timesb: timesb,
  	times: times,
  	timesd: timesd,
  	tint: tint,
  	toea: toea,
  	topbot: topbot,
  	topcir: topcir,
  	top: top,
  	Topf: Topf,
  	topf: topf,
  	topfork: topfork,
  	tosa: tosa,
  	tprime: tprime,
  	trade: trade,
  	TRADE: TRADE,
  	triangle: triangle,
  	triangledown: triangledown,
  	triangleleft: triangleleft,
  	trianglelefteq: trianglelefteq,
  	triangleq: triangleq,
  	triangleright: triangleright,
  	trianglerighteq: trianglerighteq,
  	tridot: tridot,
  	trie: trie,
  	triminus: triminus,
  	TripleDot: TripleDot,
  	triplus: triplus,
  	trisb: trisb,
  	tritime: tritime,
  	trpezium: trpezium,
  	Tscr: Tscr,
  	tscr: tscr,
  	TScy: TScy,
  	tscy: tscy,
  	TSHcy: TSHcy,
  	tshcy: tshcy,
  	Tstrok: Tstrok,
  	tstrok: tstrok,
  	twixt: twixt,
  	twoheadleftarrow: twoheadleftarrow,
  	twoheadrightarrow: twoheadrightarrow,
  	Uacute: Uacute,
  	uacute: uacute,
  	uarr: uarr,
  	Uarr: Uarr,
  	uArr: uArr,
  	Uarrocir: Uarrocir,
  	Ubrcy: Ubrcy,
  	ubrcy: ubrcy,
  	Ubreve: Ubreve,
  	ubreve: ubreve,
  	Ucirc: Ucirc,
  	ucirc: ucirc,
  	Ucy: Ucy,
  	ucy: ucy,
  	udarr: udarr,
  	Udblac: Udblac,
  	udblac: udblac,
  	udhar: udhar,
  	ufisht: ufisht,
  	Ufr: Ufr,
  	ufr: ufr,
  	Ugrave: Ugrave,
  	ugrave: ugrave,
  	uHar: uHar,
  	uharl: uharl,
  	uharr: uharr,
  	uhblk: uhblk,
  	ulcorn: ulcorn,
  	ulcorner: ulcorner,
  	ulcrop: ulcrop,
  	ultri: ultri,
  	Umacr: Umacr,
  	umacr: umacr,
  	uml: uml,
  	UnderBar: UnderBar,
  	UnderBrace: UnderBrace,
  	UnderBracket: UnderBracket,
  	UnderParenthesis: UnderParenthesis,
  	Union: Union,
  	UnionPlus: UnionPlus,
  	Uogon: Uogon,
  	uogon: uogon,
  	Uopf: Uopf,
  	uopf: uopf,
  	UpArrowBar: UpArrowBar,
  	uparrow: uparrow,
  	UpArrow: UpArrow,
  	Uparrow: Uparrow,
  	UpArrowDownArrow: UpArrowDownArrow,
  	updownarrow: updownarrow,
  	UpDownArrow: UpDownArrow,
  	Updownarrow: Updownarrow,
  	UpEquilibrium: UpEquilibrium,
  	upharpoonleft: upharpoonleft,
  	upharpoonright: upharpoonright,
  	uplus: uplus,
  	UpperLeftArrow: UpperLeftArrow,
  	UpperRightArrow: UpperRightArrow,
  	upsi: upsi,
  	Upsi: Upsi,
  	upsih: upsih,
  	Upsilon: Upsilon,
  	upsilon: upsilon,
  	UpTeeArrow: UpTeeArrow,
  	UpTee: UpTee,
  	upuparrows: upuparrows,
  	urcorn: urcorn,
  	urcorner: urcorner,
  	urcrop: urcrop,
  	Uring: Uring,
  	uring: uring,
  	urtri: urtri,
  	Uscr: Uscr,
  	uscr: uscr,
  	utdot: utdot,
  	Utilde: Utilde,
  	utilde: utilde,
  	utri: utri,
  	utrif: utrif,
  	uuarr: uuarr,
  	Uuml: Uuml,
  	uuml: uuml,
  	uwangle: uwangle,
  	vangrt: vangrt,
  	varepsilon: varepsilon,
  	varkappa: varkappa,
  	varnothing: varnothing,
  	varphi: varphi,
  	varpi: varpi,
  	varpropto: varpropto,
  	varr: varr,
  	vArr: vArr,
  	varrho: varrho,
  	varsigma: varsigma,
  	varsubsetneq: varsubsetneq,
  	varsubsetneqq: varsubsetneqq,
  	varsupsetneq: varsupsetneq,
  	varsupsetneqq: varsupsetneqq,
  	vartheta: vartheta,
  	vartriangleleft: vartriangleleft,
  	vartriangleright: vartriangleright,
  	vBar: vBar,
  	Vbar: Vbar,
  	vBarv: vBarv,
  	Vcy: Vcy,
  	vcy: vcy,
  	vdash: vdash,
  	vDash: vDash,
  	Vdash: Vdash,
  	VDash: VDash,
  	Vdashl: Vdashl,
  	veebar: veebar,
  	vee: vee,
  	Vee: Vee,
  	veeeq: veeeq,
  	vellip: vellip,
  	verbar: verbar,
  	Verbar: Verbar,
  	vert: vert,
  	Vert: Vert,
  	VerticalBar: VerticalBar,
  	VerticalLine: VerticalLine,
  	VerticalSeparator: VerticalSeparator,
  	VerticalTilde: VerticalTilde,
  	VeryThinSpace: VeryThinSpace,
  	Vfr: Vfr,
  	vfr: vfr,
  	vltri: vltri,
  	vnsub: vnsub,
  	vnsup: vnsup,
  	Vopf: Vopf,
  	vopf: vopf,
  	vprop: vprop,
  	vrtri: vrtri,
  	Vscr: Vscr,
  	vscr: vscr,
  	vsubnE: vsubnE,
  	vsubne: vsubne,
  	vsupnE: vsupnE,
  	vsupne: vsupne,
  	Vvdash: Vvdash,
  	vzigzag: vzigzag,
  	Wcirc: Wcirc,
  	wcirc: wcirc,
  	wedbar: wedbar,
  	wedge: wedge,
  	Wedge: Wedge,
  	wedgeq: wedgeq,
  	weierp: weierp,
  	Wfr: Wfr,
  	wfr: wfr,
  	Wopf: Wopf,
  	wopf: wopf,
  	wp: wp,
  	wr: wr,
  	wreath: wreath,
  	Wscr: Wscr,
  	wscr: wscr,
  	xcap: xcap,
  	xcirc: xcirc,
  	xcup: xcup,
  	xdtri: xdtri,
  	Xfr: Xfr,
  	xfr: xfr,
  	xharr: xharr,
  	xhArr: xhArr,
  	Xi: Xi,
  	xi: xi,
  	xlarr: xlarr,
  	xlArr: xlArr,
  	xmap: xmap,
  	xnis: xnis,
  	xodot: xodot,
  	Xopf: Xopf,
  	xopf: xopf,
  	xoplus: xoplus,
  	xotime: xotime,
  	xrarr: xrarr,
  	xrArr: xrArr,
  	Xscr: Xscr,
  	xscr: xscr,
  	xsqcup: xsqcup,
  	xuplus: xuplus,
  	xutri: xutri,
  	xvee: xvee,
  	xwedge: xwedge,
  	Yacute: Yacute,
  	yacute: yacute,
  	YAcy: YAcy,
  	yacy: yacy,
  	Ycirc: Ycirc,
  	ycirc: ycirc,
  	Ycy: Ycy,
  	ycy: ycy,
  	yen: yen,
  	Yfr: Yfr,
  	yfr: yfr,
  	YIcy: YIcy,
  	yicy: yicy,
  	Yopf: Yopf,
  	yopf: yopf,
  	Yscr: Yscr,
  	yscr: yscr,
  	YUcy: YUcy,
  	yucy: yucy,
  	yuml: yuml,
  	Yuml: Yuml,
  	Zacute: Zacute,
  	zacute: zacute,
  	Zcaron: Zcaron,
  	zcaron: zcaron,
  	Zcy: Zcy,
  	zcy: zcy,
  	Zdot: Zdot,
  	zdot: zdot,
  	zeetrf: zeetrf,
  	ZeroWidthSpace: ZeroWidthSpace,
  	Zeta: Zeta,
  	zeta: zeta,
  	zfr: zfr,
  	Zfr: Zfr,
  	ZHcy: ZHcy,
  	zhcy: zhcy,
  	zigrarr: zigrarr,
  	zopf: zopf,
  	Zopf: Zopf,
  	Zscr: Zscr,
  	zscr: zscr,
  	zwj: zwj,
  	zwnj: zwnj,
  	"in": "∈",
  	"Map": "⤅"
  };

  var entities = Object.freeze({
  	Aacute: Aacute,
  	aacute: aacute,
  	Abreve: Abreve,
  	abreve: abreve,
  	ac: ac,
  	acd: acd,
  	acE: acE,
  	Acirc: Acirc,
  	acirc: acirc,
  	acute: acute,
  	Acy: Acy,
  	acy: acy,
  	AElig: AElig,
  	aelig: aelig,
  	af: af,
  	Afr: Afr,
  	afr: afr,
  	Agrave: Agrave,
  	agrave: agrave,
  	alefsym: alefsym,
  	aleph: aleph,
  	Alpha: Alpha,
  	alpha: alpha,
  	Amacr: Amacr,
  	amacr: amacr,
  	amalg: amalg,
  	amp: amp$1,
  	AMP: AMP,
  	andand: andand,
  	And: And,
  	and: and,
  	andd: andd,
  	andslope: andslope,
  	andv: andv,
  	ang: ang,
  	ange: ange,
  	angle: angle,
  	angmsdaa: angmsdaa,
  	angmsdab: angmsdab,
  	angmsdac: angmsdac,
  	angmsdad: angmsdad,
  	angmsdae: angmsdae,
  	angmsdaf: angmsdaf,
  	angmsdag: angmsdag,
  	angmsdah: angmsdah,
  	angmsd: angmsd,
  	angrt: angrt,
  	angrtvb: angrtvb,
  	angrtvbd: angrtvbd,
  	angsph: angsph,
  	angst: angst,
  	angzarr: angzarr,
  	Aogon: Aogon,
  	aogon: aogon,
  	Aopf: Aopf,
  	aopf: aopf,
  	apacir: apacir,
  	ap: ap,
  	apE: apE,
  	ape: ape,
  	apid: apid,
  	apos: apos$1,
  	ApplyFunction: ApplyFunction,
  	approx: approx,
  	approxeq: approxeq,
  	Aring: Aring,
  	aring: aring,
  	Ascr: Ascr,
  	ascr: ascr,
  	Assign: Assign,
  	ast: ast,
  	asymp: asymp,
  	asympeq: asympeq,
  	Atilde: Atilde,
  	atilde: atilde,
  	Auml: Auml,
  	auml: auml,
  	awconint: awconint,
  	awint: awint,
  	backcong: backcong,
  	backepsilon: backepsilon,
  	backprime: backprime,
  	backsim: backsim,
  	backsimeq: backsimeq,
  	Backslash: Backslash,
  	Barv: Barv,
  	barvee: barvee,
  	barwed: barwed,
  	Barwed: Barwed,
  	barwedge: barwedge,
  	bbrk: bbrk,
  	bbrktbrk: bbrktbrk,
  	bcong: bcong,
  	Bcy: Bcy,
  	bcy: bcy,
  	bdquo: bdquo,
  	becaus: becaus,
  	because: because,
  	Because: Because,
  	bemptyv: bemptyv,
  	bepsi: bepsi,
  	bernou: bernou,
  	Bernoullis: Bernoullis,
  	Beta: Beta,
  	beta: beta,
  	beth: beth,
  	between: between,
  	Bfr: Bfr,
  	bfr: bfr,
  	bigcap: bigcap,
  	bigcirc: bigcirc,
  	bigcup: bigcup,
  	bigodot: bigodot,
  	bigoplus: bigoplus,
  	bigotimes: bigotimes,
  	bigsqcup: bigsqcup,
  	bigstar: bigstar,
  	bigtriangledown: bigtriangledown,
  	bigtriangleup: bigtriangleup,
  	biguplus: biguplus,
  	bigvee: bigvee,
  	bigwedge: bigwedge,
  	bkarow: bkarow,
  	blacklozenge: blacklozenge,
  	blacksquare: blacksquare,
  	blacktriangle: blacktriangle,
  	blacktriangledown: blacktriangledown,
  	blacktriangleleft: blacktriangleleft,
  	blacktriangleright: blacktriangleright,
  	blank: blank,
  	blk12: blk12,
  	blk14: blk14,
  	blk34: blk34,
  	block: block,
  	bne: bne,
  	bnequiv: bnequiv,
  	bNot: bNot,
  	bnot: bnot,
  	Bopf: Bopf,
  	bopf: bopf,
  	bot: bot,
  	bottom: bottom,
  	bowtie: bowtie,
  	boxbox: boxbox,
  	boxdl: boxdl,
  	boxdL: boxdL,
  	boxDl: boxDl,
  	boxDL: boxDL,
  	boxdr: boxdr,
  	boxdR: boxdR,
  	boxDr: boxDr,
  	boxDR: boxDR,
  	boxh: boxh,
  	boxH: boxH,
  	boxhd: boxhd,
  	boxHd: boxHd,
  	boxhD: boxhD,
  	boxHD: boxHD,
  	boxhu: boxhu,
  	boxHu: boxHu,
  	boxhU: boxhU,
  	boxHU: boxHU,
  	boxminus: boxminus,
  	boxplus: boxplus,
  	boxtimes: boxtimes,
  	boxul: boxul,
  	boxuL: boxuL,
  	boxUl: boxUl,
  	boxUL: boxUL,
  	boxur: boxur,
  	boxuR: boxuR,
  	boxUr: boxUr,
  	boxUR: boxUR,
  	boxv: boxv,
  	boxV: boxV,
  	boxvh: boxvh,
  	boxvH: boxvH,
  	boxVh: boxVh,
  	boxVH: boxVH,
  	boxvl: boxvl,
  	boxvL: boxvL,
  	boxVl: boxVl,
  	boxVL: boxVL,
  	boxvr: boxvr,
  	boxvR: boxvR,
  	boxVr: boxVr,
  	boxVR: boxVR,
  	bprime: bprime,
  	breve: breve,
  	Breve: Breve,
  	brvbar: brvbar,
  	bscr: bscr,
  	Bscr: Bscr,
  	bsemi: bsemi,
  	bsim: bsim,
  	bsime: bsime,
  	bsolb: bsolb,
  	bsol: bsol,
  	bsolhsub: bsolhsub,
  	bull: bull,
  	bullet: bullet,
  	bump: bump,
  	bumpE: bumpE,
  	bumpe: bumpe,
  	Bumpeq: Bumpeq,
  	bumpeq: bumpeq,
  	Cacute: Cacute,
  	cacute: cacute,
  	capand: capand,
  	capbrcup: capbrcup,
  	capcap: capcap,
  	cap: cap,
  	Cap: Cap,
  	capcup: capcup,
  	capdot: capdot,
  	CapitalDifferentialD: CapitalDifferentialD,
  	caps: caps,
  	caret: caret,
  	caron: caron,
  	Cayleys: Cayleys,
  	ccaps: ccaps,
  	Ccaron: Ccaron,
  	ccaron: ccaron,
  	Ccedil: Ccedil,
  	ccedil: ccedil,
  	Ccirc: Ccirc,
  	ccirc: ccirc,
  	Cconint: Cconint,
  	ccups: ccups,
  	ccupssm: ccupssm,
  	Cdot: Cdot,
  	cdot: cdot,
  	cedil: cedil,
  	Cedilla: Cedilla,
  	cemptyv: cemptyv,
  	cent: cent,
  	centerdot: centerdot,
  	CenterDot: CenterDot,
  	cfr: cfr,
  	Cfr: Cfr,
  	CHcy: CHcy,
  	chcy: chcy,
  	check: check,
  	checkmark: checkmark,
  	Chi: Chi,
  	chi: chi,
  	circ: circ,
  	circeq: circeq,
  	circlearrowleft: circlearrowleft,
  	circlearrowright: circlearrowright,
  	circledast: circledast,
  	circledcirc: circledcirc,
  	circleddash: circleddash,
  	CircleDot: CircleDot,
  	circledR: circledR,
  	circledS: circledS,
  	CircleMinus: CircleMinus,
  	CirclePlus: CirclePlus,
  	CircleTimes: CircleTimes,
  	cir: cir,
  	cirE: cirE,
  	cire: cire,
  	cirfnint: cirfnint,
  	cirmid: cirmid,
  	cirscir: cirscir,
  	ClockwiseContourIntegral: ClockwiseContourIntegral,
  	CloseCurlyDoubleQuote: CloseCurlyDoubleQuote,
  	CloseCurlyQuote: CloseCurlyQuote,
  	clubs: clubs,
  	clubsuit: clubsuit,
  	colon: colon,
  	Colon: Colon,
  	Colone: Colone,
  	colone: colone,
  	coloneq: coloneq,
  	comma: comma,
  	commat: commat,
  	comp: comp,
  	compfn: compfn,
  	complement: complement,
  	complexes: complexes,
  	cong: cong,
  	congdot: congdot,
  	Congruent: Congruent,
  	conint: conint,
  	Conint: Conint,
  	ContourIntegral: ContourIntegral,
  	copf: copf,
  	Copf: Copf,
  	coprod: coprod,
  	Coproduct: Coproduct,
  	copy: copy,
  	COPY: COPY,
  	copysr: copysr,
  	CounterClockwiseContourIntegral: CounterClockwiseContourIntegral,
  	crarr: crarr,
  	cross: cross,
  	Cross: Cross,
  	Cscr: Cscr,
  	cscr: cscr,
  	csub: csub,
  	csube: csube,
  	csup: csup,
  	csupe: csupe,
  	ctdot: ctdot,
  	cudarrl: cudarrl,
  	cudarrr: cudarrr,
  	cuepr: cuepr,
  	cuesc: cuesc,
  	cularr: cularr,
  	cularrp: cularrp,
  	cupbrcap: cupbrcap,
  	cupcap: cupcap,
  	CupCap: CupCap,
  	cup: cup,
  	Cup: Cup,
  	cupcup: cupcup,
  	cupdot: cupdot,
  	cupor: cupor,
  	cups: cups,
  	curarr: curarr,
  	curarrm: curarrm,
  	curlyeqprec: curlyeqprec,
  	curlyeqsucc: curlyeqsucc,
  	curlyvee: curlyvee,
  	curlywedge: curlywedge,
  	curren: curren,
  	curvearrowleft: curvearrowleft,
  	curvearrowright: curvearrowright,
  	cuvee: cuvee,
  	cuwed: cuwed,
  	cwconint: cwconint,
  	cwint: cwint,
  	cylcty: cylcty,
  	dagger: dagger,
  	Dagger: Dagger,
  	daleth: daleth,
  	darr: darr,
  	Darr: Darr,
  	dArr: dArr,
  	dash: dash,
  	Dashv: Dashv,
  	dashv: dashv,
  	dbkarow: dbkarow,
  	dblac: dblac,
  	Dcaron: Dcaron,
  	dcaron: dcaron,
  	Dcy: Dcy,
  	dcy: dcy,
  	ddagger: ddagger,
  	ddarr: ddarr,
  	DD: DD,
  	dd: dd,
  	DDotrahd: DDotrahd,
  	ddotseq: ddotseq,
  	deg: deg,
  	Del: Del,
  	Delta: Delta,
  	delta: delta,
  	demptyv: demptyv,
  	dfisht: dfisht,
  	Dfr: Dfr,
  	dfr: dfr,
  	dHar: dHar,
  	dharl: dharl,
  	dharr: dharr,
  	DiacriticalAcute: DiacriticalAcute,
  	DiacriticalDot: DiacriticalDot,
  	DiacriticalDoubleAcute: DiacriticalDoubleAcute,
  	DiacriticalGrave: DiacriticalGrave,
  	DiacriticalTilde: DiacriticalTilde,
  	diam: diam,
  	diamond: diamond,
  	Diamond: Diamond,
  	diamondsuit: diamondsuit,
  	diams: diams,
  	die: die,
  	DifferentialD: DifferentialD,
  	digamma: digamma,
  	disin: disin,
  	div: div,
  	divide: divide,
  	divideontimes: divideontimes,
  	divonx: divonx,
  	DJcy: DJcy,
  	djcy: djcy,
  	dlcorn: dlcorn,
  	dlcrop: dlcrop,
  	dollar: dollar,
  	Dopf: Dopf,
  	dopf: dopf,
  	Dot: Dot,
  	dot: dot,
  	DotDot: DotDot,
  	doteq: doteq,
  	doteqdot: doteqdot,
  	DotEqual: DotEqual,
  	dotminus: dotminus,
  	dotplus: dotplus,
  	dotsquare: dotsquare,
  	doublebarwedge: doublebarwedge,
  	DoubleContourIntegral: DoubleContourIntegral,
  	DoubleDot: DoubleDot,
  	DoubleDownArrow: DoubleDownArrow,
  	DoubleLeftArrow: DoubleLeftArrow,
  	DoubleLeftRightArrow: DoubleLeftRightArrow,
  	DoubleLeftTee: DoubleLeftTee,
  	DoubleLongLeftArrow: DoubleLongLeftArrow,
  	DoubleLongLeftRightArrow: DoubleLongLeftRightArrow,
  	DoubleLongRightArrow: DoubleLongRightArrow,
  	DoubleRightArrow: DoubleRightArrow,
  	DoubleRightTee: DoubleRightTee,
  	DoubleUpArrow: DoubleUpArrow,
  	DoubleUpDownArrow: DoubleUpDownArrow,
  	DoubleVerticalBar: DoubleVerticalBar,
  	DownArrowBar: DownArrowBar,
  	downarrow: downarrow,
  	DownArrow: DownArrow,
  	Downarrow: Downarrow,
  	DownArrowUpArrow: DownArrowUpArrow,
  	DownBreve: DownBreve,
  	downdownarrows: downdownarrows,
  	downharpoonleft: downharpoonleft,
  	downharpoonright: downharpoonright,
  	DownLeftRightVector: DownLeftRightVector,
  	DownLeftTeeVector: DownLeftTeeVector,
  	DownLeftVectorBar: DownLeftVectorBar,
  	DownLeftVector: DownLeftVector,
  	DownRightTeeVector: DownRightTeeVector,
  	DownRightVectorBar: DownRightVectorBar,
  	DownRightVector: DownRightVector,
  	DownTeeArrow: DownTeeArrow,
  	DownTee: DownTee,
  	drbkarow: drbkarow,
  	drcorn: drcorn,
  	drcrop: drcrop,
  	Dscr: Dscr,
  	dscr: dscr,
  	DScy: DScy,
  	dscy: dscy,
  	dsol: dsol,
  	Dstrok: Dstrok,
  	dstrok: dstrok,
  	dtdot: dtdot,
  	dtri: dtri,
  	dtrif: dtrif,
  	duarr: duarr,
  	duhar: duhar,
  	dwangle: dwangle,
  	DZcy: DZcy,
  	dzcy: dzcy,
  	dzigrarr: dzigrarr,
  	Eacute: Eacute,
  	eacute: eacute,
  	easter: easter,
  	Ecaron: Ecaron,
  	ecaron: ecaron,
  	Ecirc: Ecirc,
  	ecirc: ecirc,
  	ecir: ecir,
  	ecolon: ecolon,
  	Ecy: Ecy,
  	ecy: ecy,
  	eDDot: eDDot,
  	Edot: Edot,
  	edot: edot,
  	eDot: eDot,
  	ee: ee,
  	efDot: efDot,
  	Efr: Efr,
  	efr: efr,
  	eg: eg,
  	Egrave: Egrave,
  	egrave: egrave,
  	egs: egs,
  	egsdot: egsdot,
  	el: el,
  	Element: Element,
  	elinters: elinters,
  	ell: ell,
  	els: els,
  	elsdot: elsdot,
  	Emacr: Emacr,
  	emacr: emacr,
  	empty: empty,
  	emptyset: emptyset,
  	EmptySmallSquare: EmptySmallSquare,
  	emptyv: emptyv,
  	EmptyVerySmallSquare: EmptyVerySmallSquare,
  	emsp13: emsp13,
  	emsp14: emsp14,
  	emsp: emsp,
  	ENG: ENG,
  	eng: eng,
  	ensp: ensp,
  	Eogon: Eogon,
  	eogon: eogon,
  	Eopf: Eopf,
  	eopf: eopf,
  	epar: epar,
  	eparsl: eparsl,
  	eplus: eplus,
  	epsi: epsi,
  	Epsilon: Epsilon,
  	epsilon: epsilon,
  	epsiv: epsiv,
  	eqcirc: eqcirc,
  	eqcolon: eqcolon,
  	eqsim: eqsim,
  	eqslantgtr: eqslantgtr,
  	eqslantless: eqslantless,
  	Equal: Equal,
  	equals: equals,
  	EqualTilde: EqualTilde,
  	equest: equest,
  	Equilibrium: Equilibrium,
  	equiv: equiv,
  	equivDD: equivDD,
  	eqvparsl: eqvparsl,
  	erarr: erarr,
  	erDot: erDot,
  	escr: escr,
  	Escr: Escr,
  	esdot: esdot,
  	Esim: Esim,
  	esim: esim,
  	Eta: Eta,
  	eta: eta,
  	ETH: ETH,
  	eth: eth,
  	Euml: Euml,
  	euml: euml,
  	euro: euro,
  	excl: excl,
  	exist: exist,
  	Exists: Exists,
  	expectation: expectation,
  	exponentiale: exponentiale,
  	ExponentialE: ExponentialE,
  	fallingdotseq: fallingdotseq,
  	Fcy: Fcy,
  	fcy: fcy,
  	female: female,
  	ffilig: ffilig,
  	fflig: fflig,
  	ffllig: ffllig,
  	Ffr: Ffr,
  	ffr: ffr,
  	filig: filig,
  	FilledSmallSquare: FilledSmallSquare,
  	FilledVerySmallSquare: FilledVerySmallSquare,
  	fjlig: fjlig,
  	flat: flat,
  	fllig: fllig,
  	fltns: fltns,
  	fnof: fnof,
  	Fopf: Fopf,
  	fopf: fopf,
  	forall: forall,
  	ForAll: ForAll,
  	fork: fork,
  	forkv: forkv,
  	Fouriertrf: Fouriertrf,
  	fpartint: fpartint,
  	frac12: frac12,
  	frac13: frac13,
  	frac14: frac14,
  	frac15: frac15,
  	frac16: frac16,
  	frac18: frac18,
  	frac23: frac23,
  	frac25: frac25,
  	frac34: frac34,
  	frac35: frac35,
  	frac38: frac38,
  	frac45: frac45,
  	frac56: frac56,
  	frac58: frac58,
  	frac78: frac78,
  	frasl: frasl,
  	frown: frown,
  	fscr: fscr,
  	Fscr: Fscr,
  	gacute: gacute,
  	Gamma: Gamma,
  	gamma: gamma,
  	Gammad: Gammad,
  	gammad: gammad,
  	gap: gap,
  	Gbreve: Gbreve,
  	gbreve: gbreve,
  	Gcedil: Gcedil,
  	Gcirc: Gcirc,
  	gcirc: gcirc,
  	Gcy: Gcy,
  	gcy: gcy,
  	Gdot: Gdot,
  	gdot: gdot,
  	ge: ge,
  	gE: gE,
  	gEl: gEl,
  	gel: gel,
  	geq: geq,
  	geqq: geqq,
  	geqslant: geqslant,
  	gescc: gescc,
  	ges: ges,
  	gesdot: gesdot,
  	gesdoto: gesdoto,
  	gesdotol: gesdotol,
  	gesl: gesl,
  	gesles: gesles,
  	Gfr: Gfr,
  	gfr: gfr,
  	gg: gg,
  	Gg: Gg,
  	ggg: ggg,
  	gimel: gimel,
  	GJcy: GJcy,
  	gjcy: gjcy,
  	gla: gla,
  	gl: gl,
  	glE: glE,
  	glj: glj,
  	gnap: gnap,
  	gnapprox: gnapprox,
  	gne: gne,
  	gnE: gnE,
  	gneq: gneq,
  	gneqq: gneqq,
  	gnsim: gnsim,
  	Gopf: Gopf,
  	gopf: gopf,
  	grave: grave,
  	GreaterEqual: GreaterEqual,
  	GreaterEqualLess: GreaterEqualLess,
  	GreaterFullEqual: GreaterFullEqual,
  	GreaterGreater: GreaterGreater,
  	GreaterLess: GreaterLess,
  	GreaterSlantEqual: GreaterSlantEqual,
  	GreaterTilde: GreaterTilde,
  	Gscr: Gscr,
  	gscr: gscr,
  	gsim: gsim,
  	gsime: gsime,
  	gsiml: gsiml,
  	gtcc: gtcc,
  	gtcir: gtcir,
  	gt: gt$1,
  	GT: GT,
  	Gt: Gt,
  	gtdot: gtdot,
  	gtlPar: gtlPar,
  	gtquest: gtquest,
  	gtrapprox: gtrapprox,
  	gtrarr: gtrarr,
  	gtrdot: gtrdot,
  	gtreqless: gtreqless,
  	gtreqqless: gtreqqless,
  	gtrless: gtrless,
  	gtrsim: gtrsim,
  	gvertneqq: gvertneqq,
  	gvnE: gvnE,
  	Hacek: Hacek,
  	hairsp: hairsp,
  	half: half,
  	hamilt: hamilt,
  	HARDcy: HARDcy,
  	hardcy: hardcy,
  	harrcir: harrcir,
  	harr: harr,
  	hArr: hArr,
  	harrw: harrw,
  	Hat: Hat,
  	hbar: hbar,
  	Hcirc: Hcirc,
  	hcirc: hcirc,
  	hearts: hearts,
  	heartsuit: heartsuit,
  	hellip: hellip,
  	hercon: hercon,
  	hfr: hfr,
  	Hfr: Hfr,
  	HilbertSpace: HilbertSpace,
  	hksearow: hksearow,
  	hkswarow: hkswarow,
  	hoarr: hoarr,
  	homtht: homtht,
  	hookleftarrow: hookleftarrow,
  	hookrightarrow: hookrightarrow,
  	hopf: hopf,
  	Hopf: Hopf,
  	horbar: horbar,
  	HorizontalLine: HorizontalLine,
  	hscr: hscr,
  	Hscr: Hscr,
  	hslash: hslash,
  	Hstrok: Hstrok,
  	hstrok: hstrok,
  	HumpDownHump: HumpDownHump,
  	HumpEqual: HumpEqual,
  	hybull: hybull,
  	hyphen: hyphen,
  	Iacute: Iacute,
  	iacute: iacute,
  	ic: ic,
  	Icirc: Icirc,
  	icirc: icirc,
  	Icy: Icy,
  	icy: icy,
  	Idot: Idot,
  	IEcy: IEcy,
  	iecy: iecy,
  	iexcl: iexcl,
  	iff: iff,
  	ifr: ifr,
  	Ifr: Ifr,
  	Igrave: Igrave,
  	igrave: igrave,
  	ii: ii,
  	iiiint: iiiint,
  	iiint: iiint,
  	iinfin: iinfin,
  	iiota: iiota,
  	IJlig: IJlig,
  	ijlig: ijlig,
  	Imacr: Imacr,
  	imacr: imacr,
  	image: image,
  	ImaginaryI: ImaginaryI,
  	imagline: imagline,
  	imagpart: imagpart,
  	imath: imath,
  	Im: Im,
  	imof: imof,
  	imped: imped,
  	Implies: Implies,
  	incare: incare,
  	infin: infin,
  	infintie: infintie,
  	inodot: inodot,
  	intcal: intcal,
  	int: int,
  	Int: Int,
  	integers: integers,
  	Integral: Integral,
  	intercal: intercal,
  	Intersection: Intersection,
  	intlarhk: intlarhk,
  	intprod: intprod,
  	InvisibleComma: InvisibleComma,
  	InvisibleTimes: InvisibleTimes,
  	IOcy: IOcy,
  	iocy: iocy,
  	Iogon: Iogon,
  	iogon: iogon,
  	Iopf: Iopf,
  	iopf: iopf,
  	Iota: Iota,
  	iota: iota,
  	iprod: iprod,
  	iquest: iquest,
  	iscr: iscr,
  	Iscr: Iscr,
  	isin: isin,
  	isindot: isindot,
  	isinE: isinE,
  	isins: isins,
  	isinsv: isinsv,
  	isinv: isinv,
  	it: it,
  	Itilde: Itilde,
  	itilde: itilde,
  	Iukcy: Iukcy,
  	iukcy: iukcy,
  	Iuml: Iuml,
  	iuml: iuml,
  	Jcirc: Jcirc,
  	jcirc: jcirc,
  	Jcy: Jcy,
  	jcy: jcy,
  	Jfr: Jfr,
  	jfr: jfr,
  	jmath: jmath,
  	Jopf: Jopf,
  	jopf: jopf,
  	Jscr: Jscr,
  	jscr: jscr,
  	Jsercy: Jsercy,
  	jsercy: jsercy,
  	Jukcy: Jukcy,
  	jukcy: jukcy,
  	Kappa: Kappa,
  	kappa: kappa,
  	kappav: kappav,
  	Kcedil: Kcedil,
  	kcedil: kcedil,
  	Kcy: Kcy,
  	kcy: kcy,
  	Kfr: Kfr,
  	kfr: kfr,
  	kgreen: kgreen,
  	KHcy: KHcy,
  	khcy: khcy,
  	KJcy: KJcy,
  	kjcy: kjcy,
  	Kopf: Kopf,
  	kopf: kopf,
  	Kscr: Kscr,
  	kscr: kscr,
  	lAarr: lAarr,
  	Lacute: Lacute,
  	lacute: lacute,
  	laemptyv: laemptyv,
  	lagran: lagran,
  	Lambda: Lambda,
  	lambda: lambda,
  	lang: lang,
  	Lang: Lang,
  	langd: langd,
  	langle: langle,
  	lap: lap,
  	Laplacetrf: Laplacetrf,
  	laquo: laquo,
  	larrb: larrb,
  	larrbfs: larrbfs,
  	larr: larr,
  	Larr: Larr,
  	lArr: lArr,
  	larrfs: larrfs,
  	larrhk: larrhk,
  	larrlp: larrlp,
  	larrpl: larrpl,
  	larrsim: larrsim,
  	larrtl: larrtl,
  	latail: latail,
  	lAtail: lAtail,
  	lat: lat,
  	late: late,
  	lates: lates,
  	lbarr: lbarr,
  	lBarr: lBarr,
  	lbbrk: lbbrk,
  	lbrace: lbrace,
  	lbrack: lbrack,
  	lbrke: lbrke,
  	lbrksld: lbrksld,
  	lbrkslu: lbrkslu,
  	Lcaron: Lcaron,
  	lcaron: lcaron,
  	Lcedil: Lcedil,
  	lcedil: lcedil,
  	lceil: lceil,
  	lcub: lcub,
  	Lcy: Lcy,
  	lcy: lcy,
  	ldca: ldca,
  	ldquo: ldquo,
  	ldquor: ldquor,
  	ldrdhar: ldrdhar,
  	ldrushar: ldrushar,
  	ldsh: ldsh,
  	le: le,
  	lE: lE,
  	LeftAngleBracket: LeftAngleBracket,
  	LeftArrowBar: LeftArrowBar,
  	leftarrow: leftarrow,
  	LeftArrow: LeftArrow,
  	Leftarrow: Leftarrow,
  	LeftArrowRightArrow: LeftArrowRightArrow,
  	leftarrowtail: leftarrowtail,
  	LeftCeiling: LeftCeiling,
  	LeftDoubleBracket: LeftDoubleBracket,
  	LeftDownTeeVector: LeftDownTeeVector,
  	LeftDownVectorBar: LeftDownVectorBar,
  	LeftDownVector: LeftDownVector,
  	LeftFloor: LeftFloor,
  	leftharpoondown: leftharpoondown,
  	leftharpoonup: leftharpoonup,
  	leftleftarrows: leftleftarrows,
  	leftrightarrow: leftrightarrow,
  	LeftRightArrow: LeftRightArrow,
  	Leftrightarrow: Leftrightarrow,
  	leftrightarrows: leftrightarrows,
  	leftrightharpoons: leftrightharpoons,
  	leftrightsquigarrow: leftrightsquigarrow,
  	LeftRightVector: LeftRightVector,
  	LeftTeeArrow: LeftTeeArrow,
  	LeftTee: LeftTee,
  	LeftTeeVector: LeftTeeVector,
  	leftthreetimes: leftthreetimes,
  	LeftTriangleBar: LeftTriangleBar,
  	LeftTriangle: LeftTriangle,
  	LeftTriangleEqual: LeftTriangleEqual,
  	LeftUpDownVector: LeftUpDownVector,
  	LeftUpTeeVector: LeftUpTeeVector,
  	LeftUpVectorBar: LeftUpVectorBar,
  	LeftUpVector: LeftUpVector,
  	LeftVectorBar: LeftVectorBar,
  	LeftVector: LeftVector,
  	lEg: lEg,
  	leg: leg,
  	leq: leq,
  	leqq: leqq,
  	leqslant: leqslant,
  	lescc: lescc,
  	les: les,
  	lesdot: lesdot,
  	lesdoto: lesdoto,
  	lesdotor: lesdotor,
  	lesg: lesg,
  	lesges: lesges,
  	lessapprox: lessapprox,
  	lessdot: lessdot,
  	lesseqgtr: lesseqgtr,
  	lesseqqgtr: lesseqqgtr,
  	LessEqualGreater: LessEqualGreater,
  	LessFullEqual: LessFullEqual,
  	LessGreater: LessGreater,
  	lessgtr: lessgtr,
  	LessLess: LessLess,
  	lesssim: lesssim,
  	LessSlantEqual: LessSlantEqual,
  	LessTilde: LessTilde,
  	lfisht: lfisht,
  	lfloor: lfloor,
  	Lfr: Lfr,
  	lfr: lfr,
  	lg: lg,
  	lgE: lgE,
  	lHar: lHar,
  	lhard: lhard,
  	lharu: lharu,
  	lharul: lharul,
  	lhblk: lhblk,
  	LJcy: LJcy,
  	ljcy: ljcy,
  	llarr: llarr,
  	ll: ll,
  	Ll: Ll,
  	llcorner: llcorner,
  	Lleftarrow: Lleftarrow,
  	llhard: llhard,
  	lltri: lltri,
  	Lmidot: Lmidot,
  	lmidot: lmidot,
  	lmoustache: lmoustache,
  	lmoust: lmoust,
  	lnap: lnap,
  	lnapprox: lnapprox,
  	lne: lne,
  	lnE: lnE,
  	lneq: lneq,
  	lneqq: lneqq,
  	lnsim: lnsim,
  	loang: loang,
  	loarr: loarr,
  	lobrk: lobrk,
  	longleftarrow: longleftarrow,
  	LongLeftArrow: LongLeftArrow,
  	Longleftarrow: Longleftarrow,
  	longleftrightarrow: longleftrightarrow,
  	LongLeftRightArrow: LongLeftRightArrow,
  	Longleftrightarrow: Longleftrightarrow,
  	longmapsto: longmapsto,
  	longrightarrow: longrightarrow,
  	LongRightArrow: LongRightArrow,
  	Longrightarrow: Longrightarrow,
  	looparrowleft: looparrowleft,
  	looparrowright: looparrowright,
  	lopar: lopar,
  	Lopf: Lopf,
  	lopf: lopf,
  	loplus: loplus,
  	lotimes: lotimes,
  	lowast: lowast,
  	lowbar: lowbar,
  	LowerLeftArrow: LowerLeftArrow,
  	LowerRightArrow: LowerRightArrow,
  	loz: loz,
  	lozenge: lozenge,
  	lozf: lozf,
  	lpar: lpar,
  	lparlt: lparlt,
  	lrarr: lrarr,
  	lrcorner: lrcorner,
  	lrhar: lrhar,
  	lrhard: lrhard,
  	lrm: lrm,
  	lrtri: lrtri,
  	lsaquo: lsaquo,
  	lscr: lscr,
  	Lscr: Lscr,
  	lsh: lsh,
  	Lsh: Lsh,
  	lsim: lsim,
  	lsime: lsime,
  	lsimg: lsimg,
  	lsqb: lsqb,
  	lsquo: lsquo,
  	lsquor: lsquor,
  	Lstrok: Lstrok,
  	lstrok: lstrok,
  	ltcc: ltcc,
  	ltcir: ltcir,
  	lt: lt$1,
  	LT: LT,
  	Lt: Lt,
  	ltdot: ltdot,
  	lthree: lthree,
  	ltimes: ltimes,
  	ltlarr: ltlarr,
  	ltquest: ltquest,
  	ltri: ltri,
  	ltrie: ltrie,
  	ltrif: ltrif,
  	ltrPar: ltrPar,
  	lurdshar: lurdshar,
  	luruhar: luruhar,
  	lvertneqq: lvertneqq,
  	lvnE: lvnE,
  	macr: macr,
  	male: male,
  	malt: malt,
  	maltese: maltese,
  	map: map$1,
  	mapsto: mapsto,
  	mapstodown: mapstodown,
  	mapstoleft: mapstoleft,
  	mapstoup: mapstoup,
  	marker: marker,
  	mcomma: mcomma,
  	Mcy: Mcy,
  	mcy: mcy,
  	mdash: mdash,
  	mDDot: mDDot,
  	measuredangle: measuredangle,
  	MediumSpace: MediumSpace,
  	Mellintrf: Mellintrf,
  	Mfr: Mfr,
  	mfr: mfr,
  	mho: mho,
  	micro: micro,
  	midast: midast,
  	midcir: midcir,
  	mid: mid,
  	middot: middot,
  	minusb: minusb,
  	minus: minus,
  	minusd: minusd,
  	minusdu: minusdu,
  	MinusPlus: MinusPlus,
  	mlcp: mlcp,
  	mldr: mldr,
  	mnplus: mnplus,
  	models: models,
  	Mopf: Mopf,
  	mopf: mopf,
  	mp: mp,
  	mscr: mscr,
  	Mscr: Mscr,
  	mstpos: mstpos,
  	Mu: Mu,
  	mu: mu,
  	multimap: multimap,
  	mumap: mumap,
  	nabla: nabla,
  	Nacute: Nacute,
  	nacute: nacute,
  	nang: nang,
  	nap: nap,
  	napE: napE,
  	napid: napid,
  	napos: napos,
  	napprox: napprox,
  	natural: natural,
  	naturals: naturals,
  	natur: natur,
  	nbsp: nbsp,
  	nbump: nbump,
  	nbumpe: nbumpe,
  	ncap: ncap,
  	Ncaron: Ncaron,
  	ncaron: ncaron,
  	Ncedil: Ncedil,
  	ncedil: ncedil,
  	ncong: ncong,
  	ncongdot: ncongdot,
  	ncup: ncup,
  	Ncy: Ncy,
  	ncy: ncy,
  	ndash: ndash,
  	nearhk: nearhk,
  	nearr: nearr,
  	neArr: neArr,
  	nearrow: nearrow,
  	ne: ne,
  	nedot: nedot,
  	NegativeMediumSpace: NegativeMediumSpace,
  	NegativeThickSpace: NegativeThickSpace,
  	NegativeThinSpace: NegativeThinSpace,
  	NegativeVeryThinSpace: NegativeVeryThinSpace,
  	nequiv: nequiv,
  	nesear: nesear,
  	nesim: nesim,
  	NestedGreaterGreater: NestedGreaterGreater,
  	NestedLessLess: NestedLessLess,
  	NewLine: NewLine,
  	nexist: nexist,
  	nexists: nexists,
  	Nfr: Nfr,
  	nfr: nfr,
  	ngE: ngE,
  	nge: nge,
  	ngeq: ngeq,
  	ngeqq: ngeqq,
  	ngeqslant: ngeqslant,
  	nges: nges,
  	nGg: nGg,
  	ngsim: ngsim,
  	nGt: nGt,
  	ngt: ngt,
  	ngtr: ngtr,
  	nGtv: nGtv,
  	nharr: nharr,
  	nhArr: nhArr,
  	nhpar: nhpar,
  	ni: ni,
  	nis: nis,
  	nisd: nisd,
  	niv: niv,
  	NJcy: NJcy,
  	njcy: njcy,
  	nlarr: nlarr,
  	nlArr: nlArr,
  	nldr: nldr,
  	nlE: nlE,
  	nle: nle,
  	nleftarrow: nleftarrow,
  	nLeftarrow: nLeftarrow,
  	nleftrightarrow: nleftrightarrow,
  	nLeftrightarrow: nLeftrightarrow,
  	nleq: nleq,
  	nleqq: nleqq,
  	nleqslant: nleqslant,
  	nles: nles,
  	nless: nless,
  	nLl: nLl,
  	nlsim: nlsim,
  	nLt: nLt,
  	nlt: nlt,
  	nltri: nltri,
  	nltrie: nltrie,
  	nLtv: nLtv,
  	nmid: nmid,
  	NoBreak: NoBreak,
  	NonBreakingSpace: NonBreakingSpace,
  	nopf: nopf,
  	Nopf: Nopf,
  	Not: Not,
  	not: not,
  	NotCongruent: NotCongruent,
  	NotCupCap: NotCupCap,
  	NotDoubleVerticalBar: NotDoubleVerticalBar,
  	NotElement: NotElement,
  	NotEqual: NotEqual,
  	NotEqualTilde: NotEqualTilde,
  	NotExists: NotExists,
  	NotGreater: NotGreater,
  	NotGreaterEqual: NotGreaterEqual,
  	NotGreaterFullEqual: NotGreaterFullEqual,
  	NotGreaterGreater: NotGreaterGreater,
  	NotGreaterLess: NotGreaterLess,
  	NotGreaterSlantEqual: NotGreaterSlantEqual,
  	NotGreaterTilde: NotGreaterTilde,
  	NotHumpDownHump: NotHumpDownHump,
  	NotHumpEqual: NotHumpEqual,
  	notin: notin,
  	notindot: notindot,
  	notinE: notinE,
  	notinva: notinva,
  	notinvb: notinvb,
  	notinvc: notinvc,
  	NotLeftTriangleBar: NotLeftTriangleBar,
  	NotLeftTriangle: NotLeftTriangle,
  	NotLeftTriangleEqual: NotLeftTriangleEqual,
  	NotLess: NotLess,
  	NotLessEqual: NotLessEqual,
  	NotLessGreater: NotLessGreater,
  	NotLessLess: NotLessLess,
  	NotLessSlantEqual: NotLessSlantEqual,
  	NotLessTilde: NotLessTilde,
  	NotNestedGreaterGreater: NotNestedGreaterGreater,
  	NotNestedLessLess: NotNestedLessLess,
  	notni: notni,
  	notniva: notniva,
  	notnivb: notnivb,
  	notnivc: notnivc,
  	NotPrecedes: NotPrecedes,
  	NotPrecedesEqual: NotPrecedesEqual,
  	NotPrecedesSlantEqual: NotPrecedesSlantEqual,
  	NotReverseElement: NotReverseElement,
  	NotRightTriangleBar: NotRightTriangleBar,
  	NotRightTriangle: NotRightTriangle,
  	NotRightTriangleEqual: NotRightTriangleEqual,
  	NotSquareSubset: NotSquareSubset,
  	NotSquareSubsetEqual: NotSquareSubsetEqual,
  	NotSquareSuperset: NotSquareSuperset,
  	NotSquareSupersetEqual: NotSquareSupersetEqual,
  	NotSubset: NotSubset,
  	NotSubsetEqual: NotSubsetEqual,
  	NotSucceeds: NotSucceeds,
  	NotSucceedsEqual: NotSucceedsEqual,
  	NotSucceedsSlantEqual: NotSucceedsSlantEqual,
  	NotSucceedsTilde: NotSucceedsTilde,
  	NotSuperset: NotSuperset,
  	NotSupersetEqual: NotSupersetEqual,
  	NotTilde: NotTilde,
  	NotTildeEqual: NotTildeEqual,
  	NotTildeFullEqual: NotTildeFullEqual,
  	NotTildeTilde: NotTildeTilde,
  	NotVerticalBar: NotVerticalBar,
  	nparallel: nparallel,
  	npar: npar,
  	nparsl: nparsl,
  	npart: npart,
  	npolint: npolint,
  	npr: npr,
  	nprcue: nprcue,
  	nprec: nprec,
  	npreceq: npreceq,
  	npre: npre,
  	nrarrc: nrarrc,
  	nrarr: nrarr,
  	nrArr: nrArr,
  	nrarrw: nrarrw,
  	nrightarrow: nrightarrow,
  	nRightarrow: nRightarrow,
  	nrtri: nrtri,
  	nrtrie: nrtrie,
  	nsc: nsc,
  	nsccue: nsccue,
  	nsce: nsce,
  	Nscr: Nscr,
  	nscr: nscr,
  	nshortmid: nshortmid,
  	nshortparallel: nshortparallel,
  	nsim: nsim,
  	nsime: nsime,
  	nsimeq: nsimeq,
  	nsmid: nsmid,
  	nspar: nspar,
  	nsqsube: nsqsube,
  	nsqsupe: nsqsupe,
  	nsub: nsub,
  	nsubE: nsubE,
  	nsube: nsube,
  	nsubset: nsubset,
  	nsubseteq: nsubseteq,
  	nsubseteqq: nsubseteqq,
  	nsucc: nsucc,
  	nsucceq: nsucceq,
  	nsup: nsup,
  	nsupE: nsupE,
  	nsupe: nsupe,
  	nsupset: nsupset,
  	nsupseteq: nsupseteq,
  	nsupseteqq: nsupseteqq,
  	ntgl: ntgl,
  	Ntilde: Ntilde,
  	ntilde: ntilde,
  	ntlg: ntlg,
  	ntriangleleft: ntriangleleft,
  	ntrianglelefteq: ntrianglelefteq,
  	ntriangleright: ntriangleright,
  	ntrianglerighteq: ntrianglerighteq,
  	Nu: Nu,
  	nu: nu,
  	num: num,
  	numero: numero,
  	numsp: numsp,
  	nvap: nvap,
  	nvdash: nvdash,
  	nvDash: nvDash,
  	nVdash: nVdash,
  	nVDash: nVDash,
  	nvge: nvge,
  	nvgt: nvgt,
  	nvHarr: nvHarr,
  	nvinfin: nvinfin,
  	nvlArr: nvlArr,
  	nvle: nvle,
  	nvlt: nvlt,
  	nvltrie: nvltrie,
  	nvrArr: nvrArr,
  	nvrtrie: nvrtrie,
  	nvsim: nvsim,
  	nwarhk: nwarhk,
  	nwarr: nwarr,
  	nwArr: nwArr,
  	nwarrow: nwarrow,
  	nwnear: nwnear,
  	Oacute: Oacute,
  	oacute: oacute,
  	oast: oast,
  	Ocirc: Ocirc,
  	ocirc: ocirc,
  	ocir: ocir,
  	Ocy: Ocy,
  	ocy: ocy,
  	odash: odash,
  	Odblac: Odblac,
  	odblac: odblac,
  	odiv: odiv,
  	odot: odot,
  	odsold: odsold,
  	OElig: OElig,
  	oelig: oelig,
  	ofcir: ofcir,
  	Ofr: Ofr,
  	ofr: ofr,
  	ogon: ogon,
  	Ograve: Ograve,
  	ograve: ograve,
  	ogt: ogt,
  	ohbar: ohbar,
  	ohm: ohm,
  	oint: oint,
  	olarr: olarr,
  	olcir: olcir,
  	olcross: olcross,
  	oline: oline,
  	olt: olt,
  	Omacr: Omacr,
  	omacr: omacr,
  	Omega: Omega,
  	omega: omega,
  	Omicron: Omicron,
  	omicron: omicron,
  	omid: omid,
  	ominus: ominus,
  	Oopf: Oopf,
  	oopf: oopf,
  	opar: opar,
  	OpenCurlyDoubleQuote: OpenCurlyDoubleQuote,
  	OpenCurlyQuote: OpenCurlyQuote,
  	operp: operp,
  	oplus: oplus,
  	orarr: orarr,
  	Or: Or,
  	or: or,
  	ord: ord,
  	order: order,
  	orderof: orderof,
  	ordf: ordf,
  	ordm: ordm,
  	origof: origof,
  	oror: oror,
  	orslope: orslope,
  	orv: orv,
  	oS: oS,
  	Oscr: Oscr,
  	oscr: oscr,
  	Oslash: Oslash,
  	oslash: oslash,
  	osol: osol,
  	Otilde: Otilde,
  	otilde: otilde,
  	otimesas: otimesas,
  	Otimes: Otimes,
  	otimes: otimes,
  	Ouml: Ouml,
  	ouml: ouml,
  	ovbar: ovbar,
  	OverBar: OverBar,
  	OverBrace: OverBrace,
  	OverBracket: OverBracket,
  	OverParenthesis: OverParenthesis,
  	para: para,
  	parallel: parallel,
  	par: par,
  	parsim: parsim,
  	parsl: parsl,
  	part: part,
  	PartialD: PartialD,
  	Pcy: Pcy,
  	pcy: pcy,
  	percnt: percnt,
  	period: period,
  	permil: permil,
  	perp: perp,
  	pertenk: pertenk,
  	Pfr: Pfr,
  	pfr: pfr,
  	Phi: Phi,
  	phi: phi,
  	phiv: phiv,
  	phmmat: phmmat,
  	phone: phone,
  	Pi: Pi,
  	pi: pi,
  	pitchfork: pitchfork,
  	piv: piv,
  	planck: planck,
  	planckh: planckh,
  	plankv: plankv,
  	plusacir: plusacir,
  	plusb: plusb,
  	pluscir: pluscir,
  	plus: plus,
  	plusdo: plusdo,
  	plusdu: plusdu,
  	pluse: pluse,
  	PlusMinus: PlusMinus,
  	plusmn: plusmn,
  	plussim: plussim,
  	plustwo: plustwo,
  	pm: pm,
  	Poincareplane: Poincareplane,
  	pointint: pointint,
  	popf: popf,
  	Popf: Popf,
  	pound: pound,
  	prap: prap,
  	Pr: Pr,
  	pr: pr,
  	prcue: prcue,
  	precapprox: precapprox,
  	prec: prec,
  	preccurlyeq: preccurlyeq,
  	Precedes: Precedes,
  	PrecedesEqual: PrecedesEqual,
  	PrecedesSlantEqual: PrecedesSlantEqual,
  	PrecedesTilde: PrecedesTilde,
  	preceq: preceq,
  	precnapprox: precnapprox,
  	precneqq: precneqq,
  	precnsim: precnsim,
  	pre: pre,
  	prE: prE,
  	precsim: precsim,
  	prime: prime,
  	Prime: Prime,
  	primes: primes,
  	prnap: prnap,
  	prnE: prnE,
  	prnsim: prnsim,
  	prod: prod,
  	Product: Product,
  	profalar: profalar,
  	profline: profline,
  	profsurf: profsurf,
  	prop: prop,
  	Proportional: Proportional,
  	Proportion: Proportion,
  	propto: propto,
  	prsim: prsim,
  	prurel: prurel,
  	Pscr: Pscr,
  	pscr: pscr,
  	Psi: Psi,
  	psi: psi,
  	puncsp: puncsp,
  	Qfr: Qfr,
  	qfr: qfr,
  	qint: qint,
  	qopf: qopf,
  	Qopf: Qopf,
  	qprime: qprime,
  	Qscr: Qscr,
  	qscr: qscr,
  	quaternions: quaternions,
  	quatint: quatint,
  	quest: quest,
  	questeq: questeq,
  	quot: quot$1,
  	QUOT: QUOT,
  	rAarr: rAarr,
  	race: race,
  	Racute: Racute,
  	racute: racute,
  	radic: radic,
  	raemptyv: raemptyv,
  	rang: rang,
  	Rang: Rang,
  	rangd: rangd,
  	range: range,
  	rangle: rangle,
  	raquo: raquo,
  	rarrap: rarrap,
  	rarrb: rarrb,
  	rarrbfs: rarrbfs,
  	rarrc: rarrc,
  	rarr: rarr,
  	Rarr: Rarr,
  	rArr: rArr,
  	rarrfs: rarrfs,
  	rarrhk: rarrhk,
  	rarrlp: rarrlp,
  	rarrpl: rarrpl,
  	rarrsim: rarrsim,
  	Rarrtl: Rarrtl,
  	rarrtl: rarrtl,
  	rarrw: rarrw,
  	ratail: ratail,
  	rAtail: rAtail,
  	ratio: ratio,
  	rationals: rationals,
  	rbarr: rbarr,
  	rBarr: rBarr,
  	RBarr: RBarr,
  	rbbrk: rbbrk,
  	rbrace: rbrace,
  	rbrack: rbrack,
  	rbrke: rbrke,
  	rbrksld: rbrksld,
  	rbrkslu: rbrkslu,
  	Rcaron: Rcaron,
  	rcaron: rcaron,
  	Rcedil: Rcedil,
  	rcedil: rcedil,
  	rceil: rceil,
  	rcub: rcub,
  	Rcy: Rcy,
  	rcy: rcy,
  	rdca: rdca,
  	rdldhar: rdldhar,
  	rdquo: rdquo,
  	rdquor: rdquor,
  	rdsh: rdsh,
  	real: real,
  	realine: realine,
  	realpart: realpart,
  	reals: reals,
  	Re: Re,
  	rect: rect,
  	reg: reg,
  	REG: REG,
  	ReverseElement: ReverseElement,
  	ReverseEquilibrium: ReverseEquilibrium,
  	ReverseUpEquilibrium: ReverseUpEquilibrium,
  	rfisht: rfisht,
  	rfloor: rfloor,
  	rfr: rfr,
  	Rfr: Rfr,
  	rHar: rHar,
  	rhard: rhard,
  	rharu: rharu,
  	rharul: rharul,
  	Rho: Rho,
  	rho: rho,
  	rhov: rhov,
  	RightAngleBracket: RightAngleBracket,
  	RightArrowBar: RightArrowBar,
  	rightarrow: rightarrow,
  	RightArrow: RightArrow,
  	Rightarrow: Rightarrow,
  	RightArrowLeftArrow: RightArrowLeftArrow,
  	rightarrowtail: rightarrowtail,
  	RightCeiling: RightCeiling,
  	RightDoubleBracket: RightDoubleBracket,
  	RightDownTeeVector: RightDownTeeVector,
  	RightDownVectorBar: RightDownVectorBar,
  	RightDownVector: RightDownVector,
  	RightFloor: RightFloor,
  	rightharpoondown: rightharpoondown,
  	rightharpoonup: rightharpoonup,
  	rightleftarrows: rightleftarrows,
  	rightleftharpoons: rightleftharpoons,
  	rightrightarrows: rightrightarrows,
  	rightsquigarrow: rightsquigarrow,
  	RightTeeArrow: RightTeeArrow,
  	RightTee: RightTee,
  	RightTeeVector: RightTeeVector,
  	rightthreetimes: rightthreetimes,
  	RightTriangleBar: RightTriangleBar,
  	RightTriangle: RightTriangle,
  	RightTriangleEqual: RightTriangleEqual,
  	RightUpDownVector: RightUpDownVector,
  	RightUpTeeVector: RightUpTeeVector,
  	RightUpVectorBar: RightUpVectorBar,
  	RightUpVector: RightUpVector,
  	RightVectorBar: RightVectorBar,
  	RightVector: RightVector,
  	ring: ring,
  	risingdotseq: risingdotseq,
  	rlarr: rlarr,
  	rlhar: rlhar,
  	rlm: rlm,
  	rmoustache: rmoustache,
  	rmoust: rmoust,
  	rnmid: rnmid,
  	roang: roang,
  	roarr: roarr,
  	robrk: robrk,
  	ropar: ropar,
  	ropf: ropf,
  	Ropf: Ropf,
  	roplus: roplus,
  	rotimes: rotimes,
  	RoundImplies: RoundImplies,
  	rpar: rpar,
  	rpargt: rpargt,
  	rppolint: rppolint,
  	rrarr: rrarr,
  	Rrightarrow: Rrightarrow,
  	rsaquo: rsaquo,
  	rscr: rscr,
  	Rscr: Rscr,
  	rsh: rsh,
  	Rsh: Rsh,
  	rsqb: rsqb,
  	rsquo: rsquo,
  	rsquor: rsquor,
  	rthree: rthree,
  	rtimes: rtimes,
  	rtri: rtri,
  	rtrie: rtrie,
  	rtrif: rtrif,
  	rtriltri: rtriltri,
  	RuleDelayed: RuleDelayed,
  	ruluhar: ruluhar,
  	rx: rx,
  	Sacute: Sacute,
  	sacute: sacute,
  	sbquo: sbquo,
  	scap: scap,
  	Scaron: Scaron,
  	scaron: scaron,
  	Sc: Sc,
  	sc: sc,
  	sccue: sccue,
  	sce: sce,
  	scE: scE,
  	Scedil: Scedil,
  	scedil: scedil,
  	Scirc: Scirc,
  	scirc: scirc,
  	scnap: scnap,
  	scnE: scnE,
  	scnsim: scnsim,
  	scpolint: scpolint,
  	scsim: scsim,
  	Scy: Scy,
  	scy: scy,
  	sdotb: sdotb,
  	sdot: sdot,
  	sdote: sdote,
  	searhk: searhk,
  	searr: searr,
  	seArr: seArr,
  	searrow: searrow,
  	sect: sect,
  	semi: semi,
  	seswar: seswar,
  	setminus: setminus,
  	setmn: setmn,
  	sext: sext,
  	Sfr: Sfr,
  	sfr: sfr,
  	sfrown: sfrown,
  	sharp: sharp,
  	SHCHcy: SHCHcy,
  	shchcy: shchcy,
  	SHcy: SHcy,
  	shcy: shcy,
  	ShortDownArrow: ShortDownArrow,
  	ShortLeftArrow: ShortLeftArrow,
  	shortmid: shortmid,
  	shortparallel: shortparallel,
  	ShortRightArrow: ShortRightArrow,
  	ShortUpArrow: ShortUpArrow,
  	shy: shy,
  	Sigma: Sigma,
  	sigma: sigma,
  	sigmaf: sigmaf,
  	sigmav: sigmav,
  	sim: sim,
  	simdot: simdot,
  	sime: sime,
  	simeq: simeq,
  	simg: simg,
  	simgE: simgE,
  	siml: siml,
  	simlE: simlE,
  	simne: simne,
  	simplus: simplus,
  	simrarr: simrarr,
  	slarr: slarr,
  	SmallCircle: SmallCircle,
  	smallsetminus: smallsetminus,
  	smashp: smashp,
  	smeparsl: smeparsl,
  	smid: smid,
  	smile: smile,
  	smt: smt,
  	smte: smte,
  	smtes: smtes,
  	SOFTcy: SOFTcy,
  	softcy: softcy,
  	solbar: solbar,
  	solb: solb,
  	sol: sol,
  	Sopf: Sopf,
  	sopf: sopf,
  	spades: spades,
  	spadesuit: spadesuit,
  	spar: spar,
  	sqcap: sqcap,
  	sqcaps: sqcaps,
  	sqcup: sqcup,
  	sqcups: sqcups,
  	Sqrt: Sqrt,
  	sqsub: sqsub,
  	sqsube: sqsube,
  	sqsubset: sqsubset,
  	sqsubseteq: sqsubseteq,
  	sqsup: sqsup,
  	sqsupe: sqsupe,
  	sqsupset: sqsupset,
  	sqsupseteq: sqsupseteq,
  	square: square,
  	Square: Square,
  	SquareIntersection: SquareIntersection,
  	SquareSubset: SquareSubset,
  	SquareSubsetEqual: SquareSubsetEqual,
  	SquareSuperset: SquareSuperset,
  	SquareSupersetEqual: SquareSupersetEqual,
  	SquareUnion: SquareUnion,
  	squarf: squarf,
  	squ: squ,
  	squf: squf,
  	srarr: srarr,
  	Sscr: Sscr,
  	sscr: sscr,
  	ssetmn: ssetmn,
  	ssmile: ssmile,
  	sstarf: sstarf,
  	Star: Star,
  	star: star,
  	starf: starf,
  	straightepsilon: straightepsilon,
  	straightphi: straightphi,
  	strns: strns,
  	sub: sub,
  	Sub: Sub,
  	subdot: subdot,
  	subE: subE,
  	sube: sube,
  	subedot: subedot,
  	submult: submult,
  	subnE: subnE,
  	subne: subne,
  	subplus: subplus,
  	subrarr: subrarr,
  	subset: subset,
  	Subset: Subset,
  	subseteq: subseteq,
  	subseteqq: subseteqq,
  	SubsetEqual: SubsetEqual,
  	subsetneq: subsetneq,
  	subsetneqq: subsetneqq,
  	subsim: subsim,
  	subsub: subsub,
  	subsup: subsup,
  	succapprox: succapprox,
  	succ: succ,
  	succcurlyeq: succcurlyeq,
  	Succeeds: Succeeds,
  	SucceedsEqual: SucceedsEqual,
  	SucceedsSlantEqual: SucceedsSlantEqual,
  	SucceedsTilde: SucceedsTilde,
  	succeq: succeq,
  	succnapprox: succnapprox,
  	succneqq: succneqq,
  	succnsim: succnsim,
  	succsim: succsim,
  	SuchThat: SuchThat,
  	sum: sum,
  	Sum: Sum,
  	sung: sung,
  	sup1: sup1,
  	sup2: sup2,
  	sup3: sup3,
  	sup: sup,
  	Sup: Sup,
  	supdot: supdot,
  	supdsub: supdsub,
  	supE: supE,
  	supe: supe,
  	supedot: supedot,
  	Superset: Superset,
  	SupersetEqual: SupersetEqual,
  	suphsol: suphsol,
  	suphsub: suphsub,
  	suplarr: suplarr,
  	supmult: supmult,
  	supnE: supnE,
  	supne: supne,
  	supplus: supplus,
  	supset: supset,
  	Supset: Supset,
  	supseteq: supseteq,
  	supseteqq: supseteqq,
  	supsetneq: supsetneq,
  	supsetneqq: supsetneqq,
  	supsim: supsim,
  	supsub: supsub,
  	supsup: supsup,
  	swarhk: swarhk,
  	swarr: swarr,
  	swArr: swArr,
  	swarrow: swarrow,
  	swnwar: swnwar,
  	szlig: szlig,
  	Tab: Tab,
  	target: target,
  	Tau: Tau,
  	tau: tau,
  	tbrk: tbrk,
  	Tcaron: Tcaron,
  	tcaron: tcaron,
  	Tcedil: Tcedil,
  	tcedil: tcedil,
  	Tcy: Tcy,
  	tcy: tcy,
  	tdot: tdot,
  	telrec: telrec,
  	Tfr: Tfr,
  	tfr: tfr,
  	there4: there4,
  	therefore: therefore,
  	Therefore: Therefore,
  	Theta: Theta,
  	theta: theta,
  	thetasym: thetasym,
  	thetav: thetav,
  	thickapprox: thickapprox,
  	thicksim: thicksim,
  	ThickSpace: ThickSpace,
  	ThinSpace: ThinSpace,
  	thinsp: thinsp,
  	thkap: thkap,
  	thksim: thksim,
  	THORN: THORN,
  	thorn: thorn,
  	tilde: tilde,
  	Tilde: Tilde,
  	TildeEqual: TildeEqual,
  	TildeFullEqual: TildeFullEqual,
  	TildeTilde: TildeTilde,
  	timesbar: timesbar,
  	timesb: timesb,
  	times: times,
  	timesd: timesd,
  	tint: tint,
  	toea: toea,
  	topbot: topbot,
  	topcir: topcir,
  	top: top,
  	Topf: Topf,
  	topf: topf,
  	topfork: topfork,
  	tosa: tosa,
  	tprime: tprime,
  	trade: trade,
  	TRADE: TRADE,
  	triangle: triangle,
  	triangledown: triangledown,
  	triangleleft: triangleleft,
  	trianglelefteq: trianglelefteq,
  	triangleq: triangleq,
  	triangleright: triangleright,
  	trianglerighteq: trianglerighteq,
  	tridot: tridot,
  	trie: trie,
  	triminus: triminus,
  	TripleDot: TripleDot,
  	triplus: triplus,
  	trisb: trisb,
  	tritime: tritime,
  	trpezium: trpezium,
  	Tscr: Tscr,
  	tscr: tscr,
  	TScy: TScy,
  	tscy: tscy,
  	TSHcy: TSHcy,
  	tshcy: tshcy,
  	Tstrok: Tstrok,
  	tstrok: tstrok,
  	twixt: twixt,
  	twoheadleftarrow: twoheadleftarrow,
  	twoheadrightarrow: twoheadrightarrow,
  	Uacute: Uacute,
  	uacute: uacute,
  	uarr: uarr,
  	Uarr: Uarr,
  	uArr: uArr,
  	Uarrocir: Uarrocir,
  	Ubrcy: Ubrcy,
  	ubrcy: ubrcy,
  	Ubreve: Ubreve,
  	ubreve: ubreve,
  	Ucirc: Ucirc,
  	ucirc: ucirc,
  	Ucy: Ucy,
  	ucy: ucy,
  	udarr: udarr,
  	Udblac: Udblac,
  	udblac: udblac,
  	udhar: udhar,
  	ufisht: ufisht,
  	Ufr: Ufr,
  	ufr: ufr,
  	Ugrave: Ugrave,
  	ugrave: ugrave,
  	uHar: uHar,
  	uharl: uharl,
  	uharr: uharr,
  	uhblk: uhblk,
  	ulcorn: ulcorn,
  	ulcorner: ulcorner,
  	ulcrop: ulcrop,
  	ultri: ultri,
  	Umacr: Umacr,
  	umacr: umacr,
  	uml: uml,
  	UnderBar: UnderBar,
  	UnderBrace: UnderBrace,
  	UnderBracket: UnderBracket,
  	UnderParenthesis: UnderParenthesis,
  	Union: Union,
  	UnionPlus: UnionPlus,
  	Uogon: Uogon,
  	uogon: uogon,
  	Uopf: Uopf,
  	uopf: uopf,
  	UpArrowBar: UpArrowBar,
  	uparrow: uparrow,
  	UpArrow: UpArrow,
  	Uparrow: Uparrow,
  	UpArrowDownArrow: UpArrowDownArrow,
  	updownarrow: updownarrow,
  	UpDownArrow: UpDownArrow,
  	Updownarrow: Updownarrow,
  	UpEquilibrium: UpEquilibrium,
  	upharpoonleft: upharpoonleft,
  	upharpoonright: upharpoonright,
  	uplus: uplus,
  	UpperLeftArrow: UpperLeftArrow,
  	UpperRightArrow: UpperRightArrow,
  	upsi: upsi,
  	Upsi: Upsi,
  	upsih: upsih,
  	Upsilon: Upsilon,
  	upsilon: upsilon,
  	UpTeeArrow: UpTeeArrow,
  	UpTee: UpTee,
  	upuparrows: upuparrows,
  	urcorn: urcorn,
  	urcorner: urcorner,
  	urcrop: urcrop,
  	Uring: Uring,
  	uring: uring,
  	urtri: urtri,
  	Uscr: Uscr,
  	uscr: uscr,
  	utdot: utdot,
  	Utilde: Utilde,
  	utilde: utilde,
  	utri: utri,
  	utrif: utrif,
  	uuarr: uuarr,
  	Uuml: Uuml,
  	uuml: uuml,
  	uwangle: uwangle,
  	vangrt: vangrt,
  	varepsilon: varepsilon,
  	varkappa: varkappa,
  	varnothing: varnothing,
  	varphi: varphi,
  	varpi: varpi,
  	varpropto: varpropto,
  	varr: varr,
  	vArr: vArr,
  	varrho: varrho,
  	varsigma: varsigma,
  	varsubsetneq: varsubsetneq,
  	varsubsetneqq: varsubsetneqq,
  	varsupsetneq: varsupsetneq,
  	varsupsetneqq: varsupsetneqq,
  	vartheta: vartheta,
  	vartriangleleft: vartriangleleft,
  	vartriangleright: vartriangleright,
  	vBar: vBar,
  	Vbar: Vbar,
  	vBarv: vBarv,
  	Vcy: Vcy,
  	vcy: vcy,
  	vdash: vdash,
  	vDash: vDash,
  	Vdash: Vdash,
  	VDash: VDash,
  	Vdashl: Vdashl,
  	veebar: veebar,
  	vee: vee,
  	Vee: Vee,
  	veeeq: veeeq,
  	vellip: vellip,
  	verbar: verbar,
  	Verbar: Verbar,
  	vert: vert,
  	Vert: Vert,
  	VerticalBar: VerticalBar,
  	VerticalLine: VerticalLine,
  	VerticalSeparator: VerticalSeparator,
  	VerticalTilde: VerticalTilde,
  	VeryThinSpace: VeryThinSpace,
  	Vfr: Vfr,
  	vfr: vfr,
  	vltri: vltri,
  	vnsub: vnsub,
  	vnsup: vnsup,
  	Vopf: Vopf,
  	vopf: vopf,
  	vprop: vprop,
  	vrtri: vrtri,
  	Vscr: Vscr,
  	vscr: vscr,
  	vsubnE: vsubnE,
  	vsubne: vsubne,
  	vsupnE: vsupnE,
  	vsupne: vsupne,
  	Vvdash: Vvdash,
  	vzigzag: vzigzag,
  	Wcirc: Wcirc,
  	wcirc: wcirc,
  	wedbar: wedbar,
  	wedge: wedge,
  	Wedge: Wedge,
  	wedgeq: wedgeq,
  	weierp: weierp,
  	Wfr: Wfr,
  	wfr: wfr,
  	Wopf: Wopf,
  	wopf: wopf,
  	wp: wp,
  	wr: wr,
  	wreath: wreath,
  	Wscr: Wscr,
  	wscr: wscr,
  	xcap: xcap,
  	xcirc: xcirc,
  	xcup: xcup,
  	xdtri: xdtri,
  	Xfr: Xfr,
  	xfr: xfr,
  	xharr: xharr,
  	xhArr: xhArr,
  	Xi: Xi,
  	xi: xi,
  	xlarr: xlarr,
  	xlArr: xlArr,
  	xmap: xmap,
  	xnis: xnis,
  	xodot: xodot,
  	Xopf: Xopf,
  	xopf: xopf,
  	xoplus: xoplus,
  	xotime: xotime,
  	xrarr: xrarr,
  	xrArr: xrArr,
  	Xscr: Xscr,
  	xscr: xscr,
  	xsqcup: xsqcup,
  	xuplus: xuplus,
  	xutri: xutri,
  	xvee: xvee,
  	xwedge: xwedge,
  	Yacute: Yacute,
  	yacute: yacute,
  	YAcy: YAcy,
  	yacy: yacy,
  	Ycirc: Ycirc,
  	ycirc: ycirc,
  	Ycy: Ycy,
  	ycy: ycy,
  	yen: yen,
  	Yfr: Yfr,
  	yfr: yfr,
  	YIcy: YIcy,
  	yicy: yicy,
  	Yopf: Yopf,
  	yopf: yopf,
  	Yscr: Yscr,
  	yscr: yscr,
  	YUcy: YUcy,
  	yucy: yucy,
  	yuml: yuml,
  	Yuml: Yuml,
  	Zacute: Zacute,
  	zacute: zacute,
  	Zcaron: Zcaron,
  	zcaron: zcaron,
  	Zcy: Zcy,
  	zcy: zcy,
  	Zdot: Zdot,
  	zdot: zdot,
  	zeetrf: zeetrf,
  	ZeroWidthSpace: ZeroWidthSpace,
  	Zeta: Zeta,
  	zeta: zeta,
  	zfr: zfr,
  	Zfr: Zfr,
  	ZHcy: ZHcy,
  	zhcy: zhcy,
  	zigrarr: zigrarr,
  	zopf: zopf,
  	Zopf: Zopf,
  	Zscr: Zscr,
  	zscr: zscr,
  	zwj: zwj,
  	zwnj: zwnj,
  	default: entitiesJSON
  });

  var require$$0 = ( xml && xmlJSON ) || xml;

  var require$$1 = ( entities && entitiesJSON ) || entities;

  var inverseXML = getInverseObj$1(require$$0);
  var xmlReplacer = getInverseReplacer$1(inverseXML);

  var XML = getInverse$1(inverseXML, xmlReplacer);

  var inverseHTML = getInverseObj$1(require$$1);
  var htmlReplacer = getInverseReplacer$1(inverseHTML);

  var HTML = getInverse$1(inverseHTML, htmlReplacer);

  function getInverseObj$1(obj){
  	return Object.keys(obj).sort().reduce(function(inverse, name){
  		inverse[obj[name]] = "&" + name + ";";
  		return inverse;
  	}, {});
  }

  function getInverseReplacer$1(inverse){
  	var single = [],
  	    multiple = [];

  	Object.keys(inverse).forEach(function(k){
  		if(k.length === 1){
  			single.push("\\" + k);
  		} else {
  			multiple.push(k);
  		}
  	});

  	
  	multiple.unshift("[" + single.join("") + "]");

  	return new RegExp(multiple.join("|"), "g");
  }

  var re_nonASCII = /[^\0-\x7F]/g;
  var re_astralSymbols = /[\uD800-\uDBFF][\uDC00-\uDFFF]/g;

  function singleCharReplacer$1(c){
  	return "&#x" + c.charCodeAt(0).toString(16).toUpperCase() + ";";
  }

  function astralReplacer$1(c){
  	
  	var high = c.charCodeAt(0);
  	var low  = c.charCodeAt(1);
  	var codePoint = (high - 0xD800) * 0x400 + low - 0xDC00 + 0x10000;
  	return "&#x" + codePoint.toString(16).toUpperCase() + ";";
  }

  function getInverse$1(inverse, re){
  	function func(name){
  		return inverse[name];
  	}

  	return function(data){
  		return data
  				.replace(re, func)
  				.replace(re_astralSymbols, astralReplacer$1)
  				.replace(re_nonASCII, singleCharReplacer$1);
  	};
  }

  var re_xmlChars = getInverseReplacer$1(inverseXML);

  function escapeXML(data){
  	return data
  			.replace(re_xmlChars, singleCharReplacer$1)
  			.replace(re_astralSymbols, astralReplacer$1)
  			.replace(re_nonASCII, singleCharReplacer$1);
  }

  var escape = escapeXML;

  var encode = {
  	XML: XML,
  	HTML: HTML,
  	escape: escape
  };

  var decode = {
  	"0": 65533,
  	"128": 8364,
  	"130": 8218,
  	"131": 402,
  	"132": 8222,
  	"133": 8230,
  	"134": 8224,
  	"135": 8225,
  	"136": 710,
  	"137": 8240,
  	"138": 352,
  	"139": 8249,
  	"140": 338,
  	"142": 381,
  	"145": 8216,
  	"146": 8217,
  	"147": 8220,
  	"148": 8221,
  	"149": 8226,
  	"150": 8211,
  	"151": 8212,
  	"152": 732,
  	"153": 8482,
  	"154": 353,
  	"155": 8250,
  	"156": 339,
  	"158": 382,
  	"159": 376
  };

  var decode$1 = Object.freeze({
  	default: decode
  });

  var decodeMap = ( decode$1 && decode ) || decode$1;

  var decode_codepoint = decodeCodePoint;


  function decodeCodePoint(codePoint){

  	if((codePoint >= 0xD800 && codePoint <= 0xDFFF) || codePoint > 0x10FFFF){
  		return "\uFFFD";
  	}

  	if(codePoint in decodeMap){
  		codePoint = decodeMap[codePoint];
  	}

  	var output = "";

  	if(codePoint > 0xFFFF){
  		codePoint -= 0x10000;
  		output += String.fromCharCode(codePoint >>> 10 & 0x3FF | 0xD800);
  		codePoint = 0xDC00 | codePoint & 0x3FF;
  	}

  	output += String.fromCharCode(codePoint);
  	return output;
  }

  var Aacute$1 = "Á";
  var aacute$1 = "á";
  var Acirc$1 = "Â";
  var acirc$1 = "â";
  var acute$1 = "´";
  var AElig$1 = "Æ";
  var aelig$1 = "æ";
  var Agrave$1 = "À";
  var agrave$1 = "à";
  var amp$2 = "&";
  var AMP$1 = "&";
  var Aring$1 = "Å";
  var aring$1 = "å";
  var Atilde$1 = "Ã";
  var atilde$1 = "ã";
  var Auml$1 = "Ä";
  var auml$1 = "ä";
  var brvbar$1 = "¦";
  var Ccedil$1 = "Ç";
  var ccedil$1 = "ç";
  var cedil$1 = "¸";
  var cent$1 = "¢";
  var copy$1 = "©";
  var COPY$1 = "©";
  var curren$1 = "¤";
  var deg$1 = "°";
  var divide$1 = "÷";
  var Eacute$1 = "É";
  var eacute$1 = "é";
  var Ecirc$1 = "Ê";
  var ecirc$1 = "ê";
  var Egrave$1 = "È";
  var egrave$1 = "è";
  var ETH$1 = "Ð";
  var eth$1 = "ð";
  var Euml$1 = "Ë";
  var euml$1 = "ë";
  var frac12$1 = "½";
  var frac14$1 = "¼";
  var frac34$1 = "¾";
  var gt$2 = ">";
  var GT$1 = ">";
  var Iacute$1 = "Í";
  var iacute$1 = "í";
  var Icirc$1 = "Î";
  var icirc$1 = "î";
  var iexcl$1 = "¡";
  var Igrave$1 = "Ì";
  var igrave$1 = "ì";
  var iquest$1 = "¿";
  var Iuml$1 = "Ï";
  var iuml$1 = "ï";
  var laquo$1 = "«";
  var lt$2 = "<";
  var LT$1 = "<";
  var macr$1 = "¯";
  var micro$1 = "µ";
  var middot$1 = "·";
  var nbsp$1 = " ";
  var not$1 = "¬";
  var Ntilde$1 = "Ñ";
  var ntilde$1 = "ñ";
  var Oacute$1 = "Ó";
  var oacute$1 = "ó";
  var Ocirc$1 = "Ô";
  var ocirc$1 = "ô";
  var Ograve$1 = "Ò";
  var ograve$1 = "ò";
  var ordf$1 = "ª";
  var ordm$1 = "º";
  var Oslash$1 = "Ø";
  var oslash$1 = "ø";
  var Otilde$1 = "Õ";
  var otilde$1 = "õ";
  var Ouml$1 = "Ö";
  var ouml$1 = "ö";
  var para$1 = "¶";
  var plusmn$1 = "±";
  var pound$1 = "£";
  var quot$2 = "\"";
  var QUOT$1 = "\"";
  var raquo$1 = "»";
  var reg$1 = "®";
  var REG$1 = "®";
  var sect$1 = "§";
  var shy$1 = "­";
  var sup1$1 = "¹";
  var sup2$1 = "²";
  var sup3$1 = "³";
  var szlig$1 = "ß";
  var THORN$1 = "Þ";
  var thorn$1 = "þ";
  var times$1 = "×";
  var Uacute$1 = "Ú";
  var uacute$1 = "ú";
  var Ucirc$1 = "Û";
  var ucirc$1 = "û";
  var Ugrave$1 = "Ù";
  var ugrave$1 = "ù";
  var uml$1 = "¨";
  var Uuml$1 = "Ü";
  var uuml$1 = "ü";
  var Yacute$1 = "Ý";
  var yacute$1 = "ý";
  var yen$1 = "¥";
  var yuml$1 = "ÿ";
  var legacyJSON = {
  	Aacute: Aacute$1,
  	aacute: aacute$1,
  	Acirc: Acirc$1,
  	acirc: acirc$1,
  	acute: acute$1,
  	AElig: AElig$1,
  	aelig: aelig$1,
  	Agrave: Agrave$1,
  	agrave: agrave$1,
  	amp: amp$2,
  	AMP: AMP$1,
  	Aring: Aring$1,
  	aring: aring$1,
  	Atilde: Atilde$1,
  	atilde: atilde$1,
  	Auml: Auml$1,
  	auml: auml$1,
  	brvbar: brvbar$1,
  	Ccedil: Ccedil$1,
  	ccedil: ccedil$1,
  	cedil: cedil$1,
  	cent: cent$1,
  	copy: copy$1,
  	COPY: COPY$1,
  	curren: curren$1,
  	deg: deg$1,
  	divide: divide$1,
  	Eacute: Eacute$1,
  	eacute: eacute$1,
  	Ecirc: Ecirc$1,
  	ecirc: ecirc$1,
  	Egrave: Egrave$1,
  	egrave: egrave$1,
  	ETH: ETH$1,
  	eth: eth$1,
  	Euml: Euml$1,
  	euml: euml$1,
  	frac12: frac12$1,
  	frac14: frac14$1,
  	frac34: frac34$1,
  	gt: gt$2,
  	GT: GT$1,
  	Iacute: Iacute$1,
  	iacute: iacute$1,
  	Icirc: Icirc$1,
  	icirc: icirc$1,
  	iexcl: iexcl$1,
  	Igrave: Igrave$1,
  	igrave: igrave$1,
  	iquest: iquest$1,
  	Iuml: Iuml$1,
  	iuml: iuml$1,
  	laquo: laquo$1,
  	lt: lt$2,
  	LT: LT$1,
  	macr: macr$1,
  	micro: micro$1,
  	middot: middot$1,
  	nbsp: nbsp$1,
  	not: not$1,
  	Ntilde: Ntilde$1,
  	ntilde: ntilde$1,
  	Oacute: Oacute$1,
  	oacute: oacute$1,
  	Ocirc: Ocirc$1,
  	ocirc: ocirc$1,
  	Ograve: Ograve$1,
  	ograve: ograve$1,
  	ordf: ordf$1,
  	ordm: ordm$1,
  	Oslash: Oslash$1,
  	oslash: oslash$1,
  	Otilde: Otilde$1,
  	otilde: otilde$1,
  	Ouml: Ouml$1,
  	ouml: ouml$1,
  	para: para$1,
  	plusmn: plusmn$1,
  	pound: pound$1,
  	quot: quot$2,
  	QUOT: QUOT$1,
  	raquo: raquo$1,
  	reg: reg$1,
  	REG: REG$1,
  	sect: sect$1,
  	shy: shy$1,
  	sup1: sup1$1,
  	sup2: sup2$1,
  	sup3: sup3$1,
  	szlig: szlig$1,
  	THORN: THORN$1,
  	thorn: thorn$1,
  	times: times$1,
  	Uacute: Uacute$1,
  	uacute: uacute$1,
  	Ucirc: Ucirc$1,
  	ucirc: ucirc$1,
  	Ugrave: Ugrave$1,
  	ugrave: ugrave$1,
  	uml: uml$1,
  	Uuml: Uuml$1,
  	uuml: uuml$1,
  	Yacute: Yacute$1,
  	yacute: yacute$1,
  	yen: yen$1,
  	yuml: yuml$1
  };

  var _entities = {
    encodeXML: encode.XML,
    decodeCodepoint: decode_codepoint,
    entitiesJSON,
    legacyJSON,
    xmlJSON
  };

  var _entities_decodeCodepoint = _entities.decodeCodepoint;

  var _entities_decodeCodepoint$1 = Object.freeze({
  	default: _entities_decodeCodepoint
  });

  var _entities_entitiesJSON = _entities.entitiesJSON;


  var _entities_entitiesJSON$1 = Object.freeze({
  	default: _entities_entitiesJSON
  });

  var _entities_legacyJSON = _entities.legacyJSON;

  var _entities_legacyJSON$1 = Object.freeze({
  	default: _entities_legacyJSON
  });

  var _entities_xmlJSON = _entities.xmlJSON;

  var _entities_xmlJSON$1 = Object.freeze({
  	default: _entities_xmlJSON
  });

  var decodeCodePoint$1 = ( _entities_decodeCodepoint$1 && _entities_decodeCodepoint ) || _entities_decodeCodepoint$1;

  var entityMap = ( _entities_entitiesJSON$1 && _entities_entitiesJSON ) || _entities_entitiesJSON$1;

  var legacyMap = ( _entities_legacyJSON$1 && _entities_legacyJSON ) || _entities_legacyJSON$1;

  var xmlMap = ( _entities_xmlJSON$1 && _entities_xmlJSON ) || _entities_xmlJSON$1;

  var Tokenizer_1 = Tokenizer$1;

  var i = 0;
  var TEXT                      = i++;
  var BEFORE_TAG_NAME           = i++;
  var IN_TAG_NAME               = i++;
  var IN_SELF_CLOSING_TAG       = i++;
  var BEFORE_CLOSING_TAG_NAME   = i++;
  var IN_CLOSING_TAG_NAME       = i++;
  var AFTER_CLOSING_TAG_NAME    = i++;
  var BEFORE_ATTRIBUTE_NAME     = i++;
  var IN_ATTRIBUTE_NAME         = i++;
  var AFTER_ATTRIBUTE_NAME      = i++;
  var BEFORE_ATTRIBUTE_VALUE    = i++;
  var IN_ATTRIBUTE_VALUE_DQ     = i++;
  var IN_ATTRIBUTE_VALUE_SQ     = i++;
  var IN_ATTRIBUTE_VALUE_NQ     = i++;
  var BEFORE_DECLARATION        = i++;
  var IN_DECLARATION            = i++;
  var IN_PROCESSING_INSTRUCTION = i++;
  var BEFORE_COMMENT            = i++;
  var IN_COMMENT                = i++;
  var AFTER_COMMENT_1           = i++;
  var AFTER_COMMENT_2           = i++;
  var BEFORE_CDATA_1            = i++;
  var BEFORE_CDATA_2            = i++;
  var BEFORE_CDATA_3            = i++;
  var BEFORE_CDATA_4            = i++;
  var BEFORE_CDATA_5            = i++;
  var BEFORE_CDATA_6            = i++;
  var IN_CDATA                  = i++;
  var AFTER_CDATA_1             = i++;
  var AFTER_CDATA_2             = i++;
  var BEFORE_SPECIAL            = i++;
  var BEFORE_SPECIAL_END        = i++;
  var BEFORE_SCRIPT_1           = i++;
  var BEFORE_SCRIPT_2           = i++;
  var BEFORE_SCRIPT_3           = i++;
  var BEFORE_SCRIPT_4           = i++;
  var BEFORE_SCRIPT_5           = i++;
  var AFTER_SCRIPT_1            = i++;
  var AFTER_SCRIPT_2            = i++;
  var AFTER_SCRIPT_3            = i++;
  var AFTER_SCRIPT_4            = i++;
  var AFTER_SCRIPT_5            = i++;
  var BEFORE_STYLE_1            = i++;
  var BEFORE_STYLE_2            = i++;
  var BEFORE_STYLE_3            = i++;
  var BEFORE_STYLE_4            = i++;
  var AFTER_STYLE_1             = i++;
  var AFTER_STYLE_2             = i++;
  var AFTER_STYLE_3             = i++;
  var AFTER_STYLE_4             = i++;
  var BEFORE_ENTITY             = i++;
  var BEFORE_NUMERIC_ENTITY     = i++;
  var IN_NAMED_ENTITY           = i++;
  var IN_NUMERIC_ENTITY         = i++;
  var IN_HEX_ENTITY             = i++;
  var j = 0;
  var SPECIAL_NONE              = j++;
  var SPECIAL_SCRIPT            = j++;
  var SPECIAL_STYLE             = j++;

  function whitespace(c){
  	return c === " " || c === "\n" || c === "\t" || c === "\f" || c === "\r";
  }

  function characterState(char, SUCCESS){
  	return function(c){
  		if(c === char) this._state = SUCCESS;
  	};
  }

  function ifElseState(upper, SUCCESS, FAILURE){
  	var lower = upper.toLowerCase();

  	if(upper === lower){
  		return function(c){
  			if(c === lower){
  				this._state = SUCCESS;
  			} else {
  				this._state = FAILURE;
  				this._index--;
  			}
  		};
  	} else {
  		return function(c){
  			if(c === lower || c === upper){
  				this._state = SUCCESS;
  			} else {
  				this._state = FAILURE;
  				this._index--;
  			}
  		};
  	}
  }

  function consumeSpecialNameChar(upper, NEXT_STATE){
  	var lower = upper.toLowerCase();

  	return function(c){
  		if(c === lower || c === upper){
  			this._state = NEXT_STATE;
  		} else {
  			this._state = IN_TAG_NAME;
  			this._index--; 
  		}
  	};
  }

  function Tokenizer$1(options, cbs){
  	this._state = TEXT;
  	this._buffer = "";
  	this._sectionStart = 0;
  	this._index = 0;
  	this._bufferOffset = 0; 
  	this._baseState = TEXT;
  	this._special = SPECIAL_NONE;
  	this._cbs = cbs;
  	this._running = true;
  	this._ended = false;
  	this._xmlMode = !!(options && options.xmlMode);
  	this._decodeEntities = !!(options && options.decodeEntities);
  }

  Tokenizer$1.prototype._stateText = function(c){
  	if(c === "<"){
  		if(this._index > this._sectionStart){
  			this._cbs.ontext(this._getSection());
  		}
  		this._state = BEFORE_TAG_NAME;
  		this._sectionStart = this._index;
  	} else if(this._decodeEntities && this._special === SPECIAL_NONE && c === "&"){
  		if(this._index > this._sectionStart){
  			this._cbs.ontext(this._getSection());
  		}
  		this._baseState = TEXT;
  		this._state = BEFORE_ENTITY;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1.prototype._stateBeforeTagName = function(c){
  	if(c === "/"){
  		this._state = BEFORE_CLOSING_TAG_NAME;
  	} else if(c === "<"){
  		this._cbs.ontext(this._getSection());
  		this._sectionStart = this._index;
  	} else if(c === ">" || this._special !== SPECIAL_NONE || whitespace(c)) {
  		this._state = TEXT;
  	} else if(c === "!"){
  		this._state = BEFORE_DECLARATION;
  		this._sectionStart = this._index + 1;
  	} else if(c === "?"){
  		this._state = IN_PROCESSING_INSTRUCTION;
  		this._sectionStart = this._index + 1;
  	} else {
  		this._state = (!this._xmlMode && (c === "s" || c === "S")) ?
  						BEFORE_SPECIAL : IN_TAG_NAME;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1.prototype._stateInTagName = function(c){
  	if(c === "/" || c === ">" || whitespace(c)){
  		this._emitToken("onopentagname");
  		this._state = BEFORE_ATTRIBUTE_NAME;
  		this._index--;
  	}
  };

  Tokenizer$1.prototype._stateBeforeCloseingTagName = function(c){
  	if(whitespace(c));
  	else if(c === ">"){
  		this._state = TEXT;
  	} else if(this._special !== SPECIAL_NONE){
  		if(c === "s" || c === "S"){
  			this._state = BEFORE_SPECIAL_END;
  		} else {
  			this._state = TEXT;
  			this._index--;
  		}
  	} else {
  		this._state = IN_CLOSING_TAG_NAME;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1.prototype._stateInCloseingTagName = function(c){
  	if(c === ">" || whitespace(c)){
  		this._emitToken("onclosetag");
  		this._state = AFTER_CLOSING_TAG_NAME;
  		this._index--;
  	}
  };

  Tokenizer$1.prototype._stateAfterCloseingTagName = function(c){
  	
  	if(c === ">"){
  		this._state = TEXT;
  		this._sectionStart = this._index + 1;
  	}
  };

  Tokenizer$1.prototype._stateBeforeAttributeName = function(c){
  	if(c === ">"){
  		this._cbs.onopentagend();
  		this._state = TEXT;
  		this._sectionStart = this._index + 1;
  	} else if(c === "/"){
  		this._state = IN_SELF_CLOSING_TAG;
  	} else if(!whitespace(c)){
  		this._state = IN_ATTRIBUTE_NAME;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1.prototype._stateInSelfClosingTag = function(c){
  	if(c === ">"){
  		this._cbs.onselfclosingtag();
  		this._state = TEXT;
  		this._sectionStart = this._index + 1;
  	} else if(!whitespace(c)){
  		this._state = BEFORE_ATTRIBUTE_NAME;
  		this._index--;
  	}
  };

  Tokenizer$1.prototype._stateInAttributeName = function(c){
  	if(c === "=" || c === "/" || c === ">" || whitespace(c)){
  		this._cbs.onattribname(this._getSection());
  		this._sectionStart = -1;
  		this._state = AFTER_ATTRIBUTE_NAME;
  		this._index--;
  	}
  };

  Tokenizer$1.prototype._stateAfterAttributeName = function(c){
  	if(c === "="){
  		this._state = BEFORE_ATTRIBUTE_VALUE;
  	} else if(c === "/" || c === ">"){
  		this._cbs.onattribend();
  		this._state = BEFORE_ATTRIBUTE_NAME;
  		this._index--;
  	} else if(!whitespace(c)){
  		this._cbs.onattribend();
  		this._state = IN_ATTRIBUTE_NAME;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1.prototype._stateBeforeAttributeValue = function(c){
  	if(c === "\""){
  		this._state = IN_ATTRIBUTE_VALUE_DQ;
  		this._sectionStart = this._index + 1;
  	} else if(c === "'"){
  		this._state = IN_ATTRIBUTE_VALUE_SQ;
  		this._sectionStart = this._index + 1;
  	} else if(!whitespace(c)){
  		this._state = IN_ATTRIBUTE_VALUE_NQ;
  		this._sectionStart = this._index;
  		this._index--; 
  	}
  };

  Tokenizer$1.prototype._stateInAttributeValueDoubleQuotes = function(c){
  	if(c === "\""){
  		this._emitToken("onattribdata");
  		this._cbs.onattribend();
  		this._state = BEFORE_ATTRIBUTE_NAME;
  	} else if(this._decodeEntities && c === "&"){
  		this._emitToken("onattribdata");
  		this._baseState = this._state;
  		this._state = BEFORE_ENTITY;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1.prototype._stateInAttributeValueSingleQuotes = function(c){
  	if(c === "'"){
  		this._emitToken("onattribdata");
  		this._cbs.onattribend();
  		this._state = BEFORE_ATTRIBUTE_NAME;
  	} else if(this._decodeEntities && c === "&"){
  		this._emitToken("onattribdata");
  		this._baseState = this._state;
  		this._state = BEFORE_ENTITY;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1.prototype._stateInAttributeValueNoQuotes = function(c){
  	if(whitespace(c) || c === ">"){
  		this._emitToken("onattribdata");
  		this._cbs.onattribend();
  		this._state = BEFORE_ATTRIBUTE_NAME;
  		this._index--;
  	} else if(this._decodeEntities && c === "&"){
  		this._emitToken("onattribdata");
  		this._baseState = this._state;
  		this._state = BEFORE_ENTITY;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1.prototype._stateBeforeDeclaration = function(c){
  	this._state = c === "[" ? BEFORE_CDATA_1 :
  					c === "-" ? BEFORE_COMMENT :
  						IN_DECLARATION;
  };

  Tokenizer$1.prototype._stateInDeclaration = function(c){
  	if(c === ">"){
  		this._cbs.ondeclaration(this._getSection());
  		this._state = TEXT;
  		this._sectionStart = this._index + 1;
  	}
  };

  Tokenizer$1.prototype._stateInProcessingInstruction = function(c){
  	if(c === ">"){
  		this._cbs.onprocessinginstruction(this._getSection());
  		this._state = TEXT;
  		this._sectionStart = this._index + 1;
  	}
  };

  Tokenizer$1.prototype._stateBeforeComment = function(c){
  	if(c === "-"){
  		this._state = IN_COMMENT;
  		this._sectionStart = this._index + 1;
  	} else {
  		this._state = IN_DECLARATION;
  	}
  };

  Tokenizer$1.prototype._stateInComment = function(c){
  	if(c === "-") this._state = AFTER_COMMENT_1;
  };

  Tokenizer$1.prototype._stateAfterComment1 = function(c){
  	if(c === "-"){
  		this._state = AFTER_COMMENT_2;
  	} else {
  		this._state = IN_COMMENT;
  	}
  };

  Tokenizer$1.prototype._stateAfterComment2 = function(c){
  	if(c === ">"){
  		
  		this._cbs.oncomment(this._buffer.substring(this._sectionStart, this._index - 2));
  		this._state = TEXT;
  		this._sectionStart = this._index + 1;
  	} else if(c !== "-"){
  		this._state = IN_COMMENT;
  	}
  	
  };

  Tokenizer$1.prototype._stateBeforeCdata1 = ifElseState("C", BEFORE_CDATA_2, IN_DECLARATION);
  Tokenizer$1.prototype._stateBeforeCdata2 = ifElseState("D", BEFORE_CDATA_3, IN_DECLARATION);
  Tokenizer$1.prototype._stateBeforeCdata3 = ifElseState("A", BEFORE_CDATA_4, IN_DECLARATION);
  Tokenizer$1.prototype._stateBeforeCdata4 = ifElseState("T", BEFORE_CDATA_5, IN_DECLARATION);
  Tokenizer$1.prototype._stateBeforeCdata5 = ifElseState("A", BEFORE_CDATA_6, IN_DECLARATION);

  Tokenizer$1.prototype._stateBeforeCdata6 = function(c){
  	if(c === "["){
  		this._state = IN_CDATA;
  		this._sectionStart = this._index + 1;
  	} else {
  		this._state = IN_DECLARATION;
  		this._index--;
  	}
  };

  Tokenizer$1.prototype._stateInCdata = function(c){
  	if(c === "]") this._state = AFTER_CDATA_1;
  };

  Tokenizer$1.prototype._stateAfterCdata1 = characterState("]", AFTER_CDATA_2);

  Tokenizer$1.prototype._stateAfterCdata2 = function(c){
  	if(c === ">"){
  		
  		this._cbs.oncdata(this._buffer.substring(this._sectionStart, this._index - 2));
  		this._state = TEXT;
  		this._sectionStart = this._index + 1;
  	} else if(c !== "]") {
  		this._state = IN_CDATA;
  	}
  	
  };

  Tokenizer$1.prototype._stateBeforeSpecial = function(c){
  	if(c === "c" || c === "C"){
  		this._state = BEFORE_SCRIPT_1;
  	} else if(c === "t" || c === "T"){
  		this._state = BEFORE_STYLE_1;
  	} else {
  		this._state = IN_TAG_NAME;
  		this._index--; 
  	}
  };

  Tokenizer$1.prototype._stateBeforeSpecialEnd = function(c){
  	if(this._special === SPECIAL_SCRIPT && (c === "c" || c === "C")){
  		this._state = AFTER_SCRIPT_1;
  	} else if(this._special === SPECIAL_STYLE && (c === "t" || c === "T")){
  		this._state = AFTER_STYLE_1;
  	}
  	else this._state = TEXT;
  };

  Tokenizer$1.prototype._stateBeforeScript1 = consumeSpecialNameChar("R", BEFORE_SCRIPT_2);
  Tokenizer$1.prototype._stateBeforeScript2 = consumeSpecialNameChar("I", BEFORE_SCRIPT_3);
  Tokenizer$1.prototype._stateBeforeScript3 = consumeSpecialNameChar("P", BEFORE_SCRIPT_4);
  Tokenizer$1.prototype._stateBeforeScript4 = consumeSpecialNameChar("T", BEFORE_SCRIPT_5);

  Tokenizer$1.prototype._stateBeforeScript5 = function(c){
  	if(c === "/" || c === ">" || whitespace(c)){
  		this._special = SPECIAL_SCRIPT;
  	}
  	this._state = IN_TAG_NAME;
  	this._index--; 
  };

  Tokenizer$1.prototype._stateAfterScript1 = ifElseState("R", AFTER_SCRIPT_2, TEXT);
  Tokenizer$1.prototype._stateAfterScript2 = ifElseState("I", AFTER_SCRIPT_3, TEXT);
  Tokenizer$1.prototype._stateAfterScript3 = ifElseState("P", AFTER_SCRIPT_4, TEXT);
  Tokenizer$1.prototype._stateAfterScript4 = ifElseState("T", AFTER_SCRIPT_5, TEXT);

  Tokenizer$1.prototype._stateAfterScript5 = function(c){
  	if(c === ">" || whitespace(c)){
  		this._special = SPECIAL_NONE;
  		this._state = IN_CLOSING_TAG_NAME;
  		this._sectionStart = this._index - 6;
  		this._index--; 
  	}
  	else this._state = TEXT;
  };

  Tokenizer$1.prototype._stateBeforeStyle1 = consumeSpecialNameChar("Y", BEFORE_STYLE_2);
  Tokenizer$1.prototype._stateBeforeStyle2 = consumeSpecialNameChar("L", BEFORE_STYLE_3);
  Tokenizer$1.prototype._stateBeforeStyle3 = consumeSpecialNameChar("E", BEFORE_STYLE_4);

  Tokenizer$1.prototype._stateBeforeStyle4 = function(c){
  	if(c === "/" || c === ">" || whitespace(c)){
  		this._special = SPECIAL_STYLE;
  	}
  	this._state = IN_TAG_NAME;
  	this._index--; 
  };

  Tokenizer$1.prototype._stateAfterStyle1 = ifElseState("Y", AFTER_STYLE_2, TEXT);
  Tokenizer$1.prototype._stateAfterStyle2 = ifElseState("L", AFTER_STYLE_3, TEXT);
  Tokenizer$1.prototype._stateAfterStyle3 = ifElseState("E", AFTER_STYLE_4, TEXT);

  Tokenizer$1.prototype._stateAfterStyle4 = function(c){
  	if(c === ">" || whitespace(c)){
  		this._special = SPECIAL_NONE;
  		this._state = IN_CLOSING_TAG_NAME;
  		this._sectionStart = this._index - 5;
  		this._index--; 
  	}
  	else this._state = TEXT;
  };

  Tokenizer$1.prototype._stateBeforeEntity = ifElseState("#", BEFORE_NUMERIC_ENTITY, IN_NAMED_ENTITY);
  Tokenizer$1.prototype._stateBeforeNumericEntity = ifElseState("X", IN_HEX_ENTITY, IN_NUMERIC_ENTITY);


  Tokenizer$1.prototype._parseNamedEntityStrict = function(){
  	
  	if(this._sectionStart + 1 < this._index){
  		var entity = this._buffer.substring(this._sectionStart + 1, this._index),
  		    map = this._xmlMode ? xmlMap : entityMap;

  		if(map.hasOwnProperty(entity)){
  			this._emitPartial(map[entity]);
  			this._sectionStart = this._index + 1;
  		}
  	}
  };



  Tokenizer$1.prototype._parseLegacyEntity = function(){
  	var start = this._sectionStart + 1,
  	    limit = this._index - start;

  	if(limit > 6) limit = 6; 

  	while(limit >= 2){ 
  		var entity = this._buffer.substr(start, limit);

  		if(legacyMap.hasOwnProperty(entity)){
  			this._emitPartial(legacyMap[entity]);
  			this._sectionStart += limit + 1;
  			return;
  		} else {
  			limit--;
  		}
  	}
  };

  Tokenizer$1.prototype._stateInNamedEntity = function(c){
  	if(c === ";"){
  		this._parseNamedEntityStrict();
  		if(this._sectionStart + 1 < this._index && !this._xmlMode){
  			this._parseLegacyEntity();
  		}
  		this._state = this._baseState;
  	} else if((c < "a" || c > "z") && (c < "A" || c > "Z") && (c < "0" || c > "9")){
  		if(this._xmlMode);
  		else if(this._sectionStart + 1 === this._index);
  		else if(this._baseState !== TEXT){
  			if(c !== "="){
  				this._parseNamedEntityStrict();
  			}
  		} else {
  			this._parseLegacyEntity();
  		}

  		this._state = this._baseState;
  		this._index--;
  	}
  };

  Tokenizer$1.prototype._decodeNumericEntity = function(offset, base){
  	var sectionStart = this._sectionStart + offset;

  	if(sectionStart !== this._index){
  		
  		var entity = this._buffer.substring(sectionStart, this._index);
  		var parsed = parseInt(entity, base);

  		this._emitPartial(decodeCodePoint$1(parsed));
  		this._sectionStart = this._index;
  	} else {
  		this._sectionStart--;
  	}

  	this._state = this._baseState;
  };

  Tokenizer$1.prototype._stateInNumericEntity = function(c){
  	if(c === ";"){
  		this._decodeNumericEntity(2, 10);
  		this._sectionStart++;
  	} else if(c < "0" || c > "9"){
  		if(!this._xmlMode){
  			this._decodeNumericEntity(2, 10);
  		} else {
  			this._state = this._baseState;
  		}
  		this._index--;
  	}
  };

  Tokenizer$1.prototype._stateInHexEntity = function(c){
  	if(c === ";"){
  		this._decodeNumericEntity(3, 16);
  		this._sectionStart++;
  	} else if((c < "a" || c > "f") && (c < "A" || c > "F") && (c < "0" || c > "9")){
  		if(!this._xmlMode){
  			this._decodeNumericEntity(3, 16);
  		} else {
  			this._state = this._baseState;
  		}
  		this._index--;
  	}
  };

  Tokenizer$1.prototype._cleanup = function (){
  	if(this._sectionStart < 0){
  		this._buffer = "";
  		this._bufferOffset += this._index;
  		this._index = 0;
  	} else if(this._running){
  		if(this._state === TEXT){
  			if(this._sectionStart !== this._index){
  				this._cbs.ontext(this._buffer.substr(this._sectionStart));
  			}
  			this._buffer = "";
  			this._bufferOffset += this._index;
  			this._index = 0;
  		} else if(this._sectionStart === this._index){
  			
  			this._buffer = "";
  			this._bufferOffset += this._index;
  			this._index = 0;
  		} else {
  			
  			this._buffer = this._buffer.substr(this._sectionStart);
  			this._index -= this._sectionStart;
  			this._bufferOffset += this._sectionStart;
  		}

  		this._sectionStart = 0;
  	}
  };


  Tokenizer$1.prototype.write = function(chunk){
  	if(this._ended) this._cbs.onerror(Error(".write() after done!"));

  	this._buffer += chunk;
  	this._parse();
  };

  Tokenizer$1.prototype._parse = function(){
  	while(this._index < this._buffer.length && this._running){
  		var c = this._buffer.charAt(this._index);
  		if(this._state === TEXT) {
  			this._stateText(c);
  		} else if(this._state === BEFORE_TAG_NAME){
  			this._stateBeforeTagName(c);
  		} else if(this._state === IN_TAG_NAME) {
  			this._stateInTagName(c);
  		} else if(this._state === BEFORE_CLOSING_TAG_NAME){
  			this._stateBeforeCloseingTagName(c);
  		} else if(this._state === IN_CLOSING_TAG_NAME){
  			this._stateInCloseingTagName(c);
  		} else if(this._state === AFTER_CLOSING_TAG_NAME){
  			this._stateAfterCloseingTagName(c);
  		} else if(this._state === IN_SELF_CLOSING_TAG){
  			this._stateInSelfClosingTag(c);
  		}

  		
  		else if(this._state === BEFORE_ATTRIBUTE_NAME){
  			this._stateBeforeAttributeName(c);
  		} else if(this._state === IN_ATTRIBUTE_NAME){
  			this._stateInAttributeName(c);
  		} else if(this._state === AFTER_ATTRIBUTE_NAME){
  			this._stateAfterAttributeName(c);
  		} else if(this._state === BEFORE_ATTRIBUTE_VALUE){
  			this._stateBeforeAttributeValue(c);
  		} else if(this._state === IN_ATTRIBUTE_VALUE_DQ){
  			this._stateInAttributeValueDoubleQuotes(c);
  		} else if(this._state === IN_ATTRIBUTE_VALUE_SQ){
  			this._stateInAttributeValueSingleQuotes(c);
  		} else if(this._state === IN_ATTRIBUTE_VALUE_NQ){
  			this._stateInAttributeValueNoQuotes(c);
  		}

  		
  		else if(this._state === BEFORE_DECLARATION){
  			this._stateBeforeDeclaration(c);
  		} else if(this._state === IN_DECLARATION){
  			this._stateInDeclaration(c);
  		}

  		
  		else if(this._state === IN_PROCESSING_INSTRUCTION){
  			this._stateInProcessingInstruction(c);
  		}

  		
  		else if(this._state === BEFORE_COMMENT){
  			this._stateBeforeComment(c);
  		} else if(this._state === IN_COMMENT){
  			this._stateInComment(c);
  		} else if(this._state === AFTER_COMMENT_1){
  			this._stateAfterComment1(c);
  		} else if(this._state === AFTER_COMMENT_2){
  			this._stateAfterComment2(c);
  		}

  		
  		else if(this._state === BEFORE_CDATA_1){
  			this._stateBeforeCdata1(c);
  		} else if(this._state === BEFORE_CDATA_2){
  			this._stateBeforeCdata2(c);
  		} else if(this._state === BEFORE_CDATA_3){
  			this._stateBeforeCdata3(c);
  		} else if(this._state === BEFORE_CDATA_4){
  			this._stateBeforeCdata4(c);
  		} else if(this._state === BEFORE_CDATA_5){
  			this._stateBeforeCdata5(c);
  		} else if(this._state === BEFORE_CDATA_6){
  			this._stateBeforeCdata6(c);
  		} else if(this._state === IN_CDATA){
  			this._stateInCdata(c);
  		} else if(this._state === AFTER_CDATA_1){
  			this._stateAfterCdata1(c);
  		} else if(this._state === AFTER_CDATA_2){
  			this._stateAfterCdata2(c);
  		}

  		
  		else if(this._state === BEFORE_SPECIAL){
  			this._stateBeforeSpecial(c);
  		} else if(this._state === BEFORE_SPECIAL_END){
  			this._stateBeforeSpecialEnd(c);
  		}

  		
  		else if(this._state === BEFORE_SCRIPT_1){
  			this._stateBeforeScript1(c);
  		} else if(this._state === BEFORE_SCRIPT_2){
  			this._stateBeforeScript2(c);
  		} else if(this._state === BEFORE_SCRIPT_3){
  			this._stateBeforeScript3(c);
  		} else if(this._state === BEFORE_SCRIPT_4){
  			this._stateBeforeScript4(c);
  		} else if(this._state === BEFORE_SCRIPT_5){
  			this._stateBeforeScript5(c);
  		}

  		else if(this._state === AFTER_SCRIPT_1){
  			this._stateAfterScript1(c);
  		} else if(this._state === AFTER_SCRIPT_2){
  			this._stateAfterScript2(c);
  		} else if(this._state === AFTER_SCRIPT_3){
  			this._stateAfterScript3(c);
  		} else if(this._state === AFTER_SCRIPT_4){
  			this._stateAfterScript4(c);
  		} else if(this._state === AFTER_SCRIPT_5){
  			this._stateAfterScript5(c);
  		}

  		
  		else if(this._state === BEFORE_STYLE_1){
  			this._stateBeforeStyle1(c);
  		} else if(this._state === BEFORE_STYLE_2){
  			this._stateBeforeStyle2(c);
  		} else if(this._state === BEFORE_STYLE_3){
  			this._stateBeforeStyle3(c);
  		} else if(this._state === BEFORE_STYLE_4){
  			this._stateBeforeStyle4(c);
  		}

  		else if(this._state === AFTER_STYLE_1){
  			this._stateAfterStyle1(c);
  		} else if(this._state === AFTER_STYLE_2){
  			this._stateAfterStyle2(c);
  		} else if(this._state === AFTER_STYLE_3){
  			this._stateAfterStyle3(c);
  		} else if(this._state === AFTER_STYLE_4){
  			this._stateAfterStyle4(c);
  		}

  		
  		else if(this._state === BEFORE_ENTITY){
  			this._stateBeforeEntity(c);
  		} else if(this._state === BEFORE_NUMERIC_ENTITY){
  			this._stateBeforeNumericEntity(c);
  		} else if(this._state === IN_NAMED_ENTITY){
  			this._stateInNamedEntity(c);
  		} else if(this._state === IN_NUMERIC_ENTITY){
  			this._stateInNumericEntity(c);
  		} else if(this._state === IN_HEX_ENTITY){
  			this._stateInHexEntity(c);
  		}

  		else {
  			this._cbs.onerror(Error("unknown _state"), this._state);
  		}

  		this._index++;
  	}

  	this._cleanup();
  };

  Tokenizer$1.prototype.pause = function(){
  	this._running = false;
  };
  Tokenizer$1.prototype.resume = function(){
  	this._running = true;

  	if(this._index < this._buffer.length){
  		this._parse();
  	}
  	if(this._ended){
  		this._finish();
  	}
  };

  Tokenizer$1.prototype.end = function(chunk){
  	if(this._ended) this._cbs.onerror(Error(".end() after done!"));
  	if(chunk) this.write(chunk);

  	this._ended = true;

  	if(this._running) this._finish();
  };

  Tokenizer$1.prototype._finish = function(){
  	
  	if(this._sectionStart < this._index){
  		this._handleTrailingData();
  	}

  	this._cbs.onend();
  };

  Tokenizer$1.prototype._handleTrailingData = function(){
  	var data = this._buffer.substr(this._sectionStart);

  	if(this._state === IN_CDATA || this._state === AFTER_CDATA_1 || this._state === AFTER_CDATA_2){
  		this._cbs.oncdata(data);
  	} else if(this._state === IN_COMMENT || this._state === AFTER_COMMENT_1 || this._state === AFTER_COMMENT_2){
  		this._cbs.oncomment(data);
  	} else if(this._state === IN_NAMED_ENTITY && !this._xmlMode){
  		this._parseLegacyEntity();
  		if(this._sectionStart < this._index){
  			this._state = this._baseState;
  			this._handleTrailingData();
  		}
  	} else if(this._state === IN_NUMERIC_ENTITY && !this._xmlMode){
  		this._decodeNumericEntity(2, 10);
  		if(this._sectionStart < this._index){
  			this._state = this._baseState;
  			this._handleTrailingData();
  		}
  	} else if(this._state === IN_HEX_ENTITY && !this._xmlMode){
  		this._decodeNumericEntity(3, 16);
  		if(this._sectionStart < this._index){
  			this._state = this._baseState;
  			this._handleTrailingData();
  		}
  	} else if(
  		this._state !== IN_TAG_NAME &&
  		this._state !== BEFORE_ATTRIBUTE_NAME &&
  		this._state !== BEFORE_ATTRIBUTE_VALUE &&
  		this._state !== AFTER_ATTRIBUTE_NAME &&
  		this._state !== IN_ATTRIBUTE_NAME &&
  		this._state !== IN_ATTRIBUTE_VALUE_SQ &&
  		this._state !== IN_ATTRIBUTE_VALUE_DQ &&
  		this._state !== IN_ATTRIBUTE_VALUE_NQ &&
  		this._state !== IN_CLOSING_TAG_NAME
  	){
  		this._cbs.ontext(data);
  	}
  	
  	
  };

  Tokenizer$1.prototype.reset = function(){
  	Tokenizer$1.call(this, {xmlMode: this._xmlMode, decodeEntities: this._decodeEntities}, this._cbs);
  };

  Tokenizer$1.prototype.getAbsoluteIndex = function(){
  	return this._bufferOffset + this._index;
  };

  Tokenizer$1.prototype._getSection = function(){
  	return this._buffer.substring(this._sectionStart, this._index);
  };

  Tokenizer$1.prototype._emitToken = function(name){
  	this._cbs[name](this._getSection());
  	this._sectionStart = -1;
  };

  Tokenizer$1.prototype._emitPartial = function(value){
  	if(this._baseState !== TEXT){
  		this._cbs.onattribdata(value); 
  	} else {
  		this._cbs.ontext(value);
  	}
  };

  var _inherits = function() {};

  var _inherits$1 = Object.freeze({
  	default: _inherits
  });

  var _resolve_empty = {};

  var _resolve_empty$1 = Object.freeze({
  	default: _resolve_empty
  });

  var require$$1$1 = ( _inherits$1 && _inherits ) || _inherits$1;

  var require$$2 = ( _resolve_empty$1 && _resolve_empty ) || _resolve_empty$1;

  var Tokenizer$2 = Tokenizer_1;





  var formTags = {
  	input: true,
  	option: true,
  	optgroup: true,
  	select: true,
  	button: true,
  	datalist: true,
  	textarea: true
  };

  var openImpliesClose = {
  	tr      : { tr:true, th:true, td:true },
  	th      : { th:true },
  	td      : { thead:true, th:true, td:true },
  	body    : { head:true, link:true, script:true },
  	li      : { li:true },
  	p       : { p:true },
  	h1      : { p:true },
  	h2      : { p:true },
  	h3      : { p:true },
  	h4      : { p:true },
  	h5      : { p:true },
  	h6      : { p:true },
  	select  : formTags,
  	input   : formTags,
  	output  : formTags,
  	button  : formTags,
  	datalist: formTags,
  	textarea: formTags,
  	option  : { option:true },
  	optgroup: { optgroup:true }
  };

  var voidElements = {
  	__proto__: null,
  	area: true,
  	base: true,
  	basefont: true,
  	br: true,
  	col: true,
  	command: true,
  	embed: true,
  	frame: true,
  	hr: true,
  	img: true,
  	input: true,
  	isindex: true,
  	keygen: true,
  	link: true,
  	meta: true,
  	param: true,
  	source: true,
  	track: true,
  	wbr: true,

  	
  	path: true,
  	circle: true,
  	ellipse: true,
  	line: true,
  	rect: true,
  	use: true,
  	stop: true,
  	polyline: true,
  	polygon: true
  };

  var re_nameEnd = /\s|\//;

  function Parser$1(cbs, options){
  	this._options = options || {};
  	this._cbs = cbs || {};

  	this._tagname = "";
  	this._attribname = "";
  	this._attribvalue = "";
  	this._attribs = null;
  	this._stack = [];

  	this.startIndex = 0;
  	this.endIndex = null;

  	this._lowerCaseTagNames = "lowerCaseTags" in this._options ?
  									!!this._options.lowerCaseTags :
  									!this._options.xmlMode;
  	this._lowerCaseAttributeNames = "lowerCaseAttributeNames" in this._options ?
  									!!this._options.lowerCaseAttributeNames :
  									!this._options.xmlMode;

  	if(this._options.Tokenizer) {
  		Tokenizer$2 = this._options.Tokenizer;
  	}
  	this._tokenizer = new Tokenizer$2(this._options, this);

  	if(this._cbs.onparserinit) this._cbs.onparserinit(this);
  }

  require$$1$1(Parser$1, require$$2.EventEmitter);

  Parser$1.prototype._updatePosition = function(initialOffset){
  	if(this.endIndex === null){
  		if(this._tokenizer._sectionStart <= initialOffset){
  			this.startIndex = 0;
  		} else {
  			this.startIndex = this._tokenizer._sectionStart - initialOffset;
  		}
  	}
  	else this.startIndex = this.endIndex + 1;
  	this.endIndex = this._tokenizer.getAbsoluteIndex();
  };


  Parser$1.prototype.ontext = function(data){
  	this._updatePosition(1);
  	this.endIndex--;

  	if(this._cbs.ontext) this._cbs.ontext(data);
  };

  Parser$1.prototype.onopentagname = function(name){
  	if(this._lowerCaseTagNames){
  		name = name.toLowerCase();
  	}

  	this._tagname = name;

  	if(!this._options.xmlMode && name in openImpliesClose) {
  		for(
  			var el;
  			(el = this._stack[this._stack.length - 1]) in openImpliesClose[name];
  			this.onclosetag(el)
  		);
  	}

  	if(this._options.xmlMode || !(name in voidElements)){
  		this._stack.push(name);
  	}

  	if(this._cbs.onopentagname) this._cbs.onopentagname(name);
  	if(this._cbs.onopentag) this._attribs = {};
  };

  Parser$1.prototype.onopentagend = function(){
  	this._updatePosition(1);

  	if(this._attribs){
  		if(this._cbs.onopentag) this._cbs.onopentag(this._tagname, this._attribs);
  		this._attribs = null;
  	}

  	if(!this._options.xmlMode && this._cbs.onclosetag && this._tagname in voidElements){
  		this._cbs.onclosetag(this._tagname);
  	}

  	this._tagname = "";
  };

  Parser$1.prototype.onclosetag = function(name){
  	this._updatePosition(1);
  	if(this._lowerCaseTagNames){
  		name = name.toLowerCase();
  	}
  	
  	
  	
  	if(this._options.xmlMode) {
  		const stack = this._stack;
  		let last = stack.pop();
  		while(last !== name) {
  			if(this._cbs.onerror) {
  				this._cbs.onerror("Unclosed tag <"+last+">");
  			}
  			last = stack.pop();
  		}
  		this.onopentagend();
  		if(this._cbs.onclosetag) {
  			this._cbs.onclosetag(last);
  		}
  	} else {
  		if(this._stack.length && (!(name in voidElements))) {
  			let pos = this._stack.lastIndexOf(name);
  			if(pos !== -1){
  				if(this._cbs.onclosetag){
  					pos = this._stack.length - pos;
  					while(pos--) this._cbs.onclosetag(this._stack.pop());
  				}
  				else this._stack.length = pos;
  			} else if(name === "p"){
  				this.onopentagname(name);
  				this._closeCurrentTag();
  			}
  		} else if(name === "br" || name === "p"){
  			this.onopentagname(name);
  			this._closeCurrentTag();
  		}
  	}
  };

  Parser$1.prototype.onselfclosingtag = function(){
  	if(this._options.xmlMode || this._options.recognizeSelfClosing){
  		this._closeCurrentTag();
  	} else {
  		this.onopentagend();
  	}
  };

  Parser$1.prototype._closeCurrentTag = function(){
  	var name = this._tagname;

  	this.onopentagend();

  	
  	
  	if(this._stack[this._stack.length - 1] === name){
  		if(this._cbs.onclosetag){
  			this._cbs.onclosetag(name);
  		}
  		this._stack.pop();
  	}
  };

  Parser$1.prototype.onattribname = function(name){
  	if(this._lowerCaseAttributeNames){
  		name = name.toLowerCase();
  	}
  	this._attribname = name;
  };

  Parser$1.prototype.onattribdata = function(value){
  	this._attribvalue += value;
  };

  Parser$1.prototype.onattribend = function(){
  	if(this._cbs.onattribute) this._cbs.onattribute(this._attribname, this._attribvalue);
  	if(
  		this._attribs &&
  		!Object.prototype.hasOwnProperty.call(this._attribs, this._attribname)
  	){
  		this._attribs[this._attribname] = this._attribvalue;
  	}
  	this._attribname = "";
  	this._attribvalue = "";
  };

  Parser$1.prototype._getInstructionName = function(value){
  	var idx = value.search(re_nameEnd),
  	    name = idx < 0 ? value : value.substr(0, idx);

  	if(this._lowerCaseTagNames){
  		name = name.toLowerCase();
  	}

  	return name;
  };

  Parser$1.prototype.ondeclaration = function(value){
  	if(this._cbs.onprocessinginstruction){
  		var name = this._getInstructionName(value);
  		this._cbs.onprocessinginstruction("!" + name, "!" + value);
  	}
  };

  Parser$1.prototype.onprocessinginstruction = function(value){
  	if(this._cbs.onprocessinginstruction){
  		var name = this._getInstructionName(value);
  		this._cbs.onprocessinginstruction("?" + name, "?" + value);
  	}
  };

  Parser$1.prototype.oncomment = function(value){
  	this._updatePosition(4);

  	if(this._cbs.oncomment) this._cbs.oncomment(value);
  	if(this._cbs.oncommentend) this._cbs.oncommentend();
  };

  Parser$1.prototype.oncdata = function(value){
  	this._updatePosition(1);

  	if(this._options.xmlMode || this._options.recognizeCDATA){
  		if(this._cbs.oncdatastart) this._cbs.oncdatastart();
  		if(this._cbs.ontext) this._cbs.ontext(value);
  		if(this._cbs.oncdataend) this._cbs.oncdataend();
  	} else {
  		this.oncomment("[CDATA[" + value + "]]");
  	}
  };

  Parser$1.prototype.onerror = function(err){
  	if(this._cbs.onerror) this._cbs.onerror(err);
  };

  Parser$1.prototype.onend = function(){
  	if(this._cbs.onclosetag){
  		for(
  			var i = this._stack.length;
  			i > 0;
  			this._cbs.onclosetag(this._stack[--i])
  		);
  	}
  	if(this._cbs.onend) this._cbs.onend();
  };



  Parser$1.prototype.reset = function(){
  	if(this._cbs.onreset) this._cbs.onreset();
  	this._tokenizer.reset();

  	this._tagname = "";
  	this._attribname = "";
  	this._attribs = null;
  	this._stack = [];

  	if(this._cbs.onparserinit) this._cbs.onparserinit(this);
  };


  Parser$1.prototype.parseComplete = function(data){
  	this.reset();
  	this.end(data);
  };

  Parser$1.prototype.write = function(chunk){
  	this._tokenizer.write(chunk);
  };

  Parser$1.prototype.end = function(chunk){
  	this._tokenizer.end(chunk);
  };

  Parser$1.prototype.pause = function(){
  	this._tokenizer.pause();
  };

  Parser$1.prototype.resume = function(){
  	this._tokenizer.resume();
  };


  Parser$1.prototype.parseChunk = Parser$1.prototype.write;
  Parser$1.prototype.done = Parser$1.prototype.end;

  var Parser_1 = Parser$1;

  Parser_1.prototype.ondeclaration = function(value){
    if(this._cbs.ondeclaration){
      this._cbs.ondeclaration(value);
    } else if(this._cbs.onprocessinginstruction){
      var name = this._getInstructionName(value);
      this._cbs.onprocessinginstruction("!" + name, "!" + value);
    }
  };


  Parser_1.prototype.oncdata = function(value){
    this._updatePosition(1);

    if(this._options.xmlMode || this._options.recognizeCDATA){
      if(this._cbs.oncdatastart) this._cbs.oncdatastart(value);
      
      if(this._cbs.oncdataend) this._cbs.oncdataend();
    } else {
      this.oncomment("[CDATA[" + value + "]]");
    }
  };

  function parseMarkup (markup, options) {
    let format = options.ownerDocument ? options.ownerDocument.format : options.format;
    
    if (!format) {
      throw new Error("Either 'ownerDocument' or 'format' must be set.")
    }
    let parserOptions = Object.assign({}, options, {
      xmlMode: (format === 'xml')
    });
    let handler = new DomHandler({ format, elementFactory: options.elementFactory });
    let parser = new Parser_1(handler, parserOptions);
    parser.end(markup);
    return handler.document
  }

  const RE_WHITESPACE = /\s+/g;
  const RE_DOCTYPE = /^DOCTYPE\s+([^\s]+)(?:\s+PUBLIC\s+["]([^"]+)["](?:\s+["]([^"]+)["])?)\s*$/;


  class DomHandler {
    constructor (options = {}) {
      this.elementFactory = options.elementFactory;
      if (!this.elementFactory) throw new Error("'elementFactory' is required")
      this.options = options;
      this.document = null;
      this._tagStack = [];
    }

    
    onparserinit () {
      this.document = this.elementFactory('document', { format: this.options.format });
      this._tagStack = [this.document];
    }

    onend () {
      
      if (this._tagStack.length > 1) {
        throw new Error(`Unexpected EOF. Tag was opened but not closed.`)
      }
    }

    onerror (error) {
      throw new Error(error)
    }

    onclosetag () {
      this._tagStack.pop();
    }

    _addDomElement (element) {
      let parent = this._tagStack[this._tagStack.length - 1];
      if (!parent.childNodes) parent.childNodes = [];
      let siblings = parent.childNodes;

      let previousSibling = siblings[siblings.length - 1];
      
      element.next = null;
      if (previousSibling) {
        element.prev = previousSibling;
        previousSibling.next = element;
      } else {
        element.prev = null;
      }
      
      siblings.push(element);
      element.parent = parent || null;
    }

    onopentag (name, attributes) {
      let element = this.document.createElement(name);
      forEach(attributes, (val, key) => {
        element.setAttribute(key, val);
      });
      this._addDomElement(element);
      this._tagStack.push(element);
    }

    ontext (text) {
      if (this.options.normalizeWhitespace) {
        text = text.replace(RE_WHITESPACE, ' ');
      }
      let lastTag;
      let _top = this._tagStack[this._tagStack.length - 1];
      if (_top && _top.childNodes) lastTag = _top.childNodes[_top.childNodes.length - 1];
      if (lastTag && lastTag.type === domelementtype.Text) {
        lastTag.data += text;
      } else {
        let element = this.document.createTextNode(text);
        this._addDomElement(element);
      }
    }

    oncomment (data) {
      var lastTag = this._tagStack[this._tagStack.length - 1];
      if (lastTag && lastTag.type === domelementtype.Comment) {
        lastTag.data += data;
      } else {
        let element = this.document.createComment(data);
        this._addDomElement(element);
        this._tagStack.push(element);
      }
    }

    oncommentend () {
      this._tagStack.pop();
    }

    oncdatastart (data) {
      let element = this.document.createCDATASection(data);
      this._addDomElement(element);
      this._tagStack.push(element);
    }

    oncdataend () {
      this._tagStack.pop();
    }

    onprocessinginstruction (name, data) {
      let element = this.document.createProcessingInstruction(name, data);
      this._addDomElement(element);
    }

    ondeclaration (data) {
      if (data.startsWith('DOCTYPE')) {
        let m = RE_DOCTYPE.exec(data);
        if (!m) throw new Error('Could not parse DOCTYPE element: ' + data)
        this.document.setDoctype(m[1], m[2], m[3]);
      } else {
        throw new Error('Not implemented: parse declaration ' + data)
      }
    }
  }

  const MemoryDOMElementFactory = (type, data) => {
    return new MemoryDOMElement(type, data)
  };

  class MemoryDOMElement extends DOMElement {
    constructor (type, args = {}) {
      super();

      this.type = type;
      if (!type) throw new Error("'type' is mandatory")

      this.ownerDocument = args.ownerDocument;
      
      if (type !== 'document' && !this.ownerDocument) {
        throw new Error("'ownerDocument' is mandatory")
      }

      
      
      

      switch (type) {
        case domelementtype.Tag: {
          if (!args.name) throw new Error("'name' is mandatory.")
          this.name = this._normalizeName(args.name);
          this.nameWithoutNS = nameWithoutNS(this.name);
          this.properties = new Map();
          this.attributes = new Map();
          this.classes = new Set();
          this.styles = new Map();
          this.eventListeners = [];
          this.childNodes = args.children || args.childNodes || [];
          this._assign(args);
          break
        }
        case domelementtype.Text:
        case domelementtype.Comment: {
          this.data = args.data || '';
          break
        }
        case domelementtype.CDATA: {
          this.data = args.data || '';
          break
        }
        case domelementtype.Directive: {
          if (!args.name) throw new Error("'name' is mandatory.")
          this.name = this._normalizeName(args.name);
          this.nameWithoutNS = nameWithoutNS(this.name);
          this.data = args.data;
          break
        }
        case domelementtype.Doctype: {
          this.data = args.data;
          break
        }
        case 'document': {
          let format = args.format;
          this.format = format;
          if (!format) throw new Error("'format' is mandatory.")
          this.childNodes = args.children || args.childNodes || [];
          switch (format) {
            case 'xml':
              this.contentType = 'application/xml';
              break
            case 'html':
              this.contentType = 'text/html';
              break
            default:
              throw new Error('Unsupported format ' + format)
          }
          break
        }
        default:
          this.name = null;
          this.properties = new Map();
          this.attributes = new Map();
          this.classes = new Set();
          this.styles = new Map();
          this.eventListeners = [];
          this.childNodes = args.children || args.childNodes || [];
      }
    }

    getNativeElement () {
      return this
    }

    getNodeType () {
      switch (this.type) {
        case domelementtype.Tag:
        case domelementtype.Script:
        case domelementtype.Style:
          return 'element'
        default:
          return this.type
      }
    }

    isTextNode () {
      return this.type === 'text'
    }

    isElementNode () {
      return this.type === 'tag' || this.type === 'script'
    }

    isCommentNode () {
      return this.type === 'comment'
    }

    isDocumentNode () {
      return this.type === 'document'
    }

    isComponentNode () {
      return this.type === 'component'
    }

    clone (deep) {
      let clone = new MemoryDOMElement(this.type, this);
      if (this.childNodes) {
        clone.childNodes.length = 0;
        if (deep) {
          this.childNodes.forEach((child) => {
            clone.appendChild(child.clone(deep));
          });
        }
      }
      return clone
    }

    get tagName () {
      return this.getTagName()
    }

    set tagName (tagName) {
      this.setTagName(tagName);
    }

    getTagName () {
      return this.name
    }

    setTagName (tagName) {
      if (this._isXML()) {
        this.name = String(tagName);
      } else {
        this.name = String(tagName).toLowerCase();
      }
      this.nameWithoutNS = nameWithoutNS(this.name);
      return this
    }

    hasAttribute (name) {
      return this.attributes.has(name)
    }

    getAttribute (name) {
      return this.attributes.get(name)
    }

    setAttribute (name, value) {
      value = String(value);
      
      switch (name) {
        case 'class':
          this.classes = new Set();
          parseClasses(this.classes, value);
          break
        case 'style':
          this.styles = new Map();
          parseStyles(this.styles, value);
          break
        default:
          
      }
      this.attributes.set(name, value);
      if (this._isHTML()) {
        deriveHTMLPropertyFromAttribute(this, name, value);
      }
      return this
    }

    removeAttribute (name) {
      switch (name) {
        case 'class':
          this.classes = new Set();
          break
        case 'style':
          this.styles = new Map();
          break
        default:
          
      }
      this.attributes.delete(name);
      return this
    }

    getAttributes () {
      return this.attributes
    }

    getProperty (name) {
      if (this.properties) {
        return this.properties.get(name)
      }
    }

    setProperty (name, value) {
      if (this.properties) {
        if (this._isXML()) {
          throw new Error('setProperty() is only be used on HTML elements')
        }
        _setHTMLPropertyValue(this, name, value);
      }
      return this
    }

    hasClass (name) {
      if (this.classes) {
        return this.classes.has(name)
      }
    }

    addClass (name) {
      this.classes.add(name);
      this.attributes.set('class', stringifyClasses(this.classes));
      return this
    }

    removeClass (name) {
      if (this.classes && this.classes.has(name)) {
        this.classes.delete(name);
        this.attributes.set('class', stringifyClasses(this.classes));
      }
      return this
    }

    getContentType () {
      return this.getOwnerDocument().contentType
    }

    getDoctype () {
      if (this.isDocumentNode()) {
        return _findDocTypeElement(this)
      } else {
        return this.getOwnerDocument().getDoctype()
      }
    }

    setDoctype (qualifiedNameStr, publicId, systemId) {
      
      let doc = this.getOwnerDocument();
      let oldDocType = _findDocTypeElement(doc);
      let newDocType = this.createDocumentType(qualifiedNameStr, publicId, systemId);
      if (oldDocType) {
        doc.replaceChild(newDocType, oldDocType);
      } else {
        
        doc.insertBefore(newDocType, doc.getChildren()[0]);
      }
      doc.doctype = newDocType;
    }

    getInnerHTML () {
      return domUtils.getInnerHTML(this, { decodeEntities: true })
    }

    
    
    setInnerHTML (html) {
      if (this.childNodes) {
        let _doc = parseMarkup(html, {
          ownerDocument: this.getOwnerDocument(),
          decodeEntities: true,
          elementFactory: MemoryDOMElementFactory
        });
        this.empty();
        
        
        _doc.childNodes.slice(0).forEach((child) => {
          this.appendChild(child);
        });
      }
      return this
    }

    getOuterHTML () {
      return domUtils.getOuterHTML(this, { xmlMode: this._isXML(), decodeEntities: true })
    }

    getTextContent () {
      return domUtils.getText(this)
    }

    setTextContent (text) {
      switch (this.type) {
        case domelementtype.Text:
        case domelementtype.Comment:
        case domelementtype.CDATA: {
          this.data = text;
          break
        }
        default: {
          if (this.childNodes) {
            let child = this.createTextNode(text);
            this.empty();
            this.appendChild(child);
          }
        }
      }
      return this
    }

    getStyle (name) {
      if (this.styles) {
        return this.styles.get(name)
      }
    }

    setStyle (name, value) {
      if (this.styles) {
        if (DOMElement.pxStyles[name] && isNumber(value)) {
          value = value + 'px';
        }
        this.styles.set(name, value);
        this.attributes.set('style', stringifyStyles(this.styles));
      }
      return this
    }

    is (cssSelector) {
      return cssSelect.is(this, cssSelector, { xmlMode: this._isXML() })
    }

    find (cssSelector) {
      return cssSelect.selectOne(cssSelector, this, { xmlMode: this._isXML() })
    }

    findAll (cssSelector) {
      return cssSelect.selectAll(cssSelector, this, { xmlMode: this._isXML() })
    }

    getChildCount () {
      if (this.childNodes) {
        return this.childNodes.length
      } else {
        return 0
      }
    }

    getChildNodes () {
      return this.childNodes.slice(0)
    }

    getChildren () {
      return this.childNodes.filter(function (node) {
        return node.type === 'tag'
      })
    }

    get children () {
      return this.getChildren()
    }

    getChildAt (pos) {
      if (this.childNodes) {
        return this.childNodes[pos]
      }
    }

    getChildIndex (child) {
      if (this.childNodes) {
        return this.childNodes.indexOf(child)
      }
    }

    getLastChild () {
      if (this.childNodes) {
        return last$1(this.childNodes)
      }
    }

    getFirstChild () {
      if (this.childNodes) {
        return this.childNodes[0]
      }
    }

    getNextSibling () {
      return this.next
    }

    getPreviousSibling () {
      return this.prev
    }

    getParent () {
      return this.parent
    }

    getOwnerDocument () {
      return (this.type === 'document') ? this : this.ownerDocument
    }

    getFormat () {
      return this.getOwnerDocument().format
    }

    createDocument (format) {
      return MemoryDOMElement.createDocument(format)
    }

    createElement (tagName) {
      return new MemoryDOMElement(domelementtype.Tag, { name: tagName, ownerDocument: this.getOwnerDocument() })
    }

    createTextNode (text) {
      return new MemoryDOMElement(domelementtype.Text, { data: text, ownerDocument: this.getOwnerDocument() })
    }

    createComment (data) {
      return new MemoryDOMElement(domelementtype.Comment, { data: data, ownerDocument: this.getOwnerDocument() })
    }

    createProcessingInstruction (name, data) {
      return new MemoryDOMElement(domelementtype.Directive, { name: name, data: data, ownerDocument: this.getOwnerDocument() })
    }

    createDocumentType (qualifiedNameStr, publicId, systemId) {
      return new MemoryDOMDoctype(domelementtype.Doctype, { data: { name: qualifiedNameStr, publicId, systemId }, ownerDocument: this.getOwnerDocument() })
    }

    createCDATASection (data) {
      return new MemoryDOMElement(domelementtype.CDATA, { data: data, ownerDocument: this.getOwnerDocument() })
    }

    appendChild (child) {
      if (this.childNodes && !isNil(child)) {
        child = this._normalizeChild(child);
        if (!child) return this
        domUtils.appendChild(this, child);
        child.ownerDocument = this.getOwnerDocument();
      }
      return this
    }

    removeChild (child) {
      if (child.parentNode === this) {
        child.remove();
      }
    }

    insertAt (pos, child) {
      child = this._normalizeChild(child);
      if (!child) return this
      let childNodes = this.childNodes;
      if (childNodes) {
        
        if (pos >= childNodes.length) {
          domUtils.appendChild(this, child);
        } else {
          domUtils.prepend(childNodes[pos], child);
        }
        child.ownerDocument = this.getOwnerDocument();
      }
      return this
    }

    insertBefore (newChild, before) {
      if (isNil(before)) {
        return this.appendChild(newChild)
      } else if (this.childNodes) {
        var pos = this.childNodes.indexOf(before);
        if (pos > -1) {
          domUtils.prepend(before, newChild);
          newChild.ownerDocument = this.getOwnerDocument();
        } else {
          throw new Error('insertBefore(): reference node is not a child of this element.')
        }
      }
      return this
    }

    removeAt (pos) {
      let childNodes = this.childNodes;
      if (childNodes) {
        let child = childNodes[pos];
        child.remove();
      }
      return this
    }

    empty () {
      let childNodes = this.childNodes;
      if (childNodes) {
        childNodes.forEach((child) => {
          child.next = child.prev = child.parent = null;
        });
        childNodes.length = 0;
      }
      return this
    }

    remove () {
      domUtils.removeElement(this);
      return this
    }

    replaceChild (oldChild, newChild) {
      if (oldChild.parent === this) {
        oldChild.replaceWith(newChild);
      }
      return this
    }

    replaceWith (newEl) {
      newEl = this._normalizeChild(newEl);
      domUtils.replaceElement(this, newEl);
      newEl.ownerDocument = this.getOwnerDocument();
      return this
    }

    getEventListeners () {
      return this.eventListeners || []
    }

    click () {
      this.emit('click', { target: this });
      return this
    }

    emit (name, data) {
      this._propagateEvent(new MemoryDOMElementEvent(name, this, data));
    }

    getBoundingClientRect () {
      return { top: 0, left: 0, height: 0, width: 0 }
    }

    getClientRects () {
      return [{ top: 0, left: 0, height: 0, width: 0 }]
    }

    _propagateEvent (event) {
      let listeners = this.eventListeners;
      if (listeners) {
        let listener = listeners.find((l) => {
          return l.eventName === event.type
        });
        if (listener) listener.handler(event);
        if (event.stopped) return
        let p = this.parentNode;
        if (p) p._propagateEvent(event);
      }
    }

    removeAllEventListeners () {
      this.eventListeners = [];
      return this
    }

    _assign (other) {
      if (other.name) this.name = other.name;
      if (this.classes && other.classes) {
        other.classes.forEach((val) => {
          this.classes.add(val);
        });
      }
      if (this.styles && other.styles) {
        forEach(other.styles, (val, name) => {
          this.styles.set(name, val);
        });
      }
      
      
      let otherAttributes = other.attributes || other.attribs;
      if (this.attributes && otherAttributes) {
        forEach(otherAttributes, (val, name) => {
          switch (name) {
            case 'class': {
              parseClasses(this.classes, val);
              break
            }
            case 'style': {
              parseStyles(this.styles, val);
              break
            }
            default:
              
          }
          this.attributes.set(name, val);
        });
      }
      if (this.eventListeners && other.eventListeners) {
        this.eventListeners = this.eventListeners.concat(other.eventListeners);
      }
    }

    _normalizeChild (child) {
      if (isNil(child)) return

      if (isString(child)) {
        child = this.createTextNode(child);
      }
      
      if (!child || !child._isMemoryDOMElement) {
        throw new Error('Illegal argument: only String and MemoryDOMElement instances are valid.')
      }
      return child
    }

    _normalizeName (name) {
      if (this._isXML()) {
        return name
      } else {
        return name.toLowerCase()
      }
    }

    _isHTML () {
      return this.getFormat() === 'html'
    }

    _isXML () {
      return this.getFormat() === 'xml'
    }
  }

  MemoryDOMElement.prototype._isMemoryDOMElement = true;

  MemoryDOMElement.createDocument = function (format) {
    if (format === 'xml') {
      return new MemoryDOMElement('document', { format: format })
    } else {
      return MemoryDOMElement.parseMarkup(DOMElement.EMPTY_HTML, 'html')
    }
  };

  MemoryDOMElement.parseMarkup = function (str, format, options = {}) {
    if (!str) {
      return MemoryDOMElement.createDocument(format)
    }
    let parserOpts = Object.assign({
      format,
      decodeEntities: true,
      elementFactory: MemoryDOMElementFactory
    }, options);
    
    if (options.raw) {
      return parseMarkup(str, parserOpts)
    }
    if (options.snippet) {
      str = `<__snippet__>${str}</__snippet__>`;
    }
    let doc;
    if (format === 'html') {
      doc = parseMarkup(str, parserOpts);
      _sanitizeHTMLStructure(doc);
    } else if (format === 'xml') {
      doc = parseMarkup(str, parserOpts);
    }
    if (options.snippet) {
      let childNodes = doc.find('__snippet__').childNodes;
      if (childNodes.length === 1) {
        return childNodes[0]
      } else {
        return childNodes
      }
    } else {
      return doc
    }
  };

  MemoryDOMElement.wrap =
  MemoryDOMElement.wrapNativeElement = function (el) {
    if (inBrowser) {
      
      
      
      
      if (el === window || el === window.document) {
        return new DOMElementStub()
      
      
      
      } else if (el instanceof window.Node || el._isBrowserDOMElement) ;
    }
    
    if (!el._isMemoryDOMElement) {
      throw new Error('Illegal argument: expected MemoryDOMElement instance')
    }
    return el
  };

  MemoryDOMElement.unwrap = function (el) {
    
    if (!el._isMemoryDOMElement) {
      throw new Error('Illegal argument: expected MemoryDOMElement instance')
    }
    return el
  };




  MemoryDOMElement.isReverse = function () {
    return false
  };


  let _browserWindowStub;
  MemoryDOMElement.getBrowserWindow = function () {
    
    if (!_browserWindowStub) {
      _browserWindowStub = MemoryDOMElement.createDocument('html');
    }
    return _browserWindowStub
  };

  class MemoryDOMDoctype extends MemoryDOMElement {
    get name () { return this.data.name }
    get publicId () { return this.data.publicId }
    get systemId () { return this.data.systemId }
  }

  function parseClasses (classes, classStr) {
    classStr.split(/\s+/).forEach((name) => {
      classes.add(name);
    });
  }

  function stringifyClasses (classes) {
    return Array.from(classes).join(' ')
  }

  function parseStyles (styles, styleStr) {
    styleStr = (styleStr || '').trim();
    if (!styleStr) return
    styleStr.split(';').forEach((style) => {
      let n = style.indexOf(':');
      
      if (n < 1 || n === style.length - 1) return
      let name = style.slice(0, n).trim();
      let val = style.slice(n + 1).trim();
      styles.set(name, val);
    });
  }

  function stringifyStyles (styles) {
    if (!styles) return ''
    let str = Object.keys(styles).map((name) => {
      return name + ':' + styles[name]
    }).join(';');
    if (str.length > 0) str += ';';
    return str
  }

  const BUILTIN_EVENTS = [
    'keydown', 'keyup', 'keypress',
    'mousedown', 'mouseup', 'mouseover', 'click', 'dblclick'
  ].reduce((m, k) => { m[k] = true; return m }, {});

  class MemoryDOMElementEvent {
    constructor (type, target, detail) {
      this.type = type;
      this.timeStamp = Date.now();
      this.target = target;

      if (BUILTIN_EVENTS[type]) {
        
        if (detail) {
          Object.assign(this, detail);
        }
      } else {
        this.detail = detail;
      }
    }

    stopPropagation () {
      this.stopped = true;
    }

    preventDefault () {
      this.defaultPrevented = true;
    }
  }

  class DOMElementStub {
    on () {}
    off () {}
  }

  function nameWithoutNS (name) {
    const idx = name.indexOf(':');
    if (idx > 0) {
      return name.slice(idx + 1)
    } else {
      return name
    }
  }



  const ATTR_TO_PROPS = {
    'input': {
      'value': true,
      'checked': (el, name, value) => {
        const checked = (value !== 'off');
        el.setProperty('checked', checked);
      }
    }
  };

  function deriveHTMLPropertyFromAttribute (el, name, value) {
    const mappings = ATTR_TO_PROPS[el.tagName];
    if (mappings) {
      let mapper = mappings[name];
      if (mapper === true) {
        el.setProperty(name, value);
      } else if (mapper) {
        mapper(el, name, value);
      }
    }
  }

  const PROPERTY_TRANSFORMATIONS = {
    'input': {
      'checked': (el, name, value) => {
        if (value === true) {
          el.properties.set(name, true);
          el.properties.set('value', 'on');
        } else {
          el.properties.set(name, false);
          el.properties.set('value', 'off');
        }
      },
      'value': (el, name, value) => {
        let type = el.getAttribute('type');
        switch (type) {
          case 'checkbox':
            if (value === 'on') {
              el.properties.set(name, true);
              el.properties.set('value', 'on');
            } else {
              el.properties.set(name, false);
              el.properties.set('value', 'off');
            }
            break
          default:
            _setProperty(el, name, value);
        }
      }
    }
  };

  function _setProperty (el, name, value) {
    if (value === undefined) {
      el.properties.delete(name);
    } else {
      el.properties.set(name, String(value));
    }
  }

  function _setHTMLPropertyValue (el, name, value) {
    const trafos = PROPERTY_TRANSFORMATIONS[el.tagName];
    if (trafos) {
      let mapper = trafos[name];
      if (mapper) {
        mapper(el, name, value);
        return
      }
    }
    _setProperty(el, name, value);
  }

  function _sanitizeHTMLStructure (doc) {
    
    
    
    
    let htmlEl = doc.find('html');
    if (!htmlEl) {
      
      
      let headEl = doc.find('head');
      let titleEl = doc.find('title');
      let metaEls = doc.findAll('meta');
      let bodyEl = doc.find('body');
      if (headEl) headEl.remove();
      if (titleEl) titleEl.remove();
      metaEls.forEach(e => e.remove());
      if (bodyEl) bodyEl.remove();

      
      
      let contentNodes = doc.childNodes.slice();
      contentNodes.forEach((c) => { c.parent = null; });
      doc.childNodes.length = 0;

      htmlEl = doc.createElement('html');
      
      
      
      if (!headEl) {
        headEl = doc.createElement('head');
        headEl.appendChild(titleEl);
        headEl.append(metaEls);
        htmlEl.appendChild(headEl);
      }
      if (!bodyEl) {
        bodyEl = doc.createElement('body');
        bodyEl.append(contentNodes);
      }
      htmlEl.appendChild(bodyEl);

      doc.append(htmlEl);
    }
  }

  function _findDocTypeElement (doc) {
    
    if (doc.doctype) return doc.doctype
    const childNodes = doc.childNodes;
    for (let i = 0; i < childNodes.length; i++) {
      let child = childNodes[i];
      if (child.type === domelementtype.Doctype) {
        doc.doctype = child;
        return child
      }
    }
  }

  let DefaultDOMElement = {};

  DefaultDOMElement.createDocument = function (format) {
    return _getDefaultImpl().createDocument(format)
  };


  DefaultDOMElement.createElement = function (tagName) {
    console.error('DEPRECATED: every element should have an ownerDocument. Use DefaultDOMElement.createDocument() to create a document first');
    let doc = DefaultDOMElement.createDocument('html');
    return doc.createElement(tagName)
  };


  DefaultDOMElement.createTextNode = function (text) {
    console.error('DEPRECATED: every element should have a ownerDocument. Use DefaultDOMElement.createDocument() to create a document first');
    let doc = DefaultDOMElement.createDocument('html');
    return doc.createTextNode(text)
  };


  DefaultDOMElement.getBrowserWindow = function () {
    return _getDefaultImpl().getBrowserWindow()
  };


  DefaultDOMElement.parseHTML = function (html, options) {
    return _getDefaultImpl().parseMarkup(html, 'html', options)
  };


  DefaultDOMElement.parseXML = function (xml, options) {
    return _getDefaultImpl().parseMarkup(xml, 'xml', options)
  };

  DefaultDOMElement.parseSnippet = function (str, format) {
    return _getDefaultImpl().parseMarkup(str, format, {snippet: true})
  };

  DefaultDOMElement.wrap =
  DefaultDOMElement.wrapNativeElement = function (nativeEl) {
    if (!nativeEl) throw new Error('Illegal argument')
    return _getDefaultImpl().wrap(nativeEl)
  };

  DefaultDOMElement.unwrap = function (nativeEl) {
    if (!nativeEl) throw new Error('Illegal argument')
    return _getDefaultImpl().unwrap(nativeEl)
  };



  DefaultDOMElement.isReverse = function (anchorNode, anchorOffset, focusNode, focusOffset) {
    return _getDefaultImpl().isReverse(anchorNode, anchorOffset, focusNode, focusOffset)
  };

  function _getDefaultImpl () {
    
    if (platform.inBrowser || platform.inElectron) {
      return BrowserDOMElement
    } else {
      return MemoryDOMElement
    }
  }

  const _global = (typeof global !== 'undefined') ? global : window;
  const substanceGlobals = _global.hasOwnProperty('Substance') ? _global.Substance : _global.Substance = {
    DEBUG_RENDERING: false
  };

  class VirtualElement extends DOMElement {
    constructor (owner) {
      super();

      
      this.parent = null;
      
      this._owner = owner;
      
      this._ref = null;
    }

    getParent () {
      return this.parent
    }

    get childNodes () {
      return this.getChildNodes()
    }

    getChildCount () {
      return this.children.length
    }

    getChildAt (idx) {
      return this.children[idx]
    }

    
    getComponent () {
      return this._comp
    }

    
    ref (ref) {
      if (!ref) throw new Error('Illegal argument')
      
      if (this._ref) throw new Error('A VirtualElement can only be referenced once.')
      this._ref = ref;
      if (this._context) {
        const refs = this._context.refs;
        if (refs[ref]) {
          throw new Error('An item with reference "' + ref + '" already exists.')
        }
        refs[ref] = this;
      }
      return this
    }

    isInDocument () {
      return false
    }
  }

  VirtualElement.prototype._isVirtualElement = true;


  class VirtualHTMLElement extends VirtualElement {
    constructor (tagName) {
      super();

      this._tagName = tagName;
      this.classNames = null;
      this.attributes = null;
      this.htmlProps = null;
      this.style = null;
      this.eventListeners = null;

      
      this.children = [];
    }

    getTagName () {
      return this._tagName
    }

    setTagName (tagName) {
      this._tagName = tagName;
      return this
    }

    hasClass (className) {
      if (this.classNames) {
        return this.classNames.indexOf(className) > -1
      }
      return false
    }

    addClass (className) {
      if (!this.classNames) {
        this.classNames = [];
      }
      this.classNames.push(className);
      return this
    }

    removeClass (className) {
      if (this.classNames) {
        this.classNames = without(this.classNames, className);
      }
      return this
    }

    removeAttribute (name) {
      if (this.attributes) {
        delete this.attributes[name];
      }
      return this
    }

    getAttribute (name) {
      if (this.attributes) {
        return this.attributes[name]
      }
    }

    setAttribute (name, value) {
      if (!this.attributes) {
        this.attributes = {};
      }
      this.attributes[name] = value;
      return this
    }

    getAttributes () {
      
      
      
      var attributes = {};
      if (this.attributes) {
        Object.assign(attributes, this.attributes);
      }
      if (this.classNames) {
        attributes.class = this.classNames.join(' ');
      }
      if (this.style) {
        attributes.style = map(this.style, function (val, key) {
          return key + ':' + val
        }).join(';');
      }
      return attributes
    }

    getId () {
      return this.getAttribute('id')
    }

    setId (id) {
      this.setAttribute('id', id);
      return this
    }

    setTextContent (text) {
      text = String(text || '');
      this.empty();
      this.appendChild(text);
      return this
    }

    setInnerHTML (html) {
      html = html || '';
      this.empty();
      this._innerHTMLString = html;
      return this
    }

    getInnerHTML () {
      if (!this.hasOwnProperty('_innerHTMLString')) {
        throw new Error('Not supported.')
      } else {
        return this._innerHTMLString
      }
    }

    getValue () {
      return this.htmlProp('value')
    }

    setValue (value) {
      this.htmlProp('value', value);
      return this
    }

    getChildNodes () {
      return this.children
    }

    getChildren () {
      return this.children.filter(function (child) {
        return child.getNodeType() !== 'text'
      })
    }

    isTextNode () {
      return false
    }

    isElementNode () {
      return true
    }

    isCommentNode () {
      return false
    }

    isDocumentNode () {
      return false
    }

    append () {
      if (this._innerHTMLString) {
        throw Error('It is not possible to mix $$.html() with $$.append(). You can call $$.empty() to reset this virtual element.')
      }
      this._append(this.children, arguments);
      return this
    }

    appendChild (child) {
      if (this._innerHTMLString) {
        throw Error('It is not possible to mix $$.html() with $$.append(). You can call $$.empty() to reset this virtual element.')
      }
      this._appendChild(this.children, child);
      return this
    }

    insertAt (pos, child) {
      child = this._normalizeChild(child);
      if (!child) {
        throw new Error('Illegal child: ' + child)
      }
      if (!child._isVirtualElement) {
        throw new Error('Illegal argument for $$.insertAt():' + child)
      }
      if (pos < 0 || pos > this.children.length) {
        throw new Error('insertAt(): index out of bounds.')
      }
      this._insertAt(this.children, pos, child);
      return this
    }

    insertBefore (child, before) {
      var pos = this.children.indexOf(before);
      if (pos > -1) {
        this.insertAt(pos, child);
      } else {
        throw new Error('insertBefore(): reference node is not a child of this element.')
      }
      return this
    }

    removeAt (pos) {
      if (pos < 0 || pos >= this.children.length) {
        throw new Error('removeAt(): Index out of bounds.')
      }
      this._removeAt(pos);
      return this
    }

    removeChild (child) {
      if (!child || !child._isVirtualElement) {
        throw new Error('removeChild(): Illegal arguments. Expecting a CheerioDOMElement instance.')
      }
      var idx = this.children.indexOf(child);
      if (idx < 0) {
        throw new Error('removeChild(): element is not a child.')
      }
      this.removeAt(idx);
      return this
    }

    replaceChild (oldChild, newChild) {
      if (!newChild || !oldChild ||
          !newChild._isVirtualElement || !oldChild._isVirtualElement) {
        throw new Error('replaceChild(): Illegal arguments. Expecting BrowserDOMElement instances.')
      }
      var idx = this.children.indexOf(oldChild);
      if (idx < 0) {
        throw new Error('replaceChild(): element is not a child.')
      }
      this.removeAt(idx);
      this.insertAt(idx, newChild);
      return this
    }

    empty () {
      var children = this.children;
      while (children.length) {
        var child = children.pop();
        child.parent = null;
      }
      delete this._innerHTMLString;
      return this
    }

    getProperty (name) {
      if (this.htmlProps) {
        return this.htmlProps[name]
      }
    }

    setProperty (name, value) {
      if (!this.htmlProps) {
        this.htmlProps = {};
      }
      this.htmlProps[name] = value;
      return this
    }

    removeProperty (name) {
      if (this.htmlProps) {
        delete this.htmlProps[name];
      }
      return this
    }

    getStyle (name) {
      if (this.style) {
        return this.style[name]
      }
    }

    setStyle (name, value) {
      if (!this.style) {
        this.style = {};
      }
      if (DOMElement.pxStyles[name] && isNumber(value)) value = value + 'px';
      this.style[name] = value;
      return this
    }

    _createEventListener (eventName, handler, options) {
      options.context = options.context || this._owner._comp;
      return super._createEventListener(eventName, handler, options)
    }

    getNodeType () {
      return 'element'
    }

    hasInnerHTML () {
      return Boolean(this._innerHTMLString)
    }

    _normalizeChild (child) {
      if (isNil(child)) ; else if (child._isVirtualElement) {
        return child
      } else if (isString(child) || isBoolean(child) || isNumber(child)) {
        return new VirtualTextNode(String(child))
      } else {
        throw new Error('Unsupported child type')
      }
    }

    _append (outlet, args) {
      if (args.length === 1 && !isArray(args[0])) {
        this._appendChild(outlet, args[0]);
        return
      }
      var children;
      if (isArray(args[0])) {
        children = args[0];
      } else if (arguments.length > 1) {
        children = Array.prototype.slice.call(args, 0);
      } else {
        return
      }
      children.forEach(this._appendChild.bind(this, outlet));
    }

    _appendChild (outlet, child) {
      child = this._normalizeChild(child);
      
      
      if (!child) return
      outlet.push(child);
      this._attach(child);
      return child
    }

    _insertAt (outlet, pos, child) {
      if (!child) return
      outlet.splice(pos, 0, child);
      this._attach(child);
    }

    _removeAt (outlet, pos) {
      var child = outlet[pos];
      outlet.splice(pos, 1);
      this._detach(child);
    }

    _attach (child) {
      child.parent = this;
      if (this._context && child._owner !== this._owner && child._ref) {
        this._context.foreignRefs[child._ref] = child;
      }
    }

    _detach (child) {
      child.parent = null;
      if (this._context && child._owner !== this._owner && child._ref) {
        delete this.context.foreignRefs[child._ref];
      }
    }

    _mergeHTMLConfig (other) {
      if (other.classNames) {
        if (!this.classNames) {
          this.classNames = [];
        }
        this.classNames = this.classNames.concat(other.classNames);
      }
      if (other.attributes) {
        if (!this.attributes) {
          this.attributes = {};
        }
        Object.assign(this.attributes, other.attributes);
      }
      if (other.htmlProps) {
        if (!this.htmlProps) {
          this.htmlProps = {};
        }
        Object.assign(this.htmlProps, other.htmlProps);
      }
      if (other.style) {
        if (!this.style) {
          this.style = {};
        }
        Object.assign(this.style, other.style);
      }
      if (other.eventListeners) {
        if (!this.eventListeners) {
          this.eventListeners = [];
        }
        this.eventListeners = this.eventListeners.concat(other.eventListeners);
      }
    }
  }

  VirtualHTMLElement.prototype._isVirtualHTMLElement = true;


  class VirtualComponent extends VirtualHTMLElement {
    constructor (ComponentClass, props) {
      super();

      props = props || {};

      this.ComponentClass = ComponentClass;
      this.props = props;
      if (!props.children) {
        props.children = [];
      }
      this.children = props.children;
    }

    get _isVirtualHTMLElement () { return false }

    get _isVirtualComponent () { return true }

    getComponent () {
      return this._comp
    }

    
    
    getChildren () {
      return this.props.children
    }

    getNodeType () {
      return 'component'
    }

    outlet (name) {
      return new Outlet(this, name)
    }

    _attach (child) {
      child._preliminaryParent = this;
    }

    _detach (child) {
      child._preliminaryParent = null;
    }

    _copyHTMLConfig () {
      return {
        classNames: clone(this.classNames),
        attributes: clone(this.attributes),
        htmlProps: clone(this.htmlProps),
        style: clone(this.style),
        eventListeners: clone(this.eventListeners)
      }
    }
  }

  class Outlet {
    constructor (virtualEl, name) {
      this.virtualEl = virtualEl;
      this.name = name;
      Object.freeze(this);
    }

    _getOutlet () {
      var outlet = this.virtualEl.props[this.name];
      if (!outlet) {
        outlet = [];
        this.virtualEl.props[this.name] = outlet;
      }
      return outlet
    }

    append () {
      var outlet = this._getOutlet();
      this.virtualEl._append(outlet, arguments);
      return this
    }

    empty () {
      var arr = this.virtualEl.props[this.name];
      arr.forEach(function (el) {
        this._detach(el);
      }.bind(this));
      arr.splice(0, arr.length);
      return this
    }
  }

  class VirtualTextNode extends VirtualElement {
    constructor (text) {
      super();
      this.text = text;
    }

    get _isVirtualTextNode () { return true }
  }

  VirtualElement.Component = VirtualComponent;
  VirtualElement.TextNode = VirtualTextNode;


  VirtualElement.createElement = function () {
    var content;
    var _first = arguments[0];
    var _second = arguments[1];
    var type;
    if (isString(_first)) {
      type = 'element';
    } else if (isFunction$1(_first) && _first.prototype._isComponent) {
      type = 'component';
    } else if (isNil(_first)) {
      throw new Error('$$(null): provided argument was null or undefined.')
    } else {
      throw new Error('Illegal usage of $$()')
    }
    
    var props = {};
    var classNames, ref;
    var eventHandlers = [];
    for (var key in _second) {
      if (!_second.hasOwnProperty(key)) continue
      var val = _second[key];
      switch (key) {
        case 'class':
          classNames = val;
          break
        case 'ref':
          ref = val;
          break
        default:
          props[key] = val;
      }
    }
    if (type === 'element') {
      content = new VirtualHTMLElement(_first);
      
      
      content.attr(props);
    } else {
      content = new VirtualComponent(_first, props);
    }
    
    
    content._owner = this.owner;
    if (classNames) {
      content.addClass(classNames);
    }
    if (ref) {
      content.ref(ref);
    }
    eventHandlers.forEach(function (h) {
      if (isFunction$1(h.handler)) {
        content.on(h.name, h.handler);
      } else if (isPlainObject(h.handler)) {
        var params = h.handler;
        content.on(h.name, params.handler, params.context, params);
      } else {
        throw new Error('Illegal arguments for $$(_,{ on' + h.name + '})')
      }
    });
    
    
    if (arguments.length > 2) {
      content.append(flattenOften(Array.prototype.slice.call(arguments, 2), 3));
    }
    return content
  };

  class RenderingEngine {
    constructor (options = {}) {
      this.componentFactory = options.componentFactory;
      if (!this.componentFactory) throw new Error("'componentFactory' is mandatory")
      this.elementFactory = options.elementFactory || DefaultDOMElement.createDocument('html');
      if (!this.elementFactory) throw new Error("'elementFactory' is mandatory")
    }

    _render (comp, oldProps, oldState) {
      
      var vel = _createWrappingVirtualComponent(comp);
      var state = this._createState();
      if (oldProps) {
        state.setOldProps(vel, oldProps);
      }
      if (oldState) {
        state.setOldState(vel, oldState);
      }
      try {
        
        
        
        _capture(state, vel, 'forceCapture');
        

        
        
        _render(state, vel);
        

        _triggerUpdate(state, vel);
      } finally {
        state.dispose();
      }
      
    }

    
    _renderChild (comp, vel) {
      
      
      var state = this._createState();
      vel.parent = { _comp: comp };
      try {
        _capture(state, vel);
        _render(state, vel);
        return vel._comp
      } finally {
        state.dispose();
      }
    }

    _createState () {
      return new RenderingState(this.componentFactory, this.elementFactory)
    }
  }

  function _create (state, vel) {
    var comp = vel._comp;
    console.assert(!comp, 'Component instance should not exist when this method is used.');
    var parent = vel.parent._comp;
    
    if (!parent) {
      parent = _create(state, vel.parent);
    }
    if (vel._isVirtualComponent) {
      console.assert(parent, 'A Component should have a parent.');
      comp = state.componentFactory.createComponent(vel.ComponentClass, parent, vel.props);
      
      vel.props = comp.props;
      comp.__htmlConfig__ = vel._copyHTMLConfig();
    } else if (vel._isVirtualHTMLElement) {
      comp = state.componentFactory.createElementComponent(parent, vel);
    } else if (vel._isVirtualTextNode) {
      comp = state.componentFactory.createTextNodeComponent(parent, vel);
    }
    if (vel._ref) {
      comp._ref = vel._ref;
    }
    if (vel._owner) {
      comp._owner = vel._owner._comp;
    }
    vel._comp = comp;
    return comp
  }

  function _capture (state, vel, forceCapture) {
    if (state.isCaptured(vel)) {
      return vel
    }
    
    var comp = vel._comp;
    if (!comp) {
      comp = _create(state, vel);
      state.setNew(vel);
    }
    if (vel._isVirtualComponent) {
      var needRerender;
      
      
      
      if (forceCapture) {
        needRerender = true;
      } else {
        
        needRerender = !comp.el || comp.shouldRerender(vel.props, comp.state);
        comp.__htmlConfig__ = vel._copyHTMLConfig();
        state.setOldProps(vel, comp.props);
        state.setOldState(vel, comp.state);
        
        comp._setProps(vel.props);
        if (!state.isNew(vel)) {
          state.setUpdated(vel);
        }
      }
      if (needRerender) {
        var context = new CaptureContext(vel);
        var content = comp.render(context.$$);
        if (!content) {
          throw new Error('Component.render() returned nil.')
        }
        if (!content._isVirtualHTMLElement) {
          throw new Error('Component.render() must return a plain element.')
        }

        if (comp.__htmlConfig__) {
          content._mergeHTMLConfig(comp.__htmlConfig__);
        }
        content._comp = comp;
        vel._content = content;
        if (!state.isNew(vel) && comp.isMounted()) {
          state.setUpdated(vel);
        }
        
        _prepareVirtualComponent(state, comp, content);
        
        
        if (substanceGlobals.DEBUG_RENDERING) {
          
          
          
          
          var stack = content.children.slice(0);
          while (stack.length) {
            var child = stack.shift();
            if (state.isCaptured(child)) continue
            
            if (child._isVirtualComponent) continue
            if (!child._comp) {
              _create(state, child);
            }
            if (child._isVirtualHTMLElement && child.children.length > 0) {
              stack = stack.concat(child.children);
            }
            state.setCaptured(child);
          }
          state.setCaptured(content);
          
          
          var descendingContext = new DescendingContext(state, context);
          while (descendingContext.hasPendingCaptures()) {
            descendingContext.reset();
            comp.render(descendingContext.$$);
          }
        } else {
          
          
          _capture(state, vel._content);
        }
      } else {
        state.setSkipped(vel);
      }
    } else if (vel._isVirtualHTMLElement) {
      for (var i = 0; i < vel.children.length; i++) {
        _capture(state, vel.children[i]);
      }
    }
    state.setCaptured(vel);
    return vel
  }

  function _render (state, vel) {
    if (state.isSkipped(vel)) return
    

    
    
    
    
    
    

    let comp = vel._comp;
    console.assert(comp && comp._isComponent, 'A captured VirtualElement must have a component instance attached.');

    
    if (vel._isVirtualComponent) {
      _render(state, vel._content);

      
      const context = vel._content._context;
      let refs = {};
      let foreignRefs = {};
      forEach(context.refs, (vel, ref) => {
        refs[ref] = vel._comp;
      });
      forEach(context.foreignRefs, (vel, ref) => {
        foreignRefs[ref] = vel._comp;
      });
      comp.refs = refs;
      comp.__foreignRefs__ = foreignRefs;
      return
    }

    
    if (!comp.el) {
      comp.el = _createElement(state, vel);
      comp.el._comp = comp;
    }
    _updateElement(comp, vel);

    
    if (vel._isVirtualHTMLElement && !vel.hasInnerHTML()) {
      var newChildren = vel.children;
      var oldComp, virtualComp, newComp;
      var pos1 = 0; var pos2 = 0;

      
      
      
      var oldChildren = [];
      comp.el.getChildNodes().forEach(function (node) {
        var childComp = node._comp;

        
        
        
        

        
        if (!childComp || state.isRelocated(childComp)) {
          comp.el.removeChild(node);
        } else {
          oldChildren.push(childComp);
        }
      });

      while (pos1 < oldChildren.length || pos2 < newChildren.length) {
        
        
        
        
        do {
          oldComp = oldChildren[pos1++];
        } while (oldComp && (state.isDetached(oldComp)))

        virtualComp = newChildren[pos2++];
        
        if (oldComp && !virtualComp) {
          while (oldComp) {
            _removeChild(state, comp, oldComp);
            oldComp = oldChildren[pos1++];
          }
          break
        }

        
        if (oldComp && oldComp.el.isTextNode() &&
            virtualComp && virtualComp._isVirtualTextNode &&
            oldComp.el.textContent === virtualComp.text) {
          continue
        }

        
        if (!state.isRendered(virtualComp)) {
          _render(state, virtualComp);
        }

        newComp = virtualComp._comp;

        
        
        if (state.isRelocated(newComp)) {
          newComp._setParent(comp);
        }

        console.assert(newComp, 'Component instance should now be available.');

        
        if (virtualComp && !oldComp) {
          _appendChild(state, comp, newComp);
          continue
        
        } else if (state.isMapped(virtualComp)) {
          
          if (newComp === oldComp) ; else if (state.isMapped(oldComp)) {
            
            state.setDetached(oldComp);
            _removeChild(state, comp, oldComp);
            pos2--;
          
          } else {
            _removeChild(state, comp, oldComp);
            pos2--;
          }
        } else if (state.isMapped(oldComp)) {
          _insertChildBefore(state, comp, newComp, oldComp);
          pos1--;
        } else {
          
          
          
          
          _replaceChild(state, comp, oldComp, newComp);
        }
      }
    }

    state.setRendered(vel);
  }

  function _triggerUpdate (state, vel) {
    if (vel._isVirtualComponent) {
      if (!state.isSkipped(vel)) {
        vel._content.children.forEach(_triggerUpdate.bind(null, state));
      }
      if (state.isUpdated(vel)) {
        vel._comp.didUpdate(state.getOldProps(vel), state.getOldState(vel));
      }
    } else if (vel._isVirtualHTMLElement) {
      vel.children.forEach(_triggerUpdate.bind(null, state));
    }
  }

  function _appendChild (state, parent, child) {
    parent.el.appendChild(child.el);
    _triggerDidMount(state, parent, child);
  }

  function _replaceChild (state, parent, oldChild, newChild) {
    parent.el.replaceChild(oldChild.el, newChild.el);
    if (!state.isDetached(oldChild)) {
      oldChild.triggerDispose();
    }
    _triggerDidMount(state, parent, newChild);
  }

  function _insertChildBefore (state, parent, child, before) {
    parent.el.insertBefore(child.el, before.el);
    _triggerDidMount(state, parent, child);
  }

  function _removeChild (state, parent, child) {
    parent.el.removeChild(child.el);
    if (!state.isDetached(child)) {
      child.triggerDispose();
    }
  }

  function _triggerDidMount (state, parent, child) {
    if (!state.isDetached(child) &&
        parent.isMounted() && !child.isMounted()) {
      child.triggerDidMount(true);
    }
  }


  function _prepareVirtualComponent (state, comp, vc) {
    var newRefs = {};
    var foreignRefs = {};
    
    
    
    
    if (vc._context) {
      newRefs = vc._context.refs;
      foreignRefs = vc._context.foreignRefs;
    }
    var oldRefs = comp.refs;
    var oldForeignRefs = comp.__foreignRefs__;
    
    forEach(newRefs, function (vc, ref) {
      var comp = oldRefs[ref];
      if (comp) _mapComponents(state, comp, vc);
    });
    forEach(foreignRefs, function (vc, ref) {
      var comp = oldForeignRefs[ref];
      if (comp) _mapComponents(state, comp, vc);
    });
  }



  function _mapComponents (state, comp, vc) {
    if (!comp && !vc) return true
    if (!comp || !vc) return false
    
    
    
    
    
    if (state.isMapped(vc) || state.isMapped(comp)) {
      return vc._comp === comp
    }
    if (vc._comp) {
      if (vc._comp === comp) {
        state.setMapped(vc);
        state.setMapped(comp);
        return true
      } else {
        return false
      }
    }
    if (!_isOfSameType(comp, vc)) {
      return false
    }

    vc._comp = comp;
    state.setMapped(vc);
    state.setMapped(comp);

    var canMapParent;
    var parent = comp.getParent();
    if (vc.parent) {
      canMapParent = _mapComponents(state, parent, vc.parent);
    
    
    
    
    } else if (vc._preliminaryParent) {
      while (parent && parent._isElementComponent) {
        parent = parent.getParent();
      }
      canMapParent = _mapComponents(state, parent, vc._preliminaryParent);
    }
    if (!canMapParent) {
      state.setRelocated(vc);
      state.setRelocated(comp);
    }
    return canMapParent
  }

  function _isOfSameType (comp, vc) {
    return (
      (comp._isElementComponent && vc._isVirtualHTMLElement) ||
      (comp._isComponent && vc._isVirtualComponent && comp.constructor === vc.ComponentClass) ||
      (comp._isTextNodeComponent && vc._isVirtualTextNode)
    )
  }

  function _createElement (state, vel) {
    var el;
    if (vel._isVirtualTextNode) {
      el = state.elementFactory.createTextNode(vel.text);
    } else {
      el = state.elementFactory.createElement(vel.tagName);
    }
    return el
  }

  function _updateElement (comp, vel) {
    if (comp._isTextNodeComponent) {
      comp.setTextContent(vel.text);
      return
    }
    var el = comp.el;
    console.assert(el, "Component's element should exist at this point.");
    var tagName = el.getTagName();
    if (vel.tagName.toLowerCase() !== tagName) {
      el.setTagName(vel.tagName);
    }
    _updateHash({
      oldHash: el.getAttributes(),
      newHash: vel.getAttributes(),
      update: function (key, val) {
        el.setAttribute(key, val);
      },
      remove: function (key) {
        el.removeAttribute(key);
      }
    });
    _updateHash({
      oldHash: el.htmlProps,
      newHash: vel.htmlProps,
      update: function (key, val) {
        el.setProperty(key, val);
      },
      remove: function (key) {
        el.removeProperty(key);
      }
    });
    _updateListeners({
      el: el,
      oldListeners: el.getEventListeners(),
      newListeners: vel.getEventListeners()
    });

    
    if (vel.hasInnerHTML()) {
      if (!el._hasInnerHTML) {
        el.empty();
        el.setInnerHTML(vel.getInnerHTML());
      } else {
        var oldInnerHTML = el.getInnerHTML();
        var newInnerHTML = vel.getInnerHTML();
        if (oldInnerHTML !== newInnerHTML) {
          el.setInnerHTML(newInnerHTML);
        }
      }
      el._hasInnerHTML = true;
    }
  }

  function _updateHash (args) {
    const newHash = args.newHash;
    const oldHash = args.oldHash || {};
    const update = args.update;
    const remove = args.remove;
    let updatedKeys = {};
    for (let key in newHash) {
      if (newHash.hasOwnProperty(key)) {
        var oldVal = oldHash[key];
        var newVal = newHash[key];
        updatedKeys[key] = true;
        if (oldVal !== newVal) {
          update(key, newVal);
        }
      }
    }
    
    
    if (isFunction$1(oldHash.keys) && oldHash.size > 0) {
      let keys = Array.from(oldHash.keys());
      keys.forEach((key) => {
        if (!updatedKeys[key]) {
          remove(key);
        }
      });
    } else {
      for (let key in oldHash) {
        if (oldHash.hasOwnProperty(key) && !updatedKeys[key]) {
          remove(key);
        }
      }
    }
  }

  function _updateListeners (args) {
    var el = args.el;
    
    
    
    var newListeners = args.newListeners || [];
    el.removeAllEventListeners();
    for (var i = 0; i < newListeners.length; i++) {
      el.addEventListener(newListeners[i]);
    }
  }



  class DescendingContext {
    constructor (state, captureContext) {
      this.state = state;
      this.owner = captureContext.owner;
      this.refs = {};
      this.foreignRefs = {};
      this.elements = captureContext.elements;
      this.pos = 0;
      this.updates = captureContext.components.length;
      this.remaining = this.updates;

      this.$$ = this._createComponent.bind(this);
    }

    _createComponent () {
      var state = this.state;
      var vel = this.elements[this.pos++];
      
      
      
      if (!state.isCaptured(vel) && vel._isVirtualComponent &&
           vel.parent && state.isCaptured(vel.parent)) {
        _capture(state, vel);
        this.updates++;
        this.remaining--;
      }
      
      
      
      
      vel = VirtualElement.createElement.apply(this, arguments);
      
      vel._context = this;
      vel._owner = this.owner;
      
      
      vel._attach = function () {};
      vel._detach = function () {};
      return vel
    }

    hasPendingCaptures () {
      return this.updates > 0 && this.remaining > 0
    }

    reset () {
      this.pos = 0;
      this.updates = 0;
      this.refs = {};
    }

    _ancestorsReady (vel) {
      while (vel) {
        if (this.state.isCaptured(vel) ||
            
            vel === this.owner || vel === this.owner._content) {
          return true
        }
        vel = vel.parent;
      }
      return false
    }
  }

  RenderingEngine._internal = {
    _capture: _capture,
    _wrap: _createWrappingVirtualComponent
  };

  class CaptureContext {
    constructor (owner) {
      this.owner = owner;
      this.refs = {};
      this.foreignRefs = {};
      this.elements = [];
      this.components = [];
      this.$$ = this._createComponent.bind(this);
      this.$$.capturing = true;
    }

    _createComponent () {
      var vel = VirtualElement.createElement.apply(this, arguments);
      vel._context = this;
      vel._owner = this.owner;
      if (vel._isVirtualComponent) {
        
        this.components.push(vel);
      }
      this.elements.push(vel);
      return vel
    }
  }

  function _createWrappingVirtualComponent (comp) {
    var vel = new VirtualElement.Component(comp.constructor);
    vel._comp = comp;
    if (comp.__htmlConfig__) {
      vel._mergeHTMLConfig(comp.__htmlConfig__);
    }
    return vel
  }

  RenderingEngine.createContext = function (comp) {
    var vel = _createWrappingVirtualComponent(comp);
    return new CaptureContext(vel)
  };

  class RenderingState {
    constructor (componentFactory, elementFactory) {
      this.componentFactory = componentFactory;
      this.elementFactory = elementFactory;
      this.poluted = [];
      this.id = '__' + uuid();
    }

    dispose () {
      var id = this.id;
      this.poluted.forEach(function (obj) {
        delete obj[id];
      });
    }

    set (obj, key, val) {
      var info = obj[this.id];
      if (!info) {
        info = {};
        obj[this.id] = info;
        this.poluted.push(obj);
      }
      info[key] = val;
    }

    get (obj, key) {
      var info = obj[this.id];
      if (info) {
        return info[key]
      }
    }

    setMapped (c) {
      this.set(c, 'mapped', true);
    }

    isMapped (c) {
      return Boolean(this.get(c, 'mapped'))
    }

    
    
    setRelocated (c) {
      this.set(c, 'relocated', true);
    }

    isRelocated (c) {
      return Boolean(this.get(c, 'relocated'))
    }

    setDetached (c) {
      this.set(c, 'detached', true);
    }

    isDetached (c) {
      return Boolean(this.get(c, 'detached'))
    }

    setCaptured (vc) {
      this.set(vc, 'captured', true);
    }

    isCaptured (vc) {
      return Boolean(this.get(vc, 'captured'))
    }

    setNew (vc) {
      this.set(vc, 'created', true);
    }

    isNew (vc) {
      return Boolean(this.get(vc, 'created'))
    }

    setUpdated (vc) {
      this.set(vc, 'updated', true);
    }

    isUpdated (vc) {
      return Boolean(this.get(vc, 'updated'))
    }

    setSkipped (vc) {
      this.set(vc, 'skipped', true);
    }

    isSkipped (vc) {
      return Boolean(this.get(vc, 'skipped'))
    }

    setRendered (vc) {
      this.set(vc, 'rendered', true);
    }

    isRendered (vc) {
      return Boolean(this.get(vc, 'rendered'))
    }

    setOldProps (vc, oldProps) {
      this.set(vc, 'oldProps', oldProps);
    }

    getOldProps (vc) {
      return this.get(vc, 'oldProps')
    }

    setOldState (vc, oldState) {
      this.set(vc, 'oldState', oldState);
    }

    getOldState (vc) {
      return this.get(vc, 'oldState')
    }
  }

  const COMPONENT_FACTORY = {
    createComponent (ComponentClass, parent, props) {
      return new ComponentClass(parent, props)
    },
    createElementComponent (parent, virtualElement) {
      return new ElementComponent(parent, virtualElement)
    },
    createTextNodeComponent (parent, virtualElement) {
      return new TextNodeComponent(parent, virtualElement)
    }
  };

  class Component extends EventEmitter {
    
    constructor (parent, props = {}, options = {}) {
      super();

      
      
      
      
      

      this.parent = (parent && parent._isComponent) ? parent : null;

      
      this.el = options.el;

      
      
      let context = options.context ? options.context : this._getContext() || {};
      this.context = context;
      Object.freeze(this.context);

      
      
      
      this.renderingEngine = (parent && parent.renderingEngine) || context.renderingEngine || options.renderingEngine || new RenderingEngine({
        componentFactory: COMPONENT_FACTORY
      });

      
      
      if (this._SKIP_COMPONENT_INIT) return

      this.__id__ = uuid();

      
      this.refs = {};
      
      
      
      this.__foreignRefs__ = {};

      
      this._actionHandlers = {};

      
      this.props = props;
      Object.freeze(this.props);

      
      this.state = this.getInitialState() || {};
      Object.freeze(this.state);
    }

    getId () {
      return this.__id__
    }

    setId () {
      throw new Error("'id' is readonly")
    }

    
    getChildContext () {
      return this.childContext || {}
    }

    
    getInitialState () {
      return {}
    }

    
    getParent () {
      return this.parent
    }

    
    getRoot () {
      var comp = this;
      var parent = comp;
      while (parent) {
        comp = parent;
        parent = comp.getParent();
      }
      return comp
    }

    getElement () {
      return this.el
    }

    getNativeElement () {
      return this.el.getNativeElement()
    }

    
    getLabel (name, ...args) {
      let labelProvider = this.getLabelProvider();
      if (!labelProvider) throw new Error('Missing labelProvider.')
      return labelProvider.getLabel(name, ...args)
    }

    getLabelProvider () {
      return this.context.labelProvider
    }

    
    getComponent (componentName, maybe) {
      let componentRegistry = this.getComponentRegistry();
      if (!componentRegistry) throw new Error('Missing componentRegistry.')
      const ComponentClass = componentRegistry.get(componentName);
      if (!maybe && !ComponentClass) {
        throw new Error('No Component registered with name ' + componentName)
      }
      return ComponentClass
    }

    getComponentRegistry () {
      return this.props.componentRegistry || this.context.componentRegistry
    }

    getFlow () {
      return this.context.flow
    }

    
    render ($$) {
      
      return $$('div')
    }

    
    mount (el) {
      if (!el) {
        throw new Error('Element is required.')
      }
      el = DefaultDOMElement.wrap(el);
      
      this.el = null;
      this.renderingEngine = Component.createRenderingEngine(el.getOwnerDocument());
      this._render();
      el.appendChild(this.el);
      if (el.isInDocument()) {
        this.triggerDidMount(true);
      }
      return this
    }

    
    shouldRerender(newProps, newState) { 
      return true
    }

    
    rerender () {
      this._rerender(this.props, this.state);
    }

    _rerender (oldProps, oldState) {
      this._render(oldProps, oldState);
      
      if (!this.isMounted()) {
        this.didUpdate(oldProps, oldState);
      }
    }

    _render (oldProps, oldState) {
      if (this.__isRendering__) {
        throw new Error('Component is rendering already.')
      }
      this.__isRendering__ = true;
      try {
        this.renderingEngine._render(this, oldProps, oldState);
      } finally {
        delete this.__isRendering__;
      }
    }

    
    triggerDidMount () {
      
      
      

      
      
      if (!this.__isMounted__) {
        this.__isMounted__ = true;
        this.didMount();
      }
      
      this.getChildren().forEach(function (child) {
        
        
        child.triggerDidMount(true);
      });
    }

    
    didMount () {}

    
    didUpdate () {}

    
    isMounted () {
      return this.__isMounted__
    }

    
    triggerDispose () {
      this.getChildren().forEach(function (child) {
        child.triggerDispose();
      });
      this.dispose();
      this.__isMounted__ = false;
    }

    
    dispose () {}

    
    _setParent (newParent) {
      this.parent = newParent;
      this.context = this._getContext() || {};
      Object.freeze(this.context);
    }

    
    send (action) {
      
      var comp = this.getParent();
      while (comp) {
        if (comp._actionHandlers && comp._actionHandlers[action]) {
          comp._actionHandlers[action].apply(comp, Array.prototype.slice.call(arguments, 1));
          return true
        }
        comp = comp.getParent();
      }
      console.warn('Action', action, 'was not handled.');
      return false
    }

    
    handleActions (actionHandlers) {
      forEach(actionHandlers, function (method, actionName) {
        this.handleAction(actionName, method);
      }.bind(this));
      return this
    }

    
    handleAction (name, handler) {
      if (!name || !handler || !isFunction$1(handler)) {
        throw new Error('Illegal arguments.')
      }
      handler = handler.bind(this);
      this._actionHandlers[name] = handler;
    }

    
    getState () {
      return this.state
    }

    
    setState (newState) {
      var oldProps = this.props;
      var oldState = this.state;
      
      
      var needRerender = !this.__isSettingProps__ &&
        this.shouldRerender(this.getProps(), newState);
      
      this.willUpdateState(newState);
      this.state = newState || {};
      Object.freeze(this.state);
      if (needRerender) {
        this._rerender(oldProps, oldState);
      } else if (!this.__isSettingProps__) {
        this.didUpdate(oldProps, oldState);
      }
    }

    
    extendState (newState) {
      newState = extend({}, this.state, newState);
      this.setState(newState);
    }

    
    willUpdateState(newState) { 
    }

    
    getProps () {
      return this.props
    }

    
    setProps (newProps) {
      var oldProps = this.props;
      var oldState = this.state;
      var needRerender = this.shouldRerender(newProps, this.state);
      this._setProps(newProps);
      if (needRerender) {
        this._rerender(oldProps, oldState);
      } else {
        this.didUpdate(oldProps, oldState);
      }
    }

    _setProps (newProps) {
      newProps = newProps || {};
      
      this.__isSettingProps__ = true;
      try {
        this.willReceiveProps(newProps);
        this.props = newProps || {};
        Object.freeze(newProps);
      } finally {
        delete this.__isSettingProps__;
      }
    }

    
    extendProps (updatedProps) {
      var newProps = extend({}, this.props, updatedProps);
      this.setProps(newProps);
    }

    
    willReceiveProps(newProps) { 
    }

    getTextContent () {
      if (this.el) {
        return this.el.textContent
      }
    }

    get textContent () {
      return this.getTextContent()
    }

    getInnerHTML () {
      if (this.el) {
        return this.el.getInnerHTML()
      }
    }

    get innerHTML () {
      return this.getInnerHTML()
    }

    getOuterHTML () {
      if (this.el) {
        return this.el.getOuterHTML()
      }
    }

    get outerHTML () {
      return this.getOuterHTML()
    }

    getAttribute (name) {
      if (this.el) {
        return this.el.getAttribute(name)
      }
    }

    setAttribute (name, val) {
      if (this.el) {
        this.el.setAttribute(name, val);
      }
      return this
    }

    getProperty (name) {
      if (this.el) {
        return this.el.getProperty(name)
      }
    }

    setProperty (name, val) {
      if (this.el) {
        this.el.setProperty(name, val);
      }
      return this
    }

    hasClass (name) {
      if (this.el) {
        return this.el.hasClass(name)
      }
    }

    addClass (name) {
      if (this.el) {
        this.el.addClass(name);
      }
      return this
    }

    removeClass (name) {
      if (this.el) {
        this.el.removeClass(name);
      }
      return this
    }

    getStyle (name) {
      if (this.el) {
        return this.el.getStyle(name)
      }
    }

    setStyle (name, val) {
      if (this.el) {
        return this.el.setStyle(name, val)
      }
      return this
    }

    getValue () {
      if (this.el) {
        return this.el.getValue()
      }
    }

    setValue (val) {
      if (this.el) {
        this.el.setValue(val);
      }
      return this
    }

    getChildCount () {
      if (!this.el) return 0
      return this.el.getChildCount()
    }

    get childNodes () {
      return this.getChildNodes()
    }

    getChildNodes () {
      if (!this.el) return []
      var childNodes = this.el.getChildNodes();
      childNodes = childNodes.map(_unwrapComp).filter(Boolean);
      return childNodes
    }

    getChildren () {
      if (!this.el) return []
      let children = this.el.getChildren();
      children = children.map(_unwrapComp).filter(Boolean);
      return children
    }

    getChildAt (pos) {
      let node = this.el.getChildAt(pos);
      if (node) {
        return _unwrapCompStrict(node)
      }
    }

    find (cssSelector) {
      var el = this.el.find(cssSelector);
      return _unwrapComp(el)
    }

    findAll (cssSelector) {
      var els = this.el.findAll(cssSelector);
      return els.map(_unwrapComp).filter(Boolean)
    }

    appendChild (child) {
      this.insertAt(this.getChildCount(), child);
    }

    insertAt (pos, childEl) {
      if (isString(childEl)) {
        childEl = new VirtualElement.TextNode(childEl);
      }
      if (!childEl._isVirtualElement) {
        throw new Error('Invalid argument: "child" must be a VirtualElement.')
      }
      var child = this.renderingEngine._renderChild(this, childEl);
      this.el.insertAt(pos, child.el);
      _mountChild(this, child);
    }

    removeAt (pos) {
      var childEl = this.el.getChildAt(pos);
      if (childEl) {
        var child = _unwrapCompStrict(childEl);
        _disposeChild(child);
        this.el.removeAt(pos);
      }
    }

    removeChild (child) {
      if (!child || !child._isComponent) {
        throw new Error('removeChild(): Illegal arguments. Expecting a Component instance.')
      }
      
      _disposeChild(child);
      this.el.removeChild(child.el);
    }

    replaceChild (oldChild, newChild) {
      if (!newChild || !oldChild ||
          !newChild._isComponent || !oldChild._isComponent) {
        throw new Error('replaceChild(): Illegal arguments. Expecting BrowserDOMElement instances.')
      }
      
      _disposeChild(oldChild);
      this.el.replaceChild(newChild.el, oldChild.el);
      if (this.isMounted()) {
        newChild.triggerDidMount(true);
      }
    }

    
    
    
    
    
    
    empty () {
      this._clear();
      return this
    }

    _clear () {
      let el = this.el;
      if (el) {
        this.getChildNodes().forEach(function (child) {
          _disposeChild(child);
        });
        el.empty();
      }
      this.refs = {};
      this.__foreignRefs__ = {};
    }

    remove () {
      _disposeChild(this);
      this.el.remove();
    }

    addEventListener () {
      throw new Error('Not supported.')
    }

    removeEventListener () {
      throw new Error('Not supported.')
    }

    insertBefore () {
      throw new Error('Not supported.')
    }

    click () {
      if (this.el) {
        this.el.click();
      }
    }

    getComponentPath () {
      let path = [];
      let comp = this;
      while (comp) {
        path.unshift(comp);
        comp = comp.getParent();
      }
      return path
    }

    _getContext () {
      var context = {};
      var parent = this.getParent();
      if (parent) {
        context = extend(context, parent.context);
        if (parent.getChildContext) {
          return extend(context, parent.getChildContext())
        }
      }
      return context
    }
  }

  Component.prototype._isComponent = true;

  Component.prototype.attr = DOMElement.prototype.attr;

  Component.prototype.htmlProp = DOMElement.prototype.htmlProp;

  Component.prototype.val = DOMElement.prototype.val;

  Component.prototype.css = DOMElement.prototype.css;

  Component.prototype.text = DOMElement.prototype.text;

  Component.prototype.append = DOMElement.prototype.append;

  Component.unwrap = _unwrapComp;

  Component.render = function (props) {
    props = props || {};
    var ComponentClass = this;
    var comp = new ComponentClass(null, props);
    comp._render();
    return comp
  };

  Component.mount = function (props, el) {
    if (arguments.length === 1) {
      el = props;
      props = {};
    }
    if (!el) throw new Error("'el' is required.")
    if (isString(el)) {
      var selector = el;
      if (platform.inBrowser) {
        el = window.document.querySelector(selector);
      } else {
        throw new Error('This selector is not supported on server side.')
      }
    }
    el = DefaultDOMElement.wrap(el);
    const ComponentClass = this;
    let comp = new ComponentClass(null, props);
    comp.mount(el);
    return comp
  };

  Component.getComponentForDOMElement = function (el) {
    return _unwrapComp(el)
  };

  Component.unwrapDOMElement = function (el) {
    console.warn('DEPRECATED: Use Component.getComponentForDOMElement');
    return Component.getComponentForDOMElement(el)
  };

  Component.getComponentFromNativeElement = function (nativeEl) {
    
    
    
    return _unwrapComp(DefaultDOMElement.wrap(nativeEl))
  };

  Component.createRenderingEngine = function (elementFactory) {
    return new RenderingEngine({
      componentFactory: COMPONENT_FACTORY,
      elementFactory
    })
  };


  function _disposeChild (child) {
    child.triggerDispose();
    if (child._owner && child._ref) {
      console.assert(child._owner.refs[child._ref] === child, "Owner's ref should point to this child instance.");
      delete child._owner.refs[child._ref];
    }
  }


  function _mountChild (parent, child) {
    if (parent.isMounted()) {
      child.triggerDidMount(true);
    }
    if (child._owner && child._ref) {
      child._owner.refs[child._ref] = child;
    }
  }


  function _unwrapComp (el) {
    if (el) {
      if (!el._isDOMElement) el = DefaultDOMElement.unwrap(el);
      if (el) return el._comp
    }
  }

  function _unwrapCompStrict (el) {
    let comp = _unwrapComp(el);
    if (!comp) throw new Error('Expecting a back-link to the component instance.')
    return comp
  }

  class ElementComponent extends Component {}

  ElementComponent.prototype._isElementComponent = true;
  ElementComponent.prototype._SKIP_COMPONENT_INIT = true;

  class TextNodeComponent extends Component {
    setTextContent (text) {
      if (!this.el) {
        throw new Error('Component must be rendered first.')
      }
      if (this.el.textContent !== text) {
        this.el.textContent = text;
      }
    }

    getChildNodes () {
      return []
    }

    getChildren () {
      return []
    }
  }

  TextNodeComponent.prototype._isTextNodeComponent = true;
  TextNodeComponent.prototype._SKIP_COMPONENT_INIT = true;

  Component.Element = ElementComponent;
  Component.TextNode = TextNodeComponent;

  function getDOMRangeFromEvent (evt) {
    let range;
    let x = evt.clientX;
    let y = evt.clientY;
    
    if (document.body.createTextRange) {
      range = document.body.createTextRange();
      range.moveToPoint(x, y);
    } else if (!isNil(document.createRange)) {
      
      
      if (!isNil(evt.rangeParent)) {
        range = document.createRange();
        range.setStart(evt.rangeParent, evt.rangeOffset);
        range.collapse(true);
      
      } else if (document.caretPositionFromPoint) {
        let pos = document.caretPositionFromPoint(x, y);
        range = document.createRange();
        range.setStart(pos.offsetNode, pos.offset);
        range.collapse(true);
      
      } else if (document.caretRangeFromPoint) {
        range = document.caretRangeFromPoint(x, y);
      }
    }
    return range
  }

  const events = [ 'keydown', 'keyup', 'keypress', 'mousedown', 'mouseup', 'copy' ];

  class GlobalEventHandler {
    constructor (editorSession, surfaceManager) {
      this.editorSession = editorSession;
      this.surfaceManager = surfaceManager;
      this.listeners = [];
      this.initialize();
    }

    initialize () {
      if (platform.inBrowser) {
        let document = DefaultDOMElement.wrapNativeElement(window.document);
        events.forEach(function (name) {
          document.on(name, this._dispatch.bind(this, name), this);
        }.bind(this));
      }
    }

    dispose () {
      if (platform.inBrowser) {
        let document = DefaultDOMElement.wrapNativeElement(window.document);
        document.off(this);
      }
    }

    addEventListener (eventName, handler, options) {
      if (!options.id) {
        throw new Error("GlobalEventHandler can only be used with option 'id'")
      }
      let listener = new DOMEventListener(eventName, handler, options);
      this.listeners.push(listener);
    }

    removeEventListener (listener) {
      let idx = this.listeners.indexOf(listener);
      if (idx > -1) {
        this.listeners.splice(idx, 1);
      }
    }

    getEventListeners () {
      return this.listeners
    }

    _getActiveListener (eventName) {
      let editorSession = this.editorSession;
      let sel = editorSession.getSelection();
      if (sel) {
        let surfaceId = sel.surfaceId;
        for (let i = 0; i < this.listeners.length; i++) {
          let listener = this.listeners[i];
          if (listener.eventName === eventName && listener.options.id === surfaceId) {
            return listener
          }
        }
      }
    }

    _dispatch (eventName, e) {
      let listener = this._getActiveListener(eventName);
      if (listener) {
        listener.handler(e);
      }
    }
  }

  GlobalEventHandler.prototype.on = DOMElement.prototype.on;
  GlobalEventHandler.prototype.off = DOMElement.prototype.off;

  let keys$1 = {
    UNDEFINED: 0,
    BACKSPACE: 8,
    DELETE: 46,
    INSERT: 45,
    LEFT: 37,
    RIGHT: 39,
    UP: 38,
    DOWN: 40,
    ENTER: 13,
    RETURN: 13,
    END: 35,
    HOME: 36,
    TAB: 9,
    PAGEUP: 33,
    PAGEDOWN: 34,
    ESCAPE: 27,
    ESC: 27,
    SHIFT: 16,
    SPACE: 32,
    PLUS: 171,
    VOLUMEUP: 183,
    VOLUMEDOWN: 182,
    VOLUMEMUTE: 181,
    PRINTSCREEN: 44
  };


  for (let i = 1; i <= 24; i++) {
    keys$1['F' + i] = 111 + i;
  }

  function parseKeyEvent (event, onlyModifiers) {
    let frags = [];
    if (event.altKey) {
      if (event.code === 'AltRight') {
        frags.push('ALTGR');
      } else {
        frags.push('ALT');
      }
    }
    if (event.ctrlKey) frags.push('CTRL');
    if (event.metaKey) frags.push('META');
    if (event.shiftKey) frags.push('SHIFT');
    if (!onlyModifiers) {
      frags.push(event.keyCode);
    }
    return frags.join('+')
  }

  class Marker extends PropertyAnnotation {
    _initialize (doc, props) {
      this.document = doc;
      this.type = props.type;
      if (!props.type) {
        throw new Error("'type' is mandatory")
      }
      if (!props.start) {
        throw new Error("'start' is mandatory")
      }
      if (!props.end) {
        throw new Error("'end' is mandatory")
      }
      Object.assign(this, props);
    }

    
    containsSelection (sel) {
      if (sel.isNull()) return false
      if (sel.isPropertySelection()) {
        return (isArrayEqual(this.start.path, sel.start.path) &&
          this.start.offset <= sel.start.offset &&
          this.end.offset >= sel.end.offset)
      } else {
        console.warn('Marker.contains() does not support other selection types.');
      }
    }

    get type () {
      return this._type
    }

    set type (type) {
      this._type = type;
    }
  }


  Marker.prototype._isPropertyAnnotation = false;
  Marker.prototype._isMarker = true;

  function TextNodeMixin (SuperClass) {
    class TextNodeMixin extends SuperClass {
      getTextPath () {
        
        console.warn('DEPRECATED: use node.getPath()');
        return this.getPath()
      }

      getText () {
        return this.content
      }

      isEmpty () {
        return !this.getText()
      }

      getLength () {
        return this.getText().length
      }

      getAnnotations () {
        return this.getDocument().getIndex('annotations').get(this.getPath())
      }
    }
    return TextNodeMixin
  }

  const ENTER = 1;
  const EXIT = -1;
  const ANCHOR = -2;






























  class Fragmenter {
    constructor (options) {
      Object.assign(this, options);
    }

    start (rootContext, text, annotations) {
      if (!isString(text)) {
        throw new Error("Illegal argument: 'text' must be a String, but was " + text)
      }
      this._start(rootContext, text, annotations);
    }

    onText(context, text, entry) { 
    }

    
    onEnter(entry, parentContext) { 
      return null
    }

    onExit(entry, context, parentContext) { 
    }

    _enter (entry, parentContext) {
      entry.counter++;
      return this.onEnter(entry, parentContext)
    }

    _exit (entry, context, parentContext) {
      this.onExit(entry, context, parentContext);
    }

    _createText (context, text, entry) {
      this.onText(context, text, entry);
    }

    _start (rootContext, text, annotations) {
      var entries = _extractEntries.call(this, annotations);
      var stack = [{context: rootContext, entry: null}];

      var pos = 0;
      for (var i = 0; i < entries.length; i++) {
        var entry = entries[i];
        var textFragment = text.substring(pos, entry.pos);
        if (textFragment) {
          
          this._createText(stack[stack.length - 1].context, textFragment, entry);
        }

        pos = entry.pos;
        var stackLevel, idx, _entry;
        if (entry.mode === ENTER || entry.mode === ANCHOR) {
          
          for (stackLevel = 1; stackLevel < stack.length; stackLevel++) {
            if (entry.level < stack[stackLevel].entry.level) {
              break
            }
          }
          
          
          for (idx = stack.length - 1; idx >= stackLevel; idx--) {
            _entry = stack[idx].entry;
            
            _entry.length = pos - _entry.pos;
            this._exit(_entry, stack[idx].context, stack[idx - 1].context);
          }
          stack.splice(stackLevel, 0, {entry: entry});
          
          for (idx = stackLevel; idx < stack.length; idx++) {
            _entry = stack[idx].entry;
            
            _entry.pos = pos;
            stack[idx].context = this._enter(_entry, stack[idx - 1].context);
          }
        }
        if (entry.mode === EXIT || entry.mode === ANCHOR) {
          
          for (stackLevel = 1; stackLevel < stack.length; stackLevel++) {
            if (stack[stackLevel].entry.node === entry.node) {
              break
            }
          }
          for (idx = stack.length - 1; idx >= stackLevel; idx--) {
            _entry = stack[idx].entry;
            
            _entry.length = pos - _entry.pos;
            this._exit(_entry, stack[idx].context, stack[idx - 1].context);
          }
          stack.splice(stackLevel, 1);
          
          for (idx = stackLevel; idx < stack.length; idx++) {
            _entry = stack[idx].entry;
            
            _entry.pos = pos;
            stack[idx].context = this._enter(_entry, stack[idx - 1].context);
          }
        }
      }

      
      var trailingText = text.substring(pos);
      if (trailingText) {
        this._createText(rootContext, trailingText);
      }
    }
  }

  Fragmenter.SHOULD_NOT_SPLIT = 0;
  Fragmenter.NORMAL = 10;
  Fragmenter.ANY = 100;
  Fragmenter.ALWAYS_ON_TOP = Number.MAX_VALUE;







































  function _extractEntries (annotations) {
    var openers = [];
    var closers = [];
    forEach(annotations, function (a) {
      var isAnchor = (a.isAnchor ? a.isAnchor() : false);
      
      if (isAnchor) {
        openers.push({
          mode: ANCHOR,
          pos: a.offset,
          id: a.id,
          level: Fragmenter.ALWAYS_ON_TOP,
          type: 'anchor',
          node: a,
          counter: -1,
          length: 0
        });
      } else {
        
        
        
        
        
        
        

        
        var l = Fragmenter.NORMAL;
        var isInline = (a.isInline ? a.isInline() : false);
        if (isInline) {
          l = Number.MAX_VALUE;
        } else if (a.constructor.hasOwnProperty('fragmentation')) {
          l = a.constructor.fragmentation;
        } else if (a.hasOwnProperty('fragmentationHint')) {
          l = a.fragmentationHint;
        }
        var startOffset = Math.min(a.start.offset, a.end.offset);
        var endOffset = Math.max(a.start.offset, a.end.offset);
        var opener = {
          pos: startOffset,
          mode: ENTER,
          level: l,
          id: a.id,
          type: a.type,
          node: a,
          length: 0,
          counter: -1
        };
        openers.push(opener);
        closers.push({
          pos: endOffset,
          mode: EXIT,
          level: l,
          id: a.id,
          type: a.type,
          node: a,
          opener: opener
        });
      }
    });

    
    openers.sort(_compareOpeners);
    
    for (var i = openers.length - 1; i >= 0; i--) {
      openers[i].idx = i;
    }
    closers.sort(_compareClosers);
    
    var entries = new Array(openers.length + closers.length);
    var idx = 0;
    var idx1 = 0;
    var idx2 = 0;
    var opener = openers[idx1];
    var closer = closers[idx2];
    while (opener || closer) {
      if (opener && closer) {
        
        if (closer.pos <= opener.pos && closer.opener !== opener) {
          entries[idx] = closer;
          idx2++;
        } else {
          entries[idx] = opener;
          idx1++;
        }
      } else if (opener) {
        entries[idx] = opener;
        idx1++;
      } else if (closer) {
        entries[idx] = closer;
        idx2++;
      }
      opener = openers[idx1];
      closer = closers[idx2];
      idx++;
    }
    return entries
  }

  function _compareOpeners (a, b) {
    if (a.pos < b.pos) return -1
    if (a.pos > b.pos) return 1
    if (a.mode < b.mode) return -1
    if (a.mode > b.mode) return 1
    if (a.mode === b.mode) {
      if (a.level < b.level) return -1
      if (a.level > b.level) return 1
    }
    return 0
  }


  function _compareClosers (a, b) {
    if (a.pos < b.pos) return -1
    if (a.pos > b.pos) return 1
    
    
    
    
    
    if (a.pos === a.opener.pos && b.pos === b.opener.pos) {
      if (a.opener.idx < b.opener.idx) {
        return -1
      } else {
        return 1
      }
    }
    if (a.opener.idx > b.opener.idx) return -1
    if (a.opener.idx < b.opener.idx) return 1
    return 0
  }

  function appendChild (xmlNode, child) {
    insertAt(xmlNode, xmlNode._childNodes.length, child);
    return xmlNode
  }

  function getChildPos (xmlNode, child) {
    return xmlNode._childNodes.indexOf(child.id)
  }

  function removeChild (xmlNode, child) {
    const childPos = getChildPos(xmlNode, child);
    if (childPos >= 0) {
      removeAt(xmlNode, childPos);
    } else {
      throw new Error(`Node ${child.id} is not a child of ${xmlNode.id}`)
    }
    return xmlNode
  }

  function insertBefore (xmlNode, newChild, ref) {
    if (!ref) {
      appendChild(xmlNode, newChild);
    } else {
      let pos = getChildPos(xmlNode, ref);
      if (pos < 0) {
        throw new Error(`Node ${newChild.id} is not a child of ${xmlNode.id}`)
      }
      insertAt(xmlNode, pos, newChild);
    }
    return xmlNode
  }

  function insertAt (xmlNode, pos, child) {
    const length = xmlNode._childNodes.length;
    if (pos >= 0 && pos <= length) {
      const doc = xmlNode.getDocument();
      doc.update([xmlNode.id, '_childNodes'], { type: 'insert', pos, value: child.id });
    } else {
      throw new Error('Index out of bounds.')
    }
    return xmlNode
  }

  function removeAt (xmlNode, pos) {
    const length = xmlNode._childNodes.length;
    if (pos >= 0 && pos < length) {
      const doc = xmlNode.getDocument();
      doc.update([xmlNode.id, '_childNodes'], { type: 'delete', pos: pos });
    } else {
      throw new Error('Index out of bounds.')
    }
    return xmlNode
  }

  function getInnerXML (xmlNode) {
    if (xmlNode._childNodes) {
      return xmlNode.getChildNodes().map(child => {
        return child.toXML().serialize()
      }).join('')
    } else if (xmlNode.isText()) {
      return xmlNode.toXML().getInnerXML()
    }
    return ''
  }

  function getChildAt (xmlNode, idx) {
    let childId = xmlNode._childNodes[idx];
    if (childId) {
      return xmlNode.getDocument().get(childId)
    }
  }


  function node2element (node) {
    
    let dom = DefaultDOMElement.createDocument('xml');
    let el = _node2element(dom, node);
    return el
  }

  function _node2element (dom, node) {
    let el;
    switch (node._elementType) {
      case 'text': {
        el = _renderTextNode(dom, node);
        break
      }
      case 'element':
      case 'inline-element':
      case 'container': {
        el = _renderElementNode(dom, node);
        break
      }
      case 'anchor':
      case 'annotation': {
        el = _createElement$1(dom, node);
        break
      }
      case 'external': {
        el = DefaultDOMElement.parseSnippet(node.xml, 'xml');
        break
      }
      default:
        throw new Error('Invalid element type.')
    }
    return el
  }

  function _createElement$1 (dom, node) {
    let el = dom.createElement(node.type);
    el.attr('id', node.id);
    el.attr(node.attributes);
    return el
  }

  function _renderElementNode (dom, node) {
    let el = _createElement$1(dom, node);
    el.append(node.getChildren().map(child => _node2element(dom, child)));
    return el
  }

  function _renderTextNode (dom, node) {
    const annos = node.getAnnotations();
    const text = node.getText();
    let el = _createElement$1(dom, node);
    if (annos && annos.length > 0) {
      let fragmenter = new Fragmenter({
        onText: (context, text) => {
          const node = context.node;
          if (node.isText() || (node.isAnnotation() && !node._isInlineNode)) {
            context.el.append(text);
          }
        },
        onEnter: (fragment) => {
          return {
            el: _node2element(dom, fragment.node),
            node: fragment.node
          }
        },
        onExit: (fragment, context, parentContext) => {
          parentContext.el.append(context.el);
        }
      });
      fragmenter.start({ el, node }, text, annos);
    } else {
      el.append(text);
    }
    return el
  }

  function nameWithoutNS$1 (name) {
    const idx = name.indexOf(':');
    if (idx > 0) {
      return name.slice(idx + 1)
    } else {
      return name
    }
  }

  class Adapter extends domUtils.DomUtils {
    
    isTag () {
      return true
    }

    getChildren (elem) {
      return elem.getChildren() || []
    }

    getAttributeValue (elem, name) {
      return elem.getAttribute(name)
    }

    getAttributes (elem) {
      return ['id', elem.id].concat(map(elem.attributes, (val, key) => { return [key, val] }))
    }

    hasAttrib (elem, name) {
      return name === 'id' || elem.attributes.hasOwnProperty(name)
    }

    getName (elem) {
      return elem.type
    }

    getNameWithoutNS (elem) {
      return nameWithoutNS$1(this.getName(elem))
    }

    getText (elem) {
      if (elem._elementType === 'text') {
        return elem.getText()
      }
      
      return ''
    }
  }

  var cssSelectAdapter = new Adapter();

  class XMLDocumentNode extends DocumentNode {
    _initialize (doc, props) {
      
      
      
      if (props.attributes) {
        delete props.attributes.id;
      }
      super._initialize(doc, props);
    }

    toXML () {
      return node2element(this)
    }

    
    findChild (tagName) {
      const children = this.getChildren();
      for (let i = 0; i < children.length; i++) {
        const child = children[i];
        if (child.type === tagName) return child
      }
    }

    find (cssSelector) {
      return cssSelect.selectOne(cssSelector, this, { xmlMode: true, adapter: cssSelectAdapter })
    }

    findAll (cssSelector) {
      return cssSelect.selectAll(cssSelector, this, { xmlMode: true, adapter: cssSelectAdapter })
    }

    isContainer () {
      return false
    }

    
    isBlock () {
      const parentNode = this.parentNode;
      return (parentNode && parentNode.isContainer())
    }

    get children () {
      return this.getChildren()
    }

    getChildren () {
      
      return this.getChildNodes()
    }

    getChildNodes () {
      if (this._childNodes) {
        return getNodes(this.getDocument(), this._childNodes)
      } else {
        return []
      }
    }

    getChildCount () {
      if (this._childNodes) {
        return this._childNodes.length
      } else {
        return 0
      }
    }

    getChildPosition (child) {
      if (!child || !this._childNodes) return -1
      return this._childNodes.indexOf(child.id)
    }

    getChildNodeIterator () {
      return new ArrayIterator(this.getChildNodes())
    }

    getFirstChild () {
      if (this._childNodes) {
        return this.getDocument().get(this._childNodes[0])
      }
    }

    getLastChild () {
      if (this._childNodes) {
        return this.getDocument().get(last$1(this._childNodes))
      }
    }

    getParent () {
      return this.parentNode
    }

    setParent (parentNode) {
      this.parentNode = parentNode;
    }

    get tagName () {
      return this.type
    }

    
    get parent () {
      return this.parentNode
    }

    setAttribute (name, val) {
      if (name === 'id') {
        throw new Error("'id' is read-only and can not be changed")
      }
      this.getDocument().set([this.id, 'attributes', name], val);
      return this
    }

    getAttribute (name) {
      if (name === 'id') return this.id
      return this.attributes[name]
    }

    removeAttribute (name) {
      if (this.attributes.hasOwnProperty(name)) {
        this.getDocument().set([this.id, 'attributes', name], undefined);
        delete this.attributes[name];
      }
      return this
    }

    getAttributes () {
      return clone(this.attributes)
    }

    getElementSchema () {
      return this.getDocument().getElementSchema(this.type)
    }

    serialize () {
      return this.toXML()
    }

    isTextNode () {
      return false
    }

    isElementNode () {
      return false
    }

    getInnerXML () {
      return getInnerXML(this)
    }
  }

  XMLDocumentNode.prototype.attr = DOMElement.prototype.attr;

  XMLDocumentNode.schema = {
    attributes: { type: 'object', default: {} }
  };

  class XMLTextElement extends TextNodeMixin(XMLDocumentNode) {
    getPath () {
      return [this.id, 'content']
    }

    getText () {
      return this.content
    }

    
    getChildren () {
      const annos = this.getAnnotations();
      
      annos.sort(_byStartOffset);
      return annos
    }

    setText (text) {
      const doc = this.getDocument();
      const path = this.getPath();
      const oldText = this.getText();
      
      if (oldText.length > 0) {
        doc.update(path, { type: 'delete', start: 0, end: oldText.length });
      }
      doc.update(path, { type: 'insert', start: 0, text });
      return this
    }

    

    getTextContent () {
      return this.getText()
    }

    setTextContent (text) {
      return this.setText(text)
    }

    get textContent () {
      return this.getText()
    }

    set textContent (text) {
      this.setText(text);
    }

    appendChild(child) { 
      
      
      
      
      
      
      
      throw new Error('This is not implemented yet.')
    }

    removeChild(child) { 
      
      throw new Error('This is not implemented yet.')
    }

    
    isElementNode () {
      return true
    }
  }

  XMLTextElement.prototype._isXMLTextElement = true;

  XMLTextElement.prototype.text = DOMElement.prototype.text;

  XMLTextElement.prototype._elementType = 'text';

  XMLTextElement.isText = true;
  XMLTextElement.isBlock = true;

  XMLTextElement.type = 'text';

  XMLTextElement.schema = {
    content: 'text'
  };

  function _byStartOffset (a, b) {
    return a.start.offset - b.start.offset
  }

  class XMLElementNode extends XMLDocumentNode {
    appendChild (child) {
      return appendChild(this, child)
    }

    removeChild (child) {
      return removeChild(this, child)
    }

    insertBefore (newChild, ref) {
      return insertBefore(this, newChild, ref)
    }

    insertAt (pos, child) {
      return insertAt(this, pos, child)
    }

    removeAt (pos) {
      return removeAt(this, pos)
    }

    getChildAt (idx) {
      return getChildAt(this, idx)
    }

    isElementNode () {
      return true
    }

    
  }

  XMLElementNode.prototype.append = DOMElement.prototype.append;

  XMLElementNode.prototype._elementType = 'element';

  XMLElementNode.type = 'element';

  XMLElementNode.schema = {
    _childNodes: { type: ['array', 'id'], default: [], owned: true }
  };

  XMLElementNode.isBlock = true;

  class XMLAnnotationNode extends AnnotationMixin(XMLDocumentNode) {
    
    get parentNode () {
      const path = this.start.path;
      const doc = this.getDocument();
      return doc.get(path[0])
    }

    isPropertyAnnotation () {
      return true
    }
  }

  XMLAnnotationNode.prototype._elementType = 'annotation';


  XMLAnnotationNode.prototype._isPropertyAnnotation = true;

  XMLAnnotationNode.type = '@annotation';


  XMLAnnotationNode.schema = {};

  class XMLAnchorNode extends XMLDocumentNode {
    
    get parentNode () {
      const path = this.coor.start.path;
      const doc = this.getDocument();
      return doc.get(path[0])
    }
  }

  XMLAnchorNode.prototype._elementType = 'anchor';

  XMLAnchorNode.type = 'anchor';

  XMLAnchorNode.schema = {
    coor: { type: 'coordinate', optional: true }
  };

  class XMLInlineElementNode extends XMLAnnotationNode {
    
    get parentNode () {
      const path = this.start.path;
      if (path[0]) {
        const doc = this.getDocument();
        return doc.get(path[0])
      }
      return this._parentNode
    }

    set parentNode (parent) {
      const path = this.start.path;
      if (path[0]) {
        throw new Error('parent of inline-element is implicitly given')
      }
      this._parentNode = parent;
    }

    getChildAt (idx) {
      return getChildAt(this, idx)
    }

    appendChild (child) {
      return appendChild(this, child)
    }

    removeChild (child) {
      return removeChild(this, child)
    }

    insertBefore (newChild, ref) {
      return insertBefore(this, newChild, ref)
    }

    insertAt (pos, child) {
      return insertAt(this, pos, child)
    }

    removeAt (pos) {
      return removeAt(this, pos)
    }

    getInnerXML () {
      return this.getChildren().map(child => {
        return child.toXML().outerHTML
      }).join('')
    }
  }

  XMLInlineElementNode.prototype._elementType = 'inline-element';



  XMLInlineElementNode.prototype._isInlineNode = true;
  XMLInlineElementNode.isInline = true;

  XMLInlineElementNode.type = 'inline-element';

  XMLInlineElementNode.schema = {
    _childNodes: { type: ['array', 'id'], default: [], owned: true }
  };

  class XMLExternalNode extends XMLDocumentNode {}

  XMLExternalNode.prototype._elementType = 'external';

  XMLExternalNode.type = 'external';

  XMLExternalNode.schema = {
    xml: { type: 'string', default: '' }
  };

  XMLExternalNode.isBlock = true;

  class XMLContainerNode extends ContainerMixin(XMLElementNode) {
    getContentPath () {
      return [this.id, '_childNodes']
    }

    getContent () {
      return this._childNodes
    }

    isContainer () {
      return true
    }

    appendChild (child) {
      super.show(child.id);
    }
  }

  XMLContainerNode.prototype._elementType = 'container';

  XMLContainerNode.type = 'container';

  XMLContainerNode.schema = {};

  XMLContainerNode.isBlock = true;

  function createCountingIdGenerator () {
    var counters = {};
    return function uuid (prefix) {
      if (!counters.hasOwnProperty(prefix)) {
        counters[prefix] = 1;
      }
      var result = [prefix, '-', counters[prefix]++].join('');
      return result
    }
  }

  const WS_LEFT = /^\s+/g;


  const WS_RIGHT = /\s+$/g;
  const WS_ALL = /\s+/g;


  const SPACE = ' ';
  const TABS_OR_NL = /[\t\n\r]+/g;

  const INVISIBLE_CHARACTER = '\u200B';


  class DOMImporter {
    constructor (config, context) {
      this.context = context || {};

      if (!config.schema) {
        throw new Error('"config.schema" is mandatory')
      }
      if (!config.converters) {
        throw new Error('"config.converters" is mandatory')
      }

      this.config = Object.assign({ idAttribute: 'id' }, config);
      this.schema = config.schema;
      this.converters = config.converters;
      this.state = null;

      this._defaultBlockConverter = null;
      this._allConverters = [];
      this._blockConverters = [];
      this._propertyAnnotationConverters = [];

      this.state = new DOMImporter.State();

      this._initialize();
    }

    
    _initialize () {
      const schema = this.schema;
      const defaultTextType = schema.getDefaultTextType();
      const converters = this.converters;
      for (let i = 0; i < converters.length; i++) {
        let converter;
        if (typeof converters[i] === 'function') {
          const Converter = converters[i];
          converter = new Converter();
        } else {
          converter = converters[i];
        }
        if (!converter.type) {
          throw new Error('Converter must provide the type of the associated node.')
        }
        if (!converter.matchElement && !converter.tagName) {
          throw new Error('Converter must provide a matchElement function or a tagName property.')
        }
        if (!converter.matchElement) {
          converter.matchElement = this._defaultElementMatcher.bind(converter);
        }
        const NodeClass = schema.getNodeClass(converter.type);
        if (!NodeClass) {
          throw new Error('No node type defined for converter')
        }
        if (!this._defaultBlockConverter && defaultTextType === converter.type) {
          this._defaultBlockConverter = converter;
        }
        this._allConverters.push(converter);
        
        if (NodeClass.prototype._isPropertyAnnotation) {
          this._propertyAnnotationConverters.push(converter);
        } else {
          this._blockConverters.push(converter);
        }
      }
    }

    dispose () {
      if (this.state.doc) {
        this.state.doc.dispose();
      }
    }

    
    reset () {
      if (this.state.doc) {
        this.state.doc.dispose();
      }
      this.state.reset();
      this.state.doc = this._createDocument();
    }

    getDocument () {
      return this.state.doc
    }

    
    convertContainer (elements, containerId) {
      if (!this.state.doc) this.reset();
      const state = this.state;
      const iterator = new ArrayIterator(elements);
      const nodeIds = [];
      while (iterator.hasNext()) {
        const el = iterator.next();
        let node;
        const blockTypeConverter = this._getConverterForElement(el, 'block');
        if (blockTypeConverter) {
          state.pushContext(el.tagName, blockTypeConverter);
          let nodeData = this._createNodeData(el, blockTypeConverter.type);
          nodeData = blockTypeConverter.import(el, nodeData, this) || nodeData;
          node = this._createNode(nodeData);
          let context = state.popContext();
          context.annos.forEach((a) => {
            this._createNode(a);
          });
        } else if (el.isCommentNode()) {
          continue
        } else {
          
          if (el.isTextNode() && /^\s*$/.exec(el.textContent)) continue
          
          
          iterator.back();
          node = this._wrapInlineElementsIntoBlockElement(iterator);
        }
        if (node) {
          nodeIds.push(node.id);
        }
      }
      return this._createNode({
        type: '@container',
        id: containerId,
        nodes: nodeIds
      })
    }

    
    convertElement (el) {
      if (!this.state.doc) this.reset();
      let isTopLevel = !this.state.isConverting;
      if (isTopLevel) {
        this.state.isConverting = true;
      }

      let nodeData, annos;
      const converter = this._getConverterForElement(el);
      if (converter) {
        const NodeClass = this.schema.getNodeClass(converter.type);
        nodeData = this._createNodeData(el, converter.type);
        this.state.pushContext(el.tagName, converter);
        
        
        
        
        
        if (NodeClass.isInline) {
          nodeData = this._convertInlineNode(el, nodeData, converter);
        } else if (NodeClass.prototype._isPropertyAnnotation) {
          nodeData = this._convertPropertyAnnotation(el, nodeData);
        } else {
          nodeData = converter.import(el, nodeData, this) || nodeData;
        }
        let context = this.state.popContext();
        annos = context.annos;
      } else {
        throw new Error('No converter found for ' + el.tagName)
      }
      
      const node = this._createNode(nodeData);
      
      annos.forEach((a) => {
        this._createNode(a);
      });

      
      
      if (this.config['stand-alone'] && isTopLevel) {
        this.state.isConverting = false;
        this.reset();
      }
      return node
    }

    
    annotatedText (el, path, options = {}) {
      if (!path) {
        throw new Error('path is mandatory')
      }
      const state = this.state;
      const context = last$1(state.contexts);
      
      
      if (!context) {
        throw new Error('This should be called from within an element converter.')
      }
      
      const oldPreserveWhitespace = state.preserveWhitespace;
      if (options.preserveWhitespace) {
        state.preserveWhitespace = true;
      }
      state.stack.push({ path: path, offset: 0, text: '', annos: [] });
      
      
      
      this.state.lastChar = '';
      const iterator = this.getChildNodeIterator(el);
      const text = this._annotatedText(iterator);
      
      
      const top = state.stack.pop();
      context.annos = context.annos.concat(top.annos);

      
      state.preserveWhitespace = oldPreserveWhitespace;

      return text
    }

    
    plainText (el) {
      var state = this.state;
      var text = el.textContent;
      if (state.stack.length > 0) {
        var context = last$1(state.stack);
        context.offset += text.length;
        context.text += context.text.concat(text);
      }
      return text
    }

    
    _customText (text) {
      var state = this.state;
      if (state.stack.length > 0) {
        var context = last$1(state.stack);
        context.offset += text.length;
        context.text += context.text.concat(text);
      }
      return text
    }

    
    nextId (prefix) {
      
      
      
      
      return this.state.uuid(prefix)
    }

    _getNextId (dom, type) {
      let id = this.nextId(type);
      while (this.state.ids[id] || dom.find('#' + id)) {
        id = this.nextId(type);
      }
      return id
    }

    _getIdForElement (el, type) {
      let id = el.getAttribute(this.config.idAttribute);
      if (id && !this.state.ids[id]) return id
      return this._getNextId(el.getOwnerDocument(), type)
    }

    
    
    _createDocument () {
      
      const schema = this.config.schema;
      const DocumentClass = schema.getDocumentClass();
      return new DocumentClass(schema)
    }

    _convertPropertyAnnotation (el, nodeData) {
      const path = [nodeData.id, '_content'];
      
      
      
      nodeData._content = this.annotatedText(el, path);
      nodeData.start = { path, offset: 0 };
      nodeData.end = { offset: nodeData._content.length };
      return nodeData
    }

    _convertInlineNode (el, nodeData, converter) {
      const path = [nodeData.id, '_content'];
      if (converter.import) {
        nodeData = converter.import(el, nodeData, this) || nodeData;
      }
      nodeData._content = '$';
      nodeData.start = { path, offset: 0 };
      nodeData.end = { offset: 1 };
      return nodeData
    }

    _createNodeData (el, type) {
      if (!type) {
        throw new Error('type is mandatory.')
      }
      let nodeData = {
        type,
        id: this._getIdForElement(el, type)
      };
      this.state.ids[nodeData.id] = true;
      return nodeData
    }

    _createNode (nodeData) {
      let doc = this.state.doc;
      
      
      
      let node = doc.get(nodeData.id);
      if (node) {
        
        doc.delete(node.id);
      }
      return doc.create(nodeData)
    }

    getChildNodeIterator (el) {
      return el.getChildNodeIterator()
    }

    _defaultElementMatcher (el) {
      return el.is(this.tagName)
    }

    
    _annotatedText (iterator) {
      const state = this.state;
      const context = last$1(state.stack);
      
      if (!context) {
        throw new Error('Illegal state: context is null.')
      }
      while (iterator.hasNext()) {
        var el = iterator.next();
        var text = '';
        
        
        if (el.isTextNode()) {
          text = this._prepareText(el.textContent);
          if (text.length) {
            
            
            context.text = context.text.concat(text);
            context.offset += text.length;
          }
        } else if (el.isCommentNode()) {
          
          continue
        } else if (el.isElementNode()) {
          const annoConverter = this._getConverterForElement(el, 'inline');
          
          if (!annoConverter) {
            
            if (!this.IGNORE_DEFAULT_WARNINGS) {
              console.warn('Unsupported inline element. We will not create an annotation for it, but process its children to extract annotated text.', el.outerHTML);
            }
            
            
            const iterator = this.getChildNodeIterator(el);
            this._annotatedText(iterator);
            continue
          }
          
          
          
          var startOffset = context.offset;
          const annoType = annoConverter.type;
          const AnnoClass = this.schema.getNodeClass(annoType);
          let annoData = this._createNodeData(el, annoType);
          
          let stackFrame = {
            path: context.path,
            offset: startOffset,
            text: '',
            annos: []
          };
          state.stack.push(stackFrame);
          
          if (annoConverter.import) {
            state.pushContext(el.tagName, annoConverter);
            annoData = annoConverter.import(el, annoData, this) || annoData;
            state.popContext();
          }
          
          
          
          
          if (AnnoClass.isInline) {
            this._customText(INVISIBLE_CHARACTER);
            
            
            
            state.lastChar = '';
          } else {
            
            
            
            const iterator = this.getChildNodeIterator(el);
            this._annotatedText(iterator);
          }
          
          state.stack.pop();
          context.offset = stackFrame.offset;
          context.text = context.text.concat(stackFrame.text);
          
          const endOffset = context.offset;
          annoData.start = {
            path: context.path.slice(0),
            offset: startOffset
          };
          annoData.end = {
            offset: endOffset
          };
          
          let parentFrame = last$1(state.stack);
          parentFrame.annos = parentFrame.annos.concat(stackFrame.annos, annoData);
        } else {
          console.warn('Unknown element type. Taking plain text.', el.outerHTML);
          text = this._prepareText(el.textContent);
          context.text = context.text.concat(text);
          context.offset += text.length;
        }
      }
      
      return context.text
    }

    _getConverterForElement (el, mode) {
      var converters;
      if (mode === 'block') {
        if (!el.tagName) return null
        converters = this._blockConverters;
      } else if (mode === 'inline') {
        converters = this._propertyAnnotationConverters;
      } else {
        converters = this._allConverters;
      }
      var converter = null;
      for (var i = 0; i < converters.length; i++) {
        if (this._converterCanBeApplied(converters[i], el)) {
          converter = converters[i];
          break
        }
      }
      return converter
    }

    _converterCanBeApplied (converter, el) {
      return converter.matchElement(el, this)
    }

    
    _wrapInlineElementsIntoBlockElement (childIterator) {
      if (!childIterator.hasNext()) return

      const converter = this._defaultBlockConverter;
      if (!converter) {
        throw new Error('Wrapping inline elements automatically is not supported in this schema.')
      }

      let dom = childIterator.peek().getOwnerDocument();
      let wrapper = dom.createElement('wrapper');
      while (childIterator.hasNext()) {
        const el = childIterator.next();
        
        const blockTypeConverter = this._getConverterForElement(el, 'block');
        if (blockTypeConverter) {
          childIterator.back();
          break
        }
        wrapper.append(el.clone());
      }
      const type = this.schema.getDefaultTextType();
      const id = this._getNextId(dom, type);
      let nodeData = { type, id };
      this.state.pushContext('wrapper', converter);
      nodeData = converter.import(wrapper, nodeData, this) || nodeData;
      let context = this.state.popContext();
      let annos = context.annos;
      
      const node = this._createNode(nodeData);
      
      annos.forEach((a) => {
        this._createNode(a);
      });
      return node
    }

    
    
    
    
    _prepareText (text) {
      const state = this.state;
      if (state.preserveWhitespace) {
        return text
      }
      var repl = SPACE;
      
      text = text.replace(TABS_OR_NL, '');
      
      
      
      
      
      if (state.lastChar === SPACE) {
        
        text = text.replace(WS_LEFT, '');
      } else {
        text = text.replace(WS_LEFT, repl);
      }
      text = text.replace(WS_RIGHT, repl);
      
      
      
      if (this.config.REMOVE_INNER_WS || state.removeInnerWhitespace) {
        text = text.replace(WS_ALL, SPACE);
      }
      state.lastChar = text[text.length - 1] || state.lastChar;
      return text
    }

    
    _trimTextContent (el) {
      var nodes = el.getChildNodes();
      var firstNode = nodes[0];
      var lastNode = last$1(nodes);
      var text, trimmed;
      
      if (firstNode && firstNode.isTextNode()) {
        text = firstNode.textContent;
        trimmed = this._trimLeft(text);
        firstNode.textContent = trimmed;
      }
      if (lastNode && lastNode.isTextNode()) {
        text = lastNode.textContent;
        trimmed = this._trimRight(text);
        lastNode.textContent = trimmed;
      }
      return el
    }

    _trimLeft (text) {
      return text.replace(WS_LEFT, '')
    }

    _trimRight (text) {
      return text.replace(WS_RIGHT, '')
    }
  }

  class DOMImporterState {
    constructor () {
      this.reset();
    }

    reset () {
      this.preserveWhitespace = false;
      this.nodes = [];
      this.annotations = [];
      this.containerId = null;
      this.container = [];
      this.ids = {};
      
      this.contexts = [];
      
      this.stack = [];
      this.lastChar = '';
      this.skipTypes = {};
      this.ignoreAnnotations = false;
      this.isConverting = false;

      
      
      this.uuid = createCountingIdGenerator();
    }

    pushContext (tagName, converter) {
      this.contexts.push({ tagName: tagName, converter: converter, annos: [] });
    }

    popContext () {
      return this.contexts.pop()
    }

    getCurrentContext () {
      return last$1(this.contexts)
    }
  }

  DOMImporter.State = DOMImporterState;

  DOMImporter.INVISIBLE_CHARACTER = INVISIBLE_CHARACTER;

  const START = 'START';
  const END = 'END';
  const EPSILON = 'EPSILON';
  const TEXT$1 = 'TEXT';

  class DFA {
    constructor (transitions) {
      if (!transitions || Object.keys(transitions).length === 0) {
        transitions = { START: { EPSILON: END } };
      }
      this.transitions = transitions;
    }

    consume (state, id) {
      const T = this.transitions;
      
      
      if (!T[state]) return -1
      let nextState = T[state][id];
      if (nextState !== undefined) {
        return nextState
      }
      while (T[state][EPSILON] !== undefined) {
        state = T[state][EPSILON];
        if (state === END) {
          return -1
        }
        nextState = T[state][id];
        if (nextState !== undefined) {
          return nextState
        }
      }
      return -1
    }

    canConsume (state, id) {
      let nextState = this.consume(state, id);
      return (nextState !== -1)
    }

    isFinished (state) {
      const T = this.transitions;
      if (state === 'END') return true
      
      if (!T[state]) return false
      while (T[state][EPSILON] !== undefined) {
        state = T[state][EPSILON];
        if (state === 'END') return true
      }
      return false
    }

    

    
    _tokensByPath () {
      const result = [];
      const transitions = this.transitions;
      if (!transitions) return []

      
      let first = {};
      forEach(transitions[START], (to, token) => {
        if (!first[to]) first[to] = [];
        first[to].push(token);
      });

      let visited = {START: true, END: true};
      forEach(first, (tokens, state) => {
        
        
        
        
        let _siblings = {};
        tokens.forEach((t) => {
          if (t !== EPSILON) {
            _siblings[t] = true;
          }
        });
        let stack = [state];
        while (stack.length > 0) {
          let from = stack.pop();
          if (state === END) continue
          visited[from] = true;
          let T = transitions[from];
          if (!T) throw new Error(`Internal Error: no transition from state ${from}`)
          let tokens = Object.keys(T);
          for (let i = 0; i < tokens.length; i++) {
            const token = tokens[i];
            const to = T[token];
            if (!visited[to]) stack.push(to);
            if (token !== EPSILON) {
              _siblings[token] = true;
            }
          }
        }
        let _siblingTokens = Object.keys(_siblings);
        if (_siblingTokens.length > 0) {
          result.push(_siblingTokens);
        }
      });
      return result
    }
  }

  DFA.START = START;
  DFA.END = END;
  DFA.EPSILON = EPSILON;
  DFA.TEXT = TEXT$1;

  function _isTextNodeEmpty (el) {
    return Boolean(/^\s*$/.exec(el.textContent))
  }

  const START$1 = DFA.START;
  const END$1 = DFA.END;
  const EPSILON$1 = DFA.EPSILON;


  class DFABuilder {
    constructor (transitions) {
      this.transitions = transitions;
    }

    addTransition (from, to, tokens) {
      if (!this.transitions) this.transitions = {};
      if (!isArray(tokens)) tokens = [tokens];
      tokens.forEach(token => _addTransition(this.transitions, from, to, token));
      return this
    }

    
    append (other) {
      if (this.transitions && other.transitions) {
        let t1 = cloneDeep(this.transitions);
        let t2 = cloneDeep(other.transitions);
        
        
        let firstIsOptional = Boolean(t1[START$1][EPSILON$1]);
        let secondIsOptional = Boolean(t2[START$1][EPSILON$1]);

        if (firstIsOptional) {
          
          
          delete t1[START$1][EPSILON$1];
        }
        
        
        let newState = uuid();
        
        
        forEach(t1, (T) => {
          forEach(T, (to, token) => {
            if (to === END$1) {
              T[token] = newState;
            }
          });
        });
        
        
        if (firstIsOptional) {
          forEach(t2[START$1], (to, token) => {
            _addTransition(t1, START$1, to, token);
          });
        }
        
        
        t2[newState] = t2[START$1];
        forEach(t2, (T) => {
          forEach(T, (to, token) => {
            if (to === START$1) {
              T[token] = newState;
            }
          });
        });
        delete t2[START$1];
        
        forEach(t2, (T, from) => {
          forEach(T, (to, token) => {
            _addTransition(t1, from, to, token);
          });
        });
        
        
        if (firstIsOptional && secondIsOptional) {
          _addTransition(t1, START$1, END$1, EPSILON$1);
        }
        this.transitions = t1;
      } else if (other.transitions) {
        this.transitions = cloneDeep(other.transitions);
      }
      return this
    }

    
    merge (other) {
      if (this.transitions && other.transitions) {
        let t1 = this.transitions;
        let t2 = other.transitions;
        forEach(t2, (T, from) => {
          forEach(T, (to, token) => {
            _addTransition(t1, from, to, token);
          });
        });
      } else if (other.transitions) {
        this.transitions = cloneDeep(other.transitions);
      }
      return this
    }

    
    optional () {
      let dfa = new DFABuilder(cloneDeep(this.transitions));
      if (this.transitions) {
        dfa.addTransition(START$1, END$1, EPSILON$1);
      }
      return dfa
    }

    
    kleene () {
      let dfa = this.plus();
      return dfa.optional()
    }

    
    plus () {
      let dfa;
      if (this.transitions) {
        let t1 = cloneDeep(this.transitions);
        
        
        
        
        const isOptional = Boolean(t1[START$1][EPSILON$1]);
        delete t1[START$1][EPSILON$1];
        
        
        let newState = uuid();
        forEach(t1, (T) => {
          forEach(T, (to, token) => {
            if (to === END$1) {
              T[token] = newState;
            }
          });
        });
        
        _addTransition(t1, newState, END$1, EPSILON$1);
        
        forEach(t1[START$1], (to, token) => {
          _addTransition(t1, newState, to, token);
        });
        
        if (isOptional) {
          _addTransition(t1, START$1, END$1, EPSILON$1);
        }
        dfa = new DFABuilder(t1);
      } else {
        dfa = new DFABuilder(cloneDeep(this.transitions));
      }
      return dfa
    }

    toJSON () {
      return cloneDeep(this.transitions)
    }

    
    copy () {
      let t = cloneDeep(this.transitions);
      if (this.transitions) {
        let states = Object.keys(t);
        let map = { START: START$1, END: END$1 };
        states.forEach((id) => {
          if (id === START$1 || id === END$1) return
          map[id] = uuid();
        });
        forEach(t, (T, from) => {
          if (from !== START$1 && from !== END$1) {
            t[map[from]] = T;
            delete t[from];
          }
          forEach(T, (to, token) => {
            if (to !== START$1 && to !== END$1) {
              T[token] = map[to];
            }
          });
        });
      }
      return new DFABuilder(t)
    }
  }

  DFABuilder.singleToken = function (token) {
    let dfa = new DFABuilder();
    dfa.addTransition(START$1, END$1, token);
    return dfa
  };

  function _addTransition (transitions, from, to, token) {
    let T = transitions[from];
    if (!T) {
      transitions[from] = T = {};
    }
    if (token === EPSILON$1 && from === START$1 && to !== END$1) {
      throw new Error('The only EPSILON transition from START must be START->END')
    }
    if (T[token] && T[token] !== to) {
      console.error('Token %s already used. Ignoring this transition.', token);
      return
      
    }
    T[token] = to;
  }

  const { START: START$2, END: END$2, TEXT: TEXT$3, EPSILON: EPSILON$2 } = DFA;



  class Expression {
    constructor (name, root) {
      this.name = name;
      this.root = root;

      this._initialize();
    }

    _initialize () {
      this._compile();
    }

    toString () {
      return this.root.toString()
    }

    copy () {
      return this.root.copy()
    }

    toJSON () {
      return {
        name: this.name,
        content: this.root.toJSON()
      }
    }

    isAllowed (tagName) {
      return Boolean(this._allowedChildren[tagName])
    }

    
    _normalize () {
      this.root._normalize();
    }

    
    _compile () {
      
      this.root._compile();
    }

    _describeError (state, token) {
      let msg = [];
      if (token !== TEXT$3) {
        if (!this.isAllowed(token)) {
          msg.push(`<${token}> is not valid in <${this.name}>\nSchema: ${this.toString()}`);
        } else {
          
          msg.push(`<${token}> is not allowed at the current position in <${this.name}>.\n${this.toString()}`);
          
          
          
        }
      } else {
        msg.push(`TEXT is not allowed at the current position: ${state.trace.join(',')}\n${this.toString()}`);
      }
      return msg.join('')
    }
  }

  Expression.fromJSON = function (data) {
    const name = data.name;
    const root = _fromJSON(data.content);
    return createExpression(name, root)
  };

  function createExpression (name, root) {
    if (root instanceof Interleave) {
      return new InterleaveExpr(name, root)
    } else {
      return new DFAExpr(name, root)
    }
  }

  class DFAExpr extends Expression {
    
    getInitialState () {
      return {
        dfaState: START$2,
        errors: [],
        trace: []
      }
    }

    canConsume (state, token) {
      return this.dfa.canConsume(state.dfaState, token)
    }

    consume (state, token) {
      const dfa = this.dfa;
      let oldState = state.dfaState;
      let newState = dfa.consume(oldState, token);
      state.dfaState = newState;
      if (newState === -1) {
        state.errors.push({
          msg: this._describeError(state, token),
          
          
          el: state.el
        });
        return false
      } else {
        state.trace.push(token);
        return true
      }
    }

    isFinished (state) {
      return this.dfa.isFinished(state.dfaState)
    }

    _initialize () {
      super._initialize();

      this._computeAllowedChildren();
    }

    _compile () {
      super._compile();
      this.dfa = new DFA(this.root.dfa.transitions);
    }

    _computeAllowedChildren () {
      this._allowedChildren = _collectAllTokensFromDFA(this.dfa);
    }

    _findInsertPos (el, newTag, mode) {
      const root = this.root;
      if (root instanceof Sequence) {
        return this._findInsertPosInSequence(el, newTag, mode)
      } else if (root instanceof Plus || root instanceof Kleene) {
        if (mode === 'first') {
          return 0
        } else {
          return el.childNodes.length
        }
      }
    }

    _isValid (_tokens) {
      let state = this.getInitialState();
      for (let i = 0; i < _tokens.length; i++) {
        const token = _tokens[i];
        
        
        
        if (!token) continue
        if (!this.consume(state, token)) {
          return false
        }
      }
      return this.isFinished(state)
    }

    _findInsertPosInSequence (el, newTag, mode) {
      const childNodes = el.getChildNodes();
      
      
      
      const tokens = [];
      childNodes.forEach((child) => {
        
        
        
        const tagName = child.tagName;
        if (!tagName) {
          if (child._isDOMElement && child.isTextNode() && !_isTextNodeEmpty(child)) {
            tokens.push(TEXT$3);
          } else {
            tokens.push(null);
          }
        } else {
          tokens.push(tagName);
        }
      });
      const L = tokens.length;
      const self = this;
      function _isValid (pos) {
        let _tokens = tokens.slice(0);
        _tokens.splice(pos, 0, newTag);
        return self._isValid(_tokens)
      }
      if (mode === 'first') {
        for (let pos = 0; pos <= L; pos++) {
          if (_isValid(pos)) {
            return pos
          }
        }
      } else {
        for (let pos = L; pos >= 0; pos--) {
          if (_isValid(pos)) {
            return pos
          }
        }
      }
      return -1
    }
  }

  function _collectAllTokensFromDFA (dfa) {
    
    const children = {};
    if (dfa.transitions) {
      forEach(dfa.transitions, (T) => {
        Object.keys(T).forEach((tagName) => {
          if (tagName === EPSILON$2) return
          children[tagName] = true;
        });
      });
    }
    return children
  }

  class InterleaveExpr extends Expression {
    getInitialState () {
      const dfas = this.dfas;
      const dfaStates = new Array(dfas.length);
      dfaStates.fill(START$2);
      return {
        dfaStates,
        errors: [],
        trace: [],
        
        lastDFA: 0
      }
    }

    canConsume (state, token) {
      return (this._findNextDFA(state, token) >= 0)
    }

    consume (state, token) {
      const idx = this._findNextDFA(state, token);
      if (idx < 0) {
        state.errors.push({
          msg: this._describeError(state, token)
        });
        return false
      } else {
        const dfa = this.dfas[idx];
        const oldState = state.dfaStates[idx];
        const newState = dfa.consume(oldState, token);
        state.dfaStates[idx] = newState;
        state.trace.push(token);
        return true
      }
    }

    isFinished (state) {
      const dfas = this.dfas;
      for (let i = 0; i < dfas.length; i++) {
        const dfa = dfas[i];
        const dfaState = state.dfaStates[i];
        if (!dfa.isFinished(dfaState)) {
          return false
        }
      }
      return true
    }

    _initialize () {
      super._initialize();

      this._computeAllowedChildren();
    }

    _compile () {
      super._compile();

      this.blocks = this.root.blocks;
      this.dfas = this.blocks.map(b => new DFA(b.dfa.transitions));
    }

    _computeAllowedChildren () {
      this._allowedChildren = Object.assign(...this.blocks.map((block) => {
        return _collectAllTokensFromDFA(block.dfa)
      }));
    }

    _findNextDFA (state, token) {
      console.assert(state.dfaStates.length === this.dfas.length);
      const dfas = this.dfas;
      for (let i = 0; i < state.dfaStates.length; i++) {
        const dfa = dfas[i];
        const dfaState = state.dfaStates[i];
        if (dfa.canConsume(dfaState, token)) {
          return i
        }
      }
      return -1
    }

    _findInsertPos(el, newTag, mode) { 
      
      return el.childNodes.length
    }
  }

  class Token {
    constructor (name) {
      this.name = name;
    }

    toString () {
      return this.name
    }

    toJSON () {
      return this.name
    }

    copy () {
      return new Token(this.name)
    }

    _normalize () {}

    _compile () {
      this.dfa = DFABuilder.singleToken(this.name);
    }
  }

  Token.fromJSON = function (data) {
    return new Token(data)
  };


  class Choice {
    constructor (blocks) {
      this.blocks = blocks;
    }

    copy () {
      return new Choice(this.blocks.map(b => b.copy()))
    }

    toJSON () {
      return {
        type: '|',
        blocks: this.blocks.map(b => b.toJSON())
      }
    }

    _normalize () {
      const blocks = this.blocks;
      for (let i = blocks.length - 1; i >= 0; i--) {
        let block = blocks[i];
        block._normalize();
        
        if (block instanceof Choice) {
          blocks.splice(i, 1, ...(block.blocks));
        }
      }
    }

    _compile () {
      let dfa = new DFABuilder();
      this.blocks.forEach((block) => {
        if (block instanceof Token) {
          dfa.addTransition(START$2, END$2, block.name);
        } else if (block instanceof Interleave) {
          throw new Error('Nested interleave blocks are not supported.')
        } else {
          if (!block.dfa) {
            block._compile();
          }
          dfa.merge(block.dfa);
        }
      });
      this.dfa = dfa;
      return dfa
    }

    toString () {
      return '(' + this.blocks.map(b => b.toString()).join('|') + ')'
    }
  }

  Choice.fromJSON = function (data) {
    return new Choice(data.blocks.map((block) => {
      return _fromJSON(block)
    }))
  };


  class Sequence {
    constructor (blocks) {
      this.blocks = blocks;
    }

    copy () {
      return new Sequence(this.blocks.map(b => b.copy()))
    }

    toJSON () {
      return {
        type: ',',
        blocks: this.blocks.map(b => b.toJSON())
      }
    }

    _compile () {
      let dfa = new DFABuilder();
      this.blocks.forEach((block) => {
        if (block instanceof Token) {
          dfa.append(DFABuilder.singleToken(block.name));
        } else if (block instanceof Interleave) {
          throw new Error('Nested interleave blocks are not supported.')
        } else {
          if (!block.dfa) {
            block._compile();
          }
          dfa.append(block.dfa);
        }
      });
      this.dfa = dfa;
      return dfa
    }

    _normalize () {
      const blocks = this.blocks;
      for (let i = blocks.length - 1; i >= 0; i--) {
        let block = blocks[i];
        block._normalize();
        
        if (block instanceof Sequence) {
          blocks.splice(i, 1, ...(block.blocks));
        }
      }
    }

    toString () {
      return '(' + this.blocks.map(b => b.toString()).join(',') + ')'
    }
  }

  Sequence.fromJSON = function (data) {
    return new Sequence(data.blocks.map((block) => {
      return _fromJSON(block)
    }))
  };


  class Interleave {
    constructor (blocks) {
      this.blocks = blocks;
    }

    copy () {
      return new Interleave(this.blocks.map(b => b.copy()))
    }

    toString () {
      return '(' + this.blocks.map(b => b.toString()).join('~') + ')'
    }

    toJSON () {
      return {
        type: '~',
        blocks: this.blocks.map(b => b.toJSON())
      }
    }

    _normalize () {
      
    }

    _compile () {
      this.blocks.forEach(block => block._compile());
    }
  }

  Interleave.fromJSON = function (data) {
    return new Interleave(data.blocks.map((block) => {
      return _fromJSON(block)
    }))
  };


  class Optional {
    constructor (block) {
      this.block = block;
    }

    copy () {
      return new Optional(this.block.copy())
    }

    toJSON () {
      return {
        type: '?',
        block: this.block.toJSON()
      }
    }

    _compile () {
      const block = this.block;
      if (block instanceof Interleave) {
        throw new Error('Nested interleave blocks are not supported.')
      }
      if (!block.dfa) {
        block._compile();
      }
      this.dfa = block.dfa.optional();
      return this.dfa
    }

    _normalize () {
      const block = this.block;
      block._normalize();
      if (block instanceof Optional) {
        this.block = block.block;
      } else if (block instanceof Kleene) {
        console.error('FIXME -  <optional> is useless here', this.toString());
      }
    }

    toString () {
      return this.block.toString() + '?'
    }
  }

  Optional.fromJSON = function (data) {
    return new Optional(_fromJSON(data.block))
  };


  class Kleene {
    constructor (block) {
      this.block = block;
    }

    copy () {
      return new Kleene(this.block.copy())
    }

    toJSON () {
      return {
        type: '*',
        block: this.block.toJSON()
      }
    }

    _compile () {
      const block = this.block;
      if (block instanceof Interleave) {
        throw new Error('Nested interleave blocks are not supported.')
      }
      if (!block.dfa) {
        block._compile();
      }
      this.dfa = block.dfa.kleene();
      return this.dfa
    }

    _normalize () {
      const block = this.block;
      block._normalize();
      if (block instanceof Optional || block instanceof Kleene) {
        this.block = block.block;
      } else if (block instanceof Plus) {
        throw new Error('This does not make sense:' + this.toString())
      }
    }

    toString () {
      return this.block.toString() + '*'
    }
  }

  Kleene.fromJSON = function (data) {
    return new Kleene(_fromJSON(data.block))
  };


  class Plus {
    constructor (block) {
      this.block = block;
    }

    copy () {
      return new Plus(this.block.copy())
    }

    toJSON () {
      return {
        type: '+',
        block: this.block.toJSON()
      }
    }

    _compile () {
      const block = this.block;
      if (block instanceof Interleave) {
        throw new Error('Nested interleave blocks are not supported.')
      }
      if (!block.dfa) {
        block._compile();
      }
      this.dfa = block.dfa.plus();
      return this.dfa
    }

    _normalize () {
      const block = this.block;
      block._normalize();
      if (block instanceof Optional || block instanceof Kleene) {
        throw new Error('This does not make sense:' + this.toString())
      } else if (block instanceof Plus) {
        this.block = block.block;
      }
    }

    toString () {
      return this.block.toString() + '+'
    }
  }

  Plus.fromJSON = function (data) {
    return new Plus(_fromJSON(data.block))
  };

  function _fromJSON (data) {
    switch (data.type) {
      case ',':
        return Sequence.fromJSON(data)
      case '~':
        return Interleave.fromJSON(data)
      case '|':
        return Choice.fromJSON(data)
      case '?':
        return Optional.fromJSON(data)
      case '+':
        return Plus.fromJSON(data)
      case '*':
        return Kleene.fromJSON(data)
      default:
        if (isString(data)) {
          return new Token(data)
        }
        throw new Error('Unsupported data.')
    }
  }

  const { TEXT: TEXT$4 } = DFA;

  class XMLSchema {
    constructor (elementSchemas, startElement) {
      if (!elementSchemas[startElement]) {
        throw new Error('startElement must be a valid element.')
      }
      this._elementSchemas = {};
      this.startElement = startElement;
      
      forEach(elementSchemas, (spec, name) => {
        this._elementSchemas[name] = new ElementSchema(spec.name, spec.type, spec.attributes, spec.expr);
      });
    }

    getIdAttribute () {
      return 'id'
    }

    getTagNames () {
      return Object.keys(this._elementSchemas)
    }

    getElementSchema (name) {
      return this._elementSchemas[name]
    }

    getStartElement () {
      return this.startElement
    }

    toJSON () {
      let result = {
        start: this.getStartElement(),
        elements: {}
      };
      forEach(this._elementSchemas, (schema, name) => {
        result.elements[name] = schema.toJSON();
      });
      return result
    }

    
    toMD () {
      let result = [];
      let elementNames = Object.keys(this._elementSchemas);
      elementNames.sort();
      elementNames.forEach((name) => {
        let elementSchema = this._elementSchemas[name];
        result.push(`# <${elementSchema.name}>`);
        result.push('');
        result.push(`type: ${elementSchema.type}`);
        result.push('attributes: ' + map(elementSchema.attributes, (_, name) => { return name }).join(', '));
        result.push('children:');
        result.push('  ' + elementSchema.expr.toString());
        result.push('');
      });
      return result.join('\n')
    }

    validateElement (el) {
      let tagName = el.tagName;
      let elementSchema = this.getElementSchema(tagName);
      return _validateElement(elementSchema, el)
    }
  }

  XMLSchema.fromJSON = function (data) {
    let elementSchemas = {};
    forEach(data.elements, (elData) => {
      let elSchema = ElementSchema.fromJSON(elData);
      elementSchemas[elSchema.name] = elSchema;
    });
    return new XMLSchema(elementSchemas, data.start)
  };

  class ElementSchema {
    constructor (name, type, attributes, expr) {
      this.name = name;
      this.type = type;
      this.attributes = attributes;
      this.expr = expr;

      if (!name) {
        throw new Error("'name' is mandatory")
      }
      if (!type) {
        throw new Error("'type' is mandatory")
      }
      if (!attributes) {
        throw new Error("'attributes' is mandatory")
      }
      if (!expr) {
        throw new Error("'expr' is mandatory")
      }
    }

    toJSON () {
      return {
        name: this.name,
        type: this.type,
        attributes: this.attributes,
        elements: this.expr.toJSON()
      }
    }

    isAllowed (tagName) {
      return this.expr.isAllowed(tagName)
    }

    isTextAllowed () {
      return this.expr.isAllowed(TEXT$4)
    }

    printStructure () {
      return `${this.name} ::= ${this.expr.toString()}`
    }

    findFirstValidPos (el, newTag) {
      return this.expr._findInsertPos(el, newTag, 'first')
    }

    findLastValidPos (el, newTag) {
      return this.expr._findInsertPos(el, newTag, 'last')
    }
  }

  ElementSchema.fromJSON = function (data) {
    return new ElementSchema(
      data.name,
      data.type,
      data.attributes,
      Expression.fromJSON(data.elements)
    )
  };

  function _validateElement (elementSchema, el) {
    let errors = [];
    let valid = true;
    { 
      const res = _checkAttributes(elementSchema, el);
      if (!res.ok) {
        errors = errors.concat(res.errors);
        valid = false;
      }
    }
    
    if (elementSchema.type === 'external') ; else {
      
      let res;
      if (el._isXMLTextElement) {
        res = _checkChildren(elementSchema, el.toXML());
      } else {
        res = _checkChildren(elementSchema, el);
      }
      if (!res.ok) {
        errors = errors.concat(res.errors);
        valid = false;
      }
    }
    return {
      errors,
      ok: valid
    }
  }

  function _checkAttributes(elementSchema, el) { 
    return { ok: true }
  }

  function _checkChildren (elementSchema, el) {
    
    
    if (elementSchema.type === 'external') {
      return true
    }
    const expr = elementSchema.expr;
    const state = expr.getInitialState();
    const iterator = el.getChildNodeIterator();
    let valid = true;
    while (valid && iterator.hasNext()) {
      const childEl = iterator.next();
      let token;
      if (childEl.isTextNode()) {
        
        if (elementSchema.type !== 'text' && _isTextNodeEmpty(childEl)) {
          continue
        }
        token = TEXT$4;
      } else if (childEl.isElementNode()) {
        token = childEl.tagName;
      } else {
        continue
      }
      if (!expr.consume(state, token)) {
        valid = false;
      }
    }
    
    if (state.errors.length > 0) {
      state.errors.forEach((err) => {
        err.el = el;
      });
    }
    if (valid && !expr.isFinished(state)) {
      state.errors.push({
        msg: `<${el.tagName}> is incomplete.\nSchema: ${expr.toString()}`,
        el
      });
      valid = false;
    }
    if (valid) {
      state.ok = true;
    }
    return state
  }

  var ManifestSchemaData = {"start":"dar","elements":{"dar":{"name":"dar","type":"element","attributes":{},"elements":{"name":"dar","content":{"type":",","blocks":["documents","assets"]}}},"documents":{"name":"documents","type":"element","attributes":{},"elements":{"name":"documents","content":{"type":"*","block":"document"}}},"assets":{"name":"assets","type":"element","attributes":{},"elements":{"name":"assets","content":{"type":"*","block":"asset"}}},"document":{"name":"document","type":"element","attributes":{"id":{"name":"id"},"name":{"name":"name"},"type":{"name":"type"},"path":{"name":"path"}},"elements":{"name":"document","content":{"type":",","blocks":[]}}},"asset":{"name":"asset","type":"element","attributes":{"id":{"name":"id"},"name":{"name":"name"},"type":{"name":"type"},"path":{"name":"path"}},"elements":{"name":"asset","content":{"type":",","blocks":[]}}}}};

  const ManifestSchema = XMLSchema.fromJSON(ManifestSchemaData);


  ManifestSchema.getName = function() {
    return 'RDC-Manifest'
  };

  ManifestSchema.getVersion = function() {
    return '1.0'
  };

  ManifestSchema.getDocTypeParams = function() {
    return ['manifest', 'RDC-Manifest 1.0', ManifestSchema.uri]
  };


  ManifestSchema.getDefaultTextType = function () {
    return 'text'
  };

  ManifestSchema.uri = '//Manifest-1.0.dtd';

  class Data extends EventEmitter {
    
    constructor (schema, nodeFactory) {
      super();

      
      if (!schema) {
        throw new Error('schema is mandatory')
      }
      if (!nodeFactory) {
        throw new Error('nodeFactory is mandatory')
      }
      

      this.schema = schema;
      this.nodeFactory = nodeFactory;
      this.nodes = {};
      this.indexes = {};

      
      
      this.__QUEUE_INDEXING__ = false;
      this.queue = [];
    }

    
    contains (id) {
      return Boolean(this.nodes[id])
    }

    
    get (path, strict) {
      let result = this._get(path);
      if (strict && result === undefined) {
        if (isString(path)) {
          throw new Error("Could not find node with id '" + path + "'.")
        } else if (!this.contains(path[0])) {
          throw new Error("Could not find node with id '" + path[0] + "'.")
        } else {
          throw new Error("Property for path '" + path + "' us undefined.")
        }
      }
      return result
    }

    _get (path) {
      if (!path) return undefined
      let result;
      if (isString(path)) {
        result = this.nodes[path];
      } else if (path.length === 1) {
        result = this.nodes[path[0]];
      } else if (path.length > 1) {
        let context = this.nodes[path[0]];
        for (let i = 1; i < path.length - 1; i++) {
          if (!context) return undefined
          context = context[path[i]];
        }
        if (!context) return undefined
        result = context[path[path.length - 1]];
      }
      return result
    }

    
    getNodes () {
      return this.nodes
    }

    
    create (nodeData) {
      var node = this.nodeFactory.create(nodeData.type, nodeData);
      if (!node) {
        throw new Error('Illegal argument: could not create node for data:', nodeData)
      }
      if (this.contains(node.id)) {
        throw new Error('Node already exists: ' + node.id)
      }
      if (!node.id || !node.type) {
        throw new Error('Node id and type are mandatory.')
      }
      this.nodes[node.id] = node;

      var change = {
        type: 'create',
        node: node
      };

      if (this.__QUEUE_INDEXING__) {
        this.queue.push(change);
      } else {
        this._updateIndexes(change);
      }

      return node
    }

    
    delete (nodeId) {
      var node = this.nodes[nodeId];
      if (!node) return
      node.dispose();
      delete this.nodes[nodeId];

      var change = {
        type: 'delete',
        node: node
      };

      if (this.__QUEUE_INDEXING__) {
        this.queue.push(change);
      } else {
        this._updateIndexes(change);
      }

      return node
    }

    
    set (path, newValue) {
      let node = this.get(path[0]);
      let oldValue = this._set(path, newValue);
      var change = {
        type: 'set',
        node: node,
        path: path,
        newValue: newValue,
        oldValue: oldValue
      };
      if (this.__QUEUE_INDEXING__) {
        this.queue.push(change);
      } else {
        this._updateIndexes(change);
      }
      return oldValue
    }

    _set (path, newValue) {
      let oldValue = _setValue(this.nodes, path, newValue);
      return oldValue
    }

    
    update (path, diff) {
      var realPath = this.getRealPath(path);
      if (!realPath) {
        console.error('Could not resolve path', path);
        return
      }
      let node = this.get(realPath[0]);
      let oldValue = this._get(realPath);
      let newValue;
      if (diff.isOperation) {
        newValue = diff.apply(oldValue);
      } else {
        diff = this._normalizeDiff(oldValue, diff);
        if (isString(oldValue)) {
          switch (diff.type) {
            case 'delete': {
              newValue = oldValue.split('').splice(diff.start, diff.end - diff.start).join('');
              break
            }
            case 'insert': {
              newValue = [oldValue.substring(0, diff.start), diff.text, oldValue.substring(diff.start)].join('');
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else if (isArray(oldValue)) {
          newValue = oldValue.slice(0);
          switch (diff.type) {
            case 'delete': {
              newValue.splice(diff.pos, 1);
              break
            }
            case 'insert': {
              newValue.splice(diff.pos, 0, diff.value);
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else if (oldValue._isCoordinate) {
          switch (diff.type) {
            case 'shift': {
              
              oldValue = { path: oldValue.path, offset: oldValue.offset };
              newValue = oldValue;
              newValue.offset += diff.value;
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else {
          throw new Error('Diff is not supported:', JSON.stringify(diff))
        }
      }
      this._set(realPath, newValue);

      var change = {
        type: 'update',
        node: node,
        path: realPath,
        newValue: newValue,
        oldValue: oldValue
      };

      if (this.__QUEUE_INDEXING__) {
        this.queue.push(change);
      } else {
        this._updateIndexes(change);
      }

      return oldValue
    }

    
    _normalizeDiff (value, diff) {
      if (isString(value)) {
        
        if (diff['delete']) {
          console.warn('DEPRECATED: use doc.update(path, {type:"delete", start:s, end: e}) instead');
          diff = {
            type: 'delete',
            start: diff['delete'].start,
            end: diff['delete'].end
          };
        } else if (diff['insert']) {
          console.warn('DEPRECATED: use doc.update(path, {type:"insert", start:s, text: t}) instead');
          diff = {
            type: 'insert',
            start: diff['insert'].offset,
            text: diff['insert'].value
          };
        }
      } else if (isArray(value)) {
        
        if (diff['delete']) {
          console.warn('DEPRECATED: use doc.update(path, {type:"delete", pos:1}) instead');
          diff = {
            type: 'delete',
            pos: diff['delete'].offset
          };
        } else if (diff['insert']) {
          console.warn('DEPRECATED: use doc.update(path, {type:"insert", pos:1, value: "foo"}) instead');
          diff = {
            type: 'insert',
            pos: diff['insert'].offset,
            value: diff['insert'].value
          };
        }
      } else if (value._isCoordinate) {
        if (diff.hasOwnProperty('shift')) {
          console.warn('DEPRECATED: use doc.update(path, {type:"shift", value:2}) instead');
          diff = {
            type: 'shift',
            value: diff['shift']
          };
        }
      }
      return diff
    }

    
    toJSON () {
      let nodes = {};
      forEach(this.nodes, (node) => {
        nodes[node.id] = node.toJSON();
      });
      return {
        schema: [this.schema.id, this.schema.version],
        nodes: nodes
      }
    }

    reset () {
      this.clear();
    }

    
    clear () {
      this.nodes = {};
      forEach(this.indexes, index => index.clear());
    }

    
    addIndex (name, index) {
      if (this.indexes[name]) {
        console.error('Index with name %s already exists.', name);
      }
      index.reset(this);
      this.indexes[name] = index;
      return index
    }

    
    getIndex (name) {
      return this.indexes[name]
    }

    
    _updateIndexes (change) {
      if (!change || this.__QUEUE_INDEXING__) return
      forEach(this.indexes, function (index) {
        if (index.select(change.node)) {
          if (!index[change.type]) {
            console.error('Contract: every NodeIndex must implement ' + change.type);
          }
          index[change.type](change.node, change.path, change.newValue, change.oldValue);
        }
      });
    }

    
    _stopIndexing () {
      this.__QUEUE_INDEXING__ = true;
    }

    
    _startIndexing () {
      this.__QUEUE_INDEXING__ = false;
      while (this.queue.length > 0) {
        var change = this.queue.shift();
        this._updateIndexes(change);
      }
    }
  }

  function _setValue (root, path, newValue) {
    let ctx = root;
    let L = path.length;
    for (let i = 0; i < L - 1; i++) {
      ctx = ctx[path[i]];
      if (!ctx) throw new Error('Can not set value.')
    }
    let oldValue = ctx[path[L - 1]];
    ctx[path[L - 1]] = newValue;
    return oldValue
  }

  class IncrementalData extends Data {
    
    create (nodeData) {
      if (nodeData._isNode) {
        nodeData = nodeData.toJSON();
      }
      let op = ObjectOperation.Create([nodeData.id], nodeData);
      this.apply(op);
      return op
    }

    
    delete (nodeId) {
      var op = null;
      var node = this.get(nodeId);
      if (node) {
        var nodeData = node.toJSON();
        op = ObjectOperation.Delete([nodeId], nodeData);
        this.apply(op);
      }
      return op
    }

    
    update (path, diff) {
      var diffOp = this._getDiffOp(path, diff);
      var op = ObjectOperation.Update(path, diffOp);
      this.apply(op);
      return op
    }

    
    set (path, newValue) {
      var oldValue = this.get(path);
      var op = ObjectOperation.Set(path, oldValue, newValue);
      this.apply(op);
      return op
    }

    
    apply (op) {
      if (op.type === ObjectOperation.NOP) return
      else if (op.type === ObjectOperation.CREATE) {
        
        super.create(cloneDeep(op.val));
      } else if (op.type === ObjectOperation.DELETE) {
        super.delete(op.val.id);
      } else if (op.type === ObjectOperation.UPDATE) {
        var oldVal = this.get(op.path);
        var diff = op.diff;
        if (op.propertyType === 'array') {
          if (!(diff._isArrayOperation)) {
            diff = ArrayOperation.fromJSON(diff);
          }
          
          diff.apply(oldVal);
        } else if (op.propertyType === 'string') {
          if (!(diff._isTextOperation)) {
            diff = TextOperation.fromJSON(diff);
          }
          var newVal = diff.apply(oldVal);
          super.set(op.path, newVal);
        } else if (op.propertyType === 'coordinate') {
          if (!(diff._isCoordinateOperation)) {
            diff = CoordinateOperation.fromJSON(diff);
          }
          diff.apply(oldVal);
        } else {
          throw new Error('Unsupported type for operational update.')
        }
      } else if (op.type === ObjectOperation.SET) {
        super.set(op.path, op.val);
      } else {
        throw new Error('Illegal state.')
      }
      this.emit('operation:applied', op, this);
    }

    
    _getDiffOp (path, diff) {
      var diffOp = null;
      if (diff.isOperation) {
        diffOp = diff;
      } else {
        var value = this.get(path);
        diff = this._normalizeDiff(value, diff);
        if (value === null || value === undefined) {
          throw new Error('Property has not been initialized: ' + JSON.stringify(path))
        } else if (isString(value)) {
          switch (diff.type) {
            case 'delete': {
              diffOp = TextOperation.Delete(diff.start, value.substring(diff.start, diff.end));
              break
            }
            case 'insert': {
              diffOp = TextOperation.Insert(diff.start, diff.text);
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else if (isArray(value)) {
          switch (diff.type) {
            case 'delete': {
              diffOp = ArrayOperation.Delete(diff.pos, value[diff.pos]);
              break
            }
            case 'insert': {
              diffOp = ArrayOperation.Insert(diff.pos, diff.value);
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else if (value._isCoordinate) {
          switch (diff.type) {
            case 'shift': {
              diffOp = CoordinateOperation.Shift(diff.value);
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        }
      }
      if (!diffOp) {
        throw new Error('Unsupported diff: ' + JSON.stringify(diff))
      }
      return diffOp
    }
  }

  class PropertyIndex extends NodeIndex {
    constructor (property) {
      super();

      this._property = property || 'id';
      this.index = new TreeIndex();
    }

    
    get (path) {
      return this.index.get(path) || {}
    }

    
    getAll (path) {
      return this.index.getAll(path)
    }

    clear () {
      this.index.clear();
    }

    
    select(node) { 
      return true
    }

    
    create (node) {
      var values = node[this._property];
      if (!isArray(values)) {
        values = [values];
      }
      forEach(values, function (value) {
        this.index.set([value, node.id], node);
      }.bind(this));
    }

    
    delete (node) {
      var values = node[this._property];
      if (!isArray(values)) {
        values = [values];
      }
      forEach(values, function (value) {
        this.index.delete([value, node.id]);
      }.bind(this));
    }

    
    update (node, path, newValue, oldValue) {
      if (!this.select(node) || path[1] !== this._property) return
      var values = oldValue;
      if (!isArray(values)) {
        values = [values];
      }
      forEach(values, function (value) {
        this.index.delete([value, node.id]);
      }.bind(this));
      values = newValue;
      if (!isArray(values)) {
        values = [values];
      }
      forEach(values, function (value) {
        this.index.set([value, node.id], node);
      }.bind(this));
    }

    set (node, path, newValue, oldValue) {
      this.update(node, path, newValue, oldValue);
    }

    _initialize (data) {
      forEach(data.getNodes(), function (node) {
        if (this.select(node)) {
          this.create(node);
        }
      }.bind(this));
    }
  }

  class AnnotationIndex extends DocumentIndex {
    constructor () {
      super();

      this.byPath = new TreeIndex();
      this.byType = new TreeIndex();
    }

    select (node) {
      return node.isPropertyAnnotation()
    }

    clear () {
      this.byPath.clear();
      this.byType.clear();
    }

    
    get (path, start, end, type) {
      var annotations;
      if (isString(path) || path.length === 1) {
        annotations = this.byPath.getAll(path) || {};
      } else {
        annotations = this.byPath.get(path);
      }
      annotations = map(annotations);
      if (isNumber(start)) {
        annotations = filter(annotations, AnnotationIndex.filterByRange(start, end));
      }
      if (type) {
        annotations = filter(annotations, DocumentIndex.filterByType(type));
      }
      return annotations
    }

    create (anno) {
      const path = anno.start.path;
      this.byType.set([anno.type, anno.id], anno);
      if (path && path.length > 0) {
        this.byPath.set(anno.start.path.concat([anno.id]), anno);
      }
    }

    delete (anno) {
      this._delete(anno.type, anno.id, anno.start.path);
    }

    _delete (type, id, path) {
      this.byType.delete([type, id]);
      if (path && path.length > 0) {
        this.byPath.delete(path.concat([id]));
      }
    }

    update (node, path, newValue, oldValue) {
      
      if (this.select(node) && path[1] === 'start' && path[2] === 'path') {
        this._delete(node.type, node.id, oldValue);
        this.create(node);
      }
    }
  }

  AnnotationIndex.filterByRange = function (start, end) {
    return function (anno) {
      var aStart = anno.start.offset;
      var aEnd = anno.end.offset;
      var overlap = (aEnd >= start);
      
      if (isNumber(end)) {
        overlap = overlap && (aStart <= end);
      }
      return overlap
    }
  };

  class ContainerAnnotationIndex extends DocumentIndex {
    constructor () {
      super();
      this.byId = new TreeIndex();
    }

    select (node) {
      return node.isContainerAnnotation()
    }

    clear () {
      this.byId.clear();
    }

    get (containerId, type) {
      var annotations = map(this.byId.get(containerId));
      if (isString(type)) {
        annotations = filter(annotations, DocumentIndex.filterByType);
      }
      return annotations
    }

    create (anno) {
      this.byId.set([anno.containerId, anno.id], anno);
    }

    delete (anno) {
      this.byId.delete([anno.containerId, anno.id]);
    }

    update(node, path, newValue, oldValue) { 
      
    }
  }

  class DocumentNodeFactory {
    constructor (doc) {
      this.doc = doc;
    }

    create (nodeType, nodeData) {
      var NodeClass = this.doc.schema.getNodeClass(nodeType);
      if (!NodeClass) {
        throw new Error('No node registered by that name: ' + nodeType)
      }
      return new NodeClass(this.doc, nodeData)
    }
  }

  function copySelection (doc, selection) {
    if (!selection) throw new Error("'selection' is mandatory.")
    let copy = null;
    if (!selection.isNull() && !selection.isCollapsed()) {
      
      if (selection.isPropertySelection()) {
        copy = _copyPropertySelection(doc, selection);
      } else if (selection.isContainerSelection()) {
        copy = _copyContainerSelection(doc, selection);
      } else if (selection.isNodeSelection()) {
        copy = _copyNodeSelection(doc, selection);
      } else {
        console.error('Copy is not yet supported for selection type.');
      }
    }
    return copy
  }

  function _copyPropertySelection (doc, selection) {
    let path = selection.start.path;
    let offset = selection.start.offset;
    let endOffset = selection.end.offset;
    let text = doc.get(path);
    let snippet = doc.createSnippet();
    let containerNode = snippet.getContainer();
    snippet.create({
      type: doc.schema.getDefaultTextType(),
      id: TEXT_SNIPPET_ID,
      content: text.substring(offset, endOffset)
    });
    containerNode.show(TEXT_SNIPPET_ID);
    let annotations = doc.getIndex('annotations').get(path, offset, endOffset);
    forEach(annotations, function (anno) {
      let data = cloneDeep(anno.toJSON());
      let path = [TEXT_SNIPPET_ID, 'content'];
      data.start = {
        path: path,
        offset: Math.max(offset, anno.start.offset) - offset
      };
      data.end = {
        path: path,
        offset: Math.min(endOffset, anno.end.offset) - offset
      };
      snippet.create(data);
    });
    return snippet
  }

  function _copyContainerSelection (tx, sel) {
    let snippet = tx.createSnippet();
    let container = snippet.getContainer();

    let nodeIds = sel.getNodeIds();
    let L = nodeIds.length;
    if (L === 0) return snippet

    let start = sel.start;
    let end = sel.end;

    let skippedFirst = false;
    let skippedLast = false;

    
    let created = {};
    for (let i = 0; i < L; i++) {
      let id = nodeIds[i];
      let node = tx.get(id);
      
      if (i === 0 && isLast(tx, start)) {
        skippedFirst = true;
        continue
      }
      if (i === L - 1 && isFirst(tx, end)) {
        skippedLast = true;
        continue
      }
      if (!created[id]) {
        copyNode(node).forEach((nodeData) => {
          let copy = snippet.create(nodeData);
          created[copy.id] = true;
        });
        container.show(id);
      }
    }
    if (!skippedFirst) {
      
      let startNode = snippet.get(start.getNodeId()).getContainerRoot();
      if (startNode.isText()) {
        deleteTextRange(snippet, null, start);
      } else if (startNode.isList()) {
        deleteListRange(snippet, startNode, null, start);
      }
    }
    if (!skippedLast) {
      
      let endNode = snippet.get(end.getNodeId()).getContainerRoot();
      if (endNode.isText()) {
        deleteTextRange(snippet, end, null);
      } else if (endNode.isList()) {
        deleteListRange(snippet, endNode, end, null);
      }
    }
    return snippet
  }

  function _copyNodeSelection (doc, selection) {
    let snippet = doc.createSnippet();
    let containerNode = snippet.getContainer();
    let nodeId = selection.getNodeId();
    let node = doc.get(nodeId);
    copyNode(node).forEach((nodeData) => {
      snippet.create(nodeData);
    });
    containerNode.show(node.id);
    return snippet
  }

  function paste (tx, args) {
    let sel = tx.selection;
    if (!sel || sel.isNull() || sel.isCustomSelection()) {
      throw new Error('Can not paste, without selection or a custom selection.')
    }
    args = args || {};
    args.text = args.text || '';
    let pasteDoc = args.doc;
    
    
    let inContainer = Boolean(sel.containerId);

    
    
    
    if (!pasteDoc && !inContainer) {
      tx.insertText(args.text);
      return
    }
    if (!pasteDoc) {
      pasteDoc = _convertPlainTextToDocument(tx, args);
    }
    if (pasteDoc && !inContainer) {
      
      let nodes = [];
      let container = pasteDoc.get('snippet');
      let content = container.getContent();
      content.forEach(nodeId => {
        let text = pasteDoc.get(nodeId).getText();
        nodes.push(text);
      });
      tx.insertText(nodes.join('\r\n'));
      return
    }
    if (!sel.isCollapsed()) {
      tx.deleteSelection();
    }
    let snippet = pasteDoc.get(SNIPPET_ID);
    if (snippet.getLength() > 0) {
      let first = snippet.getChildAt(0);
      if (first.isText()) {
        _pasteAnnotatedText(tx, pasteDoc);
        
        
        
        snippet.hideAt(0);
      }
      
      if (snippet.getLength() > 0) {
        _pasteDocument(tx, pasteDoc);
      }
    }
    return args
  }


  function _convertPlainTextToDocument (tx, args) {
    let lines = args.text.split(/\s*\n\s*\n/);
    let pasteDoc = tx.getDocument().newInstance();
    let defaultTextType = pasteDoc.getSchema().getDefaultTextType();
    let container = pasteDoc.create({
      type: '@container',
      id: SNIPPET_ID,
      nodes: []
    });
    let node;
    if (lines.length === 1) {
      node = pasteDoc.create({
        id: TEXT_SNIPPET_ID,
        type: defaultTextType,
        content: lines[0]
      });
      container.show(node.id);
    } else {
      for (let i = 0; i < lines.length; i++) {
        node = pasteDoc.create({
          id: uuid(defaultTextType),
          type: defaultTextType,
          content: lines[i]
        });
        container.show(node.id);
      }
    }
    return pasteDoc
  }

  function _pasteAnnotatedText (tx, copy) {
    let sel = tx.selection;
    const nodes = copy.get(SNIPPET_ID).nodes;
    const firstId = nodes[0];
    const first = copy.get(firstId);
    const textPath = first.getPath();
    const text = copy.get(textPath);
    const annotations = copy.getIndex('annotations').get(textPath);
    
    let path = sel.start.path;
    let offset = sel.start.offset;
    tx.insertText(text);
    
    forEach(annotations, function (anno) {
      let data = anno.toJSON();
      data.start.path = path.slice(0);
      data.start.offset += offset;
      data.end.offset += offset;
      
      if (tx.get(data.id)) data.id = uuid(data.type);
      tx.create(data);
    });
  }

  function _pasteDocument (tx, pasteDoc) {
    let sel = tx.selection;
    let containerId = sel.containerId;
    let container = tx.get(containerId);
    let insertPos;
    if (sel.isPropertySelection()) {
      let startPath = sel.start.path;
      let nodeId = sel.start.getNodeId();
      let startPos = container.getPosition(nodeId, 'strict');
      let text = tx.get(startPath);
      
      
      if (text.length === 0) {
        insertPos = startPos;
        container.hide(nodeId);
        deleteNode(tx, tx.get(nodeId));
      } else if (text.length === sel.start.offset) {
        insertPos = startPos + 1;
      } else {
        tx.break();
        insertPos = startPos + 1;
      }
    } else if (sel.isNodeSelection()) {
      let nodePos = container.getPosition(sel.getNodeId(), 'strict');
      if (sel.isBefore()) {
        insertPos = nodePos;
      } else if (sel.isAfter()) {
        insertPos = nodePos + 1;
      } else {
        throw new Error('Illegal state: the selection should be collapsed.')
      }
    }
    
    let nodeIds = pasteDoc.get(SNIPPET_ID).nodes;
    let insertedNodes = [];
    let visited = {};
    for (let i = 0; i < nodeIds.length; i++) {
      let node = pasteDoc.get(nodeIds[i]);
      
      
      
      
      
      let newId = _transferWithDisambiguatedIds(node.getDocument(), tx, node.id, visited);
      
      node = tx.get(newId);
      container.showAt(insertPos++, newId);
      insertedNodes.push(node);
    }

    if (insertedNodes.length > 0) {
      let lastNode = last$1(insertedNodes);
      setCursor(tx, lastNode, containerId, 'after');
    }
  }






  function _transferWithDisambiguatedIds (sourceDoc, targetDoc, id, visited) {
    if (visited[id]) throw new Error('FIXME: dont call me twice')
    const node = sourceDoc.get(id, 'strict');
    let oldId = node.id;
    let newId;
    if (targetDoc.contains(node.id)) {
      
      newId = uuid(node.type);
      node.id = newId;
    }
    visited[id] = node.id;
    const annotationIndex = sourceDoc.getIndex('annotations');
    const nodeSchema = node.getSchema();
    
    let annos = [];
    
    
    
    for (let prop of nodeSchema) {
      const name = prop.name;
      if (name === 'id' || name === 'type') continue
      
      if ((prop.isReference() && prop.isOwned()) || (prop.type === 'file')) {
        
        
        let val = node[prop.name];
        if (prop.isArray()) {
          _transferArrayOfReferences(sourceDoc, targetDoc, val, visited);
        } else {
          let id = val;
          if (!visited[id]) {
            node[name] = _transferWithDisambiguatedIds(sourceDoc, targetDoc, id, visited);
          }
        }
      
      } else if (prop.isText()) {
        
        
        
        
        
        let _annos = annotationIndex.get([oldId, prop.name]);
        for (let i = 0; i < _annos.length; i++) {
          let anno = _annos[i];
          if (anno.start.path[0] === oldId && newId) {
            anno.start.path[0] = newId;
          }
          if (anno.end.path[0] === oldId && newId) {
            anno.end.path[0] = newId;
          }
          annos.push(anno);
        }
      }
    }
    targetDoc.create(node);
    for (let i = 0; i < annos.length; i++) {
      _transferWithDisambiguatedIds(sourceDoc, targetDoc, annos[i].id, visited);
    }
    return node.id
  }

  function _transferArrayOfReferences (sourceDoc, targetDoc, arr, visited) {
    for (let i = 0; i < arr.length; i++) {
      let val = arr[i];
      
      if (isArray(val)) {
        _transferArrayOfReferences(sourceDoc, targetDoc, val, visited);
      } else {
        let id = val;
        if (id && !visited[id]) {
          arr[i] = _transferWithDisambiguatedIds(sourceDoc, targetDoc, id, visited);
        }
      }
    }
  }

  class Editing {
    
    annotate (tx, annotation) {
      let sel = tx.selection;
      let schema = tx.getSchema();
      let AnnotationClass = schema.getNodeClass(annotation.type);
      if (!AnnotationClass) throw new Error('Unknown annotation type', annotation)
      let start = sel.start;
      let end = sel.end;
      let containerId = sel.containerId;
      let nodeData = { start, end, containerId };
      
      
      if (sel.isPropertySelection()) {
        if (!AnnotationClass.prototype._isAnnotation) {
          throw new Error('Annotation can not be created for a selection.')
        }
      } else if (sel.isContainerSelection()) {
        if (AnnotationClass.prototype._isPropertyAnnotation) {
          console.warn('NOT SUPPORTED YET: creating property annotations for a non collapsed container selection.');
        }
      }
      Object.assign(nodeData, annotation);
      return tx.create(nodeData)
    }

    break (tx) {
      let sel = tx.selection;
      if (sel.isNodeSelection()) {
        let containerId = sel.containerId;
        let container = tx.get(containerId);
        let nodeId = sel.getNodeId();
        let nodePos = container.getPosition(nodeId, 'strict');
        let textNode = this.createTextNode(tx, container);
        if (sel.isBefore()) {
          container.showAt(nodePos, textNode.id);
          
        } else {
          container.showAt(nodePos + 1, textNode.id);
          setCursor(tx, textNode, containerId, 'before');
        }
      } else if (sel.isCustomSelection()) ; else if (sel.isCollapsed() || sel.isPropertySelection()) {
        let containerId = sel.containerId;
        if (!sel.isCollapsed()) {
          
          this._deletePropertySelection(tx, sel);
          tx.setSelection(sel.collapse('left'));
        }
        
        if (containerId) {
          let container = tx.get(containerId);
          let nodeId = sel.start.path[0];
          let node = tx.get(nodeId);
          this._breakNode(tx, node, sel.start, container);
        }
      } else if (sel.isContainerSelection()) {
        let start = sel.start;
        let containerId = sel.containerId;
        let container = tx.get(containerId);
        let startNodeId = start.path[0];
        let nodePos = container.getPosition(startNodeId, 'strict');
        this._deleteContainerSelection(tx, sel, { noMerge: true });
        setCursor(tx, container.getNodeAt(nodePos + 1), containerId, 'before');
      }
    }

    createTextNode (tx, container, text) { 
      
      
      return tx.createDefaultTextNode(text)
    }

    createListNode (tx, container, params = {}) { 
      
      
      return tx.create({ type: 'list', items: [], listType: params.listType || 'bullet' })
    }

    delete (tx, direction) {
      let sel = tx.selection;
      
      
      
      
      if (sel.isNodeSelection()) {
        this._deleteNodeSelection(tx, sel, direction);
      
      } else if (sel.isCustomSelection()) ; else if (sel.isCollapsed()) {
        
        
        
        let path = sel.start.path;
        let node = tx.get(path[0]);
        let text = tx.get(path);
        let offset = sel.start.offset;
        let needsMerge = (sel.containerId && (
          (offset === 0 && direction === 'left') ||
          (offset === text.length && direction === 'right')
        ));
        if (needsMerge) {
          
          
          
          
          
          
          
          let root = node.getContainerRoot();
          if (root.isList() && offset === 0 && direction === 'left') {
            return this.toggleList(tx)
          } else {
            let container = tx.get(sel.containerId);
            this._merge(tx, root, sel.start, direction, container);
          }
        } else {
          
          if ((offset === 0 && direction === 'left') ||
            (offset === text.length && direction === 'right')) {
            return
          }
          let startOffset = (direction === 'left') ? offset - 1 : offset;
          let endOffset = startOffset + 1;
          let start = { path: path, offset: startOffset };
          let end = { path: path, offset: endOffset };
          deleteTextRange(tx, start, end);
          tx.setSelection({
            type: 'property',
            path: path,
            startOffset: startOffset,
            containerId: sel.containerId
          });
        }
      
      } else if (sel.isPropertySelection()) {
        deleteTextRange(tx, sel.start, sel.end);
        tx.setSelection(sel.collapse('left'));
      
      } else if (sel.isContainerSelection()) {
        this._deleteContainerSelection(tx, sel);
      } else {
        console.warn('Unsupported case: tx.delete(%)', direction, sel);
      }
    }

    _deleteNodeSelection (tx, sel, direction) {
      let nodeId = sel.getNodeId();
      let container = tx.get(sel.containerId);
      let nodePos = container.getPosition(nodeId, 'strict');
      if (sel.isFull() ||
          (sel.isBefore() && direction === 'right') ||
          (sel.isAfter() && direction === 'left')) {
        
        container.hideAt(nodePos);
        deleteNode(tx, tx.get(nodeId));
        let newNode = this.createTextNode(tx, container);
        container.showAt(nodePos, newNode.id);
        tx.setSelection({
          type: 'property',
          path: newNode.getPath(),
          startOffset: 0,
          containerId: container.id
        });
      } else {
        
        if (sel.isBefore() && direction === 'left') {
          if (nodePos > 0) {
            let previous = container.getNodeAt(nodePos - 1);
            if (previous.isText()) {
              tx.setSelection({
                type: 'property',
                path: previous.getPath(),
                startOffset: previous.getLength()
              });
              this.delete(tx, direction);
            } else {
              tx.setSelection({
                type: 'node',
                nodeId: previous.id,
                containerId: container.id
              });
            }
          }
        } else if (sel.isAfter() && direction === 'right') {
          if (nodePos < container.getLength() - 1) {
            let next = container.getNodeAt(nodePos + 1);
            if (next.isText()) {
              tx.setSelection({
                type: 'property',
                path: next.getPath(),
                startOffset: 0
              });
              this.delete(tx, direction);
            } else {
              tx.setSelection({
                type: 'node',
                nodeId: next.id,
                containerId: container.id
              });
            }
          }
        } else {
          console.warn('Unsupported case: delete(%s)', direction, sel);
        }
      }
    }

    _deletePropertySelection (tx, sel) {
      let path = sel.start.path;
      let start = sel.start.offset;
      let end = sel.end.offset;
      tx.update(path, { type: 'delete', start: start, end: end });
      annotationHelpers.deletedText(tx, path, start, end);
    }

    
    _deleteContainerSelection (tx, sel, options = {}) {
      let containerId = sel.containerId;
      let container = tx.get(containerId);
      let start = sel.start;
      let end = sel.end;
      let startId = start.getNodeId();
      let endId = end.getNodeId();
      let startPos = container.getPosition(startId, 'strict');
      let endPos = container.getPosition(endId, 'strict');

      
      if (startPos === endPos) {
        
        let node = tx.get(startId).getContainerRoot();
        
        if (node.isText()) {
          deleteTextRange(tx, start, end);
        } else if (node.isList()) {
          deleteListRange(tx, node, start, end);
        } else {
          throw new Error('Not supported yet.')
        }
        tx.setSelection(sel.collapse('left'));
        return
      }

      

      let firstNode = tx.get(start.getNodeId());
      let lastNode = tx.get(end.getNodeId());
      let firstEntirelySelected = isEntirelySelected(tx, firstNode, start, null);
      let lastEntirelySelected = isEntirelySelected(tx, lastNode, null, end);

      
      if (lastEntirelySelected) {
        container.hideAt(endPos);
        deleteNode(tx, lastNode);
      } else {
        
        let node = lastNode.getContainerRoot();
        
        if (node.isText()) {
          deleteTextRange(tx, null, end);
        } else if (node.isList()) {
          deleteListRange(tx, node, null, end);
        }
      }

      
      for (let i = endPos - 1; i > startPos; i--) {
        let nodeId = container.getNodeIdAt(i);
        container.hideAt(i);
        deleteNode(tx, tx.get(nodeId));
      }

      
      if (firstEntirelySelected) {
        container.hideAt(startPos);
        deleteNode(tx, firstNode);
      } else {
        
        let node = firstNode.getContainerRoot();
        
        if (node.isText()) {
          deleteTextRange(tx, start, null);
        } else if (node.isList()) {
          deleteListRange(tx, node, start, null);
        }
      }

      
      if (firstEntirelySelected && lastEntirelySelected) {
        
        let textNode = this.createTextNode(tx, container);
        container.showAt(startPos, textNode.id);
        tx.setSelection({
          type: 'property',
          path: textNode.getPath(),
          startOffset: 0,
          containerId: containerId
        });
      } else if (!firstEntirelySelected && !lastEntirelySelected) {
        if (!options.noMerge) {
          this._merge(tx, firstNode, sel.start, 'right', container);
        }
        tx.setSelection(sel.collapse('left'));
      } else if (firstEntirelySelected) {
        setCursor(tx, lastNode, container.id, 'before');
      } else {
        setCursor(tx, firstNode, container.id, 'after');
      }
    }

    insertInlineNode (tx, nodeData) {
      let sel = tx.selection;
      let text = '\uFEFF';
      this.insertText(tx, text);
      sel = tx.selection;
      let endOffset = tx.selection.end.offset;
      let startOffset = endOffset - text.length;
      nodeData = Object.assign({}, nodeData, {
        start: {
          path: sel.path,
          offset: startOffset
        },
        end: {
          path: sel.path,
          offset: endOffset
        }
      });
      return tx.create(nodeData)
    }

    insertBlockNode (tx, nodeData) {
      let sel = tx.selection;
      
      let blockNode;
      if (!nodeData._isNode || !tx.get(nodeData.id)) {
        blockNode = tx.create(nodeData);
      } else {
        blockNode = tx.get(nodeData.id);
      }
      
      if (sel.isNodeSelection()) {
        let containerId = sel.containerId;
        let container = tx.get(containerId);
        let nodeId = sel.getNodeId();
        let nodePos = container.getPosition(nodeId, 'strict');
        
        if (sel.isBefore()) {
          container.showAt(nodePos, blockNode.id);
        
        } else if (sel.isAfter()) {
          container.showAt(nodePos + 1, blockNode.id);
          tx.setSelection({
            type: 'node',
            containerId: containerId,
            nodeId: blockNode.id,
            mode: 'after'
          });
        } else {
          container.hideAt(nodePos);
          deleteNode(tx, tx.get(nodeId));
          container.showAt(nodePos, blockNode.id);
          tx.setSelection({
            type: 'node',
            containerId: containerId,
            nodeId: blockNode.id,
            mode: 'after'
          });
        }
      } else if (sel.isPropertySelection()) {
        
        if (!sel.containerId) throw new Error('insertBlockNode can only be used within a container.')
        let container = tx.get(sel.containerId);
        if (!sel.isCollapsed()) {
          this._deletePropertySelection(tx, sel);
          tx.setSelection(sel.collapse('left'));
        }
        let node = tx.get(sel.path[0]);
        
        if (!node) throw new Error('Invalid selection.')
        let nodePos = container.getPosition(node.id, 'strict');
        
        if (node.isText()) {
          let text = node.getText();
          
          if (text.length === 0) {
            container.hideAt(nodePos);
            deleteNode(tx, node);
            container.showAt(nodePos, blockNode.id);
            setCursor(tx, blockNode, container.id, 'after');
          
          } else if (sel.start.offset === 0) {
            container.showAt(nodePos, blockNode.id);
          
          } else if (sel.start.offset === text.length) {
            container.showAt(nodePos + 1, blockNode.id);
            setCursor(tx, blockNode, container.id, 'before');
          
          } else {
            this.break(tx);
            container.showAt(nodePos + 1, blockNode.id);
            setCursor(tx, blockNode, container.id, 'after');
          }
        } else {
          console.error('Not supported: insertBlockNode() on a custom node');
        }
      } else if (sel.isContainerSelection()) {
        if (sel.isCollapsed()) {
          let start = sel.start;
          
          if (start.isPropertyCoordinate()) {
            tx.setSelection({
              type: 'property',
              path: start.path,
              startOffset: start.offset,
              containerId: sel.containerId
            });
          } else if (start.isNodeCoordinate()) {
            tx.setSelection({
              type: 'node',
              containerId: sel.containerId,
              nodeId: start.path[0],
              mode: start.offset === 0 ? 'before' : 'after'
            });
          } else {
            throw new Error('Unsupported selection for insertBlockNode')
          }
          return this.insertBlockNode(tx, blockNode)
        } else {
          this.break(tx);
          return this.insertBlockNode(tx, blockNode)
        }
      }
      return blockNode
    }

    insertText (tx, text) {
      let sel = tx.selection;
      
      
      
      if (sel.isNodeSelection()) {
        let containerId = sel.containerId;
        let container = tx.get(containerId);
        let nodeId = sel.getNodeId();
        let nodePos = container.getPosition(nodeId, 'strict');
        let textNode = this.createTextNode(tx, container, text);
        if (sel.isBefore()) {
          container.showAt(nodePos, textNode);
        } else if (sel.isAfter()) {
          container.showAt(nodePos + 1, textNode);
        } else {
          container.hide(nodeId);
          deleteNode(tx, tx.get(nodeId));
          container.showAt(nodePos, textNode);
        }
        setCursor(tx, textNode, sel.containerId, 'after');
      } else if (sel.isCustomSelection()) ; else if (sel.isCollapsed() || sel.isPropertySelection()) {
        
        this._insertText(tx, sel, text);
        
      } else if (sel.isContainerSelection()) {
        this._deleteContainerSelection(tx, sel);
        this.insertText(tx, text);
      }
    }

    paste (tx, content) {
      if (!content) return
      
      if (isString(content)) {
        paste(tx, {text: content});
      } else if (content._isDocument) {
        paste(tx, {doc: content});
      } else {
        throw new Error('Illegal content for paste.')
      }
    }

    
    switchTextType (tx, data) {
      let sel = tx.selection;
      
      if (!sel.isPropertySelection()) {
        throw new Error('Selection must be a PropertySelection.')
      }
      let containerId = sel.containerId;
      
      if (!containerId) {
        throw new Error('Selection must be within a container.')
      }
      let path = sel.path;
      let nodeId = path[0];
      let node = tx.get(nodeId);
      
      if (!(node.isInstanceOf('text'))) {
        throw new Error('Trying to use switchTextType on a non text node.')
      }
      const newId = uuid(data.type);
      
      const oldPath = node.getPath();
      console.assert(oldPath.length === 2, 'Currently we assume that TextNodes store the plain-text on the first level');
      const textProp = oldPath[1];
      let newPath = [newId, textProp];
      
      let newNodeData = Object.assign({
        id: newId,
        type: data.type,
        direction: node.direction
      }, data);
      newNodeData[textProp] = node.getText();

      let newNode = tx.create(newNodeData);
      annotationHelpers.transferAnnotations(tx, path, 0, newPath, 0);

      
      let container = tx.get(sel.containerId);
      let pos = container.getPosition(nodeId, 'strict');
      container.hide(nodeId);
      deleteNode(tx, node);
      container.showAt(pos, newNode.id);

      tx.setSelection({
        type: 'property',
        path: newPath,
        startOffset: sel.start.offset,
        endOffset: sel.end.offset,
        containerId: containerId
      });

      return newNode
    }

    toggleList (tx, params) {
      let sel = tx.selection;
      let container = tx.get(sel.containerId);
      
      if (!container) {
        throw new Error('Selection must be within a container.')
      }
      if (sel.isPropertySelection()) {
        let nodeId = sel.start.path[0];
        
        let node = tx.get(nodeId).getContainerRoot();
        let nodePos = container.getPosition(node.id, 'strict');
        
        if (node.isText()) {
          container.hideAt(nodePos);
          let newList = this.createListNode(tx, container, params);
          let newItem = newList.createListItem(node.getText());
          annotationHelpers.transferAnnotations(tx, node.getPath(), 0, newItem.getPath(), 0);
          newList.appendItem(newItem);
          deleteNode(tx, node);
          container.showAt(nodePos, newList.id);
          tx.setSelection({
            type: 'property',
            path: newItem.getPath(),
            startOffset: sel.start.offset,
            containerId: sel.containerId
          });
        } else if (node.isList()) {
          let itemId = sel.start.path[0];
          let item = tx.get(itemId);
          let itemPos = node.getItemPosition(item);
          let newTextNode = this.createTextNode(tx, container, item.getText());
          annotationHelpers.transferAnnotations(tx, item.getPath(), 0, newTextNode.getPath(), 0);
          
          node.removeItemAt(itemPos);
          if (node.isEmpty()) {
            container.hideAt(nodePos);
            deleteNode(tx, node);
            container.showAt(nodePos, newTextNode.id);
          } else if (itemPos === 0) {
            container.showAt(nodePos, newTextNode.id);
          } else if (node.getLength() <= itemPos) {
            container.showAt(nodePos + 1, newTextNode.id);
          } else {
            
            let tail = [];
            const items = node.getItems();
            const L = items.length;
            for (let i = L - 1; i >= itemPos; i--) {
              tail.unshift(items[i]);
              node.removeItemAt(i);
            }
            let newList = this.createListNode(tx, container, node);
            for (let i = 0; i < tail.length; i++) {
              newList.appendItem(tail[i]);
            }
            container.showAt(nodePos + 1, newTextNode.id);
            container.showAt(nodePos + 2, newList.id);
          }
          tx.setSelection({
            type: 'property',
            path: newTextNode.getPath(),
            startOffset: sel.start.offset,
            containerId: sel.containerId
          });
        }
      } else if (sel.isContainerSelection()) {
        console.error('TODO: support toggleList with ContainerSelection');
      }
    }

    indent (tx) {
      let sel = tx.selection;
      if (sel.isPropertySelection()) {
        let nodeId = sel.start.getNodeId();
        
        let node = tx.get(nodeId).getContainerRoot();
        if (node.isList()) {
          let itemId = sel.start.path[0];
          let item = tx.get(itemId);
          let level = item.getLevel();
          
          if (item && level < 3) {
            item.setLevel(item.level + 1);
            
            tx.set([node.id, '_itemsChanged'], true);
          }
        }
      } else if (sel.isContainerSelection()) {
        console.error('TODO: support toggleList with ContainerSelection');
      }
    }

    dedent (tx) {
      let sel = tx.selection;
      if (sel.isPropertySelection()) {
        let nodeId = sel.start.getNodeId();
        
        let node = tx.get(nodeId).getContainerRoot();
        if (node.isList()) {
          let itemId = sel.start.path[0];
          let item = tx.get(itemId);
          let level = item.getLevel();
          if (item) {
            if (level > 1) {
              item.setLevel(item.level - 1);
              
              tx.set([node.id, '_itemsChanged'], true);
            }
            
            
            
            
            
          }
        }
      } else if (sel.isContainerSelection()) {
        console.error('TODO: support toggleList with ContainerSelection');
      }
    }

    
    _insertText (tx, sel, text) {
      let start = sel.start;
      let end = sel.end;
      
      if (!isArrayEqual(start.path, end.path)) {
        throw new Error('Unsupported state: range should be on one property')
      }
      let path = start.path;
      let startOffset = start.offset;
      let endOffset = end.offset;
      let typeover = !sel.isCollapsed();
      let L = text.length;
      
      if (typeover) {
        tx.update(path, { type: 'delete', start: startOffset, end: endOffset });
      }
      
      tx.update(path, { type: 'insert', start: startOffset, text: text });
      
      let annos = tx.getAnnotations(path);
      annos.forEach(function (anno) {
        let annoStart = anno.start.offset;
        let annoEnd = anno.end.offset;

        
        
        if (annoEnd < startOffset) ; else if (annoStart >= endOffset) {
          tx.update([anno.id, 'start'], { type: 'shift', value: startOffset - endOffset + L });
          tx.update([anno.id, 'end'], { type: 'shift', value: startOffset - endOffset + L });
        
        
        
        
        } else if (
          (annoStart >= startOffset && annoEnd < endOffset) ||
          (anno._isInlineNode && annoStart >= startOffset && annoEnd <= endOffset)
        ) {
          tx.delete(anno.id);
        
        } else if (annoStart >= startOffset && annoEnd >= endOffset) {
          
          if (annoStart > startOffset || !typeover) {
            tx.update([anno.id, 'start'], { type: 'shift', value: startOffset - annoStart + L });
          }
          tx.update([anno.id, 'end'], { type: 'shift', value: startOffset - endOffset + L });
        
        } else if (annoStart < startOffset && annoEnd < endOffset) {
          
          tx.update([anno.id, 'end'], { type: 'shift', value: startOffset - annoEnd + L });
        
        } else if (annoEnd === startOffset && !anno.constructor.autoExpandRight) ; else if (annoStart < startOffset && annoEnd >= endOffset) {
          if (anno._isInlineNode) ; else {
            tx.update([anno.id, 'end'], { type: 'shift', value: startOffset - endOffset + L });
          }
        } else {
          console.warn('TODO: handle annotation update case.');
        }
      });
      let offset = startOffset + text.length;
      tx.setSelection({
        type: 'property',
        path: start.path,
        startOffset: offset,
        containerId: sel.containerId,
        surfaceId: sel.surfaceId
      });
    }

    _breakNode (tx, node, coor, container) {
      
      node = node.getContainerRoot();
      
      if (node.isText()) {
        this._breakTextNode(tx, node, coor, container);
      } else if (node.isList()) {
        this._breakListNode(tx, node, coor, container);
      } else {
        throw new Error('Not supported')
      }
    }

    _breakTextNode (tx, node, coor, container) {
      let path = coor.path;
      let offset = coor.offset;
      let nodePos = container.getPosition(node.id, 'strict');
      let text = node.getText();

      
      
      if (offset === 0) {
        let newNode = tx.create({
          type: node.type,
          content: ''
        });
        
        container.showAt(nodePos, newNode.id);
        tx.setSelection({
          type: 'property',
          path: path,
          startOffset: 0,
          containerId: container.id
        });
      
      } else {
        const textPath = node.getPath();
        const textProp = textPath[1];
        const newId = uuid(node.type);
        let newNodeData = node.toJSON();
        newNodeData.id = newId;
        newNodeData[textProp] = text.substring(offset);
        
        if (offset === text.length) {
          newNodeData.type = tx.getSchema().getDefaultTextType();
        }
        let newNode = tx.create(newNodeData);
        
        if (offset < text.length) {
          
          annotationHelpers.transferAnnotations(tx, path, offset, newNode.getPath(), 0);
          
          tx.update(path, { type: 'delete', start: offset, end: text.length });
        }
        
        container.showAt(nodePos + 1, newNode.id);
        
        tx.setSelection({
          type: 'property',
          path: newNode.getPath(),
          startOffset: 0,
          containerId: container.id
        });
      }
    }

    _breakListNode (tx, node, coor, container) {
      let path = coor.path;
      let offset = coor.offset;
      let listItem = tx.get(path[0]);

      let L = node.length;
      let itemPos = node.getItemPosition(listItem);
      let text = listItem.getText();
      let textProp = listItem.getPath()[1];
      let newItemData = listItem.toJSON();
      delete newItemData.id;
      if (offset === 0) {
        
        if (!text) {
          
          
          let nodePos = container.getPosition(node.id, 'strict');
          let newTextNode = this.createTextNode(tx, container);
          
          if (L < 2) {
            container.hide(node.id);
            deleteNode(tx, node);
            container.showAt(nodePos, newTextNode.id);
          
          } else if (itemPos === 0) {
            node.removeItem(listItem);
            deleteNode(tx, listItem);
            container.showAt(nodePos, newTextNode.id);
          
          } else if (itemPos >= L - 1) {
            node.removeItem(listItem);
            deleteNode(tx, listItem);
            container.showAt(nodePos + 1, newTextNode.id);
          
          } else {
            let tail = [];
            const items = node.getItems().slice();
            for (let i = L - 1; i > itemPos; i--) {
              tail.unshift(items[i]);
              node.removeItem(items[i]);
            }
            node.removeItem(items[itemPos]);
            let newList = this.createListNode(tx, container, node);
            for (let i = 0; i < tail.length; i++) {
              newList.appendItem(tail[i]);
            }
            container.showAt(nodePos + 1, newTextNode.id);
            container.showAt(nodePos + 2, newList.id);
          }
          tx.setSelection({
            type: 'property',
            path: newTextNode.getPath(),
            startOffset: 0
          });
        
        } else {
          newItemData[textProp] = '';
          let newItem = tx.create(newItemData);
          node.insertItemAt(itemPos, newItem);
          tx.setSelection({
            type: 'property',
            path: listItem.getPath(),
            startOffset: 0
          });
        }
      
      } else {
        newItemData[textProp] = text.substring(offset);
        let newItem = tx.create(newItemData);
        
        if (offset < text.length) {
          
          annotationHelpers.transferAnnotations(tx, path, offset, newItem.getPath(), 0);
          
          tx.update(path, { type: 'delete', start: offset, end: text.length });
        }
        node.insertItemAt(itemPos + 1, newItem);
        tx.setSelection({
          type: 'property',
          path: newItem.getPath(),
          startOffset: 0
        });
      }
    }

    _merge (tx, node, coor, direction, container) {
      
      
      if (node.isList()) {
        let list = node;
        let itemId = coor.path[0];
        let item = tx.get(itemId);
        let itemPos = list.getItemPosition(item);
        let withinListNode = (
          (direction === 'left' && itemPos > 0) ||
          (direction === 'right' && itemPos < list.items.length - 1)
        );
        if (withinListNode) {
          itemPos = (direction === 'left') ? itemPos - 1 : itemPos;
          let target = list.getItemAt(itemPos);
          let targetLength = target.getLength();
          mergeListItems(tx, list.id, itemPos);
          tx.setSelection({
            type: 'property',
            path: target.getPath(),
            startOffset: targetLength,
            containerId: container.id
          });
          return
        }
      }
      
      let nodePos = container.getPosition(node, 'strict');
      if (direction === 'left' && nodePos > 0) {
        this._mergeNodes(tx, container, nodePos - 1, direction);
      } else if (direction === 'right' && nodePos < container.getLength() - 1) {
        this._mergeNodes(tx, container, nodePos, direction);
      }
    }

    _mergeNodes (tx, container, pos, direction) {
      let first = container.getChildAt(pos);
      let second = container.getChildAt(pos + 1);
      if (first.isText()) {
        
        if (first.isEmpty()) {
          container.hide(first.id);
          deleteNode(tx, first);
          
          
          setCursor(tx, second, container.id, 'before');
          return
        }
        let target = first;
        let targetPath = target.getPath();
        let targetLength = target.getLength();
        if (second.isText()) {
          let source = second;
          let sourcePath = source.getPath();
          container.hide(source.id);
          
          tx.update(targetPath, { type: 'insert', start: targetLength, text: source.getText() });
          
          annotationHelpers.transferAnnotations(tx, sourcePath, 0, targetPath, targetLength);
          deleteNode(tx, source);
          tx.setSelection({
            type: 'property',
            path: targetPath,
            startOffset: targetLength,
            containerId: container.id
          });
        } else if (second.isList()) {
          let list = second;
          if (!second.isEmpty()) {
            let source = list.getFirstItem();
            let sourcePath = source.getPath();
            
            list.removeItemAt(0);
            
            tx.update(targetPath, { type: 'insert', start: targetLength, text: source.getText() });
            
            annotationHelpers.transferAnnotations(tx, sourcePath, 0, targetPath, targetLength);
            
            deleteNode(tx, source);
          }
          if (list.isEmpty()) {
            container.hide(list.id);
            deleteNode(tx, list);
          }
          tx.setSelection({
            type: 'property',
            path: targetPath,
            startOffset: targetLength,
            containerId: container.id
          });
        } else {
          selectNode(tx, direction === 'left' ? first.id : second.id, container.id);
        }
      } else if (first.isList()) {
        if (second.isText()) {
          let target = first.getLastItem();
          let targetPath = target.getPath();
          let targetLength = target.getLength();
          let third = (container.length > pos + 2) ? container.getChildAt(pos + 2) : null;
          if (second.getLength() === 0) {
            container.hide(second.id);
            deleteNode(tx, second);
          } else {
            let source = second;
            let sourcePath = source.getPath();
            container.hide(source.id);
            tx.update(targetPath, { type: 'insert', start: targetLength, text: source.getText() });
            annotationHelpers.transferAnnotations(tx, sourcePath, 0, targetPath, targetLength);
            deleteNode(tx, source);
          }
          
          if (third && third.type === first.type) {
            this._mergeTwoLists(tx, container, first, third);
          }
          tx.setSelection({
            type: 'property',
            path: target.getPath(),
            startOffset: targetLength,
            containerId: container.id
          });
        } else if (second.isList()) {
          
          if (direction !== 'right') {
            
            
            throw new Error('Illegal state')
          }
          let item = first.getLastItem();
          this._mergeTwoLists(tx, container, first, second);
          tx.setSelection({
            type: 'property',
            path: item.getPath(),
            startOffset: item.getLength(),
            containerId: container.id
          });
        } else {
          selectNode(tx, direction === 'left' ? first.id : second.id, container.id);
        }
      } else {
        if (second.isText() && second.isEmpty()) {
          container.hide(second.id);
          deleteNode(tx, second);
          setCursor(tx, first, container.id, 'after');
        } else {
          selectNode(tx, direction === 'left' ? first.id : second.id, container.id);
        }
      }
    }

    _mergeTwoLists (tx, container, first, second) {
      container.hide(second.id);
      let secondItems = second.getItems().slice();
      for (let i = 0; i < secondItems.length; i++) {
        second.removeItemAt(0);
        first.appendItem(secondItems[i]);
      }
      deleteNode(tx, second);
    }
  }

  class EditingInterface {
    constructor (doc, options = {}) {
      this._document = doc;
      this._selection = null;
      
      this._impl = options.editing || new Editing();
      this._direction = null;
    }

    dispose () {}

    getDocument () {
      return this._document
    }

    

    get (...args) {
      return this._document.get(...args)
    }

    contains (id) {
      return this._document.contains(id)
    }

    create (nodeData) {
      return this._document.create(nodeData)
    }

    createDefaultTextNode (content) {
      return this._document.createDefaultTextNode(content, this._direction)
    }

    delete (nodeId) {
      return this._document.delete(nodeId)
    }

    set (path, value) {
      return this._document.set(path, value)
    }

    update (path, diffOp) {
      return this._document.update(path, diffOp)
    }

    updateNode (id, newProps) {
      return this._document.updateNode(id, newProps)
    }

    

    createSelection (selData) {
      
      
      
      
      
      
      
      selData = augmentSelection(selData, this._selection);
      return this._document.createSelection(selData)
    }

    setSelection (sel) {
      if (!sel) {
        sel = Selection.nullSelection;
      } else if (isPlainObject(sel)) {
        sel = this.createSelection(sel);
      } else {
        sel = augmentSelection(sel, this._selection);
      }
      this._selection = sel;
      return sel
    }

    getSelection () {
      return this._selection
    }

    get selection () {
      return this._selection
    }

    set selection (sel) {
      this.setSelection(sel);
    }

    
    get textDirection () {
      return this._direction
    }

    set textDirection (dir) {
      this._direction = dir;
    }

    

    annotate (annotationData) {
      const sel = this._selection;
      if (sel && (sel.isPropertySelection() || sel.isContainerSelection())) {
        return this._impl.annotate(this, annotationData)
      }
    }

    break () {
      if (this._selection && !this._selection.isNull()) {
        this._impl.break(this);
      }
    }

    copySelection () {
      const sel = this._selection;
      if (sel && !sel.isNull() && !sel.isCollapsed()) {
        return copySelection(this.getDocument(), this._selection)
      }
    }

    deleteSelection (options) {
      const sel = this._selection;
      if (sel && !sel.isNull() && !sel.isCollapsed()) {
        this._impl.delete(this, 'right', options);
      }
    }

    deleteCharacter (direction) {
      const sel = this._selection;
      if (!sel || sel.isNull()) ; else if (!sel.isCollapsed()) {
        this.deleteSelection();
      } else {
        this._impl.delete(this, direction);
      }
    }

    insertText (text) {
      const sel = this._selection;
      if (sel && !sel.isNull()) {
        this._impl.insertText(this, text);
      }
    }

    
    insertInlineNode (inlineNode) {
      const sel = this._selection;
      if (sel && !sel.isNull() && sel.isPropertySelection()) {
        return this._impl.insertInlineNode(this, inlineNode)
      }
    }

    insertBlockNode (blockNode) {
      const sel = this._selection;
      if (sel && !sel.isNull()) {
        return this._impl.insertBlockNode(this, blockNode)
      }
    }

    paste (content) {
      const sel = this._selection;
      if (sel && !sel.isNull() && !sel.isCustomSelection()) {
        return this._impl.paste(this, content)
      }
    }

    switchTextType (nodeData) {
      const sel = this._selection;
      if (sel && !sel.isNull()) {
        return this._impl.switchTextType(this, nodeData)
      }
    }

    toggleList (params) {
      const sel = this._selection;
      if (sel && !sel.isNull()) {
        return this._impl.toggleList(this, params)
      }
    }

    indent () {
      const sel = this._selection;
      if (sel && !sel.isNull()) {
        return this._impl.indent(this)
      }
    }

    dedent () {
      const sel = this._selection;
      if (sel && !sel.isNull()) {
        return this._impl.dedent(this)
      }
    }

    

    getIndex (...args) {
      return this._document.getIndex(...args)
    }

    getAnnotations (...args) {
      return this._document.getAnnotations(...args)
    }

    getSchema () {
      return this._document.getSchema()
    }

    createSnippet () {
      return this._document.createSnippet()
    }
  }

  class JSONConverter {
    importDocument (doc, json) {
      if (!json.nodes) {
        throw new Error('Invalid JSON format.')
      }
      var schema = doc.getSchema();
      if (json.schema && schema.name !== json.schema.name) {
        throw new Error('Incompatible schema.')
      }
      
      var nodes = json.nodes;
      
      
      
      doc.import(function (tx) {
        forEach(nodes, function (node) {
          
          if (tx.get(node.id)) {
            tx.delete(node.id);
          }
          tx.create(node);
        });
      });
      return doc
    }

    exportDocument (doc) {
      var schema = doc.getSchema();
      var json = {
        schema: {
          name: schema.name
        },
        nodes: {}
      };
      forEach(doc.getNodes(), function (node) {
        if (node._isDocumentNode) {
          json.nodes[node.id] = node.toJSON();
        }
      });
      return json
    }
  }

  class ParentNodeHook {
    constructor (doc) {
      this.doc = doc;
      
      this.parents = {};
      doc.data.on('operation:applied', this._onOperationApplied, this);
    }

    _onOperationApplied (op) {
      const doc = this.doc;
      let node = doc.get(op.path[0]);
      let nodeSchema, hasOwnedProperties;
      if (node) {
        nodeSchema = node.getSchema();
        hasOwnedProperties = nodeSchema.hasOwnedProperties();
      }
      
      
      switch (op.type) {
        case 'create': {
          if (hasOwnedProperties) {
            nodeSchema.getOwnedProperties().forEach(p => {
              let refs = node[p.name];
              if (refs) {
                this._setParent(node, refs);
              }
              this._setRegisteredParent(node);
            });
          }
          break
        }
        case 'update': {
          if (hasOwnedProperties) {
            let propName = op.path[1];
            if (nodeSchema.isOwned(propName)) {
              let update = op.diff;
              if (update.isDelete()) {
                this._setParent(null, update.getValue());
              } else {
                this._setParent(node, update.getValue());
              }
            }
          }
          break
        }
        case 'set': {
          if (hasOwnedProperties) {
            let propName = op.path[1];
            if (nodeSchema.isOwned(propName)) {
              let oldValue = op.getOldValue();
              let newValue = op.getValue();
              
              this._setParent(null, oldValue);
              this._setParent(node, newValue);
            }
          }
          break
        }
        default:
          
      }
    }

    _setParent (parent, ids) {
      if (ids) {
        if (isArray(ids)) {
          ids.forEach(id => this.__setParent(parent, id));
        } else {
          this.__setParent(parent, ids);
        }
      }
    }

    __setParent (parent, id) {
      
      
      
      
      
      
      this.parents[id] = parent;
      let child = this.doc.get(id);
      if (child) {
        child.setParent(parent);
      }
    }

    _setRegisteredParent (child) {
      let parent = this.parents[child.id];
      if (parent) {
        child.setParent(parent);
      }
    }
  }

  ParentNodeHook.register = function (doc) {
    return new ParentNodeHook(doc)
  };

  const converter = new JSONConverter();



  class Document extends EventEmitter {
    
    constructor (schema, ...args) {
      super();

      this.schema = schema;
      
      if (!schema) {
        throw new Error('A document needs a schema for reflection.')
      }

      
      this._ops = [];

      this._initialize(...args);
    }

    _initialize () {
      this.__id__ = uuid();
      this.nodeFactory = new DocumentNodeFactory(this);
      this.data = new IncrementalData(this.schema, this.nodeFactory);
      
      this.addIndex('type', new PropertyIndex('type'));
      
      this.addIndex('annotations', new AnnotationIndex());
      
      
      
      this.addIndex('container-annotations', new ContainerAnnotationIndex());
      
      
      ParentNodeHook.register(this);
    }

    dispose () {
      this.off();
      this.data.off();
    }

    get id () {
      return this.__id__
    }

    
    getSchema () {
      return this.schema
    }

    
    contains (id) {
      return this.data.contains(id)
    }

    
    get (path, strict) {
      return this.data.get(path, strict)
    }

    
    getNodes () {
      return this.data.getNodes()
    }

    getAnnotations (path) {
      return this.getIndex('annotations').get(path)
    }

    
    import (importer) {
      try {
        this.data._stopIndexing();
        importer(this);
        this.data._startIndexing();
      } finally {
        this.data.queue = [];
        this.data._startIndexing();
      }
    }

    
    create (nodeData) {
      if (!nodeData.id) {
        nodeData.id = uuid(nodeData.type);
      }
      if (!nodeData.type) {
        throw new Error('No node type provided')
      }
      const op = this._create(nodeData);
      if (op) {
        this._ops.push(op);
        if (!this._isTransactionDocument) {
          this._emitChange(op);
        }
        return this.get(nodeData.id)
      }
    }

    createDefaultTextNode (text, dir) {
      return this.create({
        type: this.getSchema().getDefaultTextType(),
        content: text || '',
        direction: dir
      })
    }

    
    delete (nodeId) {
      const node = this.get(nodeId);
      const op = this._delete(nodeId);
      if (op) {
        this._ops.push(op);
        if (!this._isTransactionDocument) {
          this._emitChange(op);
        }
      }
      return node
    }

    
    set (path, value) {
      const oldValue = this.get(path);
      const op = this._set(path, value);
      if (op) {
        this._ops.push(op);
        if (!this._isTransactionDocument) {
          this._emitChange(op);
        }
      }
      return oldValue
    }

    
    update (path, diff) {
      const op = this._update(path, diff);
      if (op) {
        this._ops.push(op);
        if (!this._isTransactionDocument) {
          this._emitChange(op);
        }
      }
      return op
    }

    
    updateNode (id, newProps) {
      let node = this.get(id);
      forEach(newProps, (value, key) => {
        if (!isEqual(node[key], newProps[key])) {
          this.set([id, key], value);
        }
      });
    }

    
    addIndex (name, index) {
      return this.data.addIndex(name, index)
    }

    
    getIndex (name) {
      return this.data.getIndex(name)
    }

    
    createSelection (data) {
      let sel;
      if (isNil(data)) return Selection.nullSelection
      if (arguments.length !== 1 || !isPlainObject(data)) {
        throw new Error('Illegal argument: call createSelection({ type: ... }')
      } else {
        switch (data.type) {
          case 'property': {
            if (isNil(data.endOffset)) {
              data.endOffset = data.startOffset;
            }
            if (!data.hasOwnProperty('reverse')) {
              if (data.startOffset > data.endOffset) {
                [data.startOffset, data.endOffset] = [data.endOffset, data.startOffset];
                data.reverse = !data.reverse;
              }
            }
            
            let text = this.get(data.path, 'strict');
            if (data.startOffset < 0 || data.startOffset > text.length) {
              throw new Error('Invalid startOffset: target property has length ' + text.length + ', given startOffset is ' + data.startOffset)
            }
            if (data.endOffset < 0 || data.endOffset > text.length) {
              throw new Error('Invalid startOffset: target property has length ' + text.length + ', given endOffset is ' + data.endOffset)
            }
            sel = new PropertySelection(data);
            break
          }
          case 'container': {
            let container = this.get(data.containerId, 'strict');
            if (!container) throw new Error('Can not create ContainerSelection: container "' + data.containerId + '" does not exist.')
            let start = this._normalizeCoor({ path: data.startPath, offset: data.startOffset });
            let end = this._normalizeCoor({ path: data.endPath, offset: data.endOffset });
            let startAddress = container.getAddress(start);
            let endAddress = container.getAddress(end);
            if (!startAddress) {
              throw new Error('Invalid arguments for ContainerSelection: ', start.toString())
            }
            if (!endAddress) {
              throw new Error('Invalid arguments for ContainerSelection: ', end.toString())
            }
            if (!data.hasOwnProperty('reverse')) {
              if (endAddress.isBefore(startAddress, 'strict')) {
                [start, end] = [end, start];
                data.reverse = true;
              }
            }
            sel = new ContainerSelection(container.id, start.path, start.offset, end.path, end.offset, data.reverse, data.surfaceId);
            break
          }
          case 'node': {
            sel = createNodeSelection({
              doc: this,
              nodeId: data.nodeId,
              mode: data.mode,
              containerId: data.containerId,
              reverse: data.reverse,
              surfaceId: data.surfaceId
            });
            break
          }
          case 'custom': {
            sel = CustomSelection.fromJSON(data);
            break
          }
          default:
            throw new Error('Illegal selection type', data)
        }
      }
      if (!sel.isNull()) {
        sel.attach(this);
      }
      return sel
    }

    newInstance () {
      var DocumentClass = this.constructor;
      return new DocumentClass(this.schema)
    }

    
    createSnippet () {
      var snippet = this.newInstance();
      var snippetContainer = snippet.create({
        type: '@container',
        id: SNIPPET_ID
      });
      snippet.getContainer = function () {
        return snippetContainer
      };
      return snippet
    }

    createFromDocument (doc) {
      
      this.clear();

      
      
      
      
      
      
      
      let nodes = Object.values(doc.getNodes());
      let levels = {};
      let visited = new Set();
      nodes.forEach(n => {
        if (!visited.has(n)) this._computeDependencyLevel(n, levels, visited);
      });
      
      nodes.sort((a, b) => {
        return levels[b.id] - levels[a.id]
      });
      nodes.forEach(n => this.create(n));
      return this
    }

    _computeDependencyLevel (node, levels, visited) {
      if (!node) throw new Error('node was nil')
      if (visited.has(node)) throw new Error('Cyclic node dependency')
      visited.add(node);
      
      
      
      let level = 0;
      if (node.isAnnotation() || node.isInline()) {
        level = -1;
      } else {
        let parent = node.getParent();
        if (parent) {
          let parentLevel;
          if (levels.hasOwnProperty(parent.id)) {
            parentLevel = levels[parent.id];
          } else {
            parentLevel = this._computeDependencyLevel(parent, levels, visited);
          }
          level = parentLevel + 1;
        }
      }
      levels[node.id] = level;
      return level
    }

    
    toJSON () {
      return converter.exportDocument(this)
    }

    clone () {
      let copy = this.newInstance();
      copy.createFromDocument(this);
      return copy
    }

    clear () {
      this.data.clear();
      this._ops.length = 0;
    }

    
    createEditingInterface () {
      return new EditingInterface(this)
    }

    invert (change) {
      return change.invert()
    }

    _apply (documentChange) {
      forEach(documentChange.ops, (op) => {
        this._applyOp(op);
      });
      
      documentChange._extractInformation(this);
    }

    _applyOp (op) {
      this.data.apply(op);
      this.emit('operation:applied', op);
    }

    _create (nodeData) {
      return this.data.create(nodeData)
    }

    _delete (nodeId) {
      return this.data.delete(nodeId)
    }

    _set (path, value) {
      return this.data.set(path, value)
    }

    _update (path, diff) {
      return this.data.update(path, diff)
    }

    _emitChange (op) {
      const change = new DocumentChange([op], {}, {});
      change._extractInformation(this);
      this._notifyChangeListeners(change, { hidden: true });
    }

    _notifyChangeListeners (change, info) {
      info = info || {};
      this.emit('document:changed', change, info, this);
    }

    
    _createSelectionFromRange (range) {
      if (!range) return Selection.nullSelection
      let inOneNode = isEqual(range.start.path, range.end.path);
      if (inOneNode) {
        if (range.start.isNodeCoordinate()) {
          
          
          return new NodeSelection(range.containerId, range.start.getNodeId(), 'full', range.reverse, range.surfaceId)
        } else {
          return this.createSelection({
            type: 'property',
            path: range.start.path,
            startOffset: range.start.offset,
            endOffset: range.end.offset,
            reverse: range.reverse,
            containerId: range.containerId,
            surfaceId: range.surfaceId
          })
        }
      } else {
        return this.createSelection({
          type: 'container',
          startPath: range.start.path,
          startOffset: range.start.offset,
          endPath: range.end.path,
          endOffset: range.end.offset,
          reverse: range.reverse,
          containerId: range.containerId,
          surfaceId: range.surfaceId
        })
      }
    }

    _normalizeCoor ({ path, offset }) {
      
      if (path.length === 1) {
        let node = this.get(path[0]).getContainerRoot();
        if (node.isText()) {
          
          return new Coordinate(node.getPath(), offset === 0 ? 0 : node.getLength())
        } else if (node.isList()) {
          
          if (offset === 0) {
            let item = node.getItemAt(0);
            return new Coordinate(item.getPath(), 0)
          } else {
            let item = this.get(last$1(node.items));
            return new Coordinate(item.getPath(), item.getLength())
          }
        }
      }
      return new Coordinate(path, offset)
    }
  }

  Document.prototype._isDocument = true;

  class XMLParentNodeHook {
    constructor (doc) {
      this.doc = doc;
      
      this.parents = {};
      doc.data.on('operation:applied', this._onOperationApplied, this);
    }

    _onOperationApplied (op) {
      const doc = this.doc;
      const parents = this.parents;
      let node = doc.get(op.path[0]);
      switch (op.type) {
        case 'create': {
          if (node._childNodes) {
            _setParent(node, node._childNodes);
          }
          _setRegisteredParent(node);
          break
        }
        case 'update': {
          
          
          let update = op.diff;
          if (op.path[1] === '_childNodes') {
            if (update.isInsert()) {
              _setParent(node, update.getValue());
            } else if (update.isDelete()) {
              _setParent(null, update.getValue());
            }
          }
          break
        }
        case 'set': {
          if (op.path[1] === '_childNodes') {
            _setParent(null, op.getOldValue());
            _setParent(node, op.getValue());
          }
          break
        }
        default:
          
      }

      function _setParent (parent, ids) {
        if (ids) {
          if (isArray(ids)) {
            ids.forEach(_set);
          } else {
            _set(ids);
          }
        }
        function _set (id) {
          
          
          
          parents[id] = parent;
          let child = doc.get(id);
          if (child) {
            child.parentNode = parent;
          }
        }
      }
      function _setRegisteredParent (child) {
        let parent = parents[child.id];
        if (parent) {
          child.parentNode = parent;
        }
      }
    }

    static register (doc) {
      return new XMLParentNodeHook(doc)
    }
  }

  class XMLEditingInterface extends EditingInterface {
    find (cssSelector) {
      return this.getDocument().find(cssSelector)
    }

    findAll (cssSelector) {
      return this.getDocument().findAll(cssSelector)
    }

    createElement (...args) {
      return this.getDocument().createElement(...args)
    }
  }

  class XMLDocument extends Document {
    _initialize () {
      this.nodeFactory = new DocumentNodeFactory(this);
      this.data = new IncrementalData(this.schema, this.nodeFactory);
      
      this.addIndex('type', new PropertyIndex('type'));
      
      this.addIndex('annotations', new AnnotationIndex());
      XMLParentNodeHook.register(this);
    }

    toXML () {
      let dom = DefaultDOMElement.createDocument('xml');
      dom.setDoctype(...this.getDocTypeParams());
      let xml = dom.createProcessingInstruction('xml', 'version="1.0" encoding="UTF-8"');
      dom.insertAt(0, xml);
      let rootElement = this.getRootNode().toXML();
      dom.append(rootElement);
      return dom
    }

    getDocTypeParams () {
      
      throw new Error('This method is abstract')
    }

    getXMLSchema () {
      
      throw new Error('This method is abstract')
    }

    getRootNode () {
      
      throw new Error('This method is abstract')
    }

    
    getDocTypeAsString () {
      return new Error('This method is abstract')
    }

    createEditingInterface () {
      return new XMLEditingInterface(this)
    }

    find (cssSelector) {
      return this.getRootNode().find(cssSelector)
    }

    findAll (cssSelector) {
      return this.getRootNode().findAll(cssSelector)
    }

    createElement (tagName, data) {
      let node = this.create(Object.assign({
        id: uuid(tagName),
        type: tagName
      }, data));
      return node
    }

    getElementSchema (type) {
      return this.getXMLSchema().getElementSchema(type)
    }

    _validateChange (change) {
      let changed = {};
      let deleted = [];
      change.ops.forEach((op) => {
        switch (op.type) {
          case 'delete': {
            deleted.push(op.val.id);
            break
          }
          case 'create': {
            changed[op.val.id] = true;
            break
          }
          default: {
            if (op.path) {
              changed[op.path[0]] = true;
            }
          }
        }
      });
      
      deleted.forEach(id => delete changed[id]);

      const xmlSchema = this.getXMLSchema();
      let errors = [];
      Object.keys(changed).forEach((id) => {
        let node = this.get(id);
        let res = xmlSchema.validateElement(node);
        if (!res.ok) {
          errors = errors.concat(res.errors);
        }
      });
      return {
        ok: errors.length === 0,
        errors
      }
    }

    
    _apply (documentChange) {
      
      if (!documentChange.ops) return

      let created = {};
      let deleted = {};
      let updated = {};

      const doc = this;

      function _recordUpdate (id, op) {
        let record = updated[id];
        if (!record) {
          record = updated[id] = { ops: [] };
        }
        if (op) record.ops.push(op);
        return record
      }

      
      function _checkAnnotation (op) {
        
        switch (op.type) {
          case 'create':
          case 'delete': {
            const annoData = op.val;
            if (annoData.hasOwnProperty('start')) {
              updated[annoData.start.path] = true;
              let node = doc.get(annoData.start.path[0]);
              if (node) {
                _recordUpdate(node.id, op);
              }
            }
            if (annoData.hasOwnProperty('end')) {
              updated[annoData.end.path] = true;
              let node = doc.get(annoData.start.path[0]);
              if (node) {
                _recordUpdate(node.id, op);
              }
            }
            break
          }
          case 'update':
          case 'set': {
            let anno = doc.get(op.path[0]);
            if (anno) {
              if (anno.isPropertyAnnotation()) {
                updated[anno.start.path] = true;
                let node = doc.get(anno.start.path[0]);
                if (node) {
                  _recordUpdate(node.id, op);
                }
              } else if (anno.isContainerAnnotation()) ;
            }
            break
          }
          default:
            
        }
      }

      documentChange.ops.forEach((op) => {
        
        
        
        switch (op.type) {
          case 'create': {
            created[op.path[0]] = op.val;
            break
          }
          case 'delete': {
            let node = this.get(op.path[0]);
            deleted[node.id] = node;
            break
          }
          case 'update':
          case 'set': {
            updated[op.path] = true;
            let node = this.get(op.path[0]);
            _recordUpdate(node.id, op);
            break
          }
          default:
            
        }
        
        _checkAnnotation(op);

        this._applyOp(op);
      });
      
      Object.keys(deleted).forEach((id) => {
        delete updated[id];
      });
      documentChange.created = created;
      documentChange.updated = updated;
      documentChange.deleted = deleted;
      documentChange._extracted = true;
    }
  }

  XMLDocument.prototype._isXMLDocument = true;

  function findParentDOMElement (nativeEl) {
    while (nativeEl) {
      let el = DefaultDOMElement.unwrap(nativeEl);
      if (el) return el
      nativeEl = nativeEl.parentNode;
    }
  }

  function findParent (el, selector) {
    while (el) {
      if (el.is(selector)) return el
      el = el.getParent();
    }
  }

  function stop (event) {
    event.stopPropagation();
  }

  function stopAndPrevent (event) {
    event.stopPropagation();
    event.preventDefault();
  }

  function walk (el, cb) {
    _walk(el, cb, 0);
  }

  function _walk (el, cb, level) {
    cb(el, level);
    if (el.getChildCount() > 0) {
      let it = el.getChildNodeIterator();
      while (it.hasNext()) {
        _walk(it.next(), cb, level + 1);
      }
    }
  }

  function isRightButton (event) {
    let isRightButton = false;
    if ('which' in event) {
      isRightButton = (event.which === 3);
    } else if ('button' in event) {
      isRightButton = (event.button === 2);
    }
    return isRightButton
  }

  function getBoundingRect (el) {
    let _rect = el.getNativeElement().getBoundingClientRect();
    return {
      top: _rect.top,
      left: _rect.left,
      height: _rect.height,
      width: _rect.width
    }
  }

  function getBoundingRectForRects (...rects) {
    let top, left, bottom, right;
    if (rects.length > 0) {
      let first = rects[0];
      top = first.top;
      left = first.left;
      bottom = top + first.height;
      right = left + first.width;
      for (let i = 1; i < rects.length; i++) {
        let r = rects[i];
        top = Math.min(top, r.top);
        left = Math.min(left, r.left);
        bottom = Math.max(bottom, r.top + r.height);
        right = Math.max(right, r.left + r.width);
      }
    }
    return {
      top,
      left,
      right,
      bottom,
      height: bottom - top,
      width: right - left
    }
  }

  function isXInside (x, rect) {
    return x >= rect.left && x <= rect.left + rect.width
  }

  function isYInside (y, rect) {
    return y >= rect.top && y <= rect.top + rect.height
  }

  var domHelpers = /*#__PURE__*/Object.freeze({
    findParentDOMElement: findParentDOMElement,
    findParent: findParent,
    stop: stop,
    stopAndPrevent: stopAndPrevent,
    walk: walk,
    isRightButton: isRightButton,
    getBoundingRect: getBoundingRect,
    getBoundingRectForRects: getBoundingRectForRects,
    isXInside: isXInside,
    isYInside: isYInside
  });

  function encodeXMLEntities (str) {
    return String(str)
      .replace(/&/g, '&amp;')
      .replace(/</g, '&lt;')
      .replace(/>/g, '&gt;')
      .replace(/"/g, '&quot;')
  }

  class DOMExporter {
    constructor (config, context) {
      this.context = context || {};
      if (!config.converters) {
        throw new Error('config.converters is mandatory')
      }
      if (!config.converters._isRegistry) {
        this.converters = new Registry();
        config.converters.forEach(function (Converter) {
          let converter = isFunction$1(Converter) ? new Converter() : Converter;
          if (!converter.type) {
            console.error('Converter must provide the type of the associated node.', converter);
            return
          }
          this.converters.add(converter.type, converter);
        }.bind(this));
      } else {
        this.converters = config.converters;
      }

      this.state = {
        doc: null
      };
      this.config = config;
      
      
      this._elementFactory = config.elementFactory;
      if (!this._elementFactory) {
        throw new Error("'elementFactory' is mandatory")
      }
      this.$$ = this.createElement.bind(this);
    }

    exportDocument (doc) {
      
      
      
      
      return this.convertDocument(doc)
    }

    
    convertDocument(doc) { 
      throw new Error('This method is abstract')
    }

    convertContainer (container) {
      if (!container) {
        throw new Error('Illegal arguments: container is mandatory.')
      }
      const doc = container.getDocument();
      this.state.doc = doc;
      let elements = [];
      container.getContent().forEach((id) => {
        const node = doc.get(id);
        const nodeEl = this.convertNode(node);
        elements.push(nodeEl);
      });
      return elements
    }

    convertNode (node) {
      if (isString(node)) {
        
        node = this.state.doc.get(node);
      } else {
        this.state.doc = node.getDocument();
      }
      var converter = this.getNodeConverter(node);
      
      
      if (node.isPropertyAnnotation() && (!converter || !converter.export)) {
        return this._convertPropertyAnnotation(node)
      }
      if (!converter) {
        converter = this.getDefaultBlockConverter();
      }
      var el;
      if (converter.tagName) {
        el = this.$$(converter.tagName);
      } else {
        el = this.$$('div');
      }
      el.attr(this.config.idAttribute, node.id);
      if (converter.export) {
        el = converter.export(node, el, this) || el;
      } else {
        el = this.getDefaultBlockConverter().export(node, el, this) || el;
      }
      return el
    }

    convertProperty (doc, path, options) {
      this.initialize(doc, options);
      var wrapper = this.$$('div')
        .append(this.annotatedText(path));
      return wrapper.innerHTML
    }

    annotatedText (path) {
      var doc = this.state.doc;
      var text = doc.get(path);
      var annotations = doc.getIndex('annotations').get(path);
      return this._annotatedText(text, annotations)
    }

    getNodeConverter (node) {
      return this.converters.get(node.type)
    }

    getDefaultBlockConverter () {
      throw new Error('This method is abstract.')
    }

    getDefaultPropertyAnnotationConverter () {
      throw new Error('This method is abstract.')
    }

    getDocument () {
      return this.state.doc
    }

    createElement (str) {
      return this._elementFactory.createElement(str)
    }

    _annotatedText (text, annotations) {
      var self = this;

      var annotator = new Fragmenter();
      annotator.onText = function (context, text) {
        context.children.push(encodeXMLEntities(text));
      };
      annotator.onEnter = function (fragment) {
        var anno = fragment.node;
        return {
          annotation: anno,
          children: []
        }
      };
      annotator.onExit = function (fragment, context, parentContext) {
        var anno = context.annotation;
        var converter = self.getNodeConverter(anno);
        if (!converter) {
          converter = self.getDefaultPropertyAnnotationConverter();
        }
        var el;
        if (converter.tagName) {
          el = this.$$(converter.tagName);
        } else {
          el = this.$$('span');
        }
        el.attr(this.config.idAttribute, anno.id);
        el.append(context.children);
        if (converter.export) {
          el = converter.export(anno, el, self) || el;
        }
        parentContext.children.push(el);
      }.bind(this);
      var wrapper = { children: [] };
      annotator.start(wrapper, text, annotations);
      return wrapper.children
    }

    
    _convertPropertyAnnotation (anno) {
      
      var wrapper = this.$$('div').append(this.annotatedText(anno.path));
      var el = wrapper.find('[' + this.config.idAttribute + '="' + anno.id + '"]');
      return el
    }
  }

  class HTMLExporter extends DOMExporter {
    constructor (config, context) {
      super(_defaultConfig(config), context);
    }

    exportDocument (doc) {
      let htmlEl = DefaultDOMElement.parseHTML('<html><head></head><body></body></html>');
      return this.convertDocument(doc, htmlEl)
    }

    getDefaultBlockConverter () {
      return defaultBlockConverter 
    }

    getDefaultPropertyAnnotationConverter () {
      return defaultAnnotationConverter 
    }
  }

  function _defaultConfig (config) {
    config = Object.assign({
      idAttribute: 'data-id'
    }, config);
    if (!config.elementFactory) {
      config.elementFactory = DefaultDOMElement.createDocument('html');
    }
    return config
  }

  const defaultAnnotationConverter = {
    tagName: 'span',
    export: function (node, el) {
      el.tagName = 'span';
      el.attr('data-type', node.type);
      var properties = node.toJSON();
      forEach(properties, function (value, name) {
        if (name === 'id' || name === 'type') return
        if (isString(value) || isNumber(value) || isBoolean(value)) {
          el.attr('data-' + name, value);
        }
      });
    }
  };

  const defaultBlockConverter = {
    export: function (node, el, converter) {
      el.attr('data-type', node.type);
      var properties = node.toJSON();
      forEach(properties, function (value, name) {
        if (name === 'id' || name === 'type') {
          return
        }
        var prop = converter.$$('div').attr('property', name);
        if (node.getPropertyType(name) === 'string' && value) {
          prop.append(converter.annotatedText([node.id, name]));
        } else {
          prop.text(value);
        }
        el.append(prop);
      });
    }
  };

  class ClipboardExporter extends HTMLExporter {
    
    exportDocument (doc) {
      this.state.doc = doc;
      let html;
      let elements = this.convertDocument(doc);
      
      if (elements.length === 1 && elements[0].attr('data-id') === TEXT_SNIPPET_ID) {
        html = elements[0].innerHTML;
      } else {
        html = elements.map(function (el) {
          return el.outerHTML
        }).join('');
      }
      let jsonConverter = new JSONConverter();
      let jsonStr = JSON.stringify(jsonConverter.exportDocument(doc));
      let substanceContent = `<script id="substance-clipboard" type="application/json">${jsonStr}</script>`;
      return '<html><head>' + substanceContent + '</head><body>' + html + '</body></html>'
    }

    
    convertDocument (doc) {
      let content = doc.get(SNIPPET_ID);
      if (!content) {
        throw new Error('Illegal clipboard document: could not find container "' + SNIPPET_ID + '"')
      }
      return this.convertContainer(content)
    }
  }

  class HTMLImporter extends DOMImporter {
    constructor (config) {
      super(Object.assign({ idAttribute: 'data-id' }, config));

      
      this._el = DefaultDOMElement.parseHTML('<html></html>');
    }

    importDocument (html) {
      this.reset();
      var parsed = DefaultDOMElement.parseHTML(html);
      this.convertDocument(parsed);
      return this.state.doc
    }

    
    convertDocument(documentEl) { 
      throw new Error('This method is abstract')
    }
  }

  const INLINENODES = ['a', 'b', 'big', 'i', 'small', 'tt', 'abbr', 'acronym', 'cite', 'code', 'dfn', 'em', 'kbd', 'strong', 'samp', 'time', 'var', 'bdo', 'br', 'img', 'map', 'object', 'q', 'script', 'span', 'sub', 'sup', 'button', 'input', 'label', 'select', 'textarea'].reduce((m, n) => { m[n] = true; return m }, {});


  class ClipboardImporter extends HTMLImporter {
    constructor (config) {
      super(_withCatchAllConverter(config));
      
      this.IGNORE_DEFAULT_WARNINGS = true;

      Object.assign(config, {
        trimWhitespaces: true,
        REMOVE_INNER_WS: true
      });

      this.editorOptions = config.editorOptions;
    }

    
    importDocument (html) {
      if (platform.isWindows) {
        
        
        let match = /<!--StartFragment-->(.*)<!--EndFragment-->/.exec(html);
        if (match) {
          html = match[1];
        }
      }

      
      
      
      if (html.search(/script id=.substance-clipboard./) >= 0) {
        let htmlDoc = DefaultDOMElement.parseHTML(html);
        let substanceData = htmlDoc.find('#substance-clipboard');
        if (substanceData) {
          let jsonStr = substanceData.textContent;
          try {
            return this.importFromJSON(jsonStr)
          } finally {
            
          }
        }
      }

      if (this.editorOptions && this.editorOptions['forcePlainTextPaste']) {
        return null
      }

      let htmlDoc = DefaultDOMElement.parseHTML(html);
      let generatorMeta = htmlDoc.find('meta[name="generator"]');
      let xmnlsw = htmlDoc.find('html').getAttribute('xmlns:w');
      if (generatorMeta) {
        let generator = generatorMeta.getAttribute('content');
        if (generator.indexOf('LibreOffice') > -1) this._isLibreOffice = true;
      } else if (xmnlsw) {
        if (xmnlsw.indexOf('office:word') > -1) {
          this._isMicrosoftWord = true;
          
          
          
          let match = /<!--StartFragment-->([\s\S]*)<!--EndFragment-->/.exec(html);
          if (match) {
            htmlDoc = DefaultDOMElement.parseHTML(match[1]);
          }
        }
      } else if (html.indexOf('docs-internal-guid') > -1) {
        this._isGoogleDoc = true;
      }

      let body = htmlDoc.find('body');
      body = this._sanitizeBody(body);
      if (!body) {
        console.warn('Invalid HTML.');
        return null
      }
      this._wrapIntoParagraph(body);
      this.reset();
      this.convertBody(body);
      const doc = this.state.doc;
      return doc
    }

    _sanitizeBody (body) {
      
      body.findAll('meta').forEach(el => el.remove());

      
      
      if (this._isLibreOffice || this._isMicrosoftWord) {
        let bodyHtml = body.getInnerHTML();
        body.setInnerHTML(bodyHtml.replace(/\r\n|\r|\n/g, ' '));
      } else if (this._isGoogleDoc) {
        body = this._fixupGoogleDocsBody(body);
      }

      return body
    }

    _fixupGoogleDocsBody (body) {
      if (!body) return
      
      
      
      
      let bold = body.find('b');
      if (bold && /^docs-internal/.exec(bold.id)) {
        body = bold;
      }

      body.findAll('span').forEach(span => {
        
        
        
        
        
        
        
        
        let nodeTypes = [];
        if (span.getStyle('font-weight') === '700') nodeTypes.push('b');
        if (span.getStyle('font-style') === 'italic') nodeTypes.push('i');
        if (span.getStyle('vertical-align') === 'super') nodeTypes.push('sup');
        if (span.getStyle('vertical-align') === 'sub') nodeTypes.push('sub');
        
        span.removeAttribute('style');

        function createInlineNodes (parentEl, isRoot) {
          if (nodeTypes.length > 0) {
            let el = parentEl.createElement(nodeTypes[0]);
            if (nodeTypes.length === 1) el.append(span.textContent);
            if (isRoot) {
              parentEl.replaceChild(span, el);
            } else {
              parentEl.appendChild(el);
            }
            nodeTypes.shift();
            createInlineNodes(el);
          }
        }
        createInlineNodes(span.getParent(), true);
      });

      let tags = ['b', 'i', 'sup', 'sub'];

      tags.forEach(tag => {
        body.findAll(tag).forEach(el => {
          
          
          let previousSiblingEl = el.getPreviousSibling();
          if (previousSiblingEl && el.tagName === previousSiblingEl.tagName) {
            let parentEl = el.getParent();
            let newEl = parentEl.createElement(tag);
            newEl.setInnerHTML(previousSiblingEl.getInnerHTML() + el.getInnerHTML());
            parentEl.replaceChild(el, newEl);
            parentEl.removeChild(previousSiblingEl);
          }

          
          
          
          
          if (previousSiblingEl && previousSiblingEl.tagName && el.getChildCount() > 0 && el.getChildAt(0).tagName === previousSiblingEl.tagName) {
            let parentEl = el.getParent();
            let childEl = el.getChildAt(0);
            let newEl = parentEl.createElement(previousSiblingEl.tagName);
            let newChildEl = newEl.createElement(tag);
            newChildEl.setTextContent(childEl.textContent);
            newEl.appendChild(newChildEl);
            parentEl.replaceChild(el, newEl);
          }
        });
      });

      return body
    }

    _wrapIntoParagraph (body) {
      let childNodes = body.getChildNodes();
      let shouldWrap = false;
      for (let i = 0; i < childNodes.length; i++) {
        const c = childNodes[i];
        if (c.isTextNode()) {
          if (!(/^\s+$/.exec(c.textContent))) {
            shouldWrap = true;
            break
          }
        } else if (INLINENODES[c.tagName]) {
          shouldWrap = true;
          break
        }
      }
      if (shouldWrap) {
        let p = body.createElement('p');
        p.append(childNodes);
        body.append(p);
      }
    }

    importFromJSON (jsonStr) {
      this.reset();
      let doc = this.getDocument();
      let jsonData = JSON.parse(jsonStr);
      let converter = new JSONConverter();
      converter.importDocument(doc, jsonData);
      return doc
    }

    
    convertBody (body) {
      this.convertContainer(body.childNodes, SNIPPET_ID);
    }

    
    _createDocument () {
      let emptyDoc = super._createDocument();
      return emptyDoc.createSnippet()
    }
  }

  function _withCatchAllConverter (config) {
    config = Object.assign({}, config);
    let defaultTextType = config.schema.getDefaultTextType();
    if (defaultTextType) {
      config.converters = config.converters.concat([{
        type: defaultTextType,
        matchElement: function (el) { return el.is('div') },
        import: function (el, node, converter) {
          node.content = converter.annotatedText(el, [node.id, 'content']);
        }
      }]);
    }
    return config
  }

  class ContainerAdapter extends Container {
    constructor (doc, path) {
      super(doc, { id: String(path) });
      this.document = doc;
      this.path = path;

      
      
      doc.data.nodes[this.id] = this;
    }

    getContentPath () {
      return this.path
    }

    get nodes () {
      return this.document.get(this.path)
    }
  }

  ContainerAdapter.prototype._isDocumentNode = false;
  ContainerAdapter.prototype._isContainer = false;

  class FileNode extends DocumentNode {
    getUrl () {
      if (this.proxy) {
        return this.proxy.getUrl()
      } else {
        
        console.warn('No file proxy attached to ', this.id);
        return ''
      }
    }

    setProxy (proxy) {
      this.proxy = proxy;
    }
  }

  FileNode.type = 'file';

  FileNode.schema = {
    url: { type: 'string', optional: true },
    fileType: { type: 'string', optional: true },
    mimeType: { type: 'string', optional: true },
    sourceFile: { type: 'object', optional: true }
  };

  FileNode.prototype._isFileNode = true;
  FileNode._isFileNode = true;

  class InlineNode extends PropertyAnnotation {}

  InlineNode.prototype._isInlineNode = true;

  InlineNode.isInline = true;

  class Range {
    constructor (start, end, reverse, containerId, surfaceId) {
      
      if (arguments[0] === 'SKIP') return
      if (arguments.length === 1 && isPlainObject(arguments[0])) {
        let data = arguments[0];
        this.start = data.start;
        this.end = data.end;
        this.reverse = Boolean(data.reverse);
        this.containerId = data.containerId;
        this.surfaceId = data.surfaceId;
      } else {
        this.start = start;
        this.end = end;
        this.reverse = Boolean(reverse);
        this.containerId = containerId;
        this.surfaceId = surfaceId;
      }
    }

    isCollapsed () {
      return this.start.equals(this.end)
    }

    equals (other) {
      if (this === other) return true
      else {
        return (
          this.containerId === other.containerId &&
          this.start.equals(other.start) &&
          this.end.equals(other.end)
        )
      }
    }

    isReverse () {
      return this.reverse
    }

    toString () {
      let str = [this.start.toString(), '->', this.end.toString()];
      if (this.isReverse()) {
        str.push('[reverse]');
      }
      if (this.containerId) {
        str.push('[container=' + this.containerId + ']');
      }
      if (this.surfaceId) {
        str.push('[surface=' + this.surfaceId + ']');
      }
      return str.join('')
    }
  }

  Range.prototype._isRange = true;

  class AbstractClipboard {
    constructor (configurator) {
      let schema = configurator.getSchema();
      let converterRegistry = configurator.converterRegistry;
      let htmlConverters = [];
      if (converterRegistry && converterRegistry.contains('html')) {
        htmlConverters = converterRegistry.get('html').values() || [];
      }
      
      
      this._config = {
        schema: schema,
        DocumentClass: schema.getDocumentClass(),
        converters: htmlConverters,
        editorOptions: configurator.getEditorOptions()
      };
    }

    
    onCopy (event) {
      
      let clipboardData = this._copy();
      substanceGlobals._clipboardData = event.clipboardData;

      if (event.clipboardData && clipboardData.doc) {
        event.preventDefault();
        
        event.clipboardData.setData('text/plain', clipboardData.text);
        event.clipboardData.setData('text/html', clipboardData.html);
      }
    }

    
    onCut (event) {
      
      event.preventDefault();
      
      this.onCopy(event);
      this._cut();
    }

    
    
    onPaste (event) {
      let clipboardData = event.clipboardData;

      let types = {};
      for (let i = 0; i < clipboardData.types.length; i++) {
        types[clipboardData.types[i]] = true;
      }
      

      event.preventDefault();
      event.stopPropagation();

      let plainText;
      let html;
      if (types['text/plain']) {
        plainText = clipboardData.getData('text/plain');
      }
      if (types['text/html']) {
        html = clipboardData.getData('text/html');
      }

      substanceGlobals.clipboardData = {
        text: plainText,
        html: html
      };

      

      
      
      if (html) {
        if (!this._pasteHtml(html, plainText)) {
          this._pastePlainText(plainText);
        }
      } else {
        this._pastePlainText(plainText);
      }
    }

    _getImporter () {
      return new ClipboardImporter(this._config)
    }

    _getExporter () {
      return new ClipboardExporter(this._config)
    }

    
    _copy () {
      throw new Error('This method is abstract.')
    }

    
    _cut () {
      throw new Error('This method is abstract.')
    }

    
    _pastePlainText(plainText) { 
      throw new Error('This method is abstract.')
    }

    
    _pasteHtml(html, text) { 
      throw new Error('This method is abstract.')
    }
  }

  function levenshtein (a, b) {
    let m = [];
    for (let i = 0; i <= b.length; i++) {
      m[i] = [i];
      if (i === 0) continue
      let ib = i - 1;
      for (let j = 0; j <= a.length; j++) {
        m[0][j] = j;
        if (j === 0) continue
        let jb = j - 1;
        m[i][j] = b.charAt(ib) === a.charAt(jb) ? m[ib][jb] : Math.min(
          m[ib][jb] + 1,
          m[i][jb] + 1,
          m[ib][j] + 1
        );
      }
    }
    return m
  }

  function diff (a, b, offset) {
    if (!isString(a) || !isString(b)) {
      throw new Error('Illegal arguments.')
    }
    offset = offset || 0;
    let changes = [];
    if (a || b) {
      if (!a && b) {
        changes.push({ type: 'insert', start: offset, text: b });
      } else if (a && !b) {
        changes.push({ type: 'delete', start: offset, end: offset + a.length });
      } else {
        let m = levenshtein(a, b);
        changes = _diff(a, b, m, offset);
      }
    }
    return changes
  }

  function _diff (a, b, m, offset) {
    let i = b.length;
    let j = a.length;
    let changes = [];
    let current;
    while (i > 0 && j > 0) {
      _next();
    }
    _commit();
    return changes

    function _next () {
      let d = m[i][j];
      let ib = i - 1;
      let jb = j - 1;
      
      if (m[ib][jb] < d) {
        if (current && current.type === 'replace') {
          current.start--;
          current.text.unshift(b.charAt(ib));
        } else {
          _commit();
          current = { type: 'replace', start: jb, end: j, text: [b.charAt(ib)] };
        }
        i--;
        j--;
      
      } else if (m[ib][j] < d) {
        if (current && current.type === 'insert') {
          current.start--;
          current.text.unshift(b.charAt(ib));
        } else {
          _commit();
          current = { type: 'insert', start: jb, text: [b.charAt(ib)] };
        }
        i--;
      
      } else if (m[i][jb] < d) {
        if (current && current.type === 'delete') {
          current.start--;
        } else {
          _commit();
          current = { type: 'delete', start: jb, end: j };
        }
        j--;
      
      } else {
        _commit();
        i--;
        j--;
      }
    }

    function _commit () {
      if (current) {
        switch (current.type) {
          case 'insert':
            current.start += offset;
            current.text = current.text.join('');
            break
          case 'delete':
            current.start += offset;
            current.end += offset;
            break
          case 'replace':
            current.start += offset;
            current.end += offset;
            current.text = current.text.join('');
            break
          default:
            throw new Error('Invalid state')
        }
        changes.push(current);
        current = null;
      }
    }
  }

  class AnnotationComponent extends Component {
    
    didMount () {
      let node = this.props.node;
      node.on('highlighted', this.onHighlightedChanged, this);
    }

    
    dispose () {
      let node = this.props.node;
      node.off(this);
    }

    render ($$) {
      let el = $$(this.getTagName())
        .attr('data-id', this.props.node.id)
        .addClass(this.getClassNames());
      if (this.props.node.highlighted) {
        el.addClass('sm-highlighted');
      }
      el.append(this.props.children);
      return el
    }

    getClassNames () {
      return 'sc-' + this.props.node.type
    }

    onHighlightedChanged () {
      if (this.props.node.highlighted) {
        this.el.addClass('sm-highlighted');
      } else {
        this.el.removeClass('sm-highlighted');
      }
    }

    getTagName () {
      return 'span'
    }
  }

  class AbstractIsolatedNodeComponent extends Component {
    constructor (...args) {
      super(...args);

      this.name = this.props.node.id;
      this._state = { selectionFragment: null };

      this.handleAction('escape', this.escape);
      this.ContentClass = this._getContentClass(this.props.node);

      
      let useBlocker = platform.isFF || !this.ContentClass.noBlocker;
      this.blockingMode = useBlocker ? 'closed' : 'open';
    }

    getInitialState () {
      let selState = this.context.editorSession.getSelectionState();
      return this._deriveStateFromSelectionState(selState)
    }

    getChildContext () {
      return {
        parentSurfaceId: this.getId(),
        isolatedNodeComponent: this,
        
        
        surface: undefined
      }
    }

    didMount () {
      super.didMount();

      let editorSession = this.context.editorSession;
      editorSession.onRender('selection', this._onSelectionChanged, this);
    }

    dispose () {
      super.dispose.call(this);

      let editorSession = this.context.editorSession;
      editorSession.off(this);
    }

    renderContent ($$, node, options = {}) {
      let ComponentClass = this.ContentClass;
      if (!ComponentClass) {
        console.error('Could not resolve a component for type: ' + node.type);
        return $$(this.__elementTag)
      } else {
        let props = Object.assign({
          disabled: this.props.disabled,
          node: node,
          isolatedNodeState: this.state.mode,
          focused: (this.state.mode === 'focused')
        }, options);
        return $$(ComponentClass, props)
      }
    }

    getId () {
      
      
      if (!this._id) {
        this._id = this.context.parentSurfaceId + '/' + this.name;
      }
      return this._id
    }

    get id () { return this.getId() }

    getMode () {
      return this.state.mode
    }

    isOpen () {
      return this.blockingMode === 'open'
    }

    isClosed () {
      return this.blockingMode === 'closed'
    }

    isNotSelected () {
      return !this.state.mode
    }

    isSelected () {
      return this.state.mode === 'selected'
    }

    isCoSelected () {
      return this.state.mode === 'co-selected'
    }

    isFocused () {
      return this.state.mode === 'focused'
    }

    isCoFocused () {
      return this.state.mode === 'co-focused'
    }

    getParentSurface () {
      return this.context.surface
    }

    escape () {
      
      this.selectNode();
    }

    _onSelectionChanged () {
      let editorSession = this.context.editorSession;
      let newState = this._deriveStateFromSelectionState(editorSession.getSelectionState());
      if (!newState && this.state.mode) {
        this.extendState({ mode: null });
      } else if (newState && newState.mode !== this.state.mode) {
        this.extendState(newState);
      }
    }

    onKeydown (event) {
      
      
      
      
      if (event.keyCode === keys$1.ESCAPE && this.state.mode === 'focused') {
        event.stopPropagation();
        event.preventDefault();
        this.escape();
      }
    }

    _getContentClass (node) {
      let ComponentClass;
      
      ComponentClass = this.getComponent(node.type, true);
      
      if (!ComponentClass) {
        ComponentClass = Component;
      }
      return ComponentClass
    }

    _getSurface (selState) {
      let surface = selState.get('surface');
      if (surface === undefined) {
        let sel = selState.getSelection();
        if (sel && sel.surfaceId) {
          let surfaceManager = this.context.surfaceManager;
          surface = surfaceManager.getSurface(sel.surfaceId);
        } else {
          surface = null;
        }
        selState.set('surface', surface);
      }
      return surface
    }

    
    
    _getIsolatedNodes (selState) {
      let isolatedNodes = selState.get('isolatedNodes');
      if (!isolatedNodes) {
        let sel = selState.getSelection();
        isolatedNodes = [];
        if (sel && sel.surfaceId) {
          let surfaceManager = this.context.surfaceManager;
          let surface = surfaceManager.getSurface(sel.surfaceId);
          if (surface) {
            isolatedNodes = surface.getComponentPath().filter(comp => comp._isAbstractIsolatedNodeComponent);
          }
        }
        selState.set('isolatedNodes', isolatedNodes);
      }
      return isolatedNodes
    }

    _shouldConsumeEvent (event) {
      let comp = Component.unwrap(event.target);
      let isolatedNodeComponent = this._getIsolatedNode(comp);
      return (isolatedNodeComponent === this)
    }

    _getIsolatedNode (comp) {
      if (comp._isAbstractIsolatedNodeComponent) {
        return this
      } else if (comp.context.isolatedNodeComponent) {
        return comp.context.isolatedNodeComponent
      } else if (comp.context.surface) {
        return comp.context.surface.context.isolatedNodeComponent
      }
    }
  }

  AbstractIsolatedNodeComponent.prototype._isAbstractIsolatedNodeComponent = true;

  class InlineNodeComponent extends AbstractIsolatedNodeComponent {
    render ($$) {
      const node = this.props.node;
      const ContentClass = this.ContentClass;
      const state = this.state;

      let el = $$('span');
      el.addClass(this.getClassNames())
        .addClass('sc-inline-node')
        .addClass('sm-' + this.props.node.type)
        .attr('data-id', node.id)
        .attr('data-inline', '1');

      let disabled = this.isDisabled();

      if (state.mode) {
        el.addClass('sm-' + state.mode);
      } else {
        el.addClass('sm-not-selected');
      }

      if (!ContentClass.noStyle) {
        el.addClass('sm-default-style');
      }

      
      
      
      el.on('keydown', this.onKeydown);

      el.append(
        this.renderContent($$, node)
          .ref('content')
          .addClass('se-content')
      );

      if (disabled) {
        el.addClass('sm-disabled')
          .attr('contenteditable', false)
          .on('click', this.onClick);
      }

      
      
      
      
      
      
      
      el.attr('draggable', true);

      return el
    }

    isDisabled () {
      return !this.state.mode || ['co-selected', 'cursor'].indexOf(this.state.mode) > -1
    }

    getClassNames () {
      return ''
    }

    onClick (event) {
      if (!this._shouldConsumeEvent(event)) {
        return
      }
      this.selectNode();
    }

    selectNode () {
      
      let editorSession = this.context.editorSession;
      let surface = this.context.surface;
      let node = this.props.node;
      editorSession.setSelection({
        type: 'property',
        path: node.start.path,
        startOffset: node.start.offset,
        endOffset: node.end.offset,
        containerId: surface.getContainerId(),
        surfaceId: surface.id
      });
    }

    _getContentClass (node) {
      let ComponentClass;
      
      ComponentClass = this.getComponent(node.type, true);
      
      
      if (!ComponentClass) {
        ComponentClass = this.getComponent('unsupported-inline-node', true);
      }
      if (!ComponentClass) {
        console.error(`No component registered for inline node '${node.type}'.`);
        ComponentClass = StubInlineNodeComponent;
      }
      return ComponentClass
    }

    
    
    _deriveStateFromSelectionState (selState) {
      let surface = this._getSurface(selState);
      if (!surface) return null
      
      if (surface === this.context.surface) {
        let sel = selState.getSelection();
        let node = this.props.node;
        if (sel.isPropertySelection() && !sel.isCollapsed() && isEqual(sel.start.path, node.start.path)) {
          let nodeSel = node.getSelection();
          if (nodeSel.equals(sel)) {
            return { mode: 'selected' }
          }
          if (sel.contains(nodeSel)) {
            return { mode: 'co-selected' }
          }
        }
      }
      let isolatedNodeComponent = surface.context.isolatedNodeComponent;
      if (!isolatedNodeComponent) return null
      if (isolatedNodeComponent === this) {
        return { mode: 'focused' }
      }
      let isolatedNodes = this._getIsolatedNodes(selState);
      if (isolatedNodes.indexOf(this) > -1) {
        return { mode: 'co-focused' }
      }
      return null
    }
  }

  InlineNodeComponent.prototype._isInlineNodeComponent = true;

  class StubInlineNodeComponent extends Component {
    render ($$) {
      const node = this.props.node;
      return $$('span').text('???').attr('data-id', node.id).attr('data-type', node.type)
    }
  }

  class AnnotatedTextComponent extends Component {
    render ($$) {
      let el = this._renderContent($$)
        .addClass('sc-annotated-text')
        .css({ whiteSpace: 'pre-wrap' });
      return el
    }

    getPath () {
      return this.props.path
    }

    getText () {
      return this.getDocument().get(this.props.path) || ''
    }

    isEmpty () {
      return !(this.getText())
    }

    getAnnotations () {
      return this.getDocument().getIndex('annotations').get(this.props.path)
    }

    getDocument () {
      return this.props.doc || this.context.doc
    }

    _getTagName () {
      return this.props.tagName
    }

    _onDocumentChange (update) {
      if (update.change && update.change.updated[this.getPath()]) {
        this.rerender();
      }
    }

    _renderContent ($$) {
      let text = this.getText();
      let annotations = this.getAnnotations();
      let el = $$(this._getTagName() || 'span');
      if (annotations && annotations.length > 0) {
        let fragmenter = new Fragmenter({
          onText: this._renderTextNode.bind(this),
          onEnter: this._renderFragment.bind(this, $$),
          onExit: this._finishFragment.bind(this)
        });
        fragmenter.start(el, text, annotations);
      } else {
        el.append(text);
      }
      return el
    }

    _renderTextNode (context, text) {
      if (text && text.length > 0) {
        context.append(text);
      }
    }

    _renderFragment ($$, fragment) {
      let doc = this.getDocument();
      let componentRegistry = this.getComponentRegistry();
      let node = fragment.node;
      
      if (node.type === 'container-annotation-fragment') ; else if (node.type === 'container-annotation-anchor') ; else {
        let ComponentClass = componentRegistry.get(node.type) || AnnotationComponent;
        if (node.constructor.isInline &&
            
            !ComponentClass.prototype._isInlineNodeComponent &&
            
            !ComponentClass.isCustom) {
          ComponentClass = InlineNodeComponent;
        }
        let el = $$(ComponentClass, { doc: doc, node: node });
        return el
      }
    }

    _finishFragment (fragment, context, parentContext) {
      parentContext.append(context);
    }
  }

  class CursorComponent extends Component {
    render ($$) {
      
      let el = $$('span').addClass('se-cursor');
      
      
      
      el.append('\uFEFF');
      el.append($$('div').addClass('se-cursor-inner'));

      if (this.props.collaborator) {
        let collaboratorIndex = this.props.collaborator.colorIndex;
        el.addClass('sm-collaborator-' + collaboratorIndex);
      } else {
        el.addClass('sm-local-user');
      }

      return el
    }
  }

  class SelectionFragmentComponent extends Component {
    render ($$) {
      
      let el = $$('span').addClass('se-selection-fragment');
      if (this.props.collaborator) {
        let collaboratorIndex = this.props.collaborator.colorIndex;
        el.addClass('sm-collaborator-' + collaboratorIndex);
      } else {
        el.addClass('sm-local-user');
      }
      el.append(this.props.children);
      return el
    }
  }

  class TextPropertyComponent extends AnnotatedTextComponent {
    getInitialState () {
      const markersManager = this.context.markersManager;
      let path = this.getPath();
      let markers;
      if (markersManager) {
        
        markersManager.register(this);
        markers = markersManager.getMarkers(path, {
          surfaceId: this.getSurfaceId(),
          containerId: this.getContainerId()
        });
      } else {
        const doc = this.getDocument();
        markers = doc.getAnnotations(path);
      }
      return {
        markers: markers
      }
    }

    didMount () {
      
      
      
      
      
    }

    dispose () {
      if (this.context.markersManager) {
        this.context.markersManager.deregister(this);
      }
      if (this.domObserver) {
        this.domObserver.disconnect();
      }
    }

    render ($$) {
      let path = this.getPath();

      let el = this._renderContent($$)
        .addClass('sc-text-property')
        .attr({
          'data-path': path.join('.')
        })
        .css({
          'white-space': 'pre-wrap'
        });

      if (this.isEmpty()) {
        el.addClass('sm-empty');
        if (this.props.placeholder) {
          el.append(
            $$('span').addClass('se-placeholder').append(
              this.props.placeholder
            )
          );
        }
      }

      if (!this.props.withoutBreak) {
        el.append($$('br'));
      }

      return el
    }

    getAnnotations () {
      if (this.props.markers) {
        return this.state.markers.concat(this.props.markers)
      } else {
        return this.state.markers
      }
    }

    _renderFragment ($$, fragment) {
      let node = fragment.node;
      let id = node.id;
      let el;
      if (node.type === 'cursor') {
        el = $$(CursorComponent, { collaborator: node.collaborator });
      } else if (node.type === 'selection-fragment') {
        el = $$(SelectionFragmentComponent, { collaborator: node.collaborator });
      } else {
        el = super._renderFragment.apply(this, arguments);
        if (id) {
          el.ref(id + '@' + fragment.counter);
        }
      }
      el.attr('data-offset', fragment.pos);
      return el
    }

    _onDomMutations (mutations) {
      
      
      if (mutations.length === 2 && mutations[0].target === mutations[1].target) {
        let textEl = DefaultDOMElement.unwrap(mutations[0].target);
        if (textEl) {
          this._applyTextMutation(textEl, mutations[0].oldValue);
          return
        }
      }
      
      this.rerender();
    }

    _applyTextMutation (textEl, oldText) {
      
      let offset = _getCharPos(textEl, 0);
      let newText = textEl.textContent;
      let changes = diff(oldText, newText, offset);

      let editorSession = this.context.editorSession;
      let path = this.getPath();
      editorSession.transaction(function (tx) {
        changes.forEach(function (change) {
          
          if (change.type === 'replace') {
            tx.update(path, { type: 'delete', start: change.start, end: change.end });
            tx.update(path, { type: 'insert', start: change.start, text: change.text });
          } else {
            tx.update(path, change);
          }
        });
      });
    }

    getSurface () {
      return this.props.surface || this.context.surface
    }

    getSurfaceId () {
      let surface = this.getSurface();
      return surface ? surface.id : null
    }

    getContainerId () {
      let surface = this.getSurface();
      return surface ? surface.getContainerId() : null
    }

    isEditable () {
      const surface = this.getSurface();
      return surface ? surface.isEditable() : false
    }

    isReadonly () {
      const surface = this.getSurface();
      return surface ? surface.isReadonly() : true
    }

    getDOMCoordinate (charPos) {
      return this._getDOMCoordinate(this.el, charPos)
    }

    _finishFragment (fragment, context, parentContext) {
      context.attr('data-length', fragment.length);
      parentContext.append(context);
    }

    _getDOMCoordinate (el, charPos) {
      let l;
      let idx = 0;
      if (charPos === 0) {
        return {
          container: el.getNativeElement(),
          offset: 0
        }
      }
      for (let child = el.getFirstChild(); child; child = child.getNextSibling(), idx++) {
        if (child.isTextNode()) {
          l = child.textContent.length;
          if (l >= charPos) {
            return {
              container: child.getNativeElement(),
              offset: charPos
            }
          } else {
            charPos -= l;
          }
        } else if (child.isElementNode()) {
          let length = child.getAttribute('data-length');
          if (length) {
            l = parseInt(length, 10);
            if (l >= charPos) {
              
              if (child.attr('data-inline')) {
                let nextSibling = child.getNextSibling();
                if (nextSibling && nextSibling.isTextNode()) {
                  return {
                    container: nextSibling.getNativeElement(),
                    offset: 0
                  }
                } else {
                  return {
                    container: el.getNativeElement(),
                    offset: el.getChildIndex(child) + 1
                  }
                }
              }
              return this._getDOMCoordinate(child, charPos, idx)
            } else {
              charPos -= l;
            }
          } else {
            console.error('FIXME: Can not map to DOM coordinates.');
            return null
          }
        }
      }
    }
  }

  TextPropertyComponent.prototype._isTextPropertyComponent = true;




  TextPropertyComponent.getCoordinate = function (root, el, offset) {
    let context = _getPropertyContext(root, el, offset);
    if (!context) {
      return null
    }
    
    
    let charPos;
    if (el.parentNode && el.parentNode.is('.se-placeholder')) {
      charPos = 0;
    } else {
      
      
      
      charPos = _getCharPos(context.node, context.offset);
    }
    if (isNumber(charPos)) {
      let coor = new Coordinate(context.path, charPos);
      coor._comp = context.comp;
      return coor
    } else {
      return null
    }
  };

  function _getPropertyContext (root, node, offset) {
    let result = {
      comp: null,
      el: null,
      path: null,
      node: node,
      offset: offset
    };
    while (node && node !== root) {
      if (node.isElementNode()) {
        let comp = Component.unwrap(node);
        if (comp && comp._isTextPropertyComponent) {
          result.comp = comp;
          result.el = node;
          result.path = comp.getPath();
          return result
        }
        
        
        
        if (node.getAttribute('data-inline')) {
          result.node = node;
          if (offset > 0) {
            result.offset = 1;
          }
        }
      }
      node = node.getParent();
    }
    return null
  }

  function _getCharPos (node, offset) {
    let charPos = offset;
    let parent, childIdx;

    

    parent = node.getParent();
    if (node.isTextNode()) {
      
      if (node === parent.firstChild) {
        
        let parentPath = parent.getAttribute('data-path');
        let parentOffset = parent.getAttribute('data-offset');
        if (parentPath) {
          charPos = offset;
        
        } else if (parentOffset) {
          charPos = parseInt(parentOffset, 10) + offset;
        
        } else {
          charPos = _getCharPos(parent, 0) + offset;
        }
      } else {
        
        childIdx = parent.getChildIndex(node);
        charPos = _getCharPos(parent, childIdx) + offset;
      }
    } else if (node.isElementNode()) {
      let pathStr = node.getAttribute('data-path');
      let offsetStr = node.getAttribute('data-offset');
      
      
      if (pathStr) {
        charPos = _countCharacters(node, offset);
      
      
      } else if (offsetStr) {
        childIdx = parent.getChildIndex(node);
        charPos = parseInt(offsetStr, 10) + _countCharacters(node, offset);
      
      
      } else {
        childIdx = parent.getChildIndex(node);
        charPos = _getCharPos(parent, childIdx) + _countCharacters(node, offset);
      }
    } else {
      
      return null
    }
    return charPos
  }

  function _countCharacters (el, maxIdx) {
    let charPos = 0;
    
    if (el.getAttribute('data-inline')) {
      return maxIdx === 0 ? 0 : 1
    }
    let l = el.getChildCount();
    if (arguments.length === 1) {
      maxIdx = l;
    }
    maxIdx = Math.min(l, maxIdx);
    for (let i = 0, child = el.getFirstChild(); i < maxIdx; child = child.getNextSibling(), i++) {
      if (child.isTextNode()) {
        charPos += child.getTextContent().length;
      } else if (child.isElementNode()) {
        let length = child.getAttribute('data-length');
        if (child.getAttribute('data-inline')) {
          charPos += 1;
        } else if (length) {
          charPos += parseInt(length, 10);
        } else {
          charPos += _countCharacters(child);
        }
      }
    }
    return charPos
  }

  const BRACKET = 'X';
  const SLASH = '/'.charCodeAt(0);


  class IsolatedNodeComponent extends AbstractIsolatedNodeComponent {
    render ($$) {
      let node = this.props.node;
      let ContentClass = this.ContentClass;
      let disabled = this.props.disabled;

      
      let el = $$('div');
      el.addClass(this.getClassNames())
        .addClass('sc-isolated-node')
        .addClass('sm-' + this.props.node.type)
        .attr('data-id', node.id);
      if (disabled) {
        el.addClass('sm-disabled');
      }
      if (this.state.mode) {
        el.addClass('sm-' + this.state.mode);
      }
      if (!ContentClass.noStyle) {
        el.addClass('sm-default-style');
      }
      
      el.on('keydown', this.onKeydown);

      
      let shouldRenderBlocker = (
        this.blockingMode === 'closed' &&
        !this.state.unblocked
      );

      
      
      el.append(
        $$('div').addClass('se-bracket sm-left').ref('left')
          .append(BRACKET)
      );

      let content = this.renderContent($$, node, {
        disabled: this.props.disabled || shouldRenderBlocker
      }).ref('content');
      content.attr('contenteditable', false);

      el.append(content);
      el.append($$(Blocker).ref('blocker'));
      el.append(
        $$('div').addClass('se-bracket sm-right').ref('right')
          .append(BRACKET)
      );

      if (!shouldRenderBlocker) {
        el.addClass('sm-no-blocker');
        el.on('click', this.onClick)
          .on('dblclick', this.onDblClick);
      }
      el.on('mousedown', this._reserveMousedown, this);

      return el
    }

    getClassNames () {
      return ''
    }

    getContent () {
      return this.refs.content
    }

    selectNode () {
      
      let editorSession = this.context.editorSession;
      let surface = this.context.surface;
      let nodeId = this.props.node.id;
      editorSession.setSelection({
        type: 'node',
        nodeId: nodeId,
        containerId: surface.getContainerId(),
        surfaceId: surface.id
      });
    }

    
    
    onClick (event) {
      
      event.stopPropagation();
    }

    onDblClick (event) {
      
      event.stopPropagation();
    }

    grabFocus (event) {
      let content = this.refs.content;
      if (content.grabFocus) {
        content.grabFocus(event);
        return true
      }
    }

    
    
    _reserveMousedown (event) {
      if (event.__reserved__) ; else {
        
        event.__reserved__ = this;
      }
    }

    _deriveStateFromSelectionState (selState) {
      let surface = this._getSurface(selState);
      let newState = { mode: null, unblocked: null };
      if (!surface) return newState
      
      if (surface === this.context.surface) {
        let sel = selState.getSelection();
        let nodeId = this.props.node.id;
        if (sel.isNodeSelection() && sel.getNodeId() === nodeId) {
          if (sel.isFull()) {
            newState.mode = 'selected';
            newState.unblocked = true;
          } else if (sel.isBefore()) {
            newState.mode = 'cursor';
            newState.position = 'before';
          } else if (sel.isAfter()) {
            newState.mode = 'cursor';
            newState.position = 'after';
          }
        }
        if (sel.isContainerSelection() && sel.containsNode(nodeId)) {
          newState.mode = 'co-selected';
        }
      } else {
        let sel = selState.getSelection();
        let surfaceId = sel.surfaceId;
        let id = this.getId();
        if (id.length < surfaceId.length && surfaceId.startsWith(id) && surfaceId.charCodeAt(id.length) === SLASH) {
          let tail = surfaceId.slice(id.length + 1);
          if (tail.indexOf('/') > 0) {
            newState.mode = 'co-focused';
            newState.unblocked = true;
          } else {
            newState.mode = 'focused';
            newState.unblocked = true;
          }
        }
      }
      return newState
    }
  }

  IsolatedNodeComponent.prototype._isIsolatedNodeComponent = true;

  IsolatedNodeComponent.prototype._isDisabled = IsolatedNodeComponent.prototype.isDisabled;

  IsolatedNodeComponent.getDOMCoordinate = function (comp, coor) {
    let { start, end } = IsolatedNodeComponent.getDOMCoordinates(comp);
    if (coor.offset === 0) return start
    else return end
  };

  IsolatedNodeComponent.getDOMCoordinates = function (comp) {
    const left = comp.refs.left;
    const right = comp.refs.right;
    return {
      start: {
        container: left.getNativeElement(),
        offset: 0
      },
      end: {
        container: right.getNativeElement(),
        offset: right.getChildCount()
      }
    }
  };

  IsolatedNodeComponent.getCoordinate = function (nodeEl, options) {
    let comp = Component.unwrap(nodeEl, 'strict').context.isolatedNodeComponent;
    let offset = null;
    if (options.direction === 'left' || nodeEl === comp.refs.left.el) {
      offset = 0;
    } else if (options.direction === 'right' || nodeEl === comp.refs.right.el) {
      offset = 1;
    }
    let coor;
    if (offset !== null) {
      coor = new Coordinate([comp.props.node.id], offset);
      coor._comp = comp;
    }
    return coor
  };

  class Blocker extends Component {
    render ($$) {
      return $$('div').addClass('sc-isolated-node-blocker')
        .attr('draggable', true)
        .attr('contenteditable', false)
        .on('click', this.onClick)
        .on('dblclick', this.onDblClick)
    }

    onClick (event) {
      if (event.target !== this.getNativeElement()) return
      
      event.stopPropagation();
      const comp = this._getIsolatedNodeComponent();
      comp.extendState({ mode: 'selected', unblocked: true });
      comp.selectNode();
    }

    onDblClick (event) {
      
      event.stopPropagation();
    }

    _getIsolatedNodeComponent () {
      return this.context.isolatedNodeComponent
    }
  }

  const DEBUG$1 = false;


  class DOMSelection {
    constructor (editor) {
      this.editor = editor;
      if (platform.inBrowser) {
        this.wRange = window.document.createRange();
      }
      
      
      this.state = { dom: null, model: null };
    }

    
    getSelection (options) {
      
      if (!platform.inBrowser) return
      let range = this.mapDOMSelection(options);
      let doc = this.editor.getDocument();
      
      return doc._createSelectionFromRange(range)
    }

    getSelectionForDOMRange (wrange) {
      let range = this.mapDOMRange(wrange);
      let doc = this.editor.getDocument();
      return doc._createSelectionFromRange(range)
    }

    
    mapDOMSelection (options) {
      let wSel = window.getSelection();
      let state = this.state;
      let range;
      
      
      if (DEBUG$1) console.info('DOM->Model: ', wSel.anchorNode, wSel.anchorOffset, wSel.focusNode, wSel.focusOffset);
      if (wSel.rangeCount === 0) return _null()
      let anchorNode = DefaultDOMElement.wrapNativeElement(wSel.anchorNode);
      if (wSel.isCollapsed) {
        let coor = this._getCoordinate(anchorNode, wSel.anchorOffset, options);
        if (!coor) return _null()
        range = _createRange({
          start: coor,
          end: coor
        });
      } else {
        let focusNode = DefaultDOMElement.wrapNativeElement(wSel.focusNode);
        range = this._getRange(anchorNode, wSel.anchorOffset, focusNode, wSel.focusOffset, options);
      }
      if (DEBUG$1) console.info('DOM->Model: range ', range ? range.toString() : null);
      state.model = range;
      return range

      function _null () {
        state.dom = null;
        state.model = null;
        return null
      }
    }

    
    setSelection (sel) {
      
      if (!platform.inBrowser) return
      let state = this.state;
      let wSel = window.getSelection();
      let wRange = this.wRange;
      if (!sel || sel.isNull()) return this.clear()
      
      let {start, end} = this.mapModelToDOMCoordinates(sel);
      if (!start) return this.clear()
      if (sel.isReverse()) {
        [start, end] = [end, start];
      }
      state.dom = {
        anchorNode: start.container,
        anchorOffset: start.offset,
        focusNode: end.container,
        focusOffset: end.offset
      };
      _set(state.dom);

      function _set ({anchorNode, anchorOffset, focusNode, focusOffset}) {
        wSel.removeAllRanges();
        wRange.setStart(anchorNode, anchorOffset);
        wRange.setEnd(anchorNode, anchorOffset);
        wSel.addRange(wRange);
        if (focusNode !== anchorOffset || focusOffset !== anchorOffset) {
          wSel.extend(focusNode, focusOffset);
        }
      }
    }

    mapModelToDOMCoordinates (sel) {
      if (DEBUG$1) console.info('Model->DOM: sel =', sel.toString());
      let rootEl;
      let surface = this.editor.getSurfaceManager().getSurface(sel.surfaceId);
      if (!surface) {
        console.warn('No surface:', sel.surfaceId);
        rootEl = this.editor.getElement();
      } else {
        rootEl = surface.el;
      }
      if (sel.isNull() || sel.isCustomSelection()) {
        return {}
      }

      let start, end;
      if (sel.isPropertySelection() || sel.isContainerSelection()) {
        start = this._getDOMCoordinate(rootEl, sel.start);
        if (!start) {
          console.warn('FIXME: selection seems to be invalid.');
          return {}
        }
        if (sel.isCollapsed()) {
          end = start;
        } else {
          end = this._getDOMCoordinate(rootEl, sel.end);
          if (!end) {
            console.warn('FIXME: selection seems to be invalid.');
            return {}
          }
        }
      } else if (sel.isNodeSelection()) {
        let comp = Component.unwrap(rootEl.find('*[data-id="' + sel.getNodeId() + '"]'));
        if (!comp) {
          console.error('Could not find component with id', sel.getNodeId());
          return {}
        }
        if (comp._isIsolatedNodeComponent) {
          let coors = IsolatedNodeComponent.getDOMCoordinates(comp, sel);
          start = coors.start;
          end = coors.end;
          
          
          
        } else {
          let _nodeEl = comp.el;
          start = {
            container: _nodeEl.getNativeElement(),
            offset: 0
          };
          end = {
            container: _nodeEl.getNativeElement(),
            offset: _nodeEl.getChildCount()
          };
        }
      }
      if (DEBUG$1) console.info('Model->DOM:', start.container, start.offset, end.container, end.offset, 'isReverse?', sel.isReverse());
      return {start, end}
    }

    _getDOMCoordinate (rootEl, coor) {
      let domCoor = null;
      let comp;
      if (coor.isNodeCoordinate()) {
        comp = Component.unwrap(rootEl.find('*[data-id="' + coor.getNodeId() + '"]'));
        if (comp) {
          if (comp._isIsolatedNodeComponent) {
            domCoor = IsolatedNodeComponent.getDOMCoordinate(comp, coor);
          } else {
            let domOffset = 0;
            if (coor.offset > 0) {
              domOffset = comp.getChildCount();
            }
            domCoor = {
              container: comp.getNativeElement(),
              offset: domOffset
            };
          }
        }
      } else {
        comp = Component.unwrap(rootEl.find('.sc-text-property[data-path="' + coor.path.join('.') + '"]'));
        if (comp) {
          domCoor = comp.getDOMCoordinate(coor.offset);
        }
      }
      return domCoor
    }

    
    mapDOMRange (wRange, options) {
      return this._getRange(
        DefaultDOMElement.wrapNativeElement(wRange.startContainer),
        wRange.startOffset,
        DefaultDOMElement.wrapNativeElement(wRange.endContainer),
        wRange.endOffset, options)
    }

    
    clear () {
      window.getSelection().removeAllRanges();
      this.state.dom = null;
      this.state.model = null;
    }

    collapse (dir) {
      let wSel = window.getSelection();
      let wRange;
      if (wSel.rangeCount > 0) {
        wRange = wSel.getRangeAt(0);
        wRange.collapse(dir === 'left');
        wSel.removeAllRanges();
        wSel.addRange(wRange);
      }
    }

    select (el) {
      let wSel = window.getSelection();
      let wRange = window.document.createRange();
      wRange.selectNode(el.getNativeElement());
      wSel.removeAllRanges();
      wSel.addRange(wRange);
    }

    extend (el, offset) {
      let wSel = window.getSelection();
      wSel.extend(el.getNativeElement(), offset);
    }

    setCursor (el, offset) {
      let wSel = window.getSelection();
      let wRange = window.document.createRange();
      wRange.setStart(el.getNativeElement(), offset);
      wSel.removeAllRanges();
      wSel.addRange(wRange);
    }

    
    _getRange (anchorNode, anchorOffset, focusNode, focusOffset, options = {}) {
      let isReverse = DefaultDOMElement.isReverse(anchorNode, anchorOffset, focusNode, focusOffset);
      let isCollapsed = (anchorNode === focusNode && anchorOffset === focusOffset);
      let start, end;
      if (isCollapsed) {
        start = end = this._getCoordinate(anchorNode, anchorOffset, options);
      } else {
        start = this._getCoordinate(anchorNode, anchorOffset, { direction: isReverse ? 'right' : 'left' });
        end = this._getCoordinate(focusNode, focusOffset, options);
      }
      if (start && end) {
        return _createRange({ start, end, isReverse })
      } else {
        return null
      }
    }

    
    _getCoordinate (nodeEl, offset, options = {}) {
      let coor = null;
      
      if (!coor) {
        coor = TextPropertyComponent.getCoordinate(this.editor.getElement(), nodeEl, offset);
      }
      let comp = Component.unwrap(nodeEl);
      if (!coor && comp) {
        
        if (comp.context.isolatedNodeComponent) {
          coor = IsolatedNodeComponent.getCoordinate(nodeEl, options);
        }
      }
      
      
      if (!coor) {
        
        if (comp && comp._isContainerEditor) {
          let childIdx = (offset === 0) ? 0 : offset - 1;
          let isBefore = (offset === 0);
          let container = comp.getContainer();
          let childNode = container.getNodeAt(childIdx);
          let childComp = comp.getChildAt(childIdx);
          coor = new Coordinate([childNode.id], isBefore ? 0 : 1);
          coor._comp = childComp;
        
        } else if (nodeEl.isElementNode() && nodeEl.getChildCount() > 0) {
          let child = (offset > 0) ? nodeEl.getChildAt(offset - 1) : nodeEl.firstChild;
          let prop;
          let childComp = Component.unwrap(child);
          if (childComp && childComp._isTextPropertyComponent) {
            prop = child;
          }
          
          if (prop) {
            coor = TextPropertyComponent.getCoordinate(nodeEl, prop, (offset > 0) ? prop.getChildCount() : 0);
          }
        }
      }
      return coor
    }
  }


  function _createRange ({start, end, isReverse}) {
    if (isReverse) {
      [start, end] = [end, start];
    }
    if (!start._comp || !end._comp) {
      console.error('FIXME: getCoordinate() should provide a component instance');
      return null
    }
    let surface = start._comp.context.surface;
    if (!surface) {
      console.error('FIXME: Editable components should have their surface in the context');
      return null
    }
    if (surface !== end._comp.context.surface) {
      console.error('Coordinates are within two different surfaces. Can not create a selection.');
      return null
    }
    return new Range(start, end, isReverse, surface.getContainerId(), surface.id)
  }

  class Command {
    
    constructor (config) {
      this.config = config || {};
      this.name = this.config.name;
      if (!this.name) {
        throw new Error("'name' is required")
      }
    }

    get isAsync () {
      return false
    }

    
    getName () {
      return this.name
    }

    
    getCommandState(params, context) { 
      throw new Error('Command.getCommandState() is abstract.')
    }

    
    execute(params, context) { 
      throw new Error('Command.execute() is abstract.')
    }

    
    isAnnotationCommand () {
      return false
    }

    
    isInsertCommand () {
      return false
    }

    
    isSwitchTypeCommand () {
      return false
    }

    _getEditorSession (params, context) {
      let editorSession = params.editorSession || context.editorSession;
      if (!editorSession) {
        throw new Error("'editorSession' is required.")
      }
      return editorSession
    }

    _getSelection (params) {
      let sel = params.selection || params.selectionState.getSelection();
      if (!sel) {
        throw new Error("'selection' is required.")
      }
      return sel
    }
  }

  Command.prototype._isCommand = true;

  class NodeComponent extends Component {
    didMount () {
      this.context.editorSession.onRender('document', this.rerender, this, { path: [this.props.node.id] });
    }

    dispose () {
      this.context.editorSession.off(this);
    }

    render ($$) {
      let tagName = this.getTagName();
      let el = $$(tagName)
        .attr('data-id', this.props.node.id)
        .addClass(this.getClassNames());
      return el
    }

    getTagName () {
      return 'div'
    }

    getClassNames () {
      return ''
    }

    rerender (...args) {
      
      if (this.props.node.isDisposed()) return
      super.rerender(...args);
    }
  }

  class Clipboard extends AbstractClipboard {
    constructor (editorSession) {
      super(editorSession.getConfigurator());

      this.editorSession = editorSession;
    }

    
    _copy () {
      let editorSession = this.getEditorSession();
      let sel = editorSession.getSelection();
      let doc = editorSession.getDocument();
      let clipboardDoc = null;
      let clipboardText = '';
      let clipboardHtml = '';
      let htmlExporter = this._getExporter();
      if (!sel.isCollapsed()) {
        clipboardText = getTextForSelection(doc, sel) || '';
        clipboardDoc = copySelection(doc, sel);
        clipboardHtml = htmlExporter.exportDocument(clipboardDoc);
      }
      return {
        doc: clipboardDoc,
        html: clipboardHtml,
        text: clipboardText
      }
    }

    _cut () {
      let editorSession = this.getEditorSession();
      editorSession.transaction((tx) => {
        tx.deleteSelection();
      });
    }

    
    _pastePlainText (plainText) {
      let editorSession = this.getEditorSession();
      editorSession.transaction(function (tx) {
        tx.paste(plainText);
      }, { action: 'paste' });
    }

    
    _pasteHtml (html, text) {
      let htmlImporter = this._getImporter();
      let content = htmlImporter.importDocument(html) || text;
      if (content) {
        let editorSession = this.getEditorSession();
        editorSession.transaction((tx) => {
          tx.paste(content);
        }, { action: 'paste' });
      }
      return true
    }

    getEditorSession () {
      return this.editorSession
    }
  }

  function startsWith (str, prefix) {
    if (!isString(str)) return false
    if (str.startsWith) return str.startsWith(prefix)
    if (!isString(prefix)) prefix = String(prefix);
    return str.slice(0, prefix.length) === prefix
  }

  class UnsupportedNodeComponent extends Component {
    render ($$) {
      return $$('pre')
        .addClass('content-node unsupported')
        .attr({
          'data-id': this.props.node.id,
          contentEditable: false
        })
        .append(
          JSON.stringify(this.props.node.properties, null, 2)
        )
    }
  }

  const BROWSER_DELAY = platform.isFF ? 1 : 0;


  class Surface extends Component {
    constructor (...args) {
      super(...args);

      this._initialize();
    }

    _initialize () {
      
      
      this.editorSession = this.props.editorSession || this.context.editorSession;
      if (!this.editorSession) {
        throw new Error('No EditorSession provided')
      }
      this.name = this.props.name;
      if (!this.name) {
        throw new Error('Surface must have a name.')
      }
      if (this.name.indexOf('/') > -1) {
        
        throw new Error("Surface.name must not contain '/'")
      }
      
      
      this._surfaceId = Surface.createSurfaceId(this);

      this.clipboard = new Clipboard(this.editorSession);

      this.domSelection = this.context.domSelection || new DOMSelection(this);

      this._state = {
        
        skipNextFocusEvent: false
      };
    }

    getChildContext () {
      return {
        surface: this,
        parentSurfaceId: this.getId(),
        doc: this.getDocument(),
        
        
        isolatedNodeComponent: null
      }
    }

    didMount () {
      const editorSession = this.getEditorSession();
      const surfaceManager = this.getSurfaceManager();
      if (surfaceManager) {
        surfaceManager.registerSurface(this);
      }
      editorSession.onRender('selection', this._onSelectionChanged, this);

      const globalEventHandler = editorSession.globalEventHandler;
      globalEventHandler.addEventListener('keydown', this._muteNativeHandlers, this);
    }

    dispose () {
      const editorSession = this.getEditorSession();
      const surfaceManager = this.getSurfaceManager();
      editorSession.off(this);
      if (surfaceManager) {
        surfaceManager.unregisterSurface(this);
      }

      const globalEventHandler = editorSession.globalEventHandler;
      globalEventHandler.removeEventListener('keydown', this._muteNativeHandlers);
    }

    didUpdate () {
      this._updateContentEditableState();
    }

    render ($$) {
      let tagName = this.props.tagName || 'div';
      let el = $$(tagName)
        .addClass('sc-surface')
        .attr('tabindex', 2)
        .attr('data-surface-id', this.id);

      if (!this.isDisabled()) {
        if (this.isEditable()) {
          
          el.on('keydown', this.onKeyDown);
          
          if (!platform.isIE) {
            el.on('compositionstart', this.onCompositionStart);
            el.on('compositionend', this.onCompositionEnd);
          }
          
          
          
          
          if (platform.inBrowser && window.TextEvent && !platform.isIE) {
            el.on('textInput', this.onTextInput);
          } else {
            el.on('keypress', this.onTextInputShim);
          }
          el.on('paste', this._onPaste);
          el.on('cut', this._onCut);
        }
        if (!this.isReadonly()) {
          
          el.on('mousedown', this.onMouseDown);
          el.on('contextmenu', this.onContextMenu);
          
          
          el.on('focus', this.onNativeFocus);
          el.on('blur', this.onNativeBlur);
        }
        el.on('copy', this._onCopy);
      }
      return el
    }

    renderNode ($$, node) {
      let componentRegistry = this.getComponentRegistry();
      let ComponentClass = componentRegistry.get(node.type);
      if (!ComponentClass) {
        console.error('Could not resolve a component for type: ' + node.type);
        ComponentClass = UnsupportedNodeComponent;
      }
      return $$(ComponentClass, this._extractNodeProps(node)).ref(node.id)
    }

    getComponentRegistry () {
      return this.context.componentRegistry || this.props.componentRegistry
    }

    getName () {
      return this.name
    }

    getId () {
      return this._surfaceId
    }

    getSurfaceId () {
      return this._surfaceId
    }

    isDisabled () {
      return this.props.disabled
    }

    isEditable () {
      return (this.props.editing === 'full' || this.props.editing === undefined)
    }

    isSelectable () {
      return (this.props.editing === 'selection' || this.props.editing === 'full')
    }

    isReadonly () {
      return this.props.editing === 'readonly'
    }

    getElement () {
      return this.el
    }

    getDocument () {
      return this.getEditorSession().getDocument()
    }

    getEditorSession () {
      return this.editorSession
    }

    getSurfaceManager () {
      return this.context.surfaceManager
    }

    isEnabled () {
      return !this.state.disabled
    }

    isContainerEditor () {
      return false
    }

    isCustomEditor () {
      return false
    }

    hasNativeSpellcheck () {
      return this.props.spellcheck === 'native'
    }

    getContainerId () {
      return null
    }

    focus () {
      if (this.getEditorSession().getFocusedSurface() !== this) {
        this.selectFirst();
      }
    }

    blur () {
      if (this.getEditorSession().getFocusedSurface() === this) {
        this.getEditorSession().setSelection(null);
      }
    }

    selectFirst () {
      throw new Error('This method is abstract.')
    }

    
    rerenderDOMSelection () {
      if (this.isDisabled()) return
      if (platform.inBrowser) {
        
        let sel = this.getEditorSession().getSelection();
        if (sel.surfaceId === this.getId()) {
          this.domSelection.setSelection(sel);
          
          
          
          const scrollPane = this.context.scrollPane;
          if (scrollPane) {
            this.context.scrollPane.onSelectionPositioned();
          }
        }
      }
    }

    getDomNodeForId (nodeId) {
      return this.el.getNativeElement().querySelector('*[data-id="' + nodeId + '"]')
    }

    

    
    onKeyDown (event) {
      if (!this._shouldConsumeEvent(event)) return
      

      
      if (event.key === 'Dead') return

      
      let custom = this.getEditorSession().keyboardManager.onKeydown(event);
      if (!custom) {
        
        switch (event.keyCode) {
          
          case keys$1.LEFT:
          case keys$1.RIGHT:
            return this._handleLeftOrRightArrowKey(event)
          case keys$1.UP:
          case keys$1.DOWN:
            return this._handleUpOrDownArrowKey(event)
          case keys$1.HOME:
          case keys$1.END:
            return this._handleHomeOrEndKey(event)
          case keys$1.PAGEUP:
          case keys$1.PAGEDOWN:
            return this._handlePageUpOrDownKey(event)
          
          case keys$1.ENTER:
            return this._handleEnterKey(event)
          case keys$1.TAB:
            return this._handleTabKey(event)
          case keys$1.BACKSPACE:
          case keys$1.DELETE:
            return this._handleDeleteKey(event)
          case keys$1.ESCAPE:
            return this._handleEscapeKey(event)
          case keys$1.SPACE:
            return this._handleSpaceKey(event)
          default:
            break
        }
      }
    }

    onTextInput (event) {
      if (!this._shouldConsumeEvent(event)) return
      
      event.preventDefault();
      event.stopPropagation();
      if (!event.data) return
      const editorSession = this.getEditorSession();
      let text = event.data;
      if (!editorSession.keyboardManager.onTextInput(text)) {
        editorSession.transaction((tx) => {
          tx.insertText(text);
        }, { action: 'type' });
      }
    }

    
    onCompositionStart (event) {
      if (!this._shouldConsumeEvent(event)) return
      
      
      
      
      
      
      
      if (event.data) {
        let l = event.data.length;
        let sel = this.getEditorSession().getSelection();
        if (sel.isPropertySelection() && sel.isCollapsed()) {
          
          let offset = sel.start.offset;
          this.getEditorSession().setSelection(sel.createWithNewRange(offset - l, offset));
        }
      }
    }

    onCompositionEnd (event) {
      if (!this._shouldConsumeEvent(event)) return
      
      
      
      if (platform.isFF) {
        event.preventDefault();
        event.stopPropagation();
        if (!event.data) return
        this._delayed(() => {
          let text = event.data;
          if (!this.getEditorSession().keyboardManager.onTextInput(text)) {
            this.getEditorSession().transaction((tx) => {
              tx.insertText(text);
            }, { action: 'type' });
          }
        });
      }
    }

    
    onTextInputShim (event) {
      if (!this._shouldConsumeEvent(event)) return
      
      if (
        
        event.which === 0 || event.charCode === 0 ||
        
        event.keyCode === keys$1.TAB || event.keyCode === keys$1.ESCAPE ||
        
        Boolean(event.metaKey) || (Boolean(event.ctrlKey) ^ Boolean(event.altKey))
      ) {
        return
      }
      let character = String.fromCharCode(event.which);
      if (!event.shiftKey) {
        character = character.toLowerCase();
      }
      event.preventDefault();
      event.stopPropagation();
      if (!this.getEditorSession().keyboardManager.onTextInput(character)) {
        if (character.length > 0) {
          this.getEditorSession().transaction((tx) => {
            tx.insertText(character);
          }, { action: 'type' });
        }
      }
    }

    
    
    
    
    
    onMouseDown (event) {
      if (!this._shouldConsumeEvent(event)) {
        
        return
      }

      
      
      
      
      
      if (event.__reserved__) {
        
        return
      } else {
        
        event.__reserved__ = this;
      }

      
      
      
      if (this.isEditable()) {
        this.el.setAttribute('contenteditable', true);
      }

      
      if (event.button !== 0) {
        return
      }

      
      if (!(platform.isIE && platform.version < 12) && event.detail >= 3) {
        let sel = this.getEditorSession().getSelection();
        if (sel.isPropertySelection()) {
          this._selectProperty(sel.path);
          event.preventDefault();
          event.stopPropagation();
          return
        } else if (sel.isContainerSelection()) {
          this._selectProperty(sel.startPath);
          event.preventDefault();
          event.stopPropagation();
          return
        }
      }
      
      
      
      
      this._state.skipNextFocusEvent = true;

      
      if (platform.inBrowser) {
        let documentEl = DefaultDOMElement.wrapNativeElement(window.document);
        documentEl.on('mouseup', this.onMouseUp, this, { once: true });
      }
    }

    onMouseUp (e) {
      
      
      
      
      
      e.stopPropagation();
      
      
      
      
      this._delayed(() => {
        let sel = this.domSelection.getSelection();
        this._setSelection(sel);
      });
    }

    
    
    
    onContextMenu (event) {
      if (!this._shouldConsumeEvent(event)) return
      let sel = this.domSelection.getSelection();
      this._setSelection(sel);
    }

    onNativeBlur () {
      
      let _state = this._state;
      _state.hasNativeFocus = false;
    }

    onNativeFocus () {
      
      let _state = this._state;
      _state.hasNativeFocus = true;
    }

    _onCopy (e) {
      this.clipboard.onCopy(e);
    }

    _onCut (e) {
      this.clipboard.onCut(e);
    }

    _onPaste (e) {
      this.clipboard.onPaste(e);
    }

    

    _onSelectionChanged (selection) {
      let newMode = this._deriveModeFromSelection(selection);
      if (this.state.mode !== newMode) {
        this.extendState({
          mode: newMode
        });
      }
    }

    
    _deriveModeFromSelection (sel) {
      if (!sel) return null
      let surfaceId = sel.surfaceId;
      let id = this.getId();
      let mode;
      if (startsWith(surfaceId, id)) {
        if (surfaceId.length === id.length) {
          mode = 'focused';
        } else {
          mode = 'co-focused';
        }
      }
      return mode
    }

    _updateContentEditableState () {
      
      
      
      
      
      let enableContenteditable = false;
      if (this.isEditable() && !this.props.disabled) {
        enableContenteditable = true;
        if (this.state.mode === 'co-focused') {
          let selState = this.context.editorSession.getSelectionState();
          let sel = selState.getSelection();
          let surface = this.getSurfaceManager().getSurface(sel.surfaceId);
          if (surface) {
            let isolatedNodeComponent = surface.context.isolatedNodeComponent;
            if (isolatedNodeComponent) {
              enableContenteditable = isolatedNodeComponent.isOpen();
            }
          }
        }
      }
      if (enableContenteditable) {
        this.el.setAttribute('contenteditable', true);
      } else {
        
        this.el.removeAttribute('contenteditable');
      }
    }

    _blur () {
      if (this.el) {
        this.el.blur();
      }
    }

    _focus () {
      if (this.isDisabled()) return
      
      
      
      if (platform.isFF) {
        this.domSelection.clear();
        this.el.getNativeElement().blur();
      }
      this._focusElement();
    }

    _focusElement () {
      this._state.hasNativeFocus = true;
      
      
      
      
      if (this.el && !platform.isWebkit) {
        this._state.skipNextFocusEvent = true;
        
        
        this.el.focus();
        this._state.skipNextFocusEvent = false;
      }
    }

    _handleLeftOrRightArrowKey (event) {
      event.stopPropagation();
      let direction = (event.keyCode === keys$1.LEFT) ? 'left' : 'right';
      
      
      this._delayed(() => {
        this._updateModelSelection({direction});
      });
    }

    _handleUpOrDownArrowKey (event) {
      event.stopPropagation();
      
      
      this._delayed(() => {
        let options = {
          direction: (event.keyCode === keys$1.UP) ? 'left' : 'right'
        };
        this._updateModelSelection(options);
      });
    }

    _handleHomeOrEndKey (event) {
      event.stopPropagation();
      
      
      this._delayed(() => {
        let options = {
          direction: (event.keyCode === keys$1.HOME) ? 'left' : 'right'
        };
        this._updateModelSelection(options);
      });
    }

    _handlePageUpOrDownKey (event) {
      event.stopPropagation();
      
      
      this._delayed(() => {
        let options = {
          direction: (event.keyCode === keys$1.PAGEUP) ? 'left' : 'right'
        };
        this._updateModelSelection(options);
      });
    }

    _handleSpaceKey (event) {
      event.stopPropagation();
      event.preventDefault();
      const text = ' ';
      if (!this.getEditorSession().keyboardManager.onTextInput(text)) {
        this.getEditorSession().transaction((tx) => {
          tx.insertText(text);
        }, { action: 'type' });
      }
    }

    _handleTabKey (event) {
      event.stopPropagation();
      this.el.emit('tab', {
        altKey: event.altKey,
        ctrlKey: event.ctrlKey,
        metaKey: event.metaKey,
        shiftKey: event.shiftKey,
        code: event.code
      });
      if (this.props.handleTab === false) {
        event.preventDefault();
      } else {
        this.__handleTab(event);
      }
    }

    __handleTab () {
      
      
      this._delayed(() => {
        this._updateModelSelection();
      });
    }

    _handleEnterKey (event) {
      event.preventDefault();
      event.stopPropagation();
      this.getEditorSession().transaction((tx) => {
        tx.break();
      }, { action: 'break' });
    }

    _handleEscapeKey () {}

    _handleDeleteKey (event) {
      event.preventDefault();
      event.stopPropagation();
      let direction = (event.keyCode === keys$1.BACKSPACE) ? 'left' : 'right';
      this.getEditorSession().transaction((tx) => {
        tx.deleteCharacter(direction);
      }, { action: 'delete' });
    }

    _hasNativeFocus () {
      return Boolean(this._state.hasNativeFocus)
    }

    _setSelection (sel) {
      
      
      
      
      
      
      
      
      if (!sel.isNull() && sel.surfaceId === this.id && platform.isFF) {
        this._focusElement();
      }
      this.getEditorSession().setSelection(sel);
    }

    _updateModelSelection (options) {
      let sel = this.domSelection.getSelection(options);
      
      
      
      this._setSelection(sel);
    }

    _selectProperty (path) {
      let doc = this.getDocument();
      let text = doc.get(path);
      this._setSelection(doc.createSelection({
        type: 'property',
        path: path,
        startOffset: 0,
        endOffset: text.length
      }));
    }

    
    
    _renderNode ($$, nodeId) {
      let doc = this.getDocument();
      let node = doc.get(nodeId);
      let componentRegistry = this.context.componentRegistry || this.props.componentRegistry;
      let ComponentClass = componentRegistry.get(node.type);
      if (!ComponentClass) {
        console.error('Could not resolve a component for type: ' + node.type);
        ComponentClass = UnsupportedNodeComponent;
      }
      return $$(ComponentClass, {
        doc: doc,
        node: node
      })
    }

    _extractNodeProps (node) {
      let doc = this.getDocument();
      return {
        placeholder: this.props.placeholder,
        doc: doc,
        node: node
      }
    }

    
    _shouldConsumeEvent (event) {
      
      let comp = Component.unwrap(event.target);
      return (comp && (comp === this || comp.context.surface === this))
    }

    
    getSelectionFromEvent (event) {
      let domRange = getDOMRangeFromEvent(event);
      let sel = this.domSelection.getSelectionForDOMRange(domRange);
      sel.surfaceId = this.getId();
      return sel
    }

    setSelectionFromEvent (event) {
      let sel = this.getSelectionFromEvent(event);
      if (sel) {
        this._state.skipNextFocusEvent = true;
        this._setSelection(sel);
      } else {
        console.error('Could not create a selection from event.');
      }
    }

    get id () {
      return this._surfaceId
    }

    _delayed (fn) {
      window.setTimeout(fn, BROWSER_DELAY);
    }

    
    _muteNativeHandlers (event) {
      let contentEditableShortcuts;

      if (platform.isMac) {
        contentEditableShortcuts = [
          'META+66', 
          'META+73', 
          'META+85' 
        ];
      } else {
        contentEditableShortcuts = [
          'CTRL+66', 
          'CTRL+73', 
          'CTRL+85' 
        ];
      }

      const key = parseKeyEvent(event);
      if (contentEditableShortcuts.indexOf(key) > -1) {
        event.preventDefault();
      }
    }
  }

  Surface.prototype._isSurface = true;


  Surface.createSurfaceId = function (surface) {
    let parentSurfaceId = surface.context.parentSurfaceId;
    if (parentSurfaceId) {
      return parentSurfaceId + '/' + surface.name
    } else {
      return surface.name
    }
  };

  class ContainerEditor extends Surface {
    constructor (parent, props, el) {
      
      props.containerId = props.containerId || props.node.id;
      props.name = props.name || props.containerId || props.node.id;

      super(parent, props, el);

      this.containerId = this.props.containerId;
      if (!isString(this.containerId)) {
        throw new Error("Property 'containerId' is mandatory.")
      }
      let doc = this.getDocument();
      this.container = doc.get(this.containerId);
      if (!this.container) {
        throw new Error('Container with id ' + this.containerId + ' does not exist.')
      }

      this.editingBehavior = this.context.editingBehavior || new EditingBehavior();

      this._deriveInternalState(this.props);
    }

    
    shouldRerender (newProps) {
      if (newProps.disabled !== this.props.disabled) return true
      
      
      return false
    }

    willReceiveProps (newProps) {
      super.willReceiveProps.apply(this, arguments);
      this._deriveInternalState(newProps);
    }

    didMount () {
      super.didMount.apply(this, arguments);
      let editorSession = this.getEditorSession();
      editorSession.onUpdate('document', this._onContainerChanged, this, {
        path: this.container.getContentPath()
      });
      this._attachPlaceholder();
    }

    dispose () {
      super.dispose.apply(this, arguments);
      let editorSession = this.getEditorSession();
      editorSession.off(this);
    }

    render ($$) {
      let el = super.render($$);

      let doc = this.getDocument();
      let containerId = this.getContainerId();
      let containerNode = doc.get(containerId);
      if (!containerNode) {
        console.warn('No container node found for ', containerId);
      }
      el.addClass('sc-container-editor container-node ' + containerId)
        .attr('data-id', containerId);

      
      el.attr('spellcheck', this.props.spellcheck === 'native');

      containerNode.getNodes().forEach(function (node, index) {
        el.append(this._renderNode($$, node, index));
      }.bind(this));

      
      if (!this.props.disabled && !this.isEmpty()) {
        el.addClass('sm-enabled');
        el.setAttribute('contenteditable', true);
      }

      return el
    }

    selectFirst () {
      const container = this.getContainer();
      if (container.getLength() > 0) {
        const editorSession = this.getEditorSession();
        const first = container.getChildAt(0);
        setCursor(editorSession, first, container.id, 'before');
      }
    }

    _renderNode ($$, node, nodeIndex) {
      let props = { node };
      if (!node) throw new Error('Illegal argument')
      if (node.isText()) {
        return this.renderNode($$, node, nodeIndex)
      } else {
        let componentRegistry = this.context.componentRegistry;
        let ComponentClass = componentRegistry.get(node.type);
        if (ComponentClass.prototype._isCustomNodeComponent || ComponentClass.prototype._isIsolatedNodeComponent) {
          return $$(ComponentClass, props).ref(node.id)
        } else {
          return $$(IsolatedNodeComponent, props).ref(node.id)
        }
      }
    }

    _extractNodeProps (node) {
      let doc = this.getDocument();
      return {
        doc: doc,
        node: node
      }
    }

    _deriveInternalState (props) {
      let _state = this._state;
      if (!props.hasOwnProperty('enabled') || props.enabled) {
        _state.enabled = true;
      } else {
        _state.enabled = false;
      }
    }

    _selectNextIsolatedNode (direction) {
      let selState = this.getEditorSession().getSelectionState();
      let node = (direction === 'left') ? selState.getPreviousNode() : selState.getNextNode();
      let isIsolatedNode = !node.isText() && !node.isList();
      if (!node || !isIsolatedNode) return false
      if (
        (direction === 'left' && selState.isFirst()) ||
        (direction === 'right' && selState.isLast())
      ) {
        this.getEditorSession().setSelection({
          type: 'node',
          nodeId: node.id,
          containerId: selState.getContainer().id,
          surfaceId: this.id
        });
        return true
      }
      return false
    }

    _handleLeftOrRightArrowKey (event) {
      event.stopPropagation();
      const doc = this.getDocument();
      const sel = this.getEditorSession().getSelection();
      const left = (event.keyCode === keys$1.LEFT);
      const right = !left;
      const direction = left ? 'left' : 'right';

      if (sel && !sel.isNull()) {
        const container = doc.get(sel.containerId, 'strict');

        
        if (sel.isNodeSelection()) {
          let nodePos = container.getPosition(doc.get(sel.getNodeId()));
          if ((left && nodePos === 0) || (right && nodePos === container.length - 1)) {
            event.preventDefault();
            return
          }
        }

        if (sel.isNodeSelection() && !event.shiftKey) {
          this.domSelection.collapse(direction);
        }
      }

      this._delayed(() => {
        this._updateModelSelection({ direction });
      });
    }

    _handleUpOrDownArrowKey (event) {
      event.stopPropagation();
      const doc = this.getDocument();
      const sel = this.getEditorSession().getSelection();
      const up = (event.keyCode === keys$1.UP);
      const down = !up;
      const direction = up ? 'left' : 'right';

      if (sel && !sel.isNull()) {
        const container = doc.get(sel.containerId, 'strict');
        
        if (sel.isNodeSelection()) {
          let nodePos = container.getPosition(doc.get(sel.getNodeId()));
          if ((up && nodePos === 0) || (down && nodePos === container.length - 1)) {
            event.preventDefault();
            return
          }
          
          
          let editorSession = this.getEditorSession();
          
          
          
          if (!event.shiftKey) {
            event.preventDefault();
            if (up) {
              let prev = container.getChildAt(nodePos - 1);
              setCursor(editorSession, prev, sel.containerId, 'after');
              return
            } else {
              let next = container.getChildAt(nodePos + 1);
              setCursor(editorSession, next, sel.containerId, 'before');
              return
            }
          }
        }
      }

      this._delayed(() => {
        this._updateModelSelection({ direction });
      });
    }

    _handleTabKey (event) {
      const editorSession = this.getEditorSession();
      const sel = editorSession.getSelection();
      
      if (sel.isNodeSelection() && sel.isFull()) {
        const comp = this.refs[sel.getNodeId()];
        if (comp && stepIntoIsolatedNode(editorSession, comp)) {
          event.preventDefault();
          event.stopPropagation();
          return
        }
      }
      super._handleTabKey(event);
    }

    __handleTab (e) {
      e.preventDefault();
      if (e.shiftKey) {
        this.getEditorSession().transaction((tx) => {
          tx.dedent();
        }, { action: 'dedent' });
      } else {
        this.getEditorSession().transaction((tx) => {
          tx.indent();
        }, { action: 'indent' });
      }
    }

    
    isContainerEditor () {
      return true
    }

    
    getContainerId () {
      return this.containerId
    }

    getContainer () {
      return this.getDocument().get(this.getContainerId())
    }

    isEmpty () {
      let containerNode = this.getContainer();
      return (containerNode && containerNode.length === 0)
    }

    
    _attachPlaceholder () {
      let firstNode = this.childNodes[0];
      
      if (this.placeholderNode) {
        this.placeholderNode.extendProps({
          placeholder: undefined
        });
      }

      if (this.childNodes.length === 1 && this.props.placeholder) {
        firstNode.extendProps({
          placeholder: this.props.placeholder
        });
        this.placeholderNode = firstNode;
      }
    }

    isEditable () {
      return super.isEditable.call(this) && !this.isEmpty()
    }

    
    _onContainerChanged (change) {
      let doc = this.getDocument();
      
      let renderContext = RenderingEngine.createContext(this);
      let $$ = renderContext.$$;
      let container = this.getContainer();
      let path = container.getContentPath();
      for (let i = 0; i < change.ops.length; i++) {
        let op = change.ops[i];
        if (op.type === 'update' && op.path[0] === path[0]) {
          let diff = op.diff;
          if (diff.type === 'insert') {
            let nodeId = diff.getValue();
            let node = doc.get(nodeId);
            let nodeEl;
            if (node) {
              nodeEl = this._renderNode($$, node);
            } else {
              
              
              
              nodeEl = $$('div');
            }
            this.insertAt(diff.getOffset(), nodeEl);
          } else if (diff.type === 'delete') {
            this.removeAt(diff.getOffset());
          }
        }
      }
      this._attachPlaceholder();
    }
  }

  ContainerEditor.prototype._isContainerEditor = true;

  class TextPropertyEditor extends Surface {
    constructor (parent, props) {
      
      props.name = props.name || props.path.join('.');
      super(parent, props);

      if (!props.path) {
        throw new Error("Property 'path' is mandatory.")
      }
    }

    render ($$) {
      let el = super.render.apply(this, arguments);
      el.addClass('sc-text-property-editor');

      if (!this.props.disabled) {
        el.addClass('sm-enabled');
        el.attr('contenteditable', true);
        
        el.attr('spellcheck', this.props.spellcheck === 'native');
      }

      el.append(
        $$(TextPropertyComponent, {
          placeholder: this.props.placeholder,
          tagName: this.props.tagName || 'div',
          path: this.props.path,
          markers: this.props.markers,
          withoutBreak: this.props.withoutBreak
        }).ref('property')
      );

      return el
    }

    selectFirst () {
      this.editorSession.setSelection({
        type: 'property',
        path: this.getPath(),
        startOffset: 0,
        surfaceId: this.id
      });
    }

    getPath () {
      return this.props.path
    }

    _handleEnterKey (event) {
      event.stopPropagation();
      if (this.props.handleEnter === false || !this.props.multiLine) {
        event.preventDefault();
      }
      this.el.emit('enter', {
        altKey: event.altKey,
        ctrlKey: event.ctrlKey,
        metaKey: event.metaKey,
        shiftKey: event.shiftKey,
        code: event.code
      });
    }

    

    _handleEscapeKey (event) {
      this.el.emit('escape', {
        altKey: event.altKey,
        ctrlKey: event.ctrlKey,
        metaKey: event.metaKey,
        shiftKey: event.shiftKey,
        code: event.code
      });
    }
  }

  TextPropertyEditor.prototype._isTextPropertyEditor = true;

  function series (tasks, cb, i) {
    i = i || 0;
    tasks[i](function (err, ...args) {
      
      if (err) return cb(err)
      if (i === tasks.length - 1) {
        cb(err, ...args); 
      } else {
        series(tasks, cb, i + 1);
      }
    });
  }

  var async = /*#__PURE__*/Object.freeze({
    series: series
  });

  const ALPHABET = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';

  function getColumnLabel (colIdx) {
    if (!isNumber(colIdx)) {
      throw new Error('Illegal argument.')
    }
    let label = '';
    while(true) { 
      let mod = colIdx % ALPHABET.length;
      colIdx = Math.floor(colIdx / ALPHABET.length);
      label = ALPHABET[mod] + label;
      if (colIdx > 0) colIdx--;
      else if (colIdx === 0) break
    }
    return label
  }

  function getRowCol (cellLabel) {
    var match = /^([A-Z]+)([1-9][0-9]*)$/.exec(cellLabel);
    return [
      parseInt(match[2], 10) - 1,
      getColumnIndex(match[1])
    ]
  }

  const A = 'A'.charCodeAt(0);
  const ALPHABET_LENGTH = ALPHABET.length;

  function getColumnIndex (colStr) {
    let index = 0;
    let rank = 0;
    for (let i = colStr.length - 1; i >= 0; i--) {
      let idx = colStr.charCodeAt(i) - A;
      if (idx < 0 || idx >= ALPHABET_LENGTH) throw new Error('Illegal column label: ' + colStr)
      
      index += Math.pow(ALPHABET_LENGTH, rank) * (idx + 1);
      rank++;
    }
    
    return index - 1
  }

  function getCellLabel (rowIdx, colIdx) {
    let colLabel = getColumnLabel(colIdx);
    let rowLabel = rowIdx + 1;
    return colLabel + rowLabel
  }

  function getIndexesFromRange (start, end) {
    let [startRow, startCol] = getRowCol(start);
    let endRow, endCol;
    if (end) {
      ([endRow, endCol] = getRowCol(end));
      if (startRow > endRow) ([startRow, endRow] = [endRow, startRow]);
      if (startCol > endCol) ([startCol, endCol] = [endCol, startCol]);
    } else {
      ([endRow, endCol] = [startRow, startCol]);
    }
    return { startRow, startCol, endRow, endCol }
  }

  function getRangeFromMatrix (cells, startRow, startCol, endRow, endCol, force2D) {
    if (!force2D) {
      if (startRow === endRow && startCol === endCol) {
        let row = cells[startRow];
        if (row) return row[startCol]
        else return undefined
      }
      if (startRow === endRow) {
        let row = cells[startRow];
        if (row) return row.slice(startCol, endCol + 1)
        else return []
      }
      if (startCol === endCol) {
        let res = [];
        for (let i = startRow; i <= endRow; i++) {
          let row = cells[i];
          if (row) res.push(row[startCol]);
        }
        return res
      }
    }
    let res = [];
    for (var i = startRow; i < endRow + 1; i++) {
      let row = cells[i];
      if (row) res.push(row.slice(startCol, endCol + 1));
    }
    return res
  }

  var tableHelpers = /*#__PURE__*/Object.freeze({
    ALPHABET: ALPHABET,
    getColumnLabel: getColumnLabel,
    getRowCol: getRowCol,
    getColumnIndex: getColumnIndex,
    getCellLabel: getCellLabel,
    getIndexesFromRange: getIndexesFromRange,
    getRangeFromMatrix: getRangeFromMatrix
  });

  function renderListNode (listNode, $$) {
    let items = listNode.getItems();
    let stack = [$$(_getTagName(1))];
    for (let i = 0; i < items.length; i++) {
      const item = items[i];
      
      const level = item.getLevel();
      if (level < stack.length) {
        for (let j = stack.length; j > level; j--) {
          stack.pop();
        }
      } else if (level > stack.length) {
        for (let j = stack.length; j < level; j++) {
          let list = stack[j - 1];
          let childCount = list.getChildCount();
          let item;
          if (childCount === 0) {
            item = $$('li');
            list.append(item);
          } else {
            item = list.getChildAt(childCount - 1);
          }
          let sublist = $$(_getTagName(j + 1));
          item.append(sublist);
          stack.push(sublist);
        }
      }
      console.assert(level === stack.length, 'item.level should now be the same as stack.length');
      
      last$1(stack).append(
        $$(item)
      );
    }
    for (let j = stack.length; j > 1; j--) {
      stack.pop();
    }

    return stack[0]

    function _getTagName (level) {
      let listType = listNode.getListType(level);
      return listType === 'order' ? 'ol' : 'ul'
    }
  }

  const DISABLED = Object.freeze({
    disabled: true
  });

  class ListItemComponent extends Component {
    render ($$) {
      const node = this.props.node;
      const path = node.getPath();

      let el = $$('li').addClass('sc-list-item');
      el.append(
        $$(TextPropertyComponent, { path }).ref('text')
      );
      
      if (this.props.children) {
        el.append(this.props.children);
      }
      return el
    }
  }

  class ListComponent extends NodeComponent {
    render ($$) {
      let node = this.props.node;
      let el = renderListNode(node, (item) => {
        
        if (isString(item)) {
          return $$(item)
        } else if (item.type === 'list-item') {
          return $$(ListItemComponent, {
            node: item
          }).ref(item.id)
        }
      });
      el.addClass('sc-list').attr('data-id', node.id);
      return el
    }
  }


  ListComponent.prototype._isCustomNodeComponent = true;

  class TableCellComponent extends Component {
    render ($$) {
      let node = this.props.node;
      let el = $$('td').addClass('sc-table-cell');
      el.append(
        $$(TextPropertyEditor, {
          path: node.getPath(),
          disabled: this.props.disabled
        }).ref('editor')
      );
      if (node.rowspan > 0) {
        el.attr('rowspan', node.rowspan);
      }
      if (node.colspan > 0) {
        el.attr('colspan', node.colspan);
      }
      return el
    }

    grabFocus () {
      let node = this.props.node;
      this.context.editorSession.setSelection({
        type: 'property',
        path: node.getPath(),
        startOffset: node.getLength(),
        surfaceId: this.refs.editor.id
      });
    }
  }

  TableCellComponent.prototype._isTableCellComponent = true;

  // not yet analysed
  const UNKNOWN = Symbol('UNKNOWN');
  // analysed
  const ANALYSED = Symbol('ANALYSED');
  // syntax or graph errors
  const BROKEN = Symbol('BROKEN');
  // runtime or validation errors
  const FAILED = Symbol('FAILED');
  // one of the inputs is broken, failed, or blocked
  const BLOCKED = Symbol('BLOCKED');
  // all inputs are ready, running, or ok
  const WAITING = Symbol('WAITING');
  // all inputs are ok
  const READY = Symbol('READY');
  // evaluation is running (READY+evaluation triggered)
  // TODO: do we really need this on this level?
  const RUNNING = Symbol('RUNNING');
  // evaluation succeeded
  const OK = Symbol('OK');

  function cellStateToInteger (state) {
    switch (state) {
      case UNKNOWN: return -2
      case ANALYSED: return -1
      case BROKEN: return 1
      case FAILED: return 2
      case BLOCKED: return 3
      case WAITING: return 4
      case READY: return 5
      case RUNNING: return 6
      case OK: return 7
      default:
        throw new Error('Illegal state.')
    }
  }

  function cellStateToString (state) {
    switch (state) {
      case UNKNOWN: return 'unknown'
      case ANALYSED: return 'analysed'
      case BROKEN: return 'broken'
      case FAILED: return 'failed'
      case BLOCKED: return 'blocked'
      case WAITING: return 'waiting'
      case READY: return 'ready'
      case RUNNING: return 'running'
      case OK: return 'ok'
      default:
        throw new Error('Illegal state.')
    }
  }

  // These characters will be replaced. Add more if needed.
  const INVALID_ID_CHARACTERS = '[^A-Za-z0-9]';
  const TICK = '\''.charCodeAt(0);

  /*
    Replaces all characters that are invalid in a variable identifier.

    Note: replacing characters one-by-one retains the original length or the string
    which is desired as this does avoid source-mapping. E.g. when a runtime error
    occurs, the error location can be applied to the original source code without
    any transformation.
  */
  function toIdentifier (str, c = '_') {
    let firstIsTick = (str.charCodeAt(0) === TICK);
    str = str.replace(new RegExp(INVALID_ID_CHARACTERS, 'g'), c);
    // ATTENTION: for transclusion symbols which start with a tick
    // we want to add character that separates the transpiled id
    // from a potentially previous one
    if (firstIsTick) {
      return ' ' + str.slice(1)
    } else {
      return str
    }
  }

  function createSymbol (m) {
    const text = m[0];
    const startPos = m.index;
    const endPos = text.length + startPos;
    const mangledStr = toIdentifier(text);
    const scope = m[1] || m[2];
    const anchor = m[3];
    const focus = m[4];
    const varName = m[5];
    let type, name;
    if (anchor) {
      if (focus && focus !== anchor) {
        type = 'range';
        name = anchor + ':' + focus;
      } else {
        type = 'cell';
        name = anchor;
      }
    } else if (varName) {
      type = 'var';
      name = varName;
    } else {
      throw new Error('Invalid symbol expression')
    }
    return { type, text, scope, name, mangledStr, startPos, endPos, anchor, focus }
  }

  const BROKEN_REF = '#BROKEN_REF';
  /*
    Matchers for transclusions and cell references

    Examples:
    - A1
    - A1:B10
    - Foo!A1
    - doc1!x
    - 'My Sheet'!A1:B10
    - 'My Doc'.x
  */
  const ID = '([_A-Za-z][_A-Za-z0-9]*)';
  const NAME = "[']([^']+)[']";
  const CELL_ID = '([A-Z]+[1-9][0-9]*)';
  const EXPRESSION_CELL = '^\\s*' + ID + '?\\s*=';
  const EXPRESSION_CELL_RE = new RegExp(EXPRESSION_CELL);
  /*
    A reference can point to a variable, a cell, or a range inside the same document
    or another one. To avoid matches inside of other symbols, '\b' (word boundary) is used in the expression.
    `[']` can not be used in combination with '\b'.get

    ```
     ( ( \b ID | ['].+['] )[!] | \b)( CELL_ID([:]CELL_ID)? | ID )
    ```
  */
  const REF = '(?:(?:(?:(?:\\b' + ID + '|' + NAME + '))[!])|\\b)(?:' + CELL_ID + '(?:[:]' + CELL_ID + ')?|' + ID + ')';
  const REF_RE = new RegExp(REF);

  function extractSymbols (code) {
    if (!code) return []
    let re = new RegExp(REF, 'g');
    let symbols = [];
    let m;
    while ((m = re.exec(code))) {
      symbols.push(createSymbol(m));
    }
    return symbols
  }

  /*
    Transpiles a piece of source code so that it does not contain
    Transclusion expressions anymore, which are usually not valid in common languages.

    @param {string} code
    @param {object} map storage for transpiled symbols so that they can be mapped back later on
    @result
  */
  function transpile (code) {
    if (!code) return code
    let symbols = extractSymbols(code);
    let map = {};
    let transpiledCode = code;
    // Note: we are transpiling without changing the length of the original source
    // i.e. `'My Sheet'!A1:B10` is transpiled into `_My_Sheet__A1_B10`
    // thus the symbol locations won't get invalid by this step
    for (let i = 0; i < symbols.length; i++) {
      const s = symbols[i];
      transpiledCode = transpiledCode.substring(0, s.startPos) + s.mangledStr + transpiledCode.slice(s.endPos);
      let transpiledName = s.mangledStr;
      let key = transpiledName.trim();
      map[key] = s;
    }
    return { transpiledCode, symbols, map }
  }

  /*
    - `type`: `variable | cell | range`
    - `id`: a qualified id such as `doc1!x`, `sheet1!A1`, `sheet1!A1:A10`
    - `mangledStr`: not longer than the orignal which is used for transpiledCode
    - `scope`: e.g `doc1`, `sheet1`, `'My Document'`"
    - `symbol`: local symbol id such as `x`, `A1`, `A1:A10`
  */
  function parseSymbol (str) {
    let m = REF_RE.exec(str);
    if (!m) throw new Error('Unrecognised symbol format.')
    return createSymbol(m)
  }

  /*
    Derives the qualified id of a cell.
  */
  function qualifiedId (doc, cell) {
    let cellId = isString(cell) ? cell : cell.id;
    if (doc) {
      let docId = isString(doc) ? doc : doc.id;
      return `${docId}!${cellId}`
    } else {
      return cellId
    }
  }

  function transformRange (start, end, pos, count) {
    if (!count) return false
    if (!isNumber(pos) || !isNumber(count)) throw new Error('pos and count must be integers')
    if (end < pos) return false
    if (count > 0) {
      if (pos <= start) {
        start += count;
      }
      if (pos <= end) {
        end += count;
      }
    } else {
      // for removal count < 0
      count = -count;
      // null means deleted
      if (start >= pos && end < pos + count) return -1
      const x1 = pos;
      const x2 = pos + count;
      if (x2 <= start) {
        start -= count;
        end -= count;
      } else {
        if (pos <= start) {
          start = start - Math.min(count, start - x1);
        }
        if (pos <= end) {
          end = end - Math.min(count, end - x1 + 1);
        }
      }
    }
    return { start, end }
  }

  function recordTransformations (cells, dim, pos, count, affectedCells, visited) {
    affectedCells = affectedCells || new Set();
    visited = visited || new Set();
    cells.forEach(cell => {
      if (visited.has(cell)) return
      visited.add(cell);
      cell.symbols.forEach(s => {
        if (s.type !== 'var') {
          let start, end;
          if (dim === 0) {
            start = s.startRow;
            end = s.endRow;
          } else {
            start = s.startCol;
            end = s.endCol;
          }
          let res = transformRange(start, end, pos, count);
          if (!res) return
          if (res === -1) {
            s._update = { type: 'broken' };
          } else {
            let type = (count < 0 ? 'delete' : 'insert') + (dim === 0 ? 'Rows' : 'Cols');
            s._update = {
              type,
              start: res.start,
              end: res.end
            };
          }
        }
      });
      affectedCells.add(cell);
    });
  }

  // TODO: need to rethink if it is possible to get rid
  // of the redundancy, or have a cleaner way to derive everything after update
  function applyCellTransformations (cell) {
    let symbols = cell.symbols;
    let source = cell._source;
    let oldSource = source.original;
    let newSource = oldSource;
    // ATTENTION: iterating symbols backwards so that code transformation does
    // not invalidate character indexes in symbols
    // For that symbols must be sorted by pos, which they are
    for (let i = symbols.length - 1; i >= 0; i--) {
      let s = symbols[i];
      let update = s._update;
      if (!update) continue
      delete s._update;
      // compute derived content according to parameters
      let oldName = s.name;
      let oldScope = s.scope;
      let oldOrigStr = s.origStr;
      let newOrigStr = oldOrigStr;
      switch (update.type) {
        case 'insertRows':
        case 'deleteRows': {
          s.startRow = update.start;
          s.endRow = update.end;
          let newName = getCellSymbolName(s);
          newOrigStr = oldOrigStr.replace(oldName, newName);
          break
        }
        case 'insertCols':
        case 'deleteCols': {
          s.startCol = update.start;
          s.endCol = update.end;
          let newName = getCellSymbolName(s);
          newOrigStr = oldOrigStr.replace(oldName, newName);
          break
        }
        case 'broken': {
          s.type = 'var';
          s.startRow = s.startCol = s.endRow = s.endCol = null;
          newOrigStr = BROKEN_REF;
          break
        }
        case 'rename': {
          if (oldScope) {
            let newScope = update.scope;
            newOrigStr = oldOrigStr.replace(oldScope, newScope);
          }
          break
        }
        default:
          throw new Error('Illegal state')
      }
      newSource = newSource.slice(0, s.startPos) + newOrigStr + newSource.slice(s.endPos);
    }
    if (newSource !== oldSource) {
      cell.source = newSource;
    }
  }

  function getCellSymbolName (s) {
    let newName = tableHelpers.getCellLabel(s.startRow, s.startCol);
    if (s.type === 'range') {
      newName += ':' + tableHelpers.getCellLabel(s.endRow, s.endCol);
    }
    return newName
  }

  // TODO: change the naming
  // This is used within sheets to distinguish constants and cells with expression
  function isExpression (source) {
    return EXPRESSION_CELL_RE.exec(source)
  }

  function parseValue (text) {
    if (text === 'false') {
      return false
    } else if (text === 'true') {
      return true
    } else if (!isNaN(text)) {
      let _int = Number.parseInt(text, 10);
      if (_int == text) { // eslint-disable-line
        return _int
      } else {
        return Number.parseFloat(text)
      }
    } else {
      return text
    }
  }

  // index for all helpers

  const getIndexesFromRange$1 = tableHelpers.getIndexesFromRange;

  class Cell$1 {
    constructor (doc, cellData) {
      const { id, lang, source, status, inputs, output, value, errors, hasSideEffects, next, prev } = cellData;
      this.doc = doc;

      // Attention: the given cell id is not necessarily globally unique
      // thus, we derive a unique id using the document id and the node id
      // localId is used later to be able to map back to the associated node
      // TODO: I would rather go for only one id, and always have a doc
      if (doc) {
        let docId = this.docId = doc.id;
        // is the id already a qualified id?
        if (id.startsWith(docId)) {
          this.id = id;
          // ATTENTION: assuming that the qualified id is joining
          // the doc id and the node id with a single character (e.g. '!')
          this.unqualifiedId = id.slice(docId.length + 1);
        } else {
          this.id = qualifiedId(doc, cellData);
          this.unqualifiedId = id;
        }
      } else {
        this.docId = null;
        this.id = id;
        this.unqualifiedId = id;
      }

      this.lang = lang;

      /*
       The source code is transpiled to an object
       - original
       - transpiledSource
       - symbols
       - symbolMapping: map from transpiled names to original names
       - isContant
      */
      this._source = this._transpile(source);

      // managed by CellGraph
      this.status = status || UNKNOWN;
      // a set of symbols ('x', 'A1', 'A1:B10', 'doc1!x', 'sheet1!A1', 'sheet1!A1:A10', 'sheet1!foo')
      this.inputs = new Set(inputs || []);
      // an output symbol (typically only used for document cells)
      this.output = output;
      // one or many CellErrors
      this.errors = errors || [];
      // the last computed value
      this.value = value;
      // for cells with side effects
      this.hasSideEffects = Boolean(hasSideEffects);
      // for cells in a linear model
      // this is particularly important for cells with side effects
      this.next = next;
      this.prev = prev;
      // used by CellGraph
      this.level = 0;
      // TODO: maybe we want to keep some stats, e.g. time of last evaluation, duration of last evaluation etc.
      this.stats = {};
    }

    clearErrors (filter$$1) {
      if (isString(filter$$1)) {
        const type = filter$$1;
        filter$$1 = (e) => {
          return e.type === type
        };
      }
      this.errors = this.errors.filter(e => !filter$$1(e));
    }

    addErrors (errors) {
      this.errors = this.errors.concat(errors);
    }

    hasErrors () {
      return this.errors.length > 0
    }

    hasError (type) {
      for (let i = 0; i < this.errors.length; i++) {
        if (this.errors[i].type === type) return true
      }
      return false
    }

    get state () {
      console.warn('DEPRECATED: use cellState.status instead.');
      return this.status
    }

    hasOutput () {
      return Boolean(this.output)
    }

    hasValue () {
      return Boolean(this.value)
    }

    getValue () {
      return this.value
    }

    getLang () {
      return this.lang || (this.doc ? this.doc.lang : 'mini')
    }

    get source () {
      return this._source.original
    }

    set source (source) {
      this._source = this._transpile(source);
    }

    get transpiledSource () {
      return this._source.transpiled
    }

    get symbols () {
      return this._source.symbols
    }

    isConstant () {
      return this._source.isConstant
    }

    isSheetCell () {
      return false
    }

    toString () {
      // sheet1!A1 <- { ... source }
      let parts = [];
      if (this.output) {
        parts.push(this.output);
        parts.push(' <- ');
      } else {
        parts.push(this.id);
        parts.push(': ');
      }
      parts.push(this._source.original);
      return parts.join('')
    }

    dump () {
      // TODO: to be able to start off from a loaded dump
      // we would need to store more information, such as 'level' etc
      // let inputs = Array.from(this.inputs).map(s => s.dump())
      // let output
      // if (this.output) output = this.output.dump()
      return {
        id: this.unqualifiedId,
        lang: this.lang,
        source: this._source.original,
        status: cellStateToString(this.status),
        // inputs,
        // output,
        errors: this.errors,
        value: this.value
      }
    }

    _getStatusString () {
      return cellStateToString(this.status)
    }

    _transpile (source) {
      let original = source;
      let transpiled;
      let symbols = [];
      let symbolMapping = {};
      let isConstant = false;
      // in sheets there is a distinction between constants and
      // expressions. Typically, expression start with an '='.
      // In addition we allow cells to register an alias such as 'x = 1'
      // Then the cell can be addressed either via cell notation, such as 'A1',
      // or by name 'x'
      if (this.isSheetCell()) {
        let m = isExpression(source);
        if (m) {
          // there is an output name if the user writes 'x = ...'
          let output = m[1];
          // if the cell is an expression without an output name
          // we must transpile the source, because a leading '='
          // is not valid in all of the languages we consider
          if (!output) {
            let L = m[0].length;
            let prefix = new Array(L);
            prefix.fill(' ');
            source = prefix + source.slice(L);
          }
        } else {
          isConstant = true;
        }
      }
      if (isConstant) {
        transpiled = original;
      } else if (source) {
        let res = transpile(source);
        transpiled = res.transpiledCode;
        symbols = res.symbols.map(s => this._augmentSymbol(s));
        symbolMapping = res.map;
      }

      return {
        original,
        transpiled,
        symbols,
        symbolMapping,
        isConstant
      }
    }

    _augmentSymbol (s) {
      if (s.type === 'cell') {
        let { startRow, startCol } = getIndexesFromRange$1(s.name);
        s.startRow = s.endRow = startRow;
        s.startCol = s.endCol = startCol;
      } else if (s.type === 'range') {
        let [start, end] = s.name.split(':');
        let { startRow, startCol, endRow, endCol } = getIndexesFromRange$1(start, end);
        s.startRow = startRow;
        s.startCol = startCol;
        s.endRow = endRow;
        s.endCol = endCol;
      }
      s.origStr = s.text;
      s.cell = this;
      return s
    }
  }

  class CellError extends Error {
    constructor (msg, details) {
      super(msg);
      this.details = details;
    }

    static cast (err) {
      if (err instanceof CellError) {
        return err
      } else {
        return new RuntimeError(err.message, err)
      }
    }
  }

  class ContextError extends CellError {
    get type () { return 'engine' }
    get name () { return 'context' }
  }

  class GraphError extends CellError {
    get type () { return 'graph' }
  }

  class SyntaxError$1 extends CellError {
    get type () { return 'engine' }
    get name () { return 'syntax' }
  }

  class UnresolvedInputError extends GraphError {
    get name () { return 'unresolved' }
  }

  class CyclicDependencyError extends GraphError {
    get trace () {
      return this.details.trace
    }
    get name () { return 'cyclic' }
  }

  class OutputCollisionError extends GraphError {
    get name () { return 'collision' }
  }

  class RuntimeError extends CellError {
    get type () { return 'runtime' }
    get name () { return 'runtime' }
  }

  class ValidationError extends CellError {
    get type () { return 'runtime' }
    get name () { return 'validation' }
  }

  const MSG_UNRESOLVED_INPUT = 'Unresolved input.';

  class CellGraph {
    constructor () {
      // cell data by id
      this._cells = {};
      // a map of names to values that are global
      this._globals = new Map();
      // symbols -> cell ids; which cell is depending on a symbol
      this._ins = {};
      // symbol -> cell id; which symbol is provided by which cell
      this._out = {};
      // to record which cells have been affected (batched update)
      // - 'stateChanged': an error has been added or the state has been updated otherwise
      // - 'structureChanged': cells that have changed w.r.t. inputs or output, and thus need to be checked for structural consistency
      // - 'valueUpdated': all cells that have a new value or a runtime error, i.e. the new state is OK or FAILED.
      // TODO: shouldn't be 'valueUpdated' just be the same as 'stateChanged'?
      this._stateChanged = new Set();
      this._structureChanged = new Set();
      this._valueUpdated = new Set();
    }

    needsUpdate () {
      return this._stateChanged.size > 0 || this._structureChanged.size > 0 || this._valueUpdated.size > 0
    }

    hasCell (id) {
      return this._cells.hasOwnProperty(id)
    }

    getCell (id) {
      return this._cells[id]
    }

    addCell (cell) {
      const id = cell.id;
      if (this._cells[id]) throw new Error(`Cell with ${id} already exists`)
      this._cells[id] = cell;
      this._structureChanged.add(id);
      if (cell.inputs && cell.inputs.size > 0) {
        this._registerInputs(id, new Set(), cell.inputs);
      }
      if (cell.output) {
        this._registerOutput(id, null, cell.output);
      }
    }

    removeCell (id) {
      const cell = this._cells[id];
      if (!cell) throw new Error('Internal error: cell does not exist.')
      cell.inputs.forEach(s => {
        if (this._ins[s]) {
          this._ins[s].delete(id);
        }
      });
      this._deregisterOutput(id, cell.output);
      delete this._cells[id];
      if (cell.prev) {
        this._setNext(cell.prev, cell.next);
      }
      if (cell.next) {
        this._setPrev(cell.next, cell.prev);
      }
      // remove the cell from all registered updates
      this._stateChanged.delete(cell.id);
      this._structureChanged.delete(cell.id);
      this._valueUpdated.delete(cell.id);
    }

    getValue (symbol) {
      let cellId = this._out[symbol];
      if (!cellId) return undefined
      // if there is a name collision return undefined
      // TODO: should we allow this at all?
      if (isArray(cellId)) {
        throw new Error('Ambigous symbol: ' + symbol)
      }

      const cell = this._cells[cellId];
      if (!cell) throw new Error('Internal error: cell does not exist.')
      // Note, that the cell value is actually not interpreted in any way by the graph
      // it is maintained by the engine.
      return cell.value
    }

    setInputsOutputs (id, newInputs, newOutput) {
      let cell = this._cells[id];
      if (!cell) throw new Error(`Unknown cell ${id}`)
      this._setInputs(cell, newInputs);
      this._setOutput(cell, newOutput);
      if (cell.status === UNKNOWN) {
        this._structureChanged.add(id);
        cell.status = ANALYSED;
      }
    }

    // Note: we use this for sheet cells
    setInputs (id, newInputs) {
      let cell = this._cells[id];
      if (!cell) throw new Error(`Unknown cell ${id}`)
      this._setInputs(cell, newInputs);
      if (cell.status === UNKNOWN) {
        this._structureChanged.add(id);
        cell.status = ANALYSED;
      }
    }

    // used to update sheet cell output symbols after structural
    // changes. In this case there are typically a lot of other changes, too
    setOutput (id, newOutput) {
      let cell = this._cells[id];
      if (!cell) throw new Error(`Unknown cell ${id}`)
      this._setOutput(cell, newOutput);
    }

    _setInputs (cell, newInputs) {
      newInputs = new Set(newInputs);
      if (this._registerInputs(cell.id, cell.inputs, newInputs)) {
        cell.inputs = newInputs;
        this._clearCyclicDependencyError(cell);
        cell.clearErrors(e => e instanceof UnresolvedInputError);
      }
    }

    _setOutput (cell, newOutput) {
      // TODO: if only the output of a cell changed, we could retain the runtime result
      // and leave the cell's state untouched
      let oldOutput = cell.output;
      if (this._registerOutput(cell.id, oldOutput, newOutput)) {
        cell.output = newOutput;
        // TODO: do we need to clear a potential old graph error
        // e.g. from a previous cyclic dependency
        this._clearCyclicDependencyError(cell);
      }
    }

    _setNext (id, nextId) {
      let cell = this._cells[id];
      cell.next = nextId;
      this._structureChanged.add(nextId);
    }

    _setPrev (id, prevId) {
      let cell = this._cells[id];
      cell.prev = prevId;
      this._structureChanged.add(id);
    }

    addError (id, error) {
      this.addErrors(id, [error]);
    }

    addErrors (id, errors) {
      let cell = this._cells[id];
      errors = errors.map(err => CellError.cast(err));
      cell.addErrors(errors);
      this._stateChanged.add(id);
    }

    clearErrors (id, type) {
      let cell = this._cells[id];
      if (type === 'graph' || !type) {
        this._clearCyclicDependencyError(cell);
      }
      if (type) {
        cell.clearErrors(type);
      } else {
        cell.errors = [];
      }
      this._stateChanged.add(id);
    }

    setValue (id, value) {
      let cell = this._cells[id];
      cell.value = value;
      if (cell.hasErrors()) {
        cell.status = FAILED;
      } else {
        cell.status = OK;
      }
      this._valueUpdated.add(id);
    }

    _registerInputs (id, oldInputs, newInputs) {
      let toAdd = new Set(newInputs);
      let toRemove = new Set();
      if (oldInputs) {
        oldInputs.forEach(s => {
          if (newInputs.has(s)) {
            toAdd.delete(s);
          } else {
            toRemove.add(s);
          }
        });
      }
      toRemove.forEach(s => {
        if (this._ins[s]) {
          this._ins[s].delete(id);
        }
      });
      toAdd.forEach(s => {
        let ins = this._ins[s];
        if (!ins) ins = this._ins[s] = new Set();
        ins.add(id);
      });
      if (toAdd.size > 0 || toRemove.size > 0) {
        this._structureChanged.add(id);
        return true
      } else {
        return false
      }
    }

    _registerOutput (id, oldOutput, newOutput) {
      // nothing to be done if no change
      if (oldOutput === newOutput) return false
      // deregister the old output first
      if (oldOutput) {
        this._deregisterOutput(id, oldOutput);
      }
      if (newOutput) {
        // TODO: detect collisions
        if (this._out[newOutput]) {
          let conflictingIds = this._out[newOutput];
          if (isString(conflictingIds)) conflictingIds = [conflictingIds];
          conflictingIds = new Set(conflictingIds);
          conflictingIds.add(id);
          conflictingIds = Array.from(conflictingIds);
          this._out[newOutput] = conflictingIds;
          // consider every related id for re-examination
          conflictingIds.forEach(_id => this._structureChanged.add(_id));
        } else {
          this._out[newOutput] = id;
        }
        // mark new deps as affected
        let ids = this._ins[newOutput] || [];
        ids.forEach(_id => {
          let cell = this._cells[_id];
          if (cell.status === BROKEN) {
            // TODO: probably we do not want to clear all graph errors, but only specific ones
            cell.clearErrors('graph');
          }
          this._structureChanged.add(_id);
        });
      }
      return true
    }

    _deregisterOutput (id, output) {
      if (this._hasOutputCollision(output)) {
        this._resolveOutputCollision(output, id);
      } else {
        delete this._out[output];
        // mark old deps as affected
        let ids = this._ins[output] || [];
        ids.forEach(_id => {
          let cell = this._cells[_id];
          if (cell.status === BROKEN) {
            // TODO: probably we do not want to clear all graph errors, but only specific ones
            cell.clearErrors('graph');
          }
          this._structureChanged.add(_id);
        });
      }
    }

    update () {
      // a set of cell ids that have been updated
      let updated = new Set();
      let stateChanged = this._stateChanged;

      // examine the graph structure
      // Note: we should not need to update the whole graph, still, we can do an
      // exhaustive update, because this is not performance critical
      let levels = {};
      this._structureChanged.forEach(id => {
        // detect unresolvable inputs
        this._detectUnresolvableInputs(id);
        // detect output collisions
        this._detectOutputCollisions(id);
        // deterimine the dependency level and check for cyclic dependencies
        // Note: in case of a cyclic dependency we want to set all involved
        // cells into BROKEN state
        this._computeDependencyLevel(id, levels, updated);
        updated.add(id);
        // mark cells with structure changes for state update
        stateChanged.add(id);
      });

      if (this._valueUpdated.size > 0) {
        this._valueUpdated.forEach(id => {
          updated.add(id);
        });
        // mark all followers for a state update
        this._getFollowSet(this._valueUpdated).forEach(id => {
          let cell = this._cells[id];
          cell.clearErrors('runtime');
          stateChanged.add(id);
        });
      }

      if (stateChanged.size > 0) {
        // then propagate state updates for all structural changes
        this._updateStates(stateChanged, updated);
      }

      this._stateChanged.clear();
      this._structureChanged.clear();
      this._valueUpdated.clear();

      return updated
    }

    _detectUnresolvableInputs (id) {
      let cell = this._cells[id];
      // detect unresolvable inputs
      let inputs = Array.from(cell.inputs);
      let unresolved = inputs.filter(s => {
        return !this._resolve(s) && !this._globals.has(s.name)
      });
      if (unresolved.length > 0) {
        cell.clearErrors(e => e instanceof UnresolvedInputError);
        cell.addErrors([new UnresolvedInputError(MSG_UNRESOLVED_INPUT, { unresolved })]);
        cell.status = BROKEN;
      }
    }

    _detectOutputCollisions (id) {
      let cell = this._cells[id];
      let output = cell.output;
      if (!output) return
      let ids = this._out[output];
      if (isArray(ids)) {
        // TODO: is there a more efficient way?
        cell.clearErrors(e => e instanceof OutputCollisionError);
        cell.addErrors([new OutputCollisionError('Competing output declarations.', { ids })]);
      }
    }

    _computeDependencyLevel (id, levels, updated, trace = new Set(), traceSymbols = {}) {
      let cell = this._cells[id];
      let inputs = Array.from(cell.inputs);
      trace = new Set(trace);
      trace.add(id);

      const _recursive = (id) => {
        if (trace.has(id)) {
          this._handleCycle(trace, traceSymbols, updated);
          return Infinity
        }
        if (levels.hasOwnProperty(id)) {
          return levels[id]
        } else {
          return this._computeDependencyLevel(id, levels, updated, trace, traceSymbols)
        }
      };

      let inputLevels = [];
      inputs.forEach(s => {
        let res = this._resolve(s);
        if (!res) return 0
        traceSymbols[id] = s;
        if (isString(res)) {
          inputLevels.push(_recursive(res));
        } else {
          res.forEach(id => {
            inputLevels.push(_recursive(id));
          });
        }
      });
      // EXPERIMENTAL: considering an explicitly set predecessor to preserve natural order where appropriate
      if (cell.prev) {
        traceSymbols[id] = cell.id;
        inputLevels.push(_recursive(cell.prev));
      }
      let level = inputLevels.length > 0 ? Math.max(...inputLevels) + 1 : 0;
      levels[id] = level;
      cell.level = level;
      return level
    }

    _getAffectedCellsSorted (ids) {
      let cells = [];
      let visited = new Set();
      let q = Array.from(ids);
      while (q.length > 0) {
        let id = q.shift();
        if (visited.has(id)) continue
        visited.add(id);
        const cell = this._cells[id];
        const level = cell.level;
        if (!cells[level]) cells[level] = [];
        cells[level].push(cell);
        let affected = this._getAffected(cell);
        q = q.concat(affected.filter(id => !visited[id]));
      }
      // Note: remove bins for levels that are not affected
      return flatten(cells.filter(Boolean))
    }

    _getAffected (cell) {
      let affected = [];
      if (this._cellProvidesOutput(cell)) {
        affected = Array.from(this._ins[cell.output] || []);
      }
      if (cell.hasSideEffects && cell.next) {
        // find next cell with side effects
        for (let nextId = cell.next; nextId; nextId = cell.next) {
          let nextCell = this._cells[nextId];
          if (nextCell && nextCell.hasSideEffects) {
            affected.push(nextId);
            break
          }
        }
      }
      return affected
    }

    _updateStates (ids, updated) {
      // get all affected cells, i.e. all cells that are depending
      // on the cells with given ids
      let cells = this._getAffectedCellsSorted(ids);
      // determine the cell state from the state of their inputs
      cells.forEach(cell => this._updateCellState(cell, updated));
    }

    _updateCellState (cell, updated) {
      // invariant detection of BROKEN state
      if (cell.hasError('engine') || cell.hasError('graph')) {
        if (cell.status === BROKEN) return
        cell.status = BROKEN;
        updated.add(cell.id);
        return
      }
      // skip cells which have not been registered yet
      if (cell.status === UNKNOWN) return
      // invariant detection of FAILED state
      if (cell.hasErrors()) {
        if (cell.status === FAILED) return
        cell.status = FAILED;
        updated.add(cell.id);
        return
      }
      let inputs = Array.from(cell.inputs);
      if (!cell.hasSideEffects && inputs.length === 0) {
        cell.status = READY;
        return
      }
      let inputStates = [];
      inputs.forEach(s => {
        let res = this._resolve(s);
        if (!res) return
        if (isString(res)) {
          let _cell = this._cells[res];
          // NOTE: for development we kept the less performant but more readable
          // representation as symbols, instead of ints
          inputStates.push(cellStateToInteger(_cell.status));
        } else {
          res.forEach(id => {
            let _cell = this._cells[id];
            inputStates.push(cellStateToInteger(_cell.status));
          });
        }
      });
      if (cell.hasSideEffects && cell.prev) {
        let _cell = this._cells[cell.prev];
        if (_cell) {
          inputStates.push(cellStateToInteger(_cell.status));
        }
      }
      let inputState = Math.min(...inputStates);
      // Note: it is easier to do this in an arithmetic way
      // than in boolean logic
      let newState;
      if (inputState <= cellStateToInteger(BLOCKED)) {
        newState = BLOCKED;
      } else if (inputState <= cellStateToInteger(RUNNING)) {
        newState = WAITING;
      } else { // if (inputState === OK) {
        newState = READY;
      }
      if (newState && newState !== cell.status) {
        cell.status = newState;
        updated.add(cell.id);
      }
    }

    _resolve (symbol) {
      return this._out[symbol]
    }

    _cellProvidesOutput (cell) {
      return (cell.output && cell.id === this._out[cell.output])
    }

    // set of cell ids that depend on the given ones
    _getFollowSet (ids) {
      let followSet = new Set();
      ids.forEach(id => {
        const cell = this._cells[id];
        this._getAffected(cell).forEach(id => followSet.add(id));
      });
      return followSet
    }

    // get a set of all ids a cell is depending on (recursively)
    _getPredecessorSet (id, set) {
      if (!set) set = new Set();
      const _recursive = (id) => {
        if (!set.has(id)) {
          set.add(id);
          this._getPredecessorSet(id, set);
        }
      };

      let cell = this.getCell(id);
      cell.inputs.forEach(s => {
        let res = this._resolve(s);
        if (!res) return
        if (isString(res)) {
          _recursive(res);
        } else {
          res.forEach(_recursive);
        }
      });
      if (cell.hasSideEffects && cell.prev) {
        _recursive(cell.prev);
      }
      return set
    }

    _handleCycle (trace, traceSymbols, updated) {
      let error = new CyclicDependencyError('Cyclic dependency', { trace, symbols: traceSymbols });
      trace.forEach(id => {
        let cell = this._cells[id];
        cell.status = BROKEN;
        cell.errors.push(error);
        cell.level = Infinity;
        updated.add(id);
      });
    }

    _clearCyclicDependencyError (cell) {
      let err = cell.errors.find(err => err instanceof CyclicDependencyError);
      if (err) {
        const trace = err.details.trace;
        trace.forEach(id => {
          let cell = this._cells[id];
          cell.clearErrors(err => (err instanceof CyclicDependencyError));
          this._structureChanged.add(id);
        });
      }
    }

    _hasOutputCollision (symbol) {
      return isArray(this._out[symbol])
    }

    _removeOutputCollisionError (id) {
      let cell = this._cells[id];
      cell.clearErrors(e => e instanceof OutputCollisionError);
      this._structureChanged.add(id);
    }

    /*
      called whenever an output variable is changed
      Removes the cell id from the list of competing cells
      and removes errors if possible.
    */
    _resolveOutputCollision (symbol, id) {
      let out = this._out[symbol];
      // in case of collisions we store the competing cell ids as array
      if (isArray(out)) {
        this._removeOutputCollisionError(id);
        let s = new Set(out);
        s.delete(id);
        s = Array.from(s);
        if (s.length > 1) {
          this._out[symbol] = s;
        } else {
          let _id = s[0];
          this._out[symbol] = _id;
          this._removeOutputCollisionError(_id);
        }
      }
    }
  }

  const getIndexesFromRange$2 = tableHelpers.getIndexesFromRange;

  /*
   * A class representing symbols used in Stencila Cells to reference other cells, or ranges.
   *
   * Examples:
   *
   * - `x` or `'My Document'!x` (variable)
   * - `A1` or `'My Sheet'!A1` (cell)
   * - `A1:B10` or `'My Sheet'!A1:B10` (range)
   *
   */
  class CellSymbol {
    /*
     * @param {Symbol} s the parsed symbol
     * @param {string} docId id of the target document where the symbol can be resolved
     * @param {Cell} cell the owner cell which has this symbol as an input dependency
     */
    constructor (type, name, docId, cell) {
      // 'var' | 'cell' | 'range'
      this.type = type;
      // For example 'x', 'A1', 'A1:B10'
      this.name = name;
      // id of the document which is owning the cell
      this.docId = docId;
      // a qualified id
      this.id = qualifiedId(docId, name);
      // a link to the owning cell
      this.cell = cell;

      // only used for cell or range symbols
      this.startRow = null;
      this.startCol = null;
      this.endRow = null;
      this.endCol = null;

      this.update();
    }

    update () {
      const { type, name } = this;
      if (type === 'cell') {
        let { startRow, startCol } = getIndexesFromRange$2(name);
        this.startRow = this.endRow = startRow;
        this.startCol = this.endCol = startCol;
      } else if (type === 'range') {
        let [start, end] = name.split(':');
        let { startRow, startCol, endRow, endCol } = getIndexesFromRange$2(start, end);
        this.startRow = startRow;
        this.startCol = startCol;
        this.endRow = endRow;
        this.endCol = endCol;
      }
    }

    toString () {
      return this.id
    }
  }

  function forEach$1 (iteratee, func) {
    if (!iteratee) return
    if (iteratee.constructor.prototype.forEach) {
      iteratee.forEach(func);
    } else {
      Object.keys(iteratee).forEach(function (key) {
        return func(iteratee[key], key)
      });
    }
  }

  function isObject$2 (val) {
    const type = typeof val;
    return Boolean(val) && (type === 'object' || type === 'function')
  }

  const DEBUG$2 = false;
  let count$1 = 0;
  const COUNT_MSG$1 = '%s listeners registered in the whole system.';


  class EventEmitter$1 {
    
    emit (event) {
      if (event in this.__events__) {
        
        
        var bindings = this.__events__[event].slice();
        var args = Array.prototype.slice.call(arguments, 1);
        for (var i = 0, len = bindings.length; i < len; i++) {
          var binding = bindings[i];
          
          binding.method.apply(binding.context, args);
        }
        return true
      }
      return false
    }

    
    on (event, method, context) {
      
      _on$1.call(this, event, method, context);
    }

    
    off (event, method, context) { 
      if (arguments.length === 1 && isObject$2(arguments[0])) {
        _disconnect$1.call(this, arguments[0]);
      } else {
        _off$1.apply(this, arguments);
      }
    }

    _debugEvents () {
      
      console.log('### EventEmitter: ', this);
      forEach$1(this.__events__, (handlers, name) => {
        console.log('- %s listeners for %s: ', handlers.length, name, handlers);
      });
      
    }

    get __events__ () {
      if (!this.___events___) {
        this.___events___ = {};
      }
      return this.___events___
    }
  }


  function _on$1 (event, method, context) {
    
    var bindings;
    validateMethod$1(method, context);
    if (this.__events__.hasOwnProperty(event)) {
      bindings = this.__events__[event];
    } else {
      
      bindings = this.__events__[event] = [];
    }
    
    bindings.push({
      method: method,
      context: context || null
    });
    if (DEBUG$2) {
      count$1++;
      console.info('_on()', event, method.name, context, this);
      console.info(COUNT_MSG$1, count$1);
    }
    return this
    
  }


  function _off$1 (event, method, context) {
    
    if (arguments.length === 0) {
      if (DEBUG$2) {
        forEach$1(this.__events__, (bindings) => {
          bindings.forEach((b) => {
            console.info('_off()', b.method.name, b.context, this);
          });
          count$1 -= bindings.length;
        });
        console.info(COUNT_MSG$1, count$1);
      }
      this.___events___ = {};
      return this
    }
    if (arguments.length === 1) {
      
      if (DEBUG$2) {
        count$1 -= (this.__events__[event] || []).length;
        console.info(COUNT_MSG$1, count$1);
      }
      delete this.__events__[event];
      return this
    }
    validateMethod$1(method, context);
    if (!(event in this.__events__) || !this.__events__[event].length) {
      if (DEBUG$2) console.info('NO MATCHING BINDINGS');
      
      return this
    }
    
    if (arguments.length < 3) {
      context = null;
    }
    
    let bindings = this.__events__[event];
    for (let i = bindings.length - 1; i >= 0; i--) {
      const b = bindings[i];
      if (b.method === method && b.context === context) {
        bindings.splice(i, 1);
        if (DEBUG$2) count$1--;
      }
    }
    
    if (bindings.length === 0) {
      delete this.__events__[event];
    }
    if (DEBUG$2) console.info(COUNT_MSG$1, count$1);
    return this
    
  }


  function _disconnect$1 (context) {
    
    
    forEach$1(this.__events__, (bindings, event) => {
      for (let i = bindings.length - 1; i >= 0; i--) {
        
        
        if (bindings[i] && bindings[i].context === context) {
          _off$1.call(this, event, bindings[i].method, context);
        }
      }
    });
    return this
    
  }

  function validateMethod$1 (method, context) {
    
    if (typeof method === 'string') {
      
      if (context === undefined || context === null) {
        throw new Error('Method name "' + method + '" has no context.')
      }
      if (!(method in context)) {
        
        
        throw new Error('Method not found: "' + method + '"')
      }
      if (typeof context[method] !== 'function') {
        
        
        throw new Error('Property "' + method + '" is not a function')
      }
    } else if (typeof method !== 'function') {
      throw new Error('Invalid callback. Function or method name expected.')
    }
  }

  function uuid$1 (prefix, len) {
    if (prefix && prefix[prefix.length - 1] !== '-') {
      prefix = prefix.concat('-');
    }
    var chars = '0123456789abcdefghijklmnopqrstuvwxyz'.split('');
    var uuid = [];
    var radix = 16;
    var idx;
    len = len || 32;
    if (len) {
      
      for (idx = 0; idx < len; idx++) uuid[idx] = chars[0 | Math.random() * radix];
    } else {
      
      var r;
      
      uuid[8] = uuid[13] = uuid[18] = uuid[23] = '-';
      uuid[14] = '4';
      
      
      for (idx = 0; idx < 36; idx++) {
        if (!uuid[idx]) {
          r = 0 | Math.random() * 16;
          uuid[idx] = chars[(idx === 19) ? (r & 0x3) | 0x8 : r];
        }
      }
    }
    return (prefix || '') + uuid.join('')
  }

  function map$2 (iteratee, func) {
    if (!iteratee) return []
    if (!func) func = function (item) { return item };
    if (Array.isArray(iteratee)) {
      return iteratee.map(func)
    }
    if (iteratee instanceof Map) {
      let result = [];
      for (let [name, val] of iteratee) {
        result.push(func(val, name));
      }
      return result
    }
    if (iteratee instanceof Set) {
      let result = [];
      let idx = 0;
      iteratee.forEach(item => {
        result.push(func(item, idx++));
      });
      return result
    }
    return Object.keys(iteratee).map(function (key) {
      return func(iteratee[key], key)
    })
  }

  function isPlainObject$2 (o) {
    return Boolean(o) && o.constructor === {}.constructor
  }

  function isArray$2 (a) {
    return Array.isArray(a)
  }

  function clone$1 (val) {
    if (isArray$2(val)) {
      return val.slice(0)
    }
    if (isObject$2(val)) {
      return Object.assign({}, val)
    }
    
    
    return val
  }

  const platform$1 = {

    inBrowser: false,

    inNodeJS: false,

    inElectron: false,

    
    isIE: false,
    

    isFF: false,

    isWebkit: false,

    
    version: -1,

    

    isWindows: false,

    isMac: false,

    devtools: false,

    
    
    _reset: detect$1
  };

  function detect$1 () {
    if (typeof window !== 'undefined') {
      platform$1.inBrowser = true;

      
      const ua = window.navigator.userAgent;
      const msie = ua.indexOf('MSIE ');
      const trident = ua.indexOf('Trident/');
      const edge = ua.indexOf('Edge/');

      if (msie > 0) {
        
        platform$1.isIE = true;
        platform$1.version = 10;
        
        
      } else if (trident > 0) {
        
        platform$1.isIE = true;
        platform$1.version = 11;
        platform$1.isTrident = true;
        
        
        
      } else if (edge > 0) {
        
        platform$1.isIE = true;
        platform$1.isEdge = true;
        platform$1.version = 12;
        
        parseInt(ua.substring(edge + 5, ua.indexOf('.', edge)), 10);
      }

      
      platform$1.isFF = window.navigator.userAgent.toLowerCase().indexOf('firefox') > -1;

      
      platform$1.isWebkit = !platform$1.isFF && !platform$1.isIE;
    } else {
      platform$1.inBrowser = false;
    }

    if (platform$1.inBrowser) {
      platform$1.isWindows = (window.navigator !== undefined && window.navigator.appVersion && window.navigator.appVersion.indexOf('Win') !== -1);
      platform$1.isMac = (window.navigator !== undefined && window.navigator.platform.indexOf('Mac') >= 0);
    }

    
    if (typeof process !== 'undefined') {
      if (platform$1.inBrowser) {
        platform$1.inElectron = true;
      } else {
        platform$1.inNodeJS = true;
      }
    }

    if (platform$1.inBrowser) {
      let widthThreshold = window.outerWidth - window.innerWidth > 160;
      let heightThreshold = window.outerHeight - window.innerHeight > 160;
      let orientation = widthThreshold ? 'vertical' : 'horizontal';
      if (!(heightThreshold && widthThreshold) &&
        ((window.Firebug && window.Firebug.chrome && window.Firebug.chrome.isInitialized) || widthThreshold || heightThreshold)) {
        platform$1.devtools = true;
        platform$1.devtoolsOrientation = orientation;
      }
    }
  }

  detect$1();

  function cloneDeep$1 (val) {
    if (isArray$2(val)) {
      return _cloneArrayDeep$1(val)
    }
    
    
    
    if (platform$1.inBrowser && val instanceof window.File) {
      return val
    }
    if (isObject$2(val)) {
      return _cloneObjectDeep$1(val)
    }
    
    
    return val
  }

  function _cloneObjectDeep$1 (obj) {
    let res = {};
    forEach$1(obj, (val, key) => {
      res[key] = cloneDeep$1(val);
    });
    return res
  }

  function _cloneArrayDeep$1 (arr) {
    return arr.map(cloneDeep$1)
  }

  function isNumber$1 (n) {
    return typeof n === 'number'
  }

  function isEqual$1 (a, b) {
    if (a === b) return true
    if (isArray$2(a) && isArray$2(b)) {
      if (a.length !== b.length) return false
      for (let i = 0; i < a.length; i++) {
        if (!isEqual$1(a[i], b[i])) return false
      }
      return true
    }
    if (isPlainObject$2(a) && isPlainObject$2(b)) {
      let akeys = Object.keys(a).sort();
      let bkeys = Object.keys(b).sort();
      if (!isEqual$1(akeys, bkeys)) return false
      for (let i = 0; i < akeys.length; i++) {
        let key = akeys[i];
        if (!isEqual$1(a[key], b[key])) return false
      }
      return true
    }
    return false
  }

  function isNil$1 (o) {
    return o === null || o === undefined
  }

  function isString$1 (s) {
    return typeof s === 'string'
  }

  var freeGlobal$1 = typeof global == 'object' && global && global.Object === Object && global;

  var freeSelf$1 = typeof self == 'object' && self && self.Object === Object && self;


  var root$1 = freeGlobal$1 || freeSelf$1 || Function('return this')();

  var Symbol$1$1 = root$1.Symbol;

  function arrayMap$1(array, iteratee) {
    var index = -1,
        length = array == null ? 0 : array.length,
        result = Array(length);

    while (++index < length) {
      result[index] = iteratee(array[index], index, array);
    }
    return result;
  }

  var isArray$1$1 = Array.isArray;

  var objectProto$15 = Object.prototype;


  var hasOwnProperty$12 = objectProto$15.hasOwnProperty;


  var nativeObjectToString$2 = objectProto$15.toString;


  var symToStringTag$2 = Symbol$1$1 ? Symbol$1$1.toStringTag : undefined;


  function getRawTag$1(value) {
    var isOwn = hasOwnProperty$12.call(value, symToStringTag$2),
        tag = value[symToStringTag$2];

    try {
      value[symToStringTag$2] = undefined;
      var unmasked = true;
    } catch (e) {}

    var result = nativeObjectToString$2.call(value);
    if (unmasked) {
      if (isOwn) {
        value[symToStringTag$2] = tag;
      } else {
        delete value[symToStringTag$2];
      }
    }
    return result;
  }

  var objectProto$1$1 = Object.prototype;


  var nativeObjectToString$1$1 = objectProto$1$1.toString;


  function objectToString$1(value) {
    return nativeObjectToString$1$1.call(value);
  }

  var nullTag$1 = '[object Null]',
      undefinedTag$1 = '[object Undefined]';


  var symToStringTag$1$1 = Symbol$1$1 ? Symbol$1$1.toStringTag : undefined;


  function baseGetTag$1(value) {
    if (value == null) {
      return value === undefined ? undefinedTag$1 : nullTag$1;
    }
    return (symToStringTag$1$1 && symToStringTag$1$1 in Object(value))
      ? getRawTag$1(value)
      : objectToString$1(value);
  }

  function isObjectLike$1(value) {
    return value != null && typeof value == 'object';
  }

  var symbolTag$2 = '[object Symbol]';


  function isSymbol$1(value) {
    return typeof value == 'symbol' ||
      (isObjectLike$1(value) && baseGetTag$1(value) == symbolTag$2);
  }

  var INFINITY$4 = 1 / 0;


  var symbolProto$2 = Symbol$1$1 ? Symbol$1$1.prototype : undefined,
      symbolToString$1 = symbolProto$2 ? symbolProto$2.toString : undefined;


  function baseToString$1(value) {
    
    if (typeof value == 'string') {
      return value;
    }
    if (isArray$1$1(value)) {
      
      return arrayMap$1(value, baseToString$1) + '';
    }
    if (isSymbol$1(value)) {
      return symbolToString$1 ? symbolToString$1.call(value) : '';
    }
    var result = (value + '');
    return (result == '0' && (1 / value) == -INFINITY$4) ? '-0' : result;
  }

  function toString$1(value) {
    return value == null ? '' : baseToString$1(value);
  }

  function baseSlice$1(array, start, end) {
    var index = -1,
        length = array.length;

    if (start < 0) {
      start = -start > length ? 0 : (length + start);
    }
    end = end > length ? length : end;
    if (end < 0) {
      end += length;
    }
    length = start > end ? 0 : ((end - start) >>> 0);
    start >>>= 0;

    var result = Array(length);
    while (++index < length) {
      result[index] = array[index + start];
    }
    return result;
  }

  function isObject$1$1(value) {
    var type = typeof value;
    return value != null && (type == 'object' || type == 'function');
  }

  var reIsDeepProp$1 = /\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/,
      reIsPlainProp$1 = /^\w*$/;


  function isKey$1(value, object) {
    if (isArray$1$1(value)) {
      return false;
    }
    var type = typeof value;
    if (type == 'number' || type == 'symbol' || type == 'boolean' ||
        value == null || isSymbol$1(value)) {
      return true;
    }
    return reIsPlainProp$1.test(value) || !reIsDeepProp$1.test(value) ||
      (object != null && value in Object(object));
  }

  var asyncTag$1 = '[object AsyncFunction]',
      funcTag$2 = '[object Function]',
      genTag$1 = '[object GeneratorFunction]',
      proxyTag$1 = '[object Proxy]';


  function isFunction$2(value) {
    if (!isObject$1$1(value)) {
      return false;
    }
    
    
    var tag = baseGetTag$1(value);
    return tag == funcTag$2 || tag == genTag$1 || tag == asyncTag$1 || tag == proxyTag$1;
  }

  var coreJsData$1 = root$1['__core-js_shared__'];

  var maskSrcKey$1 = (function() {
    var uid = /[^.]+$/.exec(coreJsData$1 && coreJsData$1.keys && coreJsData$1.keys.IE_PROTO || '');
    return uid ? ('Symbol(src)_1.' + uid) : '';
  }());


  function isMasked$1(func) {
    return !!maskSrcKey$1 && (maskSrcKey$1 in func);
  }

  var funcProto$3 = Function.prototype;


  var funcToString$3 = funcProto$3.toString;


  function toSource$1(func) {
    if (func != null) {
      try {
        return funcToString$3.call(func);
      } catch (e) {}
      try {
        return (func + '');
      } catch (e) {}
    }
    return '';
  }

  var reRegExpChar$1 = /[\\^$.*+?()[\]{}|]/g;


  var reIsHostCtor$1 = /^\[object .+?Constructor\]$/;


  var funcProto$1$1 = Function.prototype,
      objectProto$2$1 = Object.prototype;


  var funcToString$1$1 = funcProto$1$1.toString;


  var hasOwnProperty$1$1 = objectProto$2$1.hasOwnProperty;


  var reIsNative$1 = RegExp('^' +
    funcToString$1$1.call(hasOwnProperty$1$1).replace(reRegExpChar$1, '\\$&')
    .replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g, '$1.*?') + '$'
  );


  function baseIsNative$1(value) {
    if (!isObject$1$1(value) || isMasked$1(value)) {
      return false;
    }
    var pattern = isFunction$2(value) ? reIsNative$1 : reIsHostCtor$1;
    return pattern.test(toSource$1(value));
  }

  function getValue$1(object, key) {
    return object == null ? undefined : object[key];
  }

  function getNative$1(object, key) {
    var value = getValue$1(object, key);
    return baseIsNative$1(value) ? value : undefined;
  }

  var nativeCreate$1 = getNative$1(Object, 'create');

  function hashClear$1() {
    this.__data__ = nativeCreate$1 ? nativeCreate$1(null) : {};
    this.size = 0;
  }

  function hashDelete$1(key) {
    var result = this.has(key) && delete this.__data__[key];
    this.size -= result ? 1 : 0;
    return result;
  }

  var HASH_UNDEFINED$3 = '__lodash_hash_undefined__';


  var objectProto$3$1 = Object.prototype;


  var hasOwnProperty$2$1 = objectProto$3$1.hasOwnProperty;


  function hashGet$1(key) {
    var data = this.__data__;
    if (nativeCreate$1) {
      var result = data[key];
      return result === HASH_UNDEFINED$3 ? undefined : result;
    }
    return hasOwnProperty$2$1.call(data, key) ? data[key] : undefined;
  }

  var objectProto$4$1 = Object.prototype;


  var hasOwnProperty$3$1 = objectProto$4$1.hasOwnProperty;


  function hashHas$1(key) {
    var data = this.__data__;
    return nativeCreate$1 ? (data[key] !== undefined) : hasOwnProperty$3$1.call(data, key);
  }

  var HASH_UNDEFINED$1$1 = '__lodash_hash_undefined__';


  function hashSet$1(key, value) {
    var data = this.__data__;
    this.size += this.has(key) ? 0 : 1;
    data[key] = (nativeCreate$1 && value === undefined) ? HASH_UNDEFINED$1$1 : value;
    return this;
  }

  function Hash$2(entries) {
    var index = -1,
        length = entries == null ? 0 : entries.length;

    this.clear();
    while (++index < length) {
      var entry = entries[index];
      this.set(entry[0], entry[1]);
    }
  }


  Hash$2.prototype.clear = hashClear$1;
  Hash$2.prototype['delete'] = hashDelete$1;
  Hash$2.prototype.get = hashGet$1;
  Hash$2.prototype.has = hashHas$1;
  Hash$2.prototype.set = hashSet$1;

  function listCacheClear$1() {
    this.__data__ = [];
    this.size = 0;
  }

  function eq$1(value, other) {
    return value === other || (value !== value && other !== other);
  }

  function assocIndexOf$1(array, key) {
    var length = array.length;
    while (length--) {
      if (eq$1(array[length][0], key)) {
        return length;
      }
    }
    return -1;
  }

  var arrayProto$1 = Array.prototype;


  var splice$1 = arrayProto$1.splice;


  function listCacheDelete$1(key) {
    var data = this.__data__,
        index = assocIndexOf$1(data, key);

    if (index < 0) {
      return false;
    }
    var lastIndex = data.length - 1;
    if (index == lastIndex) {
      data.pop();
    } else {
      splice$1.call(data, index, 1);
    }
    --this.size;
    return true;
  }

  function listCacheGet$1(key) {
    var data = this.__data__,
        index = assocIndexOf$1(data, key);

    return index < 0 ? undefined : data[index][1];
  }

  function listCacheHas$1(key) {
    return assocIndexOf$1(this.__data__, key) > -1;
  }

  function listCacheSet$1(key, value) {
    var data = this.__data__,
        index = assocIndexOf$1(data, key);

    if (index < 0) {
      ++this.size;
      data.push([key, value]);
    } else {
      data[index][1] = value;
    }
    return this;
  }

  function ListCache$1(entries) {
    var index = -1,
        length = entries == null ? 0 : entries.length;

    this.clear();
    while (++index < length) {
      var entry = entries[index];
      this.set(entry[0], entry[1]);
    }
  }


  ListCache$1.prototype.clear = listCacheClear$1;
  ListCache$1.prototype['delete'] = listCacheDelete$1;
  ListCache$1.prototype.get = listCacheGet$1;
  ListCache$1.prototype.has = listCacheHas$1;
  ListCache$1.prototype.set = listCacheSet$1;

  var Map$1$1 = getNative$1(root$1, 'Map');

  function mapCacheClear$1() {
    this.size = 0;
    this.__data__ = {
      'hash': new Hash$2,
      'map': new (Map$1$1 || ListCache$1),
      'string': new Hash$2
    };
  }

  function isKeyable$1(value) {
    var type = typeof value;
    return (type == 'string' || type == 'number' || type == 'symbol' || type == 'boolean')
      ? (value !== '__proto__')
      : (value === null);
  }

  function getMapData$1(map, key) {
    var data = map.__data__;
    return isKeyable$1(key)
      ? data[typeof key == 'string' ? 'string' : 'hash']
      : data.map;
  }

  function mapCacheDelete$1(key) {
    var result = getMapData$1(this, key)['delete'](key);
    this.size -= result ? 1 : 0;
    return result;
  }

  function mapCacheGet$1(key) {
    return getMapData$1(this, key).get(key);
  }

  function mapCacheHas$1(key) {
    return getMapData$1(this, key).has(key);
  }

  function mapCacheSet$1(key, value) {
    var data = getMapData$1(this, key),
        size = data.size;

    data.set(key, value);
    this.size += data.size == size ? 0 : 1;
    return this;
  }

  function MapCache$1(entries) {
    var index = -1,
        length = entries == null ? 0 : entries.length;

    this.clear();
    while (++index < length) {
      var entry = entries[index];
      this.set(entry[0], entry[1]);
    }
  }


  MapCache$1.prototype.clear = mapCacheClear$1;
  MapCache$1.prototype['delete'] = mapCacheDelete$1;
  MapCache$1.prototype.get = mapCacheGet$1;
  MapCache$1.prototype.has = mapCacheHas$1;
  MapCache$1.prototype.set = mapCacheSet$1;

  var FUNC_ERROR_TEXT$1$1 = 'Expected a function';


  function memoize$1(func, resolver) {
    if (typeof func != 'function' || (resolver != null && typeof resolver != 'function')) {
      throw new TypeError(FUNC_ERROR_TEXT$1$1);
    }
    var memoized = function() {
      var args = arguments,
          key = resolver ? resolver.apply(this, args) : args[0],
          cache = memoized.cache;

      if (cache.has(key)) {
        return cache.get(key);
      }
      var result = func.apply(this, args);
      memoized.cache = cache.set(key, result) || cache;
      return result;
    };
    memoized.cache = new (memoize$1.Cache || MapCache$1);
    return memoized;
  }


  memoize$1.Cache = MapCache$1;

  var MAX_MEMOIZE_SIZE$1 = 500;


  function memoizeCapped$1(func) {
    var result = memoize$1(func, function(key) {
      if (cache.size === MAX_MEMOIZE_SIZE$1) {
        cache.clear();
      }
      return key;
    });

    var cache = result.cache;
    return result;
  }

  var rePropName$1 = /[^.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(?:\.|\[\])(?:\.|\[\]|$))/g;


  var reEscapeChar$1 = /\\(\\)?/g;


  var stringToPath$1 = memoizeCapped$1(function(string) {
    var result = [];
    if (string.charCodeAt(0) === 46 ) {
      result.push('');
    }
    string.replace(rePropName$1, function(match, number, quote, subString) {
      result.push(quote ? subString.replace(reEscapeChar$1, '$1') : (number || match));
    });
    return result;
  });

  function castPath$1(value, object) {
    if (isArray$1$1(value)) {
      return value;
    }
    return isKey$1(value, object) ? [value] : stringToPath$1(toString$1(value));
  }

  var INFINITY$1$1 = 1 / 0;


  function toKey$1(value) {
    if (typeof value == 'string' || isSymbol$1(value)) {
      return value;
    }
    var result = (value + '');
    return (result == '0' && (1 / value) == -INFINITY$1$1) ? '-0' : result;
  }

  function baseGet$1(object, path) {
    path = castPath$1(path, object);

    var index = 0,
        length = path.length;

    while (object != null && index < length) {
      object = object[toKey$1(path[index++])];
    }
    return (index && index == length) ? object : undefined;
  }

  function get$1(object, path, defaultValue) {
    var result = object == null ? undefined : baseGet$1(object, path);
    return result === undefined ? defaultValue : result;
  }

  function stackClear$1() {
    this.__data__ = new ListCache$1;
    this.size = 0;
  }

  function stackDelete$1(key) {
    var data = this.__data__,
        result = data['delete'](key);

    this.size = data.size;
    return result;
  }

  function stackGet$1(key) {
    return this.__data__.get(key);
  }

  function stackHas$1(key) {
    return this.__data__.has(key);
  }

  var LARGE_ARRAY_SIZE$3 = 200;


  function stackSet$1(key, value) {
    var data = this.__data__;
    if (data instanceof ListCache$1) {
      var pairs = data.__data__;
      if (!Map$1$1 || (pairs.length < LARGE_ARRAY_SIZE$3 - 1)) {
        pairs.push([key, value]);
        this.size = ++data.size;
        return this;
      }
      data = this.__data__ = new MapCache$1(pairs);
    }
    data.set(key, value);
    this.size = data.size;
    return this;
  }

  function Stack$1(entries) {
    var data = this.__data__ = new ListCache$1(entries);
    this.size = data.size;
  }


  Stack$1.prototype.clear = stackClear$1;
  Stack$1.prototype['delete'] = stackDelete$1;
  Stack$1.prototype.get = stackGet$1;
  Stack$1.prototype.has = stackHas$1;
  Stack$1.prototype.set = stackSet$1;

  var HASH_UNDEFINED$2$1 = '__lodash_hash_undefined__';


  function setCacheAdd$1(value) {
    this.__data__.set(value, HASH_UNDEFINED$2$1);
    return this;
  }

  function setCacheHas$1(value) {
    return this.__data__.has(value);
  }

  function SetCache$1(values) {
    var index = -1,
        length = values == null ? 0 : values.length;

    this.__data__ = new MapCache$1;
    while (++index < length) {
      this.add(values[index]);
    }
  }


  SetCache$1.prototype.add = SetCache$1.prototype.push = setCacheAdd$1;
  SetCache$1.prototype.has = setCacheHas$1;

  function cacheHas$1(cache, key) {
    return cache.has(key);
  }

  var Uint8Array$1 = root$1.Uint8Array;

  function setToArray$1(set) {
    var index = -1,
        result = Array(set.size);

    set.forEach(function(value) {
      result[++index] = value;
    });
    return result;
  }


  var symbolProto$1$1 = Symbol$1$1 ? Symbol$1$1.prototype : undefined,
      symbolValueOf$1 = symbolProto$1$1 ? symbolProto$1$1.valueOf : undefined;

  function baseTimes$1(n, iteratee) {
    var index = -1,
        result = Array(n);

    while (++index < n) {
      result[index] = iteratee(index);
    }
    return result;
  }

  var argsTag$3 = '[object Arguments]';


  function baseIsArguments$1(value) {
    return isObjectLike$1(value) && baseGetTag$1(value) == argsTag$3;
  }

  var objectProto$6$1 = Object.prototype;


  var hasOwnProperty$4$1 = objectProto$6$1.hasOwnProperty;


  var propertyIsEnumerable$1$1 = objectProto$6$1.propertyIsEnumerable;


  var isArguments$1 = baseIsArguments$1(function() { return arguments; }()) ? baseIsArguments$1 : function(value) {
    return isObjectLike$1(value) && hasOwnProperty$4$1.call(value, 'callee') &&
      !propertyIsEnumerable$1$1.call(value, 'callee');
  };

  function stubFalse$1() {
    return false;
  }

  var freeExports$3 = typeof exports == 'object' && exports && !exports.nodeType && exports;


  var freeModule$3 = freeExports$3 && typeof module == 'object' && module && !module.nodeType && module;


  var moduleExports$3 = freeModule$3 && freeModule$3.exports === freeExports$3;


  var Buffer$2 = moduleExports$3 ? root$1.Buffer : undefined;


  var nativeIsBuffer$1 = Buffer$2 ? Buffer$2.isBuffer : undefined;


  var isBuffer$1 = nativeIsBuffer$1 || stubFalse$1;

  var MAX_SAFE_INTEGER$2 = 9007199254740991;


  var reIsUint$1 = /^(?:0|[1-9]\d*)$/;


  function isIndex$1(value, length) {
    var type = typeof value;
    length = length == null ? MAX_SAFE_INTEGER$2 : length;

    return !!length &&
      (type == 'number' ||
        (type != 'symbol' && reIsUint$1.test(value))) &&
          (value > -1 && value % 1 == 0 && value < length);
  }

  var MAX_SAFE_INTEGER$1$1 = 9007199254740991;


  function isLength$1(value) {
    return typeof value == 'number' &&
      value > -1 && value % 1 == 0 && value <= MAX_SAFE_INTEGER$1$1;
  }

  var argsTag$1$1 = '[object Arguments]',
      arrayTag$2 = '[object Array]',
      boolTag$1$1 = '[object Boolean]',
      dateTag$1$1 = '[object Date]',
      errorTag$1$1 = '[object Error]',
      funcTag$1$1 = '[object Function]',
      mapTag$1$1 = '[object Map]',
      numberTag$1$1 = '[object Number]',
      objectTag$4 = '[object Object]',
      regexpTag$1$1 = '[object RegExp]',
      setTag$1$1 = '[object Set]',
      stringTag$1$1 = '[object String]',
      weakMapTag$2 = '[object WeakMap]';

  var arrayBufferTag$1$1 = '[object ArrayBuffer]',
      dataViewTag$1$1 = '[object DataView]',
      float32Tag$1 = '[object Float32Array]',
      float64Tag$1 = '[object Float64Array]',
      int8Tag$1 = '[object Int8Array]',
      int16Tag$1 = '[object Int16Array]',
      int32Tag$1 = '[object Int32Array]',
      uint8Tag$1 = '[object Uint8Array]',
      uint8ClampedTag$1 = '[object Uint8ClampedArray]',
      uint16Tag$1 = '[object Uint16Array]',
      uint32Tag$1 = '[object Uint32Array]';


  var typedArrayTags$1 = {};
  typedArrayTags$1[float32Tag$1] = typedArrayTags$1[float64Tag$1] =
  typedArrayTags$1[int8Tag$1] = typedArrayTags$1[int16Tag$1] =
  typedArrayTags$1[int32Tag$1] = typedArrayTags$1[uint8Tag$1] =
  typedArrayTags$1[uint8ClampedTag$1] = typedArrayTags$1[uint16Tag$1] =
  typedArrayTags$1[uint32Tag$1] = true;
  typedArrayTags$1[argsTag$1$1] = typedArrayTags$1[arrayTag$2] =
  typedArrayTags$1[arrayBufferTag$1$1] = typedArrayTags$1[boolTag$1$1] =
  typedArrayTags$1[dataViewTag$1$1] = typedArrayTags$1[dateTag$1$1] =
  typedArrayTags$1[errorTag$1$1] = typedArrayTags$1[funcTag$1$1] =
  typedArrayTags$1[mapTag$1$1] = typedArrayTags$1[numberTag$1$1] =
  typedArrayTags$1[objectTag$4] = typedArrayTags$1[regexpTag$1$1] =
  typedArrayTags$1[setTag$1$1] = typedArrayTags$1[stringTag$1$1] =
  typedArrayTags$1[weakMapTag$2] = false;


  function baseIsTypedArray$1(value) {
    return isObjectLike$1(value) &&
      isLength$1(value.length) && !!typedArrayTags$1[baseGetTag$1(value)];
  }

  function baseUnary$1(func) {
    return function(value) {
      return func(value);
    };
  }

  var freeExports$1$1 = typeof exports == 'object' && exports && !exports.nodeType && exports;


  var freeModule$1$1 = freeExports$1$1 && typeof module == 'object' && module && !module.nodeType && module;


  var moduleExports$1$1 = freeModule$1$1 && freeModule$1$1.exports === freeExports$1$1;


  var freeProcess$1 = moduleExports$1$1 && freeGlobal$1.process;


  var nodeUtil$1 = (function() {
    try {
      
      var types = freeModule$1$1 && freeModule$1$1.require && freeModule$1$1.require('util').types;

      if (types) {
        return types;
      }

      
      return freeProcess$1 && freeProcess$1.binding && freeProcess$1.binding('util');
    } catch (e) {}
  }());

  var nodeIsTypedArray$1 = nodeUtil$1 && nodeUtil$1.isTypedArray;


  var isTypedArray$1 = nodeIsTypedArray$1 ? baseUnary$1(nodeIsTypedArray$1) : baseIsTypedArray$1;

  var objectProto$7$1 = Object.prototype;


  var hasOwnProperty$5$1 = objectProto$7$1.hasOwnProperty;


  function arrayLikeKeys$1(value, inherited) {
    var isArr = isArray$1$1(value),
        isArg = !isArr && isArguments$1(value),
        isBuff = !isArr && !isArg && isBuffer$1(value),
        isType = !isArr && !isArg && !isBuff && isTypedArray$1(value),
        skipIndexes = isArr || isArg || isBuff || isType,
        result = skipIndexes ? baseTimes$1(value.length, String) : [],
        length = result.length;

    for (var key in value) {
      if ((inherited || hasOwnProperty$5$1.call(value, key)) &&
          !(skipIndexes && (
             
             key == 'length' ||
             
             (isBuff && (key == 'offset' || key == 'parent')) ||
             
             (isType && (key == 'buffer' || key == 'byteLength' || key == 'byteOffset')) ||
             
             isIndex$1(key, length)
          ))) {
        result.push(key);
      }
    }
    return result;
  }

  var objectProto$8$1 = Object.prototype;


  function isPrototype$1(value) {
    var Ctor = value && value.constructor,
        proto = (typeof Ctor == 'function' && Ctor.prototype) || objectProto$8$1;

    return value === proto;
  }

  function overArg$1(func, transform) {
    return function(arg) {
      return func(transform(arg));
    };
  }

  function isArrayLike$1(value) {
    return value != null && isLength$1(value.length) && !isFunction$2(value);
  }

  var DataView$1 = getNative$1(root$1, 'DataView');

  var Promise$1$1 = getNative$1(root$1, 'Promise');

  var Set$1$1 = getNative$1(root$1, 'Set');

  var WeakMap$1$1 = getNative$1(root$1, 'WeakMap');

  var mapTag$2$1 = '[object Map]',
      objectTag$1$1 = '[object Object]',
      promiseTag$1 = '[object Promise]',
      setTag$2$1 = '[object Set]',
      weakMapTag$1$1 = '[object WeakMap]';

  var dataViewTag$2$1 = '[object DataView]';


  var dataViewCtorString$1 = toSource$1(DataView$1),
      mapCtorString$1 = toSource$1(Map$1$1),
      promiseCtorString$1 = toSource$1(Promise$1$1),
      setCtorString$1 = toSource$1(Set$1$1),
      weakMapCtorString$1 = toSource$1(WeakMap$1$1);


  var getTag$2 = baseGetTag$1;


  if ((DataView$1 && getTag$2(new DataView$1(new ArrayBuffer(1))) != dataViewTag$2$1) ||
      (Map$1$1 && getTag$2(new Map$1$1) != mapTag$2$1) ||
      (Promise$1$1 && getTag$2(Promise$1$1.resolve()) != promiseTag$1) ||
      (Set$1$1 && getTag$2(new Set$1$1) != setTag$2$1) ||
      (WeakMap$1$1 && getTag$2(new WeakMap$1$1) != weakMapTag$1$1)) {
    getTag$2 = function(value) {
      var result = baseGetTag$1(value),
          Ctor = result == objectTag$1$1 ? value.constructor : undefined,
          ctorString = Ctor ? toSource$1(Ctor) : '';

      if (ctorString) {
        switch (ctorString) {
          case dataViewCtorString$1: return dataViewTag$2$1;
          case mapCtorString$1: return mapTag$2$1;
          case promiseCtorString$1: return promiseTag$1;
          case setCtorString$1: return setTag$2$1;
          case weakMapCtorString$1: return weakMapTag$1$1;
        }
      }
      return result;
    };
  }

  function identity$1(value) {
    return value;
  }

  function baseFindIndex$1(array, predicate, fromIndex, fromRight) {
    var length = array.length,
        index = fromIndex + (fromRight ? 1 : -1);

    while ((fromRight ? index-- : ++index < length)) {
      if (predicate(array[index], index, array)) {
        return index;
      }
    }
    return -1;
  }

  var defineProperty$1 = (function() {
    try {
      var func = getNative$1(Object, 'defineProperty');
      func({}, '', {});
      return func;
    } catch (e) {}
  }());

  function baseAssignValue$1(object, key, value) {
    if (key == '__proto__' && defineProperty$1) {
      defineProperty$1(object, key, {
        'configurable': true,
        'enumerable': true,
        'value': value,
        'writable': true
      });
    } else {
      object[key] = value;
    }
  }

  function assignMergeValue$1(object, key, value) {
    if ((value !== undefined && !eq$1(object[key], value)) ||
        (value === undefined && !(key in object))) {
      baseAssignValue$1(object, key, value);
    }
  }

  function createBaseFor$1(fromRight) {
    return function(object, iteratee, keysFunc) {
      var index = -1,
          iterable = Object(object),
          props = keysFunc(object),
          length = props.length;

      while (length--) {
        var key = props[fromRight ? length : ++index];
        if (iteratee(iterable[key], key, iterable) === false) {
          break;
        }
      }
      return object;
    };
  }

  var baseFor$1 = createBaseFor$1();

  var freeExports$2$1 = typeof exports == 'object' && exports && !exports.nodeType && exports;


  var freeModule$2$1 = freeExports$2$1 && typeof module == 'object' && module && !module.nodeType && module;


  var moduleExports$2$1 = freeModule$2$1 && freeModule$2$1.exports === freeExports$2$1;


  var Buffer$1$1 = moduleExports$2$1 ? root$1.Buffer : undefined,
      allocUnsafe$1 = Buffer$1$1 ? Buffer$1$1.allocUnsafe : undefined;


  function cloneBuffer$1(buffer, isDeep) {
    if (isDeep) {
      return buffer.slice();
    }
    var length = buffer.length,
        result = allocUnsafe$1 ? allocUnsafe$1(length) : new buffer.constructor(length);

    buffer.copy(result);
    return result;
  }

  function cloneArrayBuffer$1(arrayBuffer) {
    var result = new arrayBuffer.constructor(arrayBuffer.byteLength);
    new Uint8Array$1(result).set(new Uint8Array$1(arrayBuffer));
    return result;
  }

  function cloneTypedArray$1(typedArray, isDeep) {
    var buffer = isDeep ? cloneArrayBuffer$1(typedArray.buffer) : typedArray.buffer;
    return new typedArray.constructor(buffer, typedArray.byteOffset, typedArray.length);
  }

  function copyArray$1(source, array) {
    var index = -1,
        length = source.length;

    array || (array = Array(length));
    while (++index < length) {
      array[index] = source[index];
    }
    return array;
  }

  var objectCreate$1 = Object.create;


  var baseCreate$1 = (function() {
    function object() {}
    return function(proto) {
      if (!isObject$1$1(proto)) {
        return {};
      }
      if (objectCreate$1) {
        return objectCreate$1(proto);
      }
      object.prototype = proto;
      var result = new object;
      object.prototype = undefined;
      return result;
    };
  }());

  var getPrototype$1 = overArg$1(Object.getPrototypeOf, Object);

  function initCloneObject$1(object) {
    return (typeof object.constructor == 'function' && !isPrototype$1(object))
      ? baseCreate$1(getPrototype$1(object))
      : {};
  }

  function isArrayLikeObject$1(value) {
    return isObjectLike$1(value) && isArrayLike$1(value);
  }

  var objectTag$3$1 = '[object Object]';


  var funcProto$2$1 = Function.prototype,
      objectProto$12$1 = Object.prototype;


  var funcToString$2$1 = funcProto$2$1.toString;


  var hasOwnProperty$9$1 = objectProto$12$1.hasOwnProperty;


  var objectCtorString$1 = funcToString$2$1.call(Object);


  function isPlainObject$1$1(value) {
    if (!isObjectLike$1(value) || baseGetTag$1(value) != objectTag$3$1) {
      return false;
    }
    var proto = getPrototype$1(value);
    if (proto === null) {
      return true;
    }
    var Ctor = hasOwnProperty$9$1.call(proto, 'constructor') && proto.constructor;
    return typeof Ctor == 'function' && Ctor instanceof Ctor &&
      funcToString$2$1.call(Ctor) == objectCtorString$1;
  }

  function safeGet$1(object, key) {
    return key == '__proto__'
      ? undefined
      : object[key];
  }

  var objectProto$13$1 = Object.prototype;


  var hasOwnProperty$10$1 = objectProto$13$1.hasOwnProperty;


  function assignValue$1(object, key, value) {
    var objValue = object[key];
    if (!(hasOwnProperty$10$1.call(object, key) && eq$1(objValue, value)) ||
        (value === undefined && !(key in object))) {
      baseAssignValue$1(object, key, value);
    }
  }

  function copyObject$1(source, props, object, customizer) {
    var isNew = !object;
    object || (object = {});

    var index = -1,
        length = props.length;

    while (++index < length) {
      var key = props[index];

      var newValue = customizer
        ? customizer(object[key], source[key], key, object, source)
        : undefined;

      if (newValue === undefined) {
        newValue = source[key];
      }
      if (isNew) {
        baseAssignValue$1(object, key, newValue);
      } else {
        assignValue$1(object, key, newValue);
      }
    }
    return object;
  }

  function nativeKeysIn$1(object) {
    var result = [];
    if (object != null) {
      for (var key in Object(object)) {
        result.push(key);
      }
    }
    return result;
  }

  var objectProto$14$1 = Object.prototype;


  var hasOwnProperty$11$1 = objectProto$14$1.hasOwnProperty;


  function baseKeysIn$1(object) {
    if (!isObject$1$1(object)) {
      return nativeKeysIn$1(object);
    }
    var isProto = isPrototype$1(object),
        result = [];

    for (var key in object) {
      if (!(key == 'constructor' && (isProto || !hasOwnProperty$11$1.call(object, key)))) {
        result.push(key);
      }
    }
    return result;
  }

  function keysIn$1(object) {
    return isArrayLike$1(object) ? arrayLikeKeys$1(object, true) : baseKeysIn$1(object);
  }

  function toPlainObject$1(value) {
    return copyObject$1(value, keysIn$1(value));
  }

  function baseMergeDeep$1(object, source, key, srcIndex, mergeFunc, customizer, stack) {
    var objValue = safeGet$1(object, key),
        srcValue = safeGet$1(source, key),
        stacked = stack.get(srcValue);

    if (stacked) {
      assignMergeValue$1(object, key, stacked);
      return;
    }
    var newValue = customizer
      ? customizer(objValue, srcValue, (key + ''), object, source, stack)
      : undefined;

    var isCommon = newValue === undefined;

    if (isCommon) {
      var isArr = isArray$1$1(srcValue),
          isBuff = !isArr && isBuffer$1(srcValue),
          isTyped = !isArr && !isBuff && isTypedArray$1(srcValue);

      newValue = srcValue;
      if (isArr || isBuff || isTyped) {
        if (isArray$1$1(objValue)) {
          newValue = objValue;
        }
        else if (isArrayLikeObject$1(objValue)) {
          newValue = copyArray$1(objValue);
        }
        else if (isBuff) {
          isCommon = false;
          newValue = cloneBuffer$1(srcValue, true);
        }
        else if (isTyped) {
          isCommon = false;
          newValue = cloneTypedArray$1(srcValue, true);
        }
        else {
          newValue = [];
        }
      }
      else if (isPlainObject$1$1(srcValue) || isArguments$1(srcValue)) {
        newValue = objValue;
        if (isArguments$1(objValue)) {
          newValue = toPlainObject$1(objValue);
        }
        else if (!isObject$1$1(objValue) || (srcIndex && isFunction$2(objValue))) {
          newValue = initCloneObject$1(srcValue);
        }
      }
      else {
        isCommon = false;
      }
    }
    if (isCommon) {
      
      stack.set(srcValue, newValue);
      mergeFunc(newValue, srcValue, srcIndex, customizer, stack);
      stack['delete'](srcValue);
    }
    assignMergeValue$1(object, key, newValue);
  }

  function baseMerge$1(object, source, srcIndex, customizer, stack) {
    if (object === source) {
      return;
    }
    baseFor$1(source, function(srcValue, key) {
      if (isObject$1$1(srcValue)) {
        stack || (stack = new Stack$1);
        baseMergeDeep$1(object, source, key, srcIndex, baseMerge$1, customizer, stack);
      }
      else {
        var newValue = customizer
          ? customizer(safeGet$1(object, key), srcValue, (key + ''), object, source, stack)
          : undefined;

        if (newValue === undefined) {
          newValue = srcValue;
        }
        assignMergeValue$1(object, key, newValue);
      }
    }, keysIn$1);
  }

  function apply$1(func, thisArg, args) {
    switch (args.length) {
      case 0: return func.call(thisArg);
      case 1: return func.call(thisArg, args[0]);
      case 2: return func.call(thisArg, args[0], args[1]);
      case 3: return func.call(thisArg, args[0], args[1], args[2]);
    }
    return func.apply(thisArg, args);
  }

  var nativeMax$2$1 = Math.max;


  function overRest$1(func, start, transform) {
    start = nativeMax$2$1(start === undefined ? (func.length - 1) : start, 0);
    return function() {
      var args = arguments,
          index = -1,
          length = nativeMax$2$1(args.length - start, 0),
          array = Array(length);

      while (++index < length) {
        array[index] = args[start + index];
      }
      index = -1;
      var otherArgs = Array(start + 1);
      while (++index < start) {
        otherArgs[index] = args[index];
      }
      otherArgs[start] = transform(array);
      return apply$1(func, this, otherArgs);
    };
  }

  function constant$1(value) {
    return function() {
      return value;
    };
  }

  var baseSetToString$1 = !defineProperty$1 ? identity$1 : function(func, string) {
    return defineProperty$1(func, 'toString', {
      'configurable': true,
      'enumerable': false,
      'value': constant$1(string),
      'writable': true
    });
  };

  var HOT_COUNT$1 = 800,
      HOT_SPAN$1 = 16;


  var nativeNow$1 = Date.now;


  function shortOut$1(func) {
    var count = 0,
        lastCalled = 0;

    return function() {
      var stamp = nativeNow$1(),
          remaining = HOT_SPAN$1 - (stamp - lastCalled);

      lastCalled = stamp;
      if (remaining > 0) {
        if (++count >= HOT_COUNT$1) {
          return arguments[0];
        }
      } else {
        count = 0;
      }
      return func.apply(undefined, arguments);
    };
  }

  var setToString$1 = shortOut$1(baseSetToString$1);

  function baseRest$1(func, start) {
    return setToString$1(overRest$1(func, start, identity$1), func + '');
  }

  function isIterateeCall$1(value, index, object) {
    if (!isObject$1$1(object)) {
      return false;
    }
    var type = typeof index;
    if (type == 'number'
          ? (isArrayLike$1(object) && isIndex$1(index, object.length))
          : (type == 'string' && index in object)
        ) {
      return eq$1(object[index], value);
    }
    return false;
  }

  function createAssigner$1(assigner) {
    return baseRest$1(function(object, sources) {
      var index = -1,
          length = sources.length,
          customizer = length > 1 ? sources[length - 1] : undefined,
          guard = length > 2 ? sources[2] : undefined;

      customizer = (assigner.length > 3 && typeof customizer == 'function')
        ? (length--, customizer)
        : undefined;

      if (guard && isIterateeCall$1(sources[0], sources[1], guard)) {
        customizer = length < 3 ? undefined : customizer;
        length = 1;
      }
      object = Object(object);
      while (++index < length) {
        var source = sources[index];
        if (source) {
          assigner(object, source, index, customizer);
        }
      }
      return object;
    });
  }

  var merge$2 = createAssigner$1(function(object, source, srcIndex) {
    baseMerge$1(object, source, srcIndex);
  });

  var mergeWith$1 = createAssigner$1(function(object, source, srcIndex, customizer) {
    baseMerge$1(object, source, srcIndex, customizer);
  });

  function baseSet$1(object, path, value, customizer) {
    if (!isObject$1$1(object)) {
      return object;
    }
    path = castPath$1(path, object);

    var index = -1,
        length = path.length,
        lastIndex = length - 1,
        nested = object;

    while (nested != null && ++index < length) {
      var key = toKey$1(path[index]),
          newValue = value;

      if (index != lastIndex) {
        var objValue = nested[key];
        newValue = customizer ? customizer(objValue, key, nested) : undefined;
        if (newValue === undefined) {
          newValue = isObject$1$1(objValue)
            ? objValue
            : (isIndex$1(path[index + 1]) ? [] : {});
        }
      }
      assignValue$1(nested, key, newValue);
      nested = nested[key];
    }
    return object;
  }

  function setWith$1(object, path, value, customizer) {
    customizer = typeof customizer == 'function' ? customizer : undefined;
    return object == null ? object : baseSet$1(object, path, value, customizer);
  }

  function baseIsNaN$1(value) {
    return value !== value;
  }

  function strictIndexOf$1(array, value, fromIndex) {
    var index = fromIndex - 1,
        length = array.length;

    while (++index < length) {
      if (array[index] === value) {
        return index;
      }
    }
    return -1;
  }

  function baseIndexOf$1(array, value, fromIndex) {
    return value === value
      ? strictIndexOf$1(array, value, fromIndex)
      : baseFindIndex$1(array, baseIsNaN$1, fromIndex);
  }

  function arrayIncludes$1(array, value) {
    var length = array == null ? 0 : array.length;
    return !!length && baseIndexOf$1(array, value, 0) > -1;
  }

  function arrayIncludesWith$1(array, value, comparator) {
    var index = -1,
        length = array == null ? 0 : array.length;

    while (++index < length) {
      if (comparator(value, array[index])) {
        return true;
      }
    }
    return false;
  }

  function noop$1() {
    
  }

  var INFINITY$3$1 = 1 / 0;


  var createSet$1 = !(Set$1$1 && (1 / setToArray$1(new Set$1$1([,-0]))[1]) == INFINITY$3$1) ? noop$1 : function(values) {
    return new Set$1$1(values);
  };

  function last$2(array) {
    var length = array == null ? 0 : array.length;
    return length ? array[length - 1] : undefined;
  }

  function parent$2(object, path) {
    return path.length < 2 ? object : baseGet$1(object, baseSlice$1(path, 0, -1));
  }

  function baseUnset$1(object, path) {
    path = castPath$1(path, object);
    object = parent$2(object, path);
    return object == null || delete object[toKey$1(last$2(path))];
  }

  function unset$1(object, path) {
    return object == null ? true : baseUnset$1(object, path);
  }

  var LARGE_ARRAY_SIZE$2$1 = 200;


  function baseDifference$1(array, values, iteratee, comparator) {
    var index = -1,
        includes = arrayIncludes$1,
        isCommon = true,
        length = array.length,
        result = [],
        valuesLength = values.length;

    if (!length) {
      return result;
    }
    if (iteratee) {
      values = arrayMap$1(values, baseUnary$1(iteratee));
    }
    if (comparator) {
      includes = arrayIncludesWith$1;
      isCommon = false;
    }
    else if (values.length >= LARGE_ARRAY_SIZE$2$1) {
      includes = cacheHas$1;
      isCommon = false;
      values = new SetCache$1(values);
    }
    outer:
    while (++index < length) {
      var value = array[index],
          computed = iteratee == null ? value : iteratee(value);

      value = (comparator || value !== 0) ? value : 0;
      if (isCommon && computed === computed) {
        var valuesIndex = valuesLength;
        while (valuesIndex--) {
          if (values[valuesIndex] === computed) {
            continue outer;
          }
        }
        result.push(value);
      }
      else if (!includes(values, computed, comparator)) {
        result.push(value);
      }
    }
    return result;
  }

  var without$1 = baseRest$1(function(array, values) {
    return isArrayLikeObject$1(array)
      ? baseDifference$1(array, values)
      : [];
  });

  class PathObject$1 {
    
    constructor (root) {
      if (root) {
        this.__root__ = root;
      }
    }

    contains (id) {
      return Boolean(this.getRoot()[id])
    }

    getRoot () {
      if (this.__root__) {
        return this.__root__
      } else {
        return this
      }
    }

    
    get (path) {
      if (!path) {
        return undefined
      }
      if (isString$1(path)) {
        return this.getRoot()[path]
      }
      if (arguments.length > 1) {
        path = Array.prototype.slice(arguments, 0);
      }
      if (!isArray$2(path)) {
        throw new Error('Illegal argument for PathObject.get()')
      }
      return get$1(this.getRoot(), path)
    }

    set (path, value) {
      if (!path) {
        throw new Error('Illegal argument: PathObject.set(>path<, value) - path is mandatory.')
      }
      if (isString$1(path)) {
        this.getRoot()[path] = value;
      } else {
        setWith$1(this.getRoot(), path, value);
      }
    }

    delete (path) {
      if (isString$1(path)) {
        delete this.getRoot()[path];
      } else if (path.length === 1) {
        delete this.getRoot()[path[0]];
      } else {
        var success = unset$1(this.getRoot(), path);
        if (!success) {
          throw new Error('Could not delete property at path' + path)
        }
      }
    }

    clear () {
      var root = this.getRoot();
      for (var key in root) {
        if (root.hasOwnProperty(key)) {
          delete root[key];
        }
      }
    }
  }

  PathObject$1.prototype._isPathObject = true;

  class Conflict$1 extends Error {
    constructor (a, b) {
      super('Conflict: ' + JSON.stringify(a) + ' vs ' + JSON.stringify(b));
      this.a = a;
      this.b = b;
    }
  }

  const INSERT$2 = 'insert';
  const DELETE$3 = 'delete';

  class TextOperation$1 {
    constructor (data) {
      if (!data || data.type === undefined || data.pos === undefined || data.str === undefined) {
        throw new Error('Illegal argument: insufficient data.')
      }
      
      this.type = data.type;
      
      this.pos = data.pos;
      
      this.str = data.str;
      
      if (!this.isInsert() && !this.isDelete()) {
        throw new Error('Illegal type.')
      }
      if (!isString$1(this.str)) {
        throw new Error('Illegal argument: expecting string.')
      }
      if (!isNumber$1(this.pos) || this.pos < 0) {
        throw new Error('Illegal argument: expecting positive number as pos.')
      }
    }

    apply (str) {
      if (this.isEmpty()) return str
      if (this.type === INSERT$2) {
        if (str.length < this.pos) {
          throw new Error('Provided string is too short.')
        }
        if (str.splice) {
          return str.splice(this.pos, 0, this.str)
        } else {
          return str.slice(0, this.pos).concat(this.str).concat(str.slice(this.pos))
        }
      } else  {
        if (str.length < this.pos + this.str.length) {
          throw new Error('Provided string is too short.')
        }
        if (str.splice) {
          return str.splice(this.pos, this.str.length)
        } else {
          return str.slice(0, this.pos).concat(str.slice(this.pos + this.str.length))
        }
      }
    }

    clone () {
      return new TextOperation$1(this)
    }

    isNOP () {
      return this.type === 'NOP' || this.str.length === 0
    }

    isInsert () {
      return this.type === INSERT$2
    }

    isDelete () {
      return this.type === DELETE$3
    }

    getLength () {
      return this.str.length
    }

    invert () {
      var data = {
        type: this.isInsert() ? DELETE$3 : INSERT$2,
        pos: this.pos,
        str: this.str
      };
      return new TextOperation$1(data)
    }

    hasConflict (other) {
      return _hasConflict$1(this, other)
    }

    isEmpty () {
      return this.str.length === 0
    }

    toJSON () {
      return {
        type: this.type,
        pos: this.pos,
        str: this.str
      }
    }

    toString () {
      return ['(', (this.isInsert() ? INSERT$2 : DELETE$3), ',', this.pos, ",'", this.str, "')"].join('')
    }
  }

  TextOperation$1.prototype._isOperation = true;
  TextOperation$1.prototype._isTextOperation = true;

  function _hasConflict$1 (a, b) {
    
    
    
    if (a.type === INSERT$2 && b.type === INSERT$2) return (a.pos === b.pos)
    
    
    
    if (a.type === DELETE$3 && b.type === DELETE$3) {
      
      return !(a.pos >= b.pos + b.str.length || b.pos >= a.pos + a.str.length)
    }
    
    
    
    var del, ins;
    if (a.type === DELETE$3) {
      del = a; ins = b;
    } else {
      del = b; ins = a;
    }
    return (ins.pos >= del.pos && ins.pos < del.pos + del.str.length)
  }




  function transformInsertInsert$2 (a, b) {
    if (a.pos === b.pos) {
      b.pos += a.str.length;
    } else if (a.pos < b.pos) {
      b.pos += a.str.length;
    } else {
      a.pos += b.str.length;
    }
  }





  function transformDeleteDelete$3 (a, b, first) {
    
    if (a.pos > b.pos) {
      return transformDeleteDelete$3(b, a, !first)
    }
    if (a.pos === b.pos && a.str.length > b.str.length) {
      return transformDeleteDelete$3(b, a, !first)
    }
    
    if (b.pos < a.pos + a.str.length) {
      var s = b.pos - a.pos;
      var s1 = a.str.length - s;
      var s2 = s + b.str.length;
      a.str = a.str.slice(0, s) + a.str.slice(s2);
      b.str = b.str.slice(s1);
      b.pos -= s;
    } else {
      b.pos -= a.str.length;
    }
  }





  function transformInsertDelete$2 (a, b) {
    if (a.type === DELETE$3) {
      return transformInsertDelete$2(b, a)
    }
    
    
    if (a.pos <= b.pos) {
      b.pos += a.str.length;
    
    } else if (a.pos >= b.pos + b.str.length) {
      a.pos -= b.str.length;
    
    
    
    } else {
      var s = a.pos - b.pos;
      b.str = b.str.slice(0, s) + a.str + b.str.slice(s);
      a.str = '';
    }
  }

  function transform$4 (a, b, options) {
    options = options || {};
    if (options['no-conflict'] && _hasConflict$1(a, b)) {
      throw new Conflict$1(a, b)
    }
    if (!options.inplace) {
      a = a.clone();
      b = b.clone();
    }
    if (a.type === INSERT$2 && b.type === INSERT$2) {
      transformInsertInsert$2(a, b);
    } else if (a.type === DELETE$3 && b.type === DELETE$3) {
      transformDeleteDelete$3(a, b, true);
    } else {
      transformInsertDelete$2(a, b);
    }
    return [a, b]
  }

  TextOperation$1.transform = function () {
    return transform$4.apply(null, arguments)
  };



  TextOperation$1.Insert = function (pos, str) {
    return new TextOperation$1({ type: INSERT$2, pos: pos, str: str })
  };

  TextOperation$1.Delete = function (pos, str) {
    return new TextOperation$1({ type: DELETE$3, pos: pos, str: str })
  };

  TextOperation$1.INSERT = INSERT$2;
  TextOperation$1.DELETE = DELETE$3;

  TextOperation$1.fromJSON = function (data) {
    return new TextOperation$1(data)
  };

  const NOP$2 = 'NOP';
  const DELETE$1$1 = 'delete';
  const INSERT$1$1 = 'insert';

  class ArrayOperation$1 {
    constructor (data) {
      if (!data || !data.type) {
        throw new Error('Illegal argument: insufficient data.')
      }
      this.type = data.type;
      if (this.type === NOP$2) return

      if (this.type !== INSERT$1$1 && this.type !== DELETE$1$1) {
        throw new Error('Illegal type.')
      }
      
      this.pos = data.pos;
      
      this.val = data.val;
      if (!isNumber$1(this.pos) || this.pos < 0) {
        throw new Error('Illegal argument: expecting positive number as pos.')
      }
    }

    apply (array) {
      if (this.type === NOP$2) {
        return array
      }
      if (this.type === INSERT$1$1) {
        if (array.length < this.pos) {
          throw new Error('Provided array is too small.')
        }
        array.splice(this.pos, 0, this.val);
        return array
      
      } else  {
        if (array.length < this.pos) {
          throw new Error('Provided array is too small.')
        }
        if (!isEqual$1(array[this.pos], this.val)) {
          throw Error('Unexpected value at position ' + this.pos + '. Expected ' + this.val + ', found ' + array[this.pos])
        }
        array.splice(this.pos, 1);
        return array
      }
    }

    clone () {
      var data = {
        type: this.type,
        pos: this.pos,
        val: cloneDeep$1(this.val)
      };
      return new ArrayOperation$1(data)
    }

    invert () {
      var data = this.toJSON();
      if (this.type === NOP$2) data.type = NOP$2;
      else if (this.type === INSERT$1$1) data.type = DELETE$1$1;
      else  data.type = INSERT$1$1;
      return new ArrayOperation$1(data)
    }

    hasConflict (other) {
      return ArrayOperation$1.hasConflict(this, other)
    }

    toJSON () {
      var result = {
        type: this.type
      };
      if (this.type === NOP$2) return result
      result.pos = this.pos;
      result.val = cloneDeep$1(this.val);
      return result
    }

    isInsert () {
      return this.type === INSERT$1$1
    }

    isDelete () {
      return this.type === DELETE$1$1
    }

    getOffset () {
      return this.pos
    }

    getValue () {
      return this.val
    }

    isNOP () {
      return this.type === NOP$2
    }

    toString () {
      return ['(', (this.isInsert() ? INSERT$1$1 : DELETE$1$1), ',', this.getOffset(), ",'", this.getValue(), "')"].join('')
    }
  }

  ArrayOperation$1.prototype._isOperation = true;
  ArrayOperation$1.prototype._isArrayOperation = true;

  function hasConflict$2 (a, b) {
    if (a.type === NOP$2 || b.type === NOP$2) return false
    if (a.type === INSERT$1$1 && b.type === INSERT$1$1) {
      return a.pos === b.pos
    } else {
      return false
    }
  }

  function transformInsertInsert$1$1 (a, b) {
    if (a.pos === b.pos) {
      b.pos += 1;
    
    } else if (a.pos < b.pos) {
      b.pos += 1;
    
    } else {
      a.pos += 1;
    }
  }

  function transformDeleteDelete$1$1 (a, b) {
    
    if (a.pos === b.pos) {
      b.type = NOP$2;
      a.type = NOP$2;
      return
    }
    if (a.pos < b.pos) {
      b.pos -= 1;
    } else {
      a.pos -= 1;
    }
  }

  function transformInsertDelete$1$1 (a, b) {
    
    if (a.type === DELETE$1$1) {
      var tmp = a;
      a = b;
      b = tmp;
    }
    if (a.pos <= b.pos) {
      b.pos += 1;
    } else {
      a.pos -= 1;
    }
  }

  var transform$1$1 = function (a, b, options) {
    options = options || {};
    
    
    if (options['no-conflict'] && hasConflict$2(a, b)) {
      throw new Conflict$1(a, b)
    }
    
    if (!options.inplace) {
      a = a.clone();
      b = b.clone();
    }
    if (a.type === NOP$2 || b.type === NOP$2) ; else if (a.type === INSERT$1$1 && b.type === INSERT$1$1) {
      transformInsertInsert$1$1(a, b);
    } else if (a.type === DELETE$1$1 && b.type === DELETE$1$1) {
      transformDeleteDelete$1$1(a, b);
    } else {
      transformInsertDelete$1$1(a, b);
    }
    return [a, b]
  };

  ArrayOperation$1.transform = transform$1$1;
  ArrayOperation$1.hasConflict = hasConflict$2;



  ArrayOperation$1.Insert = function (pos, val) {
    return new ArrayOperation$1({type: INSERT$1$1, pos: pos, val: val})
  };

  ArrayOperation$1.Delete = function (pos, val) {
    return new ArrayOperation$1({ type: DELETE$1$1, pos: pos, val: val })
  };

  ArrayOperation$1.fromJSON = function (data) {
    return new ArrayOperation$1(data)
  };

  ArrayOperation$1.NOP = NOP$2;
  ArrayOperation$1.DELETE = DELETE$1$1;
  ArrayOperation$1.INSERT = INSERT$1$1;

  const SHIFT$1 = 'shift';

  class CoordinateOperation$1 {
    constructor (data) {
      if (!data || data.type === undefined) {
        throw new Error('Illegal argument: insufficient data.')
      }
      
      this.type = data.type;
      
      this.val = data.val;
      
      if (!this.isShift()) {
        throw new Error('Illegal type.')
      }
      if (!isNumber$1(this.val)) {
        throw new Error('Illegal argument: expecting number as shift value.')
      }
    }

    apply (coor) {
      coor.offset = coor.offset + this.val;
    }

    isShift () {
      return this.type === SHIFT$1
    }

    isNOP () {
      switch (this.type) {
        case SHIFT$1: {
          return this.val === 0
        }
        default:
          return false
      }
    }

    clone () {
      return new CoordinateOperation$1(this)
    }

    invert () {
      let data;
      switch (this.type) {
        case SHIFT$1:
          data = {
            type: SHIFT$1,
            val: -this.val
          };
          break
        default:
          throw new Error('Invalid type.')
      }
      return new CoordinateOperation$1(data)
    }

    hasConflict () {
      
      return false
    }

    toJSON () {
      return {
        type: this.type,
        val: this.val
      }
    }

    toString () {
      return ['(', (this.type), ',', this.val, "')"].join('')
    }
  }

  CoordinateOperation$1.prototype._isOperation = true;
  CoordinateOperation$1.prototype._isCoordinateOperation = true;

  function transformShiftShift$1 (a, b) {
    a.val += b.val;
    b.val += a.val;
  }

  function transform$2$1 (a, b, options) {
    options = options || {};
    
    if (!options.inplace) {
      a = a.clone();
      b = b.clone();
    }
    if (a.type === SHIFT$1 && b.type === SHIFT$1) {
      transformShiftShift$1(a, b);
    } else {
      throw new Error('Illegal type')
    }
    return [a, b]
  }

  CoordinateOperation$1.transform = function (...args) {
    return transform$2$1(...args)
  };

  CoordinateOperation$1.fromJSON = function (json) {
    return new CoordinateOperation$1(json)
  };

  CoordinateOperation$1.Shift = function (val) {
    return new CoordinateOperation$1({
      type: SHIFT$1,
      val: val
    })
  };

  const NOP$1$1 = 'NOP';
  const CREATE$1 = 'create';
  const DELETE$2$1 = 'delete';
  const UPDATE$1 = 'update';
  const SET$1 = 'set';

  class ObjectOperation$1 {
    constructor (data) {
      
      if (!data) {
        throw new Error('Data of ObjectOperation is missing.')
      }
      
      if (!data.type) {
        throw new Error('Invalid data: type is mandatory.')
      }
      this.type = data.type;
      if (data.type === NOP$1$1) {
        return
      }
      this.path = data.path;
      if (!data.path) {
        throw new Error('Invalid data: path is mandatory.')
      }
      if (this.type === CREATE$1 || this.type === DELETE$2$1) {
        if (!data.val) {
          throw new Error('Invalid data: value is missing.')
        }
        this.val = data.val;
      } else if (this.type === UPDATE$1) {
        if (data.diff) {
          this.diff = data.diff;
          if (data.diff._isTextOperation) {
            this.propertyType = 'string';
          } else if (data.diff._isArrayOperation) {
            this.propertyType = 'array';
          } else if (data.diff._isCoordinateOperation) {
            this.propertyType = 'coordinate';
          } else {
            throw new Error('Invalid data: diff must be a TextOperation or an ArrayOperation.')
          }
        } else {
          throw new Error('Invalid data: diff is mandatory for update operation.')
        }
      } else if (this.type === SET$1) {
        this.val = data.val;
        this.original = data.original;
      } else {
        throw new Error('Invalid type: ' + data.type)
      }
    }

    apply (obj) {
      if (this.type === NOP$1$1) return obj
      var adapter;
      if (obj._isPathObject) {
        adapter = obj;
      } else {
        adapter = new PathObject$1(obj);
      }
      if (this.type === CREATE$1) {
        adapter.set(this.path, cloneDeep$1(this.val));
        return obj
      }
      if (this.type === DELETE$2$1) {
        adapter.delete(this.path, 'strict');
      } else if (this.type === UPDATE$1) {
        var diff = this.diff;
        switch (this.propertyType) {
          case 'array': {
            let arr = adapter.get(this.path);
            diff.apply(arr);
            break
          }
          case 'string': {
            let str = adapter.get(this.path);
            if (isNil$1(str)) str = '';
            str = diff.apply(str);
            adapter.set(this.path, str);
            break
          }
          case 'coordinate': {
            let coor = adapter.get(this.path);
            if (!coor) throw new Error('No coordinate with path ' + this.path)
            diff.apply(coor);
            break
          }
          default:
            throw new Error('Invalid state.')
        }
      } else if (this.type === SET$1) {
        
        adapter.set(this.path, cloneDeep$1(this.val));
      } else {
        throw new Error('Invalid type.')
      }
      return obj
    }

    clone () {
      var data = {
        type: this.type,
        path: this.path
      };
      if (this.val) {
        data.val = cloneDeep$1(this.val);
      }
      if (this.diff) {
        data.diff = this.diff.clone();
      }
      return new ObjectOperation$1(data)
    }

    isNOP () {
      if (this.type === NOP$1$1) return true
      else if (this.type === UPDATE$1) return this.diff.isNOP()
    }

    isCreate () {
      return this.type === CREATE$1
    }

    isDelete () {
      return this.type === DELETE$2$1
    }

    isUpdate (propertyType) {
      if (propertyType) {
        return (this.type === UPDATE$1 && this.propertyType === propertyType)
      } else {
        return this.type === UPDATE$1
      }
    }

    isSet () {
      return this.type === SET$1
    }

    invert () {
      if (this.type === NOP$1$1) {
        return new ObjectOperation$1({ type: NOP$1$1 })
      }
      var result = new ObjectOperation$1(this);
      if (this.type === CREATE$1) {
        result.type = DELETE$2$1;
      } else if (this.type === DELETE$2$1) {
        result.type = CREATE$1;
      } else if (this.type === UPDATE$1) {
        var invertedDiff;
        if (this.diff._isTextOperation) {
          invertedDiff = TextOperation$1.fromJSON(this.diff.toJSON()).invert();
        } else if (this.diff._isArrayOperation) {
          invertedDiff = ArrayOperation$1.fromJSON(this.diff.toJSON()).invert();
        } else if (this.diff._isCoordinateOperation) {
          invertedDiff = CoordinateOperation$1.fromJSON(this.diff.toJSON()).invert();
        } else {
          throw new Error('Illegal type')
        }
        result.diff = invertedDiff;
      } else  {
        result.val = this.original;
        result.original = this.val;
      }
      return result
    }

    hasConflict (other) {
      return ObjectOperation$1.hasConflict(this, other)
    }

    toJSON () {
      if (this.type === NOP$1$1) {
        return { type: NOP$1$1 }
      }
      var data = {
        type: this.type,
        path: this.path
      };
      if (this.type === CREATE$1 || this.type === DELETE$2$1) {
        data.val = this.val;
      } else if (this.type === UPDATE$1) {
        if (this.diff._isTextOperation) {
          data.propertyType = 'string';
        } else if (this.diff._isArrayOperation) {
          data.propertyType = 'array';
        } else if (this.diff._isCoordinateOperation) {
          data.propertyType = 'coordinate';
        } else {
          throw new Error('Invalid property type.')
        }
        data.diff = this.diff.toJSON();
      } else  {
        data.val = this.val;
        data.original = this.original;
      }
      return data
    }

    getType () {
      return this.type
    }

    getPath () {
      return this.path
    }

    getValue () {
      return this.val
    }

    getOldValue () {
      return this.original
    }

    getValueOp () {
      return this.diff
    }

    
    toString () {
      switch (this.type) {
        case CREATE$1:
          return ['(+,', JSON.stringify(this.path), JSON.stringify(this.val), ')'].join('')
        case DELETE$2$1:
          return ['(-,', JSON.stringify(this.path), JSON.stringify(this.val), ')'].join('')
        case UPDATE$1:
          return ['(>>,', JSON.stringify(this.path), this.propertyType, this.diff.toString(), ')'].join('')
        case SET$1:
          return ['(=,', JSON.stringify(this.path), this.val, this.original, ')'].join('')
        case NOP$1$1:
          return 'NOP'
        default:
          throw new Error('Invalid type')
      }
    }
  }

  ObjectOperation$1.prototype._isOperation = true;
  ObjectOperation$1.prototype._isObjectOperation = true;



  function hasConflict$1$1 (a, b) {
    if (a.type === NOP$1$1 || b.type === NOP$1$1) return false
    return isEqual$1(a.path, b.path)
  }

  function transformDeleteDelete$2$1 (a, b) {
    
    
    a.type = NOP$1$1;
    b.type = NOP$1$1;
  }

  function transformCreateCreate$1 () {
    throw new Error('Can not transform two concurring creates of the same property')
  }

  function transformDeleteCreate$1 () {
    throw new Error('Illegal state: can not create and delete a value at the same time.')
  }

  function transformDeleteUpdate$1 (a, b, flipped) {
    if (a.type !== DELETE$2$1) {
      return transformDeleteUpdate$1(b, a, true)
    }
    var op;
    switch (b.propertyType) {
      case 'string':
        op = TextOperation$1.fromJSON(b.diff);
        break
      case 'array':
        op = ArrayOperation$1.fromJSON(b.diff);
        break
      case 'coordinate':
        op = CoordinateOperation$1.fromJSON(b.diff);
        break
      default:
        throw new Error('Illegal type')
    }
    
    if (!flipped) {
      a.type = NOP$1$1;
      b.type = CREATE$1;
      b.val = op.apply(a.val);
    
    } else {
      a.val = op.apply(a.val);
      b.type = NOP$1$1;
    }
  }

  function transformCreateUpdate$1 () {
    
    throw new Error('Can not transform a concurring create and update of the same property')
  }

  function transformUpdateUpdate$1 (a, b) {
    
    let opA, opB, t;
    switch (b.propertyType) {
      case 'string':
        opA = TextOperation$1.fromJSON(a.diff);
        opB = TextOperation$1.fromJSON(b.diff);
        t = TextOperation$1.transform(opA, opB, {inplace: true});
        break
      case 'array':
        opA = ArrayOperation$1.fromJSON(a.diff);
        opB = ArrayOperation$1.fromJSON(b.diff);
        t = ArrayOperation$1.transform(opA, opB, {inplace: true});
        break
      case 'coordinate':
        opA = CoordinateOperation$1.fromJSON(a.diff);
        opB = CoordinateOperation$1.fromJSON(b.diff);
        t = CoordinateOperation$1.transform(opA, opB, {inplace: true});
        break
      default:
        throw new Error('Illegal type')
    }
    a.diff = t[0];
    b.diff = t[1];
  }

  function transformCreateSet$1 () {
    throw new Error('Illegal state: can not create and set a value at the same time.')
  }

  function transformDeleteSet$1 (a, b, flipped) {
    if (a.type !== DELETE$2$1) return transformDeleteSet$1(b, a, true)
    if (!flipped) {
      a.type = NOP$1$1;
      b.type = CREATE$1;
      b.original = undefined;
    } else {
      a.val = b.val;
      b.type = NOP$1$1;
    }
  }

  function transformUpdateSet$1 () {
    throw new Error('Unresolvable conflict: update + set.')
  }

  function transformSetSet$1 (a, b) {
    a.type = NOP$1$1;
    b.original = a.val;
  }

  const _NOP$1 = 0;
  const _CREATE$1 = 1;
  const _DELETE$1 = 2;
  const _UPDATE$1 = 4;
  const _SET$1 = 8;

  const CODE$1 = (() => {
    const c = {};
    c[NOP$1$1] = _NOP$1;
    c[CREATE$1] = _CREATE$1;
    c[DELETE$2$1] = _DELETE$1;
    c[UPDATE$1] = _UPDATE$1;
    c[SET$1] = _SET$1;
    return c
  })();

  const __transform__$1 = (() => {
    
    const t = {};
    t[_DELETE$1 | _DELETE$1] = transformDeleteDelete$2$1;
    t[_DELETE$1 | _CREATE$1] = transformDeleteCreate$1;
    t[_DELETE$1 | _UPDATE$1] = transformDeleteUpdate$1;
    t[_CREATE$1 | _CREATE$1] = transformCreateCreate$1;
    t[_CREATE$1 | _UPDATE$1] = transformCreateUpdate$1;
    t[_UPDATE$1 | _UPDATE$1] = transformUpdateUpdate$1;
    t[_CREATE$1 | _SET$1] = transformCreateSet$1;
    t[_DELETE$1 | _SET$1] = transformDeleteSet$1;
    t[_UPDATE$1 | _SET$1] = transformUpdateSet$1;
    t[_SET$1    | _SET$1] = transformSetSet$1;
    
    return t
  })();

  function transform$3$1 (a, b, options) {
    options = options || {};
    if (options['no-conflict'] && hasConflict$1$1(a, b)) {
      throw new Conflict$1(a, b)
    }
    if (!options.inplace) {
      a = a.clone();
      b = b.clone();
    }
    if (a.isNOP() || b.isNOP()) {
      return [a, b]
    }
    var sameProp = isEqual$1(a.path, b.path);
    
    if (sameProp) {
      __transform__$1[CODE$1[a.type] | CODE$1[b.type]](a, b);
    }
    return [a, b]
  }

  ObjectOperation$1.transform = transform$3$1;
  ObjectOperation$1.hasConflict = hasConflict$1$1;



  ObjectOperation$1.Create = function (idOrPath, val) {
    var path;
    if (isString$1(idOrPath)) {
      path = [idOrPath];
    } else {
      path = idOrPath;
    }
    return new ObjectOperation$1({type: CREATE$1, path: path, val: val})
  };

  ObjectOperation$1.Delete = function (idOrPath, val) {
    var path;
    if (isString$1(idOrPath)) {
      path = [idOrPath];
    } else {
      path = idOrPath;
    }
    return new ObjectOperation$1({type: DELETE$2$1, path: path, val: val})
  };

  ObjectOperation$1.Update = function (path, op) {
    return new ObjectOperation$1({
      type: UPDATE$1,
      path: path,
      diff: op
    })
  };

  ObjectOperation$1.Set = function (path, oldVal, newVal) {
    return new ObjectOperation$1({
      type: SET$1,
      path: path,
      val: cloneDeep$1(newVal),
      original: cloneDeep$1(oldVal)
    })
  };

  ObjectOperation$1.fromJSON = function (data) {
    data = cloneDeep$1(data);
    if (data.type === 'update') {
      switch (data.propertyType) {
        case 'string':
          data.diff = TextOperation$1.fromJSON(data.diff);
          break
        case 'array':
          data.diff = ArrayOperation$1.fromJSON(data.diff);
          break
        case 'coordinate':
          data.diff = CoordinateOperation$1.fromJSON(data.diff);
          break
        default:
          throw new Error('Unsupported update diff:' + JSON.stringify(data.diff))
      }
    }
    var op = new ObjectOperation$1(data);
    return op
  };

  ObjectOperation$1.NOP = NOP$1$1;
  ObjectOperation$1.CREATE = CREATE$1;
  ObjectOperation$1.DELETE = DELETE$2$1;
  ObjectOperation$1.UPDATE = UPDATE$1;
  ObjectOperation$1.SET = SET$1;

  class OperationSerializer$1 {
    constructor () {
      this.SEPARATOR = '\t';
    }

    serialize (op) {
      var out = [];
      switch (op.type) {
        case 'create':
          out.push('c');
          out.push(op.val.id);
          out.push(op.val);
          break
        case 'delete':
          out.push('d');
          out.push(op.val.id);
          out.push(op.val);
          break
        case 'set':
          out.push('s');
          out.push(op.path.join('.'));
          out.push(op.val);
          out.push(op.original);
          break
        case 'update':
          out.push('u');
          out.push(op.path.join('.'));
          Array.prototype.push.apply(out, this.serializePrimitiveOp(op.diff));
          break
        default:
          throw new Error('Unsupported operation type.')
      }
      return out
    }

    serializePrimitiveOp (op) {
      var out = [];
      if (op._isTextOperation) {
        if (op.isInsert()) {
          out.push('t+');
        } else if (op.isDelete()) {
          out.push('t-');
        }
        out.push(op.pos);
        out.push(op.str);
      } else if (op._isArrayOperation) {
        if (op.isInsert()) {
          out.push('a+');
        } else if (op.isDelete()) {
          out.push('a-');
        }
        out.push(op.pos);
        out.push(op.val);
      } else if (op._isCoordinateOperation) {
        if (op.isShift()) {
          out.push('c>>');
        } else {
          throw new Error('Unsupported CoordinateOperation type.')
        }
        out.push(op.pos);
        out.push(op.val);
      } else {
        throw new Error('Unsupported operation type.')
      }
      return out
    }

    deserialize (str, tokenizer) {
      if (!tokenizer) {
        tokenizer = new Tokenizer$3(str, this.SEPARATOR);
      }
      var type = tokenizer.getString();
      var op, path, val, oldVal, diff;
      switch (type) {
        case 'c':
          path = tokenizer.getPath();
          val = tokenizer.getObject();
          op = ObjectOperation$1.Create(path, val);
          break
        case 'd':
          path = tokenizer.getPath();
          val = tokenizer.getObject();
          op = ObjectOperation$1.Delete(path, val);
          break
        case 's':
          path = tokenizer.getPath();
          val = tokenizer.getAny();
          oldVal = tokenizer.getAny();
          op = ObjectOperation$1.Set(path, oldVal, val);
          break
        case 'u':
          path = tokenizer.getPath();
          diff = this.deserializePrimitiveOp(str, tokenizer);
          op = ObjectOperation$1.Update(path, diff);
          break
        default:
          throw new Error('Illegal type for ObjectOperation: ' + type)
      }
      return op
    }

    deserializePrimitiveOp (str, tokenizer) {
      if (!tokenizer) {
        tokenizer = new Tokenizer$3(str, this.SEPARATOR);
      }
      var type = tokenizer.getString();
      var op, pos, val;
      switch (type) {
        case 't+':
          pos = tokenizer.getNumber();
          val = tokenizer.getString();
          op = TextOperation$1.Insert(pos, val);
          break
        case 't-':
          pos = tokenizer.getNumber();
          val = tokenizer.getString();
          op = TextOperation$1.Delete(pos, val);
          break
        case 'a+':
          pos = tokenizer.getNumber();
          val = tokenizer.getAny();
          op = ArrayOperation$1.Insert(pos, val);
          break
        case 'a-':
          pos = tokenizer.getNumber();
          val = tokenizer.getAny();
          op = ArrayOperation$1.Delete(pos, val);
          break
        case 'c>>':
          val = tokenizer.getNumber();
          op = CoordinateOperation$1.Shift(val);
          break
        default:
          throw new Error('Unsupported operation type: ' + type)
      }
      return op
    }
  }

  class Tokenizer$3 {
    constructor (str, sep) {
      if (isArray$2(arguments[0])) {
        this.tokens = arguments[0];
      } else {
        this.tokens = str.split(sep);
      }
      this.pos = -1;
    }

    error (msg) {
      throw new Error('Parsing error: ' + msg + '\n' + this.tokens[this.pos])
    }

    getString () {
      this.pos++;
      var str = this.tokens[this.pos];
      if (str[0] === '"') {
        str = str.slice(1, -1);
      }
      return str
    }

    getNumber () {
      this.pos++;
      var number;
      var token = this.tokens[this.pos];
      try {
        if (isNumber$1(token)) {
          number = token;
        } else {
          number = parseInt(this.tokens[this.pos], 10);
        }
        return number
      } catch (err) {
        this.error('expected number');
      }
    }

    getObject () {
      this.pos++;
      var obj;
      var token = this.tokens[this.pos];
      try {
        if (isObject$2(token)) {
          obj = token;
        } else {
          obj = JSON.parse(this.tokens[this.pos]);
        }
        return obj
      } catch (err) {
        this.error('expected object');
      }
    }

    getAny () {
      this.pos++;
      var token = this.tokens[this.pos];
      return token
    }

    getPath () {
      var str = this.getString();
      return str.split('.')
    }
  }

  OperationSerializer$1.Tokenizer = Tokenizer$3;

  function last$1$1 (arr) {
    return arr[arr.length - 1]
  }

  class Selection$1 {
    constructor () {
      
      var _internal = {};
      Object.defineProperty(this, '_internal', {
        enumerable: false,
        value: _internal
      });
      
      _internal.doc = null;
    }

    clone () {
      var newSel = this._clone();
      if (this._internal.doc) {
        newSel.attach(this._internal.doc);
      }
      return newSel
    }

    
    getDocument () {
      var doc = this._internal.doc;
      if (!doc) {
        throw new Error('Selection is not attached to a document.')
      }
      return doc
    }

    isAttached () {
      return Boolean(this._internal.doc)
    }

    
    attach (doc) {
      this._internal.doc = doc;
      return this
    }

    
    isNull () { return false }

    
    isPropertySelection () { return false }

    
    isContainerSelection () { return false }

    
    isNodeSelection () { return false }

    isCustomSelection () { return false }

    
    isCollapsed () { return true }

    
    isReverse () { return false }

    getType () {
      throw new Error('Selection.getType() is abstract.')
    }

    get type () {
      return this.getType()
    }

    
    equals (other) {
      if (this === other) {
        return true
      } else if (!other) {
        return false
      } else if (this.isNull() !== other.isNull()) {
        return false
      } else if (this.getType() !== other.getType()) {
        return false
      } else {
        
        
        return true
      }
    }

    
    toString () {
      return 'null'
    }

    
    toJSON () {
      throw new Error('This method is abstract.')
    }

    createWith (update) {
      let SelectionClass = this.constructor;
      let data = this.toJSON();
      Object.assign(data, update);
      return SelectionClass.fromJSON(data)
    }
  }


  Selection$1.prototype._isSelection = true;


  class NullSelection$1 extends Selection$1 {
    isNull () {
      return true
    }

    getType () {
      return 'null'
    }

    toJSON () {
      return null
    }

    clone () {
      return this
    }
  }



  Selection$1.nullSelection = Object.freeze(new NullSelection$1());

  function isArrayEqual$1 (arr1, arr2) {
    if (arr1 === arr2) return true
    if (!isArray$2(arr1) || !isArray$2(arr2)) return false
    if (arr1.length !== arr2.length) return false
    let L = arr1.length;
    for (var i = 0; i < L; i++) {
      if (arr1[i] !== arr2[i]) return false
    }
    return true
  }

  class Coordinate$1 {
    
    constructor (path, offset) {
      
      if (arguments[0] === 'SKIP') return
      if (arguments.length === 1) {
        let data = arguments[0];
        this.path = data.path;
        this.offset = data.offset;
      } else {
        this.path = path;
        this.offset = offset;
      }
      if (!isArray$2(this.path)) {
        throw new Error('Invalid arguments: path should be an array.')
      }
      if (!isNumber$1(this.offset) || this.offset < 0) {
        throw new Error('Invalid arguments: offset must be a positive number.')
      }
    }

    equals (other) {
      return (other === this ||
        (isArrayEqual$1(other.path, this.path) && other.offset === this.offset))
    }

    withCharPos (offset) {
      return new Coordinate$1(this.path, offset)
    }

    getNodeId () {
      return this.path[0]
    }

    getPath () {
      return this.path
    }

    getOffset () {
      return this.offset
    }

    toJSON () {
      return {
        path: this.path.slice(),
        offset: this.offset
      }
    }

    toString () {
      return '(' + this.path.join('.') + ', ' + this.offset + ')'
    }

    isPropertyCoordinate () {
      return this.path.length > 1
    }

    isNodeCoordinate () {
      return this.path.length === 1
    }

    hasSamePath (other) {
      return isArrayEqual$1(this.path, other.path)
    }
  }

  Coordinate$1.prototype._isCoordinate = true;

  class PropertySelection$1 extends Selection$1 {
    
    constructor (path, startOffset, endOffset, reverse, containerId, surfaceId) {
      super();

      if (arguments.length === 1) {
        let data = arguments[0];
        path = data.path;
        startOffset = data.startOffset;
        endOffset = data.endOffset;
        reverse = data.reverse;
        containerId = data.containerId;
        surfaceId = data.surfaceId;
      }

      if (!path || !isNumber$1(startOffset)) {
        throw new Error('Invalid arguments: `path` and `startOffset` are mandatory')
      }

      this.start = new Coordinate$1(path, startOffset);
      this.end = new Coordinate$1(path, isNumber$1(endOffset) ? endOffset : startOffset);

      
      this.reverse = Boolean(reverse);

      this.containerId = containerId;

      
      this.surfaceId = surfaceId;
    }

    get path () {
      return this.start.path
    }

    get startOffset () {
      console.warn('DEPRECATED: Use sel.start.offset instead');
      return this.start.offset
    }

    get endOffset () {
      console.warn('DEPRECATED: Use sel.end.offset instead');
      return this.end.offset
    }

    
    toJSON () {
      return {
        type: 'property',
        path: this.start.path,
        startOffset: this.start.offset,
        endOffset: this.end.offset,
        reverse: this.reverse,
        containerId: this.containerId,
        surfaceId: this.surfaceId
      }
    }

    isPropertySelection () {
      return true
    }

    getType () {
      return 'property'
    }

    isNull () {
      return false
    }

    isCollapsed () {
      return this.start.offset === this.end.offset
    }

    isReverse () {
      return this.reverse
    }

    equals (other) {
      return (
        Selection$1.prototype.equals.call(this, other) &&
        (this.start.equals(other.start) && this.end.equals(other.end))
      )
    }

    toString () {
      
      return [
        'PropertySelection(', JSON.stringify(this.path), ', ',
        this.start.offset, ' -> ', this.end.offset,
        (this.reverse ? ', reverse' : ''),
        (this.surfaceId ? (', ' + this.surfaceId) : ''),
        ')'
      ].join('')
    }

    
    collapse (direction) {
      var offset;
      if (direction === 'left') {
        offset = this.start.offset;
      } else {
        offset = this.end.offset;
      }
      return this.createWithNewRange(offset, offset)
    }

    
    

    
    getPath () {
      return this.start.path
    }

    getNodeId () {
      return this.start.path[0]
    }

    
    isInsideOf (other, strict) {
      if (other.isNull()) return false
      if (other.isContainerSelection()) {
        return other.contains(this, strict)
      }
      if (strict) {
        return (isArrayEqual$1(this.path, other.path) &&
          this.start.offset > other.start.offset &&
          this.end.offset < other.end.offset)
      } else {
        return (isArrayEqual$1(this.path, other.path) &&
          this.start.offset >= other.start.offset &&
          this.end.offset <= other.end.offset)
      }
    }

    
    contains (other, strict) {
      if (other.isNull()) return false
      return other.isInsideOf(this, strict)
    }

    
    overlaps (other, strict) {
      if (other.isNull()) return false
      if (other.isContainerSelection()) {
        
        return other.overlaps(this)
      }
      if (!isArrayEqual$1(this.path, other.path)) return false
      if (strict) {
        return (!(this.start.offset >= other.end.offset || this.end.offset <= other.start.offset))
      } else {
        return (!(this.start.offset > other.end.offset || this.end.offset < other.start.offset))
      }
    }

    
    isRightAlignedWith (other) {
      if (other.isNull()) return false
      if (other.isContainerSelection()) {
        
        return other.isRightAlignedWith(this)
      }
      return (isArrayEqual$1(this.path, other.path) &&
        this.end.offset === other.end.offset)
    }

    
    isLeftAlignedWith (other) {
      if (other.isNull()) return false
      if (other.isContainerSelection()) {
        
        return other.isLeftAlignedWith(this)
      }
      return (isArrayEqual$1(this.path, other.path) &&
        this.start.offset === other.start.offset)
    }

    
    expand (other) {
      if (other.isNull()) return this

      
      
      
      if (other.isContainerSelection()) {
        return other.expand(this)
      }
      if (!isArrayEqual$1(this.path, other.path)) {
        throw new Error('Can not expand PropertySelection to a different property.')
      }
      var newStartOffset = Math.min(this.start.offset, other.start.offset);
      var newEndOffset = Math.max(this.end.offset, other.end.offset);
      return this.createWithNewRange(newStartOffset, newEndOffset)
    }

    
    truncateWith (other) {
      if (other.isNull()) return this
      if (other.isInsideOf(this, 'strict')) {
        
        throw new Error('Can not truncate with a contained selections')
      }
      if (!this.overlaps(other)) {
        return this
      }
      var otherStartOffset, otherEndOffset;
      if (other.isPropertySelection()) {
        otherStartOffset = other.start.offset;
        otherEndOffset = other.end.offset;
      } else if (other.isContainerSelection()) {
        
        if (isArrayEqual$1(other.start.path, this.start.path)) {
          otherStartOffset = other.start.offset;
        } else {
          otherStartOffset = this.start.offset;
        }
        if (isArrayEqual$1(other.end.path, this.start.path)) {
          otherEndOffset = other.end.offset;
        } else {
          otherEndOffset = this.end.offset;
        }
      } else {
        return this
      }

      var newStartOffset;
      var newEndOffset;
      if (this.start.offset > otherStartOffset && this.end.offset > otherEndOffset) {
        newStartOffset = otherEndOffset;
        newEndOffset = this.end.offset;
      } else if (this.start.offset < otherStartOffset && this.end.offset < otherEndOffset) {
        newStartOffset = this.start.offset;
        newEndOffset = otherStartOffset;
      } else if (this.start.offset === otherStartOffset) {
        if (this.end.offset <= otherEndOffset) {
          return Selection$1.nullSelection
        } else {
          newStartOffset = otherEndOffset;
          newEndOffset = this.end.offset;
        }
      } else if (this.end.offset === otherEndOffset) {
        if (this.start.offset >= otherStartOffset) {
          return Selection$1.nullSelection
        } else {
          newStartOffset = this.start.offset;
          newEndOffset = otherStartOffset;
        }
      } else if (other.contains(this)) {
        return Selection$1.nullSelection
      } else {
        
        throw new Error('Illegal state.')
      }
      return this.createWithNewRange(newStartOffset, newEndOffset)
    }

    
    createWithNewRange (startOffset, endOffset) {
      var sel = new PropertySelection$1(this.path, startOffset, endOffset, false, this.containerId, this.surfaceId);
      var doc = this._internal.doc;
      if (doc) {
        sel.attach(doc);
      }
      return sel
    }

    _clone () {
      return new PropertySelection$1(this.start.path, this.start.offset, this.end.offset, this.reverse, this.containerId, this.surfaceId)
    }
  }

  PropertySelection$1.fromJSON = function (json) {
    return new PropertySelection$1(json)
  };

  class ContainerSelection$1 extends Selection$1 {
    constructor (containerId, startPath, startOffset, endPath, endOffset, reverse, surfaceId) {
      super();

      if (arguments.length === 1) {
        let data = arguments[0];
        containerId = data.containerId;
        startPath = data.startPath;
        startOffset = data.startOffset;
        endPath = data.endPath;
        endOffset = data.endOffset;
        reverse = data.reverse;
        surfaceId = data.surfaceId;
      }

      
      this.containerId = containerId;
      if (!this.containerId) throw new Error('Invalid arguments: `containerId` is mandatory')

      this.start = new Coordinate$1(startPath, startOffset);
      this.end = new Coordinate$1(isNil$1(endPath) ? startPath : endPath, isNil$1(endOffset) ? startOffset : endOffset);

      this.reverse = Boolean(reverse);

      this.surfaceId = surfaceId;
    }

    

    get startPath () {
      console.warn('DEPRECATED: use sel.start.path instead.');
      return this.start.path
    }

    get startOffset () {
      console.warn('DEPRECATED: use sel.start.offset instead.');
      return this.start.offset
    }

    get endPath () {
      console.warn('DEPRECATED: use sel.end.path instead.');
      return this.end.path
    }

    get endOffset () {
      console.warn('DEPRECATED: use sel.end.offset instead.');
      return this.end.offset
    }

    

    toJSON () {
      return {
        type: 'container',
        containerId: this.containerId,
        startPath: this.start.path,
        startOffset: this.start.offset,
        endPath: this.end.path,
        endOffset: this.end.offset,
        reverse: this.reverse,
        surfaceId: this.surfaceId
      }
    }

    isContainerSelection () {
      return true
    }

    getType () {
      return 'container'
    }

    isNull () {
      return false
    }

    isCollapsed () {
      return this.start.equals(this.end)
    }

    isReverse () {
      return this.reverse
    }

    equals (other) {
      return (
        Selection$1.prototype.equals.call(this, other) &&
        this.containerId === other.containerId &&
        (this.start.equals(other.start) && this.end.equals(other.end))
      )
    }

    toString () {
      
      return [
        'ContainerSelection(',
        this.containerId, ', ',
        JSON.stringify(this.start.path), ', ', this.start.offset,
        ' -> ',
        JSON.stringify(this.end.path), ', ', this.end.offset,
        (this.reverse ? ', reverse' : ''),
        (this.surfaceId ? (', ' + this.surfaceId) : ''),
        ')'
      ].join('')
    }

    
    getContainer () {
      if (!this._internal.container) {
        this._internal.container = this.getDocument().get(this.containerId);
      }
      return this._internal.container
    }

    isInsideOf (other, strict) {
      
      
      if (other.isNull()) return false
      strict = Boolean(strict);
      let r1 = this._range(this);
      let r2 = this._range(other);
      return (r2.start.isBefore(r1.start, strict) &&
        r1.end.isBefore(r2.end, strict))
    }

    contains (other, strict) {
      
      
      if (other.isNull()) return false
      strict = Boolean(strict);
      let r1 = this._range(this);
      let r2 = this._range(other);
      return (r1.start.isBefore(r2.start, strict) &&
        r2.end.isBefore(r1.end, strict))
    }

    containsNode (nodeId, strict) {
      const container = this.getContainer();
      if (!container.contains(nodeId)) return false
      const coor = new Coordinate$1([nodeId], 0);
      const address = container.getAddress(coor);
      const r = this._range(this);
      
      let contained = r.start.isBefore(address, strict);
      if (contained) {
        address.offset = 1;
        contained = r.end.isAfter(address, strict);
      }
      return contained
    }

    overlaps (other) {
      let r1 = this._range(this);
      let r2 = this._range(other);
      
      return !(r1.end.isBefore(r2.start, false) ||
        r2.end.isBefore(r1.start, false))
    }

    isLeftAlignedWith (other) {
      let r1 = this._range(this);
      let r2 = this._range(other);
      return r1.start.isEqual(r2.start)
    }

    isRightAlignedWith (other) {
      let r1 = this._range(this);
      let r2 = this._range(other);
      return r1.end.isEqual(r2.end)
    }

    
    collapse (direction) {
      let coor;
      if (direction === 'left') {
        coor = this.start;
      } else {
        coor = this.end;
      }
      return _createNewSelection$1(this, coor, coor)
    }

    expand (other) {
      let r1 = this._range(this);
      let r2 = this._range(other);
      let start;
      let end;

      if (r1.start.isEqual(r2.start)) {
        start = new Coordinate$1(this.start.path, Math.min(this.start.offset, other.start.offset));
      } else if (r1.start.isAfter(r2.start)) {
        start = new Coordinate$1(other.start.path, other.start.offset);
      } else {
        start = this.start;
      }
      if (r1.end.isEqual(r2.end)) {
        end = new Coordinate$1(this.end.path, Math.max(this.end.offset, other.end.offset));
      } else if (r1.end.isBefore(r2.end, false)) {
        end = new Coordinate$1(other.end.path, other.end.offset);
      } else {
        end = this.end;
      }

      return _createNewSelection$1(this, start, end)
    }

    truncateWith (other) {
      if (other.isInsideOf(this, 'strict')) {
        
        throw new Error('Can not truncate with a contained selections')
      }
      if (!this.overlaps(other)) {
        return this
      }
      let r1 = this._range(this);
      let r2 = this._range(other);
      let start, end;
      if (r2.start.isBefore(r1.start, 'strict') && r2.end.isBefore(r1.end, 'strict')) {
        start = other.end;
        end = this.end;
      } else if (r1.start.isBefore(r2.start, 'strict') && r1.end.isBefore(r2.end, 'strict')) {
        start = this.start;
        end = other.start;
      } else if (r1.start.isEqual(r2.start)) {
        if (r2.end.isBefore(r1.end, 'strict')) {
          start = other.end;
          end = this.end;
        } else {
          
          return Selection$1.nullSelection
        }
      } else if (r1.end.isEqual(r2.end)) {
        if (r1.start.isBefore(r2.start, 'strict')) {
          start = this.start;
          end = other.start;
        } else {
          
          return Selection$1.nullSelection
        }
      } else if (this.isInsideOf(other)) {
        return Selection$1.nullSelection
      } else {
        throw new Error('Could not determine coordinates for truncate. Check input')
      }
      return _createNewSelection$1(this, start, end)
    }

    
    getNodeIds () {
      const container = this.getContainer();
      const startPos = container.getPosition(this.start.path[0]);
      const endPos = container.getPosition(this.end.path[0]);
      return container.getContent().slice(startPos, endPos + 1)
    }

    
    splitIntoPropertySelections () {
      let sels = [];
      let fragments = this.getFragments();
      fragments.forEach(function (fragment) {
        if (fragment instanceof Selection$1.Fragment) {
          sels.push(
            new PropertySelection$1(fragment.path, fragment.startOffset,
              fragment.endOffset, false, this.containerId, this.surfaceId)
          );
        }
      }.bind(this));
      return sels
    }

    _clone () {
      return new ContainerSelection$1(this)
    }

    _range (sel) {
      
      
      
      if (sel._internal.addressRange) {
        return sel._internal.addressRange
      }

      let container = this.getContainer();
      let startAddress = container.getAddress(sel.start);
      let endAddress;
      if (sel.isCollapsed()) {
        endAddress = startAddress;
      } else {
        endAddress = container.getAddress(sel.end);
      }
      let addressRange = {
        start: startAddress,
        end: endAddress
      };
      if (sel._isContainerSelection) {
        sel._internal.addressRange = addressRange;
      }
      return addressRange
    }

    get path () {
      throw new Error('ContainerSelection has no path property. Use startPath and endPath instead')
    }
  }

  ContainerSelection$1.prototype._isContainerSelection = true;

  ContainerSelection$1.fromJSON = function (properties) {
    let sel = new ContainerSelection$1(properties);
    return sel
  };

  function _createNewSelection$1 (containerSel, start, end) {
    let newSel;

    if (start === end) {
      newSel = new PropertySelection$1({
        path: start.path,
        startOffset: start.offset,
        endOffset: start.offset,
        containerId: containerSel.containerId,
        surfaceId: containerSel.surfaceId
      });
    } else {
      newSel = new ContainerSelection$1(
        containerSel.containerId,
        start.path, start.offset, end.path, end.offset,
        false, containerSel.surfaceId
      );
    }
    
    const doc = containerSel._internal.doc;
    if (doc) {
      newSel.attach(doc);
    }
    return newSel
  }

  class NodeSelection$1 extends Selection$1 {
    constructor (containerId, nodeId, mode, reverse, surfaceId) {
      super();

      if (arguments.length === 1) {
        let data = arguments[0];
        containerId = data.containerId;
        nodeId = data.nodeId;
        mode = data.mode;
        reverse = data.reverse;
        surfaceId = data.surfaceId;
      }

      if (!isString$1(containerId)) {
        throw new Error("'containerId' is mandatory.")
      }
      if (!isString$1(nodeId)) {
        throw new Error("'nodeId' is mandatory.")
      }
      mode = mode || 'full';

      this.containerId = containerId;
      this.nodeId = nodeId;
      this.mode = mode;
      this.reverse = Boolean(reverse);
      this.surfaceId = surfaceId;

      this.start = new Coordinate$1([nodeId], 0);
      this.end = new Coordinate$1([nodeId], 1);
    }

    equals (other) {
      return (
        super.equals(other) &&
        this.nodeId === other.nodeId &&
        this.mode === other.mode
      )
    }

    isNodeSelection () {
      return true
    }

    getType () {
      return 'node'
    }

    getNodeId () {
      return this.nodeId
    }

    isFull () {
      return this.mode === 'full'
    }

    isBefore () {
      return this.mode === 'before'
    }

    isAfter () {
      return this.mode === 'after'
    }

    isCollapsed () {
      return this.mode !== 'full'
    }

    toJSON () {
      return {
        type: 'node',
        nodeId: this.nodeId,
        mode: this.mode,
        reverse: this.reverse,
        containerId: this.containerId,
        surfaceId: this.surfaceId
      }
    }

    toString () {
      
      return [
        'NodeSelection(',
        this.containerId, '.', this.nodeId, ', ',
        this.mode, ', ',
        (this.reverse ? ', reverse' : ''),
        (this.surfaceId ? (', ' + this.surfaceId) : ''),
        ')'
      ].join('')
    }

    collapse (direction) {
      if (direction === 'left') {
        if (this.isBefore()) {
          return this
        } else {
          return new NodeSelection$1(this.containerId, this.nodeId, 'before', this.reverse, this.surfaceId)
        }
      } else if (direction === 'right') {
        if (this.isAfter()) {
          return this
        } else {
          return new NodeSelection$1(this.containerId, this.nodeId, 'after', this.reverse, this.surfaceId)
        }
      } else {
        throw new Error("'direction' must be either 'left' or 'right'")
      }
    }

    _getCoordinate () {
      if (this.mode === 'before') {
        return new Coordinate$1([this.nodeId], 0)
      } else if (this.mode === 'after') {
        return new Coordinate$1([this.nodeId], 1)
      }
    }

    _clone () {
      return new NodeSelection$1(this)
    }
  }

  NodeSelection$1.prototype._isNodeSelection = true;

  NodeSelection$1.fromJSON = function (json) {
    return new NodeSelection$1(json)
  };


  NodeSelection$1._createFromCoordinate = function (coor) {
    var containerId = coor.containerId;
    var nodeId = coor.getNodeId();
    var mode = coor.offset === 0 ? 'before' : 'after';
    return new NodeSelection$1(containerId, nodeId, mode, false)
  };

  class CustomSelection$1 extends Selection$1 {
    constructor (customType, data, surfaceId) {
      super();

      if (arguments.length === 1) {
        let _data = arguments[0];
        customType = _data.customType;
        data = _data.data;
        surfaceId = _data.surfaceId;
      }

      this.customType = customType;
      this.data = data || {};
      this.surfaceId = surfaceId;
    }

    isCustomSelection () {
      return true
    }

    getType () {
      return 'custom'
    }

    getCustomType () {
      return this.customType
    }

    toJSON () {
      return {
        type: 'custom',
        customType: this.customType,
        data: cloneDeep$1(this.data),
        surfaceId: this.surfaceId
      }
    }

    toString () {
      
      return [
        'CustomSelection(',
        this.customType, ', ',
        JSON.stringify(this.data),
        ')'
      ].join('')
    }

    equals (other) {
      return (
        Selection$1.prototype.equals.call(this, other) &&
        other.isCustomSelection() &&
        isEqual$1(this.data, other.data)
      )
    }

    _clone () {
      return new CustomSelection$1(this)
    }
  }

  CustomSelection$1.prototype._isCustomSelection = true;

  CustomSelection$1.fromJSON = function (json) {
    return new CustomSelection$1(json)
  };

  function fromJSON$1 (json) {
    if (!json) return Selection$1.nullSelection
    var type = json.type;
    switch (type) {
      case 'property':
        return PropertySelection$1.fromJSON(json)
      case 'container':
        return ContainerSelection$1.fromJSON(json)
      case 'node':
        return NodeSelection$1.fromJSON(json)
      case 'custom':
        return CustomSelection$1.fromJSON(json)
      default:
        
        return Selection$1.nullSelection
    }
  }


  function isFirst$1 (doc, coor) {
    if (coor.isNodeCoordinate() && coor.offset === 0) return true
    let node = doc.get(coor.path[0]).getContainerRoot();
    if (node.isText() && coor.offset === 0) return true
    if (node.isList()) {
      let itemId = coor.path[0];
      if (node.items[0] === itemId && coor.offset === 0) return true
    }
  }


  function isLast$1 (doc, coor) {
    if (coor.isNodeCoordinate() && coor.offset > 0) return true
    let node = doc.get(coor.path[0]).getContainerRoot();
    if (node.isText() && coor.offset >= node.getLength()) return true
    if (node.isList()) {
      let itemId = coor.path[0];
      let item = doc.get(itemId);
      if (last$1$1(node.items) === itemId && coor.offset === item.getLength()) return true
    }
  }

  function isEntirelySelected$1 (doc, node, start, end) {
    let { isEntirelySelected } = getRangeInfo$1(doc, node, start, end);
    return isEntirelySelected
  }

  function getRangeInfo$1 (doc, node, start, end) {
    let isFirst = true;
    let isLast = true;
    if (node.isText()) {
      if (start && start.offset !== 0) isFirst = false;
      if (end && end.offset < node.getLength()) isLast = false;
    } else if (node.isList()) {
      if (start) {
        let itemId = start.path[0];
        let itemPos = node.getItemPosition(itemId);
        if (itemPos > 0 || start.offset !== 0) isFirst = false;
      }
      if (end) {
        let itemId = end.path[0];
        let itemPos = node.getItemPosition(itemId);
        let item = doc.get(itemId);
        if (itemPos < node.items.length - 1 || end.offset < item.getLength()) isLast = false;
      }
    }
    let isEntirelySelected = isFirst && isLast;
    return {isFirst, isLast, isEntirelySelected}
  }

  function setCursor$1 (tx, node, containerId, mode) {
    if (node.isText()) {
      let offset = 0;
      if (mode === 'after') {
        let text = node.getText();
        offset = text.length;
      }
      tx.setSelection({
        type: 'property',
        path: node.getPath(),
        startOffset: offset,
        containerId: containerId
      });
    } else if (node.isList()) {
      let item, offset;
      if (mode === 'after') {
        item = node.getLastItem();
        offset = item.getLength();
      } else {
        item = node.getFirstItem();
        offset = 0;
      }
      tx.setSelection({
        type: 'property',
        path: item.getPath(),
        startOffset: offset,
        containerId: containerId
      });
    } else {
      tx.setSelection({
        type: 'node',
        containerId: containerId,
        nodeId: node.id
        
        
        
      });
    }
  }

  function selectNode$1 (tx, nodeId, containerId) {
    tx.setSelection(createNodeSelection$1({ doc: tx, nodeId, containerId }));
  }

  function createNodeSelection$1 ({ doc, nodeId, containerId, mode, reverse, surfaceId }) {
    let node = doc.get(nodeId);
    if (!node) return Selection$1.nullSelection
    node = node.getContainerRoot();
    if (node.isText()) {
      return new PropertySelection$1({
        path: node.getPath(),
        startOffset: mode === 'after' ? node.getLength() : 0,
        endOffset: mode === 'before' ? 0 : node.getLength(),
        reverse: reverse,
        containerId: containerId,
        surfaceId: surfaceId
      })
    } else if (node.isList() && node.getLength() > 0) {
      let first = node.getFirstItem();
      let last = node.getLastItem();
      let start = {
        path: first.getPath(),
        offset: 0
      };
      let end = {
        path: last.getPath(),
        offset: last.getLength()
      };
      if (mode === 'after') start = end;
      else if (mode === 'before') end = start;
      return new ContainerSelection$1({
        startPath: start.path,
        startOffset: start.offset,
        endPath: end.path,
        endOffset: end.offset,
        reverse: reverse,
        containerId: containerId,
        surfaceId: surfaceId
      })
    } else {
      return new NodeSelection$1({ nodeId, mode, reverse, containerId, surfaceId })
    }
  }

  function stepIntoIsolatedNode$1 (editorSession, comp) {
    
    
    if (comp.grabFocus()) return true

    
    let surface = comp.find('.sc-surface');
    if (surface) {
      
      if (surface._isTextPropertyEditor) {
        const doc = editorSession.getDocument();
        const path = surface.getPath();
        const text = doc.get(path, 'strict');
        editorSession.setSelection({
          type: 'property',
          path: path,
          startOffset: text.length,
          surfaceId: surface.id
        });
        return true
      } else if (surface._isContainerEditor) {
        let container = surface.getContainer();
        if (container.length > 0) {
          let first = container.getChildAt(0);
          setCursor$1(editorSession, first, container.id, 'after');
        }
        return true
      }
    }
    return false
  }

  function augmentSelection$1 (selData, oldSel) {
    
    if (selData && oldSel && !selData.surfaceId && !oldSel.isNull()) {
      selData.containerId = selData.containerId || oldSel.containerId;
      selData.surfaceId = selData.surfaceId || oldSel.surfaceId;
    }
    return selData
  }

  var selectionHelpers$1 = /*#__PURE__*/Object.freeze({
    fromJSON: fromJSON$1,
    isFirst: isFirst$1,
    isLast: isLast$1,
    isEntirelySelected: isEntirelySelected$1,
    getRangeInfo: getRangeInfo$1,
    setCursor: setCursor$1,
    selectNode: selectNode$1,
    createNodeSelection: createNodeSelection$1,
    stepIntoIsolatedNode: stepIntoIsolatedNode$1,
    augmentSelection: augmentSelection$1
  });

  class DocumentChange$1 {
    constructor (ops, before, after) {
      if (arguments.length === 1 && isPlainObject$2(arguments[0])) {
        let data = arguments[0];
        
        this.sha = data.sha;
        
        this.timestamp = data.timestamp;
        
        this.before = data.before || {};
        
        this.ops = data.ops;
        this.info = data.info; 
        
        this.after = data.after || {};
      } else if (arguments.length === 3) {
        this.sha = uuid$1();
        this.info = {};
        this.timestamp = Date.now();
        this.ops = ops.slice(0);
        this.before = before || {};
        this.after = after || {};
      } else {
        throw new Error('Illegal arguments.')
      }
      
      this.updated = null;
      
      this.created = null;
      
      this.deleted = null;
    }

    
    _extractInformation (doc) {
      
      
      if (this._extracted) return

      let ops = this.ops;
      let created = {};
      let deleted = {};
      let updated = {};
      let affectedContainerAnnos = [];

      
      function _checkAnnotation (op) {
        switch (op.type) {
          case 'create':
          case 'delete': {
            let node = op.val;
            if (node.hasOwnProperty('start')) {
              updated[node.start.path] = true;
            }
            if (node.hasOwnProperty('end')) {
              updated[node.end.path] = true;
            }
            break
          }
          case 'update':
          case 'set': {
            
            let node = doc.get(op.path[0]);
            if (node) {
              if (node.isPropertyAnnotation()) {
                updated[node.start.path] = true;
              } else if (node.isContainerAnnotation()) {
                affectedContainerAnnos.push(node);
              }
            }
            break
          }
          default:
            
            throw new Error('Illegal state')
        }
      }

      for (let i = 0; i < ops.length; i++) {
        let op = ops[i];
        if (op.type === 'create') {
          created[op.val.id] = op.val;
          delete deleted[op.val.id];
        }
        if (op.type === 'delete') {
          delete created[op.val.id];
          deleted[op.val.id] = op.val;
        }
        if (op.type === 'set' || op.type === 'update') {
          updated[op.path] = true;
          
          updated[op.path[0]] = true;
        }
        _checkAnnotation(op);
      }

      affectedContainerAnnos.forEach(function (anno) {
        let container = doc.get(anno.containerId, 'strict');
        let startPos = container.getPosition(anno.start.path[0]);
        let endPos = container.getPosition(anno.end.path[0]);
        for (let pos = startPos; pos <= endPos; pos++) {
          let node = container.getChildAt(pos);
          let path;
          if (node.isText()) {
            path = node.getPath();
          } else {
            path = [node.id];
          }
          if (!deleted[node.id]) {
            updated[path] = true;
          }
        }
      });

      
      if (Object.keys(deleted).length > 0) {
        forEach$1(updated, function (_, key) {
          let nodeId = key.split(',')[0];
          if (deleted[nodeId]) {
            delete updated[key];
          }
        });
      }

      this.created = created;
      this.deleted = deleted;
      this.updated = updated;

      this._extracted = true;
    }

    invert () {
      
      let copy = this.toJSON();
      copy.ops = [];
      
      let tmp = copy.before;
      copy.before = copy.after;
      copy.after = tmp;
      let inverted = DocumentChange$1.fromJSON(copy);
      let ops = [];
      for (let i = this.ops.length - 1; i >= 0; i--) {
        ops.push(this.ops[i].invert());
      }
      inverted.ops = ops;
      return inverted
    }

    
    isAffected (path) {
      console.error('DEPRECATED: use change.hasUpdated() instead');
      return this.hasUpdated(path)
    }

    isUpdated (path) {
      console.error('DEPRECATED: use change.hasUpdated() instead');
      return this.hasUpdated(path)
    }
    

    hasUpdated (path) {
      return this.updated[path]
    }

    hasDeleted (id) {
      return this.deleted[id]
    }

    serialize () {
      
      

      let opSerializer = new OperationSerializer$1();
      let data = this.toJSON();
      data.ops = this.ops.map(function (op) {
        return opSerializer.serialize(op)
      });
      return JSON.stringify(data)
    }

    clone () {
      return DocumentChange$1.fromJSON(this.toJSON())
    }

    toJSON () {
      let data = {
        
        sha: this.sha,
        
        before: clone$1(this.before),
        ops: map$2(this.ops, function (op) {
          return op.toJSON()
        }),
        info: this.info,
        
        after: clone$1(this.after)
      };

      
      
      data.after.selection = undefined;
      data.before.selection = undefined;

      let sel = this.before.selection;
      if (sel && sel._isSelection) {
        data.before.selection = sel.toJSON();
      }
      sel = this.after.selection;
      if (sel && sel._isSelection) {
        data.after.selection = sel.toJSON();
      }
      return data
    }
  }

  DocumentChange$1.deserialize = function (str) {
    let opSerializer = new OperationSerializer$1();
    let data = JSON.parse(str);
    data.ops = data.ops.map(function (opData) {
      return opSerializer.deserialize(opData)
    });
    if (data.before.selection) {
      data.before.selection = fromJSON$1(data.before.selection);
    }
    if (data.after.selection) {
      data.after.selection = fromJSON$1(data.after.selection);
    }
    return new DocumentChange$1(data)
  };

  DocumentChange$1.fromJSON = function (data) {
    
    let change = cloneDeep$1(data);
    change.ops = data.ops.map(function (opData) {
      return ObjectOperation$1.fromJSON(opData)
    });
    change.before.selection = fromJSON$1(data.before.selection);
    change.after.selection = fromJSON$1(data.after.selection);
    return new DocumentChange$1(change)
  };

  function transformDocumentChange$1 (A, B) {
    _transformInplaceBatch$1(A, B);
  }

  function transformSelection$1 (sel, a) {
    let newSel = sel.clone();
    let hasChanged = _transformSelectionInplace$1(newSel, a);
    if (hasChanged) {
      return newSel
    } else {
      return sel
    }
  }

  function _transformInplaceSingle$1 (a, b) {
    for (let i = 0; i < a.ops.length; i++) {
      let opA = a.ops[i];
      for (let j = 0; j < b.ops.length; j++) {
        let opB = b.ops[j];
        
        
        ObjectOperation$1.transform(opA, opB, {inplace: true});
      }
    }
    if (a.before) {
      _transformSelectionInplace$1(a.before.selection, b);
    }
    if (a.after) {
      _transformSelectionInplace$1(a.after.selection, b);
    }
    if (b.before) {
      _transformSelectionInplace$1(b.before.selection, a);
    }
    if (b.after) {
      _transformSelectionInplace$1(b.after.selection, a);
    }
  }

  function _transformInplaceBatch$1 (A, B) {
    if (!isArray$2(A)) {
      A = [A];
    }
    if (!isArray$2(B)) {
      B = [B];
    }
    for (let i = 0; i < A.length; i++) {
      let a = A[i];
      for (let j = 0; j < B.length; j++) {
        let b = B[j];
        _transformInplaceSingle$1(a, b);
      }
    }
  }

  function _transformSelectionInplace$1 (sel, a) {
    if (!sel || (!sel.isPropertySelection() && !sel.isContainerSelection())) {
      return false
    }
    let ops = a.ops;
    let hasChanged = false;
    let isCollapsed = sel.isCollapsed();
    for (let i = 0; i < ops.length; i++) {
      let op = ops[i];
      hasChanged |= _transformCoordinateInplace$1(sel.start, op);
      if (!isCollapsed) {
        hasChanged |= _transformCoordinateInplace$1(sel.end, op);
      } else {
        if (sel.isContainerSelection()) {
          sel.end.path = sel.start.path;
        }
        sel.end.offset = sel.start.offset;
      }
    }
    return hasChanged
  }

  function _transformCoordinateInplace$1 (coor, op) {
    if (!isEqual$1(op.path, coor.path)) return false
    let hasChanged = false;
    if (op.type === 'update' && op.propertyType === 'string') {
      let diff = op.diff;
      let newOffset;
      if (diff.isInsert() && diff.pos <= coor.offset) {
        newOffset = coor.offset + diff.str.length;
        
        coor.offset = newOffset;
        hasChanged = true;
      } else if (diff.isDelete() && diff.pos <= coor.offset) {
        newOffset = Math.max(diff.pos, coor.offset - diff.str.length);
        
        coor.offset = newOffset;
        hasChanged = true;
      }
    }
    return hasChanged
  }

  var operationHelpers$1 = /*#__PURE__*/Object.freeze({
    transformDocumentChange: transformDocumentChange$1,
    transformSelection: transformSelection$1
  });

  function isFunction$1$1 (f) {
    return typeof f === 'function'
  }

  function deleteFromArray$1 (array, value) {
    if (!array) return
    for (var i = 0; i < array.length; i++) {
      if (array[i] === value) {
        array.splice(i, 1);
        i--;
      }
    }
  }

  class TreeNode$1 {}



  class TreeIndex$1 {
    
    get (path) {
      if (arguments.length > 1) {
        path = Array.prototype.slice(arguments, 0);
      }
      if (isString$1(path)) {
        path = [path];
      }
      return get$1(this, path)
    }

    getAll (path) {
      if (arguments.length > 1) {
        path = Array.prototype.slice(arguments, 0);
      }
      if (isString$1(path)) {
        path = [path];
      }
      if (!isArray$2(path)) {
        throw new Error('Illegal argument for TreeIndex.get()')
      }
      var node = get$1(this, path);
      return this._collectValues(node)
    }

    set (path, value) {
      if (isString$1(path)) {
        path = [path];
      }
      setWith$1(this, path, value, function (val) {
        if (!val) return new TreeNode$1()
      });
    }

    delete (path) {
      if (isString$1(path)) {
        delete this[path];
      } else if (path.length === 1) {
        delete this[path[0]];
      } else {
        var key = path[path.length - 1];
        path = path.slice(0, -1);
        var parent = get$1(this, path);
        if (parent) {
          delete parent[key];
        }
      }
    }

    clear () {
      var root = this;
      for (var key in root) {
        if (root.hasOwnProperty(key)) {
          delete root[key];
        }
      }
    }

    traverse (fn) {
      this._traverse(this, [], fn);
    }

    forEach (...args) {
      this.traverse(...args);
    }

    _traverse (root, path, fn) {
      var id;
      for (id in root) {
        if (!root.hasOwnProperty(id)) continue
        var child = root[id];
        var childPath = path.concat([id]);
        if (child instanceof TreeNode$1) {
          this._traverse(child, childPath, fn);
        } else {
          fn(child, childPath);
        }
      }
    }

    _collectValues (root) {
      
      
      var vals = {};
      this._traverse(root, [], function (val, path) {
        var key = path[path.length - 1];
        vals[key] = val;
      });
      return vals
    }
  }

  class TreeIndexArrays$1 extends TreeIndex$1 {
    contains (path) {
      let val = super.get(path);
      return Boolean(val)
    }

    get (path) {
      let val = super.get(path);
      if (val instanceof TreeNode$1) {
        val = val.__values__ || [];
      }
      return val
    }

    set (path, arr) {
      let val = super.get(path);
      val.__values__ = arr;
    }

    add (path, value) {
      if (isString$1(path)) {
        path = [path];
      }
      if (!isArray$2(path)) {
        throw new Error('Illegal arguments.')
      }
      var arr;

      
      
      
      
      
      
      setWith$1(this, path.concat(['__values__', '__dummy__']), undefined, function (val, key) {
        if (key === '__values__') {
          if (!val) val = [];
          arr = val;
        } else if (!val) {
          val = new TreeNode$1();
        }
        return val
      });
      delete arr.__dummy__;
      arr.push(value);
    }

    remove (path, value) {
      var arr = get$1(this, path);
      if (arr instanceof TreeNode$1) {
        if (arguments.length === 1) {
          delete arr.__values__;
        } else {
          deleteFromArray$1(arr.__values__, value);
        }
      }
    }

    _collectValues (root) {
      var vals = [];
      this._traverse(root, [], function (val) {
        vals.push(val);
      });
      vals = Array.prototype.concat.apply([], vals);
      return vals
    }
  }

  TreeIndex$1.Arrays = TreeIndexArrays$1;

  function filter$1 (iteratee, fn) {
    if (!iteratee) return []
    if (iteratee.constructor.prototype.filter && isFunction$1$1(iteratee.constructor.prototype.filter)) {
      return iteratee.filter(fn)
    }
    let result = [];
    forEach$1(iteratee, (val, key) => {
      if (fn(val, key)) {
        result.push(val);
      }
    });
    return result
  }

  function flatten$1 (arr) {
    return Array.prototype.concat.apply([], arr)
  }

  function flattenOften$1 (arr, max) {
    if (!(max > 0)) throw new Error("'max' must be a positive number")
    let l = arr.length;
    arr = flatten$1(arr);
    let round = 1;
    while (round < max && l < arr.length) {
      l = arr.length;
      arr = flatten$1(arr);
      round++;
    }
    return arr
  }

  class NodeIndex$1 {
    
    select (node) { 
      throw new Error('This method is abstract.')
    }

    clear () {
      throw new Error('This method is abstract')
    }

    
    create (node) { 
      throw new Error('This method is abstract.')
    }

    
    delete (node) { 
      throw new Error('This method is abstract.')
    }

    set (node, path, newValue, oldValue) {
      this.update(node, path, newValue, oldValue);
    }

    
    update (node, path, newValue, oldValue) { 
      throw new Error('This method is abstract.')
    }

    
    reset (data) {
      this.clear();
      this._initialize(data);
    }

    
    clone () {
      var NodeIndexClass = this.constructor;
      var clone = new NodeIndexClass();
      return clone
    }

    _initialize (data) {
      forEach$1(data.getNodes(), function (node) {
        if (this.select(node)) {
          this.create(node);
        }
      }.bind(this));
    }
  }


  NodeIndex$1.create = function (prototype) {
    var index = Object.assign(new NodeIndex$1(), prototype);
    index.clone = function () {
      return NodeIndex$1.create(prototype)
    };
    return index
  };


  NodeIndex$1.filterByType = function (type) {
    return function (node) {
      return node.isInstanceOf(type)
    }
  };

  class DocumentIndex$1 extends NodeIndex$1 {}

  var annotationHelpers$1 = {
    insertedText: insertedText$1,
    deletedText: deletedText$1,
    transferAnnotations: transferAnnotations$1,
    expandAnnotation: expandAnnotation$1,
    fuseAnnotation: fuseAnnotation$1,
    truncateAnnotation: truncateAnnotation$1
  };

  function insertedText$1 (doc, coordinate, length) {
    if (!length) return
    var index = doc.getIndex('annotations');
    var annotations = index.get(coordinate.path);
    for (let i = 0; i < annotations.length; i++) {
      let anno = annotations[i];
      var pos = coordinate.offset;
      var start = anno.start.offset;
      var end = anno.end.offset;
      var newStart = start;
      var newEnd = end;
      if ((pos < start) ||
           (pos === start)) {
        newStart += length;
      }
      
      if ((pos < end) ||
           (pos === end && !anno.isInline())) {
        newEnd += length;
      }
      
      if (newStart !== start) {
        doc.set([anno.id, 'start', 'offset'], newStart);
      }
      if (newEnd !== end) {
        doc.set([anno.id, 'end', 'offset'], newEnd);
      }
    }

    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
  }

  function deletedText$1 (doc, path, startOffset, endOffset) {
    if (startOffset === endOffset) return
    var index = doc.getIndex('annotations');
    var annotations = index.get(path);
    var length = endOffset - startOffset;
    for (let i = 0; i < annotations.length; i++) {
      let anno = annotations[i];
      var pos1 = startOffset;
      var pos2 = endOffset;
      var start = anno.start.offset;
      var end = anno.end.offset;
      var newStart = start;
      var newEnd = end;
      if (pos2 <= start) {
        newStart -= length;
        newEnd -= length;
        doc.set([anno.id, 'start', 'offset'], newStart);
        doc.set([anno.id, 'end', 'offset'], newEnd);
      } else {
        if (pos1 <= start) {
          newStart = start - Math.min(pos2 - pos1, start - pos1);
        }
        if (pos1 <= end) {
          newEnd = end - Math.min(pos2 - pos1, end - pos1);
        }
        
        if (start !== end && newStart === newEnd) {
          doc.delete(anno.id);
        } else {
          
          if (start !== newStart) {
            doc.set([anno.id, 'start', 'offset'], newStart);
          }
          if (end !== newEnd) {
            doc.set([anno.id, 'end', 'offset'], newEnd);
          }
        }
      }
    }
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
  }


  function transferAnnotations$1 (doc, path, offset, newPath, newOffset) {
    var index = doc.getIndex('annotations');
    var annotations = index.get(path, offset);
    for (let i = 0; i < annotations.length; i++) {
      let a = annotations[i];
      var isInside = (offset > a.start.offset && offset < a.end.offset);
      var start = a.start.offset;
      var end = a.end.offset;
      
      if (isInside) {
        
        if (a.canSplit()) {
          let newAnno = a.toJSON();
          newAnno.id = uuid$1(a.type + '_');
          newAnno.start.path = newPath;
          newAnno.start.offset = newOffset;
          newAnno.end.path = newPath;
          newAnno.end.offset = newOffset + a.end.offset - offset;
          doc.create(newAnno);
        }
        
        let newStartOffset = a.start.offset;
        let newEndOffset = offset;
        
        if (newEndOffset === newStartOffset) {
          doc.delete(a.id);
        
        } else {
          
          if (newStartOffset !== start) {
            doc.set([a.id, 'start', 'offset'], newStartOffset);
          }
          if (newEndOffset !== end) {
            doc.set([a.id, 'end', 'offset'], newEndOffset);
          }
        }
      
      } else if (a.start.offset >= offset) {
        
        
        
        doc.set([a.id, 'start', 'path'], newPath);
        doc.set([a.id, 'start', 'offset'], newOffset + a.start.offset - offset);
        doc.set([a.id, 'end', 'path'], newPath);
        doc.set([a.id, 'end', 'offset'], newOffset + a.end.offset - offset);
      }
    }

    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
  }


  function truncateAnnotation$1 (tx, anno, sel) {
    if (!sel || !sel._isSelection) throw new Error('Argument "selection" is required.')
    if (!anno || !anno.isAnnotation()) throw new Error('Argument "anno" is required and must be an annotation.')
    let annoSel = anno.getSelection();
    let newAnnoSel = annoSel.truncateWith(sel);
    anno._updateRange(tx, newAnnoSel);
    return anno
  }


  function expandAnnotation$1 (tx, anno, sel) {
    if (!sel || !sel._isSelection) throw new Error('Argument "selection" is required.')
    if (!anno || !anno.isAnnotation()) throw new Error('Argument "anno" is required and must be an annotation.')
    let annoSel = anno.getSelection();
    let newAnnoSel = annoSel.expand(sel);
    anno._updateRange(tx, newAnnoSel);
    return anno
  }


  function fuseAnnotation$1 (tx, annos) {
    if (!isArray$2(annos) || annos.length < 2) {
      throw new Error('fuseAnnotation(): at least two annotations are necessary.')
    }
    let sel, annoType;
    annos.forEach(function (anno, idx) {
      if (idx === 0) {
        sel = anno.getSelection();
        annoType = anno.type;
      } else {
        if (anno.type !== annoType) {
          throw new Error('fuseAnnotation(): all annotations must be of the same type.')
        }
        sel = sel.expand(anno.getSelection());
      }
    });
    
    for (var i = 1; i < annos.length; i++) {
      tx.delete(annos[i].id);
    }
    expandAnnotation$1(tx, annos[0], sel);
    tx.setSelection(sel);
  }

  function getPropertyAnnotationsForSelection$1 (doc, sel, options) {
    options = options || {};
    if (!sel.isPropertySelection()) {
      return []
    }
    let path = sel.getPath();
    let annotations = doc.getIndex('annotations').get(path, sel.start.offset, sel.end.offset);
    if (options.type) {
      annotations = filter$1(annotations, DocumentIndex$1.filterByType(options.type));
    }
    return annotations
  }


  function getContainerAnnotationsForSelection$1 (doc, sel, containerId, options) {
    
    
    
    
    
    if (!containerId) {
      throw new Error("'containerId' is required.")
    }
    options = options || {};
    let index = doc.getIndex('container-annotations');
    let annotations = [];
    if (index) {
      annotations = index.get(containerId, options.type);
      annotations = filter$1(annotations, function (anno) {
        return sel.overlaps(anno.getSelection())
      });
    }
    return annotations
  }


  function isContainerAnnotation$1 (doc, type) {
    let schema = doc.getSchema();
    return schema.isInstanceOf(type, '@container-annotation')
  }


  function getTextForSelection$1 (doc, sel) {
    if (!sel || sel.isNull()) {
      return ''
    } else if (sel.isPropertySelection()) {
      let text = doc.get(sel.start.path);
      return text.substring(sel.start.offset, sel.end.offset)
    } else if (sel.isContainerSelection()) {
      let result = [];
      let nodeIds = sel.getNodeIds();
      let L = nodeIds.length;
      for (let i = 0; i < L; i++) {
        let id = nodeIds[i];
        let node = doc.get(id);
        if (node.isText()) {
          let text = node.getText();
          if (i === L - 1) {
            text = text.slice(0, sel.end.offset);
          }
          if (i === 0) {
            text = text.slice(sel.start.offset);
          }
          result.push(text);
        }
      }
      return result.join('\n')
    }
  }

  function getMarkersForSelection$1 (doc, sel) {
    
    if (!sel || !sel.isPropertySelection()) return []
    const path = sel.getPath();
    
    let markers = doc.getIndex('markers').get(path);
    const filtered = filter$1(markers, function (m) {
      return m.containsSelection(sel)
    });
    return filtered
  }


  function deleteNode$1 (doc, node) {
    
    if (!node) {
      console.warn('Invalid arguments');
      return
    }
    
    if (node.isText()) {
      
      let annos = doc.getIndex('annotations').get(node.id);
      for (let i = 0; i < annos.length; i++) {
        doc.delete(annos[i].id);
      }
    }
    let nodeSchema = node.getSchema();
    
    
    
    doc.delete(node.id);
    
    let ownedProps = nodeSchema.getOwnedProperties();
    ownedProps.forEach(prop => {
      let value = node[prop.name];
      if (prop.isArray()) {
        let ids = value;
        if (ids.length > 0) {
          
          if (isArray$2(ids[0])) ids = flattenOften$1(ids, 2);
          ids.forEach((id) => {
            deleteNode$1(doc, doc.get(id));
          });
        }
      } else {
        deleteNode$1(doc, doc.get(value));
      }
    });
  }


  function copyNode$1 (node) {
    let nodes = [];
    
    let doc = node.getDocument();
    let nodeSchema = node.getSchema();
    for (let prop of nodeSchema) {
      
      
      if ((prop.isReference() && prop.isOwned()) || (prop.type === 'file')) {
        let val = node[prop.name];
        nodes.push(_copyChildren(val));
      }
    }
    nodes.push(node.toJSON());
    let annotationIndex = node.getDocument().getIndex('annotations');
    let annotations = annotationIndex.get([node.id]);
    forEach$1(annotations, function (anno) {
      nodes.push(anno.toJSON());
    });
    let result = flatten$1(nodes).filter(Boolean);
    
    return result

    function _copyChildren (val) {
      if (!val) return null
      if (isArray$2(val)) {
        return flatten$1(val.map(_copyChildren))
      } else {
        let id = val;
        if (!id) return null
        let child = doc.get(id);
        if (!child) return
        return copyNode$1(child)
      }
    }
  }


  function deleteTextRange$1 (doc, start, end) {
    if (!start) {
      start = {
        path: end.path,
        offset: 0
      };
    }
    let path = start.path;
    let text = doc.get(path);
    if (!end) {
      end = {
        path: start.path,
        offset: text.length
      };
    }
    
    if (!isArrayEqual$1(start.path, end.path)) {
      throw new Error('start and end must be on one property')
    }
    let startOffset = start.offset;
    if (startOffset < 0) throw new Error('start offset must be >= 0')
    let endOffset = end.offset;
    if (endOffset > text.length) throw new Error('end offset must be smaller than the text length')

    doc.update(path, { type: 'delete', start: startOffset, end: endOffset });
    
    let annos = doc.getAnnotations(path);
    annos.forEach(function (anno) {
      let annoStart = anno.start.offset;
      let annoEnd = anno.end.offset;
      
      if (annoEnd <= startOffset) ; else if (annoStart >= endOffset) {
        doc.update([anno.id, 'start'], { type: 'shift', value: startOffset - endOffset });
        doc.update([anno.id, 'end'], { type: 'shift', value: startOffset - endOffset });
      
      } else if (annoStart >= startOffset && annoEnd <= endOffset) {
        doc.delete(anno.id);
      
      } else if (annoStart >= startOffset && annoEnd >= endOffset) {
        if (annoStart > startOffset) {
          doc.update([anno.id, 'start'], { type: 'shift', value: startOffset - annoStart });
        }
        doc.update([anno.id, 'end'], { type: 'shift', value: startOffset - endOffset });
      
      } else if (annoStart <= startOffset && annoEnd <= endOffset) {
        doc.update([anno.id, 'end'], { type: 'shift', value: startOffset - annoEnd });
      
      } else if (annoStart < startOffset && annoEnd >= endOffset) {
        doc.update([anno.id, 'end'], { type: 'shift', value: startOffset - endOffset });
      } else {
        console.warn('TODO: handle annotation update case.');
      }
    });
  }

  function deleteListRange$1 (doc, list, start, end) {
    if (doc !== list.getDocument()) {
      list = doc.get(list.id);
    }
    let startItem, endItem;
    if (!start) {
      startItem = list.getItemAt(0);
      start = {
        path: startItem.getPath(),
        offset: 0
      };
    } else {
      startItem = doc.get(start.path[0]);
    }
    if (!end) {
      endItem = list.getLastItem();
      end = {
        path: endItem.getPath(),
        offset: endItem.getLength()
      };
    } else {
      endItem = doc.get(end.path[0]);
    }
    let startPos = list.getItemPosition(startItem);
    let endPos = list.getItemPosition(endItem);
    
    if (startPos === endPos) {
      deleteTextRange$1(doc, start, end);
      return
    }
    
    if (startPos > endPos) {
      [start, end] = [end, start];
      [startPos, endPos] = [endPos, startPos];
      [startItem, endItem] = [endItem, startItem];
    }
    let firstEntirelySelected = isEntirelySelected$1(doc, startItem, start, null);
    let lastEntirelySelected = isEntirelySelected$1(doc, endItem, null, end);

    
    if (lastEntirelySelected) {
      list.removeItemAt(endPos);
      deleteNode$1(doc, endItem);
    } else {
      deleteTextRange$1(doc, null, end);
    }

    
    let items = list.getItems();
    for (let i = endPos - 1; i > startPos; i--) {
      let item = items[i];
      list.removeItemAt(i);
      deleteNode$1(doc, item);
    }

    
    if (firstEntirelySelected) {
      
      
      
      
      deleteTextRange$1(doc, start, null);
    } else {
      deleteTextRange$1(doc, start, null);
    }

    if (!firstEntirelySelected && !lastEntirelySelected) {
      mergeListItems$1(doc, list.id, startPos);
    }
  }

  function mergeListItems$1 (doc, listId, itemPos) {
    
    let list = doc.get(listId);
    let targetItem = list.getItemAt(itemPos);
    let targetPath = targetItem.getPath();
    let targetLength = targetItem.getLength();
    let sourceItem = list.getItemAt(itemPos + 1);
    let sourcePath = sourceItem.getPath();
    
    list.removeItemAt(itemPos + 1);
    
    doc.update(targetPath, { type: 'insert', start: targetLength, text: sourceItem.getText() });
    
    annotationHelpers$1.transferAnnotations(doc, sourcePath, 0, targetPath, targetLength);
    deleteNode$1(doc, sourceItem);
  }

  function getNodes$1 (doc, ids) {
    return ids.map((id) => {
      return doc.get(id, 'strict')
    })
  }



  const SNIPPET_ID$1 = 'snippet';
  const TEXT_SNIPPET_ID$1 = 'text-snippet';

  var documentHelpers$1 = /*#__PURE__*/Object.freeze({
    getPropertyAnnotationsForSelection: getPropertyAnnotationsForSelection$1,
    getContainerAnnotationsForSelection: getContainerAnnotationsForSelection$1,
    isContainerAnnotation: isContainerAnnotation$1,
    getTextForSelection: getTextForSelection$1,
    getMarkersForSelection: getMarkersForSelection$1,
    deleteNode: deleteNode$1,
    copyNode: copyNode$1,
    deleteTextRange: deleteTextRange$1,
    deleteListRange: deleteListRange$1,
    mergeListItems: mergeListItems$1,
    getNodes: getNodes$1,
    SNIPPET_ID: SNIPPET_ID$1,
    TEXT_SNIPPET_ID: TEXT_SNIPPET_ID$1
  });

  const PLAINOBJ$1 = {};


  class Registry$1 {
    constructor (entries, validator) {
      this.entries = {};
      this.names = [];
      this.validator = validator;

      if (entries) {
        forEach$1(entries, function (entry, name) {
          this.add(name, entry);
        }.bind(this));
      }
    }

    
    contains (name) {
      return this.entries.hasOwnProperty(name)
    }

    
    add (name, entry) {
      if (this.validator) {
        this.validator(entry);
      }
      if (PLAINOBJ$1[name]) {
        throw new Error('Illegal key: "' + name + '" is a property of Object which is thus not allowed as a key.')
      }
      if (this.contains(name)) {
        this.remove(name);
      }
      this.entries[name] = entry;
      this.names.push(name);
    }

    
    remove (name) {
      let pos = this.names.indexOf(name);
      if (pos >= 0) {
        this.names.splice(pos, 1);
      }
      delete this.entries[name];
    }

    
    clear () {
      this.names = [];
      this.entries = {};
    }

    
    get (name, strict) {
      let result = this.entries[name];
      if (strict && !result) {
        throw new Error('No entry registered for name ' + name)
      }
      return result
    }

    
    forEach (callback) {
      for (let i = 0; i < this.names.length; i++) {
        let name = this.names[i];
        let _continue = callback(this.entries[name], name);
        if (_continue === false) {
          break
        }
      }
    }

    map (callback) {
      let result = [];
      this.forEach((entry, name) => {
        result.push(callback(entry, name));
      });
      return result
    }

    filter (callback) {
      let result = [];
      this.forEach(function (entry, name) {
        if (callback(entry, name)) {
          result.push(entry);
        }
      });
      return result
    }

    values () {
      return this.filter(() => { return true })
    }
  }

  Registry$1.prototype._isRegistry = true;

  function isBoolean$1 (val) {
    return (val === true || val === false || (val && val.constructor === Boolean))
  }

  class NodeProperty$1 {
    constructor (name, definition) {
      this.name = name;
      this.definition = definition;

      Object.freeze(this);
      Object.freeze(definition);
    }

    isArray () {
      return isArray$2(this.definition.type)
    }

    isReference () {
      if (this.isArray()) {
        return last$1$1(this.definition.type) === 'id'
      } else {
        return this.definition.type === 'id'
      }
    }

    isText () {
      return Boolean(this.definition._isText)
    }

    isOwned () {
      return Boolean(this.definition.owned)
    }

    isOptional () {
      return this.definition.optional || this.hasDefault()
    }

    isNotNull () {
      return Boolean(this.definition.notNull)
    }

    hasDefault () {
      return this.definition.hasOwnProperty('default')
    }

    getDefault () {
      return this.definition.default
    }

    createDefaultValue () {
      if (isArray$2(this.definition.type)) {
        return []
      }
      switch (this.definition.type) {
        case 'object':
          return {}
        case 'number':
          return -1
        case 'coordinate':
          return new Coordinate$1([], 0)
        case 'boolean':
          return false
        case 'id':
          return null
        case 'string':
          return ''
        default:
          return null
      }
    }

    get type () {
      return this.definition.type
    }

    get targetTypes () {
      return this.definition.targetTypes
    }
  }

  class NodeSchema$1 {
    constructor (properties) {
      this.properties = properties;

      
      
      
      
      
      this._ownedPropNames = new Set();
      this._ownedProps = [];
      forEach$1(properties, (prop) => {
        if ((prop.isReference() && prop.isOwned()) || (prop.type === 'file')) {
          this._ownedPropNames.add(prop.name);
          this._ownedProps.push(prop);
        }
      });
    }

    getProperty (name) {
      return this.properties[name]
    }

    hasOwnedProperties () {
      return this._ownedPropNames.size > 0
    }

    getOwnedProperties () {
      return this._ownedProps.slice()
    }

    isOwned (name) {
      return this._ownedPropNames.has(name)
    }

    [Symbol.iterator] () {
      const properties = this.properties;
      let ids = Object.keys(properties);
      let idx = 0;
      return { 
        next () {
          let done = idx > ids.length - 1;
          if (done) {
            return { done }
          } else {
            return {
              value: properties[ids[idx++]]
            }
          }
        }
      }
    }
  }

  class Node$1 extends EventEmitter$1 {
    
    constructor () {
      super();

      
      
      this._initialize.apply(this, arguments);
    }

    _initialize (data) {
      const NodeClass = this.constructor;

      let schema = NodeClass.schema;
      for (let property of schema) {
        let name = property.name;
        
        
        const propIsGiven = (data[name] !== undefined);
        const isOptional = property.isOptional();
        const hasDefault = property.hasDefault();
        if (!isOptional && !propIsGiven) {
          throw new Error('Property ' + name + ' is mandatory for node type ' + this.type)
        }
        if (propIsGiven) {
          this[name] = _checked$1(property, data[name]);
        } else if (hasDefault) {
          this[name] = cloneDeep$1(_checked$1(property, property.getDefault()));
        }
      }
    }

    dispose () {
      this._disposed = true;
    }

    isDisposed () {
      return Boolean(this._disposed)
    }

    
    isInstanceOf (typeName) {
      return Node$1.isInstanceOf(this.constructor, typeName)
    }

    getSchema () {
      return this.constructor.schema
    }

    
    getTypeNames () {
      var typeNames = [];
      var NodeClass = this.constructor;
      while (NodeClass.type !== '@node') {
        typeNames.push(NodeClass.type);
        NodeClass = Object.getPrototypeOf(NodeClass);
      }
      return typeNames
    }

    
    getPropertyType (propertyName) {
      return this.constructor.schema.getProperty(propertyName).type
    }

    
    toJSON () {
      var data = {
        type: this.type
      };
      const schema = this.getSchema();
      for (let prop of schema) {
        let val = this[prop.name];
        if (prop.isOptional() && val === undefined) continue
        if (isArray$2(val) || isObject$2(val)) {
          val = cloneDeep$1(val);
        }
        data[prop.name] = val;
      }
      return data
    }

    get type () {
      return this.constructor.type
    }
  }

  Node$1.prototype._isNode = true;


  Object.defineProperty(Node$1, 'schema', {
    get () { return this._schema },
    set (schema) {
      let NodeClass = this;
      
      
      if (schema.type) {
        NodeClass.type = schema.type;
      }
      
      
      
      NodeClass._schema = compileSchema$1(NodeClass, schema);
    }
  });

  Node$1.define = Node$1.defineSchema = function define (schema) {
    this.schema = schema;
  };

  Node$1.schema = {
    type: '@node',
    id: 'string'
  };


  Node$1.isInstanceOf = function (NodeClass, typeName) {
    var type = NodeClass.type;
    while (type !== '@node') {
      if (type === typeName) return true
      var _super = Object.getPrototypeOf(NodeClass.prototype).constructor;
      if (_super && _super.type) {
        NodeClass = _super;
        type = NodeClass.type;
      } else {
        break
      }
    }
    return false
  };



  function compileSchema$1 (NodeClass, schema) {
    let compiledSchema = _compileSchema$1(schema);
    let schemas = [compiledSchema];
    let clazz = NodeClass;
    while (clazz) {
      var parentProto = Object.getPrototypeOf(clazz.prototype);
      if (!parentProto) break
      clazz = parentProto.constructor;
      if (clazz && clazz._schema) {
        schemas.unshift(clazz._schema.properties);
      }
    }
    schemas.unshift({});
    return new NodeSchema$1(Object.assign.apply(null, schemas))
  }

  function _compileSchema$1 (schema) {
    let compiledSchema = {};
    forEach$1(schema, function (definition, name) {
      
      if (name === 'type') {
        return
      }
      if (isString$1(definition) || isArray$2(definition)) {
        definition = { type: definition };
      }
      definition = _compileDefintion$1(definition);
      definition.name = name;
      compiledSchema[name] = new NodeProperty$1(name, definition);
    });
    return compiledSchema
  }

  const _valueTypes$1 = new Set(['string', 'number', 'boolean', 'object', 'array']);

  function _compileDefintion$1 (definition) {
    let result = definition;
    if (isArray$2(definition.type)) {
      
      
      
      
      let defs = definition.type;
      let lastIdx = defs.length - 1;
      let first = defs[0];
      let last = defs[lastIdx];
      let isCanonical = first === 'array';
      let hasTargetType = last !== 'id' && !_valueTypes$1.has(last);
      definition.targetTypes = definition.type;
      if (isCanonical) {
        definition.type = defs.slice();
        definition.type[lastIdx] = 'id';
        if (hasTargetType) definition.targetTypes = [last];
      } else {
        if (defs.length > 1) {
          defs.forEach(t => {
            if (_valueTypes$1.has(t)) {
              throw new Error('Multi-types must consist of node types.')
            }
          });
          definition.type = [ 'array', 'id' ];
          definition.targetTypes = defs;
        } else {
          if (_valueTypes$1.has(first)) {
            definition.type = [ 'array', first ];
          } else {
            definition.type = [ 'array', 'id' ];
            definition.targetTypes = defs;
          }
        }
      }
    } else if (definition.type === 'text') {
      result = {
        type: 'string',
        default: '',
        _isText: true
      };
    }
    return result
  }

  function _checked$1 (prop, value) {
    let type;
    let name = prop.name;
    if (prop.isArray()) {
      type = 'array';
    } else {
      type = prop.type;
    }
    if (value === null) {
      if (prop.isNotNull()) {
        throw new Error('Value for property ' + name + ' is null.')
      } else {
        return value
      }
    }
    if (value === undefined) {
      throw new Error('Value for property ' + name + ' is undefined.')
    }
    if ((type === 'string' && !isString$1(value)) ||
        (type === 'boolean' && !isBoolean$1(value)) ||
        (type === 'number' && !isNumber$1(value)) ||
        (type === 'array' && !isArray$2(value)) ||
        (type === 'id' && !isString$1(value)) ||
        (type === 'object' && !isObject$2(value))) {
      throw new Error('Illegal value type for property ' + name + ': expected ' + type + ', was ' + (typeof value))
    }
    return value
  }

  class DocumentNode$1 extends Node$1 {
    _initialize (doc, props) {
      this.document = doc;
      super._initialize(props);
    }

    
    getDocument () {
      return this.document
    }

    
    hasParent () {
      return Boolean(this.parent)
    }

    
    getParent () {
      if (isString$1(this.parent)) return this.document.get(this.parent)
      return this.parent
    }

    setParent (parent) {
      if (isString$1(parent)) parent = this.document.get(parent);
      this.parent = parent;
    }

    
    getRoot () {
      let node = this;
      while (node.parent) {
        node = node.parent;
      }
      return node
    }

    getContainerRoot () {
      let node = this;
      while (node.parent) {
        
        if (node.parent.isContainer()) return node
        
        node = node.parent;
      }
      return node
    }

    
    hasChildren () {
      return false
    }

    
    getChildIndex(child) { 
      return -1
    }

    
    getChildAt(idx) { 
      return null
    }

    
    getChildCount () {
      return 0
    }

    
    

    

    
    isBlock () {
      return Boolean(this.constructor.isBlock)
    }

    
    isText () {
      return Boolean(this.constructor.isText)
    }

    isList () {
      return Boolean(this.constructor.isList)
    }

    isListItem () {
      return Boolean(this.constructor.isListItem)
    }

    isContainer () {
      return Boolean(this._isContainer)
    }

    

    isAnnotation () {
      return Boolean(this._isAnnotation)
    }

    isPropertyAnnotation () {
      return Boolean(this._isPropertyAnnotation)
    }

    isContainerAnnotation () {
      return Boolean(this._isContainerAnnotation)
    }

    
    isInline () {
      return Boolean(this.constructor.isInline)
    }
  }

  DocumentNode$1.prototype._isDocumentNode = true;


  DocumentNode$1.isBlock = false;


  DocumentNode$1.isText = false;


  DocumentNode$1.isPropertyAnnotation = false;


  DocumentNode$1.isContainerAnnotation = false;


  DocumentNode$1.isInline = false;

  class ContainerAddress$1 {
    constructor (pos, offset) {
      this.pos = pos;
      this.offset = offset;
    }

    isBefore (other, strict) {
      strict = Boolean(strict);
      if (this.pos < other.pos) {
        return true
      } else if (this.pos > other.pos) {
        return false
      } else if (this.offset < other.offset) {
        return true
      } else if (this.offset > other.offset) {
        return false
      }
      if (strict) {
        return false
      } else {
        return true
      }
    }

    isAfter (other, strict) {
      return other.isBefore(this, strict)
    }

    isEqual (other) {
      return (this.pos === other.pos && this.offset === other.offset)
    }

    toString () {
      return [this.pos, '.', this.offset].join('')
    }
  }

  function ContainerMixin$1 (DocumentNode) {
    class AbstractContainer extends DocumentNode {
      contains (nodeId) {
        return this.getPosition(nodeId) >= 0
      }

      getPosition (node, strict) {
        if (isString$1(node)) {
          node = this.document.get(node);
        }
        if (!node) return -1
        let pos = this._getPosition(node);
        if (strict && pos < 0) {
          throw new Error('Node is not within this container: ' + node.id)
        }
        return pos
      }

      getNodeAt (idx) {
        const nodeId = this.getNodeIdAt(idx);
        if (nodeId) {
          return this.getDocument().get(nodeId)
        }
      }

      getNodeIdAt (idx) {
        let content = this.getContent();
        if (idx < 0 || idx >= content.length) {
          
          return undefined
        } else {
          return content[idx]
        }
      }

      getNodes () {
        const doc = this.getDocument();
        return this.getContent().map(id => doc.get(id)).filter(Boolean)
      }

      show (nodeId, pos) {
        
        const arg1 = arguments[0];
        if (!isString$1(arg1)) {
          if (arg1._isNode) {
            nodeId = arg1.id;
          }
        }
        if (arguments.length > 1) {
          console.error('DEPRECATED: use container.showAt(pos, nodeId) instead');
        } else {
          pos = this.getLength();
        }
        return this.showAt(pos, nodeId)
      }

      showAt (pos, nodeId) {
        const doc = this.getDocument();
        const length = this.getLength();
        if (!isNumber$1(pos) || pos < 0 || pos > length) {
          throw new Error('Index out of bounds')
        }
        if (!isString$1(nodeId)) {
          if (nodeId._isNode) {
            nodeId = nodeId.id;
          } else {
            throw new Error('Invalid argument.')
          }
        }
        doc.update(this.getContentPath(), { type: 'insert', pos: pos, value: nodeId });
      }

      hide (nodeId) {
        const pos = this.getPosition(nodeId);
        this.hideAt(pos);
      }

      hideAt (pos) {
        const length = this.getLength();
        if (pos >= 0 && pos < length) {
          const doc = this.getDocument();
          doc.update(this.getContentPath(), { type: 'delete', pos: pos });
        } else {
          throw new Error('Index out of bounds.')
        }
      }

      getAddress (coor) {
        if (!coor._isCoordinate) {
          
          throw new Error('Illegal argument: Container.getAddress(coor) expects a Coordinate instance.')
        }
        var nodeId = coor.path[0];
        var nodePos = this.getPosition(nodeId);
        var offset;
        if (coor.isNodeCoordinate()) {
          if (coor.offset > 0) {
            offset = Number.MAX_VALUE;
          } else {
            offset = 0;
          }
        } else {
          offset = coor.offset;
        }
        return new ContainerAddress$1(nodePos, offset)
      }

      getLength () {
        return this.getContent().length
      }

      get length () {
        return this.getLength()
      }

      _getPosition (node) {
        if (this._isCaching) {
          return this._getCachedPosition(node)
        } else {
          return this._lookupPosition(node)
        }
      }

      _getCachedPosition (node) {
        let cache = this._cachedPositions || this._fillCache();
        let nodeId = node.id;
        let pos = -1;
        if (cache.hasOwnProperty(nodeId)) {
          pos = cache[nodeId];
        } else {
          pos = this._lookupPosition(node);
          cache[nodeId] = pos;
        }
        return pos
      }

      _fillCache () {
        let positions = {};
        this.nodes.forEach((id, pos) => {
          positions[id] = pos;
        });
        this._cachedPositions = positions;
        return positions
      }

      _invalidateCache () {
        this._cachedPositions = null;
      }

      _lookupPosition (node) {
        if (node.hasParent()) {
          node = node.getContainerRoot();
        }
        return this.getContent().indexOf(node.id)
      }

      _enableCaching () {
        
        if (this.document) {
          this.document.data.on('operation:applied', this._onOperationApplied, this);
          this._isCaching = true;
        }
      }

      _onOperationApplied (op) {
        if (op.type === 'set' || op.type === 'update') {
          if (op.path[0] === this.id) {
            this._invalidateCache();
          }
        }
      }

      _onDocumentChange (change) {
        if (change.hasUpdated(this.getContentPath())) {
          this._invalidateCache();
        }
      }

      
      

      hasChildren () {
        return this.getContent().length > 0
      }

      getChildIndex (child) {
        return this.getContent().indexOf(child.id)
      }

      getChildren () {
        return getNodes$1(this.getDocument(), this.getContent())
      }

      getChildAt (idx) {
        var childrenIds = this.getContent();
        if (idx < 0 || idx >= childrenIds.length) {
          throw new Error('Array index out of bounds: ' + idx + ', ' + childrenIds.length)
        }
        return this.getDocument().get(childrenIds[idx], 'strict')
      }

      getChildCount () {
        return this.getContent().length
      }
    }
    return AbstractContainer
  }

  class Container$1 extends ContainerMixin$1(DocumentNode$1) {
    constructor (...args) {
      super(...args);

      
      
      
      this._enableCaching();
    }

    dispose () {
      this.document.off(this);
    }

    getContentPath () {
      return [this.id, 'nodes']
    }

    getContent () {
      return this.nodes
    }
  }

  Container$1.prototype._isContainer = true;

  Container$1.schema = {
    type: '@container',
    nodes: { type: ['array', 'id'], default: [], owned: true }
  };

  function AnnotationMixin$1 (DocumentNode) {
    class AbstractAnnotation extends DocumentNode {
      constructor (doc, props) {
        super(doc, _normalizedProps$1(props));

        
        this.start = new Coordinate$1(this.start);
        this.end = new Coordinate$1(this.end);
      }

      

      get path () {
        console.warn('DEPRECATED: use annotation.start.path instead');
        return this.start.path
      }

      getPath () {
        return this.start.path
      }

      get startPath () {
        console.warn('DEPRECATED: use annotation.start.path instead.');
        return this.start.path
      }

      set startPath (path) {
        console.warn('DEPRECATED: use annotation.start.path instead.');
        this.start.path = path;
      }

      get startOffset () {
        console.warn('DEPRECATED: use annotation.start.offset instead.');
        return this.start.offset
      }

      set startOffset (offset) {
        console.warn('DEPRECATED: use annotation.start.offset instead.');
        this.start.offset = offset;
      }

      get endPath () {
        console.warn('DEPRECATED: use annotation.end.path instead.');
        return this.end.path
      }

      set endPath (path) {
        console.warn('DEPRECATED: use annotation.end.path instead.');
        this.end.path = path;
      }

      get endOffset () {
        console.warn('DEPRECATED: use annotation.end.offset instead.');
        return this.end.offset
      }

      set endOffset (offset) {
        console.warn('DEPRECATED: use annotation.end.offset instead.');
        this.end.offset = offset;
      }

      

      
      getText () {
        var doc = this.getDocument();
        if (!doc) {
          console.warn('Trying to use a Annotation which is not attached to the document.');
          return ''
        }
        return getTextForSelection$1(doc, this.getSelection())
      }

      isAnnotation () {
        return true
      }

      
      canSplit () {
        return true
      }

      
      isAnchor () {
        return false
      }

      
      getSelection () {
        const doc = this.getDocument();
        
        if (!doc) {
          console.warn('Trying to use a ContainerAnnotation which is not attached to the document.');
          return Selection$1.nullSelection()
        }
        if (this._isContainerAnnotation) {
          return doc.createSelection({
            type: 'container',
            containerId: this.containerId,
            startPath: this.start.path,
            startOffset: this.start.offset,
            endPath: this.end.path,
            endOffset: this.end.offset
          })
        } else {
          return this.getDocument().createSelection({
            type: 'property',
            path: this.start.path,
            startOffset: this.start.offset,
            endOffset: this.end.offset
          })
        }
      }

      _updateRange (tx, sel) {
        if (sel.isContainerSelection()) {
          
          if (!isEqual$1(this.start.path, sel.start.path)) {
            tx.set([this.id, 'start', 'path'], sel.start.path);
          }
          if (this.start.offset !== sel.start.offset) {
            tx.set([this.id, 'start', 'offset'], sel.start.offset);
          }
          if (!isEqual$1(this.end.path, sel.end.path)) {
            tx.set([this.id, 'end', 'path'], sel.end.path);
          }
          if (this.end.offset !== sel.end.offset) {
            tx.set([this.id, 'end', 'offset'], sel.end.offset);
          }
        } else if (sel.isPropertySelection()) {
          if (!isArrayEqual$1(this.start.path, sel.start.path)) {
            tx.set([this.id, 'path'], sel.start.path);
          }
          
          if (this.start.offset !== sel.start.offset) {
            tx.set([this.id, 'start', 'offset'], sel.start.offset);
          }
          if (this.end.offset !== sel.end.offset) {
            tx.set([this.id, 'end', 'offset'], sel.end.offset);
          }
        } else {
          throw new Error('Invalid selection.')
        }
      }
    }

    AbstractAnnotation.prototype._isAnnotation = true;

    AbstractAnnotation.schema = {
      start: { type: 'coordinate', default: { path: [], offset: 0 } },
      end: { type: 'coordinate', default: { path: [], offset: 0 } }
    };

    return AbstractAnnotation
  }

  function _normalizedProps$1 (props) {
    
    
    if (!props.hasOwnProperty('start')) {
      
      
      
      let start, end;
      if (props.hasOwnProperty('startPath') || props.hasOwnProperty('path')) {
        start = {
          path: props.startPath || props.path,
          offset: props.startOffset
        };
      }
      if (props.hasOwnProperty('endPath') || props.hasOwnProperty('endOffset')) {
        end = {
          path: props.endPath || props.path,
          offset: props.endOffset
        };
      }
      if (start && !end) {
        end = cloneDeep$1(start);
      }
      if (start) {
        props = Object.assign({}, props);
        delete props.path;
        delete props.startPath;
        delete props.endPath;
        delete props.startOffset;
        delete props.endOffset;
        props.start = start;
        props.end = end;
      }
    } else if (props.hasOwnProperty('end') && !props.end.path) {
      props.end.path = props.start.path;
    }
    return props
  }

  class PropertyAnnotation$1 extends AnnotationMixin$1(DocumentNode$1) {}

  PropertyAnnotation$1.prototype._isAnnotation = true;
  PropertyAnnotation$1.prototype._isPropertyAnnotation = true;

  PropertyAnnotation$1.isPropertyAnnotation = true;
  PropertyAnnotation$1.autoExpandRight = true;

  PropertyAnnotation$1.schema = {
    type: '@annotation',
    
    
    _content: { type: 'string', optional: true }
  };

  class ContainerAnnotation$1 extends AnnotationMixin$1(DocumentNode$1) {
    setHighlighted (highlighted, scope) {
      if (this.highlighted !== highlighted) {
        this.highlighted = highlighted;
        this.highlightedScope = scope;
        this.emit('highlighted', highlighted, scope);
        forEach$1(this.fragments, function (frag) {
          frag.emit('highlighted', highlighted, scope);
        });
      }
    }
  }

  ContainerAnnotation$1.schema = {
    type: '@container-annotation',
    containerId: 'string',
    start: 'coordinate',
    end: 'coordinate'
  };

  ContainerAnnotation$1.prototype._isAnnotation = true;
  ContainerAnnotation$1.prototype._isContainerAnnotation = true;

  class EditingBehavior$1 {
    constructor () {
      this._merge = {};
      this._mergeComponents = {};
      this._break = {};
    }

    defineMerge (firstType, secondType, impl) {
      if (!this._merge[firstType]) {
        this._merge[firstType] = {};
      }
      this._merge[firstType][secondType] = impl;
      return this
    }

    canMerge (firstType, secondType) {
      return (this._merge[firstType] && this._merge[firstType][secondType])
    }

    getMerger (firstType, secondType) {
      return this._merge[firstType][secondType]
    }

    defineComponentMerge (nodeType, impl) {
      this._mergeComponents[nodeType] = impl;
    }

    canMergeComponents (nodeType) {
      return this._mergeComponents[nodeType]
    }

    getComponentMerger (nodeType) {
      return this._mergeComponents[nodeType]
    }

    defineBreak (nodeType, impl) {
      this._break[nodeType] = impl;
      return this
    }

    canBreak (nodeType) {
      return this._break[nodeType]
    }

    getBreaker (nodeType) {
      return this._break[nodeType]
    }
  }

  function extend$1 (...args) {
    return Object.assign(...args)
  }

  class ArrayIterator$1 {
    constructor (arr) {
      this.arr = arr;
      this.pos = -1;
    }

    get _isArrayIterator () {
      return true
    }

    
    hasNext () {
      return this.pos < this.arr.length - 1
    }

    
    next () {
      this.pos += 1;
      var next = this.arr[this.pos];
      return next
    }

    
    back () {
      if (this.pos >= 0) {
        this.pos -= 1;
      }
      return this
    }

    peek () {
      return this.arr[this.pos + 1]
    }
  }

  function findIndex$1$1 (arr, predicate) {
    if (!isFunction$1$1(predicate)) return arr.indexOf(predicate)
    for (let i = 0; i < arr.length; i++) {
      if (predicate(arr[i])) return i
    }
    return -1
  }

  class DOMEventListener$1 {
    constructor (eventName, handler, options) {
      
      if (!isString$1(eventName) || !isFunction$1$1(handler)) {
        throw new Error("Illegal arguments: 'eventName' must be a String, and 'handler' must be a Function.")
      }
      options = options || {};
      var origHandler = handler;
      var context = options.context;
      var capture = Boolean(options.capture);

      if (context) {
        handler = handler.bind(context);
      }
      if (options.once === true) {
        handler = _once$1(this, handler);
      }

      this.eventName = eventName;
      this.originalHandler = origHandler;
      this.handler = handler;
      this.capture = capture;
      this.context = context;
      this.options = options;
      
      this._el = null;
    }
  }

  DOMEventListener$1.prototype._isDOMEventListener = true;

  DOMEventListener$1.findIndex = function (eventListeners, eventName, handler) {
    var idx = -1;
    if (arguments[1]._isDOMEventListener) {
      idx = eventListeners.indexOf(arguments[1]);
    } else {
      idx = findIndex$1$1(eventListeners,
        _matches$1.bind(null, {
          eventName: eventName,
          originalHandler: handler
        })
      );
    }
    return idx
  };

  function _matches$1 (l1, l2) {
    return l1.eventName === l2.eventName && l1.originalHandler === l2.originalHandler
  }

  function _once$1 (listener, handler) {
    return function (event) {
      handler(event);
      listener._el.removeEventListener(listener);
    }
  }

  const NOT_IMPLEMENTED$1 = 'This method is not implemented.';


  class DOMElement$1 {
    

    

    

    

    

    

    

    

    

    
    getNativeElement () {
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    isTextNode () {
      
      return false
    }

    
    isElementNode () {
      
      return false
    }

    
    isCommentNode () {
      
      return false
    }

    
    isDocumentNode () {
      
      return false
    }

    
    getTagName () {
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    setTagName (tagName) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    getId () {
      return this.getAttribute('id')
    }

    
    setId (id) {
      this.setAttribute('id', id);
      return this
    }

    
    hasClass (className) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    addClass (classString) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    removeClass (classString) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    hasAttribute (name) {
      return Boolean(this.getAttribute(name))
    }

    
    attr () {
      if (arguments.length === 1) {
        if (isString$1(arguments[0])) {
          return this.getAttribute(arguments[0])
        } else if (isObject$2(arguments[0])) {
          forEach$1(arguments[0], function (value, name) {
            this.setAttribute(name, value);
          }.bind(this));
        }
      } else if (arguments.length === 2) {
        this.setAttribute(arguments[0], arguments[1]);
      }
      return this
    }

    
    removeAttr (name) {
      var names = name.split(/\s+/);
      if (names.length === 1) {
        this.removeAttribute(name);
      } else {
        names.forEach(function (name) {
          this.removeAttribute(name);
        }.bind(this));
      }
      return this
    }

    
    getAttribute (name) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    setAttribute (name, value) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    removeAttribute (name) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    getAttributes () {
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    htmlProp () {
      if (arguments.length === 1) {
        if (isString$1(arguments[0])) {
          return this.getProperty(arguments[0])
        } else if (isObject$2(arguments[0])) {
          forEach$1(arguments[0], function (value, name) {
            this.setProperty(name, value);
          }.bind(this));
        }
      } else if (arguments.length === 2) {
        this.setProperty(arguments[0], arguments[1]);
      }
      return this
    }

    getProperty (name) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    setProperty (name, value) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    val (value) {
      if (arguments.length === 0) {
        return this.getValue()
      } else {
        this.setValue(value);
        return this
      }
    }

    getValue () {
      return this.getProperty('value')
    }

    setValue (value) {
      this.setProperty('value', value);
      return this
    }

    
    css () {
      
      if (arguments.length === 1) {
        
        if (isString$1(arguments[0])) {
          return this.getStyle(arguments[0])
        } else if (isObject$2(arguments[0])) {
          forEach$1(arguments[0], function (value, name) {
            this.setStyle(name, value);
          }.bind(this));
        } else {
          throw new Error('Illegal arguments.')
        }
      } else if (arguments.length === 2) {
        this.setStyle(arguments[0], arguments[1]);
      } else {
        throw new Error('Illegal arguments.')
      }
      return this
    }

    getStyle (name) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    setStyle (name, value) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    text (text) {
      if (arguments.length === 0) {
        return this.getTextContent()
      } else {
        this.setTextContent(text);
      }
      return this
    }

    
    getTextContent () {
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    setTextContent (text) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    html (html) {
      if (arguments.length === 0) {
        return this.getInnerHTML()
      } else {
        this.setInnerHTML(html);
      }
      return this
    }

    
    getInnerHTML () {
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    getInnerXML () {
      return this.getInnerHTML()
    }

    
    setInnerHTML (html) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    setInnerXML (xml) {
      return this.setInnerHTML(xml)
    }

    
    getOuterHTML () {
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    getOuterXML () {
      return this.getOuterHTML()
    }

    
    on (eventName, handler, context, options) {
      
      if (!isString$1(eventName)) {
        throw new Error('Illegal argument: "event" must be a String.')
      }
      options = options || {};
      if (context) {
        options.context = context;
      }
      
      if (!handler || !isFunction$1$1(handler)) {
        throw new Error('Illegal argument: invalid handler function for event ' + eventName)
      }
      this.addEventListener(eventName, handler, options);
      return this
    }

    
    off (eventName, handler) {
      
      if (arguments.length === 1 && !isString$1(eventName)) {
        let context = arguments[0];
        this.getEventListeners().filter(function (l) {
          return l.context === context
        }).forEach(function (l) {
          this.removeEventListener(l);
        }.bind(this));
      } else {
        this.removeEventListener(eventName, handler);
      }
      return this
    }

    addEventListener (eventName, handler, options = {}) {
      let listener;
      if (arguments.length === 1 && arguments[0]) {
        listener = arguments[0];
      } else {
        listener = this._createEventListener(eventName, handler, options);
      }
      if (!this.eventListeners) {
        this.eventListeners = [];
      }
      listener._el = this;
      this.eventListeners.push(listener);
      this._addEventListenerNative(listener);
      return this
    }

    _createEventListener (eventName, handler, options) {
      return new DOMEventListener$1(eventName, handler, options)
    }

    _addEventListenerNative(listener) {} 

    removeEventListener (eventName, handler) {
      if (!this.eventListeners) return
      
      let listener = null;
      let idx = DOMEventListener$1.findIndex(this.eventListeners, eventName, handler);
      listener = this.eventListeners[idx];
      if (idx > -1) {
        this.eventListeners.splice(idx, 1);
        
        listener._el = null;
        this._removeEventListenerNative(listener);
      }
      return this
    }

    _removeEventListenerNative(listener) {} 

    removeAllEventListeners () {
      if (!this.eventListeners) return
      for (let i = 0; i < this.eventListeners.length; i++) {
        let listener = this.eventListeners[i];
        
        listener._el = null;
        this._removeEventListenerNative(listener);
      }
      delete this.eventListeners;
    }

    getEventListeners () {
      return this.eventListeners || []
    }

    
    getNodeType () {
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    getContentType () {
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    getChildCount () {
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    getChildNodes () {
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    getChildren () {
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    getChildAt (pos) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    getChildIndex (child) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    getChildNodeIterator () {
      return new ArrayIterator$1(this.getChildNodes())
    }

    getLastChild () {
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    getFirstChild () {
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    getNextSibling () {
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    getPreviousSibling () {
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    clone(deep) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    createElement (str) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    createTextNode (text) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    createComment (data) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    createProcessingInstruction (name, data) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    createCDATASection (data) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    is (cssSelector) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    getParent () {
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    getOwnerDocument () {
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    getDoctype () {
      
      throw new Error('NOT_IMPLEMENTED')
    }

    setDoctype(qualifiedNameStr, publicId, systemId) { 
      throw new Error('NOT_IMPLEMENTED')
    }

    setDocType (...args) {
      console.error('DEPRECATED: use setDoctype() instead');
      return this.setDoctype(...args)
    }

    
    find (cssSelector) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    findAll (cssSelector) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    append (child) {
      var children;
      if (arguments.length === 1) {
        if (isArray$2(child)) {
          children = child;
        } else {
          this.appendChild(child);
          return this
        }
      } else {
        children = arguments;
      }
      if (children) {
        Array.prototype.forEach.call(children, this.appendChild.bind(this));
      }
      return this
    }

    appendChild (child) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    insertAt (pos, child) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    insertBefore (newChild, before) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    removeAt (pos) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    removeChild (child) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    replaceChild (oldChild, newChild) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    
    remove () {
      var parent = this.getParent();
      if (parent) {
        parent.removeChild(this);
      }
    }

    
    empty () {
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    serialize () {
      return this.getOuterHTML()
    }

    isInDocument () {
      let el = this;
      while (el) {
        if (el.isDocumentNode()) {
          return true
        }
        el = el.getParent();
      }
    }

    
    focus () {
      
      return this
    }

    
    select () {
      
      return this
    }

    
    blur () {
      
      return this
    }

    
    click () {
      
      return this
    }

    

    getWidth () {
      
      return 0
    }

    getHeight () {
      
      return 0
    }

    
    getOuterHeight (withMargin) { 
      
      return 0
    }

    
    getOffset () {
      
      return { top: 0, left: 0 }
    }

    
    getPosition () {
      
      return { top: 0, left: 0 }
    }

    
    getElementFactory () {
      return this.createElement.bind(this)
    }

    
    emit(name, data) { 
      
      throw new Error(NOT_IMPLEMENTED$1)
    }

    

    get id () {
      return this.getId()
    }

    set id (id) {
      this.setId(id);
    }

    get tagName () {
      return this.getTagName()
    }

    set tagName (tagName) {
      this.setTagName(tagName);
    }

    get nodeName () {
      return this.getTagName()
    }

    get nodeType () {
      return this.getNodeType()
    }

    get className () {
      return this.getAttribute('class')
    }

    set className (className) {
      this.setAttribute('class', className);
    }

    get textContent () {
      return this.getTextContent()
    }

    set textContent (text) {
      this.setTextContent(text);
    }

    get innerHTML () {
      return this.getInnerHTML()
    }

    set innerHTML (html) {
      this.setInnerHTML(html);
    }

    get outerHTML () {
      return this.getOuterHTML()
    }

    get firstChild () {
      return this.getFirstChild()
    }

    get lastChild () {
      return this.getLastChild()
    }

    get nextSibling () {
      return this.getNextSibling()
    }

    get previousSibling () {
      return this.getPreviousSibling()
    }

    get parentNode () {
      return this.getParent()
    }

    get height () {
      return this.getHeight()
    }

    get width () {
      return this.getWidth()
    }

    get value () {
      return this.getValue()
    }

    set value (value) {
      return this.setValue(value)
    }
  }

  DOMElement$1.prototype._isDOMElement = true;

  DOMElement$1.pxStyles = {
    top: true,
    bottom: true,
    left: true,
    right: true,
    height: true,
    width: true
  };

  DOMElement$1.EMPTY_HTML = '<html><head></head><body></body></html>';

  const SIGNATURE$1 = uuid$1('_BrowserDOMElement');

  function _attach$1 (nativeEl, browserDOMElement) {
    nativeEl[SIGNATURE$1] = browserDOMElement;
  }

  function _detach$1 (nativeEl) {
    delete nativeEl[SIGNATURE$1];
  }

  function _unwrap$1 (nativeEl) {
    return nativeEl[SIGNATURE$1]
  }

  class BrowserDOMElement$1 extends DOMElement$1 {
    constructor (el) {
      super();
      console.assert(el instanceof window.Node, 'Expecting native DOM node.');
      this.el = el;
      
      
      _attach$1(el, this);
    }

    getNativeElement () {
      return this.el
    }

    getNodeType () {
      switch (this.el.nodeType) {
        case window.Node.TEXT_NODE:
          return 'text'
        case window.Node.ELEMENT_NODE:
          return 'element'
        case window.Node.DOCUMENT_NODE:
          return 'document'
        case window.Node.COMMENT_NODE:
          return 'comment'
        case window.Node.PROCESSING_INSTRUCTION_NODE:
          return 'directive'
        case window.Node.CDATA_SECTION_NODE:
          return 'cdata'
        default:
          
      }
    }

    getDoctype () {
      if (this.isDocumentNode()) {
        return this.el.doctype
      } else {
        return this.getOwnerDocument().getDoctype()
      }
    }

    setDoctype (qualifiedNameStr, publicId, systemId) {
      let ownerDocument = this._getNativeOwnerDocument();
      let oldDocType = ownerDocument.doctype;
      let newDocType = ownerDocument.implementation.createDocumentType(
        qualifiedNameStr, publicId, systemId
      );
      if (oldDocType) {
        oldDocType.parentNode.replaceChild(newDocType, oldDocType);
      } else {
        ownerDocument.insertBefore(newDocType, ownerDocument.firstChild);
      }
    }

    isTextNode () {
      return (this.el.nodeType === window.Node.TEXT_NODE)
    }

    isElementNode () {
      return (this.el.nodeType === window.Node.ELEMENT_NODE)
    }

    isCommentNode () {
      return (this.el.nodeType === window.Node.COMMENT_NODE)
    }

    isDocumentNode () {
      return (this.el.nodeType === window.Node.DOCUMENT_NODE)
    }

    hasClass (className) {
      return this.el.classList.contains(className)
    }

    addClass (className) {
      this.el.classList.add(className);
      return this
    }

    removeClass (className) {
      this.el.classList.remove(className);
      return this
    }

    getAttribute (name) {
      return this.el.getAttribute(name)
    }

    setAttribute (name, value) {
      this.el.setAttribute(name, String(value));
      return this
    }

    removeAttribute (name) {
      this.el.removeAttribute(name);
      return this
    }

    getAttributes () {
      if (!this.el.attributes._mapAdapter) {
        this.el.attributes._mapAdapter = new AttributesMapAdapter$1(this.el.attributes);
      }
      return this.el.attributes._mapAdapter
    }

    getProperty (name) {
      return this.el[name]
    }

    setProperty (name, value) {
      
      
      if (this._isXML()) throw new Error('setProperty() is only supported for HTML elements.')
      this.el[name] = value;
      return this
    }

    getTagName () {
      
      
      if (this._isXML()) {
        return this.el.tagName
      } else if (this.el.tagName) {
        return this.el.tagName.toLowerCase()
      }
    }

    setTagName (tagName) {
      let newEl = this.createElement(tagName);
      let attributes = this.el.attributes;
      let l = attributes.length;
      let i;
      for (i = 0; i < l; i++) {
        let attr = attributes.item(i);
        newEl.setAttribute(attr.name, attr.value);
      }
      if (this.eventListeners) {
        this.eventListeners.forEach(function (listener) {
          newEl.addEventListener(listener.eventName, listener.handler, listener.capture);
        });
      }
      newEl.append(this.getChildNodes());

      this._replaceNativeEl(newEl.getNativeElement());
      return this
    }

    getId () {
      return this.el.id
    }

    setId (id) {
      this.el.id = id;
      return this
    }

    getStyle (name) {
      let val = this.el.style[name];
      if (!val) {
        let computedStyle = this.getComputedStyle();
        val = computedStyle[name];
      }
      return val
    }

    getComputedStyle () {
      return window.getComputedStyle(this.el)
    }

    setStyle (name, value) {
      if (DOMElement$1.pxStyles[name] && isNumber$1(value)) value = value + 'px';
      this.el.style[name] = value;
      return this
    }

    getTextContent () {
      return this.el.textContent
    }

    setTextContent (text) {
      this.el.textContent = text;
      return this
    }

    getInnerHTML () {
      if (this._isXML()) {
        let xs = new window.XMLSerializer();
        let result = Array.prototype.map.call(this.el.childNodes, c => xs.serializeToString(c));
        return result.join('')
      } else {
        return this.el.innerHTML
      }
    }

    setInnerHTML (html) {
      
      
      this.el.innerHTML = html;
      return this
    }

    getOuterHTML () {
      
      
      if (this._isXML()) {
        let xs = new window.XMLSerializer();
        return xs.serializeToString(this.el)
      } else {
        return this.el.outerHTML
      }
    }

    _addEventListenerNative (listener) {
      this.el.addEventListener(listener.eventName, listener.handler, listener.capture);
    }

    _removeEventListenerNative (listener) {
      this.el.removeEventListener(listener.eventName, listener.handler);
    }

    getEventListeners () {
      return this.eventListeners || []
    }

    getChildCount () {
      return this.el.childNodes.length
    }

    getChildNodes () {
      let childNodes = [];
      for (let node = this.el.firstChild; node; node = node.nextSibling) {
        childNodes.push(BrowserDOMElement$1.wrap(node));
      }
      return childNodes
    }

    getChildNodeIterator () {
      return new BrowserChildNodeIterator$1(this.el)
    }

    get childNodes () {
      return this.getChildNodes()
    }

    getChildren () {
      
      
      let children = [];
      for (let node = this.el.firstChild; node; node = node.nextSibling) {
        if (node.nodeType === window.Node.ELEMENT_NODE) {
          children.push(BrowserDOMElement$1.wrap(node));
        }
      }
      return children
    }

    get children () {
      return this.getChildren()
    }

    getChildAt (pos) {
      return BrowserDOMElement$1.wrap(this.el.childNodes[pos])
    }

    getChildIndex (child) {
      
      if (!child._isBrowserDOMElement) {
        throw new Error('Expecting a BrowserDOMElement instance.')
      }
      return Array.prototype.indexOf.call(this.el.childNodes, child.el)
    }

    getFirstChild () {
      let firstChild = this.el.firstChild;
      
      if (firstChild) {
        return BrowserDOMElement$1.wrap(firstChild)
      } else {
        return null
      }
    }

    getLastChild () {
      var lastChild = this.el.lastChild;
      
      if (lastChild) {
        return BrowserDOMElement$1.wrap(lastChild)
      } else {
        return null
      }
    }

    getNextSibling () {
      let next = this.el.nextSibling;
      
      if (next) {
        return BrowserDOMElement$1.wrap(next)
      } else {
        return null
      }
    }

    getPreviousSibling () {
      let previous = this.el.previousSibling;
      
      if (previous) {
        return BrowserDOMElement$1.wrap(previous)
      } else {
        return null
      }
    }

    clone (deep) {
      let clone = this.el.cloneNode(deep);
      return BrowserDOMElement$1.wrap(clone)
    }

    createDocument (format) {
      return BrowserDOMElement$1.createDocument(format)
    }

    createElement (tagName) {
      let doc = this._getNativeOwnerDocument();
      let el = doc.createElement(tagName);
      return BrowserDOMElement$1.wrap(el)
    }

    createTextNode (text) {
      let doc = this._getNativeOwnerDocument();
      let el = doc.createTextNode(text);
      return BrowserDOMElement$1.wrap(el)
    }

    createComment (data) {
      let doc = this._getNativeOwnerDocument();
      let el = doc.createComment(data);
      return BrowserDOMElement$1.wrap(el)
    }

    createProcessingInstruction (name, data) {
      let doc = this._getNativeOwnerDocument();
      let el = doc.createProcessingInstruction(name, data);
      return BrowserDOMElement$1.wrap(el)
    }

    createCDATASection (data) {
      let doc = this._getNativeOwnerDocument();
      let el = doc.createCDATASection(data);
      return BrowserDOMElement$1.wrap(el)
    }

    is (cssSelector) {
      
      
      let el = this.el;
      
      if (this.isElementNode()) {
        return matches$1(el, cssSelector)
      } else {
        return false
      }
    }

    getParent () {
      let parent = this.el.parentNode;
      
      if (parent) {
        return BrowserDOMElement$1.wrap(parent)
      } else {
        return null
      }
    }

    getOwnerDocument () {
      return BrowserDOMElement$1.wrap(this._getNativeOwnerDocument())
    }

    get ownerDocument () {
      return this.getOwnerDocument()
    }

    _getNativeOwnerDocument () {
      return (this.isDocumentNode() ? this.el : this.el.ownerDocument)
    }

    find (cssSelector) {
      let result = null;
      if (this.el.querySelector) {
        result = this.el.querySelector(cssSelector);
      }
      if (result) {
        return BrowserDOMElement$1.wrap(result)
      } else {
        return null
      }
    }

    findAll (cssSelector) {
      let result = [];
      if (this.el.querySelectorAll) {
        result = this.el.querySelectorAll(cssSelector);
      }
      return Array.prototype.map.call(result, function (el) {
        return BrowserDOMElement$1.wrap(el)
      })
    }

    _normalizeChild (child) {
      if (isNil$1(child)) return child

      if (child instanceof window.Node) {
        child = BrowserDOMElement$1.wrap(child);
      
      
      
      } else if (child._isBrowserDOMElement && !(child instanceof BrowserDOMElement$1)) {
        child = BrowserDOMElement$1.wrap(child);
      } else if (isString$1(child) || isNumber$1(child)) {
        child = this.createTextNode(child);
      }
      
      if (!child || !child._isBrowserDOMElement) {
        throw new Error('Illegal child type.')
      }
      console.assert(_unwrap$1(child.el) === child, 'The backlink to the wrapper should be consistent');
      return child.getNativeElement()
    }

    appendChild (child) {
      let nativeChild = this._normalizeChild(child);
      if (nativeChild) {
        this.el.appendChild(nativeChild);
      }
      return this
    }

    insertAt (pos, child) {
      let nativeChild = this._normalizeChild(child);
      let childNodes = this.el.childNodes;
      if (pos >= childNodes.length) {
        this.el.appendChild(nativeChild);
      } else {
        this.el.insertBefore(nativeChild, childNodes[pos]);
      }
      return this
    }

    insertBefore (child, before) {
      
      if (isNil$1(before)) {
        return this.appendChild(child)
      }
      if (!before._isBrowserDOMElement) {
        throw new Error('insertBefore(): Illegal arguments. "before" must be a BrowserDOMElement instance.')
      }
      var nativeChild = this._normalizeChild(child);
      if (nativeChild) {
        this.el.insertBefore(nativeChild, before.el);
      }
      return this
    }

    removeAt (pos) {
      this.el.removeChild(this.el.childNodes[pos]);
      return this
    }

    removeChild (child) {
      
      if (!child || !child._isBrowserDOMElement) {
        throw new Error('removeChild(): Illegal arguments. Expecting a BrowserDOMElement instance.')
      }
      this.el.removeChild(child.el);
      return this
    }

    replaceChild (oldChild, newChild) {
      
      if (!newChild || !oldChild ||
          !newChild._isBrowserDOMElement || !oldChild._isBrowserDOMElement) {
        throw new Error('replaceChild(): Illegal arguments. Expecting BrowserDOMElement instances.')
      }
      
      this.el.replaceChild(newChild.el, oldChild.el);
      return this
    }

    empty () {
      let el = this.el;
      while (el.lastChild) {
        el.removeChild(el.lastChild);
      }
      return this
    }

    remove () {
      if (this.el.parentNode) {
        this.el.parentNode.removeChild(this.el);
      }
      return this
    }

    serialize () {
      let outerHTML = this.el.outerHTML;
      if (isString$1(outerHTML)) {
        return outerHTML
      } else {
        let xs = new window.XMLSerializer();
        return xs.serializeToString(this.el)
      }
    }

    isInDocument () {
      let el = this.el;
      while (el) {
        if (el.nodeType === window.Node.DOCUMENT_NODE) {
          return true
        }
        el = el.parentNode;
      }
    }

    _replaceNativeEl (newEl) {
      console.assert(newEl instanceof window.Node, 'Expecting a native element.');
      let oldEl = this.el;
      let parentNode = oldEl.parentNode;
      if (parentNode) {
        parentNode.replaceChild(newEl, oldEl);
      }
      this.el = newEl;
      _detach$1(oldEl);
      _attach$1(newEl, this);
    }

    _getChildNodeCount () {
      return this.el.childNodes.length
    }

    focus () {
      this.el.focus();
      return this
    }

    select () {
      this.el.select();
      return this
    }

    blur () {
      this.el.blur();
      return this
    }

    click () {
      this.el.click();
      return this
    }

    getWidth () {
      let rect = this.el.getClientRects()[0];
      if (rect) {
        return rect.width
      } else {
        return 0
      }
    }

    getHeight () {
      let rect = this.el.getClientRects()[0];
      if (rect) {
        return rect.height
      } else {
        return 0
      }
    }

    getOffset () {
      let rect = this.el.getBoundingClientRect();
      return {
        top: rect.top + document.body.scrollTop,
        left: rect.left + document.body.scrollLeft
      }
    }

    getPosition () {
      return {left: this.el.offsetLeft, top: this.el.offsetTop}
    }

    getOuterHeight (withMargin) {
      let outerHeight = this.el.offsetHeight;
      if (withMargin) {
        let style = this.getComputedStyle();
        outerHeight += parseInt(style.marginTop, 10) + parseInt(style.marginBottom, 10);
      }
      return outerHeight
    }

    getContentType () {
      return this._getNativeOwnerDocument().contentType
    }

    _isXML () {
      return this.getContentType() === 'application/xml'
    }

    emit (name, data) {
      let event;
      if (data) {
        event = new window.CustomEvent(name, {
          detail: data,
          bubbles: true,
          cancelable: true
        });
      } else {
        event = new window.Event(name, {
          bubbles: true,
          cancelable: true
        });
      }
      this.el.dispatchEvent(event);
    }
  }

  BrowserDOMElement$1.prototype._isBrowserDOMElement = true;


  BrowserDOMElement$1.createDocument = function (format) {
    let doc;
    if (format === 'xml') {
      
      doc = window.document.implementation.createDocument(null, 'dummy');
      
      doc.removeChild(doc.firstChild);
    } else {
      doc = (new window.DOMParser()).parseFromString(DOMElement$1.EMPTY_HTML, 'text/html');
    }
    return BrowserDOMElement$1.wrap(doc)
  };

  BrowserDOMElement$1.parseMarkup = function (str, format, options = {}) {
    if (!str) {
      return BrowserDOMElement$1.createDocument(format)
    }
    if (options.snippet) {
      str = `<div id='__snippet__'>${str}</div>`;
    }
    let doc;
    let parser = new window.DOMParser();
    if (format === 'html') {
      doc = BrowserDOMElement$1.wrap(
        _check(
          parser.parseFromString(str, 'text/html')
        )
      );
    } else if (format === 'xml') {
      doc = BrowserDOMElement$1.wrap(
        _check(
          parser.parseFromString(str, 'application/xml')
        )
      );
    }
    if (options.snippet) {
      let childNodes = doc.find('#__snippet__').childNodes;
      if (childNodes.length === 1) {
        return childNodes[0]
      } else {
        return childNodes
      }
    } else {
      return doc
    }

    function _check (doc) {
      if (doc) {
        let parserError = doc.querySelector('parsererror');
        if (parserError) {
          throw new Error('ParserError: ' + parserError)
        }
      }
      return doc
    }
  };

  BrowserDOMElement$1.wrap =
  BrowserDOMElement$1.wrapNativeElement = function (el) {
    if (el) {
      let _el = _unwrap$1(el);
      if (_el) {
        return _el
      } else if (el instanceof window.Node) {
        return new BrowserDOMElement$1(el)
      } else if (el._isBrowserDOMElement) {
        return new BrowserDOMElement$1(el.getNativeElement())
      } else if (el === window) {
        return BrowserDOMElement$1.getBrowserWindow()
      }
    } else {
      return null
    }
  };

  BrowserDOMElement$1.unwrap = function (nativeEl) {
    return _unwrap$1(nativeEl)
  };


  class BrowserWindow$1 {
    constructor () {
      
      this.el = window;
      _attach$1(window, this);
    }
  }

  BrowserWindow$1.prototype.on = BrowserDOMElement$1.prototype.on;
  BrowserWindow$1.prototype.off = BrowserDOMElement$1.prototype.off;
  BrowserWindow$1.prototype.addEventListener = BrowserDOMElement$1.prototype.addEventListener;
  BrowserWindow$1.prototype.removeEventListener = BrowserDOMElement$1.prototype.removeEventListener;
  BrowserWindow$1.prototype._createEventListener = BrowserDOMElement$1.prototype._createEventListener;
  BrowserWindow$1.prototype._addEventListenerNative = BrowserDOMElement$1.prototype._addEventListenerNative;
  BrowserWindow$1.prototype._removeEventListenerNative = BrowserDOMElement$1.prototype._removeEventListenerNative;

  BrowserWindow$1.prototype.getEventListeners = BrowserDOMElement$1.prototype.getEventListeners;

  BrowserDOMElement$1.getBrowserWindow = function () {
    if (window[SIGNATURE$1]) return window[SIGNATURE$1]
    return new BrowserWindow$1(window)
  };

  BrowserDOMElement$1.isReverse = function (anchorNode, anchorOffset, focusNode, focusOffset) {
    
    
    if (focusNode && anchorNode) {
      if (!BrowserDOMElement$1.isReverse._r1) {
        BrowserDOMElement$1.isReverse._r1 = window.document.createRange();
        BrowserDOMElement$1.isReverse._r2 = window.document.createRange();
      }
      const _r1 = BrowserDOMElement$1.isReverse._r1;
      const _r2 = BrowserDOMElement$1.isReverse._r2;
      _r1.setStart(anchorNode.getNativeElement(), anchorOffset);
      _r2.setStart(focusNode.getNativeElement(), focusOffset);
      let cmp = _r1.compareBoundaryPoints(window.Range.START_TO_START, _r2);
      if (cmp === 1) {
        return true
      }
    }
    return false
  };

  BrowserDOMElement$1.getWindowSelection = function () {
    let nativeSel = window.getSelection();
    let result = {
      anchorNode: BrowserDOMElement$1.wrap(nativeSel.anchorNode),
      anchorOffset: nativeSel.anchorOffset,
      focusNode: BrowserDOMElement$1.wrap(nativeSel.focusNode),
      focusOffset: nativeSel.focusOffset
    };
    return result
  };

  function matches$1 (el, selector) {
    let elProto = window.Element.prototype;
    let _matches = (
      elProto.matches || elProto.matchesSelector ||
      elProto.msMatchesSelector || elProto.webkitMatchesSelector
    );
    return _matches.call(el, selector)
  }

  class AttributesMapAdapter$1 {
    constructor (attributes) {
      this.attributes = attributes;
    }

    get size () {
      return this.attributes.length
    }

    get (name) {
      let item = this.attributes.getNamedItem(name);
      if (item) {
        return item.value
      }
    }

    set (name, value) {
      this.attributes.setNamedItem(name, value);
    }

    forEach (fn) {
      const S = this.size;
      for (let i = 0; i < S; i++) {
        const item = this.attributes.item(i);
        fn(item.value, item.name);
      }
    }

    map (fn) {
      let result = [];
      this.forEach((val, key) => { result.push(fn(val, key)); });
      return result
    }

    keys () {
      return this.map((val, key) => { return key })
    }

    values () {
      return this.map((val) => { return val })
    }

    entries () {
      return this.map((val, key) => { return [key, val] })
    }
  }

  class BrowserChildNodeIterator$1 {
    constructor (el) {
      this._next = el.firstChild;
      this._curr = null;
    }

    hasNext () {
      return Boolean(this._next)
    }

    next () {
      let next = this._next;
      this._curr = next;
      this._next = next.nextSibling;
      return BrowserDOMElement$1.wrap(next)
    }

    back () {
      this._next = this._curr;
      this._curr = this._curr.previousSibling;
    }

    peek () {
      return BrowserDOMElement$1.wrap(this._curr)
    }
  }

  var inBrowser$1 = platform$1.inBrowser;

  var domelementtype$1 = {
  	Text: "text", 
  	Directive: "directive", 
  	Comment: "comment", 
  	Script: "script", 
  	Style: "style", 
  	Tag: "tag", 
  	CDATA: "cdata", 
  	Doctype: "doctype",

  	isTag: function(elem){
  		return elem.type === "tag" || elem.type === "script" || elem.type === "style";
  	}
  };

  var boolbase$1 = {
  	trueFunc: function trueFunc(){
  		return true;
  	},
  	falseFunc: function falseFunc(){
  		return false;
  	}
  };

  var parse_1$2 = parse$2;




  var re_nthElement$1 = /^([+\-]?\d*n)?\s*(?:([+\-]?)\s*(\d+))?$/;


  function parse$2(formula){
  	formula = formula.trim().toLowerCase();

  	if(formula === "even"){
  		return [2, 0];
  	} else if(formula === "odd"){
  		return [2, 1];
  	} else {
  		var parsed = formula.match(re_nthElement$1);

  		if(!parsed){
  			throw new SyntaxError("n-th rule couldn't be parsed ('" + formula + "')");
  		}

  		var a;

  		if(parsed[1]){
  			a = parseInt(parsed[1], 10);
  			if(isNaN(a)){
  				if(parsed[1].charAt(0) === "-") a = -1;
  				else a = 1;
  			}
  		} else a = 0;

  		return [
  			a,
  			parsed[3] ? parseInt((parsed[2] || "") + parsed[3], 10) : 0
  		];
  	}
  }

  var compile_1$2 = compile$2;

  var trueFunc$2  = boolbase$1.trueFunc,
      falseFunc$4 = boolbase$1.falseFunc;


  function compile$2(parsed){
  	var a = parsed[0],
  	    b = parsed[1] - 1;

  	
  	
  	if(b < 0 && a <= 0) return falseFunc$4;

  	
  	if(a ===-1) return function(pos){ return pos <= b; };
  	if(a === 0) return function(pos){ return pos === b; };
  	
  	if(a === 1) return b < 0 ? trueFunc$2 : function(pos){ return pos >= b; };

  	
  	var bMod = b % a;
  	if(bMod < 0) bMod += a;

  	if(a > 1){
  		return function(pos){
  			return pos >= b && pos % a === bMod;
  		};
  	}

  	a *= -1; 

  	return function(pos){
  		return pos <= b && pos % a === bMod;
  	};
  }

  var nthCheck$1 = function nthCheck(formula){
  	return compile_1$2(parse_1$2(formula));
  };

  var parse_1$1$1 = parse_1$2;
  var compile_1$1$1 = compile_1$2;
  nthCheck$1.parse = parse_1$1$1;
  nthCheck$1.compile = compile_1$1$1;

  var cssWhat$1 = parse$1$1;

  var re_name$1 = /^(?:\\.|[\w\-\u00c0-\uFFFF])+/,
      re_escape$1 = /\\([\da-f]{1,6}\s?|(\s)|.)/ig,
      
      re_attr$1 = /^\s*((?:\\.|[\w\u00c0-\uFFFF\-])+)\s*(?:(\S?)=\s*(?:(['"])(.*?)\3|(#?(?:\\.|[\w\u00c0-\uFFFF\-])*)|)|)\s*(i)?\]/;

  var actionTypes$1 = {
  	__proto__: null,
  	"undefined": "exists",
  	"":  "equals",
  	"~": "element",
  	"^": "start",
  	"$": "end",
  	"*": "any",
  	"!": "not",
  	"|": "hyphen"
  };

  var simpleSelectors$1 = {
  	__proto__: null,
  	">": "child",
  	"<": "parent",
  	"~": "sibling",
  	"+": "adjacent"
  };

  var attribSelectors$1 = {
  	__proto__: null,
  	"#": ["id", "equals"],
  	".": ["class", "element"]
  };


  var unpackPseudos$1 = {
  	__proto__: null,
  	"has": true,
  	"not": true,
  	"matches": true
  };

  var stripQuotesFromPseudos$1 = {
  	__proto__: null,
  	"contains": true,
  	"icontains": true
  };

  var quotes$1 = {
  	__proto__: null,
  	"\"": true,
  	"'": true
  };


  function funescape$1( _, escaped, escapedWhitespace ) {
  	var high = "0x" + escaped - 0x10000;
  	
  	
  	
  	return high !== high || escapedWhitespace ?
  		escaped :
  		
  		high < 0 ?
  			String.fromCharCode( high + 0x10000 ) :
  			
  			String.fromCharCode( high >> 10 | 0xD800, high & 0x3FF | 0xDC00 );
  }

  function unescapeCSS$1(str){
  	return str.replace(re_escape$1, funescape$1);
  }

  function isWhitespace$1(c){
  	return c === " " || c === "\n" || c === "\t" || c === "\f" || c === "\r";
  }

  function parse$1$1(selector, options){
  	var subselects = [];

  	selector = parseSelector$1(subselects, selector + "", options);

  	if(selector !== ""){
  		throw new SyntaxError("Unmatched selector: " + selector);
  	}

  	return subselects;
  }

  function parseSelector$1(subselects, selector, options){
  	var tokens = [],
  		sawWS = false,
  		data, firstChar, name, quot;

  	function getName(){
  		var sub = selector.match(re_name$1)[0];
  		selector = selector.substr(sub.length);
  		return unescapeCSS$1(sub);
  	}

  	function stripWhitespace(start){
  		while(isWhitespace$1(selector.charAt(start))) start++;
  		selector = selector.substr(start);
  	}

  	stripWhitespace(0);

  	while(selector !== ""){
  		firstChar = selector.charAt(0);

  		if(isWhitespace$1(firstChar)){
  			sawWS = true;
  			stripWhitespace(1);
  		} else if(firstChar in simpleSelectors$1){
  			tokens.push({type: simpleSelectors$1[firstChar]});
  			sawWS = false;

  			stripWhitespace(1);
  		} else if(firstChar === ","){
  			if(tokens.length === 0){
  				throw new SyntaxError("empty sub-selector");
  			}
  			subselects.push(tokens);
  			tokens = [];
  			sawWS = false;
  			stripWhitespace(1);
  		} else {
  			if(sawWS){
  				if(tokens.length > 0){
  					tokens.push({type: "descendant"});
  				}
  				sawWS = false;
  			}

  			if(firstChar === "*"){
  				selector = selector.substr(1);
  				tokens.push({type: "universal"});
  			} else if(firstChar in attribSelectors$1){
  				selector = selector.substr(1);
  				tokens.push({
  					type: "attribute",
  					name: attribSelectors$1[firstChar][0],
  					action: attribSelectors$1[firstChar][1],
  					value: getName(),
  					ignoreCase: false
  				});
  			} else if(firstChar === "["){
  				selector = selector.substr(1);
  				data = selector.match(re_attr$1);
  				if(!data){
  					throw new SyntaxError("Malformed attribute selector: " + selector);
  				}
  				selector = selector.substr(data[0].length);
  				name = unescapeCSS$1(data[1]);

  				if(
  					!options || (
  						"lowerCaseAttributeNames" in options ?
  							options.lowerCaseAttributeNames :
  							!options.xmlMode
  					)
  				){
  					name = name.toLowerCase();
  				}

  				tokens.push({
  					type: "attribute",
  					name: name,
  					action: actionTypes$1[data[2]],
  					value: unescapeCSS$1(data[4] || data[5] || ""),
  					ignoreCase: !!data[6]
  				});

  			} else if(firstChar === ":"){
  				if(selector.charAt(1) === ":"){
  					selector = selector.substr(2);
  					tokens.push({type: "pseudo-element", name: getName().toLowerCase()});
  					continue;
  				}

  				selector = selector.substr(1);

  				name = getName().toLowerCase();
  				data = null;

  				if(selector.charAt(0) === "("){
  					if(name in unpackPseudos$1){
  						quot = selector.charAt(1);
  						var quoted = quot in quotes$1;

  						selector = selector.substr(quoted + 1);

  						data = [];
  						selector = parseSelector$1(data, selector, options);

  						if(quoted){
  							if(selector.charAt(0) !== quot){
  								throw new SyntaxError("unmatched quotes in :" + name);
  							} else {
  								selector = selector.substr(1);
  							}
  						}

  						if(selector.charAt(0) !== ")"){
  							throw new SyntaxError("missing closing parenthesis in :" + name + " " + selector);
  						}

  						selector = selector.substr(1);
  					} else {
  						var pos = 1, counter = 1;

  						for(; counter > 0 && pos < selector.length; pos++){
  							if(selector.charAt(pos) === "(") counter++;
  							else if(selector.charAt(pos) === ")") counter--;
  						}

  						if(counter){
  							throw new SyntaxError("parenthesis not matched");
  						}

  						data = selector.substr(1, pos - 2);
  						selector = selector.substr(pos);

  						if(name in stripQuotesFromPseudos$1){
  							quot = data.charAt(0);

  							if(quot === data.slice(-1) && quot in quotes$1){
  								data = data.slice(1, -1);
  							}

  							data = unescapeCSS$1(data);
  						}
  					}
  				}

  				tokens.push({type: "pseudo", name: name, data: data});
  			} else if(re_name$1.test(selector)){
  				name = getName();

  				if(!options || ("lowerCaseTags" in options ? options.lowerCaseTags : !options.xmlMode)){
  					name = name.toLowerCase();
  				}

  				tokens.push({type: "tag", name: name});
  			} else {
  				if(tokens.length && tokens[tokens.length - 1].type === "descendant"){
  					tokens.pop();
  				}
  				addToken$1(subselects, tokens);
  				return selector;
  			}
  		}
  	}

  	addToken$1(subselects, tokens);

  	return selector;
  }

  function addToken$1(subselects, tokens){
  	if(subselects.length > 0 && tokens.length === 0){
  		throw new SyntaxError("empty sub-selector");
  	}

  	subselects.push(tokens);
  }

  const _encodeXMLContent$1 = ((obj) => {
    let invObj = getInverseObj$2(obj);
    let replacer = getInverseReplacer$2(invObj);
    return getInverse$2(invObj, replacer)
  })({
    amp: '&',
    gt: '>',
    lt: '<'
  });

  const _encodeXMLAttr$1 = ((obj) => {
    let invObj = getInverseObj$2(obj);
    let replacer = getInverseReplacer$2(invObj);
    return getInverse$2(invObj, replacer)
  })({
    quot: '"'
  });

  function getInverseObj$2 (obj) {
    return Object.keys(obj).sort().reduce(function (inverse, name) {
      inverse[obj[name]] = '&' + name + ';';
      return inverse
    }, {})
  }

  function getInverseReplacer$2 (inverse) {
    let single = [];
    let multiple = [];

    Object.keys(inverse).forEach(function (k) {
      if (k.length === 1) {
        single.push('\\' + k);
      } else {
        multiple.push(k);
      }
    });

    multiple.unshift('[' + single.join('') + ']');

    return new RegExp(multiple.join('|'), 'g')
  }

  const RE_NON_ASCII$1 = /[^\0-\x7F]/g;
  const RE_ASTRAL_SYMBOLS$1 = /[\uD800-\uDBFF][\uDC00-\uDFFF]/g;

  function singleCharReplacer$2 (c) {
    return '&#x' + c.charCodeAt(0).toString(16).toUpperCase() + ';'
  }

  function astralReplacer$2 (c) {
    var high = c.charCodeAt(0);
    var low = c.charCodeAt(1);
    var codePoint = (high - 0xD800) * 0x400 + low - 0xDC00 + 0x10000;
    return '&#x' + codePoint.toString(16).toUpperCase() + ';'
  }

  function getInverse$2 (inverse, re) {
    function func (name) {
      return inverse[name]
    }
    return function (data) {
      return data
        .replace(re, func)
        .replace(RE_ASTRAL_SYMBOLS$1, astralReplacer$2)
        .replace(RE_NON_ASCII$1, singleCharReplacer$2)
    }
  }

  const booleanAttributes$1 = {
    __proto__: null,
    allowfullscreen: true,
    async: true,
    autofocus: true,
    autoplay: true,
    checked: true,
    controls: true,
    default: true,
    defer: true,
    disabled: true,
    hidden: true,
    ismap: true,
    loop: true,
    multiple: true,
    muted: true,
    open: true,
    readonly: true,
    required: true,
    reversed: true,
    scoped: true,
    seamless: true,
    selected: true,
    typemustmatch: true
  };

  const unencodedElements$1 = {
    __proto__: null,
    style: true,
    script: true,
    xmp: true,
    iframe: true,
    noembed: true,
    noframes: true,
    plaintext: true,
    noscript: true
  };

  const singleTag$1 = {
    __proto__: null,
    area: true,
    base: true,
    basefont: true,
    br: true,
    col: true,
    command: true,
    embed: true,
    frame: true,
    hr: true,
    img: true,
    input: true,
    isindex: true,
    keygen: true,
    link: true,
    meta: true,
    param: true,
    source: true,
    track: true,
    wbr: true
  };

  class DomUtils$1 {
    isTag (elem) {
      return domelementtype$1.isTag(elem)
    }

    removeElement (elem) {
      if (elem.prev) elem.prev.next = elem.next;
      if (elem.next) elem.next.prev = elem.prev;
      if (elem.parent) {
        var childs = elem.parent.childNodes;
        let pos = childs.lastIndexOf(elem);
        if (pos < 0) throw new Error('Invalid state')
        childs.splice(pos, 1);
        elem.parent = null;
      }
    }

    replaceElement (elem, replacement) {
      if (replacement.parent) this.removeElement(replacement);
      var prev = replacement.prev = elem.prev;
      if (prev) {
        prev.next = replacement;
      }

      var next = replacement.next = elem.next;
      if (next) {
        next.prev = replacement;
      }

      var parent = replacement.parent = elem.parent;
      if (parent) {
        var childs = parent.childNodes;
        let pos = childs.lastIndexOf(elem);
        if (pos < 0) throw new Error('Invalid state')
        childs[pos] = replacement;
      }
    }

    appendChild (elem, child) {
      if (child.parent) this.removeElement(child);
      child.parent = elem;

      if (elem.childNodes.push(child) !== 1) {
        var sibling = elem.childNodes[elem.childNodes.length - 2];
        sibling.next = child;
        child.prev = sibling;
        child.next = null;
      }
    }

    append (elem, next) {
      if (next.parent) this.removeElement(next);
      let parent = elem.parent;
      let currNext = elem.next;

      next.next = currNext;
      next.prev = elem;
      elem.next = next;
      next.parent = parent;

      if (currNext) {
        currNext.prev = next;
        if (parent) {
          var childs = parent.childNodes;
          let pos = childs.lastIndexOf(currNext);
          if (pos < 0) throw new Error('Invalid state')
          childs.splice(pos, 0, next);
        }
      } else if (parent) {
        parent.childNodes.push(next);
      }
    }

    prepend (elem, prev) {
      if (prev.parent) this.removeElement(prev);
      var parent = elem.parent;
      if (parent) {
        var childs = parent.childNodes;
        let pos = childs.lastIndexOf(elem);
        if (pos < 0) throw new Error('Invalid state')
        childs.splice(pos, 0, prev);
      }

      if (elem.prev) {
        elem.prev.next = prev;
      }

      prev.parent = parent;
      prev.prev = elem.prev;
      prev.next = elem;
      elem.prev = prev;
    }

    filter (test, element, recurse, limit) {
      if (!Array.isArray(element)) element = [element];

      if (typeof limit !== 'number' || !isFinite(limit)) {
        limit = Infinity;
      }
      return this.find(test, element, recurse !== false, limit)
    }

    find (test, elems, recurse, limit) {
      let result = [];
      let childs;

      for (var i = 0, j = elems.length; i < j; i++) {
        if (test(elems[i])) {
          result.push(elems[i]);
          if (--limit <= 0) break
        }

        childs = this.getChildren(elems[i]);
        if (recurse && childs && childs.length > 0) {
          childs = this.find(test, childs, recurse, limit);
          result = result.concat(childs);
          limit -= childs.length;
          if (limit <= 0) break
        }
      }

      return result
    }

    findOneChild (test, elems) {
      for (var i = 0, l = elems.length; i < l; i++) {
        if (test(elems[i])) return elems[i]
      }

      return null
    }

    findOne (test, elems) {
      var elem = null;

      for (var i = 0, l = elems.length; i < l && !elem; i++) {
        const child = elems[i];
        if (!this.isTag(child)) {
          continue
        } else if (test(child)) {
          elem = child;
        } else {
          const childNodes = this.getChildren(child);
          if (childNodes.length > 0) {
            elem = this.findOne(test, childNodes);
          }
        }
      }

      return elem
    }

    existsOne (test, elems) {
      for (var i = 0, l = elems.length; i < l; i++) {
        const elem = elems[i];
        
        if (!this.isTag(elem)) continue
        
        if (test(elem)) return true
        
        const childNodes = this.getChildren(elem);
        if (childNodes.length > 0 && this.existsOne(test, childNodes)) return true
      }
      return false
    }

    findAll (test, elems) {
      var result = [];
      for (var i = 0, j = elems.length; i < j; i++) {
        const elem = elems[i];
        if (!this.isTag(elem)) continue
        if (test(elem)) result.push(elem);
        const childNodes = this.getChildren(elem);
        if (childNodes.length > 0) {
          result = result.concat(this.findAll(test, childNodes));
        }
      }
      return result
    }

    getAttributes (el) {
      let attribs = el.getAttributes();
      
      
      if (attribs instanceof Map) {
        return Array.from(attribs)
      } else if (attribs && attribs.forEach) {
        let res = [];
        attribs.forEach((val, key) => {
          res.push([key, val]);
        });
        return res
      } else {
        return []
      }
    }

    formatAttribs (el, opts = {}) {
      let output = [];
      const attributes = this.getAttributes(el);
      attributes.forEach(([key, value]) => {
        if (!value && booleanAttributes$1[key]) {
          output.push(key);
        } else {
          output.push(key + '="' + (opts.decodeEntities ? _encodeXMLAttr$1(value) : value) + '"');
        }
      });
      return output.join(' ')
    }

    render (dom, opts) {
      if (!Array.isArray(dom)) dom = [dom];
      opts = opts || {};
      let output = [];
      for (var i = 0; i < dom.length; i++) {
        let elem = dom[i];
        switch (elem.type) {
          case 'root':
          case 'document': {
            output.push(this.render(this.getChildren(elem), opts));
            break
          }
          case domelementtype$1.Tag:
          case domelementtype$1.Script:
          case domelementtype$1.Style: {
            output.push(this.renderTag(elem, opts));
            break
          }
          case domelementtype$1.CDATA: {
            output.push(this.renderCdata(elem));
            break
          }
          case domelementtype$1.Comment: {
            output.push(this.renderComment(elem));
            break
          }
          case domelementtype$1.Directive: {
            output.push(this.renderDirective(elem));
            break
          }
          case domelementtype$1.Doctype: {
            output.push(this.renderDoctype(elem));
            break
          }
          case domelementtype$1.Text: {
            output.push(this.renderText(elem, opts));
            break
          }
          default:
            throw new Error('Not implement yet: render of element type ' + elem.type)
        }
      }
      return output.join('')
    }

    renderTag (elem, opts) {
      const name = this.getName(elem);
      if (name === 'svg') opts = {decodeEntities: opts.decodeEntities, xmlMode: true};
      let tag = '<' + name;
      let attribs = this.formatAttribs(elem, opts);
      if (attribs) {
        tag += ' ' + attribs;
      }
      const childNodes = this.getChildren(elem);
      if (opts.xmlMode && childNodes.length === 0) {
        tag += '/>';
      } else {
        tag += '>';
        if (childNodes.length > 0) {
          tag += this.render(childNodes, opts);
        }
        if (!singleTag$1[name] || opts.xmlMode) {
          tag += '</' + name + '>';
        }
      }
      return tag
    }

    renderDirective (elem) {
      return '<?' + this.getData(elem) + '?>'
    }

    renderDoctype (elem) {
      const { name, publicId, systemId } = this.getData(elem);
      let frags = ['DOCTYPE', name];
      if (publicId) {
        frags.push('PUBLIC');
        frags.push('"' + publicId + '"');
        if (systemId) frags.push('"' + systemId + '"');
      }
      return '<!' + frags.join(' ') + '>'
    }

    renderText (elem, opts) {
      let text = this.getText(elem);
      if (opts.decodeEntities) {
        const parent = this.getParent(elem);
        if (!(parent && this.getName(parent) in unencodedElements$1)) {
          text = _encodeXMLContent$1(text);
        }
      }
      return text
    }

    renderCdata (elem) {
      return '<![CDATA[' + this.getData(elem) + ']]>'
    }

    renderComment (elem) {
      return '<!--' + this.getData(elem) + '-->'
    }

    getInnerHTML (elem, opts) {
      const childNodes = this.getChildren(elem);
      return childNodes.map((child) => {
        return this.render(child, opts)
      }).join('')
    }

    getOuterHTML (elem, opts) {
      return this.render(elem, opts)
    }

    getData (elem) {
      return elem.data
    }

    getText (elem, sub) {
      if (Array.isArray(elem)) return elem.map(e => this.getText(e, sub)).join('')
      switch (elem.type) {
        case domelementtype$1.Tag:
        case domelementtype$1.Script:
        case domelementtype$1.Style:
          return this.getText(this.getChildren(elem), true)
        case domelementtype$1.Text:
        case domelementtype$1.CDATA:
          return elem.data
        case domelementtype$1.Comment:
          
          
          if (sub) {
            return ''
          }
          return elem.data
        default:
          return ''
      }
    }

    getChildren (elem) {
      return elem.childNodes
    }

    getParent (elem) {
      return elem.parent
    }

    getSiblings (elem) {
      var parent = this.getParent(elem);
      return parent ? this.getChildren(parent) : [elem]
    }

    getAttributeValue (elem, name) {
      return elem.getAttribute(name)
    }

    hasAttrib (elem, name) {
      return elem.hasAttribute(name)
    }

    getName (elem) {
      return elem.name
    }

    getNameWithoutNS (elem) {
      return elem.nameWithoutNS
    }
  }

  const domUtils$1 = new DomUtils$1();
  domUtils$1.DomUtils = DomUtils$1;

  var universal$1 = 50;
  var tag$1 = 30;
  var attribute$1 = 1;
  var pseudo$1 = 0;
  var descendant$1 = -1;
  var child$1 = -1;
  var parent$1$1 = -1;
  var sibling$1 = -1;
  var adjacent$1 = -1;
  var procedure$3 = {
  	universal: universal$1,
  	tag: tag$1,
  	attribute: attribute$1,
  	pseudo: pseudo$1,
  	descendant: descendant$1,
  	child: child$1,
  	parent: parent$1$1,
  	sibling: sibling$1,
  	adjacent: adjacent$1
  };

  var procedure$1$1 = Object.freeze({
    universal: universal$1,
    tag: tag$1,
    attribute: attribute$1,
    pseudo: pseudo$1,
    descendant: descendant$1,
    child: child$1,
    parent: parent$1$1,
    sibling: sibling$1,
    adjacent: adjacent$1,
    default: procedure$3
  });

  var procedure$2$1 = ( procedure$1$1 && procedure$3 ) || procedure$1$1;

  var sort$1 = sortByProcedure$1;





  var attributes$2 = {
  	__proto__: null,
  	exists: 10,
  	equals: 8,
  	not: 7,
  	start: 6,
  	end: 6,
  	any: 5,
  	hyphen: 4,
  	element: 4
  };

  function sortByProcedure$1(arr){
  	var procs = arr.map(getProcedure$1);
  	for(var i = 1; i < arr.length; i++){
  		var procNew = procs[i];

  		if(procNew < 0) continue;

  		for(var j = i - 1; j >= 0 && procNew < procs[j]; j--){
  			var token = arr[j + 1];
  			arr[j + 1] = arr[j];
  			arr[j] = token;
  			procs[j + 1] = procs[j];
  			procs[j] = procNew;
  		}
  	}
  }

  function getProcedure$1(token){
  	var proc = procedure$2$1[token.type];

  	if(proc === procedure$2$1.attribute){
  		proc = attributes$2[token.action];

  		if(proc === attributes$2.equals && token.name === "id"){
  			
  			proc = 9;
  		}

  		if(token.ignoreCase){
  			
  			
  			proc >>= 1;
  		}
  	} else if(proc === procedure$2$1.pseudo){
  		if(!token.data){
  			proc = 3;
  		} else if(token.name === "has" || token.name === "contains"){
  			proc = 0; 
  		} else if(token.name === "matches" || token.name === "not"){
  			proc = 0;
  			for(var i = 0; i < token.data.length; i++){
  				
  				if(token.data[i].length !== 1) continue;
  				var cur = getProcedure$1(token.data[i][0]);
  				
  				if(cur === 0){
  					proc = 0;
  					break;
  				}
  				if(cur > proc) proc = cur;
  			}
  			if(token.data.length > 1 && proc > 0) proc -= 1;
  		} else {
  			proc = 1;
  		}
  	}
  	return proc;
  }

  var falseFunc$1$2 = boolbase$1.falseFunc;


  var reChars$1 = /[-[\]{}()*+?.,\\^$|#\s]/g;

  function factory$2(adapter){
  	
  	var attributeRules = {
  		__proto__: null,
  		equals: function(next, data){
  			var name  = data.name,
  				value = data.value;

  			if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function equalsIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && attr.toLowerCase() === value && next(elem);
  				};
  			}

  			return function equals(elem){
  				return adapter.getAttributeValue(elem, name) === value && next(elem);
  			};
  		},
  		hyphen: function(next, data){
  			var name  = data.name,
  				value = data.value,
  				len = value.length;

  			if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function hyphenIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null &&
  							(attr.length === len || attr.charAt(len) === "-") &&
  							attr.substr(0, len).toLowerCase() === value &&
  							next(elem);
  				};
  			}

  			return function hyphen(elem){
  				var attr = adapter.getAttributeValue(elem, name);
  				return attr != null &&
  						attr.substr(0, len) === value &&
  						(attr.length === len || attr.charAt(len) === "-") &&
  						next(elem);
  			};
  		},
  		element: function(next, data){
  			var name = data.name,
  				value = data.value;
  			if (data.name === 'class') {
  				let value = data.value;
  				if (/\s/.test(value)) return function() { return false }
  				return function clazz(elem) {
  					let classes = elem.classes;
  					return classes && classes.has(value) && next(elem)
  				}
  			} else {
  				if(/\s/.test(value)){
  					return falseFunc$1$2;
  				}

  				value = value.replace(reChars$1, "\\$&");

  				var pattern = "(?:^|\\s)" + value + "(?:$|\\s)",
  					flags = data.ignoreCase ? "i" : "",
  					regex = new RegExp(pattern, flags);

  				return function element(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && regex.test(attr) && next(elem);
  				};
  			}
  		},
  		exists: function(next, data){
  			var name = data.name;
  			return function exists(elem){
  				return adapter.hasAttrib(elem, name) && next(elem);
  			};
  		},
  		start: function(next, data){
  			var name  = data.name,
  				value = data.value,
  				len = value.length;

  			if(len === 0){
  				return falseFunc$1$2;
  			}

  			if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function startIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && attr.substr(0, len).toLowerCase() === value && next(elem);
  				};
  			}

  			return function start(elem){
  				var attr = adapter.getAttributeValue(elem, name);
  				return attr != null && attr.substr(0, len) === value && next(elem);
  			};
  		},
  		end: function(next, data){
  			var name  = data.name,
  				value = data.value,
  				len   = -value.length;

  			if(len === 0){
  				return falseFunc$1$2;
  			}

  			if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function endIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && attr.substr(len).toLowerCase() === value && next(elem);
  				};
  			}

  			return function end(elem){
  				var attr = adapter.getAttributeValue(elem, name);
  				return attr != null && attr.substr(len) === value && next(elem);
  			};
  		},
  		any: function(next, data){
  			var name  = data.name,
  				value = data.value;

  			if(value === ""){
  				return falseFunc$1$2;
  			}

  			if(data.ignoreCase){
  				var regex = new RegExp(value.replace(reChars$1, "\\$&"), "i");

  				return function anyIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && regex.test(attr) && next(elem);
  				};
  			}

  			return function any(elem){
  				var attr = adapter.getAttributeValue(elem, name);
  				return attr != null && attr.indexOf(value) >= 0 && next(elem);
  			};
  		},
  		not: function(next, data){
  			var name  = data.name,
  				value = data.value;

  			if(value === ""){
  				return function notEmpty(elem){
  					return !!adapter.getAttributeValue(elem, name) && next(elem);
  				};
  			} else if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function notIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && attr.toLowerCase() !== value && next(elem);
  				};
  			}

  			return function not(elem){
  				return adapter.getAttributeValue(elem, name) !== value && next(elem);
  			};
  		}
  	};

  	return {
  		compile: function(next, data, options){
  			if(options && options.strict && (
  				data.ignoreCase || data.action === "not"
  			)) throw new Error("Unsupported attribute selector");
  			return attributeRules[data.action](next, data);
  		},
  		rules: attributeRules
  	};
  }

  var attributes$1$1 = factory$2;

  function generalFactory$1(adapter, Pseudos){
  	
  	return {
  		__proto__: null,

  		attribute: attributes$1$1(adapter).compile,
  		pseudo: Pseudos.compile,

  		
  		tag: function(next, data){
  			var name = data.name;
  			return function tag(elem){
  				return adapter.getNameWithoutNS(elem) === name && next(elem);
  			}
  		},

  		
  		descendant: function(next){
  			return function descendant(elem){

  				var found = false;

  				while(!found && (elem = adapter.getParent(elem))){
  					found = next(elem);
  				}

  				return found;
  			};
  		},
  		_flexibleDescendant: function(next){
  			
  			return function descendant(elem){

  				var found = next(elem);

  				while(!found && (elem = adapter.getParent(elem))){
  					found = next(elem);
  				}

  				return found;
  			};
  		},
  		parent: function(next, data, options){
  			if(options && options.strict) throw new Error("Parent selector isn't part of CSS3");

  			return function parent(elem){
  				return adapter.getChildren(elem).some(test);
  			};

  			function test(elem){
  				return adapter.isTag(elem) && next(elem);
  			}
  		},
  		child: function(next){
  			return function child(elem){
  				var parent = adapter.getParent(elem);
  				return !!parent && next(parent);
  			};
  		},
  		sibling: function(next){
  			return function sibling(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var i = 0; i < siblings.length; i++){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) break;
  						if(next(siblings[i])) return true;
  					}
  				}

  				return false;
  			};
  		},
  		adjacent: function(next){
  			return function adjacent(elem){
  				var siblings = adapter.getSiblings(elem),
  					lastElement;

  				for(var i = 0; i < siblings.length; i++){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) break;
  						lastElement = siblings[i];
  					}
  				}

  				return !!lastElement && next(lastElement);
  			};
  		},
  		universal: function(next){
  			return next;
  		}
  	};
  }

  var general$1 = generalFactory$1;

  var trueFunc$1$2          = boolbase$1.trueFunc,
  	falseFunc$1$1$1         = boolbase$1.falseFunc;

  function filtersFactory$1(adapter){
  	var attributes  = attributes$1$1(adapter),
  		checkAttrib = attributes.rules.equals;

  	
  	function equals(a, b){
  		if(typeof adapter.equals === "function") return adapter.equals(a, b);

  		return a === b;
  	}

  	function getAttribFunc(name, value){
  		var data = {name: name, value: value};
  		return function attribFunc(next){
  			return checkAttrib(next, data);
  		};
  	}

  	function getChildFunc(next){
  		return function(elem){
  			return !!adapter.getParent(elem) && next(elem);
  		};
  	}

  	var filters = {
  		contains: function(next, text){
  			return function contains(elem){
  				return next(elem) && adapter.getText(elem).indexOf(text) >= 0;
  			};
  		},
  		icontains: function(next, text){
  			var itext = text.toLowerCase();
  			return function icontains(elem){
  				return next(elem) &&
  					adapter.getText(elem).toLowerCase().indexOf(itext) >= 0;
  			};
  		},

  		
  		"nth-child": function(next, rule){
  			var func = nthCheck$1(rule);

  			if(func === falseFunc$1$1$1) return func;
  			if(func === trueFunc$1$2)  return getChildFunc(next);

  			return function nthChild(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var i = 0, pos = 0; i < siblings.length; i++){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) break;
  						else pos++;
  					}
  				}

  				return func(pos) && next(elem);
  			};
  		},
  		"nth-last-child": function(next, rule){
  			var func = nthCheck$1(rule);

  			if(func === falseFunc$1$1$1) return func;
  			if(func === trueFunc$1$2)  return getChildFunc(next);

  			return function nthLastChild(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var pos = 0, i = siblings.length - 1; i >= 0; i--){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) break;
  						else pos++;
  					}
  				}

  				return func(pos) && next(elem);
  			};
  		},
  		"nth-of-type": function(next, rule){
  			var func = nthCheck$1(rule);

  			if(func === falseFunc$1$1$1) return func;
  			if(func === trueFunc$1$2)  return getChildFunc(next);

  			return function nthOfType(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var pos = 0, i = 0; i < siblings.length; i++){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) break;
  						if(adapter.getName(siblings[i]) === adapter.getName(elem)) pos++;
  					}
  				}

  				return func(pos) && next(elem);
  			};
  		},
  		"nth-last-of-type": function(next, rule){
  			var func = nthCheck$1(rule);

  			if(func === falseFunc$1$1$1) return func;
  			if(func === trueFunc$1$2)  return getChildFunc(next);

  			return function nthLastOfType(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var pos = 0, i = siblings.length - 1; i >= 0; i--){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) break;
  						if(adapter.getName(siblings[i]) === adapter.getName(elem)) pos++;
  					}
  				}

  				return func(pos) && next(elem);
  			};
  		},

  		
  		root: function(next){
  			return function(elem){
  				return !adapter.getParent(elem) && next(elem);
  			};
  		},

  		scope: function(next, rule, options, context){
  			if(!context || context.length === 0){
  				
  				return filters.root(next);
  			}

  			if(context.length === 1){
  				
  				return function(elem){
  					return equals(context[0], elem) && next(elem);
  				};
  			}

  			return function(elem){
  				return context.indexOf(elem) >= 0 && next(elem);
  			};
  		},

  		
  		checkbox: getAttribFunc("type", "checkbox"),
  		file: getAttribFunc("type", "file"),
  		password: getAttribFunc("type", "password"),
  		radio: getAttribFunc("type", "radio"),
  		reset: getAttribFunc("type", "reset"),
  		image: getAttribFunc("type", "image"),
  		submit: getAttribFunc("type", "submit")
  	};
  	return filters;
  }

  function pseudosFactory$1(adapter){
  	
  	function getFirstElement(elems){
  		for(var i = 0; elems && i < elems.length; i++){
  			if(adapter.isTag(elems[i])) return elems[i];
  		}
  	}

  	
  	var pseudos = {
  		empty: function(elem){
  			return !adapter.getChildren(elem).some(function(elem){
  				return adapter.isTag(elem) || elem.type === "text";
  			});
  		},

  		"first-child": function(elem){
  			return getFirstElement(adapter.getSiblings(elem)) === elem;
  		},
  		"last-child": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = siblings.length - 1; i >= 0; i--){
  				if(siblings[i] === elem) return true;
  				if(adapter.isTag(siblings[i])) break;
  			}

  			return false;
  		},
  		"first-of-type": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = 0; i < siblings.length; i++){
  				if(adapter.isTag(siblings[i])){
  					if(siblings[i] === elem) return true;
  					if(adapter.getName(siblings[i]) === adapter.getName(elem)) break;
  				}
  			}

  			return false;
  		},
  		"last-of-type": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = siblings.length - 1; i >= 0; i--){
  				if(adapter.isTag(siblings[i])){
  					if(siblings[i] === elem) return true;
  					if(adapter.getName(siblings[i]) === adapter.getName(elem)) break;
  				}
  			}

  			return false;
  		},
  		"only-of-type": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = 0, j = siblings.length; i < j; i++){
  				if(adapter.isTag(siblings[i])){
  					if(siblings[i] === elem) continue;
  					if(adapter.getName(siblings[i]) === adapter.getName(elem)) return false;
  				}
  			}

  			return true;
  		},
  		"only-child": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = 0; i < siblings.length; i++){
  				if(adapter.isTag(siblings[i]) && siblings[i] !== elem) return false;
  			}

  			return true;
  		},

  		
  		link: function(elem){
  			return adapter.hasAttrib(elem, "href");
  		},
  		visited: falseFunc$1$1$1, 
  		

  		
  		

  		
  		selected: function(elem){
  			if(adapter.hasAttrib(elem, "selected")) return true;
  			else if(adapter.getName(elem) !== "option") return false;

  			
  			var parent = adapter.getParent(elem);

  			if(
  				!parent ||
  				adapter.getName(parent) !== "select" ||
  				adapter.hasAttrib(parent, "multiple")
  			) return false;

  			var siblings = adapter.getChildren(parent),
  				sawElem  = false;

  			for(var i = 0; i < siblings.length; i++){
  				if(adapter.isTag(siblings[i])){
  					if(siblings[i] === elem){
  						sawElem = true;
  					} else if(!sawElem){
  						return false;
  					} else if(adapter.hasAttrib(siblings[i], "selected")){
  						return false;
  					}
  				}
  			}

  			return sawElem;
  		},
  		
  		
  		
  		
  		
  		
  		disabled: function(elem){
  			return adapter.hasAttrib(elem, "disabled");
  		},
  		enabled: function(elem){
  			return !adapter.hasAttrib(elem, "disabled");
  		},
  		
  		checked: function(elem){
  			return adapter.hasAttrib(elem, "checked") || pseudos.selected(elem);
  		},
  		
  		required: function(elem){
  			return adapter.hasAttrib(elem, "required");
  		},
  		
  		optional: function(elem){
  			return !adapter.hasAttrib(elem, "required");
  		},

  		

  		
  		parent: function(elem){
  			return !pseudos.empty(elem);
  		},
  		
  		header: function(elem){
  			var name = adapter.getName(elem);
  			return name === "h1" ||
  					name === "h2" ||
  					name === "h3" ||
  					name === "h4" ||
  					name === "h5" ||
  					name === "h6";
  		},

  		
  		button: function(elem){
  			var name = adapter.getName(elem);
  			return name === "button" ||
  					name === "input" &&
  					adapter.getAttributeValue(elem, "type") === "button";
  		},
  		
  		input: function(elem){
  			var name = adapter.getName(elem);
  			return name === "input" ||
  					name === "textarea" ||
  					name === "select" ||
  					name === "button";
  		},
  		
  		text: function(elem){
  			var attr;
  			return adapter.getName(elem) === "input" && (
  				!(attr = adapter.getAttributeValue(elem, "type")) ||
  				attr.toLowerCase() === "text"
  			);
  		}
  	};

  	return pseudos;
  }

  function verifyArgs$1(func, name, subselect){
  	if(subselect === null){
  		if(func.length > 1 && name !== "scope"){
  			throw new Error("pseudo-selector :" + name + " requires an argument");
  		}
  	} else {
  		if(func.length === 1){
  			throw new Error("pseudo-selector :" + name + " doesn't have any arguments");
  		}
  	}
  }


  var re_CSS3$1 = /^(?:(?:nth|last|first|only)-(?:child|of-type)|root|empty|(?:en|dis)abled|checked|not)$/;

  function factory$1$1(adapter){
  	var pseudos = pseudosFactory$1(adapter);
  	var filters = filtersFactory$1(adapter);

  	return {
  		compile: function(next, data, options, context){
  			var name = data.name,
  				subselect = data.data;

  			if(options && options.strict && !re_CSS3$1.test(name)){
  				throw new Error(":" + name + " isn't part of CSS3");
  			}

  			if(typeof filters[name] === "function"){
  				verifyArgs$1(filters[name], name,  subselect);
  				return filters[name](next, subselect, options, context);
  			} else if(typeof pseudos[name] === "function"){
  				var func = pseudos[name];
  				verifyArgs$1(func, name, subselect);

  				if(next === trueFunc$1$2) return func;

  				return function pseudoArgs(elem){
  					return func(elem, subselect) && next(elem);
  				};
  			} else {
  				throw new Error("unmatched pseudo-class :" + name);
  			}
  		},
  		filters: filters,
  		pseudos: pseudos
  	};
  }

  var pseudos$1 = factory$1$1;

  var compile$1$1 = compileFactory$1;

  var trueFunc$1$1$1       = boolbase$1.trueFunc,
  	falseFunc$2$1      = boolbase$1.falseFunc;

  function compileFactory$1(adapter){
  	var Pseudos     = pseudos$1(adapter),
  		filters     = Pseudos.filters,
  		Rules 			= general$1(adapter, Pseudos);

  	function compile(selector, options, context){
  		var next = compileUnsafe(selector, options, context);
  		return wrap(next);
  	}

  	function wrap(next){
  		return function base(elem){
  			return adapter.isTag(elem) && next(elem);
  		};
  	}

  	function compileUnsafe(selector, options, context){
  		var token = cssWhat$1(selector, options);
  		return compileToken(token, options, context);
  	}

  	function includesScopePseudo(t){
  		return t.type === "pseudo" && (
  			t.name === "scope" || (
  				Array.isArray(t.data) &&
  				t.data.some(function(data){
  					return data.some(includesScopePseudo);
  				})
  			)
  		);
  	}

  	var DESCENDANT_TOKEN = {type: "descendant"},
  		FLEXIBLE_DESCENDANT_TOKEN = {type: "_flexibleDescendant"},
  		SCOPE_TOKEN = {type: "pseudo", name: "scope"},
  		PLACEHOLDER_ELEMENT = {};

  	
  	
  	function absolutize(token, context){
  		
  		var hasContext = !!context && !!context.length && context.every(function(e){
  			return e === PLACEHOLDER_ELEMENT || !!adapter.getParent(e);
  		});


  		token.forEach(function(t){
  			if(t.length > 0 && isTraversal(t[0]) && t[0].type !== "descendant"); else if(hasContext && !includesScopePseudo(t)){
  				t.unshift(DESCENDANT_TOKEN);
  			} else {
  				return;
  			}

  			t.unshift(SCOPE_TOKEN);
  		});
  	}

  	function compileToken(token, options, context){
  		token = token.filter(function(t){ return t.length > 0; });

  		token.forEach(sort$1);

  		var isArrayContext = Array.isArray(context);

  		context = (options && options.context) || context;

  		if(context && !isArrayContext) context = [context];

  		absolutize(token, context);

  		var shouldTestNextSiblings = false;

  		var query = token
  			.map(function(rules){
  				if(rules[0] && rules[1] && rules[0].name === "scope"){
  					var ruleType = rules[1].type;
  					if(isArrayContext && ruleType === "descendant") rules[1] = FLEXIBLE_DESCENDANT_TOKEN;
  					else if(ruleType === "adjacent" || ruleType === "sibling") shouldTestNextSiblings = true;
  				}
  				return compileRules(rules, options, context);
  			})
  			.reduce(reduceRules, falseFunc$2$1);

  		query.shouldTestNextSiblings = shouldTestNextSiblings;

  		return query;
  	}

  	function isTraversal(t){
  		return procedure$2$1[t.type] < 0;
  	}

  	function compileRules(rules, options, context){
  		return rules.reduce(function(func, rule){
  			if(func === falseFunc$2$1) return func;
  			return Rules[rule.type](func, rule, options, context);
  		}, options && options.rootFunc || trueFunc$1$1$1);
  	}

  	function reduceRules(a, b){
  		if(b === falseFunc$2$1 || a === trueFunc$1$1$1){
  			return a;
  		}
  		if(a === falseFunc$2$1 || b === trueFunc$1$1$1){
  			return b;
  		}

  		return function combine(elem){
  			return a(elem) || b(elem);
  		};
  	}

  	function containsTraversal(t){
  		return t.some(isTraversal);
  	}

  	
  	
  	
  	filters.not = function(next, token, options, context){
  		var opts = {
  			xmlMode: !!(options && options.xmlMode),
  			strict: !!(options && options.strict)
  		};

  		if(opts.strict){
  			if(token.length > 1 || token.some(containsTraversal)){
  				throw new Error("complex selectors in :not aren't allowed in strict mode");
  			}
  		}

  		var func = compileToken(token, opts, context);

  		if(func === falseFunc$2$1) return next;
  		if(func === trueFunc$1$1$1)  return falseFunc$2$1;

  		return function(elem){
  			return !func(elem) && next(elem);
  		};
  	};

  	filters.has = function(next, token, options){
  		var opts = {
  			xmlMode: !!(options && options.xmlMode),
  			strict: !!(options && options.strict)
  		};

  		
  		var context = token.some(containsTraversal) ? [PLACEHOLDER_ELEMENT] : null;

  		var func = compileToken(token, opts, context);

  		if(func === falseFunc$2$1) return falseFunc$2$1;
  		if(func === trueFunc$1$1$1){
  			return function(elem){
  				return adapter.getChildren(elem).some(adapter.isTag) && next(elem);
  			};
  		}

  		func = wrap(func);

  		if(context){
  			return function has(elem){
  				return next(elem) && (
  					(context[0] = elem), adapter.existsOne(func, adapter.getChildren(elem))
  				);
  			};
  		}

  		return function has(elem){
  			return next(elem) && adapter.existsOne(func, adapter.getChildren(elem));
  		};
  	};

  	filters.matches = function(next, token, options, context){
  		var opts = {
  			xmlMode: !!(options && options.xmlMode),
  			strict: !!(options && options.strict),
  			rootFunc: next
  		};

  		return compileToken(token, opts, context);
  	};

  	compile.compileToken = compileToken;
  	compile.compileUnsafe = compileUnsafe;
  	compile.Pseudos = Pseudos;

  	return compile;
  }

  var cssSelect$1 = CSSselect$1;

  var falseFunc$3$1      = boolbase$1.falseFunc,
  	defaultCompile$1 = compile$1$1(domUtils$1);

  function adapterCompile$1(adapter){
  	if(!adapter.__compile__){
  		adapter.__compile__ = compile$1$1(adapter);
  	}
  	return adapter.__compile__
  }

  function getSelectorFunc$1(searchFunc){
  	return function select(query, elems, options){
  		options = options || {};
  		options.adapter = options.adapter || domUtils$1;
  		var compile$$1 = adapterCompile$1(options.adapter);

  		if(typeof query !== "function") query = compile$$1.compileUnsafe(query, options, elems);
  		if(query.shouldTestNextSiblings) elems = appendNextSiblings$1((options && options.context) || elems, options.adapter);
  		if(!Array.isArray(elems)) elems = options.adapter.getChildren(elems);
  		else elems = options.adapter.removeSubsets(elems);
  		return searchFunc(query, elems, options);
  	};
  }

  function getNextSiblings$1(elem, adapter){
  	var siblings = adapter.getSiblings(elem);
  	if(!Array.isArray(siblings)) return [];
  	siblings = siblings.slice(0);
  	while(siblings.shift() !== elem);
  	return siblings;
  }

  function appendNextSiblings$1(elems, adapter){
  	
  	if(!Array.isArray(elems)) elems = [elems];
  	var newElems = elems.slice(0);

  	for(var i = 0, len = elems.length; i < len; i++){
  		var nextSiblings = getNextSiblings$1(newElems[i], adapter);
  		newElems.push.apply(newElems, nextSiblings);
  	}
  	return newElems;
  }

  var selectAll$1 = getSelectorFunc$1(function selectAll(query, elems, options){
  	return (query === falseFunc$3$1 || !elems || elems.length === 0) ? [] : options.adapter.findAll(query, elems);
  });

  var selectOne$1 = getSelectorFunc$1(function selectOne(query, elems, options){
  	return (query === falseFunc$3$1 || !elems || elems.length === 0) ? null : options.adapter.findOne(query, elems);
  });

  function is$1(elem, query, options){
  	options = options || {};
  	options.adapter = options.adapter || domUtils$1;
  	var compile$$1 = adapterCompile$1(options.adapter);
  	return (typeof query === "function" ? query : compile$$1(query, options))(elem);
  }


  function CSSselect$1(query, elems, options){
  	return selectAll$1(query, elems, options);
  }

  CSSselect$1.compile = defaultCompile$1;
  CSSselect$1.filters = defaultCompile$1.Pseudos.filters;
  CSSselect$1.pseudos = defaultCompile$1.Pseudos.pseudos;

  CSSselect$1.selectAll = selectAll$1;
  CSSselect$1.selectOne = selectOne$1;

  CSSselect$1.is = is$1;


  CSSselect$1.parse = defaultCompile$1;
  CSSselect$1.iterate = selectAll$1;


  CSSselect$1._compileUnsafe = defaultCompile$1.compileUnsafe;
  CSSselect$1._compileToken = defaultCompile$1.compileToken;

  var amp$3 = "&";
  var apos$2 = "'";
  var gt$3 = ">";
  var lt$3 = "<";
  var quot$3 = "\"";
  var xmlJSON$1 = {
  	amp: amp$3,
  	apos: apos$2,
  	gt: gt$3,
  	lt: lt$3,
  	quot: quot$3
  };

  var xml$1 = Object.freeze({
  	amp: amp$3,
  	apos: apos$2,
  	gt: gt$3,
  	lt: lt$3,
  	quot: quot$3,
  	default: xmlJSON$1
  });

  var Aacute$2 = "Á";
  var aacute$2 = "á";
  var Abreve$1 = "Ă";
  var abreve$1 = "ă";
  var ac$1 = "∾";
  var acd$1 = "∿";
  var acE$1 = "∾̳";
  var Acirc$2 = "Â";
  var acirc$2 = "â";
  var acute$2 = "´";
  var Acy$1 = "А";
  var acy$1 = "а";
  var AElig$2 = "Æ";
  var aelig$2 = "æ";
  var af$1 = "⁡";
  var Afr$1 = "𝔄";
  var afr$1 = "𝔞";
  var Agrave$2 = "À";
  var agrave$2 = "à";
  var alefsym$1 = "ℵ";
  var aleph$1 = "ℵ";
  var Alpha$1 = "Α";
  var alpha$1 = "α";
  var Amacr$1 = "Ā";
  var amacr$1 = "ā";
  var amalg$1 = "⨿";
  var amp$1$1 = "&";
  var AMP$2 = "&";
  var andand$1 = "⩕";
  var And$1 = "⩓";
  var and$1 = "∧";
  var andd$1 = "⩜";
  var andslope$1 = "⩘";
  var andv$1 = "⩚";
  var ang$1 = "∠";
  var ange$1 = "⦤";
  var angle$1 = "∠";
  var angmsdaa$1 = "⦨";
  var angmsdab$1 = "⦩";
  var angmsdac$1 = "⦪";
  var angmsdad$1 = "⦫";
  var angmsdae$1 = "⦬";
  var angmsdaf$1 = "⦭";
  var angmsdag$1 = "⦮";
  var angmsdah$1 = "⦯";
  var angmsd$1 = "∡";
  var angrt$1 = "∟";
  var angrtvb$1 = "⊾";
  var angrtvbd$1 = "⦝";
  var angsph$1 = "∢";
  var angst$1 = "Å";
  var angzarr$1 = "⍼";
  var Aogon$1 = "Ą";
  var aogon$1 = "ą";
  var Aopf$1 = "𝔸";
  var aopf$1 = "𝕒";
  var apacir$1 = "⩯";
  var ap$1 = "≈";
  var apE$1 = "⩰";
  var ape$1 = "≊";
  var apid$1 = "≋";
  var apos$1$1 = "'";
  var ApplyFunction$1 = "⁡";
  var approx$1 = "≈";
  var approxeq$1 = "≊";
  var Aring$2 = "Å";
  var aring$2 = "å";
  var Ascr$1 = "𝒜";
  var ascr$1 = "𝒶";
  var Assign$1 = "≔";
  var ast$1 = "*";
  var asymp$1 = "≈";
  var asympeq$1 = "≍";
  var Atilde$2 = "Ã";
  var atilde$2 = "ã";
  var Auml$2 = "Ä";
  var auml$2 = "ä";
  var awconint$1 = "∳";
  var awint$1 = "⨑";
  var backcong$1 = "≌";
  var backepsilon$1 = "϶";
  var backprime$1 = "‵";
  var backsim$1 = "∽";
  var backsimeq$1 = "⋍";
  var Backslash$1 = "∖";
  var Barv$1 = "⫧";
  var barvee$1 = "⊽";
  var barwed$1 = "⌅";
  var Barwed$1 = "⌆";
  var barwedge$1 = "⌅";
  var bbrk$1 = "⎵";
  var bbrktbrk$1 = "⎶";
  var bcong$1 = "≌";
  var Bcy$1 = "Б";
  var bcy$1 = "б";
  var bdquo$1 = "„";
  var becaus$1 = "∵";
  var because$1 = "∵";
  var Because$1 = "∵";
  var bemptyv$1 = "⦰";
  var bepsi$1 = "϶";
  var bernou$1 = "ℬ";
  var Bernoullis$1 = "ℬ";
  var Beta$1 = "Β";
  var beta$1 = "β";
  var beth$1 = "ℶ";
  var between$1 = "≬";
  var Bfr$1 = "𝔅";
  var bfr$1 = "𝔟";
  var bigcap$1 = "⋂";
  var bigcirc$1 = "◯";
  var bigcup$1 = "⋃";
  var bigodot$1 = "⨀";
  var bigoplus$1 = "⨁";
  var bigotimes$1 = "⨂";
  var bigsqcup$1 = "⨆";
  var bigstar$1 = "★";
  var bigtriangledown$1 = "▽";
  var bigtriangleup$1 = "△";
  var biguplus$1 = "⨄";
  var bigvee$1 = "⋁";
  var bigwedge$1 = "⋀";
  var bkarow$1 = "⤍";
  var blacklozenge$1 = "⧫";
  var blacksquare$1 = "▪";
  var blacktriangle$1 = "▴";
  var blacktriangledown$1 = "▾";
  var blacktriangleleft$1 = "◂";
  var blacktriangleright$1 = "▸";
  var blank$1 = "␣";
  var blk12$1 = "▒";
  var blk14$1 = "░";
  var blk34$1 = "▓";
  var block$1 = "█";
  var bne$1 = "=⃥";
  var bnequiv$1 = "≡⃥";
  var bNot$1 = "⫭";
  var bnot$1 = "⌐";
  var Bopf$1 = "𝔹";
  var bopf$1 = "𝕓";
  var bot$1 = "⊥";
  var bottom$1 = "⊥";
  var bowtie$1 = "⋈";
  var boxbox$1 = "⧉";
  var boxdl$1 = "┐";
  var boxdL$1 = "╕";
  var boxDl$1 = "╖";
  var boxDL$1 = "╗";
  var boxdr$1 = "┌";
  var boxdR$1 = "╒";
  var boxDr$1 = "╓";
  var boxDR$1 = "╔";
  var boxh$1 = "─";
  var boxH$1 = "═";
  var boxhd$1 = "┬";
  var boxHd$1 = "╤";
  var boxhD$1 = "╥";
  var boxHD$1 = "╦";
  var boxhu$1 = "┴";
  var boxHu$1 = "╧";
  var boxhU$1 = "╨";
  var boxHU$1 = "╩";
  var boxminus$1 = "⊟";
  var boxplus$1 = "⊞";
  var boxtimes$1 = "⊠";
  var boxul$1 = "┘";
  var boxuL$1 = "╛";
  var boxUl$1 = "╜";
  var boxUL$1 = "╝";
  var boxur$1 = "└";
  var boxuR$1 = "╘";
  var boxUr$1 = "╙";
  var boxUR$1 = "╚";
  var boxv$1 = "│";
  var boxV$1 = "║";
  var boxvh$1 = "┼";
  var boxvH$1 = "╪";
  var boxVh$1 = "╫";
  var boxVH$1 = "╬";
  var boxvl$1 = "┤";
  var boxvL$1 = "╡";
  var boxVl$1 = "╢";
  var boxVL$1 = "╣";
  var boxvr$1 = "├";
  var boxvR$1 = "╞";
  var boxVr$1 = "╟";
  var boxVR$1 = "╠";
  var bprime$1 = "‵";
  var breve$1 = "˘";
  var Breve$1 = "˘";
  var brvbar$2 = "¦";
  var bscr$1 = "𝒷";
  var Bscr$1 = "ℬ";
  var bsemi$1 = "⁏";
  var bsim$1 = "∽";
  var bsime$1 = "⋍";
  var bsolb$1 = "⧅";
  var bsol$1 = "\\";
  var bsolhsub$1 = "⟈";
  var bull$1 = "•";
  var bullet$1 = "•";
  var bump$1 = "≎";
  var bumpE$1 = "⪮";
  var bumpe$1 = "≏";
  var Bumpeq$1 = "≎";
  var bumpeq$1 = "≏";
  var Cacute$1 = "Ć";
  var cacute$1 = "ć";
  var capand$1 = "⩄";
  var capbrcup$1 = "⩉";
  var capcap$1 = "⩋";
  var cap$1 = "∩";
  var Cap$1 = "⋒";
  var capcup$1 = "⩇";
  var capdot$1 = "⩀";
  var CapitalDifferentialD$1 = "ⅅ";
  var caps$1 = "∩︀";
  var caret$1 = "⁁";
  var caron$1 = "ˇ";
  var Cayleys$1 = "ℭ";
  var ccaps$1 = "⩍";
  var Ccaron$1 = "Č";
  var ccaron$1 = "č";
  var Ccedil$2 = "Ç";
  var ccedil$2 = "ç";
  var Ccirc$1 = "Ĉ";
  var ccirc$1 = "ĉ";
  var Cconint$1 = "∰";
  var ccups$1 = "⩌";
  var ccupssm$1 = "⩐";
  var Cdot$1 = "Ċ";
  var cdot$1 = "ċ";
  var cedil$2 = "¸";
  var Cedilla$1 = "¸";
  var cemptyv$1 = "⦲";
  var cent$2 = "¢";
  var centerdot$1 = "·";
  var CenterDot$1 = "·";
  var cfr$1 = "𝔠";
  var Cfr$1 = "ℭ";
  var CHcy$1 = "Ч";
  var chcy$1 = "ч";
  var check$1 = "✓";
  var checkmark$1 = "✓";
  var Chi$1 = "Χ";
  var chi$1 = "χ";
  var circ$1 = "ˆ";
  var circeq$1 = "≗";
  var circlearrowleft$1 = "↺";
  var circlearrowright$1 = "↻";
  var circledast$1 = "⊛";
  var circledcirc$1 = "⊚";
  var circleddash$1 = "⊝";
  var CircleDot$1 = "⊙";
  var circledR$1 = "®";
  var circledS$1 = "Ⓢ";
  var CircleMinus$1 = "⊖";
  var CirclePlus$1 = "⊕";
  var CircleTimes$1 = "⊗";
  var cir$1 = "○";
  var cirE$1 = "⧃";
  var cire$1 = "≗";
  var cirfnint$1 = "⨐";
  var cirmid$1 = "⫯";
  var cirscir$1 = "⧂";
  var ClockwiseContourIntegral$1 = "∲";
  var CloseCurlyDoubleQuote$1 = "”";
  var CloseCurlyQuote$1 = "’";
  var clubs$1 = "♣";
  var clubsuit$1 = "♣";
  var colon$1 = ":";
  var Colon$1 = "∷";
  var Colone$1 = "⩴";
  var colone$1 = "≔";
  var coloneq$1 = "≔";
  var comma$1 = ",";
  var commat$1 = "@";
  var comp$1 = "∁";
  var compfn$1 = "∘";
  var complement$1 = "∁";
  var complexes$1 = "ℂ";
  var cong$1 = "≅";
  var congdot$1 = "⩭";
  var Congruent$1 = "≡";
  var conint$1 = "∮";
  var Conint$1 = "∯";
  var ContourIntegral$1 = "∮";
  var copf$1 = "𝕔";
  var Copf$1 = "ℂ";
  var coprod$1 = "∐";
  var Coproduct$1 = "∐";
  var copy$2 = "©";
  var COPY$2 = "©";
  var copysr$1 = "℗";
  var CounterClockwiseContourIntegral$1 = "∳";
  var crarr$1 = "↵";
  var cross$1 = "✗";
  var Cross$1 = "⨯";
  var Cscr$1 = "𝒞";
  var cscr$1 = "𝒸";
  var csub$1 = "⫏";
  var csube$1 = "⫑";
  var csup$1 = "⫐";
  var csupe$1 = "⫒";
  var ctdot$1 = "⋯";
  var cudarrl$1 = "⤸";
  var cudarrr$1 = "⤵";
  var cuepr$1 = "⋞";
  var cuesc$1 = "⋟";
  var cularr$1 = "↶";
  var cularrp$1 = "⤽";
  var cupbrcap$1 = "⩈";
  var cupcap$1 = "⩆";
  var CupCap$1 = "≍";
  var cup$1 = "∪";
  var Cup$1 = "⋓";
  var cupcup$1 = "⩊";
  var cupdot$1 = "⊍";
  var cupor$1 = "⩅";
  var cups$1 = "∪︀";
  var curarr$1 = "↷";
  var curarrm$1 = "⤼";
  var curlyeqprec$1 = "⋞";
  var curlyeqsucc$1 = "⋟";
  var curlyvee$1 = "⋎";
  var curlywedge$1 = "⋏";
  var curren$2 = "¤";
  var curvearrowleft$1 = "↶";
  var curvearrowright$1 = "↷";
  var cuvee$1 = "⋎";
  var cuwed$1 = "⋏";
  var cwconint$1 = "∲";
  var cwint$1 = "∱";
  var cylcty$1 = "⌭";
  var dagger$1 = "†";
  var Dagger$1 = "‡";
  var daleth$1 = "ℸ";
  var darr$1 = "↓";
  var Darr$1 = "↡";
  var dArr$1 = "⇓";
  var dash$1 = "‐";
  var Dashv$1 = "⫤";
  var dashv$1 = "⊣";
  var dbkarow$1 = "⤏";
  var dblac$1 = "˝";
  var Dcaron$1 = "Ď";
  var dcaron$1 = "ď";
  var Dcy$1 = "Д";
  var dcy$1 = "д";
  var ddagger$1 = "‡";
  var ddarr$1 = "⇊";
  var DD$1 = "ⅅ";
  var dd$1 = "ⅆ";
  var DDotrahd$1 = "⤑";
  var ddotseq$1 = "⩷";
  var deg$2 = "°";
  var Del$1 = "∇";
  var Delta$1 = "Δ";
  var delta$1 = "δ";
  var demptyv$1 = "⦱";
  var dfisht$1 = "⥿";
  var Dfr$1 = "𝔇";
  var dfr$1 = "𝔡";
  var dHar$1 = "⥥";
  var dharl$1 = "⇃";
  var dharr$1 = "⇂";
  var DiacriticalAcute$1 = "´";
  var DiacriticalDot$1 = "˙";
  var DiacriticalDoubleAcute$1 = "˝";
  var DiacriticalGrave$1 = "`";
  var DiacriticalTilde$1 = "˜";
  var diam$1 = "⋄";
  var diamond$1 = "⋄";
  var Diamond$1 = "⋄";
  var diamondsuit$1 = "♦";
  var diams$1 = "♦";
  var die$1 = "¨";
  var DifferentialD$1 = "ⅆ";
  var digamma$1 = "ϝ";
  var disin$1 = "⋲";
  var div$1 = "÷";
  var divide$2 = "÷";
  var divideontimes$1 = "⋇";
  var divonx$1 = "⋇";
  var DJcy$1 = "Ђ";
  var djcy$1 = "ђ";
  var dlcorn$1 = "⌞";
  var dlcrop$1 = "⌍";
  var dollar$1 = "$";
  var Dopf$1 = "𝔻";
  var dopf$1 = "𝕕";
  var Dot$1 = "¨";
  var dot$1 = "˙";
  var DotDot$1 = "⃜";
  var doteq$1 = "≐";
  var doteqdot$1 = "≑";
  var DotEqual$1 = "≐";
  var dotminus$1 = "∸";
  var dotplus$1 = "∔";
  var dotsquare$1 = "⊡";
  var doublebarwedge$1 = "⌆";
  var DoubleContourIntegral$1 = "∯";
  var DoubleDot$1 = "¨";
  var DoubleDownArrow$1 = "⇓";
  var DoubleLeftArrow$1 = "⇐";
  var DoubleLeftRightArrow$1 = "⇔";
  var DoubleLeftTee$1 = "⫤";
  var DoubleLongLeftArrow$1 = "⟸";
  var DoubleLongLeftRightArrow$1 = "⟺";
  var DoubleLongRightArrow$1 = "⟹";
  var DoubleRightArrow$1 = "⇒";
  var DoubleRightTee$1 = "⊨";
  var DoubleUpArrow$1 = "⇑";
  var DoubleUpDownArrow$1 = "⇕";
  var DoubleVerticalBar$1 = "∥";
  var DownArrowBar$1 = "⤓";
  var downarrow$1 = "↓";
  var DownArrow$1 = "↓";
  var Downarrow$1 = "⇓";
  var DownArrowUpArrow$1 = "⇵";
  var DownBreve$1 = "̑";
  var downdownarrows$1 = "⇊";
  var downharpoonleft$1 = "⇃";
  var downharpoonright$1 = "⇂";
  var DownLeftRightVector$1 = "⥐";
  var DownLeftTeeVector$1 = "⥞";
  var DownLeftVectorBar$1 = "⥖";
  var DownLeftVector$1 = "↽";
  var DownRightTeeVector$1 = "⥟";
  var DownRightVectorBar$1 = "⥗";
  var DownRightVector$1 = "⇁";
  var DownTeeArrow$1 = "↧";
  var DownTee$1 = "⊤";
  var drbkarow$1 = "⤐";
  var drcorn$1 = "⌟";
  var drcrop$1 = "⌌";
  var Dscr$1 = "𝒟";
  var dscr$1 = "𝒹";
  var DScy$1 = "Ѕ";
  var dscy$1 = "ѕ";
  var dsol$1 = "⧶";
  var Dstrok$1 = "Đ";
  var dstrok$1 = "đ";
  var dtdot$1 = "⋱";
  var dtri$1 = "▿";
  var dtrif$1 = "▾";
  var duarr$1 = "⇵";
  var duhar$1 = "⥯";
  var dwangle$1 = "⦦";
  var DZcy$1 = "Џ";
  var dzcy$1 = "џ";
  var dzigrarr$1 = "⟿";
  var Eacute$2 = "É";
  var eacute$2 = "é";
  var easter$1 = "⩮";
  var Ecaron$1 = "Ě";
  var ecaron$1 = "ě";
  var Ecirc$2 = "Ê";
  var ecirc$2 = "ê";
  var ecir$1 = "≖";
  var ecolon$1 = "≕";
  var Ecy$1 = "Э";
  var ecy$1 = "э";
  var eDDot$1 = "⩷";
  var Edot$1 = "Ė";
  var edot$1 = "ė";
  var eDot$1 = "≑";
  var ee$1 = "ⅇ";
  var efDot$1 = "≒";
  var Efr$1 = "𝔈";
  var efr$1 = "𝔢";
  var eg$1 = "⪚";
  var Egrave$2 = "È";
  var egrave$2 = "è";
  var egs$1 = "⪖";
  var egsdot$1 = "⪘";
  var el$1 = "⪙";
  var Element$1 = "∈";
  var elinters$1 = "⏧";
  var ell$1 = "ℓ";
  var els$1 = "⪕";
  var elsdot$1 = "⪗";
  var Emacr$1 = "Ē";
  var emacr$1 = "ē";
  var empty$1 = "∅";
  var emptyset$1 = "∅";
  var EmptySmallSquare$1 = "◻";
  var emptyv$1 = "∅";
  var EmptyVerySmallSquare$1 = "▫";
  var emsp13$1 = " ";
  var emsp14$1 = " ";
  var emsp$1 = " ";
  var ENG$1 = "Ŋ";
  var eng$1 = "ŋ";
  var ensp$1 = " ";
  var Eogon$1 = "Ę";
  var eogon$1 = "ę";
  var Eopf$1 = "𝔼";
  var eopf$1 = "𝕖";
  var epar$1 = "⋕";
  var eparsl$1 = "⧣";
  var eplus$1 = "⩱";
  var epsi$1 = "ε";
  var Epsilon$1 = "Ε";
  var epsilon$1 = "ε";
  var epsiv$1 = "ϵ";
  var eqcirc$1 = "≖";
  var eqcolon$1 = "≕";
  var eqsim$1 = "≂";
  var eqslantgtr$1 = "⪖";
  var eqslantless$1 = "⪕";
  var Equal$1 = "⩵";
  var equals$1 = "=";
  var EqualTilde$1 = "≂";
  var equest$1 = "≟";
  var Equilibrium$1 = "⇌";
  var equiv$1 = "≡";
  var equivDD$1 = "⩸";
  var eqvparsl$1 = "⧥";
  var erarr$1 = "⥱";
  var erDot$1 = "≓";
  var escr$1 = "ℯ";
  var Escr$1 = "ℰ";
  var esdot$1 = "≐";
  var Esim$1 = "⩳";
  var esim$1 = "≂";
  var Eta$1 = "Η";
  var eta$1 = "η";
  var ETH$2 = "Ð";
  var eth$2 = "ð";
  var Euml$2 = "Ë";
  var euml$2 = "ë";
  var euro$1 = "€";
  var excl$1 = "!";
  var exist$1 = "∃";
  var Exists$1 = "∃";
  var expectation$1 = "ℰ";
  var exponentiale$1 = "ⅇ";
  var ExponentialE$1 = "ⅇ";
  var fallingdotseq$1 = "≒";
  var Fcy$1 = "Ф";
  var fcy$1 = "ф";
  var female$1 = "♀";
  var ffilig$1 = "ﬃ";
  var fflig$1 = "ﬀ";
  var ffllig$1 = "ﬄ";
  var Ffr$1 = "𝔉";
  var ffr$1 = "𝔣";
  var filig$1 = "ﬁ";
  var FilledSmallSquare$1 = "◼";
  var FilledVerySmallSquare$1 = "▪";
  var fjlig$1 = "fj";
  var flat$1 = "♭";
  var fllig$1 = "ﬂ";
  var fltns$1 = "▱";
  var fnof$1 = "ƒ";
  var Fopf$1 = "𝔽";
  var fopf$1 = "𝕗";
  var forall$1 = "∀";
  var ForAll$1 = "∀";
  var fork$1 = "⋔";
  var forkv$1 = "⫙";
  var Fouriertrf$1 = "ℱ";
  var fpartint$1 = "⨍";
  var frac12$2 = "½";
  var frac13$1 = "⅓";
  var frac14$2 = "¼";
  var frac15$1 = "⅕";
  var frac16$1 = "⅙";
  var frac18$1 = "⅛";
  var frac23$1 = "⅔";
  var frac25$1 = "⅖";
  var frac34$2 = "¾";
  var frac35$1 = "⅗";
  var frac38$1 = "⅜";
  var frac45$1 = "⅘";
  var frac56$1 = "⅚";
  var frac58$1 = "⅝";
  var frac78$1 = "⅞";
  var frasl$1 = "⁄";
  var frown$1 = "⌢";
  var fscr$1 = "𝒻";
  var Fscr$1 = "ℱ";
  var gacute$1 = "ǵ";
  var Gamma$1 = "Γ";
  var gamma$1 = "γ";
  var Gammad$1 = "Ϝ";
  var gammad$1 = "ϝ";
  var gap$1 = "⪆";
  var Gbreve$1 = "Ğ";
  var gbreve$1 = "ğ";
  var Gcedil$1 = "Ģ";
  var Gcirc$1 = "Ĝ";
  var gcirc$1 = "ĝ";
  var Gcy$1 = "Г";
  var gcy$1 = "г";
  var Gdot$1 = "Ġ";
  var gdot$1 = "ġ";
  var ge$1 = "≥";
  var gE$1 = "≧";
  var gEl$1 = "⪌";
  var gel$1 = "⋛";
  var geq$1 = "≥";
  var geqq$1 = "≧";
  var geqslant$1 = "⩾";
  var gescc$1 = "⪩";
  var ges$1 = "⩾";
  var gesdot$1 = "⪀";
  var gesdoto$1 = "⪂";
  var gesdotol$1 = "⪄";
  var gesl$1 = "⋛︀";
  var gesles$1 = "⪔";
  var Gfr$1 = "𝔊";
  var gfr$1 = "𝔤";
  var gg$1 = "≫";
  var Gg$1 = "⋙";
  var ggg$1 = "⋙";
  var gimel$1 = "ℷ";
  var GJcy$1 = "Ѓ";
  var gjcy$1 = "ѓ";
  var gla$1 = "⪥";
  var gl$1 = "≷";
  var glE$1 = "⪒";
  var glj$1 = "⪤";
  var gnap$1 = "⪊";
  var gnapprox$1 = "⪊";
  var gne$1 = "⪈";
  var gnE$1 = "≩";
  var gneq$1 = "⪈";
  var gneqq$1 = "≩";
  var gnsim$1 = "⋧";
  var Gopf$1 = "𝔾";
  var gopf$1 = "𝕘";
  var grave$1 = "`";
  var GreaterEqual$1 = "≥";
  var GreaterEqualLess$1 = "⋛";
  var GreaterFullEqual$1 = "≧";
  var GreaterGreater$1 = "⪢";
  var GreaterLess$1 = "≷";
  var GreaterSlantEqual$1 = "⩾";
  var GreaterTilde$1 = "≳";
  var Gscr$1 = "𝒢";
  var gscr$1 = "ℊ";
  var gsim$1 = "≳";
  var gsime$1 = "⪎";
  var gsiml$1 = "⪐";
  var gtcc$1 = "⪧";
  var gtcir$1 = "⩺";
  var gt$1$1 = ">";
  var GT$2 = ">";
  var Gt$1 = "≫";
  var gtdot$1 = "⋗";
  var gtlPar$1 = "⦕";
  var gtquest$1 = "⩼";
  var gtrapprox$1 = "⪆";
  var gtrarr$1 = "⥸";
  var gtrdot$1 = "⋗";
  var gtreqless$1 = "⋛";
  var gtreqqless$1 = "⪌";
  var gtrless$1 = "≷";
  var gtrsim$1 = "≳";
  var gvertneqq$1 = "≩︀";
  var gvnE$1 = "≩︀";
  var Hacek$1 = "ˇ";
  var hairsp$1 = " ";
  var half$1 = "½";
  var hamilt$1 = "ℋ";
  var HARDcy$1 = "Ъ";
  var hardcy$1 = "ъ";
  var harrcir$1 = "⥈";
  var harr$1 = "↔";
  var hArr$1 = "⇔";
  var harrw$1 = "↭";
  var Hat$1 = "^";
  var hbar$1 = "ℏ";
  var Hcirc$1 = "Ĥ";
  var hcirc$1 = "ĥ";
  var hearts$1 = "♥";
  var heartsuit$1 = "♥";
  var hellip$1 = "…";
  var hercon$1 = "⊹";
  var hfr$1 = "𝔥";
  var Hfr$1 = "ℌ";
  var HilbertSpace$1 = "ℋ";
  var hksearow$1 = "⤥";
  var hkswarow$1 = "⤦";
  var hoarr$1 = "⇿";
  var homtht$1 = "∻";
  var hookleftarrow$1 = "↩";
  var hookrightarrow$1 = "↪";
  var hopf$1 = "𝕙";
  var Hopf$1 = "ℍ";
  var horbar$1 = "―";
  var HorizontalLine$1 = "─";
  var hscr$1 = "𝒽";
  var Hscr$1 = "ℋ";
  var hslash$1 = "ℏ";
  var Hstrok$1 = "Ħ";
  var hstrok$1 = "ħ";
  var HumpDownHump$1 = "≎";
  var HumpEqual$1 = "≏";
  var hybull$1 = "⁃";
  var hyphen$1 = "‐";
  var Iacute$2 = "Í";
  var iacute$2 = "í";
  var ic$1 = "⁣";
  var Icirc$2 = "Î";
  var icirc$2 = "î";
  var Icy$1 = "И";
  var icy$1 = "и";
  var Idot$1 = "İ";
  var IEcy$1 = "Е";
  var iecy$1 = "е";
  var iexcl$2 = "¡";
  var iff$1 = "⇔";
  var ifr$1 = "𝔦";
  var Ifr$1 = "ℑ";
  var Igrave$2 = "Ì";
  var igrave$2 = "ì";
  var ii$1 = "ⅈ";
  var iiiint$1 = "⨌";
  var iiint$1 = "∭";
  var iinfin$1 = "⧜";
  var iiota$1 = "℩";
  var IJlig$1 = "Ĳ";
  var ijlig$1 = "ĳ";
  var Imacr$1 = "Ī";
  var imacr$1 = "ī";
  var image$1 = "ℑ";
  var ImaginaryI$1 = "ⅈ";
  var imagline$1 = "ℐ";
  var imagpart$1 = "ℑ";
  var imath$1 = "ı";
  var Im$1 = "ℑ";
  var imof$1 = "⊷";
  var imped$1 = "Ƶ";
  var Implies$1 = "⇒";
  var incare$1 = "℅";
  var infin$1 = "∞";
  var infintie$1 = "⧝";
  var inodot$1 = "ı";
  var intcal$1 = "⊺";
  var int$1 = "∫";
  var Int$1 = "∬";
  var integers$1 = "ℤ";
  var Integral$1 = "∫";
  var intercal$1 = "⊺";
  var Intersection$1 = "⋂";
  var intlarhk$1 = "⨗";
  var intprod$1 = "⨼";
  var InvisibleComma$1 = "⁣";
  var InvisibleTimes$1 = "⁢";
  var IOcy$1 = "Ё";
  var iocy$1 = "ё";
  var Iogon$1 = "Į";
  var iogon$1 = "į";
  var Iopf$1 = "𝕀";
  var iopf$1 = "𝕚";
  var Iota$1 = "Ι";
  var iota$1 = "ι";
  var iprod$1 = "⨼";
  var iquest$2 = "¿";
  var iscr$1 = "𝒾";
  var Iscr$1 = "ℐ";
  var isin$1 = "∈";
  var isindot$1 = "⋵";
  var isinE$1 = "⋹";
  var isins$1 = "⋴";
  var isinsv$1 = "⋳";
  var isinv$1 = "∈";
  var it$1 = "⁢";
  var Itilde$1 = "Ĩ";
  var itilde$1 = "ĩ";
  var Iukcy$1 = "І";
  var iukcy$1 = "і";
  var Iuml$2 = "Ï";
  var iuml$2 = "ï";
  var Jcirc$1 = "Ĵ";
  var jcirc$1 = "ĵ";
  var Jcy$1 = "Й";
  var jcy$1 = "й";
  var Jfr$1 = "𝔍";
  var jfr$1 = "𝔧";
  var jmath$1 = "ȷ";
  var Jopf$1 = "𝕁";
  var jopf$1 = "𝕛";
  var Jscr$1 = "𝒥";
  var jscr$1 = "𝒿";
  var Jsercy$1 = "Ј";
  var jsercy$1 = "ј";
  var Jukcy$1 = "Є";
  var jukcy$1 = "є";
  var Kappa$1 = "Κ";
  var kappa$1 = "κ";
  var kappav$1 = "ϰ";
  var Kcedil$1 = "Ķ";
  var kcedil$1 = "ķ";
  var Kcy$1 = "К";
  var kcy$1 = "к";
  var Kfr$1 = "𝔎";
  var kfr$1 = "𝔨";
  var kgreen$1 = "ĸ";
  var KHcy$1 = "Х";
  var khcy$1 = "х";
  var KJcy$1 = "Ќ";
  var kjcy$1 = "ќ";
  var Kopf$1 = "𝕂";
  var kopf$1 = "𝕜";
  var Kscr$1 = "𝒦";
  var kscr$1 = "𝓀";
  var lAarr$1 = "⇚";
  var Lacute$1 = "Ĺ";
  var lacute$1 = "ĺ";
  var laemptyv$1 = "⦴";
  var lagran$1 = "ℒ";
  var Lambda$1 = "Λ";
  var lambda$1 = "λ";
  var lang$1 = "⟨";
  var Lang$1 = "⟪";
  var langd$1 = "⦑";
  var langle$1 = "⟨";
  var lap$1 = "⪅";
  var Laplacetrf$1 = "ℒ";
  var laquo$2 = "«";
  var larrb$1 = "⇤";
  var larrbfs$1 = "⤟";
  var larr$1 = "←";
  var Larr$1 = "↞";
  var lArr$1 = "⇐";
  var larrfs$1 = "⤝";
  var larrhk$1 = "↩";
  var larrlp$1 = "↫";
  var larrpl$1 = "⤹";
  var larrsim$1 = "⥳";
  var larrtl$1 = "↢";
  var latail$1 = "⤙";
  var lAtail$1 = "⤛";
  var lat$1 = "⪫";
  var late$1 = "⪭";
  var lates$1 = "⪭︀";
  var lbarr$1 = "⤌";
  var lBarr$1 = "⤎";
  var lbbrk$1 = "❲";
  var lbrace$1 = "{";
  var lbrack$1 = "[";
  var lbrke$1 = "⦋";
  var lbrksld$1 = "⦏";
  var lbrkslu$1 = "⦍";
  var Lcaron$1 = "Ľ";
  var lcaron$1 = "ľ";
  var Lcedil$1 = "Ļ";
  var lcedil$1 = "ļ";
  var lceil$1 = "⌈";
  var lcub$1 = "{";
  var Lcy$1 = "Л";
  var lcy$1 = "л";
  var ldca$1 = "⤶";
  var ldquo$1 = "“";
  var ldquor$1 = "„";
  var ldrdhar$1 = "⥧";
  var ldrushar$1 = "⥋";
  var ldsh$1 = "↲";
  var le$1 = "≤";
  var lE$1 = "≦";
  var LeftAngleBracket$1 = "⟨";
  var LeftArrowBar$1 = "⇤";
  var leftarrow$1 = "←";
  var LeftArrow$1 = "←";
  var Leftarrow$1 = "⇐";
  var LeftArrowRightArrow$1 = "⇆";
  var leftarrowtail$1 = "↢";
  var LeftCeiling$1 = "⌈";
  var LeftDoubleBracket$1 = "⟦";
  var LeftDownTeeVector$1 = "⥡";
  var LeftDownVectorBar$1 = "⥙";
  var LeftDownVector$1 = "⇃";
  var LeftFloor$1 = "⌊";
  var leftharpoondown$1 = "↽";
  var leftharpoonup$1 = "↼";
  var leftleftarrows$1 = "⇇";
  var leftrightarrow$1 = "↔";
  var LeftRightArrow$1 = "↔";
  var Leftrightarrow$1 = "⇔";
  var leftrightarrows$1 = "⇆";
  var leftrightharpoons$1 = "⇋";
  var leftrightsquigarrow$1 = "↭";
  var LeftRightVector$1 = "⥎";
  var LeftTeeArrow$1 = "↤";
  var LeftTee$1 = "⊣";
  var LeftTeeVector$1 = "⥚";
  var leftthreetimes$1 = "⋋";
  var LeftTriangleBar$1 = "⧏";
  var LeftTriangle$1 = "⊲";
  var LeftTriangleEqual$1 = "⊴";
  var LeftUpDownVector$1 = "⥑";
  var LeftUpTeeVector$1 = "⥠";
  var LeftUpVectorBar$1 = "⥘";
  var LeftUpVector$1 = "↿";
  var LeftVectorBar$1 = "⥒";
  var LeftVector$1 = "↼";
  var lEg$1 = "⪋";
  var leg$1 = "⋚";
  var leq$1 = "≤";
  var leqq$1 = "≦";
  var leqslant$1 = "⩽";
  var lescc$1 = "⪨";
  var les$1 = "⩽";
  var lesdot$1 = "⩿";
  var lesdoto$1 = "⪁";
  var lesdotor$1 = "⪃";
  var lesg$1 = "⋚︀";
  var lesges$1 = "⪓";
  var lessapprox$1 = "⪅";
  var lessdot$1 = "⋖";
  var lesseqgtr$1 = "⋚";
  var lesseqqgtr$1 = "⪋";
  var LessEqualGreater$1 = "⋚";
  var LessFullEqual$1 = "≦";
  var LessGreater$1 = "≶";
  var lessgtr$1 = "≶";
  var LessLess$1 = "⪡";
  var lesssim$1 = "≲";
  var LessSlantEqual$1 = "⩽";
  var LessTilde$1 = "≲";
  var lfisht$1 = "⥼";
  var lfloor$1 = "⌊";
  var Lfr$1 = "𝔏";
  var lfr$1 = "𝔩";
  var lg$1 = "≶";
  var lgE$1 = "⪑";
  var lHar$1 = "⥢";
  var lhard$1 = "↽";
  var lharu$1 = "↼";
  var lharul$1 = "⥪";
  var lhblk$1 = "▄";
  var LJcy$1 = "Љ";
  var ljcy$1 = "љ";
  var llarr$1 = "⇇";
  var ll$1 = "≪";
  var Ll$1 = "⋘";
  var llcorner$1 = "⌞";
  var Lleftarrow$1 = "⇚";
  var llhard$1 = "⥫";
  var lltri$1 = "◺";
  var Lmidot$1 = "Ŀ";
  var lmidot$1 = "ŀ";
  var lmoustache$1 = "⎰";
  var lmoust$1 = "⎰";
  var lnap$1 = "⪉";
  var lnapprox$1 = "⪉";
  var lne$1 = "⪇";
  var lnE$1 = "≨";
  var lneq$1 = "⪇";
  var lneqq$1 = "≨";
  var lnsim$1 = "⋦";
  var loang$1 = "⟬";
  var loarr$1 = "⇽";
  var lobrk$1 = "⟦";
  var longleftarrow$1 = "⟵";
  var LongLeftArrow$1 = "⟵";
  var Longleftarrow$1 = "⟸";
  var longleftrightarrow$1 = "⟷";
  var LongLeftRightArrow$1 = "⟷";
  var Longleftrightarrow$1 = "⟺";
  var longmapsto$1 = "⟼";
  var longrightarrow$1 = "⟶";
  var LongRightArrow$1 = "⟶";
  var Longrightarrow$1 = "⟹";
  var looparrowleft$1 = "↫";
  var looparrowright$1 = "↬";
  var lopar$1 = "⦅";
  var Lopf$1 = "𝕃";
  var lopf$1 = "𝕝";
  var loplus$1 = "⨭";
  var lotimes$1 = "⨴";
  var lowast$1 = "∗";
  var lowbar$1 = "_";
  var LowerLeftArrow$1 = "↙";
  var LowerRightArrow$1 = "↘";
  var loz$1 = "◊";
  var lozenge$1 = "◊";
  var lozf$1 = "⧫";
  var lpar$1 = "(";
  var lparlt$1 = "⦓";
  var lrarr$1 = "⇆";
  var lrcorner$1 = "⌟";
  var lrhar$1 = "⇋";
  var lrhard$1 = "⥭";
  var lrm$1 = "‎";
  var lrtri$1 = "⊿";
  var lsaquo$1 = "‹";
  var lscr$1 = "𝓁";
  var Lscr$1 = "ℒ";
  var lsh$1 = "↰";
  var Lsh$1 = "↰";
  var lsim$1 = "≲";
  var lsime$1 = "⪍";
  var lsimg$1 = "⪏";
  var lsqb$1 = "[";
  var lsquo$1 = "‘";
  var lsquor$1 = "‚";
  var Lstrok$1 = "Ł";
  var lstrok$1 = "ł";
  var ltcc$1 = "⪦";
  var ltcir$1 = "⩹";
  var lt$1$1 = "<";
  var LT$2 = "<";
  var Lt$1 = "≪";
  var ltdot$1 = "⋖";
  var lthree$1 = "⋋";
  var ltimes$1 = "⋉";
  var ltlarr$1 = "⥶";
  var ltquest$1 = "⩻";
  var ltri$1 = "◃";
  var ltrie$1 = "⊴";
  var ltrif$1 = "◂";
  var ltrPar$1 = "⦖";
  var lurdshar$1 = "⥊";
  var luruhar$1 = "⥦";
  var lvertneqq$1 = "≨︀";
  var lvnE$1 = "≨︀";
  var macr$2 = "¯";
  var male$1 = "♂";
  var malt$1 = "✠";
  var maltese$1 = "✠";
  var map$1$1 = "↦";
  var mapsto$1 = "↦";
  var mapstodown$1 = "↧";
  var mapstoleft$1 = "↤";
  var mapstoup$1 = "↥";
  var marker$1 = "▮";
  var mcomma$1 = "⨩";
  var Mcy$1 = "М";
  var mcy$1 = "м";
  var mdash$1 = "—";
  var mDDot$1 = "∺";
  var measuredangle$1 = "∡";
  var MediumSpace$1 = " ";
  var Mellintrf$1 = "ℳ";
  var Mfr$1 = "𝔐";
  var mfr$1 = "𝔪";
  var mho$1 = "℧";
  var micro$2 = "µ";
  var midast$1 = "*";
  var midcir$1 = "⫰";
  var mid$1 = "∣";
  var middot$2 = "·";
  var minusb$1 = "⊟";
  var minus$1 = "−";
  var minusd$1 = "∸";
  var minusdu$1 = "⨪";
  var MinusPlus$1 = "∓";
  var mlcp$1 = "⫛";
  var mldr$1 = "…";
  var mnplus$1 = "∓";
  var models$1 = "⊧";
  var Mopf$1 = "𝕄";
  var mopf$1 = "𝕞";
  var mp$1 = "∓";
  var mscr$1 = "𝓂";
  var Mscr$1 = "ℳ";
  var mstpos$1 = "∾";
  var Mu$1 = "Μ";
  var mu$1 = "μ";
  var multimap$1 = "⊸";
  var mumap$1 = "⊸";
  var nabla$1 = "∇";
  var Nacute$1 = "Ń";
  var nacute$1 = "ń";
  var nang$1 = "∠⃒";
  var nap$1 = "≉";
  var napE$1 = "⩰̸";
  var napid$1 = "≋̸";
  var napos$1 = "ŉ";
  var napprox$1 = "≉";
  var natural$1 = "♮";
  var naturals$1 = "ℕ";
  var natur$1 = "♮";
  var nbsp$2 = " ";
  var nbump$1 = "≎̸";
  var nbumpe$1 = "≏̸";
  var ncap$1 = "⩃";
  var Ncaron$1 = "Ň";
  var ncaron$1 = "ň";
  var Ncedil$1 = "Ņ";
  var ncedil$1 = "ņ";
  var ncong$1 = "≇";
  var ncongdot$1 = "⩭̸";
  var ncup$1 = "⩂";
  var Ncy$1 = "Н";
  var ncy$1 = "н";
  var ndash$1 = "–";
  var nearhk$1 = "⤤";
  var nearr$1 = "↗";
  var neArr$1 = "⇗";
  var nearrow$1 = "↗";
  var ne$1 = "≠";
  var nedot$1 = "≐̸";
  var NegativeMediumSpace$1 = "​";
  var NegativeThickSpace$1 = "​";
  var NegativeThinSpace$1 = "​";
  var NegativeVeryThinSpace$1 = "​";
  var nequiv$1 = "≢";
  var nesear$1 = "⤨";
  var nesim$1 = "≂̸";
  var NestedGreaterGreater$1 = "≫";
  var NestedLessLess$1 = "≪";
  var NewLine$1 = "\n";
  var nexist$1 = "∄";
  var nexists$1 = "∄";
  var Nfr$1 = "𝔑";
  var nfr$1 = "𝔫";
  var ngE$1 = "≧̸";
  var nge$1 = "≱";
  var ngeq$1 = "≱";
  var ngeqq$1 = "≧̸";
  var ngeqslant$1 = "⩾̸";
  var nges$1 = "⩾̸";
  var nGg$1 = "⋙̸";
  var ngsim$1 = "≵";
  var nGt$1 = "≫⃒";
  var ngt$1 = "≯";
  var ngtr$1 = "≯";
  var nGtv$1 = "≫̸";
  var nharr$1 = "↮";
  var nhArr$1 = "⇎";
  var nhpar$1 = "⫲";
  var ni$1 = "∋";
  var nis$1 = "⋼";
  var nisd$1 = "⋺";
  var niv$1 = "∋";
  var NJcy$1 = "Њ";
  var njcy$1 = "њ";
  var nlarr$1 = "↚";
  var nlArr$1 = "⇍";
  var nldr$1 = "‥";
  var nlE$1 = "≦̸";
  var nle$1 = "≰";
  var nleftarrow$1 = "↚";
  var nLeftarrow$1 = "⇍";
  var nleftrightarrow$1 = "↮";
  var nLeftrightarrow$1 = "⇎";
  var nleq$1 = "≰";
  var nleqq$1 = "≦̸";
  var nleqslant$1 = "⩽̸";
  var nles$1 = "⩽̸";
  var nless$1 = "≮";
  var nLl$1 = "⋘̸";
  var nlsim$1 = "≴";
  var nLt$1 = "≪⃒";
  var nlt$1 = "≮";
  var nltri$1 = "⋪";
  var nltrie$1 = "⋬";
  var nLtv$1 = "≪̸";
  var nmid$1 = "∤";
  var NoBreak$1 = "⁠";
  var NonBreakingSpace$1 = " ";
  var nopf$1 = "𝕟";
  var Nopf$1 = "ℕ";
  var Not$1 = "⫬";
  var not$2 = "¬";
  var NotCongruent$1 = "≢";
  var NotCupCap$1 = "≭";
  var NotDoubleVerticalBar$1 = "∦";
  var NotElement$1 = "∉";
  var NotEqual$1 = "≠";
  var NotEqualTilde$1 = "≂̸";
  var NotExists$1 = "∄";
  var NotGreater$1 = "≯";
  var NotGreaterEqual$1 = "≱";
  var NotGreaterFullEqual$1 = "≧̸";
  var NotGreaterGreater$1 = "≫̸";
  var NotGreaterLess$1 = "≹";
  var NotGreaterSlantEqual$1 = "⩾̸";
  var NotGreaterTilde$1 = "≵";
  var NotHumpDownHump$1 = "≎̸";
  var NotHumpEqual$1 = "≏̸";
  var notin$1 = "∉";
  var notindot$1 = "⋵̸";
  var notinE$1 = "⋹̸";
  var notinva$1 = "∉";
  var notinvb$1 = "⋷";
  var notinvc$1 = "⋶";
  var NotLeftTriangleBar$1 = "⧏̸";
  var NotLeftTriangle$1 = "⋪";
  var NotLeftTriangleEqual$1 = "⋬";
  var NotLess$1 = "≮";
  var NotLessEqual$1 = "≰";
  var NotLessGreater$1 = "≸";
  var NotLessLess$1 = "≪̸";
  var NotLessSlantEqual$1 = "⩽̸";
  var NotLessTilde$1 = "≴";
  var NotNestedGreaterGreater$1 = "⪢̸";
  var NotNestedLessLess$1 = "⪡̸";
  var notni$1 = "∌";
  var notniva$1 = "∌";
  var notnivb$1 = "⋾";
  var notnivc$1 = "⋽";
  var NotPrecedes$1 = "⊀";
  var NotPrecedesEqual$1 = "⪯̸";
  var NotPrecedesSlantEqual$1 = "⋠";
  var NotReverseElement$1 = "∌";
  var NotRightTriangleBar$1 = "⧐̸";
  var NotRightTriangle$1 = "⋫";
  var NotRightTriangleEqual$1 = "⋭";
  var NotSquareSubset$1 = "⊏̸";
  var NotSquareSubsetEqual$1 = "⋢";
  var NotSquareSuperset$1 = "⊐̸";
  var NotSquareSupersetEqual$1 = "⋣";
  var NotSubset$1 = "⊂⃒";
  var NotSubsetEqual$1 = "⊈";
  var NotSucceeds$1 = "⊁";
  var NotSucceedsEqual$1 = "⪰̸";
  var NotSucceedsSlantEqual$1 = "⋡";
  var NotSucceedsTilde$1 = "≿̸";
  var NotSuperset$1 = "⊃⃒";
  var NotSupersetEqual$1 = "⊉";
  var NotTilde$1 = "≁";
  var NotTildeEqual$1 = "≄";
  var NotTildeFullEqual$1 = "≇";
  var NotTildeTilde$1 = "≉";
  var NotVerticalBar$1 = "∤";
  var nparallel$1 = "∦";
  var npar$1 = "∦";
  var nparsl$1 = "⫽⃥";
  var npart$1 = "∂̸";
  var npolint$1 = "⨔";
  var npr$1 = "⊀";
  var nprcue$1 = "⋠";
  var nprec$1 = "⊀";
  var npreceq$1 = "⪯̸";
  var npre$1 = "⪯̸";
  var nrarrc$1 = "⤳̸";
  var nrarr$1 = "↛";
  var nrArr$1 = "⇏";
  var nrarrw$1 = "↝̸";
  var nrightarrow$1 = "↛";
  var nRightarrow$1 = "⇏";
  var nrtri$1 = "⋫";
  var nrtrie$1 = "⋭";
  var nsc$1 = "⊁";
  var nsccue$1 = "⋡";
  var nsce$1 = "⪰̸";
  var Nscr$1 = "𝒩";
  var nscr$1 = "𝓃";
  var nshortmid$1 = "∤";
  var nshortparallel$1 = "∦";
  var nsim$1 = "≁";
  var nsime$1 = "≄";
  var nsimeq$1 = "≄";
  var nsmid$1 = "∤";
  var nspar$1 = "∦";
  var nsqsube$1 = "⋢";
  var nsqsupe$1 = "⋣";
  var nsub$1 = "⊄";
  var nsubE$1 = "⫅̸";
  var nsube$1 = "⊈";
  var nsubset$1 = "⊂⃒";
  var nsubseteq$1 = "⊈";
  var nsubseteqq$1 = "⫅̸";
  var nsucc$1 = "⊁";
  var nsucceq$1 = "⪰̸";
  var nsup$1 = "⊅";
  var nsupE$1 = "⫆̸";
  var nsupe$1 = "⊉";
  var nsupset$1 = "⊃⃒";
  var nsupseteq$1 = "⊉";
  var nsupseteqq$1 = "⫆̸";
  var ntgl$1 = "≹";
  var Ntilde$2 = "Ñ";
  var ntilde$2 = "ñ";
  var ntlg$1 = "≸";
  var ntriangleleft$1 = "⋪";
  var ntrianglelefteq$1 = "⋬";
  var ntriangleright$1 = "⋫";
  var ntrianglerighteq$1 = "⋭";
  var Nu$1 = "Ν";
  var nu$1 = "ν";
  var num$1 = "#";
  var numero$1 = "№";
  var numsp$1 = " ";
  var nvap$1 = "≍⃒";
  var nvdash$1 = "⊬";
  var nvDash$1 = "⊭";
  var nVdash$1 = "⊮";
  var nVDash$1 = "⊯";
  var nvge$1 = "≥⃒";
  var nvgt$1 = ">⃒";
  var nvHarr$1 = "⤄";
  var nvinfin$1 = "⧞";
  var nvlArr$1 = "⤂";
  var nvle$1 = "≤⃒";
  var nvlt$1 = "<⃒";
  var nvltrie$1 = "⊴⃒";
  var nvrArr$1 = "⤃";
  var nvrtrie$1 = "⊵⃒";
  var nvsim$1 = "∼⃒";
  var nwarhk$1 = "⤣";
  var nwarr$1 = "↖";
  var nwArr$1 = "⇖";
  var nwarrow$1 = "↖";
  var nwnear$1 = "⤧";
  var Oacute$2 = "Ó";
  var oacute$2 = "ó";
  var oast$1 = "⊛";
  var Ocirc$2 = "Ô";
  var ocirc$2 = "ô";
  var ocir$1 = "⊚";
  var Ocy$1 = "О";
  var ocy$1 = "о";
  var odash$1 = "⊝";
  var Odblac$1 = "Ő";
  var odblac$1 = "ő";
  var odiv$1 = "⨸";
  var odot$1 = "⊙";
  var odsold$1 = "⦼";
  var OElig$1 = "Œ";
  var oelig$1 = "œ";
  var ofcir$1 = "⦿";
  var Ofr$1 = "𝔒";
  var ofr$1 = "𝔬";
  var ogon$1 = "˛";
  var Ograve$2 = "Ò";
  var ograve$2 = "ò";
  var ogt$1 = "⧁";
  var ohbar$1 = "⦵";
  var ohm$1 = "Ω";
  var oint$1 = "∮";
  var olarr$1 = "↺";
  var olcir$1 = "⦾";
  var olcross$1 = "⦻";
  var oline$1 = "‾";
  var olt$1 = "⧀";
  var Omacr$1 = "Ō";
  var omacr$1 = "ō";
  var Omega$1 = "Ω";
  var omega$1 = "ω";
  var Omicron$1 = "Ο";
  var omicron$1 = "ο";
  var omid$1 = "⦶";
  var ominus$1 = "⊖";
  var Oopf$1 = "𝕆";
  var oopf$1 = "𝕠";
  var opar$1 = "⦷";
  var OpenCurlyDoubleQuote$1 = "“";
  var OpenCurlyQuote$1 = "‘";
  var operp$1 = "⦹";
  var oplus$1 = "⊕";
  var orarr$1 = "↻";
  var Or$1 = "⩔";
  var or$1 = "∨";
  var ord$1 = "⩝";
  var order$1 = "ℴ";
  var orderof$1 = "ℴ";
  var ordf$2 = "ª";
  var ordm$2 = "º";
  var origof$1 = "⊶";
  var oror$1 = "⩖";
  var orslope$1 = "⩗";
  var orv$1 = "⩛";
  var oS$1 = "Ⓢ";
  var Oscr$1 = "𝒪";
  var oscr$1 = "ℴ";
  var Oslash$2 = "Ø";
  var oslash$2 = "ø";
  var osol$1 = "⊘";
  var Otilde$2 = "Õ";
  var otilde$2 = "õ";
  var otimesas$1 = "⨶";
  var Otimes$1 = "⨷";
  var otimes$1 = "⊗";
  var Ouml$2 = "Ö";
  var ouml$2 = "ö";
  var ovbar$1 = "⌽";
  var OverBar$1 = "‾";
  var OverBrace$1 = "⏞";
  var OverBracket$1 = "⎴";
  var OverParenthesis$1 = "⏜";
  var para$2 = "¶";
  var parallel$1 = "∥";
  var par$1 = "∥";
  var parsim$1 = "⫳";
  var parsl$1 = "⫽";
  var part$1 = "∂";
  var PartialD$1 = "∂";
  var Pcy$1 = "П";
  var pcy$1 = "п";
  var percnt$1 = "%";
  var period$1 = ".";
  var permil$1 = "‰";
  var perp$1 = "⊥";
  var pertenk$1 = "‱";
  var Pfr$1 = "𝔓";
  var pfr$1 = "𝔭";
  var Phi$1 = "Φ";
  var phi$1 = "φ";
  var phiv$1 = "ϕ";
  var phmmat$1 = "ℳ";
  var phone$1 = "☎";
  var Pi$1 = "Π";
  var pi$1 = "π";
  var pitchfork$1 = "⋔";
  var piv$1 = "ϖ";
  var planck$1 = "ℏ";
  var planckh$1 = "ℎ";
  var plankv$1 = "ℏ";
  var plusacir$1 = "⨣";
  var plusb$1 = "⊞";
  var pluscir$1 = "⨢";
  var plus$1 = "+";
  var plusdo$1 = "∔";
  var plusdu$1 = "⨥";
  var pluse$1 = "⩲";
  var PlusMinus$1 = "±";
  var plusmn$2 = "±";
  var plussim$1 = "⨦";
  var plustwo$1 = "⨧";
  var pm$1 = "±";
  var Poincareplane$1 = "ℌ";
  var pointint$1 = "⨕";
  var popf$1 = "𝕡";
  var Popf$1 = "ℙ";
  var pound$2 = "£";
  var prap$1 = "⪷";
  var Pr$1 = "⪻";
  var pr$1 = "≺";
  var prcue$1 = "≼";
  var precapprox$1 = "⪷";
  var prec$1 = "≺";
  var preccurlyeq$1 = "≼";
  var Precedes$1 = "≺";
  var PrecedesEqual$1 = "⪯";
  var PrecedesSlantEqual$1 = "≼";
  var PrecedesTilde$1 = "≾";
  var preceq$1 = "⪯";
  var precnapprox$1 = "⪹";
  var precneqq$1 = "⪵";
  var precnsim$1 = "⋨";
  var pre$1 = "⪯";
  var prE$1 = "⪳";
  var precsim$1 = "≾";
  var prime$1 = "′";
  var Prime$1 = "″";
  var primes$1 = "ℙ";
  var prnap$1 = "⪹";
  var prnE$1 = "⪵";
  var prnsim$1 = "⋨";
  var prod$1 = "∏";
  var Product$1 = "∏";
  var profalar$1 = "⌮";
  var profline$1 = "⌒";
  var profsurf$1 = "⌓";
  var prop$1 = "∝";
  var Proportional$1 = "∝";
  var Proportion$1 = "∷";
  var propto$1 = "∝";
  var prsim$1 = "≾";
  var prurel$1 = "⊰";
  var Pscr$1 = "𝒫";
  var pscr$1 = "𝓅";
  var Psi$1 = "Ψ";
  var psi$1 = "ψ";
  var puncsp$1 = " ";
  var Qfr$1 = "𝔔";
  var qfr$1 = "𝔮";
  var qint$1 = "⨌";
  var qopf$1 = "𝕢";
  var Qopf$1 = "ℚ";
  var qprime$1 = "⁗";
  var Qscr$1 = "𝒬";
  var qscr$1 = "𝓆";
  var quaternions$1 = "ℍ";
  var quatint$1 = "⨖";
  var quest$1 = "?";
  var questeq$1 = "≟";
  var quot$1$1 = "\"";
  var QUOT$2 = "\"";
  var rAarr$1 = "⇛";
  var race$1 = "∽̱";
  var Racute$1 = "Ŕ";
  var racute$1 = "ŕ";
  var radic$1 = "√";
  var raemptyv$1 = "⦳";
  var rang$1 = "⟩";
  var Rang$1 = "⟫";
  var rangd$1 = "⦒";
  var range$1 = "⦥";
  var rangle$1 = "⟩";
  var raquo$2 = "»";
  var rarrap$1 = "⥵";
  var rarrb$1 = "⇥";
  var rarrbfs$1 = "⤠";
  var rarrc$1 = "⤳";
  var rarr$1 = "→";
  var Rarr$1 = "↠";
  var rArr$1 = "⇒";
  var rarrfs$1 = "⤞";
  var rarrhk$1 = "↪";
  var rarrlp$1 = "↬";
  var rarrpl$1 = "⥅";
  var rarrsim$1 = "⥴";
  var Rarrtl$1 = "⤖";
  var rarrtl$1 = "↣";
  var rarrw$1 = "↝";
  var ratail$1 = "⤚";
  var rAtail$1 = "⤜";
  var ratio$1 = "∶";
  var rationals$1 = "ℚ";
  var rbarr$1 = "⤍";
  var rBarr$1 = "⤏";
  var RBarr$1 = "⤐";
  var rbbrk$1 = "❳";
  var rbrace$1 = "}";
  var rbrack$1 = "]";
  var rbrke$1 = "⦌";
  var rbrksld$1 = "⦎";
  var rbrkslu$1 = "⦐";
  var Rcaron$1 = "Ř";
  var rcaron$1 = "ř";
  var Rcedil$1 = "Ŗ";
  var rcedil$1 = "ŗ";
  var rceil$1 = "⌉";
  var rcub$1 = "}";
  var Rcy$1 = "Р";
  var rcy$1 = "р";
  var rdca$1 = "⤷";
  var rdldhar$1 = "⥩";
  var rdquo$1 = "”";
  var rdquor$1 = "”";
  var rdsh$1 = "↳";
  var real$1 = "ℜ";
  var realine$1 = "ℛ";
  var realpart$1 = "ℜ";
  var reals$1 = "ℝ";
  var Re$1 = "ℜ";
  var rect$1 = "▭";
  var reg$2 = "®";
  var REG$2 = "®";
  var ReverseElement$1 = "∋";
  var ReverseEquilibrium$1 = "⇋";
  var ReverseUpEquilibrium$1 = "⥯";
  var rfisht$1 = "⥽";
  var rfloor$1 = "⌋";
  var rfr$1 = "𝔯";
  var Rfr$1 = "ℜ";
  var rHar$1 = "⥤";
  var rhard$1 = "⇁";
  var rharu$1 = "⇀";
  var rharul$1 = "⥬";
  var Rho$1 = "Ρ";
  var rho$1 = "ρ";
  var rhov$1 = "ϱ";
  var RightAngleBracket$1 = "⟩";
  var RightArrowBar$1 = "⇥";
  var rightarrow$1 = "→";
  var RightArrow$1 = "→";
  var Rightarrow$1 = "⇒";
  var RightArrowLeftArrow$1 = "⇄";
  var rightarrowtail$1 = "↣";
  var RightCeiling$1 = "⌉";
  var RightDoubleBracket$1 = "⟧";
  var RightDownTeeVector$1 = "⥝";
  var RightDownVectorBar$1 = "⥕";
  var RightDownVector$1 = "⇂";
  var RightFloor$1 = "⌋";
  var rightharpoondown$1 = "⇁";
  var rightharpoonup$1 = "⇀";
  var rightleftarrows$1 = "⇄";
  var rightleftharpoons$1 = "⇌";
  var rightrightarrows$1 = "⇉";
  var rightsquigarrow$1 = "↝";
  var RightTeeArrow$1 = "↦";
  var RightTee$1 = "⊢";
  var RightTeeVector$1 = "⥛";
  var rightthreetimes$1 = "⋌";
  var RightTriangleBar$1 = "⧐";
  var RightTriangle$1 = "⊳";
  var RightTriangleEqual$1 = "⊵";
  var RightUpDownVector$1 = "⥏";
  var RightUpTeeVector$1 = "⥜";
  var RightUpVectorBar$1 = "⥔";
  var RightUpVector$1 = "↾";
  var RightVectorBar$1 = "⥓";
  var RightVector$1 = "⇀";
  var ring$1 = "˚";
  var risingdotseq$1 = "≓";
  var rlarr$1 = "⇄";
  var rlhar$1 = "⇌";
  var rlm$1 = "‏";
  var rmoustache$1 = "⎱";
  var rmoust$1 = "⎱";
  var rnmid$1 = "⫮";
  var roang$1 = "⟭";
  var roarr$1 = "⇾";
  var robrk$1 = "⟧";
  var ropar$1 = "⦆";
  var ropf$1 = "𝕣";
  var Ropf$1 = "ℝ";
  var roplus$1 = "⨮";
  var rotimes$1 = "⨵";
  var RoundImplies$1 = "⥰";
  var rpar$1 = ")";
  var rpargt$1 = "⦔";
  var rppolint$1 = "⨒";
  var rrarr$1 = "⇉";
  var Rrightarrow$1 = "⇛";
  var rsaquo$1 = "›";
  var rscr$1 = "𝓇";
  var Rscr$1 = "ℛ";
  var rsh$1 = "↱";
  var Rsh$1 = "↱";
  var rsqb$1 = "]";
  var rsquo$1 = "’";
  var rsquor$1 = "’";
  var rthree$1 = "⋌";
  var rtimes$1 = "⋊";
  var rtri$1 = "▹";
  var rtrie$1 = "⊵";
  var rtrif$1 = "▸";
  var rtriltri$1 = "⧎";
  var RuleDelayed$1 = "⧴";
  var ruluhar$1 = "⥨";
  var rx$1 = "℞";
  var Sacute$1 = "Ś";
  var sacute$1 = "ś";
  var sbquo$1 = "‚";
  var scap$1 = "⪸";
  var Scaron$1 = "Š";
  var scaron$1 = "š";
  var Sc$1 = "⪼";
  var sc$1 = "≻";
  var sccue$1 = "≽";
  var sce$1 = "⪰";
  var scE$1 = "⪴";
  var Scedil$1 = "Ş";
  var scedil$1 = "ş";
  var Scirc$1 = "Ŝ";
  var scirc$1 = "ŝ";
  var scnap$1 = "⪺";
  var scnE$1 = "⪶";
  var scnsim$1 = "⋩";
  var scpolint$1 = "⨓";
  var scsim$1 = "≿";
  var Scy$1 = "С";
  var scy$1 = "с";
  var sdotb$1 = "⊡";
  var sdot$1 = "⋅";
  var sdote$1 = "⩦";
  var searhk$1 = "⤥";
  var searr$1 = "↘";
  var seArr$1 = "⇘";
  var searrow$1 = "↘";
  var sect$2 = "§";
  var semi$1 = ";";
  var seswar$1 = "⤩";
  var setminus$1 = "∖";
  var setmn$1 = "∖";
  var sext$1 = "✶";
  var Sfr$1 = "𝔖";
  var sfr$1 = "𝔰";
  var sfrown$1 = "⌢";
  var sharp$1 = "♯";
  var SHCHcy$1 = "Щ";
  var shchcy$1 = "щ";
  var SHcy$1 = "Ш";
  var shcy$1 = "ш";
  var ShortDownArrow$1 = "↓";
  var ShortLeftArrow$1 = "←";
  var shortmid$1 = "∣";
  var shortparallel$1 = "∥";
  var ShortRightArrow$1 = "→";
  var ShortUpArrow$1 = "↑";
  var shy$2 = "­";
  var Sigma$1 = "Σ";
  var sigma$1 = "σ";
  var sigmaf$1 = "ς";
  var sigmav$1 = "ς";
  var sim$1 = "∼";
  var simdot$1 = "⩪";
  var sime$1 = "≃";
  var simeq$1 = "≃";
  var simg$1 = "⪞";
  var simgE$1 = "⪠";
  var siml$1 = "⪝";
  var simlE$1 = "⪟";
  var simne$1 = "≆";
  var simplus$1 = "⨤";
  var simrarr$1 = "⥲";
  var slarr$1 = "←";
  var SmallCircle$1 = "∘";
  var smallsetminus$1 = "∖";
  var smashp$1 = "⨳";
  var smeparsl$1 = "⧤";
  var smid$1 = "∣";
  var smile$1 = "⌣";
  var smt$1 = "⪪";
  var smte$1 = "⪬";
  var smtes$1 = "⪬︀";
  var SOFTcy$1 = "Ь";
  var softcy$1 = "ь";
  var solbar$1 = "⌿";
  var solb$1 = "⧄";
  var sol$1 = "/";
  var Sopf$1 = "𝕊";
  var sopf$1 = "𝕤";
  var spades$1 = "♠";
  var spadesuit$1 = "♠";
  var spar$1 = "∥";
  var sqcap$1 = "⊓";
  var sqcaps$1 = "⊓︀";
  var sqcup$1 = "⊔";
  var sqcups$1 = "⊔︀";
  var Sqrt$1 = "√";
  var sqsub$1 = "⊏";
  var sqsube$1 = "⊑";
  var sqsubset$1 = "⊏";
  var sqsubseteq$1 = "⊑";
  var sqsup$1 = "⊐";
  var sqsupe$1 = "⊒";
  var sqsupset$1 = "⊐";
  var sqsupseteq$1 = "⊒";
  var square$1 = "□";
  var Square$1 = "□";
  var SquareIntersection$1 = "⊓";
  var SquareSubset$1 = "⊏";
  var SquareSubsetEqual$1 = "⊑";
  var SquareSuperset$1 = "⊐";
  var SquareSupersetEqual$1 = "⊒";
  var SquareUnion$1 = "⊔";
  var squarf$1 = "▪";
  var squ$1 = "□";
  var squf$1 = "▪";
  var srarr$1 = "→";
  var Sscr$1 = "𝒮";
  var sscr$1 = "𝓈";
  var ssetmn$1 = "∖";
  var ssmile$1 = "⌣";
  var sstarf$1 = "⋆";
  var Star$1 = "⋆";
  var star$1 = "☆";
  var starf$1 = "★";
  var straightepsilon$1 = "ϵ";
  var straightphi$1 = "ϕ";
  var strns$1 = "¯";
  var sub$1 = "⊂";
  var Sub$1 = "⋐";
  var subdot$1 = "⪽";
  var subE$1 = "⫅";
  var sube$1 = "⊆";
  var subedot$1 = "⫃";
  var submult$1 = "⫁";
  var subnE$1 = "⫋";
  var subne$1 = "⊊";
  var subplus$1 = "⪿";
  var subrarr$1 = "⥹";
  var subset$1 = "⊂";
  var Subset$1 = "⋐";
  var subseteq$1 = "⊆";
  var subseteqq$1 = "⫅";
  var SubsetEqual$1 = "⊆";
  var subsetneq$1 = "⊊";
  var subsetneqq$1 = "⫋";
  var subsim$1 = "⫇";
  var subsub$1 = "⫕";
  var subsup$1 = "⫓";
  var succapprox$1 = "⪸";
  var succ$1 = "≻";
  var succcurlyeq$1 = "≽";
  var Succeeds$1 = "≻";
  var SucceedsEqual$1 = "⪰";
  var SucceedsSlantEqual$1 = "≽";
  var SucceedsTilde$1 = "≿";
  var succeq$1 = "⪰";
  var succnapprox$1 = "⪺";
  var succneqq$1 = "⪶";
  var succnsim$1 = "⋩";
  var succsim$1 = "≿";
  var SuchThat$1 = "∋";
  var sum$1 = "∑";
  var Sum$1 = "∑";
  var sung$1 = "♪";
  var sup1$2 = "¹";
  var sup2$2 = "²";
  var sup3$2 = "³";
  var sup$1 = "⊃";
  var Sup$1 = "⋑";
  var supdot$1 = "⪾";
  var supdsub$1 = "⫘";
  var supE$1 = "⫆";
  var supe$1 = "⊇";
  var supedot$1 = "⫄";
  var Superset$1 = "⊃";
  var SupersetEqual$1 = "⊇";
  var suphsol$1 = "⟉";
  var suphsub$1 = "⫗";
  var suplarr$1 = "⥻";
  var supmult$1 = "⫂";
  var supnE$1 = "⫌";
  var supne$1 = "⊋";
  var supplus$1 = "⫀";
  var supset$1 = "⊃";
  var Supset$1 = "⋑";
  var supseteq$1 = "⊇";
  var supseteqq$1 = "⫆";
  var supsetneq$1 = "⊋";
  var supsetneqq$1 = "⫌";
  var supsim$1 = "⫈";
  var supsub$1 = "⫔";
  var supsup$1 = "⫖";
  var swarhk$1 = "⤦";
  var swarr$1 = "↙";
  var swArr$1 = "⇙";
  var swarrow$1 = "↙";
  var swnwar$1 = "⤪";
  var szlig$2 = "ß";
  var Tab$1 = "\t";
  var target$1 = "⌖";
  var Tau$1 = "Τ";
  var tau$1 = "τ";
  var tbrk$1 = "⎴";
  var Tcaron$1 = "Ť";
  var tcaron$1 = "ť";
  var Tcedil$1 = "Ţ";
  var tcedil$1 = "ţ";
  var Tcy$1 = "Т";
  var tcy$1 = "т";
  var tdot$1 = "⃛";
  var telrec$1 = "⌕";
  var Tfr$1 = "𝔗";
  var tfr$1 = "𝔱";
  var there4$1 = "∴";
  var therefore$1 = "∴";
  var Therefore$1 = "∴";
  var Theta$1 = "Θ";
  var theta$1 = "θ";
  var thetasym$1 = "ϑ";
  var thetav$1 = "ϑ";
  var thickapprox$1 = "≈";
  var thicksim$1 = "∼";
  var ThickSpace$1 = "  ";
  var ThinSpace$1 = " ";
  var thinsp$1 = " ";
  var thkap$1 = "≈";
  var thksim$1 = "∼";
  var THORN$2 = "Þ";
  var thorn$2 = "þ";
  var tilde$1 = "˜";
  var Tilde$1 = "∼";
  var TildeEqual$1 = "≃";
  var TildeFullEqual$1 = "≅";
  var TildeTilde$1 = "≈";
  var timesbar$1 = "⨱";
  var timesb$1 = "⊠";
  var times$3 = "×";
  var timesd$1 = "⨰";
  var tint$1 = "∭";
  var toea$1 = "⤨";
  var topbot$1 = "⌶";
  var topcir$1 = "⫱";
  var top$1 = "⊤";
  var Topf$1 = "𝕋";
  var topf$1 = "𝕥";
  var topfork$1 = "⫚";
  var tosa$1 = "⤩";
  var tprime$1 = "‴";
  var trade$1 = "™";
  var TRADE$1 = "™";
  var triangle$1 = "▵";
  var triangledown$1 = "▿";
  var triangleleft$1 = "◃";
  var trianglelefteq$1 = "⊴";
  var triangleq$1 = "≜";
  var triangleright$1 = "▹";
  var trianglerighteq$1 = "⊵";
  var tridot$1 = "◬";
  var trie$1 = "≜";
  var triminus$1 = "⨺";
  var TripleDot$1 = "⃛";
  var triplus$1 = "⨹";
  var trisb$1 = "⧍";
  var tritime$1 = "⨻";
  var trpezium$1 = "⏢";
  var Tscr$1 = "𝒯";
  var tscr$1 = "𝓉";
  var TScy$1 = "Ц";
  var tscy$1 = "ц";
  var TSHcy$1 = "Ћ";
  var tshcy$1 = "ћ";
  var Tstrok$1 = "Ŧ";
  var tstrok$1 = "ŧ";
  var twixt$1 = "≬";
  var twoheadleftarrow$1 = "↞";
  var twoheadrightarrow$1 = "↠";
  var Uacute$2 = "Ú";
  var uacute$2 = "ú";
  var uarr$1 = "↑";
  var Uarr$1 = "↟";
  var uArr$1 = "⇑";
  var Uarrocir$1 = "⥉";
  var Ubrcy$1 = "Ў";
  var ubrcy$1 = "ў";
  var Ubreve$1 = "Ŭ";
  var ubreve$1 = "ŭ";
  var Ucirc$2 = "Û";
  var ucirc$2 = "û";
  var Ucy$1 = "У";
  var ucy$1 = "у";
  var udarr$1 = "⇅";
  var Udblac$1 = "Ű";
  var udblac$1 = "ű";
  var udhar$1 = "⥮";
  var ufisht$1 = "⥾";
  var Ufr$1 = "𝔘";
  var ufr$1 = "𝔲";
  var Ugrave$2 = "Ù";
  var ugrave$2 = "ù";
  var uHar$1 = "⥣";
  var uharl$1 = "↿";
  var uharr$1 = "↾";
  var uhblk$1 = "▀";
  var ulcorn$1 = "⌜";
  var ulcorner$1 = "⌜";
  var ulcrop$1 = "⌏";
  var ultri$1 = "◸";
  var Umacr$1 = "Ū";
  var umacr$1 = "ū";
  var uml$2 = "¨";
  var UnderBar$1 = "_";
  var UnderBrace$1 = "⏟";
  var UnderBracket$1 = "⎵";
  var UnderParenthesis$1 = "⏝";
  var Union$1 = "⋃";
  var UnionPlus$1 = "⊎";
  var Uogon$1 = "Ų";
  var uogon$1 = "ų";
  var Uopf$1 = "𝕌";
  var uopf$1 = "𝕦";
  var UpArrowBar$1 = "⤒";
  var uparrow$1 = "↑";
  var UpArrow$1 = "↑";
  var Uparrow$1 = "⇑";
  var UpArrowDownArrow$1 = "⇅";
  var updownarrow$1 = "↕";
  var UpDownArrow$1 = "↕";
  var Updownarrow$1 = "⇕";
  var UpEquilibrium$1 = "⥮";
  var upharpoonleft$1 = "↿";
  var upharpoonright$1 = "↾";
  var uplus$1 = "⊎";
  var UpperLeftArrow$1 = "↖";
  var UpperRightArrow$1 = "↗";
  var upsi$1 = "υ";
  var Upsi$1 = "ϒ";
  var upsih$1 = "ϒ";
  var Upsilon$1 = "Υ";
  var upsilon$1 = "υ";
  var UpTeeArrow$1 = "↥";
  var UpTee$1 = "⊥";
  var upuparrows$1 = "⇈";
  var urcorn$1 = "⌝";
  var urcorner$1 = "⌝";
  var urcrop$1 = "⌎";
  var Uring$1 = "Ů";
  var uring$1 = "ů";
  var urtri$1 = "◹";
  var Uscr$1 = "𝒰";
  var uscr$1 = "𝓊";
  var utdot$1 = "⋰";
  var Utilde$1 = "Ũ";
  var utilde$1 = "ũ";
  var utri$1 = "▵";
  var utrif$1 = "▴";
  var uuarr$1 = "⇈";
  var Uuml$2 = "Ü";
  var uuml$2 = "ü";
  var uwangle$1 = "⦧";
  var vangrt$1 = "⦜";
  var varepsilon$1 = "ϵ";
  var varkappa$1 = "ϰ";
  var varnothing$1 = "∅";
  var varphi$1 = "ϕ";
  var varpi$1 = "ϖ";
  var varpropto$1 = "∝";
  var varr$1 = "↕";
  var vArr$1 = "⇕";
  var varrho$1 = "ϱ";
  var varsigma$1 = "ς";
  var varsubsetneq$1 = "⊊︀";
  var varsubsetneqq$1 = "⫋︀";
  var varsupsetneq$1 = "⊋︀";
  var varsupsetneqq$1 = "⫌︀";
  var vartheta$1 = "ϑ";
  var vartriangleleft$1 = "⊲";
  var vartriangleright$1 = "⊳";
  var vBar$1 = "⫨";
  var Vbar$1 = "⫫";
  var vBarv$1 = "⫩";
  var Vcy$1 = "В";
  var vcy$1 = "в";
  var vdash$1 = "⊢";
  var vDash$1 = "⊨";
  var Vdash$1 = "⊩";
  var VDash$1 = "⊫";
  var Vdashl$1 = "⫦";
  var veebar$1 = "⊻";
  var vee$1 = "∨";
  var Vee$1 = "⋁";
  var veeeq$1 = "≚";
  var vellip$1 = "⋮";
  var verbar$1 = "|";
  var Verbar$1 = "‖";
  var vert$1 = "|";
  var Vert$1 = "‖";
  var VerticalBar$1 = "∣";
  var VerticalLine$1 = "|";
  var VerticalSeparator$1 = "❘";
  var VerticalTilde$1 = "≀";
  var VeryThinSpace$1 = " ";
  var Vfr$1 = "𝔙";
  var vfr$1 = "𝔳";
  var vltri$1 = "⊲";
  var vnsub$1 = "⊂⃒";
  var vnsup$1 = "⊃⃒";
  var Vopf$1 = "𝕍";
  var vopf$1 = "𝕧";
  var vprop$1 = "∝";
  var vrtri$1 = "⊳";
  var Vscr$1 = "𝒱";
  var vscr$1 = "𝓋";
  var vsubnE$1 = "⫋︀";
  var vsubne$1 = "⊊︀";
  var vsupnE$1 = "⫌︀";
  var vsupne$1 = "⊋︀";
  var Vvdash$1 = "⊪";
  var vzigzag$1 = "⦚";
  var Wcirc$1 = "Ŵ";
  var wcirc$1 = "ŵ";
  var wedbar$1 = "⩟";
  var wedge$1 = "∧";
  var Wedge$1 = "⋀";
  var wedgeq$1 = "≙";
  var weierp$1 = "℘";
  var Wfr$1 = "𝔚";
  var wfr$1 = "𝔴";
  var Wopf$1 = "𝕎";
  var wopf$1 = "𝕨";
  var wp$1 = "℘";
  var wr$1 = "≀";
  var wreath$1 = "≀";
  var Wscr$1 = "𝒲";
  var wscr$1 = "𝓌";
  var xcap$1 = "⋂";
  var xcirc$1 = "◯";
  var xcup$1 = "⋃";
  var xdtri$1 = "▽";
  var Xfr$1 = "𝔛";
  var xfr$1 = "𝔵";
  var xharr$1 = "⟷";
  var xhArr$1 = "⟺";
  var Xi$1 = "Ξ";
  var xi$1 = "ξ";
  var xlarr$1 = "⟵";
  var xlArr$1 = "⟸";
  var xmap$1 = "⟼";
  var xnis$1 = "⋻";
  var xodot$1 = "⨀";
  var Xopf$1 = "𝕏";
  var xopf$1 = "𝕩";
  var xoplus$1 = "⨁";
  var xotime$1 = "⨂";
  var xrarr$1 = "⟶";
  var xrArr$1 = "⟹";
  var Xscr$1 = "𝒳";
  var xscr$1 = "𝓍";
  var xsqcup$1 = "⨆";
  var xuplus$1 = "⨄";
  var xutri$1 = "△";
  var xvee$1 = "⋁";
  var xwedge$1 = "⋀";
  var Yacute$2 = "Ý";
  var yacute$2 = "ý";
  var YAcy$1 = "Я";
  var yacy$1 = "я";
  var Ycirc$1 = "Ŷ";
  var ycirc$1 = "ŷ";
  var Ycy$1 = "Ы";
  var ycy$1 = "ы";
  var yen$2 = "¥";
  var Yfr$1 = "𝔜";
  var yfr$1 = "𝔶";
  var YIcy$1 = "Ї";
  var yicy$1 = "ї";
  var Yopf$1 = "𝕐";
  var yopf$1 = "𝕪";
  var Yscr$1 = "𝒴";
  var yscr$1 = "𝓎";
  var YUcy$1 = "Ю";
  var yucy$1 = "ю";
  var yuml$2 = "ÿ";
  var Yuml$1 = "Ÿ";
  var Zacute$1 = "Ź";
  var zacute$1 = "ź";
  var Zcaron$1 = "Ž";
  var zcaron$1 = "ž";
  var Zcy$1 = "З";
  var zcy$1 = "з";
  var Zdot$1 = "Ż";
  var zdot$1 = "ż";
  var zeetrf$1 = "ℨ";
  var ZeroWidthSpace$1 = "​";
  var Zeta$1 = "Ζ";
  var zeta$1 = "ζ";
  var zfr$1 = "𝔷";
  var Zfr$1 = "ℨ";
  var ZHcy$1 = "Ж";
  var zhcy$1 = "ж";
  var zigrarr$1 = "⇝";
  var zopf$1 = "𝕫";
  var Zopf$1 = "ℤ";
  var Zscr$1 = "𝒵";
  var zscr$1 = "𝓏";
  var zwj$1 = "‍";
  var zwnj$1 = "‌";
  var entitiesJSON$1 = {
  	Aacute: Aacute$2,
  	aacute: aacute$2,
  	Abreve: Abreve$1,
  	abreve: abreve$1,
  	ac: ac$1,
  	acd: acd$1,
  	acE: acE$1,
  	Acirc: Acirc$2,
  	acirc: acirc$2,
  	acute: acute$2,
  	Acy: Acy$1,
  	acy: acy$1,
  	AElig: AElig$2,
  	aelig: aelig$2,
  	af: af$1,
  	Afr: Afr$1,
  	afr: afr$1,
  	Agrave: Agrave$2,
  	agrave: agrave$2,
  	alefsym: alefsym$1,
  	aleph: aleph$1,
  	Alpha: Alpha$1,
  	alpha: alpha$1,
  	Amacr: Amacr$1,
  	amacr: amacr$1,
  	amalg: amalg$1,
  	amp: amp$1$1,
  	AMP: AMP$2,
  	andand: andand$1,
  	And: And$1,
  	and: and$1,
  	andd: andd$1,
  	andslope: andslope$1,
  	andv: andv$1,
  	ang: ang$1,
  	ange: ange$1,
  	angle: angle$1,
  	angmsdaa: angmsdaa$1,
  	angmsdab: angmsdab$1,
  	angmsdac: angmsdac$1,
  	angmsdad: angmsdad$1,
  	angmsdae: angmsdae$1,
  	angmsdaf: angmsdaf$1,
  	angmsdag: angmsdag$1,
  	angmsdah: angmsdah$1,
  	angmsd: angmsd$1,
  	angrt: angrt$1,
  	angrtvb: angrtvb$1,
  	angrtvbd: angrtvbd$1,
  	angsph: angsph$1,
  	angst: angst$1,
  	angzarr: angzarr$1,
  	Aogon: Aogon$1,
  	aogon: aogon$1,
  	Aopf: Aopf$1,
  	aopf: aopf$1,
  	apacir: apacir$1,
  	ap: ap$1,
  	apE: apE$1,
  	ape: ape$1,
  	apid: apid$1,
  	apos: apos$1$1,
  	ApplyFunction: ApplyFunction$1,
  	approx: approx$1,
  	approxeq: approxeq$1,
  	Aring: Aring$2,
  	aring: aring$2,
  	Ascr: Ascr$1,
  	ascr: ascr$1,
  	Assign: Assign$1,
  	ast: ast$1,
  	asymp: asymp$1,
  	asympeq: asympeq$1,
  	Atilde: Atilde$2,
  	atilde: atilde$2,
  	Auml: Auml$2,
  	auml: auml$2,
  	awconint: awconint$1,
  	awint: awint$1,
  	backcong: backcong$1,
  	backepsilon: backepsilon$1,
  	backprime: backprime$1,
  	backsim: backsim$1,
  	backsimeq: backsimeq$1,
  	Backslash: Backslash$1,
  	Barv: Barv$1,
  	barvee: barvee$1,
  	barwed: barwed$1,
  	Barwed: Barwed$1,
  	barwedge: barwedge$1,
  	bbrk: bbrk$1,
  	bbrktbrk: bbrktbrk$1,
  	bcong: bcong$1,
  	Bcy: Bcy$1,
  	bcy: bcy$1,
  	bdquo: bdquo$1,
  	becaus: becaus$1,
  	because: because$1,
  	Because: Because$1,
  	bemptyv: bemptyv$1,
  	bepsi: bepsi$1,
  	bernou: bernou$1,
  	Bernoullis: Bernoullis$1,
  	Beta: Beta$1,
  	beta: beta$1,
  	beth: beth$1,
  	between: between$1,
  	Bfr: Bfr$1,
  	bfr: bfr$1,
  	bigcap: bigcap$1,
  	bigcirc: bigcirc$1,
  	bigcup: bigcup$1,
  	bigodot: bigodot$1,
  	bigoplus: bigoplus$1,
  	bigotimes: bigotimes$1,
  	bigsqcup: bigsqcup$1,
  	bigstar: bigstar$1,
  	bigtriangledown: bigtriangledown$1,
  	bigtriangleup: bigtriangleup$1,
  	biguplus: biguplus$1,
  	bigvee: bigvee$1,
  	bigwedge: bigwedge$1,
  	bkarow: bkarow$1,
  	blacklozenge: blacklozenge$1,
  	blacksquare: blacksquare$1,
  	blacktriangle: blacktriangle$1,
  	blacktriangledown: blacktriangledown$1,
  	blacktriangleleft: blacktriangleleft$1,
  	blacktriangleright: blacktriangleright$1,
  	blank: blank$1,
  	blk12: blk12$1,
  	blk14: blk14$1,
  	blk34: blk34$1,
  	block: block$1,
  	bne: bne$1,
  	bnequiv: bnequiv$1,
  	bNot: bNot$1,
  	bnot: bnot$1,
  	Bopf: Bopf$1,
  	bopf: bopf$1,
  	bot: bot$1,
  	bottom: bottom$1,
  	bowtie: bowtie$1,
  	boxbox: boxbox$1,
  	boxdl: boxdl$1,
  	boxdL: boxdL$1,
  	boxDl: boxDl$1,
  	boxDL: boxDL$1,
  	boxdr: boxdr$1,
  	boxdR: boxdR$1,
  	boxDr: boxDr$1,
  	boxDR: boxDR$1,
  	boxh: boxh$1,
  	boxH: boxH$1,
  	boxhd: boxhd$1,
  	boxHd: boxHd$1,
  	boxhD: boxhD$1,
  	boxHD: boxHD$1,
  	boxhu: boxhu$1,
  	boxHu: boxHu$1,
  	boxhU: boxhU$1,
  	boxHU: boxHU$1,
  	boxminus: boxminus$1,
  	boxplus: boxplus$1,
  	boxtimes: boxtimes$1,
  	boxul: boxul$1,
  	boxuL: boxuL$1,
  	boxUl: boxUl$1,
  	boxUL: boxUL$1,
  	boxur: boxur$1,
  	boxuR: boxuR$1,
  	boxUr: boxUr$1,
  	boxUR: boxUR$1,
  	boxv: boxv$1,
  	boxV: boxV$1,
  	boxvh: boxvh$1,
  	boxvH: boxvH$1,
  	boxVh: boxVh$1,
  	boxVH: boxVH$1,
  	boxvl: boxvl$1,
  	boxvL: boxvL$1,
  	boxVl: boxVl$1,
  	boxVL: boxVL$1,
  	boxvr: boxvr$1,
  	boxvR: boxvR$1,
  	boxVr: boxVr$1,
  	boxVR: boxVR$1,
  	bprime: bprime$1,
  	breve: breve$1,
  	Breve: Breve$1,
  	brvbar: brvbar$2,
  	bscr: bscr$1,
  	Bscr: Bscr$1,
  	bsemi: bsemi$1,
  	bsim: bsim$1,
  	bsime: bsime$1,
  	bsolb: bsolb$1,
  	bsol: bsol$1,
  	bsolhsub: bsolhsub$1,
  	bull: bull$1,
  	bullet: bullet$1,
  	bump: bump$1,
  	bumpE: bumpE$1,
  	bumpe: bumpe$1,
  	Bumpeq: Bumpeq$1,
  	bumpeq: bumpeq$1,
  	Cacute: Cacute$1,
  	cacute: cacute$1,
  	capand: capand$1,
  	capbrcup: capbrcup$1,
  	capcap: capcap$1,
  	cap: cap$1,
  	Cap: Cap$1,
  	capcup: capcup$1,
  	capdot: capdot$1,
  	CapitalDifferentialD: CapitalDifferentialD$1,
  	caps: caps$1,
  	caret: caret$1,
  	caron: caron$1,
  	Cayleys: Cayleys$1,
  	ccaps: ccaps$1,
  	Ccaron: Ccaron$1,
  	ccaron: ccaron$1,
  	Ccedil: Ccedil$2,
  	ccedil: ccedil$2,
  	Ccirc: Ccirc$1,
  	ccirc: ccirc$1,
  	Cconint: Cconint$1,
  	ccups: ccups$1,
  	ccupssm: ccupssm$1,
  	Cdot: Cdot$1,
  	cdot: cdot$1,
  	cedil: cedil$2,
  	Cedilla: Cedilla$1,
  	cemptyv: cemptyv$1,
  	cent: cent$2,
  	centerdot: centerdot$1,
  	CenterDot: CenterDot$1,
  	cfr: cfr$1,
  	Cfr: Cfr$1,
  	CHcy: CHcy$1,
  	chcy: chcy$1,
  	check: check$1,
  	checkmark: checkmark$1,
  	Chi: Chi$1,
  	chi: chi$1,
  	circ: circ$1,
  	circeq: circeq$1,
  	circlearrowleft: circlearrowleft$1,
  	circlearrowright: circlearrowright$1,
  	circledast: circledast$1,
  	circledcirc: circledcirc$1,
  	circleddash: circleddash$1,
  	CircleDot: CircleDot$1,
  	circledR: circledR$1,
  	circledS: circledS$1,
  	CircleMinus: CircleMinus$1,
  	CirclePlus: CirclePlus$1,
  	CircleTimes: CircleTimes$1,
  	cir: cir$1,
  	cirE: cirE$1,
  	cire: cire$1,
  	cirfnint: cirfnint$1,
  	cirmid: cirmid$1,
  	cirscir: cirscir$1,
  	ClockwiseContourIntegral: ClockwiseContourIntegral$1,
  	CloseCurlyDoubleQuote: CloseCurlyDoubleQuote$1,
  	CloseCurlyQuote: CloseCurlyQuote$1,
  	clubs: clubs$1,
  	clubsuit: clubsuit$1,
  	colon: colon$1,
  	Colon: Colon$1,
  	Colone: Colone$1,
  	colone: colone$1,
  	coloneq: coloneq$1,
  	comma: comma$1,
  	commat: commat$1,
  	comp: comp$1,
  	compfn: compfn$1,
  	complement: complement$1,
  	complexes: complexes$1,
  	cong: cong$1,
  	congdot: congdot$1,
  	Congruent: Congruent$1,
  	conint: conint$1,
  	Conint: Conint$1,
  	ContourIntegral: ContourIntegral$1,
  	copf: copf$1,
  	Copf: Copf$1,
  	coprod: coprod$1,
  	Coproduct: Coproduct$1,
  	copy: copy$2,
  	COPY: COPY$2,
  	copysr: copysr$1,
  	CounterClockwiseContourIntegral: CounterClockwiseContourIntegral$1,
  	crarr: crarr$1,
  	cross: cross$1,
  	Cross: Cross$1,
  	Cscr: Cscr$1,
  	cscr: cscr$1,
  	csub: csub$1,
  	csube: csube$1,
  	csup: csup$1,
  	csupe: csupe$1,
  	ctdot: ctdot$1,
  	cudarrl: cudarrl$1,
  	cudarrr: cudarrr$1,
  	cuepr: cuepr$1,
  	cuesc: cuesc$1,
  	cularr: cularr$1,
  	cularrp: cularrp$1,
  	cupbrcap: cupbrcap$1,
  	cupcap: cupcap$1,
  	CupCap: CupCap$1,
  	cup: cup$1,
  	Cup: Cup$1,
  	cupcup: cupcup$1,
  	cupdot: cupdot$1,
  	cupor: cupor$1,
  	cups: cups$1,
  	curarr: curarr$1,
  	curarrm: curarrm$1,
  	curlyeqprec: curlyeqprec$1,
  	curlyeqsucc: curlyeqsucc$1,
  	curlyvee: curlyvee$1,
  	curlywedge: curlywedge$1,
  	curren: curren$2,
  	curvearrowleft: curvearrowleft$1,
  	curvearrowright: curvearrowright$1,
  	cuvee: cuvee$1,
  	cuwed: cuwed$1,
  	cwconint: cwconint$1,
  	cwint: cwint$1,
  	cylcty: cylcty$1,
  	dagger: dagger$1,
  	Dagger: Dagger$1,
  	daleth: daleth$1,
  	darr: darr$1,
  	Darr: Darr$1,
  	dArr: dArr$1,
  	dash: dash$1,
  	Dashv: Dashv$1,
  	dashv: dashv$1,
  	dbkarow: dbkarow$1,
  	dblac: dblac$1,
  	Dcaron: Dcaron$1,
  	dcaron: dcaron$1,
  	Dcy: Dcy$1,
  	dcy: dcy$1,
  	ddagger: ddagger$1,
  	ddarr: ddarr$1,
  	DD: DD$1,
  	dd: dd$1,
  	DDotrahd: DDotrahd$1,
  	ddotseq: ddotseq$1,
  	deg: deg$2,
  	Del: Del$1,
  	Delta: Delta$1,
  	delta: delta$1,
  	demptyv: demptyv$1,
  	dfisht: dfisht$1,
  	Dfr: Dfr$1,
  	dfr: dfr$1,
  	dHar: dHar$1,
  	dharl: dharl$1,
  	dharr: dharr$1,
  	DiacriticalAcute: DiacriticalAcute$1,
  	DiacriticalDot: DiacriticalDot$1,
  	DiacriticalDoubleAcute: DiacriticalDoubleAcute$1,
  	DiacriticalGrave: DiacriticalGrave$1,
  	DiacriticalTilde: DiacriticalTilde$1,
  	diam: diam$1,
  	diamond: diamond$1,
  	Diamond: Diamond$1,
  	diamondsuit: diamondsuit$1,
  	diams: diams$1,
  	die: die$1,
  	DifferentialD: DifferentialD$1,
  	digamma: digamma$1,
  	disin: disin$1,
  	div: div$1,
  	divide: divide$2,
  	divideontimes: divideontimes$1,
  	divonx: divonx$1,
  	DJcy: DJcy$1,
  	djcy: djcy$1,
  	dlcorn: dlcorn$1,
  	dlcrop: dlcrop$1,
  	dollar: dollar$1,
  	Dopf: Dopf$1,
  	dopf: dopf$1,
  	Dot: Dot$1,
  	dot: dot$1,
  	DotDot: DotDot$1,
  	doteq: doteq$1,
  	doteqdot: doteqdot$1,
  	DotEqual: DotEqual$1,
  	dotminus: dotminus$1,
  	dotplus: dotplus$1,
  	dotsquare: dotsquare$1,
  	doublebarwedge: doublebarwedge$1,
  	DoubleContourIntegral: DoubleContourIntegral$1,
  	DoubleDot: DoubleDot$1,
  	DoubleDownArrow: DoubleDownArrow$1,
  	DoubleLeftArrow: DoubleLeftArrow$1,
  	DoubleLeftRightArrow: DoubleLeftRightArrow$1,
  	DoubleLeftTee: DoubleLeftTee$1,
  	DoubleLongLeftArrow: DoubleLongLeftArrow$1,
  	DoubleLongLeftRightArrow: DoubleLongLeftRightArrow$1,
  	DoubleLongRightArrow: DoubleLongRightArrow$1,
  	DoubleRightArrow: DoubleRightArrow$1,
  	DoubleRightTee: DoubleRightTee$1,
  	DoubleUpArrow: DoubleUpArrow$1,
  	DoubleUpDownArrow: DoubleUpDownArrow$1,
  	DoubleVerticalBar: DoubleVerticalBar$1,
  	DownArrowBar: DownArrowBar$1,
  	downarrow: downarrow$1,
  	DownArrow: DownArrow$1,
  	Downarrow: Downarrow$1,
  	DownArrowUpArrow: DownArrowUpArrow$1,
  	DownBreve: DownBreve$1,
  	downdownarrows: downdownarrows$1,
  	downharpoonleft: downharpoonleft$1,
  	downharpoonright: downharpoonright$1,
  	DownLeftRightVector: DownLeftRightVector$1,
  	DownLeftTeeVector: DownLeftTeeVector$1,
  	DownLeftVectorBar: DownLeftVectorBar$1,
  	DownLeftVector: DownLeftVector$1,
  	DownRightTeeVector: DownRightTeeVector$1,
  	DownRightVectorBar: DownRightVectorBar$1,
  	DownRightVector: DownRightVector$1,
  	DownTeeArrow: DownTeeArrow$1,
  	DownTee: DownTee$1,
  	drbkarow: drbkarow$1,
  	drcorn: drcorn$1,
  	drcrop: drcrop$1,
  	Dscr: Dscr$1,
  	dscr: dscr$1,
  	DScy: DScy$1,
  	dscy: dscy$1,
  	dsol: dsol$1,
  	Dstrok: Dstrok$1,
  	dstrok: dstrok$1,
  	dtdot: dtdot$1,
  	dtri: dtri$1,
  	dtrif: dtrif$1,
  	duarr: duarr$1,
  	duhar: duhar$1,
  	dwangle: dwangle$1,
  	DZcy: DZcy$1,
  	dzcy: dzcy$1,
  	dzigrarr: dzigrarr$1,
  	Eacute: Eacute$2,
  	eacute: eacute$2,
  	easter: easter$1,
  	Ecaron: Ecaron$1,
  	ecaron: ecaron$1,
  	Ecirc: Ecirc$2,
  	ecirc: ecirc$2,
  	ecir: ecir$1,
  	ecolon: ecolon$1,
  	Ecy: Ecy$1,
  	ecy: ecy$1,
  	eDDot: eDDot$1,
  	Edot: Edot$1,
  	edot: edot$1,
  	eDot: eDot$1,
  	ee: ee$1,
  	efDot: efDot$1,
  	Efr: Efr$1,
  	efr: efr$1,
  	eg: eg$1,
  	Egrave: Egrave$2,
  	egrave: egrave$2,
  	egs: egs$1,
  	egsdot: egsdot$1,
  	el: el$1,
  	Element: Element$1,
  	elinters: elinters$1,
  	ell: ell$1,
  	els: els$1,
  	elsdot: elsdot$1,
  	Emacr: Emacr$1,
  	emacr: emacr$1,
  	empty: empty$1,
  	emptyset: emptyset$1,
  	EmptySmallSquare: EmptySmallSquare$1,
  	emptyv: emptyv$1,
  	EmptyVerySmallSquare: EmptyVerySmallSquare$1,
  	emsp13: emsp13$1,
  	emsp14: emsp14$1,
  	emsp: emsp$1,
  	ENG: ENG$1,
  	eng: eng$1,
  	ensp: ensp$1,
  	Eogon: Eogon$1,
  	eogon: eogon$1,
  	Eopf: Eopf$1,
  	eopf: eopf$1,
  	epar: epar$1,
  	eparsl: eparsl$1,
  	eplus: eplus$1,
  	epsi: epsi$1,
  	Epsilon: Epsilon$1,
  	epsilon: epsilon$1,
  	epsiv: epsiv$1,
  	eqcirc: eqcirc$1,
  	eqcolon: eqcolon$1,
  	eqsim: eqsim$1,
  	eqslantgtr: eqslantgtr$1,
  	eqslantless: eqslantless$1,
  	Equal: Equal$1,
  	equals: equals$1,
  	EqualTilde: EqualTilde$1,
  	equest: equest$1,
  	Equilibrium: Equilibrium$1,
  	equiv: equiv$1,
  	equivDD: equivDD$1,
  	eqvparsl: eqvparsl$1,
  	erarr: erarr$1,
  	erDot: erDot$1,
  	escr: escr$1,
  	Escr: Escr$1,
  	esdot: esdot$1,
  	Esim: Esim$1,
  	esim: esim$1,
  	Eta: Eta$1,
  	eta: eta$1,
  	ETH: ETH$2,
  	eth: eth$2,
  	Euml: Euml$2,
  	euml: euml$2,
  	euro: euro$1,
  	excl: excl$1,
  	exist: exist$1,
  	Exists: Exists$1,
  	expectation: expectation$1,
  	exponentiale: exponentiale$1,
  	ExponentialE: ExponentialE$1,
  	fallingdotseq: fallingdotseq$1,
  	Fcy: Fcy$1,
  	fcy: fcy$1,
  	female: female$1,
  	ffilig: ffilig$1,
  	fflig: fflig$1,
  	ffllig: ffllig$1,
  	Ffr: Ffr$1,
  	ffr: ffr$1,
  	filig: filig$1,
  	FilledSmallSquare: FilledSmallSquare$1,
  	FilledVerySmallSquare: FilledVerySmallSquare$1,
  	fjlig: fjlig$1,
  	flat: flat$1,
  	fllig: fllig$1,
  	fltns: fltns$1,
  	fnof: fnof$1,
  	Fopf: Fopf$1,
  	fopf: fopf$1,
  	forall: forall$1,
  	ForAll: ForAll$1,
  	fork: fork$1,
  	forkv: forkv$1,
  	Fouriertrf: Fouriertrf$1,
  	fpartint: fpartint$1,
  	frac12: frac12$2,
  	frac13: frac13$1,
  	frac14: frac14$2,
  	frac15: frac15$1,
  	frac16: frac16$1,
  	frac18: frac18$1,
  	frac23: frac23$1,
  	frac25: frac25$1,
  	frac34: frac34$2,
  	frac35: frac35$1,
  	frac38: frac38$1,
  	frac45: frac45$1,
  	frac56: frac56$1,
  	frac58: frac58$1,
  	frac78: frac78$1,
  	frasl: frasl$1,
  	frown: frown$1,
  	fscr: fscr$1,
  	Fscr: Fscr$1,
  	gacute: gacute$1,
  	Gamma: Gamma$1,
  	gamma: gamma$1,
  	Gammad: Gammad$1,
  	gammad: gammad$1,
  	gap: gap$1,
  	Gbreve: Gbreve$1,
  	gbreve: gbreve$1,
  	Gcedil: Gcedil$1,
  	Gcirc: Gcirc$1,
  	gcirc: gcirc$1,
  	Gcy: Gcy$1,
  	gcy: gcy$1,
  	Gdot: Gdot$1,
  	gdot: gdot$1,
  	ge: ge$1,
  	gE: gE$1,
  	gEl: gEl$1,
  	gel: gel$1,
  	geq: geq$1,
  	geqq: geqq$1,
  	geqslant: geqslant$1,
  	gescc: gescc$1,
  	ges: ges$1,
  	gesdot: gesdot$1,
  	gesdoto: gesdoto$1,
  	gesdotol: gesdotol$1,
  	gesl: gesl$1,
  	gesles: gesles$1,
  	Gfr: Gfr$1,
  	gfr: gfr$1,
  	gg: gg$1,
  	Gg: Gg$1,
  	ggg: ggg$1,
  	gimel: gimel$1,
  	GJcy: GJcy$1,
  	gjcy: gjcy$1,
  	gla: gla$1,
  	gl: gl$1,
  	glE: glE$1,
  	glj: glj$1,
  	gnap: gnap$1,
  	gnapprox: gnapprox$1,
  	gne: gne$1,
  	gnE: gnE$1,
  	gneq: gneq$1,
  	gneqq: gneqq$1,
  	gnsim: gnsim$1,
  	Gopf: Gopf$1,
  	gopf: gopf$1,
  	grave: grave$1,
  	GreaterEqual: GreaterEqual$1,
  	GreaterEqualLess: GreaterEqualLess$1,
  	GreaterFullEqual: GreaterFullEqual$1,
  	GreaterGreater: GreaterGreater$1,
  	GreaterLess: GreaterLess$1,
  	GreaterSlantEqual: GreaterSlantEqual$1,
  	GreaterTilde: GreaterTilde$1,
  	Gscr: Gscr$1,
  	gscr: gscr$1,
  	gsim: gsim$1,
  	gsime: gsime$1,
  	gsiml: gsiml$1,
  	gtcc: gtcc$1,
  	gtcir: gtcir$1,
  	gt: gt$1$1,
  	GT: GT$2,
  	Gt: Gt$1,
  	gtdot: gtdot$1,
  	gtlPar: gtlPar$1,
  	gtquest: gtquest$1,
  	gtrapprox: gtrapprox$1,
  	gtrarr: gtrarr$1,
  	gtrdot: gtrdot$1,
  	gtreqless: gtreqless$1,
  	gtreqqless: gtreqqless$1,
  	gtrless: gtrless$1,
  	gtrsim: gtrsim$1,
  	gvertneqq: gvertneqq$1,
  	gvnE: gvnE$1,
  	Hacek: Hacek$1,
  	hairsp: hairsp$1,
  	half: half$1,
  	hamilt: hamilt$1,
  	HARDcy: HARDcy$1,
  	hardcy: hardcy$1,
  	harrcir: harrcir$1,
  	harr: harr$1,
  	hArr: hArr$1,
  	harrw: harrw$1,
  	Hat: Hat$1,
  	hbar: hbar$1,
  	Hcirc: Hcirc$1,
  	hcirc: hcirc$1,
  	hearts: hearts$1,
  	heartsuit: heartsuit$1,
  	hellip: hellip$1,
  	hercon: hercon$1,
  	hfr: hfr$1,
  	Hfr: Hfr$1,
  	HilbertSpace: HilbertSpace$1,
  	hksearow: hksearow$1,
  	hkswarow: hkswarow$1,
  	hoarr: hoarr$1,
  	homtht: homtht$1,
  	hookleftarrow: hookleftarrow$1,
  	hookrightarrow: hookrightarrow$1,
  	hopf: hopf$1,
  	Hopf: Hopf$1,
  	horbar: horbar$1,
  	HorizontalLine: HorizontalLine$1,
  	hscr: hscr$1,
  	Hscr: Hscr$1,
  	hslash: hslash$1,
  	Hstrok: Hstrok$1,
  	hstrok: hstrok$1,
  	HumpDownHump: HumpDownHump$1,
  	HumpEqual: HumpEqual$1,
  	hybull: hybull$1,
  	hyphen: hyphen$1,
  	Iacute: Iacute$2,
  	iacute: iacute$2,
  	ic: ic$1,
  	Icirc: Icirc$2,
  	icirc: icirc$2,
  	Icy: Icy$1,
  	icy: icy$1,
  	Idot: Idot$1,
  	IEcy: IEcy$1,
  	iecy: iecy$1,
  	iexcl: iexcl$2,
  	iff: iff$1,
  	ifr: ifr$1,
  	Ifr: Ifr$1,
  	Igrave: Igrave$2,
  	igrave: igrave$2,
  	ii: ii$1,
  	iiiint: iiiint$1,
  	iiint: iiint$1,
  	iinfin: iinfin$1,
  	iiota: iiota$1,
  	IJlig: IJlig$1,
  	ijlig: ijlig$1,
  	Imacr: Imacr$1,
  	imacr: imacr$1,
  	image: image$1,
  	ImaginaryI: ImaginaryI$1,
  	imagline: imagline$1,
  	imagpart: imagpart$1,
  	imath: imath$1,
  	Im: Im$1,
  	imof: imof$1,
  	imped: imped$1,
  	Implies: Implies$1,
  	incare: incare$1,
  	infin: infin$1,
  	infintie: infintie$1,
  	inodot: inodot$1,
  	intcal: intcal$1,
  	int: int$1,
  	Int: Int$1,
  	integers: integers$1,
  	Integral: Integral$1,
  	intercal: intercal$1,
  	Intersection: Intersection$1,
  	intlarhk: intlarhk$1,
  	intprod: intprod$1,
  	InvisibleComma: InvisibleComma$1,
  	InvisibleTimes: InvisibleTimes$1,
  	IOcy: IOcy$1,
  	iocy: iocy$1,
  	Iogon: Iogon$1,
  	iogon: iogon$1,
  	Iopf: Iopf$1,
  	iopf: iopf$1,
  	Iota: Iota$1,
  	iota: iota$1,
  	iprod: iprod$1,
  	iquest: iquest$2,
  	iscr: iscr$1,
  	Iscr: Iscr$1,
  	isin: isin$1,
  	isindot: isindot$1,
  	isinE: isinE$1,
  	isins: isins$1,
  	isinsv: isinsv$1,
  	isinv: isinv$1,
  	it: it$1,
  	Itilde: Itilde$1,
  	itilde: itilde$1,
  	Iukcy: Iukcy$1,
  	iukcy: iukcy$1,
  	Iuml: Iuml$2,
  	iuml: iuml$2,
  	Jcirc: Jcirc$1,
  	jcirc: jcirc$1,
  	Jcy: Jcy$1,
  	jcy: jcy$1,
  	Jfr: Jfr$1,
  	jfr: jfr$1,
  	jmath: jmath$1,
  	Jopf: Jopf$1,
  	jopf: jopf$1,
  	Jscr: Jscr$1,
  	jscr: jscr$1,
  	Jsercy: Jsercy$1,
  	jsercy: jsercy$1,
  	Jukcy: Jukcy$1,
  	jukcy: jukcy$1,
  	Kappa: Kappa$1,
  	kappa: kappa$1,
  	kappav: kappav$1,
  	Kcedil: Kcedil$1,
  	kcedil: kcedil$1,
  	Kcy: Kcy$1,
  	kcy: kcy$1,
  	Kfr: Kfr$1,
  	kfr: kfr$1,
  	kgreen: kgreen$1,
  	KHcy: KHcy$1,
  	khcy: khcy$1,
  	KJcy: KJcy$1,
  	kjcy: kjcy$1,
  	Kopf: Kopf$1,
  	kopf: kopf$1,
  	Kscr: Kscr$1,
  	kscr: kscr$1,
  	lAarr: lAarr$1,
  	Lacute: Lacute$1,
  	lacute: lacute$1,
  	laemptyv: laemptyv$1,
  	lagran: lagran$1,
  	Lambda: Lambda$1,
  	lambda: lambda$1,
  	lang: lang$1,
  	Lang: Lang$1,
  	langd: langd$1,
  	langle: langle$1,
  	lap: lap$1,
  	Laplacetrf: Laplacetrf$1,
  	laquo: laquo$2,
  	larrb: larrb$1,
  	larrbfs: larrbfs$1,
  	larr: larr$1,
  	Larr: Larr$1,
  	lArr: lArr$1,
  	larrfs: larrfs$1,
  	larrhk: larrhk$1,
  	larrlp: larrlp$1,
  	larrpl: larrpl$1,
  	larrsim: larrsim$1,
  	larrtl: larrtl$1,
  	latail: latail$1,
  	lAtail: lAtail$1,
  	lat: lat$1,
  	late: late$1,
  	lates: lates$1,
  	lbarr: lbarr$1,
  	lBarr: lBarr$1,
  	lbbrk: lbbrk$1,
  	lbrace: lbrace$1,
  	lbrack: lbrack$1,
  	lbrke: lbrke$1,
  	lbrksld: lbrksld$1,
  	lbrkslu: lbrkslu$1,
  	Lcaron: Lcaron$1,
  	lcaron: lcaron$1,
  	Lcedil: Lcedil$1,
  	lcedil: lcedil$1,
  	lceil: lceil$1,
  	lcub: lcub$1,
  	Lcy: Lcy$1,
  	lcy: lcy$1,
  	ldca: ldca$1,
  	ldquo: ldquo$1,
  	ldquor: ldquor$1,
  	ldrdhar: ldrdhar$1,
  	ldrushar: ldrushar$1,
  	ldsh: ldsh$1,
  	le: le$1,
  	lE: lE$1,
  	LeftAngleBracket: LeftAngleBracket$1,
  	LeftArrowBar: LeftArrowBar$1,
  	leftarrow: leftarrow$1,
  	LeftArrow: LeftArrow$1,
  	Leftarrow: Leftarrow$1,
  	LeftArrowRightArrow: LeftArrowRightArrow$1,
  	leftarrowtail: leftarrowtail$1,
  	LeftCeiling: LeftCeiling$1,
  	LeftDoubleBracket: LeftDoubleBracket$1,
  	LeftDownTeeVector: LeftDownTeeVector$1,
  	LeftDownVectorBar: LeftDownVectorBar$1,
  	LeftDownVector: LeftDownVector$1,
  	LeftFloor: LeftFloor$1,
  	leftharpoondown: leftharpoondown$1,
  	leftharpoonup: leftharpoonup$1,
  	leftleftarrows: leftleftarrows$1,
  	leftrightarrow: leftrightarrow$1,
  	LeftRightArrow: LeftRightArrow$1,
  	Leftrightarrow: Leftrightarrow$1,
  	leftrightarrows: leftrightarrows$1,
  	leftrightharpoons: leftrightharpoons$1,
  	leftrightsquigarrow: leftrightsquigarrow$1,
  	LeftRightVector: LeftRightVector$1,
  	LeftTeeArrow: LeftTeeArrow$1,
  	LeftTee: LeftTee$1,
  	LeftTeeVector: LeftTeeVector$1,
  	leftthreetimes: leftthreetimes$1,
  	LeftTriangleBar: LeftTriangleBar$1,
  	LeftTriangle: LeftTriangle$1,
  	LeftTriangleEqual: LeftTriangleEqual$1,
  	LeftUpDownVector: LeftUpDownVector$1,
  	LeftUpTeeVector: LeftUpTeeVector$1,
  	LeftUpVectorBar: LeftUpVectorBar$1,
  	LeftUpVector: LeftUpVector$1,
  	LeftVectorBar: LeftVectorBar$1,
  	LeftVector: LeftVector$1,
  	lEg: lEg$1,
  	leg: leg$1,
  	leq: leq$1,
  	leqq: leqq$1,
  	leqslant: leqslant$1,
  	lescc: lescc$1,
  	les: les$1,
  	lesdot: lesdot$1,
  	lesdoto: lesdoto$1,
  	lesdotor: lesdotor$1,
  	lesg: lesg$1,
  	lesges: lesges$1,
  	lessapprox: lessapprox$1,
  	lessdot: lessdot$1,
  	lesseqgtr: lesseqgtr$1,
  	lesseqqgtr: lesseqqgtr$1,
  	LessEqualGreater: LessEqualGreater$1,
  	LessFullEqual: LessFullEqual$1,
  	LessGreater: LessGreater$1,
  	lessgtr: lessgtr$1,
  	LessLess: LessLess$1,
  	lesssim: lesssim$1,
  	LessSlantEqual: LessSlantEqual$1,
  	LessTilde: LessTilde$1,
  	lfisht: lfisht$1,
  	lfloor: lfloor$1,
  	Lfr: Lfr$1,
  	lfr: lfr$1,
  	lg: lg$1,
  	lgE: lgE$1,
  	lHar: lHar$1,
  	lhard: lhard$1,
  	lharu: lharu$1,
  	lharul: lharul$1,
  	lhblk: lhblk$1,
  	LJcy: LJcy$1,
  	ljcy: ljcy$1,
  	llarr: llarr$1,
  	ll: ll$1,
  	Ll: Ll$1,
  	llcorner: llcorner$1,
  	Lleftarrow: Lleftarrow$1,
  	llhard: llhard$1,
  	lltri: lltri$1,
  	Lmidot: Lmidot$1,
  	lmidot: lmidot$1,
  	lmoustache: lmoustache$1,
  	lmoust: lmoust$1,
  	lnap: lnap$1,
  	lnapprox: lnapprox$1,
  	lne: lne$1,
  	lnE: lnE$1,
  	lneq: lneq$1,
  	lneqq: lneqq$1,
  	lnsim: lnsim$1,
  	loang: loang$1,
  	loarr: loarr$1,
  	lobrk: lobrk$1,
  	longleftarrow: longleftarrow$1,
  	LongLeftArrow: LongLeftArrow$1,
  	Longleftarrow: Longleftarrow$1,
  	longleftrightarrow: longleftrightarrow$1,
  	LongLeftRightArrow: LongLeftRightArrow$1,
  	Longleftrightarrow: Longleftrightarrow$1,
  	longmapsto: longmapsto$1,
  	longrightarrow: longrightarrow$1,
  	LongRightArrow: LongRightArrow$1,
  	Longrightarrow: Longrightarrow$1,
  	looparrowleft: looparrowleft$1,
  	looparrowright: looparrowright$1,
  	lopar: lopar$1,
  	Lopf: Lopf$1,
  	lopf: lopf$1,
  	loplus: loplus$1,
  	lotimes: lotimes$1,
  	lowast: lowast$1,
  	lowbar: lowbar$1,
  	LowerLeftArrow: LowerLeftArrow$1,
  	LowerRightArrow: LowerRightArrow$1,
  	loz: loz$1,
  	lozenge: lozenge$1,
  	lozf: lozf$1,
  	lpar: lpar$1,
  	lparlt: lparlt$1,
  	lrarr: lrarr$1,
  	lrcorner: lrcorner$1,
  	lrhar: lrhar$1,
  	lrhard: lrhard$1,
  	lrm: lrm$1,
  	lrtri: lrtri$1,
  	lsaquo: lsaquo$1,
  	lscr: lscr$1,
  	Lscr: Lscr$1,
  	lsh: lsh$1,
  	Lsh: Lsh$1,
  	lsim: lsim$1,
  	lsime: lsime$1,
  	lsimg: lsimg$1,
  	lsqb: lsqb$1,
  	lsquo: lsquo$1,
  	lsquor: lsquor$1,
  	Lstrok: Lstrok$1,
  	lstrok: lstrok$1,
  	ltcc: ltcc$1,
  	ltcir: ltcir$1,
  	lt: lt$1$1,
  	LT: LT$2,
  	Lt: Lt$1,
  	ltdot: ltdot$1,
  	lthree: lthree$1,
  	ltimes: ltimes$1,
  	ltlarr: ltlarr$1,
  	ltquest: ltquest$1,
  	ltri: ltri$1,
  	ltrie: ltrie$1,
  	ltrif: ltrif$1,
  	ltrPar: ltrPar$1,
  	lurdshar: lurdshar$1,
  	luruhar: luruhar$1,
  	lvertneqq: lvertneqq$1,
  	lvnE: lvnE$1,
  	macr: macr$2,
  	male: male$1,
  	malt: malt$1,
  	maltese: maltese$1,
  	map: map$1$1,
  	mapsto: mapsto$1,
  	mapstodown: mapstodown$1,
  	mapstoleft: mapstoleft$1,
  	mapstoup: mapstoup$1,
  	marker: marker$1,
  	mcomma: mcomma$1,
  	Mcy: Mcy$1,
  	mcy: mcy$1,
  	mdash: mdash$1,
  	mDDot: mDDot$1,
  	measuredangle: measuredangle$1,
  	MediumSpace: MediumSpace$1,
  	Mellintrf: Mellintrf$1,
  	Mfr: Mfr$1,
  	mfr: mfr$1,
  	mho: mho$1,
  	micro: micro$2,
  	midast: midast$1,
  	midcir: midcir$1,
  	mid: mid$1,
  	middot: middot$2,
  	minusb: minusb$1,
  	minus: minus$1,
  	minusd: minusd$1,
  	minusdu: minusdu$1,
  	MinusPlus: MinusPlus$1,
  	mlcp: mlcp$1,
  	mldr: mldr$1,
  	mnplus: mnplus$1,
  	models: models$1,
  	Mopf: Mopf$1,
  	mopf: mopf$1,
  	mp: mp$1,
  	mscr: mscr$1,
  	Mscr: Mscr$1,
  	mstpos: mstpos$1,
  	Mu: Mu$1,
  	mu: mu$1,
  	multimap: multimap$1,
  	mumap: mumap$1,
  	nabla: nabla$1,
  	Nacute: Nacute$1,
  	nacute: nacute$1,
  	nang: nang$1,
  	nap: nap$1,
  	napE: napE$1,
  	napid: napid$1,
  	napos: napos$1,
  	napprox: napprox$1,
  	natural: natural$1,
  	naturals: naturals$1,
  	natur: natur$1,
  	nbsp: nbsp$2,
  	nbump: nbump$1,
  	nbumpe: nbumpe$1,
  	ncap: ncap$1,
  	Ncaron: Ncaron$1,
  	ncaron: ncaron$1,
  	Ncedil: Ncedil$1,
  	ncedil: ncedil$1,
  	ncong: ncong$1,
  	ncongdot: ncongdot$1,
  	ncup: ncup$1,
  	Ncy: Ncy$1,
  	ncy: ncy$1,
  	ndash: ndash$1,
  	nearhk: nearhk$1,
  	nearr: nearr$1,
  	neArr: neArr$1,
  	nearrow: nearrow$1,
  	ne: ne$1,
  	nedot: nedot$1,
  	NegativeMediumSpace: NegativeMediumSpace$1,
  	NegativeThickSpace: NegativeThickSpace$1,
  	NegativeThinSpace: NegativeThinSpace$1,
  	NegativeVeryThinSpace: NegativeVeryThinSpace$1,
  	nequiv: nequiv$1,
  	nesear: nesear$1,
  	nesim: nesim$1,
  	NestedGreaterGreater: NestedGreaterGreater$1,
  	NestedLessLess: NestedLessLess$1,
  	NewLine: NewLine$1,
  	nexist: nexist$1,
  	nexists: nexists$1,
  	Nfr: Nfr$1,
  	nfr: nfr$1,
  	ngE: ngE$1,
  	nge: nge$1,
  	ngeq: ngeq$1,
  	ngeqq: ngeqq$1,
  	ngeqslant: ngeqslant$1,
  	nges: nges$1,
  	nGg: nGg$1,
  	ngsim: ngsim$1,
  	nGt: nGt$1,
  	ngt: ngt$1,
  	ngtr: ngtr$1,
  	nGtv: nGtv$1,
  	nharr: nharr$1,
  	nhArr: nhArr$1,
  	nhpar: nhpar$1,
  	ni: ni$1,
  	nis: nis$1,
  	nisd: nisd$1,
  	niv: niv$1,
  	NJcy: NJcy$1,
  	njcy: njcy$1,
  	nlarr: nlarr$1,
  	nlArr: nlArr$1,
  	nldr: nldr$1,
  	nlE: nlE$1,
  	nle: nle$1,
  	nleftarrow: nleftarrow$1,
  	nLeftarrow: nLeftarrow$1,
  	nleftrightarrow: nleftrightarrow$1,
  	nLeftrightarrow: nLeftrightarrow$1,
  	nleq: nleq$1,
  	nleqq: nleqq$1,
  	nleqslant: nleqslant$1,
  	nles: nles$1,
  	nless: nless$1,
  	nLl: nLl$1,
  	nlsim: nlsim$1,
  	nLt: nLt$1,
  	nlt: nlt$1,
  	nltri: nltri$1,
  	nltrie: nltrie$1,
  	nLtv: nLtv$1,
  	nmid: nmid$1,
  	NoBreak: NoBreak$1,
  	NonBreakingSpace: NonBreakingSpace$1,
  	nopf: nopf$1,
  	Nopf: Nopf$1,
  	Not: Not$1,
  	not: not$2,
  	NotCongruent: NotCongruent$1,
  	NotCupCap: NotCupCap$1,
  	NotDoubleVerticalBar: NotDoubleVerticalBar$1,
  	NotElement: NotElement$1,
  	NotEqual: NotEqual$1,
  	NotEqualTilde: NotEqualTilde$1,
  	NotExists: NotExists$1,
  	NotGreater: NotGreater$1,
  	NotGreaterEqual: NotGreaterEqual$1,
  	NotGreaterFullEqual: NotGreaterFullEqual$1,
  	NotGreaterGreater: NotGreaterGreater$1,
  	NotGreaterLess: NotGreaterLess$1,
  	NotGreaterSlantEqual: NotGreaterSlantEqual$1,
  	NotGreaterTilde: NotGreaterTilde$1,
  	NotHumpDownHump: NotHumpDownHump$1,
  	NotHumpEqual: NotHumpEqual$1,
  	notin: notin$1,
  	notindot: notindot$1,
  	notinE: notinE$1,
  	notinva: notinva$1,
  	notinvb: notinvb$1,
  	notinvc: notinvc$1,
  	NotLeftTriangleBar: NotLeftTriangleBar$1,
  	NotLeftTriangle: NotLeftTriangle$1,
  	NotLeftTriangleEqual: NotLeftTriangleEqual$1,
  	NotLess: NotLess$1,
  	NotLessEqual: NotLessEqual$1,
  	NotLessGreater: NotLessGreater$1,
  	NotLessLess: NotLessLess$1,
  	NotLessSlantEqual: NotLessSlantEqual$1,
  	NotLessTilde: NotLessTilde$1,
  	NotNestedGreaterGreater: NotNestedGreaterGreater$1,
  	NotNestedLessLess: NotNestedLessLess$1,
  	notni: notni$1,
  	notniva: notniva$1,
  	notnivb: notnivb$1,
  	notnivc: notnivc$1,
  	NotPrecedes: NotPrecedes$1,
  	NotPrecedesEqual: NotPrecedesEqual$1,
  	NotPrecedesSlantEqual: NotPrecedesSlantEqual$1,
  	NotReverseElement: NotReverseElement$1,
  	NotRightTriangleBar: NotRightTriangleBar$1,
  	NotRightTriangle: NotRightTriangle$1,
  	NotRightTriangleEqual: NotRightTriangleEqual$1,
  	NotSquareSubset: NotSquareSubset$1,
  	NotSquareSubsetEqual: NotSquareSubsetEqual$1,
  	NotSquareSuperset: NotSquareSuperset$1,
  	NotSquareSupersetEqual: NotSquareSupersetEqual$1,
  	NotSubset: NotSubset$1,
  	NotSubsetEqual: NotSubsetEqual$1,
  	NotSucceeds: NotSucceeds$1,
  	NotSucceedsEqual: NotSucceedsEqual$1,
  	NotSucceedsSlantEqual: NotSucceedsSlantEqual$1,
  	NotSucceedsTilde: NotSucceedsTilde$1,
  	NotSuperset: NotSuperset$1,
  	NotSupersetEqual: NotSupersetEqual$1,
  	NotTilde: NotTilde$1,
  	NotTildeEqual: NotTildeEqual$1,
  	NotTildeFullEqual: NotTildeFullEqual$1,
  	NotTildeTilde: NotTildeTilde$1,
  	NotVerticalBar: NotVerticalBar$1,
  	nparallel: nparallel$1,
  	npar: npar$1,
  	nparsl: nparsl$1,
  	npart: npart$1,
  	npolint: npolint$1,
  	npr: npr$1,
  	nprcue: nprcue$1,
  	nprec: nprec$1,
  	npreceq: npreceq$1,
  	npre: npre$1,
  	nrarrc: nrarrc$1,
  	nrarr: nrarr$1,
  	nrArr: nrArr$1,
  	nrarrw: nrarrw$1,
  	nrightarrow: nrightarrow$1,
  	nRightarrow: nRightarrow$1,
  	nrtri: nrtri$1,
  	nrtrie: nrtrie$1,
  	nsc: nsc$1,
  	nsccue: nsccue$1,
  	nsce: nsce$1,
  	Nscr: Nscr$1,
  	nscr: nscr$1,
  	nshortmid: nshortmid$1,
  	nshortparallel: nshortparallel$1,
  	nsim: nsim$1,
  	nsime: nsime$1,
  	nsimeq: nsimeq$1,
  	nsmid: nsmid$1,
  	nspar: nspar$1,
  	nsqsube: nsqsube$1,
  	nsqsupe: nsqsupe$1,
  	nsub: nsub$1,
  	nsubE: nsubE$1,
  	nsube: nsube$1,
  	nsubset: nsubset$1,
  	nsubseteq: nsubseteq$1,
  	nsubseteqq: nsubseteqq$1,
  	nsucc: nsucc$1,
  	nsucceq: nsucceq$1,
  	nsup: nsup$1,
  	nsupE: nsupE$1,
  	nsupe: nsupe$1,
  	nsupset: nsupset$1,
  	nsupseteq: nsupseteq$1,
  	nsupseteqq: nsupseteqq$1,
  	ntgl: ntgl$1,
  	Ntilde: Ntilde$2,
  	ntilde: ntilde$2,
  	ntlg: ntlg$1,
  	ntriangleleft: ntriangleleft$1,
  	ntrianglelefteq: ntrianglelefteq$1,
  	ntriangleright: ntriangleright$1,
  	ntrianglerighteq: ntrianglerighteq$1,
  	Nu: Nu$1,
  	nu: nu$1,
  	num: num$1,
  	numero: numero$1,
  	numsp: numsp$1,
  	nvap: nvap$1,
  	nvdash: nvdash$1,
  	nvDash: nvDash$1,
  	nVdash: nVdash$1,
  	nVDash: nVDash$1,
  	nvge: nvge$1,
  	nvgt: nvgt$1,
  	nvHarr: nvHarr$1,
  	nvinfin: nvinfin$1,
  	nvlArr: nvlArr$1,
  	nvle: nvle$1,
  	nvlt: nvlt$1,
  	nvltrie: nvltrie$1,
  	nvrArr: nvrArr$1,
  	nvrtrie: nvrtrie$1,
  	nvsim: nvsim$1,
  	nwarhk: nwarhk$1,
  	nwarr: nwarr$1,
  	nwArr: nwArr$1,
  	nwarrow: nwarrow$1,
  	nwnear: nwnear$1,
  	Oacute: Oacute$2,
  	oacute: oacute$2,
  	oast: oast$1,
  	Ocirc: Ocirc$2,
  	ocirc: ocirc$2,
  	ocir: ocir$1,
  	Ocy: Ocy$1,
  	ocy: ocy$1,
  	odash: odash$1,
  	Odblac: Odblac$1,
  	odblac: odblac$1,
  	odiv: odiv$1,
  	odot: odot$1,
  	odsold: odsold$1,
  	OElig: OElig$1,
  	oelig: oelig$1,
  	ofcir: ofcir$1,
  	Ofr: Ofr$1,
  	ofr: ofr$1,
  	ogon: ogon$1,
  	Ograve: Ograve$2,
  	ograve: ograve$2,
  	ogt: ogt$1,
  	ohbar: ohbar$1,
  	ohm: ohm$1,
  	oint: oint$1,
  	olarr: olarr$1,
  	olcir: olcir$1,
  	olcross: olcross$1,
  	oline: oline$1,
  	olt: olt$1,
  	Omacr: Omacr$1,
  	omacr: omacr$1,
  	Omega: Omega$1,
  	omega: omega$1,
  	Omicron: Omicron$1,
  	omicron: omicron$1,
  	omid: omid$1,
  	ominus: ominus$1,
  	Oopf: Oopf$1,
  	oopf: oopf$1,
  	opar: opar$1,
  	OpenCurlyDoubleQuote: OpenCurlyDoubleQuote$1,
  	OpenCurlyQuote: OpenCurlyQuote$1,
  	operp: operp$1,
  	oplus: oplus$1,
  	orarr: orarr$1,
  	Or: Or$1,
  	or: or$1,
  	ord: ord$1,
  	order: order$1,
  	orderof: orderof$1,
  	ordf: ordf$2,
  	ordm: ordm$2,
  	origof: origof$1,
  	oror: oror$1,
  	orslope: orslope$1,
  	orv: orv$1,
  	oS: oS$1,
  	Oscr: Oscr$1,
  	oscr: oscr$1,
  	Oslash: Oslash$2,
  	oslash: oslash$2,
  	osol: osol$1,
  	Otilde: Otilde$2,
  	otilde: otilde$2,
  	otimesas: otimesas$1,
  	Otimes: Otimes$1,
  	otimes: otimes$1,
  	Ouml: Ouml$2,
  	ouml: ouml$2,
  	ovbar: ovbar$1,
  	OverBar: OverBar$1,
  	OverBrace: OverBrace$1,
  	OverBracket: OverBracket$1,
  	OverParenthesis: OverParenthesis$1,
  	para: para$2,
  	parallel: parallel$1,
  	par: par$1,
  	parsim: parsim$1,
  	parsl: parsl$1,
  	part: part$1,
  	PartialD: PartialD$1,
  	Pcy: Pcy$1,
  	pcy: pcy$1,
  	percnt: percnt$1,
  	period: period$1,
  	permil: permil$1,
  	perp: perp$1,
  	pertenk: pertenk$1,
  	Pfr: Pfr$1,
  	pfr: pfr$1,
  	Phi: Phi$1,
  	phi: phi$1,
  	phiv: phiv$1,
  	phmmat: phmmat$1,
  	phone: phone$1,
  	Pi: Pi$1,
  	pi: pi$1,
  	pitchfork: pitchfork$1,
  	piv: piv$1,
  	planck: planck$1,
  	planckh: planckh$1,
  	plankv: plankv$1,
  	plusacir: plusacir$1,
  	plusb: plusb$1,
  	pluscir: pluscir$1,
  	plus: plus$1,
  	plusdo: plusdo$1,
  	plusdu: plusdu$1,
  	pluse: pluse$1,
  	PlusMinus: PlusMinus$1,
  	plusmn: plusmn$2,
  	plussim: plussim$1,
  	plustwo: plustwo$1,
  	pm: pm$1,
  	Poincareplane: Poincareplane$1,
  	pointint: pointint$1,
  	popf: popf$1,
  	Popf: Popf$1,
  	pound: pound$2,
  	prap: prap$1,
  	Pr: Pr$1,
  	pr: pr$1,
  	prcue: prcue$1,
  	precapprox: precapprox$1,
  	prec: prec$1,
  	preccurlyeq: preccurlyeq$1,
  	Precedes: Precedes$1,
  	PrecedesEqual: PrecedesEqual$1,
  	PrecedesSlantEqual: PrecedesSlantEqual$1,
  	PrecedesTilde: PrecedesTilde$1,
  	preceq: preceq$1,
  	precnapprox: precnapprox$1,
  	precneqq: precneqq$1,
  	precnsim: precnsim$1,
  	pre: pre$1,
  	prE: prE$1,
  	precsim: precsim$1,
  	prime: prime$1,
  	Prime: Prime$1,
  	primes: primes$1,
  	prnap: prnap$1,
  	prnE: prnE$1,
  	prnsim: prnsim$1,
  	prod: prod$1,
  	Product: Product$1,
  	profalar: profalar$1,
  	profline: profline$1,
  	profsurf: profsurf$1,
  	prop: prop$1,
  	Proportional: Proportional$1,
  	Proportion: Proportion$1,
  	propto: propto$1,
  	prsim: prsim$1,
  	prurel: prurel$1,
  	Pscr: Pscr$1,
  	pscr: pscr$1,
  	Psi: Psi$1,
  	psi: psi$1,
  	puncsp: puncsp$1,
  	Qfr: Qfr$1,
  	qfr: qfr$1,
  	qint: qint$1,
  	qopf: qopf$1,
  	Qopf: Qopf$1,
  	qprime: qprime$1,
  	Qscr: Qscr$1,
  	qscr: qscr$1,
  	quaternions: quaternions$1,
  	quatint: quatint$1,
  	quest: quest$1,
  	questeq: questeq$1,
  	quot: quot$1$1,
  	QUOT: QUOT$2,
  	rAarr: rAarr$1,
  	race: race$1,
  	Racute: Racute$1,
  	racute: racute$1,
  	radic: radic$1,
  	raemptyv: raemptyv$1,
  	rang: rang$1,
  	Rang: Rang$1,
  	rangd: rangd$1,
  	range: range$1,
  	rangle: rangle$1,
  	raquo: raquo$2,
  	rarrap: rarrap$1,
  	rarrb: rarrb$1,
  	rarrbfs: rarrbfs$1,
  	rarrc: rarrc$1,
  	rarr: rarr$1,
  	Rarr: Rarr$1,
  	rArr: rArr$1,
  	rarrfs: rarrfs$1,
  	rarrhk: rarrhk$1,
  	rarrlp: rarrlp$1,
  	rarrpl: rarrpl$1,
  	rarrsim: rarrsim$1,
  	Rarrtl: Rarrtl$1,
  	rarrtl: rarrtl$1,
  	rarrw: rarrw$1,
  	ratail: ratail$1,
  	rAtail: rAtail$1,
  	ratio: ratio$1,
  	rationals: rationals$1,
  	rbarr: rbarr$1,
  	rBarr: rBarr$1,
  	RBarr: RBarr$1,
  	rbbrk: rbbrk$1,
  	rbrace: rbrace$1,
  	rbrack: rbrack$1,
  	rbrke: rbrke$1,
  	rbrksld: rbrksld$1,
  	rbrkslu: rbrkslu$1,
  	Rcaron: Rcaron$1,
  	rcaron: rcaron$1,
  	Rcedil: Rcedil$1,
  	rcedil: rcedil$1,
  	rceil: rceil$1,
  	rcub: rcub$1,
  	Rcy: Rcy$1,
  	rcy: rcy$1,
  	rdca: rdca$1,
  	rdldhar: rdldhar$1,
  	rdquo: rdquo$1,
  	rdquor: rdquor$1,
  	rdsh: rdsh$1,
  	real: real$1,
  	realine: realine$1,
  	realpart: realpart$1,
  	reals: reals$1,
  	Re: Re$1,
  	rect: rect$1,
  	reg: reg$2,
  	REG: REG$2,
  	ReverseElement: ReverseElement$1,
  	ReverseEquilibrium: ReverseEquilibrium$1,
  	ReverseUpEquilibrium: ReverseUpEquilibrium$1,
  	rfisht: rfisht$1,
  	rfloor: rfloor$1,
  	rfr: rfr$1,
  	Rfr: Rfr$1,
  	rHar: rHar$1,
  	rhard: rhard$1,
  	rharu: rharu$1,
  	rharul: rharul$1,
  	Rho: Rho$1,
  	rho: rho$1,
  	rhov: rhov$1,
  	RightAngleBracket: RightAngleBracket$1,
  	RightArrowBar: RightArrowBar$1,
  	rightarrow: rightarrow$1,
  	RightArrow: RightArrow$1,
  	Rightarrow: Rightarrow$1,
  	RightArrowLeftArrow: RightArrowLeftArrow$1,
  	rightarrowtail: rightarrowtail$1,
  	RightCeiling: RightCeiling$1,
  	RightDoubleBracket: RightDoubleBracket$1,
  	RightDownTeeVector: RightDownTeeVector$1,
  	RightDownVectorBar: RightDownVectorBar$1,
  	RightDownVector: RightDownVector$1,
  	RightFloor: RightFloor$1,
  	rightharpoondown: rightharpoondown$1,
  	rightharpoonup: rightharpoonup$1,
  	rightleftarrows: rightleftarrows$1,
  	rightleftharpoons: rightleftharpoons$1,
  	rightrightarrows: rightrightarrows$1,
  	rightsquigarrow: rightsquigarrow$1,
  	RightTeeArrow: RightTeeArrow$1,
  	RightTee: RightTee$1,
  	RightTeeVector: RightTeeVector$1,
  	rightthreetimes: rightthreetimes$1,
  	RightTriangleBar: RightTriangleBar$1,
  	RightTriangle: RightTriangle$1,
  	RightTriangleEqual: RightTriangleEqual$1,
  	RightUpDownVector: RightUpDownVector$1,
  	RightUpTeeVector: RightUpTeeVector$1,
  	RightUpVectorBar: RightUpVectorBar$1,
  	RightUpVector: RightUpVector$1,
  	RightVectorBar: RightVectorBar$1,
  	RightVector: RightVector$1,
  	ring: ring$1,
  	risingdotseq: risingdotseq$1,
  	rlarr: rlarr$1,
  	rlhar: rlhar$1,
  	rlm: rlm$1,
  	rmoustache: rmoustache$1,
  	rmoust: rmoust$1,
  	rnmid: rnmid$1,
  	roang: roang$1,
  	roarr: roarr$1,
  	robrk: robrk$1,
  	ropar: ropar$1,
  	ropf: ropf$1,
  	Ropf: Ropf$1,
  	roplus: roplus$1,
  	rotimes: rotimes$1,
  	RoundImplies: RoundImplies$1,
  	rpar: rpar$1,
  	rpargt: rpargt$1,
  	rppolint: rppolint$1,
  	rrarr: rrarr$1,
  	Rrightarrow: Rrightarrow$1,
  	rsaquo: rsaquo$1,
  	rscr: rscr$1,
  	Rscr: Rscr$1,
  	rsh: rsh$1,
  	Rsh: Rsh$1,
  	rsqb: rsqb$1,
  	rsquo: rsquo$1,
  	rsquor: rsquor$1,
  	rthree: rthree$1,
  	rtimes: rtimes$1,
  	rtri: rtri$1,
  	rtrie: rtrie$1,
  	rtrif: rtrif$1,
  	rtriltri: rtriltri$1,
  	RuleDelayed: RuleDelayed$1,
  	ruluhar: ruluhar$1,
  	rx: rx$1,
  	Sacute: Sacute$1,
  	sacute: sacute$1,
  	sbquo: sbquo$1,
  	scap: scap$1,
  	Scaron: Scaron$1,
  	scaron: scaron$1,
  	Sc: Sc$1,
  	sc: sc$1,
  	sccue: sccue$1,
  	sce: sce$1,
  	scE: scE$1,
  	Scedil: Scedil$1,
  	scedil: scedil$1,
  	Scirc: Scirc$1,
  	scirc: scirc$1,
  	scnap: scnap$1,
  	scnE: scnE$1,
  	scnsim: scnsim$1,
  	scpolint: scpolint$1,
  	scsim: scsim$1,
  	Scy: Scy$1,
  	scy: scy$1,
  	sdotb: sdotb$1,
  	sdot: sdot$1,
  	sdote: sdote$1,
  	searhk: searhk$1,
  	searr: searr$1,
  	seArr: seArr$1,
  	searrow: searrow$1,
  	sect: sect$2,
  	semi: semi$1,
  	seswar: seswar$1,
  	setminus: setminus$1,
  	setmn: setmn$1,
  	sext: sext$1,
  	Sfr: Sfr$1,
  	sfr: sfr$1,
  	sfrown: sfrown$1,
  	sharp: sharp$1,
  	SHCHcy: SHCHcy$1,
  	shchcy: shchcy$1,
  	SHcy: SHcy$1,
  	shcy: shcy$1,
  	ShortDownArrow: ShortDownArrow$1,
  	ShortLeftArrow: ShortLeftArrow$1,
  	shortmid: shortmid$1,
  	shortparallel: shortparallel$1,
  	ShortRightArrow: ShortRightArrow$1,
  	ShortUpArrow: ShortUpArrow$1,
  	shy: shy$2,
  	Sigma: Sigma$1,
  	sigma: sigma$1,
  	sigmaf: sigmaf$1,
  	sigmav: sigmav$1,
  	sim: sim$1,
  	simdot: simdot$1,
  	sime: sime$1,
  	simeq: simeq$1,
  	simg: simg$1,
  	simgE: simgE$1,
  	siml: siml$1,
  	simlE: simlE$1,
  	simne: simne$1,
  	simplus: simplus$1,
  	simrarr: simrarr$1,
  	slarr: slarr$1,
  	SmallCircle: SmallCircle$1,
  	smallsetminus: smallsetminus$1,
  	smashp: smashp$1,
  	smeparsl: smeparsl$1,
  	smid: smid$1,
  	smile: smile$1,
  	smt: smt$1,
  	smte: smte$1,
  	smtes: smtes$1,
  	SOFTcy: SOFTcy$1,
  	softcy: softcy$1,
  	solbar: solbar$1,
  	solb: solb$1,
  	sol: sol$1,
  	Sopf: Sopf$1,
  	sopf: sopf$1,
  	spades: spades$1,
  	spadesuit: spadesuit$1,
  	spar: spar$1,
  	sqcap: sqcap$1,
  	sqcaps: sqcaps$1,
  	sqcup: sqcup$1,
  	sqcups: sqcups$1,
  	Sqrt: Sqrt$1,
  	sqsub: sqsub$1,
  	sqsube: sqsube$1,
  	sqsubset: sqsubset$1,
  	sqsubseteq: sqsubseteq$1,
  	sqsup: sqsup$1,
  	sqsupe: sqsupe$1,
  	sqsupset: sqsupset$1,
  	sqsupseteq: sqsupseteq$1,
  	square: square$1,
  	Square: Square$1,
  	SquareIntersection: SquareIntersection$1,
  	SquareSubset: SquareSubset$1,
  	SquareSubsetEqual: SquareSubsetEqual$1,
  	SquareSuperset: SquareSuperset$1,
  	SquareSupersetEqual: SquareSupersetEqual$1,
  	SquareUnion: SquareUnion$1,
  	squarf: squarf$1,
  	squ: squ$1,
  	squf: squf$1,
  	srarr: srarr$1,
  	Sscr: Sscr$1,
  	sscr: sscr$1,
  	ssetmn: ssetmn$1,
  	ssmile: ssmile$1,
  	sstarf: sstarf$1,
  	Star: Star$1,
  	star: star$1,
  	starf: starf$1,
  	straightepsilon: straightepsilon$1,
  	straightphi: straightphi$1,
  	strns: strns$1,
  	sub: sub$1,
  	Sub: Sub$1,
  	subdot: subdot$1,
  	subE: subE$1,
  	sube: sube$1,
  	subedot: subedot$1,
  	submult: submult$1,
  	subnE: subnE$1,
  	subne: subne$1,
  	subplus: subplus$1,
  	subrarr: subrarr$1,
  	subset: subset$1,
  	Subset: Subset$1,
  	subseteq: subseteq$1,
  	subseteqq: subseteqq$1,
  	SubsetEqual: SubsetEqual$1,
  	subsetneq: subsetneq$1,
  	subsetneqq: subsetneqq$1,
  	subsim: subsim$1,
  	subsub: subsub$1,
  	subsup: subsup$1,
  	succapprox: succapprox$1,
  	succ: succ$1,
  	succcurlyeq: succcurlyeq$1,
  	Succeeds: Succeeds$1,
  	SucceedsEqual: SucceedsEqual$1,
  	SucceedsSlantEqual: SucceedsSlantEqual$1,
  	SucceedsTilde: SucceedsTilde$1,
  	succeq: succeq$1,
  	succnapprox: succnapprox$1,
  	succneqq: succneqq$1,
  	succnsim: succnsim$1,
  	succsim: succsim$1,
  	SuchThat: SuchThat$1,
  	sum: sum$1,
  	Sum: Sum$1,
  	sung: sung$1,
  	sup1: sup1$2,
  	sup2: sup2$2,
  	sup3: sup3$2,
  	sup: sup$1,
  	Sup: Sup$1,
  	supdot: supdot$1,
  	supdsub: supdsub$1,
  	supE: supE$1,
  	supe: supe$1,
  	supedot: supedot$1,
  	Superset: Superset$1,
  	SupersetEqual: SupersetEqual$1,
  	suphsol: suphsol$1,
  	suphsub: suphsub$1,
  	suplarr: suplarr$1,
  	supmult: supmult$1,
  	supnE: supnE$1,
  	supne: supne$1,
  	supplus: supplus$1,
  	supset: supset$1,
  	Supset: Supset$1,
  	supseteq: supseteq$1,
  	supseteqq: supseteqq$1,
  	supsetneq: supsetneq$1,
  	supsetneqq: supsetneqq$1,
  	supsim: supsim$1,
  	supsub: supsub$1,
  	supsup: supsup$1,
  	swarhk: swarhk$1,
  	swarr: swarr$1,
  	swArr: swArr$1,
  	swarrow: swarrow$1,
  	swnwar: swnwar$1,
  	szlig: szlig$2,
  	Tab: Tab$1,
  	target: target$1,
  	Tau: Tau$1,
  	tau: tau$1,
  	tbrk: tbrk$1,
  	Tcaron: Tcaron$1,
  	tcaron: tcaron$1,
  	Tcedil: Tcedil$1,
  	tcedil: tcedil$1,
  	Tcy: Tcy$1,
  	tcy: tcy$1,
  	tdot: tdot$1,
  	telrec: telrec$1,
  	Tfr: Tfr$1,
  	tfr: tfr$1,
  	there4: there4$1,
  	therefore: therefore$1,
  	Therefore: Therefore$1,
  	Theta: Theta$1,
  	theta: theta$1,
  	thetasym: thetasym$1,
  	thetav: thetav$1,
  	thickapprox: thickapprox$1,
  	thicksim: thicksim$1,
  	ThickSpace: ThickSpace$1,
  	ThinSpace: ThinSpace$1,
  	thinsp: thinsp$1,
  	thkap: thkap$1,
  	thksim: thksim$1,
  	THORN: THORN$2,
  	thorn: thorn$2,
  	tilde: tilde$1,
  	Tilde: Tilde$1,
  	TildeEqual: TildeEqual$1,
  	TildeFullEqual: TildeFullEqual$1,
  	TildeTilde: TildeTilde$1,
  	timesbar: timesbar$1,
  	timesb: timesb$1,
  	times: times$3,
  	timesd: timesd$1,
  	tint: tint$1,
  	toea: toea$1,
  	topbot: topbot$1,
  	topcir: topcir$1,
  	top: top$1,
  	Topf: Topf$1,
  	topf: topf$1,
  	topfork: topfork$1,
  	tosa: tosa$1,
  	tprime: tprime$1,
  	trade: trade$1,
  	TRADE: TRADE$1,
  	triangle: triangle$1,
  	triangledown: triangledown$1,
  	triangleleft: triangleleft$1,
  	trianglelefteq: trianglelefteq$1,
  	triangleq: triangleq$1,
  	triangleright: triangleright$1,
  	trianglerighteq: trianglerighteq$1,
  	tridot: tridot$1,
  	trie: trie$1,
  	triminus: triminus$1,
  	TripleDot: TripleDot$1,
  	triplus: triplus$1,
  	trisb: trisb$1,
  	tritime: tritime$1,
  	trpezium: trpezium$1,
  	Tscr: Tscr$1,
  	tscr: tscr$1,
  	TScy: TScy$1,
  	tscy: tscy$1,
  	TSHcy: TSHcy$1,
  	tshcy: tshcy$1,
  	Tstrok: Tstrok$1,
  	tstrok: tstrok$1,
  	twixt: twixt$1,
  	twoheadleftarrow: twoheadleftarrow$1,
  	twoheadrightarrow: twoheadrightarrow$1,
  	Uacute: Uacute$2,
  	uacute: uacute$2,
  	uarr: uarr$1,
  	Uarr: Uarr$1,
  	uArr: uArr$1,
  	Uarrocir: Uarrocir$1,
  	Ubrcy: Ubrcy$1,
  	ubrcy: ubrcy$1,
  	Ubreve: Ubreve$1,
  	ubreve: ubreve$1,
  	Ucirc: Ucirc$2,
  	ucirc: ucirc$2,
  	Ucy: Ucy$1,
  	ucy: ucy$1,
  	udarr: udarr$1,
  	Udblac: Udblac$1,
  	udblac: udblac$1,
  	udhar: udhar$1,
  	ufisht: ufisht$1,
  	Ufr: Ufr$1,
  	ufr: ufr$1,
  	Ugrave: Ugrave$2,
  	ugrave: ugrave$2,
  	uHar: uHar$1,
  	uharl: uharl$1,
  	uharr: uharr$1,
  	uhblk: uhblk$1,
  	ulcorn: ulcorn$1,
  	ulcorner: ulcorner$1,
  	ulcrop: ulcrop$1,
  	ultri: ultri$1,
  	Umacr: Umacr$1,
  	umacr: umacr$1,
  	uml: uml$2,
  	UnderBar: UnderBar$1,
  	UnderBrace: UnderBrace$1,
  	UnderBracket: UnderBracket$1,
  	UnderParenthesis: UnderParenthesis$1,
  	Union: Union$1,
  	UnionPlus: UnionPlus$1,
  	Uogon: Uogon$1,
  	uogon: uogon$1,
  	Uopf: Uopf$1,
  	uopf: uopf$1,
  	UpArrowBar: UpArrowBar$1,
  	uparrow: uparrow$1,
  	UpArrow: UpArrow$1,
  	Uparrow: Uparrow$1,
  	UpArrowDownArrow: UpArrowDownArrow$1,
  	updownarrow: updownarrow$1,
  	UpDownArrow: UpDownArrow$1,
  	Updownarrow: Updownarrow$1,
  	UpEquilibrium: UpEquilibrium$1,
  	upharpoonleft: upharpoonleft$1,
  	upharpoonright: upharpoonright$1,
  	uplus: uplus$1,
  	UpperLeftArrow: UpperLeftArrow$1,
  	UpperRightArrow: UpperRightArrow$1,
  	upsi: upsi$1,
  	Upsi: Upsi$1,
  	upsih: upsih$1,
  	Upsilon: Upsilon$1,
  	upsilon: upsilon$1,
  	UpTeeArrow: UpTeeArrow$1,
  	UpTee: UpTee$1,
  	upuparrows: upuparrows$1,
  	urcorn: urcorn$1,
  	urcorner: urcorner$1,
  	urcrop: urcrop$1,
  	Uring: Uring$1,
  	uring: uring$1,
  	urtri: urtri$1,
  	Uscr: Uscr$1,
  	uscr: uscr$1,
  	utdot: utdot$1,
  	Utilde: Utilde$1,
  	utilde: utilde$1,
  	utri: utri$1,
  	utrif: utrif$1,
  	uuarr: uuarr$1,
  	Uuml: Uuml$2,
  	uuml: uuml$2,
  	uwangle: uwangle$1,
  	vangrt: vangrt$1,
  	varepsilon: varepsilon$1,
  	varkappa: varkappa$1,
  	varnothing: varnothing$1,
  	varphi: varphi$1,
  	varpi: varpi$1,
  	varpropto: varpropto$1,
  	varr: varr$1,
  	vArr: vArr$1,
  	varrho: varrho$1,
  	varsigma: varsigma$1,
  	varsubsetneq: varsubsetneq$1,
  	varsubsetneqq: varsubsetneqq$1,
  	varsupsetneq: varsupsetneq$1,
  	varsupsetneqq: varsupsetneqq$1,
  	vartheta: vartheta$1,
  	vartriangleleft: vartriangleleft$1,
  	vartriangleright: vartriangleright$1,
  	vBar: vBar$1,
  	Vbar: Vbar$1,
  	vBarv: vBarv$1,
  	Vcy: Vcy$1,
  	vcy: vcy$1,
  	vdash: vdash$1,
  	vDash: vDash$1,
  	Vdash: Vdash$1,
  	VDash: VDash$1,
  	Vdashl: Vdashl$1,
  	veebar: veebar$1,
  	vee: vee$1,
  	Vee: Vee$1,
  	veeeq: veeeq$1,
  	vellip: vellip$1,
  	verbar: verbar$1,
  	Verbar: Verbar$1,
  	vert: vert$1,
  	Vert: Vert$1,
  	VerticalBar: VerticalBar$1,
  	VerticalLine: VerticalLine$1,
  	VerticalSeparator: VerticalSeparator$1,
  	VerticalTilde: VerticalTilde$1,
  	VeryThinSpace: VeryThinSpace$1,
  	Vfr: Vfr$1,
  	vfr: vfr$1,
  	vltri: vltri$1,
  	vnsub: vnsub$1,
  	vnsup: vnsup$1,
  	Vopf: Vopf$1,
  	vopf: vopf$1,
  	vprop: vprop$1,
  	vrtri: vrtri$1,
  	Vscr: Vscr$1,
  	vscr: vscr$1,
  	vsubnE: vsubnE$1,
  	vsubne: vsubne$1,
  	vsupnE: vsupnE$1,
  	vsupne: vsupne$1,
  	Vvdash: Vvdash$1,
  	vzigzag: vzigzag$1,
  	Wcirc: Wcirc$1,
  	wcirc: wcirc$1,
  	wedbar: wedbar$1,
  	wedge: wedge$1,
  	Wedge: Wedge$1,
  	wedgeq: wedgeq$1,
  	weierp: weierp$1,
  	Wfr: Wfr$1,
  	wfr: wfr$1,
  	Wopf: Wopf$1,
  	wopf: wopf$1,
  	wp: wp$1,
  	wr: wr$1,
  	wreath: wreath$1,
  	Wscr: Wscr$1,
  	wscr: wscr$1,
  	xcap: xcap$1,
  	xcirc: xcirc$1,
  	xcup: xcup$1,
  	xdtri: xdtri$1,
  	Xfr: Xfr$1,
  	xfr: xfr$1,
  	xharr: xharr$1,
  	xhArr: xhArr$1,
  	Xi: Xi$1,
  	xi: xi$1,
  	xlarr: xlarr$1,
  	xlArr: xlArr$1,
  	xmap: xmap$1,
  	xnis: xnis$1,
  	xodot: xodot$1,
  	Xopf: Xopf$1,
  	xopf: xopf$1,
  	xoplus: xoplus$1,
  	xotime: xotime$1,
  	xrarr: xrarr$1,
  	xrArr: xrArr$1,
  	Xscr: Xscr$1,
  	xscr: xscr$1,
  	xsqcup: xsqcup$1,
  	xuplus: xuplus$1,
  	xutri: xutri$1,
  	xvee: xvee$1,
  	xwedge: xwedge$1,
  	Yacute: Yacute$2,
  	yacute: yacute$2,
  	YAcy: YAcy$1,
  	yacy: yacy$1,
  	Ycirc: Ycirc$1,
  	ycirc: ycirc$1,
  	Ycy: Ycy$1,
  	ycy: ycy$1,
  	yen: yen$2,
  	Yfr: Yfr$1,
  	yfr: yfr$1,
  	YIcy: YIcy$1,
  	yicy: yicy$1,
  	Yopf: Yopf$1,
  	yopf: yopf$1,
  	Yscr: Yscr$1,
  	yscr: yscr$1,
  	YUcy: YUcy$1,
  	yucy: yucy$1,
  	yuml: yuml$2,
  	Yuml: Yuml$1,
  	Zacute: Zacute$1,
  	zacute: zacute$1,
  	Zcaron: Zcaron$1,
  	zcaron: zcaron$1,
  	Zcy: Zcy$1,
  	zcy: zcy$1,
  	Zdot: Zdot$1,
  	zdot: zdot$1,
  	zeetrf: zeetrf$1,
  	ZeroWidthSpace: ZeroWidthSpace$1,
  	Zeta: Zeta$1,
  	zeta: zeta$1,
  	zfr: zfr$1,
  	Zfr: Zfr$1,
  	ZHcy: ZHcy$1,
  	zhcy: zhcy$1,
  	zigrarr: zigrarr$1,
  	zopf: zopf$1,
  	Zopf: Zopf$1,
  	Zscr: Zscr$1,
  	zscr: zscr$1,
  	zwj: zwj$1,
  	zwnj: zwnj$1,
  	"in": "∈",
  	"Map": "⤅"
  };

  var entities$1 = Object.freeze({
  	Aacute: Aacute$2,
  	aacute: aacute$2,
  	Abreve: Abreve$1,
  	abreve: abreve$1,
  	ac: ac$1,
  	acd: acd$1,
  	acE: acE$1,
  	Acirc: Acirc$2,
  	acirc: acirc$2,
  	acute: acute$2,
  	Acy: Acy$1,
  	acy: acy$1,
  	AElig: AElig$2,
  	aelig: aelig$2,
  	af: af$1,
  	Afr: Afr$1,
  	afr: afr$1,
  	Agrave: Agrave$2,
  	agrave: agrave$2,
  	alefsym: alefsym$1,
  	aleph: aleph$1,
  	Alpha: Alpha$1,
  	alpha: alpha$1,
  	Amacr: Amacr$1,
  	amacr: amacr$1,
  	amalg: amalg$1,
  	amp: amp$1$1,
  	AMP: AMP$2,
  	andand: andand$1,
  	And: And$1,
  	and: and$1,
  	andd: andd$1,
  	andslope: andslope$1,
  	andv: andv$1,
  	ang: ang$1,
  	ange: ange$1,
  	angle: angle$1,
  	angmsdaa: angmsdaa$1,
  	angmsdab: angmsdab$1,
  	angmsdac: angmsdac$1,
  	angmsdad: angmsdad$1,
  	angmsdae: angmsdae$1,
  	angmsdaf: angmsdaf$1,
  	angmsdag: angmsdag$1,
  	angmsdah: angmsdah$1,
  	angmsd: angmsd$1,
  	angrt: angrt$1,
  	angrtvb: angrtvb$1,
  	angrtvbd: angrtvbd$1,
  	angsph: angsph$1,
  	angst: angst$1,
  	angzarr: angzarr$1,
  	Aogon: Aogon$1,
  	aogon: aogon$1,
  	Aopf: Aopf$1,
  	aopf: aopf$1,
  	apacir: apacir$1,
  	ap: ap$1,
  	apE: apE$1,
  	ape: ape$1,
  	apid: apid$1,
  	apos: apos$1$1,
  	ApplyFunction: ApplyFunction$1,
  	approx: approx$1,
  	approxeq: approxeq$1,
  	Aring: Aring$2,
  	aring: aring$2,
  	Ascr: Ascr$1,
  	ascr: ascr$1,
  	Assign: Assign$1,
  	ast: ast$1,
  	asymp: asymp$1,
  	asympeq: asympeq$1,
  	Atilde: Atilde$2,
  	atilde: atilde$2,
  	Auml: Auml$2,
  	auml: auml$2,
  	awconint: awconint$1,
  	awint: awint$1,
  	backcong: backcong$1,
  	backepsilon: backepsilon$1,
  	backprime: backprime$1,
  	backsim: backsim$1,
  	backsimeq: backsimeq$1,
  	Backslash: Backslash$1,
  	Barv: Barv$1,
  	barvee: barvee$1,
  	barwed: barwed$1,
  	Barwed: Barwed$1,
  	barwedge: barwedge$1,
  	bbrk: bbrk$1,
  	bbrktbrk: bbrktbrk$1,
  	bcong: bcong$1,
  	Bcy: Bcy$1,
  	bcy: bcy$1,
  	bdquo: bdquo$1,
  	becaus: becaus$1,
  	because: because$1,
  	Because: Because$1,
  	bemptyv: bemptyv$1,
  	bepsi: bepsi$1,
  	bernou: bernou$1,
  	Bernoullis: Bernoullis$1,
  	Beta: Beta$1,
  	beta: beta$1,
  	beth: beth$1,
  	between: between$1,
  	Bfr: Bfr$1,
  	bfr: bfr$1,
  	bigcap: bigcap$1,
  	bigcirc: bigcirc$1,
  	bigcup: bigcup$1,
  	bigodot: bigodot$1,
  	bigoplus: bigoplus$1,
  	bigotimes: bigotimes$1,
  	bigsqcup: bigsqcup$1,
  	bigstar: bigstar$1,
  	bigtriangledown: bigtriangledown$1,
  	bigtriangleup: bigtriangleup$1,
  	biguplus: biguplus$1,
  	bigvee: bigvee$1,
  	bigwedge: bigwedge$1,
  	bkarow: bkarow$1,
  	blacklozenge: blacklozenge$1,
  	blacksquare: blacksquare$1,
  	blacktriangle: blacktriangle$1,
  	blacktriangledown: blacktriangledown$1,
  	blacktriangleleft: blacktriangleleft$1,
  	blacktriangleright: blacktriangleright$1,
  	blank: blank$1,
  	blk12: blk12$1,
  	blk14: blk14$1,
  	blk34: blk34$1,
  	block: block$1,
  	bne: bne$1,
  	bnequiv: bnequiv$1,
  	bNot: bNot$1,
  	bnot: bnot$1,
  	Bopf: Bopf$1,
  	bopf: bopf$1,
  	bot: bot$1,
  	bottom: bottom$1,
  	bowtie: bowtie$1,
  	boxbox: boxbox$1,
  	boxdl: boxdl$1,
  	boxdL: boxdL$1,
  	boxDl: boxDl$1,
  	boxDL: boxDL$1,
  	boxdr: boxdr$1,
  	boxdR: boxdR$1,
  	boxDr: boxDr$1,
  	boxDR: boxDR$1,
  	boxh: boxh$1,
  	boxH: boxH$1,
  	boxhd: boxhd$1,
  	boxHd: boxHd$1,
  	boxhD: boxhD$1,
  	boxHD: boxHD$1,
  	boxhu: boxhu$1,
  	boxHu: boxHu$1,
  	boxhU: boxhU$1,
  	boxHU: boxHU$1,
  	boxminus: boxminus$1,
  	boxplus: boxplus$1,
  	boxtimes: boxtimes$1,
  	boxul: boxul$1,
  	boxuL: boxuL$1,
  	boxUl: boxUl$1,
  	boxUL: boxUL$1,
  	boxur: boxur$1,
  	boxuR: boxuR$1,
  	boxUr: boxUr$1,
  	boxUR: boxUR$1,
  	boxv: boxv$1,
  	boxV: boxV$1,
  	boxvh: boxvh$1,
  	boxvH: boxvH$1,
  	boxVh: boxVh$1,
  	boxVH: boxVH$1,
  	boxvl: boxvl$1,
  	boxvL: boxvL$1,
  	boxVl: boxVl$1,
  	boxVL: boxVL$1,
  	boxvr: boxvr$1,
  	boxvR: boxvR$1,
  	boxVr: boxVr$1,
  	boxVR: boxVR$1,
  	bprime: bprime$1,
  	breve: breve$1,
  	Breve: Breve$1,
  	brvbar: brvbar$2,
  	bscr: bscr$1,
  	Bscr: Bscr$1,
  	bsemi: bsemi$1,
  	bsim: bsim$1,
  	bsime: bsime$1,
  	bsolb: bsolb$1,
  	bsol: bsol$1,
  	bsolhsub: bsolhsub$1,
  	bull: bull$1,
  	bullet: bullet$1,
  	bump: bump$1,
  	bumpE: bumpE$1,
  	bumpe: bumpe$1,
  	Bumpeq: Bumpeq$1,
  	bumpeq: bumpeq$1,
  	Cacute: Cacute$1,
  	cacute: cacute$1,
  	capand: capand$1,
  	capbrcup: capbrcup$1,
  	capcap: capcap$1,
  	cap: cap$1,
  	Cap: Cap$1,
  	capcup: capcup$1,
  	capdot: capdot$1,
  	CapitalDifferentialD: CapitalDifferentialD$1,
  	caps: caps$1,
  	caret: caret$1,
  	caron: caron$1,
  	Cayleys: Cayleys$1,
  	ccaps: ccaps$1,
  	Ccaron: Ccaron$1,
  	ccaron: ccaron$1,
  	Ccedil: Ccedil$2,
  	ccedil: ccedil$2,
  	Ccirc: Ccirc$1,
  	ccirc: ccirc$1,
  	Cconint: Cconint$1,
  	ccups: ccups$1,
  	ccupssm: ccupssm$1,
  	Cdot: Cdot$1,
  	cdot: cdot$1,
  	cedil: cedil$2,
  	Cedilla: Cedilla$1,
  	cemptyv: cemptyv$1,
  	cent: cent$2,
  	centerdot: centerdot$1,
  	CenterDot: CenterDot$1,
  	cfr: cfr$1,
  	Cfr: Cfr$1,
  	CHcy: CHcy$1,
  	chcy: chcy$1,
  	check: check$1,
  	checkmark: checkmark$1,
  	Chi: Chi$1,
  	chi: chi$1,
  	circ: circ$1,
  	circeq: circeq$1,
  	circlearrowleft: circlearrowleft$1,
  	circlearrowright: circlearrowright$1,
  	circledast: circledast$1,
  	circledcirc: circledcirc$1,
  	circleddash: circleddash$1,
  	CircleDot: CircleDot$1,
  	circledR: circledR$1,
  	circledS: circledS$1,
  	CircleMinus: CircleMinus$1,
  	CirclePlus: CirclePlus$1,
  	CircleTimes: CircleTimes$1,
  	cir: cir$1,
  	cirE: cirE$1,
  	cire: cire$1,
  	cirfnint: cirfnint$1,
  	cirmid: cirmid$1,
  	cirscir: cirscir$1,
  	ClockwiseContourIntegral: ClockwiseContourIntegral$1,
  	CloseCurlyDoubleQuote: CloseCurlyDoubleQuote$1,
  	CloseCurlyQuote: CloseCurlyQuote$1,
  	clubs: clubs$1,
  	clubsuit: clubsuit$1,
  	colon: colon$1,
  	Colon: Colon$1,
  	Colone: Colone$1,
  	colone: colone$1,
  	coloneq: coloneq$1,
  	comma: comma$1,
  	commat: commat$1,
  	comp: comp$1,
  	compfn: compfn$1,
  	complement: complement$1,
  	complexes: complexes$1,
  	cong: cong$1,
  	congdot: congdot$1,
  	Congruent: Congruent$1,
  	conint: conint$1,
  	Conint: Conint$1,
  	ContourIntegral: ContourIntegral$1,
  	copf: copf$1,
  	Copf: Copf$1,
  	coprod: coprod$1,
  	Coproduct: Coproduct$1,
  	copy: copy$2,
  	COPY: COPY$2,
  	copysr: copysr$1,
  	CounterClockwiseContourIntegral: CounterClockwiseContourIntegral$1,
  	crarr: crarr$1,
  	cross: cross$1,
  	Cross: Cross$1,
  	Cscr: Cscr$1,
  	cscr: cscr$1,
  	csub: csub$1,
  	csube: csube$1,
  	csup: csup$1,
  	csupe: csupe$1,
  	ctdot: ctdot$1,
  	cudarrl: cudarrl$1,
  	cudarrr: cudarrr$1,
  	cuepr: cuepr$1,
  	cuesc: cuesc$1,
  	cularr: cularr$1,
  	cularrp: cularrp$1,
  	cupbrcap: cupbrcap$1,
  	cupcap: cupcap$1,
  	CupCap: CupCap$1,
  	cup: cup$1,
  	Cup: Cup$1,
  	cupcup: cupcup$1,
  	cupdot: cupdot$1,
  	cupor: cupor$1,
  	cups: cups$1,
  	curarr: curarr$1,
  	curarrm: curarrm$1,
  	curlyeqprec: curlyeqprec$1,
  	curlyeqsucc: curlyeqsucc$1,
  	curlyvee: curlyvee$1,
  	curlywedge: curlywedge$1,
  	curren: curren$2,
  	curvearrowleft: curvearrowleft$1,
  	curvearrowright: curvearrowright$1,
  	cuvee: cuvee$1,
  	cuwed: cuwed$1,
  	cwconint: cwconint$1,
  	cwint: cwint$1,
  	cylcty: cylcty$1,
  	dagger: dagger$1,
  	Dagger: Dagger$1,
  	daleth: daleth$1,
  	darr: darr$1,
  	Darr: Darr$1,
  	dArr: dArr$1,
  	dash: dash$1,
  	Dashv: Dashv$1,
  	dashv: dashv$1,
  	dbkarow: dbkarow$1,
  	dblac: dblac$1,
  	Dcaron: Dcaron$1,
  	dcaron: dcaron$1,
  	Dcy: Dcy$1,
  	dcy: dcy$1,
  	ddagger: ddagger$1,
  	ddarr: ddarr$1,
  	DD: DD$1,
  	dd: dd$1,
  	DDotrahd: DDotrahd$1,
  	ddotseq: ddotseq$1,
  	deg: deg$2,
  	Del: Del$1,
  	Delta: Delta$1,
  	delta: delta$1,
  	demptyv: demptyv$1,
  	dfisht: dfisht$1,
  	Dfr: Dfr$1,
  	dfr: dfr$1,
  	dHar: dHar$1,
  	dharl: dharl$1,
  	dharr: dharr$1,
  	DiacriticalAcute: DiacriticalAcute$1,
  	DiacriticalDot: DiacriticalDot$1,
  	DiacriticalDoubleAcute: DiacriticalDoubleAcute$1,
  	DiacriticalGrave: DiacriticalGrave$1,
  	DiacriticalTilde: DiacriticalTilde$1,
  	diam: diam$1,
  	diamond: diamond$1,
  	Diamond: Diamond$1,
  	diamondsuit: diamondsuit$1,
  	diams: diams$1,
  	die: die$1,
  	DifferentialD: DifferentialD$1,
  	digamma: digamma$1,
  	disin: disin$1,
  	div: div$1,
  	divide: divide$2,
  	divideontimes: divideontimes$1,
  	divonx: divonx$1,
  	DJcy: DJcy$1,
  	djcy: djcy$1,
  	dlcorn: dlcorn$1,
  	dlcrop: dlcrop$1,
  	dollar: dollar$1,
  	Dopf: Dopf$1,
  	dopf: dopf$1,
  	Dot: Dot$1,
  	dot: dot$1,
  	DotDot: DotDot$1,
  	doteq: doteq$1,
  	doteqdot: doteqdot$1,
  	DotEqual: DotEqual$1,
  	dotminus: dotminus$1,
  	dotplus: dotplus$1,
  	dotsquare: dotsquare$1,
  	doublebarwedge: doublebarwedge$1,
  	DoubleContourIntegral: DoubleContourIntegral$1,
  	DoubleDot: DoubleDot$1,
  	DoubleDownArrow: DoubleDownArrow$1,
  	DoubleLeftArrow: DoubleLeftArrow$1,
  	DoubleLeftRightArrow: DoubleLeftRightArrow$1,
  	DoubleLeftTee: DoubleLeftTee$1,
  	DoubleLongLeftArrow: DoubleLongLeftArrow$1,
  	DoubleLongLeftRightArrow: DoubleLongLeftRightArrow$1,
  	DoubleLongRightArrow: DoubleLongRightArrow$1,
  	DoubleRightArrow: DoubleRightArrow$1,
  	DoubleRightTee: DoubleRightTee$1,
  	DoubleUpArrow: DoubleUpArrow$1,
  	DoubleUpDownArrow: DoubleUpDownArrow$1,
  	DoubleVerticalBar: DoubleVerticalBar$1,
  	DownArrowBar: DownArrowBar$1,
  	downarrow: downarrow$1,
  	DownArrow: DownArrow$1,
  	Downarrow: Downarrow$1,
  	DownArrowUpArrow: DownArrowUpArrow$1,
  	DownBreve: DownBreve$1,
  	downdownarrows: downdownarrows$1,
  	downharpoonleft: downharpoonleft$1,
  	downharpoonright: downharpoonright$1,
  	DownLeftRightVector: DownLeftRightVector$1,
  	DownLeftTeeVector: DownLeftTeeVector$1,
  	DownLeftVectorBar: DownLeftVectorBar$1,
  	DownLeftVector: DownLeftVector$1,
  	DownRightTeeVector: DownRightTeeVector$1,
  	DownRightVectorBar: DownRightVectorBar$1,
  	DownRightVector: DownRightVector$1,
  	DownTeeArrow: DownTeeArrow$1,
  	DownTee: DownTee$1,
  	drbkarow: drbkarow$1,
  	drcorn: drcorn$1,
  	drcrop: drcrop$1,
  	Dscr: Dscr$1,
  	dscr: dscr$1,
  	DScy: DScy$1,
  	dscy: dscy$1,
  	dsol: dsol$1,
  	Dstrok: Dstrok$1,
  	dstrok: dstrok$1,
  	dtdot: dtdot$1,
  	dtri: dtri$1,
  	dtrif: dtrif$1,
  	duarr: duarr$1,
  	duhar: duhar$1,
  	dwangle: dwangle$1,
  	DZcy: DZcy$1,
  	dzcy: dzcy$1,
  	dzigrarr: dzigrarr$1,
  	Eacute: Eacute$2,
  	eacute: eacute$2,
  	easter: easter$1,
  	Ecaron: Ecaron$1,
  	ecaron: ecaron$1,
  	Ecirc: Ecirc$2,
  	ecirc: ecirc$2,
  	ecir: ecir$1,
  	ecolon: ecolon$1,
  	Ecy: Ecy$1,
  	ecy: ecy$1,
  	eDDot: eDDot$1,
  	Edot: Edot$1,
  	edot: edot$1,
  	eDot: eDot$1,
  	ee: ee$1,
  	efDot: efDot$1,
  	Efr: Efr$1,
  	efr: efr$1,
  	eg: eg$1,
  	Egrave: Egrave$2,
  	egrave: egrave$2,
  	egs: egs$1,
  	egsdot: egsdot$1,
  	el: el$1,
  	Element: Element$1,
  	elinters: elinters$1,
  	ell: ell$1,
  	els: els$1,
  	elsdot: elsdot$1,
  	Emacr: Emacr$1,
  	emacr: emacr$1,
  	empty: empty$1,
  	emptyset: emptyset$1,
  	EmptySmallSquare: EmptySmallSquare$1,
  	emptyv: emptyv$1,
  	EmptyVerySmallSquare: EmptyVerySmallSquare$1,
  	emsp13: emsp13$1,
  	emsp14: emsp14$1,
  	emsp: emsp$1,
  	ENG: ENG$1,
  	eng: eng$1,
  	ensp: ensp$1,
  	Eogon: Eogon$1,
  	eogon: eogon$1,
  	Eopf: Eopf$1,
  	eopf: eopf$1,
  	epar: epar$1,
  	eparsl: eparsl$1,
  	eplus: eplus$1,
  	epsi: epsi$1,
  	Epsilon: Epsilon$1,
  	epsilon: epsilon$1,
  	epsiv: epsiv$1,
  	eqcirc: eqcirc$1,
  	eqcolon: eqcolon$1,
  	eqsim: eqsim$1,
  	eqslantgtr: eqslantgtr$1,
  	eqslantless: eqslantless$1,
  	Equal: Equal$1,
  	equals: equals$1,
  	EqualTilde: EqualTilde$1,
  	equest: equest$1,
  	Equilibrium: Equilibrium$1,
  	equiv: equiv$1,
  	equivDD: equivDD$1,
  	eqvparsl: eqvparsl$1,
  	erarr: erarr$1,
  	erDot: erDot$1,
  	escr: escr$1,
  	Escr: Escr$1,
  	esdot: esdot$1,
  	Esim: Esim$1,
  	esim: esim$1,
  	Eta: Eta$1,
  	eta: eta$1,
  	ETH: ETH$2,
  	eth: eth$2,
  	Euml: Euml$2,
  	euml: euml$2,
  	euro: euro$1,
  	excl: excl$1,
  	exist: exist$1,
  	Exists: Exists$1,
  	expectation: expectation$1,
  	exponentiale: exponentiale$1,
  	ExponentialE: ExponentialE$1,
  	fallingdotseq: fallingdotseq$1,
  	Fcy: Fcy$1,
  	fcy: fcy$1,
  	female: female$1,
  	ffilig: ffilig$1,
  	fflig: fflig$1,
  	ffllig: ffllig$1,
  	Ffr: Ffr$1,
  	ffr: ffr$1,
  	filig: filig$1,
  	FilledSmallSquare: FilledSmallSquare$1,
  	FilledVerySmallSquare: FilledVerySmallSquare$1,
  	fjlig: fjlig$1,
  	flat: flat$1,
  	fllig: fllig$1,
  	fltns: fltns$1,
  	fnof: fnof$1,
  	Fopf: Fopf$1,
  	fopf: fopf$1,
  	forall: forall$1,
  	ForAll: ForAll$1,
  	fork: fork$1,
  	forkv: forkv$1,
  	Fouriertrf: Fouriertrf$1,
  	fpartint: fpartint$1,
  	frac12: frac12$2,
  	frac13: frac13$1,
  	frac14: frac14$2,
  	frac15: frac15$1,
  	frac16: frac16$1,
  	frac18: frac18$1,
  	frac23: frac23$1,
  	frac25: frac25$1,
  	frac34: frac34$2,
  	frac35: frac35$1,
  	frac38: frac38$1,
  	frac45: frac45$1,
  	frac56: frac56$1,
  	frac58: frac58$1,
  	frac78: frac78$1,
  	frasl: frasl$1,
  	frown: frown$1,
  	fscr: fscr$1,
  	Fscr: Fscr$1,
  	gacute: gacute$1,
  	Gamma: Gamma$1,
  	gamma: gamma$1,
  	Gammad: Gammad$1,
  	gammad: gammad$1,
  	gap: gap$1,
  	Gbreve: Gbreve$1,
  	gbreve: gbreve$1,
  	Gcedil: Gcedil$1,
  	Gcirc: Gcirc$1,
  	gcirc: gcirc$1,
  	Gcy: Gcy$1,
  	gcy: gcy$1,
  	Gdot: Gdot$1,
  	gdot: gdot$1,
  	ge: ge$1,
  	gE: gE$1,
  	gEl: gEl$1,
  	gel: gel$1,
  	geq: geq$1,
  	geqq: geqq$1,
  	geqslant: geqslant$1,
  	gescc: gescc$1,
  	ges: ges$1,
  	gesdot: gesdot$1,
  	gesdoto: gesdoto$1,
  	gesdotol: gesdotol$1,
  	gesl: gesl$1,
  	gesles: gesles$1,
  	Gfr: Gfr$1,
  	gfr: gfr$1,
  	gg: gg$1,
  	Gg: Gg$1,
  	ggg: ggg$1,
  	gimel: gimel$1,
  	GJcy: GJcy$1,
  	gjcy: gjcy$1,
  	gla: gla$1,
  	gl: gl$1,
  	glE: glE$1,
  	glj: glj$1,
  	gnap: gnap$1,
  	gnapprox: gnapprox$1,
  	gne: gne$1,
  	gnE: gnE$1,
  	gneq: gneq$1,
  	gneqq: gneqq$1,
  	gnsim: gnsim$1,
  	Gopf: Gopf$1,
  	gopf: gopf$1,
  	grave: grave$1,
  	GreaterEqual: GreaterEqual$1,
  	GreaterEqualLess: GreaterEqualLess$1,
  	GreaterFullEqual: GreaterFullEqual$1,
  	GreaterGreater: GreaterGreater$1,
  	GreaterLess: GreaterLess$1,
  	GreaterSlantEqual: GreaterSlantEqual$1,
  	GreaterTilde: GreaterTilde$1,
  	Gscr: Gscr$1,
  	gscr: gscr$1,
  	gsim: gsim$1,
  	gsime: gsime$1,
  	gsiml: gsiml$1,
  	gtcc: gtcc$1,
  	gtcir: gtcir$1,
  	gt: gt$1$1,
  	GT: GT$2,
  	Gt: Gt$1,
  	gtdot: gtdot$1,
  	gtlPar: gtlPar$1,
  	gtquest: gtquest$1,
  	gtrapprox: gtrapprox$1,
  	gtrarr: gtrarr$1,
  	gtrdot: gtrdot$1,
  	gtreqless: gtreqless$1,
  	gtreqqless: gtreqqless$1,
  	gtrless: gtrless$1,
  	gtrsim: gtrsim$1,
  	gvertneqq: gvertneqq$1,
  	gvnE: gvnE$1,
  	Hacek: Hacek$1,
  	hairsp: hairsp$1,
  	half: half$1,
  	hamilt: hamilt$1,
  	HARDcy: HARDcy$1,
  	hardcy: hardcy$1,
  	harrcir: harrcir$1,
  	harr: harr$1,
  	hArr: hArr$1,
  	harrw: harrw$1,
  	Hat: Hat$1,
  	hbar: hbar$1,
  	Hcirc: Hcirc$1,
  	hcirc: hcirc$1,
  	hearts: hearts$1,
  	heartsuit: heartsuit$1,
  	hellip: hellip$1,
  	hercon: hercon$1,
  	hfr: hfr$1,
  	Hfr: Hfr$1,
  	HilbertSpace: HilbertSpace$1,
  	hksearow: hksearow$1,
  	hkswarow: hkswarow$1,
  	hoarr: hoarr$1,
  	homtht: homtht$1,
  	hookleftarrow: hookleftarrow$1,
  	hookrightarrow: hookrightarrow$1,
  	hopf: hopf$1,
  	Hopf: Hopf$1,
  	horbar: horbar$1,
  	HorizontalLine: HorizontalLine$1,
  	hscr: hscr$1,
  	Hscr: Hscr$1,
  	hslash: hslash$1,
  	Hstrok: Hstrok$1,
  	hstrok: hstrok$1,
  	HumpDownHump: HumpDownHump$1,
  	HumpEqual: HumpEqual$1,
  	hybull: hybull$1,
  	hyphen: hyphen$1,
  	Iacute: Iacute$2,
  	iacute: iacute$2,
  	ic: ic$1,
  	Icirc: Icirc$2,
  	icirc: icirc$2,
  	Icy: Icy$1,
  	icy: icy$1,
  	Idot: Idot$1,
  	IEcy: IEcy$1,
  	iecy: iecy$1,
  	iexcl: iexcl$2,
  	iff: iff$1,
  	ifr: ifr$1,
  	Ifr: Ifr$1,
  	Igrave: Igrave$2,
  	igrave: igrave$2,
  	ii: ii$1,
  	iiiint: iiiint$1,
  	iiint: iiint$1,
  	iinfin: iinfin$1,
  	iiota: iiota$1,
  	IJlig: IJlig$1,
  	ijlig: ijlig$1,
  	Imacr: Imacr$1,
  	imacr: imacr$1,
  	image: image$1,
  	ImaginaryI: ImaginaryI$1,
  	imagline: imagline$1,
  	imagpart: imagpart$1,
  	imath: imath$1,
  	Im: Im$1,
  	imof: imof$1,
  	imped: imped$1,
  	Implies: Implies$1,
  	incare: incare$1,
  	infin: infin$1,
  	infintie: infintie$1,
  	inodot: inodot$1,
  	intcal: intcal$1,
  	int: int$1,
  	Int: Int$1,
  	integers: integers$1,
  	Integral: Integral$1,
  	intercal: intercal$1,
  	Intersection: Intersection$1,
  	intlarhk: intlarhk$1,
  	intprod: intprod$1,
  	InvisibleComma: InvisibleComma$1,
  	InvisibleTimes: InvisibleTimes$1,
  	IOcy: IOcy$1,
  	iocy: iocy$1,
  	Iogon: Iogon$1,
  	iogon: iogon$1,
  	Iopf: Iopf$1,
  	iopf: iopf$1,
  	Iota: Iota$1,
  	iota: iota$1,
  	iprod: iprod$1,
  	iquest: iquest$2,
  	iscr: iscr$1,
  	Iscr: Iscr$1,
  	isin: isin$1,
  	isindot: isindot$1,
  	isinE: isinE$1,
  	isins: isins$1,
  	isinsv: isinsv$1,
  	isinv: isinv$1,
  	it: it$1,
  	Itilde: Itilde$1,
  	itilde: itilde$1,
  	Iukcy: Iukcy$1,
  	iukcy: iukcy$1,
  	Iuml: Iuml$2,
  	iuml: iuml$2,
  	Jcirc: Jcirc$1,
  	jcirc: jcirc$1,
  	Jcy: Jcy$1,
  	jcy: jcy$1,
  	Jfr: Jfr$1,
  	jfr: jfr$1,
  	jmath: jmath$1,
  	Jopf: Jopf$1,
  	jopf: jopf$1,
  	Jscr: Jscr$1,
  	jscr: jscr$1,
  	Jsercy: Jsercy$1,
  	jsercy: jsercy$1,
  	Jukcy: Jukcy$1,
  	jukcy: jukcy$1,
  	Kappa: Kappa$1,
  	kappa: kappa$1,
  	kappav: kappav$1,
  	Kcedil: Kcedil$1,
  	kcedil: kcedil$1,
  	Kcy: Kcy$1,
  	kcy: kcy$1,
  	Kfr: Kfr$1,
  	kfr: kfr$1,
  	kgreen: kgreen$1,
  	KHcy: KHcy$1,
  	khcy: khcy$1,
  	KJcy: KJcy$1,
  	kjcy: kjcy$1,
  	Kopf: Kopf$1,
  	kopf: kopf$1,
  	Kscr: Kscr$1,
  	kscr: kscr$1,
  	lAarr: lAarr$1,
  	Lacute: Lacute$1,
  	lacute: lacute$1,
  	laemptyv: laemptyv$1,
  	lagran: lagran$1,
  	Lambda: Lambda$1,
  	lambda: lambda$1,
  	lang: lang$1,
  	Lang: Lang$1,
  	langd: langd$1,
  	langle: langle$1,
  	lap: lap$1,
  	Laplacetrf: Laplacetrf$1,
  	laquo: laquo$2,
  	larrb: larrb$1,
  	larrbfs: larrbfs$1,
  	larr: larr$1,
  	Larr: Larr$1,
  	lArr: lArr$1,
  	larrfs: larrfs$1,
  	larrhk: larrhk$1,
  	larrlp: larrlp$1,
  	larrpl: larrpl$1,
  	larrsim: larrsim$1,
  	larrtl: larrtl$1,
  	latail: latail$1,
  	lAtail: lAtail$1,
  	lat: lat$1,
  	late: late$1,
  	lates: lates$1,
  	lbarr: lbarr$1,
  	lBarr: lBarr$1,
  	lbbrk: lbbrk$1,
  	lbrace: lbrace$1,
  	lbrack: lbrack$1,
  	lbrke: lbrke$1,
  	lbrksld: lbrksld$1,
  	lbrkslu: lbrkslu$1,
  	Lcaron: Lcaron$1,
  	lcaron: lcaron$1,
  	Lcedil: Lcedil$1,
  	lcedil: lcedil$1,
  	lceil: lceil$1,
  	lcub: lcub$1,
  	Lcy: Lcy$1,
  	lcy: lcy$1,
  	ldca: ldca$1,
  	ldquo: ldquo$1,
  	ldquor: ldquor$1,
  	ldrdhar: ldrdhar$1,
  	ldrushar: ldrushar$1,
  	ldsh: ldsh$1,
  	le: le$1,
  	lE: lE$1,
  	LeftAngleBracket: LeftAngleBracket$1,
  	LeftArrowBar: LeftArrowBar$1,
  	leftarrow: leftarrow$1,
  	LeftArrow: LeftArrow$1,
  	Leftarrow: Leftarrow$1,
  	LeftArrowRightArrow: LeftArrowRightArrow$1,
  	leftarrowtail: leftarrowtail$1,
  	LeftCeiling: LeftCeiling$1,
  	LeftDoubleBracket: LeftDoubleBracket$1,
  	LeftDownTeeVector: LeftDownTeeVector$1,
  	LeftDownVectorBar: LeftDownVectorBar$1,
  	LeftDownVector: LeftDownVector$1,
  	LeftFloor: LeftFloor$1,
  	leftharpoondown: leftharpoondown$1,
  	leftharpoonup: leftharpoonup$1,
  	leftleftarrows: leftleftarrows$1,
  	leftrightarrow: leftrightarrow$1,
  	LeftRightArrow: LeftRightArrow$1,
  	Leftrightarrow: Leftrightarrow$1,
  	leftrightarrows: leftrightarrows$1,
  	leftrightharpoons: leftrightharpoons$1,
  	leftrightsquigarrow: leftrightsquigarrow$1,
  	LeftRightVector: LeftRightVector$1,
  	LeftTeeArrow: LeftTeeArrow$1,
  	LeftTee: LeftTee$1,
  	LeftTeeVector: LeftTeeVector$1,
  	leftthreetimes: leftthreetimes$1,
  	LeftTriangleBar: LeftTriangleBar$1,
  	LeftTriangle: LeftTriangle$1,
  	LeftTriangleEqual: LeftTriangleEqual$1,
  	LeftUpDownVector: LeftUpDownVector$1,
  	LeftUpTeeVector: LeftUpTeeVector$1,
  	LeftUpVectorBar: LeftUpVectorBar$1,
  	LeftUpVector: LeftUpVector$1,
  	LeftVectorBar: LeftVectorBar$1,
  	LeftVector: LeftVector$1,
  	lEg: lEg$1,
  	leg: leg$1,
  	leq: leq$1,
  	leqq: leqq$1,
  	leqslant: leqslant$1,
  	lescc: lescc$1,
  	les: les$1,
  	lesdot: lesdot$1,
  	lesdoto: lesdoto$1,
  	lesdotor: lesdotor$1,
  	lesg: lesg$1,
  	lesges: lesges$1,
  	lessapprox: lessapprox$1,
  	lessdot: lessdot$1,
  	lesseqgtr: lesseqgtr$1,
  	lesseqqgtr: lesseqqgtr$1,
  	LessEqualGreater: LessEqualGreater$1,
  	LessFullEqual: LessFullEqual$1,
  	LessGreater: LessGreater$1,
  	lessgtr: lessgtr$1,
  	LessLess: LessLess$1,
  	lesssim: lesssim$1,
  	LessSlantEqual: LessSlantEqual$1,
  	LessTilde: LessTilde$1,
  	lfisht: lfisht$1,
  	lfloor: lfloor$1,
  	Lfr: Lfr$1,
  	lfr: lfr$1,
  	lg: lg$1,
  	lgE: lgE$1,
  	lHar: lHar$1,
  	lhard: lhard$1,
  	lharu: lharu$1,
  	lharul: lharul$1,
  	lhblk: lhblk$1,
  	LJcy: LJcy$1,
  	ljcy: ljcy$1,
  	llarr: llarr$1,
  	ll: ll$1,
  	Ll: Ll$1,
  	llcorner: llcorner$1,
  	Lleftarrow: Lleftarrow$1,
  	llhard: llhard$1,
  	lltri: lltri$1,
  	Lmidot: Lmidot$1,
  	lmidot: lmidot$1,
  	lmoustache: lmoustache$1,
  	lmoust: lmoust$1,
  	lnap: lnap$1,
  	lnapprox: lnapprox$1,
  	lne: lne$1,
  	lnE: lnE$1,
  	lneq: lneq$1,
  	lneqq: lneqq$1,
  	lnsim: lnsim$1,
  	loang: loang$1,
  	loarr: loarr$1,
  	lobrk: lobrk$1,
  	longleftarrow: longleftarrow$1,
  	LongLeftArrow: LongLeftArrow$1,
  	Longleftarrow: Longleftarrow$1,
  	longleftrightarrow: longleftrightarrow$1,
  	LongLeftRightArrow: LongLeftRightArrow$1,
  	Longleftrightarrow: Longleftrightarrow$1,
  	longmapsto: longmapsto$1,
  	longrightarrow: longrightarrow$1,
  	LongRightArrow: LongRightArrow$1,
  	Longrightarrow: Longrightarrow$1,
  	looparrowleft: looparrowleft$1,
  	looparrowright: looparrowright$1,
  	lopar: lopar$1,
  	Lopf: Lopf$1,
  	lopf: lopf$1,
  	loplus: loplus$1,
  	lotimes: lotimes$1,
  	lowast: lowast$1,
  	lowbar: lowbar$1,
  	LowerLeftArrow: LowerLeftArrow$1,
  	LowerRightArrow: LowerRightArrow$1,
  	loz: loz$1,
  	lozenge: lozenge$1,
  	lozf: lozf$1,
  	lpar: lpar$1,
  	lparlt: lparlt$1,
  	lrarr: lrarr$1,
  	lrcorner: lrcorner$1,
  	lrhar: lrhar$1,
  	lrhard: lrhard$1,
  	lrm: lrm$1,
  	lrtri: lrtri$1,
  	lsaquo: lsaquo$1,
  	lscr: lscr$1,
  	Lscr: Lscr$1,
  	lsh: lsh$1,
  	Lsh: Lsh$1,
  	lsim: lsim$1,
  	lsime: lsime$1,
  	lsimg: lsimg$1,
  	lsqb: lsqb$1,
  	lsquo: lsquo$1,
  	lsquor: lsquor$1,
  	Lstrok: Lstrok$1,
  	lstrok: lstrok$1,
  	ltcc: ltcc$1,
  	ltcir: ltcir$1,
  	lt: lt$1$1,
  	LT: LT$2,
  	Lt: Lt$1,
  	ltdot: ltdot$1,
  	lthree: lthree$1,
  	ltimes: ltimes$1,
  	ltlarr: ltlarr$1,
  	ltquest: ltquest$1,
  	ltri: ltri$1,
  	ltrie: ltrie$1,
  	ltrif: ltrif$1,
  	ltrPar: ltrPar$1,
  	lurdshar: lurdshar$1,
  	luruhar: luruhar$1,
  	lvertneqq: lvertneqq$1,
  	lvnE: lvnE$1,
  	macr: macr$2,
  	male: male$1,
  	malt: malt$1,
  	maltese: maltese$1,
  	map: map$1$1,
  	mapsto: mapsto$1,
  	mapstodown: mapstodown$1,
  	mapstoleft: mapstoleft$1,
  	mapstoup: mapstoup$1,
  	marker: marker$1,
  	mcomma: mcomma$1,
  	Mcy: Mcy$1,
  	mcy: mcy$1,
  	mdash: mdash$1,
  	mDDot: mDDot$1,
  	measuredangle: measuredangle$1,
  	MediumSpace: MediumSpace$1,
  	Mellintrf: Mellintrf$1,
  	Mfr: Mfr$1,
  	mfr: mfr$1,
  	mho: mho$1,
  	micro: micro$2,
  	midast: midast$1,
  	midcir: midcir$1,
  	mid: mid$1,
  	middot: middot$2,
  	minusb: minusb$1,
  	minus: minus$1,
  	minusd: minusd$1,
  	minusdu: minusdu$1,
  	MinusPlus: MinusPlus$1,
  	mlcp: mlcp$1,
  	mldr: mldr$1,
  	mnplus: mnplus$1,
  	models: models$1,
  	Mopf: Mopf$1,
  	mopf: mopf$1,
  	mp: mp$1,
  	mscr: mscr$1,
  	Mscr: Mscr$1,
  	mstpos: mstpos$1,
  	Mu: Mu$1,
  	mu: mu$1,
  	multimap: multimap$1,
  	mumap: mumap$1,
  	nabla: nabla$1,
  	Nacute: Nacute$1,
  	nacute: nacute$1,
  	nang: nang$1,
  	nap: nap$1,
  	napE: napE$1,
  	napid: napid$1,
  	napos: napos$1,
  	napprox: napprox$1,
  	natural: natural$1,
  	naturals: naturals$1,
  	natur: natur$1,
  	nbsp: nbsp$2,
  	nbump: nbump$1,
  	nbumpe: nbumpe$1,
  	ncap: ncap$1,
  	Ncaron: Ncaron$1,
  	ncaron: ncaron$1,
  	Ncedil: Ncedil$1,
  	ncedil: ncedil$1,
  	ncong: ncong$1,
  	ncongdot: ncongdot$1,
  	ncup: ncup$1,
  	Ncy: Ncy$1,
  	ncy: ncy$1,
  	ndash: ndash$1,
  	nearhk: nearhk$1,
  	nearr: nearr$1,
  	neArr: neArr$1,
  	nearrow: nearrow$1,
  	ne: ne$1,
  	nedot: nedot$1,
  	NegativeMediumSpace: NegativeMediumSpace$1,
  	NegativeThickSpace: NegativeThickSpace$1,
  	NegativeThinSpace: NegativeThinSpace$1,
  	NegativeVeryThinSpace: NegativeVeryThinSpace$1,
  	nequiv: nequiv$1,
  	nesear: nesear$1,
  	nesim: nesim$1,
  	NestedGreaterGreater: NestedGreaterGreater$1,
  	NestedLessLess: NestedLessLess$1,
  	NewLine: NewLine$1,
  	nexist: nexist$1,
  	nexists: nexists$1,
  	Nfr: Nfr$1,
  	nfr: nfr$1,
  	ngE: ngE$1,
  	nge: nge$1,
  	ngeq: ngeq$1,
  	ngeqq: ngeqq$1,
  	ngeqslant: ngeqslant$1,
  	nges: nges$1,
  	nGg: nGg$1,
  	ngsim: ngsim$1,
  	nGt: nGt$1,
  	ngt: ngt$1,
  	ngtr: ngtr$1,
  	nGtv: nGtv$1,
  	nharr: nharr$1,
  	nhArr: nhArr$1,
  	nhpar: nhpar$1,
  	ni: ni$1,
  	nis: nis$1,
  	nisd: nisd$1,
  	niv: niv$1,
  	NJcy: NJcy$1,
  	njcy: njcy$1,
  	nlarr: nlarr$1,
  	nlArr: nlArr$1,
  	nldr: nldr$1,
  	nlE: nlE$1,
  	nle: nle$1,
  	nleftarrow: nleftarrow$1,
  	nLeftarrow: nLeftarrow$1,
  	nleftrightarrow: nleftrightarrow$1,
  	nLeftrightarrow: nLeftrightarrow$1,
  	nleq: nleq$1,
  	nleqq: nleqq$1,
  	nleqslant: nleqslant$1,
  	nles: nles$1,
  	nless: nless$1,
  	nLl: nLl$1,
  	nlsim: nlsim$1,
  	nLt: nLt$1,
  	nlt: nlt$1,
  	nltri: nltri$1,
  	nltrie: nltrie$1,
  	nLtv: nLtv$1,
  	nmid: nmid$1,
  	NoBreak: NoBreak$1,
  	NonBreakingSpace: NonBreakingSpace$1,
  	nopf: nopf$1,
  	Nopf: Nopf$1,
  	Not: Not$1,
  	not: not$2,
  	NotCongruent: NotCongruent$1,
  	NotCupCap: NotCupCap$1,
  	NotDoubleVerticalBar: NotDoubleVerticalBar$1,
  	NotElement: NotElement$1,
  	NotEqual: NotEqual$1,
  	NotEqualTilde: NotEqualTilde$1,
  	NotExists: NotExists$1,
  	NotGreater: NotGreater$1,
  	NotGreaterEqual: NotGreaterEqual$1,
  	NotGreaterFullEqual: NotGreaterFullEqual$1,
  	NotGreaterGreater: NotGreaterGreater$1,
  	NotGreaterLess: NotGreaterLess$1,
  	NotGreaterSlantEqual: NotGreaterSlantEqual$1,
  	NotGreaterTilde: NotGreaterTilde$1,
  	NotHumpDownHump: NotHumpDownHump$1,
  	NotHumpEqual: NotHumpEqual$1,
  	notin: notin$1,
  	notindot: notindot$1,
  	notinE: notinE$1,
  	notinva: notinva$1,
  	notinvb: notinvb$1,
  	notinvc: notinvc$1,
  	NotLeftTriangleBar: NotLeftTriangleBar$1,
  	NotLeftTriangle: NotLeftTriangle$1,
  	NotLeftTriangleEqual: NotLeftTriangleEqual$1,
  	NotLess: NotLess$1,
  	NotLessEqual: NotLessEqual$1,
  	NotLessGreater: NotLessGreater$1,
  	NotLessLess: NotLessLess$1,
  	NotLessSlantEqual: NotLessSlantEqual$1,
  	NotLessTilde: NotLessTilde$1,
  	NotNestedGreaterGreater: NotNestedGreaterGreater$1,
  	NotNestedLessLess: NotNestedLessLess$1,
  	notni: notni$1,
  	notniva: notniva$1,
  	notnivb: notnivb$1,
  	notnivc: notnivc$1,
  	NotPrecedes: NotPrecedes$1,
  	NotPrecedesEqual: NotPrecedesEqual$1,
  	NotPrecedesSlantEqual: NotPrecedesSlantEqual$1,
  	NotReverseElement: NotReverseElement$1,
  	NotRightTriangleBar: NotRightTriangleBar$1,
  	NotRightTriangle: NotRightTriangle$1,
  	NotRightTriangleEqual: NotRightTriangleEqual$1,
  	NotSquareSubset: NotSquareSubset$1,
  	NotSquareSubsetEqual: NotSquareSubsetEqual$1,
  	NotSquareSuperset: NotSquareSuperset$1,
  	NotSquareSupersetEqual: NotSquareSupersetEqual$1,
  	NotSubset: NotSubset$1,
  	NotSubsetEqual: NotSubsetEqual$1,
  	NotSucceeds: NotSucceeds$1,
  	NotSucceedsEqual: NotSucceedsEqual$1,
  	NotSucceedsSlantEqual: NotSucceedsSlantEqual$1,
  	NotSucceedsTilde: NotSucceedsTilde$1,
  	NotSuperset: NotSuperset$1,
  	NotSupersetEqual: NotSupersetEqual$1,
  	NotTilde: NotTilde$1,
  	NotTildeEqual: NotTildeEqual$1,
  	NotTildeFullEqual: NotTildeFullEqual$1,
  	NotTildeTilde: NotTildeTilde$1,
  	NotVerticalBar: NotVerticalBar$1,
  	nparallel: nparallel$1,
  	npar: npar$1,
  	nparsl: nparsl$1,
  	npart: npart$1,
  	npolint: npolint$1,
  	npr: npr$1,
  	nprcue: nprcue$1,
  	nprec: nprec$1,
  	npreceq: npreceq$1,
  	npre: npre$1,
  	nrarrc: nrarrc$1,
  	nrarr: nrarr$1,
  	nrArr: nrArr$1,
  	nrarrw: nrarrw$1,
  	nrightarrow: nrightarrow$1,
  	nRightarrow: nRightarrow$1,
  	nrtri: nrtri$1,
  	nrtrie: nrtrie$1,
  	nsc: nsc$1,
  	nsccue: nsccue$1,
  	nsce: nsce$1,
  	Nscr: Nscr$1,
  	nscr: nscr$1,
  	nshortmid: nshortmid$1,
  	nshortparallel: nshortparallel$1,
  	nsim: nsim$1,
  	nsime: nsime$1,
  	nsimeq: nsimeq$1,
  	nsmid: nsmid$1,
  	nspar: nspar$1,
  	nsqsube: nsqsube$1,
  	nsqsupe: nsqsupe$1,
  	nsub: nsub$1,
  	nsubE: nsubE$1,
  	nsube: nsube$1,
  	nsubset: nsubset$1,
  	nsubseteq: nsubseteq$1,
  	nsubseteqq: nsubseteqq$1,
  	nsucc: nsucc$1,
  	nsucceq: nsucceq$1,
  	nsup: nsup$1,
  	nsupE: nsupE$1,
  	nsupe: nsupe$1,
  	nsupset: nsupset$1,
  	nsupseteq: nsupseteq$1,
  	nsupseteqq: nsupseteqq$1,
  	ntgl: ntgl$1,
  	Ntilde: Ntilde$2,
  	ntilde: ntilde$2,
  	ntlg: ntlg$1,
  	ntriangleleft: ntriangleleft$1,
  	ntrianglelefteq: ntrianglelefteq$1,
  	ntriangleright: ntriangleright$1,
  	ntrianglerighteq: ntrianglerighteq$1,
  	Nu: Nu$1,
  	nu: nu$1,
  	num: num$1,
  	numero: numero$1,
  	numsp: numsp$1,
  	nvap: nvap$1,
  	nvdash: nvdash$1,
  	nvDash: nvDash$1,
  	nVdash: nVdash$1,
  	nVDash: nVDash$1,
  	nvge: nvge$1,
  	nvgt: nvgt$1,
  	nvHarr: nvHarr$1,
  	nvinfin: nvinfin$1,
  	nvlArr: nvlArr$1,
  	nvle: nvle$1,
  	nvlt: nvlt$1,
  	nvltrie: nvltrie$1,
  	nvrArr: nvrArr$1,
  	nvrtrie: nvrtrie$1,
  	nvsim: nvsim$1,
  	nwarhk: nwarhk$1,
  	nwarr: nwarr$1,
  	nwArr: nwArr$1,
  	nwarrow: nwarrow$1,
  	nwnear: nwnear$1,
  	Oacute: Oacute$2,
  	oacute: oacute$2,
  	oast: oast$1,
  	Ocirc: Ocirc$2,
  	ocirc: ocirc$2,
  	ocir: ocir$1,
  	Ocy: Ocy$1,
  	ocy: ocy$1,
  	odash: odash$1,
  	Odblac: Odblac$1,
  	odblac: odblac$1,
  	odiv: odiv$1,
  	odot: odot$1,
  	odsold: odsold$1,
  	OElig: OElig$1,
  	oelig: oelig$1,
  	ofcir: ofcir$1,
  	Ofr: Ofr$1,
  	ofr: ofr$1,
  	ogon: ogon$1,
  	Ograve: Ograve$2,
  	ograve: ograve$2,
  	ogt: ogt$1,
  	ohbar: ohbar$1,
  	ohm: ohm$1,
  	oint: oint$1,
  	olarr: olarr$1,
  	olcir: olcir$1,
  	olcross: olcross$1,
  	oline: oline$1,
  	olt: olt$1,
  	Omacr: Omacr$1,
  	omacr: omacr$1,
  	Omega: Omega$1,
  	omega: omega$1,
  	Omicron: Omicron$1,
  	omicron: omicron$1,
  	omid: omid$1,
  	ominus: ominus$1,
  	Oopf: Oopf$1,
  	oopf: oopf$1,
  	opar: opar$1,
  	OpenCurlyDoubleQuote: OpenCurlyDoubleQuote$1,
  	OpenCurlyQuote: OpenCurlyQuote$1,
  	operp: operp$1,
  	oplus: oplus$1,
  	orarr: orarr$1,
  	Or: Or$1,
  	or: or$1,
  	ord: ord$1,
  	order: order$1,
  	orderof: orderof$1,
  	ordf: ordf$2,
  	ordm: ordm$2,
  	origof: origof$1,
  	oror: oror$1,
  	orslope: orslope$1,
  	orv: orv$1,
  	oS: oS$1,
  	Oscr: Oscr$1,
  	oscr: oscr$1,
  	Oslash: Oslash$2,
  	oslash: oslash$2,
  	osol: osol$1,
  	Otilde: Otilde$2,
  	otilde: otilde$2,
  	otimesas: otimesas$1,
  	Otimes: Otimes$1,
  	otimes: otimes$1,
  	Ouml: Ouml$2,
  	ouml: ouml$2,
  	ovbar: ovbar$1,
  	OverBar: OverBar$1,
  	OverBrace: OverBrace$1,
  	OverBracket: OverBracket$1,
  	OverParenthesis: OverParenthesis$1,
  	para: para$2,
  	parallel: parallel$1,
  	par: par$1,
  	parsim: parsim$1,
  	parsl: parsl$1,
  	part: part$1,
  	PartialD: PartialD$1,
  	Pcy: Pcy$1,
  	pcy: pcy$1,
  	percnt: percnt$1,
  	period: period$1,
  	permil: permil$1,
  	perp: perp$1,
  	pertenk: pertenk$1,
  	Pfr: Pfr$1,
  	pfr: pfr$1,
  	Phi: Phi$1,
  	phi: phi$1,
  	phiv: phiv$1,
  	phmmat: phmmat$1,
  	phone: phone$1,
  	Pi: Pi$1,
  	pi: pi$1,
  	pitchfork: pitchfork$1,
  	piv: piv$1,
  	planck: planck$1,
  	planckh: planckh$1,
  	plankv: plankv$1,
  	plusacir: plusacir$1,
  	plusb: plusb$1,
  	pluscir: pluscir$1,
  	plus: plus$1,
  	plusdo: plusdo$1,
  	plusdu: plusdu$1,
  	pluse: pluse$1,
  	PlusMinus: PlusMinus$1,
  	plusmn: plusmn$2,
  	plussim: plussim$1,
  	plustwo: plustwo$1,
  	pm: pm$1,
  	Poincareplane: Poincareplane$1,
  	pointint: pointint$1,
  	popf: popf$1,
  	Popf: Popf$1,
  	pound: pound$2,
  	prap: prap$1,
  	Pr: Pr$1,
  	pr: pr$1,
  	prcue: prcue$1,
  	precapprox: precapprox$1,
  	prec: prec$1,
  	preccurlyeq: preccurlyeq$1,
  	Precedes: Precedes$1,
  	PrecedesEqual: PrecedesEqual$1,
  	PrecedesSlantEqual: PrecedesSlantEqual$1,
  	PrecedesTilde: PrecedesTilde$1,
  	preceq: preceq$1,
  	precnapprox: precnapprox$1,
  	precneqq: precneqq$1,
  	precnsim: precnsim$1,
  	pre: pre$1,
  	prE: prE$1,
  	precsim: precsim$1,
  	prime: prime$1,
  	Prime: Prime$1,
  	primes: primes$1,
  	prnap: prnap$1,
  	prnE: prnE$1,
  	prnsim: prnsim$1,
  	prod: prod$1,
  	Product: Product$1,
  	profalar: profalar$1,
  	profline: profline$1,
  	profsurf: profsurf$1,
  	prop: prop$1,
  	Proportional: Proportional$1,
  	Proportion: Proportion$1,
  	propto: propto$1,
  	prsim: prsim$1,
  	prurel: prurel$1,
  	Pscr: Pscr$1,
  	pscr: pscr$1,
  	Psi: Psi$1,
  	psi: psi$1,
  	puncsp: puncsp$1,
  	Qfr: Qfr$1,
  	qfr: qfr$1,
  	qint: qint$1,
  	qopf: qopf$1,
  	Qopf: Qopf$1,
  	qprime: qprime$1,
  	Qscr: Qscr$1,
  	qscr: qscr$1,
  	quaternions: quaternions$1,
  	quatint: quatint$1,
  	quest: quest$1,
  	questeq: questeq$1,
  	quot: quot$1$1,
  	QUOT: QUOT$2,
  	rAarr: rAarr$1,
  	race: race$1,
  	Racute: Racute$1,
  	racute: racute$1,
  	radic: radic$1,
  	raemptyv: raemptyv$1,
  	rang: rang$1,
  	Rang: Rang$1,
  	rangd: rangd$1,
  	range: range$1,
  	rangle: rangle$1,
  	raquo: raquo$2,
  	rarrap: rarrap$1,
  	rarrb: rarrb$1,
  	rarrbfs: rarrbfs$1,
  	rarrc: rarrc$1,
  	rarr: rarr$1,
  	Rarr: Rarr$1,
  	rArr: rArr$1,
  	rarrfs: rarrfs$1,
  	rarrhk: rarrhk$1,
  	rarrlp: rarrlp$1,
  	rarrpl: rarrpl$1,
  	rarrsim: rarrsim$1,
  	Rarrtl: Rarrtl$1,
  	rarrtl: rarrtl$1,
  	rarrw: rarrw$1,
  	ratail: ratail$1,
  	rAtail: rAtail$1,
  	ratio: ratio$1,
  	rationals: rationals$1,
  	rbarr: rbarr$1,
  	rBarr: rBarr$1,
  	RBarr: RBarr$1,
  	rbbrk: rbbrk$1,
  	rbrace: rbrace$1,
  	rbrack: rbrack$1,
  	rbrke: rbrke$1,
  	rbrksld: rbrksld$1,
  	rbrkslu: rbrkslu$1,
  	Rcaron: Rcaron$1,
  	rcaron: rcaron$1,
  	Rcedil: Rcedil$1,
  	rcedil: rcedil$1,
  	rceil: rceil$1,
  	rcub: rcub$1,
  	Rcy: Rcy$1,
  	rcy: rcy$1,
  	rdca: rdca$1,
  	rdldhar: rdldhar$1,
  	rdquo: rdquo$1,
  	rdquor: rdquor$1,
  	rdsh: rdsh$1,
  	real: real$1,
  	realine: realine$1,
  	realpart: realpart$1,
  	reals: reals$1,
  	Re: Re$1,
  	rect: rect$1,
  	reg: reg$2,
  	REG: REG$2,
  	ReverseElement: ReverseElement$1,
  	ReverseEquilibrium: ReverseEquilibrium$1,
  	ReverseUpEquilibrium: ReverseUpEquilibrium$1,
  	rfisht: rfisht$1,
  	rfloor: rfloor$1,
  	rfr: rfr$1,
  	Rfr: Rfr$1,
  	rHar: rHar$1,
  	rhard: rhard$1,
  	rharu: rharu$1,
  	rharul: rharul$1,
  	Rho: Rho$1,
  	rho: rho$1,
  	rhov: rhov$1,
  	RightAngleBracket: RightAngleBracket$1,
  	RightArrowBar: RightArrowBar$1,
  	rightarrow: rightarrow$1,
  	RightArrow: RightArrow$1,
  	Rightarrow: Rightarrow$1,
  	RightArrowLeftArrow: RightArrowLeftArrow$1,
  	rightarrowtail: rightarrowtail$1,
  	RightCeiling: RightCeiling$1,
  	RightDoubleBracket: RightDoubleBracket$1,
  	RightDownTeeVector: RightDownTeeVector$1,
  	RightDownVectorBar: RightDownVectorBar$1,
  	RightDownVector: RightDownVector$1,
  	RightFloor: RightFloor$1,
  	rightharpoondown: rightharpoondown$1,
  	rightharpoonup: rightharpoonup$1,
  	rightleftarrows: rightleftarrows$1,
  	rightleftharpoons: rightleftharpoons$1,
  	rightrightarrows: rightrightarrows$1,
  	rightsquigarrow: rightsquigarrow$1,
  	RightTeeArrow: RightTeeArrow$1,
  	RightTee: RightTee$1,
  	RightTeeVector: RightTeeVector$1,
  	rightthreetimes: rightthreetimes$1,
  	RightTriangleBar: RightTriangleBar$1,
  	RightTriangle: RightTriangle$1,
  	RightTriangleEqual: RightTriangleEqual$1,
  	RightUpDownVector: RightUpDownVector$1,
  	RightUpTeeVector: RightUpTeeVector$1,
  	RightUpVectorBar: RightUpVectorBar$1,
  	RightUpVector: RightUpVector$1,
  	RightVectorBar: RightVectorBar$1,
  	RightVector: RightVector$1,
  	ring: ring$1,
  	risingdotseq: risingdotseq$1,
  	rlarr: rlarr$1,
  	rlhar: rlhar$1,
  	rlm: rlm$1,
  	rmoustache: rmoustache$1,
  	rmoust: rmoust$1,
  	rnmid: rnmid$1,
  	roang: roang$1,
  	roarr: roarr$1,
  	robrk: robrk$1,
  	ropar: ropar$1,
  	ropf: ropf$1,
  	Ropf: Ropf$1,
  	roplus: roplus$1,
  	rotimes: rotimes$1,
  	RoundImplies: RoundImplies$1,
  	rpar: rpar$1,
  	rpargt: rpargt$1,
  	rppolint: rppolint$1,
  	rrarr: rrarr$1,
  	Rrightarrow: Rrightarrow$1,
  	rsaquo: rsaquo$1,
  	rscr: rscr$1,
  	Rscr: Rscr$1,
  	rsh: rsh$1,
  	Rsh: Rsh$1,
  	rsqb: rsqb$1,
  	rsquo: rsquo$1,
  	rsquor: rsquor$1,
  	rthree: rthree$1,
  	rtimes: rtimes$1,
  	rtri: rtri$1,
  	rtrie: rtrie$1,
  	rtrif: rtrif$1,
  	rtriltri: rtriltri$1,
  	RuleDelayed: RuleDelayed$1,
  	ruluhar: ruluhar$1,
  	rx: rx$1,
  	Sacute: Sacute$1,
  	sacute: sacute$1,
  	sbquo: sbquo$1,
  	scap: scap$1,
  	Scaron: Scaron$1,
  	scaron: scaron$1,
  	Sc: Sc$1,
  	sc: sc$1,
  	sccue: sccue$1,
  	sce: sce$1,
  	scE: scE$1,
  	Scedil: Scedil$1,
  	scedil: scedil$1,
  	Scirc: Scirc$1,
  	scirc: scirc$1,
  	scnap: scnap$1,
  	scnE: scnE$1,
  	scnsim: scnsim$1,
  	scpolint: scpolint$1,
  	scsim: scsim$1,
  	Scy: Scy$1,
  	scy: scy$1,
  	sdotb: sdotb$1,
  	sdot: sdot$1,
  	sdote: sdote$1,
  	searhk: searhk$1,
  	searr: searr$1,
  	seArr: seArr$1,
  	searrow: searrow$1,
  	sect: sect$2,
  	semi: semi$1,
  	seswar: seswar$1,
  	setminus: setminus$1,
  	setmn: setmn$1,
  	sext: sext$1,
  	Sfr: Sfr$1,
  	sfr: sfr$1,
  	sfrown: sfrown$1,
  	sharp: sharp$1,
  	SHCHcy: SHCHcy$1,
  	shchcy: shchcy$1,
  	SHcy: SHcy$1,
  	shcy: shcy$1,
  	ShortDownArrow: ShortDownArrow$1,
  	ShortLeftArrow: ShortLeftArrow$1,
  	shortmid: shortmid$1,
  	shortparallel: shortparallel$1,
  	ShortRightArrow: ShortRightArrow$1,
  	ShortUpArrow: ShortUpArrow$1,
  	shy: shy$2,
  	Sigma: Sigma$1,
  	sigma: sigma$1,
  	sigmaf: sigmaf$1,
  	sigmav: sigmav$1,
  	sim: sim$1,
  	simdot: simdot$1,
  	sime: sime$1,
  	simeq: simeq$1,
  	simg: simg$1,
  	simgE: simgE$1,
  	siml: siml$1,
  	simlE: simlE$1,
  	simne: simne$1,
  	simplus: simplus$1,
  	simrarr: simrarr$1,
  	slarr: slarr$1,
  	SmallCircle: SmallCircle$1,
  	smallsetminus: smallsetminus$1,
  	smashp: smashp$1,
  	smeparsl: smeparsl$1,
  	smid: smid$1,
  	smile: smile$1,
  	smt: smt$1,
  	smte: smte$1,
  	smtes: smtes$1,
  	SOFTcy: SOFTcy$1,
  	softcy: softcy$1,
  	solbar: solbar$1,
  	solb: solb$1,
  	sol: sol$1,
  	Sopf: Sopf$1,
  	sopf: sopf$1,
  	spades: spades$1,
  	spadesuit: spadesuit$1,
  	spar: spar$1,
  	sqcap: sqcap$1,
  	sqcaps: sqcaps$1,
  	sqcup: sqcup$1,
  	sqcups: sqcups$1,
  	Sqrt: Sqrt$1,
  	sqsub: sqsub$1,
  	sqsube: sqsube$1,
  	sqsubset: sqsubset$1,
  	sqsubseteq: sqsubseteq$1,
  	sqsup: sqsup$1,
  	sqsupe: sqsupe$1,
  	sqsupset: sqsupset$1,
  	sqsupseteq: sqsupseteq$1,
  	square: square$1,
  	Square: Square$1,
  	SquareIntersection: SquareIntersection$1,
  	SquareSubset: SquareSubset$1,
  	SquareSubsetEqual: SquareSubsetEqual$1,
  	SquareSuperset: SquareSuperset$1,
  	SquareSupersetEqual: SquareSupersetEqual$1,
  	SquareUnion: SquareUnion$1,
  	squarf: squarf$1,
  	squ: squ$1,
  	squf: squf$1,
  	srarr: srarr$1,
  	Sscr: Sscr$1,
  	sscr: sscr$1,
  	ssetmn: ssetmn$1,
  	ssmile: ssmile$1,
  	sstarf: sstarf$1,
  	Star: Star$1,
  	star: star$1,
  	starf: starf$1,
  	straightepsilon: straightepsilon$1,
  	straightphi: straightphi$1,
  	strns: strns$1,
  	sub: sub$1,
  	Sub: Sub$1,
  	subdot: subdot$1,
  	subE: subE$1,
  	sube: sube$1,
  	subedot: subedot$1,
  	submult: submult$1,
  	subnE: subnE$1,
  	subne: subne$1,
  	subplus: subplus$1,
  	subrarr: subrarr$1,
  	subset: subset$1,
  	Subset: Subset$1,
  	subseteq: subseteq$1,
  	subseteqq: subseteqq$1,
  	SubsetEqual: SubsetEqual$1,
  	subsetneq: subsetneq$1,
  	subsetneqq: subsetneqq$1,
  	subsim: subsim$1,
  	subsub: subsub$1,
  	subsup: subsup$1,
  	succapprox: succapprox$1,
  	succ: succ$1,
  	succcurlyeq: succcurlyeq$1,
  	Succeeds: Succeeds$1,
  	SucceedsEqual: SucceedsEqual$1,
  	SucceedsSlantEqual: SucceedsSlantEqual$1,
  	SucceedsTilde: SucceedsTilde$1,
  	succeq: succeq$1,
  	succnapprox: succnapprox$1,
  	succneqq: succneqq$1,
  	succnsim: succnsim$1,
  	succsim: succsim$1,
  	SuchThat: SuchThat$1,
  	sum: sum$1,
  	Sum: Sum$1,
  	sung: sung$1,
  	sup1: sup1$2,
  	sup2: sup2$2,
  	sup3: sup3$2,
  	sup: sup$1,
  	Sup: Sup$1,
  	supdot: supdot$1,
  	supdsub: supdsub$1,
  	supE: supE$1,
  	supe: supe$1,
  	supedot: supedot$1,
  	Superset: Superset$1,
  	SupersetEqual: SupersetEqual$1,
  	suphsol: suphsol$1,
  	suphsub: suphsub$1,
  	suplarr: suplarr$1,
  	supmult: supmult$1,
  	supnE: supnE$1,
  	supne: supne$1,
  	supplus: supplus$1,
  	supset: supset$1,
  	Supset: Supset$1,
  	supseteq: supseteq$1,
  	supseteqq: supseteqq$1,
  	supsetneq: supsetneq$1,
  	supsetneqq: supsetneqq$1,
  	supsim: supsim$1,
  	supsub: supsub$1,
  	supsup: supsup$1,
  	swarhk: swarhk$1,
  	swarr: swarr$1,
  	swArr: swArr$1,
  	swarrow: swarrow$1,
  	swnwar: swnwar$1,
  	szlig: szlig$2,
  	Tab: Tab$1,
  	target: target$1,
  	Tau: Tau$1,
  	tau: tau$1,
  	tbrk: tbrk$1,
  	Tcaron: Tcaron$1,
  	tcaron: tcaron$1,
  	Tcedil: Tcedil$1,
  	tcedil: tcedil$1,
  	Tcy: Tcy$1,
  	tcy: tcy$1,
  	tdot: tdot$1,
  	telrec: telrec$1,
  	Tfr: Tfr$1,
  	tfr: tfr$1,
  	there4: there4$1,
  	therefore: therefore$1,
  	Therefore: Therefore$1,
  	Theta: Theta$1,
  	theta: theta$1,
  	thetasym: thetasym$1,
  	thetav: thetav$1,
  	thickapprox: thickapprox$1,
  	thicksim: thicksim$1,
  	ThickSpace: ThickSpace$1,
  	ThinSpace: ThinSpace$1,
  	thinsp: thinsp$1,
  	thkap: thkap$1,
  	thksim: thksim$1,
  	THORN: THORN$2,
  	thorn: thorn$2,
  	tilde: tilde$1,
  	Tilde: Tilde$1,
  	TildeEqual: TildeEqual$1,
  	TildeFullEqual: TildeFullEqual$1,
  	TildeTilde: TildeTilde$1,
  	timesbar: timesbar$1,
  	timesb: timesb$1,
  	times: times$3,
  	timesd: timesd$1,
  	tint: tint$1,
  	toea: toea$1,
  	topbot: topbot$1,
  	topcir: topcir$1,
  	top: top$1,
  	Topf: Topf$1,
  	topf: topf$1,
  	topfork: topfork$1,
  	tosa: tosa$1,
  	tprime: tprime$1,
  	trade: trade$1,
  	TRADE: TRADE$1,
  	triangle: triangle$1,
  	triangledown: triangledown$1,
  	triangleleft: triangleleft$1,
  	trianglelefteq: trianglelefteq$1,
  	triangleq: triangleq$1,
  	triangleright: triangleright$1,
  	trianglerighteq: trianglerighteq$1,
  	tridot: tridot$1,
  	trie: trie$1,
  	triminus: triminus$1,
  	TripleDot: TripleDot$1,
  	triplus: triplus$1,
  	trisb: trisb$1,
  	tritime: tritime$1,
  	trpezium: trpezium$1,
  	Tscr: Tscr$1,
  	tscr: tscr$1,
  	TScy: TScy$1,
  	tscy: tscy$1,
  	TSHcy: TSHcy$1,
  	tshcy: tshcy$1,
  	Tstrok: Tstrok$1,
  	tstrok: tstrok$1,
  	twixt: twixt$1,
  	twoheadleftarrow: twoheadleftarrow$1,
  	twoheadrightarrow: twoheadrightarrow$1,
  	Uacute: Uacute$2,
  	uacute: uacute$2,
  	uarr: uarr$1,
  	Uarr: Uarr$1,
  	uArr: uArr$1,
  	Uarrocir: Uarrocir$1,
  	Ubrcy: Ubrcy$1,
  	ubrcy: ubrcy$1,
  	Ubreve: Ubreve$1,
  	ubreve: ubreve$1,
  	Ucirc: Ucirc$2,
  	ucirc: ucirc$2,
  	Ucy: Ucy$1,
  	ucy: ucy$1,
  	udarr: udarr$1,
  	Udblac: Udblac$1,
  	udblac: udblac$1,
  	udhar: udhar$1,
  	ufisht: ufisht$1,
  	Ufr: Ufr$1,
  	ufr: ufr$1,
  	Ugrave: Ugrave$2,
  	ugrave: ugrave$2,
  	uHar: uHar$1,
  	uharl: uharl$1,
  	uharr: uharr$1,
  	uhblk: uhblk$1,
  	ulcorn: ulcorn$1,
  	ulcorner: ulcorner$1,
  	ulcrop: ulcrop$1,
  	ultri: ultri$1,
  	Umacr: Umacr$1,
  	umacr: umacr$1,
  	uml: uml$2,
  	UnderBar: UnderBar$1,
  	UnderBrace: UnderBrace$1,
  	UnderBracket: UnderBracket$1,
  	UnderParenthesis: UnderParenthesis$1,
  	Union: Union$1,
  	UnionPlus: UnionPlus$1,
  	Uogon: Uogon$1,
  	uogon: uogon$1,
  	Uopf: Uopf$1,
  	uopf: uopf$1,
  	UpArrowBar: UpArrowBar$1,
  	uparrow: uparrow$1,
  	UpArrow: UpArrow$1,
  	Uparrow: Uparrow$1,
  	UpArrowDownArrow: UpArrowDownArrow$1,
  	updownarrow: updownarrow$1,
  	UpDownArrow: UpDownArrow$1,
  	Updownarrow: Updownarrow$1,
  	UpEquilibrium: UpEquilibrium$1,
  	upharpoonleft: upharpoonleft$1,
  	upharpoonright: upharpoonright$1,
  	uplus: uplus$1,
  	UpperLeftArrow: UpperLeftArrow$1,
  	UpperRightArrow: UpperRightArrow$1,
  	upsi: upsi$1,
  	Upsi: Upsi$1,
  	upsih: upsih$1,
  	Upsilon: Upsilon$1,
  	upsilon: upsilon$1,
  	UpTeeArrow: UpTeeArrow$1,
  	UpTee: UpTee$1,
  	upuparrows: upuparrows$1,
  	urcorn: urcorn$1,
  	urcorner: urcorner$1,
  	urcrop: urcrop$1,
  	Uring: Uring$1,
  	uring: uring$1,
  	urtri: urtri$1,
  	Uscr: Uscr$1,
  	uscr: uscr$1,
  	utdot: utdot$1,
  	Utilde: Utilde$1,
  	utilde: utilde$1,
  	utri: utri$1,
  	utrif: utrif$1,
  	uuarr: uuarr$1,
  	Uuml: Uuml$2,
  	uuml: uuml$2,
  	uwangle: uwangle$1,
  	vangrt: vangrt$1,
  	varepsilon: varepsilon$1,
  	varkappa: varkappa$1,
  	varnothing: varnothing$1,
  	varphi: varphi$1,
  	varpi: varpi$1,
  	varpropto: varpropto$1,
  	varr: varr$1,
  	vArr: vArr$1,
  	varrho: varrho$1,
  	varsigma: varsigma$1,
  	varsubsetneq: varsubsetneq$1,
  	varsubsetneqq: varsubsetneqq$1,
  	varsupsetneq: varsupsetneq$1,
  	varsupsetneqq: varsupsetneqq$1,
  	vartheta: vartheta$1,
  	vartriangleleft: vartriangleleft$1,
  	vartriangleright: vartriangleright$1,
  	vBar: vBar$1,
  	Vbar: Vbar$1,
  	vBarv: vBarv$1,
  	Vcy: Vcy$1,
  	vcy: vcy$1,
  	vdash: vdash$1,
  	vDash: vDash$1,
  	Vdash: Vdash$1,
  	VDash: VDash$1,
  	Vdashl: Vdashl$1,
  	veebar: veebar$1,
  	vee: vee$1,
  	Vee: Vee$1,
  	veeeq: veeeq$1,
  	vellip: vellip$1,
  	verbar: verbar$1,
  	Verbar: Verbar$1,
  	vert: vert$1,
  	Vert: Vert$1,
  	VerticalBar: VerticalBar$1,
  	VerticalLine: VerticalLine$1,
  	VerticalSeparator: VerticalSeparator$1,
  	VerticalTilde: VerticalTilde$1,
  	VeryThinSpace: VeryThinSpace$1,
  	Vfr: Vfr$1,
  	vfr: vfr$1,
  	vltri: vltri$1,
  	vnsub: vnsub$1,
  	vnsup: vnsup$1,
  	Vopf: Vopf$1,
  	vopf: vopf$1,
  	vprop: vprop$1,
  	vrtri: vrtri$1,
  	Vscr: Vscr$1,
  	vscr: vscr$1,
  	vsubnE: vsubnE$1,
  	vsubne: vsubne$1,
  	vsupnE: vsupnE$1,
  	vsupne: vsupne$1,
  	Vvdash: Vvdash$1,
  	vzigzag: vzigzag$1,
  	Wcirc: Wcirc$1,
  	wcirc: wcirc$1,
  	wedbar: wedbar$1,
  	wedge: wedge$1,
  	Wedge: Wedge$1,
  	wedgeq: wedgeq$1,
  	weierp: weierp$1,
  	Wfr: Wfr$1,
  	wfr: wfr$1,
  	Wopf: Wopf$1,
  	wopf: wopf$1,
  	wp: wp$1,
  	wr: wr$1,
  	wreath: wreath$1,
  	Wscr: Wscr$1,
  	wscr: wscr$1,
  	xcap: xcap$1,
  	xcirc: xcirc$1,
  	xcup: xcup$1,
  	xdtri: xdtri$1,
  	Xfr: Xfr$1,
  	xfr: xfr$1,
  	xharr: xharr$1,
  	xhArr: xhArr$1,
  	Xi: Xi$1,
  	xi: xi$1,
  	xlarr: xlarr$1,
  	xlArr: xlArr$1,
  	xmap: xmap$1,
  	xnis: xnis$1,
  	xodot: xodot$1,
  	Xopf: Xopf$1,
  	xopf: xopf$1,
  	xoplus: xoplus$1,
  	xotime: xotime$1,
  	xrarr: xrarr$1,
  	xrArr: xrArr$1,
  	Xscr: Xscr$1,
  	xscr: xscr$1,
  	xsqcup: xsqcup$1,
  	xuplus: xuplus$1,
  	xutri: xutri$1,
  	xvee: xvee$1,
  	xwedge: xwedge$1,
  	Yacute: Yacute$2,
  	yacute: yacute$2,
  	YAcy: YAcy$1,
  	yacy: yacy$1,
  	Ycirc: Ycirc$1,
  	ycirc: ycirc$1,
  	Ycy: Ycy$1,
  	ycy: ycy$1,
  	yen: yen$2,
  	Yfr: Yfr$1,
  	yfr: yfr$1,
  	YIcy: YIcy$1,
  	yicy: yicy$1,
  	Yopf: Yopf$1,
  	yopf: yopf$1,
  	Yscr: Yscr$1,
  	yscr: yscr$1,
  	YUcy: YUcy$1,
  	yucy: yucy$1,
  	yuml: yuml$2,
  	Yuml: Yuml$1,
  	Zacute: Zacute$1,
  	zacute: zacute$1,
  	Zcaron: Zcaron$1,
  	zcaron: zcaron$1,
  	Zcy: Zcy$1,
  	zcy: zcy$1,
  	Zdot: Zdot$1,
  	zdot: zdot$1,
  	zeetrf: zeetrf$1,
  	ZeroWidthSpace: ZeroWidthSpace$1,
  	Zeta: Zeta$1,
  	zeta: zeta$1,
  	zfr: zfr$1,
  	Zfr: Zfr$1,
  	ZHcy: ZHcy$1,
  	zhcy: zhcy$1,
  	zigrarr: zigrarr$1,
  	zopf: zopf$1,
  	Zopf: Zopf$1,
  	Zscr: Zscr$1,
  	zscr: zscr$1,
  	zwj: zwj$1,
  	zwnj: zwnj$1,
  	default: entitiesJSON$1
  });

  var require$$0$1 = ( xml$1 && xmlJSON$1 ) || xml$1;

  var require$$1$2 = ( entities$1 && entitiesJSON$1 ) || entities$1;

  var inverseXML$1 = getInverseObj$1$1(require$$0$1);
  var xmlReplacer$1 = getInverseReplacer$1$1(inverseXML$1);

  var XML$1 = getInverse$1$1(inverseXML$1, xmlReplacer$1);

  var inverseHTML$1 = getInverseObj$1$1(require$$1$2);
  var htmlReplacer$1 = getInverseReplacer$1$1(inverseHTML$1);

  var HTML$1 = getInverse$1$1(inverseHTML$1, htmlReplacer$1);

  function getInverseObj$1$1(obj){
  	return Object.keys(obj).sort().reduce(function(inverse, name){
  		inverse[obj[name]] = "&" + name + ";";
  		return inverse;
  	}, {});
  }

  function getInverseReplacer$1$1(inverse){
  	var single = [],
  	    multiple = [];

  	Object.keys(inverse).forEach(function(k){
  		if(k.length === 1){
  			single.push("\\" + k);
  		} else {
  			multiple.push(k);
  		}
  	});

  	
  	multiple.unshift("[" + single.join("") + "]");

  	return new RegExp(multiple.join("|"), "g");
  }

  var re_nonASCII$1 = /[^\0-\x7F]/g;
  var re_astralSymbols$1 = /[\uD800-\uDBFF][\uDC00-\uDFFF]/g;

  function singleCharReplacer$1$1(c){
  	return "&#x" + c.charCodeAt(0).toString(16).toUpperCase() + ";";
  }

  function astralReplacer$1$1(c){
  	
  	var high = c.charCodeAt(0);
  	var low  = c.charCodeAt(1);
  	var codePoint = (high - 0xD800) * 0x400 + low - 0xDC00 + 0x10000;
  	return "&#x" + codePoint.toString(16).toUpperCase() + ";";
  }

  function getInverse$1$1(inverse, re){
  	function func(name){
  		return inverse[name];
  	}

  	return function(data){
  		return data
  				.replace(re, func)
  				.replace(re_astralSymbols$1, astralReplacer$1$1)
  				.replace(re_nonASCII$1, singleCharReplacer$1$1);
  	};
  }

  var re_xmlChars$1 = getInverseReplacer$1$1(inverseXML$1);

  function escapeXML$1(data){
  	return data
  			.replace(re_xmlChars$1, singleCharReplacer$1$1)
  			.replace(re_astralSymbols$1, astralReplacer$1$1)
  			.replace(re_nonASCII$1, singleCharReplacer$1$1);
  }

  var escape$1 = escapeXML$1;

  var encode$1 = {
  	XML: XML$1,
  	HTML: HTML$1,
  	escape: escape$1
  };

  var decode$2 = {
  	"0": 65533,
  	"128": 8364,
  	"130": 8218,
  	"131": 402,
  	"132": 8222,
  	"133": 8230,
  	"134": 8224,
  	"135": 8225,
  	"136": 710,
  	"137": 8240,
  	"138": 352,
  	"139": 8249,
  	"140": 338,
  	"142": 381,
  	"145": 8216,
  	"146": 8217,
  	"147": 8220,
  	"148": 8221,
  	"149": 8226,
  	"150": 8211,
  	"151": 8212,
  	"152": 732,
  	"153": 8482,
  	"154": 353,
  	"155": 8250,
  	"156": 339,
  	"158": 382,
  	"159": 376
  };

  var decode$1$1 = Object.freeze({
  	default: decode$2
  });

  var decodeMap$1 = ( decode$1$1 && decode$2 ) || decode$1$1;

  var decode_codepoint$1 = decodeCodePoint$2;


  function decodeCodePoint$2(codePoint){

  	if((codePoint >= 0xD800 && codePoint <= 0xDFFF) || codePoint > 0x10FFFF){
  		return "\uFFFD";
  	}

  	if(codePoint in decodeMap$1){
  		codePoint = decodeMap$1[codePoint];
  	}

  	var output = "";

  	if(codePoint > 0xFFFF){
  		codePoint -= 0x10000;
  		output += String.fromCharCode(codePoint >>> 10 & 0x3FF | 0xD800);
  		codePoint = 0xDC00 | codePoint & 0x3FF;
  	}

  	output += String.fromCharCode(codePoint);
  	return output;
  }

  var Aacute$1$1 = "Á";
  var aacute$1$1 = "á";
  var Acirc$1$1 = "Â";
  var acirc$1$1 = "â";
  var acute$1$1 = "´";
  var AElig$1$1 = "Æ";
  var aelig$1$1 = "æ";
  var Agrave$1$1 = "À";
  var agrave$1$1 = "à";
  var amp$2$1 = "&";
  var AMP$1$1 = "&";
  var Aring$1$1 = "Å";
  var aring$1$1 = "å";
  var Atilde$1$1 = "Ã";
  var atilde$1$1 = "ã";
  var Auml$1$1 = "Ä";
  var auml$1$1 = "ä";
  var brvbar$1$1 = "¦";
  var Ccedil$1$1 = "Ç";
  var ccedil$1$1 = "ç";
  var cedil$1$1 = "¸";
  var cent$1$1 = "¢";
  var copy$1$1 = "©";
  var COPY$1$1 = "©";
  var curren$1$1 = "¤";
  var deg$1$1 = "°";
  var divide$1$1 = "÷";
  var Eacute$1$1 = "É";
  var eacute$1$1 = "é";
  var Ecirc$1$1 = "Ê";
  var ecirc$1$1 = "ê";
  var Egrave$1$1 = "È";
  var egrave$1$1 = "è";
  var ETH$1$1 = "Ð";
  var eth$1$1 = "ð";
  var Euml$1$1 = "Ë";
  var euml$1$1 = "ë";
  var frac12$1$1 = "½";
  var frac14$1$1 = "¼";
  var frac34$1$1 = "¾";
  var gt$2$1 = ">";
  var GT$1$1 = ">";
  var Iacute$1$1 = "Í";
  var iacute$1$1 = "í";
  var Icirc$1$1 = "Î";
  var icirc$1$1 = "î";
  var iexcl$1$1 = "¡";
  var Igrave$1$1 = "Ì";
  var igrave$1$1 = "ì";
  var iquest$1$1 = "¿";
  var Iuml$1$1 = "Ï";
  var iuml$1$1 = "ï";
  var laquo$1$1 = "«";
  var lt$2$1 = "<";
  var LT$1$1 = "<";
  var macr$1$1 = "¯";
  var micro$1$1 = "µ";
  var middot$1$1 = "·";
  var nbsp$1$1 = " ";
  var not$1$1 = "¬";
  var Ntilde$1$1 = "Ñ";
  var ntilde$1$1 = "ñ";
  var Oacute$1$1 = "Ó";
  var oacute$1$1 = "ó";
  var Ocirc$1$1 = "Ô";
  var ocirc$1$1 = "ô";
  var Ograve$1$1 = "Ò";
  var ograve$1$1 = "ò";
  var ordf$1$1 = "ª";
  var ordm$1$1 = "º";
  var Oslash$1$1 = "Ø";
  var oslash$1$1 = "ø";
  var Otilde$1$1 = "Õ";
  var otilde$1$1 = "õ";
  var Ouml$1$1 = "Ö";
  var ouml$1$1 = "ö";
  var para$1$1 = "¶";
  var plusmn$1$1 = "±";
  var pound$1$1 = "£";
  var quot$2$1 = "\"";
  var QUOT$1$1 = "\"";
  var raquo$1$1 = "»";
  var reg$1$1 = "®";
  var REG$1$1 = "®";
  var sect$1$1 = "§";
  var shy$1$1 = "­";
  var sup1$1$1 = "¹";
  var sup2$1$1 = "²";
  var sup3$1$1 = "³";
  var szlig$1$1 = "ß";
  var THORN$1$1 = "Þ";
  var thorn$1$1 = "þ";
  var times$1$1 = "×";
  var Uacute$1$1 = "Ú";
  var uacute$1$1 = "ú";
  var Ucirc$1$1 = "Û";
  var ucirc$1$1 = "û";
  var Ugrave$1$1 = "Ù";
  var ugrave$1$1 = "ù";
  var uml$1$1 = "¨";
  var Uuml$1$1 = "Ü";
  var uuml$1$1 = "ü";
  var Yacute$1$1 = "Ý";
  var yacute$1$1 = "ý";
  var yen$1$1 = "¥";
  var yuml$1$1 = "ÿ";
  var legacyJSON$1 = {
  	Aacute: Aacute$1$1,
  	aacute: aacute$1$1,
  	Acirc: Acirc$1$1,
  	acirc: acirc$1$1,
  	acute: acute$1$1,
  	AElig: AElig$1$1,
  	aelig: aelig$1$1,
  	Agrave: Agrave$1$1,
  	agrave: agrave$1$1,
  	amp: amp$2$1,
  	AMP: AMP$1$1,
  	Aring: Aring$1$1,
  	aring: aring$1$1,
  	Atilde: Atilde$1$1,
  	atilde: atilde$1$1,
  	Auml: Auml$1$1,
  	auml: auml$1$1,
  	brvbar: brvbar$1$1,
  	Ccedil: Ccedil$1$1,
  	ccedil: ccedil$1$1,
  	cedil: cedil$1$1,
  	cent: cent$1$1,
  	copy: copy$1$1,
  	COPY: COPY$1$1,
  	curren: curren$1$1,
  	deg: deg$1$1,
  	divide: divide$1$1,
  	Eacute: Eacute$1$1,
  	eacute: eacute$1$1,
  	Ecirc: Ecirc$1$1,
  	ecirc: ecirc$1$1,
  	Egrave: Egrave$1$1,
  	egrave: egrave$1$1,
  	ETH: ETH$1$1,
  	eth: eth$1$1,
  	Euml: Euml$1$1,
  	euml: euml$1$1,
  	frac12: frac12$1$1,
  	frac14: frac14$1$1,
  	frac34: frac34$1$1,
  	gt: gt$2$1,
  	GT: GT$1$1,
  	Iacute: Iacute$1$1,
  	iacute: iacute$1$1,
  	Icirc: Icirc$1$1,
  	icirc: icirc$1$1,
  	iexcl: iexcl$1$1,
  	Igrave: Igrave$1$1,
  	igrave: igrave$1$1,
  	iquest: iquest$1$1,
  	Iuml: Iuml$1$1,
  	iuml: iuml$1$1,
  	laquo: laquo$1$1,
  	lt: lt$2$1,
  	LT: LT$1$1,
  	macr: macr$1$1,
  	micro: micro$1$1,
  	middot: middot$1$1,
  	nbsp: nbsp$1$1,
  	not: not$1$1,
  	Ntilde: Ntilde$1$1,
  	ntilde: ntilde$1$1,
  	Oacute: Oacute$1$1,
  	oacute: oacute$1$1,
  	Ocirc: Ocirc$1$1,
  	ocirc: ocirc$1$1,
  	Ograve: Ograve$1$1,
  	ograve: ograve$1$1,
  	ordf: ordf$1$1,
  	ordm: ordm$1$1,
  	Oslash: Oslash$1$1,
  	oslash: oslash$1$1,
  	Otilde: Otilde$1$1,
  	otilde: otilde$1$1,
  	Ouml: Ouml$1$1,
  	ouml: ouml$1$1,
  	para: para$1$1,
  	plusmn: plusmn$1$1,
  	pound: pound$1$1,
  	quot: quot$2$1,
  	QUOT: QUOT$1$1,
  	raquo: raquo$1$1,
  	reg: reg$1$1,
  	REG: REG$1$1,
  	sect: sect$1$1,
  	shy: shy$1$1,
  	sup1: sup1$1$1,
  	sup2: sup2$1$1,
  	sup3: sup3$1$1,
  	szlig: szlig$1$1,
  	THORN: THORN$1$1,
  	thorn: thorn$1$1,
  	times: times$1$1,
  	Uacute: Uacute$1$1,
  	uacute: uacute$1$1,
  	Ucirc: Ucirc$1$1,
  	ucirc: ucirc$1$1,
  	Ugrave: Ugrave$1$1,
  	ugrave: ugrave$1$1,
  	uml: uml$1$1,
  	Uuml: Uuml$1$1,
  	uuml: uuml$1$1,
  	Yacute: Yacute$1$1,
  	yacute: yacute$1$1,
  	yen: yen$1$1,
  	yuml: yuml$1$1
  };

  var _entities$1 = {
    encodeXML: encode$1.XML,
    decodeCodepoint: decode_codepoint$1,
    entitiesJSON: entitiesJSON$1,
    legacyJSON: legacyJSON$1,
    xmlJSON: xmlJSON$1
  };

  var _entities_decodeCodepoint$2 = _entities$1.decodeCodepoint;

  var _entities_decodeCodepoint$1$1 = Object.freeze({
  	default: _entities_decodeCodepoint$2
  });

  var _entities_entitiesJSON$2 = _entities$1.entitiesJSON;


  var _entities_entitiesJSON$1$1 = Object.freeze({
  	default: _entities_entitiesJSON$2
  });

  var _entities_legacyJSON$2 = _entities$1.legacyJSON;

  var _entities_legacyJSON$1$1 = Object.freeze({
  	default: _entities_legacyJSON$2
  });

  var _entities_xmlJSON$2 = _entities$1.xmlJSON;

  var _entities_xmlJSON$1$1 = Object.freeze({
  	default: _entities_xmlJSON$2
  });

  var decodeCodePoint$1$1 = ( _entities_decodeCodepoint$1$1 && _entities_decodeCodepoint$2 ) || _entities_decodeCodepoint$1$1;

  var entityMap$1 = ( _entities_entitiesJSON$1$1 && _entities_entitiesJSON$2 ) || _entities_entitiesJSON$1$1;

  var legacyMap$1 = ( _entities_legacyJSON$1$1 && _entities_legacyJSON$2 ) || _entities_legacyJSON$1$1;

  var xmlMap$1 = ( _entities_xmlJSON$1$1 && _entities_xmlJSON$2 ) || _entities_xmlJSON$1$1;

  var Tokenizer_1$1 = Tokenizer$1$1;

  var i$1 = 0;
  var TEXT$7                      = i$1++;
  var BEFORE_TAG_NAME$1           = i$1++;
  var IN_TAG_NAME$1               = i$1++;
  var IN_SELF_CLOSING_TAG$1       = i$1++;
  var BEFORE_CLOSING_TAG_NAME$1   = i$1++;
  var IN_CLOSING_TAG_NAME$1       = i$1++;
  var AFTER_CLOSING_TAG_NAME$1    = i$1++;
  var BEFORE_ATTRIBUTE_NAME$1     = i$1++;
  var IN_ATTRIBUTE_NAME$1         = i$1++;
  var AFTER_ATTRIBUTE_NAME$1      = i$1++;
  var BEFORE_ATTRIBUTE_VALUE$1    = i$1++;
  var IN_ATTRIBUTE_VALUE_DQ$1     = i$1++;
  var IN_ATTRIBUTE_VALUE_SQ$1     = i$1++;
  var IN_ATTRIBUTE_VALUE_NQ$1     = i$1++;
  var BEFORE_DECLARATION$1        = i$1++;
  var IN_DECLARATION$1            = i$1++;
  var IN_PROCESSING_INSTRUCTION$1 = i$1++;
  var BEFORE_COMMENT$1            = i$1++;
  var IN_COMMENT$1                = i$1++;
  var AFTER_COMMENT_1$1           = i$1++;
  var AFTER_COMMENT_2$1           = i$1++;
  var BEFORE_CDATA_1$1            = i$1++;
  var BEFORE_CDATA_2$1            = i$1++;
  var BEFORE_CDATA_3$1            = i$1++;
  var BEFORE_CDATA_4$1            = i$1++;
  var BEFORE_CDATA_5$1            = i$1++;
  var BEFORE_CDATA_6$1            = i$1++;
  var IN_CDATA$1                  = i$1++;
  var AFTER_CDATA_1$1             = i$1++;
  var AFTER_CDATA_2$1             = i$1++;
  var BEFORE_SPECIAL$1            = i$1++;
  var BEFORE_SPECIAL_END$1        = i$1++;
  var BEFORE_SCRIPT_1$1           = i$1++;
  var BEFORE_SCRIPT_2$1           = i$1++;
  var BEFORE_SCRIPT_3$1           = i$1++;
  var BEFORE_SCRIPT_4$1           = i$1++;
  var BEFORE_SCRIPT_5$1           = i$1++;
  var AFTER_SCRIPT_1$1            = i$1++;
  var AFTER_SCRIPT_2$1            = i$1++;
  var AFTER_SCRIPT_3$1            = i$1++;
  var AFTER_SCRIPT_4$1            = i$1++;
  var AFTER_SCRIPT_5$1            = i$1++;
  var BEFORE_STYLE_1$1            = i$1++;
  var BEFORE_STYLE_2$1            = i$1++;
  var BEFORE_STYLE_3$1            = i$1++;
  var BEFORE_STYLE_4$1            = i$1++;
  var AFTER_STYLE_1$1             = i$1++;
  var AFTER_STYLE_2$1             = i$1++;
  var AFTER_STYLE_3$1             = i$1++;
  var AFTER_STYLE_4$1             = i$1++;
  var BEFORE_ENTITY$1             = i$1++;
  var BEFORE_NUMERIC_ENTITY$1     = i$1++;
  var IN_NAMED_ENTITY$1           = i$1++;
  var IN_NUMERIC_ENTITY$1         = i$1++;
  var IN_HEX_ENTITY$1             = i$1++;
  var j$1 = 0;
  var SPECIAL_NONE$1              = j$1++;
  var SPECIAL_SCRIPT$1            = j$1++;
  var SPECIAL_STYLE$1             = j$1++;

  function whitespace$1(c){
  	return c === " " || c === "\n" || c === "\t" || c === "\f" || c === "\r";
  }

  function characterState$1(char, SUCCESS){
  	return function(c){
  		if(c === char) this._state = SUCCESS;
  	};
  }

  function ifElseState$1(upper, SUCCESS, FAILURE){
  	var lower = upper.toLowerCase();

  	if(upper === lower){
  		return function(c){
  			if(c === lower){
  				this._state = SUCCESS;
  			} else {
  				this._state = FAILURE;
  				this._index--;
  			}
  		};
  	} else {
  		return function(c){
  			if(c === lower || c === upper){
  				this._state = SUCCESS;
  			} else {
  				this._state = FAILURE;
  				this._index--;
  			}
  		};
  	}
  }

  function consumeSpecialNameChar$1(upper, NEXT_STATE){
  	var lower = upper.toLowerCase();

  	return function(c){
  		if(c === lower || c === upper){
  			this._state = NEXT_STATE;
  		} else {
  			this._state = IN_TAG_NAME$1;
  			this._index--; 
  		}
  	};
  }

  function Tokenizer$1$1(options, cbs){
  	this._state = TEXT$7;
  	this._buffer = "";
  	this._sectionStart = 0;
  	this._index = 0;
  	this._bufferOffset = 0; 
  	this._baseState = TEXT$7;
  	this._special = SPECIAL_NONE$1;
  	this._cbs = cbs;
  	this._running = true;
  	this._ended = false;
  	this._xmlMode = !!(options && options.xmlMode);
  	this._decodeEntities = !!(options && options.decodeEntities);
  }

  Tokenizer$1$1.prototype._stateText = function(c){
  	if(c === "<"){
  		if(this._index > this._sectionStart){
  			this._cbs.ontext(this._getSection());
  		}
  		this._state = BEFORE_TAG_NAME$1;
  		this._sectionStart = this._index;
  	} else if(this._decodeEntities && this._special === SPECIAL_NONE$1 && c === "&"){
  		if(this._index > this._sectionStart){
  			this._cbs.ontext(this._getSection());
  		}
  		this._baseState = TEXT$7;
  		this._state = BEFORE_ENTITY$1;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1$1.prototype._stateBeforeTagName = function(c){
  	if(c === "/"){
  		this._state = BEFORE_CLOSING_TAG_NAME$1;
  	} else if(c === "<"){
  		this._cbs.ontext(this._getSection());
  		this._sectionStart = this._index;
  	} else if(c === ">" || this._special !== SPECIAL_NONE$1 || whitespace$1(c)) {
  		this._state = TEXT$7;
  	} else if(c === "!"){
  		this._state = BEFORE_DECLARATION$1;
  		this._sectionStart = this._index + 1;
  	} else if(c === "?"){
  		this._state = IN_PROCESSING_INSTRUCTION$1;
  		this._sectionStart = this._index + 1;
  	} else {
  		this._state = (!this._xmlMode && (c === "s" || c === "S")) ?
  						BEFORE_SPECIAL$1 : IN_TAG_NAME$1;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1$1.prototype._stateInTagName = function(c){
  	if(c === "/" || c === ">" || whitespace$1(c)){
  		this._emitToken("onopentagname");
  		this._state = BEFORE_ATTRIBUTE_NAME$1;
  		this._index--;
  	}
  };

  Tokenizer$1$1.prototype._stateBeforeCloseingTagName = function(c){
  	if(whitespace$1(c));
  	else if(c === ">"){
  		this._state = TEXT$7;
  	} else if(this._special !== SPECIAL_NONE$1){
  		if(c === "s" || c === "S"){
  			this._state = BEFORE_SPECIAL_END$1;
  		} else {
  			this._state = TEXT$7;
  			this._index--;
  		}
  	} else {
  		this._state = IN_CLOSING_TAG_NAME$1;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1$1.prototype._stateInCloseingTagName = function(c){
  	if(c === ">" || whitespace$1(c)){
  		this._emitToken("onclosetag");
  		this._state = AFTER_CLOSING_TAG_NAME$1;
  		this._index--;
  	}
  };

  Tokenizer$1$1.prototype._stateAfterCloseingTagName = function(c){
  	
  	if(c === ">"){
  		this._state = TEXT$7;
  		this._sectionStart = this._index + 1;
  	}
  };

  Tokenizer$1$1.prototype._stateBeforeAttributeName = function(c){
  	if(c === ">"){
  		this._cbs.onopentagend();
  		this._state = TEXT$7;
  		this._sectionStart = this._index + 1;
  	} else if(c === "/"){
  		this._state = IN_SELF_CLOSING_TAG$1;
  	} else if(!whitespace$1(c)){
  		this._state = IN_ATTRIBUTE_NAME$1;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1$1.prototype._stateInSelfClosingTag = function(c){
  	if(c === ">"){
  		this._cbs.onselfclosingtag();
  		this._state = TEXT$7;
  		this._sectionStart = this._index + 1;
  	} else if(!whitespace$1(c)){
  		this._state = BEFORE_ATTRIBUTE_NAME$1;
  		this._index--;
  	}
  };

  Tokenizer$1$1.prototype._stateInAttributeName = function(c){
  	if(c === "=" || c === "/" || c === ">" || whitespace$1(c)){
  		this._cbs.onattribname(this._getSection());
  		this._sectionStart = -1;
  		this._state = AFTER_ATTRIBUTE_NAME$1;
  		this._index--;
  	}
  };

  Tokenizer$1$1.prototype._stateAfterAttributeName = function(c){
  	if(c === "="){
  		this._state = BEFORE_ATTRIBUTE_VALUE$1;
  	} else if(c === "/" || c === ">"){
  		this._cbs.onattribend();
  		this._state = BEFORE_ATTRIBUTE_NAME$1;
  		this._index--;
  	} else if(!whitespace$1(c)){
  		this._cbs.onattribend();
  		this._state = IN_ATTRIBUTE_NAME$1;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1$1.prototype._stateBeforeAttributeValue = function(c){
  	if(c === "\""){
  		this._state = IN_ATTRIBUTE_VALUE_DQ$1;
  		this._sectionStart = this._index + 1;
  	} else if(c === "'"){
  		this._state = IN_ATTRIBUTE_VALUE_SQ$1;
  		this._sectionStart = this._index + 1;
  	} else if(!whitespace$1(c)){
  		this._state = IN_ATTRIBUTE_VALUE_NQ$1;
  		this._sectionStart = this._index;
  		this._index--; 
  	}
  };

  Tokenizer$1$1.prototype._stateInAttributeValueDoubleQuotes = function(c){
  	if(c === "\""){
  		this._emitToken("onattribdata");
  		this._cbs.onattribend();
  		this._state = BEFORE_ATTRIBUTE_NAME$1;
  	} else if(this._decodeEntities && c === "&"){
  		this._emitToken("onattribdata");
  		this._baseState = this._state;
  		this._state = BEFORE_ENTITY$1;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1$1.prototype._stateInAttributeValueSingleQuotes = function(c){
  	if(c === "'"){
  		this._emitToken("onattribdata");
  		this._cbs.onattribend();
  		this._state = BEFORE_ATTRIBUTE_NAME$1;
  	} else if(this._decodeEntities && c === "&"){
  		this._emitToken("onattribdata");
  		this._baseState = this._state;
  		this._state = BEFORE_ENTITY$1;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1$1.prototype._stateInAttributeValueNoQuotes = function(c){
  	if(whitespace$1(c) || c === ">"){
  		this._emitToken("onattribdata");
  		this._cbs.onattribend();
  		this._state = BEFORE_ATTRIBUTE_NAME$1;
  		this._index--;
  	} else if(this._decodeEntities && c === "&"){
  		this._emitToken("onattribdata");
  		this._baseState = this._state;
  		this._state = BEFORE_ENTITY$1;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1$1.prototype._stateBeforeDeclaration = function(c){
  	this._state = c === "[" ? BEFORE_CDATA_1$1 :
  					c === "-" ? BEFORE_COMMENT$1 :
  						IN_DECLARATION$1;
  };

  Tokenizer$1$1.prototype._stateInDeclaration = function(c){
  	if(c === ">"){
  		this._cbs.ondeclaration(this._getSection());
  		this._state = TEXT$7;
  		this._sectionStart = this._index + 1;
  	}
  };

  Tokenizer$1$1.prototype._stateInProcessingInstruction = function(c){
  	if(c === ">"){
  		this._cbs.onprocessinginstruction(this._getSection());
  		this._state = TEXT$7;
  		this._sectionStart = this._index + 1;
  	}
  };

  Tokenizer$1$1.prototype._stateBeforeComment = function(c){
  	if(c === "-"){
  		this._state = IN_COMMENT$1;
  		this._sectionStart = this._index + 1;
  	} else {
  		this._state = IN_DECLARATION$1;
  	}
  };

  Tokenizer$1$1.prototype._stateInComment = function(c){
  	if(c === "-") this._state = AFTER_COMMENT_1$1;
  };

  Tokenizer$1$1.prototype._stateAfterComment1 = function(c){
  	if(c === "-"){
  		this._state = AFTER_COMMENT_2$1;
  	} else {
  		this._state = IN_COMMENT$1;
  	}
  };

  Tokenizer$1$1.prototype._stateAfterComment2 = function(c){
  	if(c === ">"){
  		
  		this._cbs.oncomment(this._buffer.substring(this._sectionStart, this._index - 2));
  		this._state = TEXT$7;
  		this._sectionStart = this._index + 1;
  	} else if(c !== "-"){
  		this._state = IN_COMMENT$1;
  	}
  	
  };

  Tokenizer$1$1.prototype._stateBeforeCdata1 = ifElseState$1("C", BEFORE_CDATA_2$1, IN_DECLARATION$1);
  Tokenizer$1$1.prototype._stateBeforeCdata2 = ifElseState$1("D", BEFORE_CDATA_3$1, IN_DECLARATION$1);
  Tokenizer$1$1.prototype._stateBeforeCdata3 = ifElseState$1("A", BEFORE_CDATA_4$1, IN_DECLARATION$1);
  Tokenizer$1$1.prototype._stateBeforeCdata4 = ifElseState$1("T", BEFORE_CDATA_5$1, IN_DECLARATION$1);
  Tokenizer$1$1.prototype._stateBeforeCdata5 = ifElseState$1("A", BEFORE_CDATA_6$1, IN_DECLARATION$1);

  Tokenizer$1$1.prototype._stateBeforeCdata6 = function(c){
  	if(c === "["){
  		this._state = IN_CDATA$1;
  		this._sectionStart = this._index + 1;
  	} else {
  		this._state = IN_DECLARATION$1;
  		this._index--;
  	}
  };

  Tokenizer$1$1.prototype._stateInCdata = function(c){
  	if(c === "]") this._state = AFTER_CDATA_1$1;
  };

  Tokenizer$1$1.prototype._stateAfterCdata1 = characterState$1("]", AFTER_CDATA_2$1);

  Tokenizer$1$1.prototype._stateAfterCdata2 = function(c){
  	if(c === ">"){
  		
  		this._cbs.oncdata(this._buffer.substring(this._sectionStart, this._index - 2));
  		this._state = TEXT$7;
  		this._sectionStart = this._index + 1;
  	} else if(c !== "]") {
  		this._state = IN_CDATA$1;
  	}
  	
  };

  Tokenizer$1$1.prototype._stateBeforeSpecial = function(c){
  	if(c === "c" || c === "C"){
  		this._state = BEFORE_SCRIPT_1$1;
  	} else if(c === "t" || c === "T"){
  		this._state = BEFORE_STYLE_1$1;
  	} else {
  		this._state = IN_TAG_NAME$1;
  		this._index--; 
  	}
  };

  Tokenizer$1$1.prototype._stateBeforeSpecialEnd = function(c){
  	if(this._special === SPECIAL_SCRIPT$1 && (c === "c" || c === "C")){
  		this._state = AFTER_SCRIPT_1$1;
  	} else if(this._special === SPECIAL_STYLE$1 && (c === "t" || c === "T")){
  		this._state = AFTER_STYLE_1$1;
  	}
  	else this._state = TEXT$7;
  };

  Tokenizer$1$1.prototype._stateBeforeScript1 = consumeSpecialNameChar$1("R", BEFORE_SCRIPT_2$1);
  Tokenizer$1$1.prototype._stateBeforeScript2 = consumeSpecialNameChar$1("I", BEFORE_SCRIPT_3$1);
  Tokenizer$1$1.prototype._stateBeforeScript3 = consumeSpecialNameChar$1("P", BEFORE_SCRIPT_4$1);
  Tokenizer$1$1.prototype._stateBeforeScript4 = consumeSpecialNameChar$1("T", BEFORE_SCRIPT_5$1);

  Tokenizer$1$1.prototype._stateBeforeScript5 = function(c){
  	if(c === "/" || c === ">" || whitespace$1(c)){
  		this._special = SPECIAL_SCRIPT$1;
  	}
  	this._state = IN_TAG_NAME$1;
  	this._index--; 
  };

  Tokenizer$1$1.prototype._stateAfterScript1 = ifElseState$1("R", AFTER_SCRIPT_2$1, TEXT$7);
  Tokenizer$1$1.prototype._stateAfterScript2 = ifElseState$1("I", AFTER_SCRIPT_3$1, TEXT$7);
  Tokenizer$1$1.prototype._stateAfterScript3 = ifElseState$1("P", AFTER_SCRIPT_4$1, TEXT$7);
  Tokenizer$1$1.prototype._stateAfterScript4 = ifElseState$1("T", AFTER_SCRIPT_5$1, TEXT$7);

  Tokenizer$1$1.prototype._stateAfterScript5 = function(c){
  	if(c === ">" || whitespace$1(c)){
  		this._special = SPECIAL_NONE$1;
  		this._state = IN_CLOSING_TAG_NAME$1;
  		this._sectionStart = this._index - 6;
  		this._index--; 
  	}
  	else this._state = TEXT$7;
  };

  Tokenizer$1$1.prototype._stateBeforeStyle1 = consumeSpecialNameChar$1("Y", BEFORE_STYLE_2$1);
  Tokenizer$1$1.prototype._stateBeforeStyle2 = consumeSpecialNameChar$1("L", BEFORE_STYLE_3$1);
  Tokenizer$1$1.prototype._stateBeforeStyle3 = consumeSpecialNameChar$1("E", BEFORE_STYLE_4$1);

  Tokenizer$1$1.prototype._stateBeforeStyle4 = function(c){
  	if(c === "/" || c === ">" || whitespace$1(c)){
  		this._special = SPECIAL_STYLE$1;
  	}
  	this._state = IN_TAG_NAME$1;
  	this._index--; 
  };

  Tokenizer$1$1.prototype._stateAfterStyle1 = ifElseState$1("Y", AFTER_STYLE_2$1, TEXT$7);
  Tokenizer$1$1.prototype._stateAfterStyle2 = ifElseState$1("L", AFTER_STYLE_3$1, TEXT$7);
  Tokenizer$1$1.prototype._stateAfterStyle3 = ifElseState$1("E", AFTER_STYLE_4$1, TEXT$7);

  Tokenizer$1$1.prototype._stateAfterStyle4 = function(c){
  	if(c === ">" || whitespace$1(c)){
  		this._special = SPECIAL_NONE$1;
  		this._state = IN_CLOSING_TAG_NAME$1;
  		this._sectionStart = this._index - 5;
  		this._index--; 
  	}
  	else this._state = TEXT$7;
  };

  Tokenizer$1$1.prototype._stateBeforeEntity = ifElseState$1("#", BEFORE_NUMERIC_ENTITY$1, IN_NAMED_ENTITY$1);
  Tokenizer$1$1.prototype._stateBeforeNumericEntity = ifElseState$1("X", IN_HEX_ENTITY$1, IN_NUMERIC_ENTITY$1);


  Tokenizer$1$1.prototype._parseNamedEntityStrict = function(){
  	
  	if(this._sectionStart + 1 < this._index){
  		var entity = this._buffer.substring(this._sectionStart + 1, this._index),
  		    map = this._xmlMode ? xmlMap$1 : entityMap$1;

  		if(map.hasOwnProperty(entity)){
  			this._emitPartial(map[entity]);
  			this._sectionStart = this._index + 1;
  		}
  	}
  };



  Tokenizer$1$1.prototype._parseLegacyEntity = function(){
  	var start = this._sectionStart + 1,
  	    limit = this._index - start;

  	if(limit > 6) limit = 6; 

  	while(limit >= 2){ 
  		var entity = this._buffer.substr(start, limit);

  		if(legacyMap$1.hasOwnProperty(entity)){
  			this._emitPartial(legacyMap$1[entity]);
  			this._sectionStart += limit + 1;
  			return;
  		} else {
  			limit--;
  		}
  	}
  };

  Tokenizer$1$1.prototype._stateInNamedEntity = function(c){
  	if(c === ";"){
  		this._parseNamedEntityStrict();
  		if(this._sectionStart + 1 < this._index && !this._xmlMode){
  			this._parseLegacyEntity();
  		}
  		this._state = this._baseState;
  	} else if((c < "a" || c > "z") && (c < "A" || c > "Z") && (c < "0" || c > "9")){
  		if(this._xmlMode);
  		else if(this._sectionStart + 1 === this._index);
  		else if(this._baseState !== TEXT$7){
  			if(c !== "="){
  				this._parseNamedEntityStrict();
  			}
  		} else {
  			this._parseLegacyEntity();
  		}

  		this._state = this._baseState;
  		this._index--;
  	}
  };

  Tokenizer$1$1.prototype._decodeNumericEntity = function(offset, base){
  	var sectionStart = this._sectionStart + offset;

  	if(sectionStart !== this._index){
  		
  		var entity = this._buffer.substring(sectionStart, this._index);
  		var parsed = parseInt(entity, base);

  		this._emitPartial(decodeCodePoint$1$1(parsed));
  		this._sectionStart = this._index;
  	} else {
  		this._sectionStart--;
  	}

  	this._state = this._baseState;
  };

  Tokenizer$1$1.prototype._stateInNumericEntity = function(c){
  	if(c === ";"){
  		this._decodeNumericEntity(2, 10);
  		this._sectionStart++;
  	} else if(c < "0" || c > "9"){
  		if(!this._xmlMode){
  			this._decodeNumericEntity(2, 10);
  		} else {
  			this._state = this._baseState;
  		}
  		this._index--;
  	}
  };

  Tokenizer$1$1.prototype._stateInHexEntity = function(c){
  	if(c === ";"){
  		this._decodeNumericEntity(3, 16);
  		this._sectionStart++;
  	} else if((c < "a" || c > "f") && (c < "A" || c > "F") && (c < "0" || c > "9")){
  		if(!this._xmlMode){
  			this._decodeNumericEntity(3, 16);
  		} else {
  			this._state = this._baseState;
  		}
  		this._index--;
  	}
  };

  Tokenizer$1$1.prototype._cleanup = function (){
  	if(this._sectionStart < 0){
  		this._buffer = "";
  		this._bufferOffset += this._index;
  		this._index = 0;
  	} else if(this._running){
  		if(this._state === TEXT$7){
  			if(this._sectionStart !== this._index){
  				this._cbs.ontext(this._buffer.substr(this._sectionStart));
  			}
  			this._buffer = "";
  			this._bufferOffset += this._index;
  			this._index = 0;
  		} else if(this._sectionStart === this._index){
  			
  			this._buffer = "";
  			this._bufferOffset += this._index;
  			this._index = 0;
  		} else {
  			
  			this._buffer = this._buffer.substr(this._sectionStart);
  			this._index -= this._sectionStart;
  			this._bufferOffset += this._sectionStart;
  		}

  		this._sectionStart = 0;
  	}
  };


  Tokenizer$1$1.prototype.write = function(chunk){
  	if(this._ended) this._cbs.onerror(Error(".write() after done!"));

  	this._buffer += chunk;
  	this._parse();
  };

  Tokenizer$1$1.prototype._parse = function(){
  	while(this._index < this._buffer.length && this._running){
  		var c = this._buffer.charAt(this._index);
  		if(this._state === TEXT$7) {
  			this._stateText(c);
  		} else if(this._state === BEFORE_TAG_NAME$1){
  			this._stateBeforeTagName(c);
  		} else if(this._state === IN_TAG_NAME$1) {
  			this._stateInTagName(c);
  		} else if(this._state === BEFORE_CLOSING_TAG_NAME$1){
  			this._stateBeforeCloseingTagName(c);
  		} else if(this._state === IN_CLOSING_TAG_NAME$1){
  			this._stateInCloseingTagName(c);
  		} else if(this._state === AFTER_CLOSING_TAG_NAME$1){
  			this._stateAfterCloseingTagName(c);
  		} else if(this._state === IN_SELF_CLOSING_TAG$1){
  			this._stateInSelfClosingTag(c);
  		}

  		
  		else if(this._state === BEFORE_ATTRIBUTE_NAME$1){
  			this._stateBeforeAttributeName(c);
  		} else if(this._state === IN_ATTRIBUTE_NAME$1){
  			this._stateInAttributeName(c);
  		} else if(this._state === AFTER_ATTRIBUTE_NAME$1){
  			this._stateAfterAttributeName(c);
  		} else if(this._state === BEFORE_ATTRIBUTE_VALUE$1){
  			this._stateBeforeAttributeValue(c);
  		} else if(this._state === IN_ATTRIBUTE_VALUE_DQ$1){
  			this._stateInAttributeValueDoubleQuotes(c);
  		} else if(this._state === IN_ATTRIBUTE_VALUE_SQ$1){
  			this._stateInAttributeValueSingleQuotes(c);
  		} else if(this._state === IN_ATTRIBUTE_VALUE_NQ$1){
  			this._stateInAttributeValueNoQuotes(c);
  		}

  		
  		else if(this._state === BEFORE_DECLARATION$1){
  			this._stateBeforeDeclaration(c);
  		} else if(this._state === IN_DECLARATION$1){
  			this._stateInDeclaration(c);
  		}

  		
  		else if(this._state === IN_PROCESSING_INSTRUCTION$1){
  			this._stateInProcessingInstruction(c);
  		}

  		
  		else if(this._state === BEFORE_COMMENT$1){
  			this._stateBeforeComment(c);
  		} else if(this._state === IN_COMMENT$1){
  			this._stateInComment(c);
  		} else if(this._state === AFTER_COMMENT_1$1){
  			this._stateAfterComment1(c);
  		} else if(this._state === AFTER_COMMENT_2$1){
  			this._stateAfterComment2(c);
  		}

  		
  		else if(this._state === BEFORE_CDATA_1$1){
  			this._stateBeforeCdata1(c);
  		} else if(this._state === BEFORE_CDATA_2$1){
  			this._stateBeforeCdata2(c);
  		} else if(this._state === BEFORE_CDATA_3$1){
  			this._stateBeforeCdata3(c);
  		} else if(this._state === BEFORE_CDATA_4$1){
  			this._stateBeforeCdata4(c);
  		} else if(this._state === BEFORE_CDATA_5$1){
  			this._stateBeforeCdata5(c);
  		} else if(this._state === BEFORE_CDATA_6$1){
  			this._stateBeforeCdata6(c);
  		} else if(this._state === IN_CDATA$1){
  			this._stateInCdata(c);
  		} else if(this._state === AFTER_CDATA_1$1){
  			this._stateAfterCdata1(c);
  		} else if(this._state === AFTER_CDATA_2$1){
  			this._stateAfterCdata2(c);
  		}

  		
  		else if(this._state === BEFORE_SPECIAL$1){
  			this._stateBeforeSpecial(c);
  		} else if(this._state === BEFORE_SPECIAL_END$1){
  			this._stateBeforeSpecialEnd(c);
  		}

  		
  		else if(this._state === BEFORE_SCRIPT_1$1){
  			this._stateBeforeScript1(c);
  		} else if(this._state === BEFORE_SCRIPT_2$1){
  			this._stateBeforeScript2(c);
  		} else if(this._state === BEFORE_SCRIPT_3$1){
  			this._stateBeforeScript3(c);
  		} else if(this._state === BEFORE_SCRIPT_4$1){
  			this._stateBeforeScript4(c);
  		} else if(this._state === BEFORE_SCRIPT_5$1){
  			this._stateBeforeScript5(c);
  		}

  		else if(this._state === AFTER_SCRIPT_1$1){
  			this._stateAfterScript1(c);
  		} else if(this._state === AFTER_SCRIPT_2$1){
  			this._stateAfterScript2(c);
  		} else if(this._state === AFTER_SCRIPT_3$1){
  			this._stateAfterScript3(c);
  		} else if(this._state === AFTER_SCRIPT_4$1){
  			this._stateAfterScript4(c);
  		} else if(this._state === AFTER_SCRIPT_5$1){
  			this._stateAfterScript5(c);
  		}

  		
  		else if(this._state === BEFORE_STYLE_1$1){
  			this._stateBeforeStyle1(c);
  		} else if(this._state === BEFORE_STYLE_2$1){
  			this._stateBeforeStyle2(c);
  		} else if(this._state === BEFORE_STYLE_3$1){
  			this._stateBeforeStyle3(c);
  		} else if(this._state === BEFORE_STYLE_4$1){
  			this._stateBeforeStyle4(c);
  		}

  		else if(this._state === AFTER_STYLE_1$1){
  			this._stateAfterStyle1(c);
  		} else if(this._state === AFTER_STYLE_2$1){
  			this._stateAfterStyle2(c);
  		} else if(this._state === AFTER_STYLE_3$1){
  			this._stateAfterStyle3(c);
  		} else if(this._state === AFTER_STYLE_4$1){
  			this._stateAfterStyle4(c);
  		}

  		
  		else if(this._state === BEFORE_ENTITY$1){
  			this._stateBeforeEntity(c);
  		} else if(this._state === BEFORE_NUMERIC_ENTITY$1){
  			this._stateBeforeNumericEntity(c);
  		} else if(this._state === IN_NAMED_ENTITY$1){
  			this._stateInNamedEntity(c);
  		} else if(this._state === IN_NUMERIC_ENTITY$1){
  			this._stateInNumericEntity(c);
  		} else if(this._state === IN_HEX_ENTITY$1){
  			this._stateInHexEntity(c);
  		}

  		else {
  			this._cbs.onerror(Error("unknown _state"), this._state);
  		}

  		this._index++;
  	}

  	this._cleanup();
  };

  Tokenizer$1$1.prototype.pause = function(){
  	this._running = false;
  };
  Tokenizer$1$1.prototype.resume = function(){
  	this._running = true;

  	if(this._index < this._buffer.length){
  		this._parse();
  	}
  	if(this._ended){
  		this._finish();
  	}
  };

  Tokenizer$1$1.prototype.end = function(chunk){
  	if(this._ended) this._cbs.onerror(Error(".end() after done!"));
  	if(chunk) this.write(chunk);

  	this._ended = true;

  	if(this._running) this._finish();
  };

  Tokenizer$1$1.prototype._finish = function(){
  	
  	if(this._sectionStart < this._index){
  		this._handleTrailingData();
  	}

  	this._cbs.onend();
  };

  Tokenizer$1$1.prototype._handleTrailingData = function(){
  	var data = this._buffer.substr(this._sectionStart);

  	if(this._state === IN_CDATA$1 || this._state === AFTER_CDATA_1$1 || this._state === AFTER_CDATA_2$1){
  		this._cbs.oncdata(data);
  	} else if(this._state === IN_COMMENT$1 || this._state === AFTER_COMMENT_1$1 || this._state === AFTER_COMMENT_2$1){
  		this._cbs.oncomment(data);
  	} else if(this._state === IN_NAMED_ENTITY$1 && !this._xmlMode){
  		this._parseLegacyEntity();
  		if(this._sectionStart < this._index){
  			this._state = this._baseState;
  			this._handleTrailingData();
  		}
  	} else if(this._state === IN_NUMERIC_ENTITY$1 && !this._xmlMode){
  		this._decodeNumericEntity(2, 10);
  		if(this._sectionStart < this._index){
  			this._state = this._baseState;
  			this._handleTrailingData();
  		}
  	} else if(this._state === IN_HEX_ENTITY$1 && !this._xmlMode){
  		this._decodeNumericEntity(3, 16);
  		if(this._sectionStart < this._index){
  			this._state = this._baseState;
  			this._handleTrailingData();
  		}
  	} else if(
  		this._state !== IN_TAG_NAME$1 &&
  		this._state !== BEFORE_ATTRIBUTE_NAME$1 &&
  		this._state !== BEFORE_ATTRIBUTE_VALUE$1 &&
  		this._state !== AFTER_ATTRIBUTE_NAME$1 &&
  		this._state !== IN_ATTRIBUTE_NAME$1 &&
  		this._state !== IN_ATTRIBUTE_VALUE_SQ$1 &&
  		this._state !== IN_ATTRIBUTE_VALUE_DQ$1 &&
  		this._state !== IN_ATTRIBUTE_VALUE_NQ$1 &&
  		this._state !== IN_CLOSING_TAG_NAME$1
  	){
  		this._cbs.ontext(data);
  	}
  	
  	
  };

  Tokenizer$1$1.prototype.reset = function(){
  	Tokenizer$1$1.call(this, {xmlMode: this._xmlMode, decodeEntities: this._decodeEntities}, this._cbs);
  };

  Tokenizer$1$1.prototype.getAbsoluteIndex = function(){
  	return this._bufferOffset + this._index;
  };

  Tokenizer$1$1.prototype._getSection = function(){
  	return this._buffer.substring(this._sectionStart, this._index);
  };

  Tokenizer$1$1.prototype._emitToken = function(name){
  	this._cbs[name](this._getSection());
  	this._sectionStart = -1;
  };

  Tokenizer$1$1.prototype._emitPartial = function(value){
  	if(this._baseState !== TEXT$7){
  		this._cbs.onattribdata(value); 
  	} else {
  		this._cbs.ontext(value);
  	}
  };

  var _inherits$2 = function() {};

  var _inherits$1$1 = Object.freeze({
  	default: _inherits$2
  });

  var _resolve_empty$2 = {};

  var _resolve_empty$1$1 = Object.freeze({
  	default: _resolve_empty$2
  });

  var require$$1$1$1 = ( _inherits$1$1 && _inherits$2 ) || _inherits$1$1;

  var require$$2$1 = ( _resolve_empty$1$1 && _resolve_empty$2 ) || _resolve_empty$1$1;

  var Tokenizer$2$1 = Tokenizer_1$1;





  var formTags$1 = {
  	input: true,
  	option: true,
  	optgroup: true,
  	select: true,
  	button: true,
  	datalist: true,
  	textarea: true
  };

  var openImpliesClose$1 = {
  	tr      : { tr:true, th:true, td:true },
  	th      : { th:true },
  	td      : { thead:true, th:true, td:true },
  	body    : { head:true, link:true, script:true },
  	li      : { li:true },
  	p       : { p:true },
  	h1      : { p:true },
  	h2      : { p:true },
  	h3      : { p:true },
  	h4      : { p:true },
  	h5      : { p:true },
  	h6      : { p:true },
  	select  : formTags$1,
  	input   : formTags$1,
  	output  : formTags$1,
  	button  : formTags$1,
  	datalist: formTags$1,
  	textarea: formTags$1,
  	option  : { option:true },
  	optgroup: { optgroup:true }
  };

  var voidElements$1 = {
  	__proto__: null,
  	area: true,
  	base: true,
  	basefont: true,
  	br: true,
  	col: true,
  	command: true,
  	embed: true,
  	frame: true,
  	hr: true,
  	img: true,
  	input: true,
  	isindex: true,
  	keygen: true,
  	link: true,
  	meta: true,
  	param: true,
  	source: true,
  	track: true,
  	wbr: true,

  	
  	path: true,
  	circle: true,
  	ellipse: true,
  	line: true,
  	rect: true,
  	use: true,
  	stop: true,
  	polyline: true,
  	polygon: true
  };

  var re_nameEnd$1 = /\s|\//;

  function Parser$1$1(cbs, options){
  	this._options = options || {};
  	this._cbs = cbs || {};

  	this._tagname = "";
  	this._attribname = "";
  	this._attribvalue = "";
  	this._attribs = null;
  	this._stack = [];

  	this.startIndex = 0;
  	this.endIndex = null;

  	this._lowerCaseTagNames = "lowerCaseTags" in this._options ?
  									!!this._options.lowerCaseTags :
  									!this._options.xmlMode;
  	this._lowerCaseAttributeNames = "lowerCaseAttributeNames" in this._options ?
  									!!this._options.lowerCaseAttributeNames :
  									!this._options.xmlMode;

  	if(this._options.Tokenizer) {
  		Tokenizer$2$1 = this._options.Tokenizer;
  	}
  	this._tokenizer = new Tokenizer$2$1(this._options, this);

  	if(this._cbs.onparserinit) this._cbs.onparserinit(this);
  }

  require$$1$1$1(Parser$1$1, require$$2$1.EventEmitter);

  Parser$1$1.prototype._updatePosition = function(initialOffset){
  	if(this.endIndex === null){
  		if(this._tokenizer._sectionStart <= initialOffset){
  			this.startIndex = 0;
  		} else {
  			this.startIndex = this._tokenizer._sectionStart - initialOffset;
  		}
  	}
  	else this.startIndex = this.endIndex + 1;
  	this.endIndex = this._tokenizer.getAbsoluteIndex();
  };


  Parser$1$1.prototype.ontext = function(data){
  	this._updatePosition(1);
  	this.endIndex--;

  	if(this._cbs.ontext) this._cbs.ontext(data);
  };

  Parser$1$1.prototype.onopentagname = function(name){
  	if(this._lowerCaseTagNames){
  		name = name.toLowerCase();
  	}

  	this._tagname = name;

  	if(!this._options.xmlMode && name in openImpliesClose$1) {
  		for(
  			var el;
  			(el = this._stack[this._stack.length - 1]) in openImpliesClose$1[name];
  			this.onclosetag(el)
  		);
  	}

  	if(this._options.xmlMode || !(name in voidElements$1)){
  		this._stack.push(name);
  	}

  	if(this._cbs.onopentagname) this._cbs.onopentagname(name);
  	if(this._cbs.onopentag) this._attribs = {};
  };

  Parser$1$1.prototype.onopentagend = function(){
  	this._updatePosition(1);

  	if(this._attribs){
  		if(this._cbs.onopentag) this._cbs.onopentag(this._tagname, this._attribs);
  		this._attribs = null;
  	}

  	if(!this._options.xmlMode && this._cbs.onclosetag && this._tagname in voidElements$1){
  		this._cbs.onclosetag(this._tagname);
  	}

  	this._tagname = "";
  };

  Parser$1$1.prototype.onclosetag = function(name){
  	this._updatePosition(1);
  	if(this._lowerCaseTagNames){
  		name = name.toLowerCase();
  	}
  	
  	
  	
  	if(this._options.xmlMode) {
  		const stack = this._stack;
  		let last = stack.pop();
  		while(last !== name) {
  			if(this._cbs.onerror) {
  				this._cbs.onerror("Unclosed tag <"+last+">");
  			}
  			last = stack.pop();
  		}
  		this.onopentagend();
  		if(this._cbs.onclosetag) {
  			this._cbs.onclosetag(last);
  		}
  	} else {
  		if(this._stack.length && (!(name in voidElements$1))) {
  			let pos = this._stack.lastIndexOf(name);
  			if(pos !== -1){
  				if(this._cbs.onclosetag){
  					pos = this._stack.length - pos;
  					while(pos--) this._cbs.onclosetag(this._stack.pop());
  				}
  				else this._stack.length = pos;
  			} else if(name === "p"){
  				this.onopentagname(name);
  				this._closeCurrentTag();
  			}
  		} else if(name === "br" || name === "p"){
  			this.onopentagname(name);
  			this._closeCurrentTag();
  		}
  	}
  };

  Parser$1$1.prototype.onselfclosingtag = function(){
  	if(this._options.xmlMode || this._options.recognizeSelfClosing){
  		this._closeCurrentTag();
  	} else {
  		this.onopentagend();
  	}
  };

  Parser$1$1.prototype._closeCurrentTag = function(){
  	var name = this._tagname;

  	this.onopentagend();

  	
  	
  	if(this._stack[this._stack.length - 1] === name){
  		if(this._cbs.onclosetag){
  			this._cbs.onclosetag(name);
  		}
  		this._stack.pop();
  	}
  };

  Parser$1$1.prototype.onattribname = function(name){
  	if(this._lowerCaseAttributeNames){
  		name = name.toLowerCase();
  	}
  	this._attribname = name;
  };

  Parser$1$1.prototype.onattribdata = function(value){
  	this._attribvalue += value;
  };

  Parser$1$1.prototype.onattribend = function(){
  	if(this._cbs.onattribute) this._cbs.onattribute(this._attribname, this._attribvalue);
  	if(
  		this._attribs &&
  		!Object.prototype.hasOwnProperty.call(this._attribs, this._attribname)
  	){
  		this._attribs[this._attribname] = this._attribvalue;
  	}
  	this._attribname = "";
  	this._attribvalue = "";
  };

  Parser$1$1.prototype._getInstructionName = function(value){
  	var idx = value.search(re_nameEnd$1),
  	    name = idx < 0 ? value : value.substr(0, idx);

  	if(this._lowerCaseTagNames){
  		name = name.toLowerCase();
  	}

  	return name;
  };

  Parser$1$1.prototype.ondeclaration = function(value){
  	if(this._cbs.onprocessinginstruction){
  		var name = this._getInstructionName(value);
  		this._cbs.onprocessinginstruction("!" + name, "!" + value);
  	}
  };

  Parser$1$1.prototype.onprocessinginstruction = function(value){
  	if(this._cbs.onprocessinginstruction){
  		var name = this._getInstructionName(value);
  		this._cbs.onprocessinginstruction("?" + name, "?" + value);
  	}
  };

  Parser$1$1.prototype.oncomment = function(value){
  	this._updatePosition(4);

  	if(this._cbs.oncomment) this._cbs.oncomment(value);
  	if(this._cbs.oncommentend) this._cbs.oncommentend();
  };

  Parser$1$1.prototype.oncdata = function(value){
  	this._updatePosition(1);

  	if(this._options.xmlMode || this._options.recognizeCDATA){
  		if(this._cbs.oncdatastart) this._cbs.oncdatastart();
  		if(this._cbs.ontext) this._cbs.ontext(value);
  		if(this._cbs.oncdataend) this._cbs.oncdataend();
  	} else {
  		this.oncomment("[CDATA[" + value + "]]");
  	}
  };

  Parser$1$1.prototype.onerror = function(err){
  	if(this._cbs.onerror) this._cbs.onerror(err);
  };

  Parser$1$1.prototype.onend = function(){
  	if(this._cbs.onclosetag){
  		for(
  			var i = this._stack.length;
  			i > 0;
  			this._cbs.onclosetag(this._stack[--i])
  		);
  	}
  	if(this._cbs.onend) this._cbs.onend();
  };



  Parser$1$1.prototype.reset = function(){
  	if(this._cbs.onreset) this._cbs.onreset();
  	this._tokenizer.reset();

  	this._tagname = "";
  	this._attribname = "";
  	this._attribs = null;
  	this._stack = [];

  	if(this._cbs.onparserinit) this._cbs.onparserinit(this);
  };


  Parser$1$1.prototype.parseComplete = function(data){
  	this.reset();
  	this.end(data);
  };

  Parser$1$1.prototype.write = function(chunk){
  	this._tokenizer.write(chunk);
  };

  Parser$1$1.prototype.end = function(chunk){
  	this._tokenizer.end(chunk);
  };

  Parser$1$1.prototype.pause = function(){
  	this._tokenizer.pause();
  };

  Parser$1$1.prototype.resume = function(){
  	this._tokenizer.resume();
  };


  Parser$1$1.prototype.parseChunk = Parser$1$1.prototype.write;
  Parser$1$1.prototype.done = Parser$1$1.prototype.end;

  var Parser_1$1 = Parser$1$1;

  Parser_1$1.prototype.ondeclaration = function(value){
    if(this._cbs.ondeclaration){
      this._cbs.ondeclaration(value);
    } else if(this._cbs.onprocessinginstruction){
      var name = this._getInstructionName(value);
      this._cbs.onprocessinginstruction("!" + name, "!" + value);
    }
  };


  Parser_1$1.prototype.oncdata = function(value){
    this._updatePosition(1);

    if(this._options.xmlMode || this._options.recognizeCDATA){
      if(this._cbs.oncdatastart) this._cbs.oncdatastart(value);
      
      if(this._cbs.oncdataend) this._cbs.oncdataend();
    } else {
      this.oncomment("[CDATA[" + value + "]]");
    }
  };

  function parseMarkup$1 (markup, options) {
    let format = options.ownerDocument ? options.ownerDocument.format : options.format;
    
    if (!format) {
      throw new Error("Either 'ownerDocument' or 'format' must be set.")
    }
    let parserOptions = Object.assign({}, options, {
      xmlMode: (format === 'xml')
    });
    let handler = new DomHandler$1({ format, elementFactory: options.elementFactory });
    let parser = new Parser_1$1(handler, parserOptions);
    parser.end(markup);
    return handler.document
  }

  const RE_WHITESPACE$1 = /\s+/g;
  const RE_DOCTYPE$1 = /^DOCTYPE\s+([^\s]+)(?:\s+PUBLIC\s+["]([^"]+)["](?:\s+["]([^"]+)["])?)\s*$/;


  class DomHandler$1 {
    constructor (options = {}) {
      this.elementFactory = options.elementFactory;
      if (!this.elementFactory) throw new Error("'elementFactory' is required")
      this.options = options;
      this.document = null;
      this._tagStack = [];
    }

    
    onparserinit () {
      this.document = this.elementFactory('document', { format: this.options.format });
      this._tagStack = [this.document];
    }

    onend () {
      
      if (this._tagStack.length > 1) {
        throw new Error(`Unexpected EOF. Tag was opened but not closed.`)
      }
    }

    onerror (error) {
      throw new Error(error)
    }

    onclosetag () {
      this._tagStack.pop();
    }

    _addDomElement (element) {
      let parent = this._tagStack[this._tagStack.length - 1];
      if (!parent.childNodes) parent.childNodes = [];
      let siblings = parent.childNodes;

      let previousSibling = siblings[siblings.length - 1];
      
      element.next = null;
      if (previousSibling) {
        element.prev = previousSibling;
        previousSibling.next = element;
      } else {
        element.prev = null;
      }
      
      siblings.push(element);
      element.parent = parent || null;
    }

    onopentag (name, attributes) {
      let element = this.document.createElement(name);
      forEach$1(attributes, (val, key) => {
        element.setAttribute(key, val);
      });
      this._addDomElement(element);
      this._tagStack.push(element);
    }

    ontext (text) {
      if (this.options.normalizeWhitespace) {
        text = text.replace(RE_WHITESPACE$1, ' ');
      }
      let lastTag;
      let _top = this._tagStack[this._tagStack.length - 1];
      if (_top && _top.childNodes) lastTag = _top.childNodes[_top.childNodes.length - 1];
      if (lastTag && lastTag.type === domelementtype$1.Text) {
        lastTag.data += text;
      } else {
        let element = this.document.createTextNode(text);
        this._addDomElement(element);
      }
    }

    oncomment (data) {
      var lastTag = this._tagStack[this._tagStack.length - 1];
      if (lastTag && lastTag.type === domelementtype$1.Comment) {
        lastTag.data += data;
      } else {
        let element = this.document.createComment(data);
        this._addDomElement(element);
        this._tagStack.push(element);
      }
    }

    oncommentend () {
      this._tagStack.pop();
    }

    oncdatastart (data) {
      let element = this.document.createCDATASection(data);
      this._addDomElement(element);
      this._tagStack.push(element);
    }

    oncdataend () {
      this._tagStack.pop();
    }

    onprocessinginstruction (name, data) {
      let element = this.document.createProcessingInstruction(name, data);
      this._addDomElement(element);
    }

    ondeclaration (data) {
      if (data.startsWith('DOCTYPE')) {
        let m = RE_DOCTYPE$1.exec(data);
        if (!m) throw new Error('Could not parse DOCTYPE element: ' + data)
        this.document.setDoctype(m[1], m[2], m[3]);
      } else {
        throw new Error('Not implemented: parse declaration ' + data)
      }
    }
  }

  const MemoryDOMElementFactory$1 = (type, data) => {
    return new MemoryDOMElement$1(type, data)
  };

  class MemoryDOMElement$1 extends DOMElement$1 {
    constructor (type, args = {}) {
      super();

      this.type = type;
      if (!type) throw new Error("'type' is mandatory")

      this.ownerDocument = args.ownerDocument;
      
      if (type !== 'document' && !this.ownerDocument) {
        throw new Error("'ownerDocument' is mandatory")
      }

      
      
      

      switch (type) {
        case domelementtype$1.Tag: {
          if (!args.name) throw new Error("'name' is mandatory.")
          this.name = this._normalizeName(args.name);
          this.nameWithoutNS = nameWithoutNS$2(this.name);
          this.properties = new Map();
          this.attributes = new Map();
          this.classes = new Set();
          this.styles = new Map();
          this.eventListeners = [];
          this.childNodes = args.children || args.childNodes || [];
          this._assign(args);
          break
        }
        case domelementtype$1.Text:
        case domelementtype$1.Comment: {
          this.data = args.data || '';
          break
        }
        case domelementtype$1.CDATA: {
          this.data = args.data || '';
          break
        }
        case domelementtype$1.Directive: {
          if (!args.name) throw new Error("'name' is mandatory.")
          this.name = this._normalizeName(args.name);
          this.nameWithoutNS = nameWithoutNS$2(this.name);
          this.data = args.data;
          break
        }
        case domelementtype$1.Doctype: {
          this.data = args.data;
          break
        }
        case 'document': {
          let format = args.format;
          this.format = format;
          if (!format) throw new Error("'format' is mandatory.")
          this.childNodes = args.children || args.childNodes || [];
          switch (format) {
            case 'xml':
              this.contentType = 'application/xml';
              break
            case 'html':
              this.contentType = 'text/html';
              break
            default:
              throw new Error('Unsupported format ' + format)
          }
          break
        }
        default:
          this.name = null;
          this.properties = new Map();
          this.attributes = new Map();
          this.classes = new Set();
          this.styles = new Map();
          this.eventListeners = [];
          this.childNodes = args.children || args.childNodes || [];
      }
    }

    getNativeElement () {
      return this
    }

    getNodeType () {
      switch (this.type) {
        case domelementtype$1.Tag:
        case domelementtype$1.Script:
        case domelementtype$1.Style:
          return 'element'
        default:
          return this.type
      }
    }

    isTextNode () {
      return this.type === 'text'
    }

    isElementNode () {
      return this.type === 'tag' || this.type === 'script'
    }

    isCommentNode () {
      return this.type === 'comment'
    }

    isDocumentNode () {
      return this.type === 'document'
    }

    isComponentNode () {
      return this.type === 'component'
    }

    clone (deep) {
      let clone = new MemoryDOMElement$1(this.type, this);
      if (this.childNodes) {
        clone.childNodes.length = 0;
        if (deep) {
          this.childNodes.forEach((child) => {
            clone.appendChild(child.clone(deep));
          });
        }
      }
      return clone
    }

    get tagName () {
      return this.getTagName()
    }

    set tagName (tagName) {
      this.setTagName(tagName);
    }

    getTagName () {
      return this.name
    }

    setTagName (tagName) {
      if (this._isXML()) {
        this.name = String(tagName);
      } else {
        this.name = String(tagName).toLowerCase();
      }
      this.nameWithoutNS = nameWithoutNS$2(this.name);
      return this
    }

    hasAttribute (name) {
      return this.attributes.has(name)
    }

    getAttribute (name) {
      return this.attributes.get(name)
    }

    setAttribute (name, value) {
      value = String(value);
      
      switch (name) {
        case 'class':
          this.classes = new Set();
          parseClasses$1(this.classes, value);
          break
        case 'style':
          this.styles = new Map();
          parseStyles$1(this.styles, value);
          break
        default:
          
      }
      this.attributes.set(name, value);
      if (this._isHTML()) {
        deriveHTMLPropertyFromAttribute$1(this, name, value);
      }
      return this
    }

    removeAttribute (name) {
      switch (name) {
        case 'class':
          this.classes = new Set();
          break
        case 'style':
          this.styles = new Map();
          break
        default:
          
      }
      this.attributes.delete(name);
      return this
    }

    getAttributes () {
      return this.attributes
    }

    getProperty (name) {
      if (this.properties) {
        return this.properties.get(name)
      }
    }

    setProperty (name, value) {
      if (this.properties) {
        if (this._isXML()) {
          throw new Error('setProperty() is only be used on HTML elements')
        }
        _setHTMLPropertyValue$1(this, name, value);
      }
      return this
    }

    hasClass (name) {
      if (this.classes) {
        return this.classes.has(name)
      }
    }

    addClass (name) {
      this.classes.add(name);
      this.attributes.set('class', stringifyClasses$1(this.classes));
      return this
    }

    removeClass (name) {
      if (this.classes && this.classes.has(name)) {
        this.classes.delete(name);
        this.attributes.set('class', stringifyClasses$1(this.classes));
      }
      return this
    }

    getContentType () {
      return this.getOwnerDocument().contentType
    }

    getDoctype () {
      if (this.isDocumentNode()) {
        return _findDocTypeElement$1(this)
      } else {
        return this.getOwnerDocument().getDoctype()
      }
    }

    setDoctype (qualifiedNameStr, publicId, systemId) {
      
      let doc = this.getOwnerDocument();
      let oldDocType = _findDocTypeElement$1(doc);
      let newDocType = this.createDocumentType(qualifiedNameStr, publicId, systemId);
      if (oldDocType) {
        doc.replaceChild(newDocType, oldDocType);
      } else {
        
        doc.insertBefore(newDocType, doc.getChildren()[0]);
      }
      doc.doctype = newDocType;
    }

    getInnerHTML () {
      return domUtils$1.getInnerHTML(this, { decodeEntities: true })
    }

    
    
    setInnerHTML (html) {
      if (this.childNodes) {
        let _doc = parseMarkup$1(html, {
          ownerDocument: this.getOwnerDocument(),
          decodeEntities: true,
          elementFactory: MemoryDOMElementFactory$1
        });
        this.empty();
        
        
        _doc.childNodes.slice(0).forEach((child) => {
          this.appendChild(child);
        });
      }
      return this
    }

    getOuterHTML () {
      return domUtils$1.getOuterHTML(this, { xmlMode: this._isXML(), decodeEntities: true })
    }

    getTextContent () {
      return domUtils$1.getText(this)
    }

    setTextContent (text) {
      switch (this.type) {
        case domelementtype$1.Text:
        case domelementtype$1.Comment:
        case domelementtype$1.CDATA: {
          this.data = text;
          break
        }
        default: {
          if (this.childNodes) {
            let child = this.createTextNode(text);
            this.empty();
            this.appendChild(child);
          }
        }
      }
      return this
    }

    getStyle (name) {
      if (this.styles) {
        return this.styles.get(name)
      }
    }

    setStyle (name, value) {
      if (this.styles) {
        if (DOMElement$1.pxStyles[name] && isNumber$1(value)) {
          value = value + 'px';
        }
        this.styles.set(name, value);
        this.attributes.set('style', stringifyStyles$1(this.styles));
      }
      return this
    }

    is (cssSelector) {
      return cssSelect$1.is(this, cssSelector, { xmlMode: this._isXML() })
    }

    find (cssSelector) {
      return cssSelect$1.selectOne(cssSelector, this, { xmlMode: this._isXML() })
    }

    findAll (cssSelector) {
      return cssSelect$1.selectAll(cssSelector, this, { xmlMode: this._isXML() })
    }

    getChildCount () {
      if (this.childNodes) {
        return this.childNodes.length
      } else {
        return 0
      }
    }

    getChildNodes () {
      return this.childNodes.slice(0)
    }

    getChildren () {
      return this.childNodes.filter(function (node) {
        return node.type === 'tag'
      })
    }

    get children () {
      return this.getChildren()
    }

    getChildAt (pos) {
      if (this.childNodes) {
        return this.childNodes[pos]
      }
    }

    getChildIndex (child) {
      if (this.childNodes) {
        return this.childNodes.indexOf(child)
      }
    }

    getLastChild () {
      if (this.childNodes) {
        return last$1$1(this.childNodes)
      }
    }

    getFirstChild () {
      if (this.childNodes) {
        return this.childNodes[0]
      }
    }

    getNextSibling () {
      return this.next
    }

    getPreviousSibling () {
      return this.prev
    }

    getParent () {
      return this.parent
    }

    getOwnerDocument () {
      return (this.type === 'document') ? this : this.ownerDocument
    }

    getFormat () {
      return this.getOwnerDocument().format
    }

    createDocument (format) {
      return MemoryDOMElement$1.createDocument(format)
    }

    createElement (tagName) {
      return new MemoryDOMElement$1(domelementtype$1.Tag, { name: tagName, ownerDocument: this.getOwnerDocument() })
    }

    createTextNode (text) {
      return new MemoryDOMElement$1(domelementtype$1.Text, { data: text, ownerDocument: this.getOwnerDocument() })
    }

    createComment (data) {
      return new MemoryDOMElement$1(domelementtype$1.Comment, { data: data, ownerDocument: this.getOwnerDocument() })
    }

    createProcessingInstruction (name, data) {
      return new MemoryDOMElement$1(domelementtype$1.Directive, { name: name, data: data, ownerDocument: this.getOwnerDocument() })
    }

    createDocumentType (qualifiedNameStr, publicId, systemId) {
      return new MemoryDOMDoctype$1(domelementtype$1.Doctype, { data: { name: qualifiedNameStr, publicId, systemId }, ownerDocument: this.getOwnerDocument() })
    }

    createCDATASection (data) {
      return new MemoryDOMElement$1(domelementtype$1.CDATA, { data: data, ownerDocument: this.getOwnerDocument() })
    }

    appendChild (child) {
      if (this.childNodes && !isNil$1(child)) {
        child = this._normalizeChild(child);
        if (!child) return this
        domUtils$1.appendChild(this, child);
        child.ownerDocument = this.getOwnerDocument();
      }
      return this
    }

    removeChild (child) {
      if (child.parentNode === this) {
        child.remove();
      }
    }

    insertAt (pos, child) {
      child = this._normalizeChild(child);
      if (!child) return this
      let childNodes = this.childNodes;
      if (childNodes) {
        
        if (pos >= childNodes.length) {
          domUtils$1.appendChild(this, child);
        } else {
          domUtils$1.prepend(childNodes[pos], child);
        }
        child.ownerDocument = this.getOwnerDocument();
      }
      return this
    }

    insertBefore (newChild, before) {
      if (isNil$1(before)) {
        return this.appendChild(newChild)
      } else if (this.childNodes) {
        var pos = this.childNodes.indexOf(before);
        if (pos > -1) {
          domUtils$1.prepend(before, newChild);
          newChild.ownerDocument = this.getOwnerDocument();
        } else {
          throw new Error('insertBefore(): reference node is not a child of this element.')
        }
      }
      return this
    }

    removeAt (pos) {
      let childNodes = this.childNodes;
      if (childNodes) {
        let child = childNodes[pos];
        child.remove();
      }
      return this
    }

    empty () {
      let childNodes = this.childNodes;
      if (childNodes) {
        childNodes.forEach((child) => {
          child.next = child.prev = child.parent = null;
        });
        childNodes.length = 0;
      }
      return this
    }

    remove () {
      domUtils$1.removeElement(this);
      return this
    }

    replaceChild (oldChild, newChild) {
      if (oldChild.parent === this) {
        oldChild.replaceWith(newChild);
      }
      return this
    }

    replaceWith (newEl) {
      newEl = this._normalizeChild(newEl);
      domUtils$1.replaceElement(this, newEl);
      newEl.ownerDocument = this.getOwnerDocument();
      return this
    }

    getEventListeners () {
      return this.eventListeners || []
    }

    click () {
      this.emit('click', { target: this });
      return this
    }

    emit (name, data) {
      this._propagateEvent(new MemoryDOMElementEvent$1(name, this, data));
    }

    getBoundingClientRect () {
      return { top: 0, left: 0, height: 0, width: 0 }
    }

    getClientRects () {
      return [{ top: 0, left: 0, height: 0, width: 0 }]
    }

    _propagateEvent (event) {
      let listeners = this.eventListeners;
      if (listeners) {
        let listener = listeners.find((l) => {
          return l.eventName === event.type
        });
        if (listener) listener.handler(event);
        if (event.stopped) return
        let p = this.parentNode;
        if (p) p._propagateEvent(event);
      }
    }

    removeAllEventListeners () {
      this.eventListeners = [];
      return this
    }

    _assign (other) {
      if (other.name) this.name = other.name;
      if (this.classes && other.classes) {
        other.classes.forEach((val) => {
          this.classes.add(val);
        });
      }
      if (this.styles && other.styles) {
        forEach$1(other.styles, (val, name) => {
          this.styles.set(name, val);
        });
      }
      
      
      let otherAttributes = other.attributes || other.attribs;
      if (this.attributes && otherAttributes) {
        forEach$1(otherAttributes, (val, name) => {
          switch (name) {
            case 'class': {
              parseClasses$1(this.classes, val);
              break
            }
            case 'style': {
              parseStyles$1(this.styles, val);
              break
            }
            default:
              
          }
          this.attributes.set(name, val);
        });
      }
      if (this.eventListeners && other.eventListeners) {
        this.eventListeners = this.eventListeners.concat(other.eventListeners);
      }
    }

    _normalizeChild (child) {
      if (isNil$1(child)) return

      if (isString$1(child)) {
        child = this.createTextNode(child);
      }
      
      if (!child || !child._isMemoryDOMElement) {
        throw new Error('Illegal argument: only String and MemoryDOMElement instances are valid.')
      }
      return child
    }

    _normalizeName (name) {
      if (this._isXML()) {
        return name
      } else {
        return name.toLowerCase()
      }
    }

    _isHTML () {
      return this.getFormat() === 'html'
    }

    _isXML () {
      return this.getFormat() === 'xml'
    }
  }

  MemoryDOMElement$1.prototype._isMemoryDOMElement = true;

  MemoryDOMElement$1.createDocument = function (format) {
    if (format === 'xml') {
      return new MemoryDOMElement$1('document', { format: format })
    } else {
      return MemoryDOMElement$1.parseMarkup(DOMElement$1.EMPTY_HTML, 'html')
    }
  };

  MemoryDOMElement$1.parseMarkup = function (str, format, options = {}) {
    if (!str) {
      return MemoryDOMElement$1.createDocument(format)
    }
    let parserOpts = Object.assign({
      format,
      decodeEntities: true,
      elementFactory: MemoryDOMElementFactory$1
    }, options);
    
    if (options.raw) {
      return parseMarkup$1(str, parserOpts)
    }
    if (options.snippet) {
      str = `<__snippet__>${str}</__snippet__>`;
    }
    let doc;
    if (format === 'html') {
      doc = parseMarkup$1(str, parserOpts);
      _sanitizeHTMLStructure$1(doc);
    } else if (format === 'xml') {
      doc = parseMarkup$1(str, parserOpts);
    }
    if (options.snippet) {
      let childNodes = doc.find('__snippet__').childNodes;
      if (childNodes.length === 1) {
        return childNodes[0]
      } else {
        return childNodes
      }
    } else {
      return doc
    }
  };

  MemoryDOMElement$1.wrap =
  MemoryDOMElement$1.wrapNativeElement = function (el) {
    if (inBrowser$1) {
      
      
      
      
      if (el === window || el === window.document) {
        return new DOMElementStub$1()
      
      
      
      } else if (el instanceof window.Node || el._isBrowserDOMElement) ;
    }
    
    if (!el._isMemoryDOMElement) {
      throw new Error('Illegal argument: expected MemoryDOMElement instance')
    }
    return el
  };

  MemoryDOMElement$1.unwrap = function (el) {
    
    if (!el._isMemoryDOMElement) {
      throw new Error('Illegal argument: expected MemoryDOMElement instance')
    }
    return el
  };




  MemoryDOMElement$1.isReverse = function () {
    return false
  };


  let _browserWindowStub$1;
  MemoryDOMElement$1.getBrowserWindow = function () {
    
    if (!_browserWindowStub$1) {
      _browserWindowStub$1 = MemoryDOMElement$1.createDocument('html');
    }
    return _browserWindowStub$1
  };

  class MemoryDOMDoctype$1 extends MemoryDOMElement$1 {
    get name () { return this.data.name }
    get publicId () { return this.data.publicId }
    get systemId () { return this.data.systemId }
  }

  function parseClasses$1 (classes, classStr) {
    classStr.split(/\s+/).forEach((name) => {
      classes.add(name);
    });
  }

  function stringifyClasses$1 (classes) {
    return Array.from(classes).join(' ')
  }

  function parseStyles$1 (styles, styleStr) {
    styleStr = (styleStr || '').trim();
    if (!styleStr) return
    styleStr.split(';').forEach((style) => {
      let n = style.indexOf(':');
      
      if (n < 1 || n === style.length - 1) return
      let name = style.slice(0, n).trim();
      let val = style.slice(n + 1).trim();
      styles.set(name, val);
    });
  }

  function stringifyStyles$1 (styles) {
    if (!styles) return ''
    let str = Object.keys(styles).map((name) => {
      return name + ':' + styles[name]
    }).join(';');
    if (str.length > 0) str += ';';
    return str
  }

  const BUILTIN_EVENTS$1 = [
    'keydown', 'keyup', 'keypress',
    'mousedown', 'mouseup', 'mouseover', 'click', 'dblclick'
  ].reduce((m, k) => { m[k] = true; return m }, {});

  class MemoryDOMElementEvent$1 {
    constructor (type, target, detail) {
      this.type = type;
      this.timeStamp = Date.now();
      this.target = target;

      if (BUILTIN_EVENTS$1[type]) {
        
        if (detail) {
          Object.assign(this, detail);
        }
      } else {
        this.detail = detail;
      }
    }

    stopPropagation () {
      this.stopped = true;
    }

    preventDefault () {
      this.defaultPrevented = true;
    }
  }

  class DOMElementStub$1 {
    on () {}
    off () {}
  }

  function nameWithoutNS$2 (name) {
    const idx = name.indexOf(':');
    if (idx > 0) {
      return name.slice(idx + 1)
    } else {
      return name
    }
  }



  const ATTR_TO_PROPS$1 = {
    'input': {
      'value': true,
      'checked': (el, name, value) => {
        const checked = (value !== 'off');
        el.setProperty('checked', checked);
      }
    }
  };

  function deriveHTMLPropertyFromAttribute$1 (el, name, value) {
    const mappings = ATTR_TO_PROPS$1[el.tagName];
    if (mappings) {
      let mapper = mappings[name];
      if (mapper === true) {
        el.setProperty(name, value);
      } else if (mapper) {
        mapper(el, name, value);
      }
    }
  }

  const PROPERTY_TRANSFORMATIONS$1 = {
    'input': {
      'checked': (el, name, value) => {
        if (value === true) {
          el.properties.set(name, true);
          el.properties.set('value', 'on');
        } else {
          el.properties.set(name, false);
          el.properties.set('value', 'off');
        }
      },
      'value': (el, name, value) => {
        let type = el.getAttribute('type');
        switch (type) {
          case 'checkbox':
            if (value === 'on') {
              el.properties.set(name, true);
              el.properties.set('value', 'on');
            } else {
              el.properties.set(name, false);
              el.properties.set('value', 'off');
            }
            break
          default:
            _setProperty$1(el, name, value);
        }
      }
    }
  };

  function _setProperty$1 (el, name, value) {
    if (value === undefined) {
      el.properties.delete(name);
    } else {
      el.properties.set(name, String(value));
    }
  }

  function _setHTMLPropertyValue$1 (el, name, value) {
    const trafos = PROPERTY_TRANSFORMATIONS$1[el.tagName];
    if (trafos) {
      let mapper = trafos[name];
      if (mapper) {
        mapper(el, name, value);
        return
      }
    }
    _setProperty$1(el, name, value);
  }

  function _sanitizeHTMLStructure$1 (doc) {
    
    
    
    
    let htmlEl = doc.find('html');
    if (!htmlEl) {
      
      
      let headEl = doc.find('head');
      let titleEl = doc.find('title');
      let metaEls = doc.findAll('meta');
      let bodyEl = doc.find('body');
      if (headEl) headEl.remove();
      if (titleEl) titleEl.remove();
      metaEls.forEach(e => e.remove());
      if (bodyEl) bodyEl.remove();

      
      
      let contentNodes = doc.childNodes.slice();
      contentNodes.forEach((c) => { c.parent = null; });
      doc.childNodes.length = 0;

      htmlEl = doc.createElement('html');
      
      
      
      if (!headEl) {
        headEl = doc.createElement('head');
        headEl.appendChild(titleEl);
        headEl.append(metaEls);
        htmlEl.appendChild(headEl);
      }
      if (!bodyEl) {
        bodyEl = doc.createElement('body');
        bodyEl.append(contentNodes);
      }
      htmlEl.appendChild(bodyEl);

      doc.append(htmlEl);
    }
  }

  function _findDocTypeElement$1 (doc) {
    
    if (doc.doctype) return doc.doctype
    const childNodes = doc.childNodes;
    for (let i = 0; i < childNodes.length; i++) {
      let child = childNodes[i];
      if (child.type === domelementtype$1.Doctype) {
        doc.doctype = child;
        return child
      }
    }
  }

  let DefaultDOMElement$1 = {};

  DefaultDOMElement$1.createDocument = function (format) {
    return _getDefaultImpl$1().createDocument(format)
  };


  DefaultDOMElement$1.createElement = function (tagName) {
    console.error('DEPRECATED: every element should have an ownerDocument. Use DefaultDOMElement.createDocument() to create a document first');
    let doc = DefaultDOMElement$1.createDocument('html');
    return doc.createElement(tagName)
  };


  DefaultDOMElement$1.createTextNode = function (text) {
    console.error('DEPRECATED: every element should have a ownerDocument. Use DefaultDOMElement.createDocument() to create a document first');
    let doc = DefaultDOMElement$1.createDocument('html');
    return doc.createTextNode(text)
  };


  DefaultDOMElement$1.getBrowserWindow = function () {
    return _getDefaultImpl$1().getBrowserWindow()
  };


  DefaultDOMElement$1.parseHTML = function (html, options) {
    return _getDefaultImpl$1().parseMarkup(html, 'html', options)
  };


  DefaultDOMElement$1.parseXML = function (xml, options) {
    return _getDefaultImpl$1().parseMarkup(xml, 'xml', options)
  };

  DefaultDOMElement$1.parseSnippet = function (str, format) {
    return _getDefaultImpl$1().parseMarkup(str, format, {snippet: true})
  };

  DefaultDOMElement$1.wrap =
  DefaultDOMElement$1.wrapNativeElement = function (nativeEl) {
    if (!nativeEl) throw new Error('Illegal argument')
    return _getDefaultImpl$1().wrap(nativeEl)
  };

  DefaultDOMElement$1.unwrap = function (nativeEl) {
    if (!nativeEl) throw new Error('Illegal argument')
    return _getDefaultImpl$1().unwrap(nativeEl)
  };



  DefaultDOMElement$1.isReverse = function (anchorNode, anchorOffset, focusNode, focusOffset) {
    return _getDefaultImpl$1().isReverse(anchorNode, anchorOffset, focusNode, focusOffset)
  };

  function _getDefaultImpl$1 () {
    
    if (platform$1.inBrowser || platform$1.inElectron) {
      return BrowserDOMElement$1
    } else {
      return MemoryDOMElement$1
    }
  }

  const _global$1 = (typeof global !== 'undefined') ? global : window;
  const substanceGlobals$1 = _global$1.hasOwnProperty('Substance') ? _global$1.Substance : _global$1.Substance = {
    DEBUG_RENDERING: false
  };

  class VirtualElement$1 extends DOMElement$1 {
    constructor (owner) {
      super();

      
      this.parent = null;
      
      this._owner = owner;
      
      this._ref = null;
    }

    getParent () {
      return this.parent
    }

    get childNodes () {
      return this.getChildNodes()
    }

    getChildCount () {
      return this.children.length
    }

    getChildAt (idx) {
      return this.children[idx]
    }

    
    getComponent () {
      return this._comp
    }

    
    ref (ref) {
      if (!ref) throw new Error('Illegal argument')
      
      if (this._ref) throw new Error('A VirtualElement can only be referenced once.')
      this._ref = ref;
      if (this._context) {
        const refs = this._context.refs;
        if (refs[ref]) {
          throw new Error('An item with reference "' + ref + '" already exists.')
        }
        refs[ref] = this;
      }
      return this
    }

    isInDocument () {
      return false
    }
  }

  VirtualElement$1.prototype._isVirtualElement = true;


  class VirtualHTMLElement$1 extends VirtualElement$1 {
    constructor (tagName) {
      super();

      this._tagName = tagName;
      this.classNames = null;
      this.attributes = null;
      this.htmlProps = null;
      this.style = null;
      this.eventListeners = null;

      
      this.children = [];
    }

    getTagName () {
      return this._tagName
    }

    setTagName (tagName) {
      this._tagName = tagName;
      return this
    }

    hasClass (className) {
      if (this.classNames) {
        return this.classNames.indexOf(className) > -1
      }
      return false
    }

    addClass (className) {
      if (!this.classNames) {
        this.classNames = [];
      }
      this.classNames.push(className);
      return this
    }

    removeClass (className) {
      if (this.classNames) {
        this.classNames = without$1(this.classNames, className);
      }
      return this
    }

    removeAttribute (name) {
      if (this.attributes) {
        delete this.attributes[name];
      }
      return this
    }

    getAttribute (name) {
      if (this.attributes) {
        return this.attributes[name]
      }
    }

    setAttribute (name, value) {
      if (!this.attributes) {
        this.attributes = {};
      }
      this.attributes[name] = value;
      return this
    }

    getAttributes () {
      
      
      
      var attributes = {};
      if (this.attributes) {
        Object.assign(attributes, this.attributes);
      }
      if (this.classNames) {
        attributes.class = this.classNames.join(' ');
      }
      if (this.style) {
        attributes.style = map$2(this.style, function (val, key) {
          return key + ':' + val
        }).join(';');
      }
      return attributes
    }

    getId () {
      return this.getAttribute('id')
    }

    setId (id) {
      this.setAttribute('id', id);
      return this
    }

    setTextContent (text) {
      text = String(text || '');
      this.empty();
      this.appendChild(text);
      return this
    }

    setInnerHTML (html) {
      html = html || '';
      this.empty();
      this._innerHTMLString = html;
      return this
    }

    getInnerHTML () {
      if (!this.hasOwnProperty('_innerHTMLString')) {
        throw new Error('Not supported.')
      } else {
        return this._innerHTMLString
      }
    }

    getValue () {
      return this.htmlProp('value')
    }

    setValue (value) {
      this.htmlProp('value', value);
      return this
    }

    getChildNodes () {
      return this.children
    }

    getChildren () {
      return this.children.filter(function (child) {
        return child.getNodeType() !== 'text'
      })
    }

    isTextNode () {
      return false
    }

    isElementNode () {
      return true
    }

    isCommentNode () {
      return false
    }

    isDocumentNode () {
      return false
    }

    append () {
      if (this._innerHTMLString) {
        throw Error('It is not possible to mix $$.html() with $$.append(). You can call $$.empty() to reset this virtual element.')
      }
      this._append(this.children, arguments);
      return this
    }

    appendChild (child) {
      if (this._innerHTMLString) {
        throw Error('It is not possible to mix $$.html() with $$.append(). You can call $$.empty() to reset this virtual element.')
      }
      this._appendChild(this.children, child);
      return this
    }

    insertAt (pos, child) {
      child = this._normalizeChild(child);
      if (!child) {
        throw new Error('Illegal child: ' + child)
      }
      if (!child._isVirtualElement) {
        throw new Error('Illegal argument for $$.insertAt():' + child)
      }
      if (pos < 0 || pos > this.children.length) {
        throw new Error('insertAt(): index out of bounds.')
      }
      this._insertAt(this.children, pos, child);
      return this
    }

    insertBefore (child, before) {
      var pos = this.children.indexOf(before);
      if (pos > -1) {
        this.insertAt(pos, child);
      } else {
        throw new Error('insertBefore(): reference node is not a child of this element.')
      }
      return this
    }

    removeAt (pos) {
      if (pos < 0 || pos >= this.children.length) {
        throw new Error('removeAt(): Index out of bounds.')
      }
      this._removeAt(pos);
      return this
    }

    removeChild (child) {
      if (!child || !child._isVirtualElement) {
        throw new Error('removeChild(): Illegal arguments. Expecting a CheerioDOMElement instance.')
      }
      var idx = this.children.indexOf(child);
      if (idx < 0) {
        throw new Error('removeChild(): element is not a child.')
      }
      this.removeAt(idx);
      return this
    }

    replaceChild (oldChild, newChild) {
      if (!newChild || !oldChild ||
          !newChild._isVirtualElement || !oldChild._isVirtualElement) {
        throw new Error('replaceChild(): Illegal arguments. Expecting BrowserDOMElement instances.')
      }
      var idx = this.children.indexOf(oldChild);
      if (idx < 0) {
        throw new Error('replaceChild(): element is not a child.')
      }
      this.removeAt(idx);
      this.insertAt(idx, newChild);
      return this
    }

    empty () {
      var children = this.children;
      while (children.length) {
        var child = children.pop();
        child.parent = null;
      }
      delete this._innerHTMLString;
      return this
    }

    getProperty (name) {
      if (this.htmlProps) {
        return this.htmlProps[name]
      }
    }

    setProperty (name, value) {
      if (!this.htmlProps) {
        this.htmlProps = {};
      }
      this.htmlProps[name] = value;
      return this
    }

    removeProperty (name) {
      if (this.htmlProps) {
        delete this.htmlProps[name];
      }
      return this
    }

    getStyle (name) {
      if (this.style) {
        return this.style[name]
      }
    }

    setStyle (name, value) {
      if (!this.style) {
        this.style = {};
      }
      if (DOMElement$1.pxStyles[name] && isNumber$1(value)) value = value + 'px';
      this.style[name] = value;
      return this
    }

    _createEventListener (eventName, handler, options) {
      options.context = options.context || this._owner._comp;
      return super._createEventListener(eventName, handler, options)
    }

    getNodeType () {
      return 'element'
    }

    hasInnerHTML () {
      return Boolean(this._innerHTMLString)
    }

    _normalizeChild (child) {
      if (isNil$1(child)) ; else if (child._isVirtualElement) {
        return child
      } else if (isString$1(child) || isBoolean$1(child) || isNumber$1(child)) {
        return new VirtualTextNode$1(String(child))
      } else {
        throw new Error('Unsupported child type')
      }
    }

    _append (outlet, args) {
      if (args.length === 1 && !isArray$2(args[0])) {
        this._appendChild(outlet, args[0]);
        return
      }
      var children;
      if (isArray$2(args[0])) {
        children = args[0];
      } else if (arguments.length > 1) {
        children = Array.prototype.slice.call(args, 0);
      } else {
        return
      }
      children.forEach(this._appendChild.bind(this, outlet));
    }

    _appendChild (outlet, child) {
      child = this._normalizeChild(child);
      
      
      if (!child) return
      outlet.push(child);
      this._attach(child);
      return child
    }

    _insertAt (outlet, pos, child) {
      if (!child) return
      outlet.splice(pos, 0, child);
      this._attach(child);
    }

    _removeAt (outlet, pos) {
      var child = outlet[pos];
      outlet.splice(pos, 1);
      this._detach(child);
    }

    _attach (child) {
      child.parent = this;
      if (this._context && child._owner !== this._owner && child._ref) {
        this._context.foreignRefs[child._ref] = child;
      }
    }

    _detach (child) {
      child.parent = null;
      if (this._context && child._owner !== this._owner && child._ref) {
        delete this.context.foreignRefs[child._ref];
      }
    }

    _mergeHTMLConfig (other) {
      if (other.classNames) {
        if (!this.classNames) {
          this.classNames = [];
        }
        this.classNames = this.classNames.concat(other.classNames);
      }
      if (other.attributes) {
        if (!this.attributes) {
          this.attributes = {};
        }
        Object.assign(this.attributes, other.attributes);
      }
      if (other.htmlProps) {
        if (!this.htmlProps) {
          this.htmlProps = {};
        }
        Object.assign(this.htmlProps, other.htmlProps);
      }
      if (other.style) {
        if (!this.style) {
          this.style = {};
        }
        Object.assign(this.style, other.style);
      }
      if (other.eventListeners) {
        if (!this.eventListeners) {
          this.eventListeners = [];
        }
        this.eventListeners = this.eventListeners.concat(other.eventListeners);
      }
    }
  }

  VirtualHTMLElement$1.prototype._isVirtualHTMLElement = true;


  class VirtualComponent$1 extends VirtualHTMLElement$1 {
    constructor (ComponentClass, props) {
      super();

      props = props || {};

      this.ComponentClass = ComponentClass;
      this.props = props;
      if (!props.children) {
        props.children = [];
      }
      this.children = props.children;
    }

    get _isVirtualHTMLElement () { return false }

    get _isVirtualComponent () { return true }

    getComponent () {
      return this._comp
    }

    
    
    getChildren () {
      return this.props.children
    }

    getNodeType () {
      return 'component'
    }

    outlet (name) {
      return new Outlet$1(this, name)
    }

    _attach (child) {
      child._preliminaryParent = this;
    }

    _detach (child) {
      child._preliminaryParent = null;
    }

    _copyHTMLConfig () {
      return {
        classNames: clone$1(this.classNames),
        attributes: clone$1(this.attributes),
        htmlProps: clone$1(this.htmlProps),
        style: clone$1(this.style),
        eventListeners: clone$1(this.eventListeners)
      }
    }
  }

  class Outlet$1 {
    constructor (virtualEl, name) {
      this.virtualEl = virtualEl;
      this.name = name;
      Object.freeze(this);
    }

    _getOutlet () {
      var outlet = this.virtualEl.props[this.name];
      if (!outlet) {
        outlet = [];
        this.virtualEl.props[this.name] = outlet;
      }
      return outlet
    }

    append () {
      var outlet = this._getOutlet();
      this.virtualEl._append(outlet, arguments);
      return this
    }

    empty () {
      var arr = this.virtualEl.props[this.name];
      arr.forEach(function (el) {
        this._detach(el);
      }.bind(this));
      arr.splice(0, arr.length);
      return this
    }
  }

  class VirtualTextNode$1 extends VirtualElement$1 {
    constructor (text) {
      super();
      this.text = text;
    }

    get _isVirtualTextNode () { return true }
  }

  VirtualElement$1.Component = VirtualComponent$1;
  VirtualElement$1.TextNode = VirtualTextNode$1;


  VirtualElement$1.createElement = function () {
    var content;
    var _first = arguments[0];
    var _second = arguments[1];
    var type;
    if (isString$1(_first)) {
      type = 'element';
    } else if (isFunction$1$1(_first) && _first.prototype._isComponent) {
      type = 'component';
    } else if (isNil$1(_first)) {
      throw new Error('$$(null): provided argument was null or undefined.')
    } else {
      throw new Error('Illegal usage of $$()')
    }
    
    var props = {};
    var classNames, ref;
    var eventHandlers = [];
    for (var key in _second) {
      if (!_second.hasOwnProperty(key)) continue
      var val = _second[key];
      switch (key) {
        case 'class':
          classNames = val;
          break
        case 'ref':
          ref = val;
          break
        default:
          props[key] = val;
      }
    }
    if (type === 'element') {
      content = new VirtualHTMLElement$1(_first);
      
      
      content.attr(props);
    } else {
      content = new VirtualComponent$1(_first, props);
    }
    
    
    content._owner = this.owner;
    if (classNames) {
      content.addClass(classNames);
    }
    if (ref) {
      content.ref(ref);
    }
    eventHandlers.forEach(function (h) {
      if (isFunction$1$1(h.handler)) {
        content.on(h.name, h.handler);
      } else if (isPlainObject$2(h.handler)) {
        var params = h.handler;
        content.on(h.name, params.handler, params.context, params);
      } else {
        throw new Error('Illegal arguments for $$(_,{ on' + h.name + '})')
      }
    });
    
    
    if (arguments.length > 2) {
      content.append(flattenOften$1(Array.prototype.slice.call(arguments, 2), 3));
    }
    return content
  };

  class RenderingEngine$1 {
    constructor (options = {}) {
      this.componentFactory = options.componentFactory;
      if (!this.componentFactory) throw new Error("'componentFactory' is mandatory")
      this.elementFactory = options.elementFactory || DefaultDOMElement$1.createDocument('html');
      if (!this.elementFactory) throw new Error("'elementFactory' is mandatory")
    }

    _render (comp, oldProps, oldState) {
      
      var vel = _createWrappingVirtualComponent$1(comp);
      var state = this._createState();
      if (oldProps) {
        state.setOldProps(vel, oldProps);
      }
      if (oldState) {
        state.setOldState(vel, oldState);
      }
      try {
        
        
        
        _capture$1(state, vel, 'forceCapture');
        

        
        
        _render$1(state, vel);
        

        _triggerUpdate$1(state, vel);
      } finally {
        state.dispose();
      }
      
    }

    
    _renderChild (comp, vel) {
      
      
      var state = this._createState();
      vel.parent = { _comp: comp };
      try {
        _capture$1(state, vel);
        _render$1(state, vel);
        return vel._comp
      } finally {
        state.dispose();
      }
    }

    _createState () {
      return new RenderingState$1(this.componentFactory, this.elementFactory)
    }
  }

  function _create$1 (state, vel) {
    var comp = vel._comp;
    console.assert(!comp, 'Component instance should not exist when this method is used.');
    var parent = vel.parent._comp;
    
    if (!parent) {
      parent = _create$1(state, vel.parent);
    }
    if (vel._isVirtualComponent) {
      console.assert(parent, 'A Component should have a parent.');
      comp = state.componentFactory.createComponent(vel.ComponentClass, parent, vel.props);
      
      vel.props = comp.props;
      comp.__htmlConfig__ = vel._copyHTMLConfig();
    } else if (vel._isVirtualHTMLElement) {
      comp = state.componentFactory.createElementComponent(parent, vel);
    } else if (vel._isVirtualTextNode) {
      comp = state.componentFactory.createTextNodeComponent(parent, vel);
    }
    if (vel._ref) {
      comp._ref = vel._ref;
    }
    if (vel._owner) {
      comp._owner = vel._owner._comp;
    }
    vel._comp = comp;
    return comp
  }

  function _capture$1 (state, vel, forceCapture) {
    if (state.isCaptured(vel)) {
      return vel
    }
    
    var comp = vel._comp;
    if (!comp) {
      comp = _create$1(state, vel);
      state.setNew(vel);
    }
    if (vel._isVirtualComponent) {
      var needRerender;
      
      
      
      if (forceCapture) {
        needRerender = true;
      } else {
        
        needRerender = !comp.el || comp.shouldRerender(vel.props, comp.state);
        comp.__htmlConfig__ = vel._copyHTMLConfig();
        state.setOldProps(vel, comp.props);
        state.setOldState(vel, comp.state);
        
        comp._setProps(vel.props);
        if (!state.isNew(vel)) {
          state.setUpdated(vel);
        }
      }
      if (needRerender) {
        var context = new CaptureContext$1(vel);
        var content = comp.render(context.$$);
        if (!content) {
          throw new Error('Component.render() returned nil.')
        }
        if (!content._isVirtualHTMLElement) {
          throw new Error('Component.render() must return a plain element.')
        }

        if (comp.__htmlConfig__) {
          content._mergeHTMLConfig(comp.__htmlConfig__);
        }
        content._comp = comp;
        vel._content = content;
        if (!state.isNew(vel) && comp.isMounted()) {
          state.setUpdated(vel);
        }
        
        _prepareVirtualComponent$1(state, comp, content);
        
        
        if (substanceGlobals$1.DEBUG_RENDERING) {
          
          
          
          
          var stack = content.children.slice(0);
          while (stack.length) {
            var child = stack.shift();
            if (state.isCaptured(child)) continue
            
            if (child._isVirtualComponent) continue
            if (!child._comp) {
              _create$1(state, child);
            }
            if (child._isVirtualHTMLElement && child.children.length > 0) {
              stack = stack.concat(child.children);
            }
            state.setCaptured(child);
          }
          state.setCaptured(content);
          
          
          var descendingContext = new DescendingContext$1(state, context);
          while (descendingContext.hasPendingCaptures()) {
            descendingContext.reset();
            comp.render(descendingContext.$$);
          }
        } else {
          
          
          _capture$1(state, vel._content);
        }
      } else {
        state.setSkipped(vel);
      }
    } else if (vel._isVirtualHTMLElement) {
      for (var i = 0; i < vel.children.length; i++) {
        _capture$1(state, vel.children[i]);
      }
    }
    state.setCaptured(vel);
    return vel
  }

  function _render$1 (state, vel) {
    if (state.isSkipped(vel)) return
    

    
    
    
    
    
    

    let comp = vel._comp;
    console.assert(comp && comp._isComponent, 'A captured VirtualElement must have a component instance attached.');

    
    if (vel._isVirtualComponent) {
      _render$1(state, vel._content);

      
      const context = vel._content._context;
      let refs = {};
      let foreignRefs = {};
      forEach$1(context.refs, (vel, ref) => {
        refs[ref] = vel._comp;
      });
      forEach$1(context.foreignRefs, (vel, ref) => {
        foreignRefs[ref] = vel._comp;
      });
      comp.refs = refs;
      comp.__foreignRefs__ = foreignRefs;
      return
    }

    
    if (!comp.el) {
      comp.el = _createElement$2(state, vel);
      comp.el._comp = comp;
    }
    _updateElement$1(comp, vel);

    
    if (vel._isVirtualHTMLElement && !vel.hasInnerHTML()) {
      var newChildren = vel.children;
      var oldComp, virtualComp, newComp;
      var pos1 = 0; var pos2 = 0;

      
      
      
      var oldChildren = [];
      comp.el.getChildNodes().forEach(function (node) {
        var childComp = node._comp;

        
        
        
        

        
        if (!childComp || state.isRelocated(childComp)) {
          comp.el.removeChild(node);
        } else {
          oldChildren.push(childComp);
        }
      });

      while (pos1 < oldChildren.length || pos2 < newChildren.length) {
        
        
        
        
        do {
          oldComp = oldChildren[pos1++];
        } while (oldComp && (state.isDetached(oldComp)))

        virtualComp = newChildren[pos2++];
        
        if (oldComp && !virtualComp) {
          while (oldComp) {
            _removeChild$1(state, comp, oldComp);
            oldComp = oldChildren[pos1++];
          }
          break
        }

        
        if (oldComp && oldComp.el.isTextNode() &&
            virtualComp && virtualComp._isVirtualTextNode &&
            oldComp.el.textContent === virtualComp.text) {
          continue
        }

        
        if (!state.isRendered(virtualComp)) {
          _render$1(state, virtualComp);
        }

        newComp = virtualComp._comp;

        
        
        if (state.isRelocated(newComp)) {
          newComp._setParent(comp);
        }

        console.assert(newComp, 'Component instance should now be available.');

        
        if (virtualComp && !oldComp) {
          _appendChild$1(state, comp, newComp);
          continue
        
        } else if (state.isMapped(virtualComp)) {
          
          if (newComp === oldComp) ; else if (state.isMapped(oldComp)) {
            
            state.setDetached(oldComp);
            _removeChild$1(state, comp, oldComp);
            pos2--;
          
          } else {
            _removeChild$1(state, comp, oldComp);
            pos2--;
          }
        } else if (state.isMapped(oldComp)) {
          _insertChildBefore$1(state, comp, newComp, oldComp);
          pos1--;
        } else {
          
          
          
          
          _replaceChild$1(state, comp, oldComp, newComp);
        }
      }
    }

    state.setRendered(vel);
  }

  function _triggerUpdate$1 (state, vel) {
    if (vel._isVirtualComponent) {
      if (!state.isSkipped(vel)) {
        vel._content.children.forEach(_triggerUpdate$1.bind(null, state));
      }
      if (state.isUpdated(vel)) {
        vel._comp.didUpdate(state.getOldProps(vel), state.getOldState(vel));
      }
    } else if (vel._isVirtualHTMLElement) {
      vel.children.forEach(_triggerUpdate$1.bind(null, state));
    }
  }

  function _appendChild$1 (state, parent, child) {
    parent.el.appendChild(child.el);
    _triggerDidMount$1(state, parent, child);
  }

  function _replaceChild$1 (state, parent, oldChild, newChild) {
    parent.el.replaceChild(oldChild.el, newChild.el);
    if (!state.isDetached(oldChild)) {
      oldChild.triggerDispose();
    }
    _triggerDidMount$1(state, parent, newChild);
  }

  function _insertChildBefore$1 (state, parent, child, before) {
    parent.el.insertBefore(child.el, before.el);
    _triggerDidMount$1(state, parent, child);
  }

  function _removeChild$1 (state, parent, child) {
    parent.el.removeChild(child.el);
    if (!state.isDetached(child)) {
      child.triggerDispose();
    }
  }

  function _triggerDidMount$1 (state, parent, child) {
    if (!state.isDetached(child) &&
        parent.isMounted() && !child.isMounted()) {
      child.triggerDidMount(true);
    }
  }


  function _prepareVirtualComponent$1 (state, comp, vc) {
    var newRefs = {};
    var foreignRefs = {};
    
    
    
    
    if (vc._context) {
      newRefs = vc._context.refs;
      foreignRefs = vc._context.foreignRefs;
    }
    var oldRefs = comp.refs;
    var oldForeignRefs = comp.__foreignRefs__;
    
    forEach$1(newRefs, function (vc, ref) {
      var comp = oldRefs[ref];
      if (comp) _mapComponents$1(state, comp, vc);
    });
    forEach$1(foreignRefs, function (vc, ref) {
      var comp = oldForeignRefs[ref];
      if (comp) _mapComponents$1(state, comp, vc);
    });
  }



  function _mapComponents$1 (state, comp, vc) {
    if (!comp && !vc) return true
    if (!comp || !vc) return false
    
    
    
    
    
    if (state.isMapped(vc) || state.isMapped(comp)) {
      return vc._comp === comp
    }
    if (vc._comp) {
      if (vc._comp === comp) {
        state.setMapped(vc);
        state.setMapped(comp);
        return true
      } else {
        return false
      }
    }
    if (!_isOfSameType$1(comp, vc)) {
      return false
    }

    vc._comp = comp;
    state.setMapped(vc);
    state.setMapped(comp);

    var canMapParent;
    var parent = comp.getParent();
    if (vc.parent) {
      canMapParent = _mapComponents$1(state, parent, vc.parent);
    
    
    
    
    } else if (vc._preliminaryParent) {
      while (parent && parent._isElementComponent) {
        parent = parent.getParent();
      }
      canMapParent = _mapComponents$1(state, parent, vc._preliminaryParent);
    }
    if (!canMapParent) {
      state.setRelocated(vc);
      state.setRelocated(comp);
    }
    return canMapParent
  }

  function _isOfSameType$1 (comp, vc) {
    return (
      (comp._isElementComponent && vc._isVirtualHTMLElement) ||
      (comp._isComponent && vc._isVirtualComponent && comp.constructor === vc.ComponentClass) ||
      (comp._isTextNodeComponent && vc._isVirtualTextNode)
    )
  }

  function _createElement$2 (state, vel) {
    var el;
    if (vel._isVirtualTextNode) {
      el = state.elementFactory.createTextNode(vel.text);
    } else {
      el = state.elementFactory.createElement(vel.tagName);
    }
    return el
  }

  function _updateElement$1 (comp, vel) {
    if (comp._isTextNodeComponent) {
      comp.setTextContent(vel.text);
      return
    }
    var el = comp.el;
    console.assert(el, "Component's element should exist at this point.");
    var tagName = el.getTagName();
    if (vel.tagName.toLowerCase() !== tagName) {
      el.setTagName(vel.tagName);
    }
    _updateHash$1({
      oldHash: el.getAttributes(),
      newHash: vel.getAttributes(),
      update: function (key, val) {
        el.setAttribute(key, val);
      },
      remove: function (key) {
        el.removeAttribute(key);
      }
    });
    _updateHash$1({
      oldHash: el.htmlProps,
      newHash: vel.htmlProps,
      update: function (key, val) {
        el.setProperty(key, val);
      },
      remove: function (key) {
        el.removeProperty(key);
      }
    });
    _updateListeners$1({
      el: el,
      oldListeners: el.getEventListeners(),
      newListeners: vel.getEventListeners()
    });

    
    if (vel.hasInnerHTML()) {
      if (!el._hasInnerHTML) {
        el.empty();
        el.setInnerHTML(vel.getInnerHTML());
      } else {
        var oldInnerHTML = el.getInnerHTML();
        var newInnerHTML = vel.getInnerHTML();
        if (oldInnerHTML !== newInnerHTML) {
          el.setInnerHTML(newInnerHTML);
        }
      }
      el._hasInnerHTML = true;
    }
  }

  function _updateHash$1 (args) {
    const newHash = args.newHash;
    const oldHash = args.oldHash || {};
    const update = args.update;
    const remove = args.remove;
    let updatedKeys = {};
    for (let key in newHash) {
      if (newHash.hasOwnProperty(key)) {
        var oldVal = oldHash[key];
        var newVal = newHash[key];
        updatedKeys[key] = true;
        if (oldVal !== newVal) {
          update(key, newVal);
        }
      }
    }
    
    
    if (isFunction$1$1(oldHash.keys) && oldHash.size > 0) {
      let keys = Array.from(oldHash.keys());
      keys.forEach((key) => {
        if (!updatedKeys[key]) {
          remove(key);
        }
      });
    } else {
      for (let key in oldHash) {
        if (oldHash.hasOwnProperty(key) && !updatedKeys[key]) {
          remove(key);
        }
      }
    }
  }

  function _updateListeners$1 (args) {
    var el = args.el;
    
    
    
    var newListeners = args.newListeners || [];
    el.removeAllEventListeners();
    for (var i = 0; i < newListeners.length; i++) {
      el.addEventListener(newListeners[i]);
    }
  }



  class DescendingContext$1 {
    constructor (state, captureContext) {
      this.state = state;
      this.owner = captureContext.owner;
      this.refs = {};
      this.foreignRefs = {};
      this.elements = captureContext.elements;
      this.pos = 0;
      this.updates = captureContext.components.length;
      this.remaining = this.updates;

      this.$$ = this._createComponent.bind(this);
    }

    _createComponent () {
      var state = this.state;
      var vel = this.elements[this.pos++];
      
      
      
      if (!state.isCaptured(vel) && vel._isVirtualComponent &&
           vel.parent && state.isCaptured(vel.parent)) {
        _capture$1(state, vel);
        this.updates++;
        this.remaining--;
      }
      
      
      
      
      vel = VirtualElement$1.createElement.apply(this, arguments);
      
      vel._context = this;
      vel._owner = this.owner;
      
      
      vel._attach = function () {};
      vel._detach = function () {};
      return vel
    }

    hasPendingCaptures () {
      return this.updates > 0 && this.remaining > 0
    }

    reset () {
      this.pos = 0;
      this.updates = 0;
      this.refs = {};
    }

    _ancestorsReady (vel) {
      while (vel) {
        if (this.state.isCaptured(vel) ||
            
            vel === this.owner || vel === this.owner._content) {
          return true
        }
        vel = vel.parent;
      }
      return false
    }
  }

  RenderingEngine$1._internal = {
    _capture: _capture$1,
    _wrap: _createWrappingVirtualComponent$1
  };

  class CaptureContext$1 {
    constructor (owner) {
      this.owner = owner;
      this.refs = {};
      this.foreignRefs = {};
      this.elements = [];
      this.components = [];
      this.$$ = this._createComponent.bind(this);
      this.$$.capturing = true;
    }

    _createComponent () {
      var vel = VirtualElement$1.createElement.apply(this, arguments);
      vel._context = this;
      vel._owner = this.owner;
      if (vel._isVirtualComponent) {
        
        this.components.push(vel);
      }
      this.elements.push(vel);
      return vel
    }
  }

  function _createWrappingVirtualComponent$1 (comp) {
    var vel = new VirtualElement$1.Component(comp.constructor);
    vel._comp = comp;
    if (comp.__htmlConfig__) {
      vel._mergeHTMLConfig(comp.__htmlConfig__);
    }
    return vel
  }

  RenderingEngine$1.createContext = function (comp) {
    var vel = _createWrappingVirtualComponent$1(comp);
    return new CaptureContext$1(vel)
  };

  class RenderingState$1 {
    constructor (componentFactory, elementFactory) {
      this.componentFactory = componentFactory;
      this.elementFactory = elementFactory;
      this.poluted = [];
      this.id = '__' + uuid$1();
    }

    dispose () {
      var id = this.id;
      this.poluted.forEach(function (obj) {
        delete obj[id];
      });
    }

    set (obj, key, val) {
      var info = obj[this.id];
      if (!info) {
        info = {};
        obj[this.id] = info;
        this.poluted.push(obj);
      }
      info[key] = val;
    }

    get (obj, key) {
      var info = obj[this.id];
      if (info) {
        return info[key]
      }
    }

    setMapped (c) {
      this.set(c, 'mapped', true);
    }

    isMapped (c) {
      return Boolean(this.get(c, 'mapped'))
    }

    
    
    setRelocated (c) {
      this.set(c, 'relocated', true);
    }

    isRelocated (c) {
      return Boolean(this.get(c, 'relocated'))
    }

    setDetached (c) {
      this.set(c, 'detached', true);
    }

    isDetached (c) {
      return Boolean(this.get(c, 'detached'))
    }

    setCaptured (vc) {
      this.set(vc, 'captured', true);
    }

    isCaptured (vc) {
      return Boolean(this.get(vc, 'captured'))
    }

    setNew (vc) {
      this.set(vc, 'created', true);
    }

    isNew (vc) {
      return Boolean(this.get(vc, 'created'))
    }

    setUpdated (vc) {
      this.set(vc, 'updated', true);
    }

    isUpdated (vc) {
      return Boolean(this.get(vc, 'updated'))
    }

    setSkipped (vc) {
      this.set(vc, 'skipped', true);
    }

    isSkipped (vc) {
      return Boolean(this.get(vc, 'skipped'))
    }

    setRendered (vc) {
      this.set(vc, 'rendered', true);
    }

    isRendered (vc) {
      return Boolean(this.get(vc, 'rendered'))
    }

    setOldProps (vc, oldProps) {
      this.set(vc, 'oldProps', oldProps);
    }

    getOldProps (vc) {
      return this.get(vc, 'oldProps')
    }

    setOldState (vc, oldState) {
      this.set(vc, 'oldState', oldState);
    }

    getOldState (vc) {
      return this.get(vc, 'oldState')
    }
  }

  const COMPONENT_FACTORY$1 = {
    createComponent (ComponentClass, parent, props) {
      return new ComponentClass(parent, props)
    },
    createElementComponent (parent, virtualElement) {
      return new ElementComponent$1(parent, virtualElement)
    },
    createTextNodeComponent (parent, virtualElement) {
      return new TextNodeComponent$1(parent, virtualElement)
    }
  };

  class Component$1 extends EventEmitter$1 {
    
    constructor (parent, props = {}, options = {}) {
      super();

      
      
      
      
      

      this.parent = (parent && parent._isComponent) ? parent : null;

      
      this.el = options.el;

      
      
      let context = options.context ? options.context : this._getContext() || {};
      this.context = context;
      Object.freeze(this.context);

      
      
      
      this.renderingEngine = (parent && parent.renderingEngine) || context.renderingEngine || options.renderingEngine || new RenderingEngine$1({
        componentFactory: COMPONENT_FACTORY$1
      });

      
      
      if (this._SKIP_COMPONENT_INIT) return

      this.__id__ = uuid$1();

      
      this.refs = {};
      
      
      
      this.__foreignRefs__ = {};

      
      this._actionHandlers = {};

      
      this.props = props;
      Object.freeze(this.props);

      
      this.state = this.getInitialState() || {};
      Object.freeze(this.state);
    }

    getId () {
      return this.__id__
    }

    setId () {
      throw new Error("'id' is readonly")
    }

    
    getChildContext () {
      return this.childContext || {}
    }

    
    getInitialState () {
      return {}
    }

    
    getParent () {
      return this.parent
    }

    
    getRoot () {
      var comp = this;
      var parent = comp;
      while (parent) {
        comp = parent;
        parent = comp.getParent();
      }
      return comp
    }

    getElement () {
      return this.el
    }

    getNativeElement () {
      return this.el.getNativeElement()
    }

    
    getLabel (name, ...args) {
      let labelProvider = this.getLabelProvider();
      if (!labelProvider) throw new Error('Missing labelProvider.')
      return labelProvider.getLabel(name, ...args)
    }

    getLabelProvider () {
      return this.context.labelProvider
    }

    
    getComponent (componentName, maybe) {
      let componentRegistry = this.getComponentRegistry();
      if (!componentRegistry) throw new Error('Missing componentRegistry.')
      const ComponentClass = componentRegistry.get(componentName);
      if (!maybe && !ComponentClass) {
        throw new Error('No Component registered with name ' + componentName)
      }
      return ComponentClass
    }

    getComponentRegistry () {
      return this.props.componentRegistry || this.context.componentRegistry
    }

    getFlow () {
      return this.context.flow
    }

    
    render ($$) {
      
      return $$('div')
    }

    
    mount (el) {
      if (!el) {
        throw new Error('Element is required.')
      }
      el = DefaultDOMElement$1.wrap(el);
      
      this.el = null;
      this.renderingEngine = Component$1.createRenderingEngine(el.getOwnerDocument());
      this._render();
      el.appendChild(this.el);
      if (el.isInDocument()) {
        this.triggerDidMount(true);
      }
      return this
    }

    
    shouldRerender(newProps, newState) { 
      return true
    }

    
    rerender () {
      this._rerender(this.props, this.state);
    }

    _rerender (oldProps, oldState) {
      this._render(oldProps, oldState);
      
      if (!this.isMounted()) {
        this.didUpdate(oldProps, oldState);
      }
    }

    _render (oldProps, oldState) {
      if (this.__isRendering__) {
        throw new Error('Component is rendering already.')
      }
      this.__isRendering__ = true;
      try {
        this.renderingEngine._render(this, oldProps, oldState);
      } finally {
        delete this.__isRendering__;
      }
    }

    
    triggerDidMount () {
      
      
      

      
      
      if (!this.__isMounted__) {
        this.__isMounted__ = true;
        this.didMount();
      }
      
      this.getChildren().forEach(function (child) {
        
        
        child.triggerDidMount(true);
      });
    }

    
    didMount () {}

    
    didUpdate () {}

    
    isMounted () {
      return this.__isMounted__
    }

    
    triggerDispose () {
      this.getChildren().forEach(function (child) {
        child.triggerDispose();
      });
      this.dispose();
      this.__isMounted__ = false;
    }

    
    dispose () {}

    
    _setParent (newParent) {
      this.parent = newParent;
      this.context = this._getContext() || {};
      Object.freeze(this.context);
    }

    
    send (action) {
      
      var comp = this.getParent();
      while (comp) {
        if (comp._actionHandlers && comp._actionHandlers[action]) {
          comp._actionHandlers[action].apply(comp, Array.prototype.slice.call(arguments, 1));
          return true
        }
        comp = comp.getParent();
      }
      console.warn('Action', action, 'was not handled.');
      return false
    }

    
    handleActions (actionHandlers) {
      forEach$1(actionHandlers, function (method, actionName) {
        this.handleAction(actionName, method);
      }.bind(this));
      return this
    }

    
    handleAction (name, handler) {
      if (!name || !handler || !isFunction$1$1(handler)) {
        throw new Error('Illegal arguments.')
      }
      handler = handler.bind(this);
      this._actionHandlers[name] = handler;
    }

    
    getState () {
      return this.state
    }

    
    setState (newState) {
      var oldProps = this.props;
      var oldState = this.state;
      
      
      var needRerender = !this.__isSettingProps__ &&
        this.shouldRerender(this.getProps(), newState);
      
      this.willUpdateState(newState);
      this.state = newState || {};
      Object.freeze(this.state);
      if (needRerender) {
        this._rerender(oldProps, oldState);
      } else if (!this.__isSettingProps__) {
        this.didUpdate(oldProps, oldState);
      }
    }

    
    extendState (newState) {
      newState = extend$1({}, this.state, newState);
      this.setState(newState);
    }

    
    willUpdateState(newState) { 
    }

    
    getProps () {
      return this.props
    }

    
    setProps (newProps) {
      var oldProps = this.props;
      var oldState = this.state;
      var needRerender = this.shouldRerender(newProps, this.state);
      this._setProps(newProps);
      if (needRerender) {
        this._rerender(oldProps, oldState);
      } else {
        this.didUpdate(oldProps, oldState);
      }
    }

    _setProps (newProps) {
      newProps = newProps || {};
      
      this.__isSettingProps__ = true;
      try {
        this.willReceiveProps(newProps);
        this.props = newProps || {};
        Object.freeze(newProps);
      } finally {
        delete this.__isSettingProps__;
      }
    }

    
    extendProps (updatedProps) {
      var newProps = extend$1({}, this.props, updatedProps);
      this.setProps(newProps);
    }

    
    willReceiveProps(newProps) { 
    }

    getTextContent () {
      if (this.el) {
        return this.el.textContent
      }
    }

    get textContent () {
      return this.getTextContent()
    }

    getInnerHTML () {
      if (this.el) {
        return this.el.getInnerHTML()
      }
    }

    get innerHTML () {
      return this.getInnerHTML()
    }

    getOuterHTML () {
      if (this.el) {
        return this.el.getOuterHTML()
      }
    }

    get outerHTML () {
      return this.getOuterHTML()
    }

    getAttribute (name) {
      if (this.el) {
        return this.el.getAttribute(name)
      }
    }

    setAttribute (name, val) {
      if (this.el) {
        this.el.setAttribute(name, val);
      }
      return this
    }

    getProperty (name) {
      if (this.el) {
        return this.el.getProperty(name)
      }
    }

    setProperty (name, val) {
      if (this.el) {
        this.el.setProperty(name, val);
      }
      return this
    }

    hasClass (name) {
      if (this.el) {
        return this.el.hasClass(name)
      }
    }

    addClass (name) {
      if (this.el) {
        this.el.addClass(name);
      }
      return this
    }

    removeClass (name) {
      if (this.el) {
        this.el.removeClass(name);
      }
      return this
    }

    getStyle (name) {
      if (this.el) {
        return this.el.getStyle(name)
      }
    }

    setStyle (name, val) {
      if (this.el) {
        return this.el.setStyle(name, val)
      }
      return this
    }

    getValue () {
      if (this.el) {
        return this.el.getValue()
      }
    }

    setValue (val) {
      if (this.el) {
        this.el.setValue(val);
      }
      return this
    }

    getChildCount () {
      if (!this.el) return 0
      return this.el.getChildCount()
    }

    get childNodes () {
      return this.getChildNodes()
    }

    getChildNodes () {
      if (!this.el) return []
      var childNodes = this.el.getChildNodes();
      childNodes = childNodes.map(_unwrapComp$1).filter(Boolean);
      return childNodes
    }

    getChildren () {
      if (!this.el) return []
      let children = this.el.getChildren();
      children = children.map(_unwrapComp$1).filter(Boolean);
      return children
    }

    getChildAt (pos) {
      let node = this.el.getChildAt(pos);
      if (node) {
        return _unwrapCompStrict$1(node)
      }
    }

    find (cssSelector) {
      var el = this.el.find(cssSelector);
      return _unwrapComp$1(el)
    }

    findAll (cssSelector) {
      var els = this.el.findAll(cssSelector);
      return els.map(_unwrapComp$1).filter(Boolean)
    }

    appendChild (child) {
      this.insertAt(this.getChildCount(), child);
    }

    insertAt (pos, childEl) {
      if (isString$1(childEl)) {
        childEl = new VirtualElement$1.TextNode(childEl);
      }
      if (!childEl._isVirtualElement) {
        throw new Error('Invalid argument: "child" must be a VirtualElement.')
      }
      var child = this.renderingEngine._renderChild(this, childEl);
      this.el.insertAt(pos, child.el);
      _mountChild$1(this, child);
    }

    removeAt (pos) {
      var childEl = this.el.getChildAt(pos);
      if (childEl) {
        var child = _unwrapCompStrict$1(childEl);
        _disposeChild$1(child);
        this.el.removeAt(pos);
      }
    }

    removeChild (child) {
      if (!child || !child._isComponent) {
        throw new Error('removeChild(): Illegal arguments. Expecting a Component instance.')
      }
      
      _disposeChild$1(child);
      this.el.removeChild(child.el);
    }

    replaceChild (oldChild, newChild) {
      if (!newChild || !oldChild ||
          !newChild._isComponent || !oldChild._isComponent) {
        throw new Error('replaceChild(): Illegal arguments. Expecting BrowserDOMElement instances.')
      }
      
      _disposeChild$1(oldChild);
      this.el.replaceChild(newChild.el, oldChild.el);
      if (this.isMounted()) {
        newChild.triggerDidMount(true);
      }
    }

    
    
    
    
    
    
    empty () {
      this._clear();
      return this
    }

    _clear () {
      let el = this.el;
      if (el) {
        this.getChildNodes().forEach(function (child) {
          _disposeChild$1(child);
        });
        el.empty();
      }
      this.refs = {};
      this.__foreignRefs__ = {};
    }

    remove () {
      _disposeChild$1(this);
      this.el.remove();
    }

    addEventListener () {
      throw new Error('Not supported.')
    }

    removeEventListener () {
      throw new Error('Not supported.')
    }

    insertBefore () {
      throw new Error('Not supported.')
    }

    click () {
      if (this.el) {
        this.el.click();
      }
    }

    getComponentPath () {
      let path = [];
      let comp = this;
      while (comp) {
        path.unshift(comp);
        comp = comp.getParent();
      }
      return path
    }

    _getContext () {
      var context = {};
      var parent = this.getParent();
      if (parent) {
        context = extend$1(context, parent.context);
        if (parent.getChildContext) {
          return extend$1(context, parent.getChildContext())
        }
      }
      return context
    }
  }

  Component$1.prototype._isComponent = true;

  Component$1.prototype.attr = DOMElement$1.prototype.attr;

  Component$1.prototype.htmlProp = DOMElement$1.prototype.htmlProp;

  Component$1.prototype.val = DOMElement$1.prototype.val;

  Component$1.prototype.css = DOMElement$1.prototype.css;

  Component$1.prototype.text = DOMElement$1.prototype.text;

  Component$1.prototype.append = DOMElement$1.prototype.append;

  Component$1.unwrap = _unwrapComp$1;

  Component$1.render = function (props) {
    props = props || {};
    var ComponentClass = this;
    var comp = new ComponentClass(null, props);
    comp._render();
    return comp
  };

  Component$1.mount = function (props, el) {
    if (arguments.length === 1) {
      el = props;
      props = {};
    }
    if (!el) throw new Error("'el' is required.")
    if (isString$1(el)) {
      var selector = el;
      if (platform$1.inBrowser) {
        el = window.document.querySelector(selector);
      } else {
        throw new Error('This selector is not supported on server side.')
      }
    }
    el = DefaultDOMElement$1.wrap(el);
    const ComponentClass = this;
    let comp = new ComponentClass(null, props);
    comp.mount(el);
    return comp
  };

  Component$1.getComponentForDOMElement = function (el) {
    return _unwrapComp$1(el)
  };

  Component$1.unwrapDOMElement = function (el) {
    console.warn('DEPRECATED: Use Component.getComponentForDOMElement');
    return Component$1.getComponentForDOMElement(el)
  };

  Component$1.getComponentFromNativeElement = function (nativeEl) {
    
    
    
    return _unwrapComp$1(DefaultDOMElement$1.wrap(nativeEl))
  };

  Component$1.createRenderingEngine = function (elementFactory) {
    return new RenderingEngine$1({
      componentFactory: COMPONENT_FACTORY$1,
      elementFactory
    })
  };


  function _disposeChild$1 (child) {
    child.triggerDispose();
    if (child._owner && child._ref) {
      console.assert(child._owner.refs[child._ref] === child, "Owner's ref should point to this child instance.");
      delete child._owner.refs[child._ref];
    }
  }


  function _mountChild$1 (parent, child) {
    if (parent.isMounted()) {
      child.triggerDidMount(true);
    }
    if (child._owner && child._ref) {
      child._owner.refs[child._ref] = child;
    }
  }


  function _unwrapComp$1 (el) {
    if (el) {
      if (!el._isDOMElement) el = DefaultDOMElement$1.unwrap(el);
      if (el) return el._comp
    }
  }

  function _unwrapCompStrict$1 (el) {
    let comp = _unwrapComp$1(el);
    if (!comp) throw new Error('Expecting a back-link to the component instance.')
    return comp
  }

  class ElementComponent$1 extends Component$1 {}

  ElementComponent$1.prototype._isElementComponent = true;
  ElementComponent$1.prototype._SKIP_COMPONENT_INIT = true;

  class TextNodeComponent$1 extends Component$1 {
    setTextContent (text) {
      if (!this.el) {
        throw new Error('Component must be rendered first.')
      }
      if (this.el.textContent !== text) {
        this.el.textContent = text;
      }
    }

    getChildNodes () {
      return []
    }

    getChildren () {
      return []
    }
  }

  TextNodeComponent$1.prototype._isTextNodeComponent = true;
  TextNodeComponent$1.prototype._SKIP_COMPONENT_INIT = true;

  Component$1.Element = ElementComponent$1;
  Component$1.TextNode = TextNodeComponent$1;

  function getDOMRangeFromEvent$1 (evt) {
    let range;
    let x = evt.clientX;
    let y = evt.clientY;
    
    if (document.body.createTextRange) {
      range = document.body.createTextRange();
      range.moveToPoint(x, y);
    } else if (!isNil$1(document.createRange)) {
      
      
      if (!isNil$1(evt.rangeParent)) {
        range = document.createRange();
        range.setStart(evt.rangeParent, evt.rangeOffset);
        range.collapse(true);
      
      } else if (document.caretPositionFromPoint) {
        let pos = document.caretPositionFromPoint(x, y);
        range = document.createRange();
        range.setStart(pos.offsetNode, pos.offset);
        range.collapse(true);
      
      } else if (document.caretRangeFromPoint) {
        range = document.caretRangeFromPoint(x, y);
      }
    }
    return range
  }

  const events$1 = [ 'keydown', 'keyup', 'keypress', 'mousedown', 'mouseup', 'copy' ];

  class GlobalEventHandler$1 {
    constructor (editorSession, surfaceManager) {
      this.editorSession = editorSession;
      this.surfaceManager = surfaceManager;
      this.listeners = [];
      this.initialize();
    }

    initialize () {
      if (platform$1.inBrowser) {
        let document = DefaultDOMElement$1.wrapNativeElement(window.document);
        events$1.forEach(function (name) {
          document.on(name, this._dispatch.bind(this, name), this);
        }.bind(this));
      }
    }

    dispose () {
      if (platform$1.inBrowser) {
        let document = DefaultDOMElement$1.wrapNativeElement(window.document);
        document.off(this);
      }
    }

    addEventListener (eventName, handler, options) {
      if (!options.id) {
        throw new Error("GlobalEventHandler can only be used with option 'id'")
      }
      let listener = new DOMEventListener$1(eventName, handler, options);
      this.listeners.push(listener);
    }

    removeEventListener (listener) {
      let idx = this.listeners.indexOf(listener);
      if (idx > -1) {
        this.listeners.splice(idx, 1);
      }
    }

    getEventListeners () {
      return this.listeners
    }

    _getActiveListener (eventName) {
      let editorSession = this.editorSession;
      let sel = editorSession.getSelection();
      if (sel) {
        let surfaceId = sel.surfaceId;
        for (let i = 0; i < this.listeners.length; i++) {
          let listener = this.listeners[i];
          if (listener.eventName === eventName && listener.options.id === surfaceId) {
            return listener
          }
        }
      }
    }

    _dispatch (eventName, e) {
      let listener = this._getActiveListener(eventName);
      if (listener) {
        listener.handler(e);
      }
    }
  }

  GlobalEventHandler$1.prototype.on = DOMElement$1.prototype.on;
  GlobalEventHandler$1.prototype.off = DOMElement$1.prototype.off;

  let keys$1$1 = {
    UNDEFINED: 0,
    BACKSPACE: 8,
    DELETE: 46,
    INSERT: 45,
    LEFT: 37,
    RIGHT: 39,
    UP: 38,
    DOWN: 40,
    ENTER: 13,
    RETURN: 13,
    END: 35,
    HOME: 36,
    TAB: 9,
    PAGEUP: 33,
    PAGEDOWN: 34,
    ESCAPE: 27,
    ESC: 27,
    SHIFT: 16,
    SPACE: 32,
    PLUS: 171,
    VOLUMEUP: 183,
    VOLUMEDOWN: 182,
    VOLUMEMUTE: 181,
    PRINTSCREEN: 44
  };


  for (let i = 1; i <= 24; i++) {
    keys$1$1['F' + i] = 111 + i;
  }

  function parseKeyEvent$1 (event, onlyModifiers) {
    let frags = [];
    if (event.altKey) {
      if (event.code === 'AltRight') {
        frags.push('ALTGR');
      } else {
        frags.push('ALT');
      }
    }
    if (event.ctrlKey) frags.push('CTRL');
    if (event.metaKey) frags.push('META');
    if (event.shiftKey) frags.push('SHIFT');
    if (!onlyModifiers) {
      frags.push(event.keyCode);
    }
    return frags.join('+')
  }

  class Marker$1 extends PropertyAnnotation$1 {
    _initialize (doc, props) {
      this.document = doc;
      this.type = props.type;
      if (!props.type) {
        throw new Error("'type' is mandatory")
      }
      if (!props.start) {
        throw new Error("'start' is mandatory")
      }
      if (!props.end) {
        throw new Error("'end' is mandatory")
      }
      Object.assign(this, props);
    }

    
    containsSelection (sel) {
      if (sel.isNull()) return false
      if (sel.isPropertySelection()) {
        return (isArrayEqual$1(this.start.path, sel.start.path) &&
          this.start.offset <= sel.start.offset &&
          this.end.offset >= sel.end.offset)
      } else {
        console.warn('Marker.contains() does not support other selection types.');
      }
    }

    get type () {
      return this._type
    }

    set type (type) {
      this._type = type;
    }
  }


  Marker$1.prototype._isPropertyAnnotation = false;
  Marker$1.prototype._isMarker = true;

  function TextNodeMixin$1 (SuperClass) {
    class TextNodeMixin extends SuperClass {
      getTextPath () {
        
        console.warn('DEPRECATED: use node.getPath()');
        return this.getPath()
      }

      getText () {
        return this.content
      }

      isEmpty () {
        return !this.getText()
      }

      getLength () {
        return this.getText().length
      }

      getAnnotations () {
        return this.getDocument().getIndex('annotations').get(this.getPath())
      }
    }
    return TextNodeMixin
  }

  const ENTER$1 = 1;
  const EXIT$1 = -1;
  const ANCHOR$1 = -2;






























  class Fragmenter$1 {
    constructor (options) {
      Object.assign(this, options);
    }

    start (rootContext, text, annotations) {
      if (!isString$1(text)) {
        throw new Error("Illegal argument: 'text' must be a String, but was " + text)
      }
      this._start(rootContext, text, annotations);
    }

    onText(context, text, entry) { 
    }

    
    onEnter(entry, parentContext) { 
      return null
    }

    onExit(entry, context, parentContext) { 
    }

    _enter (entry, parentContext) {
      entry.counter++;
      return this.onEnter(entry, parentContext)
    }

    _exit (entry, context, parentContext) {
      this.onExit(entry, context, parentContext);
    }

    _createText (context, text, entry) {
      this.onText(context, text, entry);
    }

    _start (rootContext, text, annotations) {
      var entries = _extractEntries$1.call(this, annotations);
      var stack = [{context: rootContext, entry: null}];

      var pos = 0;
      for (var i = 0; i < entries.length; i++) {
        var entry = entries[i];
        var textFragment = text.substring(pos, entry.pos);
        if (textFragment) {
          
          this._createText(stack[stack.length - 1].context, textFragment, entry);
        }

        pos = entry.pos;
        var stackLevel, idx, _entry;
        if (entry.mode === ENTER$1 || entry.mode === ANCHOR$1) {
          
          for (stackLevel = 1; stackLevel < stack.length; stackLevel++) {
            if (entry.level < stack[stackLevel].entry.level) {
              break
            }
          }
          
          
          for (idx = stack.length - 1; idx >= stackLevel; idx--) {
            _entry = stack[idx].entry;
            
            _entry.length = pos - _entry.pos;
            this._exit(_entry, stack[idx].context, stack[idx - 1].context);
          }
          stack.splice(stackLevel, 0, {entry: entry});
          
          for (idx = stackLevel; idx < stack.length; idx++) {
            _entry = stack[idx].entry;
            
            _entry.pos = pos;
            stack[idx].context = this._enter(_entry, stack[idx - 1].context);
          }
        }
        if (entry.mode === EXIT$1 || entry.mode === ANCHOR$1) {
          
          for (stackLevel = 1; stackLevel < stack.length; stackLevel++) {
            if (stack[stackLevel].entry.node === entry.node) {
              break
            }
          }
          for (idx = stack.length - 1; idx >= stackLevel; idx--) {
            _entry = stack[idx].entry;
            
            _entry.length = pos - _entry.pos;
            this._exit(_entry, stack[idx].context, stack[idx - 1].context);
          }
          stack.splice(stackLevel, 1);
          
          for (idx = stackLevel; idx < stack.length; idx++) {
            _entry = stack[idx].entry;
            
            _entry.pos = pos;
            stack[idx].context = this._enter(_entry, stack[idx - 1].context);
          }
        }
      }

      
      var trailingText = text.substring(pos);
      if (trailingText) {
        this._createText(rootContext, trailingText);
      }
    }
  }

  Fragmenter$1.SHOULD_NOT_SPLIT = 0;
  Fragmenter$1.NORMAL = 10;
  Fragmenter$1.ANY = 100;
  Fragmenter$1.ALWAYS_ON_TOP = Number.MAX_VALUE;







































  function _extractEntries$1 (annotations) {
    var openers = [];
    var closers = [];
    forEach$1(annotations, function (a) {
      var isAnchor = (a.isAnchor ? a.isAnchor() : false);
      
      if (isAnchor) {
        openers.push({
          mode: ANCHOR$1,
          pos: a.offset,
          id: a.id,
          level: Fragmenter$1.ALWAYS_ON_TOP,
          type: 'anchor',
          node: a,
          counter: -1,
          length: 0
        });
      } else {
        
        
        
        
        
        
        

        
        var l = Fragmenter$1.NORMAL;
        var isInline = (a.isInline ? a.isInline() : false);
        if (isInline) {
          l = Number.MAX_VALUE;
        } else if (a.constructor.hasOwnProperty('fragmentation')) {
          l = a.constructor.fragmentation;
        } else if (a.hasOwnProperty('fragmentationHint')) {
          l = a.fragmentationHint;
        }
        var startOffset = Math.min(a.start.offset, a.end.offset);
        var endOffset = Math.max(a.start.offset, a.end.offset);
        var opener = {
          pos: startOffset,
          mode: ENTER$1,
          level: l,
          id: a.id,
          type: a.type,
          node: a,
          length: 0,
          counter: -1
        };
        openers.push(opener);
        closers.push({
          pos: endOffset,
          mode: EXIT$1,
          level: l,
          id: a.id,
          type: a.type,
          node: a,
          opener: opener
        });
      }
    });

    
    openers.sort(_compareOpeners$1);
    
    for (var i = openers.length - 1; i >= 0; i--) {
      openers[i].idx = i;
    }
    closers.sort(_compareClosers$1);
    
    var entries = new Array(openers.length + closers.length);
    var idx = 0;
    var idx1 = 0;
    var idx2 = 0;
    var opener = openers[idx1];
    var closer = closers[idx2];
    while (opener || closer) {
      if (opener && closer) {
        
        if (closer.pos <= opener.pos && closer.opener !== opener) {
          entries[idx] = closer;
          idx2++;
        } else {
          entries[idx] = opener;
          idx1++;
        }
      } else if (opener) {
        entries[idx] = opener;
        idx1++;
      } else if (closer) {
        entries[idx] = closer;
        idx2++;
      }
      opener = openers[idx1];
      closer = closers[idx2];
      idx++;
    }
    return entries
  }

  function _compareOpeners$1 (a, b) {
    if (a.pos < b.pos) return -1
    if (a.pos > b.pos) return 1
    if (a.mode < b.mode) return -1
    if (a.mode > b.mode) return 1
    if (a.mode === b.mode) {
      if (a.level < b.level) return -1
      if (a.level > b.level) return 1
    }
    return 0
  }


  function _compareClosers$1 (a, b) {
    if (a.pos < b.pos) return -1
    if (a.pos > b.pos) return 1
    
    
    
    
    
    if (a.pos === a.opener.pos && b.pos === b.opener.pos) {
      if (a.opener.idx < b.opener.idx) {
        return -1
      } else {
        return 1
      }
    }
    if (a.opener.idx > b.opener.idx) return -1
    if (a.opener.idx < b.opener.idx) return 1
    return 0
  }

  function appendChild$1 (xmlNode, child) {
    insertAt$1(xmlNode, xmlNode._childNodes.length, child);
    return xmlNode
  }

  function getChildPos$1 (xmlNode, child) {
    return xmlNode._childNodes.indexOf(child.id)
  }

  function removeChild$1 (xmlNode, child) {
    const childPos = getChildPos$1(xmlNode, child);
    if (childPos >= 0) {
      removeAt$1(xmlNode, childPos);
    } else {
      throw new Error(`Node ${child.id} is not a child of ${xmlNode.id}`)
    }
    return xmlNode
  }

  function insertBefore$1 (xmlNode, newChild, ref) {
    if (!ref) {
      appendChild$1(xmlNode, newChild);
    } else {
      let pos = getChildPos$1(xmlNode, ref);
      if (pos < 0) {
        throw new Error(`Node ${newChild.id} is not a child of ${xmlNode.id}`)
      }
      insertAt$1(xmlNode, pos, newChild);
    }
    return xmlNode
  }

  function insertAt$1 (xmlNode, pos, child) {
    const length = xmlNode._childNodes.length;
    if (pos >= 0 && pos <= length) {
      const doc = xmlNode.getDocument();
      doc.update([xmlNode.id, '_childNodes'], { type: 'insert', pos, value: child.id });
    } else {
      throw new Error('Index out of bounds.')
    }
    return xmlNode
  }

  function removeAt$1 (xmlNode, pos) {
    const length = xmlNode._childNodes.length;
    if (pos >= 0 && pos < length) {
      const doc = xmlNode.getDocument();
      doc.update([xmlNode.id, '_childNodes'], { type: 'delete', pos: pos });
    } else {
      throw new Error('Index out of bounds.')
    }
    return xmlNode
  }

  function getInnerXML$1 (xmlNode) {
    if (xmlNode._childNodes) {
      return xmlNode.getChildNodes().map(child => {
        return child.toXML().serialize()
      }).join('')
    } else if (xmlNode.isText()) {
      return xmlNode.toXML().getInnerXML()
    }
    return ''
  }

  function getChildAt$1 (xmlNode, idx) {
    let childId = xmlNode._childNodes[idx];
    if (childId) {
      return xmlNode.getDocument().get(childId)
    }
  }


  function node2element$1 (node) {
    
    let dom = DefaultDOMElement$1.createDocument('xml');
    let el = _node2element$1(dom, node);
    return el
  }

  function _node2element$1 (dom, node) {
    let el;
    switch (node._elementType) {
      case 'text': {
        el = _renderTextNode$1(dom, node);
        break
      }
      case 'element':
      case 'inline-element':
      case 'container': {
        el = _renderElementNode$1(dom, node);
        break
      }
      case 'anchor':
      case 'annotation': {
        el = _createElement$1$1(dom, node);
        break
      }
      case 'external': {
        el = DefaultDOMElement$1.parseSnippet(node.xml, 'xml');
        break
      }
      default:
        throw new Error('Invalid element type.')
    }
    return el
  }

  function _createElement$1$1 (dom, node) {
    let el = dom.createElement(node.type);
    el.attr('id', node.id);
    el.attr(node.attributes);
    return el
  }

  function _renderElementNode$1 (dom, node) {
    let el = _createElement$1$1(dom, node);
    el.append(node.getChildren().map(child => _node2element$1(dom, child)));
    return el
  }

  function _renderTextNode$1 (dom, node) {
    const annos = node.getAnnotations();
    const text = node.getText();
    let el = _createElement$1$1(dom, node);
    if (annos && annos.length > 0) {
      let fragmenter = new Fragmenter$1({
        onText: (context, text) => {
          const node = context.node;
          if (node.isText() || (node.isAnnotation() && !node._isInlineNode)) {
            context.el.append(text);
          }
        },
        onEnter: (fragment) => {
          return {
            el: _node2element$1(dom, fragment.node),
            node: fragment.node
          }
        },
        onExit: (fragment, context, parentContext) => {
          parentContext.el.append(context.el);
        }
      });
      fragmenter.start({ el, node }, text, annos);
    } else {
      el.append(text);
    }
    return el
  }

  function nameWithoutNS$1$1 (name) {
    const idx = name.indexOf(':');
    if (idx > 0) {
      return name.slice(idx + 1)
    } else {
      return name
    }
  }

  class Adapter$1 extends domUtils$1.DomUtils {
    
    isTag () {
      return true
    }

    getChildren (elem) {
      return elem.getChildren() || []
    }

    getAttributeValue (elem, name) {
      return elem.getAttribute(name)
    }

    getAttributes (elem) {
      return ['id', elem.id].concat(map$2(elem.attributes, (val, key) => { return [key, val] }))
    }

    hasAttrib (elem, name) {
      return name === 'id' || elem.attributes.hasOwnProperty(name)
    }

    getName (elem) {
      return elem.type
    }

    getNameWithoutNS (elem) {
      return nameWithoutNS$1$1(this.getName(elem))
    }

    getText (elem) {
      if (elem._elementType === 'text') {
        return elem.getText()
      }
      
      return ''
    }
  }

  var cssSelectAdapter$1 = new Adapter$1();

  class XMLDocumentNode$1 extends DocumentNode$1 {
    _initialize (doc, props) {
      
      
      
      if (props.attributes) {
        delete props.attributes.id;
      }
      super._initialize(doc, props);
    }

    toXML () {
      return node2element$1(this)
    }

    
    findChild (tagName) {
      const children = this.getChildren();
      for (let i = 0; i < children.length; i++) {
        const child = children[i];
        if (child.type === tagName) return child
      }
    }

    find (cssSelector) {
      return cssSelect$1.selectOne(cssSelector, this, { xmlMode: true, adapter: cssSelectAdapter$1 })
    }

    findAll (cssSelector) {
      return cssSelect$1.selectAll(cssSelector, this, { xmlMode: true, adapter: cssSelectAdapter$1 })
    }

    isContainer () {
      return false
    }

    
    isBlock () {
      const parentNode = this.parentNode;
      return (parentNode && parentNode.isContainer())
    }

    get children () {
      return this.getChildren()
    }

    getChildren () {
      
      return this.getChildNodes()
    }

    getChildNodes () {
      if (this._childNodes) {
        return getNodes$1(this.getDocument(), this._childNodes)
      } else {
        return []
      }
    }

    getChildCount () {
      if (this._childNodes) {
        return this._childNodes.length
      } else {
        return 0
      }
    }

    getChildPosition (child) {
      if (!child || !this._childNodes) return -1
      return this._childNodes.indexOf(child.id)
    }

    getChildNodeIterator () {
      return new ArrayIterator$1(this.getChildNodes())
    }

    getFirstChild () {
      if (this._childNodes) {
        return this.getDocument().get(this._childNodes[0])
      }
    }

    getLastChild () {
      if (this._childNodes) {
        return this.getDocument().get(last$1$1(this._childNodes))
      }
    }

    getParent () {
      return this.parentNode
    }

    setParent (parentNode) {
      this.parentNode = parentNode;
    }

    get tagName () {
      return this.type
    }

    
    get parent () {
      return this.parentNode
    }

    setAttribute (name, val) {
      if (name === 'id') {
        throw new Error("'id' is read-only and can not be changed")
      }
      this.getDocument().set([this.id, 'attributes', name], val);
      return this
    }

    getAttribute (name) {
      if (name === 'id') return this.id
      return this.attributes[name]
    }

    removeAttribute (name) {
      if (this.attributes.hasOwnProperty(name)) {
        this.getDocument().set([this.id, 'attributes', name], undefined);
        delete this.attributes[name];
      }
      return this
    }

    getAttributes () {
      return clone$1(this.attributes)
    }

    getElementSchema () {
      return this.getDocument().getElementSchema(this.type)
    }

    serialize () {
      return this.toXML()
    }

    isTextNode () {
      return false
    }

    isElementNode () {
      return false
    }

    getInnerXML () {
      return getInnerXML$1(this)
    }
  }

  XMLDocumentNode$1.prototype.attr = DOMElement$1.prototype.attr;

  XMLDocumentNode$1.schema = {
    attributes: { type: 'object', default: {} }
  };

  class XMLTextElement$1 extends TextNodeMixin$1(XMLDocumentNode$1) {
    getPath () {
      return [this.id, 'content']
    }

    getText () {
      return this.content
    }

    
    getChildren () {
      const annos = this.getAnnotations();
      
      annos.sort(_byStartOffset$1);
      return annos
    }

    setText (text) {
      const doc = this.getDocument();
      const path = this.getPath();
      const oldText = this.getText();
      
      if (oldText.length > 0) {
        doc.update(path, { type: 'delete', start: 0, end: oldText.length });
      }
      doc.update(path, { type: 'insert', start: 0, text });
      return this
    }

    

    getTextContent () {
      return this.getText()
    }

    setTextContent (text) {
      return this.setText(text)
    }

    get textContent () {
      return this.getText()
    }

    set textContent (text) {
      this.setText(text);
    }

    appendChild(child) { 
      
      
      
      
      
      
      
      throw new Error('This is not implemented yet.')
    }

    removeChild(child) { 
      
      throw new Error('This is not implemented yet.')
    }

    
    isElementNode () {
      return true
    }
  }

  XMLTextElement$1.prototype._isXMLTextElement = true;

  XMLTextElement$1.prototype.text = DOMElement$1.prototype.text;

  XMLTextElement$1.prototype._elementType = 'text';

  XMLTextElement$1.isText = true;
  XMLTextElement$1.isBlock = true;

  XMLTextElement$1.type = 'text';

  XMLTextElement$1.schema = {
    content: 'text'
  };

  function _byStartOffset$1 (a, b) {
    return a.start.offset - b.start.offset
  }

  class XMLElementNode$1 extends XMLDocumentNode$1 {
    appendChild (child) {
      return appendChild$1(this, child)
    }

    removeChild (child) {
      return removeChild$1(this, child)
    }

    insertBefore (newChild, ref) {
      return insertBefore$1(this, newChild, ref)
    }

    insertAt (pos, child) {
      return insertAt$1(this, pos, child)
    }

    removeAt (pos) {
      return removeAt$1(this, pos)
    }

    getChildAt (idx) {
      return getChildAt$1(this, idx)
    }

    isElementNode () {
      return true
    }

    
  }

  XMLElementNode$1.prototype.append = DOMElement$1.prototype.append;

  XMLElementNode$1.prototype._elementType = 'element';

  XMLElementNode$1.type = 'element';

  XMLElementNode$1.schema = {
    _childNodes: { type: ['array', 'id'], default: [], owned: true }
  };

  XMLElementNode$1.isBlock = true;

  class XMLAnnotationNode$1 extends AnnotationMixin$1(XMLDocumentNode$1) {
    
    get parentNode () {
      const path = this.start.path;
      const doc = this.getDocument();
      return doc.get(path[0])
    }

    isPropertyAnnotation () {
      return true
    }
  }

  XMLAnnotationNode$1.prototype._elementType = 'annotation';


  XMLAnnotationNode$1.prototype._isPropertyAnnotation = true;

  XMLAnnotationNode$1.type = '@annotation';


  XMLAnnotationNode$1.schema = {};

  class XMLAnchorNode$1 extends XMLDocumentNode$1 {
    
    get parentNode () {
      const path = this.coor.start.path;
      const doc = this.getDocument();
      return doc.get(path[0])
    }
  }

  XMLAnchorNode$1.prototype._elementType = 'anchor';

  XMLAnchorNode$1.type = 'anchor';

  XMLAnchorNode$1.schema = {
    coor: { type: 'coordinate', optional: true }
  };

  class XMLInlineElementNode$1 extends XMLAnnotationNode$1 {
    
    get parentNode () {
      const path = this.start.path;
      if (path[0]) {
        const doc = this.getDocument();
        return doc.get(path[0])
      }
      return this._parentNode
    }

    set parentNode (parent) {
      const path = this.start.path;
      if (path[0]) {
        throw new Error('parent of inline-element is implicitly given')
      }
      this._parentNode = parent;
    }

    getChildAt (idx) {
      return getChildAt$1(this, idx)
    }

    appendChild (child) {
      return appendChild$1(this, child)
    }

    removeChild (child) {
      return removeChild$1(this, child)
    }

    insertBefore (newChild, ref) {
      return insertBefore$1(this, newChild, ref)
    }

    insertAt (pos, child) {
      return insertAt$1(this, pos, child)
    }

    removeAt (pos) {
      return removeAt$1(this, pos)
    }

    getInnerXML () {
      return this.getChildren().map(child => {
        return child.toXML().outerHTML
      }).join('')
    }
  }

  XMLInlineElementNode$1.prototype._elementType = 'inline-element';



  XMLInlineElementNode$1.prototype._isInlineNode = true;
  XMLInlineElementNode$1.isInline = true;

  XMLInlineElementNode$1.type = 'inline-element';

  XMLInlineElementNode$1.schema = {
    _childNodes: { type: ['array', 'id'], default: [], owned: true }
  };

  class XMLExternalNode$1 extends XMLDocumentNode$1 {}

  XMLExternalNode$1.prototype._elementType = 'external';

  XMLExternalNode$1.type = 'external';

  XMLExternalNode$1.schema = {
    xml: { type: 'string', default: '' }
  };

  XMLExternalNode$1.isBlock = true;

  class XMLContainerNode$1 extends ContainerMixin$1(XMLElementNode$1) {
    getContentPath () {
      return [this.id, '_childNodes']
    }

    getContent () {
      return this._childNodes
    }

    isContainer () {
      return true
    }

    appendChild (child) {
      super.show(child.id);
    }
  }

  XMLContainerNode$1.prototype._elementType = 'container';

  XMLContainerNode$1.type = 'container';

  XMLContainerNode$1.schema = {};

  XMLContainerNode$1.isBlock = true;

  function createCountingIdGenerator$1 () {
    var counters = {};
    return function uuid (prefix) {
      if (!counters.hasOwnProperty(prefix)) {
        counters[prefix] = 1;
      }
      var result = [prefix, '-', counters[prefix]++].join('');
      return result
    }
  }

  const WS_LEFT$1 = /^\s+/g;


  const WS_RIGHT$1 = /\s+$/g;
  const WS_ALL$1 = /\s+/g;


  const SPACE$1 = ' ';
  const TABS_OR_NL$1 = /[\t\n\r]+/g;

  const INVISIBLE_CHARACTER$1 = '\u200B';


  class DOMImporter$1 {
    constructor (config, context) {
      this.context = context || {};

      if (!config.schema) {
        throw new Error('"config.schema" is mandatory')
      }
      if (!config.converters) {
        throw new Error('"config.converters" is mandatory')
      }

      this.config = Object.assign({ idAttribute: 'id' }, config);
      this.schema = config.schema;
      this.converters = config.converters;
      this.state = null;

      this._defaultBlockConverter = null;
      this._allConverters = [];
      this._blockConverters = [];
      this._propertyAnnotationConverters = [];

      this.state = new DOMImporter$1.State();

      this._initialize();
    }

    
    _initialize () {
      const schema = this.schema;
      const defaultTextType = schema.getDefaultTextType();
      const converters = this.converters;
      for (let i = 0; i < converters.length; i++) {
        let converter;
        if (typeof converters[i] === 'function') {
          const Converter = converters[i];
          converter = new Converter();
        } else {
          converter = converters[i];
        }
        if (!converter.type) {
          throw new Error('Converter must provide the type of the associated node.')
        }
        if (!converter.matchElement && !converter.tagName) {
          throw new Error('Converter must provide a matchElement function or a tagName property.')
        }
        if (!converter.matchElement) {
          converter.matchElement = this._defaultElementMatcher.bind(converter);
        }
        const NodeClass = schema.getNodeClass(converter.type);
        if (!NodeClass) {
          throw new Error('No node type defined for converter')
        }
        if (!this._defaultBlockConverter && defaultTextType === converter.type) {
          this._defaultBlockConverter = converter;
        }
        this._allConverters.push(converter);
        
        if (NodeClass.prototype._isPropertyAnnotation) {
          this._propertyAnnotationConverters.push(converter);
        } else {
          this._blockConverters.push(converter);
        }
      }
    }

    dispose () {
      if (this.state.doc) {
        this.state.doc.dispose();
      }
    }

    
    reset () {
      if (this.state.doc) {
        this.state.doc.dispose();
      }
      this.state.reset();
      this.state.doc = this._createDocument();
    }

    getDocument () {
      return this.state.doc
    }

    
    convertContainer (elements, containerId) {
      if (!this.state.doc) this.reset();
      const state = this.state;
      const iterator = new ArrayIterator$1(elements);
      const nodeIds = [];
      while (iterator.hasNext()) {
        const el = iterator.next();
        let node;
        const blockTypeConverter = this._getConverterForElement(el, 'block');
        if (blockTypeConverter) {
          state.pushContext(el.tagName, blockTypeConverter);
          let nodeData = this._createNodeData(el, blockTypeConverter.type);
          nodeData = blockTypeConverter.import(el, nodeData, this) || nodeData;
          node = this._createNode(nodeData);
          let context = state.popContext();
          context.annos.forEach((a) => {
            this._createNode(a);
          });
        } else if (el.isCommentNode()) {
          continue
        } else {
          
          if (el.isTextNode() && /^\s*$/.exec(el.textContent)) continue
          
          
          iterator.back();
          node = this._wrapInlineElementsIntoBlockElement(iterator);
        }
        if (node) {
          nodeIds.push(node.id);
        }
      }
      return this._createNode({
        type: '@container',
        id: containerId,
        nodes: nodeIds
      })
    }

    
    convertElement (el) {
      if (!this.state.doc) this.reset();
      let isTopLevel = !this.state.isConverting;
      if (isTopLevel) {
        this.state.isConverting = true;
      }

      let nodeData, annos;
      const converter = this._getConverterForElement(el);
      if (converter) {
        const NodeClass = this.schema.getNodeClass(converter.type);
        nodeData = this._createNodeData(el, converter.type);
        this.state.pushContext(el.tagName, converter);
        
        
        
        
        
        if (NodeClass.isInline) {
          nodeData = this._convertInlineNode(el, nodeData, converter);
        } else if (NodeClass.prototype._isPropertyAnnotation) {
          nodeData = this._convertPropertyAnnotation(el, nodeData);
        } else {
          nodeData = converter.import(el, nodeData, this) || nodeData;
        }
        let context = this.state.popContext();
        annos = context.annos;
      } else {
        throw new Error('No converter found for ' + el.tagName)
      }
      
      const node = this._createNode(nodeData);
      
      annos.forEach((a) => {
        this._createNode(a);
      });

      
      
      if (this.config['stand-alone'] && isTopLevel) {
        this.state.isConverting = false;
        this.reset();
      }
      return node
    }

    
    annotatedText (el, path, options = {}) {
      if (!path) {
        throw new Error('path is mandatory')
      }
      const state = this.state;
      const context = last$1$1(state.contexts);
      
      
      if (!context) {
        throw new Error('This should be called from within an element converter.')
      }
      
      const oldPreserveWhitespace = state.preserveWhitespace;
      if (options.preserveWhitespace) {
        state.preserveWhitespace = true;
      }
      state.stack.push({ path: path, offset: 0, text: '', annos: [] });
      
      
      
      this.state.lastChar = '';
      const iterator = this.getChildNodeIterator(el);
      const text = this._annotatedText(iterator);
      
      
      const top = state.stack.pop();
      context.annos = context.annos.concat(top.annos);

      
      state.preserveWhitespace = oldPreserveWhitespace;

      return text
    }

    
    plainText (el) {
      var state = this.state;
      var text = el.textContent;
      if (state.stack.length > 0) {
        var context = last$1$1(state.stack);
        context.offset += text.length;
        context.text += context.text.concat(text);
      }
      return text
    }

    
    _customText (text) {
      var state = this.state;
      if (state.stack.length > 0) {
        var context = last$1$1(state.stack);
        context.offset += text.length;
        context.text += context.text.concat(text);
      }
      return text
    }

    
    nextId (prefix) {
      
      
      
      
      return this.state.uuid(prefix)
    }

    _getNextId (dom, type) {
      let id = this.nextId(type);
      while (this.state.ids[id] || dom.find('#' + id)) {
        id = this.nextId(type);
      }
      return id
    }

    _getIdForElement (el, type) {
      let id = el.getAttribute(this.config.idAttribute);
      if (id && !this.state.ids[id]) return id
      return this._getNextId(el.getOwnerDocument(), type)
    }

    
    
    _createDocument () {
      
      const schema = this.config.schema;
      const DocumentClass = schema.getDocumentClass();
      return new DocumentClass(schema)
    }

    _convertPropertyAnnotation (el, nodeData) {
      const path = [nodeData.id, '_content'];
      
      
      
      nodeData._content = this.annotatedText(el, path);
      nodeData.start = { path, offset: 0 };
      nodeData.end = { offset: nodeData._content.length };
      return nodeData
    }

    _convertInlineNode (el, nodeData, converter) {
      const path = [nodeData.id, '_content'];
      if (converter.import) {
        nodeData = converter.import(el, nodeData, this) || nodeData;
      }
      nodeData._content = '$';
      nodeData.start = { path, offset: 0 };
      nodeData.end = { offset: 1 };
      return nodeData
    }

    _createNodeData (el, type) {
      if (!type) {
        throw new Error('type is mandatory.')
      }
      let nodeData = {
        type,
        id: this._getIdForElement(el, type)
      };
      this.state.ids[nodeData.id] = true;
      return nodeData
    }

    _createNode (nodeData) {
      let doc = this.state.doc;
      
      
      
      let node = doc.get(nodeData.id);
      if (node) {
        
        doc.delete(node.id);
      }
      return doc.create(nodeData)
    }

    getChildNodeIterator (el) {
      return el.getChildNodeIterator()
    }

    _defaultElementMatcher (el) {
      return el.is(this.tagName)
    }

    
    _annotatedText (iterator) {
      const state = this.state;
      const context = last$1$1(state.stack);
      
      if (!context) {
        throw new Error('Illegal state: context is null.')
      }
      while (iterator.hasNext()) {
        var el = iterator.next();
        var text = '';
        
        
        if (el.isTextNode()) {
          text = this._prepareText(el.textContent);
          if (text.length) {
            
            
            context.text = context.text.concat(text);
            context.offset += text.length;
          }
        } else if (el.isCommentNode()) {
          
          continue
        } else if (el.isElementNode()) {
          const annoConverter = this._getConverterForElement(el, 'inline');
          
          if (!annoConverter) {
            
            if (!this.IGNORE_DEFAULT_WARNINGS) {
              console.warn('Unsupported inline element. We will not create an annotation for it, but process its children to extract annotated text.', el.outerHTML);
            }
            
            
            const iterator = this.getChildNodeIterator(el);
            this._annotatedText(iterator);
            continue
          }
          
          
          
          var startOffset = context.offset;
          const annoType = annoConverter.type;
          const AnnoClass = this.schema.getNodeClass(annoType);
          let annoData = this._createNodeData(el, annoType);
          
          let stackFrame = {
            path: context.path,
            offset: startOffset,
            text: '',
            annos: []
          };
          state.stack.push(stackFrame);
          
          if (annoConverter.import) {
            state.pushContext(el.tagName, annoConverter);
            annoData = annoConverter.import(el, annoData, this) || annoData;
            state.popContext();
          }
          
          
          
          
          if (AnnoClass.isInline) {
            this._customText(INVISIBLE_CHARACTER$1);
            
            
            
            state.lastChar = '';
          } else {
            
            
            
            const iterator = this.getChildNodeIterator(el);
            this._annotatedText(iterator);
          }
          
          state.stack.pop();
          context.offset = stackFrame.offset;
          context.text = context.text.concat(stackFrame.text);
          
          const endOffset = context.offset;
          annoData.start = {
            path: context.path.slice(0),
            offset: startOffset
          };
          annoData.end = {
            offset: endOffset
          };
          
          let parentFrame = last$1$1(state.stack);
          parentFrame.annos = parentFrame.annos.concat(stackFrame.annos, annoData);
        } else {
          console.warn('Unknown element type. Taking plain text.', el.outerHTML);
          text = this._prepareText(el.textContent);
          context.text = context.text.concat(text);
          context.offset += text.length;
        }
      }
      
      return context.text
    }

    _getConverterForElement (el, mode) {
      var converters;
      if (mode === 'block') {
        if (!el.tagName) return null
        converters = this._blockConverters;
      } else if (mode === 'inline') {
        converters = this._propertyAnnotationConverters;
      } else {
        converters = this._allConverters;
      }
      var converter = null;
      for (var i = 0; i < converters.length; i++) {
        if (this._converterCanBeApplied(converters[i], el)) {
          converter = converters[i];
          break
        }
      }
      return converter
    }

    _converterCanBeApplied (converter, el) {
      return converter.matchElement(el, this)
    }

    
    _wrapInlineElementsIntoBlockElement (childIterator) {
      if (!childIterator.hasNext()) return

      const converter = this._defaultBlockConverter;
      if (!converter) {
        throw new Error('Wrapping inline elements automatically is not supported in this schema.')
      }

      let dom = childIterator.peek().getOwnerDocument();
      let wrapper = dom.createElement('wrapper');
      while (childIterator.hasNext()) {
        const el = childIterator.next();
        
        const blockTypeConverter = this._getConverterForElement(el, 'block');
        if (blockTypeConverter) {
          childIterator.back();
          break
        }
        wrapper.append(el.clone());
      }
      const type = this.schema.getDefaultTextType();
      const id = this._getNextId(dom, type);
      let nodeData = { type, id };
      this.state.pushContext('wrapper', converter);
      nodeData = converter.import(wrapper, nodeData, this) || nodeData;
      let context = this.state.popContext();
      let annos = context.annos;
      
      const node = this._createNode(nodeData);
      
      annos.forEach((a) => {
        this._createNode(a);
      });
      return node
    }

    
    
    
    
    _prepareText (text) {
      const state = this.state;
      if (state.preserveWhitespace) {
        return text
      }
      var repl = SPACE$1;
      
      text = text.replace(TABS_OR_NL$1, '');
      
      
      
      
      
      if (state.lastChar === SPACE$1) {
        
        text = text.replace(WS_LEFT$1, '');
      } else {
        text = text.replace(WS_LEFT$1, repl);
      }
      text = text.replace(WS_RIGHT$1, repl);
      
      
      
      if (this.config.REMOVE_INNER_WS || state.removeInnerWhitespace) {
        text = text.replace(WS_ALL$1, SPACE$1);
      }
      state.lastChar = text[text.length - 1] || state.lastChar;
      return text
    }

    
    _trimTextContent (el) {
      var nodes = el.getChildNodes();
      var firstNode = nodes[0];
      var lastNode = last$1$1(nodes);
      var text, trimmed;
      
      if (firstNode && firstNode.isTextNode()) {
        text = firstNode.textContent;
        trimmed = this._trimLeft(text);
        firstNode.textContent = trimmed;
      }
      if (lastNode && lastNode.isTextNode()) {
        text = lastNode.textContent;
        trimmed = this._trimRight(text);
        lastNode.textContent = trimmed;
      }
      return el
    }

    _trimLeft (text) {
      return text.replace(WS_LEFT$1, '')
    }

    _trimRight (text) {
      return text.replace(WS_RIGHT$1, '')
    }
  }

  class DOMImporterState$1 {
    constructor () {
      this.reset();
    }

    reset () {
      this.preserveWhitespace = false;
      this.nodes = [];
      this.annotations = [];
      this.containerId = null;
      this.container = [];
      this.ids = {};
      
      this.contexts = [];
      
      this.stack = [];
      this.lastChar = '';
      this.skipTypes = {};
      this.ignoreAnnotations = false;
      this.isConverting = false;

      
      
      this.uuid = createCountingIdGenerator$1();
    }

    pushContext (tagName, converter) {
      this.contexts.push({ tagName: tagName, converter: converter, annos: [] });
    }

    popContext () {
      return this.contexts.pop()
    }

    getCurrentContext () {
      return last$1$1(this.contexts)
    }
  }

  DOMImporter$1.State = DOMImporterState$1;

  DOMImporter$1.INVISIBLE_CHARACTER = INVISIBLE_CHARACTER$1;

  const START$3 = 'START';
  const END$3 = 'END';
  const EPSILON$3 = 'EPSILON';
  const TEXT$1$1 = 'TEXT';

  class DFA$1 {
    constructor (transitions) {
      if (!transitions || Object.keys(transitions).length === 0) {
        transitions = { START: { EPSILON: END$3 } };
      }
      this.transitions = transitions;
    }

    consume (state, id) {
      const T = this.transitions;
      
      
      if (!T[state]) return -1
      let nextState = T[state][id];
      if (nextState !== undefined) {
        return nextState
      }
      while (T[state][EPSILON$3] !== undefined) {
        state = T[state][EPSILON$3];
        if (state === END$3) {
          return -1
        }
        nextState = T[state][id];
        if (nextState !== undefined) {
          return nextState
        }
      }
      return -1
    }

    canConsume (state, id) {
      let nextState = this.consume(state, id);
      return (nextState !== -1)
    }

    isFinished (state) {
      const T = this.transitions;
      if (state === 'END') return true
      
      if (!T[state]) return false
      while (T[state][EPSILON$3] !== undefined) {
        state = T[state][EPSILON$3];
        if (state === 'END') return true
      }
      return false
    }

    

    
    _tokensByPath () {
      const result = [];
      const transitions = this.transitions;
      if (!transitions) return []

      
      let first = {};
      forEach$1(transitions[START$3], (to, token) => {
        if (!first[to]) first[to] = [];
        first[to].push(token);
      });

      let visited = {START: true, END: true};
      forEach$1(first, (tokens, state) => {
        
        
        
        
        let _siblings = {};
        tokens.forEach((t) => {
          if (t !== EPSILON$3) {
            _siblings[t] = true;
          }
        });
        let stack = [state];
        while (stack.length > 0) {
          let from = stack.pop();
          if (state === END$3) continue
          visited[from] = true;
          let T = transitions[from];
          if (!T) throw new Error(`Internal Error: no transition from state ${from}`)
          let tokens = Object.keys(T);
          for (let i = 0; i < tokens.length; i++) {
            const token = tokens[i];
            const to = T[token];
            if (!visited[to]) stack.push(to);
            if (token !== EPSILON$3) {
              _siblings[token] = true;
            }
          }
        }
        let _siblingTokens = Object.keys(_siblings);
        if (_siblingTokens.length > 0) {
          result.push(_siblingTokens);
        }
      });
      return result
    }
  }

  DFA$1.START = START$3;
  DFA$1.END = END$3;
  DFA$1.EPSILON = EPSILON$3;
  DFA$1.TEXT = TEXT$1$1;

  function _isTextNodeEmpty$1 (el) {
    return Boolean(/^\s*$/.exec(el.textContent))
  }

  const START$1$1 = DFA$1.START;
  const END$1$1 = DFA$1.END;
  const EPSILON$1$1 = DFA$1.EPSILON;


  class DFABuilder$1 {
    constructor (transitions) {
      this.transitions = transitions;
    }

    addTransition (from, to, tokens) {
      if (!this.transitions) this.transitions = {};
      if (!isArray$2(tokens)) tokens = [tokens];
      tokens.forEach(token => _addTransition$1(this.transitions, from, to, token));
      return this
    }

    
    append (other) {
      if (this.transitions && other.transitions) {
        let t1 = cloneDeep$1(this.transitions);
        let t2 = cloneDeep$1(other.transitions);
        
        
        let firstIsOptional = Boolean(t1[START$1$1][EPSILON$1$1]);
        let secondIsOptional = Boolean(t2[START$1$1][EPSILON$1$1]);

        if (firstIsOptional) {
          
          
          delete t1[START$1$1][EPSILON$1$1];
        }
        
        
        let newState = uuid$1();
        
        
        forEach$1(t1, (T) => {
          forEach$1(T, (to, token) => {
            if (to === END$1$1) {
              T[token] = newState;
            }
          });
        });
        
        
        if (firstIsOptional) {
          forEach$1(t2[START$1$1], (to, token) => {
            _addTransition$1(t1, START$1$1, to, token);
          });
        }
        
        
        t2[newState] = t2[START$1$1];
        forEach$1(t2, (T) => {
          forEach$1(T, (to, token) => {
            if (to === START$1$1) {
              T[token] = newState;
            }
          });
        });
        delete t2[START$1$1];
        
        forEach$1(t2, (T, from) => {
          forEach$1(T, (to, token) => {
            _addTransition$1(t1, from, to, token);
          });
        });
        
        
        if (firstIsOptional && secondIsOptional) {
          _addTransition$1(t1, START$1$1, END$1$1, EPSILON$1$1);
        }
        this.transitions = t1;
      } else if (other.transitions) {
        this.transitions = cloneDeep$1(other.transitions);
      }
      return this
    }

    
    merge (other) {
      if (this.transitions && other.transitions) {
        let t1 = this.transitions;
        let t2 = other.transitions;
        forEach$1(t2, (T, from) => {
          forEach$1(T, (to, token) => {
            _addTransition$1(t1, from, to, token);
          });
        });
      } else if (other.transitions) {
        this.transitions = cloneDeep$1(other.transitions);
      }
      return this
    }

    
    optional () {
      let dfa = new DFABuilder$1(cloneDeep$1(this.transitions));
      if (this.transitions) {
        dfa.addTransition(START$1$1, END$1$1, EPSILON$1$1);
      }
      return dfa
    }

    
    kleene () {
      let dfa = this.plus();
      return dfa.optional()
    }

    
    plus () {
      let dfa;
      if (this.transitions) {
        let t1 = cloneDeep$1(this.transitions);
        
        
        
        
        const isOptional = Boolean(t1[START$1$1][EPSILON$1$1]);
        delete t1[START$1$1][EPSILON$1$1];
        
        
        let newState = uuid$1();
        forEach$1(t1, (T) => {
          forEach$1(T, (to, token) => {
            if (to === END$1$1) {
              T[token] = newState;
            }
          });
        });
        
        _addTransition$1(t1, newState, END$1$1, EPSILON$1$1);
        
        forEach$1(t1[START$1$1], (to, token) => {
          _addTransition$1(t1, newState, to, token);
        });
        
        if (isOptional) {
          _addTransition$1(t1, START$1$1, END$1$1, EPSILON$1$1);
        }
        dfa = new DFABuilder$1(t1);
      } else {
        dfa = new DFABuilder$1(cloneDeep$1(this.transitions));
      }
      return dfa
    }

    toJSON () {
      return cloneDeep$1(this.transitions)
    }

    
    copy () {
      let t = cloneDeep$1(this.transitions);
      if (this.transitions) {
        let states = Object.keys(t);
        let map = { START: START$1$1, END: END$1$1 };
        states.forEach((id) => {
          if (id === START$1$1 || id === END$1$1) return
          map[id] = uuid$1();
        });
        forEach$1(t, (T, from) => {
          if (from !== START$1$1 && from !== END$1$1) {
            t[map[from]] = T;
            delete t[from];
          }
          forEach$1(T, (to, token) => {
            if (to !== START$1$1 && to !== END$1$1) {
              T[token] = map[to];
            }
          });
        });
      }
      return new DFABuilder$1(t)
    }
  }

  DFABuilder$1.singleToken = function (token) {
    let dfa = new DFABuilder$1();
    dfa.addTransition(START$1$1, END$1$1, token);
    return dfa
  };

  function _addTransition$1 (transitions, from, to, token) {
    let T = transitions[from];
    if (!T) {
      transitions[from] = T = {};
    }
    if (token === EPSILON$1$1 && from === START$1$1 && to !== END$1$1) {
      throw new Error('The only EPSILON transition from START must be START->END')
    }
    if (T[token] && T[token] !== to) {
      console.error('Token %s already used. Ignoring this transition.', token);
      return
      
    }
    T[token] = to;
  }

  const { START: START$2$1, END: END$2$1, TEXT: TEXT$3$1, EPSILON: EPSILON$2$1 } = DFA$1;



  class Expression$1 {
    constructor (name, root) {
      this.name = name;
      this.root = root;

      this._initialize();
    }

    _initialize () {
      this._compile();
    }

    toString () {
      return this.root.toString()
    }

    copy () {
      return this.root.copy()
    }

    toJSON () {
      return {
        name: this.name,
        content: this.root.toJSON()
      }
    }

    isAllowed (tagName) {
      return Boolean(this._allowedChildren[tagName])
    }

    
    _normalize () {
      this.root._normalize();
    }

    
    _compile () {
      
      this.root._compile();
    }

    _describeError (state, token) {
      let msg = [];
      if (token !== TEXT$3$1) {
        if (!this.isAllowed(token)) {
          msg.push(`<${token}> is not valid in <${this.name}>\nSchema: ${this.toString()}`);
        } else {
          
          msg.push(`<${token}> is not allowed at the current position in <${this.name}>.\n${this.toString()}`);
          
          
          
        }
      } else {
        msg.push(`TEXT is not allowed at the current position: ${state.trace.join(',')}\n${this.toString()}`);
      }
      return msg.join('')
    }
  }

  Expression$1.fromJSON = function (data) {
    const name = data.name;
    const root = _fromJSON$1(data.content);
    return createExpression$1(name, root)
  };

  function createExpression$1 (name, root) {
    if (root instanceof Interleave$1) {
      return new InterleaveExpr$1(name, root)
    } else {
      return new DFAExpr$1(name, root)
    }
  }

  class DFAExpr$1 extends Expression$1 {
    
    getInitialState () {
      return {
        dfaState: START$2$1,
        errors: [],
        trace: []
      }
    }

    canConsume (state, token) {
      return this.dfa.canConsume(state.dfaState, token)
    }

    consume (state, token) {
      const dfa = this.dfa;
      let oldState = state.dfaState;
      let newState = dfa.consume(oldState, token);
      state.dfaState = newState;
      if (newState === -1) {
        state.errors.push({
          msg: this._describeError(state, token),
          
          
          el: state.el
        });
        return false
      } else {
        state.trace.push(token);
        return true
      }
    }

    isFinished (state) {
      return this.dfa.isFinished(state.dfaState)
    }

    _initialize () {
      super._initialize();

      this._computeAllowedChildren();
    }

    _compile () {
      super._compile();
      this.dfa = new DFA$1(this.root.dfa.transitions);
    }

    _computeAllowedChildren () {
      this._allowedChildren = _collectAllTokensFromDFA$1(this.dfa);
    }

    _findInsertPos (el, newTag, mode) {
      const root = this.root;
      if (root instanceof Sequence$1) {
        return this._findInsertPosInSequence(el, newTag, mode)
      } else if (root instanceof Plus$1 || root instanceof Kleene$1) {
        if (mode === 'first') {
          return 0
        } else {
          return el.childNodes.length
        }
      }
    }

    _isValid (_tokens) {
      let state = this.getInitialState();
      for (let i = 0; i < _tokens.length; i++) {
        const token = _tokens[i];
        
        
        
        if (!token) continue
        if (!this.consume(state, token)) {
          return false
        }
      }
      return this.isFinished(state)
    }

    _findInsertPosInSequence (el, newTag, mode) {
      const childNodes = el.getChildNodes();
      
      
      
      const tokens = [];
      childNodes.forEach((child) => {
        
        
        
        const tagName = child.tagName;
        if (!tagName) {
          if (child._isDOMElement && child.isTextNode() && !_isTextNodeEmpty$1(child)) {
            tokens.push(TEXT$3$1);
          } else {
            tokens.push(null);
          }
        } else {
          tokens.push(tagName);
        }
      });
      const L = tokens.length;
      const self = this;
      function _isValid (pos) {
        let _tokens = tokens.slice(0);
        _tokens.splice(pos, 0, newTag);
        return self._isValid(_tokens)
      }
      if (mode === 'first') {
        for (let pos = 0; pos <= L; pos++) {
          if (_isValid(pos)) {
            return pos
          }
        }
      } else {
        for (let pos = L; pos >= 0; pos--) {
          if (_isValid(pos)) {
            return pos
          }
        }
      }
      return -1
    }
  }

  function _collectAllTokensFromDFA$1 (dfa) {
    
    const children = {};
    if (dfa.transitions) {
      forEach$1(dfa.transitions, (T) => {
        Object.keys(T).forEach((tagName) => {
          if (tagName === EPSILON$2$1) return
          children[tagName] = true;
        });
      });
    }
    return children
  }

  class InterleaveExpr$1 extends Expression$1 {
    getInitialState () {
      const dfas = this.dfas;
      const dfaStates = new Array(dfas.length);
      dfaStates.fill(START$2$1);
      return {
        dfaStates,
        errors: [],
        trace: [],
        
        lastDFA: 0
      }
    }

    canConsume (state, token) {
      return (this._findNextDFA(state, token) >= 0)
    }

    consume (state, token) {
      const idx = this._findNextDFA(state, token);
      if (idx < 0) {
        state.errors.push({
          msg: this._describeError(state, token)
        });
        return false
      } else {
        const dfa = this.dfas[idx];
        const oldState = state.dfaStates[idx];
        const newState = dfa.consume(oldState, token);
        state.dfaStates[idx] = newState;
        state.trace.push(token);
        return true
      }
    }

    isFinished (state) {
      const dfas = this.dfas;
      for (let i = 0; i < dfas.length; i++) {
        const dfa = dfas[i];
        const dfaState = state.dfaStates[i];
        if (!dfa.isFinished(dfaState)) {
          return false
        }
      }
      return true
    }

    _initialize () {
      super._initialize();

      this._computeAllowedChildren();
    }

    _compile () {
      super._compile();

      this.blocks = this.root.blocks;
      this.dfas = this.blocks.map(b => new DFA$1(b.dfa.transitions));
    }

    _computeAllowedChildren () {
      this._allowedChildren = Object.assign(...this.blocks.map((block) => {
        return _collectAllTokensFromDFA$1(block.dfa)
      }));
    }

    _findNextDFA (state, token) {
      console.assert(state.dfaStates.length === this.dfas.length);
      const dfas = this.dfas;
      for (let i = 0; i < state.dfaStates.length; i++) {
        const dfa = dfas[i];
        const dfaState = state.dfaStates[i];
        if (dfa.canConsume(dfaState, token)) {
          return i
        }
      }
      return -1
    }

    _findInsertPos(el, newTag, mode) { 
      
      return el.childNodes.length
    }
  }

  class Token$1 {
    constructor (name) {
      this.name = name;
    }

    toString () {
      return this.name
    }

    toJSON () {
      return this.name
    }

    copy () {
      return new Token$1(this.name)
    }

    _normalize () {}

    _compile () {
      this.dfa = DFABuilder$1.singleToken(this.name);
    }
  }

  Token$1.fromJSON = function (data) {
    return new Token$1(data)
  };


  class Choice$1 {
    constructor (blocks) {
      this.blocks = blocks;
    }

    copy () {
      return new Choice$1(this.blocks.map(b => b.copy()))
    }

    toJSON () {
      return {
        type: '|',
        blocks: this.blocks.map(b => b.toJSON())
      }
    }

    _normalize () {
      const blocks = this.blocks;
      for (let i = blocks.length - 1; i >= 0; i--) {
        let block = blocks[i];
        block._normalize();
        
        if (block instanceof Choice$1) {
          blocks.splice(i, 1, ...(block.blocks));
        }
      }
    }

    _compile () {
      let dfa = new DFABuilder$1();
      this.blocks.forEach((block) => {
        if (block instanceof Token$1) {
          dfa.addTransition(START$2$1, END$2$1, block.name);
        } else if (block instanceof Interleave$1) {
          throw new Error('Nested interleave blocks are not supported.')
        } else {
          if (!block.dfa) {
            block._compile();
          }
          dfa.merge(block.dfa);
        }
      });
      this.dfa = dfa;
      return dfa
    }

    toString () {
      return '(' + this.blocks.map(b => b.toString()).join('|') + ')'
    }
  }

  Choice$1.fromJSON = function (data) {
    return new Choice$1(data.blocks.map((block) => {
      return _fromJSON$1(block)
    }))
  };


  class Sequence$1 {
    constructor (blocks) {
      this.blocks = blocks;
    }

    copy () {
      return new Sequence$1(this.blocks.map(b => b.copy()))
    }

    toJSON () {
      return {
        type: ',',
        blocks: this.blocks.map(b => b.toJSON())
      }
    }

    _compile () {
      let dfa = new DFABuilder$1();
      this.blocks.forEach((block) => {
        if (block instanceof Token$1) {
          dfa.append(DFABuilder$1.singleToken(block.name));
        } else if (block instanceof Interleave$1) {
          throw new Error('Nested interleave blocks are not supported.')
        } else {
          if (!block.dfa) {
            block._compile();
          }
          dfa.append(block.dfa);
        }
      });
      this.dfa = dfa;
      return dfa
    }

    _normalize () {
      const blocks = this.blocks;
      for (let i = blocks.length - 1; i >= 0; i--) {
        let block = blocks[i];
        block._normalize();
        
        if (block instanceof Sequence$1) {
          blocks.splice(i, 1, ...(block.blocks));
        }
      }
    }

    toString () {
      return '(' + this.blocks.map(b => b.toString()).join(',') + ')'
    }
  }

  Sequence$1.fromJSON = function (data) {
    return new Sequence$1(data.blocks.map((block) => {
      return _fromJSON$1(block)
    }))
  };


  class Interleave$1 {
    constructor (blocks) {
      this.blocks = blocks;
    }

    copy () {
      return new Interleave$1(this.blocks.map(b => b.copy()))
    }

    toString () {
      return '(' + this.blocks.map(b => b.toString()).join('~') + ')'
    }

    toJSON () {
      return {
        type: '~',
        blocks: this.blocks.map(b => b.toJSON())
      }
    }

    _normalize () {
      
    }

    _compile () {
      this.blocks.forEach(block => block._compile());
    }
  }

  Interleave$1.fromJSON = function (data) {
    return new Interleave$1(data.blocks.map((block) => {
      return _fromJSON$1(block)
    }))
  };


  class Optional$1 {
    constructor (block) {
      this.block = block;
    }

    copy () {
      return new Optional$1(this.block.copy())
    }

    toJSON () {
      return {
        type: '?',
        block: this.block.toJSON()
      }
    }

    _compile () {
      const block = this.block;
      if (block instanceof Interleave$1) {
        throw new Error('Nested interleave blocks are not supported.')
      }
      if (!block.dfa) {
        block._compile();
      }
      this.dfa = block.dfa.optional();
      return this.dfa
    }

    _normalize () {
      const block = this.block;
      block._normalize();
      if (block instanceof Optional$1) {
        this.block = block.block;
      } else if (block instanceof Kleene$1) {
        console.error('FIXME -  <optional> is useless here', this.toString());
      }
    }

    toString () {
      return this.block.toString() + '?'
    }
  }

  Optional$1.fromJSON = function (data) {
    return new Optional$1(_fromJSON$1(data.block))
  };


  class Kleene$1 {
    constructor (block) {
      this.block = block;
    }

    copy () {
      return new Kleene$1(this.block.copy())
    }

    toJSON () {
      return {
        type: '*',
        block: this.block.toJSON()
      }
    }

    _compile () {
      const block = this.block;
      if (block instanceof Interleave$1) {
        throw new Error('Nested interleave blocks are not supported.')
      }
      if (!block.dfa) {
        block._compile();
      }
      this.dfa = block.dfa.kleene();
      return this.dfa
    }

    _normalize () {
      const block = this.block;
      block._normalize();
      if (block instanceof Optional$1 || block instanceof Kleene$1) {
        this.block = block.block;
      } else if (block instanceof Plus$1) {
        throw new Error('This does not make sense:' + this.toString())
      }
    }

    toString () {
      return this.block.toString() + '*'
    }
  }

  Kleene$1.fromJSON = function (data) {
    return new Kleene$1(_fromJSON$1(data.block))
  };


  class Plus$1 {
    constructor (block) {
      this.block = block;
    }

    copy () {
      return new Plus$1(this.block.copy())
    }

    toJSON () {
      return {
        type: '+',
        block: this.block.toJSON()
      }
    }

    _compile () {
      const block = this.block;
      if (block instanceof Interleave$1) {
        throw new Error('Nested interleave blocks are not supported.')
      }
      if (!block.dfa) {
        block._compile();
      }
      this.dfa = block.dfa.plus();
      return this.dfa
    }

    _normalize () {
      const block = this.block;
      block._normalize();
      if (block instanceof Optional$1 || block instanceof Kleene$1) {
        throw new Error('This does not make sense:' + this.toString())
      } else if (block instanceof Plus$1) {
        this.block = block.block;
      }
    }

    toString () {
      return this.block.toString() + '+'
    }
  }

  Plus$1.fromJSON = function (data) {
    return new Plus$1(_fromJSON$1(data.block))
  };

  function _fromJSON$1 (data) {
    switch (data.type) {
      case ',':
        return Sequence$1.fromJSON(data)
      case '~':
        return Interleave$1.fromJSON(data)
      case '|':
        return Choice$1.fromJSON(data)
      case '?':
        return Optional$1.fromJSON(data)
      case '+':
        return Plus$1.fromJSON(data)
      case '*':
        return Kleene$1.fromJSON(data)
      default:
        if (isString$1(data)) {
          return new Token$1(data)
        }
        throw new Error('Unsupported data.')
    }
  }

  const { TEXT: TEXT$4$1 } = DFA$1;

  class XMLSchema$1 {
    constructor (elementSchemas, startElement) {
      if (!elementSchemas[startElement]) {
        throw new Error('startElement must be a valid element.')
      }
      this._elementSchemas = {};
      this.startElement = startElement;
      
      forEach$1(elementSchemas, (spec, name) => {
        this._elementSchemas[name] = new ElementSchema$1(spec.name, spec.type, spec.attributes, spec.expr);
      });
    }

    getIdAttribute () {
      return 'id'
    }

    getTagNames () {
      return Object.keys(this._elementSchemas)
    }

    getElementSchema (name) {
      return this._elementSchemas[name]
    }

    getStartElement () {
      return this.startElement
    }

    toJSON () {
      let result = {
        start: this.getStartElement(),
        elements: {}
      };
      forEach$1(this._elementSchemas, (schema, name) => {
        result.elements[name] = schema.toJSON();
      });
      return result
    }

    
    toMD () {
      let result = [];
      let elementNames = Object.keys(this._elementSchemas);
      elementNames.sort();
      elementNames.forEach((name) => {
        let elementSchema = this._elementSchemas[name];
        result.push(`# <${elementSchema.name}>`);
        result.push('');
        result.push(`type: ${elementSchema.type}`);
        result.push('attributes: ' + map$2(elementSchema.attributes, (_, name) => { return name }).join(', '));
        result.push('children:');
        result.push('  ' + elementSchema.expr.toString());
        result.push('');
      });
      return result.join('\n')
    }

    validateElement (el) {
      let tagName = el.tagName;
      let elementSchema = this.getElementSchema(tagName);
      return _validateElement$1(elementSchema, el)
    }
  }

  XMLSchema$1.fromJSON = function (data) {
    let elementSchemas = {};
    forEach$1(data.elements, (elData) => {
      let elSchema = ElementSchema$1.fromJSON(elData);
      elementSchemas[elSchema.name] = elSchema;
    });
    return new XMLSchema$1(elementSchemas, data.start)
  };

  class ElementSchema$1 {
    constructor (name, type, attributes, expr) {
      this.name = name;
      this.type = type;
      this.attributes = attributes;
      this.expr = expr;

      if (!name) {
        throw new Error("'name' is mandatory")
      }
      if (!type) {
        throw new Error("'type' is mandatory")
      }
      if (!attributes) {
        throw new Error("'attributes' is mandatory")
      }
      if (!expr) {
        throw new Error("'expr' is mandatory")
      }
    }

    toJSON () {
      return {
        name: this.name,
        type: this.type,
        attributes: this.attributes,
        elements: this.expr.toJSON()
      }
    }

    isAllowed (tagName) {
      return this.expr.isAllowed(tagName)
    }

    isTextAllowed () {
      return this.expr.isAllowed(TEXT$4$1)
    }

    printStructure () {
      return `${this.name} ::= ${this.expr.toString()}`
    }

    findFirstValidPos (el, newTag) {
      return this.expr._findInsertPos(el, newTag, 'first')
    }

    findLastValidPos (el, newTag) {
      return this.expr._findInsertPos(el, newTag, 'last')
    }
  }

  ElementSchema$1.fromJSON = function (data) {
    return new ElementSchema$1(
      data.name,
      data.type,
      data.attributes,
      Expression$1.fromJSON(data.elements)
    )
  };

  function _validateElement$1 (elementSchema, el) {
    let errors = [];
    let valid = true;
    { 
      const res = _checkAttributes$1(elementSchema, el);
      if (!res.ok) {
        errors = errors.concat(res.errors);
        valid = false;
      }
    }
    
    if (elementSchema.type === 'external') ; else {
      
      let res;
      if (el._isXMLTextElement) {
        res = _checkChildren$1(elementSchema, el.toXML());
      } else {
        res = _checkChildren$1(elementSchema, el);
      }
      if (!res.ok) {
        errors = errors.concat(res.errors);
        valid = false;
      }
    }
    return {
      errors,
      ok: valid
    }
  }

  function _checkAttributes$1(elementSchema, el) { 
    return { ok: true }
  }

  function _checkChildren$1 (elementSchema, el) {
    
    
    if (elementSchema.type === 'external') {
      return true
    }
    const expr = elementSchema.expr;
    const state = expr.getInitialState();
    const iterator = el.getChildNodeIterator();
    let valid = true;
    while (valid && iterator.hasNext()) {
      const childEl = iterator.next();
      let token;
      if (childEl.isTextNode()) {
        
        if (elementSchema.type !== 'text' && _isTextNodeEmpty$1(childEl)) {
          continue
        }
        token = TEXT$4$1;
      } else if (childEl.isElementNode()) {
        token = childEl.tagName;
      } else {
        continue
      }
      if (!expr.consume(state, token)) {
        valid = false;
      }
    }
    
    if (state.errors.length > 0) {
      state.errors.forEach((err) => {
        err.el = el;
      });
    }
    if (valid && !expr.isFinished(state)) {
      state.errors.push({
        msg: `<${el.tagName}> is incomplete.\nSchema: ${expr.toString()}`,
        el
      });
      valid = false;
    }
    if (valid) {
      state.ok = true;
    }
    return state
  }

  var ManifestSchemaData$1 = {"start":"dar","elements":{"dar":{"name":"dar","type":"element","attributes":{},"elements":{"name":"dar","content":{"type":",","blocks":["documents","assets"]}}},"documents":{"name":"documents","type":"element","attributes":{},"elements":{"name":"documents","content":{"type":"*","block":"document"}}},"assets":{"name":"assets","type":"element","attributes":{},"elements":{"name":"assets","content":{"type":"*","block":"asset"}}},"document":{"name":"document","type":"element","attributes":{"id":{"name":"id"},"name":{"name":"name"},"type":{"name":"type"},"path":{"name":"path"}},"elements":{"name":"document","content":{"type":",","blocks":[]}}},"asset":{"name":"asset","type":"element","attributes":{"id":{"name":"id"},"name":{"name":"name"},"type":{"name":"type"},"path":{"name":"path"}},"elements":{"name":"asset","content":{"type":",","blocks":[]}}}}};

  const ManifestSchema$1 = XMLSchema$1.fromJSON(ManifestSchemaData$1);


  ManifestSchema$1.getName = function() {
    return 'RDC-Manifest'
  };

  ManifestSchema$1.getVersion = function() {
    return '1.0'
  };

  ManifestSchema$1.getDocTypeParams = function() {
    return ['manifest', 'RDC-Manifest 1.0', ManifestSchema$1.uri]
  };


  ManifestSchema$1.getDefaultTextType = function () {
    return 'text'
  };

  ManifestSchema$1.uri = '//Manifest-1.0.dtd';

  class Data$1 extends EventEmitter$1 {
    
    constructor (schema, nodeFactory) {
      super();

      
      if (!schema) {
        throw new Error('schema is mandatory')
      }
      if (!nodeFactory) {
        throw new Error('nodeFactory is mandatory')
      }
      

      this.schema = schema;
      this.nodeFactory = nodeFactory;
      this.nodes = {};
      this.indexes = {};

      
      
      this.__QUEUE_INDEXING__ = false;
      this.queue = [];
    }

    
    contains (id) {
      return Boolean(this.nodes[id])
    }

    
    get (path, strict) {
      let result = this._get(path);
      if (strict && result === undefined) {
        if (isString$1(path)) {
          throw new Error("Could not find node with id '" + path + "'.")
        } else if (!this.contains(path[0])) {
          throw new Error("Could not find node with id '" + path[0] + "'.")
        } else {
          throw new Error("Property for path '" + path + "' us undefined.")
        }
      }
      return result
    }

    _get (path) {
      if (!path) return undefined
      let result;
      if (isString$1(path)) {
        result = this.nodes[path];
      } else if (path.length === 1) {
        result = this.nodes[path[0]];
      } else if (path.length > 1) {
        let context = this.nodes[path[0]];
        for (let i = 1; i < path.length - 1; i++) {
          if (!context) return undefined
          context = context[path[i]];
        }
        if (!context) return undefined
        result = context[path[path.length - 1]];
      }
      return result
    }

    
    getNodes () {
      return this.nodes
    }

    
    create (nodeData) {
      var node = this.nodeFactory.create(nodeData.type, nodeData);
      if (!node) {
        throw new Error('Illegal argument: could not create node for data:', nodeData)
      }
      if (this.contains(node.id)) {
        throw new Error('Node already exists: ' + node.id)
      }
      if (!node.id || !node.type) {
        throw new Error('Node id and type are mandatory.')
      }
      this.nodes[node.id] = node;

      var change = {
        type: 'create',
        node: node
      };

      if (this.__QUEUE_INDEXING__) {
        this.queue.push(change);
      } else {
        this._updateIndexes(change);
      }

      return node
    }

    
    delete (nodeId) {
      var node = this.nodes[nodeId];
      if (!node) return
      node.dispose();
      delete this.nodes[nodeId];

      var change = {
        type: 'delete',
        node: node
      };

      if (this.__QUEUE_INDEXING__) {
        this.queue.push(change);
      } else {
        this._updateIndexes(change);
      }

      return node
    }

    
    set (path, newValue) {
      let node = this.get(path[0]);
      let oldValue = this._set(path, newValue);
      var change = {
        type: 'set',
        node: node,
        path: path,
        newValue: newValue,
        oldValue: oldValue
      };
      if (this.__QUEUE_INDEXING__) {
        this.queue.push(change);
      } else {
        this._updateIndexes(change);
      }
      return oldValue
    }

    _set (path, newValue) {
      let oldValue = _setValue$1(this.nodes, path, newValue);
      return oldValue
    }

    
    update (path, diff) {
      var realPath = this.getRealPath(path);
      if (!realPath) {
        console.error('Could not resolve path', path);
        return
      }
      let node = this.get(realPath[0]);
      let oldValue = this._get(realPath);
      let newValue;
      if (diff.isOperation) {
        newValue = diff.apply(oldValue);
      } else {
        diff = this._normalizeDiff(oldValue, diff);
        if (isString$1(oldValue)) {
          switch (diff.type) {
            case 'delete': {
              newValue = oldValue.split('').splice(diff.start, diff.end - diff.start).join('');
              break
            }
            case 'insert': {
              newValue = [oldValue.substring(0, diff.start), diff.text, oldValue.substring(diff.start)].join('');
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else if (isArray$2(oldValue)) {
          newValue = oldValue.slice(0);
          switch (diff.type) {
            case 'delete': {
              newValue.splice(diff.pos, 1);
              break
            }
            case 'insert': {
              newValue.splice(diff.pos, 0, diff.value);
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else if (oldValue._isCoordinate) {
          switch (diff.type) {
            case 'shift': {
              
              oldValue = { path: oldValue.path, offset: oldValue.offset };
              newValue = oldValue;
              newValue.offset += diff.value;
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else {
          throw new Error('Diff is not supported:', JSON.stringify(diff))
        }
      }
      this._set(realPath, newValue);

      var change = {
        type: 'update',
        node: node,
        path: realPath,
        newValue: newValue,
        oldValue: oldValue
      };

      if (this.__QUEUE_INDEXING__) {
        this.queue.push(change);
      } else {
        this._updateIndexes(change);
      }

      return oldValue
    }

    
    _normalizeDiff (value, diff) {
      if (isString$1(value)) {
        
        if (diff['delete']) {
          console.warn('DEPRECATED: use doc.update(path, {type:"delete", start:s, end: e}) instead');
          diff = {
            type: 'delete',
            start: diff['delete'].start,
            end: diff['delete'].end
          };
        } else if (diff['insert']) {
          console.warn('DEPRECATED: use doc.update(path, {type:"insert", start:s, text: t}) instead');
          diff = {
            type: 'insert',
            start: diff['insert'].offset,
            text: diff['insert'].value
          };
        }
      } else if (isArray$2(value)) {
        
        if (diff['delete']) {
          console.warn('DEPRECATED: use doc.update(path, {type:"delete", pos:1}) instead');
          diff = {
            type: 'delete',
            pos: diff['delete'].offset
          };
        } else if (diff['insert']) {
          console.warn('DEPRECATED: use doc.update(path, {type:"insert", pos:1, value: "foo"}) instead');
          diff = {
            type: 'insert',
            pos: diff['insert'].offset,
            value: diff['insert'].value
          };
        }
      } else if (value._isCoordinate) {
        if (diff.hasOwnProperty('shift')) {
          console.warn('DEPRECATED: use doc.update(path, {type:"shift", value:2}) instead');
          diff = {
            type: 'shift',
            value: diff['shift']
          };
        }
      }
      return diff
    }

    
    toJSON () {
      let nodes = {};
      forEach$1(this.nodes, (node) => {
        nodes[node.id] = node.toJSON();
      });
      return {
        schema: [this.schema.id, this.schema.version],
        nodes: nodes
      }
    }

    reset () {
      this.clear();
    }

    
    clear () {
      this.nodes = {};
      forEach$1(this.indexes, index => index.clear());
    }

    
    addIndex (name, index) {
      if (this.indexes[name]) {
        console.error('Index with name %s already exists.', name);
      }
      index.reset(this);
      this.indexes[name] = index;
      return index
    }

    
    getIndex (name) {
      return this.indexes[name]
    }

    
    _updateIndexes (change) {
      if (!change || this.__QUEUE_INDEXING__) return
      forEach$1(this.indexes, function (index) {
        if (index.select(change.node)) {
          if (!index[change.type]) {
            console.error('Contract: every NodeIndex must implement ' + change.type);
          }
          index[change.type](change.node, change.path, change.newValue, change.oldValue);
        }
      });
    }

    
    _stopIndexing () {
      this.__QUEUE_INDEXING__ = true;
    }

    
    _startIndexing () {
      this.__QUEUE_INDEXING__ = false;
      while (this.queue.length > 0) {
        var change = this.queue.shift();
        this._updateIndexes(change);
      }
    }
  }

  function _setValue$1 (root, path, newValue) {
    let ctx = root;
    let L = path.length;
    for (let i = 0; i < L - 1; i++) {
      ctx = ctx[path[i]];
      if (!ctx) throw new Error('Can not set value.')
    }
    let oldValue = ctx[path[L - 1]];
    ctx[path[L - 1]] = newValue;
    return oldValue
  }

  class IncrementalData$1 extends Data$1 {
    
    create (nodeData) {
      if (nodeData._isNode) {
        nodeData = nodeData.toJSON();
      }
      let op = ObjectOperation$1.Create([nodeData.id], nodeData);
      this.apply(op);
      return op
    }

    
    delete (nodeId) {
      var op = null;
      var node = this.get(nodeId);
      if (node) {
        var nodeData = node.toJSON();
        op = ObjectOperation$1.Delete([nodeId], nodeData);
        this.apply(op);
      }
      return op
    }

    
    update (path, diff) {
      var diffOp = this._getDiffOp(path, diff);
      var op = ObjectOperation$1.Update(path, diffOp);
      this.apply(op);
      return op
    }

    
    set (path, newValue) {
      var oldValue = this.get(path);
      var op = ObjectOperation$1.Set(path, oldValue, newValue);
      this.apply(op);
      return op
    }

    
    apply (op) {
      if (op.type === ObjectOperation$1.NOP) return
      else if (op.type === ObjectOperation$1.CREATE) {
        
        super.create(cloneDeep$1(op.val));
      } else if (op.type === ObjectOperation$1.DELETE) {
        super.delete(op.val.id);
      } else if (op.type === ObjectOperation$1.UPDATE) {
        var oldVal = this.get(op.path);
        var diff = op.diff;
        if (op.propertyType === 'array') {
          if (!(diff._isArrayOperation)) {
            diff = ArrayOperation$1.fromJSON(diff);
          }
          
          diff.apply(oldVal);
        } else if (op.propertyType === 'string') {
          if (!(diff._isTextOperation)) {
            diff = TextOperation$1.fromJSON(diff);
          }
          var newVal = diff.apply(oldVal);
          super.set(op.path, newVal);
        } else if (op.propertyType === 'coordinate') {
          if (!(diff._isCoordinateOperation)) {
            diff = CoordinateOperation$1.fromJSON(diff);
          }
          diff.apply(oldVal);
        } else {
          throw new Error('Unsupported type for operational update.')
        }
      } else if (op.type === ObjectOperation$1.SET) {
        super.set(op.path, op.val);
      } else {
        throw new Error('Illegal state.')
      }
      this.emit('operation:applied', op, this);
    }

    
    _getDiffOp (path, diff) {
      var diffOp = null;
      if (diff.isOperation) {
        diffOp = diff;
      } else {
        var value = this.get(path);
        diff = this._normalizeDiff(value, diff);
        if (value === null || value === undefined) {
          throw new Error('Property has not been initialized: ' + JSON.stringify(path))
        } else if (isString$1(value)) {
          switch (diff.type) {
            case 'delete': {
              diffOp = TextOperation$1.Delete(diff.start, value.substring(diff.start, diff.end));
              break
            }
            case 'insert': {
              diffOp = TextOperation$1.Insert(diff.start, diff.text);
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else if (isArray$2(value)) {
          switch (diff.type) {
            case 'delete': {
              diffOp = ArrayOperation$1.Delete(diff.pos, value[diff.pos]);
              break
            }
            case 'insert': {
              diffOp = ArrayOperation$1.Insert(diff.pos, diff.value);
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else if (value._isCoordinate) {
          switch (diff.type) {
            case 'shift': {
              diffOp = CoordinateOperation$1.Shift(diff.value);
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        }
      }
      if (!diffOp) {
        throw new Error('Unsupported diff: ' + JSON.stringify(diff))
      }
      return diffOp
    }
  }

  class PropertyIndex$1 extends NodeIndex$1 {
    constructor (property) {
      super();

      this._property = property || 'id';
      this.index = new TreeIndex$1();
    }

    
    get (path) {
      return this.index.get(path) || {}
    }

    
    getAll (path) {
      return this.index.getAll(path)
    }

    clear () {
      this.index.clear();
    }

    
    select(node) { 
      return true
    }

    
    create (node) {
      var values = node[this._property];
      if (!isArray$2(values)) {
        values = [values];
      }
      forEach$1(values, function (value) {
        this.index.set([value, node.id], node);
      }.bind(this));
    }

    
    delete (node) {
      var values = node[this._property];
      if (!isArray$2(values)) {
        values = [values];
      }
      forEach$1(values, function (value) {
        this.index.delete([value, node.id]);
      }.bind(this));
    }

    
    update (node, path, newValue, oldValue) {
      if (!this.select(node) || path[1] !== this._property) return
      var values = oldValue;
      if (!isArray$2(values)) {
        values = [values];
      }
      forEach$1(values, function (value) {
        this.index.delete([value, node.id]);
      }.bind(this));
      values = newValue;
      if (!isArray$2(values)) {
        values = [values];
      }
      forEach$1(values, function (value) {
        this.index.set([value, node.id], node);
      }.bind(this));
    }

    set (node, path, newValue, oldValue) {
      this.update(node, path, newValue, oldValue);
    }

    _initialize (data) {
      forEach$1(data.getNodes(), function (node) {
        if (this.select(node)) {
          this.create(node);
        }
      }.bind(this));
    }
  }

  class AnnotationIndex$1 extends DocumentIndex$1 {
    constructor () {
      super();

      this.byPath = new TreeIndex$1();
      this.byType = new TreeIndex$1();
    }

    select (node) {
      return node.isPropertyAnnotation()
    }

    clear () {
      this.byPath.clear();
      this.byType.clear();
    }

    
    get (path, start, end, type) {
      var annotations;
      if (isString$1(path) || path.length === 1) {
        annotations = this.byPath.getAll(path) || {};
      } else {
        annotations = this.byPath.get(path);
      }
      annotations = map$2(annotations);
      if (isNumber$1(start)) {
        annotations = filter$1(annotations, AnnotationIndex$1.filterByRange(start, end));
      }
      if (type) {
        annotations = filter$1(annotations, DocumentIndex$1.filterByType(type));
      }
      return annotations
    }

    create (anno) {
      const path = anno.start.path;
      this.byType.set([anno.type, anno.id], anno);
      if (path && path.length > 0) {
        this.byPath.set(anno.start.path.concat([anno.id]), anno);
      }
    }

    delete (anno) {
      this._delete(anno.type, anno.id, anno.start.path);
    }

    _delete (type, id, path) {
      this.byType.delete([type, id]);
      if (path && path.length > 0) {
        this.byPath.delete(path.concat([id]));
      }
    }

    update (node, path, newValue, oldValue) {
      
      if (this.select(node) && path[1] === 'start' && path[2] === 'path') {
        this._delete(node.type, node.id, oldValue);
        this.create(node);
      }
    }
  }

  AnnotationIndex$1.filterByRange = function (start, end) {
    return function (anno) {
      var aStart = anno.start.offset;
      var aEnd = anno.end.offset;
      var overlap = (aEnd >= start);
      
      if (isNumber$1(end)) {
        overlap = overlap && (aStart <= end);
      }
      return overlap
    }
  };

  class ContainerAnnotationIndex$1 extends DocumentIndex$1 {
    constructor () {
      super();
      this.byId = new TreeIndex$1();
    }

    select (node) {
      return node.isContainerAnnotation()
    }

    clear () {
      this.byId.clear();
    }

    get (containerId, type) {
      var annotations = map$2(this.byId.get(containerId));
      if (isString$1(type)) {
        annotations = filter$1(annotations, DocumentIndex$1.filterByType);
      }
      return annotations
    }

    create (anno) {
      this.byId.set([anno.containerId, anno.id], anno);
    }

    delete (anno) {
      this.byId.delete([anno.containerId, anno.id]);
    }

    update(node, path, newValue, oldValue) { 
      
    }
  }

  class DocumentNodeFactory$1 {
    constructor (doc) {
      this.doc = doc;
    }

    create (nodeType, nodeData) {
      var NodeClass = this.doc.schema.getNodeClass(nodeType);
      if (!NodeClass) {
        throw new Error('No node registered by that name: ' + nodeType)
      }
      return new NodeClass(this.doc, nodeData)
    }
  }

  function copySelection$1 (doc, selection) {
    if (!selection) throw new Error("'selection' is mandatory.")
    let copy = null;
    if (!selection.isNull() && !selection.isCollapsed()) {
      
      if (selection.isPropertySelection()) {
        copy = _copyPropertySelection$1(doc, selection);
      } else if (selection.isContainerSelection()) {
        copy = _copyContainerSelection$1(doc, selection);
      } else if (selection.isNodeSelection()) {
        copy = _copyNodeSelection$1(doc, selection);
      } else {
        console.error('Copy is not yet supported for selection type.');
      }
    }
    return copy
  }

  function _copyPropertySelection$1 (doc, selection) {
    let path = selection.start.path;
    let offset = selection.start.offset;
    let endOffset = selection.end.offset;
    let text = doc.get(path);
    let snippet = doc.createSnippet();
    let containerNode = snippet.getContainer();
    snippet.create({
      type: doc.schema.getDefaultTextType(),
      id: TEXT_SNIPPET_ID$1,
      content: text.substring(offset, endOffset)
    });
    containerNode.show(TEXT_SNIPPET_ID$1);
    let annotations = doc.getIndex('annotations').get(path, offset, endOffset);
    forEach$1(annotations, function (anno) {
      let data = cloneDeep$1(anno.toJSON());
      let path = [TEXT_SNIPPET_ID$1, 'content'];
      data.start = {
        path: path,
        offset: Math.max(offset, anno.start.offset) - offset
      };
      data.end = {
        path: path,
        offset: Math.min(endOffset, anno.end.offset) - offset
      };
      snippet.create(data);
    });
    return snippet
  }

  function _copyContainerSelection$1 (tx, sel) {
    let snippet = tx.createSnippet();
    let container = snippet.getContainer();

    let nodeIds = sel.getNodeIds();
    let L = nodeIds.length;
    if (L === 0) return snippet

    let start = sel.start;
    let end = sel.end;

    let skippedFirst = false;
    let skippedLast = false;

    
    let created = {};
    for (let i = 0; i < L; i++) {
      let id = nodeIds[i];
      let node = tx.get(id);
      
      if (i === 0 && isLast$1(tx, start)) {
        skippedFirst = true;
        continue
      }
      if (i === L - 1 && isFirst$1(tx, end)) {
        skippedLast = true;
        continue
      }
      if (!created[id]) {
        copyNode$1(node).forEach((nodeData) => {
          let copy = snippet.create(nodeData);
          created[copy.id] = true;
        });
        container.show(id);
      }
    }
    if (!skippedFirst) {
      
      let startNode = snippet.get(start.getNodeId()).getContainerRoot();
      if (startNode.isText()) {
        deleteTextRange$1(snippet, null, start);
      } else if (startNode.isList()) {
        deleteListRange$1(snippet, startNode, null, start);
      }
    }
    if (!skippedLast) {
      
      let endNode = snippet.get(end.getNodeId()).getContainerRoot();
      if (endNode.isText()) {
        deleteTextRange$1(snippet, end, null);
      } else if (endNode.isList()) {
        deleteListRange$1(snippet, endNode, end, null);
      }
    }
    return snippet
  }

  function _copyNodeSelection$1 (doc, selection) {
    let snippet = doc.createSnippet();
    let containerNode = snippet.getContainer();
    let nodeId = selection.getNodeId();
    let node = doc.get(nodeId);
    copyNode$1(node).forEach((nodeData) => {
      snippet.create(nodeData);
    });
    containerNode.show(node.id);
    return snippet
  }

  function paste$1 (tx, args) {
    let sel = tx.selection;
    if (!sel || sel.isNull() || sel.isCustomSelection()) {
      throw new Error('Can not paste, without selection or a custom selection.')
    }
    args = args || {};
    args.text = args.text || '';
    let pasteDoc = args.doc;
    
    
    let inContainer = Boolean(sel.containerId);

    
    
    
    if (!pasteDoc && !inContainer) {
      tx.insertText(args.text);
      return
    }
    if (!pasteDoc) {
      pasteDoc = _convertPlainTextToDocument$1(tx, args);
    }
    if (pasteDoc && !inContainer) {
      
      let nodes = [];
      let container = pasteDoc.get('snippet');
      let content = container.getContent();
      content.forEach(nodeId => {
        let text = pasteDoc.get(nodeId).getText();
        nodes.push(text);
      });
      tx.insertText(nodes.join('\r\n'));
      return
    }
    if (!sel.isCollapsed()) {
      tx.deleteSelection();
    }
    let snippet = pasteDoc.get(SNIPPET_ID$1);
    if (snippet.getLength() > 0) {
      let first = snippet.getChildAt(0);
      if (first.isText()) {
        _pasteAnnotatedText$1(tx, pasteDoc);
        
        
        
        snippet.hideAt(0);
      }
      
      if (snippet.getLength() > 0) {
        _pasteDocument$1(tx, pasteDoc);
      }
    }
    return args
  }


  function _convertPlainTextToDocument$1 (tx, args) {
    let lines = args.text.split(/\s*\n\s*\n/);
    let pasteDoc = tx.getDocument().newInstance();
    let defaultTextType = pasteDoc.getSchema().getDefaultTextType();
    let container = pasteDoc.create({
      type: '@container',
      id: SNIPPET_ID$1,
      nodes: []
    });
    let node;
    if (lines.length === 1) {
      node = pasteDoc.create({
        id: TEXT_SNIPPET_ID$1,
        type: defaultTextType,
        content: lines[0]
      });
      container.show(node.id);
    } else {
      for (let i = 0; i < lines.length; i++) {
        node = pasteDoc.create({
          id: uuid$1(defaultTextType),
          type: defaultTextType,
          content: lines[i]
        });
        container.show(node.id);
      }
    }
    return pasteDoc
  }

  function _pasteAnnotatedText$1 (tx, copy) {
    let sel = tx.selection;
    const nodes = copy.get(SNIPPET_ID$1).nodes;
    const firstId = nodes[0];
    const first = copy.get(firstId);
    const textPath = first.getPath();
    const text = copy.get(textPath);
    const annotations = copy.getIndex('annotations').get(textPath);
    
    let path = sel.start.path;
    let offset = sel.start.offset;
    tx.insertText(text);
    
    forEach$1(annotations, function (anno) {
      let data = anno.toJSON();
      data.start.path = path.slice(0);
      data.start.offset += offset;
      data.end.offset += offset;
      
      if (tx.get(data.id)) data.id = uuid$1(data.type);
      tx.create(data);
    });
  }

  function _pasteDocument$1 (tx, pasteDoc) {
    let sel = tx.selection;
    let containerId = sel.containerId;
    let container = tx.get(containerId);
    let insertPos;
    if (sel.isPropertySelection()) {
      let startPath = sel.start.path;
      let nodeId = sel.start.getNodeId();
      let startPos = container.getPosition(nodeId, 'strict');
      let text = tx.get(startPath);
      
      
      if (text.length === 0) {
        insertPos = startPos;
        container.hide(nodeId);
        deleteNode$1(tx, tx.get(nodeId));
      } else if (text.length === sel.start.offset) {
        insertPos = startPos + 1;
      } else {
        tx.break();
        insertPos = startPos + 1;
      }
    } else if (sel.isNodeSelection()) {
      let nodePos = container.getPosition(sel.getNodeId(), 'strict');
      if (sel.isBefore()) {
        insertPos = nodePos;
      } else if (sel.isAfter()) {
        insertPos = nodePos + 1;
      } else {
        throw new Error('Illegal state: the selection should be collapsed.')
      }
    }
    
    let nodeIds = pasteDoc.get(SNIPPET_ID$1).nodes;
    let insertedNodes = [];
    let visited = {};
    for (let i = 0; i < nodeIds.length; i++) {
      let node = pasteDoc.get(nodeIds[i]);
      
      
      
      
      
      let newId = _transferWithDisambiguatedIds$1(node.getDocument(), tx, node.id, visited);
      
      node = tx.get(newId);
      container.showAt(insertPos++, newId);
      insertedNodes.push(node);
    }

    if (insertedNodes.length > 0) {
      let lastNode = last$1$1(insertedNodes);
      setCursor$1(tx, lastNode, containerId, 'after');
    }
  }






  function _transferWithDisambiguatedIds$1 (sourceDoc, targetDoc, id, visited) {
    if (visited[id]) throw new Error('FIXME: dont call me twice')
    const node = sourceDoc.get(id, 'strict');
    let oldId = node.id;
    let newId;
    if (targetDoc.contains(node.id)) {
      
      newId = uuid$1(node.type);
      node.id = newId;
    }
    visited[id] = node.id;
    const annotationIndex = sourceDoc.getIndex('annotations');
    const nodeSchema = node.getSchema();
    
    let annos = [];
    
    
    
    for (let prop of nodeSchema) {
      const name = prop.name;
      if (name === 'id' || name === 'type') continue
      
      if ((prop.isReference() && prop.isOwned()) || (prop.type === 'file')) {
        
        
        let val = node[prop.name];
        if (prop.isArray()) {
          _transferArrayOfReferences$1(sourceDoc, targetDoc, val, visited);
        } else {
          let id = val;
          if (!visited[id]) {
            node[name] = _transferWithDisambiguatedIds$1(sourceDoc, targetDoc, id, visited);
          }
        }
      
      } else if (prop.isText()) {
        
        
        
        
        
        let _annos = annotationIndex.get([oldId, prop.name]);
        for (let i = 0; i < _annos.length; i++) {
          let anno = _annos[i];
          if (anno.start.path[0] === oldId && newId) {
            anno.start.path[0] = newId;
          }
          if (anno.end.path[0] === oldId && newId) {
            anno.end.path[0] = newId;
          }
          annos.push(anno);
        }
      }
    }
    targetDoc.create(node);
    for (let i = 0; i < annos.length; i++) {
      _transferWithDisambiguatedIds$1(sourceDoc, targetDoc, annos[i].id, visited);
    }
    return node.id
  }

  function _transferArrayOfReferences$1 (sourceDoc, targetDoc, arr, visited) {
    for (let i = 0; i < arr.length; i++) {
      let val = arr[i];
      
      if (isArray$2(val)) {
        _transferArrayOfReferences$1(sourceDoc, targetDoc, val, visited);
      } else {
        let id = val;
        if (id && !visited[id]) {
          arr[i] = _transferWithDisambiguatedIds$1(sourceDoc, targetDoc, id, visited);
        }
      }
    }
  }

  class Editing$1 {
    
    annotate (tx, annotation) {
      let sel = tx.selection;
      let schema = tx.getSchema();
      let AnnotationClass = schema.getNodeClass(annotation.type);
      if (!AnnotationClass) throw new Error('Unknown annotation type', annotation)
      let start = sel.start;
      let end = sel.end;
      let containerId = sel.containerId;
      let nodeData = { start, end, containerId };
      
      
      if (sel.isPropertySelection()) {
        if (!AnnotationClass.prototype._isAnnotation) {
          throw new Error('Annotation can not be created for a selection.')
        }
      } else if (sel.isContainerSelection()) {
        if (AnnotationClass.prototype._isPropertyAnnotation) {
          console.warn('NOT SUPPORTED YET: creating property annotations for a non collapsed container selection.');
        }
      }
      Object.assign(nodeData, annotation);
      return tx.create(nodeData)
    }

    break (tx) {
      let sel = tx.selection;
      if (sel.isNodeSelection()) {
        let containerId = sel.containerId;
        let container = tx.get(containerId);
        let nodeId = sel.getNodeId();
        let nodePos = container.getPosition(nodeId, 'strict');
        let textNode = this.createTextNode(tx, container);
        if (sel.isBefore()) {
          container.showAt(nodePos, textNode.id);
          
        } else {
          container.showAt(nodePos + 1, textNode.id);
          setCursor$1(tx, textNode, containerId, 'before');
        }
      } else if (sel.isCustomSelection()) ; else if (sel.isCollapsed() || sel.isPropertySelection()) {
        let containerId = sel.containerId;
        if (!sel.isCollapsed()) {
          
          this._deletePropertySelection(tx, sel);
          tx.setSelection(sel.collapse('left'));
        }
        
        if (containerId) {
          let container = tx.get(containerId);
          let nodeId = sel.start.path[0];
          let node = tx.get(nodeId);
          this._breakNode(tx, node, sel.start, container);
        }
      } else if (sel.isContainerSelection()) {
        let start = sel.start;
        let containerId = sel.containerId;
        let container = tx.get(containerId);
        let startNodeId = start.path[0];
        let nodePos = container.getPosition(startNodeId, 'strict');
        this._deleteContainerSelection(tx, sel, { noMerge: true });
        setCursor$1(tx, container.getNodeAt(nodePos + 1), containerId, 'before');
      }
    }

    createTextNode (tx, container, text) { 
      
      
      return tx.createDefaultTextNode(text)
    }

    createListNode (tx, container, params = {}) { 
      
      
      return tx.create({ type: 'list', items: [], listType: params.listType || 'bullet' })
    }

    delete (tx, direction) {
      let sel = tx.selection;
      
      
      
      
      if (sel.isNodeSelection()) {
        this._deleteNodeSelection(tx, sel, direction);
      
      } else if (sel.isCustomSelection()) ; else if (sel.isCollapsed()) {
        
        
        
        let path = sel.start.path;
        let node = tx.get(path[0]);
        let text = tx.get(path);
        let offset = sel.start.offset;
        let needsMerge = (sel.containerId && (
          (offset === 0 && direction === 'left') ||
          (offset === text.length && direction === 'right')
        ));
        if (needsMerge) {
          
          
          
          
          
          
          
          let root = node.getContainerRoot();
          if (root.isList() && offset === 0 && direction === 'left') {
            return this.toggleList(tx)
          } else {
            let container = tx.get(sel.containerId);
            this._merge(tx, root, sel.start, direction, container);
          }
        } else {
          
          if ((offset === 0 && direction === 'left') ||
            (offset === text.length && direction === 'right')) {
            return
          }
          let startOffset = (direction === 'left') ? offset - 1 : offset;
          let endOffset = startOffset + 1;
          let start = { path: path, offset: startOffset };
          let end = { path: path, offset: endOffset };
          deleteTextRange$1(tx, start, end);
          tx.setSelection({
            type: 'property',
            path: path,
            startOffset: startOffset,
            containerId: sel.containerId
          });
        }
      
      } else if (sel.isPropertySelection()) {
        deleteTextRange$1(tx, sel.start, sel.end);
        tx.setSelection(sel.collapse('left'));
      
      } else if (sel.isContainerSelection()) {
        this._deleteContainerSelection(tx, sel);
      } else {
        console.warn('Unsupported case: tx.delete(%)', direction, sel);
      }
    }

    _deleteNodeSelection (tx, sel, direction) {
      let nodeId = sel.getNodeId();
      let container = tx.get(sel.containerId);
      let nodePos = container.getPosition(nodeId, 'strict');
      if (sel.isFull() ||
          (sel.isBefore() && direction === 'right') ||
          (sel.isAfter() && direction === 'left')) {
        
        container.hideAt(nodePos);
        deleteNode$1(tx, tx.get(nodeId));
        let newNode = this.createTextNode(tx, container);
        container.showAt(nodePos, newNode.id);
        tx.setSelection({
          type: 'property',
          path: newNode.getPath(),
          startOffset: 0,
          containerId: container.id
        });
      } else {
        
        if (sel.isBefore() && direction === 'left') {
          if (nodePos > 0) {
            let previous = container.getNodeAt(nodePos - 1);
            if (previous.isText()) {
              tx.setSelection({
                type: 'property',
                path: previous.getPath(),
                startOffset: previous.getLength()
              });
              this.delete(tx, direction);
            } else {
              tx.setSelection({
                type: 'node',
                nodeId: previous.id,
                containerId: container.id
              });
            }
          }
        } else if (sel.isAfter() && direction === 'right') {
          if (nodePos < container.getLength() - 1) {
            let next = container.getNodeAt(nodePos + 1);
            if (next.isText()) {
              tx.setSelection({
                type: 'property',
                path: next.getPath(),
                startOffset: 0
              });
              this.delete(tx, direction);
            } else {
              tx.setSelection({
                type: 'node',
                nodeId: next.id,
                containerId: container.id
              });
            }
          }
        } else {
          console.warn('Unsupported case: delete(%s)', direction, sel);
        }
      }
    }

    _deletePropertySelection (tx, sel) {
      let path = sel.start.path;
      let start = sel.start.offset;
      let end = sel.end.offset;
      tx.update(path, { type: 'delete', start: start, end: end });
      annotationHelpers$1.deletedText(tx, path, start, end);
    }

    
    _deleteContainerSelection (tx, sel, options = {}) {
      let containerId = sel.containerId;
      let container = tx.get(containerId);
      let start = sel.start;
      let end = sel.end;
      let startId = start.getNodeId();
      let endId = end.getNodeId();
      let startPos = container.getPosition(startId, 'strict');
      let endPos = container.getPosition(endId, 'strict');

      
      if (startPos === endPos) {
        
        let node = tx.get(startId).getContainerRoot();
        
        if (node.isText()) {
          deleteTextRange$1(tx, start, end);
        } else if (node.isList()) {
          deleteListRange$1(tx, node, start, end);
        } else {
          throw new Error('Not supported yet.')
        }
        tx.setSelection(sel.collapse('left'));
        return
      }

      

      let firstNode = tx.get(start.getNodeId());
      let lastNode = tx.get(end.getNodeId());
      let firstEntirelySelected = isEntirelySelected$1(tx, firstNode, start, null);
      let lastEntirelySelected = isEntirelySelected$1(tx, lastNode, null, end);

      
      if (lastEntirelySelected) {
        container.hideAt(endPos);
        deleteNode$1(tx, lastNode);
      } else {
        
        let node = lastNode.getContainerRoot();
        
        if (node.isText()) {
          deleteTextRange$1(tx, null, end);
        } else if (node.isList()) {
          deleteListRange$1(tx, node, null, end);
        }
      }

      
      for (let i = endPos - 1; i > startPos; i--) {
        let nodeId = container.getNodeIdAt(i);
        container.hideAt(i);
        deleteNode$1(tx, tx.get(nodeId));
      }

      
      if (firstEntirelySelected) {
        container.hideAt(startPos);
        deleteNode$1(tx, firstNode);
      } else {
        
        let node = firstNode.getContainerRoot();
        
        if (node.isText()) {
          deleteTextRange$1(tx, start, null);
        } else if (node.isList()) {
          deleteListRange$1(tx, node, start, null);
        }
      }

      
      if (firstEntirelySelected && lastEntirelySelected) {
        
        let textNode = this.createTextNode(tx, container);
        container.showAt(startPos, textNode.id);
        tx.setSelection({
          type: 'property',
          path: textNode.getPath(),
          startOffset: 0,
          containerId: containerId
        });
      } else if (!firstEntirelySelected && !lastEntirelySelected) {
        if (!options.noMerge) {
          this._merge(tx, firstNode, sel.start, 'right', container);
        }
        tx.setSelection(sel.collapse('left'));
      } else if (firstEntirelySelected) {
        setCursor$1(tx, lastNode, container.id, 'before');
      } else {
        setCursor$1(tx, firstNode, container.id, 'after');
      }
    }

    insertInlineNode (tx, nodeData) {
      let sel = tx.selection;
      let text = '\uFEFF';
      this.insertText(tx, text);
      sel = tx.selection;
      let endOffset = tx.selection.end.offset;
      let startOffset = endOffset - text.length;
      nodeData = Object.assign({}, nodeData, {
        start: {
          path: sel.path,
          offset: startOffset
        },
        end: {
          path: sel.path,
          offset: endOffset
        }
      });
      return tx.create(nodeData)
    }

    insertBlockNode (tx, nodeData) {
      let sel = tx.selection;
      
      let blockNode;
      if (!nodeData._isNode || !tx.get(nodeData.id)) {
        blockNode = tx.create(nodeData);
      } else {
        blockNode = tx.get(nodeData.id);
      }
      
      if (sel.isNodeSelection()) {
        let containerId = sel.containerId;
        let container = tx.get(containerId);
        let nodeId = sel.getNodeId();
        let nodePos = container.getPosition(nodeId, 'strict');
        
        if (sel.isBefore()) {
          container.showAt(nodePos, blockNode.id);
        
        } else if (sel.isAfter()) {
          container.showAt(nodePos + 1, blockNode.id);
          tx.setSelection({
            type: 'node',
            containerId: containerId,
            nodeId: blockNode.id,
            mode: 'after'
          });
        } else {
          container.hideAt(nodePos);
          deleteNode$1(tx, tx.get(nodeId));
          container.showAt(nodePos, blockNode.id);
          tx.setSelection({
            type: 'node',
            containerId: containerId,
            nodeId: blockNode.id,
            mode: 'after'
          });
        }
      } else if (sel.isPropertySelection()) {
        
        if (!sel.containerId) throw new Error('insertBlockNode can only be used within a container.')
        let container = tx.get(sel.containerId);
        if (!sel.isCollapsed()) {
          this._deletePropertySelection(tx, sel);
          tx.setSelection(sel.collapse('left'));
        }
        let node = tx.get(sel.path[0]);
        
        if (!node) throw new Error('Invalid selection.')
        let nodePos = container.getPosition(node.id, 'strict');
        
        if (node.isText()) {
          let text = node.getText();
          
          if (text.length === 0) {
            container.hideAt(nodePos);
            deleteNode$1(tx, node);
            container.showAt(nodePos, blockNode.id);
            setCursor$1(tx, blockNode, container.id, 'after');
          
          } else if (sel.start.offset === 0) {
            container.showAt(nodePos, blockNode.id);
          
          } else if (sel.start.offset === text.length) {
            container.showAt(nodePos + 1, blockNode.id);
            setCursor$1(tx, blockNode, container.id, 'before');
          
          } else {
            this.break(tx);
            container.showAt(nodePos + 1, blockNode.id);
            setCursor$1(tx, blockNode, container.id, 'after');
          }
        } else {
          console.error('Not supported: insertBlockNode() on a custom node');
        }
      } else if (sel.isContainerSelection()) {
        if (sel.isCollapsed()) {
          let start = sel.start;
          
          if (start.isPropertyCoordinate()) {
            tx.setSelection({
              type: 'property',
              path: start.path,
              startOffset: start.offset,
              containerId: sel.containerId
            });
          } else if (start.isNodeCoordinate()) {
            tx.setSelection({
              type: 'node',
              containerId: sel.containerId,
              nodeId: start.path[0],
              mode: start.offset === 0 ? 'before' : 'after'
            });
          } else {
            throw new Error('Unsupported selection for insertBlockNode')
          }
          return this.insertBlockNode(tx, blockNode)
        } else {
          this.break(tx);
          return this.insertBlockNode(tx, blockNode)
        }
      }
      return blockNode
    }

    insertText (tx, text) {
      let sel = tx.selection;
      
      
      
      if (sel.isNodeSelection()) {
        let containerId = sel.containerId;
        let container = tx.get(containerId);
        let nodeId = sel.getNodeId();
        let nodePos = container.getPosition(nodeId, 'strict');
        let textNode = this.createTextNode(tx, container, text);
        if (sel.isBefore()) {
          container.showAt(nodePos, textNode);
        } else if (sel.isAfter()) {
          container.showAt(nodePos + 1, textNode);
        } else {
          container.hide(nodeId);
          deleteNode$1(tx, tx.get(nodeId));
          container.showAt(nodePos, textNode);
        }
        setCursor$1(tx, textNode, sel.containerId, 'after');
      } else if (sel.isCustomSelection()) ; else if (sel.isCollapsed() || sel.isPropertySelection()) {
        
        this._insertText(tx, sel, text);
        
      } else if (sel.isContainerSelection()) {
        this._deleteContainerSelection(tx, sel);
        this.insertText(tx, text);
      }
    }

    paste (tx, content) {
      if (!content) return
      
      if (isString$1(content)) {
        paste$1(tx, {text: content});
      } else if (content._isDocument) {
        paste$1(tx, {doc: content});
      } else {
        throw new Error('Illegal content for paste.')
      }
    }

    
    switchTextType (tx, data) {
      let sel = tx.selection;
      
      if (!sel.isPropertySelection()) {
        throw new Error('Selection must be a PropertySelection.')
      }
      let containerId = sel.containerId;
      
      if (!containerId) {
        throw new Error('Selection must be within a container.')
      }
      let path = sel.path;
      let nodeId = path[0];
      let node = tx.get(nodeId);
      
      if (!(node.isInstanceOf('text'))) {
        throw new Error('Trying to use switchTextType on a non text node.')
      }
      const newId = uuid$1(data.type);
      
      const oldPath = node.getPath();
      console.assert(oldPath.length === 2, 'Currently we assume that TextNodes store the plain-text on the first level');
      const textProp = oldPath[1];
      let newPath = [newId, textProp];
      
      let newNodeData = Object.assign({
        id: newId,
        type: data.type,
        direction: node.direction
      }, data);
      newNodeData[textProp] = node.getText();

      let newNode = tx.create(newNodeData);
      annotationHelpers$1.transferAnnotations(tx, path, 0, newPath, 0);

      
      let container = tx.get(sel.containerId);
      let pos = container.getPosition(nodeId, 'strict');
      container.hide(nodeId);
      deleteNode$1(tx, node);
      container.showAt(pos, newNode.id);

      tx.setSelection({
        type: 'property',
        path: newPath,
        startOffset: sel.start.offset,
        endOffset: sel.end.offset,
        containerId: containerId
      });

      return newNode
    }

    toggleList (tx, params) {
      let sel = tx.selection;
      let container = tx.get(sel.containerId);
      
      if (!container) {
        throw new Error('Selection must be within a container.')
      }
      if (sel.isPropertySelection()) {
        let nodeId = sel.start.path[0];
        
        let node = tx.get(nodeId).getContainerRoot();
        let nodePos = container.getPosition(node.id, 'strict');
        
        if (node.isText()) {
          container.hideAt(nodePos);
          let newList = this.createListNode(tx, container, params);
          let newItem = newList.createListItem(node.getText());
          annotationHelpers$1.transferAnnotations(tx, node.getPath(), 0, newItem.getPath(), 0);
          newList.appendItem(newItem);
          deleteNode$1(tx, node);
          container.showAt(nodePos, newList.id);
          tx.setSelection({
            type: 'property',
            path: newItem.getPath(),
            startOffset: sel.start.offset,
            containerId: sel.containerId
          });
        } else if (node.isList()) {
          let itemId = sel.start.path[0];
          let item = tx.get(itemId);
          let itemPos = node.getItemPosition(item);
          let newTextNode = this.createTextNode(tx, container, item.getText());
          annotationHelpers$1.transferAnnotations(tx, item.getPath(), 0, newTextNode.getPath(), 0);
          
          node.removeItemAt(itemPos);
          if (node.isEmpty()) {
            container.hideAt(nodePos);
            deleteNode$1(tx, node);
            container.showAt(nodePos, newTextNode.id);
          } else if (itemPos === 0) {
            container.showAt(nodePos, newTextNode.id);
          } else if (node.getLength() <= itemPos) {
            container.showAt(nodePos + 1, newTextNode.id);
          } else {
            
            let tail = [];
            const items = node.getItems();
            const L = items.length;
            for (let i = L - 1; i >= itemPos; i--) {
              tail.unshift(items[i]);
              node.removeItemAt(i);
            }
            let newList = this.createListNode(tx, container, node);
            for (let i = 0; i < tail.length; i++) {
              newList.appendItem(tail[i]);
            }
            container.showAt(nodePos + 1, newTextNode.id);
            container.showAt(nodePos + 2, newList.id);
          }
          tx.setSelection({
            type: 'property',
            path: newTextNode.getPath(),
            startOffset: sel.start.offset,
            containerId: sel.containerId
          });
        }
      } else if (sel.isContainerSelection()) {
        console.error('TODO: support toggleList with ContainerSelection');
      }
    }

    indent (tx) {
      let sel = tx.selection;
      if (sel.isPropertySelection()) {
        let nodeId = sel.start.getNodeId();
        
        let node = tx.get(nodeId).getContainerRoot();
        if (node.isList()) {
          let itemId = sel.start.path[0];
          let item = tx.get(itemId);
          let level = item.getLevel();
          
          if (item && level < 3) {
            item.setLevel(item.level + 1);
            
            tx.set([node.id, '_itemsChanged'], true);
          }
        }
      } else if (sel.isContainerSelection()) {
        console.error('TODO: support toggleList with ContainerSelection');
      }
    }

    dedent (tx) {
      let sel = tx.selection;
      if (sel.isPropertySelection()) {
        let nodeId = sel.start.getNodeId();
        
        let node = tx.get(nodeId).getContainerRoot();
        if (node.isList()) {
          let itemId = sel.start.path[0];
          let item = tx.get(itemId);
          let level = item.getLevel();
          if (item) {
            if (level > 1) {
              item.setLevel(item.level - 1);
              
              tx.set([node.id, '_itemsChanged'], true);
            }
            
            
            
            
            
          }
        }
      } else if (sel.isContainerSelection()) {
        console.error('TODO: support toggleList with ContainerSelection');
      }
    }

    
    _insertText (tx, sel, text) {
      let start = sel.start;
      let end = sel.end;
      
      if (!isArrayEqual$1(start.path, end.path)) {
        throw new Error('Unsupported state: range should be on one property')
      }
      let path = start.path;
      let startOffset = start.offset;
      let endOffset = end.offset;
      let typeover = !sel.isCollapsed();
      let L = text.length;
      
      if (typeover) {
        tx.update(path, { type: 'delete', start: startOffset, end: endOffset });
      }
      
      tx.update(path, { type: 'insert', start: startOffset, text: text });
      
      let annos = tx.getAnnotations(path);
      annos.forEach(function (anno) {
        let annoStart = anno.start.offset;
        let annoEnd = anno.end.offset;

        
        
        if (annoEnd < startOffset) ; else if (annoStart >= endOffset) {
          tx.update([anno.id, 'start'], { type: 'shift', value: startOffset - endOffset + L });
          tx.update([anno.id, 'end'], { type: 'shift', value: startOffset - endOffset + L });
        
        
        
        
        } else if (
          (annoStart >= startOffset && annoEnd < endOffset) ||
          (anno._isInlineNode && annoStart >= startOffset && annoEnd <= endOffset)
        ) {
          tx.delete(anno.id);
        
        } else if (annoStart >= startOffset && annoEnd >= endOffset) {
          
          if (annoStart > startOffset || !typeover) {
            tx.update([anno.id, 'start'], { type: 'shift', value: startOffset - annoStart + L });
          }
          tx.update([anno.id, 'end'], { type: 'shift', value: startOffset - endOffset + L });
        
        } else if (annoStart < startOffset && annoEnd < endOffset) {
          
          tx.update([anno.id, 'end'], { type: 'shift', value: startOffset - annoEnd + L });
        
        } else if (annoEnd === startOffset && !anno.constructor.autoExpandRight) ; else if (annoStart < startOffset && annoEnd >= endOffset) {
          if (anno._isInlineNode) ; else {
            tx.update([anno.id, 'end'], { type: 'shift', value: startOffset - endOffset + L });
          }
        } else {
          console.warn('TODO: handle annotation update case.');
        }
      });
      let offset = startOffset + text.length;
      tx.setSelection({
        type: 'property',
        path: start.path,
        startOffset: offset,
        containerId: sel.containerId,
        surfaceId: sel.surfaceId
      });
    }

    _breakNode (tx, node, coor, container) {
      
      node = node.getContainerRoot();
      
      if (node.isText()) {
        this._breakTextNode(tx, node, coor, container);
      } else if (node.isList()) {
        this._breakListNode(tx, node, coor, container);
      } else {
        throw new Error('Not supported')
      }
    }

    _breakTextNode (tx, node, coor, container) {
      let path = coor.path;
      let offset = coor.offset;
      let nodePos = container.getPosition(node.id, 'strict');
      let text = node.getText();

      
      
      if (offset === 0) {
        let newNode = tx.create({
          type: node.type,
          content: ''
        });
        
        container.showAt(nodePos, newNode.id);
        tx.setSelection({
          type: 'property',
          path: path,
          startOffset: 0,
          containerId: container.id
        });
      
      } else {
        const textPath = node.getPath();
        const textProp = textPath[1];
        const newId = uuid$1(node.type);
        let newNodeData = node.toJSON();
        newNodeData.id = newId;
        newNodeData[textProp] = text.substring(offset);
        
        if (offset === text.length) {
          newNodeData.type = tx.getSchema().getDefaultTextType();
        }
        let newNode = tx.create(newNodeData);
        
        if (offset < text.length) {
          
          annotationHelpers$1.transferAnnotations(tx, path, offset, newNode.getPath(), 0);
          
          tx.update(path, { type: 'delete', start: offset, end: text.length });
        }
        
        container.showAt(nodePos + 1, newNode.id);
        
        tx.setSelection({
          type: 'property',
          path: newNode.getPath(),
          startOffset: 0,
          containerId: container.id
        });
      }
    }

    _breakListNode (tx, node, coor, container) {
      let path = coor.path;
      let offset = coor.offset;
      let listItem = tx.get(path[0]);

      let L = node.length;
      let itemPos = node.getItemPosition(listItem);
      let text = listItem.getText();
      let textProp = listItem.getPath()[1];
      let newItemData = listItem.toJSON();
      delete newItemData.id;
      if (offset === 0) {
        
        if (!text) {
          
          
          let nodePos = container.getPosition(node.id, 'strict');
          let newTextNode = this.createTextNode(tx, container);
          
          if (L < 2) {
            container.hide(node.id);
            deleteNode$1(tx, node);
            container.showAt(nodePos, newTextNode.id);
          
          } else if (itemPos === 0) {
            node.removeItem(listItem);
            deleteNode$1(tx, listItem);
            container.showAt(nodePos, newTextNode.id);
          
          } else if (itemPos >= L - 1) {
            node.removeItem(listItem);
            deleteNode$1(tx, listItem);
            container.showAt(nodePos + 1, newTextNode.id);
          
          } else {
            let tail = [];
            const items = node.getItems().slice();
            for (let i = L - 1; i > itemPos; i--) {
              tail.unshift(items[i]);
              node.removeItem(items[i]);
            }
            node.removeItem(items[itemPos]);
            let newList = this.createListNode(tx, container, node);
            for (let i = 0; i < tail.length; i++) {
              newList.appendItem(tail[i]);
            }
            container.showAt(nodePos + 1, newTextNode.id);
            container.showAt(nodePos + 2, newList.id);
          }
          tx.setSelection({
            type: 'property',
            path: newTextNode.getPath(),
            startOffset: 0
          });
        
        } else {
          newItemData[textProp] = '';
          let newItem = tx.create(newItemData);
          node.insertItemAt(itemPos, newItem);
          tx.setSelection({
            type: 'property',
            path: listItem.getPath(),
            startOffset: 0
          });
        }
      
      } else {
        newItemData[textProp] = text.substring(offset);
        let newItem = tx.create(newItemData);
        
        if (offset < text.length) {
          
          annotationHelpers$1.transferAnnotations(tx, path, offset, newItem.getPath(), 0);
          
          tx.update(path, { type: 'delete', start: offset, end: text.length });
        }
        node.insertItemAt(itemPos + 1, newItem);
        tx.setSelection({
          type: 'property',
          path: newItem.getPath(),
          startOffset: 0
        });
      }
    }

    _merge (tx, node, coor, direction, container) {
      
      
      if (node.isList()) {
        let list = node;
        let itemId = coor.path[0];
        let item = tx.get(itemId);
        let itemPos = list.getItemPosition(item);
        let withinListNode = (
          (direction === 'left' && itemPos > 0) ||
          (direction === 'right' && itemPos < list.items.length - 1)
        );
        if (withinListNode) {
          itemPos = (direction === 'left') ? itemPos - 1 : itemPos;
          let target = list.getItemAt(itemPos);
          let targetLength = target.getLength();
          mergeListItems$1(tx, list.id, itemPos);
          tx.setSelection({
            type: 'property',
            path: target.getPath(),
            startOffset: targetLength,
            containerId: container.id
          });
          return
        }
      }
      
      let nodePos = container.getPosition(node, 'strict');
      if (direction === 'left' && nodePos > 0) {
        this._mergeNodes(tx, container, nodePos - 1, direction);
      } else if (direction === 'right' && nodePos < container.getLength() - 1) {
        this._mergeNodes(tx, container, nodePos, direction);
      }
    }

    _mergeNodes (tx, container, pos, direction) {
      let first = container.getChildAt(pos);
      let second = container.getChildAt(pos + 1);
      if (first.isText()) {
        
        if (first.isEmpty()) {
          container.hide(first.id);
          deleteNode$1(tx, first);
          
          
          setCursor$1(tx, second, container.id, 'before');
          return
        }
        let target = first;
        let targetPath = target.getPath();
        let targetLength = target.getLength();
        if (second.isText()) {
          let source = second;
          let sourcePath = source.getPath();
          container.hide(source.id);
          
          tx.update(targetPath, { type: 'insert', start: targetLength, text: source.getText() });
          
          annotationHelpers$1.transferAnnotations(tx, sourcePath, 0, targetPath, targetLength);
          deleteNode$1(tx, source);
          tx.setSelection({
            type: 'property',
            path: targetPath,
            startOffset: targetLength,
            containerId: container.id
          });
        } else if (second.isList()) {
          let list = second;
          if (!second.isEmpty()) {
            let source = list.getFirstItem();
            let sourcePath = source.getPath();
            
            list.removeItemAt(0);
            
            tx.update(targetPath, { type: 'insert', start: targetLength, text: source.getText() });
            
            annotationHelpers$1.transferAnnotations(tx, sourcePath, 0, targetPath, targetLength);
            
            deleteNode$1(tx, source);
          }
          if (list.isEmpty()) {
            container.hide(list.id);
            deleteNode$1(tx, list);
          }
          tx.setSelection({
            type: 'property',
            path: targetPath,
            startOffset: targetLength,
            containerId: container.id
          });
        } else {
          selectNode$1(tx, direction === 'left' ? first.id : second.id, container.id);
        }
      } else if (first.isList()) {
        if (second.isText()) {
          let target = first.getLastItem();
          let targetPath = target.getPath();
          let targetLength = target.getLength();
          let third = (container.length > pos + 2) ? container.getChildAt(pos + 2) : null;
          if (second.getLength() === 0) {
            container.hide(second.id);
            deleteNode$1(tx, second);
          } else {
            let source = second;
            let sourcePath = source.getPath();
            container.hide(source.id);
            tx.update(targetPath, { type: 'insert', start: targetLength, text: source.getText() });
            annotationHelpers$1.transferAnnotations(tx, sourcePath, 0, targetPath, targetLength);
            deleteNode$1(tx, source);
          }
          
          if (third && third.type === first.type) {
            this._mergeTwoLists(tx, container, first, third);
          }
          tx.setSelection({
            type: 'property',
            path: target.getPath(),
            startOffset: targetLength,
            containerId: container.id
          });
        } else if (second.isList()) {
          
          if (direction !== 'right') {
            
            
            throw new Error('Illegal state')
          }
          let item = first.getLastItem();
          this._mergeTwoLists(tx, container, first, second);
          tx.setSelection({
            type: 'property',
            path: item.getPath(),
            startOffset: item.getLength(),
            containerId: container.id
          });
        } else {
          selectNode$1(tx, direction === 'left' ? first.id : second.id, container.id);
        }
      } else {
        if (second.isText() && second.isEmpty()) {
          container.hide(second.id);
          deleteNode$1(tx, second);
          setCursor$1(tx, first, container.id, 'after');
        } else {
          selectNode$1(tx, direction === 'left' ? first.id : second.id, container.id);
        }
      }
    }

    _mergeTwoLists (tx, container, first, second) {
      container.hide(second.id);
      let secondItems = second.getItems().slice();
      for (let i = 0; i < secondItems.length; i++) {
        second.removeItemAt(0);
        first.appendItem(secondItems[i]);
      }
      deleteNode$1(tx, second);
    }
  }

  class EditingInterface$1 {
    constructor (doc, options = {}) {
      this._document = doc;
      this._selection = null;
      
      this._impl = options.editing || new Editing$1();
      this._direction = null;
    }

    dispose () {}

    getDocument () {
      return this._document
    }

    

    get (...args) {
      return this._document.get(...args)
    }

    contains (id) {
      return this._document.contains(id)
    }

    create (nodeData) {
      return this._document.create(nodeData)
    }

    createDefaultTextNode (content) {
      return this._document.createDefaultTextNode(content, this._direction)
    }

    delete (nodeId) {
      return this._document.delete(nodeId)
    }

    set (path, value) {
      return this._document.set(path, value)
    }

    update (path, diffOp) {
      return this._document.update(path, diffOp)
    }

    updateNode (id, newProps) {
      return this._document.updateNode(id, newProps)
    }

    

    createSelection (selData) {
      
      
      
      
      
      
      
      selData = augmentSelection$1(selData, this._selection);
      return this._document.createSelection(selData)
    }

    setSelection (sel) {
      if (!sel) {
        sel = Selection$1.nullSelection;
      } else if (isPlainObject$2(sel)) {
        sel = this.createSelection(sel);
      } else {
        sel = augmentSelection$1(sel, this._selection);
      }
      this._selection = sel;
      return sel
    }

    getSelection () {
      return this._selection
    }

    get selection () {
      return this._selection
    }

    set selection (sel) {
      this.setSelection(sel);
    }

    
    get textDirection () {
      return this._direction
    }

    set textDirection (dir) {
      this._direction = dir;
    }

    

    annotate (annotationData) {
      const sel = this._selection;
      if (sel && (sel.isPropertySelection() || sel.isContainerSelection())) {
        return this._impl.annotate(this, annotationData)
      }
    }

    break () {
      if (this._selection && !this._selection.isNull()) {
        this._impl.break(this);
      }
    }

    copySelection () {
      const sel = this._selection;
      if (sel && !sel.isNull() && !sel.isCollapsed()) {
        return copySelection$1(this.getDocument(), this._selection)
      }
    }

    deleteSelection (options) {
      const sel = this._selection;
      if (sel && !sel.isNull() && !sel.isCollapsed()) {
        this._impl.delete(this, 'right', options);
      }
    }

    deleteCharacter (direction) {
      const sel = this._selection;
      if (!sel || sel.isNull()) ; else if (!sel.isCollapsed()) {
        this.deleteSelection();
      } else {
        this._impl.delete(this, direction);
      }
    }

    insertText (text) {
      const sel = this._selection;
      if (sel && !sel.isNull()) {
        this._impl.insertText(this, text);
      }
    }

    
    insertInlineNode (inlineNode) {
      const sel = this._selection;
      if (sel && !sel.isNull() && sel.isPropertySelection()) {
        return this._impl.insertInlineNode(this, inlineNode)
      }
    }

    insertBlockNode (blockNode) {
      const sel = this._selection;
      if (sel && !sel.isNull()) {
        return this._impl.insertBlockNode(this, blockNode)
      }
    }

    paste (content) {
      const sel = this._selection;
      if (sel && !sel.isNull() && !sel.isCustomSelection()) {
        return this._impl.paste(this, content)
      }
    }

    switchTextType (nodeData) {
      const sel = this._selection;
      if (sel && !sel.isNull()) {
        return this._impl.switchTextType(this, nodeData)
      }
    }

    toggleList (params) {
      const sel = this._selection;
      if (sel && !sel.isNull()) {
        return this._impl.toggleList(this, params)
      }
    }

    indent () {
      const sel = this._selection;
      if (sel && !sel.isNull()) {
        return this._impl.indent(this)
      }
    }

    dedent () {
      const sel = this._selection;
      if (sel && !sel.isNull()) {
        return this._impl.dedent(this)
      }
    }

    

    getIndex (...args) {
      return this._document.getIndex(...args)
    }

    getAnnotations (...args) {
      return this._document.getAnnotations(...args)
    }

    getSchema () {
      return this._document.getSchema()
    }

    createSnippet () {
      return this._document.createSnippet()
    }
  }

  class JSONConverter$1 {
    importDocument (doc, json) {
      if (!json.nodes) {
        throw new Error('Invalid JSON format.')
      }
      var schema = doc.getSchema();
      if (json.schema && schema.name !== json.schema.name) {
        throw new Error('Incompatible schema.')
      }
      
      var nodes = json.nodes;
      
      
      
      doc.import(function (tx) {
        forEach$1(nodes, function (node) {
          
          if (tx.get(node.id)) {
            tx.delete(node.id);
          }
          tx.create(node);
        });
      });
      return doc
    }

    exportDocument (doc) {
      var schema = doc.getSchema();
      var json = {
        schema: {
          name: schema.name
        },
        nodes: {}
      };
      forEach$1(doc.getNodes(), function (node) {
        if (node._isDocumentNode) {
          json.nodes[node.id] = node.toJSON();
        }
      });
      return json
    }
  }

  class ParentNodeHook$1 {
    constructor (doc) {
      this.doc = doc;
      
      this.parents = {};
      doc.data.on('operation:applied', this._onOperationApplied, this);
    }

    _onOperationApplied (op) {
      const doc = this.doc;
      let node = doc.get(op.path[0]);
      let nodeSchema, hasOwnedProperties;
      if (node) {
        nodeSchema = node.getSchema();
        hasOwnedProperties = nodeSchema.hasOwnedProperties();
      }
      
      
      switch (op.type) {
        case 'create': {
          if (hasOwnedProperties) {
            nodeSchema.getOwnedProperties().forEach(p => {
              let refs = node[p.name];
              if (refs) {
                this._setParent(node, refs);
              }
              this._setRegisteredParent(node);
            });
          }
          break
        }
        case 'update': {
          if (hasOwnedProperties) {
            let propName = op.path[1];
            if (nodeSchema.isOwned(propName)) {
              let update = op.diff;
              if (update.isDelete()) {
                this._setParent(null, update.getValue());
              } else {
                this._setParent(node, update.getValue());
              }
            }
          }
          break
        }
        case 'set': {
          if (hasOwnedProperties) {
            let propName = op.path[1];
            if (nodeSchema.isOwned(propName)) {
              let oldValue = op.getOldValue();
              let newValue = op.getValue();
              
              this._setParent(null, oldValue);
              this._setParent(node, newValue);
            }
          }
          break
        }
        default:
          
      }
    }

    _setParent (parent, ids) {
      if (ids) {
        if (isArray$2(ids)) {
          ids.forEach(id => this.__setParent(parent, id));
        } else {
          this.__setParent(parent, ids);
        }
      }
    }

    __setParent (parent, id) {
      
      
      
      
      
      
      this.parents[id] = parent;
      let child = this.doc.get(id);
      if (child) {
        child.setParent(parent);
      }
    }

    _setRegisteredParent (child) {
      let parent = this.parents[child.id];
      if (parent) {
        child.setParent(parent);
      }
    }
  }

  ParentNodeHook$1.register = function (doc) {
    return new ParentNodeHook$1(doc)
  };

  const converter$1 = new JSONConverter$1();



  class Document$1 extends EventEmitter$1 {
    
    constructor (schema, ...args) {
      super();

      this.schema = schema;
      
      if (!schema) {
        throw new Error('A document needs a schema for reflection.')
      }

      
      this._ops = [];

      this._initialize(...args);
    }

    _initialize () {
      this.__id__ = uuid$1();
      this.nodeFactory = new DocumentNodeFactory$1(this);
      this.data = new IncrementalData$1(this.schema, this.nodeFactory);
      
      this.addIndex('type', new PropertyIndex$1('type'));
      
      this.addIndex('annotations', new AnnotationIndex$1());
      
      
      
      this.addIndex('container-annotations', new ContainerAnnotationIndex$1());
      
      
      ParentNodeHook$1.register(this);
    }

    dispose () {
      this.off();
      this.data.off();
    }

    get id () {
      return this.__id__
    }

    
    getSchema () {
      return this.schema
    }

    
    contains (id) {
      return this.data.contains(id)
    }

    
    get (path, strict) {
      return this.data.get(path, strict)
    }

    
    getNodes () {
      return this.data.getNodes()
    }

    getAnnotations (path) {
      return this.getIndex('annotations').get(path)
    }

    
    import (importer) {
      try {
        this.data._stopIndexing();
        importer(this);
        this.data._startIndexing();
      } finally {
        this.data.queue = [];
        this.data._startIndexing();
      }
    }

    
    create (nodeData) {
      if (!nodeData.id) {
        nodeData.id = uuid$1(nodeData.type);
      }
      if (!nodeData.type) {
        throw new Error('No node type provided')
      }
      const op = this._create(nodeData);
      if (op) {
        this._ops.push(op);
        if (!this._isTransactionDocument) {
          this._emitChange(op);
        }
        return this.get(nodeData.id)
      }
    }

    createDefaultTextNode (text, dir) {
      return this.create({
        type: this.getSchema().getDefaultTextType(),
        content: text || '',
        direction: dir
      })
    }

    
    delete (nodeId) {
      const node = this.get(nodeId);
      const op = this._delete(nodeId);
      if (op) {
        this._ops.push(op);
        if (!this._isTransactionDocument) {
          this._emitChange(op);
        }
      }
      return node
    }

    
    set (path, value) {
      const oldValue = this.get(path);
      const op = this._set(path, value);
      if (op) {
        this._ops.push(op);
        if (!this._isTransactionDocument) {
          this._emitChange(op);
        }
      }
      return oldValue
    }

    
    update (path, diff) {
      const op = this._update(path, diff);
      if (op) {
        this._ops.push(op);
        if (!this._isTransactionDocument) {
          this._emitChange(op);
        }
      }
      return op
    }

    
    updateNode (id, newProps) {
      let node = this.get(id);
      forEach$1(newProps, (value, key) => {
        if (!isEqual$1(node[key], newProps[key])) {
          this.set([id, key], value);
        }
      });
    }

    
    addIndex (name, index) {
      return this.data.addIndex(name, index)
    }

    
    getIndex (name) {
      return this.data.getIndex(name)
    }

    
    createSelection (data) {
      let sel;
      if (isNil$1(data)) return Selection$1.nullSelection
      if (arguments.length !== 1 || !isPlainObject$2(data)) {
        throw new Error('Illegal argument: call createSelection({ type: ... }')
      } else {
        switch (data.type) {
          case 'property': {
            if (isNil$1(data.endOffset)) {
              data.endOffset = data.startOffset;
            }
            if (!data.hasOwnProperty('reverse')) {
              if (data.startOffset > data.endOffset) {
                [data.startOffset, data.endOffset] = [data.endOffset, data.startOffset];
                data.reverse = !data.reverse;
              }
            }
            
            let text = this.get(data.path, 'strict');
            if (data.startOffset < 0 || data.startOffset > text.length) {
              throw new Error('Invalid startOffset: target property has length ' + text.length + ', given startOffset is ' + data.startOffset)
            }
            if (data.endOffset < 0 || data.endOffset > text.length) {
              throw new Error('Invalid startOffset: target property has length ' + text.length + ', given endOffset is ' + data.endOffset)
            }
            sel = new PropertySelection$1(data);
            break
          }
          case 'container': {
            let container = this.get(data.containerId, 'strict');
            if (!container) throw new Error('Can not create ContainerSelection: container "' + data.containerId + '" does not exist.')
            let start = this._normalizeCoor({ path: data.startPath, offset: data.startOffset });
            let end = this._normalizeCoor({ path: data.endPath, offset: data.endOffset });
            let startAddress = container.getAddress(start);
            let endAddress = container.getAddress(end);
            if (!startAddress) {
              throw new Error('Invalid arguments for ContainerSelection: ', start.toString())
            }
            if (!endAddress) {
              throw new Error('Invalid arguments for ContainerSelection: ', end.toString())
            }
            if (!data.hasOwnProperty('reverse')) {
              if (endAddress.isBefore(startAddress, 'strict')) {
                [start, end] = [end, start];
                data.reverse = true;
              }
            }
            sel = new ContainerSelection$1(container.id, start.path, start.offset, end.path, end.offset, data.reverse, data.surfaceId);
            break
          }
          case 'node': {
            sel = createNodeSelection$1({
              doc: this,
              nodeId: data.nodeId,
              mode: data.mode,
              containerId: data.containerId,
              reverse: data.reverse,
              surfaceId: data.surfaceId
            });
            break
          }
          case 'custom': {
            sel = CustomSelection$1.fromJSON(data);
            break
          }
          default:
            throw new Error('Illegal selection type', data)
        }
      }
      if (!sel.isNull()) {
        sel.attach(this);
      }
      return sel
    }

    newInstance () {
      var DocumentClass = this.constructor;
      return new DocumentClass(this.schema)
    }

    
    createSnippet () {
      var snippet = this.newInstance();
      var snippetContainer = snippet.create({
        type: '@container',
        id: SNIPPET_ID$1
      });
      snippet.getContainer = function () {
        return snippetContainer
      };
      return snippet
    }

    createFromDocument (doc) {
      
      this.clear();

      
      
      
      
      
      
      
      let nodes = Object.values(doc.getNodes());
      let levels = {};
      let visited = new Set();
      nodes.forEach(n => {
        if (!visited.has(n)) this._computeDependencyLevel(n, levels, visited);
      });
      
      nodes.sort((a, b) => {
        return levels[b.id] - levels[a.id]
      });
      nodes.forEach(n => this.create(n));
      return this
    }

    _computeDependencyLevel (node, levels, visited) {
      if (!node) throw new Error('node was nil')
      if (visited.has(node)) throw new Error('Cyclic node dependency')
      visited.add(node);
      
      
      
      let level = 0;
      if (node.isAnnotation() || node.isInline()) {
        level = -1;
      } else {
        let parent = node.getParent();
        if (parent) {
          let parentLevel;
          if (levels.hasOwnProperty(parent.id)) {
            parentLevel = levels[parent.id];
          } else {
            parentLevel = this._computeDependencyLevel(parent, levels, visited);
          }
          level = parentLevel + 1;
        }
      }
      levels[node.id] = level;
      return level
    }

    
    toJSON () {
      return converter$1.exportDocument(this)
    }

    clone () {
      let copy = this.newInstance();
      copy.createFromDocument(this);
      return copy
    }

    clear () {
      this.data.clear();
      this._ops.length = 0;
    }

    
    createEditingInterface () {
      return new EditingInterface$1(this)
    }

    invert (change) {
      return change.invert()
    }

    _apply (documentChange) {
      forEach$1(documentChange.ops, (op) => {
        this._applyOp(op);
      });
      
      documentChange._extractInformation(this);
    }

    _applyOp (op) {
      this.data.apply(op);
      this.emit('operation:applied', op);
    }

    _create (nodeData) {
      return this.data.create(nodeData)
    }

    _delete (nodeId) {
      return this.data.delete(nodeId)
    }

    _set (path, value) {
      return this.data.set(path, value)
    }

    _update (path, diff) {
      return this.data.update(path, diff)
    }

    _emitChange (op) {
      const change = new DocumentChange$1([op], {}, {});
      change._extractInformation(this);
      this._notifyChangeListeners(change, { hidden: true });
    }

    _notifyChangeListeners (change, info) {
      info = info || {};
      this.emit('document:changed', change, info, this);
    }

    
    _createSelectionFromRange (range) {
      if (!range) return Selection$1.nullSelection
      let inOneNode = isEqual$1(range.start.path, range.end.path);
      if (inOneNode) {
        if (range.start.isNodeCoordinate()) {
          
          
          return new NodeSelection$1(range.containerId, range.start.getNodeId(), 'full', range.reverse, range.surfaceId)
        } else {
          return this.createSelection({
            type: 'property',
            path: range.start.path,
            startOffset: range.start.offset,
            endOffset: range.end.offset,
            reverse: range.reverse,
            containerId: range.containerId,
            surfaceId: range.surfaceId
          })
        }
      } else {
        return this.createSelection({
          type: 'container',
          startPath: range.start.path,
          startOffset: range.start.offset,
          endPath: range.end.path,
          endOffset: range.end.offset,
          reverse: range.reverse,
          containerId: range.containerId,
          surfaceId: range.surfaceId
        })
      }
    }

    _normalizeCoor ({ path, offset }) {
      
      if (path.length === 1) {
        let node = this.get(path[0]).getContainerRoot();
        if (node.isText()) {
          
          return new Coordinate$1(node.getPath(), offset === 0 ? 0 : node.getLength())
        } else if (node.isList()) {
          
          if (offset === 0) {
            let item = node.getItemAt(0);
            return new Coordinate$1(item.getPath(), 0)
          } else {
            let item = this.get(last$1$1(node.items));
            return new Coordinate$1(item.getPath(), item.getLength())
          }
        }
      }
      return new Coordinate$1(path, offset)
    }
  }

  Document$1.prototype._isDocument = true;

  class XMLParentNodeHook$1 {
    constructor (doc) {
      this.doc = doc;
      
      this.parents = {};
      doc.data.on('operation:applied', this._onOperationApplied, this);
    }

    _onOperationApplied (op) {
      const doc = this.doc;
      const parents = this.parents;
      let node = doc.get(op.path[0]);
      switch (op.type) {
        case 'create': {
          if (node._childNodes) {
            _setParent(node, node._childNodes);
          }
          _setRegisteredParent(node);
          break
        }
        case 'update': {
          
          
          let update = op.diff;
          if (op.path[1] === '_childNodes') {
            if (update.isInsert()) {
              _setParent(node, update.getValue());
            } else if (update.isDelete()) {
              _setParent(null, update.getValue());
            }
          }
          break
        }
        case 'set': {
          if (op.path[1] === '_childNodes') {
            _setParent(null, op.getOldValue());
            _setParent(node, op.getValue());
          }
          break
        }
        default:
          
      }

      function _setParent (parent, ids) {
        if (ids) {
          if (isArray$2(ids)) {
            ids.forEach(_set);
          } else {
            _set(ids);
          }
        }
        function _set (id) {
          
          
          
          parents[id] = parent;
          let child = doc.get(id);
          if (child) {
            child.parentNode = parent;
          }
        }
      }
      function _setRegisteredParent (child) {
        let parent = parents[child.id];
        if (parent) {
          child.parentNode = parent;
        }
      }
    }

    static register (doc) {
      return new XMLParentNodeHook$1(doc)
    }
  }

  class XMLEditingInterface$1 extends EditingInterface$1 {
    find (cssSelector) {
      return this.getDocument().find(cssSelector)
    }

    findAll (cssSelector) {
      return this.getDocument().findAll(cssSelector)
    }

    createElement (...args) {
      return this.getDocument().createElement(...args)
    }
  }

  class XMLDocument$1 extends Document$1 {
    _initialize () {
      this.nodeFactory = new DocumentNodeFactory$1(this);
      this.data = new IncrementalData$1(this.schema, this.nodeFactory);
      
      this.addIndex('type', new PropertyIndex$1('type'));
      
      this.addIndex('annotations', new AnnotationIndex$1());
      XMLParentNodeHook$1.register(this);
    }

    toXML () {
      let dom = DefaultDOMElement$1.createDocument('xml');
      dom.setDoctype(...this.getDocTypeParams());
      let xml = dom.createProcessingInstruction('xml', 'version="1.0" encoding="UTF-8"');
      dom.insertAt(0, xml);
      let rootElement = this.getRootNode().toXML();
      dom.append(rootElement);
      return dom
    }

    getDocTypeParams () {
      
      throw new Error('This method is abstract')
    }

    getXMLSchema () {
      
      throw new Error('This method is abstract')
    }

    getRootNode () {
      
      throw new Error('This method is abstract')
    }

    
    getDocTypeAsString () {
      return new Error('This method is abstract')
    }

    createEditingInterface () {
      return new XMLEditingInterface$1(this)
    }

    find (cssSelector) {
      return this.getRootNode().find(cssSelector)
    }

    findAll (cssSelector) {
      return this.getRootNode().findAll(cssSelector)
    }

    createElement (tagName, data) {
      let node = this.create(Object.assign({
        id: uuid$1(tagName),
        type: tagName
      }, data));
      return node
    }

    getElementSchema (type) {
      return this.getXMLSchema().getElementSchema(type)
    }

    _validateChange (change) {
      let changed = {};
      let deleted = [];
      change.ops.forEach((op) => {
        switch (op.type) {
          case 'delete': {
            deleted.push(op.val.id);
            break
          }
          case 'create': {
            changed[op.val.id] = true;
            break
          }
          default: {
            if (op.path) {
              changed[op.path[0]] = true;
            }
          }
        }
      });
      
      deleted.forEach(id => delete changed[id]);

      const xmlSchema = this.getXMLSchema();
      let errors = [];
      Object.keys(changed).forEach((id) => {
        let node = this.get(id);
        let res = xmlSchema.validateElement(node);
        if (!res.ok) {
          errors = errors.concat(res.errors);
        }
      });
      return {
        ok: errors.length === 0,
        errors
      }
    }

    
    _apply (documentChange) {
      
      if (!documentChange.ops) return

      let created = {};
      let deleted = {};
      let updated = {};

      const doc = this;

      function _recordUpdate (id, op) {
        let record = updated[id];
        if (!record) {
          record = updated[id] = { ops: [] };
        }
        if (op) record.ops.push(op);
        return record
      }

      
      function _checkAnnotation (op) {
        
        switch (op.type) {
          case 'create':
          case 'delete': {
            const annoData = op.val;
            if (annoData.hasOwnProperty('start')) {
              updated[annoData.start.path] = true;
              let node = doc.get(annoData.start.path[0]);
              if (node) {
                _recordUpdate(node.id, op);
              }
            }
            if (annoData.hasOwnProperty('end')) {
              updated[annoData.end.path] = true;
              let node = doc.get(annoData.start.path[0]);
              if (node) {
                _recordUpdate(node.id, op);
              }
            }
            break
          }
          case 'update':
          case 'set': {
            let anno = doc.get(op.path[0]);
            if (anno) {
              if (anno.isPropertyAnnotation()) {
                updated[anno.start.path] = true;
                let node = doc.get(anno.start.path[0]);
                if (node) {
                  _recordUpdate(node.id, op);
                }
              } else if (anno.isContainerAnnotation()) ;
            }
            break
          }
          default:
            
        }
      }

      documentChange.ops.forEach((op) => {
        
        
        
        switch (op.type) {
          case 'create': {
            created[op.path[0]] = op.val;
            break
          }
          case 'delete': {
            let node = this.get(op.path[0]);
            deleted[node.id] = node;
            break
          }
          case 'update':
          case 'set': {
            updated[op.path] = true;
            let node = this.get(op.path[0]);
            _recordUpdate(node.id, op);
            break
          }
          default:
            
        }
        
        _checkAnnotation(op);

        this._applyOp(op);
      });
      
      Object.keys(deleted).forEach((id) => {
        delete updated[id];
      });
      documentChange.created = created;
      documentChange.updated = updated;
      documentChange.deleted = deleted;
      documentChange._extracted = true;
    }
  }

  XMLDocument$1.prototype._isXMLDocument = true;

  function findParentDOMElement$1 (nativeEl) {
    while (nativeEl) {
      let el = DefaultDOMElement$1.unwrap(nativeEl);
      if (el) return el
      nativeEl = nativeEl.parentNode;
    }
  }

  function findParent$1 (el, selector) {
    while (el) {
      if (el.is(selector)) return el
      el = el.getParent();
    }
  }

  function stop$1 (event) {
    event.stopPropagation();
  }

  function stopAndPrevent$1 (event) {
    event.stopPropagation();
    event.preventDefault();
  }

  function walk$1 (el, cb) {
    _walk$1(el, cb, 0);
  }

  function _walk$1 (el, cb, level) {
    cb(el, level);
    if (el.getChildCount() > 0) {
      let it = el.getChildNodeIterator();
      while (it.hasNext()) {
        _walk$1(it.next(), cb, level + 1);
      }
    }
  }

  function isRightButton$1 (event) {
    let isRightButton = false;
    if ('which' in event) {
      isRightButton = (event.which === 3);
    } else if ('button' in event) {
      isRightButton = (event.button === 2);
    }
    return isRightButton
  }

  function getBoundingRect$1 (el) {
    let _rect = el.getNativeElement().getBoundingClientRect();
    return {
      top: _rect.top,
      left: _rect.left,
      height: _rect.height,
      width: _rect.width
    }
  }

  function getBoundingRectForRects$1 (...rects) {
    let top, left, bottom, right;
    if (rects.length > 0) {
      let first = rects[0];
      top = first.top;
      left = first.left;
      bottom = top + first.height;
      right = left + first.width;
      for (let i = 1; i < rects.length; i++) {
        let r = rects[i];
        top = Math.min(top, r.top);
        left = Math.min(left, r.left);
        bottom = Math.max(bottom, r.top + r.height);
        right = Math.max(right, r.left + r.width);
      }
    }
    return {
      top,
      left,
      right,
      bottom,
      height: bottom - top,
      width: right - left
    }
  }

  function isXInside$1 (x, rect) {
    return x >= rect.left && x <= rect.left + rect.width
  }

  function isYInside$1 (y, rect) {
    return y >= rect.top && y <= rect.top + rect.height
  }

  var domHelpers$1 = /*#__PURE__*/Object.freeze({
    findParentDOMElement: findParentDOMElement$1,
    findParent: findParent$1,
    stop: stop$1,
    stopAndPrevent: stopAndPrevent$1,
    walk: walk$1,
    isRightButton: isRightButton$1,
    getBoundingRect: getBoundingRect$1,
    getBoundingRectForRects: getBoundingRectForRects$1,
    isXInside: isXInside$1,
    isYInside: isYInside$1
  });

  function encodeXMLEntities$1 (str) {
    return String(str)
      .replace(/&/g, '&amp;')
      .replace(/</g, '&lt;')
      .replace(/>/g, '&gt;')
      .replace(/"/g, '&quot;')
  }

  class DOMExporter$1 {
    constructor (config, context) {
      this.context = context || {};
      if (!config.converters) {
        throw new Error('config.converters is mandatory')
      }
      if (!config.converters._isRegistry) {
        this.converters = new Registry$1();
        config.converters.forEach(function (Converter) {
          let converter = isFunction$1$1(Converter) ? new Converter() : Converter;
          if (!converter.type) {
            console.error('Converter must provide the type of the associated node.', converter);
            return
          }
          this.converters.add(converter.type, converter);
        }.bind(this));
      } else {
        this.converters = config.converters;
      }

      this.state = {
        doc: null
      };
      this.config = config;
      
      
      this._elementFactory = config.elementFactory;
      if (!this._elementFactory) {
        throw new Error("'elementFactory' is mandatory")
      }
      this.$$ = this.createElement.bind(this);
    }

    exportDocument (doc) {
      
      
      
      
      return this.convertDocument(doc)
    }

    
    convertDocument(doc) { 
      throw new Error('This method is abstract')
    }

    convertContainer (container) {
      if (!container) {
        throw new Error('Illegal arguments: container is mandatory.')
      }
      const doc = container.getDocument();
      this.state.doc = doc;
      let elements = [];
      container.getContent().forEach((id) => {
        const node = doc.get(id);
        const nodeEl = this.convertNode(node);
        elements.push(nodeEl);
      });
      return elements
    }

    convertNode (node) {
      if (isString$1(node)) {
        
        node = this.state.doc.get(node);
      } else {
        this.state.doc = node.getDocument();
      }
      var converter = this.getNodeConverter(node);
      
      
      if (node.isPropertyAnnotation() && (!converter || !converter.export)) {
        return this._convertPropertyAnnotation(node)
      }
      if (!converter) {
        converter = this.getDefaultBlockConverter();
      }
      var el;
      if (converter.tagName) {
        el = this.$$(converter.tagName);
      } else {
        el = this.$$('div');
      }
      el.attr(this.config.idAttribute, node.id);
      if (converter.export) {
        el = converter.export(node, el, this) || el;
      } else {
        el = this.getDefaultBlockConverter().export(node, el, this) || el;
      }
      return el
    }

    convertProperty (doc, path, options) {
      this.initialize(doc, options);
      var wrapper = this.$$('div')
        .append(this.annotatedText(path));
      return wrapper.innerHTML
    }

    annotatedText (path) {
      var doc = this.state.doc;
      var text = doc.get(path);
      var annotations = doc.getIndex('annotations').get(path);
      return this._annotatedText(text, annotations)
    }

    getNodeConverter (node) {
      return this.converters.get(node.type)
    }

    getDefaultBlockConverter () {
      throw new Error('This method is abstract.')
    }

    getDefaultPropertyAnnotationConverter () {
      throw new Error('This method is abstract.')
    }

    getDocument () {
      return this.state.doc
    }

    createElement (str) {
      return this._elementFactory.createElement(str)
    }

    _annotatedText (text, annotations) {
      var self = this;

      var annotator = new Fragmenter$1();
      annotator.onText = function (context, text) {
        context.children.push(encodeXMLEntities$1(text));
      };
      annotator.onEnter = function (fragment) {
        var anno = fragment.node;
        return {
          annotation: anno,
          children: []
        }
      };
      annotator.onExit = function (fragment, context, parentContext) {
        var anno = context.annotation;
        var converter = self.getNodeConverter(anno);
        if (!converter) {
          converter = self.getDefaultPropertyAnnotationConverter();
        }
        var el;
        if (converter.tagName) {
          el = this.$$(converter.tagName);
        } else {
          el = this.$$('span');
        }
        el.attr(this.config.idAttribute, anno.id);
        el.append(context.children);
        if (converter.export) {
          el = converter.export(anno, el, self) || el;
        }
        parentContext.children.push(el);
      }.bind(this);
      var wrapper = { children: [] };
      annotator.start(wrapper, text, annotations);
      return wrapper.children
    }

    
    _convertPropertyAnnotation (anno) {
      
      var wrapper = this.$$('div').append(this.annotatedText(anno.path));
      var el = wrapper.find('[' + this.config.idAttribute + '="' + anno.id + '"]');
      return el
    }
  }

  class HTMLExporter$1 extends DOMExporter$1 {
    constructor (config, context) {
      super(_defaultConfig$2(config), context);
    }

    exportDocument (doc) {
      let htmlEl = DefaultDOMElement$1.parseHTML('<html><head></head><body></body></html>');
      return this.convertDocument(doc, htmlEl)
    }

    getDefaultBlockConverter () {
      return defaultBlockConverter$2 
    }

    getDefaultPropertyAnnotationConverter () {
      return defaultAnnotationConverter$2 
    }
  }

  function _defaultConfig$2 (config) {
    config = Object.assign({
      idAttribute: 'data-id'
    }, config);
    if (!config.elementFactory) {
      config.elementFactory = DefaultDOMElement$1.createDocument('html');
    }
    return config
  }

  const defaultAnnotationConverter$2 = {
    tagName: 'span',
    export: function (node, el) {
      el.tagName = 'span';
      el.attr('data-type', node.type);
      var properties = node.toJSON();
      forEach$1(properties, function (value, name) {
        if (name === 'id' || name === 'type') return
        if (isString$1(value) || isNumber$1(value) || isBoolean$1(value)) {
          el.attr('data-' + name, value);
        }
      });
    }
  };

  const defaultBlockConverter$2 = {
    export: function (node, el, converter) {
      el.attr('data-type', node.type);
      var properties = node.toJSON();
      forEach$1(properties, function (value, name) {
        if (name === 'id' || name === 'type') {
          return
        }
        var prop = converter.$$('div').attr('property', name);
        if (node.getPropertyType(name) === 'string' && value) {
          prop.append(converter.annotatedText([node.id, name]));
        } else {
          prop.text(value);
        }
        el.append(prop);
      });
    }
  };

  class ClipboardExporter$1 extends HTMLExporter$1 {
    
    exportDocument (doc) {
      this.state.doc = doc;
      let html;
      let elements = this.convertDocument(doc);
      
      if (elements.length === 1 && elements[0].attr('data-id') === TEXT_SNIPPET_ID$1) {
        html = elements[0].innerHTML;
      } else {
        html = elements.map(function (el) {
          return el.outerHTML
        }).join('');
      }
      let jsonConverter = new JSONConverter$1();
      let jsonStr = JSON.stringify(jsonConverter.exportDocument(doc));
      let substanceContent = `<script id="substance-clipboard" type="application/json">${jsonStr}</script>`;
      return '<html><head>' + substanceContent + '</head><body>' + html + '</body></html>'
    }

    
    convertDocument (doc) {
      let content = doc.get(SNIPPET_ID$1);
      if (!content) {
        throw new Error('Illegal clipboard document: could not find container "' + SNIPPET_ID$1 + '"')
      }
      return this.convertContainer(content)
    }
  }

  class HTMLImporter$1 extends DOMImporter$1 {
    constructor (config) {
      super(Object.assign({ idAttribute: 'data-id' }, config));

      
      this._el = DefaultDOMElement$1.parseHTML('<html></html>');
    }

    importDocument (html) {
      this.reset();
      var parsed = DefaultDOMElement$1.parseHTML(html);
      this.convertDocument(parsed);
      return this.state.doc
    }

    
    convertDocument(documentEl) { 
      throw new Error('This method is abstract')
    }
  }

  const INLINENODES$1 = ['a', 'b', 'big', 'i', 'small', 'tt', 'abbr', 'acronym', 'cite', 'code', 'dfn', 'em', 'kbd', 'strong', 'samp', 'time', 'var', 'bdo', 'br', 'img', 'map', 'object', 'q', 'script', 'span', 'sub', 'sup', 'button', 'input', 'label', 'select', 'textarea'].reduce((m, n) => { m[n] = true; return m }, {});


  class ClipboardImporter$1 extends HTMLImporter$1 {
    constructor (config) {
      super(_withCatchAllConverter$1(config));
      
      this.IGNORE_DEFAULT_WARNINGS = true;

      Object.assign(config, {
        trimWhitespaces: true,
        REMOVE_INNER_WS: true
      });

      this.editorOptions = config.editorOptions;
    }

    
    importDocument (html) {
      if (platform$1.isWindows) {
        
        
        let match = /<!--StartFragment-->(.*)<!--EndFragment-->/.exec(html);
        if (match) {
          html = match[1];
        }
      }

      
      
      
      if (html.search(/script id=.substance-clipboard./) >= 0) {
        let htmlDoc = DefaultDOMElement$1.parseHTML(html);
        let substanceData = htmlDoc.find('#substance-clipboard');
        if (substanceData) {
          let jsonStr = substanceData.textContent;
          try {
            return this.importFromJSON(jsonStr)
          } finally {
            
          }
        }
      }

      if (this.editorOptions && this.editorOptions['forcePlainTextPaste']) {
        return null
      }

      let htmlDoc = DefaultDOMElement$1.parseHTML(html);
      let generatorMeta = htmlDoc.find('meta[name="generator"]');
      let xmnlsw = htmlDoc.find('html').getAttribute('xmlns:w');
      if (generatorMeta) {
        let generator = generatorMeta.getAttribute('content');
        if (generator.indexOf('LibreOffice') > -1) this._isLibreOffice = true;
      } else if (xmnlsw) {
        if (xmnlsw.indexOf('office:word') > -1) {
          this._isMicrosoftWord = true;
          
          
          
          let match = /<!--StartFragment-->([\s\S]*)<!--EndFragment-->/.exec(html);
          if (match) {
            htmlDoc = DefaultDOMElement$1.parseHTML(match[1]);
          }
        }
      } else if (html.indexOf('docs-internal-guid') > -1) {
        this._isGoogleDoc = true;
      }

      let body = htmlDoc.find('body');
      body = this._sanitizeBody(body);
      if (!body) {
        console.warn('Invalid HTML.');
        return null
      }
      this._wrapIntoParagraph(body);
      this.reset();
      this.convertBody(body);
      const doc = this.state.doc;
      return doc
    }

    _sanitizeBody (body) {
      
      body.findAll('meta').forEach(el => el.remove());

      
      
      if (this._isLibreOffice || this._isMicrosoftWord) {
        let bodyHtml = body.getInnerHTML();
        body.setInnerHTML(bodyHtml.replace(/\r\n|\r|\n/g, ' '));
      } else if (this._isGoogleDoc) {
        body = this._fixupGoogleDocsBody(body);
      }

      return body
    }

    _fixupGoogleDocsBody (body) {
      if (!body) return
      
      
      
      
      let bold = body.find('b');
      if (bold && /^docs-internal/.exec(bold.id)) {
        body = bold;
      }

      body.findAll('span').forEach(span => {
        
        
        
        
        
        
        
        
        let nodeTypes = [];
        if (span.getStyle('font-weight') === '700') nodeTypes.push('b');
        if (span.getStyle('font-style') === 'italic') nodeTypes.push('i');
        if (span.getStyle('vertical-align') === 'super') nodeTypes.push('sup');
        if (span.getStyle('vertical-align') === 'sub') nodeTypes.push('sub');
        
        span.removeAttribute('style');

        function createInlineNodes (parentEl, isRoot) {
          if (nodeTypes.length > 0) {
            let el = parentEl.createElement(nodeTypes[0]);
            if (nodeTypes.length === 1) el.append(span.textContent);
            if (isRoot) {
              parentEl.replaceChild(span, el);
            } else {
              parentEl.appendChild(el);
            }
            nodeTypes.shift();
            createInlineNodes(el);
          }
        }
        createInlineNodes(span.getParent(), true);
      });

      let tags = ['b', 'i', 'sup', 'sub'];

      tags.forEach(tag => {
        body.findAll(tag).forEach(el => {
          
          
          let previousSiblingEl = el.getPreviousSibling();
          if (previousSiblingEl && el.tagName === previousSiblingEl.tagName) {
            let parentEl = el.getParent();
            let newEl = parentEl.createElement(tag);
            newEl.setInnerHTML(previousSiblingEl.getInnerHTML() + el.getInnerHTML());
            parentEl.replaceChild(el, newEl);
            parentEl.removeChild(previousSiblingEl);
          }

          
          
          
          
          if (previousSiblingEl && previousSiblingEl.tagName && el.getChildCount() > 0 && el.getChildAt(0).tagName === previousSiblingEl.tagName) {
            let parentEl = el.getParent();
            let childEl = el.getChildAt(0);
            let newEl = parentEl.createElement(previousSiblingEl.tagName);
            let newChildEl = newEl.createElement(tag);
            newChildEl.setTextContent(childEl.textContent);
            newEl.appendChild(newChildEl);
            parentEl.replaceChild(el, newEl);
          }
        });
      });

      return body
    }

    _wrapIntoParagraph (body) {
      let childNodes = body.getChildNodes();
      let shouldWrap = false;
      for (let i = 0; i < childNodes.length; i++) {
        const c = childNodes[i];
        if (c.isTextNode()) {
          if (!(/^\s+$/.exec(c.textContent))) {
            shouldWrap = true;
            break
          }
        } else if (INLINENODES$1[c.tagName]) {
          shouldWrap = true;
          break
        }
      }
      if (shouldWrap) {
        let p = body.createElement('p');
        p.append(childNodes);
        body.append(p);
      }
    }

    importFromJSON (jsonStr) {
      this.reset();
      let doc = this.getDocument();
      let jsonData = JSON.parse(jsonStr);
      let converter = new JSONConverter$1();
      converter.importDocument(doc, jsonData);
      return doc
    }

    
    convertBody (body) {
      this.convertContainer(body.childNodes, SNIPPET_ID$1);
    }

    
    _createDocument () {
      let emptyDoc = super._createDocument();
      return emptyDoc.createSnippet()
    }
  }

  function _withCatchAllConverter$1 (config) {
    config = Object.assign({}, config);
    let defaultTextType = config.schema.getDefaultTextType();
    if (defaultTextType) {
      config.converters = config.converters.concat([{
        type: defaultTextType,
        matchElement: function (el) { return el.is('div') },
        import: function (el, node, converter) {
          node.content = converter.annotatedText(el, [node.id, 'content']);
        }
      }]);
    }
    return config
  }

  class ContainerAdapter$1 extends Container$1 {
    constructor (doc, path) {
      super(doc, { id: String(path) });
      this.document = doc;
      this.path = path;

      
      
      doc.data.nodes[this.id] = this;
    }

    getContentPath () {
      return this.path
    }

    get nodes () {
      return this.document.get(this.path)
    }
  }

  ContainerAdapter$1.prototype._isDocumentNode = false;
  ContainerAdapter$1.prototype._isContainer = false;

  class FileNode$1 extends DocumentNode$1 {
    getUrl () {
      if (this.proxy) {
        return this.proxy.getUrl()
      } else {
        
        console.warn('No file proxy attached to ', this.id);
        return ''
      }
    }

    setProxy (proxy) {
      this.proxy = proxy;
    }
  }

  FileNode$1.type = 'file';

  FileNode$1.schema = {
    url: { type: 'string', optional: true },
    fileType: { type: 'string', optional: true },
    mimeType: { type: 'string', optional: true },
    sourceFile: { type: 'object', optional: true }
  };

  FileNode$1.prototype._isFileNode = true;
  FileNode$1._isFileNode = true;

  class InlineNode$1 extends PropertyAnnotation$1 {}

  InlineNode$1.prototype._isInlineNode = true;

  InlineNode$1.isInline = true;

  class Range$1 {
    constructor (start, end, reverse, containerId, surfaceId) {
      
      if (arguments[0] === 'SKIP') return
      if (arguments.length === 1 && isPlainObject$2(arguments[0])) {
        let data = arguments[0];
        this.start = data.start;
        this.end = data.end;
        this.reverse = Boolean(data.reverse);
        this.containerId = data.containerId;
        this.surfaceId = data.surfaceId;
      } else {
        this.start = start;
        this.end = end;
        this.reverse = Boolean(reverse);
        this.containerId = containerId;
        this.surfaceId = surfaceId;
      }
    }

    isCollapsed () {
      return this.start.equals(this.end)
    }

    equals (other) {
      if (this === other) return true
      else {
        return (
          this.containerId === other.containerId &&
          this.start.equals(other.start) &&
          this.end.equals(other.end)
        )
      }
    }

    isReverse () {
      return this.reverse
    }

    toString () {
      let str = [this.start.toString(), '->', this.end.toString()];
      if (this.isReverse()) {
        str.push('[reverse]');
      }
      if (this.containerId) {
        str.push('[container=' + this.containerId + ']');
      }
      if (this.surfaceId) {
        str.push('[surface=' + this.surfaceId + ']');
      }
      return str.join('')
    }
  }

  Range$1.prototype._isRange = true;

  class AbstractClipboard$1 {
    constructor (configurator) {
      let schema = configurator.getSchema();
      let converterRegistry = configurator.converterRegistry;
      let htmlConverters = [];
      if (converterRegistry && converterRegistry.contains('html')) {
        htmlConverters = converterRegistry.get('html').values() || [];
      }
      
      
      this._config = {
        schema: schema,
        DocumentClass: schema.getDocumentClass(),
        converters: htmlConverters,
        editorOptions: configurator.getEditorOptions()
      };
    }

    
    onCopy (event) {
      
      let clipboardData = this._copy();
      substanceGlobals$1._clipboardData = event.clipboardData;

      if (event.clipboardData && clipboardData.doc) {
        event.preventDefault();
        
        event.clipboardData.setData('text/plain', clipboardData.text);
        event.clipboardData.setData('text/html', clipboardData.html);
      }
    }

    
    onCut (event) {
      
      event.preventDefault();
      
      this.onCopy(event);
      this._cut();
    }

    
    
    onPaste (event) {
      let clipboardData = event.clipboardData;

      let types = {};
      for (let i = 0; i < clipboardData.types.length; i++) {
        types[clipboardData.types[i]] = true;
      }
      

      event.preventDefault();
      event.stopPropagation();

      let plainText;
      let html;
      if (types['text/plain']) {
        plainText = clipboardData.getData('text/plain');
      }
      if (types['text/html']) {
        html = clipboardData.getData('text/html');
      }

      substanceGlobals$1.clipboardData = {
        text: plainText,
        html: html
      };

      

      
      
      if (html) {
        if (!this._pasteHtml(html, plainText)) {
          this._pastePlainText(plainText);
        }
      } else {
        this._pastePlainText(plainText);
      }
    }

    _getImporter () {
      return new ClipboardImporter$1(this._config)
    }

    _getExporter () {
      return new ClipboardExporter$1(this._config)
    }

    
    _copy () {
      throw new Error('This method is abstract.')
    }

    
    _cut () {
      throw new Error('This method is abstract.')
    }

    
    _pastePlainText(plainText) { 
      throw new Error('This method is abstract.')
    }

    
    _pasteHtml(html, text) { 
      throw new Error('This method is abstract.')
    }
  }

  function levenshtein$1 (a, b) {
    let m = [];
    for (let i = 0; i <= b.length; i++) {
      m[i] = [i];
      if (i === 0) continue
      let ib = i - 1;
      for (let j = 0; j <= a.length; j++) {
        m[0][j] = j;
        if (j === 0) continue
        let jb = j - 1;
        m[i][j] = b.charAt(ib) === a.charAt(jb) ? m[ib][jb] : Math.min(
          m[ib][jb] + 1,
          m[i][jb] + 1,
          m[ib][j] + 1
        );
      }
    }
    return m
  }

  function diff$1 (a, b, offset) {
    if (!isString$1(a) || !isString$1(b)) {
      throw new Error('Illegal arguments.')
    }
    offset = offset || 0;
    let changes = [];
    if (a || b) {
      if (!a && b) {
        changes.push({ type: 'insert', start: offset, text: b });
      } else if (a && !b) {
        changes.push({ type: 'delete', start: offset, end: offset + a.length });
      } else {
        let m = levenshtein$1(a, b);
        changes = _diff$1(a, b, m, offset);
      }
    }
    return changes
  }

  function _diff$1 (a, b, m, offset) {
    let i = b.length;
    let j = a.length;
    let changes = [];
    let current;
    while (i > 0 && j > 0) {
      _next();
    }
    _commit();
    return changes

    function _next () {
      let d = m[i][j];
      let ib = i - 1;
      let jb = j - 1;
      
      if (m[ib][jb] < d) {
        if (current && current.type === 'replace') {
          current.start--;
          current.text.unshift(b.charAt(ib));
        } else {
          _commit();
          current = { type: 'replace', start: jb, end: j, text: [b.charAt(ib)] };
        }
        i--;
        j--;
      
      } else if (m[ib][j] < d) {
        if (current && current.type === 'insert') {
          current.start--;
          current.text.unshift(b.charAt(ib));
        } else {
          _commit();
          current = { type: 'insert', start: jb, text: [b.charAt(ib)] };
        }
        i--;
      
      } else if (m[i][jb] < d) {
        if (current && current.type === 'delete') {
          current.start--;
        } else {
          _commit();
          current = { type: 'delete', start: jb, end: j };
        }
        j--;
      
      } else {
        _commit();
        i--;
        j--;
      }
    }

    function _commit () {
      if (current) {
        switch (current.type) {
          case 'insert':
            current.start += offset;
            current.text = current.text.join('');
            break
          case 'delete':
            current.start += offset;
            current.end += offset;
            break
          case 'replace':
            current.start += offset;
            current.end += offset;
            current.text = current.text.join('');
            break
          default:
            throw new Error('Invalid state')
        }
        changes.push(current);
        current = null;
      }
    }
  }

  class AnnotationComponent$1 extends Component$1 {
    
    didMount () {
      let node = this.props.node;
      node.on('highlighted', this.onHighlightedChanged, this);
    }

    
    dispose () {
      let node = this.props.node;
      node.off(this);
    }

    render ($$) {
      let el = $$(this.getTagName())
        .attr('data-id', this.props.node.id)
        .addClass(this.getClassNames());
      if (this.props.node.highlighted) {
        el.addClass('sm-highlighted');
      }
      el.append(this.props.children);
      return el
    }

    getClassNames () {
      return 'sc-' + this.props.node.type
    }

    onHighlightedChanged () {
      if (this.props.node.highlighted) {
        this.el.addClass('sm-highlighted');
      } else {
        this.el.removeClass('sm-highlighted');
      }
    }

    getTagName () {
      return 'span'
    }
  }

  class AbstractIsolatedNodeComponent$1 extends Component$1 {
    constructor (...args) {
      super(...args);

      this.name = this.props.node.id;
      this._state = { selectionFragment: null };

      this.handleAction('escape', this.escape);
      this.ContentClass = this._getContentClass(this.props.node);

      
      let useBlocker = platform$1.isFF || !this.ContentClass.noBlocker;
      this.blockingMode = useBlocker ? 'closed' : 'open';
    }

    getInitialState () {
      let selState = this.context.editorSession.getSelectionState();
      return this._deriveStateFromSelectionState(selState)
    }

    getChildContext () {
      return {
        parentSurfaceId: this.getId(),
        isolatedNodeComponent: this,
        
        
        surface: undefined
      }
    }

    didMount () {
      super.didMount();

      let editorSession = this.context.editorSession;
      editorSession.onRender('selection', this._onSelectionChanged, this);
    }

    dispose () {
      super.dispose.call(this);

      let editorSession = this.context.editorSession;
      editorSession.off(this);
    }

    renderContent ($$, node, options = {}) {
      let ComponentClass = this.ContentClass;
      if (!ComponentClass) {
        console.error('Could not resolve a component for type: ' + node.type);
        return $$(this.__elementTag)
      } else {
        let props = Object.assign({
          disabled: this.props.disabled,
          node: node,
          isolatedNodeState: this.state.mode,
          focused: (this.state.mode === 'focused')
        }, options);
        return $$(ComponentClass, props)
      }
    }

    getId () {
      
      
      if (!this._id) {
        this._id = this.context.parentSurfaceId + '/' + this.name;
      }
      return this._id
    }

    get id () { return this.getId() }

    getMode () {
      return this.state.mode
    }

    isOpen () {
      return this.blockingMode === 'open'
    }

    isClosed () {
      return this.blockingMode === 'closed'
    }

    isNotSelected () {
      return !this.state.mode
    }

    isSelected () {
      return this.state.mode === 'selected'
    }

    isCoSelected () {
      return this.state.mode === 'co-selected'
    }

    isFocused () {
      return this.state.mode === 'focused'
    }

    isCoFocused () {
      return this.state.mode === 'co-focused'
    }

    getParentSurface () {
      return this.context.surface
    }

    escape () {
      
      this.selectNode();
    }

    _onSelectionChanged () {
      let editorSession = this.context.editorSession;
      let newState = this._deriveStateFromSelectionState(editorSession.getSelectionState());
      if (!newState && this.state.mode) {
        this.extendState({ mode: null });
      } else if (newState && newState.mode !== this.state.mode) {
        this.extendState(newState);
      }
    }

    onKeydown (event) {
      
      
      
      
      if (event.keyCode === keys$1$1.ESCAPE && this.state.mode === 'focused') {
        event.stopPropagation();
        event.preventDefault();
        this.escape();
      }
    }

    _getContentClass (node) {
      let ComponentClass;
      
      ComponentClass = this.getComponent(node.type, true);
      
      if (!ComponentClass) {
        ComponentClass = Component$1;
      }
      return ComponentClass
    }

    _getSurface (selState) {
      let surface = selState.get('surface');
      if (surface === undefined) {
        let sel = selState.getSelection();
        if (sel && sel.surfaceId) {
          let surfaceManager = this.context.surfaceManager;
          surface = surfaceManager.getSurface(sel.surfaceId);
        } else {
          surface = null;
        }
        selState.set('surface', surface);
      }
      return surface
    }

    
    
    _getIsolatedNodes (selState) {
      let isolatedNodes = selState.get('isolatedNodes');
      if (!isolatedNodes) {
        let sel = selState.getSelection();
        isolatedNodes = [];
        if (sel && sel.surfaceId) {
          let surfaceManager = this.context.surfaceManager;
          let surface = surfaceManager.getSurface(sel.surfaceId);
          if (surface) {
            isolatedNodes = surface.getComponentPath().filter(comp => comp._isAbstractIsolatedNodeComponent);
          }
        }
        selState.set('isolatedNodes', isolatedNodes);
      }
      return isolatedNodes
    }

    _shouldConsumeEvent (event) {
      let comp = Component$1.unwrap(event.target);
      let isolatedNodeComponent = this._getIsolatedNode(comp);
      return (isolatedNodeComponent === this)
    }

    _getIsolatedNode (comp) {
      if (comp._isAbstractIsolatedNodeComponent) {
        return this
      } else if (comp.context.isolatedNodeComponent) {
        return comp.context.isolatedNodeComponent
      } else if (comp.context.surface) {
        return comp.context.surface.context.isolatedNodeComponent
      }
    }
  }

  AbstractIsolatedNodeComponent$1.prototype._isAbstractIsolatedNodeComponent = true;

  class InlineNodeComponent$1 extends AbstractIsolatedNodeComponent$1 {
    render ($$) {
      const node = this.props.node;
      const ContentClass = this.ContentClass;
      const state = this.state;

      let el = $$('span');
      el.addClass(this.getClassNames())
        .addClass('sc-inline-node')
        .addClass('sm-' + this.props.node.type)
        .attr('data-id', node.id)
        .attr('data-inline', '1');

      let disabled = this.isDisabled();

      if (state.mode) {
        el.addClass('sm-' + state.mode);
      } else {
        el.addClass('sm-not-selected');
      }

      if (!ContentClass.noStyle) {
        el.addClass('sm-default-style');
      }

      
      
      
      el.on('keydown', this.onKeydown);

      el.append(
        this.renderContent($$, node)
          .ref('content')
          .addClass('se-content')
      );

      if (disabled) {
        el.addClass('sm-disabled')
          .attr('contenteditable', false)
          .on('click', this.onClick);
      }

      
      
      
      
      
      
      
      el.attr('draggable', true);

      return el
    }

    isDisabled () {
      return !this.state.mode || ['co-selected', 'cursor'].indexOf(this.state.mode) > -1
    }

    getClassNames () {
      return ''
    }

    onClick (event) {
      if (!this._shouldConsumeEvent(event)) {
        return
      }
      this.selectNode();
    }

    selectNode () {
      
      let editorSession = this.context.editorSession;
      let surface = this.context.surface;
      let node = this.props.node;
      editorSession.setSelection({
        type: 'property',
        path: node.start.path,
        startOffset: node.start.offset,
        endOffset: node.end.offset,
        containerId: surface.getContainerId(),
        surfaceId: surface.id
      });
    }

    _getContentClass (node) {
      let ComponentClass;
      
      ComponentClass = this.getComponent(node.type, true);
      
      
      if (!ComponentClass) {
        ComponentClass = this.getComponent('unsupported-inline-node', true);
      }
      if (!ComponentClass) {
        console.error(`No component registered for inline node '${node.type}'.`);
        ComponentClass = StubInlineNodeComponent$1;
      }
      return ComponentClass
    }

    
    
    _deriveStateFromSelectionState (selState) {
      let surface = this._getSurface(selState);
      if (!surface) return null
      
      if (surface === this.context.surface) {
        let sel = selState.getSelection();
        let node = this.props.node;
        if (sel.isPropertySelection() && !sel.isCollapsed() && isEqual$1(sel.start.path, node.start.path)) {
          let nodeSel = node.getSelection();
          if (nodeSel.equals(sel)) {
            return { mode: 'selected' }
          }
          if (sel.contains(nodeSel)) {
            return { mode: 'co-selected' }
          }
        }
      }
      let isolatedNodeComponent = surface.context.isolatedNodeComponent;
      if (!isolatedNodeComponent) return null
      if (isolatedNodeComponent === this) {
        return { mode: 'focused' }
      }
      let isolatedNodes = this._getIsolatedNodes(selState);
      if (isolatedNodes.indexOf(this) > -1) {
        return { mode: 'co-focused' }
      }
      return null
    }
  }

  InlineNodeComponent$1.prototype._isInlineNodeComponent = true;

  class StubInlineNodeComponent$1 extends Component$1 {
    render ($$) {
      const node = this.props.node;
      return $$('span').text('???').attr('data-id', node.id).attr('data-type', node.type)
    }
  }

  class AnnotatedTextComponent$1 extends Component$1 {
    render ($$) {
      let el = this._renderContent($$)
        .addClass('sc-annotated-text')
        .css({ whiteSpace: 'pre-wrap' });
      return el
    }

    getPath () {
      return this.props.path
    }

    getText () {
      return this.getDocument().get(this.props.path) || ''
    }

    isEmpty () {
      return !(this.getText())
    }

    getAnnotations () {
      return this.getDocument().getIndex('annotations').get(this.props.path)
    }

    getDocument () {
      return this.props.doc || this.context.doc
    }

    _getTagName () {
      return this.props.tagName
    }

    _onDocumentChange (update) {
      if (update.change && update.change.updated[this.getPath()]) {
        this.rerender();
      }
    }

    _renderContent ($$) {
      let text = this.getText();
      let annotations = this.getAnnotations();
      let el = $$(this._getTagName() || 'span');
      if (annotations && annotations.length > 0) {
        let fragmenter = new Fragmenter$1({
          onText: this._renderTextNode.bind(this),
          onEnter: this._renderFragment.bind(this, $$),
          onExit: this._finishFragment.bind(this)
        });
        fragmenter.start(el, text, annotations);
      } else {
        el.append(text);
      }
      return el
    }

    _renderTextNode (context, text) {
      if (text && text.length > 0) {
        context.append(text);
      }
    }

    _renderFragment ($$, fragment) {
      let doc = this.getDocument();
      let componentRegistry = this.getComponentRegistry();
      let node = fragment.node;
      
      if (node.type === 'container-annotation-fragment') ; else if (node.type === 'container-annotation-anchor') ; else {
        let ComponentClass = componentRegistry.get(node.type) || AnnotationComponent$1;
        if (node.constructor.isInline &&
            
            !ComponentClass.prototype._isInlineNodeComponent &&
            
            !ComponentClass.isCustom) {
          ComponentClass = InlineNodeComponent$1;
        }
        let el = $$(ComponentClass, { doc: doc, node: node });
        return el
      }
    }

    _finishFragment (fragment, context, parentContext) {
      parentContext.append(context);
    }
  }

  class CursorComponent$1 extends Component$1 {
    render ($$) {
      
      let el = $$('span').addClass('se-cursor');
      
      
      
      el.append('\uFEFF');
      el.append($$('div').addClass('se-cursor-inner'));

      if (this.props.collaborator) {
        let collaboratorIndex = this.props.collaborator.colorIndex;
        el.addClass('sm-collaborator-' + collaboratorIndex);
      } else {
        el.addClass('sm-local-user');
      }

      return el
    }
  }

  class SelectionFragmentComponent$1 extends Component$1 {
    render ($$) {
      
      let el = $$('span').addClass('se-selection-fragment');
      if (this.props.collaborator) {
        let collaboratorIndex = this.props.collaborator.colorIndex;
        el.addClass('sm-collaborator-' + collaboratorIndex);
      } else {
        el.addClass('sm-local-user');
      }
      el.append(this.props.children);
      return el
    }
  }

  class TextPropertyComponent$1 extends AnnotatedTextComponent$1 {
    getInitialState () {
      const markersManager = this.context.markersManager;
      let path = this.getPath();
      let markers;
      if (markersManager) {
        
        markersManager.register(this);
        markers = markersManager.getMarkers(path, {
          surfaceId: this.getSurfaceId(),
          containerId: this.getContainerId()
        });
      } else {
        const doc = this.getDocument();
        markers = doc.getAnnotations(path);
      }
      return {
        markers: markers
      }
    }

    didMount () {
      
      
      
      
      
    }

    dispose () {
      if (this.context.markersManager) {
        this.context.markersManager.deregister(this);
      }
      if (this.domObserver) {
        this.domObserver.disconnect();
      }
    }

    render ($$) {
      let path = this.getPath();

      let el = this._renderContent($$)
        .addClass('sc-text-property')
        .attr({
          'data-path': path.join('.')
        })
        .css({
          'white-space': 'pre-wrap'
        });

      if (this.isEmpty()) {
        el.addClass('sm-empty');
        if (this.props.placeholder) {
          el.append(
            $$('span').addClass('se-placeholder').append(
              this.props.placeholder
            )
          );
        }
      }

      if (!this.props.withoutBreak) {
        el.append($$('br'));
      }

      return el
    }

    getAnnotations () {
      if (this.props.markers) {
        return this.state.markers.concat(this.props.markers)
      } else {
        return this.state.markers
      }
    }

    _renderFragment ($$, fragment) {
      let node = fragment.node;
      let id = node.id;
      let el;
      if (node.type === 'cursor') {
        el = $$(CursorComponent$1, { collaborator: node.collaborator });
      } else if (node.type === 'selection-fragment') {
        el = $$(SelectionFragmentComponent$1, { collaborator: node.collaborator });
      } else {
        el = super._renderFragment.apply(this, arguments);
        if (id) {
          el.ref(id + '@' + fragment.counter);
        }
      }
      el.attr('data-offset', fragment.pos);
      return el
    }

    _onDomMutations (mutations) {
      
      
      if (mutations.length === 2 && mutations[0].target === mutations[1].target) {
        let textEl = DefaultDOMElement$1.unwrap(mutations[0].target);
        if (textEl) {
          this._applyTextMutation(textEl, mutations[0].oldValue);
          return
        }
      }
      
      this.rerender();
    }

    _applyTextMutation (textEl, oldText) {
      
      let offset = _getCharPos$1(textEl, 0);
      let newText = textEl.textContent;
      let changes = diff$1(oldText, newText, offset);

      let editorSession = this.context.editorSession;
      let path = this.getPath();
      editorSession.transaction(function (tx) {
        changes.forEach(function (change) {
          
          if (change.type === 'replace') {
            tx.update(path, { type: 'delete', start: change.start, end: change.end });
            tx.update(path, { type: 'insert', start: change.start, text: change.text });
          } else {
            tx.update(path, change);
          }
        });
      });
    }

    getSurface () {
      return this.props.surface || this.context.surface
    }

    getSurfaceId () {
      let surface = this.getSurface();
      return surface ? surface.id : null
    }

    getContainerId () {
      let surface = this.getSurface();
      return surface ? surface.getContainerId() : null
    }

    isEditable () {
      const surface = this.getSurface();
      return surface ? surface.isEditable() : false
    }

    isReadonly () {
      const surface = this.getSurface();
      return surface ? surface.isReadonly() : true
    }

    getDOMCoordinate (charPos) {
      return this._getDOMCoordinate(this.el, charPos)
    }

    _finishFragment (fragment, context, parentContext) {
      context.attr('data-length', fragment.length);
      parentContext.append(context);
    }

    _getDOMCoordinate (el, charPos) {
      let l;
      let idx = 0;
      if (charPos === 0) {
        return {
          container: el.getNativeElement(),
          offset: 0
        }
      }
      for (let child = el.getFirstChild(); child; child = child.getNextSibling(), idx++) {
        if (child.isTextNode()) {
          l = child.textContent.length;
          if (l >= charPos) {
            return {
              container: child.getNativeElement(),
              offset: charPos
            }
          } else {
            charPos -= l;
          }
        } else if (child.isElementNode()) {
          let length = child.getAttribute('data-length');
          if (length) {
            l = parseInt(length, 10);
            if (l >= charPos) {
              
              if (child.attr('data-inline')) {
                let nextSibling = child.getNextSibling();
                if (nextSibling && nextSibling.isTextNode()) {
                  return {
                    container: nextSibling.getNativeElement(),
                    offset: 0
                  }
                } else {
                  return {
                    container: el.getNativeElement(),
                    offset: el.getChildIndex(child) + 1
                  }
                }
              }
              return this._getDOMCoordinate(child, charPos, idx)
            } else {
              charPos -= l;
            }
          } else {
            console.error('FIXME: Can not map to DOM coordinates.');
            return null
          }
        }
      }
    }
  }

  TextPropertyComponent$1.prototype._isTextPropertyComponent = true;




  TextPropertyComponent$1.getCoordinate = function (root, el, offset) {
    let context = _getPropertyContext$1(root, el, offset);
    if (!context) {
      return null
    }
    
    
    let charPos;
    if (el.parentNode && el.parentNode.is('.se-placeholder')) {
      charPos = 0;
    } else {
      
      
      
      charPos = _getCharPos$1(context.node, context.offset);
    }
    if (isNumber$1(charPos)) {
      let coor = new Coordinate$1(context.path, charPos);
      coor._comp = context.comp;
      return coor
    } else {
      return null
    }
  };

  function _getPropertyContext$1 (root, node, offset) {
    let result = {
      comp: null,
      el: null,
      path: null,
      node: node,
      offset: offset
    };
    while (node && node !== root) {
      if (node.isElementNode()) {
        let comp = Component$1.unwrap(node);
        if (comp && comp._isTextPropertyComponent) {
          result.comp = comp;
          result.el = node;
          result.path = comp.getPath();
          return result
        }
        
        
        
        if (node.getAttribute('data-inline')) {
          result.node = node;
          if (offset > 0) {
            result.offset = 1;
          }
        }
      }
      node = node.getParent();
    }
    return null
  }

  function _getCharPos$1 (node, offset) {
    let charPos = offset;
    let parent, childIdx;

    

    parent = node.getParent();
    if (node.isTextNode()) {
      
      if (node === parent.firstChild) {
        
        let parentPath = parent.getAttribute('data-path');
        let parentOffset = parent.getAttribute('data-offset');
        if (parentPath) {
          charPos = offset;
        
        } else if (parentOffset) {
          charPos = parseInt(parentOffset, 10) + offset;
        
        } else {
          charPos = _getCharPos$1(parent, 0) + offset;
        }
      } else {
        
        childIdx = parent.getChildIndex(node);
        charPos = _getCharPos$1(parent, childIdx) + offset;
      }
    } else if (node.isElementNode()) {
      let pathStr = node.getAttribute('data-path');
      let offsetStr = node.getAttribute('data-offset');
      
      
      if (pathStr) {
        charPos = _countCharacters$1(node, offset);
      
      
      } else if (offsetStr) {
        childIdx = parent.getChildIndex(node);
        charPos = parseInt(offsetStr, 10) + _countCharacters$1(node, offset);
      
      
      } else {
        childIdx = parent.getChildIndex(node);
        charPos = _getCharPos$1(parent, childIdx) + _countCharacters$1(node, offset);
      }
    } else {
      
      return null
    }
    return charPos
  }

  function _countCharacters$1 (el, maxIdx) {
    let charPos = 0;
    
    if (el.getAttribute('data-inline')) {
      return maxIdx === 0 ? 0 : 1
    }
    let l = el.getChildCount();
    if (arguments.length === 1) {
      maxIdx = l;
    }
    maxIdx = Math.min(l, maxIdx);
    for (let i = 0, child = el.getFirstChild(); i < maxIdx; child = child.getNextSibling(), i++) {
      if (child.isTextNode()) {
        charPos += child.getTextContent().length;
      } else if (child.isElementNode()) {
        let length = child.getAttribute('data-length');
        if (child.getAttribute('data-inline')) {
          charPos += 1;
        } else if (length) {
          charPos += parseInt(length, 10);
        } else {
          charPos += _countCharacters$1(child);
        }
      }
    }
    return charPos
  }

  const BRACKET$1 = 'X';
  const SLASH$1 = '/'.charCodeAt(0);


  class IsolatedNodeComponent$1 extends AbstractIsolatedNodeComponent$1 {
    render ($$) {
      let node = this.props.node;
      let ContentClass = this.ContentClass;
      let disabled = this.props.disabled;

      
      let el = $$('div');
      el.addClass(this.getClassNames())
        .addClass('sc-isolated-node')
        .addClass('sm-' + this.props.node.type)
        .attr('data-id', node.id);
      if (disabled) {
        el.addClass('sm-disabled');
      }
      if (this.state.mode) {
        el.addClass('sm-' + this.state.mode);
      }
      if (!ContentClass.noStyle) {
        el.addClass('sm-default-style');
      }
      
      el.on('keydown', this.onKeydown);

      
      let shouldRenderBlocker = (
        this.blockingMode === 'closed' &&
        !this.state.unblocked
      );

      
      
      el.append(
        $$('div').addClass('se-bracket sm-left').ref('left')
          .append(BRACKET$1)
      );

      let content = this.renderContent($$, node, {
        disabled: this.props.disabled || shouldRenderBlocker
      }).ref('content');
      content.attr('contenteditable', false);

      el.append(content);
      el.append($$(Blocker$1).ref('blocker'));
      el.append(
        $$('div').addClass('se-bracket sm-right').ref('right')
          .append(BRACKET$1)
      );

      if (!shouldRenderBlocker) {
        el.addClass('sm-no-blocker');
        el.on('click', this.onClick)
          .on('dblclick', this.onDblClick);
      }
      el.on('mousedown', this._reserveMousedown, this);

      return el
    }

    getClassNames () {
      return ''
    }

    getContent () {
      return this.refs.content
    }

    selectNode () {
      
      let editorSession = this.context.editorSession;
      let surface = this.context.surface;
      let nodeId = this.props.node.id;
      editorSession.setSelection({
        type: 'node',
        nodeId: nodeId,
        containerId: surface.getContainerId(),
        surfaceId: surface.id
      });
    }

    
    
    onClick (event) {
      
      event.stopPropagation();
    }

    onDblClick (event) {
      
      event.stopPropagation();
    }

    grabFocus (event) {
      let content = this.refs.content;
      if (content.grabFocus) {
        content.grabFocus(event);
        return true
      }
    }

    
    
    _reserveMousedown (event) {
      if (event.__reserved__) ; else {
        
        event.__reserved__ = this;
      }
    }

    _deriveStateFromSelectionState (selState) {
      let surface = this._getSurface(selState);
      let newState = { mode: null, unblocked: null };
      if (!surface) return newState
      
      if (surface === this.context.surface) {
        let sel = selState.getSelection();
        let nodeId = this.props.node.id;
        if (sel.isNodeSelection() && sel.getNodeId() === nodeId) {
          if (sel.isFull()) {
            newState.mode = 'selected';
            newState.unblocked = true;
          } else if (sel.isBefore()) {
            newState.mode = 'cursor';
            newState.position = 'before';
          } else if (sel.isAfter()) {
            newState.mode = 'cursor';
            newState.position = 'after';
          }
        }
        if (sel.isContainerSelection() && sel.containsNode(nodeId)) {
          newState.mode = 'co-selected';
        }
      } else {
        let sel = selState.getSelection();
        let surfaceId = sel.surfaceId;
        let id = this.getId();
        if (id.length < surfaceId.length && surfaceId.startsWith(id) && surfaceId.charCodeAt(id.length) === SLASH$1) {
          let tail = surfaceId.slice(id.length + 1);
          if (tail.indexOf('/') > 0) {
            newState.mode = 'co-focused';
            newState.unblocked = true;
          } else {
            newState.mode = 'focused';
            newState.unblocked = true;
          }
        }
      }
      return newState
    }
  }

  IsolatedNodeComponent$1.prototype._isIsolatedNodeComponent = true;

  IsolatedNodeComponent$1.prototype._isDisabled = IsolatedNodeComponent$1.prototype.isDisabled;

  IsolatedNodeComponent$1.getDOMCoordinate = function (comp, coor) {
    let { start, end } = IsolatedNodeComponent$1.getDOMCoordinates(comp);
    if (coor.offset === 0) return start
    else return end
  };

  IsolatedNodeComponent$1.getDOMCoordinates = function (comp) {
    const left = comp.refs.left;
    const right = comp.refs.right;
    return {
      start: {
        container: left.getNativeElement(),
        offset: 0
      },
      end: {
        container: right.getNativeElement(),
        offset: right.getChildCount()
      }
    }
  };

  IsolatedNodeComponent$1.getCoordinate = function (nodeEl, options) {
    let comp = Component$1.unwrap(nodeEl, 'strict').context.isolatedNodeComponent;
    let offset = null;
    if (options.direction === 'left' || nodeEl === comp.refs.left.el) {
      offset = 0;
    } else if (options.direction === 'right' || nodeEl === comp.refs.right.el) {
      offset = 1;
    }
    let coor;
    if (offset !== null) {
      coor = new Coordinate$1([comp.props.node.id], offset);
      coor._comp = comp;
    }
    return coor
  };

  class Blocker$1 extends Component$1 {
    render ($$) {
      return $$('div').addClass('sc-isolated-node-blocker')
        .attr('draggable', true)
        .attr('contenteditable', false)
        .on('click', this.onClick)
        .on('dblclick', this.onDblClick)
    }

    onClick (event) {
      if (event.target !== this.getNativeElement()) return
      
      event.stopPropagation();
      const comp = this._getIsolatedNodeComponent();
      comp.extendState({ mode: 'selected', unblocked: true });
      comp.selectNode();
    }

    onDblClick (event) {
      
      event.stopPropagation();
    }

    _getIsolatedNodeComponent () {
      return this.context.isolatedNodeComponent
    }
  }

  const DEBUG$1$1 = false;


  class DOMSelection$1 {
    constructor (editor) {
      this.editor = editor;
      if (platform$1.inBrowser) {
        this.wRange = window.document.createRange();
      }
      
      
      this.state = { dom: null, model: null };
    }

    
    getSelection (options) {
      
      if (!platform$1.inBrowser) return
      let range = this.mapDOMSelection(options);
      let doc = this.editor.getDocument();
      
      return doc._createSelectionFromRange(range)
    }

    getSelectionForDOMRange (wrange) {
      let range = this.mapDOMRange(wrange);
      let doc = this.editor.getDocument();
      return doc._createSelectionFromRange(range)
    }

    
    mapDOMSelection (options) {
      let wSel = window.getSelection();
      let state = this.state;
      let range;
      
      
      if (DEBUG$1$1) console.info('DOM->Model: ', wSel.anchorNode, wSel.anchorOffset, wSel.focusNode, wSel.focusOffset);
      if (wSel.rangeCount === 0) return _null()
      let anchorNode = DefaultDOMElement$1.wrapNativeElement(wSel.anchorNode);
      if (wSel.isCollapsed) {
        let coor = this._getCoordinate(anchorNode, wSel.anchorOffset, options);
        if (!coor) return _null()
        range = _createRange$1({
          start: coor,
          end: coor
        });
      } else {
        let focusNode = DefaultDOMElement$1.wrapNativeElement(wSel.focusNode);
        range = this._getRange(anchorNode, wSel.anchorOffset, focusNode, wSel.focusOffset, options);
      }
      if (DEBUG$1$1) console.info('DOM->Model: range ', range ? range.toString() : null);
      state.model = range;
      return range

      function _null () {
        state.dom = null;
        state.model = null;
        return null
      }
    }

    
    setSelection (sel) {
      
      if (!platform$1.inBrowser) return
      let state = this.state;
      let wSel = window.getSelection();
      let wRange = this.wRange;
      if (!sel || sel.isNull()) return this.clear()
      
      let {start, end} = this.mapModelToDOMCoordinates(sel);
      if (!start) return this.clear()
      if (sel.isReverse()) {
        [start, end] = [end, start];
      }
      state.dom = {
        anchorNode: start.container,
        anchorOffset: start.offset,
        focusNode: end.container,
        focusOffset: end.offset
      };
      _set(state.dom);

      function _set ({anchorNode, anchorOffset, focusNode, focusOffset}) {
        wSel.removeAllRanges();
        wRange.setStart(anchorNode, anchorOffset);
        wRange.setEnd(anchorNode, anchorOffset);
        wSel.addRange(wRange);
        if (focusNode !== anchorOffset || focusOffset !== anchorOffset) {
          wSel.extend(focusNode, focusOffset);
        }
      }
    }

    mapModelToDOMCoordinates (sel) {
      if (DEBUG$1$1) console.info('Model->DOM: sel =', sel.toString());
      let rootEl;
      let surface = this.editor.getSurfaceManager().getSurface(sel.surfaceId);
      if (!surface) {
        console.warn('No surface:', sel.surfaceId);
        rootEl = this.editor.getElement();
      } else {
        rootEl = surface.el;
      }
      if (sel.isNull() || sel.isCustomSelection()) {
        return {}
      }

      let start, end;
      if (sel.isPropertySelection() || sel.isContainerSelection()) {
        start = this._getDOMCoordinate(rootEl, sel.start);
        if (!start) {
          console.warn('FIXME: selection seems to be invalid.');
          return {}
        }
        if (sel.isCollapsed()) {
          end = start;
        } else {
          end = this._getDOMCoordinate(rootEl, sel.end);
          if (!end) {
            console.warn('FIXME: selection seems to be invalid.');
            return {}
          }
        }
      } else if (sel.isNodeSelection()) {
        let comp = Component$1.unwrap(rootEl.find('*[data-id="' + sel.getNodeId() + '"]'));
        if (!comp) {
          console.error('Could not find component with id', sel.getNodeId());
          return {}
        }
        if (comp._isIsolatedNodeComponent) {
          let coors = IsolatedNodeComponent$1.getDOMCoordinates(comp, sel);
          start = coors.start;
          end = coors.end;
          
          
          
        } else {
          let _nodeEl = comp.el;
          start = {
            container: _nodeEl.getNativeElement(),
            offset: 0
          };
          end = {
            container: _nodeEl.getNativeElement(),
            offset: _nodeEl.getChildCount()
          };
        }
      }
      if (DEBUG$1$1) console.info('Model->DOM:', start.container, start.offset, end.container, end.offset, 'isReverse?', sel.isReverse());
      return {start, end}
    }

    _getDOMCoordinate (rootEl, coor) {
      let domCoor = null;
      let comp;
      if (coor.isNodeCoordinate()) {
        comp = Component$1.unwrap(rootEl.find('*[data-id="' + coor.getNodeId() + '"]'));
        if (comp) {
          if (comp._isIsolatedNodeComponent) {
            domCoor = IsolatedNodeComponent$1.getDOMCoordinate(comp, coor);
          } else {
            let domOffset = 0;
            if (coor.offset > 0) {
              domOffset = comp.getChildCount();
            }
            domCoor = {
              container: comp.getNativeElement(),
              offset: domOffset
            };
          }
        }
      } else {
        comp = Component$1.unwrap(rootEl.find('.sc-text-property[data-path="' + coor.path.join('.') + '"]'));
        if (comp) {
          domCoor = comp.getDOMCoordinate(coor.offset);
        }
      }
      return domCoor
    }

    
    mapDOMRange (wRange, options) {
      return this._getRange(
        DefaultDOMElement$1.wrapNativeElement(wRange.startContainer),
        wRange.startOffset,
        DefaultDOMElement$1.wrapNativeElement(wRange.endContainer),
        wRange.endOffset, options)
    }

    
    clear () {
      window.getSelection().removeAllRanges();
      this.state.dom = null;
      this.state.model = null;
    }

    collapse (dir) {
      let wSel = window.getSelection();
      let wRange;
      if (wSel.rangeCount > 0) {
        wRange = wSel.getRangeAt(0);
        wRange.collapse(dir === 'left');
        wSel.removeAllRanges();
        wSel.addRange(wRange);
      }
    }

    select (el) {
      let wSel = window.getSelection();
      let wRange = window.document.createRange();
      wRange.selectNode(el.getNativeElement());
      wSel.removeAllRanges();
      wSel.addRange(wRange);
    }

    extend (el, offset) {
      let wSel = window.getSelection();
      wSel.extend(el.getNativeElement(), offset);
    }

    setCursor (el, offset) {
      let wSel = window.getSelection();
      let wRange = window.document.createRange();
      wRange.setStart(el.getNativeElement(), offset);
      wSel.removeAllRanges();
      wSel.addRange(wRange);
    }

    
    _getRange (anchorNode, anchorOffset, focusNode, focusOffset, options = {}) {
      let isReverse = DefaultDOMElement$1.isReverse(anchorNode, anchorOffset, focusNode, focusOffset);
      let isCollapsed = (anchorNode === focusNode && anchorOffset === focusOffset);
      let start, end;
      if (isCollapsed) {
        start = end = this._getCoordinate(anchorNode, anchorOffset, options);
      } else {
        start = this._getCoordinate(anchorNode, anchorOffset, { direction: isReverse ? 'right' : 'left' });
        end = this._getCoordinate(focusNode, focusOffset, options);
      }
      if (start && end) {
        return _createRange$1({ start, end, isReverse })
      } else {
        return null
      }
    }

    
    _getCoordinate (nodeEl, offset, options = {}) {
      let coor = null;
      
      if (!coor) {
        coor = TextPropertyComponent$1.getCoordinate(this.editor.getElement(), nodeEl, offset);
      }
      let comp = Component$1.unwrap(nodeEl);
      if (!coor && comp) {
        
        if (comp.context.isolatedNodeComponent) {
          coor = IsolatedNodeComponent$1.getCoordinate(nodeEl, options);
        }
      }
      
      
      if (!coor) {
        
        if (comp && comp._isContainerEditor) {
          let childIdx = (offset === 0) ? 0 : offset - 1;
          let isBefore = (offset === 0);
          let container = comp.getContainer();
          let childNode = container.getNodeAt(childIdx);
          let childComp = comp.getChildAt(childIdx);
          coor = new Coordinate$1([childNode.id], isBefore ? 0 : 1);
          coor._comp = childComp;
        
        } else if (nodeEl.isElementNode() && nodeEl.getChildCount() > 0) {
          let child = (offset > 0) ? nodeEl.getChildAt(offset - 1) : nodeEl.firstChild;
          let prop;
          let childComp = Component$1.unwrap(child);
          if (childComp && childComp._isTextPropertyComponent) {
            prop = child;
          }
          
          if (prop) {
            coor = TextPropertyComponent$1.getCoordinate(nodeEl, prop, (offset > 0) ? prop.getChildCount() : 0);
          }
        }
      }
      return coor
    }
  }


  function _createRange$1 ({start, end, isReverse}) {
    if (isReverse) {
      [start, end] = [end, start];
    }
    if (!start._comp || !end._comp) {
      console.error('FIXME: getCoordinate() should provide a component instance');
      return null
    }
    let surface = start._comp.context.surface;
    if (!surface) {
      console.error('FIXME: Editable components should have their surface in the context');
      return null
    }
    if (surface !== end._comp.context.surface) {
      console.error('Coordinates are within two different surfaces. Can not create a selection.');
      return null
    }
    return new Range$1(start, end, isReverse, surface.getContainerId(), surface.id)
  }

  class Command$1 {
    
    constructor (config) {
      this.config = config || {};
      this.name = this.config.name;
      if (!this.name) {
        throw new Error("'name' is required")
      }
    }

    get isAsync () {
      return false
    }

    
    getName () {
      return this.name
    }

    
    getCommandState(params, context) { 
      throw new Error('Command.getCommandState() is abstract.')
    }

    
    execute(params, context) { 
      throw new Error('Command.execute() is abstract.')
    }

    
    isAnnotationCommand () {
      return false
    }

    
    isInsertCommand () {
      return false
    }

    
    isSwitchTypeCommand () {
      return false
    }

    _getEditorSession (params, context) {
      let editorSession = params.editorSession || context.editorSession;
      if (!editorSession) {
        throw new Error("'editorSession' is required.")
      }
      return editorSession
    }

    _getSelection (params) {
      let sel = params.selection || params.selectionState.getSelection();
      if (!sel) {
        throw new Error("'selection' is required.")
      }
      return sel
    }
  }

  Command$1.prototype._isCommand = true;

  class NodeComponent$1 extends Component$1 {
    didMount () {
      this.context.editorSession.onRender('document', this.rerender, this, { path: [this.props.node.id] });
    }

    dispose () {
      this.context.editorSession.off(this);
    }

    render ($$) {
      let tagName = this.getTagName();
      let el = $$(tagName)
        .attr('data-id', this.props.node.id)
        .addClass(this.getClassNames());
      return el
    }

    getTagName () {
      return 'div'
    }

    getClassNames () {
      return ''
    }

    rerender (...args) {
      
      if (this.props.node.isDisposed()) return
      super.rerender(...args);
    }
  }

  class Clipboard$1 extends AbstractClipboard$1 {
    constructor (editorSession) {
      super(editorSession.getConfigurator());

      this.editorSession = editorSession;
    }

    
    _copy () {
      let editorSession = this.getEditorSession();
      let sel = editorSession.getSelection();
      let doc = editorSession.getDocument();
      let clipboardDoc = null;
      let clipboardText = '';
      let clipboardHtml = '';
      let htmlExporter = this._getExporter();
      if (!sel.isCollapsed()) {
        clipboardText = getTextForSelection$1(doc, sel) || '';
        clipboardDoc = copySelection$1(doc, sel);
        clipboardHtml = htmlExporter.exportDocument(clipboardDoc);
      }
      return {
        doc: clipboardDoc,
        html: clipboardHtml,
        text: clipboardText
      }
    }

    _cut () {
      let editorSession = this.getEditorSession();
      editorSession.transaction((tx) => {
        tx.deleteSelection();
      });
    }

    
    _pastePlainText (plainText) {
      let editorSession = this.getEditorSession();
      editorSession.transaction(function (tx) {
        tx.paste(plainText);
      }, { action: 'paste' });
    }

    
    _pasteHtml (html, text) {
      let htmlImporter = this._getImporter();
      let content = htmlImporter.importDocument(html) || text;
      if (content) {
        let editorSession = this.getEditorSession();
        editorSession.transaction((tx) => {
          tx.paste(content);
        }, { action: 'paste' });
      }
      return true
    }

    getEditorSession () {
      return this.editorSession
    }
  }

  function startsWith$1 (str, prefix) {
    if (!isString$1(str)) return false
    if (str.startsWith) return str.startsWith(prefix)
    if (!isString$1(prefix)) prefix = String(prefix);
    return str.slice(0, prefix.length) === prefix
  }

  class UnsupportedNodeComponent$1 extends Component$1 {
    render ($$) {
      return $$('pre')
        .addClass('content-node unsupported')
        .attr({
          'data-id': this.props.node.id,
          contentEditable: false
        })
        .append(
          JSON.stringify(this.props.node.properties, null, 2)
        )
    }
  }

  const BROWSER_DELAY$1 = platform$1.isFF ? 1 : 0;


  class Surface$1 extends Component$1 {
    constructor (...args) {
      super(...args);

      this._initialize();
    }

    _initialize () {
      
      
      this.editorSession = this.props.editorSession || this.context.editorSession;
      if (!this.editorSession) {
        throw new Error('No EditorSession provided')
      }
      this.name = this.props.name;
      if (!this.name) {
        throw new Error('Surface must have a name.')
      }
      if (this.name.indexOf('/') > -1) {
        
        throw new Error("Surface.name must not contain '/'")
      }
      
      
      this._surfaceId = Surface$1.createSurfaceId(this);

      this.clipboard = new Clipboard$1(this.editorSession);

      this.domSelection = this.context.domSelection || new DOMSelection$1(this);

      this._state = {
        
        skipNextFocusEvent: false
      };
    }

    getChildContext () {
      return {
        surface: this,
        parentSurfaceId: this.getId(),
        doc: this.getDocument(),
        
        
        isolatedNodeComponent: null
      }
    }

    didMount () {
      const editorSession = this.getEditorSession();
      const surfaceManager = this.getSurfaceManager();
      if (surfaceManager) {
        surfaceManager.registerSurface(this);
      }
      editorSession.onRender('selection', this._onSelectionChanged, this);

      const globalEventHandler = editorSession.globalEventHandler;
      globalEventHandler.addEventListener('keydown', this._muteNativeHandlers, this);
    }

    dispose () {
      const editorSession = this.getEditorSession();
      const surfaceManager = this.getSurfaceManager();
      editorSession.off(this);
      if (surfaceManager) {
        surfaceManager.unregisterSurface(this);
      }

      const globalEventHandler = editorSession.globalEventHandler;
      globalEventHandler.removeEventListener('keydown', this._muteNativeHandlers);
    }

    didUpdate () {
      this._updateContentEditableState();
    }

    render ($$) {
      let tagName = this.props.tagName || 'div';
      let el = $$(tagName)
        .addClass('sc-surface')
        .attr('tabindex', 2)
        .attr('data-surface-id', this.id);

      if (!this.isDisabled()) {
        if (this.isEditable()) {
          
          el.on('keydown', this.onKeyDown);
          
          if (!platform$1.isIE) {
            el.on('compositionstart', this.onCompositionStart);
            el.on('compositionend', this.onCompositionEnd);
          }
          
          
          
          
          if (platform$1.inBrowser && window.TextEvent && !platform$1.isIE) {
            el.on('textInput', this.onTextInput);
          } else {
            el.on('keypress', this.onTextInputShim);
          }
          el.on('paste', this._onPaste);
          el.on('cut', this._onCut);
        }
        if (!this.isReadonly()) {
          
          el.on('mousedown', this.onMouseDown);
          el.on('contextmenu', this.onContextMenu);
          
          
          el.on('focus', this.onNativeFocus);
          el.on('blur', this.onNativeBlur);
        }
        el.on('copy', this._onCopy);
      }
      return el
    }

    renderNode ($$, node) {
      let componentRegistry = this.getComponentRegistry();
      let ComponentClass = componentRegistry.get(node.type);
      if (!ComponentClass) {
        console.error('Could not resolve a component for type: ' + node.type);
        ComponentClass = UnsupportedNodeComponent$1;
      }
      return $$(ComponentClass, this._extractNodeProps(node)).ref(node.id)
    }

    getComponentRegistry () {
      return this.context.componentRegistry || this.props.componentRegistry
    }

    getName () {
      return this.name
    }

    getId () {
      return this._surfaceId
    }

    getSurfaceId () {
      return this._surfaceId
    }

    isDisabled () {
      return this.props.disabled
    }

    isEditable () {
      return (this.props.editing === 'full' || this.props.editing === undefined)
    }

    isSelectable () {
      return (this.props.editing === 'selection' || this.props.editing === 'full')
    }

    isReadonly () {
      return this.props.editing === 'readonly'
    }

    getElement () {
      return this.el
    }

    getDocument () {
      return this.getEditorSession().getDocument()
    }

    getEditorSession () {
      return this.editorSession
    }

    getSurfaceManager () {
      return this.context.surfaceManager
    }

    isEnabled () {
      return !this.state.disabled
    }

    isContainerEditor () {
      return false
    }

    isCustomEditor () {
      return false
    }

    hasNativeSpellcheck () {
      return this.props.spellcheck === 'native'
    }

    getContainerId () {
      return null
    }

    focus () {
      if (this.getEditorSession().getFocusedSurface() !== this) {
        this.selectFirst();
      }
    }

    blur () {
      if (this.getEditorSession().getFocusedSurface() === this) {
        this.getEditorSession().setSelection(null);
      }
    }

    selectFirst () {
      throw new Error('This method is abstract.')
    }

    
    rerenderDOMSelection () {
      if (this.isDisabled()) return
      if (platform$1.inBrowser) {
        
        let sel = this.getEditorSession().getSelection();
        if (sel.surfaceId === this.getId()) {
          this.domSelection.setSelection(sel);
          
          
          
          const scrollPane = this.context.scrollPane;
          if (scrollPane) {
            this.context.scrollPane.onSelectionPositioned();
          }
        }
      }
    }

    getDomNodeForId (nodeId) {
      return this.el.getNativeElement().querySelector('*[data-id="' + nodeId + '"]')
    }

    

    
    onKeyDown (event) {
      if (!this._shouldConsumeEvent(event)) return
      

      
      if (event.key === 'Dead') return

      
      let custom = this.getEditorSession().keyboardManager.onKeydown(event);
      if (!custom) {
        
        switch (event.keyCode) {
          
          case keys$1$1.LEFT:
          case keys$1$1.RIGHT:
            return this._handleLeftOrRightArrowKey(event)
          case keys$1$1.UP:
          case keys$1$1.DOWN:
            return this._handleUpOrDownArrowKey(event)
          case keys$1$1.HOME:
          case keys$1$1.END:
            return this._handleHomeOrEndKey(event)
          case keys$1$1.PAGEUP:
          case keys$1$1.PAGEDOWN:
            return this._handlePageUpOrDownKey(event)
          
          case keys$1$1.ENTER:
            return this._handleEnterKey(event)
          case keys$1$1.TAB:
            return this._handleTabKey(event)
          case keys$1$1.BACKSPACE:
          case keys$1$1.DELETE:
            return this._handleDeleteKey(event)
          case keys$1$1.ESCAPE:
            return this._handleEscapeKey(event)
          case keys$1$1.SPACE:
            return this._handleSpaceKey(event)
          default:
            break
        }
      }
    }

    onTextInput (event) {
      if (!this._shouldConsumeEvent(event)) return
      
      event.preventDefault();
      event.stopPropagation();
      if (!event.data) return
      const editorSession = this.getEditorSession();
      let text = event.data;
      if (!editorSession.keyboardManager.onTextInput(text)) {
        editorSession.transaction((tx) => {
          tx.insertText(text);
        }, { action: 'type' });
      }
    }

    
    onCompositionStart (event) {
      if (!this._shouldConsumeEvent(event)) return
      
      
      
      
      
      
      
      if (event.data) {
        let l = event.data.length;
        let sel = this.getEditorSession().getSelection();
        if (sel.isPropertySelection() && sel.isCollapsed()) {
          
          let offset = sel.start.offset;
          this.getEditorSession().setSelection(sel.createWithNewRange(offset - l, offset));
        }
      }
    }

    onCompositionEnd (event) {
      if (!this._shouldConsumeEvent(event)) return
      
      
      
      if (platform$1.isFF) {
        event.preventDefault();
        event.stopPropagation();
        if (!event.data) return
        this._delayed(() => {
          let text = event.data;
          if (!this.getEditorSession().keyboardManager.onTextInput(text)) {
            this.getEditorSession().transaction((tx) => {
              tx.insertText(text);
            }, { action: 'type' });
          }
        });
      }
    }

    
    onTextInputShim (event) {
      if (!this._shouldConsumeEvent(event)) return
      
      if (
        
        event.which === 0 || event.charCode === 0 ||
        
        event.keyCode === keys$1$1.TAB || event.keyCode === keys$1$1.ESCAPE ||
        
        Boolean(event.metaKey) || (Boolean(event.ctrlKey) ^ Boolean(event.altKey))
      ) {
        return
      }
      let character = String.fromCharCode(event.which);
      if (!event.shiftKey) {
        character = character.toLowerCase();
      }
      event.preventDefault();
      event.stopPropagation();
      if (!this.getEditorSession().keyboardManager.onTextInput(character)) {
        if (character.length > 0) {
          this.getEditorSession().transaction((tx) => {
            tx.insertText(character);
          }, { action: 'type' });
        }
      }
    }

    
    
    
    
    
    onMouseDown (event) {
      if (!this._shouldConsumeEvent(event)) {
        
        return
      }

      
      
      
      
      
      if (event.__reserved__) {
        
        return
      } else {
        
        event.__reserved__ = this;
      }

      
      
      
      if (this.isEditable()) {
        this.el.setAttribute('contenteditable', true);
      }

      
      if (event.button !== 0) {
        return
      }

      
      if (!(platform$1.isIE && platform$1.version < 12) && event.detail >= 3) {
        let sel = this.getEditorSession().getSelection();
        if (sel.isPropertySelection()) {
          this._selectProperty(sel.path);
          event.preventDefault();
          event.stopPropagation();
          return
        } else if (sel.isContainerSelection()) {
          this._selectProperty(sel.startPath);
          event.preventDefault();
          event.stopPropagation();
          return
        }
      }
      
      
      
      
      this._state.skipNextFocusEvent = true;

      
      if (platform$1.inBrowser) {
        let documentEl = DefaultDOMElement$1.wrapNativeElement(window.document);
        documentEl.on('mouseup', this.onMouseUp, this, { once: true });
      }
    }

    onMouseUp (e) {
      
      
      
      
      
      e.stopPropagation();
      
      
      
      
      this._delayed(() => {
        let sel = this.domSelection.getSelection();
        this._setSelection(sel);
      });
    }

    
    
    
    onContextMenu (event) {
      if (!this._shouldConsumeEvent(event)) return
      let sel = this.domSelection.getSelection();
      this._setSelection(sel);
    }

    onNativeBlur () {
      
      let _state = this._state;
      _state.hasNativeFocus = false;
    }

    onNativeFocus () {
      
      let _state = this._state;
      _state.hasNativeFocus = true;
    }

    _onCopy (e) {
      this.clipboard.onCopy(e);
    }

    _onCut (e) {
      this.clipboard.onCut(e);
    }

    _onPaste (e) {
      this.clipboard.onPaste(e);
    }

    

    _onSelectionChanged (selection) {
      let newMode = this._deriveModeFromSelection(selection);
      if (this.state.mode !== newMode) {
        this.extendState({
          mode: newMode
        });
      }
    }

    
    _deriveModeFromSelection (sel) {
      if (!sel) return null
      let surfaceId = sel.surfaceId;
      let id = this.getId();
      let mode;
      if (startsWith$1(surfaceId, id)) {
        if (surfaceId.length === id.length) {
          mode = 'focused';
        } else {
          mode = 'co-focused';
        }
      }
      return mode
    }

    _updateContentEditableState () {
      
      
      
      
      
      let enableContenteditable = false;
      if (this.isEditable() && !this.props.disabled) {
        enableContenteditable = true;
        if (this.state.mode === 'co-focused') {
          let selState = this.context.editorSession.getSelectionState();
          let sel = selState.getSelection();
          let surface = this.getSurfaceManager().getSurface(sel.surfaceId);
          if (surface) {
            let isolatedNodeComponent = surface.context.isolatedNodeComponent;
            if (isolatedNodeComponent) {
              enableContenteditable = isolatedNodeComponent.isOpen();
            }
          }
        }
      }
      if (enableContenteditable) {
        this.el.setAttribute('contenteditable', true);
      } else {
        
        this.el.removeAttribute('contenteditable');
      }
    }

    _blur () {
      if (this.el) {
        this.el.blur();
      }
    }

    _focus () {
      if (this.isDisabled()) return
      
      
      
      if (platform$1.isFF) {
        this.domSelection.clear();
        this.el.getNativeElement().blur();
      }
      this._focusElement();
    }

    _focusElement () {
      this._state.hasNativeFocus = true;
      
      
      
      
      if (this.el && !platform$1.isWebkit) {
        this._state.skipNextFocusEvent = true;
        
        
        this.el.focus();
        this._state.skipNextFocusEvent = false;
      }
    }

    _handleLeftOrRightArrowKey (event) {
      event.stopPropagation();
      let direction = (event.keyCode === keys$1$1.LEFT) ? 'left' : 'right';
      
      
      this._delayed(() => {
        this._updateModelSelection({direction});
      });
    }

    _handleUpOrDownArrowKey (event) {
      event.stopPropagation();
      
      
      this._delayed(() => {
        let options = {
          direction: (event.keyCode === keys$1$1.UP) ? 'left' : 'right'
        };
        this._updateModelSelection(options);
      });
    }

    _handleHomeOrEndKey (event) {
      event.stopPropagation();
      
      
      this._delayed(() => {
        let options = {
          direction: (event.keyCode === keys$1$1.HOME) ? 'left' : 'right'
        };
        this._updateModelSelection(options);
      });
    }

    _handlePageUpOrDownKey (event) {
      event.stopPropagation();
      
      
      this._delayed(() => {
        let options = {
          direction: (event.keyCode === keys$1$1.PAGEUP) ? 'left' : 'right'
        };
        this._updateModelSelection(options);
      });
    }

    _handleSpaceKey (event) {
      event.stopPropagation();
      event.preventDefault();
      const text = ' ';
      if (!this.getEditorSession().keyboardManager.onTextInput(text)) {
        this.getEditorSession().transaction((tx) => {
          tx.insertText(text);
        }, { action: 'type' });
      }
    }

    _handleTabKey (event) {
      event.stopPropagation();
      this.el.emit('tab', {
        altKey: event.altKey,
        ctrlKey: event.ctrlKey,
        metaKey: event.metaKey,
        shiftKey: event.shiftKey,
        code: event.code
      });
      if (this.props.handleTab === false) {
        event.preventDefault();
      } else {
        this.__handleTab(event);
      }
    }

    __handleTab () {
      
      
      this._delayed(() => {
        this._updateModelSelection();
      });
    }

    _handleEnterKey (event) {
      event.preventDefault();
      event.stopPropagation();
      this.getEditorSession().transaction((tx) => {
        tx.break();
      }, { action: 'break' });
    }

    _handleEscapeKey () {}

    _handleDeleteKey (event) {
      event.preventDefault();
      event.stopPropagation();
      let direction = (event.keyCode === keys$1$1.BACKSPACE) ? 'left' : 'right';
      this.getEditorSession().transaction((tx) => {
        tx.deleteCharacter(direction);
      }, { action: 'delete' });
    }

    _hasNativeFocus () {
      return Boolean(this._state.hasNativeFocus)
    }

    _setSelection (sel) {
      
      
      
      
      
      
      
      
      if (!sel.isNull() && sel.surfaceId === this.id && platform$1.isFF) {
        this._focusElement();
      }
      this.getEditorSession().setSelection(sel);
    }

    _updateModelSelection (options) {
      let sel = this.domSelection.getSelection(options);
      
      
      
      this._setSelection(sel);
    }

    _selectProperty (path) {
      let doc = this.getDocument();
      let text = doc.get(path);
      this._setSelection(doc.createSelection({
        type: 'property',
        path: path,
        startOffset: 0,
        endOffset: text.length
      }));
    }

    
    
    _renderNode ($$, nodeId) {
      let doc = this.getDocument();
      let node = doc.get(nodeId);
      let componentRegistry = this.context.componentRegistry || this.props.componentRegistry;
      let ComponentClass = componentRegistry.get(node.type);
      if (!ComponentClass) {
        console.error('Could not resolve a component for type: ' + node.type);
        ComponentClass = UnsupportedNodeComponent$1;
      }
      return $$(ComponentClass, {
        doc: doc,
        node: node
      })
    }

    _extractNodeProps (node) {
      let doc = this.getDocument();
      return {
        placeholder: this.props.placeholder,
        doc: doc,
        node: node
      }
    }

    
    _shouldConsumeEvent (event) {
      
      let comp = Component$1.unwrap(event.target);
      return (comp && (comp === this || comp.context.surface === this))
    }

    
    getSelectionFromEvent (event) {
      let domRange = getDOMRangeFromEvent$1(event);
      let sel = this.domSelection.getSelectionForDOMRange(domRange);
      sel.surfaceId = this.getId();
      return sel
    }

    setSelectionFromEvent (event) {
      let sel = this.getSelectionFromEvent(event);
      if (sel) {
        this._state.skipNextFocusEvent = true;
        this._setSelection(sel);
      } else {
        console.error('Could not create a selection from event.');
      }
    }

    get id () {
      return this._surfaceId
    }

    _delayed (fn) {
      window.setTimeout(fn, BROWSER_DELAY$1);
    }

    
    _muteNativeHandlers (event) {
      let contentEditableShortcuts;

      if (platform$1.isMac) {
        contentEditableShortcuts = [
          'META+66', 
          'META+73', 
          'META+85' 
        ];
      } else {
        contentEditableShortcuts = [
          'CTRL+66', 
          'CTRL+73', 
          'CTRL+85' 
        ];
      }

      const key = parseKeyEvent$1(event);
      if (contentEditableShortcuts.indexOf(key) > -1) {
        event.preventDefault();
      }
    }
  }

  Surface$1.prototype._isSurface = true;


  Surface$1.createSurfaceId = function (surface) {
    let parentSurfaceId = surface.context.parentSurfaceId;
    if (parentSurfaceId) {
      return parentSurfaceId + '/' + surface.name
    } else {
      return surface.name
    }
  };

  class ContainerEditor$1 extends Surface$1 {
    constructor (parent, props, el) {
      
      props.containerId = props.containerId || props.node.id;
      props.name = props.name || props.containerId || props.node.id;

      super(parent, props, el);

      this.containerId = this.props.containerId;
      if (!isString$1(this.containerId)) {
        throw new Error("Property 'containerId' is mandatory.")
      }
      let doc = this.getDocument();
      this.container = doc.get(this.containerId);
      if (!this.container) {
        throw new Error('Container with id ' + this.containerId + ' does not exist.')
      }

      this.editingBehavior = this.context.editingBehavior || new EditingBehavior$1();

      this._deriveInternalState(this.props);
    }

    
    shouldRerender (newProps) {
      if (newProps.disabled !== this.props.disabled) return true
      
      
      return false
    }

    willReceiveProps (newProps) {
      super.willReceiveProps.apply(this, arguments);
      this._deriveInternalState(newProps);
    }

    didMount () {
      super.didMount.apply(this, arguments);
      let editorSession = this.getEditorSession();
      editorSession.onUpdate('document', this._onContainerChanged, this, {
        path: this.container.getContentPath()
      });
      this._attachPlaceholder();
    }

    dispose () {
      super.dispose.apply(this, arguments);
      let editorSession = this.getEditorSession();
      editorSession.off(this);
    }

    render ($$) {
      let el = super.render($$);

      let doc = this.getDocument();
      let containerId = this.getContainerId();
      let containerNode = doc.get(containerId);
      if (!containerNode) {
        console.warn('No container node found for ', containerId);
      }
      el.addClass('sc-container-editor container-node ' + containerId)
        .attr('data-id', containerId);

      
      el.attr('spellcheck', this.props.spellcheck === 'native');

      containerNode.getNodes().forEach(function (node, index) {
        el.append(this._renderNode($$, node, index));
      }.bind(this));

      
      if (!this.props.disabled && !this.isEmpty()) {
        el.addClass('sm-enabled');
        el.setAttribute('contenteditable', true);
      }

      return el
    }

    selectFirst () {
      const container = this.getContainer();
      if (container.getLength() > 0) {
        const editorSession = this.getEditorSession();
        const first = container.getChildAt(0);
        setCursor$1(editorSession, first, container.id, 'before');
      }
    }

    _renderNode ($$, node, nodeIndex) {
      let props = { node };
      if (!node) throw new Error('Illegal argument')
      if (node.isText()) {
        return this.renderNode($$, node, nodeIndex)
      } else {
        let componentRegistry = this.context.componentRegistry;
        let ComponentClass = componentRegistry.get(node.type);
        if (ComponentClass.prototype._isCustomNodeComponent || ComponentClass.prototype._isIsolatedNodeComponent) {
          return $$(ComponentClass, props).ref(node.id)
        } else {
          return $$(IsolatedNodeComponent$1, props).ref(node.id)
        }
      }
    }

    _extractNodeProps (node) {
      let doc = this.getDocument();
      return {
        doc: doc,
        node: node
      }
    }

    _deriveInternalState (props) {
      let _state = this._state;
      if (!props.hasOwnProperty('enabled') || props.enabled) {
        _state.enabled = true;
      } else {
        _state.enabled = false;
      }
    }

    _selectNextIsolatedNode (direction) {
      let selState = this.getEditorSession().getSelectionState();
      let node = (direction === 'left') ? selState.getPreviousNode() : selState.getNextNode();
      let isIsolatedNode = !node.isText() && !node.isList();
      if (!node || !isIsolatedNode) return false
      if (
        (direction === 'left' && selState.isFirst()) ||
        (direction === 'right' && selState.isLast())
      ) {
        this.getEditorSession().setSelection({
          type: 'node',
          nodeId: node.id,
          containerId: selState.getContainer().id,
          surfaceId: this.id
        });
        return true
      }
      return false
    }

    _handleLeftOrRightArrowKey (event) {
      event.stopPropagation();
      const doc = this.getDocument();
      const sel = this.getEditorSession().getSelection();
      const left = (event.keyCode === keys$1$1.LEFT);
      const right = !left;
      const direction = left ? 'left' : 'right';

      if (sel && !sel.isNull()) {
        const container = doc.get(sel.containerId, 'strict');

        
        if (sel.isNodeSelection()) {
          let nodePos = container.getPosition(doc.get(sel.getNodeId()));
          if ((left && nodePos === 0) || (right && nodePos === container.length - 1)) {
            event.preventDefault();
            return
          }
        }

        if (sel.isNodeSelection() && !event.shiftKey) {
          this.domSelection.collapse(direction);
        }
      }

      this._delayed(() => {
        this._updateModelSelection({ direction });
      });
    }

    _handleUpOrDownArrowKey (event) {
      event.stopPropagation();
      const doc = this.getDocument();
      const sel = this.getEditorSession().getSelection();
      const up = (event.keyCode === keys$1$1.UP);
      const down = !up;
      const direction = up ? 'left' : 'right';

      if (sel && !sel.isNull()) {
        const container = doc.get(sel.containerId, 'strict');
        
        if (sel.isNodeSelection()) {
          let nodePos = container.getPosition(doc.get(sel.getNodeId()));
          if ((up && nodePos === 0) || (down && nodePos === container.length - 1)) {
            event.preventDefault();
            return
          }
          
          
          let editorSession = this.getEditorSession();
          
          
          
          if (!event.shiftKey) {
            event.preventDefault();
            if (up) {
              let prev = container.getChildAt(nodePos - 1);
              setCursor$1(editorSession, prev, sel.containerId, 'after');
              return
            } else {
              let next = container.getChildAt(nodePos + 1);
              setCursor$1(editorSession, next, sel.containerId, 'before');
              return
            }
          }
        }
      }

      this._delayed(() => {
        this._updateModelSelection({ direction });
      });
    }

    _handleTabKey (event) {
      const editorSession = this.getEditorSession();
      const sel = editorSession.getSelection();
      
      if (sel.isNodeSelection() && sel.isFull()) {
        const comp = this.refs[sel.getNodeId()];
        if (comp && stepIntoIsolatedNode$1(editorSession, comp)) {
          event.preventDefault();
          event.stopPropagation();
          return
        }
      }
      super._handleTabKey(event);
    }

    __handleTab (e) {
      e.preventDefault();
      if (e.shiftKey) {
        this.getEditorSession().transaction((tx) => {
          tx.dedent();
        }, { action: 'dedent' });
      } else {
        this.getEditorSession().transaction((tx) => {
          tx.indent();
        }, { action: 'indent' });
      }
    }

    
    isContainerEditor () {
      return true
    }

    
    getContainerId () {
      return this.containerId
    }

    getContainer () {
      return this.getDocument().get(this.getContainerId())
    }

    isEmpty () {
      let containerNode = this.getContainer();
      return (containerNode && containerNode.length === 0)
    }

    
    _attachPlaceholder () {
      let firstNode = this.childNodes[0];
      
      if (this.placeholderNode) {
        this.placeholderNode.extendProps({
          placeholder: undefined
        });
      }

      if (this.childNodes.length === 1 && this.props.placeholder) {
        firstNode.extendProps({
          placeholder: this.props.placeholder
        });
        this.placeholderNode = firstNode;
      }
    }

    isEditable () {
      return super.isEditable.call(this) && !this.isEmpty()
    }

    
    _onContainerChanged (change) {
      let doc = this.getDocument();
      
      let renderContext = RenderingEngine$1.createContext(this);
      let $$ = renderContext.$$;
      let container = this.getContainer();
      let path = container.getContentPath();
      for (let i = 0; i < change.ops.length; i++) {
        let op = change.ops[i];
        if (op.type === 'update' && op.path[0] === path[0]) {
          let diff = op.diff;
          if (diff.type === 'insert') {
            let nodeId = diff.getValue();
            let node = doc.get(nodeId);
            let nodeEl;
            if (node) {
              nodeEl = this._renderNode($$, node);
            } else {
              
              
              
              nodeEl = $$('div');
            }
            this.insertAt(diff.getOffset(), nodeEl);
          } else if (diff.type === 'delete') {
            this.removeAt(diff.getOffset());
          }
        }
      }
      this._attachPlaceholder();
    }
  }

  ContainerEditor$1.prototype._isContainerEditor = true;

  class TextPropertyEditor$1 extends Surface$1 {
    constructor (parent, props) {
      
      props.name = props.name || props.path.join('.');
      super(parent, props);

      if (!props.path) {
        throw new Error("Property 'path' is mandatory.")
      }
    }

    render ($$) {
      let el = super.render.apply(this, arguments);
      el.addClass('sc-text-property-editor');

      if (!this.props.disabled) {
        el.addClass('sm-enabled');
        el.attr('contenteditable', true);
        
        el.attr('spellcheck', this.props.spellcheck === 'native');
      }

      el.append(
        $$(TextPropertyComponent$1, {
          placeholder: this.props.placeholder,
          tagName: this.props.tagName || 'div',
          path: this.props.path,
          markers: this.props.markers,
          withoutBreak: this.props.withoutBreak
        }).ref('property')
      );

      return el
    }

    selectFirst () {
      this.editorSession.setSelection({
        type: 'property',
        path: this.getPath(),
        startOffset: 0,
        surfaceId: this.id
      });
    }

    getPath () {
      return this.props.path
    }

    _handleEnterKey (event) {
      event.stopPropagation();
      if (this.props.handleEnter === false || !this.props.multiLine) {
        event.preventDefault();
      }
      this.el.emit('enter', {
        altKey: event.altKey,
        ctrlKey: event.ctrlKey,
        metaKey: event.metaKey,
        shiftKey: event.shiftKey,
        code: event.code
      });
    }

    

    _handleEscapeKey (event) {
      this.el.emit('escape', {
        altKey: event.altKey,
        ctrlKey: event.ctrlKey,
        metaKey: event.metaKey,
        shiftKey: event.shiftKey,
        code: event.code
      });
    }
  }

  TextPropertyEditor$1.prototype._isTextPropertyEditor = true;

  function series$1 (tasks, cb, i) {
    i = i || 0;
    tasks[i](function (err, ...args) {
      
      if (err) return cb(err)
      if (i === tasks.length - 1) {
        cb(err, ...args); 
      } else {
        series$1(tasks, cb, i + 1);
      }
    });
  }

  var async$1 = /*#__PURE__*/Object.freeze({
    series: series$1
  });

  const ALPHABET$1 = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';

  function getColumnLabel$1 (colIdx) {
    if (!isNumber$1(colIdx)) {
      throw new Error('Illegal argument.')
    }
    let label = '';
    while(true) { 
      let mod = colIdx % ALPHABET$1.length;
      colIdx = Math.floor(colIdx / ALPHABET$1.length);
      label = ALPHABET$1[mod] + label;
      if (colIdx > 0) colIdx--;
      else if (colIdx === 0) break
    }
    return label
  }

  function getRowCol$1 (cellLabel) {
    var match = /^([A-Z]+)([1-9][0-9]*)$/.exec(cellLabel);
    return [
      parseInt(match[2], 10) - 1,
      getColumnIndex$1(match[1])
    ]
  }

  const A$1 = 'A'.charCodeAt(0);
  const ALPHABET_LENGTH$1 = ALPHABET$1.length;

  function getColumnIndex$1 (colStr) {
    let index = 0;
    let rank = 0;
    for (let i = colStr.length - 1; i >= 0; i--) {
      let idx = colStr.charCodeAt(i) - A$1;
      if (idx < 0 || idx >= ALPHABET_LENGTH$1) throw new Error('Illegal column label: ' + colStr)
      
      index += Math.pow(ALPHABET_LENGTH$1, rank) * (idx + 1);
      rank++;
    }
    
    return index - 1
  }

  function getCellLabel$1 (rowIdx, colIdx) {
    let colLabel = getColumnLabel$1(colIdx);
    let rowLabel = rowIdx + 1;
    return colLabel + rowLabel
  }

  function getIndexesFromRange$3 (start, end) {
    let [startRow, startCol] = getRowCol$1(start);
    let endRow, endCol;
    if (end) {
      ([endRow, endCol] = getRowCol$1(end));
      if (startRow > endRow) ([startRow, endRow] = [endRow, startRow]);
      if (startCol > endCol) ([startCol, endCol] = [endCol, startCol]);
    } else {
      ([endRow, endCol] = [startRow, startCol]);
    }
    return { startRow, startCol, endRow, endCol }
  }

  function getRangeFromMatrix$2 (cells, startRow, startCol, endRow, endCol, force2D) {
    if (!force2D) {
      if (startRow === endRow && startCol === endCol) {
        let row = cells[startRow];
        if (row) return row[startCol]
        else return undefined
      }
      if (startRow === endRow) {
        let row = cells[startRow];
        if (row) return row.slice(startCol, endCol + 1)
        else return []
      }
      if (startCol === endCol) {
        let res = [];
        for (let i = startRow; i <= endRow; i++) {
          let row = cells[i];
          if (row) res.push(row[startCol]);
        }
        return res
      }
    }
    let res = [];
    for (var i = startRow; i < endRow + 1; i++) {
      let row = cells[i];
      if (row) res.push(row.slice(startCol, endCol + 1));
    }
    return res
  }

  var tableHelpers$1 = /*#__PURE__*/Object.freeze({
    ALPHABET: ALPHABET$1,
    getColumnLabel: getColumnLabel$1,
    getRowCol: getRowCol$1,
    getColumnIndex: getColumnIndex$1,
    getCellLabel: getCellLabel$1,
    getIndexesFromRange: getIndexesFromRange$3,
    getRangeFromMatrix: getRangeFromMatrix$2
  });

  function renderListNode$1 (listNode, $$) {
    let items = listNode.getItems();
    let stack = [$$(_getTagName(1))];
    for (let i = 0; i < items.length; i++) {
      const item = items[i];
      
      const level = item.getLevel();
      if (level < stack.length) {
        for (let j = stack.length; j > level; j--) {
          stack.pop();
        }
      } else if (level > stack.length) {
        for (let j = stack.length; j < level; j++) {
          let list = stack[j - 1];
          let childCount = list.getChildCount();
          let item;
          if (childCount === 0) {
            item = $$('li');
            list.append(item);
          } else {
            item = list.getChildAt(childCount - 1);
          }
          let sublist = $$(_getTagName(j + 1));
          item.append(sublist);
          stack.push(sublist);
        }
      }
      console.assert(level === stack.length, 'item.level should now be the same as stack.length');
      
      last$1$1(stack).append(
        $$(item)
      );
    }
    for (let j = stack.length; j > 1; j--) {
      stack.pop();
    }

    return stack[0]

    function _getTagName (level) {
      let listType = listNode.getListType(level);
      return listType === 'order' ? 'ol' : 'ul'
    }
  }

  const DISABLED$1 = Object.freeze({
    disabled: true
  });

  class ListItemComponent$1 extends Component$1 {
    render ($$) {
      const node = this.props.node;
      const path = node.getPath();

      let el = $$('li').addClass('sc-list-item');
      el.append(
        $$(TextPropertyComponent$1, { path }).ref('text')
      );
      
      if (this.props.children) {
        el.append(this.props.children);
      }
      return el
    }
  }

  class ListComponent$1 extends NodeComponent$1 {
    render ($$) {
      let node = this.props.node;
      let el = renderListNode$1(node, (item) => {
        
        if (isString$1(item)) {
          return $$(item)
        } else if (item.type === 'list-item') {
          return $$(ListItemComponent$1, {
            node: item
          }).ref(item.id)
        }
      });
      el.addClass('sc-list').attr('data-id', node.id);
      return el
    }
  }


  ListComponent$1.prototype._isCustomNodeComponent = true;

  class TableCellComponent$1 extends Component$1 {
    render ($$) {
      let node = this.props.node;
      let el = $$('td').addClass('sc-table-cell');
      el.append(
        $$(TextPropertyEditor$1, {
          path: node.getPath(),
          disabled: this.props.disabled
        }).ref('editor')
      );
      if (node.rowspan > 0) {
        el.attr('rowspan', node.rowspan);
      }
      if (node.colspan > 0) {
        el.attr('colspan', node.colspan);
      }
      return el
    }

    grabFocus () {
      let node = this.props.node;
      this.context.editorSession.setSelection({
        type: 'property',
        path: node.getPath(),
        startOffset: node.getLength(),
        surfaceId: this.refs.editor.id
      });
    }
  }

  TableCellComponent$1.prototype._isTableCellComponent = true;

  // TODO: this should go into its own repo: `stencila/types`

  // Parent of each type
  const parentTypes = {
    'any': null,
    'null': 'any',
    'boolean': 'any',
    'number': 'any',
    'integer': 'number',
    'string': 'any',
    'object': 'any',
    'array': 'any',
    'array[boolean]': 'array',
    'array[number]': 'array',
    'array[integer]': 'array[number]',
    'array[string]': 'array',
    'array[object]': 'array',
    'table': 'any'
  };

  // Children of each type
  const childrenTypes = {};
  for (let type of Object.keys(parentTypes)) {
    if (!childrenTypes[type]) childrenTypes[type] = [];
    let base = parentTypes[type];
    if (!base) continue
    if (childrenTypes[base]) childrenTypes[base].push(type);
    else childrenTypes[base] = [type];
  }

  // Descendants (children, grandchildren etc) of each type
  const descendantTypes = {};
  for (let type of Object.keys(parentTypes)) {
    if (!descendantTypes[type]) descendantTypes[type] = [];
    let parent = parentTypes[type];
    while (parent) {
      if (descendantTypes[parent]) descendantTypes[parent].push(type);
      else descendantTypes[parent] = [type];
      parent = parentTypes[parent];
    }
  }

  function coercedArrayType (arr) {
    let valType = arr.reduce(_mostSpecificType, undefined);
    if (valType === 'any') {
      return 'array'
    } else {
      return `array[${valType}]`
    }
  }

  /**
   * Get the type code for a value
   *
   * @memberof value
   * @param {*} value - A JavaScript value
   * @return {string} - Type code for value
   */
  function type (value) {
    let type = typeof value;

    if (isNil$1(value) === null) {
      return 'null'
    } else if (type === 'boolean') {
      return 'boolean'
    } else if (type === 'number') {
      let isInteger = false;
      if (value.isInteger) isInteger = value.isInteger();
      else isInteger = (value % 1) === 0;
      return isInteger ? 'integer' : 'number'
    } else if (type === 'string') {
      return 'string'
    } else if (type === 'object') {
      if (value.constructor === Array) {
        return 'array'
      }
      if (value.type) return value.type
      else return 'object'
    } else {
      return 'unknown'
    }
  }

  /*
    A helper to get a coerced array value from a given array of values.
  */
  function coerceArray (arr) {
    let type = coercedArrayType(arr);
    let data = arr.map(v => {
      if (v) {
        return v.data
      } else {
        return undefined
      }
    });
    return {
      type,
      data
    }
  }

  // used to reduce an array type by 'bubbling' a type
  // through an array of types coercing to the most specific possible
  // TODO: what is type actually?
  function _mostSpecificType (type, val) {
    // treat nil is any
    if (isNil$1(val)) return 'any'
    let valType = val.type;
    if (!type) return valType || type(val)
    if (type === valType) return type
    switch (type) {
      case 'number': {
        if (valType === 'integer') {
          return 'number'
        }
        break
      }
      case 'integer': {
        if (valType === 'number') {
          return 'number'
        }
        break
      }
      default:
        //
    }
    return 'any'
  }

  function pack (value, opts = {}) {
    if (isNil$1(value)) {
      return { type: 'null', data: null }
    }
    if (typeof value === 'function') {
      let data = { name: value.name };
      if (opts.context) data.context = opts.context.id;
      if (opts.cell) data.id = opts.cell.id;
      return {
        type: 'function',
        data
      }
    }
    if (value.type === 'image') {
      return { type: 'image', src: value.src }
    } else {
      let _type = value.type || type(value);
      return { type: _type, data: value }
    }
  }

  function unpack (pkg) {
    if (isNil$1(pkg)) return undefined
    return pkg.data
  }

  // TODO: a shared repo for this would be nice

  class CompositeContext {
    constructor () {
      this._contexts = {};
      this._lang2context = {};
    }

    configure (config) {
      let contexts = {};
      let lang2context = {};
      config.contexts.forEach(spec => {
        let lang = spec.lang;
        let id = spec.id;
        let ClientClass = spec.client;
        let args = spec.args || [];
        let context = new ClientClass(this, id, ...args);
        contexts[id] = context;
        lang2context[lang] = context;
      });
      this._contexts = contexts;
      this._lang2context = lang2context;

      config.libraries.forEach(({lang, lib}) => {
        let langContext = this.getLanguageContext(lang);
        langContext.importLibrary(lib);
      });
    }

    getLanguageContext (lang) {
      return this._lang2context[lang]
    }

    async compile (cell) {
      const lang = cell.lang;
      const context = this.getLanguageContext(lang);
      if (!context) {
        // TODO: need to think about how initialization should be done
        Object.assign(cell, {messages: [], inputs: [], outputs: []});
        cell.messages.push(new ContextError(`No context for language '${lang}'`));
        return cell
      } else {
        return context.compile(cell)
      }
    }

    async execute (cell) {
      const lang = cell.lang;
      const context = this.getLanguageContext(lang);
      if (!context) {
        if (!cell.messages) cell.messages = [];
        cell.messages.push(new ContextError(`No context for language '${lang}'`));
        return cell
      } else {
        return context.execute(cell)
      }
    }

    /**
      @param {object} a function value (TODO: link to documentation)
      @param {array} an array of packed values
      @param {object} a Map with packed values by name
    */
    async callFunction (funcValue, args, namedArgs) {
      // HACK: this is being used inconsistently
      let data = funcValue.data || funcValue.value.data;
      const contextId = data.context;
      if (!contextId) {
        throw new Error('context is mandatory')
      }
      // TODO: instead of storing contexts via language
      // I would prefer to store them by id
      // and do the language mapping extra
      let context = this._contexts[contextId];
      if (!context) {
        throw new Error('No context registered for language')
      }
      let call = {
        type: 'call',
        func: funcValue,
        args,
        namedArgs
      };
      return context.evaluateCall(call)
    }

    pack (value, options) {
      return pack(value, options)
    }

    unpack (pkg) {
      return unpack(pkg)
    }
  }

  class EngineCellGraph extends CellGraph {
    constructor (engine) {
      super();

      this._engine = engine;
      // forward dependencies / impact
      // for a every sheet cell all other cells that depend on it via cell or range ref
      // id -> Set
      this._sheetCellOuts = {};
      // id -> Set
      this._sheetCellIns = {};
    }

    _getDoc (s) {
      return this._engine._docs[s.docId]
    }

    // Note: this is overridden to register and deregister
    // cell dependencies expressed by cell or range references
    _setInputs (cell, newInputs) {
      super._setInputs(cell, newInputs);

      // first remove the cell from all the cells it has depended on first
      let oldDeps = this._sheetCellIns[cell.id];
      if (oldDeps) {
        oldDeps.forEach(dep => {
          let _deps = this._sheetCellOuts[dep.id];
          if (_deps) {
            _deps.delete(cell);
          }
        });
      }

      newInputs.forEach(s => {
        if (s.type !== 'var') {
          let sheet = this._getDoc(s);
          if (sheet) {
            let deps = sheet._getCellsForRange(s);
            this._addDependencies(cell, deps);
          }
        }
      });
    }

    _addDependencies (cell, deps) {
      let ins = this._sheetCellIns[cell.id];
      if (!ins) {
        ins = this._sheetCellIns[cell.id] = new Set();
      }
      deps.forEach(dep => {
        let outs = this._sheetCellOuts[dep.id];
        if (!outs) {
          outs = this._sheetCellOuts[dep.id] = new Set();
        }
        outs.add(cell);
        ins.add(dep);
      });
    }

    _resolve (s) {
      switch (s.type) {
        case 'cell': {
          let sheet = this._getDoc(s);
          if (sheet) {
            let row = sheet.cells[s.startRow];
            if (row) {
              let cell = row[s.startCol];
              if (cell) return cell.id
            }
          }
          break
        }
        case 'range': {
          let sheet = this._getDoc(s);
          if (sheet) {
            let cells = tableHelpers.getRangeFromMatrix(sheet.cells, s.startRow, s.startRow, s.endRow, s.endCol);
            return flatten(cells).map(c => c.id)
          }
          break
        }
        default:
          return super._resolve(s)
      }
    }

    _getAffected (cell) {
      let affected = super._getAffected(cell);
      // Note: in addition to explicit dependencies of sheet cells
      // we add all cells that depend on this cell via a cell or range expression
      if (cell.isSheetCell()) {
        let cells = this._sheetCellOuts[cell.id];
        if (cells) cells.forEach(cell => affected.push(cell.id));
      }
      return affected
    }
  }

  class SheetCell extends Cell$1 {
    isSheetCell () { return true }
  }

  const { getCellLabel: getCellLabel$2, getColumnLabel: getColumnLabel$2 } = tableHelpers;

  /*
    Engine's internal model of a Spreadsheet.
  */
  class Sheet {
    constructor (engine, data) {
      this.engine = engine;
      const docId = data.id;
      if (!docId) throw new Error("'id' is required")
      this.id = docId;
      this.name = data.name;
      // default language
      const defaultLang = data.lang || 'mini';
      this.lang = defaultLang;
      if (data.hasOwnProperty('autorun')) {
        this.autorun = data.autorun;
      } else {
        // TODO: using auto/ cells automatically by default
        this.autorun = true;
      }
      // TODO: we can revise this as we move on
      // for now, data.cells must be present being a sequence of rows of cells.
      // data.columns is optional, but if present every data row have corresponding dimensions
      if (!data.cells) throw new Error("'cells' is mandatory")
      let ncols;
      if (data.columns) {
        this.columns = data.columns;
      } else {
        ncols = data.cells[0].length;
        let columns = [];
        for (let i = 0; i < ncols; i++) {
          columns.push({ type: 'auto' });
        }
        this.columns = columns;
      }
      ncols = this.columns.length;
      this.cells = data.cells.map((rowData) => {
        if (rowData.length !== ncols) throw new Error('Invalid data')
        return rowData.map(cellData => this._createCell(cellData))
      });

      if (data.onCellRegister) this.onCellRegister = data.onCellRegister;
    }

    get type () { return 'sheet' }

    setAutorun (val) {
      this.autorun = val;
    }

    getColumnName (colIdx) {
      let columnMeta = this.columns[colIdx];
      if (columnMeta && columnMeta.name) {
        return columnMeta.name
      } else {
        return getColumnLabel$2(colIdx)
      }
    }

    getCells () {
      return this.cells
    }

    updateCell (id, cellData) {
      let qualifiedId$$1 = qualifiedId(this.id, id);
      if (isString(cellData)) {
        cellData = { source: cellData };
      }
      this.engine._updateCell(qualifiedId$$1, cellData);
    }

    insertRows (pos, dataBlock) {
      const graph = this.engine._graph;
      // TODO: what if all columns and all rows had been removed
      const count = dataBlock.length;
      if (count === 0) return
      const ncols = this.columns.length;
      let block = dataBlock.map((rowData) => {
        if (rowData.length !== ncols) throw new Error('Invalid data')
        return rowData.map(cellData => this._createCell(cellData))
      });
      let affectedCells = new Set();
      let spans = transformCells(this.engine, this.cells, 0, pos, count, affectedCells);
      // add the spanning symbols to the deps of the new cells
      for (let i = 0; i < block.length; i++) {
        let row = block[i];
        for (let j = 0; j < row.length; j++) {
          let cell = row[j];
          if (spans[j]) {
            graph._addDependencies(cell, spans[j]);
          }
        }
      }
      // update sheet structure
      this.cells.splice(pos, 0, ...block);
      this._registerCells(block);
      this._sendSourceUpdate(affectedCells);
    }

    deleteRows (pos, count) {
      if (count === 0) return
      let affectedCells = new Set();
      let block = this.cells.slice(pos, pos + count);
      transformCells(this.engine, this.cells, 0, pos, -count, affectedCells);
      this.cells.splice(pos, count);
      this._unregisterCells(block);
      this._sendSourceUpdate(affectedCells);
    }

    insertCols (pos, dataBlock) {
      const graph = this.engine._graph;
      const nrows = this.cells.length;
      if (dataBlock.length !== nrows) throw new Error('Invalid dimensions')
      let count = dataBlock[0].length;
      if (count === 0) return
      let affectedCells = new Set();
      // transform cells
      let spans = transformCells(this.engine, this.cells, 1, pos, count, affectedCells);
      let block = dataBlock.map((rowData) => {
        if (rowData.length !== count) throw new Error('Invalid data')
        return rowData.map(cellData => this._createCell(cellData))
      });
      let cols = [];
      for (let i = 0; i < count; i++) {
        cols.push({ type: 'auto' });
      }
      this.columns.splice(pos, 0, ...cols);
      for (let i = 0; i < nrows; i++) {
        let row = this.cells[i];
        row.splice(pos, 0, ...block[i]);
      }
      // add the spanning symbols to the deps of the new cells
      for (let i = 0; i < block.length; i++) {
        let row = block[i];
        for (let j = 0; j < row.length; j++) {
          let cell = row[j];
          if (spans[i]) {
            graph._addDependencies(cell, spans[i]);
          }
        }
      }
      this._registerCells(block);
      this._sendSourceUpdate(affectedCells);
    }

    deleteCols (pos, count) {
      if (count === 0) return
      let affectedCells = new Set();
      transformCells(this.engine, this.cells, 1, pos, -count, affectedCells);
      const N = this.cells.length;
      let block = [];
      this.columns.splice(pos, count);
      for (var i = 0; i < N; i++) {
        let row = this.cells[i];
        block.push(row.slice(pos, pos + count));
        row.splice(pos, count);
      }
      this._unregisterCells(block);
      this._sendSourceUpdate(affectedCells);
    }

    rename (newName) {
      if (newName === this.name) return
      const graph = this.engine._graph;
      let cells = this.cells;
      let affectedCells = new Set();
      for (let i = 0; i < cells.length; i++) {
        let row = cells[i];
        for (let j = 0; j < row.length; j++) {
          let cell = row[j];
          let deps = graph._sheetCellOuts[cell.id];
          deps.forEach(dep => {
            // update all symbols that point to this document
            dep.symbols.forEach(s => {
              if (s.scope === this.name) {
                s._update = { type: 'rename', scope: newName };
              }
            });
            affectedCells.add(dep);
          });
        }
      }
      affectedCells.forEach(cell => {
        applyCellTransformations(cell);
        this.engine._resetCell(cell);
      });
      this.name = newName;
      this._sendSourceUpdate(affectedCells);
    }

    onCellRegister(cell) { // eslint-disable-line
    }

    dump () {
      let columns = this.columns.map(c => clone(c));
      let cells = this.cells.map(row => row.map(cell => cell.dump()));
      return {
        type: 'sheet',
        id: this.id,
        name: this.name,
        lang: this.lang,
        autorun: this.autorun,
        columns,
        cells
      }
    }

    _getCellSymbol (rowIdx, colIdx) {
      return `${this.id}!${getCellLabel$2(rowIdx, colIdx)}`
    }

    _createCell (cellData) {
      // simple format: just the expression
      if (isString(cellData)) {
        let source = cellData;
        cellData = {
          id: uuid(),
          docId: this.id,
          source
        };
      }
      let cell = new SheetCell(this, cellData);
      return cell
    }

    _registerCell (cell) {
      const engine = this.engine;
      engine._registerCell(cell);
      this.onCellRegister(cell);
    }

    _unregisterCell (cell) {
      const engine = this.engine;
      engine._unregisterCell(cell);
    }

    _registerCells (block) {
      if (!block) block = this.cells;
      block.forEach(row => row.forEach(cell => this._registerCell(cell)));
    }

    _unregisterCells (block) {
      if (!block) block = this.cells;
      block.forEach(row => row.forEach(cell => this._unregisterCell(cell)));
    }

    _getCellsForRange (s) {
      const cells = this.cells;
      let result = new Set();
      for (let i = s.startRow; i <= s.endRow; i++) {
        let row = cells[i];
        for (let j = s.startCol; j <= s.endCol; j++) {
          let cell = row[j];
          if (!cell) {
            console.error('FIXME: SOMETHING IS BROKEN HERE');
          } else {
            result.add(cell);
          }
        }
      }
      return result
    }

    _sendSourceUpdate (cells) {
      if (cells.size > 0) {
        this.engine._sendUpdate('source', cells);
      }
    }
  }

  /*
    Records symbol updates and applies to cells.
    Additionally determines cells that have a symbol that is spanning
    over the insertion position.
  */
  function transformCells (engine, cells, dim, pos, count, affected) {
    if (count === 0) return
    // track updates for symbols and affected cells
    let startRow = 0;
    let startCol = 0;
    if (dim === 0) {
      startRow = pos;
    } else {
      startCol = pos;
    }
    const graph = engine._graph;
    let visited = new Set();
    for (let i = startRow; i < cells.length; i++) {
      let row = cells[i];
      for (let j = startCol; j < row.length; j++) {
        let cell = row[j];
        let deps = graph._sheetCellOuts[cell.id];
        if (deps) {
          recordTransformations(deps, dim, pos, count, affected, visited);
        }
      }
    }
    // NOTE: this seems a bit weird, but this needs to be done before the cell
    // symbols are changed.
    let spans = _getCellsWithSpanningSymbols(cells, dim, pos);
    // update the source for all affected cells
    affected.forEach(cell => {
      applyCellTransformations(cell);
      if (engine) {
        engine._resetCell(cell);
      }
    });
    return spans
  }

  /*
  * Some symbols are spanning the insert position. The cells with these
  * symbols need to be added the the dependencies for the inserted cells.
  * This method is only used after symbol updates have been recorded, i.e.
  * stored into `s._update`
  */
  function _getCellsWithSpanningSymbols (graph, cells, dim, pos) {
    // Note: it is enough to walk along the row/col where cells are inserted
    // and investigate their symbols
    let spans = [];
    if (pos > 0) {
      if (cells.length === 0 || cells[0].length === 0) return
      let size = [cells.length, cells[0].length];
      if (pos >= size[dim]) return
      // check cells along the other dimension
      let L = dim === 0 ? size[1] : size[0];
      for (let i = 0; i < L; i++) {
        let cell = dim === 0 ? cells[pos][i] : cells[i][pos];
        let deps = graph._sheetCellIns[cell.id];
        if (deps) {
          for (let dep of deps) {
            for (let s of dep.symbols) {
              let update = s._update;
              if (update && update.start <= pos) {
                if (!spans[i]) spans[i] = new Set();
                spans[i].add(dep);
              }
            }
          }
        }
      }
    }
    return spans
  }

  /*
    Engine's internal model of a Document.
  */
  class Document$2 {
    constructor (engine, data) {
      this.engine = engine;
      this.data = data;
      if (!data.id) throw new Error("'id' is required")
      this.id = data.id;
      this.name = data.name;
      this.lang = data.lang || 'mini';
      if (data.hasOwnProperty('autorun')) {
        this.autorun = data.autorun;
      } else {
        // TODO: using manual execution as a default for now
        this.autorun = true;
      }
      this.cells = data.cells.map(cellData => this._createCell(cellData));
      // registration hook used for propagating initial cell state to the application
      if (data.onCellRegister) this.onCellRegister = data.onCellRegister;
    }

    get type () { return 'document' }

    getCells () {
      return this.cells
    }

    setAutorun (val) {
      this.autorun = val;
    }

    insertCellAt (pos, cellData) {
      let cell = this._createCell(cellData);
      this._registerCell(cell);
      this.cells.splice(pos, 0, cell);
      return cell
    }

    removeCell (id) {
      const qualifiedId$$1 = qualifiedId(this.id, id);
      const cells = this.cells;
      let pos = cells.findIndex(cell => cell.id === qualifiedId$$1);
      if (pos >= 0) {
        let cell = cells[pos];
        this.cells.splice(pos, 1);
        this.engine._unregisterCell(cell);
      } else {
        console.error('Unknown cell', id);
      }
    }

    updateCell (id, cellData) {
      let qualifiedId$$1 = qualifiedId(this.id, id);
      if (isString(cellData)) {
        cellData = { source: cellData };
      }
      this.engine._updateCell(qualifiedId$$1, cellData);
    }

    rename (newName) {
      if (newName === this.name) return
      const docId = this.id;
      let graph = this.engine._graph;
      let cells = this.cells;
      let affectedCells = new Set();
      for (let i = 0; i < cells.length; i++) {
        let cell = cells[i];
        if (graph._cellProvidesOutput(cell)) {
          let ids = graph._ins[cell.output];
          if (ids) {
            ids.forEach(cellId => {
              let cell = graph.getCell(cellId);
              cell.inputs.forEach(s => {
                if (s.docId === docId) {
                  s._update = { type: 'rename', scope: newName };
                }
              });
              affectedCells.add(cell);
            });
          }
        }
      }
      affectedCells.forEach(cell => {
        applyCellTransformations(cell);
        this.engine._resetCell(cell);
      });
      this.name = newName;
      this._sendSourceUpdate(affectedCells);
    }

    onCellRegister(cell) { // eslint-disable-line
    }

    dump () {
      let cells = this.cells.map(cell => cell.dump());
      return {
        type: 'document',
        id: this.id,
        name: this.name,
        lang: this.lang,
        autorun: this.autorun,
        cells
      }
    }

    _createCell (cellData) {
      if (isString(cellData)) {
        let source = cellData;
        cellData = {
          id: uuid(),
          docId: this.id,
          source,
          lang: this.lang
        };
      }
      return new Cell$1(this, cellData)
    }

    _registerCell (cell) {
      const engine = this.engine;
      engine._registerCell(cell);
      this.onCellRegister(cell);
    }

    _registerCells (block) {
      if (!block) block = this.cells;
      block.forEach(cell => this._registerCell(cell));
    }

    _sendSourceUpdate (cells) {
      if (cells.size > 0) {
        this.engine._sendUpdate('source', cells);
      }
    }
  }

  /**
    The Engine implements the Stencila Execution Model.

    The Engine can be run independently, and thus has its own internal model.

    There are two types of resources containing cells, Documents and Sheets.
    Every document defines a variable scope. Variables are produced by cells.
    A document has an id but also a human readable name.
    Sheets have a tabular layout, while Documents have a sequential layout.

    Document cells can define variables which can be referenced within the same document,
    like in `x + 1`.

    Sheet cells can be referenced via cell- and range expressions, such as
    `A1`, or `A1:B10`.

    Across documents and sheets, cells are referenced using a transclusion syntax, prefixed with the document id or the document name, such as in
    `'My Document'!x` or `sheet1!A1:B10`.
  */
  class Engine extends EventEmitter {
    constructor (context) {
      super();

      if (!context) throw new Error('context is required')
      this.context = context;

      this._docs = {};
      this._graph = new EngineCellGraph(this);

      // for every (actionable) cell there is information what to do next
      // There are several steps that need to be done, to complete a cell:
      // - code analysis (context)
      // - registration of inputs/output (graph)
      // - cell evaluation (context)
      // - validation (engine)
      // - graph update
      this._nextActions = new Map();
      this._currentActions = new Map();
    }

    runOnce () {
      const self = this;
      return new Promise(resolve => {
        function step () {
          if (self._needsUpdate()) {
            self.cycle().then(step);
          } else {
            resolve();
          }
        }
        step();
      })
    }

    cycle () {
      return Promise.all(this._cycle())
    }

    _needsUpdate () {
      const graph = this._graph;
      if (graph.needsUpdate()) return true
      const nextActions = this._nextActions;
      if (nextActions.size === 0) return false
      // update is required if there is an action that has not been suspended
      for (let [, a] of nextActions) {
        if (!a.suspended) return true
      }
      return false
    }

    runForEver (refreshInterval) {
      // TODO: does this only work in the browser?
      if (this._runner) {
        clearInterval(this._runner);
      }
      this._runner = setInterval(() => {
        if (this.needsUpdate()) {
          this._cycle();
        }
      }, refreshInterval);
    }

    run (interval) {
      this.runForEver(interval);
    }

    /*
      Registers a document via id.

      @param {object} data
        - `type`: 'document' | 'sheet'
        - `name`: a human readable name used for transclusions
        - `columns`: (for sheets) initial column data
        - 'sequence': (for documents) initial order of cells
    */
    addDocument (data) {
      let doc = new Document$2(this, data);
      this._registerResource(doc);
      return doc
    }

    addSheet (data) {
      let sheet = new Sheet(this, data);
      this._registerResource(sheet);
      return sheet
    }

    dump () {
      let resources = [];
      forEach(this._docs, (doc, id) => {
        resources.push(doc.dump());
      });
      return {
        resources
      }
    }

    hasResource (id) {
      return this._docs.hasOwnProperty(id)
    }

    getResource (id) {
      return this._docs[id]
    }

    needsUpdate () {
      return this._nextActions.size > 0 || this._graph.needsUpdate()
    }

    _cycle () {
      let res = [];
      const graph = this._graph;
      const nextActions = this._nextActions;
      if (nextActions.size > 0) {
        // console.log('executing cycle')
        // clearing next actions so that we can record new next actions
        this._nextActions = new Map();

        // group actions by type
        let actions = {
          analyse: [],
          register: [],
          evaluate: [],
          update: []
        };
        nextActions.forEach(a => actions[a.type].push(a));
        actions.update.forEach(a => {
          if (a.errors && a.errors.length > 0) {
            graph.addErrors(a.id, a.errors);
          } else {
            graph.setValue(a.id, a.value);
          }
        });
        actions.register.forEach(a => {
          let cell = graph.getCell(a.id);
          graph.setInputsOutputs(cell.id, a.inputs, a.output);
        });

        this._updateGraph();

        let A = actions.analyse.map(a => this._analyse(a));
        let B = actions.evaluate.map(a => {
          let cell = graph.getCell(a.id);
          // This is necessary because we make sure the cell still exists
          if (cell) {
            if (this._canRunCell(cell)) {
              return this._evaluate(a)
            } else {
              // otherwise keep this as a next action
              a.suspended = true;
              this._setAction(a.id, a);
              return false
            }
          } else {
            return false
          }
        });
        res = A.concat(B);
      } else if (graph.needsUpdate()) {
        this._updateGraph();
      }
      return res
    }

    getNextActions () {
      return this._nextActions
    }

    _registerResource (doc) {
      const id = doc.id;
      if (this._docs.hasOwnProperty(id)) throw new Error(`document with id ${id} already exists`)
      this._docs[id] = doc;
      doc._registerCells();
    }

    /*
      Registers a cell.

      A cell is registered independent from the topology it resides in.

      Cells are treated differently w.r.t. their parent document.

      For instance, in a document cells can be block expressions,
      and can define a variable. In a sheet every cell must be a simple expression
      and it is is assigned to a variable implicitly (such as `sheet1!A1`).
    */
    _registerCell (cell) {
      this._graph.addCell(cell);
      this._resetCell(cell);
    }

    /*
      Removes a cell from the engine.
    */
    _unregisterCell(cellOrId) { // eslint-disable-line
      let id = isString(cellOrId) ? cellOrId : cellOrId.id;
      let cell = this._graph.getCell(id);
      if (cell) {
        this._graph.removeCell(id);
      }
    }

    _updateCell (id, cellData) {
      const graph = this._graph;
      let cell = graph.getCell(id);
      Object.assign(cell, cellData);
      cell.status = UNKNOWN;
      this._setAction(id, {
        id,
        type: 'analyse'
      });
    }

    _resetCell (cell) {
      const id = cell.id;
      cell.status = UNKNOWN;
      this._setAction(id, {
        id,
        type: 'analyse'
      });
    }

    _sendUpdate (type$$1, cells) {
      let cellsByDocId = {};
      cells.forEach(cell => {
        let _cells = cellsByDocId[cell.docId];
        if (!_cells) _cells = cellsByDocId[cell.docId] = [];
        _cells.push(cell);
      });
      this.emit('update', type$$1, cellsByDocId);
    }

    _updateGraph () {
      const graph = this._graph;
      let updatedIds = graph.update();
      let cells = new Set();
      updatedIds.forEach(id => {
        let cell = graph.getCell(id);
        if (cell) {
          // WIP: adding support for RangeCells
          // Instead of registering an evaluation, we just update the graph.
          // TODO: this requires another cycle to propagate the result of the RangeCell,
          // which would not be necessary in theory
          if (cell.status === READY) {
            this._setAction(cell.id, {
              type: 'evaluate',
              id: cell.id
            });
          }
          cells.add(cell);
        }
      });
      if (cells.size > 0) {
        this._sendUpdate('state', cells);
      }
    }

    _analyse (action) {
      const graph = this._graph;
      const id = action.id;
      const cell = graph.getCell(id);
      // if the cell has been removed in the meantime
      if (!cell) return
      // clear all errors which are not managed by the CellGraph
      cell.clearErrors(e => {
        return e.type !== 'graph'
      });
      // in case of constants, casting the string into a value,
      // updating the cell graph and returning without further evaluation
      if (cell.isConstant()) {
        // TODO: we might want to coerce to the type from the cell
        let value = pack(parseValue(cell.source));
        // constants can't have inputs, so deregister them
        if (cell.inputs && cell.inputs.size > 0) {
          graph.setInputs(id, new Set());
        }
        // constants can't have errors at this stage (later on maybe validation errors)
        if (cell.errors.length > 0) {
          graph.clearErrors(id);
        }
        graph.setValue(id, value);
        return
      }
      // TODO: we need to reset the cell status. Should we let CellGraph do this?
      cell.status = UNKNOWN;
      // leave a mark that we are currently running this action
      this._currentActions.set(id, action);

      const lang = cell.getLang();
      const isExpr = cell.isSheetCell();
      const transpiledSource = cell.transpiledSource;
      return this.context.compile({
        id: cell.id,
        code: transpiledSource,
        lang,
        expr: isExpr
      }).then(res => {
        if (this._isSuperseded(id, action)) {
        // console.log('action has been superseded')
          return
        }
        this._currentActions.delete(id);
        // storing the cell representation of the context
        cell.data = res;
        // make sure that the cell id is set
        res.id = id;

        // Note: treating all errors coming from analyseCode() as SyntaxErrors
        // TODO: we might want to be more specific here
        if (res.messages && res.messages.length > 0) {
          // TODO: we should not need to set this manually
          cell.status = ANALYSED;
          graph.addErrors(id, res.messages.map(err => {
            console.error(err);
            if (err instanceof CellError) {
              return err
            } else {
              return new SyntaxError$1(err.message)
            }
          }));
        }
        // console.log('analysed cell', cell, res)

        // mapping the result from the context to the engine's internal format
        let inputs = new Set();
        // TODO: at some point we want to allow for multiple outputs
        let output = null;
        if (res.inputs.length > 0 || res.outputs.length > 0) {
          // transform the extracted symbols into fully-qualified symbols
          // e.g. in `x` in `sheet1` is compiled into `sheet1.x`
          // At this point symbols are bound to a specific scope
          ({ inputs, output } = this._compileSymbols(res, cell));

          // TODO: this was originally here to make the app more robust
          // but trying to get rid it
          // try {
          //   ({ inputs, output } = this._compileSymbols(res, cell))
          // } catch (error) {
          //   console.error(error)
          //   cell.status = ANALYSED
          //   graph.addErrors(id, [new SyntaxError('Invalid syntax')])
          // }
        }
        this._setAction(id, {
          type: 'register',
          id,
          // Note: these symbols are in plain-text analysed by the context
          // based on the transpiled source
          inputs,
          output
        });
      })
    }

    _evaluate (action) {
      const graph = this._graph;
      const id = action.id;
      const cell = graph.getCell(id);
      cell.clearErrors(e => {
        return e.type !== 'graph'
      });
      // console.log('evaluating cell', cell.toString())
      this._currentActions.set(id, action);
      // EXPERIMENTAL: remove 'autorun' so that the cell is not updated forever
      delete cell.autorun;
      // prepare inputs for the context
      this._getInputValues(cell);
      // execute
      let p = this.context.execute(cell.data).then(res => {
        if (this._isSuperseded(id, action)) {
        // console.log('action has been superseded')
          return
        }
        this._currentActions.delete(id);
        let value;
        let output = res.outputs[0];
        if (output) value = output.value;
        this._setAction(id, {
          type: 'update',
          id,
          errors: res.messages,
          value
        });
      });
      return p.catch(err => {
        console.error(err);
        graph.addError(id, new RuntimeError('Internal error', err));
      })
    }

    // create symbols that can be passed to the cell graph
    _compileSymbols (res, cell) {
      const sourceSymbolMapping = cell._source.symbolMapping;
      const docId = cell.docId;
      const symbolMapping = {};
      let inputs = new Set();
      // Note: the inputs here are given as mangledStr
      // typically we have detected these already during transpilation
      // Let's wait for it to happen where this is not the case
      res.inputs.forEach(input => {
        let name = input.name;
        // HACK
        // TODO: when do we need this?
        if (isString(input)) {
          console.error('FIXME: input is in an unexpected format');
          name = input;
        }
        let symbol = sourceSymbolMapping[name];
        // Note: the engine does not track function names as symbols
        // which are returned as input
        // in this case we create a locally bound symbol
        if (!symbol) {
          symbol = new CellSymbol('var', name, docId, cell);
        } else {
          // if there is a scope given explicily try to lookup the doc
          // otherwise it is a local reference, i.e. within the same document as the cell
          let targetDocId = symbol.scope ? this._lookupDocumentId(symbol.scope) : docId;
          symbol = new CellSymbol(symbol.type, symbol.name, targetDocId, cell);
        }
        symbolMapping[name] = symbol;
        inputs.add(symbol);
      });
      cell._symbolMapping = symbolMapping;
      // turn the output into a qualified id
      let output = res.outputs[0];
      if (output) output = qualifiedId(docId, output.name);
      return { inputs, output }
    }

    /*
      Provides packed values stored in a hash by their name.
      Ranges and transcluded symbols are stored via their mangled name.

      > Attention: this requires that cell code is being transpiled accordingly.

      ```
      $ graph._getInputValues(['x', 'sheet1!A1:B3'])
      {
        'x': ...,
        'sheet1_A1_B3': ...
      }
      ```
    */
    _getInputValues (cell) {
      const graph = this._graph;
      for (let input of cell.data.inputs) {
        let symbolMapping = cell._symbolMapping;
        let s = symbolMapping[input.name];
        let val;
        switch (s.type) {
          case 'cell': {
            let sheet = this._docs[s.docId];
            if (sheet) {
              let cell = sheet.cells[s.startRow][s.startCol];
              val = cell.value;
            }
            break
          }
          case 'range': {
            let sheet = this._docs[s.docId];
            if (sheet) {
              val = _getValueForRange(sheet, s.startRow, s.startCol, s.endRow, s.endCol);
            }
            break
          }
          default:
            val = graph.getValue(s) || graph._globals.get(s.name);
        }
        input.value = val;
      }
    }

    _lookupDocumentId (name) {
      for (var id in this._docs) { // eslint-disable-line guard-for-in
        let doc = this._docs[id];
        if (doc.name === name || id === name) {
          return doc.id
        }
      }
    }

    _lookupDocument (name) {
      let docId = this._lookupDocumentId(name);
      return this._docs[docId]
    }

    _canRunCell (cell) {
      if (cell.hasOwnProperty('autorun')) {
        return cell.autorun
      }
      return cell.doc.autorun
    }

    _allowRunningCellAndPredecessors (id) {
      const graph = this._graph;
      let predecessors = graph._getPredecessorSet(id);
      this._allowRunningCell(id, true);
      predecessors.forEach(_id => {
        this._allowRunningCell(_id);
      });
    }

    _allowRunningCell (id, reset) {
      const graph = this._graph;
      let cell = graph.getCell(id);
      cell.autorun = true;
      if (reset && cellStateToInteger(cell.status) > cellStateToInteger(ANALYSED)) {
        cell.status = ANALYSED;
        graph._structureChanged.add(id);
      }
      let action = this._nextActions.get(id);
      if (action) {
        delete action.suspended;
      }
    }

    _allowRunningAllCellsOfDocument (docId) {
      const graph = this._graph;
      let doc = this._docs[docId];
      let cells = doc.getCells();
      if (doc instanceof Sheet) {
        cells = flatten(cells);
      }
      let ids = new Set();
      cells.forEach(cell => {
        ids.add(cell.id);
      });
      cells.forEach(cell => {
        graph._getPredecessorSet(cell.id, ids);
      });
      ids.forEach(id => {
        this._allowRunningCell(id);
      });
    }

    _setAction (id, action) {
      let currentAction = this._currentActions.get(id);
      if (!currentAction || currentAction.type !== action.type) {
        // console.log('Scheduling action', id, action)
        this._nextActions.set(id, action);
        // supersede the current action
        this._currentActions.delete(id);
      }
    }

    _isSuperseded (id, action) {
      return (this._currentActions.get(id) !== action)
    }

    // EXPERIMENTAL: allow to set some global values
    // This is not dynamic yet, i.e. cells can not produce globals
    // It is used for registering global library functions
    _addGlobal (name, value) {
      this._graph._globals.set(name, value);
    }
  }

  function getCellValue (cell) {
    return cell ? cell.value : undefined
  }

  function _getArrayValueForCells (cells) {
    let arr = cells.map(c => getCellValue(c));
    return coerceArray(arr)
  }

  /*
    Gathers the value for a cell range
    - `A1:A1`: value
    - `A1:A10`: array
    - `A1:E1`: array
    - `A1:B10`: table

    TODO: we should try to avoid using specific coercion here
  */
  function _getValueForRange (sheet, startRow, startCol, endRow, endCol) {
    let matrix = sheet.getCells();
    let val;
    // range is a single cell
    // NOTE: with the current implementation of parseSymbol this should not happen
    /* istanbul ignore if */
    if (startRow === endRow && startCol === endCol) {
      val = getCellValue(matrix[startRow][startCol]);
    // range is 1D
    } else if (startRow === endRow) {
      let cells = matrix[startRow].slice(startCol, endCol + 1);
      val = _getArrayValueForCells(cells);
    } else if (startCol === endCol) {
      let cells = [];
      for (let i = startRow; i <= endRow; i++) {
        cells.push(matrix[i][startCol]);
      }
      val = _getArrayValueForCells(cells);
    // range is 2D (-> creating a table)
    } else {
      let data = {};
      for (let j = startCol; j <= endCol; j++) {
        let name = sheet.getColumnName(j) || tableHelpers.getColumnLabel(j);
        let cells = [];
        for (let i = startRow; i <= endRow; i++) {
          cells.push(matrix[i][j]);
        }
        // TODO: why is it necessary to extract the primitive value here, instead of just using getCellValue()?
        data[name] = cells.map(c => {
          let val = getCellValue(c);
          if (val) {
            return val.data
          } else {
            return undefined
          }
        });
      }
      val = {
        // Note: first 'type' is for packing
        // and second type for diambiguation against other complex types
        type: 'table',
        data: {
          type: 'table',
          data,
          columns: endCol - startCol + 1,
          rows: endRow - startRow + 1
        }
      };
    }
    return val
  }

  function forEach$2 (iteratee, func) {
    if (!iteratee) return
    if (iteratee.constructor.prototype.forEach) {
      iteratee.forEach(func);
    } else {
      Object.keys(iteratee).forEach(function (key) {
        return func(iteratee[key], key)
      });
    }
  }

  function isObject$3 (val) {
    const type = typeof val;
    return Boolean(val) && (type === 'object' || type === 'function')
  }

  const DEBUG$3 = false;
  let count$2 = 0;
  const COUNT_MSG$2 = '%s listeners registered in the whole system.';


  class EventEmitter$2 {
    
    emit (event) {
      if (event in this.__events__) {
        
        
        var bindings = this.__events__[event].slice();
        var args = Array.prototype.slice.call(arguments, 1);
        for (var i = 0, len = bindings.length; i < len; i++) {
          var binding = bindings[i];
          
          binding.method.apply(binding.context, args);
        }
        return true
      }
      return false
    }

    
    on (event, method, context) {
      
      _on$2.call(this, event, method, context);
    }

    
    off (event, method, context) { 
      if (arguments.length === 1 && isObject$3(arguments[0])) {
        _disconnect$2.call(this, arguments[0]);
      } else {
        _off$2.apply(this, arguments);
      }
    }

    _debugEvents () {
      
      console.log('### EventEmitter: ', this);
      forEach$2(this.__events__, (handlers, name) => {
        console.log('- %s listeners for %s: ', handlers.length, name, handlers);
      });
      
    }

    get __events__ () {
      if (!this.___events___) {
        this.___events___ = {};
      }
      return this.___events___
    }
  }


  function _on$2 (event, method, context) {
    
    var bindings;
    validateMethod$2(method, context);
    if (this.__events__.hasOwnProperty(event)) {
      bindings = this.__events__[event];
    } else {
      
      bindings = this.__events__[event] = [];
    }
    
    bindings.push({
      method: method,
      context: context || null
    });
    if (DEBUG$3) {
      count$2++;
      console.info('_on()', event, method.name, context, this);
      console.info(COUNT_MSG$2, count$2);
    }
    return this
    
  }


  function _off$2 (event, method, context) {
    
    if (arguments.length === 0) {
      if (DEBUG$3) {
        forEach$2(this.__events__, (bindings) => {
          bindings.forEach((b) => {
            console.info('_off()', b.method.name, b.context, this);
          });
          count$2 -= bindings.length;
        });
        console.info(COUNT_MSG$2, count$2);
      }
      this.___events___ = {};
      return this
    }
    if (arguments.length === 1) {
      
      if (DEBUG$3) {
        count$2 -= (this.__events__[event] || []).length;
        console.info(COUNT_MSG$2, count$2);
      }
      delete this.__events__[event];
      return this
    }
    validateMethod$2(method, context);
    if (!(event in this.__events__) || !this.__events__[event].length) {
      if (DEBUG$3) console.info('NO MATCHING BINDINGS');
      
      return this
    }
    
    if (arguments.length < 3) {
      context = null;
    }
    
    let bindings = this.__events__[event];
    for (let i = bindings.length - 1; i >= 0; i--) {
      const b = bindings[i];
      if (b.method === method && b.context === context) {
        bindings.splice(i, 1);
        if (DEBUG$3) count$2--;
      }
    }
    
    if (bindings.length === 0) {
      delete this.__events__[event];
    }
    if (DEBUG$3) console.info(COUNT_MSG$2, count$2);
    return this
    
  }


  function _disconnect$2 (context) {
    
    
    forEach$2(this.__events__, (bindings, event) => {
      for (let i = bindings.length - 1; i >= 0; i--) {
        
        
        if (bindings[i] && bindings[i].context === context) {
          _off$2.call(this, event, bindings[i].method, context);
        }
      }
    });
    return this
    
  }

  function validateMethod$2 (method, context) {
    
    if (typeof method === 'string') {
      
      if (context === undefined || context === null) {
        throw new Error('Method name "' + method + '" has no context.')
      }
      if (!(method in context)) {
        
        
        throw new Error('Method not found: "' + method + '"')
      }
      if (typeof context[method] !== 'function') {
        
        
        throw new Error('Property "' + method + '" is not a function')
      }
    } else if (typeof method !== 'function') {
      throw new Error('Invalid callback. Function or method name expected.')
    }
  }

  function uuid$2 (prefix, len) {
    if (prefix && prefix[prefix.length - 1] !== '-') {
      prefix = prefix.concat('-');
    }
    var chars = '0123456789abcdefghijklmnopqrstuvwxyz'.split('');
    var uuid = [];
    var radix = 16;
    var idx;
    len = len || 32;
    if (len) {
      
      for (idx = 0; idx < len; idx++) uuid[idx] = chars[0 | Math.random() * radix];
    } else {
      
      var r;
      
      uuid[8] = uuid[13] = uuid[18] = uuid[23] = '-';
      uuid[14] = '4';
      
      
      for (idx = 0; idx < 36; idx++) {
        if (!uuid[idx]) {
          r = 0 | Math.random() * 16;
          uuid[idx] = chars[(idx === 19) ? (r & 0x3) | 0x8 : r];
        }
      }
    }
    return (prefix || '') + uuid.join('')
  }

  function map$3 (iteratee, func) {
    if (!iteratee) return []
    if (!func) func = function (item) { return item };
    if (Array.isArray(iteratee)) {
      return iteratee.map(func)
    }
    if (iteratee instanceof Map) {
      let result = [];
      for (let [name, val] of iteratee) {
        result.push(func(val, name));
      }
      return result
    }
    if (iteratee instanceof Set) {
      let result = [];
      let idx = 0;
      iteratee.forEach(item => {
        result.push(func(item, idx++));
      });
      return result
    }
    return Object.keys(iteratee).map(function (key) {
      return func(iteratee[key], key)
    })
  }

  function isPlainObject$3 (o) {
    return Boolean(o) && o.constructor === {}.constructor
  }

  function isArray$3 (a) {
    return Array.isArray(a)
  }

  function clone$2 (val) {
    if (isArray$3(val)) {
      return val.slice(0)
    }
    if (isObject$3(val)) {
      return Object.assign({}, val)
    }
    
    
    return val
  }

  const platform$2 = {

    inBrowser: false,

    inNodeJS: false,

    inElectron: false,

    
    isIE: false,
    

    isFF: false,

    isWebkit: false,

    
    version: -1,

    

    isWindows: false,

    isMac: false,

    devtools: false,

    
    
    _reset: detect$2
  };

  function detect$2 () {
    if (typeof window !== 'undefined') {
      platform$2.inBrowser = true;

      
      const ua = window.navigator.userAgent;
      const msie = ua.indexOf('MSIE ');
      const trident = ua.indexOf('Trident/');
      const edge = ua.indexOf('Edge/');

      if (msie > 0) {
        
        platform$2.isIE = true;
        platform$2.version = 10;
        
        
      } else if (trident > 0) {
        
        platform$2.isIE = true;
        platform$2.version = 11;
        platform$2.isTrident = true;
        
        
        
      } else if (edge > 0) {
        
        platform$2.isIE = true;
        platform$2.isEdge = true;
        platform$2.version = 12;
        
        parseInt(ua.substring(edge + 5, ua.indexOf('.', edge)), 10);
      }

      
      platform$2.isFF = window.navigator.userAgent.toLowerCase().indexOf('firefox') > -1;

      
      platform$2.isWebkit = !platform$2.isFF && !platform$2.isIE;
    } else {
      platform$2.inBrowser = false;
    }

    if (platform$2.inBrowser) {
      platform$2.isWindows = (window.navigator !== undefined && window.navigator.appVersion && window.navigator.appVersion.indexOf('Win') !== -1);
      platform$2.isMac = (window.navigator !== undefined && window.navigator.platform.indexOf('Mac') >= 0);
    }

    
    if (typeof process !== 'undefined') {
      if (platform$2.inBrowser) {
        platform$2.inElectron = true;
      } else {
        platform$2.inNodeJS = true;
      }
    }

    if (platform$2.inBrowser) {
      let widthThreshold = window.outerWidth - window.innerWidth > 160;
      let heightThreshold = window.outerHeight - window.innerHeight > 160;
      let orientation = widthThreshold ? 'vertical' : 'horizontal';
      if (!(heightThreshold && widthThreshold) &&
        ((window.Firebug && window.Firebug.chrome && window.Firebug.chrome.isInitialized) || widthThreshold || heightThreshold)) {
        platform$2.devtools = true;
        platform$2.devtoolsOrientation = orientation;
      }
    }
  }

  detect$2();

  function cloneDeep$2 (val) {
    if (isArray$3(val)) {
      return _cloneArrayDeep$2(val)
    }
    
    
    
    if (platform$2.inBrowser && val instanceof window.File) {
      return val
    }
    if (isObject$3(val)) {
      return _cloneObjectDeep$2(val)
    }
    
    
    return val
  }

  function _cloneObjectDeep$2 (obj) {
    let res = {};
    forEach$2(obj, (val, key) => {
      res[key] = cloneDeep$2(val);
    });
    return res
  }

  function _cloneArrayDeep$2 (arr) {
    return arr.map(cloneDeep$2)
  }

  function isNumber$2 (n) {
    return typeof n === 'number'
  }

  function isEqual$2 (a, b) {
    if (a === b) return true
    if (isArray$3(a) && isArray$3(b)) {
      if (a.length !== b.length) return false
      for (let i = 0; i < a.length; i++) {
        if (!isEqual$2(a[i], b[i])) return false
      }
      return true
    }
    if (isPlainObject$3(a) && isPlainObject$3(b)) {
      let akeys = Object.keys(a).sort();
      let bkeys = Object.keys(b).sort();
      if (!isEqual$2(akeys, bkeys)) return false
      for (let i = 0; i < akeys.length; i++) {
        let key = akeys[i];
        if (!isEqual$2(a[key], b[key])) return false
      }
      return true
    }
    return false
  }

  function isNil$2 (o) {
    return o === null || o === undefined
  }

  function isString$2 (s) {
    return typeof s === 'string'
  }

  var freeGlobal$2 = typeof global == 'object' && global && global.Object === Object && global;

  var freeSelf$2 = typeof self == 'object' && self && self.Object === Object && self;


  var root$2 = freeGlobal$2 || freeSelf$2 || Function('return this')();

  var Symbol$1$2 = root$2.Symbol;

  function arrayMap$2(array, iteratee) {
    var index = -1,
        length = array == null ? 0 : array.length,
        result = Array(length);

    while (++index < length) {
      result[index] = iteratee(array[index], index, array);
    }
    return result;
  }

  var isArray$1$2 = Array.isArray;

  var objectProto$16 = Object.prototype;


  var hasOwnProperty$13 = objectProto$16.hasOwnProperty;


  var nativeObjectToString$3 = objectProto$16.toString;


  var symToStringTag$3 = Symbol$1$2 ? Symbol$1$2.toStringTag : undefined;


  function getRawTag$2(value) {
    var isOwn = hasOwnProperty$13.call(value, symToStringTag$3),
        tag = value[symToStringTag$3];

    try {
      value[symToStringTag$3] = undefined;
      var unmasked = true;
    } catch (e) {}

    var result = nativeObjectToString$3.call(value);
    if (unmasked) {
      if (isOwn) {
        value[symToStringTag$3] = tag;
      } else {
        delete value[symToStringTag$3];
      }
    }
    return result;
  }

  var objectProto$1$2 = Object.prototype;


  var nativeObjectToString$1$2 = objectProto$1$2.toString;


  function objectToString$2(value) {
    return nativeObjectToString$1$2.call(value);
  }

  var nullTag$2 = '[object Null]',
      undefinedTag$2 = '[object Undefined]';


  var symToStringTag$1$2 = Symbol$1$2 ? Symbol$1$2.toStringTag : undefined;


  function baseGetTag$2(value) {
    if (value == null) {
      return value === undefined ? undefinedTag$2 : nullTag$2;
    }
    return (symToStringTag$1$2 && symToStringTag$1$2 in Object(value))
      ? getRawTag$2(value)
      : objectToString$2(value);
  }

  function isObjectLike$2(value) {
    return value != null && typeof value == 'object';
  }

  var symbolTag$3 = '[object Symbol]';


  function isSymbol$2(value) {
    return typeof value == 'symbol' ||
      (isObjectLike$2(value) && baseGetTag$2(value) == symbolTag$3);
  }

  var INFINITY$5 = 1 / 0;


  var symbolProto$3 = Symbol$1$2 ? Symbol$1$2.prototype : undefined,
      symbolToString$2 = symbolProto$3 ? symbolProto$3.toString : undefined;


  function baseToString$2(value) {
    
    if (typeof value == 'string') {
      return value;
    }
    if (isArray$1$2(value)) {
      
      return arrayMap$2(value, baseToString$2) + '';
    }
    if (isSymbol$2(value)) {
      return symbolToString$2 ? symbolToString$2.call(value) : '';
    }
    var result = (value + '');
    return (result == '0' && (1 / value) == -INFINITY$5) ? '-0' : result;
  }

  function toString$2(value) {
    return value == null ? '' : baseToString$2(value);
  }

  function baseSlice$2(array, start, end) {
    var index = -1,
        length = array.length;

    if (start < 0) {
      start = -start > length ? 0 : (length + start);
    }
    end = end > length ? length : end;
    if (end < 0) {
      end += length;
    }
    length = start > end ? 0 : ((end - start) >>> 0);
    start >>>= 0;

    var result = Array(length);
    while (++index < length) {
      result[index] = array[index + start];
    }
    return result;
  }

  function isObject$1$2(value) {
    var type = typeof value;
    return value != null && (type == 'object' || type == 'function');
  }

  var reIsDeepProp$2 = /\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/,
      reIsPlainProp$2 = /^\w*$/;


  function isKey$2(value, object) {
    if (isArray$1$2(value)) {
      return false;
    }
    var type = typeof value;
    if (type == 'number' || type == 'symbol' || type == 'boolean' ||
        value == null || isSymbol$2(value)) {
      return true;
    }
    return reIsPlainProp$2.test(value) || !reIsDeepProp$2.test(value) ||
      (object != null && value in Object(object));
  }

  var asyncTag$2 = '[object AsyncFunction]',
      funcTag$3 = '[object Function]',
      genTag$2 = '[object GeneratorFunction]',
      proxyTag$2 = '[object Proxy]';


  function isFunction$3(value) {
    if (!isObject$1$2(value)) {
      return false;
    }
    
    
    var tag = baseGetTag$2(value);
    return tag == funcTag$3 || tag == genTag$2 || tag == asyncTag$2 || tag == proxyTag$2;
  }

  var coreJsData$2 = root$2['__core-js_shared__'];

  var maskSrcKey$2 = (function() {
    var uid = /[^.]+$/.exec(coreJsData$2 && coreJsData$2.keys && coreJsData$2.keys.IE_PROTO || '');
    return uid ? ('Symbol(src)_1.' + uid) : '';
  }());


  function isMasked$2(func) {
    return !!maskSrcKey$2 && (maskSrcKey$2 in func);
  }

  var funcProto$4 = Function.prototype;


  var funcToString$4 = funcProto$4.toString;


  function toSource$2(func) {
    if (func != null) {
      try {
        return funcToString$4.call(func);
      } catch (e) {}
      try {
        return (func + '');
      } catch (e) {}
    }
    return '';
  }

  var reRegExpChar$2 = /[\\^$.*+?()[\]{}|]/g;


  var reIsHostCtor$2 = /^\[object .+?Constructor\]$/;


  var funcProto$1$2 = Function.prototype,
      objectProto$2$2 = Object.prototype;


  var funcToString$1$2 = funcProto$1$2.toString;


  var hasOwnProperty$1$2 = objectProto$2$2.hasOwnProperty;


  var reIsNative$2 = RegExp('^' +
    funcToString$1$2.call(hasOwnProperty$1$2).replace(reRegExpChar$2, '\\$&')
    .replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g, '$1.*?') + '$'
  );


  function baseIsNative$2(value) {
    if (!isObject$1$2(value) || isMasked$2(value)) {
      return false;
    }
    var pattern = isFunction$3(value) ? reIsNative$2 : reIsHostCtor$2;
    return pattern.test(toSource$2(value));
  }

  function getValue$2(object, key) {
    return object == null ? undefined : object[key];
  }

  function getNative$2(object, key) {
    var value = getValue$2(object, key);
    return baseIsNative$2(value) ? value : undefined;
  }

  var nativeCreate$2 = getNative$2(Object, 'create');

  function hashClear$2() {
    this.__data__ = nativeCreate$2 ? nativeCreate$2(null) : {};
    this.size = 0;
  }

  function hashDelete$2(key) {
    var result = this.has(key) && delete this.__data__[key];
    this.size -= result ? 1 : 0;
    return result;
  }

  var HASH_UNDEFINED$4 = '__lodash_hash_undefined__';


  var objectProto$3$2 = Object.prototype;


  var hasOwnProperty$2$2 = objectProto$3$2.hasOwnProperty;


  function hashGet$2(key) {
    var data = this.__data__;
    if (nativeCreate$2) {
      var result = data[key];
      return result === HASH_UNDEFINED$4 ? undefined : result;
    }
    return hasOwnProperty$2$2.call(data, key) ? data[key] : undefined;
  }

  var objectProto$4$2 = Object.prototype;


  var hasOwnProperty$3$2 = objectProto$4$2.hasOwnProperty;


  function hashHas$2(key) {
    var data = this.__data__;
    return nativeCreate$2 ? (data[key] !== undefined) : hasOwnProperty$3$2.call(data, key);
  }

  var HASH_UNDEFINED$1$2 = '__lodash_hash_undefined__';


  function hashSet$2(key, value) {
    var data = this.__data__;
    this.size += this.has(key) ? 0 : 1;
    data[key] = (nativeCreate$2 && value === undefined) ? HASH_UNDEFINED$1$2 : value;
    return this;
  }

  function Hash$3(entries) {
    var index = -1,
        length = entries == null ? 0 : entries.length;

    this.clear();
    while (++index < length) {
      var entry = entries[index];
      this.set(entry[0], entry[1]);
    }
  }


  Hash$3.prototype.clear = hashClear$2;
  Hash$3.prototype['delete'] = hashDelete$2;
  Hash$3.prototype.get = hashGet$2;
  Hash$3.prototype.has = hashHas$2;
  Hash$3.prototype.set = hashSet$2;

  function listCacheClear$2() {
    this.__data__ = [];
    this.size = 0;
  }

  function eq$2(value, other) {
    return value === other || (value !== value && other !== other);
  }

  function assocIndexOf$2(array, key) {
    var length = array.length;
    while (length--) {
      if (eq$2(array[length][0], key)) {
        return length;
      }
    }
    return -1;
  }

  var arrayProto$2 = Array.prototype;


  var splice$2 = arrayProto$2.splice;


  function listCacheDelete$2(key) {
    var data = this.__data__,
        index = assocIndexOf$2(data, key);

    if (index < 0) {
      return false;
    }
    var lastIndex = data.length - 1;
    if (index == lastIndex) {
      data.pop();
    } else {
      splice$2.call(data, index, 1);
    }
    --this.size;
    return true;
  }

  function listCacheGet$2(key) {
    var data = this.__data__,
        index = assocIndexOf$2(data, key);

    return index < 0 ? undefined : data[index][1];
  }

  function listCacheHas$2(key) {
    return assocIndexOf$2(this.__data__, key) > -1;
  }

  function listCacheSet$2(key, value) {
    var data = this.__data__,
        index = assocIndexOf$2(data, key);

    if (index < 0) {
      ++this.size;
      data.push([key, value]);
    } else {
      data[index][1] = value;
    }
    return this;
  }

  function ListCache$2(entries) {
    var index = -1,
        length = entries == null ? 0 : entries.length;

    this.clear();
    while (++index < length) {
      var entry = entries[index];
      this.set(entry[0], entry[1]);
    }
  }


  ListCache$2.prototype.clear = listCacheClear$2;
  ListCache$2.prototype['delete'] = listCacheDelete$2;
  ListCache$2.prototype.get = listCacheGet$2;
  ListCache$2.prototype.has = listCacheHas$2;
  ListCache$2.prototype.set = listCacheSet$2;

  var Map$1$2 = getNative$2(root$2, 'Map');

  function mapCacheClear$2() {
    this.size = 0;
    this.__data__ = {
      'hash': new Hash$3,
      'map': new (Map$1$2 || ListCache$2),
      'string': new Hash$3
    };
  }

  function isKeyable$2(value) {
    var type = typeof value;
    return (type == 'string' || type == 'number' || type == 'symbol' || type == 'boolean')
      ? (value !== '__proto__')
      : (value === null);
  }

  function getMapData$2(map, key) {
    var data = map.__data__;
    return isKeyable$2(key)
      ? data[typeof key == 'string' ? 'string' : 'hash']
      : data.map;
  }

  function mapCacheDelete$2(key) {
    var result = getMapData$2(this, key)['delete'](key);
    this.size -= result ? 1 : 0;
    return result;
  }

  function mapCacheGet$2(key) {
    return getMapData$2(this, key).get(key);
  }

  function mapCacheHas$2(key) {
    return getMapData$2(this, key).has(key);
  }

  function mapCacheSet$2(key, value) {
    var data = getMapData$2(this, key),
        size = data.size;

    data.set(key, value);
    this.size += data.size == size ? 0 : 1;
    return this;
  }

  function MapCache$2(entries) {
    var index = -1,
        length = entries == null ? 0 : entries.length;

    this.clear();
    while (++index < length) {
      var entry = entries[index];
      this.set(entry[0], entry[1]);
    }
  }


  MapCache$2.prototype.clear = mapCacheClear$2;
  MapCache$2.prototype['delete'] = mapCacheDelete$2;
  MapCache$2.prototype.get = mapCacheGet$2;
  MapCache$2.prototype.has = mapCacheHas$2;
  MapCache$2.prototype.set = mapCacheSet$2;

  var FUNC_ERROR_TEXT$1$2 = 'Expected a function';


  function memoize$2(func, resolver) {
    if (typeof func != 'function' || (resolver != null && typeof resolver != 'function')) {
      throw new TypeError(FUNC_ERROR_TEXT$1$2);
    }
    var memoized = function() {
      var args = arguments,
          key = resolver ? resolver.apply(this, args) : args[0],
          cache = memoized.cache;

      if (cache.has(key)) {
        return cache.get(key);
      }
      var result = func.apply(this, args);
      memoized.cache = cache.set(key, result) || cache;
      return result;
    };
    memoized.cache = new (memoize$2.Cache || MapCache$2);
    return memoized;
  }


  memoize$2.Cache = MapCache$2;

  var MAX_MEMOIZE_SIZE$2 = 500;


  function memoizeCapped$2(func) {
    var result = memoize$2(func, function(key) {
      if (cache.size === MAX_MEMOIZE_SIZE$2) {
        cache.clear();
      }
      return key;
    });

    var cache = result.cache;
    return result;
  }

  var rePropName$2 = /[^.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(?:\.|\[\])(?:\.|\[\]|$))/g;


  var reEscapeChar$2 = /\\(\\)?/g;


  var stringToPath$2 = memoizeCapped$2(function(string) {
    var result = [];
    if (string.charCodeAt(0) === 46 ) {
      result.push('');
    }
    string.replace(rePropName$2, function(match, number, quote, subString) {
      result.push(quote ? subString.replace(reEscapeChar$2, '$1') : (number || match));
    });
    return result;
  });

  function castPath$2(value, object) {
    if (isArray$1$2(value)) {
      return value;
    }
    return isKey$2(value, object) ? [value] : stringToPath$2(toString$2(value));
  }

  var INFINITY$1$2 = 1 / 0;


  function toKey$2(value) {
    if (typeof value == 'string' || isSymbol$2(value)) {
      return value;
    }
    var result = (value + '');
    return (result == '0' && (1 / value) == -INFINITY$1$2) ? '-0' : result;
  }

  function baseGet$2(object, path) {
    path = castPath$2(path, object);

    var index = 0,
        length = path.length;

    while (object != null && index < length) {
      object = object[toKey$2(path[index++])];
    }
    return (index && index == length) ? object : undefined;
  }

  function get$2(object, path, defaultValue) {
    var result = object == null ? undefined : baseGet$2(object, path);
    return result === undefined ? defaultValue : result;
  }

  function stackClear$2() {
    this.__data__ = new ListCache$2;
    this.size = 0;
  }

  function stackDelete$2(key) {
    var data = this.__data__,
        result = data['delete'](key);

    this.size = data.size;
    return result;
  }

  function stackGet$2(key) {
    return this.__data__.get(key);
  }

  function stackHas$2(key) {
    return this.__data__.has(key);
  }

  var LARGE_ARRAY_SIZE$4 = 200;


  function stackSet$2(key, value) {
    var data = this.__data__;
    if (data instanceof ListCache$2) {
      var pairs = data.__data__;
      if (!Map$1$2 || (pairs.length < LARGE_ARRAY_SIZE$4 - 1)) {
        pairs.push([key, value]);
        this.size = ++data.size;
        return this;
      }
      data = this.__data__ = new MapCache$2(pairs);
    }
    data.set(key, value);
    this.size = data.size;
    return this;
  }

  function Stack$2(entries) {
    var data = this.__data__ = new ListCache$2(entries);
    this.size = data.size;
  }


  Stack$2.prototype.clear = stackClear$2;
  Stack$2.prototype['delete'] = stackDelete$2;
  Stack$2.prototype.get = stackGet$2;
  Stack$2.prototype.has = stackHas$2;
  Stack$2.prototype.set = stackSet$2;

  var HASH_UNDEFINED$2$2 = '__lodash_hash_undefined__';


  function setCacheAdd$2(value) {
    this.__data__.set(value, HASH_UNDEFINED$2$2);
    return this;
  }

  function setCacheHas$2(value) {
    return this.__data__.has(value);
  }

  function SetCache$2(values) {
    var index = -1,
        length = values == null ? 0 : values.length;

    this.__data__ = new MapCache$2;
    while (++index < length) {
      this.add(values[index]);
    }
  }


  SetCache$2.prototype.add = SetCache$2.prototype.push = setCacheAdd$2;
  SetCache$2.prototype.has = setCacheHas$2;

  function cacheHas$2(cache, key) {
    return cache.has(key);
  }

  var Uint8Array$2 = root$2.Uint8Array;

  function setToArray$2(set) {
    var index = -1,
        result = Array(set.size);

    set.forEach(function(value) {
      result[++index] = value;
    });
    return result;
  }


  var symbolProto$1$2 = Symbol$1$2 ? Symbol$1$2.prototype : undefined,
      symbolValueOf$2 = symbolProto$1$2 ? symbolProto$1$2.valueOf : undefined;

  function baseTimes$2(n, iteratee) {
    var index = -1,
        result = Array(n);

    while (++index < n) {
      result[index] = iteratee(index);
    }
    return result;
  }

  var argsTag$4 = '[object Arguments]';


  function baseIsArguments$2(value) {
    return isObjectLike$2(value) && baseGetTag$2(value) == argsTag$4;
  }

  var objectProto$6$2 = Object.prototype;


  var hasOwnProperty$4$2 = objectProto$6$2.hasOwnProperty;


  var propertyIsEnumerable$1$2 = objectProto$6$2.propertyIsEnumerable;


  var isArguments$2 = baseIsArguments$2(function() { return arguments; }()) ? baseIsArguments$2 : function(value) {
    return isObjectLike$2(value) && hasOwnProperty$4$2.call(value, 'callee') &&
      !propertyIsEnumerable$1$2.call(value, 'callee');
  };

  function stubFalse$2() {
    return false;
  }

  var freeExports$4 = typeof exports == 'object' && exports && !exports.nodeType && exports;


  var freeModule$4 = freeExports$4 && typeof module == 'object' && module && !module.nodeType && module;


  var moduleExports$4 = freeModule$4 && freeModule$4.exports === freeExports$4;


  var Buffer$3 = moduleExports$4 ? root$2.Buffer : undefined;


  var nativeIsBuffer$2 = Buffer$3 ? Buffer$3.isBuffer : undefined;


  var isBuffer$2 = nativeIsBuffer$2 || stubFalse$2;

  var MAX_SAFE_INTEGER$3 = 9007199254740991;


  var reIsUint$2 = /^(?:0|[1-9]\d*)$/;


  function isIndex$2(value, length) {
    var type = typeof value;
    length = length == null ? MAX_SAFE_INTEGER$3 : length;

    return !!length &&
      (type == 'number' ||
        (type != 'symbol' && reIsUint$2.test(value))) &&
          (value > -1 && value % 1 == 0 && value < length);
  }

  var MAX_SAFE_INTEGER$1$2 = 9007199254740991;


  function isLength$2(value) {
    return typeof value == 'number' &&
      value > -1 && value % 1 == 0 && value <= MAX_SAFE_INTEGER$1$2;
  }

  var argsTag$1$2 = '[object Arguments]',
      arrayTag$3 = '[object Array]',
      boolTag$1$2 = '[object Boolean]',
      dateTag$1$2 = '[object Date]',
      errorTag$1$2 = '[object Error]',
      funcTag$1$2 = '[object Function]',
      mapTag$1$2 = '[object Map]',
      numberTag$1$2 = '[object Number]',
      objectTag$5 = '[object Object]',
      regexpTag$1$2 = '[object RegExp]',
      setTag$1$2 = '[object Set]',
      stringTag$1$2 = '[object String]',
      weakMapTag$3 = '[object WeakMap]';

  var arrayBufferTag$1$2 = '[object ArrayBuffer]',
      dataViewTag$1$2 = '[object DataView]',
      float32Tag$2 = '[object Float32Array]',
      float64Tag$2 = '[object Float64Array]',
      int8Tag$2 = '[object Int8Array]',
      int16Tag$2 = '[object Int16Array]',
      int32Tag$2 = '[object Int32Array]',
      uint8Tag$2 = '[object Uint8Array]',
      uint8ClampedTag$2 = '[object Uint8ClampedArray]',
      uint16Tag$2 = '[object Uint16Array]',
      uint32Tag$2 = '[object Uint32Array]';


  var typedArrayTags$2 = {};
  typedArrayTags$2[float32Tag$2] = typedArrayTags$2[float64Tag$2] =
  typedArrayTags$2[int8Tag$2] = typedArrayTags$2[int16Tag$2] =
  typedArrayTags$2[int32Tag$2] = typedArrayTags$2[uint8Tag$2] =
  typedArrayTags$2[uint8ClampedTag$2] = typedArrayTags$2[uint16Tag$2] =
  typedArrayTags$2[uint32Tag$2] = true;
  typedArrayTags$2[argsTag$1$2] = typedArrayTags$2[arrayTag$3] =
  typedArrayTags$2[arrayBufferTag$1$2] = typedArrayTags$2[boolTag$1$2] =
  typedArrayTags$2[dataViewTag$1$2] = typedArrayTags$2[dateTag$1$2] =
  typedArrayTags$2[errorTag$1$2] = typedArrayTags$2[funcTag$1$2] =
  typedArrayTags$2[mapTag$1$2] = typedArrayTags$2[numberTag$1$2] =
  typedArrayTags$2[objectTag$5] = typedArrayTags$2[regexpTag$1$2] =
  typedArrayTags$2[setTag$1$2] = typedArrayTags$2[stringTag$1$2] =
  typedArrayTags$2[weakMapTag$3] = false;


  function baseIsTypedArray$2(value) {
    return isObjectLike$2(value) &&
      isLength$2(value.length) && !!typedArrayTags$2[baseGetTag$2(value)];
  }

  function baseUnary$2(func) {
    return function(value) {
      return func(value);
    };
  }

  var freeExports$1$2 = typeof exports == 'object' && exports && !exports.nodeType && exports;


  var freeModule$1$2 = freeExports$1$2 && typeof module == 'object' && module && !module.nodeType && module;


  var moduleExports$1$2 = freeModule$1$2 && freeModule$1$2.exports === freeExports$1$2;


  var freeProcess$2 = moduleExports$1$2 && freeGlobal$2.process;


  var nodeUtil$2 = (function() {
    try {
      
      var types = freeModule$1$2 && freeModule$1$2.require && freeModule$1$2.require('util').types;

      if (types) {
        return types;
      }

      
      return freeProcess$2 && freeProcess$2.binding && freeProcess$2.binding('util');
    } catch (e) {}
  }());

  var nodeIsTypedArray$2 = nodeUtil$2 && nodeUtil$2.isTypedArray;


  var isTypedArray$2 = nodeIsTypedArray$2 ? baseUnary$2(nodeIsTypedArray$2) : baseIsTypedArray$2;

  var objectProto$7$2 = Object.prototype;


  var hasOwnProperty$5$2 = objectProto$7$2.hasOwnProperty;


  function arrayLikeKeys$2(value, inherited) {
    var isArr = isArray$1$2(value),
        isArg = !isArr && isArguments$2(value),
        isBuff = !isArr && !isArg && isBuffer$2(value),
        isType = !isArr && !isArg && !isBuff && isTypedArray$2(value),
        skipIndexes = isArr || isArg || isBuff || isType,
        result = skipIndexes ? baseTimes$2(value.length, String) : [],
        length = result.length;

    for (var key in value) {
      if ((inherited || hasOwnProperty$5$2.call(value, key)) &&
          !(skipIndexes && (
             
             key == 'length' ||
             
             (isBuff && (key == 'offset' || key == 'parent')) ||
             
             (isType && (key == 'buffer' || key == 'byteLength' || key == 'byteOffset')) ||
             
             isIndex$2(key, length)
          ))) {
        result.push(key);
      }
    }
    return result;
  }

  var objectProto$8$2 = Object.prototype;


  function isPrototype$2(value) {
    var Ctor = value && value.constructor,
        proto = (typeof Ctor == 'function' && Ctor.prototype) || objectProto$8$2;

    return value === proto;
  }

  function overArg$2(func, transform) {
    return function(arg) {
      return func(transform(arg));
    };
  }

  function isArrayLike$2(value) {
    return value != null && isLength$2(value.length) && !isFunction$3(value);
  }

  var DataView$2 = getNative$2(root$2, 'DataView');

  var Promise$1$2 = getNative$2(root$2, 'Promise');

  var Set$1$2 = getNative$2(root$2, 'Set');

  var WeakMap$1$2 = getNative$2(root$2, 'WeakMap');

  var mapTag$2$2 = '[object Map]',
      objectTag$1$2 = '[object Object]',
      promiseTag$2 = '[object Promise]',
      setTag$2$2 = '[object Set]',
      weakMapTag$1$2 = '[object WeakMap]';

  var dataViewTag$2$2 = '[object DataView]';


  var dataViewCtorString$2 = toSource$2(DataView$2),
      mapCtorString$2 = toSource$2(Map$1$2),
      promiseCtorString$2 = toSource$2(Promise$1$2),
      setCtorString$2 = toSource$2(Set$1$2),
      weakMapCtorString$2 = toSource$2(WeakMap$1$2);


  var getTag$3 = baseGetTag$2;


  if ((DataView$2 && getTag$3(new DataView$2(new ArrayBuffer(1))) != dataViewTag$2$2) ||
      (Map$1$2 && getTag$3(new Map$1$2) != mapTag$2$2) ||
      (Promise$1$2 && getTag$3(Promise$1$2.resolve()) != promiseTag$2) ||
      (Set$1$2 && getTag$3(new Set$1$2) != setTag$2$2) ||
      (WeakMap$1$2 && getTag$3(new WeakMap$1$2) != weakMapTag$1$2)) {
    getTag$3 = function(value) {
      var result = baseGetTag$2(value),
          Ctor = result == objectTag$1$2 ? value.constructor : undefined,
          ctorString = Ctor ? toSource$2(Ctor) : '';

      if (ctorString) {
        switch (ctorString) {
          case dataViewCtorString$2: return dataViewTag$2$2;
          case mapCtorString$2: return mapTag$2$2;
          case promiseCtorString$2: return promiseTag$2;
          case setCtorString$2: return setTag$2$2;
          case weakMapCtorString$2: return weakMapTag$1$2;
        }
      }
      return result;
    };
  }

  function identity$2(value) {
    return value;
  }

  function baseFindIndex$2(array, predicate, fromIndex, fromRight) {
    var length = array.length,
        index = fromIndex + (fromRight ? 1 : -1);

    while ((fromRight ? index-- : ++index < length)) {
      if (predicate(array[index], index, array)) {
        return index;
      }
    }
    return -1;
  }

  var defineProperty$2 = (function() {
    try {
      var func = getNative$2(Object, 'defineProperty');
      func({}, '', {});
      return func;
    } catch (e) {}
  }());

  function baseAssignValue$2(object, key, value) {
    if (key == '__proto__' && defineProperty$2) {
      defineProperty$2(object, key, {
        'configurable': true,
        'enumerable': true,
        'value': value,
        'writable': true
      });
    } else {
      object[key] = value;
    }
  }

  function assignMergeValue$2(object, key, value) {
    if ((value !== undefined && !eq$2(object[key], value)) ||
        (value === undefined && !(key in object))) {
      baseAssignValue$2(object, key, value);
    }
  }

  function createBaseFor$2(fromRight) {
    return function(object, iteratee, keysFunc) {
      var index = -1,
          iterable = Object(object),
          props = keysFunc(object),
          length = props.length;

      while (length--) {
        var key = props[fromRight ? length : ++index];
        if (iteratee(iterable[key], key, iterable) === false) {
          break;
        }
      }
      return object;
    };
  }

  var baseFor$2 = createBaseFor$2();

  var freeExports$2$2 = typeof exports == 'object' && exports && !exports.nodeType && exports;


  var freeModule$2$2 = freeExports$2$2 && typeof module == 'object' && module && !module.nodeType && module;


  var moduleExports$2$2 = freeModule$2$2 && freeModule$2$2.exports === freeExports$2$2;


  var Buffer$1$2 = moduleExports$2$2 ? root$2.Buffer : undefined,
      allocUnsafe$2 = Buffer$1$2 ? Buffer$1$2.allocUnsafe : undefined;


  function cloneBuffer$2(buffer, isDeep) {
    if (isDeep) {
      return buffer.slice();
    }
    var length = buffer.length,
        result = allocUnsafe$2 ? allocUnsafe$2(length) : new buffer.constructor(length);

    buffer.copy(result);
    return result;
  }

  function cloneArrayBuffer$2(arrayBuffer) {
    var result = new arrayBuffer.constructor(arrayBuffer.byteLength);
    new Uint8Array$2(result).set(new Uint8Array$2(arrayBuffer));
    return result;
  }

  function cloneTypedArray$2(typedArray, isDeep) {
    var buffer = isDeep ? cloneArrayBuffer$2(typedArray.buffer) : typedArray.buffer;
    return new typedArray.constructor(buffer, typedArray.byteOffset, typedArray.length);
  }

  function copyArray$2(source, array) {
    var index = -1,
        length = source.length;

    array || (array = Array(length));
    while (++index < length) {
      array[index] = source[index];
    }
    return array;
  }

  var objectCreate$2 = Object.create;


  var baseCreate$2 = (function() {
    function object() {}
    return function(proto) {
      if (!isObject$1$2(proto)) {
        return {};
      }
      if (objectCreate$2) {
        return objectCreate$2(proto);
      }
      object.prototype = proto;
      var result = new object;
      object.prototype = undefined;
      return result;
    };
  }());

  var getPrototype$2 = overArg$2(Object.getPrototypeOf, Object);

  function initCloneObject$2(object) {
    return (typeof object.constructor == 'function' && !isPrototype$2(object))
      ? baseCreate$2(getPrototype$2(object))
      : {};
  }

  function isArrayLikeObject$2(value) {
    return isObjectLike$2(value) && isArrayLike$2(value);
  }

  var objectTag$3$2 = '[object Object]';


  var funcProto$2$2 = Function.prototype,
      objectProto$12$2 = Object.prototype;


  var funcToString$2$2 = funcProto$2$2.toString;


  var hasOwnProperty$9$2 = objectProto$12$2.hasOwnProperty;


  var objectCtorString$2 = funcToString$2$2.call(Object);


  function isPlainObject$1$2(value) {
    if (!isObjectLike$2(value) || baseGetTag$2(value) != objectTag$3$2) {
      return false;
    }
    var proto = getPrototype$2(value);
    if (proto === null) {
      return true;
    }
    var Ctor = hasOwnProperty$9$2.call(proto, 'constructor') && proto.constructor;
    return typeof Ctor == 'function' && Ctor instanceof Ctor &&
      funcToString$2$2.call(Ctor) == objectCtorString$2;
  }

  function safeGet$2(object, key) {
    return key == '__proto__'
      ? undefined
      : object[key];
  }

  var objectProto$13$2 = Object.prototype;


  var hasOwnProperty$10$2 = objectProto$13$2.hasOwnProperty;


  function assignValue$2(object, key, value) {
    var objValue = object[key];
    if (!(hasOwnProperty$10$2.call(object, key) && eq$2(objValue, value)) ||
        (value === undefined && !(key in object))) {
      baseAssignValue$2(object, key, value);
    }
  }

  function copyObject$2(source, props, object, customizer) {
    var isNew = !object;
    object || (object = {});

    var index = -1,
        length = props.length;

    while (++index < length) {
      var key = props[index];

      var newValue = customizer
        ? customizer(object[key], source[key], key, object, source)
        : undefined;

      if (newValue === undefined) {
        newValue = source[key];
      }
      if (isNew) {
        baseAssignValue$2(object, key, newValue);
      } else {
        assignValue$2(object, key, newValue);
      }
    }
    return object;
  }

  function nativeKeysIn$2(object) {
    var result = [];
    if (object != null) {
      for (var key in Object(object)) {
        result.push(key);
      }
    }
    return result;
  }

  var objectProto$14$2 = Object.prototype;


  var hasOwnProperty$11$2 = objectProto$14$2.hasOwnProperty;


  function baseKeysIn$2(object) {
    if (!isObject$1$2(object)) {
      return nativeKeysIn$2(object);
    }
    var isProto = isPrototype$2(object),
        result = [];

    for (var key in object) {
      if (!(key == 'constructor' && (isProto || !hasOwnProperty$11$2.call(object, key)))) {
        result.push(key);
      }
    }
    return result;
  }

  function keysIn$2(object) {
    return isArrayLike$2(object) ? arrayLikeKeys$2(object, true) : baseKeysIn$2(object);
  }

  function toPlainObject$2(value) {
    return copyObject$2(value, keysIn$2(value));
  }

  function baseMergeDeep$2(object, source, key, srcIndex, mergeFunc, customizer, stack) {
    var objValue = safeGet$2(object, key),
        srcValue = safeGet$2(source, key),
        stacked = stack.get(srcValue);

    if (stacked) {
      assignMergeValue$2(object, key, stacked);
      return;
    }
    var newValue = customizer
      ? customizer(objValue, srcValue, (key + ''), object, source, stack)
      : undefined;

    var isCommon = newValue === undefined;

    if (isCommon) {
      var isArr = isArray$1$2(srcValue),
          isBuff = !isArr && isBuffer$2(srcValue),
          isTyped = !isArr && !isBuff && isTypedArray$2(srcValue);

      newValue = srcValue;
      if (isArr || isBuff || isTyped) {
        if (isArray$1$2(objValue)) {
          newValue = objValue;
        }
        else if (isArrayLikeObject$2(objValue)) {
          newValue = copyArray$2(objValue);
        }
        else if (isBuff) {
          isCommon = false;
          newValue = cloneBuffer$2(srcValue, true);
        }
        else if (isTyped) {
          isCommon = false;
          newValue = cloneTypedArray$2(srcValue, true);
        }
        else {
          newValue = [];
        }
      }
      else if (isPlainObject$1$2(srcValue) || isArguments$2(srcValue)) {
        newValue = objValue;
        if (isArguments$2(objValue)) {
          newValue = toPlainObject$2(objValue);
        }
        else if (!isObject$1$2(objValue) || (srcIndex && isFunction$3(objValue))) {
          newValue = initCloneObject$2(srcValue);
        }
      }
      else {
        isCommon = false;
      }
    }
    if (isCommon) {
      
      stack.set(srcValue, newValue);
      mergeFunc(newValue, srcValue, srcIndex, customizer, stack);
      stack['delete'](srcValue);
    }
    assignMergeValue$2(object, key, newValue);
  }

  function baseMerge$2(object, source, srcIndex, customizer, stack) {
    if (object === source) {
      return;
    }
    baseFor$2(source, function(srcValue, key) {
      if (isObject$1$2(srcValue)) {
        stack || (stack = new Stack$2);
        baseMergeDeep$2(object, source, key, srcIndex, baseMerge$2, customizer, stack);
      }
      else {
        var newValue = customizer
          ? customizer(safeGet$2(object, key), srcValue, (key + ''), object, source, stack)
          : undefined;

        if (newValue === undefined) {
          newValue = srcValue;
        }
        assignMergeValue$2(object, key, newValue);
      }
    }, keysIn$2);
  }

  function apply$2(func, thisArg, args) {
    switch (args.length) {
      case 0: return func.call(thisArg);
      case 1: return func.call(thisArg, args[0]);
      case 2: return func.call(thisArg, args[0], args[1]);
      case 3: return func.call(thisArg, args[0], args[1], args[2]);
    }
    return func.apply(thisArg, args);
  }

  var nativeMax$2$2 = Math.max;


  function overRest$2(func, start, transform) {
    start = nativeMax$2$2(start === undefined ? (func.length - 1) : start, 0);
    return function() {
      var args = arguments,
          index = -1,
          length = nativeMax$2$2(args.length - start, 0),
          array = Array(length);

      while (++index < length) {
        array[index] = args[start + index];
      }
      index = -1;
      var otherArgs = Array(start + 1);
      while (++index < start) {
        otherArgs[index] = args[index];
      }
      otherArgs[start] = transform(array);
      return apply$2(func, this, otherArgs);
    };
  }

  function constant$2(value) {
    return function() {
      return value;
    };
  }

  var baseSetToString$2 = !defineProperty$2 ? identity$2 : function(func, string) {
    return defineProperty$2(func, 'toString', {
      'configurable': true,
      'enumerable': false,
      'value': constant$2(string),
      'writable': true
    });
  };

  var HOT_COUNT$2 = 800,
      HOT_SPAN$2 = 16;


  var nativeNow$2 = Date.now;


  function shortOut$2(func) {
    var count = 0,
        lastCalled = 0;

    return function() {
      var stamp = nativeNow$2(),
          remaining = HOT_SPAN$2 - (stamp - lastCalled);

      lastCalled = stamp;
      if (remaining > 0) {
        if (++count >= HOT_COUNT$2) {
          return arguments[0];
        }
      } else {
        count = 0;
      }
      return func.apply(undefined, arguments);
    };
  }

  var setToString$2 = shortOut$2(baseSetToString$2);

  function baseRest$2(func, start) {
    return setToString$2(overRest$2(func, start, identity$2), func + '');
  }

  function isIterateeCall$2(value, index, object) {
    if (!isObject$1$2(object)) {
      return false;
    }
    var type = typeof index;
    if (type == 'number'
          ? (isArrayLike$2(object) && isIndex$2(index, object.length))
          : (type == 'string' && index in object)
        ) {
      return eq$2(object[index], value);
    }
    return false;
  }

  function createAssigner$2(assigner) {
    return baseRest$2(function(object, sources) {
      var index = -1,
          length = sources.length,
          customizer = length > 1 ? sources[length - 1] : undefined,
          guard = length > 2 ? sources[2] : undefined;

      customizer = (assigner.length > 3 && typeof customizer == 'function')
        ? (length--, customizer)
        : undefined;

      if (guard && isIterateeCall$2(sources[0], sources[1], guard)) {
        customizer = length < 3 ? undefined : customizer;
        length = 1;
      }
      object = Object(object);
      while (++index < length) {
        var source = sources[index];
        if (source) {
          assigner(object, source, index, customizer);
        }
      }
      return object;
    });
  }

  var merge$3 = createAssigner$2(function(object, source, srcIndex) {
    baseMerge$2(object, source, srcIndex);
  });

  var mergeWith$2 = createAssigner$2(function(object, source, srcIndex, customizer) {
    baseMerge$2(object, source, srcIndex, customizer);
  });

  function baseSet$2(object, path, value, customizer) {
    if (!isObject$1$2(object)) {
      return object;
    }
    path = castPath$2(path, object);

    var index = -1,
        length = path.length,
        lastIndex = length - 1,
        nested = object;

    while (nested != null && ++index < length) {
      var key = toKey$2(path[index]),
          newValue = value;

      if (index != lastIndex) {
        var objValue = nested[key];
        newValue = customizer ? customizer(objValue, key, nested) : undefined;
        if (newValue === undefined) {
          newValue = isObject$1$2(objValue)
            ? objValue
            : (isIndex$2(path[index + 1]) ? [] : {});
        }
      }
      assignValue$2(nested, key, newValue);
      nested = nested[key];
    }
    return object;
  }

  function setWith$2(object, path, value, customizer) {
    customizer = typeof customizer == 'function' ? customizer : undefined;
    return object == null ? object : baseSet$2(object, path, value, customizer);
  }

  function baseIsNaN$2(value) {
    return value !== value;
  }

  function strictIndexOf$2(array, value, fromIndex) {
    var index = fromIndex - 1,
        length = array.length;

    while (++index < length) {
      if (array[index] === value) {
        return index;
      }
    }
    return -1;
  }

  function baseIndexOf$2(array, value, fromIndex) {
    return value === value
      ? strictIndexOf$2(array, value, fromIndex)
      : baseFindIndex$2(array, baseIsNaN$2, fromIndex);
  }

  function arrayIncludes$2(array, value) {
    var length = array == null ? 0 : array.length;
    return !!length && baseIndexOf$2(array, value, 0) > -1;
  }

  function arrayIncludesWith$2(array, value, comparator) {
    var index = -1,
        length = array == null ? 0 : array.length;

    while (++index < length) {
      if (comparator(value, array[index])) {
        return true;
      }
    }
    return false;
  }

  function noop$2() {
    
  }

  var INFINITY$3$2 = 1 / 0;


  var createSet$2 = !(Set$1$2 && (1 / setToArray$2(new Set$1$2([,-0]))[1]) == INFINITY$3$2) ? noop$2 : function(values) {
    return new Set$1$2(values);
  };

  function last$3(array) {
    var length = array == null ? 0 : array.length;
    return length ? array[length - 1] : undefined;
  }

  function parent$3(object, path) {
    return path.length < 2 ? object : baseGet$2(object, baseSlice$2(path, 0, -1));
  }

  function baseUnset$2(object, path) {
    path = castPath$2(path, object);
    object = parent$3(object, path);
    return object == null || delete object[toKey$2(last$3(path))];
  }

  function unset$2(object, path) {
    return object == null ? true : baseUnset$2(object, path);
  }

  var LARGE_ARRAY_SIZE$2$2 = 200;


  function baseDifference$2(array, values, iteratee, comparator) {
    var index = -1,
        includes = arrayIncludes$2,
        isCommon = true,
        length = array.length,
        result = [],
        valuesLength = values.length;

    if (!length) {
      return result;
    }
    if (iteratee) {
      values = arrayMap$2(values, baseUnary$2(iteratee));
    }
    if (comparator) {
      includes = arrayIncludesWith$2;
      isCommon = false;
    }
    else if (values.length >= LARGE_ARRAY_SIZE$2$2) {
      includes = cacheHas$2;
      isCommon = false;
      values = new SetCache$2(values);
    }
    outer:
    while (++index < length) {
      var value = array[index],
          computed = iteratee == null ? value : iteratee(value);

      value = (comparator || value !== 0) ? value : 0;
      if (isCommon && computed === computed) {
        var valuesIndex = valuesLength;
        while (valuesIndex--) {
          if (values[valuesIndex] === computed) {
            continue outer;
          }
        }
        result.push(value);
      }
      else if (!includes(values, computed, comparator)) {
        result.push(value);
      }
    }
    return result;
  }

  var without$2 = baseRest$2(function(array, values) {
    return isArrayLikeObject$2(array)
      ? baseDifference$2(array, values)
      : [];
  });

  class PathObject$2 {
    
    constructor (root) {
      if (root) {
        this.__root__ = root;
      }
    }

    contains (id) {
      return Boolean(this.getRoot()[id])
    }

    getRoot () {
      if (this.__root__) {
        return this.__root__
      } else {
        return this
      }
    }

    
    get (path) {
      if (!path) {
        return undefined
      }
      if (isString$2(path)) {
        return this.getRoot()[path]
      }
      if (arguments.length > 1) {
        path = Array.prototype.slice(arguments, 0);
      }
      if (!isArray$3(path)) {
        throw new Error('Illegal argument for PathObject.get()')
      }
      return get$2(this.getRoot(), path)
    }

    set (path, value) {
      if (!path) {
        throw new Error('Illegal argument: PathObject.set(>path<, value) - path is mandatory.')
      }
      if (isString$2(path)) {
        this.getRoot()[path] = value;
      } else {
        setWith$2(this.getRoot(), path, value);
      }
    }

    delete (path) {
      if (isString$2(path)) {
        delete this.getRoot()[path];
      } else if (path.length === 1) {
        delete this.getRoot()[path[0]];
      } else {
        var success = unset$2(this.getRoot(), path);
        if (!success) {
          throw new Error('Could not delete property at path' + path)
        }
      }
    }

    clear () {
      var root = this.getRoot();
      for (var key in root) {
        if (root.hasOwnProperty(key)) {
          delete root[key];
        }
      }
    }
  }

  PathObject$2.prototype._isPathObject = true;

  class Conflict$2 extends Error {
    constructor (a, b) {
      super('Conflict: ' + JSON.stringify(a) + ' vs ' + JSON.stringify(b));
      this.a = a;
      this.b = b;
    }
  }

  const INSERT$3 = 'insert';
  const DELETE$4 = 'delete';

  class TextOperation$2 {
    constructor (data) {
      if (!data || data.type === undefined || data.pos === undefined || data.str === undefined) {
        throw new Error('Illegal argument: insufficient data.')
      }
      
      this.type = data.type;
      
      this.pos = data.pos;
      
      this.str = data.str;
      
      if (!this.isInsert() && !this.isDelete()) {
        throw new Error('Illegal type.')
      }
      if (!isString$2(this.str)) {
        throw new Error('Illegal argument: expecting string.')
      }
      if (!isNumber$2(this.pos) || this.pos < 0) {
        throw new Error('Illegal argument: expecting positive number as pos.')
      }
    }

    apply (str) {
      if (this.isEmpty()) return str
      if (this.type === INSERT$3) {
        if (str.length < this.pos) {
          throw new Error('Provided string is too short.')
        }
        if (str.splice) {
          return str.splice(this.pos, 0, this.str)
        } else {
          return str.slice(0, this.pos).concat(this.str).concat(str.slice(this.pos))
        }
      } else  {
        if (str.length < this.pos + this.str.length) {
          throw new Error('Provided string is too short.')
        }
        if (str.splice) {
          return str.splice(this.pos, this.str.length)
        } else {
          return str.slice(0, this.pos).concat(str.slice(this.pos + this.str.length))
        }
      }
    }

    clone () {
      return new TextOperation$2(this)
    }

    isNOP () {
      return this.type === 'NOP' || this.str.length === 0
    }

    isInsert () {
      return this.type === INSERT$3
    }

    isDelete () {
      return this.type === DELETE$4
    }

    getLength () {
      return this.str.length
    }

    invert () {
      var data = {
        type: this.isInsert() ? DELETE$4 : INSERT$3,
        pos: this.pos,
        str: this.str
      };
      return new TextOperation$2(data)
    }

    hasConflict (other) {
      return _hasConflict$2(this, other)
    }

    isEmpty () {
      return this.str.length === 0
    }

    toJSON () {
      return {
        type: this.type,
        pos: this.pos,
        str: this.str
      }
    }

    toString () {
      return ['(', (this.isInsert() ? INSERT$3 : DELETE$4), ',', this.pos, ",'", this.str, "')"].join('')
    }
  }

  TextOperation$2.prototype._isOperation = true;
  TextOperation$2.prototype._isTextOperation = true;

  function _hasConflict$2 (a, b) {
    
    
    
    if (a.type === INSERT$3 && b.type === INSERT$3) return (a.pos === b.pos)
    
    
    
    if (a.type === DELETE$4 && b.type === DELETE$4) {
      
      return !(a.pos >= b.pos + b.str.length || b.pos >= a.pos + a.str.length)
    }
    
    
    
    var del, ins;
    if (a.type === DELETE$4) {
      del = a; ins = b;
    } else {
      del = b; ins = a;
    }
    return (ins.pos >= del.pos && ins.pos < del.pos + del.str.length)
  }




  function transformInsertInsert$3 (a, b) {
    if (a.pos === b.pos) {
      b.pos += a.str.length;
    } else if (a.pos < b.pos) {
      b.pos += a.str.length;
    } else {
      a.pos += b.str.length;
    }
  }





  function transformDeleteDelete$4 (a, b, first) {
    
    if (a.pos > b.pos) {
      return transformDeleteDelete$4(b, a, !first)
    }
    if (a.pos === b.pos && a.str.length > b.str.length) {
      return transformDeleteDelete$4(b, a, !first)
    }
    
    if (b.pos < a.pos + a.str.length) {
      var s = b.pos - a.pos;
      var s1 = a.str.length - s;
      var s2 = s + b.str.length;
      a.str = a.str.slice(0, s) + a.str.slice(s2);
      b.str = b.str.slice(s1);
      b.pos -= s;
    } else {
      b.pos -= a.str.length;
    }
  }





  function transformInsertDelete$3 (a, b) {
    if (a.type === DELETE$4) {
      return transformInsertDelete$3(b, a)
    }
    
    
    if (a.pos <= b.pos) {
      b.pos += a.str.length;
    
    } else if (a.pos >= b.pos + b.str.length) {
      a.pos -= b.str.length;
    
    
    
    } else {
      var s = a.pos - b.pos;
      b.str = b.str.slice(0, s) + a.str + b.str.slice(s);
      a.str = '';
    }
  }

  function transform$5 (a, b, options) {
    options = options || {};
    if (options['no-conflict'] && _hasConflict$2(a, b)) {
      throw new Conflict$2(a, b)
    }
    if (!options.inplace) {
      a = a.clone();
      b = b.clone();
    }
    if (a.type === INSERT$3 && b.type === INSERT$3) {
      transformInsertInsert$3(a, b);
    } else if (a.type === DELETE$4 && b.type === DELETE$4) {
      transformDeleteDelete$4(a, b, true);
    } else {
      transformInsertDelete$3(a, b);
    }
    return [a, b]
  }

  TextOperation$2.transform = function () {
    return transform$5.apply(null, arguments)
  };



  TextOperation$2.Insert = function (pos, str) {
    return new TextOperation$2({ type: INSERT$3, pos: pos, str: str })
  };

  TextOperation$2.Delete = function (pos, str) {
    return new TextOperation$2({ type: DELETE$4, pos: pos, str: str })
  };

  TextOperation$2.INSERT = INSERT$3;
  TextOperation$2.DELETE = DELETE$4;

  TextOperation$2.fromJSON = function (data) {
    return new TextOperation$2(data)
  };

  const NOP$3 = 'NOP';
  const DELETE$1$2 = 'delete';
  const INSERT$1$2 = 'insert';

  class ArrayOperation$2 {
    constructor (data) {
      if (!data || !data.type) {
        throw new Error('Illegal argument: insufficient data.')
      }
      this.type = data.type;
      if (this.type === NOP$3) return

      if (this.type !== INSERT$1$2 && this.type !== DELETE$1$2) {
        throw new Error('Illegal type.')
      }
      
      this.pos = data.pos;
      
      this.val = data.val;
      if (!isNumber$2(this.pos) || this.pos < 0) {
        throw new Error('Illegal argument: expecting positive number as pos.')
      }
    }

    apply (array) {
      if (this.type === NOP$3) {
        return array
      }
      if (this.type === INSERT$1$2) {
        if (array.length < this.pos) {
          throw new Error('Provided array is too small.')
        }
        array.splice(this.pos, 0, this.val);
        return array
      
      } else  {
        if (array.length < this.pos) {
          throw new Error('Provided array is too small.')
        }
        if (!isEqual$2(array[this.pos], this.val)) {
          throw Error('Unexpected value at position ' + this.pos + '. Expected ' + this.val + ', found ' + array[this.pos])
        }
        array.splice(this.pos, 1);
        return array
      }
    }

    clone () {
      var data = {
        type: this.type,
        pos: this.pos,
        val: cloneDeep$2(this.val)
      };
      return new ArrayOperation$2(data)
    }

    invert () {
      var data = this.toJSON();
      if (this.type === NOP$3) data.type = NOP$3;
      else if (this.type === INSERT$1$2) data.type = DELETE$1$2;
      else  data.type = INSERT$1$2;
      return new ArrayOperation$2(data)
    }

    hasConflict (other) {
      return ArrayOperation$2.hasConflict(this, other)
    }

    toJSON () {
      var result = {
        type: this.type
      };
      if (this.type === NOP$3) return result
      result.pos = this.pos;
      result.val = cloneDeep$2(this.val);
      return result
    }

    isInsert () {
      return this.type === INSERT$1$2
    }

    isDelete () {
      return this.type === DELETE$1$2
    }

    getOffset () {
      return this.pos
    }

    getValue () {
      return this.val
    }

    isNOP () {
      return this.type === NOP$3
    }

    toString () {
      return ['(', (this.isInsert() ? INSERT$1$2 : DELETE$1$2), ',', this.getOffset(), ",'", this.getValue(), "')"].join('')
    }
  }

  ArrayOperation$2.prototype._isOperation = true;
  ArrayOperation$2.prototype._isArrayOperation = true;

  function hasConflict$3 (a, b) {
    if (a.type === NOP$3 || b.type === NOP$3) return false
    if (a.type === INSERT$1$2 && b.type === INSERT$1$2) {
      return a.pos === b.pos
    } else {
      return false
    }
  }

  function transformInsertInsert$1$2 (a, b) {
    if (a.pos === b.pos) {
      b.pos += 1;
    
    } else if (a.pos < b.pos) {
      b.pos += 1;
    
    } else {
      a.pos += 1;
    }
  }

  function transformDeleteDelete$1$2 (a, b) {
    
    if (a.pos === b.pos) {
      b.type = NOP$3;
      a.type = NOP$3;
      return
    }
    if (a.pos < b.pos) {
      b.pos -= 1;
    } else {
      a.pos -= 1;
    }
  }

  function transformInsertDelete$1$2 (a, b) {
    
    if (a.type === DELETE$1$2) {
      var tmp = a;
      a = b;
      b = tmp;
    }
    if (a.pos <= b.pos) {
      b.pos += 1;
    } else {
      a.pos -= 1;
    }
  }

  var transform$1$2 = function (a, b, options) {
    options = options || {};
    
    
    if (options['no-conflict'] && hasConflict$3(a, b)) {
      throw new Conflict$2(a, b)
    }
    
    if (!options.inplace) {
      a = a.clone();
      b = b.clone();
    }
    if (a.type === NOP$3 || b.type === NOP$3) ; else if (a.type === INSERT$1$2 && b.type === INSERT$1$2) {
      transformInsertInsert$1$2(a, b);
    } else if (a.type === DELETE$1$2 && b.type === DELETE$1$2) {
      transformDeleteDelete$1$2(a, b);
    } else {
      transformInsertDelete$1$2(a, b);
    }
    return [a, b]
  };

  ArrayOperation$2.transform = transform$1$2;
  ArrayOperation$2.hasConflict = hasConflict$3;



  ArrayOperation$2.Insert = function (pos, val) {
    return new ArrayOperation$2({type: INSERT$1$2, pos: pos, val: val})
  };

  ArrayOperation$2.Delete = function (pos, val) {
    return new ArrayOperation$2({ type: DELETE$1$2, pos: pos, val: val })
  };

  ArrayOperation$2.fromJSON = function (data) {
    return new ArrayOperation$2(data)
  };

  ArrayOperation$2.NOP = NOP$3;
  ArrayOperation$2.DELETE = DELETE$1$2;
  ArrayOperation$2.INSERT = INSERT$1$2;

  const SHIFT$2 = 'shift';

  class CoordinateOperation$2 {
    constructor (data) {
      if (!data || data.type === undefined) {
        throw new Error('Illegal argument: insufficient data.')
      }
      
      this.type = data.type;
      
      this.val = data.val;
      
      if (!this.isShift()) {
        throw new Error('Illegal type.')
      }
      if (!isNumber$2(this.val)) {
        throw new Error('Illegal argument: expecting number as shift value.')
      }
    }

    apply (coor) {
      coor.offset = coor.offset + this.val;
    }

    isShift () {
      return this.type === SHIFT$2
    }

    isNOP () {
      switch (this.type) {
        case SHIFT$2: {
          return this.val === 0
        }
        default:
          return false
      }
    }

    clone () {
      return new CoordinateOperation$2(this)
    }

    invert () {
      let data;
      switch (this.type) {
        case SHIFT$2:
          data = {
            type: SHIFT$2,
            val: -this.val
          };
          break
        default:
          throw new Error('Invalid type.')
      }
      return new CoordinateOperation$2(data)
    }

    hasConflict () {
      
      return false
    }

    toJSON () {
      return {
        type: this.type,
        val: this.val
      }
    }

    toString () {
      return ['(', (this.type), ',', this.val, "')"].join('')
    }
  }

  CoordinateOperation$2.prototype._isOperation = true;
  CoordinateOperation$2.prototype._isCoordinateOperation = true;

  function transformShiftShift$2 (a, b) {
    a.val += b.val;
    b.val += a.val;
  }

  function transform$2$2 (a, b, options) {
    options = options || {};
    
    if (!options.inplace) {
      a = a.clone();
      b = b.clone();
    }
    if (a.type === SHIFT$2 && b.type === SHIFT$2) {
      transformShiftShift$2(a, b);
    } else {
      throw new Error('Illegal type')
    }
    return [a, b]
  }

  CoordinateOperation$2.transform = function (...args) {
    return transform$2$2(...args)
  };

  CoordinateOperation$2.fromJSON = function (json) {
    return new CoordinateOperation$2(json)
  };

  CoordinateOperation$2.Shift = function (val) {
    return new CoordinateOperation$2({
      type: SHIFT$2,
      val: val
    })
  };

  const NOP$1$2 = 'NOP';
  const CREATE$2 = 'create';
  const DELETE$2$2 = 'delete';
  const UPDATE$2 = 'update';
  const SET$2 = 'set';

  class ObjectOperation$2 {
    constructor (data) {
      
      if (!data) {
        throw new Error('Data of ObjectOperation is missing.')
      }
      
      if (!data.type) {
        throw new Error('Invalid data: type is mandatory.')
      }
      this.type = data.type;
      if (data.type === NOP$1$2) {
        return
      }
      this.path = data.path;
      if (!data.path) {
        throw new Error('Invalid data: path is mandatory.')
      }
      if (this.type === CREATE$2 || this.type === DELETE$2$2) {
        if (!data.val) {
          throw new Error('Invalid data: value is missing.')
        }
        this.val = data.val;
      } else if (this.type === UPDATE$2) {
        if (data.diff) {
          this.diff = data.diff;
          if (data.diff._isTextOperation) {
            this.propertyType = 'string';
          } else if (data.diff._isArrayOperation) {
            this.propertyType = 'array';
          } else if (data.diff._isCoordinateOperation) {
            this.propertyType = 'coordinate';
          } else {
            throw new Error('Invalid data: diff must be a TextOperation or an ArrayOperation.')
          }
        } else {
          throw new Error('Invalid data: diff is mandatory for update operation.')
        }
      } else if (this.type === SET$2) {
        this.val = data.val;
        this.original = data.original;
      } else {
        throw new Error('Invalid type: ' + data.type)
      }
    }

    apply (obj) {
      if (this.type === NOP$1$2) return obj
      var adapter;
      if (obj._isPathObject) {
        adapter = obj;
      } else {
        adapter = new PathObject$2(obj);
      }
      if (this.type === CREATE$2) {
        adapter.set(this.path, cloneDeep$2(this.val));
        return obj
      }
      if (this.type === DELETE$2$2) {
        adapter.delete(this.path, 'strict');
      } else if (this.type === UPDATE$2) {
        var diff = this.diff;
        switch (this.propertyType) {
          case 'array': {
            let arr = adapter.get(this.path);
            diff.apply(arr);
            break
          }
          case 'string': {
            let str = adapter.get(this.path);
            if (isNil$2(str)) str = '';
            str = diff.apply(str);
            adapter.set(this.path, str);
            break
          }
          case 'coordinate': {
            let coor = adapter.get(this.path);
            if (!coor) throw new Error('No coordinate with path ' + this.path)
            diff.apply(coor);
            break
          }
          default:
            throw new Error('Invalid state.')
        }
      } else if (this.type === SET$2) {
        
        adapter.set(this.path, cloneDeep$2(this.val));
      } else {
        throw new Error('Invalid type.')
      }
      return obj
    }

    clone () {
      var data = {
        type: this.type,
        path: this.path
      };
      if (this.val) {
        data.val = cloneDeep$2(this.val);
      }
      if (this.diff) {
        data.diff = this.diff.clone();
      }
      return new ObjectOperation$2(data)
    }

    isNOP () {
      if (this.type === NOP$1$2) return true
      else if (this.type === UPDATE$2) return this.diff.isNOP()
    }

    isCreate () {
      return this.type === CREATE$2
    }

    isDelete () {
      return this.type === DELETE$2$2
    }

    isUpdate (propertyType) {
      if (propertyType) {
        return (this.type === UPDATE$2 && this.propertyType === propertyType)
      } else {
        return this.type === UPDATE$2
      }
    }

    isSet () {
      return this.type === SET$2
    }

    invert () {
      if (this.type === NOP$1$2) {
        return new ObjectOperation$2({ type: NOP$1$2 })
      }
      var result = new ObjectOperation$2(this);
      if (this.type === CREATE$2) {
        result.type = DELETE$2$2;
      } else if (this.type === DELETE$2$2) {
        result.type = CREATE$2;
      } else if (this.type === UPDATE$2) {
        var invertedDiff;
        if (this.diff._isTextOperation) {
          invertedDiff = TextOperation$2.fromJSON(this.diff.toJSON()).invert();
        } else if (this.diff._isArrayOperation) {
          invertedDiff = ArrayOperation$2.fromJSON(this.diff.toJSON()).invert();
        } else if (this.diff._isCoordinateOperation) {
          invertedDiff = CoordinateOperation$2.fromJSON(this.diff.toJSON()).invert();
        } else {
          throw new Error('Illegal type')
        }
        result.diff = invertedDiff;
      } else  {
        result.val = this.original;
        result.original = this.val;
      }
      return result
    }

    hasConflict (other) {
      return ObjectOperation$2.hasConflict(this, other)
    }

    toJSON () {
      if (this.type === NOP$1$2) {
        return { type: NOP$1$2 }
      }
      var data = {
        type: this.type,
        path: this.path
      };
      if (this.type === CREATE$2 || this.type === DELETE$2$2) {
        data.val = this.val;
      } else if (this.type === UPDATE$2) {
        if (this.diff._isTextOperation) {
          data.propertyType = 'string';
        } else if (this.diff._isArrayOperation) {
          data.propertyType = 'array';
        } else if (this.diff._isCoordinateOperation) {
          data.propertyType = 'coordinate';
        } else {
          throw new Error('Invalid property type.')
        }
        data.diff = this.diff.toJSON();
      } else  {
        data.val = this.val;
        data.original = this.original;
      }
      return data
    }

    getType () {
      return this.type
    }

    getPath () {
      return this.path
    }

    getValue () {
      return this.val
    }

    getOldValue () {
      return this.original
    }

    getValueOp () {
      return this.diff
    }

    
    toString () {
      switch (this.type) {
        case CREATE$2:
          return ['(+,', JSON.stringify(this.path), JSON.stringify(this.val), ')'].join('')
        case DELETE$2$2:
          return ['(-,', JSON.stringify(this.path), JSON.stringify(this.val), ')'].join('')
        case UPDATE$2:
          return ['(>>,', JSON.stringify(this.path), this.propertyType, this.diff.toString(), ')'].join('')
        case SET$2:
          return ['(=,', JSON.stringify(this.path), this.val, this.original, ')'].join('')
        case NOP$1$2:
          return 'NOP'
        default:
          throw new Error('Invalid type')
      }
    }
  }

  ObjectOperation$2.prototype._isOperation = true;
  ObjectOperation$2.prototype._isObjectOperation = true;



  function hasConflict$1$2 (a, b) {
    if (a.type === NOP$1$2 || b.type === NOP$1$2) return false
    return isEqual$2(a.path, b.path)
  }

  function transformDeleteDelete$2$2 (a, b) {
    
    
    a.type = NOP$1$2;
    b.type = NOP$1$2;
  }

  function transformCreateCreate$2 () {
    throw new Error('Can not transform two concurring creates of the same property')
  }

  function transformDeleteCreate$2 () {
    throw new Error('Illegal state: can not create and delete a value at the same time.')
  }

  function transformDeleteUpdate$2 (a, b, flipped) {
    if (a.type !== DELETE$2$2) {
      return transformDeleteUpdate$2(b, a, true)
    }
    var op;
    switch (b.propertyType) {
      case 'string':
        op = TextOperation$2.fromJSON(b.diff);
        break
      case 'array':
        op = ArrayOperation$2.fromJSON(b.diff);
        break
      case 'coordinate':
        op = CoordinateOperation$2.fromJSON(b.diff);
        break
      default:
        throw new Error('Illegal type')
    }
    
    if (!flipped) {
      a.type = NOP$1$2;
      b.type = CREATE$2;
      b.val = op.apply(a.val);
    
    } else {
      a.val = op.apply(a.val);
      b.type = NOP$1$2;
    }
  }

  function transformCreateUpdate$2 () {
    
    throw new Error('Can not transform a concurring create and update of the same property')
  }

  function transformUpdateUpdate$2 (a, b) {
    
    let opA, opB, t;
    switch (b.propertyType) {
      case 'string':
        opA = TextOperation$2.fromJSON(a.diff);
        opB = TextOperation$2.fromJSON(b.diff);
        t = TextOperation$2.transform(opA, opB, {inplace: true});
        break
      case 'array':
        opA = ArrayOperation$2.fromJSON(a.diff);
        opB = ArrayOperation$2.fromJSON(b.diff);
        t = ArrayOperation$2.transform(opA, opB, {inplace: true});
        break
      case 'coordinate':
        opA = CoordinateOperation$2.fromJSON(a.diff);
        opB = CoordinateOperation$2.fromJSON(b.diff);
        t = CoordinateOperation$2.transform(opA, opB, {inplace: true});
        break
      default:
        throw new Error('Illegal type')
    }
    a.diff = t[0];
    b.diff = t[1];
  }

  function transformCreateSet$2 () {
    throw new Error('Illegal state: can not create and set a value at the same time.')
  }

  function transformDeleteSet$2 (a, b, flipped) {
    if (a.type !== DELETE$2$2) return transformDeleteSet$2(b, a, true)
    if (!flipped) {
      a.type = NOP$1$2;
      b.type = CREATE$2;
      b.original = undefined;
    } else {
      a.val = b.val;
      b.type = NOP$1$2;
    }
  }

  function transformUpdateSet$2 () {
    throw new Error('Unresolvable conflict: update + set.')
  }

  function transformSetSet$2 (a, b) {
    a.type = NOP$1$2;
    b.original = a.val;
  }

  const _NOP$2 = 0;
  const _CREATE$2 = 1;
  const _DELETE$2 = 2;
  const _UPDATE$2 = 4;
  const _SET$2 = 8;

  const CODE$2 = (() => {
    const c = {};
    c[NOP$1$2] = _NOP$2;
    c[CREATE$2] = _CREATE$2;
    c[DELETE$2$2] = _DELETE$2;
    c[UPDATE$2] = _UPDATE$2;
    c[SET$2] = _SET$2;
    return c
  })();

  const __transform__$2 = (() => {
    
    const t = {};
    t[_DELETE$2 | _DELETE$2] = transformDeleteDelete$2$2;
    t[_DELETE$2 | _CREATE$2] = transformDeleteCreate$2;
    t[_DELETE$2 | _UPDATE$2] = transformDeleteUpdate$2;
    t[_CREATE$2 | _CREATE$2] = transformCreateCreate$2;
    t[_CREATE$2 | _UPDATE$2] = transformCreateUpdate$2;
    t[_UPDATE$2 | _UPDATE$2] = transformUpdateUpdate$2;
    t[_CREATE$2 | _SET$2] = transformCreateSet$2;
    t[_DELETE$2 | _SET$2] = transformDeleteSet$2;
    t[_UPDATE$2 | _SET$2] = transformUpdateSet$2;
    t[_SET$2    | _SET$2] = transformSetSet$2;
    
    return t
  })();

  function transform$3$2 (a, b, options) {
    options = options || {};
    if (options['no-conflict'] && hasConflict$1$2(a, b)) {
      throw new Conflict$2(a, b)
    }
    if (!options.inplace) {
      a = a.clone();
      b = b.clone();
    }
    if (a.isNOP() || b.isNOP()) {
      return [a, b]
    }
    var sameProp = isEqual$2(a.path, b.path);
    
    if (sameProp) {
      __transform__$2[CODE$2[a.type] | CODE$2[b.type]](a, b);
    }
    return [a, b]
  }

  ObjectOperation$2.transform = transform$3$2;
  ObjectOperation$2.hasConflict = hasConflict$1$2;



  ObjectOperation$2.Create = function (idOrPath, val) {
    var path;
    if (isString$2(idOrPath)) {
      path = [idOrPath];
    } else {
      path = idOrPath;
    }
    return new ObjectOperation$2({type: CREATE$2, path: path, val: val})
  };

  ObjectOperation$2.Delete = function (idOrPath, val) {
    var path;
    if (isString$2(idOrPath)) {
      path = [idOrPath];
    } else {
      path = idOrPath;
    }
    return new ObjectOperation$2({type: DELETE$2$2, path: path, val: val})
  };

  ObjectOperation$2.Update = function (path, op) {
    return new ObjectOperation$2({
      type: UPDATE$2,
      path: path,
      diff: op
    })
  };

  ObjectOperation$2.Set = function (path, oldVal, newVal) {
    return new ObjectOperation$2({
      type: SET$2,
      path: path,
      val: cloneDeep$2(newVal),
      original: cloneDeep$2(oldVal)
    })
  };

  ObjectOperation$2.fromJSON = function (data) {
    data = cloneDeep$2(data);
    if (data.type === 'update') {
      switch (data.propertyType) {
        case 'string':
          data.diff = TextOperation$2.fromJSON(data.diff);
          break
        case 'array':
          data.diff = ArrayOperation$2.fromJSON(data.diff);
          break
        case 'coordinate':
          data.diff = CoordinateOperation$2.fromJSON(data.diff);
          break
        default:
          throw new Error('Unsupported update diff:' + JSON.stringify(data.diff))
      }
    }
    var op = new ObjectOperation$2(data);
    return op
  };

  ObjectOperation$2.NOP = NOP$1$2;
  ObjectOperation$2.CREATE = CREATE$2;
  ObjectOperation$2.DELETE = DELETE$2$2;
  ObjectOperation$2.UPDATE = UPDATE$2;
  ObjectOperation$2.SET = SET$2;

  class OperationSerializer$2 {
    constructor () {
      this.SEPARATOR = '\t';
    }

    serialize (op) {
      var out = [];
      switch (op.type) {
        case 'create':
          out.push('c');
          out.push(op.val.id);
          out.push(op.val);
          break
        case 'delete':
          out.push('d');
          out.push(op.val.id);
          out.push(op.val);
          break
        case 'set':
          out.push('s');
          out.push(op.path.join('.'));
          out.push(op.val);
          out.push(op.original);
          break
        case 'update':
          out.push('u');
          out.push(op.path.join('.'));
          Array.prototype.push.apply(out, this.serializePrimitiveOp(op.diff));
          break
        default:
          throw new Error('Unsupported operation type.')
      }
      return out
    }

    serializePrimitiveOp (op) {
      var out = [];
      if (op._isTextOperation) {
        if (op.isInsert()) {
          out.push('t+');
        } else if (op.isDelete()) {
          out.push('t-');
        }
        out.push(op.pos);
        out.push(op.str);
      } else if (op._isArrayOperation) {
        if (op.isInsert()) {
          out.push('a+');
        } else if (op.isDelete()) {
          out.push('a-');
        }
        out.push(op.pos);
        out.push(op.val);
      } else if (op._isCoordinateOperation) {
        if (op.isShift()) {
          out.push('c>>');
        } else {
          throw new Error('Unsupported CoordinateOperation type.')
        }
        out.push(op.pos);
        out.push(op.val);
      } else {
        throw new Error('Unsupported operation type.')
      }
      return out
    }

    deserialize (str, tokenizer) {
      if (!tokenizer) {
        tokenizer = new Tokenizer$4(str, this.SEPARATOR);
      }
      var type = tokenizer.getString();
      var op, path, val, oldVal, diff;
      switch (type) {
        case 'c':
          path = tokenizer.getPath();
          val = tokenizer.getObject();
          op = ObjectOperation$2.Create(path, val);
          break
        case 'd':
          path = tokenizer.getPath();
          val = tokenizer.getObject();
          op = ObjectOperation$2.Delete(path, val);
          break
        case 's':
          path = tokenizer.getPath();
          val = tokenizer.getAny();
          oldVal = tokenizer.getAny();
          op = ObjectOperation$2.Set(path, oldVal, val);
          break
        case 'u':
          path = tokenizer.getPath();
          diff = this.deserializePrimitiveOp(str, tokenizer);
          op = ObjectOperation$2.Update(path, diff);
          break
        default:
          throw new Error('Illegal type for ObjectOperation: ' + type)
      }
      return op
    }

    deserializePrimitiveOp (str, tokenizer) {
      if (!tokenizer) {
        tokenizer = new Tokenizer$4(str, this.SEPARATOR);
      }
      var type = tokenizer.getString();
      var op, pos, val;
      switch (type) {
        case 't+':
          pos = tokenizer.getNumber();
          val = tokenizer.getString();
          op = TextOperation$2.Insert(pos, val);
          break
        case 't-':
          pos = tokenizer.getNumber();
          val = tokenizer.getString();
          op = TextOperation$2.Delete(pos, val);
          break
        case 'a+':
          pos = tokenizer.getNumber();
          val = tokenizer.getAny();
          op = ArrayOperation$2.Insert(pos, val);
          break
        case 'a-':
          pos = tokenizer.getNumber();
          val = tokenizer.getAny();
          op = ArrayOperation$2.Delete(pos, val);
          break
        case 'c>>':
          val = tokenizer.getNumber();
          op = CoordinateOperation$2.Shift(val);
          break
        default:
          throw new Error('Unsupported operation type: ' + type)
      }
      return op
    }
  }

  class Tokenizer$4 {
    constructor (str, sep) {
      if (isArray$3(arguments[0])) {
        this.tokens = arguments[0];
      } else {
        this.tokens = str.split(sep);
      }
      this.pos = -1;
    }

    error (msg) {
      throw new Error('Parsing error: ' + msg + '\n' + this.tokens[this.pos])
    }

    getString () {
      this.pos++;
      var str = this.tokens[this.pos];
      if (str[0] === '"') {
        str = str.slice(1, -1);
      }
      return str
    }

    getNumber () {
      this.pos++;
      var number;
      var token = this.tokens[this.pos];
      try {
        if (isNumber$2(token)) {
          number = token;
        } else {
          number = parseInt(this.tokens[this.pos], 10);
        }
        return number
      } catch (err) {
        this.error('expected number');
      }
    }

    getObject () {
      this.pos++;
      var obj;
      var token = this.tokens[this.pos];
      try {
        if (isObject$3(token)) {
          obj = token;
        } else {
          obj = JSON.parse(this.tokens[this.pos]);
        }
        return obj
      } catch (err) {
        this.error('expected object');
      }
    }

    getAny () {
      this.pos++;
      var token = this.tokens[this.pos];
      return token
    }

    getPath () {
      var str = this.getString();
      return str.split('.')
    }
  }

  OperationSerializer$2.Tokenizer = Tokenizer$4;

  function last$1$2 (arr) {
    return arr[arr.length - 1]
  }

  class Selection$2 {
    constructor () {
      
      var _internal = {};
      Object.defineProperty(this, '_internal', {
        enumerable: false,
        value: _internal
      });
      
      _internal.doc = null;
    }

    clone () {
      var newSel = this._clone();
      if (this._internal.doc) {
        newSel.attach(this._internal.doc);
      }
      return newSel
    }

    
    getDocument () {
      var doc = this._internal.doc;
      if (!doc) {
        throw new Error('Selection is not attached to a document.')
      }
      return doc
    }

    isAttached () {
      return Boolean(this._internal.doc)
    }

    
    attach (doc) {
      this._internal.doc = doc;
      return this
    }

    
    isNull () { return false }

    
    isPropertySelection () { return false }

    
    isContainerSelection () { return false }

    
    isNodeSelection () { return false }

    isCustomSelection () { return false }

    
    isCollapsed () { return true }

    
    isReverse () { return false }

    getType () {
      throw new Error('Selection.getType() is abstract.')
    }

    get type () {
      return this.getType()
    }

    
    equals (other) {
      if (this === other) {
        return true
      } else if (!other) {
        return false
      } else if (this.isNull() !== other.isNull()) {
        return false
      } else if (this.getType() !== other.getType()) {
        return false
      } else {
        
        
        return true
      }
    }

    
    toString () {
      return 'null'
    }

    
    toJSON () {
      throw new Error('This method is abstract.')
    }

    createWith (update) {
      let SelectionClass = this.constructor;
      let data = this.toJSON();
      Object.assign(data, update);
      return SelectionClass.fromJSON(data)
    }
  }


  Selection$2.prototype._isSelection = true;


  class NullSelection$2 extends Selection$2 {
    isNull () {
      return true
    }

    getType () {
      return 'null'
    }

    toJSON () {
      return null
    }

    clone () {
      return this
    }
  }



  Selection$2.nullSelection = Object.freeze(new NullSelection$2());

  function isArrayEqual$2 (arr1, arr2) {
    if (arr1 === arr2) return true
    if (!isArray$3(arr1) || !isArray$3(arr2)) return false
    if (arr1.length !== arr2.length) return false
    let L = arr1.length;
    for (var i = 0; i < L; i++) {
      if (arr1[i] !== arr2[i]) return false
    }
    return true
  }

  class Coordinate$2 {
    
    constructor (path, offset) {
      
      if (arguments[0] === 'SKIP') return
      if (arguments.length === 1) {
        let data = arguments[0];
        this.path = data.path;
        this.offset = data.offset;
      } else {
        this.path = path;
        this.offset = offset;
      }
      if (!isArray$3(this.path)) {
        throw new Error('Invalid arguments: path should be an array.')
      }
      if (!isNumber$2(this.offset) || this.offset < 0) {
        throw new Error('Invalid arguments: offset must be a positive number.')
      }
    }

    equals (other) {
      return (other === this ||
        (isArrayEqual$2(other.path, this.path) && other.offset === this.offset))
    }

    withCharPos (offset) {
      return new Coordinate$2(this.path, offset)
    }

    getNodeId () {
      return this.path[0]
    }

    getPath () {
      return this.path
    }

    getOffset () {
      return this.offset
    }

    toJSON () {
      return {
        path: this.path.slice(),
        offset: this.offset
      }
    }

    toString () {
      return '(' + this.path.join('.') + ', ' + this.offset + ')'
    }

    isPropertyCoordinate () {
      return this.path.length > 1
    }

    isNodeCoordinate () {
      return this.path.length === 1
    }

    hasSamePath (other) {
      return isArrayEqual$2(this.path, other.path)
    }
  }

  Coordinate$2.prototype._isCoordinate = true;

  class PropertySelection$2 extends Selection$2 {
    
    constructor (path, startOffset, endOffset, reverse, containerId, surfaceId) {
      super();

      if (arguments.length === 1) {
        let data = arguments[0];
        path = data.path;
        startOffset = data.startOffset;
        endOffset = data.endOffset;
        reverse = data.reverse;
        containerId = data.containerId;
        surfaceId = data.surfaceId;
      }

      if (!path || !isNumber$2(startOffset)) {
        throw new Error('Invalid arguments: `path` and `startOffset` are mandatory')
      }

      this.start = new Coordinate$2(path, startOffset);
      this.end = new Coordinate$2(path, isNumber$2(endOffset) ? endOffset : startOffset);

      
      this.reverse = Boolean(reverse);

      this.containerId = containerId;

      
      this.surfaceId = surfaceId;
    }

    get path () {
      return this.start.path
    }

    get startOffset () {
      console.warn('DEPRECATED: Use sel.start.offset instead');
      return this.start.offset
    }

    get endOffset () {
      console.warn('DEPRECATED: Use sel.end.offset instead');
      return this.end.offset
    }

    
    toJSON () {
      return {
        type: 'property',
        path: this.start.path,
        startOffset: this.start.offset,
        endOffset: this.end.offset,
        reverse: this.reverse,
        containerId: this.containerId,
        surfaceId: this.surfaceId
      }
    }

    isPropertySelection () {
      return true
    }

    getType () {
      return 'property'
    }

    isNull () {
      return false
    }

    isCollapsed () {
      return this.start.offset === this.end.offset
    }

    isReverse () {
      return this.reverse
    }

    equals (other) {
      return (
        Selection$2.prototype.equals.call(this, other) &&
        (this.start.equals(other.start) && this.end.equals(other.end))
      )
    }

    toString () {
      
      return [
        'PropertySelection(', JSON.stringify(this.path), ', ',
        this.start.offset, ' -> ', this.end.offset,
        (this.reverse ? ', reverse' : ''),
        (this.surfaceId ? (', ' + this.surfaceId) : ''),
        ')'
      ].join('')
    }

    
    collapse (direction) {
      var offset;
      if (direction === 'left') {
        offset = this.start.offset;
      } else {
        offset = this.end.offset;
      }
      return this.createWithNewRange(offset, offset)
    }

    
    

    
    getPath () {
      return this.start.path
    }

    getNodeId () {
      return this.start.path[0]
    }

    
    isInsideOf (other, strict) {
      if (other.isNull()) return false
      if (other.isContainerSelection()) {
        return other.contains(this, strict)
      }
      if (strict) {
        return (isArrayEqual$2(this.path, other.path) &&
          this.start.offset > other.start.offset &&
          this.end.offset < other.end.offset)
      } else {
        return (isArrayEqual$2(this.path, other.path) &&
          this.start.offset >= other.start.offset &&
          this.end.offset <= other.end.offset)
      }
    }

    
    contains (other, strict) {
      if (other.isNull()) return false
      return other.isInsideOf(this, strict)
    }

    
    overlaps (other, strict) {
      if (other.isNull()) return false
      if (other.isContainerSelection()) {
        
        return other.overlaps(this)
      }
      if (!isArrayEqual$2(this.path, other.path)) return false
      if (strict) {
        return (!(this.start.offset >= other.end.offset || this.end.offset <= other.start.offset))
      } else {
        return (!(this.start.offset > other.end.offset || this.end.offset < other.start.offset))
      }
    }

    
    isRightAlignedWith (other) {
      if (other.isNull()) return false
      if (other.isContainerSelection()) {
        
        return other.isRightAlignedWith(this)
      }
      return (isArrayEqual$2(this.path, other.path) &&
        this.end.offset === other.end.offset)
    }

    
    isLeftAlignedWith (other) {
      if (other.isNull()) return false
      if (other.isContainerSelection()) {
        
        return other.isLeftAlignedWith(this)
      }
      return (isArrayEqual$2(this.path, other.path) &&
        this.start.offset === other.start.offset)
    }

    
    expand (other) {
      if (other.isNull()) return this

      
      
      
      if (other.isContainerSelection()) {
        return other.expand(this)
      }
      if (!isArrayEqual$2(this.path, other.path)) {
        throw new Error('Can not expand PropertySelection to a different property.')
      }
      var newStartOffset = Math.min(this.start.offset, other.start.offset);
      var newEndOffset = Math.max(this.end.offset, other.end.offset);
      return this.createWithNewRange(newStartOffset, newEndOffset)
    }

    
    truncateWith (other) {
      if (other.isNull()) return this
      if (other.isInsideOf(this, 'strict')) {
        
        throw new Error('Can not truncate with a contained selections')
      }
      if (!this.overlaps(other)) {
        return this
      }
      var otherStartOffset, otherEndOffset;
      if (other.isPropertySelection()) {
        otherStartOffset = other.start.offset;
        otherEndOffset = other.end.offset;
      } else if (other.isContainerSelection()) {
        
        if (isArrayEqual$2(other.start.path, this.start.path)) {
          otherStartOffset = other.start.offset;
        } else {
          otherStartOffset = this.start.offset;
        }
        if (isArrayEqual$2(other.end.path, this.start.path)) {
          otherEndOffset = other.end.offset;
        } else {
          otherEndOffset = this.end.offset;
        }
      } else {
        return this
      }

      var newStartOffset;
      var newEndOffset;
      if (this.start.offset > otherStartOffset && this.end.offset > otherEndOffset) {
        newStartOffset = otherEndOffset;
        newEndOffset = this.end.offset;
      } else if (this.start.offset < otherStartOffset && this.end.offset < otherEndOffset) {
        newStartOffset = this.start.offset;
        newEndOffset = otherStartOffset;
      } else if (this.start.offset === otherStartOffset) {
        if (this.end.offset <= otherEndOffset) {
          return Selection$2.nullSelection
        } else {
          newStartOffset = otherEndOffset;
          newEndOffset = this.end.offset;
        }
      } else if (this.end.offset === otherEndOffset) {
        if (this.start.offset >= otherStartOffset) {
          return Selection$2.nullSelection
        } else {
          newStartOffset = this.start.offset;
          newEndOffset = otherStartOffset;
        }
      } else if (other.contains(this)) {
        return Selection$2.nullSelection
      } else {
        
        throw new Error('Illegal state.')
      }
      return this.createWithNewRange(newStartOffset, newEndOffset)
    }

    
    createWithNewRange (startOffset, endOffset) {
      var sel = new PropertySelection$2(this.path, startOffset, endOffset, false, this.containerId, this.surfaceId);
      var doc = this._internal.doc;
      if (doc) {
        sel.attach(doc);
      }
      return sel
    }

    _clone () {
      return new PropertySelection$2(this.start.path, this.start.offset, this.end.offset, this.reverse, this.containerId, this.surfaceId)
    }
  }

  PropertySelection$2.fromJSON = function (json) {
    return new PropertySelection$2(json)
  };

  class ContainerSelection$2 extends Selection$2 {
    constructor (containerId, startPath, startOffset, endPath, endOffset, reverse, surfaceId) {
      super();

      if (arguments.length === 1) {
        let data = arguments[0];
        containerId = data.containerId;
        startPath = data.startPath;
        startOffset = data.startOffset;
        endPath = data.endPath;
        endOffset = data.endOffset;
        reverse = data.reverse;
        surfaceId = data.surfaceId;
      }

      
      this.containerId = containerId;
      if (!this.containerId) throw new Error('Invalid arguments: `containerId` is mandatory')

      this.start = new Coordinate$2(startPath, startOffset);
      this.end = new Coordinate$2(isNil$2(endPath) ? startPath : endPath, isNil$2(endOffset) ? startOffset : endOffset);

      this.reverse = Boolean(reverse);

      this.surfaceId = surfaceId;
    }

    

    get startPath () {
      console.warn('DEPRECATED: use sel.start.path instead.');
      return this.start.path
    }

    get startOffset () {
      console.warn('DEPRECATED: use sel.start.offset instead.');
      return this.start.offset
    }

    get endPath () {
      console.warn('DEPRECATED: use sel.end.path instead.');
      return this.end.path
    }

    get endOffset () {
      console.warn('DEPRECATED: use sel.end.offset instead.');
      return this.end.offset
    }

    

    toJSON () {
      return {
        type: 'container',
        containerId: this.containerId,
        startPath: this.start.path,
        startOffset: this.start.offset,
        endPath: this.end.path,
        endOffset: this.end.offset,
        reverse: this.reverse,
        surfaceId: this.surfaceId
      }
    }

    isContainerSelection () {
      return true
    }

    getType () {
      return 'container'
    }

    isNull () {
      return false
    }

    isCollapsed () {
      return this.start.equals(this.end)
    }

    isReverse () {
      return this.reverse
    }

    equals (other) {
      return (
        Selection$2.prototype.equals.call(this, other) &&
        this.containerId === other.containerId &&
        (this.start.equals(other.start) && this.end.equals(other.end))
      )
    }

    toString () {
      
      return [
        'ContainerSelection(',
        this.containerId, ', ',
        JSON.stringify(this.start.path), ', ', this.start.offset,
        ' -> ',
        JSON.stringify(this.end.path), ', ', this.end.offset,
        (this.reverse ? ', reverse' : ''),
        (this.surfaceId ? (', ' + this.surfaceId) : ''),
        ')'
      ].join('')
    }

    
    getContainer () {
      if (!this._internal.container) {
        this._internal.container = this.getDocument().get(this.containerId);
      }
      return this._internal.container
    }

    isInsideOf (other, strict) {
      
      
      if (other.isNull()) return false
      strict = Boolean(strict);
      let r1 = this._range(this);
      let r2 = this._range(other);
      return (r2.start.isBefore(r1.start, strict) &&
        r1.end.isBefore(r2.end, strict))
    }

    contains (other, strict) {
      
      
      if (other.isNull()) return false
      strict = Boolean(strict);
      let r1 = this._range(this);
      let r2 = this._range(other);
      return (r1.start.isBefore(r2.start, strict) &&
        r2.end.isBefore(r1.end, strict))
    }

    containsNode (nodeId, strict) {
      const container = this.getContainer();
      if (!container.contains(nodeId)) return false
      const coor = new Coordinate$2([nodeId], 0);
      const address = container.getAddress(coor);
      const r = this._range(this);
      
      let contained = r.start.isBefore(address, strict);
      if (contained) {
        address.offset = 1;
        contained = r.end.isAfter(address, strict);
      }
      return contained
    }

    overlaps (other) {
      let r1 = this._range(this);
      let r2 = this._range(other);
      
      return !(r1.end.isBefore(r2.start, false) ||
        r2.end.isBefore(r1.start, false))
    }

    isLeftAlignedWith (other) {
      let r1 = this._range(this);
      let r2 = this._range(other);
      return r1.start.isEqual(r2.start)
    }

    isRightAlignedWith (other) {
      let r1 = this._range(this);
      let r2 = this._range(other);
      return r1.end.isEqual(r2.end)
    }

    
    collapse (direction) {
      let coor;
      if (direction === 'left') {
        coor = this.start;
      } else {
        coor = this.end;
      }
      return _createNewSelection$2(this, coor, coor)
    }

    expand (other) {
      let r1 = this._range(this);
      let r2 = this._range(other);
      let start;
      let end;

      if (r1.start.isEqual(r2.start)) {
        start = new Coordinate$2(this.start.path, Math.min(this.start.offset, other.start.offset));
      } else if (r1.start.isAfter(r2.start)) {
        start = new Coordinate$2(other.start.path, other.start.offset);
      } else {
        start = this.start;
      }
      if (r1.end.isEqual(r2.end)) {
        end = new Coordinate$2(this.end.path, Math.max(this.end.offset, other.end.offset));
      } else if (r1.end.isBefore(r2.end, false)) {
        end = new Coordinate$2(other.end.path, other.end.offset);
      } else {
        end = this.end;
      }

      return _createNewSelection$2(this, start, end)
    }

    truncateWith (other) {
      if (other.isInsideOf(this, 'strict')) {
        
        throw new Error('Can not truncate with a contained selections')
      }
      if (!this.overlaps(other)) {
        return this
      }
      let r1 = this._range(this);
      let r2 = this._range(other);
      let start, end;
      if (r2.start.isBefore(r1.start, 'strict') && r2.end.isBefore(r1.end, 'strict')) {
        start = other.end;
        end = this.end;
      } else if (r1.start.isBefore(r2.start, 'strict') && r1.end.isBefore(r2.end, 'strict')) {
        start = this.start;
        end = other.start;
      } else if (r1.start.isEqual(r2.start)) {
        if (r2.end.isBefore(r1.end, 'strict')) {
          start = other.end;
          end = this.end;
        } else {
          
          return Selection$2.nullSelection
        }
      } else if (r1.end.isEqual(r2.end)) {
        if (r1.start.isBefore(r2.start, 'strict')) {
          start = this.start;
          end = other.start;
        } else {
          
          return Selection$2.nullSelection
        }
      } else if (this.isInsideOf(other)) {
        return Selection$2.nullSelection
      } else {
        throw new Error('Could not determine coordinates for truncate. Check input')
      }
      return _createNewSelection$2(this, start, end)
    }

    
    getNodeIds () {
      const container = this.getContainer();
      const startPos = container.getPosition(this.start.path[0]);
      const endPos = container.getPosition(this.end.path[0]);
      return container.getContent().slice(startPos, endPos + 1)
    }

    
    splitIntoPropertySelections () {
      let sels = [];
      let fragments = this.getFragments();
      fragments.forEach(function (fragment) {
        if (fragment instanceof Selection$2.Fragment) {
          sels.push(
            new PropertySelection$2(fragment.path, fragment.startOffset,
              fragment.endOffset, false, this.containerId, this.surfaceId)
          );
        }
      }.bind(this));
      return sels
    }

    _clone () {
      return new ContainerSelection$2(this)
    }

    _range (sel) {
      
      
      
      if (sel._internal.addressRange) {
        return sel._internal.addressRange
      }

      let container = this.getContainer();
      let startAddress = container.getAddress(sel.start);
      let endAddress;
      if (sel.isCollapsed()) {
        endAddress = startAddress;
      } else {
        endAddress = container.getAddress(sel.end);
      }
      let addressRange = {
        start: startAddress,
        end: endAddress
      };
      if (sel._isContainerSelection) {
        sel._internal.addressRange = addressRange;
      }
      return addressRange
    }

    get path () {
      throw new Error('ContainerSelection has no path property. Use startPath and endPath instead')
    }
  }

  ContainerSelection$2.prototype._isContainerSelection = true;

  ContainerSelection$2.fromJSON = function (properties) {
    let sel = new ContainerSelection$2(properties);
    return sel
  };

  function _createNewSelection$2 (containerSel, start, end) {
    let newSel;

    if (start === end) {
      newSel = new PropertySelection$2({
        path: start.path,
        startOffset: start.offset,
        endOffset: start.offset,
        containerId: containerSel.containerId,
        surfaceId: containerSel.surfaceId
      });
    } else {
      newSel = new ContainerSelection$2(
        containerSel.containerId,
        start.path, start.offset, end.path, end.offset,
        false, containerSel.surfaceId
      );
    }
    
    const doc = containerSel._internal.doc;
    if (doc) {
      newSel.attach(doc);
    }
    return newSel
  }

  class NodeSelection$2 extends Selection$2 {
    constructor (containerId, nodeId, mode, reverse, surfaceId) {
      super();

      if (arguments.length === 1) {
        let data = arguments[0];
        containerId = data.containerId;
        nodeId = data.nodeId;
        mode = data.mode;
        reverse = data.reverse;
        surfaceId = data.surfaceId;
      }

      if (!isString$2(containerId)) {
        throw new Error("'containerId' is mandatory.")
      }
      if (!isString$2(nodeId)) {
        throw new Error("'nodeId' is mandatory.")
      }
      mode = mode || 'full';

      this.containerId = containerId;
      this.nodeId = nodeId;
      this.mode = mode;
      this.reverse = Boolean(reverse);
      this.surfaceId = surfaceId;

      this.start = new Coordinate$2([nodeId], 0);
      this.end = new Coordinate$2([nodeId], 1);
    }

    equals (other) {
      return (
        super.equals(other) &&
        this.nodeId === other.nodeId &&
        this.mode === other.mode
      )
    }

    isNodeSelection () {
      return true
    }

    getType () {
      return 'node'
    }

    getNodeId () {
      return this.nodeId
    }

    isFull () {
      return this.mode === 'full'
    }

    isBefore () {
      return this.mode === 'before'
    }

    isAfter () {
      return this.mode === 'after'
    }

    isCollapsed () {
      return this.mode !== 'full'
    }

    toJSON () {
      return {
        type: 'node',
        nodeId: this.nodeId,
        mode: this.mode,
        reverse: this.reverse,
        containerId: this.containerId,
        surfaceId: this.surfaceId
      }
    }

    toString () {
      
      return [
        'NodeSelection(',
        this.containerId, '.', this.nodeId, ', ',
        this.mode, ', ',
        (this.reverse ? ', reverse' : ''),
        (this.surfaceId ? (', ' + this.surfaceId) : ''),
        ')'
      ].join('')
    }

    collapse (direction) {
      if (direction === 'left') {
        if (this.isBefore()) {
          return this
        } else {
          return new NodeSelection$2(this.containerId, this.nodeId, 'before', this.reverse, this.surfaceId)
        }
      } else if (direction === 'right') {
        if (this.isAfter()) {
          return this
        } else {
          return new NodeSelection$2(this.containerId, this.nodeId, 'after', this.reverse, this.surfaceId)
        }
      } else {
        throw new Error("'direction' must be either 'left' or 'right'")
      }
    }

    _getCoordinate () {
      if (this.mode === 'before') {
        return new Coordinate$2([this.nodeId], 0)
      } else if (this.mode === 'after') {
        return new Coordinate$2([this.nodeId], 1)
      }
    }

    _clone () {
      return new NodeSelection$2(this)
    }
  }

  NodeSelection$2.prototype._isNodeSelection = true;

  NodeSelection$2.fromJSON = function (json) {
    return new NodeSelection$2(json)
  };


  NodeSelection$2._createFromCoordinate = function (coor) {
    var containerId = coor.containerId;
    var nodeId = coor.getNodeId();
    var mode = coor.offset === 0 ? 'before' : 'after';
    return new NodeSelection$2(containerId, nodeId, mode, false)
  };

  class CustomSelection$2 extends Selection$2 {
    constructor (customType, data, surfaceId) {
      super();

      if (arguments.length === 1) {
        let _data = arguments[0];
        customType = _data.customType;
        data = _data.data;
        surfaceId = _data.surfaceId;
      }

      this.customType = customType;
      this.data = data || {};
      this.surfaceId = surfaceId;
    }

    isCustomSelection () {
      return true
    }

    getType () {
      return 'custom'
    }

    getCustomType () {
      return this.customType
    }

    toJSON () {
      return {
        type: 'custom',
        customType: this.customType,
        data: cloneDeep$2(this.data),
        surfaceId: this.surfaceId
      }
    }

    toString () {
      
      return [
        'CustomSelection(',
        this.customType, ', ',
        JSON.stringify(this.data),
        ')'
      ].join('')
    }

    equals (other) {
      return (
        Selection$2.prototype.equals.call(this, other) &&
        other.isCustomSelection() &&
        isEqual$2(this.data, other.data)
      )
    }

    _clone () {
      return new CustomSelection$2(this)
    }
  }

  CustomSelection$2.prototype._isCustomSelection = true;

  CustomSelection$2.fromJSON = function (json) {
    return new CustomSelection$2(json)
  };

  function fromJSON$2 (json) {
    if (!json) return Selection$2.nullSelection
    var type = json.type;
    switch (type) {
      case 'property':
        return PropertySelection$2.fromJSON(json)
      case 'container':
        return ContainerSelection$2.fromJSON(json)
      case 'node':
        return NodeSelection$2.fromJSON(json)
      case 'custom':
        return CustomSelection$2.fromJSON(json)
      default:
        
        return Selection$2.nullSelection
    }
  }


  function isFirst$2 (doc, coor) {
    if (coor.isNodeCoordinate() && coor.offset === 0) return true
    let node = doc.get(coor.path[0]).getContainerRoot();
    if (node.isText() && coor.offset === 0) return true
    if (node.isList()) {
      let itemId = coor.path[0];
      if (node.items[0] === itemId && coor.offset === 0) return true
    }
  }


  function isLast$2 (doc, coor) {
    if (coor.isNodeCoordinate() && coor.offset > 0) return true
    let node = doc.get(coor.path[0]).getContainerRoot();
    if (node.isText() && coor.offset >= node.getLength()) return true
    if (node.isList()) {
      let itemId = coor.path[0];
      let item = doc.get(itemId);
      if (last$1$2(node.items) === itemId && coor.offset === item.getLength()) return true
    }
  }

  function isEntirelySelected$2 (doc, node, start, end) {
    let { isEntirelySelected } = getRangeInfo$2(doc, node, start, end);
    return isEntirelySelected
  }

  function getRangeInfo$2 (doc, node, start, end) {
    let isFirst = true;
    let isLast = true;
    if (node.isText()) {
      if (start && start.offset !== 0) isFirst = false;
      if (end && end.offset < node.getLength()) isLast = false;
    } else if (node.isList()) {
      if (start) {
        let itemId = start.path[0];
        let itemPos = node.getItemPosition(itemId);
        if (itemPos > 0 || start.offset !== 0) isFirst = false;
      }
      if (end) {
        let itemId = end.path[0];
        let itemPos = node.getItemPosition(itemId);
        let item = doc.get(itemId);
        if (itemPos < node.items.length - 1 || end.offset < item.getLength()) isLast = false;
      }
    }
    let isEntirelySelected = isFirst && isLast;
    return {isFirst, isLast, isEntirelySelected}
  }

  function setCursor$2 (tx, node, containerId, mode) {
    if (node.isText()) {
      let offset = 0;
      if (mode === 'after') {
        let text = node.getText();
        offset = text.length;
      }
      tx.setSelection({
        type: 'property',
        path: node.getPath(),
        startOffset: offset,
        containerId: containerId
      });
    } else if (node.isList()) {
      let item, offset;
      if (mode === 'after') {
        item = node.getLastItem();
        offset = item.getLength();
      } else {
        item = node.getFirstItem();
        offset = 0;
      }
      tx.setSelection({
        type: 'property',
        path: item.getPath(),
        startOffset: offset,
        containerId: containerId
      });
    } else {
      tx.setSelection({
        type: 'node',
        containerId: containerId,
        nodeId: node.id
        
        
        
      });
    }
  }

  function selectNode$2 (tx, nodeId, containerId) {
    tx.setSelection(createNodeSelection$2({ doc: tx, nodeId, containerId }));
  }

  function createNodeSelection$2 ({ doc, nodeId, containerId, mode, reverse, surfaceId }) {
    let node = doc.get(nodeId);
    if (!node) return Selection$2.nullSelection
    node = node.getContainerRoot();
    if (node.isText()) {
      return new PropertySelection$2({
        path: node.getPath(),
        startOffset: mode === 'after' ? node.getLength() : 0,
        endOffset: mode === 'before' ? 0 : node.getLength(),
        reverse: reverse,
        containerId: containerId,
        surfaceId: surfaceId
      })
    } else if (node.isList() && node.getLength() > 0) {
      let first = node.getFirstItem();
      let last = node.getLastItem();
      let start = {
        path: first.getPath(),
        offset: 0
      };
      let end = {
        path: last.getPath(),
        offset: last.getLength()
      };
      if (mode === 'after') start = end;
      else if (mode === 'before') end = start;
      return new ContainerSelection$2({
        startPath: start.path,
        startOffset: start.offset,
        endPath: end.path,
        endOffset: end.offset,
        reverse: reverse,
        containerId: containerId,
        surfaceId: surfaceId
      })
    } else {
      return new NodeSelection$2({ nodeId, mode, reverse, containerId, surfaceId })
    }
  }

  function stepIntoIsolatedNode$2 (editorSession, comp) {
    
    
    if (comp.grabFocus()) return true

    
    let surface = comp.find('.sc-surface');
    if (surface) {
      
      if (surface._isTextPropertyEditor) {
        const doc = editorSession.getDocument();
        const path = surface.getPath();
        const text = doc.get(path, 'strict');
        editorSession.setSelection({
          type: 'property',
          path: path,
          startOffset: text.length,
          surfaceId: surface.id
        });
        return true
      } else if (surface._isContainerEditor) {
        let container = surface.getContainer();
        if (container.length > 0) {
          let first = container.getChildAt(0);
          setCursor$2(editorSession, first, container.id, 'after');
        }
        return true
      }
    }
    return false
  }

  function augmentSelection$2 (selData, oldSel) {
    
    if (selData && oldSel && !selData.surfaceId && !oldSel.isNull()) {
      selData.containerId = selData.containerId || oldSel.containerId;
      selData.surfaceId = selData.surfaceId || oldSel.surfaceId;
    }
    return selData
  }

  var selectionHelpers$2 = /*#__PURE__*/Object.freeze({
    fromJSON: fromJSON$2,
    isFirst: isFirst$2,
    isLast: isLast$2,
    isEntirelySelected: isEntirelySelected$2,
    getRangeInfo: getRangeInfo$2,
    setCursor: setCursor$2,
    selectNode: selectNode$2,
    createNodeSelection: createNodeSelection$2,
    stepIntoIsolatedNode: stepIntoIsolatedNode$2,
    augmentSelection: augmentSelection$2
  });

  class DocumentChange$2 {
    constructor (ops, before, after) {
      if (arguments.length === 1 && isPlainObject$3(arguments[0])) {
        let data = arguments[0];
        
        this.sha = data.sha;
        
        this.timestamp = data.timestamp;
        
        this.before = data.before || {};
        
        this.ops = data.ops;
        this.info = data.info; 
        
        this.after = data.after || {};
      } else if (arguments.length === 3) {
        this.sha = uuid$2();
        this.info = {};
        this.timestamp = Date.now();
        this.ops = ops.slice(0);
        this.before = before || {};
        this.after = after || {};
      } else {
        throw new Error('Illegal arguments.')
      }
      
      this.updated = null;
      
      this.created = null;
      
      this.deleted = null;
    }

    
    _extractInformation (doc) {
      
      
      if (this._extracted) return

      let ops = this.ops;
      let created = {};
      let deleted = {};
      let updated = {};
      let affectedContainerAnnos = [];

      
      function _checkAnnotation (op) {
        switch (op.type) {
          case 'create':
          case 'delete': {
            let node = op.val;
            if (node.hasOwnProperty('start')) {
              updated[node.start.path] = true;
            }
            if (node.hasOwnProperty('end')) {
              updated[node.end.path] = true;
            }
            break
          }
          case 'update':
          case 'set': {
            
            let node = doc.get(op.path[0]);
            if (node) {
              if (node.isPropertyAnnotation()) {
                updated[node.start.path] = true;
              } else if (node.isContainerAnnotation()) {
                affectedContainerAnnos.push(node);
              }
            }
            break
          }
          default:
            
            throw new Error('Illegal state')
        }
      }

      for (let i = 0; i < ops.length; i++) {
        let op = ops[i];
        if (op.type === 'create') {
          created[op.val.id] = op.val;
          delete deleted[op.val.id];
        }
        if (op.type === 'delete') {
          delete created[op.val.id];
          deleted[op.val.id] = op.val;
        }
        if (op.type === 'set' || op.type === 'update') {
          updated[op.path] = true;
          
          updated[op.path[0]] = true;
        }
        _checkAnnotation(op);
      }

      affectedContainerAnnos.forEach(function (anno) {
        let container = doc.get(anno.containerId, 'strict');
        let startPos = container.getPosition(anno.start.path[0]);
        let endPos = container.getPosition(anno.end.path[0]);
        for (let pos = startPos; pos <= endPos; pos++) {
          let node = container.getChildAt(pos);
          let path;
          if (node.isText()) {
            path = node.getPath();
          } else {
            path = [node.id];
          }
          if (!deleted[node.id]) {
            updated[path] = true;
          }
        }
      });

      
      if (Object.keys(deleted).length > 0) {
        forEach$2(updated, function (_, key) {
          let nodeId = key.split(',')[0];
          if (deleted[nodeId]) {
            delete updated[key];
          }
        });
      }

      this.created = created;
      this.deleted = deleted;
      this.updated = updated;

      this._extracted = true;
    }

    invert () {
      
      let copy = this.toJSON();
      copy.ops = [];
      
      let tmp = copy.before;
      copy.before = copy.after;
      copy.after = tmp;
      let inverted = DocumentChange$2.fromJSON(copy);
      let ops = [];
      for (let i = this.ops.length - 1; i >= 0; i--) {
        ops.push(this.ops[i].invert());
      }
      inverted.ops = ops;
      return inverted
    }

    
    isAffected (path) {
      console.error('DEPRECATED: use change.hasUpdated() instead');
      return this.hasUpdated(path)
    }

    isUpdated (path) {
      console.error('DEPRECATED: use change.hasUpdated() instead');
      return this.hasUpdated(path)
    }
    

    hasUpdated (path) {
      return this.updated[path]
    }

    hasDeleted (id) {
      return this.deleted[id]
    }

    serialize () {
      
      

      let opSerializer = new OperationSerializer$2();
      let data = this.toJSON();
      data.ops = this.ops.map(function (op) {
        return opSerializer.serialize(op)
      });
      return JSON.stringify(data)
    }

    clone () {
      return DocumentChange$2.fromJSON(this.toJSON())
    }

    toJSON () {
      let data = {
        
        sha: this.sha,
        
        before: clone$2(this.before),
        ops: map$3(this.ops, function (op) {
          return op.toJSON()
        }),
        info: this.info,
        
        after: clone$2(this.after)
      };

      
      
      data.after.selection = undefined;
      data.before.selection = undefined;

      let sel = this.before.selection;
      if (sel && sel._isSelection) {
        data.before.selection = sel.toJSON();
      }
      sel = this.after.selection;
      if (sel && sel._isSelection) {
        data.after.selection = sel.toJSON();
      }
      return data
    }
  }

  DocumentChange$2.deserialize = function (str) {
    let opSerializer = new OperationSerializer$2();
    let data = JSON.parse(str);
    data.ops = data.ops.map(function (opData) {
      return opSerializer.deserialize(opData)
    });
    if (data.before.selection) {
      data.before.selection = fromJSON$2(data.before.selection);
    }
    if (data.after.selection) {
      data.after.selection = fromJSON$2(data.after.selection);
    }
    return new DocumentChange$2(data)
  };

  DocumentChange$2.fromJSON = function (data) {
    
    let change = cloneDeep$2(data);
    change.ops = data.ops.map(function (opData) {
      return ObjectOperation$2.fromJSON(opData)
    });
    change.before.selection = fromJSON$2(data.before.selection);
    change.after.selection = fromJSON$2(data.after.selection);
    return new DocumentChange$2(change)
  };

  function transformDocumentChange$2 (A, B) {
    _transformInplaceBatch$2(A, B);
  }

  function transformSelection$2 (sel, a) {
    let newSel = sel.clone();
    let hasChanged = _transformSelectionInplace$2(newSel, a);
    if (hasChanged) {
      return newSel
    } else {
      return sel
    }
  }

  function _transformInplaceSingle$2 (a, b) {
    for (let i = 0; i < a.ops.length; i++) {
      let opA = a.ops[i];
      for (let j = 0; j < b.ops.length; j++) {
        let opB = b.ops[j];
        
        
        ObjectOperation$2.transform(opA, opB, {inplace: true});
      }
    }
    if (a.before) {
      _transformSelectionInplace$2(a.before.selection, b);
    }
    if (a.after) {
      _transformSelectionInplace$2(a.after.selection, b);
    }
    if (b.before) {
      _transformSelectionInplace$2(b.before.selection, a);
    }
    if (b.after) {
      _transformSelectionInplace$2(b.after.selection, a);
    }
  }

  function _transformInplaceBatch$2 (A, B) {
    if (!isArray$3(A)) {
      A = [A];
    }
    if (!isArray$3(B)) {
      B = [B];
    }
    for (let i = 0; i < A.length; i++) {
      let a = A[i];
      for (let j = 0; j < B.length; j++) {
        let b = B[j];
        _transformInplaceSingle$2(a, b);
      }
    }
  }

  function _transformSelectionInplace$2 (sel, a) {
    if (!sel || (!sel.isPropertySelection() && !sel.isContainerSelection())) {
      return false
    }
    let ops = a.ops;
    let hasChanged = false;
    let isCollapsed = sel.isCollapsed();
    for (let i = 0; i < ops.length; i++) {
      let op = ops[i];
      hasChanged |= _transformCoordinateInplace$2(sel.start, op);
      if (!isCollapsed) {
        hasChanged |= _transformCoordinateInplace$2(sel.end, op);
      } else {
        if (sel.isContainerSelection()) {
          sel.end.path = sel.start.path;
        }
        sel.end.offset = sel.start.offset;
      }
    }
    return hasChanged
  }

  function _transformCoordinateInplace$2 (coor, op) {
    if (!isEqual$2(op.path, coor.path)) return false
    let hasChanged = false;
    if (op.type === 'update' && op.propertyType === 'string') {
      let diff = op.diff;
      let newOffset;
      if (diff.isInsert() && diff.pos <= coor.offset) {
        newOffset = coor.offset + diff.str.length;
        
        coor.offset = newOffset;
        hasChanged = true;
      } else if (diff.isDelete() && diff.pos <= coor.offset) {
        newOffset = Math.max(diff.pos, coor.offset - diff.str.length);
        
        coor.offset = newOffset;
        hasChanged = true;
      }
    }
    return hasChanged
  }

  var operationHelpers$2 = /*#__PURE__*/Object.freeze({
    transformDocumentChange: transformDocumentChange$2,
    transformSelection: transformSelection$2
  });

  function isFunction$1$2 (f) {
    return typeof f === 'function'
  }

  function deleteFromArray$2 (array, value) {
    if (!array) return
    for (var i = 0; i < array.length; i++) {
      if (array[i] === value) {
        array.splice(i, 1);
        i--;
      }
    }
  }

  class TreeNode$2 {}



  class TreeIndex$2 {
    
    get (path) {
      if (arguments.length > 1) {
        path = Array.prototype.slice(arguments, 0);
      }
      if (isString$2(path)) {
        path = [path];
      }
      return get$2(this, path)
    }

    getAll (path) {
      if (arguments.length > 1) {
        path = Array.prototype.slice(arguments, 0);
      }
      if (isString$2(path)) {
        path = [path];
      }
      if (!isArray$3(path)) {
        throw new Error('Illegal argument for TreeIndex.get()')
      }
      var node = get$2(this, path);
      return this._collectValues(node)
    }

    set (path, value) {
      if (isString$2(path)) {
        path = [path];
      }
      setWith$2(this, path, value, function (val) {
        if (!val) return new TreeNode$2()
      });
    }

    delete (path) {
      if (isString$2(path)) {
        delete this[path];
      } else if (path.length === 1) {
        delete this[path[0]];
      } else {
        var key = path[path.length - 1];
        path = path.slice(0, -1);
        var parent = get$2(this, path);
        if (parent) {
          delete parent[key];
        }
      }
    }

    clear () {
      var root = this;
      for (var key in root) {
        if (root.hasOwnProperty(key)) {
          delete root[key];
        }
      }
    }

    traverse (fn) {
      this._traverse(this, [], fn);
    }

    forEach (...args) {
      this.traverse(...args);
    }

    _traverse (root, path, fn) {
      var id;
      for (id in root) {
        if (!root.hasOwnProperty(id)) continue
        var child = root[id];
        var childPath = path.concat([id]);
        if (child instanceof TreeNode$2) {
          this._traverse(child, childPath, fn);
        } else {
          fn(child, childPath);
        }
      }
    }

    _collectValues (root) {
      
      
      var vals = {};
      this._traverse(root, [], function (val, path) {
        var key = path[path.length - 1];
        vals[key] = val;
      });
      return vals
    }
  }

  class TreeIndexArrays$2 extends TreeIndex$2 {
    contains (path) {
      let val = super.get(path);
      return Boolean(val)
    }

    get (path) {
      let val = super.get(path);
      if (val instanceof TreeNode$2) {
        val = val.__values__ || [];
      }
      return val
    }

    set (path, arr) {
      let val = super.get(path);
      val.__values__ = arr;
    }

    add (path, value) {
      if (isString$2(path)) {
        path = [path];
      }
      if (!isArray$3(path)) {
        throw new Error('Illegal arguments.')
      }
      var arr;

      
      
      
      
      
      
      setWith$2(this, path.concat(['__values__', '__dummy__']), undefined, function (val, key) {
        if (key === '__values__') {
          if (!val) val = [];
          arr = val;
        } else if (!val) {
          val = new TreeNode$2();
        }
        return val
      });
      delete arr.__dummy__;
      arr.push(value);
    }

    remove (path, value) {
      var arr = get$2(this, path);
      if (arr instanceof TreeNode$2) {
        if (arguments.length === 1) {
          delete arr.__values__;
        } else {
          deleteFromArray$2(arr.__values__, value);
        }
      }
    }

    _collectValues (root) {
      var vals = [];
      this._traverse(root, [], function (val) {
        vals.push(val);
      });
      vals = Array.prototype.concat.apply([], vals);
      return vals
    }
  }

  TreeIndex$2.Arrays = TreeIndexArrays$2;

  function filter$2 (iteratee, fn) {
    if (!iteratee) return []
    if (iteratee.constructor.prototype.filter && isFunction$1$2(iteratee.constructor.prototype.filter)) {
      return iteratee.filter(fn)
    }
    let result = [];
    forEach$2(iteratee, (val, key) => {
      if (fn(val, key)) {
        result.push(val);
      }
    });
    return result
  }

  function flatten$2 (arr) {
    return Array.prototype.concat.apply([], arr)
  }

  function flattenOften$2 (arr, max) {
    if (!(max > 0)) throw new Error("'max' must be a positive number")
    let l = arr.length;
    arr = flatten$2(arr);
    let round = 1;
    while (round < max && l < arr.length) {
      l = arr.length;
      arr = flatten$2(arr);
      round++;
    }
    return arr
  }

  class NodeIndex$2 {
    
    select (node) { 
      throw new Error('This method is abstract.')
    }

    clear () {
      throw new Error('This method is abstract')
    }

    
    create (node) { 
      throw new Error('This method is abstract.')
    }

    
    delete (node) { 
      throw new Error('This method is abstract.')
    }

    set (node, path, newValue, oldValue) {
      this.update(node, path, newValue, oldValue);
    }

    
    update (node, path, newValue, oldValue) { 
      throw new Error('This method is abstract.')
    }

    
    reset (data) {
      this.clear();
      this._initialize(data);
    }

    
    clone () {
      var NodeIndexClass = this.constructor;
      var clone = new NodeIndexClass();
      return clone
    }

    _initialize (data) {
      forEach$2(data.getNodes(), function (node) {
        if (this.select(node)) {
          this.create(node);
        }
      }.bind(this));
    }
  }


  NodeIndex$2.create = function (prototype) {
    var index = Object.assign(new NodeIndex$2(), prototype);
    index.clone = function () {
      return NodeIndex$2.create(prototype)
    };
    return index
  };


  NodeIndex$2.filterByType = function (type) {
    return function (node) {
      return node.isInstanceOf(type)
    }
  };

  class DocumentIndex$2 extends NodeIndex$2 {}

  var annotationHelpers$2 = {
    insertedText: insertedText$2,
    deletedText: deletedText$2,
    transferAnnotations: transferAnnotations$2,
    expandAnnotation: expandAnnotation$2,
    fuseAnnotation: fuseAnnotation$2,
    truncateAnnotation: truncateAnnotation$2
  };

  function insertedText$2 (doc, coordinate, length) {
    if (!length) return
    var index = doc.getIndex('annotations');
    var annotations = index.get(coordinate.path);
    for (let i = 0; i < annotations.length; i++) {
      let anno = annotations[i];
      var pos = coordinate.offset;
      var start = anno.start.offset;
      var end = anno.end.offset;
      var newStart = start;
      var newEnd = end;
      if ((pos < start) ||
           (pos === start)) {
        newStart += length;
      }
      
      if ((pos < end) ||
           (pos === end && !anno.isInline())) {
        newEnd += length;
      }
      
      if (newStart !== start) {
        doc.set([anno.id, 'start', 'offset'], newStart);
      }
      if (newEnd !== end) {
        doc.set([anno.id, 'end', 'offset'], newEnd);
      }
    }

    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
  }

  function deletedText$2 (doc, path, startOffset, endOffset) {
    if (startOffset === endOffset) return
    var index = doc.getIndex('annotations');
    var annotations = index.get(path);
    var length = endOffset - startOffset;
    for (let i = 0; i < annotations.length; i++) {
      let anno = annotations[i];
      var pos1 = startOffset;
      var pos2 = endOffset;
      var start = anno.start.offset;
      var end = anno.end.offset;
      var newStart = start;
      var newEnd = end;
      if (pos2 <= start) {
        newStart -= length;
        newEnd -= length;
        doc.set([anno.id, 'start', 'offset'], newStart);
        doc.set([anno.id, 'end', 'offset'], newEnd);
      } else {
        if (pos1 <= start) {
          newStart = start - Math.min(pos2 - pos1, start - pos1);
        }
        if (pos1 <= end) {
          newEnd = end - Math.min(pos2 - pos1, end - pos1);
        }
        
        if (start !== end && newStart === newEnd) {
          doc.delete(anno.id);
        } else {
          
          if (start !== newStart) {
            doc.set([anno.id, 'start', 'offset'], newStart);
          }
          if (end !== newEnd) {
            doc.set([anno.id, 'end', 'offset'], newEnd);
          }
        }
      }
    }
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
  }


  function transferAnnotations$2 (doc, path, offset, newPath, newOffset) {
    var index = doc.getIndex('annotations');
    var annotations = index.get(path, offset);
    for (let i = 0; i < annotations.length; i++) {
      let a = annotations[i];
      var isInside = (offset > a.start.offset && offset < a.end.offset);
      var start = a.start.offset;
      var end = a.end.offset;
      
      if (isInside) {
        
        if (a.canSplit()) {
          let newAnno = a.toJSON();
          newAnno.id = uuid$2(a.type + '_');
          newAnno.start.path = newPath;
          newAnno.start.offset = newOffset;
          newAnno.end.path = newPath;
          newAnno.end.offset = newOffset + a.end.offset - offset;
          doc.create(newAnno);
        }
        
        let newStartOffset = a.start.offset;
        let newEndOffset = offset;
        
        if (newEndOffset === newStartOffset) {
          doc.delete(a.id);
        
        } else {
          
          if (newStartOffset !== start) {
            doc.set([a.id, 'start', 'offset'], newStartOffset);
          }
          if (newEndOffset !== end) {
            doc.set([a.id, 'end', 'offset'], newEndOffset);
          }
        }
      
      } else if (a.start.offset >= offset) {
        
        
        
        doc.set([a.id, 'start', 'path'], newPath);
        doc.set([a.id, 'start', 'offset'], newOffset + a.start.offset - offset);
        doc.set([a.id, 'end', 'path'], newPath);
        doc.set([a.id, 'end', 'offset'], newOffset + a.end.offset - offset);
      }
    }

    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
  }


  function truncateAnnotation$2 (tx, anno, sel) {
    if (!sel || !sel._isSelection) throw new Error('Argument "selection" is required.')
    if (!anno || !anno.isAnnotation()) throw new Error('Argument "anno" is required and must be an annotation.')
    let annoSel = anno.getSelection();
    let newAnnoSel = annoSel.truncateWith(sel);
    anno._updateRange(tx, newAnnoSel);
    return anno
  }


  function expandAnnotation$2 (tx, anno, sel) {
    if (!sel || !sel._isSelection) throw new Error('Argument "selection" is required.')
    if (!anno || !anno.isAnnotation()) throw new Error('Argument "anno" is required and must be an annotation.')
    let annoSel = anno.getSelection();
    let newAnnoSel = annoSel.expand(sel);
    anno._updateRange(tx, newAnnoSel);
    return anno
  }


  function fuseAnnotation$2 (tx, annos) {
    if (!isArray$3(annos) || annos.length < 2) {
      throw new Error('fuseAnnotation(): at least two annotations are necessary.')
    }
    let sel, annoType;
    annos.forEach(function (anno, idx) {
      if (idx === 0) {
        sel = anno.getSelection();
        annoType = anno.type;
      } else {
        if (anno.type !== annoType) {
          throw new Error('fuseAnnotation(): all annotations must be of the same type.')
        }
        sel = sel.expand(anno.getSelection());
      }
    });
    
    for (var i = 1; i < annos.length; i++) {
      tx.delete(annos[i].id);
    }
    expandAnnotation$2(tx, annos[0], sel);
    tx.setSelection(sel);
  }

  function getPropertyAnnotationsForSelection$2 (doc, sel, options) {
    options = options || {};
    if (!sel.isPropertySelection()) {
      return []
    }
    let path = sel.getPath();
    let annotations = doc.getIndex('annotations').get(path, sel.start.offset, sel.end.offset);
    if (options.type) {
      annotations = filter$2(annotations, DocumentIndex$2.filterByType(options.type));
    }
    return annotations
  }


  function getContainerAnnotationsForSelection$2 (doc, sel, containerId, options) {
    
    
    
    
    
    if (!containerId) {
      throw new Error("'containerId' is required.")
    }
    options = options || {};
    let index = doc.getIndex('container-annotations');
    let annotations = [];
    if (index) {
      annotations = index.get(containerId, options.type);
      annotations = filter$2(annotations, function (anno) {
        return sel.overlaps(anno.getSelection())
      });
    }
    return annotations
  }


  function isContainerAnnotation$2 (doc, type) {
    let schema = doc.getSchema();
    return schema.isInstanceOf(type, '@container-annotation')
  }


  function getTextForSelection$2 (doc, sel) {
    if (!sel || sel.isNull()) {
      return ''
    } else if (sel.isPropertySelection()) {
      let text = doc.get(sel.start.path);
      return text.substring(sel.start.offset, sel.end.offset)
    } else if (sel.isContainerSelection()) {
      let result = [];
      let nodeIds = sel.getNodeIds();
      let L = nodeIds.length;
      for (let i = 0; i < L; i++) {
        let id = nodeIds[i];
        let node = doc.get(id);
        if (node.isText()) {
          let text = node.getText();
          if (i === L - 1) {
            text = text.slice(0, sel.end.offset);
          }
          if (i === 0) {
            text = text.slice(sel.start.offset);
          }
          result.push(text);
        }
      }
      return result.join('\n')
    }
  }

  function getMarkersForSelection$2 (doc, sel) {
    
    if (!sel || !sel.isPropertySelection()) return []
    const path = sel.getPath();
    
    let markers = doc.getIndex('markers').get(path);
    const filtered = filter$2(markers, function (m) {
      return m.containsSelection(sel)
    });
    return filtered
  }


  function deleteNode$2 (doc, node) {
    
    if (!node) {
      console.warn('Invalid arguments');
      return
    }
    
    if (node.isText()) {
      
      let annos = doc.getIndex('annotations').get(node.id);
      for (let i = 0; i < annos.length; i++) {
        doc.delete(annos[i].id);
      }
    }
    let nodeSchema = node.getSchema();
    
    
    
    doc.delete(node.id);
    
    let ownedProps = nodeSchema.getOwnedProperties();
    ownedProps.forEach(prop => {
      let value = node[prop.name];
      if (prop.isArray()) {
        let ids = value;
        if (ids.length > 0) {
          
          if (isArray$3(ids[0])) ids = flattenOften$2(ids, 2);
          ids.forEach((id) => {
            deleteNode$2(doc, doc.get(id));
          });
        }
      } else {
        deleteNode$2(doc, doc.get(value));
      }
    });
  }


  function copyNode$2 (node) {
    let nodes = [];
    
    let doc = node.getDocument();
    let nodeSchema = node.getSchema();
    for (let prop of nodeSchema) {
      
      
      if ((prop.isReference() && prop.isOwned()) || (prop.type === 'file')) {
        let val = node[prop.name];
        nodes.push(_copyChildren(val));
      }
    }
    nodes.push(node.toJSON());
    let annotationIndex = node.getDocument().getIndex('annotations');
    let annotations = annotationIndex.get([node.id]);
    forEach$2(annotations, function (anno) {
      nodes.push(anno.toJSON());
    });
    let result = flatten$2(nodes).filter(Boolean);
    
    return result

    function _copyChildren (val) {
      if (!val) return null
      if (isArray$3(val)) {
        return flatten$2(val.map(_copyChildren))
      } else {
        let id = val;
        if (!id) return null
        let child = doc.get(id);
        if (!child) return
        return copyNode$2(child)
      }
    }
  }


  function deleteTextRange$2 (doc, start, end) {
    if (!start) {
      start = {
        path: end.path,
        offset: 0
      };
    }
    let path = start.path;
    let text = doc.get(path);
    if (!end) {
      end = {
        path: start.path,
        offset: text.length
      };
    }
    
    if (!isArrayEqual$2(start.path, end.path)) {
      throw new Error('start and end must be on one property')
    }
    let startOffset = start.offset;
    if (startOffset < 0) throw new Error('start offset must be >= 0')
    let endOffset = end.offset;
    if (endOffset > text.length) throw new Error('end offset must be smaller than the text length')

    doc.update(path, { type: 'delete', start: startOffset, end: endOffset });
    
    let annos = doc.getAnnotations(path);
    annos.forEach(function (anno) {
      let annoStart = anno.start.offset;
      let annoEnd = anno.end.offset;
      
      if (annoEnd <= startOffset) ; else if (annoStart >= endOffset) {
        doc.update([anno.id, 'start'], { type: 'shift', value: startOffset - endOffset });
        doc.update([anno.id, 'end'], { type: 'shift', value: startOffset - endOffset });
      
      } else if (annoStart >= startOffset && annoEnd <= endOffset) {
        doc.delete(anno.id);
      
      } else if (annoStart >= startOffset && annoEnd >= endOffset) {
        if (annoStart > startOffset) {
          doc.update([anno.id, 'start'], { type: 'shift', value: startOffset - annoStart });
        }
        doc.update([anno.id, 'end'], { type: 'shift', value: startOffset - endOffset });
      
      } else if (annoStart <= startOffset && annoEnd <= endOffset) {
        doc.update([anno.id, 'end'], { type: 'shift', value: startOffset - annoEnd });
      
      } else if (annoStart < startOffset && annoEnd >= endOffset) {
        doc.update([anno.id, 'end'], { type: 'shift', value: startOffset - endOffset });
      } else {
        console.warn('TODO: handle annotation update case.');
      }
    });
  }

  function deleteListRange$2 (doc, list, start, end) {
    if (doc !== list.getDocument()) {
      list = doc.get(list.id);
    }
    let startItem, endItem;
    if (!start) {
      startItem = list.getItemAt(0);
      start = {
        path: startItem.getPath(),
        offset: 0
      };
    } else {
      startItem = doc.get(start.path[0]);
    }
    if (!end) {
      endItem = list.getLastItem();
      end = {
        path: endItem.getPath(),
        offset: endItem.getLength()
      };
    } else {
      endItem = doc.get(end.path[0]);
    }
    let startPos = list.getItemPosition(startItem);
    let endPos = list.getItemPosition(endItem);
    
    if (startPos === endPos) {
      deleteTextRange$2(doc, start, end);
      return
    }
    
    if (startPos > endPos) {
      [start, end] = [end, start];
      [startPos, endPos] = [endPos, startPos];
      [startItem, endItem] = [endItem, startItem];
    }
    let firstEntirelySelected = isEntirelySelected$2(doc, startItem, start, null);
    let lastEntirelySelected = isEntirelySelected$2(doc, endItem, null, end);

    
    if (lastEntirelySelected) {
      list.removeItemAt(endPos);
      deleteNode$2(doc, endItem);
    } else {
      deleteTextRange$2(doc, null, end);
    }

    
    let items = list.getItems();
    for (let i = endPos - 1; i > startPos; i--) {
      let item = items[i];
      list.removeItemAt(i);
      deleteNode$2(doc, item);
    }

    
    if (firstEntirelySelected) {
      
      
      
      
      deleteTextRange$2(doc, start, null);
    } else {
      deleteTextRange$2(doc, start, null);
    }

    if (!firstEntirelySelected && !lastEntirelySelected) {
      mergeListItems$2(doc, list.id, startPos);
    }
  }

  function mergeListItems$2 (doc, listId, itemPos) {
    
    let list = doc.get(listId);
    let targetItem = list.getItemAt(itemPos);
    let targetPath = targetItem.getPath();
    let targetLength = targetItem.getLength();
    let sourceItem = list.getItemAt(itemPos + 1);
    let sourcePath = sourceItem.getPath();
    
    list.removeItemAt(itemPos + 1);
    
    doc.update(targetPath, { type: 'insert', start: targetLength, text: sourceItem.getText() });
    
    annotationHelpers$2.transferAnnotations(doc, sourcePath, 0, targetPath, targetLength);
    deleteNode$2(doc, sourceItem);
  }

  function getNodes$2 (doc, ids) {
    return ids.map((id) => {
      return doc.get(id, 'strict')
    })
  }



  const SNIPPET_ID$2 = 'snippet';
  const TEXT_SNIPPET_ID$2 = 'text-snippet';

  var documentHelpers$2 = /*#__PURE__*/Object.freeze({
    getPropertyAnnotationsForSelection: getPropertyAnnotationsForSelection$2,
    getContainerAnnotationsForSelection: getContainerAnnotationsForSelection$2,
    isContainerAnnotation: isContainerAnnotation$2,
    getTextForSelection: getTextForSelection$2,
    getMarkersForSelection: getMarkersForSelection$2,
    deleteNode: deleteNode$2,
    copyNode: copyNode$2,
    deleteTextRange: deleteTextRange$2,
    deleteListRange: deleteListRange$2,
    mergeListItems: mergeListItems$2,
    getNodes: getNodes$2,
    SNIPPET_ID: SNIPPET_ID$2,
    TEXT_SNIPPET_ID: TEXT_SNIPPET_ID$2
  });

  const PLAINOBJ$2 = {};


  class Registry$2 {
    constructor (entries, validator) {
      this.entries = {};
      this.names = [];
      this.validator = validator;

      if (entries) {
        forEach$2(entries, function (entry, name) {
          this.add(name, entry);
        }.bind(this));
      }
    }

    
    contains (name) {
      return this.entries.hasOwnProperty(name)
    }

    
    add (name, entry) {
      if (this.validator) {
        this.validator(entry);
      }
      if (PLAINOBJ$2[name]) {
        throw new Error('Illegal key: "' + name + '" is a property of Object which is thus not allowed as a key.')
      }
      if (this.contains(name)) {
        this.remove(name);
      }
      this.entries[name] = entry;
      this.names.push(name);
    }

    
    remove (name) {
      let pos = this.names.indexOf(name);
      if (pos >= 0) {
        this.names.splice(pos, 1);
      }
      delete this.entries[name];
    }

    
    clear () {
      this.names = [];
      this.entries = {};
    }

    
    get (name, strict) {
      let result = this.entries[name];
      if (strict && !result) {
        throw new Error('No entry registered for name ' + name)
      }
      return result
    }

    
    forEach (callback) {
      for (let i = 0; i < this.names.length; i++) {
        let name = this.names[i];
        let _continue = callback(this.entries[name], name);
        if (_continue === false) {
          break
        }
      }
    }

    map (callback) {
      let result = [];
      this.forEach((entry, name) => {
        result.push(callback(entry, name));
      });
      return result
    }

    filter (callback) {
      let result = [];
      this.forEach(function (entry, name) {
        if (callback(entry, name)) {
          result.push(entry);
        }
      });
      return result
    }

    values () {
      return this.filter(() => { return true })
    }
  }

  Registry$2.prototype._isRegistry = true;

  function isBoolean$2 (val) {
    return (val === true || val === false || (val && val.constructor === Boolean))
  }

  class NodeProperty$2 {
    constructor (name, definition) {
      this.name = name;
      this.definition = definition;

      Object.freeze(this);
      Object.freeze(definition);
    }

    isArray () {
      return isArray$3(this.definition.type)
    }

    isReference () {
      if (this.isArray()) {
        return last$1$2(this.definition.type) === 'id'
      } else {
        return this.definition.type === 'id'
      }
    }

    isText () {
      return Boolean(this.definition._isText)
    }

    isOwned () {
      return Boolean(this.definition.owned)
    }

    isOptional () {
      return this.definition.optional || this.hasDefault()
    }

    isNotNull () {
      return Boolean(this.definition.notNull)
    }

    hasDefault () {
      return this.definition.hasOwnProperty('default')
    }

    getDefault () {
      return this.definition.default
    }

    createDefaultValue () {
      if (isArray$3(this.definition.type)) {
        return []
      }
      switch (this.definition.type) {
        case 'object':
          return {}
        case 'number':
          return -1
        case 'coordinate':
          return new Coordinate$2([], 0)
        case 'boolean':
          return false
        case 'id':
          return null
        case 'string':
          return ''
        default:
          return null
      }
    }

    get type () {
      return this.definition.type
    }

    get targetTypes () {
      return this.definition.targetTypes
    }
  }

  class NodeSchema$2 {
    constructor (properties) {
      this.properties = properties;

      
      
      
      
      
      this._ownedPropNames = new Set();
      this._ownedProps = [];
      forEach$2(properties, (prop) => {
        if ((prop.isReference() && prop.isOwned()) || (prop.type === 'file')) {
          this._ownedPropNames.add(prop.name);
          this._ownedProps.push(prop);
        }
      });
    }

    getProperty (name) {
      return this.properties[name]
    }

    hasOwnedProperties () {
      return this._ownedPropNames.size > 0
    }

    getOwnedProperties () {
      return this._ownedProps.slice()
    }

    isOwned (name) {
      return this._ownedPropNames.has(name)
    }

    [Symbol.iterator] () {
      const properties = this.properties;
      let ids = Object.keys(properties);
      let idx = 0;
      return { 
        next () {
          let done = idx > ids.length - 1;
          if (done) {
            return { done }
          } else {
            return {
              value: properties[ids[idx++]]
            }
          }
        }
      }
    }
  }

  class Node$2 extends EventEmitter$2 {
    
    constructor () {
      super();

      
      
      this._initialize.apply(this, arguments);
    }

    _initialize (data) {
      const NodeClass = this.constructor;

      let schema = NodeClass.schema;
      for (let property of schema) {
        let name = property.name;
        
        
        const propIsGiven = (data[name] !== undefined);
        const isOptional = property.isOptional();
        const hasDefault = property.hasDefault();
        if (!isOptional && !propIsGiven) {
          throw new Error('Property ' + name + ' is mandatory for node type ' + this.type)
        }
        if (propIsGiven) {
          this[name] = _checked$2(property, data[name]);
        } else if (hasDefault) {
          this[name] = cloneDeep$2(_checked$2(property, property.getDefault()));
        }
      }
    }

    dispose () {
      this._disposed = true;
    }

    isDisposed () {
      return Boolean(this._disposed)
    }

    
    isInstanceOf (typeName) {
      return Node$2.isInstanceOf(this.constructor, typeName)
    }

    getSchema () {
      return this.constructor.schema
    }

    
    getTypeNames () {
      var typeNames = [];
      var NodeClass = this.constructor;
      while (NodeClass.type !== '@node') {
        typeNames.push(NodeClass.type);
        NodeClass = Object.getPrototypeOf(NodeClass);
      }
      return typeNames
    }

    
    getPropertyType (propertyName) {
      return this.constructor.schema.getProperty(propertyName).type
    }

    
    toJSON () {
      var data = {
        type: this.type
      };
      const schema = this.getSchema();
      for (let prop of schema) {
        let val = this[prop.name];
        if (prop.isOptional() && val === undefined) continue
        if (isArray$3(val) || isObject$3(val)) {
          val = cloneDeep$2(val);
        }
        data[prop.name] = val;
      }
      return data
    }

    get type () {
      return this.constructor.type
    }
  }

  Node$2.prototype._isNode = true;


  Object.defineProperty(Node$2, 'schema', {
    get () { return this._schema },
    set (schema) {
      let NodeClass = this;
      
      
      if (schema.type) {
        NodeClass.type = schema.type;
      }
      
      
      
      NodeClass._schema = compileSchema$2(NodeClass, schema);
    }
  });

  Node$2.define = Node$2.defineSchema = function define (schema) {
    this.schema = schema;
  };

  Node$2.schema = {
    type: '@node',
    id: 'string'
  };


  Node$2.isInstanceOf = function (NodeClass, typeName) {
    var type = NodeClass.type;
    while (type !== '@node') {
      if (type === typeName) return true
      var _super = Object.getPrototypeOf(NodeClass.prototype).constructor;
      if (_super && _super.type) {
        NodeClass = _super;
        type = NodeClass.type;
      } else {
        break
      }
    }
    return false
  };



  function compileSchema$2 (NodeClass, schema) {
    let compiledSchema = _compileSchema$2(schema);
    let schemas = [compiledSchema];
    let clazz = NodeClass;
    while (clazz) {
      var parentProto = Object.getPrototypeOf(clazz.prototype);
      if (!parentProto) break
      clazz = parentProto.constructor;
      if (clazz && clazz._schema) {
        schemas.unshift(clazz._schema.properties);
      }
    }
    schemas.unshift({});
    return new NodeSchema$2(Object.assign.apply(null, schemas))
  }

  function _compileSchema$2 (schema) {
    let compiledSchema = {};
    forEach$2(schema, function (definition, name) {
      
      if (name === 'type') {
        return
      }
      if (isString$2(definition) || isArray$3(definition)) {
        definition = { type: definition };
      }
      definition = _compileDefintion$2(definition);
      definition.name = name;
      compiledSchema[name] = new NodeProperty$2(name, definition);
    });
    return compiledSchema
  }

  const _valueTypes$2 = new Set(['string', 'number', 'boolean', 'object', 'array']);

  function _compileDefintion$2 (definition) {
    let result = definition;
    if (isArray$3(definition.type)) {
      
      
      
      
      let defs = definition.type;
      let lastIdx = defs.length - 1;
      let first = defs[0];
      let last = defs[lastIdx];
      let isCanonical = first === 'array';
      let hasTargetType = last !== 'id' && !_valueTypes$2.has(last);
      definition.targetTypes = definition.type;
      if (isCanonical) {
        definition.type = defs.slice();
        definition.type[lastIdx] = 'id';
        if (hasTargetType) definition.targetTypes = [last];
      } else {
        if (defs.length > 1) {
          defs.forEach(t => {
            if (_valueTypes$2.has(t)) {
              throw new Error('Multi-types must consist of node types.')
            }
          });
          definition.type = [ 'array', 'id' ];
          definition.targetTypes = defs;
        } else {
          if (_valueTypes$2.has(first)) {
            definition.type = [ 'array', first ];
          } else {
            definition.type = [ 'array', 'id' ];
            definition.targetTypes = defs;
          }
        }
      }
    } else if (definition.type === 'text') {
      result = {
        type: 'string',
        default: '',
        _isText: true
      };
    }
    return result
  }

  function _checked$2 (prop, value) {
    let type;
    let name = prop.name;
    if (prop.isArray()) {
      type = 'array';
    } else {
      type = prop.type;
    }
    if (value === null) {
      if (prop.isNotNull()) {
        throw new Error('Value for property ' + name + ' is null.')
      } else {
        return value
      }
    }
    if (value === undefined) {
      throw new Error('Value for property ' + name + ' is undefined.')
    }
    if ((type === 'string' && !isString$2(value)) ||
        (type === 'boolean' && !isBoolean$2(value)) ||
        (type === 'number' && !isNumber$2(value)) ||
        (type === 'array' && !isArray$3(value)) ||
        (type === 'id' && !isString$2(value)) ||
        (type === 'object' && !isObject$3(value))) {
      throw new Error('Illegal value type for property ' + name + ': expected ' + type + ', was ' + (typeof value))
    }
    return value
  }

  class DocumentNode$2 extends Node$2 {
    _initialize (doc, props) {
      this.document = doc;
      super._initialize(props);
    }

    
    getDocument () {
      return this.document
    }

    
    hasParent () {
      return Boolean(this.parent)
    }

    
    getParent () {
      if (isString$2(this.parent)) return this.document.get(this.parent)
      return this.parent
    }

    setParent (parent) {
      if (isString$2(parent)) parent = this.document.get(parent);
      this.parent = parent;
    }

    
    getRoot () {
      let node = this;
      while (node.parent) {
        node = node.parent;
      }
      return node
    }

    getContainerRoot () {
      let node = this;
      while (node.parent) {
        
        if (node.parent.isContainer()) return node
        
        node = node.parent;
      }
      return node
    }

    
    hasChildren () {
      return false
    }

    
    getChildIndex(child) { 
      return -1
    }

    
    getChildAt(idx) { 
      return null
    }

    
    getChildCount () {
      return 0
    }

    
    

    

    
    isBlock () {
      return Boolean(this.constructor.isBlock)
    }

    
    isText () {
      return Boolean(this.constructor.isText)
    }

    isList () {
      return Boolean(this.constructor.isList)
    }

    isListItem () {
      return Boolean(this.constructor.isListItem)
    }

    isContainer () {
      return Boolean(this._isContainer)
    }

    

    isAnnotation () {
      return Boolean(this._isAnnotation)
    }

    isPropertyAnnotation () {
      return Boolean(this._isPropertyAnnotation)
    }

    isContainerAnnotation () {
      return Boolean(this._isContainerAnnotation)
    }

    
    isInline () {
      return Boolean(this.constructor.isInline)
    }
  }

  DocumentNode$2.prototype._isDocumentNode = true;


  DocumentNode$2.isBlock = false;


  DocumentNode$2.isText = false;


  DocumentNode$2.isPropertyAnnotation = false;


  DocumentNode$2.isContainerAnnotation = false;


  DocumentNode$2.isInline = false;

  class ContainerAddress$2 {
    constructor (pos, offset) {
      this.pos = pos;
      this.offset = offset;
    }

    isBefore (other, strict) {
      strict = Boolean(strict);
      if (this.pos < other.pos) {
        return true
      } else if (this.pos > other.pos) {
        return false
      } else if (this.offset < other.offset) {
        return true
      } else if (this.offset > other.offset) {
        return false
      }
      if (strict) {
        return false
      } else {
        return true
      }
    }

    isAfter (other, strict) {
      return other.isBefore(this, strict)
    }

    isEqual (other) {
      return (this.pos === other.pos && this.offset === other.offset)
    }

    toString () {
      return [this.pos, '.', this.offset].join('')
    }
  }

  function ContainerMixin$2 (DocumentNode) {
    class AbstractContainer extends DocumentNode {
      contains (nodeId) {
        return this.getPosition(nodeId) >= 0
      }

      getPosition (node, strict) {
        if (isString$2(node)) {
          node = this.document.get(node);
        }
        if (!node) return -1
        let pos = this._getPosition(node);
        if (strict && pos < 0) {
          throw new Error('Node is not within this container: ' + node.id)
        }
        return pos
      }

      getNodeAt (idx) {
        const nodeId = this.getNodeIdAt(idx);
        if (nodeId) {
          return this.getDocument().get(nodeId)
        }
      }

      getNodeIdAt (idx) {
        let content = this.getContent();
        if (idx < 0 || idx >= content.length) {
          
          return undefined
        } else {
          return content[idx]
        }
      }

      getNodes () {
        const doc = this.getDocument();
        return this.getContent().map(id => doc.get(id)).filter(Boolean)
      }

      show (nodeId, pos) {
        
        const arg1 = arguments[0];
        if (!isString$2(arg1)) {
          if (arg1._isNode) {
            nodeId = arg1.id;
          }
        }
        if (arguments.length > 1) {
          console.error('DEPRECATED: use container.showAt(pos, nodeId) instead');
        } else {
          pos = this.getLength();
        }
        return this.showAt(pos, nodeId)
      }

      showAt (pos, nodeId) {
        const doc = this.getDocument();
        const length = this.getLength();
        if (!isNumber$2(pos) || pos < 0 || pos > length) {
          throw new Error('Index out of bounds')
        }
        if (!isString$2(nodeId)) {
          if (nodeId._isNode) {
            nodeId = nodeId.id;
          } else {
            throw new Error('Invalid argument.')
          }
        }
        doc.update(this.getContentPath(), { type: 'insert', pos: pos, value: nodeId });
      }

      hide (nodeId) {
        const pos = this.getPosition(nodeId);
        this.hideAt(pos);
      }

      hideAt (pos) {
        const length = this.getLength();
        if (pos >= 0 && pos < length) {
          const doc = this.getDocument();
          doc.update(this.getContentPath(), { type: 'delete', pos: pos });
        } else {
          throw new Error('Index out of bounds.')
        }
      }

      getAddress (coor) {
        if (!coor._isCoordinate) {
          
          throw new Error('Illegal argument: Container.getAddress(coor) expects a Coordinate instance.')
        }
        var nodeId = coor.path[0];
        var nodePos = this.getPosition(nodeId);
        var offset;
        if (coor.isNodeCoordinate()) {
          if (coor.offset > 0) {
            offset = Number.MAX_VALUE;
          } else {
            offset = 0;
          }
        } else {
          offset = coor.offset;
        }
        return new ContainerAddress$2(nodePos, offset)
      }

      getLength () {
        return this.getContent().length
      }

      get length () {
        return this.getLength()
      }

      _getPosition (node) {
        if (this._isCaching) {
          return this._getCachedPosition(node)
        } else {
          return this._lookupPosition(node)
        }
      }

      _getCachedPosition (node) {
        let cache = this._cachedPositions || this._fillCache();
        let nodeId = node.id;
        let pos = -1;
        if (cache.hasOwnProperty(nodeId)) {
          pos = cache[nodeId];
        } else {
          pos = this._lookupPosition(node);
          cache[nodeId] = pos;
        }
        return pos
      }

      _fillCache () {
        let positions = {};
        this.nodes.forEach((id, pos) => {
          positions[id] = pos;
        });
        this._cachedPositions = positions;
        return positions
      }

      _invalidateCache () {
        this._cachedPositions = null;
      }

      _lookupPosition (node) {
        if (node.hasParent()) {
          node = node.getContainerRoot();
        }
        return this.getContent().indexOf(node.id)
      }

      _enableCaching () {
        
        if (this.document) {
          this.document.data.on('operation:applied', this._onOperationApplied, this);
          this._isCaching = true;
        }
      }

      _onOperationApplied (op) {
        if (op.type === 'set' || op.type === 'update') {
          if (op.path[0] === this.id) {
            this._invalidateCache();
          }
        }
      }

      _onDocumentChange (change) {
        if (change.hasUpdated(this.getContentPath())) {
          this._invalidateCache();
        }
      }

      
      

      hasChildren () {
        return this.getContent().length > 0
      }

      getChildIndex (child) {
        return this.getContent().indexOf(child.id)
      }

      getChildren () {
        return getNodes$2(this.getDocument(), this.getContent())
      }

      getChildAt (idx) {
        var childrenIds = this.getContent();
        if (idx < 0 || idx >= childrenIds.length) {
          throw new Error('Array index out of bounds: ' + idx + ', ' + childrenIds.length)
        }
        return this.getDocument().get(childrenIds[idx], 'strict')
      }

      getChildCount () {
        return this.getContent().length
      }
    }
    return AbstractContainer
  }

  class Container$2 extends ContainerMixin$2(DocumentNode$2) {
    constructor (...args) {
      super(...args);

      
      
      
      this._enableCaching();
    }

    dispose () {
      this.document.off(this);
    }

    getContentPath () {
      return [this.id, 'nodes']
    }

    getContent () {
      return this.nodes
    }
  }

  Container$2.prototype._isContainer = true;

  Container$2.schema = {
    type: '@container',
    nodes: { type: ['array', 'id'], default: [], owned: true }
  };

  function AnnotationMixin$2 (DocumentNode) {
    class AbstractAnnotation extends DocumentNode {
      constructor (doc, props) {
        super(doc, _normalizedProps$2(props));

        
        this.start = new Coordinate$2(this.start);
        this.end = new Coordinate$2(this.end);
      }

      

      get path () {
        console.warn('DEPRECATED: use annotation.start.path instead');
        return this.start.path
      }

      getPath () {
        return this.start.path
      }

      get startPath () {
        console.warn('DEPRECATED: use annotation.start.path instead.');
        return this.start.path
      }

      set startPath (path) {
        console.warn('DEPRECATED: use annotation.start.path instead.');
        this.start.path = path;
      }

      get startOffset () {
        console.warn('DEPRECATED: use annotation.start.offset instead.');
        return this.start.offset
      }

      set startOffset (offset) {
        console.warn('DEPRECATED: use annotation.start.offset instead.');
        this.start.offset = offset;
      }

      get endPath () {
        console.warn('DEPRECATED: use annotation.end.path instead.');
        return this.end.path
      }

      set endPath (path) {
        console.warn('DEPRECATED: use annotation.end.path instead.');
        this.end.path = path;
      }

      get endOffset () {
        console.warn('DEPRECATED: use annotation.end.offset instead.');
        return this.end.offset
      }

      set endOffset (offset) {
        console.warn('DEPRECATED: use annotation.end.offset instead.');
        this.end.offset = offset;
      }

      

      
      getText () {
        var doc = this.getDocument();
        if (!doc) {
          console.warn('Trying to use a Annotation which is not attached to the document.');
          return ''
        }
        return getTextForSelection$2(doc, this.getSelection())
      }

      isAnnotation () {
        return true
      }

      
      canSplit () {
        return true
      }

      
      isAnchor () {
        return false
      }

      
      getSelection () {
        const doc = this.getDocument();
        
        if (!doc) {
          console.warn('Trying to use a ContainerAnnotation which is not attached to the document.');
          return Selection$2.nullSelection()
        }
        if (this._isContainerAnnotation) {
          return doc.createSelection({
            type: 'container',
            containerId: this.containerId,
            startPath: this.start.path,
            startOffset: this.start.offset,
            endPath: this.end.path,
            endOffset: this.end.offset
          })
        } else {
          return this.getDocument().createSelection({
            type: 'property',
            path: this.start.path,
            startOffset: this.start.offset,
            endOffset: this.end.offset
          })
        }
      }

      _updateRange (tx, sel) {
        if (sel.isContainerSelection()) {
          
          if (!isEqual$2(this.start.path, sel.start.path)) {
            tx.set([this.id, 'start', 'path'], sel.start.path);
          }
          if (this.start.offset !== sel.start.offset) {
            tx.set([this.id, 'start', 'offset'], sel.start.offset);
          }
          if (!isEqual$2(this.end.path, sel.end.path)) {
            tx.set([this.id, 'end', 'path'], sel.end.path);
          }
          if (this.end.offset !== sel.end.offset) {
            tx.set([this.id, 'end', 'offset'], sel.end.offset);
          }
        } else if (sel.isPropertySelection()) {
          if (!isArrayEqual$2(this.start.path, sel.start.path)) {
            tx.set([this.id, 'path'], sel.start.path);
          }
          
          if (this.start.offset !== sel.start.offset) {
            tx.set([this.id, 'start', 'offset'], sel.start.offset);
          }
          if (this.end.offset !== sel.end.offset) {
            tx.set([this.id, 'end', 'offset'], sel.end.offset);
          }
        } else {
          throw new Error('Invalid selection.')
        }
      }
    }

    AbstractAnnotation.prototype._isAnnotation = true;

    AbstractAnnotation.schema = {
      start: { type: 'coordinate', default: { path: [], offset: 0 } },
      end: { type: 'coordinate', default: { path: [], offset: 0 } }
    };

    return AbstractAnnotation
  }

  function _normalizedProps$2 (props) {
    
    
    if (!props.hasOwnProperty('start')) {
      
      
      
      let start, end;
      if (props.hasOwnProperty('startPath') || props.hasOwnProperty('path')) {
        start = {
          path: props.startPath || props.path,
          offset: props.startOffset
        };
      }
      if (props.hasOwnProperty('endPath') || props.hasOwnProperty('endOffset')) {
        end = {
          path: props.endPath || props.path,
          offset: props.endOffset
        };
      }
      if (start && !end) {
        end = cloneDeep$2(start);
      }
      if (start) {
        props = Object.assign({}, props);
        delete props.path;
        delete props.startPath;
        delete props.endPath;
        delete props.startOffset;
        delete props.endOffset;
        props.start = start;
        props.end = end;
      }
    } else if (props.hasOwnProperty('end') && !props.end.path) {
      props.end.path = props.start.path;
    }
    return props
  }

  class PropertyAnnotation$2 extends AnnotationMixin$2(DocumentNode$2) {}

  PropertyAnnotation$2.prototype._isAnnotation = true;
  PropertyAnnotation$2.prototype._isPropertyAnnotation = true;

  PropertyAnnotation$2.isPropertyAnnotation = true;
  PropertyAnnotation$2.autoExpandRight = true;

  PropertyAnnotation$2.schema = {
    type: '@annotation',
    
    
    _content: { type: 'string', optional: true }
  };

  class ContainerAnnotation$2 extends AnnotationMixin$2(DocumentNode$2) {
    setHighlighted (highlighted, scope) {
      if (this.highlighted !== highlighted) {
        this.highlighted = highlighted;
        this.highlightedScope = scope;
        this.emit('highlighted', highlighted, scope);
        forEach$2(this.fragments, function (frag) {
          frag.emit('highlighted', highlighted, scope);
        });
      }
    }
  }

  ContainerAnnotation$2.schema = {
    type: '@container-annotation',
    containerId: 'string',
    start: 'coordinate',
    end: 'coordinate'
  };

  ContainerAnnotation$2.prototype._isAnnotation = true;
  ContainerAnnotation$2.prototype._isContainerAnnotation = true;

  class EditingBehavior$2 {
    constructor () {
      this._merge = {};
      this._mergeComponents = {};
      this._break = {};
    }

    defineMerge (firstType, secondType, impl) {
      if (!this._merge[firstType]) {
        this._merge[firstType] = {};
      }
      this._merge[firstType][secondType] = impl;
      return this
    }

    canMerge (firstType, secondType) {
      return (this._merge[firstType] && this._merge[firstType][secondType])
    }

    getMerger (firstType, secondType) {
      return this._merge[firstType][secondType]
    }

    defineComponentMerge (nodeType, impl) {
      this._mergeComponents[nodeType] = impl;
    }

    canMergeComponents (nodeType) {
      return this._mergeComponents[nodeType]
    }

    getComponentMerger (nodeType) {
      return this._mergeComponents[nodeType]
    }

    defineBreak (nodeType, impl) {
      this._break[nodeType] = impl;
      return this
    }

    canBreak (nodeType) {
      return this._break[nodeType]
    }

    getBreaker (nodeType) {
      return this._break[nodeType]
    }
  }

  function extend$2 (...args) {
    return Object.assign(...args)
  }

  class ArrayIterator$2 {
    constructor (arr) {
      this.arr = arr;
      this.pos = -1;
    }

    get _isArrayIterator () {
      return true
    }

    
    hasNext () {
      return this.pos < this.arr.length - 1
    }

    
    next () {
      this.pos += 1;
      var next = this.arr[this.pos];
      return next
    }

    
    back () {
      if (this.pos >= 0) {
        this.pos -= 1;
      }
      return this
    }

    peek () {
      return this.arr[this.pos + 1]
    }
  }

  function findIndex$1$2 (arr, predicate) {
    if (!isFunction$1$2(predicate)) return arr.indexOf(predicate)
    for (let i = 0; i < arr.length; i++) {
      if (predicate(arr[i])) return i
    }
    return -1
  }

  class DOMEventListener$2 {
    constructor (eventName, handler, options) {
      
      if (!isString$2(eventName) || !isFunction$1$2(handler)) {
        throw new Error("Illegal arguments: 'eventName' must be a String, and 'handler' must be a Function.")
      }
      options = options || {};
      var origHandler = handler;
      var context = options.context;
      var capture = Boolean(options.capture);

      if (context) {
        handler = handler.bind(context);
      }
      if (options.once === true) {
        handler = _once$2(this, handler);
      }

      this.eventName = eventName;
      this.originalHandler = origHandler;
      this.handler = handler;
      this.capture = capture;
      this.context = context;
      this.options = options;
      
      this._el = null;
    }
  }

  DOMEventListener$2.prototype._isDOMEventListener = true;

  DOMEventListener$2.findIndex = function (eventListeners, eventName, handler) {
    var idx = -1;
    if (arguments[1]._isDOMEventListener) {
      idx = eventListeners.indexOf(arguments[1]);
    } else {
      idx = findIndex$1$2(eventListeners,
        _matches$2.bind(null, {
          eventName: eventName,
          originalHandler: handler
        })
      );
    }
    return idx
  };

  function _matches$2 (l1, l2) {
    return l1.eventName === l2.eventName && l1.originalHandler === l2.originalHandler
  }

  function _once$2 (listener, handler) {
    return function (event) {
      handler(event);
      listener._el.removeEventListener(listener);
    }
  }

  const NOT_IMPLEMENTED$2 = 'This method is not implemented.';


  class DOMElement$2 {
    

    

    

    

    

    

    

    

    

    
    getNativeElement () {
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    isTextNode () {
      
      return false
    }

    
    isElementNode () {
      
      return false
    }

    
    isCommentNode () {
      
      return false
    }

    
    isDocumentNode () {
      
      return false
    }

    
    getTagName () {
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    setTagName (tagName) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    getId () {
      return this.getAttribute('id')
    }

    
    setId (id) {
      this.setAttribute('id', id);
      return this
    }

    
    hasClass (className) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    addClass (classString) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    removeClass (classString) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    hasAttribute (name) {
      return Boolean(this.getAttribute(name))
    }

    
    attr () {
      if (arguments.length === 1) {
        if (isString$2(arguments[0])) {
          return this.getAttribute(arguments[0])
        } else if (isObject$3(arguments[0])) {
          forEach$2(arguments[0], function (value, name) {
            this.setAttribute(name, value);
          }.bind(this));
        }
      } else if (arguments.length === 2) {
        this.setAttribute(arguments[0], arguments[1]);
      }
      return this
    }

    
    removeAttr (name) {
      var names = name.split(/\s+/);
      if (names.length === 1) {
        this.removeAttribute(name);
      } else {
        names.forEach(function (name) {
          this.removeAttribute(name);
        }.bind(this));
      }
      return this
    }

    
    getAttribute (name) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    setAttribute (name, value) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    removeAttribute (name) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    getAttributes () {
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    htmlProp () {
      if (arguments.length === 1) {
        if (isString$2(arguments[0])) {
          return this.getProperty(arguments[0])
        } else if (isObject$3(arguments[0])) {
          forEach$2(arguments[0], function (value, name) {
            this.setProperty(name, value);
          }.bind(this));
        }
      } else if (arguments.length === 2) {
        this.setProperty(arguments[0], arguments[1]);
      }
      return this
    }

    getProperty (name) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    setProperty (name, value) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    val (value) {
      if (arguments.length === 0) {
        return this.getValue()
      } else {
        this.setValue(value);
        return this
      }
    }

    getValue () {
      return this.getProperty('value')
    }

    setValue (value) {
      this.setProperty('value', value);
      return this
    }

    
    css () {
      
      if (arguments.length === 1) {
        
        if (isString$2(arguments[0])) {
          return this.getStyle(arguments[0])
        } else if (isObject$3(arguments[0])) {
          forEach$2(arguments[0], function (value, name) {
            this.setStyle(name, value);
          }.bind(this));
        } else {
          throw new Error('Illegal arguments.')
        }
      } else if (arguments.length === 2) {
        this.setStyle(arguments[0], arguments[1]);
      } else {
        throw new Error('Illegal arguments.')
      }
      return this
    }

    getStyle (name) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    setStyle (name, value) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    text (text) {
      if (arguments.length === 0) {
        return this.getTextContent()
      } else {
        this.setTextContent(text);
      }
      return this
    }

    
    getTextContent () {
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    setTextContent (text) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    html (html) {
      if (arguments.length === 0) {
        return this.getInnerHTML()
      } else {
        this.setInnerHTML(html);
      }
      return this
    }

    
    getInnerHTML () {
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    getInnerXML () {
      return this.getInnerHTML()
    }

    
    setInnerHTML (html) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    setInnerXML (xml) {
      return this.setInnerHTML(xml)
    }

    
    getOuterHTML () {
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    getOuterXML () {
      return this.getOuterHTML()
    }

    
    on (eventName, handler, context, options) {
      
      if (!isString$2(eventName)) {
        throw new Error('Illegal argument: "event" must be a String.')
      }
      options = options || {};
      if (context) {
        options.context = context;
      }
      
      if (!handler || !isFunction$1$2(handler)) {
        throw new Error('Illegal argument: invalid handler function for event ' + eventName)
      }
      this.addEventListener(eventName, handler, options);
      return this
    }

    
    off (eventName, handler) {
      
      if (arguments.length === 1 && !isString$2(eventName)) {
        let context = arguments[0];
        this.getEventListeners().filter(function (l) {
          return l.context === context
        }).forEach(function (l) {
          this.removeEventListener(l);
        }.bind(this));
      } else {
        this.removeEventListener(eventName, handler);
      }
      return this
    }

    addEventListener (eventName, handler, options = {}) {
      let listener;
      if (arguments.length === 1 && arguments[0]) {
        listener = arguments[0];
      } else {
        listener = this._createEventListener(eventName, handler, options);
      }
      if (!this.eventListeners) {
        this.eventListeners = [];
      }
      listener._el = this;
      this.eventListeners.push(listener);
      this._addEventListenerNative(listener);
      return this
    }

    _createEventListener (eventName, handler, options) {
      return new DOMEventListener$2(eventName, handler, options)
    }

    _addEventListenerNative(listener) {} 

    removeEventListener (eventName, handler) {
      if (!this.eventListeners) return
      
      let listener = null;
      let idx = DOMEventListener$2.findIndex(this.eventListeners, eventName, handler);
      listener = this.eventListeners[idx];
      if (idx > -1) {
        this.eventListeners.splice(idx, 1);
        
        listener._el = null;
        this._removeEventListenerNative(listener);
      }
      return this
    }

    _removeEventListenerNative(listener) {} 

    removeAllEventListeners () {
      if (!this.eventListeners) return
      for (let i = 0; i < this.eventListeners.length; i++) {
        let listener = this.eventListeners[i];
        
        listener._el = null;
        this._removeEventListenerNative(listener);
      }
      delete this.eventListeners;
    }

    getEventListeners () {
      return this.eventListeners || []
    }

    
    getNodeType () {
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    getContentType () {
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    getChildCount () {
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    getChildNodes () {
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    getChildren () {
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    getChildAt (pos) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    getChildIndex (child) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    getChildNodeIterator () {
      return new ArrayIterator$2(this.getChildNodes())
    }

    getLastChild () {
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    getFirstChild () {
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    getNextSibling () {
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    getPreviousSibling () {
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    clone(deep) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    createElement (str) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    createTextNode (text) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    createComment (data) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    createProcessingInstruction (name, data) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    createCDATASection (data) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    is (cssSelector) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    getParent () {
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    getOwnerDocument () {
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    getDoctype () {
      
      throw new Error('NOT_IMPLEMENTED')
    }

    setDoctype(qualifiedNameStr, publicId, systemId) { 
      throw new Error('NOT_IMPLEMENTED')
    }

    setDocType (...args) {
      console.error('DEPRECATED: use setDoctype() instead');
      return this.setDoctype(...args)
    }

    
    find (cssSelector) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    findAll (cssSelector) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    append (child) {
      var children;
      if (arguments.length === 1) {
        if (isArray$3(child)) {
          children = child;
        } else {
          this.appendChild(child);
          return this
        }
      } else {
        children = arguments;
      }
      if (children) {
        Array.prototype.forEach.call(children, this.appendChild.bind(this));
      }
      return this
    }

    appendChild (child) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    insertAt (pos, child) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    insertBefore (newChild, before) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    removeAt (pos) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    removeChild (child) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    replaceChild (oldChild, newChild) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    
    remove () {
      var parent = this.getParent();
      if (parent) {
        parent.removeChild(this);
      }
    }

    
    empty () {
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    serialize () {
      return this.getOuterHTML()
    }

    isInDocument () {
      let el = this;
      while (el) {
        if (el.isDocumentNode()) {
          return true
        }
        el = el.getParent();
      }
    }

    
    focus () {
      
      return this
    }

    
    select () {
      
      return this
    }

    
    blur () {
      
      return this
    }

    
    click () {
      
      return this
    }

    

    getWidth () {
      
      return 0
    }

    getHeight () {
      
      return 0
    }

    
    getOuterHeight (withMargin) { 
      
      return 0
    }

    
    getOffset () {
      
      return { top: 0, left: 0 }
    }

    
    getPosition () {
      
      return { top: 0, left: 0 }
    }

    
    getElementFactory () {
      return this.createElement.bind(this)
    }

    
    emit(name, data) { 
      
      throw new Error(NOT_IMPLEMENTED$2)
    }

    

    get id () {
      return this.getId()
    }

    set id (id) {
      this.setId(id);
    }

    get tagName () {
      return this.getTagName()
    }

    set tagName (tagName) {
      this.setTagName(tagName);
    }

    get nodeName () {
      return this.getTagName()
    }

    get nodeType () {
      return this.getNodeType()
    }

    get className () {
      return this.getAttribute('class')
    }

    set className (className) {
      this.setAttribute('class', className);
    }

    get textContent () {
      return this.getTextContent()
    }

    set textContent (text) {
      this.setTextContent(text);
    }

    get innerHTML () {
      return this.getInnerHTML()
    }

    set innerHTML (html) {
      this.setInnerHTML(html);
    }

    get outerHTML () {
      return this.getOuterHTML()
    }

    get firstChild () {
      return this.getFirstChild()
    }

    get lastChild () {
      return this.getLastChild()
    }

    get nextSibling () {
      return this.getNextSibling()
    }

    get previousSibling () {
      return this.getPreviousSibling()
    }

    get parentNode () {
      return this.getParent()
    }

    get height () {
      return this.getHeight()
    }

    get width () {
      return this.getWidth()
    }

    get value () {
      return this.getValue()
    }

    set value (value) {
      return this.setValue(value)
    }
  }

  DOMElement$2.prototype._isDOMElement = true;

  DOMElement$2.pxStyles = {
    top: true,
    bottom: true,
    left: true,
    right: true,
    height: true,
    width: true
  };

  DOMElement$2.EMPTY_HTML = '<html><head></head><body></body></html>';

  const SIGNATURE$2 = uuid$2('_BrowserDOMElement');

  function _attach$2 (nativeEl, browserDOMElement) {
    nativeEl[SIGNATURE$2] = browserDOMElement;
  }

  function _detach$2 (nativeEl) {
    delete nativeEl[SIGNATURE$2];
  }

  function _unwrap$2 (nativeEl) {
    return nativeEl[SIGNATURE$2]
  }

  class BrowserDOMElement$2 extends DOMElement$2 {
    constructor (el) {
      super();
      console.assert(el instanceof window.Node, 'Expecting native DOM node.');
      this.el = el;
      
      
      _attach$2(el, this);
    }

    getNativeElement () {
      return this.el
    }

    getNodeType () {
      switch (this.el.nodeType) {
        case window.Node.TEXT_NODE:
          return 'text'
        case window.Node.ELEMENT_NODE:
          return 'element'
        case window.Node.DOCUMENT_NODE:
          return 'document'
        case window.Node.COMMENT_NODE:
          return 'comment'
        case window.Node.PROCESSING_INSTRUCTION_NODE:
          return 'directive'
        case window.Node.CDATA_SECTION_NODE:
          return 'cdata'
        default:
          
      }
    }

    getDoctype () {
      if (this.isDocumentNode()) {
        return this.el.doctype
      } else {
        return this.getOwnerDocument().getDoctype()
      }
    }

    setDoctype (qualifiedNameStr, publicId, systemId) {
      let ownerDocument = this._getNativeOwnerDocument();
      let oldDocType = ownerDocument.doctype;
      let newDocType = ownerDocument.implementation.createDocumentType(
        qualifiedNameStr, publicId, systemId
      );
      if (oldDocType) {
        oldDocType.parentNode.replaceChild(newDocType, oldDocType);
      } else {
        ownerDocument.insertBefore(newDocType, ownerDocument.firstChild);
      }
    }

    isTextNode () {
      return (this.el.nodeType === window.Node.TEXT_NODE)
    }

    isElementNode () {
      return (this.el.nodeType === window.Node.ELEMENT_NODE)
    }

    isCommentNode () {
      return (this.el.nodeType === window.Node.COMMENT_NODE)
    }

    isDocumentNode () {
      return (this.el.nodeType === window.Node.DOCUMENT_NODE)
    }

    hasClass (className) {
      return this.el.classList.contains(className)
    }

    addClass (className) {
      this.el.classList.add(className);
      return this
    }

    removeClass (className) {
      this.el.classList.remove(className);
      return this
    }

    getAttribute (name) {
      return this.el.getAttribute(name)
    }

    setAttribute (name, value) {
      this.el.setAttribute(name, String(value));
      return this
    }

    removeAttribute (name) {
      this.el.removeAttribute(name);
      return this
    }

    getAttributes () {
      if (!this.el.attributes._mapAdapter) {
        this.el.attributes._mapAdapter = new AttributesMapAdapter$2(this.el.attributes);
      }
      return this.el.attributes._mapAdapter
    }

    getProperty (name) {
      return this.el[name]
    }

    setProperty (name, value) {
      
      
      if (this._isXML()) throw new Error('setProperty() is only supported for HTML elements.')
      this.el[name] = value;
      return this
    }

    getTagName () {
      
      
      if (this._isXML()) {
        return this.el.tagName
      } else if (this.el.tagName) {
        return this.el.tagName.toLowerCase()
      }
    }

    setTagName (tagName) {
      let newEl = this.createElement(tagName);
      let attributes = this.el.attributes;
      let l = attributes.length;
      let i;
      for (i = 0; i < l; i++) {
        let attr = attributes.item(i);
        newEl.setAttribute(attr.name, attr.value);
      }
      if (this.eventListeners) {
        this.eventListeners.forEach(function (listener) {
          newEl.addEventListener(listener.eventName, listener.handler, listener.capture);
        });
      }
      newEl.append(this.getChildNodes());

      this._replaceNativeEl(newEl.getNativeElement());
      return this
    }

    getId () {
      return this.el.id
    }

    setId (id) {
      this.el.id = id;
      return this
    }

    getStyle (name) {
      let val = this.el.style[name];
      if (!val) {
        let computedStyle = this.getComputedStyle();
        val = computedStyle[name];
      }
      return val
    }

    getComputedStyle () {
      return window.getComputedStyle(this.el)
    }

    setStyle (name, value) {
      if (DOMElement$2.pxStyles[name] && isNumber$2(value)) value = value + 'px';
      this.el.style[name] = value;
      return this
    }

    getTextContent () {
      return this.el.textContent
    }

    setTextContent (text) {
      this.el.textContent = text;
      return this
    }

    getInnerHTML () {
      if (this._isXML()) {
        let xs = new window.XMLSerializer();
        let result = Array.prototype.map.call(this.el.childNodes, c => xs.serializeToString(c));
        return result.join('')
      } else {
        return this.el.innerHTML
      }
    }

    setInnerHTML (html) {
      
      
      this.el.innerHTML = html;
      return this
    }

    getOuterHTML () {
      
      
      if (this._isXML()) {
        let xs = new window.XMLSerializer();
        return xs.serializeToString(this.el)
      } else {
        return this.el.outerHTML
      }
    }

    _addEventListenerNative (listener) {
      this.el.addEventListener(listener.eventName, listener.handler, listener.capture);
    }

    _removeEventListenerNative (listener) {
      this.el.removeEventListener(listener.eventName, listener.handler);
    }

    getEventListeners () {
      return this.eventListeners || []
    }

    getChildCount () {
      return this.el.childNodes.length
    }

    getChildNodes () {
      let childNodes = [];
      for (let node = this.el.firstChild; node; node = node.nextSibling) {
        childNodes.push(BrowserDOMElement$2.wrap(node));
      }
      return childNodes
    }

    getChildNodeIterator () {
      return new BrowserChildNodeIterator$2(this.el)
    }

    get childNodes () {
      return this.getChildNodes()
    }

    getChildren () {
      
      
      let children = [];
      for (let node = this.el.firstChild; node; node = node.nextSibling) {
        if (node.nodeType === window.Node.ELEMENT_NODE) {
          children.push(BrowserDOMElement$2.wrap(node));
        }
      }
      return children
    }

    get children () {
      return this.getChildren()
    }

    getChildAt (pos) {
      return BrowserDOMElement$2.wrap(this.el.childNodes[pos])
    }

    getChildIndex (child) {
      
      if (!child._isBrowserDOMElement) {
        throw new Error('Expecting a BrowserDOMElement instance.')
      }
      return Array.prototype.indexOf.call(this.el.childNodes, child.el)
    }

    getFirstChild () {
      let firstChild = this.el.firstChild;
      
      if (firstChild) {
        return BrowserDOMElement$2.wrap(firstChild)
      } else {
        return null
      }
    }

    getLastChild () {
      var lastChild = this.el.lastChild;
      
      if (lastChild) {
        return BrowserDOMElement$2.wrap(lastChild)
      } else {
        return null
      }
    }

    getNextSibling () {
      let next = this.el.nextSibling;
      
      if (next) {
        return BrowserDOMElement$2.wrap(next)
      } else {
        return null
      }
    }

    getPreviousSibling () {
      let previous = this.el.previousSibling;
      
      if (previous) {
        return BrowserDOMElement$2.wrap(previous)
      } else {
        return null
      }
    }

    clone (deep) {
      let clone = this.el.cloneNode(deep);
      return BrowserDOMElement$2.wrap(clone)
    }

    createDocument (format) {
      return BrowserDOMElement$2.createDocument(format)
    }

    createElement (tagName) {
      let doc = this._getNativeOwnerDocument();
      let el = doc.createElement(tagName);
      return BrowserDOMElement$2.wrap(el)
    }

    createTextNode (text) {
      let doc = this._getNativeOwnerDocument();
      let el = doc.createTextNode(text);
      return BrowserDOMElement$2.wrap(el)
    }

    createComment (data) {
      let doc = this._getNativeOwnerDocument();
      let el = doc.createComment(data);
      return BrowserDOMElement$2.wrap(el)
    }

    createProcessingInstruction (name, data) {
      let doc = this._getNativeOwnerDocument();
      let el = doc.createProcessingInstruction(name, data);
      return BrowserDOMElement$2.wrap(el)
    }

    createCDATASection (data) {
      let doc = this._getNativeOwnerDocument();
      let el = doc.createCDATASection(data);
      return BrowserDOMElement$2.wrap(el)
    }

    is (cssSelector) {
      
      
      let el = this.el;
      
      if (this.isElementNode()) {
        return matches$2(el, cssSelector)
      } else {
        return false
      }
    }

    getParent () {
      let parent = this.el.parentNode;
      
      if (parent) {
        return BrowserDOMElement$2.wrap(parent)
      } else {
        return null
      }
    }

    getOwnerDocument () {
      return BrowserDOMElement$2.wrap(this._getNativeOwnerDocument())
    }

    get ownerDocument () {
      return this.getOwnerDocument()
    }

    _getNativeOwnerDocument () {
      return (this.isDocumentNode() ? this.el : this.el.ownerDocument)
    }

    find (cssSelector) {
      let result = null;
      if (this.el.querySelector) {
        result = this.el.querySelector(cssSelector);
      }
      if (result) {
        return BrowserDOMElement$2.wrap(result)
      } else {
        return null
      }
    }

    findAll (cssSelector) {
      let result = [];
      if (this.el.querySelectorAll) {
        result = this.el.querySelectorAll(cssSelector);
      }
      return Array.prototype.map.call(result, function (el) {
        return BrowserDOMElement$2.wrap(el)
      })
    }

    _normalizeChild (child) {
      if (isNil$2(child)) return child

      if (child instanceof window.Node) {
        child = BrowserDOMElement$2.wrap(child);
      
      
      
      } else if (child._isBrowserDOMElement && !(child instanceof BrowserDOMElement$2)) {
        child = BrowserDOMElement$2.wrap(child);
      } else if (isString$2(child) || isNumber$2(child)) {
        child = this.createTextNode(child);
      }
      
      if (!child || !child._isBrowserDOMElement) {
        throw new Error('Illegal child type.')
      }
      console.assert(_unwrap$2(child.el) === child, 'The backlink to the wrapper should be consistent');
      return child.getNativeElement()
    }

    appendChild (child) {
      let nativeChild = this._normalizeChild(child);
      if (nativeChild) {
        this.el.appendChild(nativeChild);
      }
      return this
    }

    insertAt (pos, child) {
      let nativeChild = this._normalizeChild(child);
      let childNodes = this.el.childNodes;
      if (pos >= childNodes.length) {
        this.el.appendChild(nativeChild);
      } else {
        this.el.insertBefore(nativeChild, childNodes[pos]);
      }
      return this
    }

    insertBefore (child, before) {
      
      if (isNil$2(before)) {
        return this.appendChild(child)
      }
      if (!before._isBrowserDOMElement) {
        throw new Error('insertBefore(): Illegal arguments. "before" must be a BrowserDOMElement instance.')
      }
      var nativeChild = this._normalizeChild(child);
      if (nativeChild) {
        this.el.insertBefore(nativeChild, before.el);
      }
      return this
    }

    removeAt (pos) {
      this.el.removeChild(this.el.childNodes[pos]);
      return this
    }

    removeChild (child) {
      
      if (!child || !child._isBrowserDOMElement) {
        throw new Error('removeChild(): Illegal arguments. Expecting a BrowserDOMElement instance.')
      }
      this.el.removeChild(child.el);
      return this
    }

    replaceChild (oldChild, newChild) {
      
      if (!newChild || !oldChild ||
          !newChild._isBrowserDOMElement || !oldChild._isBrowserDOMElement) {
        throw new Error('replaceChild(): Illegal arguments. Expecting BrowserDOMElement instances.')
      }
      
      this.el.replaceChild(newChild.el, oldChild.el);
      return this
    }

    empty () {
      let el = this.el;
      while (el.lastChild) {
        el.removeChild(el.lastChild);
      }
      return this
    }

    remove () {
      if (this.el.parentNode) {
        this.el.parentNode.removeChild(this.el);
      }
      return this
    }

    serialize () {
      let outerHTML = this.el.outerHTML;
      if (isString$2(outerHTML)) {
        return outerHTML
      } else {
        let xs = new window.XMLSerializer();
        return xs.serializeToString(this.el)
      }
    }

    isInDocument () {
      let el = this.el;
      while (el) {
        if (el.nodeType === window.Node.DOCUMENT_NODE) {
          return true
        }
        el = el.parentNode;
      }
    }

    _replaceNativeEl (newEl) {
      console.assert(newEl instanceof window.Node, 'Expecting a native element.');
      let oldEl = this.el;
      let parentNode = oldEl.parentNode;
      if (parentNode) {
        parentNode.replaceChild(newEl, oldEl);
      }
      this.el = newEl;
      _detach$2(oldEl);
      _attach$2(newEl, this);
    }

    _getChildNodeCount () {
      return this.el.childNodes.length
    }

    focus () {
      this.el.focus();
      return this
    }

    select () {
      this.el.select();
      return this
    }

    blur () {
      this.el.blur();
      return this
    }

    click () {
      this.el.click();
      return this
    }

    getWidth () {
      let rect = this.el.getClientRects()[0];
      if (rect) {
        return rect.width
      } else {
        return 0
      }
    }

    getHeight () {
      let rect = this.el.getClientRects()[0];
      if (rect) {
        return rect.height
      } else {
        return 0
      }
    }

    getOffset () {
      let rect = this.el.getBoundingClientRect();
      return {
        top: rect.top + document.body.scrollTop,
        left: rect.left + document.body.scrollLeft
      }
    }

    getPosition () {
      return {left: this.el.offsetLeft, top: this.el.offsetTop}
    }

    getOuterHeight (withMargin) {
      let outerHeight = this.el.offsetHeight;
      if (withMargin) {
        let style = this.getComputedStyle();
        outerHeight += parseInt(style.marginTop, 10) + parseInt(style.marginBottom, 10);
      }
      return outerHeight
    }

    getContentType () {
      return this._getNativeOwnerDocument().contentType
    }

    _isXML () {
      return this.getContentType() === 'application/xml'
    }

    emit (name, data) {
      let event;
      if (data) {
        event = new window.CustomEvent(name, {
          detail: data,
          bubbles: true,
          cancelable: true
        });
      } else {
        event = new window.Event(name, {
          bubbles: true,
          cancelable: true
        });
      }
      this.el.dispatchEvent(event);
    }
  }

  BrowserDOMElement$2.prototype._isBrowserDOMElement = true;


  BrowserDOMElement$2.createDocument = function (format) {
    let doc;
    if (format === 'xml') {
      
      doc = window.document.implementation.createDocument(null, 'dummy');
      
      doc.removeChild(doc.firstChild);
    } else {
      doc = (new window.DOMParser()).parseFromString(DOMElement$2.EMPTY_HTML, 'text/html');
    }
    return BrowserDOMElement$2.wrap(doc)
  };

  BrowserDOMElement$2.parseMarkup = function (str, format, options = {}) {
    if (!str) {
      return BrowserDOMElement$2.createDocument(format)
    }
    if (options.snippet) {
      str = `<div id='__snippet__'>${str}</div>`;
    }
    let doc;
    let parser = new window.DOMParser();
    if (format === 'html') {
      doc = BrowserDOMElement$2.wrap(
        _check(
          parser.parseFromString(str, 'text/html')
        )
      );
    } else if (format === 'xml') {
      doc = BrowserDOMElement$2.wrap(
        _check(
          parser.parseFromString(str, 'application/xml')
        )
      );
    }
    if (options.snippet) {
      let childNodes = doc.find('#__snippet__').childNodes;
      if (childNodes.length === 1) {
        return childNodes[0]
      } else {
        return childNodes
      }
    } else {
      return doc
    }

    function _check (doc) {
      if (doc) {
        let parserError = doc.querySelector('parsererror');
        if (parserError) {
          throw new Error('ParserError: ' + parserError)
        }
      }
      return doc
    }
  };

  BrowserDOMElement$2.wrap =
  BrowserDOMElement$2.wrapNativeElement = function (el) {
    if (el) {
      let _el = _unwrap$2(el);
      if (_el) {
        return _el
      } else if (el instanceof window.Node) {
        return new BrowserDOMElement$2(el)
      } else if (el._isBrowserDOMElement) {
        return new BrowserDOMElement$2(el.getNativeElement())
      } else if (el === window) {
        return BrowserDOMElement$2.getBrowserWindow()
      }
    } else {
      return null
    }
  };

  BrowserDOMElement$2.unwrap = function (nativeEl) {
    return _unwrap$2(nativeEl)
  };


  class BrowserWindow$2 {
    constructor () {
      
      this.el = window;
      _attach$2(window, this);
    }
  }

  BrowserWindow$2.prototype.on = BrowserDOMElement$2.prototype.on;
  BrowserWindow$2.prototype.off = BrowserDOMElement$2.prototype.off;
  BrowserWindow$2.prototype.addEventListener = BrowserDOMElement$2.prototype.addEventListener;
  BrowserWindow$2.prototype.removeEventListener = BrowserDOMElement$2.prototype.removeEventListener;
  BrowserWindow$2.prototype._createEventListener = BrowserDOMElement$2.prototype._createEventListener;
  BrowserWindow$2.prototype._addEventListenerNative = BrowserDOMElement$2.prototype._addEventListenerNative;
  BrowserWindow$2.prototype._removeEventListenerNative = BrowserDOMElement$2.prototype._removeEventListenerNative;

  BrowserWindow$2.prototype.getEventListeners = BrowserDOMElement$2.prototype.getEventListeners;

  BrowserDOMElement$2.getBrowserWindow = function () {
    if (window[SIGNATURE$2]) return window[SIGNATURE$2]
    return new BrowserWindow$2(window)
  };

  BrowserDOMElement$2.isReverse = function (anchorNode, anchorOffset, focusNode, focusOffset) {
    
    
    if (focusNode && anchorNode) {
      if (!BrowserDOMElement$2.isReverse._r1) {
        BrowserDOMElement$2.isReverse._r1 = window.document.createRange();
        BrowserDOMElement$2.isReverse._r2 = window.document.createRange();
      }
      const _r1 = BrowserDOMElement$2.isReverse._r1;
      const _r2 = BrowserDOMElement$2.isReverse._r2;
      _r1.setStart(anchorNode.getNativeElement(), anchorOffset);
      _r2.setStart(focusNode.getNativeElement(), focusOffset);
      let cmp = _r1.compareBoundaryPoints(window.Range.START_TO_START, _r2);
      if (cmp === 1) {
        return true
      }
    }
    return false
  };

  BrowserDOMElement$2.getWindowSelection = function () {
    let nativeSel = window.getSelection();
    let result = {
      anchorNode: BrowserDOMElement$2.wrap(nativeSel.anchorNode),
      anchorOffset: nativeSel.anchorOffset,
      focusNode: BrowserDOMElement$2.wrap(nativeSel.focusNode),
      focusOffset: nativeSel.focusOffset
    };
    return result
  };

  function matches$2 (el, selector) {
    let elProto = window.Element.prototype;
    let _matches = (
      elProto.matches || elProto.matchesSelector ||
      elProto.msMatchesSelector || elProto.webkitMatchesSelector
    );
    return _matches.call(el, selector)
  }

  class AttributesMapAdapter$2 {
    constructor (attributes) {
      this.attributes = attributes;
    }

    get size () {
      return this.attributes.length
    }

    get (name) {
      let item = this.attributes.getNamedItem(name);
      if (item) {
        return item.value
      }
    }

    set (name, value) {
      this.attributes.setNamedItem(name, value);
    }

    forEach (fn) {
      const S = this.size;
      for (let i = 0; i < S; i++) {
        const item = this.attributes.item(i);
        fn(item.value, item.name);
      }
    }

    map (fn) {
      let result = [];
      this.forEach((val, key) => { result.push(fn(val, key)); });
      return result
    }

    keys () {
      return this.map((val, key) => { return key })
    }

    values () {
      return this.map((val) => { return val })
    }

    entries () {
      return this.map((val, key) => { return [key, val] })
    }
  }

  class BrowserChildNodeIterator$2 {
    constructor (el) {
      this._next = el.firstChild;
      this._curr = null;
    }

    hasNext () {
      return Boolean(this._next)
    }

    next () {
      let next = this._next;
      this._curr = next;
      this._next = next.nextSibling;
      return BrowserDOMElement$2.wrap(next)
    }

    back () {
      this._next = this._curr;
      this._curr = this._curr.previousSibling;
    }

    peek () {
      return BrowserDOMElement$2.wrap(this._curr)
    }
  }

  var inBrowser$2 = platform$2.inBrowser;

  var domelementtype$2 = {
  	Text: "text", 
  	Directive: "directive", 
  	Comment: "comment", 
  	Script: "script", 
  	Style: "style", 
  	Tag: "tag", 
  	CDATA: "cdata", 
  	Doctype: "doctype",

  	isTag: function(elem){
  		return elem.type === "tag" || elem.type === "script" || elem.type === "style";
  	}
  };

  var boolbase$2 = {
  	trueFunc: function trueFunc(){
  		return true;
  	},
  	falseFunc: function falseFunc(){
  		return false;
  	}
  };

  var parse_1$3 = parse$3;




  var re_nthElement$2 = /^([+\-]?\d*n)?\s*(?:([+\-]?)\s*(\d+))?$/;


  function parse$3(formula){
  	formula = formula.trim().toLowerCase();

  	if(formula === "even"){
  		return [2, 0];
  	} else if(formula === "odd"){
  		return [2, 1];
  	} else {
  		var parsed = formula.match(re_nthElement$2);

  		if(!parsed){
  			throw new SyntaxError("n-th rule couldn't be parsed ('" + formula + "')");
  		}

  		var a;

  		if(parsed[1]){
  			a = parseInt(parsed[1], 10);
  			if(isNaN(a)){
  				if(parsed[1].charAt(0) === "-") a = -1;
  				else a = 1;
  			}
  		} else a = 0;

  		return [
  			a,
  			parsed[3] ? parseInt((parsed[2] || "") + parsed[3], 10) : 0
  		];
  	}
  }

  var compile_1$3 = compile$3;

  var trueFunc$3  = boolbase$2.trueFunc,
      falseFunc$5 = boolbase$2.falseFunc;


  function compile$3(parsed){
  	var a = parsed[0],
  	    b = parsed[1] - 1;

  	
  	
  	if(b < 0 && a <= 0) return falseFunc$5;

  	
  	if(a ===-1) return function(pos){ return pos <= b; };
  	if(a === 0) return function(pos){ return pos === b; };
  	
  	if(a === 1) return b < 0 ? trueFunc$3 : function(pos){ return pos >= b; };

  	
  	var bMod = b % a;
  	if(bMod < 0) bMod += a;

  	if(a > 1){
  		return function(pos){
  			return pos >= b && pos % a === bMod;
  		};
  	}

  	a *= -1; 

  	return function(pos){
  		return pos <= b && pos % a === bMod;
  	};
  }

  var nthCheck$2 = function nthCheck(formula){
  	return compile_1$3(parse_1$3(formula));
  };

  var parse_1$1$2 = parse_1$3;
  var compile_1$1$2 = compile_1$3;
  nthCheck$2.parse = parse_1$1$2;
  nthCheck$2.compile = compile_1$1$2;

  var cssWhat$2 = parse$1$2;

  var re_name$2 = /^(?:\\.|[\w\-\u00c0-\uFFFF])+/,
      re_escape$2 = /\\([\da-f]{1,6}\s?|(\s)|.)/ig,
      
      re_attr$2 = /^\s*((?:\\.|[\w\u00c0-\uFFFF\-])+)\s*(?:(\S?)=\s*(?:(['"])(.*?)\3|(#?(?:\\.|[\w\u00c0-\uFFFF\-])*)|)|)\s*(i)?\]/;

  var actionTypes$2 = {
  	__proto__: null,
  	"undefined": "exists",
  	"":  "equals",
  	"~": "element",
  	"^": "start",
  	"$": "end",
  	"*": "any",
  	"!": "not",
  	"|": "hyphen"
  };

  var simpleSelectors$2 = {
  	__proto__: null,
  	">": "child",
  	"<": "parent",
  	"~": "sibling",
  	"+": "adjacent"
  };

  var attribSelectors$2 = {
  	__proto__: null,
  	"#": ["id", "equals"],
  	".": ["class", "element"]
  };


  var unpackPseudos$2 = {
  	__proto__: null,
  	"has": true,
  	"not": true,
  	"matches": true
  };

  var stripQuotesFromPseudos$2 = {
  	__proto__: null,
  	"contains": true,
  	"icontains": true
  };

  var quotes$2 = {
  	__proto__: null,
  	"\"": true,
  	"'": true
  };


  function funescape$2( _, escaped, escapedWhitespace ) {
  	var high = "0x" + escaped - 0x10000;
  	
  	
  	
  	return high !== high || escapedWhitespace ?
  		escaped :
  		
  		high < 0 ?
  			String.fromCharCode( high + 0x10000 ) :
  			
  			String.fromCharCode( high >> 10 | 0xD800, high & 0x3FF | 0xDC00 );
  }

  function unescapeCSS$2(str){
  	return str.replace(re_escape$2, funescape$2);
  }

  function isWhitespace$2(c){
  	return c === " " || c === "\n" || c === "\t" || c === "\f" || c === "\r";
  }

  function parse$1$2(selector, options){
  	var subselects = [];

  	selector = parseSelector$2(subselects, selector + "", options);

  	if(selector !== ""){
  		throw new SyntaxError("Unmatched selector: " + selector);
  	}

  	return subselects;
  }

  function parseSelector$2(subselects, selector, options){
  	var tokens = [],
  		sawWS = false,
  		data, firstChar, name, quot;

  	function getName(){
  		var sub = selector.match(re_name$2)[0];
  		selector = selector.substr(sub.length);
  		return unescapeCSS$2(sub);
  	}

  	function stripWhitespace(start){
  		while(isWhitespace$2(selector.charAt(start))) start++;
  		selector = selector.substr(start);
  	}

  	stripWhitespace(0);

  	while(selector !== ""){
  		firstChar = selector.charAt(0);

  		if(isWhitespace$2(firstChar)){
  			sawWS = true;
  			stripWhitespace(1);
  		} else if(firstChar in simpleSelectors$2){
  			tokens.push({type: simpleSelectors$2[firstChar]});
  			sawWS = false;

  			stripWhitespace(1);
  		} else if(firstChar === ","){
  			if(tokens.length === 0){
  				throw new SyntaxError("empty sub-selector");
  			}
  			subselects.push(tokens);
  			tokens = [];
  			sawWS = false;
  			stripWhitespace(1);
  		} else {
  			if(sawWS){
  				if(tokens.length > 0){
  					tokens.push({type: "descendant"});
  				}
  				sawWS = false;
  			}

  			if(firstChar === "*"){
  				selector = selector.substr(1);
  				tokens.push({type: "universal"});
  			} else if(firstChar in attribSelectors$2){
  				selector = selector.substr(1);
  				tokens.push({
  					type: "attribute",
  					name: attribSelectors$2[firstChar][0],
  					action: attribSelectors$2[firstChar][1],
  					value: getName(),
  					ignoreCase: false
  				});
  			} else if(firstChar === "["){
  				selector = selector.substr(1);
  				data = selector.match(re_attr$2);
  				if(!data){
  					throw new SyntaxError("Malformed attribute selector: " + selector);
  				}
  				selector = selector.substr(data[0].length);
  				name = unescapeCSS$2(data[1]);

  				if(
  					!options || (
  						"lowerCaseAttributeNames" in options ?
  							options.lowerCaseAttributeNames :
  							!options.xmlMode
  					)
  				){
  					name = name.toLowerCase();
  				}

  				tokens.push({
  					type: "attribute",
  					name: name,
  					action: actionTypes$2[data[2]],
  					value: unescapeCSS$2(data[4] || data[5] || ""),
  					ignoreCase: !!data[6]
  				});

  			} else if(firstChar === ":"){
  				if(selector.charAt(1) === ":"){
  					selector = selector.substr(2);
  					tokens.push({type: "pseudo-element", name: getName().toLowerCase()});
  					continue;
  				}

  				selector = selector.substr(1);

  				name = getName().toLowerCase();
  				data = null;

  				if(selector.charAt(0) === "("){
  					if(name in unpackPseudos$2){
  						quot = selector.charAt(1);
  						var quoted = quot in quotes$2;

  						selector = selector.substr(quoted + 1);

  						data = [];
  						selector = parseSelector$2(data, selector, options);

  						if(quoted){
  							if(selector.charAt(0) !== quot){
  								throw new SyntaxError("unmatched quotes in :" + name);
  							} else {
  								selector = selector.substr(1);
  							}
  						}

  						if(selector.charAt(0) !== ")"){
  							throw new SyntaxError("missing closing parenthesis in :" + name + " " + selector);
  						}

  						selector = selector.substr(1);
  					} else {
  						var pos = 1, counter = 1;

  						for(; counter > 0 && pos < selector.length; pos++){
  							if(selector.charAt(pos) === "(") counter++;
  							else if(selector.charAt(pos) === ")") counter--;
  						}

  						if(counter){
  							throw new SyntaxError("parenthesis not matched");
  						}

  						data = selector.substr(1, pos - 2);
  						selector = selector.substr(pos);

  						if(name in stripQuotesFromPseudos$2){
  							quot = data.charAt(0);

  							if(quot === data.slice(-1) && quot in quotes$2){
  								data = data.slice(1, -1);
  							}

  							data = unescapeCSS$2(data);
  						}
  					}
  				}

  				tokens.push({type: "pseudo", name: name, data: data});
  			} else if(re_name$2.test(selector)){
  				name = getName();

  				if(!options || ("lowerCaseTags" in options ? options.lowerCaseTags : !options.xmlMode)){
  					name = name.toLowerCase();
  				}

  				tokens.push({type: "tag", name: name});
  			} else {
  				if(tokens.length && tokens[tokens.length - 1].type === "descendant"){
  					tokens.pop();
  				}
  				addToken$2(subselects, tokens);
  				return selector;
  			}
  		}
  	}

  	addToken$2(subselects, tokens);

  	return selector;
  }

  function addToken$2(subselects, tokens){
  	if(subselects.length > 0 && tokens.length === 0){
  		throw new SyntaxError("empty sub-selector");
  	}

  	subselects.push(tokens);
  }

  const _encodeXMLContent$2 = ((obj) => {
    let invObj = getInverseObj$3(obj);
    let replacer = getInverseReplacer$3(invObj);
    return getInverse$3(invObj, replacer)
  })({
    amp: '&',
    gt: '>',
    lt: '<'
  });

  const _encodeXMLAttr$2 = ((obj) => {
    let invObj = getInverseObj$3(obj);
    let replacer = getInverseReplacer$3(invObj);
    return getInverse$3(invObj, replacer)
  })({
    quot: '"'
  });

  function getInverseObj$3 (obj) {
    return Object.keys(obj).sort().reduce(function (inverse, name) {
      inverse[obj[name]] = '&' + name + ';';
      return inverse
    }, {})
  }

  function getInverseReplacer$3 (inverse) {
    let single = [];
    let multiple = [];

    Object.keys(inverse).forEach(function (k) {
      if (k.length === 1) {
        single.push('\\' + k);
      } else {
        multiple.push(k);
      }
    });

    multiple.unshift('[' + single.join('') + ']');

    return new RegExp(multiple.join('|'), 'g')
  }

  const RE_NON_ASCII$2 = /[^\0-\x7F]/g;
  const RE_ASTRAL_SYMBOLS$2 = /[\uD800-\uDBFF][\uDC00-\uDFFF]/g;

  function singleCharReplacer$3 (c) {
    return '&#x' + c.charCodeAt(0).toString(16).toUpperCase() + ';'
  }

  function astralReplacer$3 (c) {
    var high = c.charCodeAt(0);
    var low = c.charCodeAt(1);
    var codePoint = (high - 0xD800) * 0x400 + low - 0xDC00 + 0x10000;
    return '&#x' + codePoint.toString(16).toUpperCase() + ';'
  }

  function getInverse$3 (inverse, re) {
    function func (name) {
      return inverse[name]
    }
    return function (data) {
      return data
        .replace(re, func)
        .replace(RE_ASTRAL_SYMBOLS$2, astralReplacer$3)
        .replace(RE_NON_ASCII$2, singleCharReplacer$3)
    }
  }

  const booleanAttributes$2 = {
    __proto__: null,
    allowfullscreen: true,
    async: true,
    autofocus: true,
    autoplay: true,
    checked: true,
    controls: true,
    default: true,
    defer: true,
    disabled: true,
    hidden: true,
    ismap: true,
    loop: true,
    multiple: true,
    muted: true,
    open: true,
    readonly: true,
    required: true,
    reversed: true,
    scoped: true,
    seamless: true,
    selected: true,
    typemustmatch: true
  };

  const unencodedElements$2 = {
    __proto__: null,
    style: true,
    script: true,
    xmp: true,
    iframe: true,
    noembed: true,
    noframes: true,
    plaintext: true,
    noscript: true
  };

  const singleTag$2 = {
    __proto__: null,
    area: true,
    base: true,
    basefont: true,
    br: true,
    col: true,
    command: true,
    embed: true,
    frame: true,
    hr: true,
    img: true,
    input: true,
    isindex: true,
    keygen: true,
    link: true,
    meta: true,
    param: true,
    source: true,
    track: true,
    wbr: true
  };

  class DomUtils$2 {
    isTag (elem) {
      return domelementtype$2.isTag(elem)
    }

    removeElement (elem) {
      if (elem.prev) elem.prev.next = elem.next;
      if (elem.next) elem.next.prev = elem.prev;
      if (elem.parent) {
        var childs = elem.parent.childNodes;
        let pos = childs.lastIndexOf(elem);
        if (pos < 0) throw new Error('Invalid state')
        childs.splice(pos, 1);
        elem.parent = null;
      }
    }

    replaceElement (elem, replacement) {
      if (replacement.parent) this.removeElement(replacement);
      var prev = replacement.prev = elem.prev;
      if (prev) {
        prev.next = replacement;
      }

      var next = replacement.next = elem.next;
      if (next) {
        next.prev = replacement;
      }

      var parent = replacement.parent = elem.parent;
      if (parent) {
        var childs = parent.childNodes;
        let pos = childs.lastIndexOf(elem);
        if (pos < 0) throw new Error('Invalid state')
        childs[pos] = replacement;
      }
    }

    appendChild (elem, child) {
      if (child.parent) this.removeElement(child);
      child.parent = elem;

      if (elem.childNodes.push(child) !== 1) {
        var sibling = elem.childNodes[elem.childNodes.length - 2];
        sibling.next = child;
        child.prev = sibling;
        child.next = null;
      }
    }

    append (elem, next) {
      if (next.parent) this.removeElement(next);
      let parent = elem.parent;
      let currNext = elem.next;

      next.next = currNext;
      next.prev = elem;
      elem.next = next;
      next.parent = parent;

      if (currNext) {
        currNext.prev = next;
        if (parent) {
          var childs = parent.childNodes;
          let pos = childs.lastIndexOf(currNext);
          if (pos < 0) throw new Error('Invalid state')
          childs.splice(pos, 0, next);
        }
      } else if (parent) {
        parent.childNodes.push(next);
      }
    }

    prepend (elem, prev) {
      if (prev.parent) this.removeElement(prev);
      var parent = elem.parent;
      if (parent) {
        var childs = parent.childNodes;
        let pos = childs.lastIndexOf(elem);
        if (pos < 0) throw new Error('Invalid state')
        childs.splice(pos, 0, prev);
      }

      if (elem.prev) {
        elem.prev.next = prev;
      }

      prev.parent = parent;
      prev.prev = elem.prev;
      prev.next = elem;
      elem.prev = prev;
    }

    filter (test, element, recurse, limit) {
      if (!Array.isArray(element)) element = [element];

      if (typeof limit !== 'number' || !isFinite(limit)) {
        limit = Infinity;
      }
      return this.find(test, element, recurse !== false, limit)
    }

    find (test, elems, recurse, limit) {
      let result = [];
      let childs;

      for (var i = 0, j = elems.length; i < j; i++) {
        if (test(elems[i])) {
          result.push(elems[i]);
          if (--limit <= 0) break
        }

        childs = this.getChildren(elems[i]);
        if (recurse && childs && childs.length > 0) {
          childs = this.find(test, childs, recurse, limit);
          result = result.concat(childs);
          limit -= childs.length;
          if (limit <= 0) break
        }
      }

      return result
    }

    findOneChild (test, elems) {
      for (var i = 0, l = elems.length; i < l; i++) {
        if (test(elems[i])) return elems[i]
      }

      return null
    }

    findOne (test, elems) {
      var elem = null;

      for (var i = 0, l = elems.length; i < l && !elem; i++) {
        const child = elems[i];
        if (!this.isTag(child)) {
          continue
        } else if (test(child)) {
          elem = child;
        } else {
          const childNodes = this.getChildren(child);
          if (childNodes.length > 0) {
            elem = this.findOne(test, childNodes);
          }
        }
      }

      return elem
    }

    existsOne (test, elems) {
      for (var i = 0, l = elems.length; i < l; i++) {
        const elem = elems[i];
        
        if (!this.isTag(elem)) continue
        
        if (test(elem)) return true
        
        const childNodes = this.getChildren(elem);
        if (childNodes.length > 0 && this.existsOne(test, childNodes)) return true
      }
      return false
    }

    findAll (test, elems) {
      var result = [];
      for (var i = 0, j = elems.length; i < j; i++) {
        const elem = elems[i];
        if (!this.isTag(elem)) continue
        if (test(elem)) result.push(elem);
        const childNodes = this.getChildren(elem);
        if (childNodes.length > 0) {
          result = result.concat(this.findAll(test, childNodes));
        }
      }
      return result
    }

    getAttributes (el) {
      let attribs = el.getAttributes();
      
      
      if (attribs instanceof Map) {
        return Array.from(attribs)
      } else if (attribs && attribs.forEach) {
        let res = [];
        attribs.forEach((val, key) => {
          res.push([key, val]);
        });
        return res
      } else {
        return []
      }
    }

    formatAttribs (el, opts = {}) {
      let output = [];
      const attributes = this.getAttributes(el);
      attributes.forEach(([key, value]) => {
        if (!value && booleanAttributes$2[key]) {
          output.push(key);
        } else {
          output.push(key + '="' + (opts.decodeEntities ? _encodeXMLAttr$2(value) : value) + '"');
        }
      });
      return output.join(' ')
    }

    render (dom, opts) {
      if (!Array.isArray(dom)) dom = [dom];
      opts = opts || {};
      let output = [];
      for (var i = 0; i < dom.length; i++) {
        let elem = dom[i];
        switch (elem.type) {
          case 'root':
          case 'document': {
            output.push(this.render(this.getChildren(elem), opts));
            break
          }
          case domelementtype$2.Tag:
          case domelementtype$2.Script:
          case domelementtype$2.Style: {
            output.push(this.renderTag(elem, opts));
            break
          }
          case domelementtype$2.CDATA: {
            output.push(this.renderCdata(elem));
            break
          }
          case domelementtype$2.Comment: {
            output.push(this.renderComment(elem));
            break
          }
          case domelementtype$2.Directive: {
            output.push(this.renderDirective(elem));
            break
          }
          case domelementtype$2.Doctype: {
            output.push(this.renderDoctype(elem));
            break
          }
          case domelementtype$2.Text: {
            output.push(this.renderText(elem, opts));
            break
          }
          default:
            throw new Error('Not implement yet: render of element type ' + elem.type)
        }
      }
      return output.join('')
    }

    renderTag (elem, opts) {
      const name = this.getName(elem);
      if (name === 'svg') opts = {decodeEntities: opts.decodeEntities, xmlMode: true};
      let tag = '<' + name;
      let attribs = this.formatAttribs(elem, opts);
      if (attribs) {
        tag += ' ' + attribs;
      }
      const childNodes = this.getChildren(elem);
      if (opts.xmlMode && childNodes.length === 0) {
        tag += '/>';
      } else {
        tag += '>';
        if (childNodes.length > 0) {
          tag += this.render(childNodes, opts);
        }
        if (!singleTag$2[name] || opts.xmlMode) {
          tag += '</' + name + '>';
        }
      }
      return tag
    }

    renderDirective (elem) {
      return '<?' + this.getData(elem) + '?>'
    }

    renderDoctype (elem) {
      const { name, publicId, systemId } = this.getData(elem);
      let frags = ['DOCTYPE', name];
      if (publicId) {
        frags.push('PUBLIC');
        frags.push('"' + publicId + '"');
        if (systemId) frags.push('"' + systemId + '"');
      }
      return '<!' + frags.join(' ') + '>'
    }

    renderText (elem, opts) {
      let text = this.getText(elem);
      if (opts.decodeEntities) {
        const parent = this.getParent(elem);
        if (!(parent && this.getName(parent) in unencodedElements$2)) {
          text = _encodeXMLContent$2(text);
        }
      }
      return text
    }

    renderCdata (elem) {
      return '<![CDATA[' + this.getData(elem) + ']]>'
    }

    renderComment (elem) {
      return '<!--' + this.getData(elem) + '-->'
    }

    getInnerHTML (elem, opts) {
      const childNodes = this.getChildren(elem);
      return childNodes.map((child) => {
        return this.render(child, opts)
      }).join('')
    }

    getOuterHTML (elem, opts) {
      return this.render(elem, opts)
    }

    getData (elem) {
      return elem.data
    }

    getText (elem, sub) {
      if (Array.isArray(elem)) return elem.map(e => this.getText(e, sub)).join('')
      switch (elem.type) {
        case domelementtype$2.Tag:
        case domelementtype$2.Script:
        case domelementtype$2.Style:
          return this.getText(this.getChildren(elem), true)
        case domelementtype$2.Text:
        case domelementtype$2.CDATA:
          return elem.data
        case domelementtype$2.Comment:
          
          
          if (sub) {
            return ''
          }
          return elem.data
        default:
          return ''
      }
    }

    getChildren (elem) {
      return elem.childNodes
    }

    getParent (elem) {
      return elem.parent
    }

    getSiblings (elem) {
      var parent = this.getParent(elem);
      return parent ? this.getChildren(parent) : [elem]
    }

    getAttributeValue (elem, name) {
      return elem.getAttribute(name)
    }

    hasAttrib (elem, name) {
      return elem.hasAttribute(name)
    }

    getName (elem) {
      return elem.name
    }

    getNameWithoutNS (elem) {
      return elem.nameWithoutNS
    }
  }

  const domUtils$2 = new DomUtils$2();
  domUtils$2.DomUtils = DomUtils$2;

  var universal$2 = 50;
  var tag$2 = 30;
  var attribute$2 = 1;
  var pseudo$2 = 0;
  var descendant$2 = -1;
  var child$2 = -1;
  var parent$1$2 = -1;
  var sibling$2 = -1;
  var adjacent$2 = -1;
  var procedure$4 = {
  	universal: universal$2,
  	tag: tag$2,
  	attribute: attribute$2,
  	pseudo: pseudo$2,
  	descendant: descendant$2,
  	child: child$2,
  	parent: parent$1$2,
  	sibling: sibling$2,
  	adjacent: adjacent$2
  };

  var procedure$1$2 = Object.freeze({
    universal: universal$2,
    tag: tag$2,
    attribute: attribute$2,
    pseudo: pseudo$2,
    descendant: descendant$2,
    child: child$2,
    parent: parent$1$2,
    sibling: sibling$2,
    adjacent: adjacent$2,
    default: procedure$4
  });

  var procedure$2$2 = ( procedure$1$2 && procedure$4 ) || procedure$1$2;

  var sort$2 = sortByProcedure$2;





  var attributes$3 = {
  	__proto__: null,
  	exists: 10,
  	equals: 8,
  	not: 7,
  	start: 6,
  	end: 6,
  	any: 5,
  	hyphen: 4,
  	element: 4
  };

  function sortByProcedure$2(arr){
  	var procs = arr.map(getProcedure$2);
  	for(var i = 1; i < arr.length; i++){
  		var procNew = procs[i];

  		if(procNew < 0) continue;

  		for(var j = i - 1; j >= 0 && procNew < procs[j]; j--){
  			var token = arr[j + 1];
  			arr[j + 1] = arr[j];
  			arr[j] = token;
  			procs[j + 1] = procs[j];
  			procs[j] = procNew;
  		}
  	}
  }

  function getProcedure$2(token){
  	var proc = procedure$2$2[token.type];

  	if(proc === procedure$2$2.attribute){
  		proc = attributes$3[token.action];

  		if(proc === attributes$3.equals && token.name === "id"){
  			
  			proc = 9;
  		}

  		if(token.ignoreCase){
  			
  			
  			proc >>= 1;
  		}
  	} else if(proc === procedure$2$2.pseudo){
  		if(!token.data){
  			proc = 3;
  		} else if(token.name === "has" || token.name === "contains"){
  			proc = 0; 
  		} else if(token.name === "matches" || token.name === "not"){
  			proc = 0;
  			for(var i = 0; i < token.data.length; i++){
  				
  				if(token.data[i].length !== 1) continue;
  				var cur = getProcedure$2(token.data[i][0]);
  				
  				if(cur === 0){
  					proc = 0;
  					break;
  				}
  				if(cur > proc) proc = cur;
  			}
  			if(token.data.length > 1 && proc > 0) proc -= 1;
  		} else {
  			proc = 1;
  		}
  	}
  	return proc;
  }

  var falseFunc$1$3 = boolbase$2.falseFunc;


  var reChars$2 = /[-[\]{}()*+?.,\\^$|#\s]/g;

  function factory$3(adapter){
  	
  	var attributeRules = {
  		__proto__: null,
  		equals: function(next, data){
  			var name  = data.name,
  				value = data.value;

  			if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function equalsIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && attr.toLowerCase() === value && next(elem);
  				};
  			}

  			return function equals(elem){
  				return adapter.getAttributeValue(elem, name) === value && next(elem);
  			};
  		},
  		hyphen: function(next, data){
  			var name  = data.name,
  				value = data.value,
  				len = value.length;

  			if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function hyphenIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null &&
  							(attr.length === len || attr.charAt(len) === "-") &&
  							attr.substr(0, len).toLowerCase() === value &&
  							next(elem);
  				};
  			}

  			return function hyphen(elem){
  				var attr = adapter.getAttributeValue(elem, name);
  				return attr != null &&
  						attr.substr(0, len) === value &&
  						(attr.length === len || attr.charAt(len) === "-") &&
  						next(elem);
  			};
  		},
  		element: function(next, data){
  			var name = data.name,
  				value = data.value;
  			if (data.name === 'class') {
  				let value = data.value;
  				if (/\s/.test(value)) return function() { return false }
  				return function clazz(elem) {
  					let classes = elem.classes;
  					return classes && classes.has(value) && next(elem)
  				}
  			} else {
  				if(/\s/.test(value)){
  					return falseFunc$1$3;
  				}

  				value = value.replace(reChars$2, "\\$&");

  				var pattern = "(?:^|\\s)" + value + "(?:$|\\s)",
  					flags = data.ignoreCase ? "i" : "",
  					regex = new RegExp(pattern, flags);

  				return function element(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && regex.test(attr) && next(elem);
  				};
  			}
  		},
  		exists: function(next, data){
  			var name = data.name;
  			return function exists(elem){
  				return adapter.hasAttrib(elem, name) && next(elem);
  			};
  		},
  		start: function(next, data){
  			var name  = data.name,
  				value = data.value,
  				len = value.length;

  			if(len === 0){
  				return falseFunc$1$3;
  			}

  			if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function startIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && attr.substr(0, len).toLowerCase() === value && next(elem);
  				};
  			}

  			return function start(elem){
  				var attr = adapter.getAttributeValue(elem, name);
  				return attr != null && attr.substr(0, len) === value && next(elem);
  			};
  		},
  		end: function(next, data){
  			var name  = data.name,
  				value = data.value,
  				len   = -value.length;

  			if(len === 0){
  				return falseFunc$1$3;
  			}

  			if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function endIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && attr.substr(len).toLowerCase() === value && next(elem);
  				};
  			}

  			return function end(elem){
  				var attr = adapter.getAttributeValue(elem, name);
  				return attr != null && attr.substr(len) === value && next(elem);
  			};
  		},
  		any: function(next, data){
  			var name  = data.name,
  				value = data.value;

  			if(value === ""){
  				return falseFunc$1$3;
  			}

  			if(data.ignoreCase){
  				var regex = new RegExp(value.replace(reChars$2, "\\$&"), "i");

  				return function anyIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && regex.test(attr) && next(elem);
  				};
  			}

  			return function any(elem){
  				var attr = adapter.getAttributeValue(elem, name);
  				return attr != null && attr.indexOf(value) >= 0 && next(elem);
  			};
  		},
  		not: function(next, data){
  			var name  = data.name,
  				value = data.value;

  			if(value === ""){
  				return function notEmpty(elem){
  					return !!adapter.getAttributeValue(elem, name) && next(elem);
  				};
  			} else if(data.ignoreCase){
  				value = value.toLowerCase();

  				return function notIC(elem){
  					var attr = adapter.getAttributeValue(elem, name);
  					return attr != null && attr.toLowerCase() !== value && next(elem);
  				};
  			}

  			return function not(elem){
  				return adapter.getAttributeValue(elem, name) !== value && next(elem);
  			};
  		}
  	};

  	return {
  		compile: function(next, data, options){
  			if(options && options.strict && (
  				data.ignoreCase || data.action === "not"
  			)) throw new Error("Unsupported attribute selector");
  			return attributeRules[data.action](next, data);
  		},
  		rules: attributeRules
  	};
  }

  var attributes$1$2 = factory$3;

  function generalFactory$2(adapter, Pseudos){
  	
  	return {
  		__proto__: null,

  		attribute: attributes$1$2(adapter).compile,
  		pseudo: Pseudos.compile,

  		
  		tag: function(next, data){
  			var name = data.name;
  			return function tag(elem){
  				return adapter.getNameWithoutNS(elem) === name && next(elem);
  			}
  		},

  		
  		descendant: function(next){
  			return function descendant(elem){

  				var found = false;

  				while(!found && (elem = adapter.getParent(elem))){
  					found = next(elem);
  				}

  				return found;
  			};
  		},
  		_flexibleDescendant: function(next){
  			
  			return function descendant(elem){

  				var found = next(elem);

  				while(!found && (elem = adapter.getParent(elem))){
  					found = next(elem);
  				}

  				return found;
  			};
  		},
  		parent: function(next, data, options){
  			if(options && options.strict) throw new Error("Parent selector isn't part of CSS3");

  			return function parent(elem){
  				return adapter.getChildren(elem).some(test);
  			};

  			function test(elem){
  				return adapter.isTag(elem) && next(elem);
  			}
  		},
  		child: function(next){
  			return function child(elem){
  				var parent = adapter.getParent(elem);
  				return !!parent && next(parent);
  			};
  		},
  		sibling: function(next){
  			return function sibling(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var i = 0; i < siblings.length; i++){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) break;
  						if(next(siblings[i])) return true;
  					}
  				}

  				return false;
  			};
  		},
  		adjacent: function(next){
  			return function adjacent(elem){
  				var siblings = adapter.getSiblings(elem),
  					lastElement;

  				for(var i = 0; i < siblings.length; i++){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) break;
  						lastElement = siblings[i];
  					}
  				}

  				return !!lastElement && next(lastElement);
  			};
  		},
  		universal: function(next){
  			return next;
  		}
  	};
  }

  var general$2 = generalFactory$2;

  var trueFunc$1$3          = boolbase$2.trueFunc,
  	falseFunc$1$1$2         = boolbase$2.falseFunc;

  function filtersFactory$2(adapter){
  	var attributes  = attributes$1$2(adapter),
  		checkAttrib = attributes.rules.equals;

  	
  	function equals(a, b){
  		if(typeof adapter.equals === "function") return adapter.equals(a, b);

  		return a === b;
  	}

  	function getAttribFunc(name, value){
  		var data = {name: name, value: value};
  		return function attribFunc(next){
  			return checkAttrib(next, data);
  		};
  	}

  	function getChildFunc(next){
  		return function(elem){
  			return !!adapter.getParent(elem) && next(elem);
  		};
  	}

  	var filters = {
  		contains: function(next, text){
  			return function contains(elem){
  				return next(elem) && adapter.getText(elem).indexOf(text) >= 0;
  			};
  		},
  		icontains: function(next, text){
  			var itext = text.toLowerCase();
  			return function icontains(elem){
  				return next(elem) &&
  					adapter.getText(elem).toLowerCase().indexOf(itext) >= 0;
  			};
  		},

  		
  		"nth-child": function(next, rule){
  			var func = nthCheck$2(rule);

  			if(func === falseFunc$1$1$2) return func;
  			if(func === trueFunc$1$3)  return getChildFunc(next);

  			return function nthChild(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var i = 0, pos = 0; i < siblings.length; i++){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) break;
  						else pos++;
  					}
  				}

  				return func(pos) && next(elem);
  			};
  		},
  		"nth-last-child": function(next, rule){
  			var func = nthCheck$2(rule);

  			if(func === falseFunc$1$1$2) return func;
  			if(func === trueFunc$1$3)  return getChildFunc(next);

  			return function nthLastChild(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var pos = 0, i = siblings.length - 1; i >= 0; i--){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) break;
  						else pos++;
  					}
  				}

  				return func(pos) && next(elem);
  			};
  		},
  		"nth-of-type": function(next, rule){
  			var func = nthCheck$2(rule);

  			if(func === falseFunc$1$1$2) return func;
  			if(func === trueFunc$1$3)  return getChildFunc(next);

  			return function nthOfType(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var pos = 0, i = 0; i < siblings.length; i++){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) break;
  						if(adapter.getName(siblings[i]) === adapter.getName(elem)) pos++;
  					}
  				}

  				return func(pos) && next(elem);
  			};
  		},
  		"nth-last-of-type": function(next, rule){
  			var func = nthCheck$2(rule);

  			if(func === falseFunc$1$1$2) return func;
  			if(func === trueFunc$1$3)  return getChildFunc(next);

  			return function nthLastOfType(elem){
  				var siblings = adapter.getSiblings(elem);

  				for(var pos = 0, i = siblings.length - 1; i >= 0; i--){
  					if(adapter.isTag(siblings[i])){
  						if(siblings[i] === elem) break;
  						if(adapter.getName(siblings[i]) === adapter.getName(elem)) pos++;
  					}
  				}

  				return func(pos) && next(elem);
  			};
  		},

  		
  		root: function(next){
  			return function(elem){
  				return !adapter.getParent(elem) && next(elem);
  			};
  		},

  		scope: function(next, rule, options, context){
  			if(!context || context.length === 0){
  				
  				return filters.root(next);
  			}

  			if(context.length === 1){
  				
  				return function(elem){
  					return equals(context[0], elem) && next(elem);
  				};
  			}

  			return function(elem){
  				return context.indexOf(elem) >= 0 && next(elem);
  			};
  		},

  		
  		checkbox: getAttribFunc("type", "checkbox"),
  		file: getAttribFunc("type", "file"),
  		password: getAttribFunc("type", "password"),
  		radio: getAttribFunc("type", "radio"),
  		reset: getAttribFunc("type", "reset"),
  		image: getAttribFunc("type", "image"),
  		submit: getAttribFunc("type", "submit")
  	};
  	return filters;
  }

  function pseudosFactory$2(adapter){
  	
  	function getFirstElement(elems){
  		for(var i = 0; elems && i < elems.length; i++){
  			if(adapter.isTag(elems[i])) return elems[i];
  		}
  	}

  	
  	var pseudos = {
  		empty: function(elem){
  			return !adapter.getChildren(elem).some(function(elem){
  				return adapter.isTag(elem) || elem.type === "text";
  			});
  		},

  		"first-child": function(elem){
  			return getFirstElement(adapter.getSiblings(elem)) === elem;
  		},
  		"last-child": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = siblings.length - 1; i >= 0; i--){
  				if(siblings[i] === elem) return true;
  				if(adapter.isTag(siblings[i])) break;
  			}

  			return false;
  		},
  		"first-of-type": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = 0; i < siblings.length; i++){
  				if(adapter.isTag(siblings[i])){
  					if(siblings[i] === elem) return true;
  					if(adapter.getName(siblings[i]) === adapter.getName(elem)) break;
  				}
  			}

  			return false;
  		},
  		"last-of-type": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = siblings.length - 1; i >= 0; i--){
  				if(adapter.isTag(siblings[i])){
  					if(siblings[i] === elem) return true;
  					if(adapter.getName(siblings[i]) === adapter.getName(elem)) break;
  				}
  			}

  			return false;
  		},
  		"only-of-type": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = 0, j = siblings.length; i < j; i++){
  				if(adapter.isTag(siblings[i])){
  					if(siblings[i] === elem) continue;
  					if(adapter.getName(siblings[i]) === adapter.getName(elem)) return false;
  				}
  			}

  			return true;
  		},
  		"only-child": function(elem){
  			var siblings = adapter.getSiblings(elem);

  			for(var i = 0; i < siblings.length; i++){
  				if(adapter.isTag(siblings[i]) && siblings[i] !== elem) return false;
  			}

  			return true;
  		},

  		
  		link: function(elem){
  			return adapter.hasAttrib(elem, "href");
  		},
  		visited: falseFunc$1$1$2, 
  		

  		
  		

  		
  		selected: function(elem){
  			if(adapter.hasAttrib(elem, "selected")) return true;
  			else if(adapter.getName(elem) !== "option") return false;

  			
  			var parent = adapter.getParent(elem);

  			if(
  				!parent ||
  				adapter.getName(parent) !== "select" ||
  				adapter.hasAttrib(parent, "multiple")
  			) return false;

  			var siblings = adapter.getChildren(parent),
  				sawElem  = false;

  			for(var i = 0; i < siblings.length; i++){
  				if(adapter.isTag(siblings[i])){
  					if(siblings[i] === elem){
  						sawElem = true;
  					} else if(!sawElem){
  						return false;
  					} else if(adapter.hasAttrib(siblings[i], "selected")){
  						return false;
  					}
  				}
  			}

  			return sawElem;
  		},
  		
  		
  		
  		
  		
  		
  		disabled: function(elem){
  			return adapter.hasAttrib(elem, "disabled");
  		},
  		enabled: function(elem){
  			return !adapter.hasAttrib(elem, "disabled");
  		},
  		
  		checked: function(elem){
  			return adapter.hasAttrib(elem, "checked") || pseudos.selected(elem);
  		},
  		
  		required: function(elem){
  			return adapter.hasAttrib(elem, "required");
  		},
  		
  		optional: function(elem){
  			return !adapter.hasAttrib(elem, "required");
  		},

  		

  		
  		parent: function(elem){
  			return !pseudos.empty(elem);
  		},
  		
  		header: function(elem){
  			var name = adapter.getName(elem);
  			return name === "h1" ||
  					name === "h2" ||
  					name === "h3" ||
  					name === "h4" ||
  					name === "h5" ||
  					name === "h6";
  		},

  		
  		button: function(elem){
  			var name = adapter.getName(elem);
  			return name === "button" ||
  					name === "input" &&
  					adapter.getAttributeValue(elem, "type") === "button";
  		},
  		
  		input: function(elem){
  			var name = adapter.getName(elem);
  			return name === "input" ||
  					name === "textarea" ||
  					name === "select" ||
  					name === "button";
  		},
  		
  		text: function(elem){
  			var attr;
  			return adapter.getName(elem) === "input" && (
  				!(attr = adapter.getAttributeValue(elem, "type")) ||
  				attr.toLowerCase() === "text"
  			);
  		}
  	};

  	return pseudos;
  }

  function verifyArgs$2(func, name, subselect){
  	if(subselect === null){
  		if(func.length > 1 && name !== "scope"){
  			throw new Error("pseudo-selector :" + name + " requires an argument");
  		}
  	} else {
  		if(func.length === 1){
  			throw new Error("pseudo-selector :" + name + " doesn't have any arguments");
  		}
  	}
  }


  var re_CSS3$2 = /^(?:(?:nth|last|first|only)-(?:child|of-type)|root|empty|(?:en|dis)abled|checked|not)$/;

  function factory$1$2(adapter){
  	var pseudos = pseudosFactory$2(adapter);
  	var filters = filtersFactory$2(adapter);

  	return {
  		compile: function(next, data, options, context){
  			var name = data.name,
  				subselect = data.data;

  			if(options && options.strict && !re_CSS3$2.test(name)){
  				throw new Error(":" + name + " isn't part of CSS3");
  			}

  			if(typeof filters[name] === "function"){
  				verifyArgs$2(filters[name], name,  subselect);
  				return filters[name](next, subselect, options, context);
  			} else if(typeof pseudos[name] === "function"){
  				var func = pseudos[name];
  				verifyArgs$2(func, name, subselect);

  				if(next === trueFunc$1$3) return func;

  				return function pseudoArgs(elem){
  					return func(elem, subselect) && next(elem);
  				};
  			} else {
  				throw new Error("unmatched pseudo-class :" + name);
  			}
  		},
  		filters: filters,
  		pseudos: pseudos
  	};
  }

  var pseudos$2 = factory$1$2;

  var compile$1$2 = compileFactory$2;

  var trueFunc$1$1$2       = boolbase$2.trueFunc,
  	falseFunc$2$2      = boolbase$2.falseFunc;

  function compileFactory$2(adapter){
  	var Pseudos     = pseudos$2(adapter),
  		filters     = Pseudos.filters,
  		Rules 			= general$2(adapter, Pseudos);

  	function compile(selector, options, context){
  		var next = compileUnsafe(selector, options, context);
  		return wrap(next);
  	}

  	function wrap(next){
  		return function base(elem){
  			return adapter.isTag(elem) && next(elem);
  		};
  	}

  	function compileUnsafe(selector, options, context){
  		var token = cssWhat$2(selector, options);
  		return compileToken(token, options, context);
  	}

  	function includesScopePseudo(t){
  		return t.type === "pseudo" && (
  			t.name === "scope" || (
  				Array.isArray(t.data) &&
  				t.data.some(function(data){
  					return data.some(includesScopePseudo);
  				})
  			)
  		);
  	}

  	var DESCENDANT_TOKEN = {type: "descendant"},
  		FLEXIBLE_DESCENDANT_TOKEN = {type: "_flexibleDescendant"},
  		SCOPE_TOKEN = {type: "pseudo", name: "scope"},
  		PLACEHOLDER_ELEMENT = {};

  	
  	
  	function absolutize(token, context){
  		
  		var hasContext = !!context && !!context.length && context.every(function(e){
  			return e === PLACEHOLDER_ELEMENT || !!adapter.getParent(e);
  		});


  		token.forEach(function(t){
  			if(t.length > 0 && isTraversal(t[0]) && t[0].type !== "descendant"); else if(hasContext && !includesScopePseudo(t)){
  				t.unshift(DESCENDANT_TOKEN);
  			} else {
  				return;
  			}

  			t.unshift(SCOPE_TOKEN);
  		});
  	}

  	function compileToken(token, options, context){
  		token = token.filter(function(t){ return t.length > 0; });

  		token.forEach(sort$2);

  		var isArrayContext = Array.isArray(context);

  		context = (options && options.context) || context;

  		if(context && !isArrayContext) context = [context];

  		absolutize(token, context);

  		var shouldTestNextSiblings = false;

  		var query = token
  			.map(function(rules){
  				if(rules[0] && rules[1] && rules[0].name === "scope"){
  					var ruleType = rules[1].type;
  					if(isArrayContext && ruleType === "descendant") rules[1] = FLEXIBLE_DESCENDANT_TOKEN;
  					else if(ruleType === "adjacent" || ruleType === "sibling") shouldTestNextSiblings = true;
  				}
  				return compileRules(rules, options, context);
  			})
  			.reduce(reduceRules, falseFunc$2$2);

  		query.shouldTestNextSiblings = shouldTestNextSiblings;

  		return query;
  	}

  	function isTraversal(t){
  		return procedure$2$2[t.type] < 0;
  	}

  	function compileRules(rules, options, context){
  		return rules.reduce(function(func, rule){
  			if(func === falseFunc$2$2) return func;
  			return Rules[rule.type](func, rule, options, context);
  		}, options && options.rootFunc || trueFunc$1$1$2);
  	}

  	function reduceRules(a, b){
  		if(b === falseFunc$2$2 || a === trueFunc$1$1$2){
  			return a;
  		}
  		if(a === falseFunc$2$2 || b === trueFunc$1$1$2){
  			return b;
  		}

  		return function combine(elem){
  			return a(elem) || b(elem);
  		};
  	}

  	function containsTraversal(t){
  		return t.some(isTraversal);
  	}

  	
  	
  	
  	filters.not = function(next, token, options, context){
  		var opts = {
  			xmlMode: !!(options && options.xmlMode),
  			strict: !!(options && options.strict)
  		};

  		if(opts.strict){
  			if(token.length > 1 || token.some(containsTraversal)){
  				throw new Error("complex selectors in :not aren't allowed in strict mode");
  			}
  		}

  		var func = compileToken(token, opts, context);

  		if(func === falseFunc$2$2) return next;
  		if(func === trueFunc$1$1$2)  return falseFunc$2$2;

  		return function(elem){
  			return !func(elem) && next(elem);
  		};
  	};

  	filters.has = function(next, token, options){
  		var opts = {
  			xmlMode: !!(options && options.xmlMode),
  			strict: !!(options && options.strict)
  		};

  		
  		var context = token.some(containsTraversal) ? [PLACEHOLDER_ELEMENT] : null;

  		var func = compileToken(token, opts, context);

  		if(func === falseFunc$2$2) return falseFunc$2$2;
  		if(func === trueFunc$1$1$2){
  			return function(elem){
  				return adapter.getChildren(elem).some(adapter.isTag) && next(elem);
  			};
  		}

  		func = wrap(func);

  		if(context){
  			return function has(elem){
  				return next(elem) && (
  					(context[0] = elem), adapter.existsOne(func, adapter.getChildren(elem))
  				);
  			};
  		}

  		return function has(elem){
  			return next(elem) && adapter.existsOne(func, adapter.getChildren(elem));
  		};
  	};

  	filters.matches = function(next, token, options, context){
  		var opts = {
  			xmlMode: !!(options && options.xmlMode),
  			strict: !!(options && options.strict),
  			rootFunc: next
  		};

  		return compileToken(token, opts, context);
  	};

  	compile.compileToken = compileToken;
  	compile.compileUnsafe = compileUnsafe;
  	compile.Pseudos = Pseudos;

  	return compile;
  }

  var cssSelect$2 = CSSselect$2;

  var falseFunc$3$2      = boolbase$2.falseFunc,
  	defaultCompile$2 = compile$1$2(domUtils$2);

  function adapterCompile$2(adapter){
  	if(!adapter.__compile__){
  		adapter.__compile__ = compile$1$2(adapter);
  	}
  	return adapter.__compile__
  }

  function getSelectorFunc$2(searchFunc){
  	return function select(query, elems, options){
  		options = options || {};
  		options.adapter = options.adapter || domUtils$2;
  		var compile$$1 = adapterCompile$2(options.adapter);

  		if(typeof query !== "function") query = compile$$1.compileUnsafe(query, options, elems);
  		if(query.shouldTestNextSiblings) elems = appendNextSiblings$2((options && options.context) || elems, options.adapter);
  		if(!Array.isArray(elems)) elems = options.adapter.getChildren(elems);
  		else elems = options.adapter.removeSubsets(elems);
  		return searchFunc(query, elems, options);
  	};
  }

  function getNextSiblings$2(elem, adapter){
  	var siblings = adapter.getSiblings(elem);
  	if(!Array.isArray(siblings)) return [];
  	siblings = siblings.slice(0);
  	while(siblings.shift() !== elem);
  	return siblings;
  }

  function appendNextSiblings$2(elems, adapter){
  	
  	if(!Array.isArray(elems)) elems = [elems];
  	var newElems = elems.slice(0);

  	for(var i = 0, len = elems.length; i < len; i++){
  		var nextSiblings = getNextSiblings$2(newElems[i], adapter);
  		newElems.push.apply(newElems, nextSiblings);
  	}
  	return newElems;
  }

  var selectAll$2 = getSelectorFunc$2(function selectAll(query, elems, options){
  	return (query === falseFunc$3$2 || !elems || elems.length === 0) ? [] : options.adapter.findAll(query, elems);
  });

  var selectOne$2 = getSelectorFunc$2(function selectOne(query, elems, options){
  	return (query === falseFunc$3$2 || !elems || elems.length === 0) ? null : options.adapter.findOne(query, elems);
  });

  function is$2(elem, query, options){
  	options = options || {};
  	options.adapter = options.adapter || domUtils$2;
  	var compile$$1 = adapterCompile$2(options.adapter);
  	return (typeof query === "function" ? query : compile$$1(query, options))(elem);
  }


  function CSSselect$2(query, elems, options){
  	return selectAll$2(query, elems, options);
  }

  CSSselect$2.compile = defaultCompile$2;
  CSSselect$2.filters = defaultCompile$2.Pseudos.filters;
  CSSselect$2.pseudos = defaultCompile$2.Pseudos.pseudos;

  CSSselect$2.selectAll = selectAll$2;
  CSSselect$2.selectOne = selectOne$2;

  CSSselect$2.is = is$2;


  CSSselect$2.parse = defaultCompile$2;
  CSSselect$2.iterate = selectAll$2;


  CSSselect$2._compileUnsafe = defaultCompile$2.compileUnsafe;
  CSSselect$2._compileToken = defaultCompile$2.compileToken;

  var amp$4 = "&";
  var apos$3 = "'";
  var gt$4 = ">";
  var lt$4 = "<";
  var quot$4 = "\"";
  var xmlJSON$2 = {
  	amp: amp$4,
  	apos: apos$3,
  	gt: gt$4,
  	lt: lt$4,
  	quot: quot$4
  };

  var xml$2 = Object.freeze({
  	amp: amp$4,
  	apos: apos$3,
  	gt: gt$4,
  	lt: lt$4,
  	quot: quot$4,
  	default: xmlJSON$2
  });

  var Aacute$3 = "Á";
  var aacute$3 = "á";
  var Abreve$2 = "Ă";
  var abreve$2 = "ă";
  var ac$2 = "∾";
  var acd$2 = "∿";
  var acE$2 = "∾̳";
  var Acirc$3 = "Â";
  var acirc$3 = "â";
  var acute$3 = "´";
  var Acy$2 = "А";
  var acy$2 = "а";
  var AElig$3 = "Æ";
  var aelig$3 = "æ";
  var af$2 = "⁡";
  var Afr$2 = "𝔄";
  var afr$2 = "𝔞";
  var Agrave$3 = "À";
  var agrave$3 = "à";
  var alefsym$2 = "ℵ";
  var aleph$2 = "ℵ";
  var Alpha$2 = "Α";
  var alpha$2 = "α";
  var Amacr$2 = "Ā";
  var amacr$2 = "ā";
  var amalg$2 = "⨿";
  var amp$1$2 = "&";
  var AMP$3 = "&";
  var andand$2 = "⩕";
  var And$2 = "⩓";
  var and$2 = "∧";
  var andd$2 = "⩜";
  var andslope$2 = "⩘";
  var andv$2 = "⩚";
  var ang$2 = "∠";
  var ange$2 = "⦤";
  var angle$2 = "∠";
  var angmsdaa$2 = "⦨";
  var angmsdab$2 = "⦩";
  var angmsdac$2 = "⦪";
  var angmsdad$2 = "⦫";
  var angmsdae$2 = "⦬";
  var angmsdaf$2 = "⦭";
  var angmsdag$2 = "⦮";
  var angmsdah$2 = "⦯";
  var angmsd$2 = "∡";
  var angrt$2 = "∟";
  var angrtvb$2 = "⊾";
  var angrtvbd$2 = "⦝";
  var angsph$2 = "∢";
  var angst$2 = "Å";
  var angzarr$2 = "⍼";
  var Aogon$2 = "Ą";
  var aogon$2 = "ą";
  var Aopf$2 = "𝔸";
  var aopf$2 = "𝕒";
  var apacir$2 = "⩯";
  var ap$2 = "≈";
  var apE$2 = "⩰";
  var ape$2 = "≊";
  var apid$2 = "≋";
  var apos$1$2 = "'";
  var ApplyFunction$2 = "⁡";
  var approx$2 = "≈";
  var approxeq$2 = "≊";
  var Aring$3 = "Å";
  var aring$3 = "å";
  var Ascr$2 = "𝒜";
  var ascr$2 = "𝒶";
  var Assign$2 = "≔";
  var ast$2 = "*";
  var asymp$2 = "≈";
  var asympeq$2 = "≍";
  var Atilde$3 = "Ã";
  var atilde$3 = "ã";
  var Auml$3 = "Ä";
  var auml$3 = "ä";
  var awconint$2 = "∳";
  var awint$2 = "⨑";
  var backcong$2 = "≌";
  var backepsilon$2 = "϶";
  var backprime$2 = "‵";
  var backsim$2 = "∽";
  var backsimeq$2 = "⋍";
  var Backslash$2 = "∖";
  var Barv$2 = "⫧";
  var barvee$2 = "⊽";
  var barwed$2 = "⌅";
  var Barwed$2 = "⌆";
  var barwedge$2 = "⌅";
  var bbrk$2 = "⎵";
  var bbrktbrk$2 = "⎶";
  var bcong$2 = "≌";
  var Bcy$2 = "Б";
  var bcy$2 = "б";
  var bdquo$2 = "„";
  var becaus$2 = "∵";
  var because$2 = "∵";
  var Because$2 = "∵";
  var bemptyv$2 = "⦰";
  var bepsi$2 = "϶";
  var bernou$2 = "ℬ";
  var Bernoullis$2 = "ℬ";
  var Beta$2 = "Β";
  var beta$2 = "β";
  var beth$2 = "ℶ";
  var between$2 = "≬";
  var Bfr$2 = "𝔅";
  var bfr$2 = "𝔟";
  var bigcap$2 = "⋂";
  var bigcirc$2 = "◯";
  var bigcup$2 = "⋃";
  var bigodot$2 = "⨀";
  var bigoplus$2 = "⨁";
  var bigotimes$2 = "⨂";
  var bigsqcup$2 = "⨆";
  var bigstar$2 = "★";
  var bigtriangledown$2 = "▽";
  var bigtriangleup$2 = "△";
  var biguplus$2 = "⨄";
  var bigvee$2 = "⋁";
  var bigwedge$2 = "⋀";
  var bkarow$2 = "⤍";
  var blacklozenge$2 = "⧫";
  var blacksquare$2 = "▪";
  var blacktriangle$2 = "▴";
  var blacktriangledown$2 = "▾";
  var blacktriangleleft$2 = "◂";
  var blacktriangleright$2 = "▸";
  var blank$2 = "␣";
  var blk12$2 = "▒";
  var blk14$2 = "░";
  var blk34$2 = "▓";
  var block$2 = "█";
  var bne$2 = "=⃥";
  var bnequiv$2 = "≡⃥";
  var bNot$2 = "⫭";
  var bnot$2 = "⌐";
  var Bopf$2 = "𝔹";
  var bopf$2 = "𝕓";
  var bot$2 = "⊥";
  var bottom$2 = "⊥";
  var bowtie$2 = "⋈";
  var boxbox$2 = "⧉";
  var boxdl$2 = "┐";
  var boxdL$2 = "╕";
  var boxDl$2 = "╖";
  var boxDL$2 = "╗";
  var boxdr$2 = "┌";
  var boxdR$2 = "╒";
  var boxDr$2 = "╓";
  var boxDR$2 = "╔";
  var boxh$2 = "─";
  var boxH$2 = "═";
  var boxhd$2 = "┬";
  var boxHd$2 = "╤";
  var boxhD$2 = "╥";
  var boxHD$2 = "╦";
  var boxhu$2 = "┴";
  var boxHu$2 = "╧";
  var boxhU$2 = "╨";
  var boxHU$2 = "╩";
  var boxminus$2 = "⊟";
  var boxplus$2 = "⊞";
  var boxtimes$2 = "⊠";
  var boxul$2 = "┘";
  var boxuL$2 = "╛";
  var boxUl$2 = "╜";
  var boxUL$2 = "╝";
  var boxur$2 = "└";
  var boxuR$2 = "╘";
  var boxUr$2 = "╙";
  var boxUR$2 = "╚";
  var boxv$2 = "│";
  var boxV$2 = "║";
  var boxvh$2 = "┼";
  var boxvH$2 = "╪";
  var boxVh$2 = "╫";
  var boxVH$2 = "╬";
  var boxvl$2 = "┤";
  var boxvL$2 = "╡";
  var boxVl$2 = "╢";
  var boxVL$2 = "╣";
  var boxvr$2 = "├";
  var boxvR$2 = "╞";
  var boxVr$2 = "╟";
  var boxVR$2 = "╠";
  var bprime$2 = "‵";
  var breve$2 = "˘";
  var Breve$2 = "˘";
  var brvbar$3 = "¦";
  var bscr$2 = "𝒷";
  var Bscr$2 = "ℬ";
  var bsemi$2 = "⁏";
  var bsim$2 = "∽";
  var bsime$2 = "⋍";
  var bsolb$2 = "⧅";
  var bsol$2 = "\\";
  var bsolhsub$2 = "⟈";
  var bull$2 = "•";
  var bullet$2 = "•";
  var bump$2 = "≎";
  var bumpE$2 = "⪮";
  var bumpe$2 = "≏";
  var Bumpeq$2 = "≎";
  var bumpeq$2 = "≏";
  var Cacute$2 = "Ć";
  var cacute$2 = "ć";
  var capand$2 = "⩄";
  var capbrcup$2 = "⩉";
  var capcap$2 = "⩋";
  var cap$2 = "∩";
  var Cap$2 = "⋒";
  var capcup$2 = "⩇";
  var capdot$2 = "⩀";
  var CapitalDifferentialD$2 = "ⅅ";
  var caps$2 = "∩︀";
  var caret$2 = "⁁";
  var caron$2 = "ˇ";
  var Cayleys$2 = "ℭ";
  var ccaps$2 = "⩍";
  var Ccaron$2 = "Č";
  var ccaron$2 = "č";
  var Ccedil$3 = "Ç";
  var ccedil$3 = "ç";
  var Ccirc$2 = "Ĉ";
  var ccirc$2 = "ĉ";
  var Cconint$2 = "∰";
  var ccups$2 = "⩌";
  var ccupssm$2 = "⩐";
  var Cdot$2 = "Ċ";
  var cdot$2 = "ċ";
  var cedil$3 = "¸";
  var Cedilla$2 = "¸";
  var cemptyv$2 = "⦲";
  var cent$3 = "¢";
  var centerdot$2 = "·";
  var CenterDot$2 = "·";
  var cfr$2 = "𝔠";
  var Cfr$2 = "ℭ";
  var CHcy$2 = "Ч";
  var chcy$2 = "ч";
  var check$2 = "✓";
  var checkmark$2 = "✓";
  var Chi$2 = "Χ";
  var chi$2 = "χ";
  var circ$2 = "ˆ";
  var circeq$2 = "≗";
  var circlearrowleft$2 = "↺";
  var circlearrowright$2 = "↻";
  var circledast$2 = "⊛";
  var circledcirc$2 = "⊚";
  var circleddash$2 = "⊝";
  var CircleDot$2 = "⊙";
  var circledR$2 = "®";
  var circledS$2 = "Ⓢ";
  var CircleMinus$2 = "⊖";
  var CirclePlus$2 = "⊕";
  var CircleTimes$2 = "⊗";
  var cir$2 = "○";
  var cirE$2 = "⧃";
  var cire$2 = "≗";
  var cirfnint$2 = "⨐";
  var cirmid$2 = "⫯";
  var cirscir$2 = "⧂";
  var ClockwiseContourIntegral$2 = "∲";
  var CloseCurlyDoubleQuote$2 = "”";
  var CloseCurlyQuote$2 = "’";
  var clubs$2 = "♣";
  var clubsuit$2 = "♣";
  var colon$2 = ":";
  var Colon$2 = "∷";
  var Colone$2 = "⩴";
  var colone$2 = "≔";
  var coloneq$2 = "≔";
  var comma$2 = ",";
  var commat$2 = "@";
  var comp$2 = "∁";
  var compfn$2 = "∘";
  var complement$2 = "∁";
  var complexes$2 = "ℂ";
  var cong$2 = "≅";
  var congdot$2 = "⩭";
  var Congruent$2 = "≡";
  var conint$2 = "∮";
  var Conint$2 = "∯";
  var ContourIntegral$2 = "∮";
  var copf$2 = "𝕔";
  var Copf$2 = "ℂ";
  var coprod$2 = "∐";
  var Coproduct$2 = "∐";
  var copy$3 = "©";
  var COPY$3 = "©";
  var copysr$2 = "℗";
  var CounterClockwiseContourIntegral$2 = "∳";
  var crarr$2 = "↵";
  var cross$2 = "✗";
  var Cross$2 = "⨯";
  var Cscr$2 = "𝒞";
  var cscr$2 = "𝒸";
  var csub$2 = "⫏";
  var csube$2 = "⫑";
  var csup$2 = "⫐";
  var csupe$2 = "⫒";
  var ctdot$2 = "⋯";
  var cudarrl$2 = "⤸";
  var cudarrr$2 = "⤵";
  var cuepr$2 = "⋞";
  var cuesc$2 = "⋟";
  var cularr$2 = "↶";
  var cularrp$2 = "⤽";
  var cupbrcap$2 = "⩈";
  var cupcap$2 = "⩆";
  var CupCap$2 = "≍";
  var cup$2 = "∪";
  var Cup$2 = "⋓";
  var cupcup$2 = "⩊";
  var cupdot$2 = "⊍";
  var cupor$2 = "⩅";
  var cups$2 = "∪︀";
  var curarr$2 = "↷";
  var curarrm$2 = "⤼";
  var curlyeqprec$2 = "⋞";
  var curlyeqsucc$2 = "⋟";
  var curlyvee$2 = "⋎";
  var curlywedge$2 = "⋏";
  var curren$3 = "¤";
  var curvearrowleft$2 = "↶";
  var curvearrowright$2 = "↷";
  var cuvee$2 = "⋎";
  var cuwed$2 = "⋏";
  var cwconint$2 = "∲";
  var cwint$2 = "∱";
  var cylcty$2 = "⌭";
  var dagger$2 = "†";
  var Dagger$2 = "‡";
  var daleth$2 = "ℸ";
  var darr$2 = "↓";
  var Darr$2 = "↡";
  var dArr$2 = "⇓";
  var dash$2 = "‐";
  var Dashv$2 = "⫤";
  var dashv$2 = "⊣";
  var dbkarow$2 = "⤏";
  var dblac$2 = "˝";
  var Dcaron$2 = "Ď";
  var dcaron$2 = "ď";
  var Dcy$2 = "Д";
  var dcy$2 = "д";
  var ddagger$2 = "‡";
  var ddarr$2 = "⇊";
  var DD$2 = "ⅅ";
  var dd$2 = "ⅆ";
  var DDotrahd$2 = "⤑";
  var ddotseq$2 = "⩷";
  var deg$3 = "°";
  var Del$2 = "∇";
  var Delta$2 = "Δ";
  var delta$2 = "δ";
  var demptyv$2 = "⦱";
  var dfisht$2 = "⥿";
  var Dfr$2 = "𝔇";
  var dfr$2 = "𝔡";
  var dHar$2 = "⥥";
  var dharl$2 = "⇃";
  var dharr$2 = "⇂";
  var DiacriticalAcute$2 = "´";
  var DiacriticalDot$2 = "˙";
  var DiacriticalDoubleAcute$2 = "˝";
  var DiacriticalGrave$2 = "`";
  var DiacriticalTilde$2 = "˜";
  var diam$2 = "⋄";
  var diamond$2 = "⋄";
  var Diamond$2 = "⋄";
  var diamondsuit$2 = "♦";
  var diams$2 = "♦";
  var die$2 = "¨";
  var DifferentialD$2 = "ⅆ";
  var digamma$2 = "ϝ";
  var disin$2 = "⋲";
  var div$2 = "÷";
  var divide$3 = "÷";
  var divideontimes$2 = "⋇";
  var divonx$2 = "⋇";
  var DJcy$2 = "Ђ";
  var djcy$2 = "ђ";
  var dlcorn$2 = "⌞";
  var dlcrop$2 = "⌍";
  var dollar$2 = "$";
  var Dopf$2 = "𝔻";
  var dopf$2 = "𝕕";
  var Dot$2 = "¨";
  var dot$2 = "˙";
  var DotDot$2 = "⃜";
  var doteq$2 = "≐";
  var doteqdot$2 = "≑";
  var DotEqual$2 = "≐";
  var dotminus$2 = "∸";
  var dotplus$2 = "∔";
  var dotsquare$2 = "⊡";
  var doublebarwedge$2 = "⌆";
  var DoubleContourIntegral$2 = "∯";
  var DoubleDot$2 = "¨";
  var DoubleDownArrow$2 = "⇓";
  var DoubleLeftArrow$2 = "⇐";
  var DoubleLeftRightArrow$2 = "⇔";
  var DoubleLeftTee$2 = "⫤";
  var DoubleLongLeftArrow$2 = "⟸";
  var DoubleLongLeftRightArrow$2 = "⟺";
  var DoubleLongRightArrow$2 = "⟹";
  var DoubleRightArrow$2 = "⇒";
  var DoubleRightTee$2 = "⊨";
  var DoubleUpArrow$2 = "⇑";
  var DoubleUpDownArrow$2 = "⇕";
  var DoubleVerticalBar$2 = "∥";
  var DownArrowBar$2 = "⤓";
  var downarrow$2 = "↓";
  var DownArrow$2 = "↓";
  var Downarrow$2 = "⇓";
  var DownArrowUpArrow$2 = "⇵";
  var DownBreve$2 = "̑";
  var downdownarrows$2 = "⇊";
  var downharpoonleft$2 = "⇃";
  var downharpoonright$2 = "⇂";
  var DownLeftRightVector$2 = "⥐";
  var DownLeftTeeVector$2 = "⥞";
  var DownLeftVectorBar$2 = "⥖";
  var DownLeftVector$2 = "↽";
  var DownRightTeeVector$2 = "⥟";
  var DownRightVectorBar$2 = "⥗";
  var DownRightVector$2 = "⇁";
  var DownTeeArrow$2 = "↧";
  var DownTee$2 = "⊤";
  var drbkarow$2 = "⤐";
  var drcorn$2 = "⌟";
  var drcrop$2 = "⌌";
  var Dscr$2 = "𝒟";
  var dscr$2 = "𝒹";
  var DScy$2 = "Ѕ";
  var dscy$2 = "ѕ";
  var dsol$2 = "⧶";
  var Dstrok$2 = "Đ";
  var dstrok$2 = "đ";
  var dtdot$2 = "⋱";
  var dtri$2 = "▿";
  var dtrif$2 = "▾";
  var duarr$2 = "⇵";
  var duhar$2 = "⥯";
  var dwangle$2 = "⦦";
  var DZcy$2 = "Џ";
  var dzcy$2 = "џ";
  var dzigrarr$2 = "⟿";
  var Eacute$3 = "É";
  var eacute$3 = "é";
  var easter$2 = "⩮";
  var Ecaron$2 = "Ě";
  var ecaron$2 = "ě";
  var Ecirc$3 = "Ê";
  var ecirc$3 = "ê";
  var ecir$2 = "≖";
  var ecolon$2 = "≕";
  var Ecy$2 = "Э";
  var ecy$2 = "э";
  var eDDot$2 = "⩷";
  var Edot$2 = "Ė";
  var edot$2 = "ė";
  var eDot$2 = "≑";
  var ee$2 = "ⅇ";
  var efDot$2 = "≒";
  var Efr$2 = "𝔈";
  var efr$2 = "𝔢";
  var eg$2 = "⪚";
  var Egrave$3 = "È";
  var egrave$3 = "è";
  var egs$2 = "⪖";
  var egsdot$2 = "⪘";
  var el$2 = "⪙";
  var Element$2 = "∈";
  var elinters$2 = "⏧";
  var ell$2 = "ℓ";
  var els$2 = "⪕";
  var elsdot$2 = "⪗";
  var Emacr$2 = "Ē";
  var emacr$2 = "ē";
  var empty$2 = "∅";
  var emptyset$2 = "∅";
  var EmptySmallSquare$2 = "◻";
  var emptyv$2 = "∅";
  var EmptyVerySmallSquare$2 = "▫";
  var emsp13$2 = " ";
  var emsp14$2 = " ";
  var emsp$2 = " ";
  var ENG$2 = "Ŋ";
  var eng$2 = "ŋ";
  var ensp$2 = " ";
  var Eogon$2 = "Ę";
  var eogon$2 = "ę";
  var Eopf$2 = "𝔼";
  var eopf$2 = "𝕖";
  var epar$2 = "⋕";
  var eparsl$2 = "⧣";
  var eplus$2 = "⩱";
  var epsi$2 = "ε";
  var Epsilon$2 = "Ε";
  var epsilon$2 = "ε";
  var epsiv$2 = "ϵ";
  var eqcirc$2 = "≖";
  var eqcolon$2 = "≕";
  var eqsim$2 = "≂";
  var eqslantgtr$2 = "⪖";
  var eqslantless$2 = "⪕";
  var Equal$2 = "⩵";
  var equals$2 = "=";
  var EqualTilde$2 = "≂";
  var equest$2 = "≟";
  var Equilibrium$2 = "⇌";
  var equiv$2 = "≡";
  var equivDD$2 = "⩸";
  var eqvparsl$2 = "⧥";
  var erarr$2 = "⥱";
  var erDot$2 = "≓";
  var escr$2 = "ℯ";
  var Escr$2 = "ℰ";
  var esdot$2 = "≐";
  var Esim$2 = "⩳";
  var esim$2 = "≂";
  var Eta$2 = "Η";
  var eta$2 = "η";
  var ETH$3 = "Ð";
  var eth$3 = "ð";
  var Euml$3 = "Ë";
  var euml$3 = "ë";
  var euro$2 = "€";
  var excl$2 = "!";
  var exist$2 = "∃";
  var Exists$2 = "∃";
  var expectation$2 = "ℰ";
  var exponentiale$2 = "ⅇ";
  var ExponentialE$2 = "ⅇ";
  var fallingdotseq$2 = "≒";
  var Fcy$2 = "Ф";
  var fcy$2 = "ф";
  var female$2 = "♀";
  var ffilig$2 = "ﬃ";
  var fflig$2 = "ﬀ";
  var ffllig$2 = "ﬄ";
  var Ffr$2 = "𝔉";
  var ffr$2 = "𝔣";
  var filig$2 = "ﬁ";
  var FilledSmallSquare$2 = "◼";
  var FilledVerySmallSquare$2 = "▪";
  var fjlig$2 = "fj";
  var flat$2 = "♭";
  var fllig$2 = "ﬂ";
  var fltns$2 = "▱";
  var fnof$2 = "ƒ";
  var Fopf$2 = "𝔽";
  var fopf$2 = "𝕗";
  var forall$2 = "∀";
  var ForAll$2 = "∀";
  var fork$2 = "⋔";
  var forkv$2 = "⫙";
  var Fouriertrf$2 = "ℱ";
  var fpartint$2 = "⨍";
  var frac12$3 = "½";
  var frac13$2 = "⅓";
  var frac14$3 = "¼";
  var frac15$2 = "⅕";
  var frac16$2 = "⅙";
  var frac18$2 = "⅛";
  var frac23$2 = "⅔";
  var frac25$2 = "⅖";
  var frac34$3 = "¾";
  var frac35$2 = "⅗";
  var frac38$2 = "⅜";
  var frac45$2 = "⅘";
  var frac56$2 = "⅚";
  var frac58$2 = "⅝";
  var frac78$2 = "⅞";
  var frasl$2 = "⁄";
  var frown$2 = "⌢";
  var fscr$2 = "𝒻";
  var Fscr$2 = "ℱ";
  var gacute$2 = "ǵ";
  var Gamma$2 = "Γ";
  var gamma$2 = "γ";
  var Gammad$2 = "Ϝ";
  var gammad$2 = "ϝ";
  var gap$2 = "⪆";
  var Gbreve$2 = "Ğ";
  var gbreve$2 = "ğ";
  var Gcedil$2 = "Ģ";
  var Gcirc$2 = "Ĝ";
  var gcirc$2 = "ĝ";
  var Gcy$2 = "Г";
  var gcy$2 = "г";
  var Gdot$2 = "Ġ";
  var gdot$2 = "ġ";
  var ge$2 = "≥";
  var gE$2 = "≧";
  var gEl$2 = "⪌";
  var gel$2 = "⋛";
  var geq$2 = "≥";
  var geqq$2 = "≧";
  var geqslant$2 = "⩾";
  var gescc$2 = "⪩";
  var ges$2 = "⩾";
  var gesdot$2 = "⪀";
  var gesdoto$2 = "⪂";
  var gesdotol$2 = "⪄";
  var gesl$2 = "⋛︀";
  var gesles$2 = "⪔";
  var Gfr$2 = "𝔊";
  var gfr$2 = "𝔤";
  var gg$2 = "≫";
  var Gg$2 = "⋙";
  var ggg$2 = "⋙";
  var gimel$2 = "ℷ";
  var GJcy$2 = "Ѓ";
  var gjcy$2 = "ѓ";
  var gla$2 = "⪥";
  var gl$2 = "≷";
  var glE$2 = "⪒";
  var glj$2 = "⪤";
  var gnap$2 = "⪊";
  var gnapprox$2 = "⪊";
  var gne$2 = "⪈";
  var gnE$2 = "≩";
  var gneq$2 = "⪈";
  var gneqq$2 = "≩";
  var gnsim$2 = "⋧";
  var Gopf$2 = "𝔾";
  var gopf$2 = "𝕘";
  var grave$2 = "`";
  var GreaterEqual$2 = "≥";
  var GreaterEqualLess$2 = "⋛";
  var GreaterFullEqual$2 = "≧";
  var GreaterGreater$2 = "⪢";
  var GreaterLess$2 = "≷";
  var GreaterSlantEqual$2 = "⩾";
  var GreaterTilde$2 = "≳";
  var Gscr$2 = "𝒢";
  var gscr$2 = "ℊ";
  var gsim$2 = "≳";
  var gsime$2 = "⪎";
  var gsiml$2 = "⪐";
  var gtcc$2 = "⪧";
  var gtcir$2 = "⩺";
  var gt$1$2 = ">";
  var GT$3 = ">";
  var Gt$2 = "≫";
  var gtdot$2 = "⋗";
  var gtlPar$2 = "⦕";
  var gtquest$2 = "⩼";
  var gtrapprox$2 = "⪆";
  var gtrarr$2 = "⥸";
  var gtrdot$2 = "⋗";
  var gtreqless$2 = "⋛";
  var gtreqqless$2 = "⪌";
  var gtrless$2 = "≷";
  var gtrsim$2 = "≳";
  var gvertneqq$2 = "≩︀";
  var gvnE$2 = "≩︀";
  var Hacek$2 = "ˇ";
  var hairsp$2 = " ";
  var half$2 = "½";
  var hamilt$2 = "ℋ";
  var HARDcy$2 = "Ъ";
  var hardcy$2 = "ъ";
  var harrcir$2 = "⥈";
  var harr$2 = "↔";
  var hArr$2 = "⇔";
  var harrw$2 = "↭";
  var Hat$2 = "^";
  var hbar$2 = "ℏ";
  var Hcirc$2 = "Ĥ";
  var hcirc$2 = "ĥ";
  var hearts$2 = "♥";
  var heartsuit$2 = "♥";
  var hellip$2 = "…";
  var hercon$2 = "⊹";
  var hfr$2 = "𝔥";
  var Hfr$2 = "ℌ";
  var HilbertSpace$2 = "ℋ";
  var hksearow$2 = "⤥";
  var hkswarow$2 = "⤦";
  var hoarr$2 = "⇿";
  var homtht$2 = "∻";
  var hookleftarrow$2 = "↩";
  var hookrightarrow$2 = "↪";
  var hopf$2 = "𝕙";
  var Hopf$2 = "ℍ";
  var horbar$2 = "―";
  var HorizontalLine$2 = "─";
  var hscr$2 = "𝒽";
  var Hscr$2 = "ℋ";
  var hslash$2 = "ℏ";
  var Hstrok$2 = "Ħ";
  var hstrok$2 = "ħ";
  var HumpDownHump$2 = "≎";
  var HumpEqual$2 = "≏";
  var hybull$2 = "⁃";
  var hyphen$2 = "‐";
  var Iacute$3 = "Í";
  var iacute$3 = "í";
  var ic$2 = "⁣";
  var Icirc$3 = "Î";
  var icirc$3 = "î";
  var Icy$2 = "И";
  var icy$2 = "и";
  var Idot$2 = "İ";
  var IEcy$2 = "Е";
  var iecy$2 = "е";
  var iexcl$3 = "¡";
  var iff$2 = "⇔";
  var ifr$2 = "𝔦";
  var Ifr$2 = "ℑ";
  var Igrave$3 = "Ì";
  var igrave$3 = "ì";
  var ii$2 = "ⅈ";
  var iiiint$2 = "⨌";
  var iiint$2 = "∭";
  var iinfin$2 = "⧜";
  var iiota$2 = "℩";
  var IJlig$2 = "Ĳ";
  var ijlig$2 = "ĳ";
  var Imacr$2 = "Ī";
  var imacr$2 = "ī";
  var image$2 = "ℑ";
  var ImaginaryI$2 = "ⅈ";
  var imagline$2 = "ℐ";
  var imagpart$2 = "ℑ";
  var imath$2 = "ı";
  var Im$2 = "ℑ";
  var imof$2 = "⊷";
  var imped$2 = "Ƶ";
  var Implies$2 = "⇒";
  var incare$2 = "℅";
  var infin$2 = "∞";
  var infintie$2 = "⧝";
  var inodot$2 = "ı";
  var intcal$2 = "⊺";
  var int$2 = "∫";
  var Int$2 = "∬";
  var integers$2 = "ℤ";
  var Integral$2 = "∫";
  var intercal$2 = "⊺";
  var Intersection$2 = "⋂";
  var intlarhk$2 = "⨗";
  var intprod$2 = "⨼";
  var InvisibleComma$2 = "⁣";
  var InvisibleTimes$2 = "⁢";
  var IOcy$2 = "Ё";
  var iocy$2 = "ё";
  var Iogon$2 = "Į";
  var iogon$2 = "į";
  var Iopf$2 = "𝕀";
  var iopf$2 = "𝕚";
  var Iota$2 = "Ι";
  var iota$2 = "ι";
  var iprod$2 = "⨼";
  var iquest$3 = "¿";
  var iscr$2 = "𝒾";
  var Iscr$2 = "ℐ";
  var isin$2 = "∈";
  var isindot$2 = "⋵";
  var isinE$2 = "⋹";
  var isins$2 = "⋴";
  var isinsv$2 = "⋳";
  var isinv$2 = "∈";
  var it$2 = "⁢";
  var Itilde$2 = "Ĩ";
  var itilde$2 = "ĩ";
  var Iukcy$2 = "І";
  var iukcy$2 = "і";
  var Iuml$3 = "Ï";
  var iuml$3 = "ï";
  var Jcirc$2 = "Ĵ";
  var jcirc$2 = "ĵ";
  var Jcy$2 = "Й";
  var jcy$2 = "й";
  var Jfr$2 = "𝔍";
  var jfr$2 = "𝔧";
  var jmath$2 = "ȷ";
  var Jopf$2 = "𝕁";
  var jopf$2 = "𝕛";
  var Jscr$2 = "𝒥";
  var jscr$2 = "𝒿";
  var Jsercy$2 = "Ј";
  var jsercy$2 = "ј";
  var Jukcy$2 = "Є";
  var jukcy$2 = "є";
  var Kappa$2 = "Κ";
  var kappa$2 = "κ";
  var kappav$2 = "ϰ";
  var Kcedil$2 = "Ķ";
  var kcedil$2 = "ķ";
  var Kcy$2 = "К";
  var kcy$2 = "к";
  var Kfr$2 = "𝔎";
  var kfr$2 = "𝔨";
  var kgreen$2 = "ĸ";
  var KHcy$2 = "Х";
  var khcy$2 = "х";
  var KJcy$2 = "Ќ";
  var kjcy$2 = "ќ";
  var Kopf$2 = "𝕂";
  var kopf$2 = "𝕜";
  var Kscr$2 = "𝒦";
  var kscr$2 = "𝓀";
  var lAarr$2 = "⇚";
  var Lacute$2 = "Ĺ";
  var lacute$2 = "ĺ";
  var laemptyv$2 = "⦴";
  var lagran$2 = "ℒ";
  var Lambda$2 = "Λ";
  var lambda$2 = "λ";
  var lang$2 = "⟨";
  var Lang$2 = "⟪";
  var langd$2 = "⦑";
  var langle$2 = "⟨";
  var lap$2 = "⪅";
  var Laplacetrf$2 = "ℒ";
  var laquo$3 = "«";
  var larrb$2 = "⇤";
  var larrbfs$2 = "⤟";
  var larr$2 = "←";
  var Larr$2 = "↞";
  var lArr$2 = "⇐";
  var larrfs$2 = "⤝";
  var larrhk$2 = "↩";
  var larrlp$2 = "↫";
  var larrpl$2 = "⤹";
  var larrsim$2 = "⥳";
  var larrtl$2 = "↢";
  var latail$2 = "⤙";
  var lAtail$2 = "⤛";
  var lat$2 = "⪫";
  var late$2 = "⪭";
  var lates$2 = "⪭︀";
  var lbarr$2 = "⤌";
  var lBarr$2 = "⤎";
  var lbbrk$2 = "❲";
  var lbrace$2 = "{";
  var lbrack$2 = "[";
  var lbrke$2 = "⦋";
  var lbrksld$2 = "⦏";
  var lbrkslu$2 = "⦍";
  var Lcaron$2 = "Ľ";
  var lcaron$2 = "ľ";
  var Lcedil$2 = "Ļ";
  var lcedil$2 = "ļ";
  var lceil$2 = "⌈";
  var lcub$2 = "{";
  var Lcy$2 = "Л";
  var lcy$2 = "л";
  var ldca$2 = "⤶";
  var ldquo$2 = "“";
  var ldquor$2 = "„";
  var ldrdhar$2 = "⥧";
  var ldrushar$2 = "⥋";
  var ldsh$2 = "↲";
  var le$2 = "≤";
  var lE$2 = "≦";
  var LeftAngleBracket$2 = "⟨";
  var LeftArrowBar$2 = "⇤";
  var leftarrow$2 = "←";
  var LeftArrow$2 = "←";
  var Leftarrow$2 = "⇐";
  var LeftArrowRightArrow$2 = "⇆";
  var leftarrowtail$2 = "↢";
  var LeftCeiling$2 = "⌈";
  var LeftDoubleBracket$2 = "⟦";
  var LeftDownTeeVector$2 = "⥡";
  var LeftDownVectorBar$2 = "⥙";
  var LeftDownVector$2 = "⇃";
  var LeftFloor$2 = "⌊";
  var leftharpoondown$2 = "↽";
  var leftharpoonup$2 = "↼";
  var leftleftarrows$2 = "⇇";
  var leftrightarrow$2 = "↔";
  var LeftRightArrow$2 = "↔";
  var Leftrightarrow$2 = "⇔";
  var leftrightarrows$2 = "⇆";
  var leftrightharpoons$2 = "⇋";
  var leftrightsquigarrow$2 = "↭";
  var LeftRightVector$2 = "⥎";
  var LeftTeeArrow$2 = "↤";
  var LeftTee$2 = "⊣";
  var LeftTeeVector$2 = "⥚";
  var leftthreetimes$2 = "⋋";
  var LeftTriangleBar$2 = "⧏";
  var LeftTriangle$2 = "⊲";
  var LeftTriangleEqual$2 = "⊴";
  var LeftUpDownVector$2 = "⥑";
  var LeftUpTeeVector$2 = "⥠";
  var LeftUpVectorBar$2 = "⥘";
  var LeftUpVector$2 = "↿";
  var LeftVectorBar$2 = "⥒";
  var LeftVector$2 = "↼";
  var lEg$2 = "⪋";
  var leg$2 = "⋚";
  var leq$2 = "≤";
  var leqq$2 = "≦";
  var leqslant$2 = "⩽";
  var lescc$2 = "⪨";
  var les$2 = "⩽";
  var lesdot$2 = "⩿";
  var lesdoto$2 = "⪁";
  var lesdotor$2 = "⪃";
  var lesg$2 = "⋚︀";
  var lesges$2 = "⪓";
  var lessapprox$2 = "⪅";
  var lessdot$2 = "⋖";
  var lesseqgtr$2 = "⋚";
  var lesseqqgtr$2 = "⪋";
  var LessEqualGreater$2 = "⋚";
  var LessFullEqual$2 = "≦";
  var LessGreater$2 = "≶";
  var lessgtr$2 = "≶";
  var LessLess$2 = "⪡";
  var lesssim$2 = "≲";
  var LessSlantEqual$2 = "⩽";
  var LessTilde$2 = "≲";
  var lfisht$2 = "⥼";
  var lfloor$2 = "⌊";
  var Lfr$2 = "𝔏";
  var lfr$2 = "𝔩";
  var lg$2 = "≶";
  var lgE$2 = "⪑";
  var lHar$2 = "⥢";
  var lhard$2 = "↽";
  var lharu$2 = "↼";
  var lharul$2 = "⥪";
  var lhblk$2 = "▄";
  var LJcy$2 = "Љ";
  var ljcy$2 = "љ";
  var llarr$2 = "⇇";
  var ll$2 = "≪";
  var Ll$2 = "⋘";
  var llcorner$2 = "⌞";
  var Lleftarrow$2 = "⇚";
  var llhard$2 = "⥫";
  var lltri$2 = "◺";
  var Lmidot$2 = "Ŀ";
  var lmidot$2 = "ŀ";
  var lmoustache$2 = "⎰";
  var lmoust$2 = "⎰";
  var lnap$2 = "⪉";
  var lnapprox$2 = "⪉";
  var lne$2 = "⪇";
  var lnE$2 = "≨";
  var lneq$2 = "⪇";
  var lneqq$2 = "≨";
  var lnsim$2 = "⋦";
  var loang$2 = "⟬";
  var loarr$2 = "⇽";
  var lobrk$2 = "⟦";
  var longleftarrow$2 = "⟵";
  var LongLeftArrow$2 = "⟵";
  var Longleftarrow$2 = "⟸";
  var longleftrightarrow$2 = "⟷";
  var LongLeftRightArrow$2 = "⟷";
  var Longleftrightarrow$2 = "⟺";
  var longmapsto$2 = "⟼";
  var longrightarrow$2 = "⟶";
  var LongRightArrow$2 = "⟶";
  var Longrightarrow$2 = "⟹";
  var looparrowleft$2 = "↫";
  var looparrowright$2 = "↬";
  var lopar$2 = "⦅";
  var Lopf$2 = "𝕃";
  var lopf$2 = "𝕝";
  var loplus$2 = "⨭";
  var lotimes$2 = "⨴";
  var lowast$2 = "∗";
  var lowbar$2 = "_";
  var LowerLeftArrow$2 = "↙";
  var LowerRightArrow$2 = "↘";
  var loz$2 = "◊";
  var lozenge$2 = "◊";
  var lozf$2 = "⧫";
  var lpar$2 = "(";
  var lparlt$2 = "⦓";
  var lrarr$2 = "⇆";
  var lrcorner$2 = "⌟";
  var lrhar$2 = "⇋";
  var lrhard$2 = "⥭";
  var lrm$2 = "‎";
  var lrtri$2 = "⊿";
  var lsaquo$2 = "‹";
  var lscr$2 = "𝓁";
  var Lscr$2 = "ℒ";
  var lsh$2 = "↰";
  var Lsh$2 = "↰";
  var lsim$2 = "≲";
  var lsime$2 = "⪍";
  var lsimg$2 = "⪏";
  var lsqb$2 = "[";
  var lsquo$2 = "‘";
  var lsquor$2 = "‚";
  var Lstrok$2 = "Ł";
  var lstrok$2 = "ł";
  var ltcc$2 = "⪦";
  var ltcir$2 = "⩹";
  var lt$1$2 = "<";
  var LT$3 = "<";
  var Lt$2 = "≪";
  var ltdot$2 = "⋖";
  var lthree$2 = "⋋";
  var ltimes$2 = "⋉";
  var ltlarr$2 = "⥶";
  var ltquest$2 = "⩻";
  var ltri$2 = "◃";
  var ltrie$2 = "⊴";
  var ltrif$2 = "◂";
  var ltrPar$2 = "⦖";
  var lurdshar$2 = "⥊";
  var luruhar$2 = "⥦";
  var lvertneqq$2 = "≨︀";
  var lvnE$2 = "≨︀";
  var macr$3 = "¯";
  var male$2 = "♂";
  var malt$2 = "✠";
  var maltese$2 = "✠";
  var map$1$2 = "↦";
  var mapsto$2 = "↦";
  var mapstodown$2 = "↧";
  var mapstoleft$2 = "↤";
  var mapstoup$2 = "↥";
  var marker$2 = "▮";
  var mcomma$2 = "⨩";
  var Mcy$2 = "М";
  var mcy$2 = "м";
  var mdash$2 = "—";
  var mDDot$2 = "∺";
  var measuredangle$2 = "∡";
  var MediumSpace$2 = " ";
  var Mellintrf$2 = "ℳ";
  var Mfr$2 = "𝔐";
  var mfr$2 = "𝔪";
  var mho$2 = "℧";
  var micro$3 = "µ";
  var midast$2 = "*";
  var midcir$2 = "⫰";
  var mid$2 = "∣";
  var middot$3 = "·";
  var minusb$2 = "⊟";
  var minus$2 = "−";
  var minusd$2 = "∸";
  var minusdu$2 = "⨪";
  var MinusPlus$2 = "∓";
  var mlcp$2 = "⫛";
  var mldr$2 = "…";
  var mnplus$2 = "∓";
  var models$2 = "⊧";
  var Mopf$2 = "𝕄";
  var mopf$2 = "𝕞";
  var mp$2 = "∓";
  var mscr$2 = "𝓂";
  var Mscr$2 = "ℳ";
  var mstpos$2 = "∾";
  var Mu$2 = "Μ";
  var mu$2 = "μ";
  var multimap$2 = "⊸";
  var mumap$2 = "⊸";
  var nabla$2 = "∇";
  var Nacute$2 = "Ń";
  var nacute$2 = "ń";
  var nang$2 = "∠⃒";
  var nap$2 = "≉";
  var napE$2 = "⩰̸";
  var napid$2 = "≋̸";
  var napos$2 = "ŉ";
  var napprox$2 = "≉";
  var natural$2 = "♮";
  var naturals$2 = "ℕ";
  var natur$2 = "♮";
  var nbsp$3 = " ";
  var nbump$2 = "≎̸";
  var nbumpe$2 = "≏̸";
  var ncap$2 = "⩃";
  var Ncaron$2 = "Ň";
  var ncaron$2 = "ň";
  var Ncedil$2 = "Ņ";
  var ncedil$2 = "ņ";
  var ncong$2 = "≇";
  var ncongdot$2 = "⩭̸";
  var ncup$2 = "⩂";
  var Ncy$2 = "Н";
  var ncy$2 = "н";
  var ndash$2 = "–";
  var nearhk$2 = "⤤";
  var nearr$2 = "↗";
  var neArr$2 = "⇗";
  var nearrow$2 = "↗";
  var ne$2 = "≠";
  var nedot$2 = "≐̸";
  var NegativeMediumSpace$2 = "​";
  var NegativeThickSpace$2 = "​";
  var NegativeThinSpace$2 = "​";
  var NegativeVeryThinSpace$2 = "​";
  var nequiv$2 = "≢";
  var nesear$2 = "⤨";
  var nesim$2 = "≂̸";
  var NestedGreaterGreater$2 = "≫";
  var NestedLessLess$2 = "≪";
  var NewLine$2 = "\n";
  var nexist$2 = "∄";
  var nexists$2 = "∄";
  var Nfr$2 = "𝔑";
  var nfr$2 = "𝔫";
  var ngE$2 = "≧̸";
  var nge$2 = "≱";
  var ngeq$2 = "≱";
  var ngeqq$2 = "≧̸";
  var ngeqslant$2 = "⩾̸";
  var nges$2 = "⩾̸";
  var nGg$2 = "⋙̸";
  var ngsim$2 = "≵";
  var nGt$2 = "≫⃒";
  var ngt$2 = "≯";
  var ngtr$2 = "≯";
  var nGtv$2 = "≫̸";
  var nharr$2 = "↮";
  var nhArr$2 = "⇎";
  var nhpar$2 = "⫲";
  var ni$2 = "∋";
  var nis$2 = "⋼";
  var nisd$2 = "⋺";
  var niv$2 = "∋";
  var NJcy$2 = "Њ";
  var njcy$2 = "њ";
  var nlarr$2 = "↚";
  var nlArr$2 = "⇍";
  var nldr$2 = "‥";
  var nlE$2 = "≦̸";
  var nle$2 = "≰";
  var nleftarrow$2 = "↚";
  var nLeftarrow$2 = "⇍";
  var nleftrightarrow$2 = "↮";
  var nLeftrightarrow$2 = "⇎";
  var nleq$2 = "≰";
  var nleqq$2 = "≦̸";
  var nleqslant$2 = "⩽̸";
  var nles$2 = "⩽̸";
  var nless$2 = "≮";
  var nLl$2 = "⋘̸";
  var nlsim$2 = "≴";
  var nLt$2 = "≪⃒";
  var nlt$2 = "≮";
  var nltri$2 = "⋪";
  var nltrie$2 = "⋬";
  var nLtv$2 = "≪̸";
  var nmid$2 = "∤";
  var NoBreak$2 = "⁠";
  var NonBreakingSpace$2 = " ";
  var nopf$2 = "𝕟";
  var Nopf$2 = "ℕ";
  var Not$2 = "⫬";
  var not$3 = "¬";
  var NotCongruent$2 = "≢";
  var NotCupCap$2 = "≭";
  var NotDoubleVerticalBar$2 = "∦";
  var NotElement$2 = "∉";
  var NotEqual$2 = "≠";
  var NotEqualTilde$2 = "≂̸";
  var NotExists$2 = "∄";
  var NotGreater$2 = "≯";
  var NotGreaterEqual$2 = "≱";
  var NotGreaterFullEqual$2 = "≧̸";
  var NotGreaterGreater$2 = "≫̸";
  var NotGreaterLess$2 = "≹";
  var NotGreaterSlantEqual$2 = "⩾̸";
  var NotGreaterTilde$2 = "≵";
  var NotHumpDownHump$2 = "≎̸";
  var NotHumpEqual$2 = "≏̸";
  var notin$2 = "∉";
  var notindot$2 = "⋵̸";
  var notinE$2 = "⋹̸";
  var notinva$2 = "∉";
  var notinvb$2 = "⋷";
  var notinvc$2 = "⋶";
  var NotLeftTriangleBar$2 = "⧏̸";
  var NotLeftTriangle$2 = "⋪";
  var NotLeftTriangleEqual$2 = "⋬";
  var NotLess$2 = "≮";
  var NotLessEqual$2 = "≰";
  var NotLessGreater$2 = "≸";
  var NotLessLess$2 = "≪̸";
  var NotLessSlantEqual$2 = "⩽̸";
  var NotLessTilde$2 = "≴";
  var NotNestedGreaterGreater$2 = "⪢̸";
  var NotNestedLessLess$2 = "⪡̸";
  var notni$2 = "∌";
  var notniva$2 = "∌";
  var notnivb$2 = "⋾";
  var notnivc$2 = "⋽";
  var NotPrecedes$2 = "⊀";
  var NotPrecedesEqual$2 = "⪯̸";
  var NotPrecedesSlantEqual$2 = "⋠";
  var NotReverseElement$2 = "∌";
  var NotRightTriangleBar$2 = "⧐̸";
  var NotRightTriangle$2 = "⋫";
  var NotRightTriangleEqual$2 = "⋭";
  var NotSquareSubset$2 = "⊏̸";
  var NotSquareSubsetEqual$2 = "⋢";
  var NotSquareSuperset$2 = "⊐̸";
  var NotSquareSupersetEqual$2 = "⋣";
  var NotSubset$2 = "⊂⃒";
  var NotSubsetEqual$2 = "⊈";
  var NotSucceeds$2 = "⊁";
  var NotSucceedsEqual$2 = "⪰̸";
  var NotSucceedsSlantEqual$2 = "⋡";
  var NotSucceedsTilde$2 = "≿̸";
  var NotSuperset$2 = "⊃⃒";
  var NotSupersetEqual$2 = "⊉";
  var NotTilde$2 = "≁";
  var NotTildeEqual$2 = "≄";
  var NotTildeFullEqual$2 = "≇";
  var NotTildeTilde$2 = "≉";
  var NotVerticalBar$2 = "∤";
  var nparallel$2 = "∦";
  var npar$2 = "∦";
  var nparsl$2 = "⫽⃥";
  var npart$2 = "∂̸";
  var npolint$2 = "⨔";
  var npr$2 = "⊀";
  var nprcue$2 = "⋠";
  var nprec$2 = "⊀";
  var npreceq$2 = "⪯̸";
  var npre$2 = "⪯̸";
  var nrarrc$2 = "⤳̸";
  var nrarr$2 = "↛";
  var nrArr$2 = "⇏";
  var nrarrw$2 = "↝̸";
  var nrightarrow$2 = "↛";
  var nRightarrow$2 = "⇏";
  var nrtri$2 = "⋫";
  var nrtrie$2 = "⋭";
  var nsc$2 = "⊁";
  var nsccue$2 = "⋡";
  var nsce$2 = "⪰̸";
  var Nscr$2 = "𝒩";
  var nscr$2 = "𝓃";
  var nshortmid$2 = "∤";
  var nshortparallel$2 = "∦";
  var nsim$2 = "≁";
  var nsime$2 = "≄";
  var nsimeq$2 = "≄";
  var nsmid$2 = "∤";
  var nspar$2 = "∦";
  var nsqsube$2 = "⋢";
  var nsqsupe$2 = "⋣";
  var nsub$2 = "⊄";
  var nsubE$2 = "⫅̸";
  var nsube$2 = "⊈";
  var nsubset$2 = "⊂⃒";
  var nsubseteq$2 = "⊈";
  var nsubseteqq$2 = "⫅̸";
  var nsucc$2 = "⊁";
  var nsucceq$2 = "⪰̸";
  var nsup$2 = "⊅";
  var nsupE$2 = "⫆̸";
  var nsupe$2 = "⊉";
  var nsupset$2 = "⊃⃒";
  var nsupseteq$2 = "⊉";
  var nsupseteqq$2 = "⫆̸";
  var ntgl$2 = "≹";
  var Ntilde$3 = "Ñ";
  var ntilde$3 = "ñ";
  var ntlg$2 = "≸";
  var ntriangleleft$2 = "⋪";
  var ntrianglelefteq$2 = "⋬";
  var ntriangleright$2 = "⋫";
  var ntrianglerighteq$2 = "⋭";
  var Nu$2 = "Ν";
  var nu$2 = "ν";
  var num$2 = "#";
  var numero$2 = "№";
  var numsp$2 = " ";
  var nvap$2 = "≍⃒";
  var nvdash$2 = "⊬";
  var nvDash$2 = "⊭";
  var nVdash$2 = "⊮";
  var nVDash$2 = "⊯";
  var nvge$2 = "≥⃒";
  var nvgt$2 = ">⃒";
  var nvHarr$2 = "⤄";
  var nvinfin$2 = "⧞";
  var nvlArr$2 = "⤂";
  var nvle$2 = "≤⃒";
  var nvlt$2 = "<⃒";
  var nvltrie$2 = "⊴⃒";
  var nvrArr$2 = "⤃";
  var nvrtrie$2 = "⊵⃒";
  var nvsim$2 = "∼⃒";
  var nwarhk$2 = "⤣";
  var nwarr$2 = "↖";
  var nwArr$2 = "⇖";
  var nwarrow$2 = "↖";
  var nwnear$2 = "⤧";
  var Oacute$3 = "Ó";
  var oacute$3 = "ó";
  var oast$2 = "⊛";
  var Ocirc$3 = "Ô";
  var ocirc$3 = "ô";
  var ocir$2 = "⊚";
  var Ocy$2 = "О";
  var ocy$2 = "о";
  var odash$2 = "⊝";
  var Odblac$2 = "Ő";
  var odblac$2 = "ő";
  var odiv$2 = "⨸";
  var odot$2 = "⊙";
  var odsold$2 = "⦼";
  var OElig$2 = "Œ";
  var oelig$2 = "œ";
  var ofcir$2 = "⦿";
  var Ofr$2 = "𝔒";
  var ofr$2 = "𝔬";
  var ogon$2 = "˛";
  var Ograve$3 = "Ò";
  var ograve$3 = "ò";
  var ogt$2 = "⧁";
  var ohbar$2 = "⦵";
  var ohm$2 = "Ω";
  var oint$2 = "∮";
  var olarr$2 = "↺";
  var olcir$2 = "⦾";
  var olcross$2 = "⦻";
  var oline$2 = "‾";
  var olt$2 = "⧀";
  var Omacr$2 = "Ō";
  var omacr$2 = "ō";
  var Omega$2 = "Ω";
  var omega$2 = "ω";
  var Omicron$2 = "Ο";
  var omicron$2 = "ο";
  var omid$2 = "⦶";
  var ominus$2 = "⊖";
  var Oopf$2 = "𝕆";
  var oopf$2 = "𝕠";
  var opar$2 = "⦷";
  var OpenCurlyDoubleQuote$2 = "“";
  var OpenCurlyQuote$2 = "‘";
  var operp$2 = "⦹";
  var oplus$2 = "⊕";
  var orarr$2 = "↻";
  var Or$2 = "⩔";
  var or$2 = "∨";
  var ord$2 = "⩝";
  var order$2 = "ℴ";
  var orderof$2 = "ℴ";
  var ordf$3 = "ª";
  var ordm$3 = "º";
  var origof$2 = "⊶";
  var oror$2 = "⩖";
  var orslope$2 = "⩗";
  var orv$2 = "⩛";
  var oS$2 = "Ⓢ";
  var Oscr$2 = "𝒪";
  var oscr$2 = "ℴ";
  var Oslash$3 = "Ø";
  var oslash$3 = "ø";
  var osol$2 = "⊘";
  var Otilde$3 = "Õ";
  var otilde$3 = "õ";
  var otimesas$2 = "⨶";
  var Otimes$2 = "⨷";
  var otimes$2 = "⊗";
  var Ouml$3 = "Ö";
  var ouml$3 = "ö";
  var ovbar$2 = "⌽";
  var OverBar$2 = "‾";
  var OverBrace$2 = "⏞";
  var OverBracket$2 = "⎴";
  var OverParenthesis$2 = "⏜";
  var para$3 = "¶";
  var parallel$2 = "∥";
  var par$2 = "∥";
  var parsim$2 = "⫳";
  var parsl$2 = "⫽";
  var part$2 = "∂";
  var PartialD$2 = "∂";
  var Pcy$2 = "П";
  var pcy$2 = "п";
  var percnt$2 = "%";
  var period$2 = ".";
  var permil$2 = "‰";
  var perp$2 = "⊥";
  var pertenk$2 = "‱";
  var Pfr$2 = "𝔓";
  var pfr$2 = "𝔭";
  var Phi$2 = "Φ";
  var phi$2 = "φ";
  var phiv$2 = "ϕ";
  var phmmat$2 = "ℳ";
  var phone$2 = "☎";
  var Pi$2 = "Π";
  var pi$2 = "π";
  var pitchfork$2 = "⋔";
  var piv$2 = "ϖ";
  var planck$2 = "ℏ";
  var planckh$2 = "ℎ";
  var plankv$2 = "ℏ";
  var plusacir$2 = "⨣";
  var plusb$2 = "⊞";
  var pluscir$2 = "⨢";
  var plus$2 = "+";
  var plusdo$2 = "∔";
  var plusdu$2 = "⨥";
  var pluse$2 = "⩲";
  var PlusMinus$2 = "±";
  var plusmn$3 = "±";
  var plussim$2 = "⨦";
  var plustwo$2 = "⨧";
  var pm$2 = "±";
  var Poincareplane$2 = "ℌ";
  var pointint$2 = "⨕";
  var popf$2 = "𝕡";
  var Popf$2 = "ℙ";
  var pound$3 = "£";
  var prap$2 = "⪷";
  var Pr$2 = "⪻";
  var pr$2 = "≺";
  var prcue$2 = "≼";
  var precapprox$2 = "⪷";
  var prec$2 = "≺";
  var preccurlyeq$2 = "≼";
  var Precedes$2 = "≺";
  var PrecedesEqual$2 = "⪯";
  var PrecedesSlantEqual$2 = "≼";
  var PrecedesTilde$2 = "≾";
  var preceq$2 = "⪯";
  var precnapprox$2 = "⪹";
  var precneqq$2 = "⪵";
  var precnsim$2 = "⋨";
  var pre$2 = "⪯";
  var prE$2 = "⪳";
  var precsim$2 = "≾";
  var prime$2 = "′";
  var Prime$2 = "″";
  var primes$2 = "ℙ";
  var prnap$2 = "⪹";
  var prnE$2 = "⪵";
  var prnsim$2 = "⋨";
  var prod$2 = "∏";
  var Product$2 = "∏";
  var profalar$2 = "⌮";
  var profline$2 = "⌒";
  var profsurf$2 = "⌓";
  var prop$2 = "∝";
  var Proportional$2 = "∝";
  var Proportion$2 = "∷";
  var propto$2 = "∝";
  var prsim$2 = "≾";
  var prurel$2 = "⊰";
  var Pscr$2 = "𝒫";
  var pscr$2 = "𝓅";
  var Psi$2 = "Ψ";
  var psi$2 = "ψ";
  var puncsp$2 = " ";
  var Qfr$2 = "𝔔";
  var qfr$2 = "𝔮";
  var qint$2 = "⨌";
  var qopf$2 = "𝕢";
  var Qopf$2 = "ℚ";
  var qprime$2 = "⁗";
  var Qscr$2 = "𝒬";
  var qscr$2 = "𝓆";
  var quaternions$2 = "ℍ";
  var quatint$2 = "⨖";
  var quest$2 = "?";
  var questeq$2 = "≟";
  var quot$1$2 = "\"";
  var QUOT$3 = "\"";
  var rAarr$2 = "⇛";
  var race$2 = "∽̱";
  var Racute$2 = "Ŕ";
  var racute$2 = "ŕ";
  var radic$2 = "√";
  var raemptyv$2 = "⦳";
  var rang$2 = "⟩";
  var Rang$2 = "⟫";
  var rangd$2 = "⦒";
  var range$2 = "⦥";
  var rangle$2 = "⟩";
  var raquo$3 = "»";
  var rarrap$2 = "⥵";
  var rarrb$2 = "⇥";
  var rarrbfs$2 = "⤠";
  var rarrc$2 = "⤳";
  var rarr$2 = "→";
  var Rarr$2 = "↠";
  var rArr$2 = "⇒";
  var rarrfs$2 = "⤞";
  var rarrhk$2 = "↪";
  var rarrlp$2 = "↬";
  var rarrpl$2 = "⥅";
  var rarrsim$2 = "⥴";
  var Rarrtl$2 = "⤖";
  var rarrtl$2 = "↣";
  var rarrw$2 = "↝";
  var ratail$2 = "⤚";
  var rAtail$2 = "⤜";
  var ratio$2 = "∶";
  var rationals$2 = "ℚ";
  var rbarr$2 = "⤍";
  var rBarr$2 = "⤏";
  var RBarr$2 = "⤐";
  var rbbrk$2 = "❳";
  var rbrace$2 = "}";
  var rbrack$2 = "]";
  var rbrke$2 = "⦌";
  var rbrksld$2 = "⦎";
  var rbrkslu$2 = "⦐";
  var Rcaron$2 = "Ř";
  var rcaron$2 = "ř";
  var Rcedil$2 = "Ŗ";
  var rcedil$2 = "ŗ";
  var rceil$2 = "⌉";
  var rcub$2 = "}";
  var Rcy$2 = "Р";
  var rcy$2 = "р";
  var rdca$2 = "⤷";
  var rdldhar$2 = "⥩";
  var rdquo$2 = "”";
  var rdquor$2 = "”";
  var rdsh$2 = "↳";
  var real$2 = "ℜ";
  var realine$2 = "ℛ";
  var realpart$2 = "ℜ";
  var reals$2 = "ℝ";
  var Re$2 = "ℜ";
  var rect$2 = "▭";
  var reg$3 = "®";
  var REG$3 = "®";
  var ReverseElement$2 = "∋";
  var ReverseEquilibrium$2 = "⇋";
  var ReverseUpEquilibrium$2 = "⥯";
  var rfisht$2 = "⥽";
  var rfloor$2 = "⌋";
  var rfr$2 = "𝔯";
  var Rfr$2 = "ℜ";
  var rHar$2 = "⥤";
  var rhard$2 = "⇁";
  var rharu$2 = "⇀";
  var rharul$2 = "⥬";
  var Rho$2 = "Ρ";
  var rho$2 = "ρ";
  var rhov$2 = "ϱ";
  var RightAngleBracket$2 = "⟩";
  var RightArrowBar$2 = "⇥";
  var rightarrow$2 = "→";
  var RightArrow$2 = "→";
  var Rightarrow$2 = "⇒";
  var RightArrowLeftArrow$2 = "⇄";
  var rightarrowtail$2 = "↣";
  var RightCeiling$2 = "⌉";
  var RightDoubleBracket$2 = "⟧";
  var RightDownTeeVector$2 = "⥝";
  var RightDownVectorBar$2 = "⥕";
  var RightDownVector$2 = "⇂";
  var RightFloor$2 = "⌋";
  var rightharpoondown$2 = "⇁";
  var rightharpoonup$2 = "⇀";
  var rightleftarrows$2 = "⇄";
  var rightleftharpoons$2 = "⇌";
  var rightrightarrows$2 = "⇉";
  var rightsquigarrow$2 = "↝";
  var RightTeeArrow$2 = "↦";
  var RightTee$2 = "⊢";
  var RightTeeVector$2 = "⥛";
  var rightthreetimes$2 = "⋌";
  var RightTriangleBar$2 = "⧐";
  var RightTriangle$2 = "⊳";
  var RightTriangleEqual$2 = "⊵";
  var RightUpDownVector$2 = "⥏";
  var RightUpTeeVector$2 = "⥜";
  var RightUpVectorBar$2 = "⥔";
  var RightUpVector$2 = "↾";
  var RightVectorBar$2 = "⥓";
  var RightVector$2 = "⇀";
  var ring$2 = "˚";
  var risingdotseq$2 = "≓";
  var rlarr$2 = "⇄";
  var rlhar$2 = "⇌";
  var rlm$2 = "‏";
  var rmoustache$2 = "⎱";
  var rmoust$2 = "⎱";
  var rnmid$2 = "⫮";
  var roang$2 = "⟭";
  var roarr$2 = "⇾";
  var robrk$2 = "⟧";
  var ropar$2 = "⦆";
  var ropf$2 = "𝕣";
  var Ropf$2 = "ℝ";
  var roplus$2 = "⨮";
  var rotimes$2 = "⨵";
  var RoundImplies$2 = "⥰";
  var rpar$2 = ")";
  var rpargt$2 = "⦔";
  var rppolint$2 = "⨒";
  var rrarr$2 = "⇉";
  var Rrightarrow$2 = "⇛";
  var rsaquo$2 = "›";
  var rscr$2 = "𝓇";
  var Rscr$2 = "ℛ";
  var rsh$2 = "↱";
  var Rsh$2 = "↱";
  var rsqb$2 = "]";
  var rsquo$2 = "’";
  var rsquor$2 = "’";
  var rthree$2 = "⋌";
  var rtimes$2 = "⋊";
  var rtri$2 = "▹";
  var rtrie$2 = "⊵";
  var rtrif$2 = "▸";
  var rtriltri$2 = "⧎";
  var RuleDelayed$2 = "⧴";
  var ruluhar$2 = "⥨";
  var rx$2 = "℞";
  var Sacute$2 = "Ś";
  var sacute$2 = "ś";
  var sbquo$2 = "‚";
  var scap$2 = "⪸";
  var Scaron$2 = "Š";
  var scaron$2 = "š";
  var Sc$2 = "⪼";
  var sc$2 = "≻";
  var sccue$2 = "≽";
  var sce$2 = "⪰";
  var scE$2 = "⪴";
  var Scedil$2 = "Ş";
  var scedil$2 = "ş";
  var Scirc$2 = "Ŝ";
  var scirc$2 = "ŝ";
  var scnap$2 = "⪺";
  var scnE$2 = "⪶";
  var scnsim$2 = "⋩";
  var scpolint$2 = "⨓";
  var scsim$2 = "≿";
  var Scy$2 = "С";
  var scy$2 = "с";
  var sdotb$2 = "⊡";
  var sdot$2 = "⋅";
  var sdote$2 = "⩦";
  var searhk$2 = "⤥";
  var searr$2 = "↘";
  var seArr$2 = "⇘";
  var searrow$2 = "↘";
  var sect$3 = "§";
  var semi$2 = ";";
  var seswar$2 = "⤩";
  var setminus$2 = "∖";
  var setmn$2 = "∖";
  var sext$2 = "✶";
  var Sfr$2 = "𝔖";
  var sfr$2 = "𝔰";
  var sfrown$2 = "⌢";
  var sharp$2 = "♯";
  var SHCHcy$2 = "Щ";
  var shchcy$2 = "щ";
  var SHcy$2 = "Ш";
  var shcy$2 = "ш";
  var ShortDownArrow$2 = "↓";
  var ShortLeftArrow$2 = "←";
  var shortmid$2 = "∣";
  var shortparallel$2 = "∥";
  var ShortRightArrow$2 = "→";
  var ShortUpArrow$2 = "↑";
  var shy$3 = "­";
  var Sigma$2 = "Σ";
  var sigma$2 = "σ";
  var sigmaf$2 = "ς";
  var sigmav$2 = "ς";
  var sim$2 = "∼";
  var simdot$2 = "⩪";
  var sime$2 = "≃";
  var simeq$2 = "≃";
  var simg$2 = "⪞";
  var simgE$2 = "⪠";
  var siml$2 = "⪝";
  var simlE$2 = "⪟";
  var simne$2 = "≆";
  var simplus$2 = "⨤";
  var simrarr$2 = "⥲";
  var slarr$2 = "←";
  var SmallCircle$2 = "∘";
  var smallsetminus$2 = "∖";
  var smashp$2 = "⨳";
  var smeparsl$2 = "⧤";
  var smid$2 = "∣";
  var smile$2 = "⌣";
  var smt$2 = "⪪";
  var smte$2 = "⪬";
  var smtes$2 = "⪬︀";
  var SOFTcy$2 = "Ь";
  var softcy$2 = "ь";
  var solbar$2 = "⌿";
  var solb$2 = "⧄";
  var sol$2 = "/";
  var Sopf$2 = "𝕊";
  var sopf$2 = "𝕤";
  var spades$2 = "♠";
  var spadesuit$2 = "♠";
  var spar$2 = "∥";
  var sqcap$2 = "⊓";
  var sqcaps$2 = "⊓︀";
  var sqcup$2 = "⊔";
  var sqcups$2 = "⊔︀";
  var Sqrt$2 = "√";
  var sqsub$2 = "⊏";
  var sqsube$2 = "⊑";
  var sqsubset$2 = "⊏";
  var sqsubseteq$2 = "⊑";
  var sqsup$2 = "⊐";
  var sqsupe$2 = "⊒";
  var sqsupset$2 = "⊐";
  var sqsupseteq$2 = "⊒";
  var square$2 = "□";
  var Square$2 = "□";
  var SquareIntersection$2 = "⊓";
  var SquareSubset$2 = "⊏";
  var SquareSubsetEqual$2 = "⊑";
  var SquareSuperset$2 = "⊐";
  var SquareSupersetEqual$2 = "⊒";
  var SquareUnion$2 = "⊔";
  var squarf$2 = "▪";
  var squ$2 = "□";
  var squf$2 = "▪";
  var srarr$2 = "→";
  var Sscr$2 = "𝒮";
  var sscr$2 = "𝓈";
  var ssetmn$2 = "∖";
  var ssmile$2 = "⌣";
  var sstarf$2 = "⋆";
  var Star$2 = "⋆";
  var star$2 = "☆";
  var starf$2 = "★";
  var straightepsilon$2 = "ϵ";
  var straightphi$2 = "ϕ";
  var strns$2 = "¯";
  var sub$2 = "⊂";
  var Sub$2 = "⋐";
  var subdot$2 = "⪽";
  var subE$2 = "⫅";
  var sube$2 = "⊆";
  var subedot$2 = "⫃";
  var submult$2 = "⫁";
  var subnE$2 = "⫋";
  var subne$2 = "⊊";
  var subplus$2 = "⪿";
  var subrarr$2 = "⥹";
  var subset$2 = "⊂";
  var Subset$2 = "⋐";
  var subseteq$2 = "⊆";
  var subseteqq$2 = "⫅";
  var SubsetEqual$2 = "⊆";
  var subsetneq$2 = "⊊";
  var subsetneqq$2 = "⫋";
  var subsim$2 = "⫇";
  var subsub$2 = "⫕";
  var subsup$2 = "⫓";
  var succapprox$2 = "⪸";
  var succ$2 = "≻";
  var succcurlyeq$2 = "≽";
  var Succeeds$2 = "≻";
  var SucceedsEqual$2 = "⪰";
  var SucceedsSlantEqual$2 = "≽";
  var SucceedsTilde$2 = "≿";
  var succeq$2 = "⪰";
  var succnapprox$2 = "⪺";
  var succneqq$2 = "⪶";
  var succnsim$2 = "⋩";
  var succsim$2 = "≿";
  var SuchThat$2 = "∋";
  var sum$2 = "∑";
  var Sum$2 = "∑";
  var sung$2 = "♪";
  var sup1$3 = "¹";
  var sup2$3 = "²";
  var sup3$3 = "³";
  var sup$2 = "⊃";
  var Sup$2 = "⋑";
  var supdot$2 = "⪾";
  var supdsub$2 = "⫘";
  var supE$2 = "⫆";
  var supe$2 = "⊇";
  var supedot$2 = "⫄";
  var Superset$2 = "⊃";
  var SupersetEqual$2 = "⊇";
  var suphsol$2 = "⟉";
  var suphsub$2 = "⫗";
  var suplarr$2 = "⥻";
  var supmult$2 = "⫂";
  var supnE$2 = "⫌";
  var supne$2 = "⊋";
  var supplus$2 = "⫀";
  var supset$2 = "⊃";
  var Supset$2 = "⋑";
  var supseteq$2 = "⊇";
  var supseteqq$2 = "⫆";
  var supsetneq$2 = "⊋";
  var supsetneqq$2 = "⫌";
  var supsim$2 = "⫈";
  var supsub$2 = "⫔";
  var supsup$2 = "⫖";
  var swarhk$2 = "⤦";
  var swarr$2 = "↙";
  var swArr$2 = "⇙";
  var swarrow$2 = "↙";
  var swnwar$2 = "⤪";
  var szlig$3 = "ß";
  var Tab$2 = "\t";
  var target$2 = "⌖";
  var Tau$2 = "Τ";
  var tau$2 = "τ";
  var tbrk$2 = "⎴";
  var Tcaron$2 = "Ť";
  var tcaron$2 = "ť";
  var Tcedil$2 = "Ţ";
  var tcedil$2 = "ţ";
  var Tcy$2 = "Т";
  var tcy$2 = "т";
  var tdot$2 = "⃛";
  var telrec$2 = "⌕";
  var Tfr$2 = "𝔗";
  var tfr$2 = "𝔱";
  var there4$2 = "∴";
  var therefore$2 = "∴";
  var Therefore$2 = "∴";
  var Theta$2 = "Θ";
  var theta$2 = "θ";
  var thetasym$2 = "ϑ";
  var thetav$2 = "ϑ";
  var thickapprox$2 = "≈";
  var thicksim$2 = "∼";
  var ThickSpace$2 = "  ";
  var ThinSpace$2 = " ";
  var thinsp$2 = " ";
  var thkap$2 = "≈";
  var thksim$2 = "∼";
  var THORN$3 = "Þ";
  var thorn$3 = "þ";
  var tilde$2 = "˜";
  var Tilde$2 = "∼";
  var TildeEqual$2 = "≃";
  var TildeFullEqual$2 = "≅";
  var TildeTilde$2 = "≈";
  var timesbar$2 = "⨱";
  var timesb$2 = "⊠";
  var times$4 = "×";
  var timesd$2 = "⨰";
  var tint$2 = "∭";
  var toea$2 = "⤨";
  var topbot$2 = "⌶";
  var topcir$2 = "⫱";
  var top$2 = "⊤";
  var Topf$2 = "𝕋";
  var topf$2 = "𝕥";
  var topfork$2 = "⫚";
  var tosa$2 = "⤩";
  var tprime$2 = "‴";
  var trade$2 = "™";
  var TRADE$2 = "™";
  var triangle$2 = "▵";
  var triangledown$2 = "▿";
  var triangleleft$2 = "◃";
  var trianglelefteq$2 = "⊴";
  var triangleq$2 = "≜";
  var triangleright$2 = "▹";
  var trianglerighteq$2 = "⊵";
  var tridot$2 = "◬";
  var trie$2 = "≜";
  var triminus$2 = "⨺";
  var TripleDot$2 = "⃛";
  var triplus$2 = "⨹";
  var trisb$2 = "⧍";
  var tritime$2 = "⨻";
  var trpezium$2 = "⏢";
  var Tscr$2 = "𝒯";
  var tscr$2 = "𝓉";
  var TScy$2 = "Ц";
  var tscy$2 = "ц";
  var TSHcy$2 = "Ћ";
  var tshcy$2 = "ћ";
  var Tstrok$2 = "Ŧ";
  var tstrok$2 = "ŧ";
  var twixt$2 = "≬";
  var twoheadleftarrow$2 = "↞";
  var twoheadrightarrow$2 = "↠";
  var Uacute$3 = "Ú";
  var uacute$3 = "ú";
  var uarr$2 = "↑";
  var Uarr$2 = "↟";
  var uArr$2 = "⇑";
  var Uarrocir$2 = "⥉";
  var Ubrcy$2 = "Ў";
  var ubrcy$2 = "ў";
  var Ubreve$2 = "Ŭ";
  var ubreve$2 = "ŭ";
  var Ucirc$3 = "Û";
  var ucirc$3 = "û";
  var Ucy$2 = "У";
  var ucy$2 = "у";
  var udarr$2 = "⇅";
  var Udblac$2 = "Ű";
  var udblac$2 = "ű";
  var udhar$2 = "⥮";
  var ufisht$2 = "⥾";
  var Ufr$2 = "𝔘";
  var ufr$2 = "𝔲";
  var Ugrave$3 = "Ù";
  var ugrave$3 = "ù";
  var uHar$2 = "⥣";
  var uharl$2 = "↿";
  var uharr$2 = "↾";
  var uhblk$2 = "▀";
  var ulcorn$2 = "⌜";
  var ulcorner$2 = "⌜";
  var ulcrop$2 = "⌏";
  var ultri$2 = "◸";
  var Umacr$2 = "Ū";
  var umacr$2 = "ū";
  var uml$3 = "¨";
  var UnderBar$2 = "_";
  var UnderBrace$2 = "⏟";
  var UnderBracket$2 = "⎵";
  var UnderParenthesis$2 = "⏝";
  var Union$2 = "⋃";
  var UnionPlus$2 = "⊎";
  var Uogon$2 = "Ų";
  var uogon$2 = "ų";
  var Uopf$2 = "𝕌";
  var uopf$2 = "𝕦";
  var UpArrowBar$2 = "⤒";
  var uparrow$2 = "↑";
  var UpArrow$2 = "↑";
  var Uparrow$2 = "⇑";
  var UpArrowDownArrow$2 = "⇅";
  var updownarrow$2 = "↕";
  var UpDownArrow$2 = "↕";
  var Updownarrow$2 = "⇕";
  var UpEquilibrium$2 = "⥮";
  var upharpoonleft$2 = "↿";
  var upharpoonright$2 = "↾";
  var uplus$2 = "⊎";
  var UpperLeftArrow$2 = "↖";
  var UpperRightArrow$2 = "↗";
  var upsi$2 = "υ";
  var Upsi$2 = "ϒ";
  var upsih$2 = "ϒ";
  var Upsilon$2 = "Υ";
  var upsilon$2 = "υ";
  var UpTeeArrow$2 = "↥";
  var UpTee$2 = "⊥";
  var upuparrows$2 = "⇈";
  var urcorn$2 = "⌝";
  var urcorner$2 = "⌝";
  var urcrop$2 = "⌎";
  var Uring$2 = "Ů";
  var uring$2 = "ů";
  var urtri$2 = "◹";
  var Uscr$2 = "𝒰";
  var uscr$2 = "𝓊";
  var utdot$2 = "⋰";
  var Utilde$2 = "Ũ";
  var utilde$2 = "ũ";
  var utri$2 = "▵";
  var utrif$2 = "▴";
  var uuarr$2 = "⇈";
  var Uuml$3 = "Ü";
  var uuml$3 = "ü";
  var uwangle$2 = "⦧";
  var vangrt$2 = "⦜";
  var varepsilon$2 = "ϵ";
  var varkappa$2 = "ϰ";
  var varnothing$2 = "∅";
  var varphi$2 = "ϕ";
  var varpi$2 = "ϖ";
  var varpropto$2 = "∝";
  var varr$2 = "↕";
  var vArr$2 = "⇕";
  var varrho$2 = "ϱ";
  var varsigma$2 = "ς";
  var varsubsetneq$2 = "⊊︀";
  var varsubsetneqq$2 = "⫋︀";
  var varsupsetneq$2 = "⊋︀";
  var varsupsetneqq$2 = "⫌︀";
  var vartheta$2 = "ϑ";
  var vartriangleleft$2 = "⊲";
  var vartriangleright$2 = "⊳";
  var vBar$2 = "⫨";
  var Vbar$2 = "⫫";
  var vBarv$2 = "⫩";
  var Vcy$2 = "В";
  var vcy$2 = "в";
  var vdash$2 = "⊢";
  var vDash$2 = "⊨";
  var Vdash$2 = "⊩";
  var VDash$2 = "⊫";
  var Vdashl$2 = "⫦";
  var veebar$2 = "⊻";
  var vee$2 = "∨";
  var Vee$2 = "⋁";
  var veeeq$2 = "≚";
  var vellip$2 = "⋮";
  var verbar$2 = "|";
  var Verbar$2 = "‖";
  var vert$2 = "|";
  var Vert$2 = "‖";
  var VerticalBar$2 = "∣";
  var VerticalLine$2 = "|";
  var VerticalSeparator$2 = "❘";
  var VerticalTilde$2 = "≀";
  var VeryThinSpace$2 = " ";
  var Vfr$2 = "𝔙";
  var vfr$2 = "𝔳";
  var vltri$2 = "⊲";
  var vnsub$2 = "⊂⃒";
  var vnsup$2 = "⊃⃒";
  var Vopf$2 = "𝕍";
  var vopf$2 = "𝕧";
  var vprop$2 = "∝";
  var vrtri$2 = "⊳";
  var Vscr$2 = "𝒱";
  var vscr$2 = "𝓋";
  var vsubnE$2 = "⫋︀";
  var vsubne$2 = "⊊︀";
  var vsupnE$2 = "⫌︀";
  var vsupne$2 = "⊋︀";
  var Vvdash$2 = "⊪";
  var vzigzag$2 = "⦚";
  var Wcirc$2 = "Ŵ";
  var wcirc$2 = "ŵ";
  var wedbar$2 = "⩟";
  var wedge$2 = "∧";
  var Wedge$2 = "⋀";
  var wedgeq$2 = "≙";
  var weierp$2 = "℘";
  var Wfr$2 = "𝔚";
  var wfr$2 = "𝔴";
  var Wopf$2 = "𝕎";
  var wopf$2 = "𝕨";
  var wp$2 = "℘";
  var wr$2 = "≀";
  var wreath$2 = "≀";
  var Wscr$2 = "𝒲";
  var wscr$2 = "𝓌";
  var xcap$2 = "⋂";
  var xcirc$2 = "◯";
  var xcup$2 = "⋃";
  var xdtri$2 = "▽";
  var Xfr$2 = "𝔛";
  var xfr$2 = "𝔵";
  var xharr$2 = "⟷";
  var xhArr$2 = "⟺";
  var Xi$2 = "Ξ";
  var xi$2 = "ξ";
  var xlarr$2 = "⟵";
  var xlArr$2 = "⟸";
  var xmap$2 = "⟼";
  var xnis$2 = "⋻";
  var xodot$2 = "⨀";
  var Xopf$2 = "𝕏";
  var xopf$2 = "𝕩";
  var xoplus$2 = "⨁";
  var xotime$2 = "⨂";
  var xrarr$2 = "⟶";
  var xrArr$2 = "⟹";
  var Xscr$2 = "𝒳";
  var xscr$2 = "𝓍";
  var xsqcup$2 = "⨆";
  var xuplus$2 = "⨄";
  var xutri$2 = "△";
  var xvee$2 = "⋁";
  var xwedge$2 = "⋀";
  var Yacute$3 = "Ý";
  var yacute$3 = "ý";
  var YAcy$2 = "Я";
  var yacy$2 = "я";
  var Ycirc$2 = "Ŷ";
  var ycirc$2 = "ŷ";
  var Ycy$2 = "Ы";
  var ycy$2 = "ы";
  var yen$3 = "¥";
  var Yfr$2 = "𝔜";
  var yfr$2 = "𝔶";
  var YIcy$2 = "Ї";
  var yicy$2 = "ї";
  var Yopf$2 = "𝕐";
  var yopf$2 = "𝕪";
  var Yscr$2 = "𝒴";
  var yscr$2 = "𝓎";
  var YUcy$2 = "Ю";
  var yucy$2 = "ю";
  var yuml$3 = "ÿ";
  var Yuml$2 = "Ÿ";
  var Zacute$2 = "Ź";
  var zacute$2 = "ź";
  var Zcaron$2 = "Ž";
  var zcaron$2 = "ž";
  var Zcy$2 = "З";
  var zcy$2 = "з";
  var Zdot$2 = "Ż";
  var zdot$2 = "ż";
  var zeetrf$2 = "ℨ";
  var ZeroWidthSpace$2 = "​";
  var Zeta$2 = "Ζ";
  var zeta$2 = "ζ";
  var zfr$2 = "𝔷";
  var Zfr$2 = "ℨ";
  var ZHcy$2 = "Ж";
  var zhcy$2 = "ж";
  var zigrarr$2 = "⇝";
  var zopf$2 = "𝕫";
  var Zopf$2 = "ℤ";
  var Zscr$2 = "𝒵";
  var zscr$2 = "𝓏";
  var zwj$2 = "‍";
  var zwnj$2 = "‌";
  var entitiesJSON$2 = {
  	Aacute: Aacute$3,
  	aacute: aacute$3,
  	Abreve: Abreve$2,
  	abreve: abreve$2,
  	ac: ac$2,
  	acd: acd$2,
  	acE: acE$2,
  	Acirc: Acirc$3,
  	acirc: acirc$3,
  	acute: acute$3,
  	Acy: Acy$2,
  	acy: acy$2,
  	AElig: AElig$3,
  	aelig: aelig$3,
  	af: af$2,
  	Afr: Afr$2,
  	afr: afr$2,
  	Agrave: Agrave$3,
  	agrave: agrave$3,
  	alefsym: alefsym$2,
  	aleph: aleph$2,
  	Alpha: Alpha$2,
  	alpha: alpha$2,
  	Amacr: Amacr$2,
  	amacr: amacr$2,
  	amalg: amalg$2,
  	amp: amp$1$2,
  	AMP: AMP$3,
  	andand: andand$2,
  	And: And$2,
  	and: and$2,
  	andd: andd$2,
  	andslope: andslope$2,
  	andv: andv$2,
  	ang: ang$2,
  	ange: ange$2,
  	angle: angle$2,
  	angmsdaa: angmsdaa$2,
  	angmsdab: angmsdab$2,
  	angmsdac: angmsdac$2,
  	angmsdad: angmsdad$2,
  	angmsdae: angmsdae$2,
  	angmsdaf: angmsdaf$2,
  	angmsdag: angmsdag$2,
  	angmsdah: angmsdah$2,
  	angmsd: angmsd$2,
  	angrt: angrt$2,
  	angrtvb: angrtvb$2,
  	angrtvbd: angrtvbd$2,
  	angsph: angsph$2,
  	angst: angst$2,
  	angzarr: angzarr$2,
  	Aogon: Aogon$2,
  	aogon: aogon$2,
  	Aopf: Aopf$2,
  	aopf: aopf$2,
  	apacir: apacir$2,
  	ap: ap$2,
  	apE: apE$2,
  	ape: ape$2,
  	apid: apid$2,
  	apos: apos$1$2,
  	ApplyFunction: ApplyFunction$2,
  	approx: approx$2,
  	approxeq: approxeq$2,
  	Aring: Aring$3,
  	aring: aring$3,
  	Ascr: Ascr$2,
  	ascr: ascr$2,
  	Assign: Assign$2,
  	ast: ast$2,
  	asymp: asymp$2,
  	asympeq: asympeq$2,
  	Atilde: Atilde$3,
  	atilde: atilde$3,
  	Auml: Auml$3,
  	auml: auml$3,
  	awconint: awconint$2,
  	awint: awint$2,
  	backcong: backcong$2,
  	backepsilon: backepsilon$2,
  	backprime: backprime$2,
  	backsim: backsim$2,
  	backsimeq: backsimeq$2,
  	Backslash: Backslash$2,
  	Barv: Barv$2,
  	barvee: barvee$2,
  	barwed: barwed$2,
  	Barwed: Barwed$2,
  	barwedge: barwedge$2,
  	bbrk: bbrk$2,
  	bbrktbrk: bbrktbrk$2,
  	bcong: bcong$2,
  	Bcy: Bcy$2,
  	bcy: bcy$2,
  	bdquo: bdquo$2,
  	becaus: becaus$2,
  	because: because$2,
  	Because: Because$2,
  	bemptyv: bemptyv$2,
  	bepsi: bepsi$2,
  	bernou: bernou$2,
  	Bernoullis: Bernoullis$2,
  	Beta: Beta$2,
  	beta: beta$2,
  	beth: beth$2,
  	between: between$2,
  	Bfr: Bfr$2,
  	bfr: bfr$2,
  	bigcap: bigcap$2,
  	bigcirc: bigcirc$2,
  	bigcup: bigcup$2,
  	bigodot: bigodot$2,
  	bigoplus: bigoplus$2,
  	bigotimes: bigotimes$2,
  	bigsqcup: bigsqcup$2,
  	bigstar: bigstar$2,
  	bigtriangledown: bigtriangledown$2,
  	bigtriangleup: bigtriangleup$2,
  	biguplus: biguplus$2,
  	bigvee: bigvee$2,
  	bigwedge: bigwedge$2,
  	bkarow: bkarow$2,
  	blacklozenge: blacklozenge$2,
  	blacksquare: blacksquare$2,
  	blacktriangle: blacktriangle$2,
  	blacktriangledown: blacktriangledown$2,
  	blacktriangleleft: blacktriangleleft$2,
  	blacktriangleright: blacktriangleright$2,
  	blank: blank$2,
  	blk12: blk12$2,
  	blk14: blk14$2,
  	blk34: blk34$2,
  	block: block$2,
  	bne: bne$2,
  	bnequiv: bnequiv$2,
  	bNot: bNot$2,
  	bnot: bnot$2,
  	Bopf: Bopf$2,
  	bopf: bopf$2,
  	bot: bot$2,
  	bottom: bottom$2,
  	bowtie: bowtie$2,
  	boxbox: boxbox$2,
  	boxdl: boxdl$2,
  	boxdL: boxdL$2,
  	boxDl: boxDl$2,
  	boxDL: boxDL$2,
  	boxdr: boxdr$2,
  	boxdR: boxdR$2,
  	boxDr: boxDr$2,
  	boxDR: boxDR$2,
  	boxh: boxh$2,
  	boxH: boxH$2,
  	boxhd: boxhd$2,
  	boxHd: boxHd$2,
  	boxhD: boxhD$2,
  	boxHD: boxHD$2,
  	boxhu: boxhu$2,
  	boxHu: boxHu$2,
  	boxhU: boxhU$2,
  	boxHU: boxHU$2,
  	boxminus: boxminus$2,
  	boxplus: boxplus$2,
  	boxtimes: boxtimes$2,
  	boxul: boxul$2,
  	boxuL: boxuL$2,
  	boxUl: boxUl$2,
  	boxUL: boxUL$2,
  	boxur: boxur$2,
  	boxuR: boxuR$2,
  	boxUr: boxUr$2,
  	boxUR: boxUR$2,
  	boxv: boxv$2,
  	boxV: boxV$2,
  	boxvh: boxvh$2,
  	boxvH: boxvH$2,
  	boxVh: boxVh$2,
  	boxVH: boxVH$2,
  	boxvl: boxvl$2,
  	boxvL: boxvL$2,
  	boxVl: boxVl$2,
  	boxVL: boxVL$2,
  	boxvr: boxvr$2,
  	boxvR: boxvR$2,
  	boxVr: boxVr$2,
  	boxVR: boxVR$2,
  	bprime: bprime$2,
  	breve: breve$2,
  	Breve: Breve$2,
  	brvbar: brvbar$3,
  	bscr: bscr$2,
  	Bscr: Bscr$2,
  	bsemi: bsemi$2,
  	bsim: bsim$2,
  	bsime: bsime$2,
  	bsolb: bsolb$2,
  	bsol: bsol$2,
  	bsolhsub: bsolhsub$2,
  	bull: bull$2,
  	bullet: bullet$2,
  	bump: bump$2,
  	bumpE: bumpE$2,
  	bumpe: bumpe$2,
  	Bumpeq: Bumpeq$2,
  	bumpeq: bumpeq$2,
  	Cacute: Cacute$2,
  	cacute: cacute$2,
  	capand: capand$2,
  	capbrcup: capbrcup$2,
  	capcap: capcap$2,
  	cap: cap$2,
  	Cap: Cap$2,
  	capcup: capcup$2,
  	capdot: capdot$2,
  	CapitalDifferentialD: CapitalDifferentialD$2,
  	caps: caps$2,
  	caret: caret$2,
  	caron: caron$2,
  	Cayleys: Cayleys$2,
  	ccaps: ccaps$2,
  	Ccaron: Ccaron$2,
  	ccaron: ccaron$2,
  	Ccedil: Ccedil$3,
  	ccedil: ccedil$3,
  	Ccirc: Ccirc$2,
  	ccirc: ccirc$2,
  	Cconint: Cconint$2,
  	ccups: ccups$2,
  	ccupssm: ccupssm$2,
  	Cdot: Cdot$2,
  	cdot: cdot$2,
  	cedil: cedil$3,
  	Cedilla: Cedilla$2,
  	cemptyv: cemptyv$2,
  	cent: cent$3,
  	centerdot: centerdot$2,
  	CenterDot: CenterDot$2,
  	cfr: cfr$2,
  	Cfr: Cfr$2,
  	CHcy: CHcy$2,
  	chcy: chcy$2,
  	check: check$2,
  	checkmark: checkmark$2,
  	Chi: Chi$2,
  	chi: chi$2,
  	circ: circ$2,
  	circeq: circeq$2,
  	circlearrowleft: circlearrowleft$2,
  	circlearrowright: circlearrowright$2,
  	circledast: circledast$2,
  	circledcirc: circledcirc$2,
  	circleddash: circleddash$2,
  	CircleDot: CircleDot$2,
  	circledR: circledR$2,
  	circledS: circledS$2,
  	CircleMinus: CircleMinus$2,
  	CirclePlus: CirclePlus$2,
  	CircleTimes: CircleTimes$2,
  	cir: cir$2,
  	cirE: cirE$2,
  	cire: cire$2,
  	cirfnint: cirfnint$2,
  	cirmid: cirmid$2,
  	cirscir: cirscir$2,
  	ClockwiseContourIntegral: ClockwiseContourIntegral$2,
  	CloseCurlyDoubleQuote: CloseCurlyDoubleQuote$2,
  	CloseCurlyQuote: CloseCurlyQuote$2,
  	clubs: clubs$2,
  	clubsuit: clubsuit$2,
  	colon: colon$2,
  	Colon: Colon$2,
  	Colone: Colone$2,
  	colone: colone$2,
  	coloneq: coloneq$2,
  	comma: comma$2,
  	commat: commat$2,
  	comp: comp$2,
  	compfn: compfn$2,
  	complement: complement$2,
  	complexes: complexes$2,
  	cong: cong$2,
  	congdot: congdot$2,
  	Congruent: Congruent$2,
  	conint: conint$2,
  	Conint: Conint$2,
  	ContourIntegral: ContourIntegral$2,
  	copf: copf$2,
  	Copf: Copf$2,
  	coprod: coprod$2,
  	Coproduct: Coproduct$2,
  	copy: copy$3,
  	COPY: COPY$3,
  	copysr: copysr$2,
  	CounterClockwiseContourIntegral: CounterClockwiseContourIntegral$2,
  	crarr: crarr$2,
  	cross: cross$2,
  	Cross: Cross$2,
  	Cscr: Cscr$2,
  	cscr: cscr$2,
  	csub: csub$2,
  	csube: csube$2,
  	csup: csup$2,
  	csupe: csupe$2,
  	ctdot: ctdot$2,
  	cudarrl: cudarrl$2,
  	cudarrr: cudarrr$2,
  	cuepr: cuepr$2,
  	cuesc: cuesc$2,
  	cularr: cularr$2,
  	cularrp: cularrp$2,
  	cupbrcap: cupbrcap$2,
  	cupcap: cupcap$2,
  	CupCap: CupCap$2,
  	cup: cup$2,
  	Cup: Cup$2,
  	cupcup: cupcup$2,
  	cupdot: cupdot$2,
  	cupor: cupor$2,
  	cups: cups$2,
  	curarr: curarr$2,
  	curarrm: curarrm$2,
  	curlyeqprec: curlyeqprec$2,
  	curlyeqsucc: curlyeqsucc$2,
  	curlyvee: curlyvee$2,
  	curlywedge: curlywedge$2,
  	curren: curren$3,
  	curvearrowleft: curvearrowleft$2,
  	curvearrowright: curvearrowright$2,
  	cuvee: cuvee$2,
  	cuwed: cuwed$2,
  	cwconint: cwconint$2,
  	cwint: cwint$2,
  	cylcty: cylcty$2,
  	dagger: dagger$2,
  	Dagger: Dagger$2,
  	daleth: daleth$2,
  	darr: darr$2,
  	Darr: Darr$2,
  	dArr: dArr$2,
  	dash: dash$2,
  	Dashv: Dashv$2,
  	dashv: dashv$2,
  	dbkarow: dbkarow$2,
  	dblac: dblac$2,
  	Dcaron: Dcaron$2,
  	dcaron: dcaron$2,
  	Dcy: Dcy$2,
  	dcy: dcy$2,
  	ddagger: ddagger$2,
  	ddarr: ddarr$2,
  	DD: DD$2,
  	dd: dd$2,
  	DDotrahd: DDotrahd$2,
  	ddotseq: ddotseq$2,
  	deg: deg$3,
  	Del: Del$2,
  	Delta: Delta$2,
  	delta: delta$2,
  	demptyv: demptyv$2,
  	dfisht: dfisht$2,
  	Dfr: Dfr$2,
  	dfr: dfr$2,
  	dHar: dHar$2,
  	dharl: dharl$2,
  	dharr: dharr$2,
  	DiacriticalAcute: DiacriticalAcute$2,
  	DiacriticalDot: DiacriticalDot$2,
  	DiacriticalDoubleAcute: DiacriticalDoubleAcute$2,
  	DiacriticalGrave: DiacriticalGrave$2,
  	DiacriticalTilde: DiacriticalTilde$2,
  	diam: diam$2,
  	diamond: diamond$2,
  	Diamond: Diamond$2,
  	diamondsuit: diamondsuit$2,
  	diams: diams$2,
  	die: die$2,
  	DifferentialD: DifferentialD$2,
  	digamma: digamma$2,
  	disin: disin$2,
  	div: div$2,
  	divide: divide$3,
  	divideontimes: divideontimes$2,
  	divonx: divonx$2,
  	DJcy: DJcy$2,
  	djcy: djcy$2,
  	dlcorn: dlcorn$2,
  	dlcrop: dlcrop$2,
  	dollar: dollar$2,
  	Dopf: Dopf$2,
  	dopf: dopf$2,
  	Dot: Dot$2,
  	dot: dot$2,
  	DotDot: DotDot$2,
  	doteq: doteq$2,
  	doteqdot: doteqdot$2,
  	DotEqual: DotEqual$2,
  	dotminus: dotminus$2,
  	dotplus: dotplus$2,
  	dotsquare: dotsquare$2,
  	doublebarwedge: doublebarwedge$2,
  	DoubleContourIntegral: DoubleContourIntegral$2,
  	DoubleDot: DoubleDot$2,
  	DoubleDownArrow: DoubleDownArrow$2,
  	DoubleLeftArrow: DoubleLeftArrow$2,
  	DoubleLeftRightArrow: DoubleLeftRightArrow$2,
  	DoubleLeftTee: DoubleLeftTee$2,
  	DoubleLongLeftArrow: DoubleLongLeftArrow$2,
  	DoubleLongLeftRightArrow: DoubleLongLeftRightArrow$2,
  	DoubleLongRightArrow: DoubleLongRightArrow$2,
  	DoubleRightArrow: DoubleRightArrow$2,
  	DoubleRightTee: DoubleRightTee$2,
  	DoubleUpArrow: DoubleUpArrow$2,
  	DoubleUpDownArrow: DoubleUpDownArrow$2,
  	DoubleVerticalBar: DoubleVerticalBar$2,
  	DownArrowBar: DownArrowBar$2,
  	downarrow: downarrow$2,
  	DownArrow: DownArrow$2,
  	Downarrow: Downarrow$2,
  	DownArrowUpArrow: DownArrowUpArrow$2,
  	DownBreve: DownBreve$2,
  	downdownarrows: downdownarrows$2,
  	downharpoonleft: downharpoonleft$2,
  	downharpoonright: downharpoonright$2,
  	DownLeftRightVector: DownLeftRightVector$2,
  	DownLeftTeeVector: DownLeftTeeVector$2,
  	DownLeftVectorBar: DownLeftVectorBar$2,
  	DownLeftVector: DownLeftVector$2,
  	DownRightTeeVector: DownRightTeeVector$2,
  	DownRightVectorBar: DownRightVectorBar$2,
  	DownRightVector: DownRightVector$2,
  	DownTeeArrow: DownTeeArrow$2,
  	DownTee: DownTee$2,
  	drbkarow: drbkarow$2,
  	drcorn: drcorn$2,
  	drcrop: drcrop$2,
  	Dscr: Dscr$2,
  	dscr: dscr$2,
  	DScy: DScy$2,
  	dscy: dscy$2,
  	dsol: dsol$2,
  	Dstrok: Dstrok$2,
  	dstrok: dstrok$2,
  	dtdot: dtdot$2,
  	dtri: dtri$2,
  	dtrif: dtrif$2,
  	duarr: duarr$2,
  	duhar: duhar$2,
  	dwangle: dwangle$2,
  	DZcy: DZcy$2,
  	dzcy: dzcy$2,
  	dzigrarr: dzigrarr$2,
  	Eacute: Eacute$3,
  	eacute: eacute$3,
  	easter: easter$2,
  	Ecaron: Ecaron$2,
  	ecaron: ecaron$2,
  	Ecirc: Ecirc$3,
  	ecirc: ecirc$3,
  	ecir: ecir$2,
  	ecolon: ecolon$2,
  	Ecy: Ecy$2,
  	ecy: ecy$2,
  	eDDot: eDDot$2,
  	Edot: Edot$2,
  	edot: edot$2,
  	eDot: eDot$2,
  	ee: ee$2,
  	efDot: efDot$2,
  	Efr: Efr$2,
  	efr: efr$2,
  	eg: eg$2,
  	Egrave: Egrave$3,
  	egrave: egrave$3,
  	egs: egs$2,
  	egsdot: egsdot$2,
  	el: el$2,
  	Element: Element$2,
  	elinters: elinters$2,
  	ell: ell$2,
  	els: els$2,
  	elsdot: elsdot$2,
  	Emacr: Emacr$2,
  	emacr: emacr$2,
  	empty: empty$2,
  	emptyset: emptyset$2,
  	EmptySmallSquare: EmptySmallSquare$2,
  	emptyv: emptyv$2,
  	EmptyVerySmallSquare: EmptyVerySmallSquare$2,
  	emsp13: emsp13$2,
  	emsp14: emsp14$2,
  	emsp: emsp$2,
  	ENG: ENG$2,
  	eng: eng$2,
  	ensp: ensp$2,
  	Eogon: Eogon$2,
  	eogon: eogon$2,
  	Eopf: Eopf$2,
  	eopf: eopf$2,
  	epar: epar$2,
  	eparsl: eparsl$2,
  	eplus: eplus$2,
  	epsi: epsi$2,
  	Epsilon: Epsilon$2,
  	epsilon: epsilon$2,
  	epsiv: epsiv$2,
  	eqcirc: eqcirc$2,
  	eqcolon: eqcolon$2,
  	eqsim: eqsim$2,
  	eqslantgtr: eqslantgtr$2,
  	eqslantless: eqslantless$2,
  	Equal: Equal$2,
  	equals: equals$2,
  	EqualTilde: EqualTilde$2,
  	equest: equest$2,
  	Equilibrium: Equilibrium$2,
  	equiv: equiv$2,
  	equivDD: equivDD$2,
  	eqvparsl: eqvparsl$2,
  	erarr: erarr$2,
  	erDot: erDot$2,
  	escr: escr$2,
  	Escr: Escr$2,
  	esdot: esdot$2,
  	Esim: Esim$2,
  	esim: esim$2,
  	Eta: Eta$2,
  	eta: eta$2,
  	ETH: ETH$3,
  	eth: eth$3,
  	Euml: Euml$3,
  	euml: euml$3,
  	euro: euro$2,
  	excl: excl$2,
  	exist: exist$2,
  	Exists: Exists$2,
  	expectation: expectation$2,
  	exponentiale: exponentiale$2,
  	ExponentialE: ExponentialE$2,
  	fallingdotseq: fallingdotseq$2,
  	Fcy: Fcy$2,
  	fcy: fcy$2,
  	female: female$2,
  	ffilig: ffilig$2,
  	fflig: fflig$2,
  	ffllig: ffllig$2,
  	Ffr: Ffr$2,
  	ffr: ffr$2,
  	filig: filig$2,
  	FilledSmallSquare: FilledSmallSquare$2,
  	FilledVerySmallSquare: FilledVerySmallSquare$2,
  	fjlig: fjlig$2,
  	flat: flat$2,
  	fllig: fllig$2,
  	fltns: fltns$2,
  	fnof: fnof$2,
  	Fopf: Fopf$2,
  	fopf: fopf$2,
  	forall: forall$2,
  	ForAll: ForAll$2,
  	fork: fork$2,
  	forkv: forkv$2,
  	Fouriertrf: Fouriertrf$2,
  	fpartint: fpartint$2,
  	frac12: frac12$3,
  	frac13: frac13$2,
  	frac14: frac14$3,
  	frac15: frac15$2,
  	frac16: frac16$2,
  	frac18: frac18$2,
  	frac23: frac23$2,
  	frac25: frac25$2,
  	frac34: frac34$3,
  	frac35: frac35$2,
  	frac38: frac38$2,
  	frac45: frac45$2,
  	frac56: frac56$2,
  	frac58: frac58$2,
  	frac78: frac78$2,
  	frasl: frasl$2,
  	frown: frown$2,
  	fscr: fscr$2,
  	Fscr: Fscr$2,
  	gacute: gacute$2,
  	Gamma: Gamma$2,
  	gamma: gamma$2,
  	Gammad: Gammad$2,
  	gammad: gammad$2,
  	gap: gap$2,
  	Gbreve: Gbreve$2,
  	gbreve: gbreve$2,
  	Gcedil: Gcedil$2,
  	Gcirc: Gcirc$2,
  	gcirc: gcirc$2,
  	Gcy: Gcy$2,
  	gcy: gcy$2,
  	Gdot: Gdot$2,
  	gdot: gdot$2,
  	ge: ge$2,
  	gE: gE$2,
  	gEl: gEl$2,
  	gel: gel$2,
  	geq: geq$2,
  	geqq: geqq$2,
  	geqslant: geqslant$2,
  	gescc: gescc$2,
  	ges: ges$2,
  	gesdot: gesdot$2,
  	gesdoto: gesdoto$2,
  	gesdotol: gesdotol$2,
  	gesl: gesl$2,
  	gesles: gesles$2,
  	Gfr: Gfr$2,
  	gfr: gfr$2,
  	gg: gg$2,
  	Gg: Gg$2,
  	ggg: ggg$2,
  	gimel: gimel$2,
  	GJcy: GJcy$2,
  	gjcy: gjcy$2,
  	gla: gla$2,
  	gl: gl$2,
  	glE: glE$2,
  	glj: glj$2,
  	gnap: gnap$2,
  	gnapprox: gnapprox$2,
  	gne: gne$2,
  	gnE: gnE$2,
  	gneq: gneq$2,
  	gneqq: gneqq$2,
  	gnsim: gnsim$2,
  	Gopf: Gopf$2,
  	gopf: gopf$2,
  	grave: grave$2,
  	GreaterEqual: GreaterEqual$2,
  	GreaterEqualLess: GreaterEqualLess$2,
  	GreaterFullEqual: GreaterFullEqual$2,
  	GreaterGreater: GreaterGreater$2,
  	GreaterLess: GreaterLess$2,
  	GreaterSlantEqual: GreaterSlantEqual$2,
  	GreaterTilde: GreaterTilde$2,
  	Gscr: Gscr$2,
  	gscr: gscr$2,
  	gsim: gsim$2,
  	gsime: gsime$2,
  	gsiml: gsiml$2,
  	gtcc: gtcc$2,
  	gtcir: gtcir$2,
  	gt: gt$1$2,
  	GT: GT$3,
  	Gt: Gt$2,
  	gtdot: gtdot$2,
  	gtlPar: gtlPar$2,
  	gtquest: gtquest$2,
  	gtrapprox: gtrapprox$2,
  	gtrarr: gtrarr$2,
  	gtrdot: gtrdot$2,
  	gtreqless: gtreqless$2,
  	gtreqqless: gtreqqless$2,
  	gtrless: gtrless$2,
  	gtrsim: gtrsim$2,
  	gvertneqq: gvertneqq$2,
  	gvnE: gvnE$2,
  	Hacek: Hacek$2,
  	hairsp: hairsp$2,
  	half: half$2,
  	hamilt: hamilt$2,
  	HARDcy: HARDcy$2,
  	hardcy: hardcy$2,
  	harrcir: harrcir$2,
  	harr: harr$2,
  	hArr: hArr$2,
  	harrw: harrw$2,
  	Hat: Hat$2,
  	hbar: hbar$2,
  	Hcirc: Hcirc$2,
  	hcirc: hcirc$2,
  	hearts: hearts$2,
  	heartsuit: heartsuit$2,
  	hellip: hellip$2,
  	hercon: hercon$2,
  	hfr: hfr$2,
  	Hfr: Hfr$2,
  	HilbertSpace: HilbertSpace$2,
  	hksearow: hksearow$2,
  	hkswarow: hkswarow$2,
  	hoarr: hoarr$2,
  	homtht: homtht$2,
  	hookleftarrow: hookleftarrow$2,
  	hookrightarrow: hookrightarrow$2,
  	hopf: hopf$2,
  	Hopf: Hopf$2,
  	horbar: horbar$2,
  	HorizontalLine: HorizontalLine$2,
  	hscr: hscr$2,
  	Hscr: Hscr$2,
  	hslash: hslash$2,
  	Hstrok: Hstrok$2,
  	hstrok: hstrok$2,
  	HumpDownHump: HumpDownHump$2,
  	HumpEqual: HumpEqual$2,
  	hybull: hybull$2,
  	hyphen: hyphen$2,
  	Iacute: Iacute$3,
  	iacute: iacute$3,
  	ic: ic$2,
  	Icirc: Icirc$3,
  	icirc: icirc$3,
  	Icy: Icy$2,
  	icy: icy$2,
  	Idot: Idot$2,
  	IEcy: IEcy$2,
  	iecy: iecy$2,
  	iexcl: iexcl$3,
  	iff: iff$2,
  	ifr: ifr$2,
  	Ifr: Ifr$2,
  	Igrave: Igrave$3,
  	igrave: igrave$3,
  	ii: ii$2,
  	iiiint: iiiint$2,
  	iiint: iiint$2,
  	iinfin: iinfin$2,
  	iiota: iiota$2,
  	IJlig: IJlig$2,
  	ijlig: ijlig$2,
  	Imacr: Imacr$2,
  	imacr: imacr$2,
  	image: image$2,
  	ImaginaryI: ImaginaryI$2,
  	imagline: imagline$2,
  	imagpart: imagpart$2,
  	imath: imath$2,
  	Im: Im$2,
  	imof: imof$2,
  	imped: imped$2,
  	Implies: Implies$2,
  	incare: incare$2,
  	infin: infin$2,
  	infintie: infintie$2,
  	inodot: inodot$2,
  	intcal: intcal$2,
  	int: int$2,
  	Int: Int$2,
  	integers: integers$2,
  	Integral: Integral$2,
  	intercal: intercal$2,
  	Intersection: Intersection$2,
  	intlarhk: intlarhk$2,
  	intprod: intprod$2,
  	InvisibleComma: InvisibleComma$2,
  	InvisibleTimes: InvisibleTimes$2,
  	IOcy: IOcy$2,
  	iocy: iocy$2,
  	Iogon: Iogon$2,
  	iogon: iogon$2,
  	Iopf: Iopf$2,
  	iopf: iopf$2,
  	Iota: Iota$2,
  	iota: iota$2,
  	iprod: iprod$2,
  	iquest: iquest$3,
  	iscr: iscr$2,
  	Iscr: Iscr$2,
  	isin: isin$2,
  	isindot: isindot$2,
  	isinE: isinE$2,
  	isins: isins$2,
  	isinsv: isinsv$2,
  	isinv: isinv$2,
  	it: it$2,
  	Itilde: Itilde$2,
  	itilde: itilde$2,
  	Iukcy: Iukcy$2,
  	iukcy: iukcy$2,
  	Iuml: Iuml$3,
  	iuml: iuml$3,
  	Jcirc: Jcirc$2,
  	jcirc: jcirc$2,
  	Jcy: Jcy$2,
  	jcy: jcy$2,
  	Jfr: Jfr$2,
  	jfr: jfr$2,
  	jmath: jmath$2,
  	Jopf: Jopf$2,
  	jopf: jopf$2,
  	Jscr: Jscr$2,
  	jscr: jscr$2,
  	Jsercy: Jsercy$2,
  	jsercy: jsercy$2,
  	Jukcy: Jukcy$2,
  	jukcy: jukcy$2,
  	Kappa: Kappa$2,
  	kappa: kappa$2,
  	kappav: kappav$2,
  	Kcedil: Kcedil$2,
  	kcedil: kcedil$2,
  	Kcy: Kcy$2,
  	kcy: kcy$2,
  	Kfr: Kfr$2,
  	kfr: kfr$2,
  	kgreen: kgreen$2,
  	KHcy: KHcy$2,
  	khcy: khcy$2,
  	KJcy: KJcy$2,
  	kjcy: kjcy$2,
  	Kopf: Kopf$2,
  	kopf: kopf$2,
  	Kscr: Kscr$2,
  	kscr: kscr$2,
  	lAarr: lAarr$2,
  	Lacute: Lacute$2,
  	lacute: lacute$2,
  	laemptyv: laemptyv$2,
  	lagran: lagran$2,
  	Lambda: Lambda$2,
  	lambda: lambda$2,
  	lang: lang$2,
  	Lang: Lang$2,
  	langd: langd$2,
  	langle: langle$2,
  	lap: lap$2,
  	Laplacetrf: Laplacetrf$2,
  	laquo: laquo$3,
  	larrb: larrb$2,
  	larrbfs: larrbfs$2,
  	larr: larr$2,
  	Larr: Larr$2,
  	lArr: lArr$2,
  	larrfs: larrfs$2,
  	larrhk: larrhk$2,
  	larrlp: larrlp$2,
  	larrpl: larrpl$2,
  	larrsim: larrsim$2,
  	larrtl: larrtl$2,
  	latail: latail$2,
  	lAtail: lAtail$2,
  	lat: lat$2,
  	late: late$2,
  	lates: lates$2,
  	lbarr: lbarr$2,
  	lBarr: lBarr$2,
  	lbbrk: lbbrk$2,
  	lbrace: lbrace$2,
  	lbrack: lbrack$2,
  	lbrke: lbrke$2,
  	lbrksld: lbrksld$2,
  	lbrkslu: lbrkslu$2,
  	Lcaron: Lcaron$2,
  	lcaron: lcaron$2,
  	Lcedil: Lcedil$2,
  	lcedil: lcedil$2,
  	lceil: lceil$2,
  	lcub: lcub$2,
  	Lcy: Lcy$2,
  	lcy: lcy$2,
  	ldca: ldca$2,
  	ldquo: ldquo$2,
  	ldquor: ldquor$2,
  	ldrdhar: ldrdhar$2,
  	ldrushar: ldrushar$2,
  	ldsh: ldsh$2,
  	le: le$2,
  	lE: lE$2,
  	LeftAngleBracket: LeftAngleBracket$2,
  	LeftArrowBar: LeftArrowBar$2,
  	leftarrow: leftarrow$2,
  	LeftArrow: LeftArrow$2,
  	Leftarrow: Leftarrow$2,
  	LeftArrowRightArrow: LeftArrowRightArrow$2,
  	leftarrowtail: leftarrowtail$2,
  	LeftCeiling: LeftCeiling$2,
  	LeftDoubleBracket: LeftDoubleBracket$2,
  	LeftDownTeeVector: LeftDownTeeVector$2,
  	LeftDownVectorBar: LeftDownVectorBar$2,
  	LeftDownVector: LeftDownVector$2,
  	LeftFloor: LeftFloor$2,
  	leftharpoondown: leftharpoondown$2,
  	leftharpoonup: leftharpoonup$2,
  	leftleftarrows: leftleftarrows$2,
  	leftrightarrow: leftrightarrow$2,
  	LeftRightArrow: LeftRightArrow$2,
  	Leftrightarrow: Leftrightarrow$2,
  	leftrightarrows: leftrightarrows$2,
  	leftrightharpoons: leftrightharpoons$2,
  	leftrightsquigarrow: leftrightsquigarrow$2,
  	LeftRightVector: LeftRightVector$2,
  	LeftTeeArrow: LeftTeeArrow$2,
  	LeftTee: LeftTee$2,
  	LeftTeeVector: LeftTeeVector$2,
  	leftthreetimes: leftthreetimes$2,
  	LeftTriangleBar: LeftTriangleBar$2,
  	LeftTriangle: LeftTriangle$2,
  	LeftTriangleEqual: LeftTriangleEqual$2,
  	LeftUpDownVector: LeftUpDownVector$2,
  	LeftUpTeeVector: LeftUpTeeVector$2,
  	LeftUpVectorBar: LeftUpVectorBar$2,
  	LeftUpVector: LeftUpVector$2,
  	LeftVectorBar: LeftVectorBar$2,
  	LeftVector: LeftVector$2,
  	lEg: lEg$2,
  	leg: leg$2,
  	leq: leq$2,
  	leqq: leqq$2,
  	leqslant: leqslant$2,
  	lescc: lescc$2,
  	les: les$2,
  	lesdot: lesdot$2,
  	lesdoto: lesdoto$2,
  	lesdotor: lesdotor$2,
  	lesg: lesg$2,
  	lesges: lesges$2,
  	lessapprox: lessapprox$2,
  	lessdot: lessdot$2,
  	lesseqgtr: lesseqgtr$2,
  	lesseqqgtr: lesseqqgtr$2,
  	LessEqualGreater: LessEqualGreater$2,
  	LessFullEqual: LessFullEqual$2,
  	LessGreater: LessGreater$2,
  	lessgtr: lessgtr$2,
  	LessLess: LessLess$2,
  	lesssim: lesssim$2,
  	LessSlantEqual: LessSlantEqual$2,
  	LessTilde: LessTilde$2,
  	lfisht: lfisht$2,
  	lfloor: lfloor$2,
  	Lfr: Lfr$2,
  	lfr: lfr$2,
  	lg: lg$2,
  	lgE: lgE$2,
  	lHar: lHar$2,
  	lhard: lhard$2,
  	lharu: lharu$2,
  	lharul: lharul$2,
  	lhblk: lhblk$2,
  	LJcy: LJcy$2,
  	ljcy: ljcy$2,
  	llarr: llarr$2,
  	ll: ll$2,
  	Ll: Ll$2,
  	llcorner: llcorner$2,
  	Lleftarrow: Lleftarrow$2,
  	llhard: llhard$2,
  	lltri: lltri$2,
  	Lmidot: Lmidot$2,
  	lmidot: lmidot$2,
  	lmoustache: lmoustache$2,
  	lmoust: lmoust$2,
  	lnap: lnap$2,
  	lnapprox: lnapprox$2,
  	lne: lne$2,
  	lnE: lnE$2,
  	lneq: lneq$2,
  	lneqq: lneqq$2,
  	lnsim: lnsim$2,
  	loang: loang$2,
  	loarr: loarr$2,
  	lobrk: lobrk$2,
  	longleftarrow: longleftarrow$2,
  	LongLeftArrow: LongLeftArrow$2,
  	Longleftarrow: Longleftarrow$2,
  	longleftrightarrow: longleftrightarrow$2,
  	LongLeftRightArrow: LongLeftRightArrow$2,
  	Longleftrightarrow: Longleftrightarrow$2,
  	longmapsto: longmapsto$2,
  	longrightarrow: longrightarrow$2,
  	LongRightArrow: LongRightArrow$2,
  	Longrightarrow: Longrightarrow$2,
  	looparrowleft: looparrowleft$2,
  	looparrowright: looparrowright$2,
  	lopar: lopar$2,
  	Lopf: Lopf$2,
  	lopf: lopf$2,
  	loplus: loplus$2,
  	lotimes: lotimes$2,
  	lowast: lowast$2,
  	lowbar: lowbar$2,
  	LowerLeftArrow: LowerLeftArrow$2,
  	LowerRightArrow: LowerRightArrow$2,
  	loz: loz$2,
  	lozenge: lozenge$2,
  	lozf: lozf$2,
  	lpar: lpar$2,
  	lparlt: lparlt$2,
  	lrarr: lrarr$2,
  	lrcorner: lrcorner$2,
  	lrhar: lrhar$2,
  	lrhard: lrhard$2,
  	lrm: lrm$2,
  	lrtri: lrtri$2,
  	lsaquo: lsaquo$2,
  	lscr: lscr$2,
  	Lscr: Lscr$2,
  	lsh: lsh$2,
  	Lsh: Lsh$2,
  	lsim: lsim$2,
  	lsime: lsime$2,
  	lsimg: lsimg$2,
  	lsqb: lsqb$2,
  	lsquo: lsquo$2,
  	lsquor: lsquor$2,
  	Lstrok: Lstrok$2,
  	lstrok: lstrok$2,
  	ltcc: ltcc$2,
  	ltcir: ltcir$2,
  	lt: lt$1$2,
  	LT: LT$3,
  	Lt: Lt$2,
  	ltdot: ltdot$2,
  	lthree: lthree$2,
  	ltimes: ltimes$2,
  	ltlarr: ltlarr$2,
  	ltquest: ltquest$2,
  	ltri: ltri$2,
  	ltrie: ltrie$2,
  	ltrif: ltrif$2,
  	ltrPar: ltrPar$2,
  	lurdshar: lurdshar$2,
  	luruhar: luruhar$2,
  	lvertneqq: lvertneqq$2,
  	lvnE: lvnE$2,
  	macr: macr$3,
  	male: male$2,
  	malt: malt$2,
  	maltese: maltese$2,
  	map: map$1$2,
  	mapsto: mapsto$2,
  	mapstodown: mapstodown$2,
  	mapstoleft: mapstoleft$2,
  	mapstoup: mapstoup$2,
  	marker: marker$2,
  	mcomma: mcomma$2,
  	Mcy: Mcy$2,
  	mcy: mcy$2,
  	mdash: mdash$2,
  	mDDot: mDDot$2,
  	measuredangle: measuredangle$2,
  	MediumSpace: MediumSpace$2,
  	Mellintrf: Mellintrf$2,
  	Mfr: Mfr$2,
  	mfr: mfr$2,
  	mho: mho$2,
  	micro: micro$3,
  	midast: midast$2,
  	midcir: midcir$2,
  	mid: mid$2,
  	middot: middot$3,
  	minusb: minusb$2,
  	minus: minus$2,
  	minusd: minusd$2,
  	minusdu: minusdu$2,
  	MinusPlus: MinusPlus$2,
  	mlcp: mlcp$2,
  	mldr: mldr$2,
  	mnplus: mnplus$2,
  	models: models$2,
  	Mopf: Mopf$2,
  	mopf: mopf$2,
  	mp: mp$2,
  	mscr: mscr$2,
  	Mscr: Mscr$2,
  	mstpos: mstpos$2,
  	Mu: Mu$2,
  	mu: mu$2,
  	multimap: multimap$2,
  	mumap: mumap$2,
  	nabla: nabla$2,
  	Nacute: Nacute$2,
  	nacute: nacute$2,
  	nang: nang$2,
  	nap: nap$2,
  	napE: napE$2,
  	napid: napid$2,
  	napos: napos$2,
  	napprox: napprox$2,
  	natural: natural$2,
  	naturals: naturals$2,
  	natur: natur$2,
  	nbsp: nbsp$3,
  	nbump: nbump$2,
  	nbumpe: nbumpe$2,
  	ncap: ncap$2,
  	Ncaron: Ncaron$2,
  	ncaron: ncaron$2,
  	Ncedil: Ncedil$2,
  	ncedil: ncedil$2,
  	ncong: ncong$2,
  	ncongdot: ncongdot$2,
  	ncup: ncup$2,
  	Ncy: Ncy$2,
  	ncy: ncy$2,
  	ndash: ndash$2,
  	nearhk: nearhk$2,
  	nearr: nearr$2,
  	neArr: neArr$2,
  	nearrow: nearrow$2,
  	ne: ne$2,
  	nedot: nedot$2,
  	NegativeMediumSpace: NegativeMediumSpace$2,
  	NegativeThickSpace: NegativeThickSpace$2,
  	NegativeThinSpace: NegativeThinSpace$2,
  	NegativeVeryThinSpace: NegativeVeryThinSpace$2,
  	nequiv: nequiv$2,
  	nesear: nesear$2,
  	nesim: nesim$2,
  	NestedGreaterGreater: NestedGreaterGreater$2,
  	NestedLessLess: NestedLessLess$2,
  	NewLine: NewLine$2,
  	nexist: nexist$2,
  	nexists: nexists$2,
  	Nfr: Nfr$2,
  	nfr: nfr$2,
  	ngE: ngE$2,
  	nge: nge$2,
  	ngeq: ngeq$2,
  	ngeqq: ngeqq$2,
  	ngeqslant: ngeqslant$2,
  	nges: nges$2,
  	nGg: nGg$2,
  	ngsim: ngsim$2,
  	nGt: nGt$2,
  	ngt: ngt$2,
  	ngtr: ngtr$2,
  	nGtv: nGtv$2,
  	nharr: nharr$2,
  	nhArr: nhArr$2,
  	nhpar: nhpar$2,
  	ni: ni$2,
  	nis: nis$2,
  	nisd: nisd$2,
  	niv: niv$2,
  	NJcy: NJcy$2,
  	njcy: njcy$2,
  	nlarr: nlarr$2,
  	nlArr: nlArr$2,
  	nldr: nldr$2,
  	nlE: nlE$2,
  	nle: nle$2,
  	nleftarrow: nleftarrow$2,
  	nLeftarrow: nLeftarrow$2,
  	nleftrightarrow: nleftrightarrow$2,
  	nLeftrightarrow: nLeftrightarrow$2,
  	nleq: nleq$2,
  	nleqq: nleqq$2,
  	nleqslant: nleqslant$2,
  	nles: nles$2,
  	nless: nless$2,
  	nLl: nLl$2,
  	nlsim: nlsim$2,
  	nLt: nLt$2,
  	nlt: nlt$2,
  	nltri: nltri$2,
  	nltrie: nltrie$2,
  	nLtv: nLtv$2,
  	nmid: nmid$2,
  	NoBreak: NoBreak$2,
  	NonBreakingSpace: NonBreakingSpace$2,
  	nopf: nopf$2,
  	Nopf: Nopf$2,
  	Not: Not$2,
  	not: not$3,
  	NotCongruent: NotCongruent$2,
  	NotCupCap: NotCupCap$2,
  	NotDoubleVerticalBar: NotDoubleVerticalBar$2,
  	NotElement: NotElement$2,
  	NotEqual: NotEqual$2,
  	NotEqualTilde: NotEqualTilde$2,
  	NotExists: NotExists$2,
  	NotGreater: NotGreater$2,
  	NotGreaterEqual: NotGreaterEqual$2,
  	NotGreaterFullEqual: NotGreaterFullEqual$2,
  	NotGreaterGreater: NotGreaterGreater$2,
  	NotGreaterLess: NotGreaterLess$2,
  	NotGreaterSlantEqual: NotGreaterSlantEqual$2,
  	NotGreaterTilde: NotGreaterTilde$2,
  	NotHumpDownHump: NotHumpDownHump$2,
  	NotHumpEqual: NotHumpEqual$2,
  	notin: notin$2,
  	notindot: notindot$2,
  	notinE: notinE$2,
  	notinva: notinva$2,
  	notinvb: notinvb$2,
  	notinvc: notinvc$2,
  	NotLeftTriangleBar: NotLeftTriangleBar$2,
  	NotLeftTriangle: NotLeftTriangle$2,
  	NotLeftTriangleEqual: NotLeftTriangleEqual$2,
  	NotLess: NotLess$2,
  	NotLessEqual: NotLessEqual$2,
  	NotLessGreater: NotLessGreater$2,
  	NotLessLess: NotLessLess$2,
  	NotLessSlantEqual: NotLessSlantEqual$2,
  	NotLessTilde: NotLessTilde$2,
  	NotNestedGreaterGreater: NotNestedGreaterGreater$2,
  	NotNestedLessLess: NotNestedLessLess$2,
  	notni: notni$2,
  	notniva: notniva$2,
  	notnivb: notnivb$2,
  	notnivc: notnivc$2,
  	NotPrecedes: NotPrecedes$2,
  	NotPrecedesEqual: NotPrecedesEqual$2,
  	NotPrecedesSlantEqual: NotPrecedesSlantEqual$2,
  	NotReverseElement: NotReverseElement$2,
  	NotRightTriangleBar: NotRightTriangleBar$2,
  	NotRightTriangle: NotRightTriangle$2,
  	NotRightTriangleEqual: NotRightTriangleEqual$2,
  	NotSquareSubset: NotSquareSubset$2,
  	NotSquareSubsetEqual: NotSquareSubsetEqual$2,
  	NotSquareSuperset: NotSquareSuperset$2,
  	NotSquareSupersetEqual: NotSquareSupersetEqual$2,
  	NotSubset: NotSubset$2,
  	NotSubsetEqual: NotSubsetEqual$2,
  	NotSucceeds: NotSucceeds$2,
  	NotSucceedsEqual: NotSucceedsEqual$2,
  	NotSucceedsSlantEqual: NotSucceedsSlantEqual$2,
  	NotSucceedsTilde: NotSucceedsTilde$2,
  	NotSuperset: NotSuperset$2,
  	NotSupersetEqual: NotSupersetEqual$2,
  	NotTilde: NotTilde$2,
  	NotTildeEqual: NotTildeEqual$2,
  	NotTildeFullEqual: NotTildeFullEqual$2,
  	NotTildeTilde: NotTildeTilde$2,
  	NotVerticalBar: NotVerticalBar$2,
  	nparallel: nparallel$2,
  	npar: npar$2,
  	nparsl: nparsl$2,
  	npart: npart$2,
  	npolint: npolint$2,
  	npr: npr$2,
  	nprcue: nprcue$2,
  	nprec: nprec$2,
  	npreceq: npreceq$2,
  	npre: npre$2,
  	nrarrc: nrarrc$2,
  	nrarr: nrarr$2,
  	nrArr: nrArr$2,
  	nrarrw: nrarrw$2,
  	nrightarrow: nrightarrow$2,
  	nRightarrow: nRightarrow$2,
  	nrtri: nrtri$2,
  	nrtrie: nrtrie$2,
  	nsc: nsc$2,
  	nsccue: nsccue$2,
  	nsce: nsce$2,
  	Nscr: Nscr$2,
  	nscr: nscr$2,
  	nshortmid: nshortmid$2,
  	nshortparallel: nshortparallel$2,
  	nsim: nsim$2,
  	nsime: nsime$2,
  	nsimeq: nsimeq$2,
  	nsmid: nsmid$2,
  	nspar: nspar$2,
  	nsqsube: nsqsube$2,
  	nsqsupe: nsqsupe$2,
  	nsub: nsub$2,
  	nsubE: nsubE$2,
  	nsube: nsube$2,
  	nsubset: nsubset$2,
  	nsubseteq: nsubseteq$2,
  	nsubseteqq: nsubseteqq$2,
  	nsucc: nsucc$2,
  	nsucceq: nsucceq$2,
  	nsup: nsup$2,
  	nsupE: nsupE$2,
  	nsupe: nsupe$2,
  	nsupset: nsupset$2,
  	nsupseteq: nsupseteq$2,
  	nsupseteqq: nsupseteqq$2,
  	ntgl: ntgl$2,
  	Ntilde: Ntilde$3,
  	ntilde: ntilde$3,
  	ntlg: ntlg$2,
  	ntriangleleft: ntriangleleft$2,
  	ntrianglelefteq: ntrianglelefteq$2,
  	ntriangleright: ntriangleright$2,
  	ntrianglerighteq: ntrianglerighteq$2,
  	Nu: Nu$2,
  	nu: nu$2,
  	num: num$2,
  	numero: numero$2,
  	numsp: numsp$2,
  	nvap: nvap$2,
  	nvdash: nvdash$2,
  	nvDash: nvDash$2,
  	nVdash: nVdash$2,
  	nVDash: nVDash$2,
  	nvge: nvge$2,
  	nvgt: nvgt$2,
  	nvHarr: nvHarr$2,
  	nvinfin: nvinfin$2,
  	nvlArr: nvlArr$2,
  	nvle: nvle$2,
  	nvlt: nvlt$2,
  	nvltrie: nvltrie$2,
  	nvrArr: nvrArr$2,
  	nvrtrie: nvrtrie$2,
  	nvsim: nvsim$2,
  	nwarhk: nwarhk$2,
  	nwarr: nwarr$2,
  	nwArr: nwArr$2,
  	nwarrow: nwarrow$2,
  	nwnear: nwnear$2,
  	Oacute: Oacute$3,
  	oacute: oacute$3,
  	oast: oast$2,
  	Ocirc: Ocirc$3,
  	ocirc: ocirc$3,
  	ocir: ocir$2,
  	Ocy: Ocy$2,
  	ocy: ocy$2,
  	odash: odash$2,
  	Odblac: Odblac$2,
  	odblac: odblac$2,
  	odiv: odiv$2,
  	odot: odot$2,
  	odsold: odsold$2,
  	OElig: OElig$2,
  	oelig: oelig$2,
  	ofcir: ofcir$2,
  	Ofr: Ofr$2,
  	ofr: ofr$2,
  	ogon: ogon$2,
  	Ograve: Ograve$3,
  	ograve: ograve$3,
  	ogt: ogt$2,
  	ohbar: ohbar$2,
  	ohm: ohm$2,
  	oint: oint$2,
  	olarr: olarr$2,
  	olcir: olcir$2,
  	olcross: olcross$2,
  	oline: oline$2,
  	olt: olt$2,
  	Omacr: Omacr$2,
  	omacr: omacr$2,
  	Omega: Omega$2,
  	omega: omega$2,
  	Omicron: Omicron$2,
  	omicron: omicron$2,
  	omid: omid$2,
  	ominus: ominus$2,
  	Oopf: Oopf$2,
  	oopf: oopf$2,
  	opar: opar$2,
  	OpenCurlyDoubleQuote: OpenCurlyDoubleQuote$2,
  	OpenCurlyQuote: OpenCurlyQuote$2,
  	operp: operp$2,
  	oplus: oplus$2,
  	orarr: orarr$2,
  	Or: Or$2,
  	or: or$2,
  	ord: ord$2,
  	order: order$2,
  	orderof: orderof$2,
  	ordf: ordf$3,
  	ordm: ordm$3,
  	origof: origof$2,
  	oror: oror$2,
  	orslope: orslope$2,
  	orv: orv$2,
  	oS: oS$2,
  	Oscr: Oscr$2,
  	oscr: oscr$2,
  	Oslash: Oslash$3,
  	oslash: oslash$3,
  	osol: osol$2,
  	Otilde: Otilde$3,
  	otilde: otilde$3,
  	otimesas: otimesas$2,
  	Otimes: Otimes$2,
  	otimes: otimes$2,
  	Ouml: Ouml$3,
  	ouml: ouml$3,
  	ovbar: ovbar$2,
  	OverBar: OverBar$2,
  	OverBrace: OverBrace$2,
  	OverBracket: OverBracket$2,
  	OverParenthesis: OverParenthesis$2,
  	para: para$3,
  	parallel: parallel$2,
  	par: par$2,
  	parsim: parsim$2,
  	parsl: parsl$2,
  	part: part$2,
  	PartialD: PartialD$2,
  	Pcy: Pcy$2,
  	pcy: pcy$2,
  	percnt: percnt$2,
  	period: period$2,
  	permil: permil$2,
  	perp: perp$2,
  	pertenk: pertenk$2,
  	Pfr: Pfr$2,
  	pfr: pfr$2,
  	Phi: Phi$2,
  	phi: phi$2,
  	phiv: phiv$2,
  	phmmat: phmmat$2,
  	phone: phone$2,
  	Pi: Pi$2,
  	pi: pi$2,
  	pitchfork: pitchfork$2,
  	piv: piv$2,
  	planck: planck$2,
  	planckh: planckh$2,
  	plankv: plankv$2,
  	plusacir: plusacir$2,
  	plusb: plusb$2,
  	pluscir: pluscir$2,
  	plus: plus$2,
  	plusdo: plusdo$2,
  	plusdu: plusdu$2,
  	pluse: pluse$2,
  	PlusMinus: PlusMinus$2,
  	plusmn: plusmn$3,
  	plussim: plussim$2,
  	plustwo: plustwo$2,
  	pm: pm$2,
  	Poincareplane: Poincareplane$2,
  	pointint: pointint$2,
  	popf: popf$2,
  	Popf: Popf$2,
  	pound: pound$3,
  	prap: prap$2,
  	Pr: Pr$2,
  	pr: pr$2,
  	prcue: prcue$2,
  	precapprox: precapprox$2,
  	prec: prec$2,
  	preccurlyeq: preccurlyeq$2,
  	Precedes: Precedes$2,
  	PrecedesEqual: PrecedesEqual$2,
  	PrecedesSlantEqual: PrecedesSlantEqual$2,
  	PrecedesTilde: PrecedesTilde$2,
  	preceq: preceq$2,
  	precnapprox: precnapprox$2,
  	precneqq: precneqq$2,
  	precnsim: precnsim$2,
  	pre: pre$2,
  	prE: prE$2,
  	precsim: precsim$2,
  	prime: prime$2,
  	Prime: Prime$2,
  	primes: primes$2,
  	prnap: prnap$2,
  	prnE: prnE$2,
  	prnsim: prnsim$2,
  	prod: prod$2,
  	Product: Product$2,
  	profalar: profalar$2,
  	profline: profline$2,
  	profsurf: profsurf$2,
  	prop: prop$2,
  	Proportional: Proportional$2,
  	Proportion: Proportion$2,
  	propto: propto$2,
  	prsim: prsim$2,
  	prurel: prurel$2,
  	Pscr: Pscr$2,
  	pscr: pscr$2,
  	Psi: Psi$2,
  	psi: psi$2,
  	puncsp: puncsp$2,
  	Qfr: Qfr$2,
  	qfr: qfr$2,
  	qint: qint$2,
  	qopf: qopf$2,
  	Qopf: Qopf$2,
  	qprime: qprime$2,
  	Qscr: Qscr$2,
  	qscr: qscr$2,
  	quaternions: quaternions$2,
  	quatint: quatint$2,
  	quest: quest$2,
  	questeq: questeq$2,
  	quot: quot$1$2,
  	QUOT: QUOT$3,
  	rAarr: rAarr$2,
  	race: race$2,
  	Racute: Racute$2,
  	racute: racute$2,
  	radic: radic$2,
  	raemptyv: raemptyv$2,
  	rang: rang$2,
  	Rang: Rang$2,
  	rangd: rangd$2,
  	range: range$2,
  	rangle: rangle$2,
  	raquo: raquo$3,
  	rarrap: rarrap$2,
  	rarrb: rarrb$2,
  	rarrbfs: rarrbfs$2,
  	rarrc: rarrc$2,
  	rarr: rarr$2,
  	Rarr: Rarr$2,
  	rArr: rArr$2,
  	rarrfs: rarrfs$2,
  	rarrhk: rarrhk$2,
  	rarrlp: rarrlp$2,
  	rarrpl: rarrpl$2,
  	rarrsim: rarrsim$2,
  	Rarrtl: Rarrtl$2,
  	rarrtl: rarrtl$2,
  	rarrw: rarrw$2,
  	ratail: ratail$2,
  	rAtail: rAtail$2,
  	ratio: ratio$2,
  	rationals: rationals$2,
  	rbarr: rbarr$2,
  	rBarr: rBarr$2,
  	RBarr: RBarr$2,
  	rbbrk: rbbrk$2,
  	rbrace: rbrace$2,
  	rbrack: rbrack$2,
  	rbrke: rbrke$2,
  	rbrksld: rbrksld$2,
  	rbrkslu: rbrkslu$2,
  	Rcaron: Rcaron$2,
  	rcaron: rcaron$2,
  	Rcedil: Rcedil$2,
  	rcedil: rcedil$2,
  	rceil: rceil$2,
  	rcub: rcub$2,
  	Rcy: Rcy$2,
  	rcy: rcy$2,
  	rdca: rdca$2,
  	rdldhar: rdldhar$2,
  	rdquo: rdquo$2,
  	rdquor: rdquor$2,
  	rdsh: rdsh$2,
  	real: real$2,
  	realine: realine$2,
  	realpart: realpart$2,
  	reals: reals$2,
  	Re: Re$2,
  	rect: rect$2,
  	reg: reg$3,
  	REG: REG$3,
  	ReverseElement: ReverseElement$2,
  	ReverseEquilibrium: ReverseEquilibrium$2,
  	ReverseUpEquilibrium: ReverseUpEquilibrium$2,
  	rfisht: rfisht$2,
  	rfloor: rfloor$2,
  	rfr: rfr$2,
  	Rfr: Rfr$2,
  	rHar: rHar$2,
  	rhard: rhard$2,
  	rharu: rharu$2,
  	rharul: rharul$2,
  	Rho: Rho$2,
  	rho: rho$2,
  	rhov: rhov$2,
  	RightAngleBracket: RightAngleBracket$2,
  	RightArrowBar: RightArrowBar$2,
  	rightarrow: rightarrow$2,
  	RightArrow: RightArrow$2,
  	Rightarrow: Rightarrow$2,
  	RightArrowLeftArrow: RightArrowLeftArrow$2,
  	rightarrowtail: rightarrowtail$2,
  	RightCeiling: RightCeiling$2,
  	RightDoubleBracket: RightDoubleBracket$2,
  	RightDownTeeVector: RightDownTeeVector$2,
  	RightDownVectorBar: RightDownVectorBar$2,
  	RightDownVector: RightDownVector$2,
  	RightFloor: RightFloor$2,
  	rightharpoondown: rightharpoondown$2,
  	rightharpoonup: rightharpoonup$2,
  	rightleftarrows: rightleftarrows$2,
  	rightleftharpoons: rightleftharpoons$2,
  	rightrightarrows: rightrightarrows$2,
  	rightsquigarrow: rightsquigarrow$2,
  	RightTeeArrow: RightTeeArrow$2,
  	RightTee: RightTee$2,
  	RightTeeVector: RightTeeVector$2,
  	rightthreetimes: rightthreetimes$2,
  	RightTriangleBar: RightTriangleBar$2,
  	RightTriangle: RightTriangle$2,
  	RightTriangleEqual: RightTriangleEqual$2,
  	RightUpDownVector: RightUpDownVector$2,
  	RightUpTeeVector: RightUpTeeVector$2,
  	RightUpVectorBar: RightUpVectorBar$2,
  	RightUpVector: RightUpVector$2,
  	RightVectorBar: RightVectorBar$2,
  	RightVector: RightVector$2,
  	ring: ring$2,
  	risingdotseq: risingdotseq$2,
  	rlarr: rlarr$2,
  	rlhar: rlhar$2,
  	rlm: rlm$2,
  	rmoustache: rmoustache$2,
  	rmoust: rmoust$2,
  	rnmid: rnmid$2,
  	roang: roang$2,
  	roarr: roarr$2,
  	robrk: robrk$2,
  	ropar: ropar$2,
  	ropf: ropf$2,
  	Ropf: Ropf$2,
  	roplus: roplus$2,
  	rotimes: rotimes$2,
  	RoundImplies: RoundImplies$2,
  	rpar: rpar$2,
  	rpargt: rpargt$2,
  	rppolint: rppolint$2,
  	rrarr: rrarr$2,
  	Rrightarrow: Rrightarrow$2,
  	rsaquo: rsaquo$2,
  	rscr: rscr$2,
  	Rscr: Rscr$2,
  	rsh: rsh$2,
  	Rsh: Rsh$2,
  	rsqb: rsqb$2,
  	rsquo: rsquo$2,
  	rsquor: rsquor$2,
  	rthree: rthree$2,
  	rtimes: rtimes$2,
  	rtri: rtri$2,
  	rtrie: rtrie$2,
  	rtrif: rtrif$2,
  	rtriltri: rtriltri$2,
  	RuleDelayed: RuleDelayed$2,
  	ruluhar: ruluhar$2,
  	rx: rx$2,
  	Sacute: Sacute$2,
  	sacute: sacute$2,
  	sbquo: sbquo$2,
  	scap: scap$2,
  	Scaron: Scaron$2,
  	scaron: scaron$2,
  	Sc: Sc$2,
  	sc: sc$2,
  	sccue: sccue$2,
  	sce: sce$2,
  	scE: scE$2,
  	Scedil: Scedil$2,
  	scedil: scedil$2,
  	Scirc: Scirc$2,
  	scirc: scirc$2,
  	scnap: scnap$2,
  	scnE: scnE$2,
  	scnsim: scnsim$2,
  	scpolint: scpolint$2,
  	scsim: scsim$2,
  	Scy: Scy$2,
  	scy: scy$2,
  	sdotb: sdotb$2,
  	sdot: sdot$2,
  	sdote: sdote$2,
  	searhk: searhk$2,
  	searr: searr$2,
  	seArr: seArr$2,
  	searrow: searrow$2,
  	sect: sect$3,
  	semi: semi$2,
  	seswar: seswar$2,
  	setminus: setminus$2,
  	setmn: setmn$2,
  	sext: sext$2,
  	Sfr: Sfr$2,
  	sfr: sfr$2,
  	sfrown: sfrown$2,
  	sharp: sharp$2,
  	SHCHcy: SHCHcy$2,
  	shchcy: shchcy$2,
  	SHcy: SHcy$2,
  	shcy: shcy$2,
  	ShortDownArrow: ShortDownArrow$2,
  	ShortLeftArrow: ShortLeftArrow$2,
  	shortmid: shortmid$2,
  	shortparallel: shortparallel$2,
  	ShortRightArrow: ShortRightArrow$2,
  	ShortUpArrow: ShortUpArrow$2,
  	shy: shy$3,
  	Sigma: Sigma$2,
  	sigma: sigma$2,
  	sigmaf: sigmaf$2,
  	sigmav: sigmav$2,
  	sim: sim$2,
  	simdot: simdot$2,
  	sime: sime$2,
  	simeq: simeq$2,
  	simg: simg$2,
  	simgE: simgE$2,
  	siml: siml$2,
  	simlE: simlE$2,
  	simne: simne$2,
  	simplus: simplus$2,
  	simrarr: simrarr$2,
  	slarr: slarr$2,
  	SmallCircle: SmallCircle$2,
  	smallsetminus: smallsetminus$2,
  	smashp: smashp$2,
  	smeparsl: smeparsl$2,
  	smid: smid$2,
  	smile: smile$2,
  	smt: smt$2,
  	smte: smte$2,
  	smtes: smtes$2,
  	SOFTcy: SOFTcy$2,
  	softcy: softcy$2,
  	solbar: solbar$2,
  	solb: solb$2,
  	sol: sol$2,
  	Sopf: Sopf$2,
  	sopf: sopf$2,
  	spades: spades$2,
  	spadesuit: spadesuit$2,
  	spar: spar$2,
  	sqcap: sqcap$2,
  	sqcaps: sqcaps$2,
  	sqcup: sqcup$2,
  	sqcups: sqcups$2,
  	Sqrt: Sqrt$2,
  	sqsub: sqsub$2,
  	sqsube: sqsube$2,
  	sqsubset: sqsubset$2,
  	sqsubseteq: sqsubseteq$2,
  	sqsup: sqsup$2,
  	sqsupe: sqsupe$2,
  	sqsupset: sqsupset$2,
  	sqsupseteq: sqsupseteq$2,
  	square: square$2,
  	Square: Square$2,
  	SquareIntersection: SquareIntersection$2,
  	SquareSubset: SquareSubset$2,
  	SquareSubsetEqual: SquareSubsetEqual$2,
  	SquareSuperset: SquareSuperset$2,
  	SquareSupersetEqual: SquareSupersetEqual$2,
  	SquareUnion: SquareUnion$2,
  	squarf: squarf$2,
  	squ: squ$2,
  	squf: squf$2,
  	srarr: srarr$2,
  	Sscr: Sscr$2,
  	sscr: sscr$2,
  	ssetmn: ssetmn$2,
  	ssmile: ssmile$2,
  	sstarf: sstarf$2,
  	Star: Star$2,
  	star: star$2,
  	starf: starf$2,
  	straightepsilon: straightepsilon$2,
  	straightphi: straightphi$2,
  	strns: strns$2,
  	sub: sub$2,
  	Sub: Sub$2,
  	subdot: subdot$2,
  	subE: subE$2,
  	sube: sube$2,
  	subedot: subedot$2,
  	submult: submult$2,
  	subnE: subnE$2,
  	subne: subne$2,
  	subplus: subplus$2,
  	subrarr: subrarr$2,
  	subset: subset$2,
  	Subset: Subset$2,
  	subseteq: subseteq$2,
  	subseteqq: subseteqq$2,
  	SubsetEqual: SubsetEqual$2,
  	subsetneq: subsetneq$2,
  	subsetneqq: subsetneqq$2,
  	subsim: subsim$2,
  	subsub: subsub$2,
  	subsup: subsup$2,
  	succapprox: succapprox$2,
  	succ: succ$2,
  	succcurlyeq: succcurlyeq$2,
  	Succeeds: Succeeds$2,
  	SucceedsEqual: SucceedsEqual$2,
  	SucceedsSlantEqual: SucceedsSlantEqual$2,
  	SucceedsTilde: SucceedsTilde$2,
  	succeq: succeq$2,
  	succnapprox: succnapprox$2,
  	succneqq: succneqq$2,
  	succnsim: succnsim$2,
  	succsim: succsim$2,
  	SuchThat: SuchThat$2,
  	sum: sum$2,
  	Sum: Sum$2,
  	sung: sung$2,
  	sup1: sup1$3,
  	sup2: sup2$3,
  	sup3: sup3$3,
  	sup: sup$2,
  	Sup: Sup$2,
  	supdot: supdot$2,
  	supdsub: supdsub$2,
  	supE: supE$2,
  	supe: supe$2,
  	supedot: supedot$2,
  	Superset: Superset$2,
  	SupersetEqual: SupersetEqual$2,
  	suphsol: suphsol$2,
  	suphsub: suphsub$2,
  	suplarr: suplarr$2,
  	supmult: supmult$2,
  	supnE: supnE$2,
  	supne: supne$2,
  	supplus: supplus$2,
  	supset: supset$2,
  	Supset: Supset$2,
  	supseteq: supseteq$2,
  	supseteqq: supseteqq$2,
  	supsetneq: supsetneq$2,
  	supsetneqq: supsetneqq$2,
  	supsim: supsim$2,
  	supsub: supsub$2,
  	supsup: supsup$2,
  	swarhk: swarhk$2,
  	swarr: swarr$2,
  	swArr: swArr$2,
  	swarrow: swarrow$2,
  	swnwar: swnwar$2,
  	szlig: szlig$3,
  	Tab: Tab$2,
  	target: target$2,
  	Tau: Tau$2,
  	tau: tau$2,
  	tbrk: tbrk$2,
  	Tcaron: Tcaron$2,
  	tcaron: tcaron$2,
  	Tcedil: Tcedil$2,
  	tcedil: tcedil$2,
  	Tcy: Tcy$2,
  	tcy: tcy$2,
  	tdot: tdot$2,
  	telrec: telrec$2,
  	Tfr: Tfr$2,
  	tfr: tfr$2,
  	there4: there4$2,
  	therefore: therefore$2,
  	Therefore: Therefore$2,
  	Theta: Theta$2,
  	theta: theta$2,
  	thetasym: thetasym$2,
  	thetav: thetav$2,
  	thickapprox: thickapprox$2,
  	thicksim: thicksim$2,
  	ThickSpace: ThickSpace$2,
  	ThinSpace: ThinSpace$2,
  	thinsp: thinsp$2,
  	thkap: thkap$2,
  	thksim: thksim$2,
  	THORN: THORN$3,
  	thorn: thorn$3,
  	tilde: tilde$2,
  	Tilde: Tilde$2,
  	TildeEqual: TildeEqual$2,
  	TildeFullEqual: TildeFullEqual$2,
  	TildeTilde: TildeTilde$2,
  	timesbar: timesbar$2,
  	timesb: timesb$2,
  	times: times$4,
  	timesd: timesd$2,
  	tint: tint$2,
  	toea: toea$2,
  	topbot: topbot$2,
  	topcir: topcir$2,
  	top: top$2,
  	Topf: Topf$2,
  	topf: topf$2,
  	topfork: topfork$2,
  	tosa: tosa$2,
  	tprime: tprime$2,
  	trade: trade$2,
  	TRADE: TRADE$2,
  	triangle: triangle$2,
  	triangledown: triangledown$2,
  	triangleleft: triangleleft$2,
  	trianglelefteq: trianglelefteq$2,
  	triangleq: triangleq$2,
  	triangleright: triangleright$2,
  	trianglerighteq: trianglerighteq$2,
  	tridot: tridot$2,
  	trie: trie$2,
  	triminus: triminus$2,
  	TripleDot: TripleDot$2,
  	triplus: triplus$2,
  	trisb: trisb$2,
  	tritime: tritime$2,
  	trpezium: trpezium$2,
  	Tscr: Tscr$2,
  	tscr: tscr$2,
  	TScy: TScy$2,
  	tscy: tscy$2,
  	TSHcy: TSHcy$2,
  	tshcy: tshcy$2,
  	Tstrok: Tstrok$2,
  	tstrok: tstrok$2,
  	twixt: twixt$2,
  	twoheadleftarrow: twoheadleftarrow$2,
  	twoheadrightarrow: twoheadrightarrow$2,
  	Uacute: Uacute$3,
  	uacute: uacute$3,
  	uarr: uarr$2,
  	Uarr: Uarr$2,
  	uArr: uArr$2,
  	Uarrocir: Uarrocir$2,
  	Ubrcy: Ubrcy$2,
  	ubrcy: ubrcy$2,
  	Ubreve: Ubreve$2,
  	ubreve: ubreve$2,
  	Ucirc: Ucirc$3,
  	ucirc: ucirc$3,
  	Ucy: Ucy$2,
  	ucy: ucy$2,
  	udarr: udarr$2,
  	Udblac: Udblac$2,
  	udblac: udblac$2,
  	udhar: udhar$2,
  	ufisht: ufisht$2,
  	Ufr: Ufr$2,
  	ufr: ufr$2,
  	Ugrave: Ugrave$3,
  	ugrave: ugrave$3,
  	uHar: uHar$2,
  	uharl: uharl$2,
  	uharr: uharr$2,
  	uhblk: uhblk$2,
  	ulcorn: ulcorn$2,
  	ulcorner: ulcorner$2,
  	ulcrop: ulcrop$2,
  	ultri: ultri$2,
  	Umacr: Umacr$2,
  	umacr: umacr$2,
  	uml: uml$3,
  	UnderBar: UnderBar$2,
  	UnderBrace: UnderBrace$2,
  	UnderBracket: UnderBracket$2,
  	UnderParenthesis: UnderParenthesis$2,
  	Union: Union$2,
  	UnionPlus: UnionPlus$2,
  	Uogon: Uogon$2,
  	uogon: uogon$2,
  	Uopf: Uopf$2,
  	uopf: uopf$2,
  	UpArrowBar: UpArrowBar$2,
  	uparrow: uparrow$2,
  	UpArrow: UpArrow$2,
  	Uparrow: Uparrow$2,
  	UpArrowDownArrow: UpArrowDownArrow$2,
  	updownarrow: updownarrow$2,
  	UpDownArrow: UpDownArrow$2,
  	Updownarrow: Updownarrow$2,
  	UpEquilibrium: UpEquilibrium$2,
  	upharpoonleft: upharpoonleft$2,
  	upharpoonright: upharpoonright$2,
  	uplus: uplus$2,
  	UpperLeftArrow: UpperLeftArrow$2,
  	UpperRightArrow: UpperRightArrow$2,
  	upsi: upsi$2,
  	Upsi: Upsi$2,
  	upsih: upsih$2,
  	Upsilon: Upsilon$2,
  	upsilon: upsilon$2,
  	UpTeeArrow: UpTeeArrow$2,
  	UpTee: UpTee$2,
  	upuparrows: upuparrows$2,
  	urcorn: urcorn$2,
  	urcorner: urcorner$2,
  	urcrop: urcrop$2,
  	Uring: Uring$2,
  	uring: uring$2,
  	urtri: urtri$2,
  	Uscr: Uscr$2,
  	uscr: uscr$2,
  	utdot: utdot$2,
  	Utilde: Utilde$2,
  	utilde: utilde$2,
  	utri: utri$2,
  	utrif: utrif$2,
  	uuarr: uuarr$2,
  	Uuml: Uuml$3,
  	uuml: uuml$3,
  	uwangle: uwangle$2,
  	vangrt: vangrt$2,
  	varepsilon: varepsilon$2,
  	varkappa: varkappa$2,
  	varnothing: varnothing$2,
  	varphi: varphi$2,
  	varpi: varpi$2,
  	varpropto: varpropto$2,
  	varr: varr$2,
  	vArr: vArr$2,
  	varrho: varrho$2,
  	varsigma: varsigma$2,
  	varsubsetneq: varsubsetneq$2,
  	varsubsetneqq: varsubsetneqq$2,
  	varsupsetneq: varsupsetneq$2,
  	varsupsetneqq: varsupsetneqq$2,
  	vartheta: vartheta$2,
  	vartriangleleft: vartriangleleft$2,
  	vartriangleright: vartriangleright$2,
  	vBar: vBar$2,
  	Vbar: Vbar$2,
  	vBarv: vBarv$2,
  	Vcy: Vcy$2,
  	vcy: vcy$2,
  	vdash: vdash$2,
  	vDash: vDash$2,
  	Vdash: Vdash$2,
  	VDash: VDash$2,
  	Vdashl: Vdashl$2,
  	veebar: veebar$2,
  	vee: vee$2,
  	Vee: Vee$2,
  	veeeq: veeeq$2,
  	vellip: vellip$2,
  	verbar: verbar$2,
  	Verbar: Verbar$2,
  	vert: vert$2,
  	Vert: Vert$2,
  	VerticalBar: VerticalBar$2,
  	VerticalLine: VerticalLine$2,
  	VerticalSeparator: VerticalSeparator$2,
  	VerticalTilde: VerticalTilde$2,
  	VeryThinSpace: VeryThinSpace$2,
  	Vfr: Vfr$2,
  	vfr: vfr$2,
  	vltri: vltri$2,
  	vnsub: vnsub$2,
  	vnsup: vnsup$2,
  	Vopf: Vopf$2,
  	vopf: vopf$2,
  	vprop: vprop$2,
  	vrtri: vrtri$2,
  	Vscr: Vscr$2,
  	vscr: vscr$2,
  	vsubnE: vsubnE$2,
  	vsubne: vsubne$2,
  	vsupnE: vsupnE$2,
  	vsupne: vsupne$2,
  	Vvdash: Vvdash$2,
  	vzigzag: vzigzag$2,
  	Wcirc: Wcirc$2,
  	wcirc: wcirc$2,
  	wedbar: wedbar$2,
  	wedge: wedge$2,
  	Wedge: Wedge$2,
  	wedgeq: wedgeq$2,
  	weierp: weierp$2,
  	Wfr: Wfr$2,
  	wfr: wfr$2,
  	Wopf: Wopf$2,
  	wopf: wopf$2,
  	wp: wp$2,
  	wr: wr$2,
  	wreath: wreath$2,
  	Wscr: Wscr$2,
  	wscr: wscr$2,
  	xcap: xcap$2,
  	xcirc: xcirc$2,
  	xcup: xcup$2,
  	xdtri: xdtri$2,
  	Xfr: Xfr$2,
  	xfr: xfr$2,
  	xharr: xharr$2,
  	xhArr: xhArr$2,
  	Xi: Xi$2,
  	xi: xi$2,
  	xlarr: xlarr$2,
  	xlArr: xlArr$2,
  	xmap: xmap$2,
  	xnis: xnis$2,
  	xodot: xodot$2,
  	Xopf: Xopf$2,
  	xopf: xopf$2,
  	xoplus: xoplus$2,
  	xotime: xotime$2,
  	xrarr: xrarr$2,
  	xrArr: xrArr$2,
  	Xscr: Xscr$2,
  	xscr: xscr$2,
  	xsqcup: xsqcup$2,
  	xuplus: xuplus$2,
  	xutri: xutri$2,
  	xvee: xvee$2,
  	xwedge: xwedge$2,
  	Yacute: Yacute$3,
  	yacute: yacute$3,
  	YAcy: YAcy$2,
  	yacy: yacy$2,
  	Ycirc: Ycirc$2,
  	ycirc: ycirc$2,
  	Ycy: Ycy$2,
  	ycy: ycy$2,
  	yen: yen$3,
  	Yfr: Yfr$2,
  	yfr: yfr$2,
  	YIcy: YIcy$2,
  	yicy: yicy$2,
  	Yopf: Yopf$2,
  	yopf: yopf$2,
  	Yscr: Yscr$2,
  	yscr: yscr$2,
  	YUcy: YUcy$2,
  	yucy: yucy$2,
  	yuml: yuml$3,
  	Yuml: Yuml$2,
  	Zacute: Zacute$2,
  	zacute: zacute$2,
  	Zcaron: Zcaron$2,
  	zcaron: zcaron$2,
  	Zcy: Zcy$2,
  	zcy: zcy$2,
  	Zdot: Zdot$2,
  	zdot: zdot$2,
  	zeetrf: zeetrf$2,
  	ZeroWidthSpace: ZeroWidthSpace$2,
  	Zeta: Zeta$2,
  	zeta: zeta$2,
  	zfr: zfr$2,
  	Zfr: Zfr$2,
  	ZHcy: ZHcy$2,
  	zhcy: zhcy$2,
  	zigrarr: zigrarr$2,
  	zopf: zopf$2,
  	Zopf: Zopf$2,
  	Zscr: Zscr$2,
  	zscr: zscr$2,
  	zwj: zwj$2,
  	zwnj: zwnj$2,
  	"in": "∈",
  	"Map": "⤅"
  };

  var entities$2 = Object.freeze({
  	Aacute: Aacute$3,
  	aacute: aacute$3,
  	Abreve: Abreve$2,
  	abreve: abreve$2,
  	ac: ac$2,
  	acd: acd$2,
  	acE: acE$2,
  	Acirc: Acirc$3,
  	acirc: acirc$3,
  	acute: acute$3,
  	Acy: Acy$2,
  	acy: acy$2,
  	AElig: AElig$3,
  	aelig: aelig$3,
  	af: af$2,
  	Afr: Afr$2,
  	afr: afr$2,
  	Agrave: Agrave$3,
  	agrave: agrave$3,
  	alefsym: alefsym$2,
  	aleph: aleph$2,
  	Alpha: Alpha$2,
  	alpha: alpha$2,
  	Amacr: Amacr$2,
  	amacr: amacr$2,
  	amalg: amalg$2,
  	amp: amp$1$2,
  	AMP: AMP$3,
  	andand: andand$2,
  	And: And$2,
  	and: and$2,
  	andd: andd$2,
  	andslope: andslope$2,
  	andv: andv$2,
  	ang: ang$2,
  	ange: ange$2,
  	angle: angle$2,
  	angmsdaa: angmsdaa$2,
  	angmsdab: angmsdab$2,
  	angmsdac: angmsdac$2,
  	angmsdad: angmsdad$2,
  	angmsdae: angmsdae$2,
  	angmsdaf: angmsdaf$2,
  	angmsdag: angmsdag$2,
  	angmsdah: angmsdah$2,
  	angmsd: angmsd$2,
  	angrt: angrt$2,
  	angrtvb: angrtvb$2,
  	angrtvbd: angrtvbd$2,
  	angsph: angsph$2,
  	angst: angst$2,
  	angzarr: angzarr$2,
  	Aogon: Aogon$2,
  	aogon: aogon$2,
  	Aopf: Aopf$2,
  	aopf: aopf$2,
  	apacir: apacir$2,
  	ap: ap$2,
  	apE: apE$2,
  	ape: ape$2,
  	apid: apid$2,
  	apos: apos$1$2,
  	ApplyFunction: ApplyFunction$2,
  	approx: approx$2,
  	approxeq: approxeq$2,
  	Aring: Aring$3,
  	aring: aring$3,
  	Ascr: Ascr$2,
  	ascr: ascr$2,
  	Assign: Assign$2,
  	ast: ast$2,
  	asymp: asymp$2,
  	asympeq: asympeq$2,
  	Atilde: Atilde$3,
  	atilde: atilde$3,
  	Auml: Auml$3,
  	auml: auml$3,
  	awconint: awconint$2,
  	awint: awint$2,
  	backcong: backcong$2,
  	backepsilon: backepsilon$2,
  	backprime: backprime$2,
  	backsim: backsim$2,
  	backsimeq: backsimeq$2,
  	Backslash: Backslash$2,
  	Barv: Barv$2,
  	barvee: barvee$2,
  	barwed: barwed$2,
  	Barwed: Barwed$2,
  	barwedge: barwedge$2,
  	bbrk: bbrk$2,
  	bbrktbrk: bbrktbrk$2,
  	bcong: bcong$2,
  	Bcy: Bcy$2,
  	bcy: bcy$2,
  	bdquo: bdquo$2,
  	becaus: becaus$2,
  	because: because$2,
  	Because: Because$2,
  	bemptyv: bemptyv$2,
  	bepsi: bepsi$2,
  	bernou: bernou$2,
  	Bernoullis: Bernoullis$2,
  	Beta: Beta$2,
  	beta: beta$2,
  	beth: beth$2,
  	between: between$2,
  	Bfr: Bfr$2,
  	bfr: bfr$2,
  	bigcap: bigcap$2,
  	bigcirc: bigcirc$2,
  	bigcup: bigcup$2,
  	bigodot: bigodot$2,
  	bigoplus: bigoplus$2,
  	bigotimes: bigotimes$2,
  	bigsqcup: bigsqcup$2,
  	bigstar: bigstar$2,
  	bigtriangledown: bigtriangledown$2,
  	bigtriangleup: bigtriangleup$2,
  	biguplus: biguplus$2,
  	bigvee: bigvee$2,
  	bigwedge: bigwedge$2,
  	bkarow: bkarow$2,
  	blacklozenge: blacklozenge$2,
  	blacksquare: blacksquare$2,
  	blacktriangle: blacktriangle$2,
  	blacktriangledown: blacktriangledown$2,
  	blacktriangleleft: blacktriangleleft$2,
  	blacktriangleright: blacktriangleright$2,
  	blank: blank$2,
  	blk12: blk12$2,
  	blk14: blk14$2,
  	blk34: blk34$2,
  	block: block$2,
  	bne: bne$2,
  	bnequiv: bnequiv$2,
  	bNot: bNot$2,
  	bnot: bnot$2,
  	Bopf: Bopf$2,
  	bopf: bopf$2,
  	bot: bot$2,
  	bottom: bottom$2,
  	bowtie: bowtie$2,
  	boxbox: boxbox$2,
  	boxdl: boxdl$2,
  	boxdL: boxdL$2,
  	boxDl: boxDl$2,
  	boxDL: boxDL$2,
  	boxdr: boxdr$2,
  	boxdR: boxdR$2,
  	boxDr: boxDr$2,
  	boxDR: boxDR$2,
  	boxh: boxh$2,
  	boxH: boxH$2,
  	boxhd: boxhd$2,
  	boxHd: boxHd$2,
  	boxhD: boxhD$2,
  	boxHD: boxHD$2,
  	boxhu: boxhu$2,
  	boxHu: boxHu$2,
  	boxhU: boxhU$2,
  	boxHU: boxHU$2,
  	boxminus: boxminus$2,
  	boxplus: boxplus$2,
  	boxtimes: boxtimes$2,
  	boxul: boxul$2,
  	boxuL: boxuL$2,
  	boxUl: boxUl$2,
  	boxUL: boxUL$2,
  	boxur: boxur$2,
  	boxuR: boxuR$2,
  	boxUr: boxUr$2,
  	boxUR: boxUR$2,
  	boxv: boxv$2,
  	boxV: boxV$2,
  	boxvh: boxvh$2,
  	boxvH: boxvH$2,
  	boxVh: boxVh$2,
  	boxVH: boxVH$2,
  	boxvl: boxvl$2,
  	boxvL: boxvL$2,
  	boxVl: boxVl$2,
  	boxVL: boxVL$2,
  	boxvr: boxvr$2,
  	boxvR: boxvR$2,
  	boxVr: boxVr$2,
  	boxVR: boxVR$2,
  	bprime: bprime$2,
  	breve: breve$2,
  	Breve: Breve$2,
  	brvbar: brvbar$3,
  	bscr: bscr$2,
  	Bscr: Bscr$2,
  	bsemi: bsemi$2,
  	bsim: bsim$2,
  	bsime: bsime$2,
  	bsolb: bsolb$2,
  	bsol: bsol$2,
  	bsolhsub: bsolhsub$2,
  	bull: bull$2,
  	bullet: bullet$2,
  	bump: bump$2,
  	bumpE: bumpE$2,
  	bumpe: bumpe$2,
  	Bumpeq: Bumpeq$2,
  	bumpeq: bumpeq$2,
  	Cacute: Cacute$2,
  	cacute: cacute$2,
  	capand: capand$2,
  	capbrcup: capbrcup$2,
  	capcap: capcap$2,
  	cap: cap$2,
  	Cap: Cap$2,
  	capcup: capcup$2,
  	capdot: capdot$2,
  	CapitalDifferentialD: CapitalDifferentialD$2,
  	caps: caps$2,
  	caret: caret$2,
  	caron: caron$2,
  	Cayleys: Cayleys$2,
  	ccaps: ccaps$2,
  	Ccaron: Ccaron$2,
  	ccaron: ccaron$2,
  	Ccedil: Ccedil$3,
  	ccedil: ccedil$3,
  	Ccirc: Ccirc$2,
  	ccirc: ccirc$2,
  	Cconint: Cconint$2,
  	ccups: ccups$2,
  	ccupssm: ccupssm$2,
  	Cdot: Cdot$2,
  	cdot: cdot$2,
  	cedil: cedil$3,
  	Cedilla: Cedilla$2,
  	cemptyv: cemptyv$2,
  	cent: cent$3,
  	centerdot: centerdot$2,
  	CenterDot: CenterDot$2,
  	cfr: cfr$2,
  	Cfr: Cfr$2,
  	CHcy: CHcy$2,
  	chcy: chcy$2,
  	check: check$2,
  	checkmark: checkmark$2,
  	Chi: Chi$2,
  	chi: chi$2,
  	circ: circ$2,
  	circeq: circeq$2,
  	circlearrowleft: circlearrowleft$2,
  	circlearrowright: circlearrowright$2,
  	circledast: circledast$2,
  	circledcirc: circledcirc$2,
  	circleddash: circleddash$2,
  	CircleDot: CircleDot$2,
  	circledR: circledR$2,
  	circledS: circledS$2,
  	CircleMinus: CircleMinus$2,
  	CirclePlus: CirclePlus$2,
  	CircleTimes: CircleTimes$2,
  	cir: cir$2,
  	cirE: cirE$2,
  	cire: cire$2,
  	cirfnint: cirfnint$2,
  	cirmid: cirmid$2,
  	cirscir: cirscir$2,
  	ClockwiseContourIntegral: ClockwiseContourIntegral$2,
  	CloseCurlyDoubleQuote: CloseCurlyDoubleQuote$2,
  	CloseCurlyQuote: CloseCurlyQuote$2,
  	clubs: clubs$2,
  	clubsuit: clubsuit$2,
  	colon: colon$2,
  	Colon: Colon$2,
  	Colone: Colone$2,
  	colone: colone$2,
  	coloneq: coloneq$2,
  	comma: comma$2,
  	commat: commat$2,
  	comp: comp$2,
  	compfn: compfn$2,
  	complement: complement$2,
  	complexes: complexes$2,
  	cong: cong$2,
  	congdot: congdot$2,
  	Congruent: Congruent$2,
  	conint: conint$2,
  	Conint: Conint$2,
  	ContourIntegral: ContourIntegral$2,
  	copf: copf$2,
  	Copf: Copf$2,
  	coprod: coprod$2,
  	Coproduct: Coproduct$2,
  	copy: copy$3,
  	COPY: COPY$3,
  	copysr: copysr$2,
  	CounterClockwiseContourIntegral: CounterClockwiseContourIntegral$2,
  	crarr: crarr$2,
  	cross: cross$2,
  	Cross: Cross$2,
  	Cscr: Cscr$2,
  	cscr: cscr$2,
  	csub: csub$2,
  	csube: csube$2,
  	csup: csup$2,
  	csupe: csupe$2,
  	ctdot: ctdot$2,
  	cudarrl: cudarrl$2,
  	cudarrr: cudarrr$2,
  	cuepr: cuepr$2,
  	cuesc: cuesc$2,
  	cularr: cularr$2,
  	cularrp: cularrp$2,
  	cupbrcap: cupbrcap$2,
  	cupcap: cupcap$2,
  	CupCap: CupCap$2,
  	cup: cup$2,
  	Cup: Cup$2,
  	cupcup: cupcup$2,
  	cupdot: cupdot$2,
  	cupor: cupor$2,
  	cups: cups$2,
  	curarr: curarr$2,
  	curarrm: curarrm$2,
  	curlyeqprec: curlyeqprec$2,
  	curlyeqsucc: curlyeqsucc$2,
  	curlyvee: curlyvee$2,
  	curlywedge: curlywedge$2,
  	curren: curren$3,
  	curvearrowleft: curvearrowleft$2,
  	curvearrowright: curvearrowright$2,
  	cuvee: cuvee$2,
  	cuwed: cuwed$2,
  	cwconint: cwconint$2,
  	cwint: cwint$2,
  	cylcty: cylcty$2,
  	dagger: dagger$2,
  	Dagger: Dagger$2,
  	daleth: daleth$2,
  	darr: darr$2,
  	Darr: Darr$2,
  	dArr: dArr$2,
  	dash: dash$2,
  	Dashv: Dashv$2,
  	dashv: dashv$2,
  	dbkarow: dbkarow$2,
  	dblac: dblac$2,
  	Dcaron: Dcaron$2,
  	dcaron: dcaron$2,
  	Dcy: Dcy$2,
  	dcy: dcy$2,
  	ddagger: ddagger$2,
  	ddarr: ddarr$2,
  	DD: DD$2,
  	dd: dd$2,
  	DDotrahd: DDotrahd$2,
  	ddotseq: ddotseq$2,
  	deg: deg$3,
  	Del: Del$2,
  	Delta: Delta$2,
  	delta: delta$2,
  	demptyv: demptyv$2,
  	dfisht: dfisht$2,
  	Dfr: Dfr$2,
  	dfr: dfr$2,
  	dHar: dHar$2,
  	dharl: dharl$2,
  	dharr: dharr$2,
  	DiacriticalAcute: DiacriticalAcute$2,
  	DiacriticalDot: DiacriticalDot$2,
  	DiacriticalDoubleAcute: DiacriticalDoubleAcute$2,
  	DiacriticalGrave: DiacriticalGrave$2,
  	DiacriticalTilde: DiacriticalTilde$2,
  	diam: diam$2,
  	diamond: diamond$2,
  	Diamond: Diamond$2,
  	diamondsuit: diamondsuit$2,
  	diams: diams$2,
  	die: die$2,
  	DifferentialD: DifferentialD$2,
  	digamma: digamma$2,
  	disin: disin$2,
  	div: div$2,
  	divide: divide$3,
  	divideontimes: divideontimes$2,
  	divonx: divonx$2,
  	DJcy: DJcy$2,
  	djcy: djcy$2,
  	dlcorn: dlcorn$2,
  	dlcrop: dlcrop$2,
  	dollar: dollar$2,
  	Dopf: Dopf$2,
  	dopf: dopf$2,
  	Dot: Dot$2,
  	dot: dot$2,
  	DotDot: DotDot$2,
  	doteq: doteq$2,
  	doteqdot: doteqdot$2,
  	DotEqual: DotEqual$2,
  	dotminus: dotminus$2,
  	dotplus: dotplus$2,
  	dotsquare: dotsquare$2,
  	doublebarwedge: doublebarwedge$2,
  	DoubleContourIntegral: DoubleContourIntegral$2,
  	DoubleDot: DoubleDot$2,
  	DoubleDownArrow: DoubleDownArrow$2,
  	DoubleLeftArrow: DoubleLeftArrow$2,
  	DoubleLeftRightArrow: DoubleLeftRightArrow$2,
  	DoubleLeftTee: DoubleLeftTee$2,
  	DoubleLongLeftArrow: DoubleLongLeftArrow$2,
  	DoubleLongLeftRightArrow: DoubleLongLeftRightArrow$2,
  	DoubleLongRightArrow: DoubleLongRightArrow$2,
  	DoubleRightArrow: DoubleRightArrow$2,
  	DoubleRightTee: DoubleRightTee$2,
  	DoubleUpArrow: DoubleUpArrow$2,
  	DoubleUpDownArrow: DoubleUpDownArrow$2,
  	DoubleVerticalBar: DoubleVerticalBar$2,
  	DownArrowBar: DownArrowBar$2,
  	downarrow: downarrow$2,
  	DownArrow: DownArrow$2,
  	Downarrow: Downarrow$2,
  	DownArrowUpArrow: DownArrowUpArrow$2,
  	DownBreve: DownBreve$2,
  	downdownarrows: downdownarrows$2,
  	downharpoonleft: downharpoonleft$2,
  	downharpoonright: downharpoonright$2,
  	DownLeftRightVector: DownLeftRightVector$2,
  	DownLeftTeeVector: DownLeftTeeVector$2,
  	DownLeftVectorBar: DownLeftVectorBar$2,
  	DownLeftVector: DownLeftVector$2,
  	DownRightTeeVector: DownRightTeeVector$2,
  	DownRightVectorBar: DownRightVectorBar$2,
  	DownRightVector: DownRightVector$2,
  	DownTeeArrow: DownTeeArrow$2,
  	DownTee: DownTee$2,
  	drbkarow: drbkarow$2,
  	drcorn: drcorn$2,
  	drcrop: drcrop$2,
  	Dscr: Dscr$2,
  	dscr: dscr$2,
  	DScy: DScy$2,
  	dscy: dscy$2,
  	dsol: dsol$2,
  	Dstrok: Dstrok$2,
  	dstrok: dstrok$2,
  	dtdot: dtdot$2,
  	dtri: dtri$2,
  	dtrif: dtrif$2,
  	duarr: duarr$2,
  	duhar: duhar$2,
  	dwangle: dwangle$2,
  	DZcy: DZcy$2,
  	dzcy: dzcy$2,
  	dzigrarr: dzigrarr$2,
  	Eacute: Eacute$3,
  	eacute: eacute$3,
  	easter: easter$2,
  	Ecaron: Ecaron$2,
  	ecaron: ecaron$2,
  	Ecirc: Ecirc$3,
  	ecirc: ecirc$3,
  	ecir: ecir$2,
  	ecolon: ecolon$2,
  	Ecy: Ecy$2,
  	ecy: ecy$2,
  	eDDot: eDDot$2,
  	Edot: Edot$2,
  	edot: edot$2,
  	eDot: eDot$2,
  	ee: ee$2,
  	efDot: efDot$2,
  	Efr: Efr$2,
  	efr: efr$2,
  	eg: eg$2,
  	Egrave: Egrave$3,
  	egrave: egrave$3,
  	egs: egs$2,
  	egsdot: egsdot$2,
  	el: el$2,
  	Element: Element$2,
  	elinters: elinters$2,
  	ell: ell$2,
  	els: els$2,
  	elsdot: elsdot$2,
  	Emacr: Emacr$2,
  	emacr: emacr$2,
  	empty: empty$2,
  	emptyset: emptyset$2,
  	EmptySmallSquare: EmptySmallSquare$2,
  	emptyv: emptyv$2,
  	EmptyVerySmallSquare: EmptyVerySmallSquare$2,
  	emsp13: emsp13$2,
  	emsp14: emsp14$2,
  	emsp: emsp$2,
  	ENG: ENG$2,
  	eng: eng$2,
  	ensp: ensp$2,
  	Eogon: Eogon$2,
  	eogon: eogon$2,
  	Eopf: Eopf$2,
  	eopf: eopf$2,
  	epar: epar$2,
  	eparsl: eparsl$2,
  	eplus: eplus$2,
  	epsi: epsi$2,
  	Epsilon: Epsilon$2,
  	epsilon: epsilon$2,
  	epsiv: epsiv$2,
  	eqcirc: eqcirc$2,
  	eqcolon: eqcolon$2,
  	eqsim: eqsim$2,
  	eqslantgtr: eqslantgtr$2,
  	eqslantless: eqslantless$2,
  	Equal: Equal$2,
  	equals: equals$2,
  	EqualTilde: EqualTilde$2,
  	equest: equest$2,
  	Equilibrium: Equilibrium$2,
  	equiv: equiv$2,
  	equivDD: equivDD$2,
  	eqvparsl: eqvparsl$2,
  	erarr: erarr$2,
  	erDot: erDot$2,
  	escr: escr$2,
  	Escr: Escr$2,
  	esdot: esdot$2,
  	Esim: Esim$2,
  	esim: esim$2,
  	Eta: Eta$2,
  	eta: eta$2,
  	ETH: ETH$3,
  	eth: eth$3,
  	Euml: Euml$3,
  	euml: euml$3,
  	euro: euro$2,
  	excl: excl$2,
  	exist: exist$2,
  	Exists: Exists$2,
  	expectation: expectation$2,
  	exponentiale: exponentiale$2,
  	ExponentialE: ExponentialE$2,
  	fallingdotseq: fallingdotseq$2,
  	Fcy: Fcy$2,
  	fcy: fcy$2,
  	female: female$2,
  	ffilig: ffilig$2,
  	fflig: fflig$2,
  	ffllig: ffllig$2,
  	Ffr: Ffr$2,
  	ffr: ffr$2,
  	filig: filig$2,
  	FilledSmallSquare: FilledSmallSquare$2,
  	FilledVerySmallSquare: FilledVerySmallSquare$2,
  	fjlig: fjlig$2,
  	flat: flat$2,
  	fllig: fllig$2,
  	fltns: fltns$2,
  	fnof: fnof$2,
  	Fopf: Fopf$2,
  	fopf: fopf$2,
  	forall: forall$2,
  	ForAll: ForAll$2,
  	fork: fork$2,
  	forkv: forkv$2,
  	Fouriertrf: Fouriertrf$2,
  	fpartint: fpartint$2,
  	frac12: frac12$3,
  	frac13: frac13$2,
  	frac14: frac14$3,
  	frac15: frac15$2,
  	frac16: frac16$2,
  	frac18: frac18$2,
  	frac23: frac23$2,
  	frac25: frac25$2,
  	frac34: frac34$3,
  	frac35: frac35$2,
  	frac38: frac38$2,
  	frac45: frac45$2,
  	frac56: frac56$2,
  	frac58: frac58$2,
  	frac78: frac78$2,
  	frasl: frasl$2,
  	frown: frown$2,
  	fscr: fscr$2,
  	Fscr: Fscr$2,
  	gacute: gacute$2,
  	Gamma: Gamma$2,
  	gamma: gamma$2,
  	Gammad: Gammad$2,
  	gammad: gammad$2,
  	gap: gap$2,
  	Gbreve: Gbreve$2,
  	gbreve: gbreve$2,
  	Gcedil: Gcedil$2,
  	Gcirc: Gcirc$2,
  	gcirc: gcirc$2,
  	Gcy: Gcy$2,
  	gcy: gcy$2,
  	Gdot: Gdot$2,
  	gdot: gdot$2,
  	ge: ge$2,
  	gE: gE$2,
  	gEl: gEl$2,
  	gel: gel$2,
  	geq: geq$2,
  	geqq: geqq$2,
  	geqslant: geqslant$2,
  	gescc: gescc$2,
  	ges: ges$2,
  	gesdot: gesdot$2,
  	gesdoto: gesdoto$2,
  	gesdotol: gesdotol$2,
  	gesl: gesl$2,
  	gesles: gesles$2,
  	Gfr: Gfr$2,
  	gfr: gfr$2,
  	gg: gg$2,
  	Gg: Gg$2,
  	ggg: ggg$2,
  	gimel: gimel$2,
  	GJcy: GJcy$2,
  	gjcy: gjcy$2,
  	gla: gla$2,
  	gl: gl$2,
  	glE: glE$2,
  	glj: glj$2,
  	gnap: gnap$2,
  	gnapprox: gnapprox$2,
  	gne: gne$2,
  	gnE: gnE$2,
  	gneq: gneq$2,
  	gneqq: gneqq$2,
  	gnsim: gnsim$2,
  	Gopf: Gopf$2,
  	gopf: gopf$2,
  	grave: grave$2,
  	GreaterEqual: GreaterEqual$2,
  	GreaterEqualLess: GreaterEqualLess$2,
  	GreaterFullEqual: GreaterFullEqual$2,
  	GreaterGreater: GreaterGreater$2,
  	GreaterLess: GreaterLess$2,
  	GreaterSlantEqual: GreaterSlantEqual$2,
  	GreaterTilde: GreaterTilde$2,
  	Gscr: Gscr$2,
  	gscr: gscr$2,
  	gsim: gsim$2,
  	gsime: gsime$2,
  	gsiml: gsiml$2,
  	gtcc: gtcc$2,
  	gtcir: gtcir$2,
  	gt: gt$1$2,
  	GT: GT$3,
  	Gt: Gt$2,
  	gtdot: gtdot$2,
  	gtlPar: gtlPar$2,
  	gtquest: gtquest$2,
  	gtrapprox: gtrapprox$2,
  	gtrarr: gtrarr$2,
  	gtrdot: gtrdot$2,
  	gtreqless: gtreqless$2,
  	gtreqqless: gtreqqless$2,
  	gtrless: gtrless$2,
  	gtrsim: gtrsim$2,
  	gvertneqq: gvertneqq$2,
  	gvnE: gvnE$2,
  	Hacek: Hacek$2,
  	hairsp: hairsp$2,
  	half: half$2,
  	hamilt: hamilt$2,
  	HARDcy: HARDcy$2,
  	hardcy: hardcy$2,
  	harrcir: harrcir$2,
  	harr: harr$2,
  	hArr: hArr$2,
  	harrw: harrw$2,
  	Hat: Hat$2,
  	hbar: hbar$2,
  	Hcirc: Hcirc$2,
  	hcirc: hcirc$2,
  	hearts: hearts$2,
  	heartsuit: heartsuit$2,
  	hellip: hellip$2,
  	hercon: hercon$2,
  	hfr: hfr$2,
  	Hfr: Hfr$2,
  	HilbertSpace: HilbertSpace$2,
  	hksearow: hksearow$2,
  	hkswarow: hkswarow$2,
  	hoarr: hoarr$2,
  	homtht: homtht$2,
  	hookleftarrow: hookleftarrow$2,
  	hookrightarrow: hookrightarrow$2,
  	hopf: hopf$2,
  	Hopf: Hopf$2,
  	horbar: horbar$2,
  	HorizontalLine: HorizontalLine$2,
  	hscr: hscr$2,
  	Hscr: Hscr$2,
  	hslash: hslash$2,
  	Hstrok: Hstrok$2,
  	hstrok: hstrok$2,
  	HumpDownHump: HumpDownHump$2,
  	HumpEqual: HumpEqual$2,
  	hybull: hybull$2,
  	hyphen: hyphen$2,
  	Iacute: Iacute$3,
  	iacute: iacute$3,
  	ic: ic$2,
  	Icirc: Icirc$3,
  	icirc: icirc$3,
  	Icy: Icy$2,
  	icy: icy$2,
  	Idot: Idot$2,
  	IEcy: IEcy$2,
  	iecy: iecy$2,
  	iexcl: iexcl$3,
  	iff: iff$2,
  	ifr: ifr$2,
  	Ifr: Ifr$2,
  	Igrave: Igrave$3,
  	igrave: igrave$3,
  	ii: ii$2,
  	iiiint: iiiint$2,
  	iiint: iiint$2,
  	iinfin: iinfin$2,
  	iiota: iiota$2,
  	IJlig: IJlig$2,
  	ijlig: ijlig$2,
  	Imacr: Imacr$2,
  	imacr: imacr$2,
  	image: image$2,
  	ImaginaryI: ImaginaryI$2,
  	imagline: imagline$2,
  	imagpart: imagpart$2,
  	imath: imath$2,
  	Im: Im$2,
  	imof: imof$2,
  	imped: imped$2,
  	Implies: Implies$2,
  	incare: incare$2,
  	infin: infin$2,
  	infintie: infintie$2,
  	inodot: inodot$2,
  	intcal: intcal$2,
  	int: int$2,
  	Int: Int$2,
  	integers: integers$2,
  	Integral: Integral$2,
  	intercal: intercal$2,
  	Intersection: Intersection$2,
  	intlarhk: intlarhk$2,
  	intprod: intprod$2,
  	InvisibleComma: InvisibleComma$2,
  	InvisibleTimes: InvisibleTimes$2,
  	IOcy: IOcy$2,
  	iocy: iocy$2,
  	Iogon: Iogon$2,
  	iogon: iogon$2,
  	Iopf: Iopf$2,
  	iopf: iopf$2,
  	Iota: Iota$2,
  	iota: iota$2,
  	iprod: iprod$2,
  	iquest: iquest$3,
  	iscr: iscr$2,
  	Iscr: Iscr$2,
  	isin: isin$2,
  	isindot: isindot$2,
  	isinE: isinE$2,
  	isins: isins$2,
  	isinsv: isinsv$2,
  	isinv: isinv$2,
  	it: it$2,
  	Itilde: Itilde$2,
  	itilde: itilde$2,
  	Iukcy: Iukcy$2,
  	iukcy: iukcy$2,
  	Iuml: Iuml$3,
  	iuml: iuml$3,
  	Jcirc: Jcirc$2,
  	jcirc: jcirc$2,
  	Jcy: Jcy$2,
  	jcy: jcy$2,
  	Jfr: Jfr$2,
  	jfr: jfr$2,
  	jmath: jmath$2,
  	Jopf: Jopf$2,
  	jopf: jopf$2,
  	Jscr: Jscr$2,
  	jscr: jscr$2,
  	Jsercy: Jsercy$2,
  	jsercy: jsercy$2,
  	Jukcy: Jukcy$2,
  	jukcy: jukcy$2,
  	Kappa: Kappa$2,
  	kappa: kappa$2,
  	kappav: kappav$2,
  	Kcedil: Kcedil$2,
  	kcedil: kcedil$2,
  	Kcy: Kcy$2,
  	kcy: kcy$2,
  	Kfr: Kfr$2,
  	kfr: kfr$2,
  	kgreen: kgreen$2,
  	KHcy: KHcy$2,
  	khcy: khcy$2,
  	KJcy: KJcy$2,
  	kjcy: kjcy$2,
  	Kopf: Kopf$2,
  	kopf: kopf$2,
  	Kscr: Kscr$2,
  	kscr: kscr$2,
  	lAarr: lAarr$2,
  	Lacute: Lacute$2,
  	lacute: lacute$2,
  	laemptyv: laemptyv$2,
  	lagran: lagran$2,
  	Lambda: Lambda$2,
  	lambda: lambda$2,
  	lang: lang$2,
  	Lang: Lang$2,
  	langd: langd$2,
  	langle: langle$2,
  	lap: lap$2,
  	Laplacetrf: Laplacetrf$2,
  	laquo: laquo$3,
  	larrb: larrb$2,
  	larrbfs: larrbfs$2,
  	larr: larr$2,
  	Larr: Larr$2,
  	lArr: lArr$2,
  	larrfs: larrfs$2,
  	larrhk: larrhk$2,
  	larrlp: larrlp$2,
  	larrpl: larrpl$2,
  	larrsim: larrsim$2,
  	larrtl: larrtl$2,
  	latail: latail$2,
  	lAtail: lAtail$2,
  	lat: lat$2,
  	late: late$2,
  	lates: lates$2,
  	lbarr: lbarr$2,
  	lBarr: lBarr$2,
  	lbbrk: lbbrk$2,
  	lbrace: lbrace$2,
  	lbrack: lbrack$2,
  	lbrke: lbrke$2,
  	lbrksld: lbrksld$2,
  	lbrkslu: lbrkslu$2,
  	Lcaron: Lcaron$2,
  	lcaron: lcaron$2,
  	Lcedil: Lcedil$2,
  	lcedil: lcedil$2,
  	lceil: lceil$2,
  	lcub: lcub$2,
  	Lcy: Lcy$2,
  	lcy: lcy$2,
  	ldca: ldca$2,
  	ldquo: ldquo$2,
  	ldquor: ldquor$2,
  	ldrdhar: ldrdhar$2,
  	ldrushar: ldrushar$2,
  	ldsh: ldsh$2,
  	le: le$2,
  	lE: lE$2,
  	LeftAngleBracket: LeftAngleBracket$2,
  	LeftArrowBar: LeftArrowBar$2,
  	leftarrow: leftarrow$2,
  	LeftArrow: LeftArrow$2,
  	Leftarrow: Leftarrow$2,
  	LeftArrowRightArrow: LeftArrowRightArrow$2,
  	leftarrowtail: leftarrowtail$2,
  	LeftCeiling: LeftCeiling$2,
  	LeftDoubleBracket: LeftDoubleBracket$2,
  	LeftDownTeeVector: LeftDownTeeVector$2,
  	LeftDownVectorBar: LeftDownVectorBar$2,
  	LeftDownVector: LeftDownVector$2,
  	LeftFloor: LeftFloor$2,
  	leftharpoondown: leftharpoondown$2,
  	leftharpoonup: leftharpoonup$2,
  	leftleftarrows: leftleftarrows$2,
  	leftrightarrow: leftrightarrow$2,
  	LeftRightArrow: LeftRightArrow$2,
  	Leftrightarrow: Leftrightarrow$2,
  	leftrightarrows: leftrightarrows$2,
  	leftrightharpoons: leftrightharpoons$2,
  	leftrightsquigarrow: leftrightsquigarrow$2,
  	LeftRightVector: LeftRightVector$2,
  	LeftTeeArrow: LeftTeeArrow$2,
  	LeftTee: LeftTee$2,
  	LeftTeeVector: LeftTeeVector$2,
  	leftthreetimes: leftthreetimes$2,
  	LeftTriangleBar: LeftTriangleBar$2,
  	LeftTriangle: LeftTriangle$2,
  	LeftTriangleEqual: LeftTriangleEqual$2,
  	LeftUpDownVector: LeftUpDownVector$2,
  	LeftUpTeeVector: LeftUpTeeVector$2,
  	LeftUpVectorBar: LeftUpVectorBar$2,
  	LeftUpVector: LeftUpVector$2,
  	LeftVectorBar: LeftVectorBar$2,
  	LeftVector: LeftVector$2,
  	lEg: lEg$2,
  	leg: leg$2,
  	leq: leq$2,
  	leqq: leqq$2,
  	leqslant: leqslant$2,
  	lescc: lescc$2,
  	les: les$2,
  	lesdot: lesdot$2,
  	lesdoto: lesdoto$2,
  	lesdotor: lesdotor$2,
  	lesg: lesg$2,
  	lesges: lesges$2,
  	lessapprox: lessapprox$2,
  	lessdot: lessdot$2,
  	lesseqgtr: lesseqgtr$2,
  	lesseqqgtr: lesseqqgtr$2,
  	LessEqualGreater: LessEqualGreater$2,
  	LessFullEqual: LessFullEqual$2,
  	LessGreater: LessGreater$2,
  	lessgtr: lessgtr$2,
  	LessLess: LessLess$2,
  	lesssim: lesssim$2,
  	LessSlantEqual: LessSlantEqual$2,
  	LessTilde: LessTilde$2,
  	lfisht: lfisht$2,
  	lfloor: lfloor$2,
  	Lfr: Lfr$2,
  	lfr: lfr$2,
  	lg: lg$2,
  	lgE: lgE$2,
  	lHar: lHar$2,
  	lhard: lhard$2,
  	lharu: lharu$2,
  	lharul: lharul$2,
  	lhblk: lhblk$2,
  	LJcy: LJcy$2,
  	ljcy: ljcy$2,
  	llarr: llarr$2,
  	ll: ll$2,
  	Ll: Ll$2,
  	llcorner: llcorner$2,
  	Lleftarrow: Lleftarrow$2,
  	llhard: llhard$2,
  	lltri: lltri$2,
  	Lmidot: Lmidot$2,
  	lmidot: lmidot$2,
  	lmoustache: lmoustache$2,
  	lmoust: lmoust$2,
  	lnap: lnap$2,
  	lnapprox: lnapprox$2,
  	lne: lne$2,
  	lnE: lnE$2,
  	lneq: lneq$2,
  	lneqq: lneqq$2,
  	lnsim: lnsim$2,
  	loang: loang$2,
  	loarr: loarr$2,
  	lobrk: lobrk$2,
  	longleftarrow: longleftarrow$2,
  	LongLeftArrow: LongLeftArrow$2,
  	Longleftarrow: Longleftarrow$2,
  	longleftrightarrow: longleftrightarrow$2,
  	LongLeftRightArrow: LongLeftRightArrow$2,
  	Longleftrightarrow: Longleftrightarrow$2,
  	longmapsto: longmapsto$2,
  	longrightarrow: longrightarrow$2,
  	LongRightArrow: LongRightArrow$2,
  	Longrightarrow: Longrightarrow$2,
  	looparrowleft: looparrowleft$2,
  	looparrowright: looparrowright$2,
  	lopar: lopar$2,
  	Lopf: Lopf$2,
  	lopf: lopf$2,
  	loplus: loplus$2,
  	lotimes: lotimes$2,
  	lowast: lowast$2,
  	lowbar: lowbar$2,
  	LowerLeftArrow: LowerLeftArrow$2,
  	LowerRightArrow: LowerRightArrow$2,
  	loz: loz$2,
  	lozenge: lozenge$2,
  	lozf: lozf$2,
  	lpar: lpar$2,
  	lparlt: lparlt$2,
  	lrarr: lrarr$2,
  	lrcorner: lrcorner$2,
  	lrhar: lrhar$2,
  	lrhard: lrhard$2,
  	lrm: lrm$2,
  	lrtri: lrtri$2,
  	lsaquo: lsaquo$2,
  	lscr: lscr$2,
  	Lscr: Lscr$2,
  	lsh: lsh$2,
  	Lsh: Lsh$2,
  	lsim: lsim$2,
  	lsime: lsime$2,
  	lsimg: lsimg$2,
  	lsqb: lsqb$2,
  	lsquo: lsquo$2,
  	lsquor: lsquor$2,
  	Lstrok: Lstrok$2,
  	lstrok: lstrok$2,
  	ltcc: ltcc$2,
  	ltcir: ltcir$2,
  	lt: lt$1$2,
  	LT: LT$3,
  	Lt: Lt$2,
  	ltdot: ltdot$2,
  	lthree: lthree$2,
  	ltimes: ltimes$2,
  	ltlarr: ltlarr$2,
  	ltquest: ltquest$2,
  	ltri: ltri$2,
  	ltrie: ltrie$2,
  	ltrif: ltrif$2,
  	ltrPar: ltrPar$2,
  	lurdshar: lurdshar$2,
  	luruhar: luruhar$2,
  	lvertneqq: lvertneqq$2,
  	lvnE: lvnE$2,
  	macr: macr$3,
  	male: male$2,
  	malt: malt$2,
  	maltese: maltese$2,
  	map: map$1$2,
  	mapsto: mapsto$2,
  	mapstodown: mapstodown$2,
  	mapstoleft: mapstoleft$2,
  	mapstoup: mapstoup$2,
  	marker: marker$2,
  	mcomma: mcomma$2,
  	Mcy: Mcy$2,
  	mcy: mcy$2,
  	mdash: mdash$2,
  	mDDot: mDDot$2,
  	measuredangle: measuredangle$2,
  	MediumSpace: MediumSpace$2,
  	Mellintrf: Mellintrf$2,
  	Mfr: Mfr$2,
  	mfr: mfr$2,
  	mho: mho$2,
  	micro: micro$3,
  	midast: midast$2,
  	midcir: midcir$2,
  	mid: mid$2,
  	middot: middot$3,
  	minusb: minusb$2,
  	minus: minus$2,
  	minusd: minusd$2,
  	minusdu: minusdu$2,
  	MinusPlus: MinusPlus$2,
  	mlcp: mlcp$2,
  	mldr: mldr$2,
  	mnplus: mnplus$2,
  	models: models$2,
  	Mopf: Mopf$2,
  	mopf: mopf$2,
  	mp: mp$2,
  	mscr: mscr$2,
  	Mscr: Mscr$2,
  	mstpos: mstpos$2,
  	Mu: Mu$2,
  	mu: mu$2,
  	multimap: multimap$2,
  	mumap: mumap$2,
  	nabla: nabla$2,
  	Nacute: Nacute$2,
  	nacute: nacute$2,
  	nang: nang$2,
  	nap: nap$2,
  	napE: napE$2,
  	napid: napid$2,
  	napos: napos$2,
  	napprox: napprox$2,
  	natural: natural$2,
  	naturals: naturals$2,
  	natur: natur$2,
  	nbsp: nbsp$3,
  	nbump: nbump$2,
  	nbumpe: nbumpe$2,
  	ncap: ncap$2,
  	Ncaron: Ncaron$2,
  	ncaron: ncaron$2,
  	Ncedil: Ncedil$2,
  	ncedil: ncedil$2,
  	ncong: ncong$2,
  	ncongdot: ncongdot$2,
  	ncup: ncup$2,
  	Ncy: Ncy$2,
  	ncy: ncy$2,
  	ndash: ndash$2,
  	nearhk: nearhk$2,
  	nearr: nearr$2,
  	neArr: neArr$2,
  	nearrow: nearrow$2,
  	ne: ne$2,
  	nedot: nedot$2,
  	NegativeMediumSpace: NegativeMediumSpace$2,
  	NegativeThickSpace: NegativeThickSpace$2,
  	NegativeThinSpace: NegativeThinSpace$2,
  	NegativeVeryThinSpace: NegativeVeryThinSpace$2,
  	nequiv: nequiv$2,
  	nesear: nesear$2,
  	nesim: nesim$2,
  	NestedGreaterGreater: NestedGreaterGreater$2,
  	NestedLessLess: NestedLessLess$2,
  	NewLine: NewLine$2,
  	nexist: nexist$2,
  	nexists: nexists$2,
  	Nfr: Nfr$2,
  	nfr: nfr$2,
  	ngE: ngE$2,
  	nge: nge$2,
  	ngeq: ngeq$2,
  	ngeqq: ngeqq$2,
  	ngeqslant: ngeqslant$2,
  	nges: nges$2,
  	nGg: nGg$2,
  	ngsim: ngsim$2,
  	nGt: nGt$2,
  	ngt: ngt$2,
  	ngtr: ngtr$2,
  	nGtv: nGtv$2,
  	nharr: nharr$2,
  	nhArr: nhArr$2,
  	nhpar: nhpar$2,
  	ni: ni$2,
  	nis: nis$2,
  	nisd: nisd$2,
  	niv: niv$2,
  	NJcy: NJcy$2,
  	njcy: njcy$2,
  	nlarr: nlarr$2,
  	nlArr: nlArr$2,
  	nldr: nldr$2,
  	nlE: nlE$2,
  	nle: nle$2,
  	nleftarrow: nleftarrow$2,
  	nLeftarrow: nLeftarrow$2,
  	nleftrightarrow: nleftrightarrow$2,
  	nLeftrightarrow: nLeftrightarrow$2,
  	nleq: nleq$2,
  	nleqq: nleqq$2,
  	nleqslant: nleqslant$2,
  	nles: nles$2,
  	nless: nless$2,
  	nLl: nLl$2,
  	nlsim: nlsim$2,
  	nLt: nLt$2,
  	nlt: nlt$2,
  	nltri: nltri$2,
  	nltrie: nltrie$2,
  	nLtv: nLtv$2,
  	nmid: nmid$2,
  	NoBreak: NoBreak$2,
  	NonBreakingSpace: NonBreakingSpace$2,
  	nopf: nopf$2,
  	Nopf: Nopf$2,
  	Not: Not$2,
  	not: not$3,
  	NotCongruent: NotCongruent$2,
  	NotCupCap: NotCupCap$2,
  	NotDoubleVerticalBar: NotDoubleVerticalBar$2,
  	NotElement: NotElement$2,
  	NotEqual: NotEqual$2,
  	NotEqualTilde: NotEqualTilde$2,
  	NotExists: NotExists$2,
  	NotGreater: NotGreater$2,
  	NotGreaterEqual: NotGreaterEqual$2,
  	NotGreaterFullEqual: NotGreaterFullEqual$2,
  	NotGreaterGreater: NotGreaterGreater$2,
  	NotGreaterLess: NotGreaterLess$2,
  	NotGreaterSlantEqual: NotGreaterSlantEqual$2,
  	NotGreaterTilde: NotGreaterTilde$2,
  	NotHumpDownHump: NotHumpDownHump$2,
  	NotHumpEqual: NotHumpEqual$2,
  	notin: notin$2,
  	notindot: notindot$2,
  	notinE: notinE$2,
  	notinva: notinva$2,
  	notinvb: notinvb$2,
  	notinvc: notinvc$2,
  	NotLeftTriangleBar: NotLeftTriangleBar$2,
  	NotLeftTriangle: NotLeftTriangle$2,
  	NotLeftTriangleEqual: NotLeftTriangleEqual$2,
  	NotLess: NotLess$2,
  	NotLessEqual: NotLessEqual$2,
  	NotLessGreater: NotLessGreater$2,
  	NotLessLess: NotLessLess$2,
  	NotLessSlantEqual: NotLessSlantEqual$2,
  	NotLessTilde: NotLessTilde$2,
  	NotNestedGreaterGreater: NotNestedGreaterGreater$2,
  	NotNestedLessLess: NotNestedLessLess$2,
  	notni: notni$2,
  	notniva: notniva$2,
  	notnivb: notnivb$2,
  	notnivc: notnivc$2,
  	NotPrecedes: NotPrecedes$2,
  	NotPrecedesEqual: NotPrecedesEqual$2,
  	NotPrecedesSlantEqual: NotPrecedesSlantEqual$2,
  	NotReverseElement: NotReverseElement$2,
  	NotRightTriangleBar: NotRightTriangleBar$2,
  	NotRightTriangle: NotRightTriangle$2,
  	NotRightTriangleEqual: NotRightTriangleEqual$2,
  	NotSquareSubset: NotSquareSubset$2,
  	NotSquareSubsetEqual: NotSquareSubsetEqual$2,
  	NotSquareSuperset: NotSquareSuperset$2,
  	NotSquareSupersetEqual: NotSquareSupersetEqual$2,
  	NotSubset: NotSubset$2,
  	NotSubsetEqual: NotSubsetEqual$2,
  	NotSucceeds: NotSucceeds$2,
  	NotSucceedsEqual: NotSucceedsEqual$2,
  	NotSucceedsSlantEqual: NotSucceedsSlantEqual$2,
  	NotSucceedsTilde: NotSucceedsTilde$2,
  	NotSuperset: NotSuperset$2,
  	NotSupersetEqual: NotSupersetEqual$2,
  	NotTilde: NotTilde$2,
  	NotTildeEqual: NotTildeEqual$2,
  	NotTildeFullEqual: NotTildeFullEqual$2,
  	NotTildeTilde: NotTildeTilde$2,
  	NotVerticalBar: NotVerticalBar$2,
  	nparallel: nparallel$2,
  	npar: npar$2,
  	nparsl: nparsl$2,
  	npart: npart$2,
  	npolint: npolint$2,
  	npr: npr$2,
  	nprcue: nprcue$2,
  	nprec: nprec$2,
  	npreceq: npreceq$2,
  	npre: npre$2,
  	nrarrc: nrarrc$2,
  	nrarr: nrarr$2,
  	nrArr: nrArr$2,
  	nrarrw: nrarrw$2,
  	nrightarrow: nrightarrow$2,
  	nRightarrow: nRightarrow$2,
  	nrtri: nrtri$2,
  	nrtrie: nrtrie$2,
  	nsc: nsc$2,
  	nsccue: nsccue$2,
  	nsce: nsce$2,
  	Nscr: Nscr$2,
  	nscr: nscr$2,
  	nshortmid: nshortmid$2,
  	nshortparallel: nshortparallel$2,
  	nsim: nsim$2,
  	nsime: nsime$2,
  	nsimeq: nsimeq$2,
  	nsmid: nsmid$2,
  	nspar: nspar$2,
  	nsqsube: nsqsube$2,
  	nsqsupe: nsqsupe$2,
  	nsub: nsub$2,
  	nsubE: nsubE$2,
  	nsube: nsube$2,
  	nsubset: nsubset$2,
  	nsubseteq: nsubseteq$2,
  	nsubseteqq: nsubseteqq$2,
  	nsucc: nsucc$2,
  	nsucceq: nsucceq$2,
  	nsup: nsup$2,
  	nsupE: nsupE$2,
  	nsupe: nsupe$2,
  	nsupset: nsupset$2,
  	nsupseteq: nsupseteq$2,
  	nsupseteqq: nsupseteqq$2,
  	ntgl: ntgl$2,
  	Ntilde: Ntilde$3,
  	ntilde: ntilde$3,
  	ntlg: ntlg$2,
  	ntriangleleft: ntriangleleft$2,
  	ntrianglelefteq: ntrianglelefteq$2,
  	ntriangleright: ntriangleright$2,
  	ntrianglerighteq: ntrianglerighteq$2,
  	Nu: Nu$2,
  	nu: nu$2,
  	num: num$2,
  	numero: numero$2,
  	numsp: numsp$2,
  	nvap: nvap$2,
  	nvdash: nvdash$2,
  	nvDash: nvDash$2,
  	nVdash: nVdash$2,
  	nVDash: nVDash$2,
  	nvge: nvge$2,
  	nvgt: nvgt$2,
  	nvHarr: nvHarr$2,
  	nvinfin: nvinfin$2,
  	nvlArr: nvlArr$2,
  	nvle: nvle$2,
  	nvlt: nvlt$2,
  	nvltrie: nvltrie$2,
  	nvrArr: nvrArr$2,
  	nvrtrie: nvrtrie$2,
  	nvsim: nvsim$2,
  	nwarhk: nwarhk$2,
  	nwarr: nwarr$2,
  	nwArr: nwArr$2,
  	nwarrow: nwarrow$2,
  	nwnear: nwnear$2,
  	Oacute: Oacute$3,
  	oacute: oacute$3,
  	oast: oast$2,
  	Ocirc: Ocirc$3,
  	ocirc: ocirc$3,
  	ocir: ocir$2,
  	Ocy: Ocy$2,
  	ocy: ocy$2,
  	odash: odash$2,
  	Odblac: Odblac$2,
  	odblac: odblac$2,
  	odiv: odiv$2,
  	odot: odot$2,
  	odsold: odsold$2,
  	OElig: OElig$2,
  	oelig: oelig$2,
  	ofcir: ofcir$2,
  	Ofr: Ofr$2,
  	ofr: ofr$2,
  	ogon: ogon$2,
  	Ograve: Ograve$3,
  	ograve: ograve$3,
  	ogt: ogt$2,
  	ohbar: ohbar$2,
  	ohm: ohm$2,
  	oint: oint$2,
  	olarr: olarr$2,
  	olcir: olcir$2,
  	olcross: olcross$2,
  	oline: oline$2,
  	olt: olt$2,
  	Omacr: Omacr$2,
  	omacr: omacr$2,
  	Omega: Omega$2,
  	omega: omega$2,
  	Omicron: Omicron$2,
  	omicron: omicron$2,
  	omid: omid$2,
  	ominus: ominus$2,
  	Oopf: Oopf$2,
  	oopf: oopf$2,
  	opar: opar$2,
  	OpenCurlyDoubleQuote: OpenCurlyDoubleQuote$2,
  	OpenCurlyQuote: OpenCurlyQuote$2,
  	operp: operp$2,
  	oplus: oplus$2,
  	orarr: orarr$2,
  	Or: Or$2,
  	or: or$2,
  	ord: ord$2,
  	order: order$2,
  	orderof: orderof$2,
  	ordf: ordf$3,
  	ordm: ordm$3,
  	origof: origof$2,
  	oror: oror$2,
  	orslope: orslope$2,
  	orv: orv$2,
  	oS: oS$2,
  	Oscr: Oscr$2,
  	oscr: oscr$2,
  	Oslash: Oslash$3,
  	oslash: oslash$3,
  	osol: osol$2,
  	Otilde: Otilde$3,
  	otilde: otilde$3,
  	otimesas: otimesas$2,
  	Otimes: Otimes$2,
  	otimes: otimes$2,
  	Ouml: Ouml$3,
  	ouml: ouml$3,
  	ovbar: ovbar$2,
  	OverBar: OverBar$2,
  	OverBrace: OverBrace$2,
  	OverBracket: OverBracket$2,
  	OverParenthesis: OverParenthesis$2,
  	para: para$3,
  	parallel: parallel$2,
  	par: par$2,
  	parsim: parsim$2,
  	parsl: parsl$2,
  	part: part$2,
  	PartialD: PartialD$2,
  	Pcy: Pcy$2,
  	pcy: pcy$2,
  	percnt: percnt$2,
  	period: period$2,
  	permil: permil$2,
  	perp: perp$2,
  	pertenk: pertenk$2,
  	Pfr: Pfr$2,
  	pfr: pfr$2,
  	Phi: Phi$2,
  	phi: phi$2,
  	phiv: phiv$2,
  	phmmat: phmmat$2,
  	phone: phone$2,
  	Pi: Pi$2,
  	pi: pi$2,
  	pitchfork: pitchfork$2,
  	piv: piv$2,
  	planck: planck$2,
  	planckh: planckh$2,
  	plankv: plankv$2,
  	plusacir: plusacir$2,
  	plusb: plusb$2,
  	pluscir: pluscir$2,
  	plus: plus$2,
  	plusdo: plusdo$2,
  	plusdu: plusdu$2,
  	pluse: pluse$2,
  	PlusMinus: PlusMinus$2,
  	plusmn: plusmn$3,
  	plussim: plussim$2,
  	plustwo: plustwo$2,
  	pm: pm$2,
  	Poincareplane: Poincareplane$2,
  	pointint: pointint$2,
  	popf: popf$2,
  	Popf: Popf$2,
  	pound: pound$3,
  	prap: prap$2,
  	Pr: Pr$2,
  	pr: pr$2,
  	prcue: prcue$2,
  	precapprox: precapprox$2,
  	prec: prec$2,
  	preccurlyeq: preccurlyeq$2,
  	Precedes: Precedes$2,
  	PrecedesEqual: PrecedesEqual$2,
  	PrecedesSlantEqual: PrecedesSlantEqual$2,
  	PrecedesTilde: PrecedesTilde$2,
  	preceq: preceq$2,
  	precnapprox: precnapprox$2,
  	precneqq: precneqq$2,
  	precnsim: precnsim$2,
  	pre: pre$2,
  	prE: prE$2,
  	precsim: precsim$2,
  	prime: prime$2,
  	Prime: Prime$2,
  	primes: primes$2,
  	prnap: prnap$2,
  	prnE: prnE$2,
  	prnsim: prnsim$2,
  	prod: prod$2,
  	Product: Product$2,
  	profalar: profalar$2,
  	profline: profline$2,
  	profsurf: profsurf$2,
  	prop: prop$2,
  	Proportional: Proportional$2,
  	Proportion: Proportion$2,
  	propto: propto$2,
  	prsim: prsim$2,
  	prurel: prurel$2,
  	Pscr: Pscr$2,
  	pscr: pscr$2,
  	Psi: Psi$2,
  	psi: psi$2,
  	puncsp: puncsp$2,
  	Qfr: Qfr$2,
  	qfr: qfr$2,
  	qint: qint$2,
  	qopf: qopf$2,
  	Qopf: Qopf$2,
  	qprime: qprime$2,
  	Qscr: Qscr$2,
  	qscr: qscr$2,
  	quaternions: quaternions$2,
  	quatint: quatint$2,
  	quest: quest$2,
  	questeq: questeq$2,
  	quot: quot$1$2,
  	QUOT: QUOT$3,
  	rAarr: rAarr$2,
  	race: race$2,
  	Racute: Racute$2,
  	racute: racute$2,
  	radic: radic$2,
  	raemptyv: raemptyv$2,
  	rang: rang$2,
  	Rang: Rang$2,
  	rangd: rangd$2,
  	range: range$2,
  	rangle: rangle$2,
  	raquo: raquo$3,
  	rarrap: rarrap$2,
  	rarrb: rarrb$2,
  	rarrbfs: rarrbfs$2,
  	rarrc: rarrc$2,
  	rarr: rarr$2,
  	Rarr: Rarr$2,
  	rArr: rArr$2,
  	rarrfs: rarrfs$2,
  	rarrhk: rarrhk$2,
  	rarrlp: rarrlp$2,
  	rarrpl: rarrpl$2,
  	rarrsim: rarrsim$2,
  	Rarrtl: Rarrtl$2,
  	rarrtl: rarrtl$2,
  	rarrw: rarrw$2,
  	ratail: ratail$2,
  	rAtail: rAtail$2,
  	ratio: ratio$2,
  	rationals: rationals$2,
  	rbarr: rbarr$2,
  	rBarr: rBarr$2,
  	RBarr: RBarr$2,
  	rbbrk: rbbrk$2,
  	rbrace: rbrace$2,
  	rbrack: rbrack$2,
  	rbrke: rbrke$2,
  	rbrksld: rbrksld$2,
  	rbrkslu: rbrkslu$2,
  	Rcaron: Rcaron$2,
  	rcaron: rcaron$2,
  	Rcedil: Rcedil$2,
  	rcedil: rcedil$2,
  	rceil: rceil$2,
  	rcub: rcub$2,
  	Rcy: Rcy$2,
  	rcy: rcy$2,
  	rdca: rdca$2,
  	rdldhar: rdldhar$2,
  	rdquo: rdquo$2,
  	rdquor: rdquor$2,
  	rdsh: rdsh$2,
  	real: real$2,
  	realine: realine$2,
  	realpart: realpart$2,
  	reals: reals$2,
  	Re: Re$2,
  	rect: rect$2,
  	reg: reg$3,
  	REG: REG$3,
  	ReverseElement: ReverseElement$2,
  	ReverseEquilibrium: ReverseEquilibrium$2,
  	ReverseUpEquilibrium: ReverseUpEquilibrium$2,
  	rfisht: rfisht$2,
  	rfloor: rfloor$2,
  	rfr: rfr$2,
  	Rfr: Rfr$2,
  	rHar: rHar$2,
  	rhard: rhard$2,
  	rharu: rharu$2,
  	rharul: rharul$2,
  	Rho: Rho$2,
  	rho: rho$2,
  	rhov: rhov$2,
  	RightAngleBracket: RightAngleBracket$2,
  	RightArrowBar: RightArrowBar$2,
  	rightarrow: rightarrow$2,
  	RightArrow: RightArrow$2,
  	Rightarrow: Rightarrow$2,
  	RightArrowLeftArrow: RightArrowLeftArrow$2,
  	rightarrowtail: rightarrowtail$2,
  	RightCeiling: RightCeiling$2,
  	RightDoubleBracket: RightDoubleBracket$2,
  	RightDownTeeVector: RightDownTeeVector$2,
  	RightDownVectorBar: RightDownVectorBar$2,
  	RightDownVector: RightDownVector$2,
  	RightFloor: RightFloor$2,
  	rightharpoondown: rightharpoondown$2,
  	rightharpoonup: rightharpoonup$2,
  	rightleftarrows: rightleftarrows$2,
  	rightleftharpoons: rightleftharpoons$2,
  	rightrightarrows: rightrightarrows$2,
  	rightsquigarrow: rightsquigarrow$2,
  	RightTeeArrow: RightTeeArrow$2,
  	RightTee: RightTee$2,
  	RightTeeVector: RightTeeVector$2,
  	rightthreetimes: rightthreetimes$2,
  	RightTriangleBar: RightTriangleBar$2,
  	RightTriangle: RightTriangle$2,
  	RightTriangleEqual: RightTriangleEqual$2,
  	RightUpDownVector: RightUpDownVector$2,
  	RightUpTeeVector: RightUpTeeVector$2,
  	RightUpVectorBar: RightUpVectorBar$2,
  	RightUpVector: RightUpVector$2,
  	RightVectorBar: RightVectorBar$2,
  	RightVector: RightVector$2,
  	ring: ring$2,
  	risingdotseq: risingdotseq$2,
  	rlarr: rlarr$2,
  	rlhar: rlhar$2,
  	rlm: rlm$2,
  	rmoustache: rmoustache$2,
  	rmoust: rmoust$2,
  	rnmid: rnmid$2,
  	roang: roang$2,
  	roarr: roarr$2,
  	robrk: robrk$2,
  	ropar: ropar$2,
  	ropf: ropf$2,
  	Ropf: Ropf$2,
  	roplus: roplus$2,
  	rotimes: rotimes$2,
  	RoundImplies: RoundImplies$2,
  	rpar: rpar$2,
  	rpargt: rpargt$2,
  	rppolint: rppolint$2,
  	rrarr: rrarr$2,
  	Rrightarrow: Rrightarrow$2,
  	rsaquo: rsaquo$2,
  	rscr: rscr$2,
  	Rscr: Rscr$2,
  	rsh: rsh$2,
  	Rsh: Rsh$2,
  	rsqb: rsqb$2,
  	rsquo: rsquo$2,
  	rsquor: rsquor$2,
  	rthree: rthree$2,
  	rtimes: rtimes$2,
  	rtri: rtri$2,
  	rtrie: rtrie$2,
  	rtrif: rtrif$2,
  	rtriltri: rtriltri$2,
  	RuleDelayed: RuleDelayed$2,
  	ruluhar: ruluhar$2,
  	rx: rx$2,
  	Sacute: Sacute$2,
  	sacute: sacute$2,
  	sbquo: sbquo$2,
  	scap: scap$2,
  	Scaron: Scaron$2,
  	scaron: scaron$2,
  	Sc: Sc$2,
  	sc: sc$2,
  	sccue: sccue$2,
  	sce: sce$2,
  	scE: scE$2,
  	Scedil: Scedil$2,
  	scedil: scedil$2,
  	Scirc: Scirc$2,
  	scirc: scirc$2,
  	scnap: scnap$2,
  	scnE: scnE$2,
  	scnsim: scnsim$2,
  	scpolint: scpolint$2,
  	scsim: scsim$2,
  	Scy: Scy$2,
  	scy: scy$2,
  	sdotb: sdotb$2,
  	sdot: sdot$2,
  	sdote: sdote$2,
  	searhk: searhk$2,
  	searr: searr$2,
  	seArr: seArr$2,
  	searrow: searrow$2,
  	sect: sect$3,
  	semi: semi$2,
  	seswar: seswar$2,
  	setminus: setminus$2,
  	setmn: setmn$2,
  	sext: sext$2,
  	Sfr: Sfr$2,
  	sfr: sfr$2,
  	sfrown: sfrown$2,
  	sharp: sharp$2,
  	SHCHcy: SHCHcy$2,
  	shchcy: shchcy$2,
  	SHcy: SHcy$2,
  	shcy: shcy$2,
  	ShortDownArrow: ShortDownArrow$2,
  	ShortLeftArrow: ShortLeftArrow$2,
  	shortmid: shortmid$2,
  	shortparallel: shortparallel$2,
  	ShortRightArrow: ShortRightArrow$2,
  	ShortUpArrow: ShortUpArrow$2,
  	shy: shy$3,
  	Sigma: Sigma$2,
  	sigma: sigma$2,
  	sigmaf: sigmaf$2,
  	sigmav: sigmav$2,
  	sim: sim$2,
  	simdot: simdot$2,
  	sime: sime$2,
  	simeq: simeq$2,
  	simg: simg$2,
  	simgE: simgE$2,
  	siml: siml$2,
  	simlE: simlE$2,
  	simne: simne$2,
  	simplus: simplus$2,
  	simrarr: simrarr$2,
  	slarr: slarr$2,
  	SmallCircle: SmallCircle$2,
  	smallsetminus: smallsetminus$2,
  	smashp: smashp$2,
  	smeparsl: smeparsl$2,
  	smid: smid$2,
  	smile: smile$2,
  	smt: smt$2,
  	smte: smte$2,
  	smtes: smtes$2,
  	SOFTcy: SOFTcy$2,
  	softcy: softcy$2,
  	solbar: solbar$2,
  	solb: solb$2,
  	sol: sol$2,
  	Sopf: Sopf$2,
  	sopf: sopf$2,
  	spades: spades$2,
  	spadesuit: spadesuit$2,
  	spar: spar$2,
  	sqcap: sqcap$2,
  	sqcaps: sqcaps$2,
  	sqcup: sqcup$2,
  	sqcups: sqcups$2,
  	Sqrt: Sqrt$2,
  	sqsub: sqsub$2,
  	sqsube: sqsube$2,
  	sqsubset: sqsubset$2,
  	sqsubseteq: sqsubseteq$2,
  	sqsup: sqsup$2,
  	sqsupe: sqsupe$2,
  	sqsupset: sqsupset$2,
  	sqsupseteq: sqsupseteq$2,
  	square: square$2,
  	Square: Square$2,
  	SquareIntersection: SquareIntersection$2,
  	SquareSubset: SquareSubset$2,
  	SquareSubsetEqual: SquareSubsetEqual$2,
  	SquareSuperset: SquareSuperset$2,
  	SquareSupersetEqual: SquareSupersetEqual$2,
  	SquareUnion: SquareUnion$2,
  	squarf: squarf$2,
  	squ: squ$2,
  	squf: squf$2,
  	srarr: srarr$2,
  	Sscr: Sscr$2,
  	sscr: sscr$2,
  	ssetmn: ssetmn$2,
  	ssmile: ssmile$2,
  	sstarf: sstarf$2,
  	Star: Star$2,
  	star: star$2,
  	starf: starf$2,
  	straightepsilon: straightepsilon$2,
  	straightphi: straightphi$2,
  	strns: strns$2,
  	sub: sub$2,
  	Sub: Sub$2,
  	subdot: subdot$2,
  	subE: subE$2,
  	sube: sube$2,
  	subedot: subedot$2,
  	submult: submult$2,
  	subnE: subnE$2,
  	subne: subne$2,
  	subplus: subplus$2,
  	subrarr: subrarr$2,
  	subset: subset$2,
  	Subset: Subset$2,
  	subseteq: subseteq$2,
  	subseteqq: subseteqq$2,
  	SubsetEqual: SubsetEqual$2,
  	subsetneq: subsetneq$2,
  	subsetneqq: subsetneqq$2,
  	subsim: subsim$2,
  	subsub: subsub$2,
  	subsup: subsup$2,
  	succapprox: succapprox$2,
  	succ: succ$2,
  	succcurlyeq: succcurlyeq$2,
  	Succeeds: Succeeds$2,
  	SucceedsEqual: SucceedsEqual$2,
  	SucceedsSlantEqual: SucceedsSlantEqual$2,
  	SucceedsTilde: SucceedsTilde$2,
  	succeq: succeq$2,
  	succnapprox: succnapprox$2,
  	succneqq: succneqq$2,
  	succnsim: succnsim$2,
  	succsim: succsim$2,
  	SuchThat: SuchThat$2,
  	sum: sum$2,
  	Sum: Sum$2,
  	sung: sung$2,
  	sup1: sup1$3,
  	sup2: sup2$3,
  	sup3: sup3$3,
  	sup: sup$2,
  	Sup: Sup$2,
  	supdot: supdot$2,
  	supdsub: supdsub$2,
  	supE: supE$2,
  	supe: supe$2,
  	supedot: supedot$2,
  	Superset: Superset$2,
  	SupersetEqual: SupersetEqual$2,
  	suphsol: suphsol$2,
  	suphsub: suphsub$2,
  	suplarr: suplarr$2,
  	supmult: supmult$2,
  	supnE: supnE$2,
  	supne: supne$2,
  	supplus: supplus$2,
  	supset: supset$2,
  	Supset: Supset$2,
  	supseteq: supseteq$2,
  	supseteqq: supseteqq$2,
  	supsetneq: supsetneq$2,
  	supsetneqq: supsetneqq$2,
  	supsim: supsim$2,
  	supsub: supsub$2,
  	supsup: supsup$2,
  	swarhk: swarhk$2,
  	swarr: swarr$2,
  	swArr: swArr$2,
  	swarrow: swarrow$2,
  	swnwar: swnwar$2,
  	szlig: szlig$3,
  	Tab: Tab$2,
  	target: target$2,
  	Tau: Tau$2,
  	tau: tau$2,
  	tbrk: tbrk$2,
  	Tcaron: Tcaron$2,
  	tcaron: tcaron$2,
  	Tcedil: Tcedil$2,
  	tcedil: tcedil$2,
  	Tcy: Tcy$2,
  	tcy: tcy$2,
  	tdot: tdot$2,
  	telrec: telrec$2,
  	Tfr: Tfr$2,
  	tfr: tfr$2,
  	there4: there4$2,
  	therefore: therefore$2,
  	Therefore: Therefore$2,
  	Theta: Theta$2,
  	theta: theta$2,
  	thetasym: thetasym$2,
  	thetav: thetav$2,
  	thickapprox: thickapprox$2,
  	thicksim: thicksim$2,
  	ThickSpace: ThickSpace$2,
  	ThinSpace: ThinSpace$2,
  	thinsp: thinsp$2,
  	thkap: thkap$2,
  	thksim: thksim$2,
  	THORN: THORN$3,
  	thorn: thorn$3,
  	tilde: tilde$2,
  	Tilde: Tilde$2,
  	TildeEqual: TildeEqual$2,
  	TildeFullEqual: TildeFullEqual$2,
  	TildeTilde: TildeTilde$2,
  	timesbar: timesbar$2,
  	timesb: timesb$2,
  	times: times$4,
  	timesd: timesd$2,
  	tint: tint$2,
  	toea: toea$2,
  	topbot: topbot$2,
  	topcir: topcir$2,
  	top: top$2,
  	Topf: Topf$2,
  	topf: topf$2,
  	topfork: topfork$2,
  	tosa: tosa$2,
  	tprime: tprime$2,
  	trade: trade$2,
  	TRADE: TRADE$2,
  	triangle: triangle$2,
  	triangledown: triangledown$2,
  	triangleleft: triangleleft$2,
  	trianglelefteq: trianglelefteq$2,
  	triangleq: triangleq$2,
  	triangleright: triangleright$2,
  	trianglerighteq: trianglerighteq$2,
  	tridot: tridot$2,
  	trie: trie$2,
  	triminus: triminus$2,
  	TripleDot: TripleDot$2,
  	triplus: triplus$2,
  	trisb: trisb$2,
  	tritime: tritime$2,
  	trpezium: trpezium$2,
  	Tscr: Tscr$2,
  	tscr: tscr$2,
  	TScy: TScy$2,
  	tscy: tscy$2,
  	TSHcy: TSHcy$2,
  	tshcy: tshcy$2,
  	Tstrok: Tstrok$2,
  	tstrok: tstrok$2,
  	twixt: twixt$2,
  	twoheadleftarrow: twoheadleftarrow$2,
  	twoheadrightarrow: twoheadrightarrow$2,
  	Uacute: Uacute$3,
  	uacute: uacute$3,
  	uarr: uarr$2,
  	Uarr: Uarr$2,
  	uArr: uArr$2,
  	Uarrocir: Uarrocir$2,
  	Ubrcy: Ubrcy$2,
  	ubrcy: ubrcy$2,
  	Ubreve: Ubreve$2,
  	ubreve: ubreve$2,
  	Ucirc: Ucirc$3,
  	ucirc: ucirc$3,
  	Ucy: Ucy$2,
  	ucy: ucy$2,
  	udarr: udarr$2,
  	Udblac: Udblac$2,
  	udblac: udblac$2,
  	udhar: udhar$2,
  	ufisht: ufisht$2,
  	Ufr: Ufr$2,
  	ufr: ufr$2,
  	Ugrave: Ugrave$3,
  	ugrave: ugrave$3,
  	uHar: uHar$2,
  	uharl: uharl$2,
  	uharr: uharr$2,
  	uhblk: uhblk$2,
  	ulcorn: ulcorn$2,
  	ulcorner: ulcorner$2,
  	ulcrop: ulcrop$2,
  	ultri: ultri$2,
  	Umacr: Umacr$2,
  	umacr: umacr$2,
  	uml: uml$3,
  	UnderBar: UnderBar$2,
  	UnderBrace: UnderBrace$2,
  	UnderBracket: UnderBracket$2,
  	UnderParenthesis: UnderParenthesis$2,
  	Union: Union$2,
  	UnionPlus: UnionPlus$2,
  	Uogon: Uogon$2,
  	uogon: uogon$2,
  	Uopf: Uopf$2,
  	uopf: uopf$2,
  	UpArrowBar: UpArrowBar$2,
  	uparrow: uparrow$2,
  	UpArrow: UpArrow$2,
  	Uparrow: Uparrow$2,
  	UpArrowDownArrow: UpArrowDownArrow$2,
  	updownarrow: updownarrow$2,
  	UpDownArrow: UpDownArrow$2,
  	Updownarrow: Updownarrow$2,
  	UpEquilibrium: UpEquilibrium$2,
  	upharpoonleft: upharpoonleft$2,
  	upharpoonright: upharpoonright$2,
  	uplus: uplus$2,
  	UpperLeftArrow: UpperLeftArrow$2,
  	UpperRightArrow: UpperRightArrow$2,
  	upsi: upsi$2,
  	Upsi: Upsi$2,
  	upsih: upsih$2,
  	Upsilon: Upsilon$2,
  	upsilon: upsilon$2,
  	UpTeeArrow: UpTeeArrow$2,
  	UpTee: UpTee$2,
  	upuparrows: upuparrows$2,
  	urcorn: urcorn$2,
  	urcorner: urcorner$2,
  	urcrop: urcrop$2,
  	Uring: Uring$2,
  	uring: uring$2,
  	urtri: urtri$2,
  	Uscr: Uscr$2,
  	uscr: uscr$2,
  	utdot: utdot$2,
  	Utilde: Utilde$2,
  	utilde: utilde$2,
  	utri: utri$2,
  	utrif: utrif$2,
  	uuarr: uuarr$2,
  	Uuml: Uuml$3,
  	uuml: uuml$3,
  	uwangle: uwangle$2,
  	vangrt: vangrt$2,
  	varepsilon: varepsilon$2,
  	varkappa: varkappa$2,
  	varnothing: varnothing$2,
  	varphi: varphi$2,
  	varpi: varpi$2,
  	varpropto: varpropto$2,
  	varr: varr$2,
  	vArr: vArr$2,
  	varrho: varrho$2,
  	varsigma: varsigma$2,
  	varsubsetneq: varsubsetneq$2,
  	varsubsetneqq: varsubsetneqq$2,
  	varsupsetneq: varsupsetneq$2,
  	varsupsetneqq: varsupsetneqq$2,
  	vartheta: vartheta$2,
  	vartriangleleft: vartriangleleft$2,
  	vartriangleright: vartriangleright$2,
  	vBar: vBar$2,
  	Vbar: Vbar$2,
  	vBarv: vBarv$2,
  	Vcy: Vcy$2,
  	vcy: vcy$2,
  	vdash: vdash$2,
  	vDash: vDash$2,
  	Vdash: Vdash$2,
  	VDash: VDash$2,
  	Vdashl: Vdashl$2,
  	veebar: veebar$2,
  	vee: vee$2,
  	Vee: Vee$2,
  	veeeq: veeeq$2,
  	vellip: vellip$2,
  	verbar: verbar$2,
  	Verbar: Verbar$2,
  	vert: vert$2,
  	Vert: Vert$2,
  	VerticalBar: VerticalBar$2,
  	VerticalLine: VerticalLine$2,
  	VerticalSeparator: VerticalSeparator$2,
  	VerticalTilde: VerticalTilde$2,
  	VeryThinSpace: VeryThinSpace$2,
  	Vfr: Vfr$2,
  	vfr: vfr$2,
  	vltri: vltri$2,
  	vnsub: vnsub$2,
  	vnsup: vnsup$2,
  	Vopf: Vopf$2,
  	vopf: vopf$2,
  	vprop: vprop$2,
  	vrtri: vrtri$2,
  	Vscr: Vscr$2,
  	vscr: vscr$2,
  	vsubnE: vsubnE$2,
  	vsubne: vsubne$2,
  	vsupnE: vsupnE$2,
  	vsupne: vsupne$2,
  	Vvdash: Vvdash$2,
  	vzigzag: vzigzag$2,
  	Wcirc: Wcirc$2,
  	wcirc: wcirc$2,
  	wedbar: wedbar$2,
  	wedge: wedge$2,
  	Wedge: Wedge$2,
  	wedgeq: wedgeq$2,
  	weierp: weierp$2,
  	Wfr: Wfr$2,
  	wfr: wfr$2,
  	Wopf: Wopf$2,
  	wopf: wopf$2,
  	wp: wp$2,
  	wr: wr$2,
  	wreath: wreath$2,
  	Wscr: Wscr$2,
  	wscr: wscr$2,
  	xcap: xcap$2,
  	xcirc: xcirc$2,
  	xcup: xcup$2,
  	xdtri: xdtri$2,
  	Xfr: Xfr$2,
  	xfr: xfr$2,
  	xharr: xharr$2,
  	xhArr: xhArr$2,
  	Xi: Xi$2,
  	xi: xi$2,
  	xlarr: xlarr$2,
  	xlArr: xlArr$2,
  	xmap: xmap$2,
  	xnis: xnis$2,
  	xodot: xodot$2,
  	Xopf: Xopf$2,
  	xopf: xopf$2,
  	xoplus: xoplus$2,
  	xotime: xotime$2,
  	xrarr: xrarr$2,
  	xrArr: xrArr$2,
  	Xscr: Xscr$2,
  	xscr: xscr$2,
  	xsqcup: xsqcup$2,
  	xuplus: xuplus$2,
  	xutri: xutri$2,
  	xvee: xvee$2,
  	xwedge: xwedge$2,
  	Yacute: Yacute$3,
  	yacute: yacute$3,
  	YAcy: YAcy$2,
  	yacy: yacy$2,
  	Ycirc: Ycirc$2,
  	ycirc: ycirc$2,
  	Ycy: Ycy$2,
  	ycy: ycy$2,
  	yen: yen$3,
  	Yfr: Yfr$2,
  	yfr: yfr$2,
  	YIcy: YIcy$2,
  	yicy: yicy$2,
  	Yopf: Yopf$2,
  	yopf: yopf$2,
  	Yscr: Yscr$2,
  	yscr: yscr$2,
  	YUcy: YUcy$2,
  	yucy: yucy$2,
  	yuml: yuml$3,
  	Yuml: Yuml$2,
  	Zacute: Zacute$2,
  	zacute: zacute$2,
  	Zcaron: Zcaron$2,
  	zcaron: zcaron$2,
  	Zcy: Zcy$2,
  	zcy: zcy$2,
  	Zdot: Zdot$2,
  	zdot: zdot$2,
  	zeetrf: zeetrf$2,
  	ZeroWidthSpace: ZeroWidthSpace$2,
  	Zeta: Zeta$2,
  	zeta: zeta$2,
  	zfr: zfr$2,
  	Zfr: Zfr$2,
  	ZHcy: ZHcy$2,
  	zhcy: zhcy$2,
  	zigrarr: zigrarr$2,
  	zopf: zopf$2,
  	Zopf: Zopf$2,
  	Zscr: Zscr$2,
  	zscr: zscr$2,
  	zwj: zwj$2,
  	zwnj: zwnj$2,
  	default: entitiesJSON$2
  });

  var require$$0$2 = ( xml$2 && xmlJSON$2 ) || xml$2;

  var require$$1$3 = ( entities$2 && entitiesJSON$2 ) || entities$2;

  var inverseXML$2 = getInverseObj$1$2(require$$0$2);
  var xmlReplacer$2 = getInverseReplacer$1$2(inverseXML$2);

  var XML$2 = getInverse$1$2(inverseXML$2, xmlReplacer$2);

  var inverseHTML$2 = getInverseObj$1$2(require$$1$3);
  var htmlReplacer$2 = getInverseReplacer$1$2(inverseHTML$2);

  var HTML$2 = getInverse$1$2(inverseHTML$2, htmlReplacer$2);

  function getInverseObj$1$2(obj){
  	return Object.keys(obj).sort().reduce(function(inverse, name){
  		inverse[obj[name]] = "&" + name + ";";
  		return inverse;
  	}, {});
  }

  function getInverseReplacer$1$2(inverse){
  	var single = [],
  	    multiple = [];

  	Object.keys(inverse).forEach(function(k){
  		if(k.length === 1){
  			single.push("\\" + k);
  		} else {
  			multiple.push(k);
  		}
  	});

  	
  	multiple.unshift("[" + single.join("") + "]");

  	return new RegExp(multiple.join("|"), "g");
  }

  var re_nonASCII$2 = /[^\0-\x7F]/g;
  var re_astralSymbols$2 = /[\uD800-\uDBFF][\uDC00-\uDFFF]/g;

  function singleCharReplacer$1$2(c){
  	return "&#x" + c.charCodeAt(0).toString(16).toUpperCase() + ";";
  }

  function astralReplacer$1$2(c){
  	
  	var high = c.charCodeAt(0);
  	var low  = c.charCodeAt(1);
  	var codePoint = (high - 0xD800) * 0x400 + low - 0xDC00 + 0x10000;
  	return "&#x" + codePoint.toString(16).toUpperCase() + ";";
  }

  function getInverse$1$2(inverse, re){
  	function func(name){
  		return inverse[name];
  	}

  	return function(data){
  		return data
  				.replace(re, func)
  				.replace(re_astralSymbols$2, astralReplacer$1$2)
  				.replace(re_nonASCII$2, singleCharReplacer$1$2);
  	};
  }

  var re_xmlChars$2 = getInverseReplacer$1$2(inverseXML$2);

  function escapeXML$2(data){
  	return data
  			.replace(re_xmlChars$2, singleCharReplacer$1$2)
  			.replace(re_astralSymbols$2, astralReplacer$1$2)
  			.replace(re_nonASCII$2, singleCharReplacer$1$2);
  }

  var escape$2 = escapeXML$2;

  var encode$2 = {
  	XML: XML$2,
  	HTML: HTML$2,
  	escape: escape$2
  };

  var decode$3 = {
  	"0": 65533,
  	"128": 8364,
  	"130": 8218,
  	"131": 402,
  	"132": 8222,
  	"133": 8230,
  	"134": 8224,
  	"135": 8225,
  	"136": 710,
  	"137": 8240,
  	"138": 352,
  	"139": 8249,
  	"140": 338,
  	"142": 381,
  	"145": 8216,
  	"146": 8217,
  	"147": 8220,
  	"148": 8221,
  	"149": 8226,
  	"150": 8211,
  	"151": 8212,
  	"152": 732,
  	"153": 8482,
  	"154": 353,
  	"155": 8250,
  	"156": 339,
  	"158": 382,
  	"159": 376
  };

  var decode$1$2 = Object.freeze({
  	default: decode$3
  });

  var decodeMap$2 = ( decode$1$2 && decode$3 ) || decode$1$2;

  var decode_codepoint$2 = decodeCodePoint$3;


  function decodeCodePoint$3(codePoint){

  	if((codePoint >= 0xD800 && codePoint <= 0xDFFF) || codePoint > 0x10FFFF){
  		return "\uFFFD";
  	}

  	if(codePoint in decodeMap$2){
  		codePoint = decodeMap$2[codePoint];
  	}

  	var output = "";

  	if(codePoint > 0xFFFF){
  		codePoint -= 0x10000;
  		output += String.fromCharCode(codePoint >>> 10 & 0x3FF | 0xD800);
  		codePoint = 0xDC00 | codePoint & 0x3FF;
  	}

  	output += String.fromCharCode(codePoint);
  	return output;
  }

  var Aacute$1$2 = "Á";
  var aacute$1$2 = "á";
  var Acirc$1$2 = "Â";
  var acirc$1$2 = "â";
  var acute$1$2 = "´";
  var AElig$1$2 = "Æ";
  var aelig$1$2 = "æ";
  var Agrave$1$2 = "À";
  var agrave$1$2 = "à";
  var amp$2$2 = "&";
  var AMP$1$2 = "&";
  var Aring$1$2 = "Å";
  var aring$1$2 = "å";
  var Atilde$1$2 = "Ã";
  var atilde$1$2 = "ã";
  var Auml$1$2 = "Ä";
  var auml$1$2 = "ä";
  var brvbar$1$2 = "¦";
  var Ccedil$1$2 = "Ç";
  var ccedil$1$2 = "ç";
  var cedil$1$2 = "¸";
  var cent$1$2 = "¢";
  var copy$1$2 = "©";
  var COPY$1$2 = "©";
  var curren$1$2 = "¤";
  var deg$1$2 = "°";
  var divide$1$2 = "÷";
  var Eacute$1$2 = "É";
  var eacute$1$2 = "é";
  var Ecirc$1$2 = "Ê";
  var ecirc$1$2 = "ê";
  var Egrave$1$2 = "È";
  var egrave$1$2 = "è";
  var ETH$1$2 = "Ð";
  var eth$1$2 = "ð";
  var Euml$1$2 = "Ë";
  var euml$1$2 = "ë";
  var frac12$1$2 = "½";
  var frac14$1$2 = "¼";
  var frac34$1$2 = "¾";
  var gt$2$2 = ">";
  var GT$1$2 = ">";
  var Iacute$1$2 = "Í";
  var iacute$1$2 = "í";
  var Icirc$1$2 = "Î";
  var icirc$1$2 = "î";
  var iexcl$1$2 = "¡";
  var Igrave$1$2 = "Ì";
  var igrave$1$2 = "ì";
  var iquest$1$2 = "¿";
  var Iuml$1$2 = "Ï";
  var iuml$1$2 = "ï";
  var laquo$1$2 = "«";
  var lt$2$2 = "<";
  var LT$1$2 = "<";
  var macr$1$2 = "¯";
  var micro$1$2 = "µ";
  var middot$1$2 = "·";
  var nbsp$1$2 = " ";
  var not$1$2 = "¬";
  var Ntilde$1$2 = "Ñ";
  var ntilde$1$2 = "ñ";
  var Oacute$1$2 = "Ó";
  var oacute$1$2 = "ó";
  var Ocirc$1$2 = "Ô";
  var ocirc$1$2 = "ô";
  var Ograve$1$2 = "Ò";
  var ograve$1$2 = "ò";
  var ordf$1$2 = "ª";
  var ordm$1$2 = "º";
  var Oslash$1$2 = "Ø";
  var oslash$1$2 = "ø";
  var Otilde$1$2 = "Õ";
  var otilde$1$2 = "õ";
  var Ouml$1$2 = "Ö";
  var ouml$1$2 = "ö";
  var para$1$2 = "¶";
  var plusmn$1$2 = "±";
  var pound$1$2 = "£";
  var quot$2$2 = "\"";
  var QUOT$1$2 = "\"";
  var raquo$1$2 = "»";
  var reg$1$2 = "®";
  var REG$1$2 = "®";
  var sect$1$2 = "§";
  var shy$1$2 = "­";
  var sup1$1$2 = "¹";
  var sup2$1$2 = "²";
  var sup3$1$2 = "³";
  var szlig$1$2 = "ß";
  var THORN$1$2 = "Þ";
  var thorn$1$2 = "þ";
  var times$1$2 = "×";
  var Uacute$1$2 = "Ú";
  var uacute$1$2 = "ú";
  var Ucirc$1$2 = "Û";
  var ucirc$1$2 = "û";
  var Ugrave$1$2 = "Ù";
  var ugrave$1$2 = "ù";
  var uml$1$2 = "¨";
  var Uuml$1$2 = "Ü";
  var uuml$1$2 = "ü";
  var Yacute$1$2 = "Ý";
  var yacute$1$2 = "ý";
  var yen$1$2 = "¥";
  var yuml$1$2 = "ÿ";
  var legacyJSON$2 = {
  	Aacute: Aacute$1$2,
  	aacute: aacute$1$2,
  	Acirc: Acirc$1$2,
  	acirc: acirc$1$2,
  	acute: acute$1$2,
  	AElig: AElig$1$2,
  	aelig: aelig$1$2,
  	Agrave: Agrave$1$2,
  	agrave: agrave$1$2,
  	amp: amp$2$2,
  	AMP: AMP$1$2,
  	Aring: Aring$1$2,
  	aring: aring$1$2,
  	Atilde: Atilde$1$2,
  	atilde: atilde$1$2,
  	Auml: Auml$1$2,
  	auml: auml$1$2,
  	brvbar: brvbar$1$2,
  	Ccedil: Ccedil$1$2,
  	ccedil: ccedil$1$2,
  	cedil: cedil$1$2,
  	cent: cent$1$2,
  	copy: copy$1$2,
  	COPY: COPY$1$2,
  	curren: curren$1$2,
  	deg: deg$1$2,
  	divide: divide$1$2,
  	Eacute: Eacute$1$2,
  	eacute: eacute$1$2,
  	Ecirc: Ecirc$1$2,
  	ecirc: ecirc$1$2,
  	Egrave: Egrave$1$2,
  	egrave: egrave$1$2,
  	ETH: ETH$1$2,
  	eth: eth$1$2,
  	Euml: Euml$1$2,
  	euml: euml$1$2,
  	frac12: frac12$1$2,
  	frac14: frac14$1$2,
  	frac34: frac34$1$2,
  	gt: gt$2$2,
  	GT: GT$1$2,
  	Iacute: Iacute$1$2,
  	iacute: iacute$1$2,
  	Icirc: Icirc$1$2,
  	icirc: icirc$1$2,
  	iexcl: iexcl$1$2,
  	Igrave: Igrave$1$2,
  	igrave: igrave$1$2,
  	iquest: iquest$1$2,
  	Iuml: Iuml$1$2,
  	iuml: iuml$1$2,
  	laquo: laquo$1$2,
  	lt: lt$2$2,
  	LT: LT$1$2,
  	macr: macr$1$2,
  	micro: micro$1$2,
  	middot: middot$1$2,
  	nbsp: nbsp$1$2,
  	not: not$1$2,
  	Ntilde: Ntilde$1$2,
  	ntilde: ntilde$1$2,
  	Oacute: Oacute$1$2,
  	oacute: oacute$1$2,
  	Ocirc: Ocirc$1$2,
  	ocirc: ocirc$1$2,
  	Ograve: Ograve$1$2,
  	ograve: ograve$1$2,
  	ordf: ordf$1$2,
  	ordm: ordm$1$2,
  	Oslash: Oslash$1$2,
  	oslash: oslash$1$2,
  	Otilde: Otilde$1$2,
  	otilde: otilde$1$2,
  	Ouml: Ouml$1$2,
  	ouml: ouml$1$2,
  	para: para$1$2,
  	plusmn: plusmn$1$2,
  	pound: pound$1$2,
  	quot: quot$2$2,
  	QUOT: QUOT$1$2,
  	raquo: raquo$1$2,
  	reg: reg$1$2,
  	REG: REG$1$2,
  	sect: sect$1$2,
  	shy: shy$1$2,
  	sup1: sup1$1$2,
  	sup2: sup2$1$2,
  	sup3: sup3$1$2,
  	szlig: szlig$1$2,
  	THORN: THORN$1$2,
  	thorn: thorn$1$2,
  	times: times$1$2,
  	Uacute: Uacute$1$2,
  	uacute: uacute$1$2,
  	Ucirc: Ucirc$1$2,
  	ucirc: ucirc$1$2,
  	Ugrave: Ugrave$1$2,
  	ugrave: ugrave$1$2,
  	uml: uml$1$2,
  	Uuml: Uuml$1$2,
  	uuml: uuml$1$2,
  	Yacute: Yacute$1$2,
  	yacute: yacute$1$2,
  	yen: yen$1$2,
  	yuml: yuml$1$2
  };

  var _entities$2 = {
    encodeXML: encode$2.XML,
    decodeCodepoint: decode_codepoint$2,
    entitiesJSON: entitiesJSON$2,
    legacyJSON: legacyJSON$2,
    xmlJSON: xmlJSON$2
  };

  var _entities_decodeCodepoint$3 = _entities$2.decodeCodepoint;

  var _entities_decodeCodepoint$1$2 = Object.freeze({
  	default: _entities_decodeCodepoint$3
  });

  var _entities_entitiesJSON$3 = _entities$2.entitiesJSON;


  var _entities_entitiesJSON$1$2 = Object.freeze({
  	default: _entities_entitiesJSON$3
  });

  var _entities_legacyJSON$3 = _entities$2.legacyJSON;

  var _entities_legacyJSON$1$2 = Object.freeze({
  	default: _entities_legacyJSON$3
  });

  var _entities_xmlJSON$3 = _entities$2.xmlJSON;

  var _entities_xmlJSON$1$2 = Object.freeze({
  	default: _entities_xmlJSON$3
  });

  var decodeCodePoint$1$2 = ( _entities_decodeCodepoint$1$2 && _entities_decodeCodepoint$3 ) || _entities_decodeCodepoint$1$2;

  var entityMap$2 = ( _entities_entitiesJSON$1$2 && _entities_entitiesJSON$3 ) || _entities_entitiesJSON$1$2;

  var legacyMap$2 = ( _entities_legacyJSON$1$2 && _entities_legacyJSON$3 ) || _entities_legacyJSON$1$2;

  var xmlMap$2 = ( _entities_xmlJSON$1$2 && _entities_xmlJSON$3 ) || _entities_xmlJSON$1$2;

  var Tokenizer_1$2 = Tokenizer$1$2;

  var i$2 = 0;
  var TEXT$8                      = i$2++;
  var BEFORE_TAG_NAME$2           = i$2++;
  var IN_TAG_NAME$2               = i$2++;
  var IN_SELF_CLOSING_TAG$2       = i$2++;
  var BEFORE_CLOSING_TAG_NAME$2   = i$2++;
  var IN_CLOSING_TAG_NAME$2       = i$2++;
  var AFTER_CLOSING_TAG_NAME$2    = i$2++;
  var BEFORE_ATTRIBUTE_NAME$2     = i$2++;
  var IN_ATTRIBUTE_NAME$2         = i$2++;
  var AFTER_ATTRIBUTE_NAME$2      = i$2++;
  var BEFORE_ATTRIBUTE_VALUE$2    = i$2++;
  var IN_ATTRIBUTE_VALUE_DQ$2     = i$2++;
  var IN_ATTRIBUTE_VALUE_SQ$2     = i$2++;
  var IN_ATTRIBUTE_VALUE_NQ$2     = i$2++;
  var BEFORE_DECLARATION$2        = i$2++;
  var IN_DECLARATION$2            = i$2++;
  var IN_PROCESSING_INSTRUCTION$2 = i$2++;
  var BEFORE_COMMENT$2            = i$2++;
  var IN_COMMENT$2                = i$2++;
  var AFTER_COMMENT_1$2           = i$2++;
  var AFTER_COMMENT_2$2           = i$2++;
  var BEFORE_CDATA_1$2            = i$2++;
  var BEFORE_CDATA_2$2            = i$2++;
  var BEFORE_CDATA_3$2            = i$2++;
  var BEFORE_CDATA_4$2            = i$2++;
  var BEFORE_CDATA_5$2            = i$2++;
  var BEFORE_CDATA_6$2            = i$2++;
  var IN_CDATA$2                  = i$2++;
  var AFTER_CDATA_1$2             = i$2++;
  var AFTER_CDATA_2$2             = i$2++;
  var BEFORE_SPECIAL$2            = i$2++;
  var BEFORE_SPECIAL_END$2        = i$2++;
  var BEFORE_SCRIPT_1$2           = i$2++;
  var BEFORE_SCRIPT_2$2           = i$2++;
  var BEFORE_SCRIPT_3$2           = i$2++;
  var BEFORE_SCRIPT_4$2           = i$2++;
  var BEFORE_SCRIPT_5$2           = i$2++;
  var AFTER_SCRIPT_1$2            = i$2++;
  var AFTER_SCRIPT_2$2            = i$2++;
  var AFTER_SCRIPT_3$2            = i$2++;
  var AFTER_SCRIPT_4$2            = i$2++;
  var AFTER_SCRIPT_5$2            = i$2++;
  var BEFORE_STYLE_1$2            = i$2++;
  var BEFORE_STYLE_2$2            = i$2++;
  var BEFORE_STYLE_3$2            = i$2++;
  var BEFORE_STYLE_4$2            = i$2++;
  var AFTER_STYLE_1$2             = i$2++;
  var AFTER_STYLE_2$2             = i$2++;
  var AFTER_STYLE_3$2             = i$2++;
  var AFTER_STYLE_4$2             = i$2++;
  var BEFORE_ENTITY$2             = i$2++;
  var BEFORE_NUMERIC_ENTITY$2     = i$2++;
  var IN_NAMED_ENTITY$2           = i$2++;
  var IN_NUMERIC_ENTITY$2         = i$2++;
  var IN_HEX_ENTITY$2             = i$2++;
  var j$2 = 0;
  var SPECIAL_NONE$2              = j$2++;
  var SPECIAL_SCRIPT$2            = j$2++;
  var SPECIAL_STYLE$2             = j$2++;

  function whitespace$2(c){
  	return c === " " || c === "\n" || c === "\t" || c === "\f" || c === "\r";
  }

  function characterState$2(char, SUCCESS){
  	return function(c){
  		if(c === char) this._state = SUCCESS;
  	};
  }

  function ifElseState$2(upper, SUCCESS, FAILURE){
  	var lower = upper.toLowerCase();

  	if(upper === lower){
  		return function(c){
  			if(c === lower){
  				this._state = SUCCESS;
  			} else {
  				this._state = FAILURE;
  				this._index--;
  			}
  		};
  	} else {
  		return function(c){
  			if(c === lower || c === upper){
  				this._state = SUCCESS;
  			} else {
  				this._state = FAILURE;
  				this._index--;
  			}
  		};
  	}
  }

  function consumeSpecialNameChar$2(upper, NEXT_STATE){
  	var lower = upper.toLowerCase();

  	return function(c){
  		if(c === lower || c === upper){
  			this._state = NEXT_STATE;
  		} else {
  			this._state = IN_TAG_NAME$2;
  			this._index--; 
  		}
  	};
  }

  function Tokenizer$1$2(options, cbs){
  	this._state = TEXT$8;
  	this._buffer = "";
  	this._sectionStart = 0;
  	this._index = 0;
  	this._bufferOffset = 0; 
  	this._baseState = TEXT$8;
  	this._special = SPECIAL_NONE$2;
  	this._cbs = cbs;
  	this._running = true;
  	this._ended = false;
  	this._xmlMode = !!(options && options.xmlMode);
  	this._decodeEntities = !!(options && options.decodeEntities);
  }

  Tokenizer$1$2.prototype._stateText = function(c){
  	if(c === "<"){
  		if(this._index > this._sectionStart){
  			this._cbs.ontext(this._getSection());
  		}
  		this._state = BEFORE_TAG_NAME$2;
  		this._sectionStart = this._index;
  	} else if(this._decodeEntities && this._special === SPECIAL_NONE$2 && c === "&"){
  		if(this._index > this._sectionStart){
  			this._cbs.ontext(this._getSection());
  		}
  		this._baseState = TEXT$8;
  		this._state = BEFORE_ENTITY$2;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1$2.prototype._stateBeforeTagName = function(c){
  	if(c === "/"){
  		this._state = BEFORE_CLOSING_TAG_NAME$2;
  	} else if(c === "<"){
  		this._cbs.ontext(this._getSection());
  		this._sectionStart = this._index;
  	} else if(c === ">" || this._special !== SPECIAL_NONE$2 || whitespace$2(c)) {
  		this._state = TEXT$8;
  	} else if(c === "!"){
  		this._state = BEFORE_DECLARATION$2;
  		this._sectionStart = this._index + 1;
  	} else if(c === "?"){
  		this._state = IN_PROCESSING_INSTRUCTION$2;
  		this._sectionStart = this._index + 1;
  	} else {
  		this._state = (!this._xmlMode && (c === "s" || c === "S")) ?
  						BEFORE_SPECIAL$2 : IN_TAG_NAME$2;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1$2.prototype._stateInTagName = function(c){
  	if(c === "/" || c === ">" || whitespace$2(c)){
  		this._emitToken("onopentagname");
  		this._state = BEFORE_ATTRIBUTE_NAME$2;
  		this._index--;
  	}
  };

  Tokenizer$1$2.prototype._stateBeforeCloseingTagName = function(c){
  	if(whitespace$2(c));
  	else if(c === ">"){
  		this._state = TEXT$8;
  	} else if(this._special !== SPECIAL_NONE$2){
  		if(c === "s" || c === "S"){
  			this._state = BEFORE_SPECIAL_END$2;
  		} else {
  			this._state = TEXT$8;
  			this._index--;
  		}
  	} else {
  		this._state = IN_CLOSING_TAG_NAME$2;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1$2.prototype._stateInCloseingTagName = function(c){
  	if(c === ">" || whitespace$2(c)){
  		this._emitToken("onclosetag");
  		this._state = AFTER_CLOSING_TAG_NAME$2;
  		this._index--;
  	}
  };

  Tokenizer$1$2.prototype._stateAfterCloseingTagName = function(c){
  	
  	if(c === ">"){
  		this._state = TEXT$8;
  		this._sectionStart = this._index + 1;
  	}
  };

  Tokenizer$1$2.prototype._stateBeforeAttributeName = function(c){
  	if(c === ">"){
  		this._cbs.onopentagend();
  		this._state = TEXT$8;
  		this._sectionStart = this._index + 1;
  	} else if(c === "/"){
  		this._state = IN_SELF_CLOSING_TAG$2;
  	} else if(!whitespace$2(c)){
  		this._state = IN_ATTRIBUTE_NAME$2;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1$2.prototype._stateInSelfClosingTag = function(c){
  	if(c === ">"){
  		this._cbs.onselfclosingtag();
  		this._state = TEXT$8;
  		this._sectionStart = this._index + 1;
  	} else if(!whitespace$2(c)){
  		this._state = BEFORE_ATTRIBUTE_NAME$2;
  		this._index--;
  	}
  };

  Tokenizer$1$2.prototype._stateInAttributeName = function(c){
  	if(c === "=" || c === "/" || c === ">" || whitespace$2(c)){
  		this._cbs.onattribname(this._getSection());
  		this._sectionStart = -1;
  		this._state = AFTER_ATTRIBUTE_NAME$2;
  		this._index--;
  	}
  };

  Tokenizer$1$2.prototype._stateAfterAttributeName = function(c){
  	if(c === "="){
  		this._state = BEFORE_ATTRIBUTE_VALUE$2;
  	} else if(c === "/" || c === ">"){
  		this._cbs.onattribend();
  		this._state = BEFORE_ATTRIBUTE_NAME$2;
  		this._index--;
  	} else if(!whitespace$2(c)){
  		this._cbs.onattribend();
  		this._state = IN_ATTRIBUTE_NAME$2;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1$2.prototype._stateBeforeAttributeValue = function(c){
  	if(c === "\""){
  		this._state = IN_ATTRIBUTE_VALUE_DQ$2;
  		this._sectionStart = this._index + 1;
  	} else if(c === "'"){
  		this._state = IN_ATTRIBUTE_VALUE_SQ$2;
  		this._sectionStart = this._index + 1;
  	} else if(!whitespace$2(c)){
  		this._state = IN_ATTRIBUTE_VALUE_NQ$2;
  		this._sectionStart = this._index;
  		this._index--; 
  	}
  };

  Tokenizer$1$2.prototype._stateInAttributeValueDoubleQuotes = function(c){
  	if(c === "\""){
  		this._emitToken("onattribdata");
  		this._cbs.onattribend();
  		this._state = BEFORE_ATTRIBUTE_NAME$2;
  	} else if(this._decodeEntities && c === "&"){
  		this._emitToken("onattribdata");
  		this._baseState = this._state;
  		this._state = BEFORE_ENTITY$2;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1$2.prototype._stateInAttributeValueSingleQuotes = function(c){
  	if(c === "'"){
  		this._emitToken("onattribdata");
  		this._cbs.onattribend();
  		this._state = BEFORE_ATTRIBUTE_NAME$2;
  	} else if(this._decodeEntities && c === "&"){
  		this._emitToken("onattribdata");
  		this._baseState = this._state;
  		this._state = BEFORE_ENTITY$2;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1$2.prototype._stateInAttributeValueNoQuotes = function(c){
  	if(whitespace$2(c) || c === ">"){
  		this._emitToken("onattribdata");
  		this._cbs.onattribend();
  		this._state = BEFORE_ATTRIBUTE_NAME$2;
  		this._index--;
  	} else if(this._decodeEntities && c === "&"){
  		this._emitToken("onattribdata");
  		this._baseState = this._state;
  		this._state = BEFORE_ENTITY$2;
  		this._sectionStart = this._index;
  	}
  };

  Tokenizer$1$2.prototype._stateBeforeDeclaration = function(c){
  	this._state = c === "[" ? BEFORE_CDATA_1$2 :
  					c === "-" ? BEFORE_COMMENT$2 :
  						IN_DECLARATION$2;
  };

  Tokenizer$1$2.prototype._stateInDeclaration = function(c){
  	if(c === ">"){
  		this._cbs.ondeclaration(this._getSection());
  		this._state = TEXT$8;
  		this._sectionStart = this._index + 1;
  	}
  };

  Tokenizer$1$2.prototype._stateInProcessingInstruction = function(c){
  	if(c === ">"){
  		this._cbs.onprocessinginstruction(this._getSection());
  		this._state = TEXT$8;
  		this._sectionStart = this._index + 1;
  	}
  };

  Tokenizer$1$2.prototype._stateBeforeComment = function(c){
  	if(c === "-"){
  		this._state = IN_COMMENT$2;
  		this._sectionStart = this._index + 1;
  	} else {
  		this._state = IN_DECLARATION$2;
  	}
  };

  Tokenizer$1$2.prototype._stateInComment = function(c){
  	if(c === "-") this._state = AFTER_COMMENT_1$2;
  };

  Tokenizer$1$2.prototype._stateAfterComment1 = function(c){
  	if(c === "-"){
  		this._state = AFTER_COMMENT_2$2;
  	} else {
  		this._state = IN_COMMENT$2;
  	}
  };

  Tokenizer$1$2.prototype._stateAfterComment2 = function(c){
  	if(c === ">"){
  		
  		this._cbs.oncomment(this._buffer.substring(this._sectionStart, this._index - 2));
  		this._state = TEXT$8;
  		this._sectionStart = this._index + 1;
  	} else if(c !== "-"){
  		this._state = IN_COMMENT$2;
  	}
  	
  };

  Tokenizer$1$2.prototype._stateBeforeCdata1 = ifElseState$2("C", BEFORE_CDATA_2$2, IN_DECLARATION$2);
  Tokenizer$1$2.prototype._stateBeforeCdata2 = ifElseState$2("D", BEFORE_CDATA_3$2, IN_DECLARATION$2);
  Tokenizer$1$2.prototype._stateBeforeCdata3 = ifElseState$2("A", BEFORE_CDATA_4$2, IN_DECLARATION$2);
  Tokenizer$1$2.prototype._stateBeforeCdata4 = ifElseState$2("T", BEFORE_CDATA_5$2, IN_DECLARATION$2);
  Tokenizer$1$2.prototype._stateBeforeCdata5 = ifElseState$2("A", BEFORE_CDATA_6$2, IN_DECLARATION$2);

  Tokenizer$1$2.prototype._stateBeforeCdata6 = function(c){
  	if(c === "["){
  		this._state = IN_CDATA$2;
  		this._sectionStart = this._index + 1;
  	} else {
  		this._state = IN_DECLARATION$2;
  		this._index--;
  	}
  };

  Tokenizer$1$2.prototype._stateInCdata = function(c){
  	if(c === "]") this._state = AFTER_CDATA_1$2;
  };

  Tokenizer$1$2.prototype._stateAfterCdata1 = characterState$2("]", AFTER_CDATA_2$2);

  Tokenizer$1$2.prototype._stateAfterCdata2 = function(c){
  	if(c === ">"){
  		
  		this._cbs.oncdata(this._buffer.substring(this._sectionStart, this._index - 2));
  		this._state = TEXT$8;
  		this._sectionStart = this._index + 1;
  	} else if(c !== "]") {
  		this._state = IN_CDATA$2;
  	}
  	
  };

  Tokenizer$1$2.prototype._stateBeforeSpecial = function(c){
  	if(c === "c" || c === "C"){
  		this._state = BEFORE_SCRIPT_1$2;
  	} else if(c === "t" || c === "T"){
  		this._state = BEFORE_STYLE_1$2;
  	} else {
  		this._state = IN_TAG_NAME$2;
  		this._index--; 
  	}
  };

  Tokenizer$1$2.prototype._stateBeforeSpecialEnd = function(c){
  	if(this._special === SPECIAL_SCRIPT$2 && (c === "c" || c === "C")){
  		this._state = AFTER_SCRIPT_1$2;
  	} else if(this._special === SPECIAL_STYLE$2 && (c === "t" || c === "T")){
  		this._state = AFTER_STYLE_1$2;
  	}
  	else this._state = TEXT$8;
  };

  Tokenizer$1$2.prototype._stateBeforeScript1 = consumeSpecialNameChar$2("R", BEFORE_SCRIPT_2$2);
  Tokenizer$1$2.prototype._stateBeforeScript2 = consumeSpecialNameChar$2("I", BEFORE_SCRIPT_3$2);
  Tokenizer$1$2.prototype._stateBeforeScript3 = consumeSpecialNameChar$2("P", BEFORE_SCRIPT_4$2);
  Tokenizer$1$2.prototype._stateBeforeScript4 = consumeSpecialNameChar$2("T", BEFORE_SCRIPT_5$2);

  Tokenizer$1$2.prototype._stateBeforeScript5 = function(c){
  	if(c === "/" || c === ">" || whitespace$2(c)){
  		this._special = SPECIAL_SCRIPT$2;
  	}
  	this._state = IN_TAG_NAME$2;
  	this._index--; 
  };

  Tokenizer$1$2.prototype._stateAfterScript1 = ifElseState$2("R", AFTER_SCRIPT_2$2, TEXT$8);
  Tokenizer$1$2.prototype._stateAfterScript2 = ifElseState$2("I", AFTER_SCRIPT_3$2, TEXT$8);
  Tokenizer$1$2.prototype._stateAfterScript3 = ifElseState$2("P", AFTER_SCRIPT_4$2, TEXT$8);
  Tokenizer$1$2.prototype._stateAfterScript4 = ifElseState$2("T", AFTER_SCRIPT_5$2, TEXT$8);

  Tokenizer$1$2.prototype._stateAfterScript5 = function(c){
  	if(c === ">" || whitespace$2(c)){
  		this._special = SPECIAL_NONE$2;
  		this._state = IN_CLOSING_TAG_NAME$2;
  		this._sectionStart = this._index - 6;
  		this._index--; 
  	}
  	else this._state = TEXT$8;
  };

  Tokenizer$1$2.prototype._stateBeforeStyle1 = consumeSpecialNameChar$2("Y", BEFORE_STYLE_2$2);
  Tokenizer$1$2.prototype._stateBeforeStyle2 = consumeSpecialNameChar$2("L", BEFORE_STYLE_3$2);
  Tokenizer$1$2.prototype._stateBeforeStyle3 = consumeSpecialNameChar$2("E", BEFORE_STYLE_4$2);

  Tokenizer$1$2.prototype._stateBeforeStyle4 = function(c){
  	if(c === "/" || c === ">" || whitespace$2(c)){
  		this._special = SPECIAL_STYLE$2;
  	}
  	this._state = IN_TAG_NAME$2;
  	this._index--; 
  };

  Tokenizer$1$2.prototype._stateAfterStyle1 = ifElseState$2("Y", AFTER_STYLE_2$2, TEXT$8);
  Tokenizer$1$2.prototype._stateAfterStyle2 = ifElseState$2("L", AFTER_STYLE_3$2, TEXT$8);
  Tokenizer$1$2.prototype._stateAfterStyle3 = ifElseState$2("E", AFTER_STYLE_4$2, TEXT$8);

  Tokenizer$1$2.prototype._stateAfterStyle4 = function(c){
  	if(c === ">" || whitespace$2(c)){
  		this._special = SPECIAL_NONE$2;
  		this._state = IN_CLOSING_TAG_NAME$2;
  		this._sectionStart = this._index - 5;
  		this._index--; 
  	}
  	else this._state = TEXT$8;
  };

  Tokenizer$1$2.prototype._stateBeforeEntity = ifElseState$2("#", BEFORE_NUMERIC_ENTITY$2, IN_NAMED_ENTITY$2);
  Tokenizer$1$2.prototype._stateBeforeNumericEntity = ifElseState$2("X", IN_HEX_ENTITY$2, IN_NUMERIC_ENTITY$2);


  Tokenizer$1$2.prototype._parseNamedEntityStrict = function(){
  	
  	if(this._sectionStart + 1 < this._index){
  		var entity = this._buffer.substring(this._sectionStart + 1, this._index),
  		    map = this._xmlMode ? xmlMap$2 : entityMap$2;

  		if(map.hasOwnProperty(entity)){
  			this._emitPartial(map[entity]);
  			this._sectionStart = this._index + 1;
  		}
  	}
  };



  Tokenizer$1$2.prototype._parseLegacyEntity = function(){
  	var start = this._sectionStart + 1,
  	    limit = this._index - start;

  	if(limit > 6) limit = 6; 

  	while(limit >= 2){ 
  		var entity = this._buffer.substr(start, limit);

  		if(legacyMap$2.hasOwnProperty(entity)){
  			this._emitPartial(legacyMap$2[entity]);
  			this._sectionStart += limit + 1;
  			return;
  		} else {
  			limit--;
  		}
  	}
  };

  Tokenizer$1$2.prototype._stateInNamedEntity = function(c){
  	if(c === ";"){
  		this._parseNamedEntityStrict();
  		if(this._sectionStart + 1 < this._index && !this._xmlMode){
  			this._parseLegacyEntity();
  		}
  		this._state = this._baseState;
  	} else if((c < "a" || c > "z") && (c < "A" || c > "Z") && (c < "0" || c > "9")){
  		if(this._xmlMode);
  		else if(this._sectionStart + 1 === this._index);
  		else if(this._baseState !== TEXT$8){
  			if(c !== "="){
  				this._parseNamedEntityStrict();
  			}
  		} else {
  			this._parseLegacyEntity();
  		}

  		this._state = this._baseState;
  		this._index--;
  	}
  };

  Tokenizer$1$2.prototype._decodeNumericEntity = function(offset, base){
  	var sectionStart = this._sectionStart + offset;

  	if(sectionStart !== this._index){
  		
  		var entity = this._buffer.substring(sectionStart, this._index);
  		var parsed = parseInt(entity, base);

  		this._emitPartial(decodeCodePoint$1$2(parsed));
  		this._sectionStart = this._index;
  	} else {
  		this._sectionStart--;
  	}

  	this._state = this._baseState;
  };

  Tokenizer$1$2.prototype._stateInNumericEntity = function(c){
  	if(c === ";"){
  		this._decodeNumericEntity(2, 10);
  		this._sectionStart++;
  	} else if(c < "0" || c > "9"){
  		if(!this._xmlMode){
  			this._decodeNumericEntity(2, 10);
  		} else {
  			this._state = this._baseState;
  		}
  		this._index--;
  	}
  };

  Tokenizer$1$2.prototype._stateInHexEntity = function(c){
  	if(c === ";"){
  		this._decodeNumericEntity(3, 16);
  		this._sectionStart++;
  	} else if((c < "a" || c > "f") && (c < "A" || c > "F") && (c < "0" || c > "9")){
  		if(!this._xmlMode){
  			this._decodeNumericEntity(3, 16);
  		} else {
  			this._state = this._baseState;
  		}
  		this._index--;
  	}
  };

  Tokenizer$1$2.prototype._cleanup = function (){
  	if(this._sectionStart < 0){
  		this._buffer = "";
  		this._bufferOffset += this._index;
  		this._index = 0;
  	} else if(this._running){
  		if(this._state === TEXT$8){
  			if(this._sectionStart !== this._index){
  				this._cbs.ontext(this._buffer.substr(this._sectionStart));
  			}
  			this._buffer = "";
  			this._bufferOffset += this._index;
  			this._index = 0;
  		} else if(this._sectionStart === this._index){
  			
  			this._buffer = "";
  			this._bufferOffset += this._index;
  			this._index = 0;
  		} else {
  			
  			this._buffer = this._buffer.substr(this._sectionStart);
  			this._index -= this._sectionStart;
  			this._bufferOffset += this._sectionStart;
  		}

  		this._sectionStart = 0;
  	}
  };


  Tokenizer$1$2.prototype.write = function(chunk){
  	if(this._ended) this._cbs.onerror(Error(".write() after done!"));

  	this._buffer += chunk;
  	this._parse();
  };

  Tokenizer$1$2.prototype._parse = function(){
  	while(this._index < this._buffer.length && this._running){
  		var c = this._buffer.charAt(this._index);
  		if(this._state === TEXT$8) {
  			this._stateText(c);
  		} else if(this._state === BEFORE_TAG_NAME$2){
  			this._stateBeforeTagName(c);
  		} else if(this._state === IN_TAG_NAME$2) {
  			this._stateInTagName(c);
  		} else if(this._state === BEFORE_CLOSING_TAG_NAME$2){
  			this._stateBeforeCloseingTagName(c);
  		} else if(this._state === IN_CLOSING_TAG_NAME$2){
  			this._stateInCloseingTagName(c);
  		} else if(this._state === AFTER_CLOSING_TAG_NAME$2){
  			this._stateAfterCloseingTagName(c);
  		} else if(this._state === IN_SELF_CLOSING_TAG$2){
  			this._stateInSelfClosingTag(c);
  		}

  		
  		else if(this._state === BEFORE_ATTRIBUTE_NAME$2){
  			this._stateBeforeAttributeName(c);
  		} else if(this._state === IN_ATTRIBUTE_NAME$2){
  			this._stateInAttributeName(c);
  		} else if(this._state === AFTER_ATTRIBUTE_NAME$2){
  			this._stateAfterAttributeName(c);
  		} else if(this._state === BEFORE_ATTRIBUTE_VALUE$2){
  			this._stateBeforeAttributeValue(c);
  		} else if(this._state === IN_ATTRIBUTE_VALUE_DQ$2){
  			this._stateInAttributeValueDoubleQuotes(c);
  		} else if(this._state === IN_ATTRIBUTE_VALUE_SQ$2){
  			this._stateInAttributeValueSingleQuotes(c);
  		} else if(this._state === IN_ATTRIBUTE_VALUE_NQ$2){
  			this._stateInAttributeValueNoQuotes(c);
  		}

  		
  		else if(this._state === BEFORE_DECLARATION$2){
  			this._stateBeforeDeclaration(c);
  		} else if(this._state === IN_DECLARATION$2){
  			this._stateInDeclaration(c);
  		}

  		
  		else if(this._state === IN_PROCESSING_INSTRUCTION$2){
  			this._stateInProcessingInstruction(c);
  		}

  		
  		else if(this._state === BEFORE_COMMENT$2){
  			this._stateBeforeComment(c);
  		} else if(this._state === IN_COMMENT$2){
  			this._stateInComment(c);
  		} else if(this._state === AFTER_COMMENT_1$2){
  			this._stateAfterComment1(c);
  		} else if(this._state === AFTER_COMMENT_2$2){
  			this._stateAfterComment2(c);
  		}

  		
  		else if(this._state === BEFORE_CDATA_1$2){
  			this._stateBeforeCdata1(c);
  		} else if(this._state === BEFORE_CDATA_2$2){
  			this._stateBeforeCdata2(c);
  		} else if(this._state === BEFORE_CDATA_3$2){
  			this._stateBeforeCdata3(c);
  		} else if(this._state === BEFORE_CDATA_4$2){
  			this._stateBeforeCdata4(c);
  		} else if(this._state === BEFORE_CDATA_5$2){
  			this._stateBeforeCdata5(c);
  		} else if(this._state === BEFORE_CDATA_6$2){
  			this._stateBeforeCdata6(c);
  		} else if(this._state === IN_CDATA$2){
  			this._stateInCdata(c);
  		} else if(this._state === AFTER_CDATA_1$2){
  			this._stateAfterCdata1(c);
  		} else if(this._state === AFTER_CDATA_2$2){
  			this._stateAfterCdata2(c);
  		}

  		
  		else if(this._state === BEFORE_SPECIAL$2){
  			this._stateBeforeSpecial(c);
  		} else if(this._state === BEFORE_SPECIAL_END$2){
  			this._stateBeforeSpecialEnd(c);
  		}

  		
  		else if(this._state === BEFORE_SCRIPT_1$2){
  			this._stateBeforeScript1(c);
  		} else if(this._state === BEFORE_SCRIPT_2$2){
  			this._stateBeforeScript2(c);
  		} else if(this._state === BEFORE_SCRIPT_3$2){
  			this._stateBeforeScript3(c);
  		} else if(this._state === BEFORE_SCRIPT_4$2){
  			this._stateBeforeScript4(c);
  		} else if(this._state === BEFORE_SCRIPT_5$2){
  			this._stateBeforeScript5(c);
  		}

  		else if(this._state === AFTER_SCRIPT_1$2){
  			this._stateAfterScript1(c);
  		} else if(this._state === AFTER_SCRIPT_2$2){
  			this._stateAfterScript2(c);
  		} else if(this._state === AFTER_SCRIPT_3$2){
  			this._stateAfterScript3(c);
  		} else if(this._state === AFTER_SCRIPT_4$2){
  			this._stateAfterScript4(c);
  		} else if(this._state === AFTER_SCRIPT_5$2){
  			this._stateAfterScript5(c);
  		}

  		
  		else if(this._state === BEFORE_STYLE_1$2){
  			this._stateBeforeStyle1(c);
  		} else if(this._state === BEFORE_STYLE_2$2){
  			this._stateBeforeStyle2(c);
  		} else if(this._state === BEFORE_STYLE_3$2){
  			this._stateBeforeStyle3(c);
  		} else if(this._state === BEFORE_STYLE_4$2){
  			this._stateBeforeStyle4(c);
  		}

  		else if(this._state === AFTER_STYLE_1$2){
  			this._stateAfterStyle1(c);
  		} else if(this._state === AFTER_STYLE_2$2){
  			this._stateAfterStyle2(c);
  		} else if(this._state === AFTER_STYLE_3$2){
  			this._stateAfterStyle3(c);
  		} else if(this._state === AFTER_STYLE_4$2){
  			this._stateAfterStyle4(c);
  		}

  		
  		else if(this._state === BEFORE_ENTITY$2){
  			this._stateBeforeEntity(c);
  		} else if(this._state === BEFORE_NUMERIC_ENTITY$2){
  			this._stateBeforeNumericEntity(c);
  		} else if(this._state === IN_NAMED_ENTITY$2){
  			this._stateInNamedEntity(c);
  		} else if(this._state === IN_NUMERIC_ENTITY$2){
  			this._stateInNumericEntity(c);
  		} else if(this._state === IN_HEX_ENTITY$2){
  			this._stateInHexEntity(c);
  		}

  		else {
  			this._cbs.onerror(Error("unknown _state"), this._state);
  		}

  		this._index++;
  	}

  	this._cleanup();
  };

  Tokenizer$1$2.prototype.pause = function(){
  	this._running = false;
  };
  Tokenizer$1$2.prototype.resume = function(){
  	this._running = true;

  	if(this._index < this._buffer.length){
  		this._parse();
  	}
  	if(this._ended){
  		this._finish();
  	}
  };

  Tokenizer$1$2.prototype.end = function(chunk){
  	if(this._ended) this._cbs.onerror(Error(".end() after done!"));
  	if(chunk) this.write(chunk);

  	this._ended = true;

  	if(this._running) this._finish();
  };

  Tokenizer$1$2.prototype._finish = function(){
  	
  	if(this._sectionStart < this._index){
  		this._handleTrailingData();
  	}

  	this._cbs.onend();
  };

  Tokenizer$1$2.prototype._handleTrailingData = function(){
  	var data = this._buffer.substr(this._sectionStart);

  	if(this._state === IN_CDATA$2 || this._state === AFTER_CDATA_1$2 || this._state === AFTER_CDATA_2$2){
  		this._cbs.oncdata(data);
  	} else if(this._state === IN_COMMENT$2 || this._state === AFTER_COMMENT_1$2 || this._state === AFTER_COMMENT_2$2){
  		this._cbs.oncomment(data);
  	} else if(this._state === IN_NAMED_ENTITY$2 && !this._xmlMode){
  		this._parseLegacyEntity();
  		if(this._sectionStart < this._index){
  			this._state = this._baseState;
  			this._handleTrailingData();
  		}
  	} else if(this._state === IN_NUMERIC_ENTITY$2 && !this._xmlMode){
  		this._decodeNumericEntity(2, 10);
  		if(this._sectionStart < this._index){
  			this._state = this._baseState;
  			this._handleTrailingData();
  		}
  	} else if(this._state === IN_HEX_ENTITY$2 && !this._xmlMode){
  		this._decodeNumericEntity(3, 16);
  		if(this._sectionStart < this._index){
  			this._state = this._baseState;
  			this._handleTrailingData();
  		}
  	} else if(
  		this._state !== IN_TAG_NAME$2 &&
  		this._state !== BEFORE_ATTRIBUTE_NAME$2 &&
  		this._state !== BEFORE_ATTRIBUTE_VALUE$2 &&
  		this._state !== AFTER_ATTRIBUTE_NAME$2 &&
  		this._state !== IN_ATTRIBUTE_NAME$2 &&
  		this._state !== IN_ATTRIBUTE_VALUE_SQ$2 &&
  		this._state !== IN_ATTRIBUTE_VALUE_DQ$2 &&
  		this._state !== IN_ATTRIBUTE_VALUE_NQ$2 &&
  		this._state !== IN_CLOSING_TAG_NAME$2
  	){
  		this._cbs.ontext(data);
  	}
  	
  	
  };

  Tokenizer$1$2.prototype.reset = function(){
  	Tokenizer$1$2.call(this, {xmlMode: this._xmlMode, decodeEntities: this._decodeEntities}, this._cbs);
  };

  Tokenizer$1$2.prototype.getAbsoluteIndex = function(){
  	return this._bufferOffset + this._index;
  };

  Tokenizer$1$2.prototype._getSection = function(){
  	return this._buffer.substring(this._sectionStart, this._index);
  };

  Tokenizer$1$2.prototype._emitToken = function(name){
  	this._cbs[name](this._getSection());
  	this._sectionStart = -1;
  };

  Tokenizer$1$2.prototype._emitPartial = function(value){
  	if(this._baseState !== TEXT$8){
  		this._cbs.onattribdata(value); 
  	} else {
  		this._cbs.ontext(value);
  	}
  };

  var _inherits$3 = function() {};

  var _inherits$1$2 = Object.freeze({
  	default: _inherits$3
  });

  var _resolve_empty$3 = {};

  var _resolve_empty$1$2 = Object.freeze({
  	default: _resolve_empty$3
  });

  var require$$1$1$2 = ( _inherits$1$2 && _inherits$3 ) || _inherits$1$2;

  var require$$2$2 = ( _resolve_empty$1$2 && _resolve_empty$3 ) || _resolve_empty$1$2;

  var Tokenizer$2$2 = Tokenizer_1$2;





  var formTags$2 = {
  	input: true,
  	option: true,
  	optgroup: true,
  	select: true,
  	button: true,
  	datalist: true,
  	textarea: true
  };

  var openImpliesClose$2 = {
  	tr      : { tr:true, th:true, td:true },
  	th      : { th:true },
  	td      : { thead:true, th:true, td:true },
  	body    : { head:true, link:true, script:true },
  	li      : { li:true },
  	p       : { p:true },
  	h1      : { p:true },
  	h2      : { p:true },
  	h3      : { p:true },
  	h4      : { p:true },
  	h5      : { p:true },
  	h6      : { p:true },
  	select  : formTags$2,
  	input   : formTags$2,
  	output  : formTags$2,
  	button  : formTags$2,
  	datalist: formTags$2,
  	textarea: formTags$2,
  	option  : { option:true },
  	optgroup: { optgroup:true }
  };

  var voidElements$2 = {
  	__proto__: null,
  	area: true,
  	base: true,
  	basefont: true,
  	br: true,
  	col: true,
  	command: true,
  	embed: true,
  	frame: true,
  	hr: true,
  	img: true,
  	input: true,
  	isindex: true,
  	keygen: true,
  	link: true,
  	meta: true,
  	param: true,
  	source: true,
  	track: true,
  	wbr: true,

  	
  	path: true,
  	circle: true,
  	ellipse: true,
  	line: true,
  	rect: true,
  	use: true,
  	stop: true,
  	polyline: true,
  	polygon: true
  };

  var re_nameEnd$2 = /\s|\//;

  function Parser$1$2(cbs, options){
  	this._options = options || {};
  	this._cbs = cbs || {};

  	this._tagname = "";
  	this._attribname = "";
  	this._attribvalue = "";
  	this._attribs = null;
  	this._stack = [];

  	this.startIndex = 0;
  	this.endIndex = null;

  	this._lowerCaseTagNames = "lowerCaseTags" in this._options ?
  									!!this._options.lowerCaseTags :
  									!this._options.xmlMode;
  	this._lowerCaseAttributeNames = "lowerCaseAttributeNames" in this._options ?
  									!!this._options.lowerCaseAttributeNames :
  									!this._options.xmlMode;

  	if(this._options.Tokenizer) {
  		Tokenizer$2$2 = this._options.Tokenizer;
  	}
  	this._tokenizer = new Tokenizer$2$2(this._options, this);

  	if(this._cbs.onparserinit) this._cbs.onparserinit(this);
  }

  require$$1$1$2(Parser$1$2, require$$2$2.EventEmitter);

  Parser$1$2.prototype._updatePosition = function(initialOffset){
  	if(this.endIndex === null){
  		if(this._tokenizer._sectionStart <= initialOffset){
  			this.startIndex = 0;
  		} else {
  			this.startIndex = this._tokenizer._sectionStart - initialOffset;
  		}
  	}
  	else this.startIndex = this.endIndex + 1;
  	this.endIndex = this._tokenizer.getAbsoluteIndex();
  };


  Parser$1$2.prototype.ontext = function(data){
  	this._updatePosition(1);
  	this.endIndex--;

  	if(this._cbs.ontext) this._cbs.ontext(data);
  };

  Parser$1$2.prototype.onopentagname = function(name){
  	if(this._lowerCaseTagNames){
  		name = name.toLowerCase();
  	}

  	this._tagname = name;

  	if(!this._options.xmlMode && name in openImpliesClose$2) {
  		for(
  			var el;
  			(el = this._stack[this._stack.length - 1]) in openImpliesClose$2[name];
  			this.onclosetag(el)
  		);
  	}

  	if(this._options.xmlMode || !(name in voidElements$2)){
  		this._stack.push(name);
  	}

  	if(this._cbs.onopentagname) this._cbs.onopentagname(name);
  	if(this._cbs.onopentag) this._attribs = {};
  };

  Parser$1$2.prototype.onopentagend = function(){
  	this._updatePosition(1);

  	if(this._attribs){
  		if(this._cbs.onopentag) this._cbs.onopentag(this._tagname, this._attribs);
  		this._attribs = null;
  	}

  	if(!this._options.xmlMode && this._cbs.onclosetag && this._tagname in voidElements$2){
  		this._cbs.onclosetag(this._tagname);
  	}

  	this._tagname = "";
  };

  Parser$1$2.prototype.onclosetag = function(name){
  	this._updatePosition(1);
  	if(this._lowerCaseTagNames){
  		name = name.toLowerCase();
  	}
  	
  	
  	
  	if(this._options.xmlMode) {
  		const stack = this._stack;
  		let last = stack.pop();
  		while(last !== name) {
  			if(this._cbs.onerror) {
  				this._cbs.onerror("Unclosed tag <"+last+">");
  			}
  			last = stack.pop();
  		}
  		this.onopentagend();
  		if(this._cbs.onclosetag) {
  			this._cbs.onclosetag(last);
  		}
  	} else {
  		if(this._stack.length && (!(name in voidElements$2))) {
  			let pos = this._stack.lastIndexOf(name);
  			if(pos !== -1){
  				if(this._cbs.onclosetag){
  					pos = this._stack.length - pos;
  					while(pos--) this._cbs.onclosetag(this._stack.pop());
  				}
  				else this._stack.length = pos;
  			} else if(name === "p"){
  				this.onopentagname(name);
  				this._closeCurrentTag();
  			}
  		} else if(name === "br" || name === "p"){
  			this.onopentagname(name);
  			this._closeCurrentTag();
  		}
  	}
  };

  Parser$1$2.prototype.onselfclosingtag = function(){
  	if(this._options.xmlMode || this._options.recognizeSelfClosing){
  		this._closeCurrentTag();
  	} else {
  		this.onopentagend();
  	}
  };

  Parser$1$2.prototype._closeCurrentTag = function(){
  	var name = this._tagname;

  	this.onopentagend();

  	
  	
  	if(this._stack[this._stack.length - 1] === name){
  		if(this._cbs.onclosetag){
  			this._cbs.onclosetag(name);
  		}
  		this._stack.pop();
  	}
  };

  Parser$1$2.prototype.onattribname = function(name){
  	if(this._lowerCaseAttributeNames){
  		name = name.toLowerCase();
  	}
  	this._attribname = name;
  };

  Parser$1$2.prototype.onattribdata = function(value){
  	this._attribvalue += value;
  };

  Parser$1$2.prototype.onattribend = function(){
  	if(this._cbs.onattribute) this._cbs.onattribute(this._attribname, this._attribvalue);
  	if(
  		this._attribs &&
  		!Object.prototype.hasOwnProperty.call(this._attribs, this._attribname)
  	){
  		this._attribs[this._attribname] = this._attribvalue;
  	}
  	this._attribname = "";
  	this._attribvalue = "";
  };

  Parser$1$2.prototype._getInstructionName = function(value){
  	var idx = value.search(re_nameEnd$2),
  	    name = idx < 0 ? value : value.substr(0, idx);

  	if(this._lowerCaseTagNames){
  		name = name.toLowerCase();
  	}

  	return name;
  };

  Parser$1$2.prototype.ondeclaration = function(value){
  	if(this._cbs.onprocessinginstruction){
  		var name = this._getInstructionName(value);
  		this._cbs.onprocessinginstruction("!" + name, "!" + value);
  	}
  };

  Parser$1$2.prototype.onprocessinginstruction = function(value){
  	if(this._cbs.onprocessinginstruction){
  		var name = this._getInstructionName(value);
  		this._cbs.onprocessinginstruction("?" + name, "?" + value);
  	}
  };

  Parser$1$2.prototype.oncomment = function(value){
  	this._updatePosition(4);

  	if(this._cbs.oncomment) this._cbs.oncomment(value);
  	if(this._cbs.oncommentend) this._cbs.oncommentend();
  };

  Parser$1$2.prototype.oncdata = function(value){
  	this._updatePosition(1);

  	if(this._options.xmlMode || this._options.recognizeCDATA){
  		if(this._cbs.oncdatastart) this._cbs.oncdatastart();
  		if(this._cbs.ontext) this._cbs.ontext(value);
  		if(this._cbs.oncdataend) this._cbs.oncdataend();
  	} else {
  		this.oncomment("[CDATA[" + value + "]]");
  	}
  };

  Parser$1$2.prototype.onerror = function(err){
  	if(this._cbs.onerror) this._cbs.onerror(err);
  };

  Parser$1$2.prototype.onend = function(){
  	if(this._cbs.onclosetag){
  		for(
  			var i = this._stack.length;
  			i > 0;
  			this._cbs.onclosetag(this._stack[--i])
  		);
  	}
  	if(this._cbs.onend) this._cbs.onend();
  };



  Parser$1$2.prototype.reset = function(){
  	if(this._cbs.onreset) this._cbs.onreset();
  	this._tokenizer.reset();

  	this._tagname = "";
  	this._attribname = "";
  	this._attribs = null;
  	this._stack = [];

  	if(this._cbs.onparserinit) this._cbs.onparserinit(this);
  };


  Parser$1$2.prototype.parseComplete = function(data){
  	this.reset();
  	this.end(data);
  };

  Parser$1$2.prototype.write = function(chunk){
  	this._tokenizer.write(chunk);
  };

  Parser$1$2.prototype.end = function(chunk){
  	this._tokenizer.end(chunk);
  };

  Parser$1$2.prototype.pause = function(){
  	this._tokenizer.pause();
  };

  Parser$1$2.prototype.resume = function(){
  	this._tokenizer.resume();
  };


  Parser$1$2.prototype.parseChunk = Parser$1$2.prototype.write;
  Parser$1$2.prototype.done = Parser$1$2.prototype.end;

  var Parser_1$2 = Parser$1$2;

  Parser_1$2.prototype.ondeclaration = function(value){
    if(this._cbs.ondeclaration){
      this._cbs.ondeclaration(value);
    } else if(this._cbs.onprocessinginstruction){
      var name = this._getInstructionName(value);
      this._cbs.onprocessinginstruction("!" + name, "!" + value);
    }
  };


  Parser_1$2.prototype.oncdata = function(value){
    this._updatePosition(1);

    if(this._options.xmlMode || this._options.recognizeCDATA){
      if(this._cbs.oncdatastart) this._cbs.oncdatastart(value);
      
      if(this._cbs.oncdataend) this._cbs.oncdataend();
    } else {
      this.oncomment("[CDATA[" + value + "]]");
    }
  };

  function parseMarkup$2 (markup, options) {
    let format = options.ownerDocument ? options.ownerDocument.format : options.format;
    
    if (!format) {
      throw new Error("Either 'ownerDocument' or 'format' must be set.")
    }
    let parserOptions = Object.assign({}, options, {
      xmlMode: (format === 'xml')
    });
    let handler = new DomHandler$2({ format, elementFactory: options.elementFactory });
    let parser = new Parser_1$2(handler, parserOptions);
    parser.end(markup);
    return handler.document
  }

  const RE_WHITESPACE$2 = /\s+/g;
  const RE_DOCTYPE$2 = /^DOCTYPE\s+([^\s]+)(?:\s+PUBLIC\s+["]([^"]+)["](?:\s+["]([^"]+)["])?)\s*$/;


  class DomHandler$2 {
    constructor (options = {}) {
      this.elementFactory = options.elementFactory;
      if (!this.elementFactory) throw new Error("'elementFactory' is required")
      this.options = options;
      this.document = null;
      this._tagStack = [];
    }

    
    onparserinit () {
      this.document = this.elementFactory('document', { format: this.options.format });
      this._tagStack = [this.document];
    }

    onend () {
      
      if (this._tagStack.length > 1) {
        throw new Error(`Unexpected EOF. Tag was opened but not closed.`)
      }
    }

    onerror (error) {
      throw new Error(error)
    }

    onclosetag () {
      this._tagStack.pop();
    }

    _addDomElement (element) {
      let parent = this._tagStack[this._tagStack.length - 1];
      if (!parent.childNodes) parent.childNodes = [];
      let siblings = parent.childNodes;

      let previousSibling = siblings[siblings.length - 1];
      
      element.next = null;
      if (previousSibling) {
        element.prev = previousSibling;
        previousSibling.next = element;
      } else {
        element.prev = null;
      }
      
      siblings.push(element);
      element.parent = parent || null;
    }

    onopentag (name, attributes) {
      let element = this.document.createElement(name);
      forEach$2(attributes, (val, key) => {
        element.setAttribute(key, val);
      });
      this._addDomElement(element);
      this._tagStack.push(element);
    }

    ontext (text) {
      if (this.options.normalizeWhitespace) {
        text = text.replace(RE_WHITESPACE$2, ' ');
      }
      let lastTag;
      let _top = this._tagStack[this._tagStack.length - 1];
      if (_top && _top.childNodes) lastTag = _top.childNodes[_top.childNodes.length - 1];
      if (lastTag && lastTag.type === domelementtype$2.Text) {
        lastTag.data += text;
      } else {
        let element = this.document.createTextNode(text);
        this._addDomElement(element);
      }
    }

    oncomment (data) {
      var lastTag = this._tagStack[this._tagStack.length - 1];
      if (lastTag && lastTag.type === domelementtype$2.Comment) {
        lastTag.data += data;
      } else {
        let element = this.document.createComment(data);
        this._addDomElement(element);
        this._tagStack.push(element);
      }
    }

    oncommentend () {
      this._tagStack.pop();
    }

    oncdatastart (data) {
      let element = this.document.createCDATASection(data);
      this._addDomElement(element);
      this._tagStack.push(element);
    }

    oncdataend () {
      this._tagStack.pop();
    }

    onprocessinginstruction (name, data) {
      let element = this.document.createProcessingInstruction(name, data);
      this._addDomElement(element);
    }

    ondeclaration (data) {
      if (data.startsWith('DOCTYPE')) {
        let m = RE_DOCTYPE$2.exec(data);
        if (!m) throw new Error('Could not parse DOCTYPE element: ' + data)
        this.document.setDoctype(m[1], m[2], m[3]);
      } else {
        throw new Error('Not implemented: parse declaration ' + data)
      }
    }
  }

  const MemoryDOMElementFactory$2 = (type, data) => {
    return new MemoryDOMElement$2(type, data)
  };

  class MemoryDOMElement$2 extends DOMElement$2 {
    constructor (type, args = {}) {
      super();

      this.type = type;
      if (!type) throw new Error("'type' is mandatory")

      this.ownerDocument = args.ownerDocument;
      
      if (type !== 'document' && !this.ownerDocument) {
        throw new Error("'ownerDocument' is mandatory")
      }

      
      
      

      switch (type) {
        case domelementtype$2.Tag: {
          if (!args.name) throw new Error("'name' is mandatory.")
          this.name = this._normalizeName(args.name);
          this.nameWithoutNS = nameWithoutNS$3(this.name);
          this.properties = new Map();
          this.attributes = new Map();
          this.classes = new Set();
          this.styles = new Map();
          this.eventListeners = [];
          this.childNodes = args.children || args.childNodes || [];
          this._assign(args);
          break
        }
        case domelementtype$2.Text:
        case domelementtype$2.Comment: {
          this.data = args.data || '';
          break
        }
        case domelementtype$2.CDATA: {
          this.data = args.data || '';
          break
        }
        case domelementtype$2.Directive: {
          if (!args.name) throw new Error("'name' is mandatory.")
          this.name = this._normalizeName(args.name);
          this.nameWithoutNS = nameWithoutNS$3(this.name);
          this.data = args.data;
          break
        }
        case domelementtype$2.Doctype: {
          this.data = args.data;
          break
        }
        case 'document': {
          let format = args.format;
          this.format = format;
          if (!format) throw new Error("'format' is mandatory.")
          this.childNodes = args.children || args.childNodes || [];
          switch (format) {
            case 'xml':
              this.contentType = 'application/xml';
              break
            case 'html':
              this.contentType = 'text/html';
              break
            default:
              throw new Error('Unsupported format ' + format)
          }
          break
        }
        default:
          this.name = null;
          this.properties = new Map();
          this.attributes = new Map();
          this.classes = new Set();
          this.styles = new Map();
          this.eventListeners = [];
          this.childNodes = args.children || args.childNodes || [];
      }
    }

    getNativeElement () {
      return this
    }

    getNodeType () {
      switch (this.type) {
        case domelementtype$2.Tag:
        case domelementtype$2.Script:
        case domelementtype$2.Style:
          return 'element'
        default:
          return this.type
      }
    }

    isTextNode () {
      return this.type === 'text'
    }

    isElementNode () {
      return this.type === 'tag' || this.type === 'script'
    }

    isCommentNode () {
      return this.type === 'comment'
    }

    isDocumentNode () {
      return this.type === 'document'
    }

    isComponentNode () {
      return this.type === 'component'
    }

    clone (deep) {
      let clone = new MemoryDOMElement$2(this.type, this);
      if (this.childNodes) {
        clone.childNodes.length = 0;
        if (deep) {
          this.childNodes.forEach((child) => {
            clone.appendChild(child.clone(deep));
          });
        }
      }
      return clone
    }

    get tagName () {
      return this.getTagName()
    }

    set tagName (tagName) {
      this.setTagName(tagName);
    }

    getTagName () {
      return this.name
    }

    setTagName (tagName) {
      if (this._isXML()) {
        this.name = String(tagName);
      } else {
        this.name = String(tagName).toLowerCase();
      }
      this.nameWithoutNS = nameWithoutNS$3(this.name);
      return this
    }

    hasAttribute (name) {
      return this.attributes.has(name)
    }

    getAttribute (name) {
      return this.attributes.get(name)
    }

    setAttribute (name, value) {
      value = String(value);
      
      switch (name) {
        case 'class':
          this.classes = new Set();
          parseClasses$2(this.classes, value);
          break
        case 'style':
          this.styles = new Map();
          parseStyles$2(this.styles, value);
          break
        default:
          
      }
      this.attributes.set(name, value);
      if (this._isHTML()) {
        deriveHTMLPropertyFromAttribute$2(this, name, value);
      }
      return this
    }

    removeAttribute (name) {
      switch (name) {
        case 'class':
          this.classes = new Set();
          break
        case 'style':
          this.styles = new Map();
          break
        default:
          
      }
      this.attributes.delete(name);
      return this
    }

    getAttributes () {
      return this.attributes
    }

    getProperty (name) {
      if (this.properties) {
        return this.properties.get(name)
      }
    }

    setProperty (name, value) {
      if (this.properties) {
        if (this._isXML()) {
          throw new Error('setProperty() is only be used on HTML elements')
        }
        _setHTMLPropertyValue$2(this, name, value);
      }
      return this
    }

    hasClass (name) {
      if (this.classes) {
        return this.classes.has(name)
      }
    }

    addClass (name) {
      this.classes.add(name);
      this.attributes.set('class', stringifyClasses$2(this.classes));
      return this
    }

    removeClass (name) {
      if (this.classes && this.classes.has(name)) {
        this.classes.delete(name);
        this.attributes.set('class', stringifyClasses$2(this.classes));
      }
      return this
    }

    getContentType () {
      return this.getOwnerDocument().contentType
    }

    getDoctype () {
      if (this.isDocumentNode()) {
        return _findDocTypeElement$2(this)
      } else {
        return this.getOwnerDocument().getDoctype()
      }
    }

    setDoctype (qualifiedNameStr, publicId, systemId) {
      
      let doc = this.getOwnerDocument();
      let oldDocType = _findDocTypeElement$2(doc);
      let newDocType = this.createDocumentType(qualifiedNameStr, publicId, systemId);
      if (oldDocType) {
        doc.replaceChild(newDocType, oldDocType);
      } else {
        
        doc.insertBefore(newDocType, doc.getChildren()[0]);
      }
      doc.doctype = newDocType;
    }

    getInnerHTML () {
      return domUtils$2.getInnerHTML(this, { decodeEntities: true })
    }

    
    
    setInnerHTML (html) {
      if (this.childNodes) {
        let _doc = parseMarkup$2(html, {
          ownerDocument: this.getOwnerDocument(),
          decodeEntities: true,
          elementFactory: MemoryDOMElementFactory$2
        });
        this.empty();
        
        
        _doc.childNodes.slice(0).forEach((child) => {
          this.appendChild(child);
        });
      }
      return this
    }

    getOuterHTML () {
      return domUtils$2.getOuterHTML(this, { xmlMode: this._isXML(), decodeEntities: true })
    }

    getTextContent () {
      return domUtils$2.getText(this)
    }

    setTextContent (text) {
      switch (this.type) {
        case domelementtype$2.Text:
        case domelementtype$2.Comment:
        case domelementtype$2.CDATA: {
          this.data = text;
          break
        }
        default: {
          if (this.childNodes) {
            let child = this.createTextNode(text);
            this.empty();
            this.appendChild(child);
          }
        }
      }
      return this
    }

    getStyle (name) {
      if (this.styles) {
        return this.styles.get(name)
      }
    }

    setStyle (name, value) {
      if (this.styles) {
        if (DOMElement$2.pxStyles[name] && isNumber$2(value)) {
          value = value + 'px';
        }
        this.styles.set(name, value);
        this.attributes.set('style', stringifyStyles$2(this.styles));
      }
      return this
    }

    is (cssSelector) {
      return cssSelect$2.is(this, cssSelector, { xmlMode: this._isXML() })
    }

    find (cssSelector) {
      return cssSelect$2.selectOne(cssSelector, this, { xmlMode: this._isXML() })
    }

    findAll (cssSelector) {
      return cssSelect$2.selectAll(cssSelector, this, { xmlMode: this._isXML() })
    }

    getChildCount () {
      if (this.childNodes) {
        return this.childNodes.length
      } else {
        return 0
      }
    }

    getChildNodes () {
      return this.childNodes.slice(0)
    }

    getChildren () {
      return this.childNodes.filter(function (node) {
        return node.type === 'tag'
      })
    }

    get children () {
      return this.getChildren()
    }

    getChildAt (pos) {
      if (this.childNodes) {
        return this.childNodes[pos]
      }
    }

    getChildIndex (child) {
      if (this.childNodes) {
        return this.childNodes.indexOf(child)
      }
    }

    getLastChild () {
      if (this.childNodes) {
        return last$1$2(this.childNodes)
      }
    }

    getFirstChild () {
      if (this.childNodes) {
        return this.childNodes[0]
      }
    }

    getNextSibling () {
      return this.next
    }

    getPreviousSibling () {
      return this.prev
    }

    getParent () {
      return this.parent
    }

    getOwnerDocument () {
      return (this.type === 'document') ? this : this.ownerDocument
    }

    getFormat () {
      return this.getOwnerDocument().format
    }

    createDocument (format) {
      return MemoryDOMElement$2.createDocument(format)
    }

    createElement (tagName) {
      return new MemoryDOMElement$2(domelementtype$2.Tag, { name: tagName, ownerDocument: this.getOwnerDocument() })
    }

    createTextNode (text) {
      return new MemoryDOMElement$2(domelementtype$2.Text, { data: text, ownerDocument: this.getOwnerDocument() })
    }

    createComment (data) {
      return new MemoryDOMElement$2(domelementtype$2.Comment, { data: data, ownerDocument: this.getOwnerDocument() })
    }

    createProcessingInstruction (name, data) {
      return new MemoryDOMElement$2(domelementtype$2.Directive, { name: name, data: data, ownerDocument: this.getOwnerDocument() })
    }

    createDocumentType (qualifiedNameStr, publicId, systemId) {
      return new MemoryDOMDoctype$2(domelementtype$2.Doctype, { data: { name: qualifiedNameStr, publicId, systemId }, ownerDocument: this.getOwnerDocument() })
    }

    createCDATASection (data) {
      return new MemoryDOMElement$2(domelementtype$2.CDATA, { data: data, ownerDocument: this.getOwnerDocument() })
    }

    appendChild (child) {
      if (this.childNodes && !isNil$2(child)) {
        child = this._normalizeChild(child);
        if (!child) return this
        domUtils$2.appendChild(this, child);
        child.ownerDocument = this.getOwnerDocument();
      }
      return this
    }

    removeChild (child) {
      if (child.parentNode === this) {
        child.remove();
      }
    }

    insertAt (pos, child) {
      child = this._normalizeChild(child);
      if (!child) return this
      let childNodes = this.childNodes;
      if (childNodes) {
        
        if (pos >= childNodes.length) {
          domUtils$2.appendChild(this, child);
        } else {
          domUtils$2.prepend(childNodes[pos], child);
        }
        child.ownerDocument = this.getOwnerDocument();
      }
      return this
    }

    insertBefore (newChild, before) {
      if (isNil$2(before)) {
        return this.appendChild(newChild)
      } else if (this.childNodes) {
        var pos = this.childNodes.indexOf(before);
        if (pos > -1) {
          domUtils$2.prepend(before, newChild);
          newChild.ownerDocument = this.getOwnerDocument();
        } else {
          throw new Error('insertBefore(): reference node is not a child of this element.')
        }
      }
      return this
    }

    removeAt (pos) {
      let childNodes = this.childNodes;
      if (childNodes) {
        let child = childNodes[pos];
        child.remove();
      }
      return this
    }

    empty () {
      let childNodes = this.childNodes;
      if (childNodes) {
        childNodes.forEach((child) => {
          child.next = child.prev = child.parent = null;
        });
        childNodes.length = 0;
      }
      return this
    }

    remove () {
      domUtils$2.removeElement(this);
      return this
    }

    replaceChild (oldChild, newChild) {
      if (oldChild.parent === this) {
        oldChild.replaceWith(newChild);
      }
      return this
    }

    replaceWith (newEl) {
      newEl = this._normalizeChild(newEl);
      domUtils$2.replaceElement(this, newEl);
      newEl.ownerDocument = this.getOwnerDocument();
      return this
    }

    getEventListeners () {
      return this.eventListeners || []
    }

    click () {
      this.emit('click', { target: this });
      return this
    }

    emit (name, data) {
      this._propagateEvent(new MemoryDOMElementEvent$2(name, this, data));
    }

    getBoundingClientRect () {
      return { top: 0, left: 0, height: 0, width: 0 }
    }

    getClientRects () {
      return [{ top: 0, left: 0, height: 0, width: 0 }]
    }

    _propagateEvent (event) {
      let listeners = this.eventListeners;
      if (listeners) {
        let listener = listeners.find((l) => {
          return l.eventName === event.type
        });
        if (listener) listener.handler(event);
        if (event.stopped) return
        let p = this.parentNode;
        if (p) p._propagateEvent(event);
      }
    }

    removeAllEventListeners () {
      this.eventListeners = [];
      return this
    }

    _assign (other) {
      if (other.name) this.name = other.name;
      if (this.classes && other.classes) {
        other.classes.forEach((val) => {
          this.classes.add(val);
        });
      }
      if (this.styles && other.styles) {
        forEach$2(other.styles, (val, name) => {
          this.styles.set(name, val);
        });
      }
      
      
      let otherAttributes = other.attributes || other.attribs;
      if (this.attributes && otherAttributes) {
        forEach$2(otherAttributes, (val, name) => {
          switch (name) {
            case 'class': {
              parseClasses$2(this.classes, val);
              break
            }
            case 'style': {
              parseStyles$2(this.styles, val);
              break
            }
            default:
              
          }
          this.attributes.set(name, val);
        });
      }
      if (this.eventListeners && other.eventListeners) {
        this.eventListeners = this.eventListeners.concat(other.eventListeners);
      }
    }

    _normalizeChild (child) {
      if (isNil$2(child)) return

      if (isString$2(child)) {
        child = this.createTextNode(child);
      }
      
      if (!child || !child._isMemoryDOMElement) {
        throw new Error('Illegal argument: only String and MemoryDOMElement instances are valid.')
      }
      return child
    }

    _normalizeName (name) {
      if (this._isXML()) {
        return name
      } else {
        return name.toLowerCase()
      }
    }

    _isHTML () {
      return this.getFormat() === 'html'
    }

    _isXML () {
      return this.getFormat() === 'xml'
    }
  }

  MemoryDOMElement$2.prototype._isMemoryDOMElement = true;

  MemoryDOMElement$2.createDocument = function (format) {
    if (format === 'xml') {
      return new MemoryDOMElement$2('document', { format: format })
    } else {
      return MemoryDOMElement$2.parseMarkup(DOMElement$2.EMPTY_HTML, 'html')
    }
  };

  MemoryDOMElement$2.parseMarkup = function (str, format, options = {}) {
    if (!str) {
      return MemoryDOMElement$2.createDocument(format)
    }
    let parserOpts = Object.assign({
      format,
      decodeEntities: true,
      elementFactory: MemoryDOMElementFactory$2
    }, options);
    
    if (options.raw) {
      return parseMarkup$2(str, parserOpts)
    }
    if (options.snippet) {
      str = `<__snippet__>${str}</__snippet__>`;
    }
    let doc;
    if (format === 'html') {
      doc = parseMarkup$2(str, parserOpts);
      _sanitizeHTMLStructure$2(doc);
    } else if (format === 'xml') {
      doc = parseMarkup$2(str, parserOpts);
    }
    if (options.snippet) {
      let childNodes = doc.find('__snippet__').childNodes;
      if (childNodes.length === 1) {
        return childNodes[0]
      } else {
        return childNodes
      }
    } else {
      return doc
    }
  };

  MemoryDOMElement$2.wrap =
  MemoryDOMElement$2.wrapNativeElement = function (el) {
    if (inBrowser$2) {
      
      
      
      
      if (el === window || el === window.document) {
        return new DOMElementStub$2()
      
      
      
      } else if (el instanceof window.Node || el._isBrowserDOMElement) ;
    }
    
    if (!el._isMemoryDOMElement) {
      throw new Error('Illegal argument: expected MemoryDOMElement instance')
    }
    return el
  };

  MemoryDOMElement$2.unwrap = function (el) {
    
    if (!el._isMemoryDOMElement) {
      throw new Error('Illegal argument: expected MemoryDOMElement instance')
    }
    return el
  };




  MemoryDOMElement$2.isReverse = function () {
    return false
  };


  let _browserWindowStub$2;
  MemoryDOMElement$2.getBrowserWindow = function () {
    
    if (!_browserWindowStub$2) {
      _browserWindowStub$2 = MemoryDOMElement$2.createDocument('html');
    }
    return _browserWindowStub$2
  };

  class MemoryDOMDoctype$2 extends MemoryDOMElement$2 {
    get name () { return this.data.name }
    get publicId () { return this.data.publicId }
    get systemId () { return this.data.systemId }
  }

  function parseClasses$2 (classes, classStr) {
    classStr.split(/\s+/).forEach((name) => {
      classes.add(name);
    });
  }

  function stringifyClasses$2 (classes) {
    return Array.from(classes).join(' ')
  }

  function parseStyles$2 (styles, styleStr) {
    styleStr = (styleStr || '').trim();
    if (!styleStr) return
    styleStr.split(';').forEach((style) => {
      let n = style.indexOf(':');
      
      if (n < 1 || n === style.length - 1) return
      let name = style.slice(0, n).trim();
      let val = style.slice(n + 1).trim();
      styles.set(name, val);
    });
  }

  function stringifyStyles$2 (styles) {
    if (!styles) return ''
    let str = Object.keys(styles).map((name) => {
      return name + ':' + styles[name]
    }).join(';');
    if (str.length > 0) str += ';';
    return str
  }

  const BUILTIN_EVENTS$2 = [
    'keydown', 'keyup', 'keypress',
    'mousedown', 'mouseup', 'mouseover', 'click', 'dblclick'
  ].reduce((m, k) => { m[k] = true; return m }, {});

  class MemoryDOMElementEvent$2 {
    constructor (type, target, detail) {
      this.type = type;
      this.timeStamp = Date.now();
      this.target = target;

      if (BUILTIN_EVENTS$2[type]) {
        
        if (detail) {
          Object.assign(this, detail);
        }
      } else {
        this.detail = detail;
      }
    }

    stopPropagation () {
      this.stopped = true;
    }

    preventDefault () {
      this.defaultPrevented = true;
    }
  }

  class DOMElementStub$2 {
    on () {}
    off () {}
  }

  function nameWithoutNS$3 (name) {
    const idx = name.indexOf(':');
    if (idx > 0) {
      return name.slice(idx + 1)
    } else {
      return name
    }
  }



  const ATTR_TO_PROPS$2 = {
    'input': {
      'value': true,
      'checked': (el, name, value) => {
        const checked = (value !== 'off');
        el.setProperty('checked', checked);
      }
    }
  };

  function deriveHTMLPropertyFromAttribute$2 (el, name, value) {
    const mappings = ATTR_TO_PROPS$2[el.tagName];
    if (mappings) {
      let mapper = mappings[name];
      if (mapper === true) {
        el.setProperty(name, value);
      } else if (mapper) {
        mapper(el, name, value);
      }
    }
  }

  const PROPERTY_TRANSFORMATIONS$2 = {
    'input': {
      'checked': (el, name, value) => {
        if (value === true) {
          el.properties.set(name, true);
          el.properties.set('value', 'on');
        } else {
          el.properties.set(name, false);
          el.properties.set('value', 'off');
        }
      },
      'value': (el, name, value) => {
        let type = el.getAttribute('type');
        switch (type) {
          case 'checkbox':
            if (value === 'on') {
              el.properties.set(name, true);
              el.properties.set('value', 'on');
            } else {
              el.properties.set(name, false);
              el.properties.set('value', 'off');
            }
            break
          default:
            _setProperty$2(el, name, value);
        }
      }
    }
  };

  function _setProperty$2 (el, name, value) {
    if (value === undefined) {
      el.properties.delete(name);
    } else {
      el.properties.set(name, String(value));
    }
  }

  function _setHTMLPropertyValue$2 (el, name, value) {
    const trafos = PROPERTY_TRANSFORMATIONS$2[el.tagName];
    if (trafos) {
      let mapper = trafos[name];
      if (mapper) {
        mapper(el, name, value);
        return
      }
    }
    _setProperty$2(el, name, value);
  }

  function _sanitizeHTMLStructure$2 (doc) {
    
    
    
    
    let htmlEl = doc.find('html');
    if (!htmlEl) {
      
      
      let headEl = doc.find('head');
      let titleEl = doc.find('title');
      let metaEls = doc.findAll('meta');
      let bodyEl = doc.find('body');
      if (headEl) headEl.remove();
      if (titleEl) titleEl.remove();
      metaEls.forEach(e => e.remove());
      if (bodyEl) bodyEl.remove();

      
      
      let contentNodes = doc.childNodes.slice();
      contentNodes.forEach((c) => { c.parent = null; });
      doc.childNodes.length = 0;

      htmlEl = doc.createElement('html');
      
      
      
      if (!headEl) {
        headEl = doc.createElement('head');
        headEl.appendChild(titleEl);
        headEl.append(metaEls);
        htmlEl.appendChild(headEl);
      }
      if (!bodyEl) {
        bodyEl = doc.createElement('body');
        bodyEl.append(contentNodes);
      }
      htmlEl.appendChild(bodyEl);

      doc.append(htmlEl);
    }
  }

  function _findDocTypeElement$2 (doc) {
    
    if (doc.doctype) return doc.doctype
    const childNodes = doc.childNodes;
    for (let i = 0; i < childNodes.length; i++) {
      let child = childNodes[i];
      if (child.type === domelementtype$2.Doctype) {
        doc.doctype = child;
        return child
      }
    }
  }

  let DefaultDOMElement$2 = {};

  DefaultDOMElement$2.createDocument = function (format) {
    return _getDefaultImpl$2().createDocument(format)
  };


  DefaultDOMElement$2.createElement = function (tagName) {
    console.error('DEPRECATED: every element should have an ownerDocument. Use DefaultDOMElement.createDocument() to create a document first');
    let doc = DefaultDOMElement$2.createDocument('html');
    return doc.createElement(tagName)
  };


  DefaultDOMElement$2.createTextNode = function (text) {
    console.error('DEPRECATED: every element should have a ownerDocument. Use DefaultDOMElement.createDocument() to create a document first');
    let doc = DefaultDOMElement$2.createDocument('html');
    return doc.createTextNode(text)
  };


  DefaultDOMElement$2.getBrowserWindow = function () {
    return _getDefaultImpl$2().getBrowserWindow()
  };


  DefaultDOMElement$2.parseHTML = function (html, options) {
    return _getDefaultImpl$2().parseMarkup(html, 'html', options)
  };


  DefaultDOMElement$2.parseXML = function (xml, options) {
    return _getDefaultImpl$2().parseMarkup(xml, 'xml', options)
  };

  DefaultDOMElement$2.parseSnippet = function (str, format) {
    return _getDefaultImpl$2().parseMarkup(str, format, {snippet: true})
  };

  DefaultDOMElement$2.wrap =
  DefaultDOMElement$2.wrapNativeElement = function (nativeEl) {
    if (!nativeEl) throw new Error('Illegal argument')
    return _getDefaultImpl$2().wrap(nativeEl)
  };

  DefaultDOMElement$2.unwrap = function (nativeEl) {
    if (!nativeEl) throw new Error('Illegal argument')
    return _getDefaultImpl$2().unwrap(nativeEl)
  };



  DefaultDOMElement$2.isReverse = function (anchorNode, anchorOffset, focusNode, focusOffset) {
    return _getDefaultImpl$2().isReverse(anchorNode, anchorOffset, focusNode, focusOffset)
  };

  function _getDefaultImpl$2 () {
    
    if (platform$2.inBrowser || platform$2.inElectron) {
      return BrowserDOMElement$2
    } else {
      return MemoryDOMElement$2
    }
  }

  const _global$2 = (typeof global !== 'undefined') ? global : window;
  const substanceGlobals$2 = _global$2.hasOwnProperty('Substance') ? _global$2.Substance : _global$2.Substance = {
    DEBUG_RENDERING: false
  };

  class VirtualElement$2 extends DOMElement$2 {
    constructor (owner) {
      super();

      
      this.parent = null;
      
      this._owner = owner;
      
      this._ref = null;
    }

    getParent () {
      return this.parent
    }

    get childNodes () {
      return this.getChildNodes()
    }

    getChildCount () {
      return this.children.length
    }

    getChildAt (idx) {
      return this.children[idx]
    }

    
    getComponent () {
      return this._comp
    }

    
    ref (ref) {
      if (!ref) throw new Error('Illegal argument')
      
      if (this._ref) throw new Error('A VirtualElement can only be referenced once.')
      this._ref = ref;
      if (this._context) {
        const refs = this._context.refs;
        if (refs[ref]) {
          throw new Error('An item with reference "' + ref + '" already exists.')
        }
        refs[ref] = this;
      }
      return this
    }

    isInDocument () {
      return false
    }
  }

  VirtualElement$2.prototype._isVirtualElement = true;


  class VirtualHTMLElement$2 extends VirtualElement$2 {
    constructor (tagName) {
      super();

      this._tagName = tagName;
      this.classNames = null;
      this.attributes = null;
      this.htmlProps = null;
      this.style = null;
      this.eventListeners = null;

      
      this.children = [];
    }

    getTagName () {
      return this._tagName
    }

    setTagName (tagName) {
      this._tagName = tagName;
      return this
    }

    hasClass (className) {
      if (this.classNames) {
        return this.classNames.indexOf(className) > -1
      }
      return false
    }

    addClass (className) {
      if (!this.classNames) {
        this.classNames = [];
      }
      this.classNames.push(className);
      return this
    }

    removeClass (className) {
      if (this.classNames) {
        this.classNames = without$2(this.classNames, className);
      }
      return this
    }

    removeAttribute (name) {
      if (this.attributes) {
        delete this.attributes[name];
      }
      return this
    }

    getAttribute (name) {
      if (this.attributes) {
        return this.attributes[name]
      }
    }

    setAttribute (name, value) {
      if (!this.attributes) {
        this.attributes = {};
      }
      this.attributes[name] = value;
      return this
    }

    getAttributes () {
      
      
      
      var attributes = {};
      if (this.attributes) {
        Object.assign(attributes, this.attributes);
      }
      if (this.classNames) {
        attributes.class = this.classNames.join(' ');
      }
      if (this.style) {
        attributes.style = map$3(this.style, function (val, key) {
          return key + ':' + val
        }).join(';');
      }
      return attributes
    }

    getId () {
      return this.getAttribute('id')
    }

    setId (id) {
      this.setAttribute('id', id);
      return this
    }

    setTextContent (text) {
      text = String(text || '');
      this.empty();
      this.appendChild(text);
      return this
    }

    setInnerHTML (html) {
      html = html || '';
      this.empty();
      this._innerHTMLString = html;
      return this
    }

    getInnerHTML () {
      if (!this.hasOwnProperty('_innerHTMLString')) {
        throw new Error('Not supported.')
      } else {
        return this._innerHTMLString
      }
    }

    getValue () {
      return this.htmlProp('value')
    }

    setValue (value) {
      this.htmlProp('value', value);
      return this
    }

    getChildNodes () {
      return this.children
    }

    getChildren () {
      return this.children.filter(function (child) {
        return child.getNodeType() !== 'text'
      })
    }

    isTextNode () {
      return false
    }

    isElementNode () {
      return true
    }

    isCommentNode () {
      return false
    }

    isDocumentNode () {
      return false
    }

    append () {
      if (this._innerHTMLString) {
        throw Error('It is not possible to mix $$.html() with $$.append(). You can call $$.empty() to reset this virtual element.')
      }
      this._append(this.children, arguments);
      return this
    }

    appendChild (child) {
      if (this._innerHTMLString) {
        throw Error('It is not possible to mix $$.html() with $$.append(). You can call $$.empty() to reset this virtual element.')
      }
      this._appendChild(this.children, child);
      return this
    }

    insertAt (pos, child) {
      child = this._normalizeChild(child);
      if (!child) {
        throw new Error('Illegal child: ' + child)
      }
      if (!child._isVirtualElement) {
        throw new Error('Illegal argument for $$.insertAt():' + child)
      }
      if (pos < 0 || pos > this.children.length) {
        throw new Error('insertAt(): index out of bounds.')
      }
      this._insertAt(this.children, pos, child);
      return this
    }

    insertBefore (child, before) {
      var pos = this.children.indexOf(before);
      if (pos > -1) {
        this.insertAt(pos, child);
      } else {
        throw new Error('insertBefore(): reference node is not a child of this element.')
      }
      return this
    }

    removeAt (pos) {
      if (pos < 0 || pos >= this.children.length) {
        throw new Error('removeAt(): Index out of bounds.')
      }
      this._removeAt(pos);
      return this
    }

    removeChild (child) {
      if (!child || !child._isVirtualElement) {
        throw new Error('removeChild(): Illegal arguments. Expecting a CheerioDOMElement instance.')
      }
      var idx = this.children.indexOf(child);
      if (idx < 0) {
        throw new Error('removeChild(): element is not a child.')
      }
      this.removeAt(idx);
      return this
    }

    replaceChild (oldChild, newChild) {
      if (!newChild || !oldChild ||
          !newChild._isVirtualElement || !oldChild._isVirtualElement) {
        throw new Error('replaceChild(): Illegal arguments. Expecting BrowserDOMElement instances.')
      }
      var idx = this.children.indexOf(oldChild);
      if (idx < 0) {
        throw new Error('replaceChild(): element is not a child.')
      }
      this.removeAt(idx);
      this.insertAt(idx, newChild);
      return this
    }

    empty () {
      var children = this.children;
      while (children.length) {
        var child = children.pop();
        child.parent = null;
      }
      delete this._innerHTMLString;
      return this
    }

    getProperty (name) {
      if (this.htmlProps) {
        return this.htmlProps[name]
      }
    }

    setProperty (name, value) {
      if (!this.htmlProps) {
        this.htmlProps = {};
      }
      this.htmlProps[name] = value;
      return this
    }

    removeProperty (name) {
      if (this.htmlProps) {
        delete this.htmlProps[name];
      }
      return this
    }

    getStyle (name) {
      if (this.style) {
        return this.style[name]
      }
    }

    setStyle (name, value) {
      if (!this.style) {
        this.style = {};
      }
      if (DOMElement$2.pxStyles[name] && isNumber$2(value)) value = value + 'px';
      this.style[name] = value;
      return this
    }

    _createEventListener (eventName, handler, options) {
      options.context = options.context || this._owner._comp;
      return super._createEventListener(eventName, handler, options)
    }

    getNodeType () {
      return 'element'
    }

    hasInnerHTML () {
      return Boolean(this._innerHTMLString)
    }

    _normalizeChild (child) {
      if (isNil$2(child)) ; else if (child._isVirtualElement) {
        return child
      } else if (isString$2(child) || isBoolean$2(child) || isNumber$2(child)) {
        return new VirtualTextNode$2(String(child))
      } else {
        throw new Error('Unsupported child type')
      }
    }

    _append (outlet, args) {
      if (args.length === 1 && !isArray$3(args[0])) {
        this._appendChild(outlet, args[0]);
        return
      }
      var children;
      if (isArray$3(args[0])) {
        children = args[0];
      } else if (arguments.length > 1) {
        children = Array.prototype.slice.call(args, 0);
      } else {
        return
      }
      children.forEach(this._appendChild.bind(this, outlet));
    }

    _appendChild (outlet, child) {
      child = this._normalizeChild(child);
      
      
      if (!child) return
      outlet.push(child);
      this._attach(child);
      return child
    }

    _insertAt (outlet, pos, child) {
      if (!child) return
      outlet.splice(pos, 0, child);
      this._attach(child);
    }

    _removeAt (outlet, pos) {
      var child = outlet[pos];
      outlet.splice(pos, 1);
      this._detach(child);
    }

    _attach (child) {
      child.parent = this;
      if (this._context && child._owner !== this._owner && child._ref) {
        this._context.foreignRefs[child._ref] = child;
      }
    }

    _detach (child) {
      child.parent = null;
      if (this._context && child._owner !== this._owner && child._ref) {
        delete this.context.foreignRefs[child._ref];
      }
    }

    _mergeHTMLConfig (other) {
      if (other.classNames) {
        if (!this.classNames) {
          this.classNames = [];
        }
        this.classNames = this.classNames.concat(other.classNames);
      }
      if (other.attributes) {
        if (!this.attributes) {
          this.attributes = {};
        }
        Object.assign(this.attributes, other.attributes);
      }
      if (other.htmlProps) {
        if (!this.htmlProps) {
          this.htmlProps = {};
        }
        Object.assign(this.htmlProps, other.htmlProps);
      }
      if (other.style) {
        if (!this.style) {
          this.style = {};
        }
        Object.assign(this.style, other.style);
      }
      if (other.eventListeners) {
        if (!this.eventListeners) {
          this.eventListeners = [];
        }
        this.eventListeners = this.eventListeners.concat(other.eventListeners);
      }
    }
  }

  VirtualHTMLElement$2.prototype._isVirtualHTMLElement = true;


  class VirtualComponent$2 extends VirtualHTMLElement$2 {
    constructor (ComponentClass, props) {
      super();

      props = props || {};

      this.ComponentClass = ComponentClass;
      this.props = props;
      if (!props.children) {
        props.children = [];
      }
      this.children = props.children;
    }

    get _isVirtualHTMLElement () { return false }

    get _isVirtualComponent () { return true }

    getComponent () {
      return this._comp
    }

    
    
    getChildren () {
      return this.props.children
    }

    getNodeType () {
      return 'component'
    }

    outlet (name) {
      return new Outlet$2(this, name)
    }

    _attach (child) {
      child._preliminaryParent = this;
    }

    _detach (child) {
      child._preliminaryParent = null;
    }

    _copyHTMLConfig () {
      return {
        classNames: clone$2(this.classNames),
        attributes: clone$2(this.attributes),
        htmlProps: clone$2(this.htmlProps),
        style: clone$2(this.style),
        eventListeners: clone$2(this.eventListeners)
      }
    }
  }

  class Outlet$2 {
    constructor (virtualEl, name) {
      this.virtualEl = virtualEl;
      this.name = name;
      Object.freeze(this);
    }

    _getOutlet () {
      var outlet = this.virtualEl.props[this.name];
      if (!outlet) {
        outlet = [];
        this.virtualEl.props[this.name] = outlet;
      }
      return outlet
    }

    append () {
      var outlet = this._getOutlet();
      this.virtualEl._append(outlet, arguments);
      return this
    }

    empty () {
      var arr = this.virtualEl.props[this.name];
      arr.forEach(function (el) {
        this._detach(el);
      }.bind(this));
      arr.splice(0, arr.length);
      return this
    }
  }

  class VirtualTextNode$2 extends VirtualElement$2 {
    constructor (text) {
      super();
      this.text = text;
    }

    get _isVirtualTextNode () { return true }
  }

  VirtualElement$2.Component = VirtualComponent$2;
  VirtualElement$2.TextNode = VirtualTextNode$2;


  VirtualElement$2.createElement = function () {
    var content;
    var _first = arguments[0];
    var _second = arguments[1];
    var type;
    if (isString$2(_first)) {
      type = 'element';
    } else if (isFunction$1$2(_first) && _first.prototype._isComponent) {
      type = 'component';
    } else if (isNil$2(_first)) {
      throw new Error('$$(null): provided argument was null or undefined.')
    } else {
      throw new Error('Illegal usage of $$()')
    }
    
    var props = {};
    var classNames, ref;
    var eventHandlers = [];
    for (var key in _second) {
      if (!_second.hasOwnProperty(key)) continue
      var val = _second[key];
      switch (key) {
        case 'class':
          classNames = val;
          break
        case 'ref':
          ref = val;
          break
        default:
          props[key] = val;
      }
    }
    if (type === 'element') {
      content = new VirtualHTMLElement$2(_first);
      
      
      content.attr(props);
    } else {
      content = new VirtualComponent$2(_first, props);
    }
    
    
    content._owner = this.owner;
    if (classNames) {
      content.addClass(classNames);
    }
    if (ref) {
      content.ref(ref);
    }
    eventHandlers.forEach(function (h) {
      if (isFunction$1$2(h.handler)) {
        content.on(h.name, h.handler);
      } else if (isPlainObject$3(h.handler)) {
        var params = h.handler;
        content.on(h.name, params.handler, params.context, params);
      } else {
        throw new Error('Illegal arguments for $$(_,{ on' + h.name + '})')
      }
    });
    
    
    if (arguments.length > 2) {
      content.append(flattenOften$2(Array.prototype.slice.call(arguments, 2), 3));
    }
    return content
  };

  class RenderingEngine$2 {
    constructor (options = {}) {
      this.componentFactory = options.componentFactory;
      if (!this.componentFactory) throw new Error("'componentFactory' is mandatory")
      this.elementFactory = options.elementFactory || DefaultDOMElement$2.createDocument('html');
      if (!this.elementFactory) throw new Error("'elementFactory' is mandatory")
    }

    _render (comp, oldProps, oldState) {
      
      var vel = _createWrappingVirtualComponent$2(comp);
      var state = this._createState();
      if (oldProps) {
        state.setOldProps(vel, oldProps);
      }
      if (oldState) {
        state.setOldState(vel, oldState);
      }
      try {
        
        
        
        _capture$2(state, vel, 'forceCapture');
        

        
        
        _render$2(state, vel);
        

        _triggerUpdate$2(state, vel);
      } finally {
        state.dispose();
      }
      
    }

    
    _renderChild (comp, vel) {
      
      
      var state = this._createState();
      vel.parent = { _comp: comp };
      try {
        _capture$2(state, vel);
        _render$2(state, vel);
        return vel._comp
      } finally {
        state.dispose();
      }
    }

    _createState () {
      return new RenderingState$2(this.componentFactory, this.elementFactory)
    }
  }

  function _create$2 (state, vel) {
    var comp = vel._comp;
    console.assert(!comp, 'Component instance should not exist when this method is used.');
    var parent = vel.parent._comp;
    
    if (!parent) {
      parent = _create$2(state, vel.parent);
    }
    if (vel._isVirtualComponent) {
      console.assert(parent, 'A Component should have a parent.');
      comp = state.componentFactory.createComponent(vel.ComponentClass, parent, vel.props);
      
      vel.props = comp.props;
      comp.__htmlConfig__ = vel._copyHTMLConfig();
    } else if (vel._isVirtualHTMLElement) {
      comp = state.componentFactory.createElementComponent(parent, vel);
    } else if (vel._isVirtualTextNode) {
      comp = state.componentFactory.createTextNodeComponent(parent, vel);
    }
    if (vel._ref) {
      comp._ref = vel._ref;
    }
    if (vel._owner) {
      comp._owner = vel._owner._comp;
    }
    vel._comp = comp;
    return comp
  }

  function _capture$2 (state, vel, forceCapture) {
    if (state.isCaptured(vel)) {
      return vel
    }
    
    var comp = vel._comp;
    if (!comp) {
      comp = _create$2(state, vel);
      state.setNew(vel);
    }
    if (vel._isVirtualComponent) {
      var needRerender;
      
      
      
      if (forceCapture) {
        needRerender = true;
      } else {
        
        needRerender = !comp.el || comp.shouldRerender(vel.props, comp.state);
        comp.__htmlConfig__ = vel._copyHTMLConfig();
        state.setOldProps(vel, comp.props);
        state.setOldState(vel, comp.state);
        
        comp._setProps(vel.props);
        if (!state.isNew(vel)) {
          state.setUpdated(vel);
        }
      }
      if (needRerender) {
        var context = new CaptureContext$2(vel);
        var content = comp.render(context.$$);
        if (!content) {
          throw new Error('Component.render() returned nil.')
        }
        if (!content._isVirtualHTMLElement) {
          throw new Error('Component.render() must return a plain element.')
        }

        if (comp.__htmlConfig__) {
          content._mergeHTMLConfig(comp.__htmlConfig__);
        }
        content._comp = comp;
        vel._content = content;
        if (!state.isNew(vel) && comp.isMounted()) {
          state.setUpdated(vel);
        }
        
        _prepareVirtualComponent$2(state, comp, content);
        
        
        if (substanceGlobals$2.DEBUG_RENDERING) {
          
          
          
          
          var stack = content.children.slice(0);
          while (stack.length) {
            var child = stack.shift();
            if (state.isCaptured(child)) continue
            
            if (child._isVirtualComponent) continue
            if (!child._comp) {
              _create$2(state, child);
            }
            if (child._isVirtualHTMLElement && child.children.length > 0) {
              stack = stack.concat(child.children);
            }
            state.setCaptured(child);
          }
          state.setCaptured(content);
          
          
          var descendingContext = new DescendingContext$2(state, context);
          while (descendingContext.hasPendingCaptures()) {
            descendingContext.reset();
            comp.render(descendingContext.$$);
          }
        } else {
          
          
          _capture$2(state, vel._content);
        }
      } else {
        state.setSkipped(vel);
      }
    } else if (vel._isVirtualHTMLElement) {
      for (var i = 0; i < vel.children.length; i++) {
        _capture$2(state, vel.children[i]);
      }
    }
    state.setCaptured(vel);
    return vel
  }

  function _render$2 (state, vel) {
    if (state.isSkipped(vel)) return
    

    
    
    
    
    
    

    let comp = vel._comp;
    console.assert(comp && comp._isComponent, 'A captured VirtualElement must have a component instance attached.');

    
    if (vel._isVirtualComponent) {
      _render$2(state, vel._content);

      
      const context = vel._content._context;
      let refs = {};
      let foreignRefs = {};
      forEach$2(context.refs, (vel, ref) => {
        refs[ref] = vel._comp;
      });
      forEach$2(context.foreignRefs, (vel, ref) => {
        foreignRefs[ref] = vel._comp;
      });
      comp.refs = refs;
      comp.__foreignRefs__ = foreignRefs;
      return
    }

    
    if (!comp.el) {
      comp.el = _createElement$3(state, vel);
      comp.el._comp = comp;
    }
    _updateElement$2(comp, vel);

    
    if (vel._isVirtualHTMLElement && !vel.hasInnerHTML()) {
      var newChildren = vel.children;
      var oldComp, virtualComp, newComp;
      var pos1 = 0; var pos2 = 0;

      
      
      
      var oldChildren = [];
      comp.el.getChildNodes().forEach(function (node) {
        var childComp = node._comp;

        
        
        
        

        
        if (!childComp || state.isRelocated(childComp)) {
          comp.el.removeChild(node);
        } else {
          oldChildren.push(childComp);
        }
      });

      while (pos1 < oldChildren.length || pos2 < newChildren.length) {
        
        
        
        
        do {
          oldComp = oldChildren[pos1++];
        } while (oldComp && (state.isDetached(oldComp)))

        virtualComp = newChildren[pos2++];
        
        if (oldComp && !virtualComp) {
          while (oldComp) {
            _removeChild$2(state, comp, oldComp);
            oldComp = oldChildren[pos1++];
          }
          break
        }

        
        if (oldComp && oldComp.el.isTextNode() &&
            virtualComp && virtualComp._isVirtualTextNode &&
            oldComp.el.textContent === virtualComp.text) {
          continue
        }

        
        if (!state.isRendered(virtualComp)) {
          _render$2(state, virtualComp);
        }

        newComp = virtualComp._comp;

        
        
        if (state.isRelocated(newComp)) {
          newComp._setParent(comp);
        }

        console.assert(newComp, 'Component instance should now be available.');

        
        if (virtualComp && !oldComp) {
          _appendChild$2(state, comp, newComp);
          continue
        
        } else if (state.isMapped(virtualComp)) {
          
          if (newComp === oldComp) ; else if (state.isMapped(oldComp)) {
            
            state.setDetached(oldComp);
            _removeChild$2(state, comp, oldComp);
            pos2--;
          
          } else {
            _removeChild$2(state, comp, oldComp);
            pos2--;
          }
        } else if (state.isMapped(oldComp)) {
          _insertChildBefore$2(state, comp, newComp, oldComp);
          pos1--;
        } else {
          
          
          
          
          _replaceChild$2(state, comp, oldComp, newComp);
        }
      }
    }

    state.setRendered(vel);
  }

  function _triggerUpdate$2 (state, vel) {
    if (vel._isVirtualComponent) {
      if (!state.isSkipped(vel)) {
        vel._content.children.forEach(_triggerUpdate$2.bind(null, state));
      }
      if (state.isUpdated(vel)) {
        vel._comp.didUpdate(state.getOldProps(vel), state.getOldState(vel));
      }
    } else if (vel._isVirtualHTMLElement) {
      vel.children.forEach(_triggerUpdate$2.bind(null, state));
    }
  }

  function _appendChild$2 (state, parent, child) {
    parent.el.appendChild(child.el);
    _triggerDidMount$2(state, parent, child);
  }

  function _replaceChild$2 (state, parent, oldChild, newChild) {
    parent.el.replaceChild(oldChild.el, newChild.el);
    if (!state.isDetached(oldChild)) {
      oldChild.triggerDispose();
    }
    _triggerDidMount$2(state, parent, newChild);
  }

  function _insertChildBefore$2 (state, parent, child, before) {
    parent.el.insertBefore(child.el, before.el);
    _triggerDidMount$2(state, parent, child);
  }

  function _removeChild$2 (state, parent, child) {
    parent.el.removeChild(child.el);
    if (!state.isDetached(child)) {
      child.triggerDispose();
    }
  }

  function _triggerDidMount$2 (state, parent, child) {
    if (!state.isDetached(child) &&
        parent.isMounted() && !child.isMounted()) {
      child.triggerDidMount(true);
    }
  }


  function _prepareVirtualComponent$2 (state, comp, vc) {
    var newRefs = {};
    var foreignRefs = {};
    
    
    
    
    if (vc._context) {
      newRefs = vc._context.refs;
      foreignRefs = vc._context.foreignRefs;
    }
    var oldRefs = comp.refs;
    var oldForeignRefs = comp.__foreignRefs__;
    
    forEach$2(newRefs, function (vc, ref) {
      var comp = oldRefs[ref];
      if (comp) _mapComponents$2(state, comp, vc);
    });
    forEach$2(foreignRefs, function (vc, ref) {
      var comp = oldForeignRefs[ref];
      if (comp) _mapComponents$2(state, comp, vc);
    });
  }



  function _mapComponents$2 (state, comp, vc) {
    if (!comp && !vc) return true
    if (!comp || !vc) return false
    
    
    
    
    
    if (state.isMapped(vc) || state.isMapped(comp)) {
      return vc._comp === comp
    }
    if (vc._comp) {
      if (vc._comp === comp) {
        state.setMapped(vc);
        state.setMapped(comp);
        return true
      } else {
        return false
      }
    }
    if (!_isOfSameType$2(comp, vc)) {
      return false
    }

    vc._comp = comp;
    state.setMapped(vc);
    state.setMapped(comp);

    var canMapParent;
    var parent = comp.getParent();
    if (vc.parent) {
      canMapParent = _mapComponents$2(state, parent, vc.parent);
    
    
    
    
    } else if (vc._preliminaryParent) {
      while (parent && parent._isElementComponent) {
        parent = parent.getParent();
      }
      canMapParent = _mapComponents$2(state, parent, vc._preliminaryParent);
    }
    if (!canMapParent) {
      state.setRelocated(vc);
      state.setRelocated(comp);
    }
    return canMapParent
  }

  function _isOfSameType$2 (comp, vc) {
    return (
      (comp._isElementComponent && vc._isVirtualHTMLElement) ||
      (comp._isComponent && vc._isVirtualComponent && comp.constructor === vc.ComponentClass) ||
      (comp._isTextNodeComponent && vc._isVirtualTextNode)
    )
  }

  function _createElement$3 (state, vel) {
    var el;
    if (vel._isVirtualTextNode) {
      el = state.elementFactory.createTextNode(vel.text);
    } else {
      el = state.elementFactory.createElement(vel.tagName);
    }
    return el
  }

  function _updateElement$2 (comp, vel) {
    if (comp._isTextNodeComponent) {
      comp.setTextContent(vel.text);
      return
    }
    var el = comp.el;
    console.assert(el, "Component's element should exist at this point.");
    var tagName = el.getTagName();
    if (vel.tagName.toLowerCase() !== tagName) {
      el.setTagName(vel.tagName);
    }
    _updateHash$2({
      oldHash: el.getAttributes(),
      newHash: vel.getAttributes(),
      update: function (key, val) {
        el.setAttribute(key, val);
      },
      remove: function (key) {
        el.removeAttribute(key);
      }
    });
    _updateHash$2({
      oldHash: el.htmlProps,
      newHash: vel.htmlProps,
      update: function (key, val) {
        el.setProperty(key, val);
      },
      remove: function (key) {
        el.removeProperty(key);
      }
    });
    _updateListeners$2({
      el: el,
      oldListeners: el.getEventListeners(),
      newListeners: vel.getEventListeners()
    });

    
    if (vel.hasInnerHTML()) {
      if (!el._hasInnerHTML) {
        el.empty();
        el.setInnerHTML(vel.getInnerHTML());
      } else {
        var oldInnerHTML = el.getInnerHTML();
        var newInnerHTML = vel.getInnerHTML();
        if (oldInnerHTML !== newInnerHTML) {
          el.setInnerHTML(newInnerHTML);
        }
      }
      el._hasInnerHTML = true;
    }
  }

  function _updateHash$2 (args) {
    const newHash = args.newHash;
    const oldHash = args.oldHash || {};
    const update = args.update;
    const remove = args.remove;
    let updatedKeys = {};
    for (let key in newHash) {
      if (newHash.hasOwnProperty(key)) {
        var oldVal = oldHash[key];
        var newVal = newHash[key];
        updatedKeys[key] = true;
        if (oldVal !== newVal) {
          update(key, newVal);
        }
      }
    }
    
    
    if (isFunction$1$2(oldHash.keys) && oldHash.size > 0) {
      let keys = Array.from(oldHash.keys());
      keys.forEach((key) => {
        if (!updatedKeys[key]) {
          remove(key);
        }
      });
    } else {
      for (let key in oldHash) {
        if (oldHash.hasOwnProperty(key) && !updatedKeys[key]) {
          remove(key);
        }
      }
    }
  }

  function _updateListeners$2 (args) {
    var el = args.el;
    
    
    
    var newListeners = args.newListeners || [];
    el.removeAllEventListeners();
    for (var i = 0; i < newListeners.length; i++) {
      el.addEventListener(newListeners[i]);
    }
  }



  class DescendingContext$2 {
    constructor (state, captureContext) {
      this.state = state;
      this.owner = captureContext.owner;
      this.refs = {};
      this.foreignRefs = {};
      this.elements = captureContext.elements;
      this.pos = 0;
      this.updates = captureContext.components.length;
      this.remaining = this.updates;

      this.$$ = this._createComponent.bind(this);
    }

    _createComponent () {
      var state = this.state;
      var vel = this.elements[this.pos++];
      
      
      
      if (!state.isCaptured(vel) && vel._isVirtualComponent &&
           vel.parent && state.isCaptured(vel.parent)) {
        _capture$2(state, vel);
        this.updates++;
        this.remaining--;
      }
      
      
      
      
      vel = VirtualElement$2.createElement.apply(this, arguments);
      
      vel._context = this;
      vel._owner = this.owner;
      
      
      vel._attach = function () {};
      vel._detach = function () {};
      return vel
    }

    hasPendingCaptures () {
      return this.updates > 0 && this.remaining > 0
    }

    reset () {
      this.pos = 0;
      this.updates = 0;
      this.refs = {};
    }

    _ancestorsReady (vel) {
      while (vel) {
        if (this.state.isCaptured(vel) ||
            
            vel === this.owner || vel === this.owner._content) {
          return true
        }
        vel = vel.parent;
      }
      return false
    }
  }

  RenderingEngine$2._internal = {
    _capture: _capture$2,
    _wrap: _createWrappingVirtualComponent$2
  };

  class CaptureContext$2 {
    constructor (owner) {
      this.owner = owner;
      this.refs = {};
      this.foreignRefs = {};
      this.elements = [];
      this.components = [];
      this.$$ = this._createComponent.bind(this);
      this.$$.capturing = true;
    }

    _createComponent () {
      var vel = VirtualElement$2.createElement.apply(this, arguments);
      vel._context = this;
      vel._owner = this.owner;
      if (vel._isVirtualComponent) {
        
        this.components.push(vel);
      }
      this.elements.push(vel);
      return vel
    }
  }

  function _createWrappingVirtualComponent$2 (comp) {
    var vel = new VirtualElement$2.Component(comp.constructor);
    vel._comp = comp;
    if (comp.__htmlConfig__) {
      vel._mergeHTMLConfig(comp.__htmlConfig__);
    }
    return vel
  }

  RenderingEngine$2.createContext = function (comp) {
    var vel = _createWrappingVirtualComponent$2(comp);
    return new CaptureContext$2(vel)
  };

  class RenderingState$2 {
    constructor (componentFactory, elementFactory) {
      this.componentFactory = componentFactory;
      this.elementFactory = elementFactory;
      this.poluted = [];
      this.id = '__' + uuid$2();
    }

    dispose () {
      var id = this.id;
      this.poluted.forEach(function (obj) {
        delete obj[id];
      });
    }

    set (obj, key, val) {
      var info = obj[this.id];
      if (!info) {
        info = {};
        obj[this.id] = info;
        this.poluted.push(obj);
      }
      info[key] = val;
    }

    get (obj, key) {
      var info = obj[this.id];
      if (info) {
        return info[key]
      }
    }

    setMapped (c) {
      this.set(c, 'mapped', true);
    }

    isMapped (c) {
      return Boolean(this.get(c, 'mapped'))
    }

    
    
    setRelocated (c) {
      this.set(c, 'relocated', true);
    }

    isRelocated (c) {
      return Boolean(this.get(c, 'relocated'))
    }

    setDetached (c) {
      this.set(c, 'detached', true);
    }

    isDetached (c) {
      return Boolean(this.get(c, 'detached'))
    }

    setCaptured (vc) {
      this.set(vc, 'captured', true);
    }

    isCaptured (vc) {
      return Boolean(this.get(vc, 'captured'))
    }

    setNew (vc) {
      this.set(vc, 'created', true);
    }

    isNew (vc) {
      return Boolean(this.get(vc, 'created'))
    }

    setUpdated (vc) {
      this.set(vc, 'updated', true);
    }

    isUpdated (vc) {
      return Boolean(this.get(vc, 'updated'))
    }

    setSkipped (vc) {
      this.set(vc, 'skipped', true);
    }

    isSkipped (vc) {
      return Boolean(this.get(vc, 'skipped'))
    }

    setRendered (vc) {
      this.set(vc, 'rendered', true);
    }

    isRendered (vc) {
      return Boolean(this.get(vc, 'rendered'))
    }

    setOldProps (vc, oldProps) {
      this.set(vc, 'oldProps', oldProps);
    }

    getOldProps (vc) {
      return this.get(vc, 'oldProps')
    }

    setOldState (vc, oldState) {
      this.set(vc, 'oldState', oldState);
    }

    getOldState (vc) {
      return this.get(vc, 'oldState')
    }
  }

  const COMPONENT_FACTORY$2 = {
    createComponent (ComponentClass, parent, props) {
      return new ComponentClass(parent, props)
    },
    createElementComponent (parent, virtualElement) {
      return new ElementComponent$2(parent, virtualElement)
    },
    createTextNodeComponent (parent, virtualElement) {
      return new TextNodeComponent$2(parent, virtualElement)
    }
  };

  class Component$2 extends EventEmitter$2 {
    
    constructor (parent, props = {}, options = {}) {
      super();

      
      
      
      
      

      this.parent = (parent && parent._isComponent) ? parent : null;

      
      this.el = options.el;

      
      
      let context = options.context ? options.context : this._getContext() || {};
      this.context = context;
      Object.freeze(this.context);

      
      
      
      this.renderingEngine = (parent && parent.renderingEngine) || context.renderingEngine || options.renderingEngine || new RenderingEngine$2({
        componentFactory: COMPONENT_FACTORY$2
      });

      
      
      if (this._SKIP_COMPONENT_INIT) return

      this.__id__ = uuid$2();

      
      this.refs = {};
      
      
      
      this.__foreignRefs__ = {};

      
      this._actionHandlers = {};

      
      this.props = props;
      Object.freeze(this.props);

      
      this.state = this.getInitialState() || {};
      Object.freeze(this.state);
    }

    getId () {
      return this.__id__
    }

    setId () {
      throw new Error("'id' is readonly")
    }

    
    getChildContext () {
      return this.childContext || {}
    }

    
    getInitialState () {
      return {}
    }

    
    getParent () {
      return this.parent
    }

    
    getRoot () {
      var comp = this;
      var parent = comp;
      while (parent) {
        comp = parent;
        parent = comp.getParent();
      }
      return comp
    }

    getElement () {
      return this.el
    }

    getNativeElement () {
      return this.el.getNativeElement()
    }

    
    getLabel (name, ...args) {
      let labelProvider = this.getLabelProvider();
      if (!labelProvider) throw new Error('Missing labelProvider.')
      return labelProvider.getLabel(name, ...args)
    }

    getLabelProvider () {
      return this.context.labelProvider
    }

    
    getComponent (componentName, maybe) {
      let componentRegistry = this.getComponentRegistry();
      if (!componentRegistry) throw new Error('Missing componentRegistry.')
      const ComponentClass = componentRegistry.get(componentName);
      if (!maybe && !ComponentClass) {
        throw new Error('No Component registered with name ' + componentName)
      }
      return ComponentClass
    }

    getComponentRegistry () {
      return this.props.componentRegistry || this.context.componentRegistry
    }

    getFlow () {
      return this.context.flow
    }

    
    render ($$) {
      
      return $$('div')
    }

    
    mount (el) {
      if (!el) {
        throw new Error('Element is required.')
      }
      el = DefaultDOMElement$2.wrap(el);
      
      this.el = null;
      this.renderingEngine = Component$2.createRenderingEngine(el.getOwnerDocument());
      this._render();
      el.appendChild(this.el);
      if (el.isInDocument()) {
        this.triggerDidMount(true);
      }
      return this
    }

    
    shouldRerender(newProps, newState) { 
      return true
    }

    
    rerender () {
      this._rerender(this.props, this.state);
    }

    _rerender (oldProps, oldState) {
      this._render(oldProps, oldState);
      
      if (!this.isMounted()) {
        this.didUpdate(oldProps, oldState);
      }
    }

    _render (oldProps, oldState) {
      if (this.__isRendering__) {
        throw new Error('Component is rendering already.')
      }
      this.__isRendering__ = true;
      try {
        this.renderingEngine._render(this, oldProps, oldState);
      } finally {
        delete this.__isRendering__;
      }
    }

    
    triggerDidMount () {
      
      
      

      
      
      if (!this.__isMounted__) {
        this.__isMounted__ = true;
        this.didMount();
      }
      
      this.getChildren().forEach(function (child) {
        
        
        child.triggerDidMount(true);
      });
    }

    
    didMount () {}

    
    didUpdate () {}

    
    isMounted () {
      return this.__isMounted__
    }

    
    triggerDispose () {
      this.getChildren().forEach(function (child) {
        child.triggerDispose();
      });
      this.dispose();
      this.__isMounted__ = false;
    }

    
    dispose () {}

    
    _setParent (newParent) {
      this.parent = newParent;
      this.context = this._getContext() || {};
      Object.freeze(this.context);
    }

    
    send (action) {
      
      var comp = this.getParent();
      while (comp) {
        if (comp._actionHandlers && comp._actionHandlers[action]) {
          comp._actionHandlers[action].apply(comp, Array.prototype.slice.call(arguments, 1));
          return true
        }
        comp = comp.getParent();
      }
      console.warn('Action', action, 'was not handled.');
      return false
    }

    
    handleActions (actionHandlers) {
      forEach$2(actionHandlers, function (method, actionName) {
        this.handleAction(actionName, method);
      }.bind(this));
      return this
    }

    
    handleAction (name, handler) {
      if (!name || !handler || !isFunction$1$2(handler)) {
        throw new Error('Illegal arguments.')
      }
      handler = handler.bind(this);
      this._actionHandlers[name] = handler;
    }

    
    getState () {
      return this.state
    }

    
    setState (newState) {
      var oldProps = this.props;
      var oldState = this.state;
      
      
      var needRerender = !this.__isSettingProps__ &&
        this.shouldRerender(this.getProps(), newState);
      
      this.willUpdateState(newState);
      this.state = newState || {};
      Object.freeze(this.state);
      if (needRerender) {
        this._rerender(oldProps, oldState);
      } else if (!this.__isSettingProps__) {
        this.didUpdate(oldProps, oldState);
      }
    }

    
    extendState (newState) {
      newState = extend$2({}, this.state, newState);
      this.setState(newState);
    }

    
    willUpdateState(newState) { 
    }

    
    getProps () {
      return this.props
    }

    
    setProps (newProps) {
      var oldProps = this.props;
      var oldState = this.state;
      var needRerender = this.shouldRerender(newProps, this.state);
      this._setProps(newProps);
      if (needRerender) {
        this._rerender(oldProps, oldState);
      } else {
        this.didUpdate(oldProps, oldState);
      }
    }

    _setProps (newProps) {
      newProps = newProps || {};
      
      this.__isSettingProps__ = true;
      try {
        this.willReceiveProps(newProps);
        this.props = newProps || {};
        Object.freeze(newProps);
      } finally {
        delete this.__isSettingProps__;
      }
    }

    
    extendProps (updatedProps) {
      var newProps = extend$2({}, this.props, updatedProps);
      this.setProps(newProps);
    }

    
    willReceiveProps(newProps) { 
    }

    getTextContent () {
      if (this.el) {
        return this.el.textContent
      }
    }

    get textContent () {
      return this.getTextContent()
    }

    getInnerHTML () {
      if (this.el) {
        return this.el.getInnerHTML()
      }
    }

    get innerHTML () {
      return this.getInnerHTML()
    }

    getOuterHTML () {
      if (this.el) {
        return this.el.getOuterHTML()
      }
    }

    get outerHTML () {
      return this.getOuterHTML()
    }

    getAttribute (name) {
      if (this.el) {
        return this.el.getAttribute(name)
      }
    }

    setAttribute (name, val) {
      if (this.el) {
        this.el.setAttribute(name, val);
      }
      return this
    }

    getProperty (name) {
      if (this.el) {
        return this.el.getProperty(name)
      }
    }

    setProperty (name, val) {
      if (this.el) {
        this.el.setProperty(name, val);
      }
      return this
    }

    hasClass (name) {
      if (this.el) {
        return this.el.hasClass(name)
      }
    }

    addClass (name) {
      if (this.el) {
        this.el.addClass(name);
      }
      return this
    }

    removeClass (name) {
      if (this.el) {
        this.el.removeClass(name);
      }
      return this
    }

    getStyle (name) {
      if (this.el) {
        return this.el.getStyle(name)
      }
    }

    setStyle (name, val) {
      if (this.el) {
        return this.el.setStyle(name, val)
      }
      return this
    }

    getValue () {
      if (this.el) {
        return this.el.getValue()
      }
    }

    setValue (val) {
      if (this.el) {
        this.el.setValue(val);
      }
      return this
    }

    getChildCount () {
      if (!this.el) return 0
      return this.el.getChildCount()
    }

    get childNodes () {
      return this.getChildNodes()
    }

    getChildNodes () {
      if (!this.el) return []
      var childNodes = this.el.getChildNodes();
      childNodes = childNodes.map(_unwrapComp$2).filter(Boolean);
      return childNodes
    }

    getChildren () {
      if (!this.el) return []
      let children = this.el.getChildren();
      children = children.map(_unwrapComp$2).filter(Boolean);
      return children
    }

    getChildAt (pos) {
      let node = this.el.getChildAt(pos);
      if (node) {
        return _unwrapCompStrict$2(node)
      }
    }

    find (cssSelector) {
      var el = this.el.find(cssSelector);
      return _unwrapComp$2(el)
    }

    findAll (cssSelector) {
      var els = this.el.findAll(cssSelector);
      return els.map(_unwrapComp$2).filter(Boolean)
    }

    appendChild (child) {
      this.insertAt(this.getChildCount(), child);
    }

    insertAt (pos, childEl) {
      if (isString$2(childEl)) {
        childEl = new VirtualElement$2.TextNode(childEl);
      }
      if (!childEl._isVirtualElement) {
        throw new Error('Invalid argument: "child" must be a VirtualElement.')
      }
      var child = this.renderingEngine._renderChild(this, childEl);
      this.el.insertAt(pos, child.el);
      _mountChild$2(this, child);
    }

    removeAt (pos) {
      var childEl = this.el.getChildAt(pos);
      if (childEl) {
        var child = _unwrapCompStrict$2(childEl);
        _disposeChild$2(child);
        this.el.removeAt(pos);
      }
    }

    removeChild (child) {
      if (!child || !child._isComponent) {
        throw new Error('removeChild(): Illegal arguments. Expecting a Component instance.')
      }
      
      _disposeChild$2(child);
      this.el.removeChild(child.el);
    }

    replaceChild (oldChild, newChild) {
      if (!newChild || !oldChild ||
          !newChild._isComponent || !oldChild._isComponent) {
        throw new Error('replaceChild(): Illegal arguments. Expecting BrowserDOMElement instances.')
      }
      
      _disposeChild$2(oldChild);
      this.el.replaceChild(newChild.el, oldChild.el);
      if (this.isMounted()) {
        newChild.triggerDidMount(true);
      }
    }

    
    
    
    
    
    
    empty () {
      this._clear();
      return this
    }

    _clear () {
      let el = this.el;
      if (el) {
        this.getChildNodes().forEach(function (child) {
          _disposeChild$2(child);
        });
        el.empty();
      }
      this.refs = {};
      this.__foreignRefs__ = {};
    }

    remove () {
      _disposeChild$2(this);
      this.el.remove();
    }

    addEventListener () {
      throw new Error('Not supported.')
    }

    removeEventListener () {
      throw new Error('Not supported.')
    }

    insertBefore () {
      throw new Error('Not supported.')
    }

    click () {
      if (this.el) {
        this.el.click();
      }
    }

    getComponentPath () {
      let path = [];
      let comp = this;
      while (comp) {
        path.unshift(comp);
        comp = comp.getParent();
      }
      return path
    }

    _getContext () {
      var context = {};
      var parent = this.getParent();
      if (parent) {
        context = extend$2(context, parent.context);
        if (parent.getChildContext) {
          return extend$2(context, parent.getChildContext())
        }
      }
      return context
    }
  }

  Component$2.prototype._isComponent = true;

  Component$2.prototype.attr = DOMElement$2.prototype.attr;

  Component$2.prototype.htmlProp = DOMElement$2.prototype.htmlProp;

  Component$2.prototype.val = DOMElement$2.prototype.val;

  Component$2.prototype.css = DOMElement$2.prototype.css;

  Component$2.prototype.text = DOMElement$2.prototype.text;

  Component$2.prototype.append = DOMElement$2.prototype.append;

  Component$2.unwrap = _unwrapComp$2;

  Component$2.render = function (props) {
    props = props || {};
    var ComponentClass = this;
    var comp = new ComponentClass(null, props);
    comp._render();
    return comp
  };

  Component$2.mount = function (props, el) {
    if (arguments.length === 1) {
      el = props;
      props = {};
    }
    if (!el) throw new Error("'el' is required.")
    if (isString$2(el)) {
      var selector = el;
      if (platform$2.inBrowser) {
        el = window.document.querySelector(selector);
      } else {
        throw new Error('This selector is not supported on server side.')
      }
    }
    el = DefaultDOMElement$2.wrap(el);
    const ComponentClass = this;
    let comp = new ComponentClass(null, props);
    comp.mount(el);
    return comp
  };

  Component$2.getComponentForDOMElement = function (el) {
    return _unwrapComp$2(el)
  };

  Component$2.unwrapDOMElement = function (el) {
    console.warn('DEPRECATED: Use Component.getComponentForDOMElement');
    return Component$2.getComponentForDOMElement(el)
  };

  Component$2.getComponentFromNativeElement = function (nativeEl) {
    
    
    
    return _unwrapComp$2(DefaultDOMElement$2.wrap(nativeEl))
  };

  Component$2.createRenderingEngine = function (elementFactory) {
    return new RenderingEngine$2({
      componentFactory: COMPONENT_FACTORY$2,
      elementFactory
    })
  };


  function _disposeChild$2 (child) {
    child.triggerDispose();
    if (child._owner && child._ref) {
      console.assert(child._owner.refs[child._ref] === child, "Owner's ref should point to this child instance.");
      delete child._owner.refs[child._ref];
    }
  }


  function _mountChild$2 (parent, child) {
    if (parent.isMounted()) {
      child.triggerDidMount(true);
    }
    if (child._owner && child._ref) {
      child._owner.refs[child._ref] = child;
    }
  }


  function _unwrapComp$2 (el) {
    if (el) {
      if (!el._isDOMElement) el = DefaultDOMElement$2.unwrap(el);
      if (el) return el._comp
    }
  }

  function _unwrapCompStrict$2 (el) {
    let comp = _unwrapComp$2(el);
    if (!comp) throw new Error('Expecting a back-link to the component instance.')
    return comp
  }

  class ElementComponent$2 extends Component$2 {}

  ElementComponent$2.prototype._isElementComponent = true;
  ElementComponent$2.prototype._SKIP_COMPONENT_INIT = true;

  class TextNodeComponent$2 extends Component$2 {
    setTextContent (text) {
      if (!this.el) {
        throw new Error('Component must be rendered first.')
      }
      if (this.el.textContent !== text) {
        this.el.textContent = text;
      }
    }

    getChildNodes () {
      return []
    }

    getChildren () {
      return []
    }
  }

  TextNodeComponent$2.prototype._isTextNodeComponent = true;
  TextNodeComponent$2.prototype._SKIP_COMPONENT_INIT = true;

  Component$2.Element = ElementComponent$2;
  Component$2.TextNode = TextNodeComponent$2;

  function getDOMRangeFromEvent$2 (evt) {
    let range;
    let x = evt.clientX;
    let y = evt.clientY;
    
    if (document.body.createTextRange) {
      range = document.body.createTextRange();
      range.moveToPoint(x, y);
    } else if (!isNil$2(document.createRange)) {
      
      
      if (!isNil$2(evt.rangeParent)) {
        range = document.createRange();
        range.setStart(evt.rangeParent, evt.rangeOffset);
        range.collapse(true);
      
      } else if (document.caretPositionFromPoint) {
        let pos = document.caretPositionFromPoint(x, y);
        range = document.createRange();
        range.setStart(pos.offsetNode, pos.offset);
        range.collapse(true);
      
      } else if (document.caretRangeFromPoint) {
        range = document.caretRangeFromPoint(x, y);
      }
    }
    return range
  }

  const events$2 = [ 'keydown', 'keyup', 'keypress', 'mousedown', 'mouseup', 'copy' ];

  class GlobalEventHandler$2 {
    constructor (editorSession, surfaceManager) {
      this.editorSession = editorSession;
      this.surfaceManager = surfaceManager;
      this.listeners = [];
      this.initialize();
    }

    initialize () {
      if (platform$2.inBrowser) {
        let document = DefaultDOMElement$2.wrapNativeElement(window.document);
        events$2.forEach(function (name) {
          document.on(name, this._dispatch.bind(this, name), this);
        }.bind(this));
      }
    }

    dispose () {
      if (platform$2.inBrowser) {
        let document = DefaultDOMElement$2.wrapNativeElement(window.document);
        document.off(this);
      }
    }

    addEventListener (eventName, handler, options) {
      if (!options.id) {
        throw new Error("GlobalEventHandler can only be used with option 'id'")
      }
      let listener = new DOMEventListener$2(eventName, handler, options);
      this.listeners.push(listener);
    }

    removeEventListener (listener) {
      let idx = this.listeners.indexOf(listener);
      if (idx > -1) {
        this.listeners.splice(idx, 1);
      }
    }

    getEventListeners () {
      return this.listeners
    }

    _getActiveListener (eventName) {
      let editorSession = this.editorSession;
      let sel = editorSession.getSelection();
      if (sel) {
        let surfaceId = sel.surfaceId;
        for (let i = 0; i < this.listeners.length; i++) {
          let listener = this.listeners[i];
          if (listener.eventName === eventName && listener.options.id === surfaceId) {
            return listener
          }
        }
      }
    }

    _dispatch (eventName, e) {
      let listener = this._getActiveListener(eventName);
      if (listener) {
        listener.handler(e);
      }
    }
  }

  GlobalEventHandler$2.prototype.on = DOMElement$2.prototype.on;
  GlobalEventHandler$2.prototype.off = DOMElement$2.prototype.off;

  let keys$1$2 = {
    UNDEFINED: 0,
    BACKSPACE: 8,
    DELETE: 46,
    INSERT: 45,
    LEFT: 37,
    RIGHT: 39,
    UP: 38,
    DOWN: 40,
    ENTER: 13,
    RETURN: 13,
    END: 35,
    HOME: 36,
    TAB: 9,
    PAGEUP: 33,
    PAGEDOWN: 34,
    ESCAPE: 27,
    ESC: 27,
    SHIFT: 16,
    SPACE: 32,
    PLUS: 171,
    VOLUMEUP: 183,
    VOLUMEDOWN: 182,
    VOLUMEMUTE: 181,
    PRINTSCREEN: 44
  };


  for (let i = 1; i <= 24; i++) {
    keys$1$2['F' + i] = 111 + i;
  }

  function parseKeyEvent$2 (event, onlyModifiers) {
    let frags = [];
    if (event.altKey) {
      if (event.code === 'AltRight') {
        frags.push('ALTGR');
      } else {
        frags.push('ALT');
      }
    }
    if (event.ctrlKey) frags.push('CTRL');
    if (event.metaKey) frags.push('META');
    if (event.shiftKey) frags.push('SHIFT');
    if (!onlyModifiers) {
      frags.push(event.keyCode);
    }
    return frags.join('+')
  }

  class Marker$2 extends PropertyAnnotation$2 {
    _initialize (doc, props) {
      this.document = doc;
      this.type = props.type;
      if (!props.type) {
        throw new Error("'type' is mandatory")
      }
      if (!props.start) {
        throw new Error("'start' is mandatory")
      }
      if (!props.end) {
        throw new Error("'end' is mandatory")
      }
      Object.assign(this, props);
    }

    
    containsSelection (sel) {
      if (sel.isNull()) return false
      if (sel.isPropertySelection()) {
        return (isArrayEqual$2(this.start.path, sel.start.path) &&
          this.start.offset <= sel.start.offset &&
          this.end.offset >= sel.end.offset)
      } else {
        console.warn('Marker.contains() does not support other selection types.');
      }
    }

    get type () {
      return this._type
    }

    set type (type) {
      this._type = type;
    }
  }


  Marker$2.prototype._isPropertyAnnotation = false;
  Marker$2.prototype._isMarker = true;

  function TextNodeMixin$2 (SuperClass) {
    class TextNodeMixin extends SuperClass {
      getTextPath () {
        
        console.warn('DEPRECATED: use node.getPath()');
        return this.getPath()
      }

      getText () {
        return this.content
      }

      isEmpty () {
        return !this.getText()
      }

      getLength () {
        return this.getText().length
      }

      getAnnotations () {
        return this.getDocument().getIndex('annotations').get(this.getPath())
      }
    }
    return TextNodeMixin
  }

  const ENTER$2 = 1;
  const EXIT$2 = -1;
  const ANCHOR$2 = -2;






























  class Fragmenter$2 {
    constructor (options) {
      Object.assign(this, options);
    }

    start (rootContext, text, annotations) {
      if (!isString$2(text)) {
        throw new Error("Illegal argument: 'text' must be a String, but was " + text)
      }
      this._start(rootContext, text, annotations);
    }

    onText(context, text, entry) { 
    }

    
    onEnter(entry, parentContext) { 
      return null
    }

    onExit(entry, context, parentContext) { 
    }

    _enter (entry, parentContext) {
      entry.counter++;
      return this.onEnter(entry, parentContext)
    }

    _exit (entry, context, parentContext) {
      this.onExit(entry, context, parentContext);
    }

    _createText (context, text, entry) {
      this.onText(context, text, entry);
    }

    _start (rootContext, text, annotations) {
      var entries = _extractEntries$2.call(this, annotations);
      var stack = [{context: rootContext, entry: null}];

      var pos = 0;
      for (var i = 0; i < entries.length; i++) {
        var entry = entries[i];
        var textFragment = text.substring(pos, entry.pos);
        if (textFragment) {
          
          this._createText(stack[stack.length - 1].context, textFragment, entry);
        }

        pos = entry.pos;
        var stackLevel, idx, _entry;
        if (entry.mode === ENTER$2 || entry.mode === ANCHOR$2) {
          
          for (stackLevel = 1; stackLevel < stack.length; stackLevel++) {
            if (entry.level < stack[stackLevel].entry.level) {
              break
            }
          }
          
          
          for (idx = stack.length - 1; idx >= stackLevel; idx--) {
            _entry = stack[idx].entry;
            
            _entry.length = pos - _entry.pos;
            this._exit(_entry, stack[idx].context, stack[idx - 1].context);
          }
          stack.splice(stackLevel, 0, {entry: entry});
          
          for (idx = stackLevel; idx < stack.length; idx++) {
            _entry = stack[idx].entry;
            
            _entry.pos = pos;
            stack[idx].context = this._enter(_entry, stack[idx - 1].context);
          }
        }
        if (entry.mode === EXIT$2 || entry.mode === ANCHOR$2) {
          
          for (stackLevel = 1; stackLevel < stack.length; stackLevel++) {
            if (stack[stackLevel].entry.node === entry.node) {
              break
            }
          }
          for (idx = stack.length - 1; idx >= stackLevel; idx--) {
            _entry = stack[idx].entry;
            
            _entry.length = pos - _entry.pos;
            this._exit(_entry, stack[idx].context, stack[idx - 1].context);
          }
          stack.splice(stackLevel, 1);
          
          for (idx = stackLevel; idx < stack.length; idx++) {
            _entry = stack[idx].entry;
            
            _entry.pos = pos;
            stack[idx].context = this._enter(_entry, stack[idx - 1].context);
          }
        }
      }

      
      var trailingText = text.substring(pos);
      if (trailingText) {
        this._createText(rootContext, trailingText);
      }
    }
  }

  Fragmenter$2.SHOULD_NOT_SPLIT = 0;
  Fragmenter$2.NORMAL = 10;
  Fragmenter$2.ANY = 100;
  Fragmenter$2.ALWAYS_ON_TOP = Number.MAX_VALUE;







































  function _extractEntries$2 (annotations) {
    var openers = [];
    var closers = [];
    forEach$2(annotations, function (a) {
      var isAnchor = (a.isAnchor ? a.isAnchor() : false);
      
      if (isAnchor) {
        openers.push({
          mode: ANCHOR$2,
          pos: a.offset,
          id: a.id,
          level: Fragmenter$2.ALWAYS_ON_TOP,
          type: 'anchor',
          node: a,
          counter: -1,
          length: 0
        });
      } else {
        
        
        
        
        
        
        

        
        var l = Fragmenter$2.NORMAL;
        var isInline = (a.isInline ? a.isInline() : false);
        if (isInline) {
          l = Number.MAX_VALUE;
        } else if (a.constructor.hasOwnProperty('fragmentation')) {
          l = a.constructor.fragmentation;
        } else if (a.hasOwnProperty('fragmentationHint')) {
          l = a.fragmentationHint;
        }
        var startOffset = Math.min(a.start.offset, a.end.offset);
        var endOffset = Math.max(a.start.offset, a.end.offset);
        var opener = {
          pos: startOffset,
          mode: ENTER$2,
          level: l,
          id: a.id,
          type: a.type,
          node: a,
          length: 0,
          counter: -1
        };
        openers.push(opener);
        closers.push({
          pos: endOffset,
          mode: EXIT$2,
          level: l,
          id: a.id,
          type: a.type,
          node: a,
          opener: opener
        });
      }
    });

    
    openers.sort(_compareOpeners$2);
    
    for (var i = openers.length - 1; i >= 0; i--) {
      openers[i].idx = i;
    }
    closers.sort(_compareClosers$2);
    
    var entries = new Array(openers.length + closers.length);
    var idx = 0;
    var idx1 = 0;
    var idx2 = 0;
    var opener = openers[idx1];
    var closer = closers[idx2];
    while (opener || closer) {
      if (opener && closer) {
        
        if (closer.pos <= opener.pos && closer.opener !== opener) {
          entries[idx] = closer;
          idx2++;
        } else {
          entries[idx] = opener;
          idx1++;
        }
      } else if (opener) {
        entries[idx] = opener;
        idx1++;
      } else if (closer) {
        entries[idx] = closer;
        idx2++;
      }
      opener = openers[idx1];
      closer = closers[idx2];
      idx++;
    }
    return entries
  }

  function _compareOpeners$2 (a, b) {
    if (a.pos < b.pos) return -1
    if (a.pos > b.pos) return 1
    if (a.mode < b.mode) return -1
    if (a.mode > b.mode) return 1
    if (a.mode === b.mode) {
      if (a.level < b.level) return -1
      if (a.level > b.level) return 1
    }
    return 0
  }


  function _compareClosers$2 (a, b) {
    if (a.pos < b.pos) return -1
    if (a.pos > b.pos) return 1
    
    
    
    
    
    if (a.pos === a.opener.pos && b.pos === b.opener.pos) {
      if (a.opener.idx < b.opener.idx) {
        return -1
      } else {
        return 1
      }
    }
    if (a.opener.idx > b.opener.idx) return -1
    if (a.opener.idx < b.opener.idx) return 1
    return 0
  }

  function appendChild$2 (xmlNode, child) {
    insertAt$2(xmlNode, xmlNode._childNodes.length, child);
    return xmlNode
  }

  function getChildPos$2 (xmlNode, child) {
    return xmlNode._childNodes.indexOf(child.id)
  }

  function removeChild$2 (xmlNode, child) {
    const childPos = getChildPos$2(xmlNode, child);
    if (childPos >= 0) {
      removeAt$2(xmlNode, childPos);
    } else {
      throw new Error(`Node ${child.id} is not a child of ${xmlNode.id}`)
    }
    return xmlNode
  }

  function insertBefore$2 (xmlNode, newChild, ref) {
    if (!ref) {
      appendChild$2(xmlNode, newChild);
    } else {
      let pos = getChildPos$2(xmlNode, ref);
      if (pos < 0) {
        throw new Error(`Node ${newChild.id} is not a child of ${xmlNode.id}`)
      }
      insertAt$2(xmlNode, pos, newChild);
    }
    return xmlNode
  }

  function insertAt$2 (xmlNode, pos, child) {
    const length = xmlNode._childNodes.length;
    if (pos >= 0 && pos <= length) {
      const doc = xmlNode.getDocument();
      doc.update([xmlNode.id, '_childNodes'], { type: 'insert', pos, value: child.id });
    } else {
      throw new Error('Index out of bounds.')
    }
    return xmlNode
  }

  function removeAt$2 (xmlNode, pos) {
    const length = xmlNode._childNodes.length;
    if (pos >= 0 && pos < length) {
      const doc = xmlNode.getDocument();
      doc.update([xmlNode.id, '_childNodes'], { type: 'delete', pos: pos });
    } else {
      throw new Error('Index out of bounds.')
    }
    return xmlNode
  }

  function getInnerXML$2 (xmlNode) {
    if (xmlNode._childNodes) {
      return xmlNode.getChildNodes().map(child => {
        return child.toXML().serialize()
      }).join('')
    } else if (xmlNode.isText()) {
      return xmlNode.toXML().getInnerXML()
    }
    return ''
  }

  function getChildAt$2 (xmlNode, idx) {
    let childId = xmlNode._childNodes[idx];
    if (childId) {
      return xmlNode.getDocument().get(childId)
    }
  }


  function node2element$2 (node) {
    
    let dom = DefaultDOMElement$2.createDocument('xml');
    let el = _node2element$2(dom, node);
    return el
  }

  function _node2element$2 (dom, node) {
    let el;
    switch (node._elementType) {
      case 'text': {
        el = _renderTextNode$2(dom, node);
        break
      }
      case 'element':
      case 'inline-element':
      case 'container': {
        el = _renderElementNode$2(dom, node);
        break
      }
      case 'anchor':
      case 'annotation': {
        el = _createElement$1$2(dom, node);
        break
      }
      case 'external': {
        el = DefaultDOMElement$2.parseSnippet(node.xml, 'xml');
        break
      }
      default:
        throw new Error('Invalid element type.')
    }
    return el
  }

  function _createElement$1$2 (dom, node) {
    let el = dom.createElement(node.type);
    el.attr('id', node.id);
    el.attr(node.attributes);
    return el
  }

  function _renderElementNode$2 (dom, node) {
    let el = _createElement$1$2(dom, node);
    el.append(node.getChildren().map(child => _node2element$2(dom, child)));
    return el
  }

  function _renderTextNode$2 (dom, node) {
    const annos = node.getAnnotations();
    const text = node.getText();
    let el = _createElement$1$2(dom, node);
    if (annos && annos.length > 0) {
      let fragmenter = new Fragmenter$2({
        onText: (context, text) => {
          const node = context.node;
          if (node.isText() || (node.isAnnotation() && !node._isInlineNode)) {
            context.el.append(text);
          }
        },
        onEnter: (fragment) => {
          return {
            el: _node2element$2(dom, fragment.node),
            node: fragment.node
          }
        },
        onExit: (fragment, context, parentContext) => {
          parentContext.el.append(context.el);
        }
      });
      fragmenter.start({ el, node }, text, annos);
    } else {
      el.append(text);
    }
    return el
  }

  function nameWithoutNS$1$2 (name) {
    const idx = name.indexOf(':');
    if (idx > 0) {
      return name.slice(idx + 1)
    } else {
      return name
    }
  }

  class Adapter$2 extends domUtils$2.DomUtils {
    
    isTag () {
      return true
    }

    getChildren (elem) {
      return elem.getChildren() || []
    }

    getAttributeValue (elem, name) {
      return elem.getAttribute(name)
    }

    getAttributes (elem) {
      return ['id', elem.id].concat(map$3(elem.attributes, (val, key) => { return [key, val] }))
    }

    hasAttrib (elem, name) {
      return name === 'id' || elem.attributes.hasOwnProperty(name)
    }

    getName (elem) {
      return elem.type
    }

    getNameWithoutNS (elem) {
      return nameWithoutNS$1$2(this.getName(elem))
    }

    getText (elem) {
      if (elem._elementType === 'text') {
        return elem.getText()
      }
      
      return ''
    }
  }

  var cssSelectAdapter$2 = new Adapter$2();

  class XMLDocumentNode$2 extends DocumentNode$2 {
    _initialize (doc, props) {
      
      
      
      if (props.attributes) {
        delete props.attributes.id;
      }
      super._initialize(doc, props);
    }

    toXML () {
      return node2element$2(this)
    }

    
    findChild (tagName) {
      const children = this.getChildren();
      for (let i = 0; i < children.length; i++) {
        const child = children[i];
        if (child.type === tagName) return child
      }
    }

    find (cssSelector) {
      return cssSelect$2.selectOne(cssSelector, this, { xmlMode: true, adapter: cssSelectAdapter$2 })
    }

    findAll (cssSelector) {
      return cssSelect$2.selectAll(cssSelector, this, { xmlMode: true, adapter: cssSelectAdapter$2 })
    }

    isContainer () {
      return false
    }

    
    isBlock () {
      const parentNode = this.parentNode;
      return (parentNode && parentNode.isContainer())
    }

    get children () {
      return this.getChildren()
    }

    getChildren () {
      
      return this.getChildNodes()
    }

    getChildNodes () {
      if (this._childNodes) {
        return getNodes$2(this.getDocument(), this._childNodes)
      } else {
        return []
      }
    }

    getChildCount () {
      if (this._childNodes) {
        return this._childNodes.length
      } else {
        return 0
      }
    }

    getChildPosition (child) {
      if (!child || !this._childNodes) return -1
      return this._childNodes.indexOf(child.id)
    }

    getChildNodeIterator () {
      return new ArrayIterator$2(this.getChildNodes())
    }

    getFirstChild () {
      if (this._childNodes) {
        return this.getDocument().get(this._childNodes[0])
      }
    }

    getLastChild () {
      if (this._childNodes) {
        return this.getDocument().get(last$1$2(this._childNodes))
      }
    }

    getParent () {
      return this.parentNode
    }

    setParent (parentNode) {
      this.parentNode = parentNode;
    }

    get tagName () {
      return this.type
    }

    
    get parent () {
      return this.parentNode
    }

    setAttribute (name, val) {
      if (name === 'id') {
        throw new Error("'id' is read-only and can not be changed")
      }
      this.getDocument().set([this.id, 'attributes', name], val);
      return this
    }

    getAttribute (name) {
      if (name === 'id') return this.id
      return this.attributes[name]
    }

    removeAttribute (name) {
      if (this.attributes.hasOwnProperty(name)) {
        this.getDocument().set([this.id, 'attributes', name], undefined);
        delete this.attributes[name];
      }
      return this
    }

    getAttributes () {
      return clone$2(this.attributes)
    }

    getElementSchema () {
      return this.getDocument().getElementSchema(this.type)
    }

    serialize () {
      return this.toXML()
    }

    isTextNode () {
      return false
    }

    isElementNode () {
      return false
    }

    getInnerXML () {
      return getInnerXML$2(this)
    }
  }

  XMLDocumentNode$2.prototype.attr = DOMElement$2.prototype.attr;

  XMLDocumentNode$2.schema = {
    attributes: { type: 'object', default: {} }
  };

  class XMLTextElement$2 extends TextNodeMixin$2(XMLDocumentNode$2) {
    getPath () {
      return [this.id, 'content']
    }

    getText () {
      return this.content
    }

    
    getChildren () {
      const annos = this.getAnnotations();
      
      annos.sort(_byStartOffset$2);
      return annos
    }

    setText (text) {
      const doc = this.getDocument();
      const path = this.getPath();
      const oldText = this.getText();
      
      if (oldText.length > 0) {
        doc.update(path, { type: 'delete', start: 0, end: oldText.length });
      }
      doc.update(path, { type: 'insert', start: 0, text });
      return this
    }

    

    getTextContent () {
      return this.getText()
    }

    setTextContent (text) {
      return this.setText(text)
    }

    get textContent () {
      return this.getText()
    }

    set textContent (text) {
      this.setText(text);
    }

    appendChild(child) { 
      
      
      
      
      
      
      
      throw new Error('This is not implemented yet.')
    }

    removeChild(child) { 
      
      throw new Error('This is not implemented yet.')
    }

    
    isElementNode () {
      return true
    }
  }

  XMLTextElement$2.prototype._isXMLTextElement = true;

  XMLTextElement$2.prototype.text = DOMElement$2.prototype.text;

  XMLTextElement$2.prototype._elementType = 'text';

  XMLTextElement$2.isText = true;
  XMLTextElement$2.isBlock = true;

  XMLTextElement$2.type = 'text';

  XMLTextElement$2.schema = {
    content: 'text'
  };

  function _byStartOffset$2 (a, b) {
    return a.start.offset - b.start.offset
  }

  class XMLElementNode$2 extends XMLDocumentNode$2 {
    appendChild (child) {
      return appendChild$2(this, child)
    }

    removeChild (child) {
      return removeChild$2(this, child)
    }

    insertBefore (newChild, ref) {
      return insertBefore$2(this, newChild, ref)
    }

    insertAt (pos, child) {
      return insertAt$2(this, pos, child)
    }

    removeAt (pos) {
      return removeAt$2(this, pos)
    }

    getChildAt (idx) {
      return getChildAt$2(this, idx)
    }

    isElementNode () {
      return true
    }

    
  }

  XMLElementNode$2.prototype.append = DOMElement$2.prototype.append;

  XMLElementNode$2.prototype._elementType = 'element';

  XMLElementNode$2.type = 'element';

  XMLElementNode$2.schema = {
    _childNodes: { type: ['array', 'id'], default: [], owned: true }
  };

  XMLElementNode$2.isBlock = true;

  class XMLAnnotationNode$2 extends AnnotationMixin$2(XMLDocumentNode$2) {
    
    get parentNode () {
      const path = this.start.path;
      const doc = this.getDocument();
      return doc.get(path[0])
    }

    isPropertyAnnotation () {
      return true
    }
  }

  XMLAnnotationNode$2.prototype._elementType = 'annotation';


  XMLAnnotationNode$2.prototype._isPropertyAnnotation = true;

  XMLAnnotationNode$2.type = '@annotation';


  XMLAnnotationNode$2.schema = {};

  class XMLAnchorNode$2 extends XMLDocumentNode$2 {
    
    get parentNode () {
      const path = this.coor.start.path;
      const doc = this.getDocument();
      return doc.get(path[0])
    }
  }

  XMLAnchorNode$2.prototype._elementType = 'anchor';

  XMLAnchorNode$2.type = 'anchor';

  XMLAnchorNode$2.schema = {
    coor: { type: 'coordinate', optional: true }
  };

  class XMLInlineElementNode$2 extends XMLAnnotationNode$2 {
    
    get parentNode () {
      const path = this.start.path;
      if (path[0]) {
        const doc = this.getDocument();
        return doc.get(path[0])
      }
      return this._parentNode
    }

    set parentNode (parent) {
      const path = this.start.path;
      if (path[0]) {
        throw new Error('parent of inline-element is implicitly given')
      }
      this._parentNode = parent;
    }

    getChildAt (idx) {
      return getChildAt$2(this, idx)
    }

    appendChild (child) {
      return appendChild$2(this, child)
    }

    removeChild (child) {
      return removeChild$2(this, child)
    }

    insertBefore (newChild, ref) {
      return insertBefore$2(this, newChild, ref)
    }

    insertAt (pos, child) {
      return insertAt$2(this, pos, child)
    }

    removeAt (pos) {
      return removeAt$2(this, pos)
    }

    getInnerXML () {
      return this.getChildren().map(child => {
        return child.toXML().outerHTML
      }).join('')
    }
  }

  XMLInlineElementNode$2.prototype._elementType = 'inline-element';



  XMLInlineElementNode$2.prototype._isInlineNode = true;
  XMLInlineElementNode$2.isInline = true;

  XMLInlineElementNode$2.type = 'inline-element';

  XMLInlineElementNode$2.schema = {
    _childNodes: { type: ['array', 'id'], default: [], owned: true }
  };

  class XMLExternalNode$2 extends XMLDocumentNode$2 {}

  XMLExternalNode$2.prototype._elementType = 'external';

  XMLExternalNode$2.type = 'external';

  XMLExternalNode$2.schema = {
    xml: { type: 'string', default: '' }
  };

  XMLExternalNode$2.isBlock = true;

  class XMLContainerNode$2 extends ContainerMixin$2(XMLElementNode$2) {
    getContentPath () {
      return [this.id, '_childNodes']
    }

    getContent () {
      return this._childNodes
    }

    isContainer () {
      return true
    }

    appendChild (child) {
      super.show(child.id);
    }
  }

  XMLContainerNode$2.prototype._elementType = 'container';

  XMLContainerNode$2.type = 'container';

  XMLContainerNode$2.schema = {};

  XMLContainerNode$2.isBlock = true;

  function createCountingIdGenerator$2 () {
    var counters = {};
    return function uuid (prefix) {
      if (!counters.hasOwnProperty(prefix)) {
        counters[prefix] = 1;
      }
      var result = [prefix, '-', counters[prefix]++].join('');
      return result
    }
  }

  const WS_LEFT$2 = /^\s+/g;


  const WS_RIGHT$2 = /\s+$/g;
  const WS_ALL$2 = /\s+/g;


  const SPACE$2 = ' ';
  const TABS_OR_NL$2 = /[\t\n\r]+/g;

  const INVISIBLE_CHARACTER$2 = '\u200B';


  class DOMImporter$2 {
    constructor (config, context) {
      this.context = context || {};

      if (!config.schema) {
        throw new Error('"config.schema" is mandatory')
      }
      if (!config.converters) {
        throw new Error('"config.converters" is mandatory')
      }

      this.config = Object.assign({ idAttribute: 'id' }, config);
      this.schema = config.schema;
      this.converters = config.converters;
      this.state = null;

      this._defaultBlockConverter = null;
      this._allConverters = [];
      this._blockConverters = [];
      this._propertyAnnotationConverters = [];

      this.state = new DOMImporter$2.State();

      this._initialize();
    }

    
    _initialize () {
      const schema = this.schema;
      const defaultTextType = schema.getDefaultTextType();
      const converters = this.converters;
      for (let i = 0; i < converters.length; i++) {
        let converter;
        if (typeof converters[i] === 'function') {
          const Converter = converters[i];
          converter = new Converter();
        } else {
          converter = converters[i];
        }
        if (!converter.type) {
          throw new Error('Converter must provide the type of the associated node.')
        }
        if (!converter.matchElement && !converter.tagName) {
          throw new Error('Converter must provide a matchElement function or a tagName property.')
        }
        if (!converter.matchElement) {
          converter.matchElement = this._defaultElementMatcher.bind(converter);
        }
        const NodeClass = schema.getNodeClass(converter.type);
        if (!NodeClass) {
          throw new Error('No node type defined for converter')
        }
        if (!this._defaultBlockConverter && defaultTextType === converter.type) {
          this._defaultBlockConverter = converter;
        }
        this._allConverters.push(converter);
        
        if (NodeClass.prototype._isPropertyAnnotation) {
          this._propertyAnnotationConverters.push(converter);
        } else {
          this._blockConverters.push(converter);
        }
      }
    }

    dispose () {
      if (this.state.doc) {
        this.state.doc.dispose();
      }
    }

    
    reset () {
      if (this.state.doc) {
        this.state.doc.dispose();
      }
      this.state.reset();
      this.state.doc = this._createDocument();
    }

    getDocument () {
      return this.state.doc
    }

    
    convertContainer (elements, containerId) {
      if (!this.state.doc) this.reset();
      const state = this.state;
      const iterator = new ArrayIterator$2(elements);
      const nodeIds = [];
      while (iterator.hasNext()) {
        const el = iterator.next();
        let node;
        const blockTypeConverter = this._getConverterForElement(el, 'block');
        if (blockTypeConverter) {
          state.pushContext(el.tagName, blockTypeConverter);
          let nodeData = this._createNodeData(el, blockTypeConverter.type);
          nodeData = blockTypeConverter.import(el, nodeData, this) || nodeData;
          node = this._createNode(nodeData);
          let context = state.popContext();
          context.annos.forEach((a) => {
            this._createNode(a);
          });
        } else if (el.isCommentNode()) {
          continue
        } else {
          
          if (el.isTextNode() && /^\s*$/.exec(el.textContent)) continue
          
          
          iterator.back();
          node = this._wrapInlineElementsIntoBlockElement(iterator);
        }
        if (node) {
          nodeIds.push(node.id);
        }
      }
      return this._createNode({
        type: '@container',
        id: containerId,
        nodes: nodeIds
      })
    }

    
    convertElement (el) {
      if (!this.state.doc) this.reset();
      let isTopLevel = !this.state.isConverting;
      if (isTopLevel) {
        this.state.isConverting = true;
      }

      let nodeData, annos;
      const converter = this._getConverterForElement(el);
      if (converter) {
        const NodeClass = this.schema.getNodeClass(converter.type);
        nodeData = this._createNodeData(el, converter.type);
        this.state.pushContext(el.tagName, converter);
        
        
        
        
        
        if (NodeClass.isInline) {
          nodeData = this._convertInlineNode(el, nodeData, converter);
        } else if (NodeClass.prototype._isPropertyAnnotation) {
          nodeData = this._convertPropertyAnnotation(el, nodeData);
        } else {
          nodeData = converter.import(el, nodeData, this) || nodeData;
        }
        let context = this.state.popContext();
        annos = context.annos;
      } else {
        throw new Error('No converter found for ' + el.tagName)
      }
      
      const node = this._createNode(nodeData);
      
      annos.forEach((a) => {
        this._createNode(a);
      });

      
      
      if (this.config['stand-alone'] && isTopLevel) {
        this.state.isConverting = false;
        this.reset();
      }
      return node
    }

    
    annotatedText (el, path, options = {}) {
      if (!path) {
        throw new Error('path is mandatory')
      }
      const state = this.state;
      const context = last$1$2(state.contexts);
      
      
      if (!context) {
        throw new Error('This should be called from within an element converter.')
      }
      
      const oldPreserveWhitespace = state.preserveWhitespace;
      if (options.preserveWhitespace) {
        state.preserveWhitespace = true;
      }
      state.stack.push({ path: path, offset: 0, text: '', annos: [] });
      
      
      
      this.state.lastChar = '';
      const iterator = this.getChildNodeIterator(el);
      const text = this._annotatedText(iterator);
      
      
      const top = state.stack.pop();
      context.annos = context.annos.concat(top.annos);

      
      state.preserveWhitespace = oldPreserveWhitespace;

      return text
    }

    
    plainText (el) {
      var state = this.state;
      var text = el.textContent;
      if (state.stack.length > 0) {
        var context = last$1$2(state.stack);
        context.offset += text.length;
        context.text += context.text.concat(text);
      }
      return text
    }

    
    _customText (text) {
      var state = this.state;
      if (state.stack.length > 0) {
        var context = last$1$2(state.stack);
        context.offset += text.length;
        context.text += context.text.concat(text);
      }
      return text
    }

    
    nextId (prefix) {
      
      
      
      
      return this.state.uuid(prefix)
    }

    _getNextId (dom, type) {
      let id = this.nextId(type);
      while (this.state.ids[id] || dom.find('#' + id)) {
        id = this.nextId(type);
      }
      return id
    }

    _getIdForElement (el, type) {
      let id = el.getAttribute(this.config.idAttribute);
      if (id && !this.state.ids[id]) return id
      return this._getNextId(el.getOwnerDocument(), type)
    }

    
    
    _createDocument () {
      
      const schema = this.config.schema;
      const DocumentClass = schema.getDocumentClass();
      return new DocumentClass(schema)
    }

    _convertPropertyAnnotation (el, nodeData) {
      const path = [nodeData.id, '_content'];
      
      
      
      nodeData._content = this.annotatedText(el, path);
      nodeData.start = { path, offset: 0 };
      nodeData.end = { offset: nodeData._content.length };
      return nodeData
    }

    _convertInlineNode (el, nodeData, converter) {
      const path = [nodeData.id, '_content'];
      if (converter.import) {
        nodeData = converter.import(el, nodeData, this) || nodeData;
      }
      nodeData._content = '$';
      nodeData.start = { path, offset: 0 };
      nodeData.end = { offset: 1 };
      return nodeData
    }

    _createNodeData (el, type) {
      if (!type) {
        throw new Error('type is mandatory.')
      }
      let nodeData = {
        type,
        id: this._getIdForElement(el, type)
      };
      this.state.ids[nodeData.id] = true;
      return nodeData
    }

    _createNode (nodeData) {
      let doc = this.state.doc;
      
      
      
      let node = doc.get(nodeData.id);
      if (node) {
        
        doc.delete(node.id);
      }
      return doc.create(nodeData)
    }

    getChildNodeIterator (el) {
      return el.getChildNodeIterator()
    }

    _defaultElementMatcher (el) {
      return el.is(this.tagName)
    }

    
    _annotatedText (iterator) {
      const state = this.state;
      const context = last$1$2(state.stack);
      
      if (!context) {
        throw new Error('Illegal state: context is null.')
      }
      while (iterator.hasNext()) {
        var el = iterator.next();
        var text = '';
        
        
        if (el.isTextNode()) {
          text = this._prepareText(el.textContent);
          if (text.length) {
            
            
            context.text = context.text.concat(text);
            context.offset += text.length;
          }
        } else if (el.isCommentNode()) {
          
          continue
        } else if (el.isElementNode()) {
          const annoConverter = this._getConverterForElement(el, 'inline');
          
          if (!annoConverter) {
            
            if (!this.IGNORE_DEFAULT_WARNINGS) {
              console.warn('Unsupported inline element. We will not create an annotation for it, but process its children to extract annotated text.', el.outerHTML);
            }
            
            
            const iterator = this.getChildNodeIterator(el);
            this._annotatedText(iterator);
            continue
          }
          
          
          
          var startOffset = context.offset;
          const annoType = annoConverter.type;
          const AnnoClass = this.schema.getNodeClass(annoType);
          let annoData = this._createNodeData(el, annoType);
          
          let stackFrame = {
            path: context.path,
            offset: startOffset,
            text: '',
            annos: []
          };
          state.stack.push(stackFrame);
          
          if (annoConverter.import) {
            state.pushContext(el.tagName, annoConverter);
            annoData = annoConverter.import(el, annoData, this) || annoData;
            state.popContext();
          }
          
          
          
          
          if (AnnoClass.isInline) {
            this._customText(INVISIBLE_CHARACTER$2);
            
            
            
            state.lastChar = '';
          } else {
            
            
            
            const iterator = this.getChildNodeIterator(el);
            this._annotatedText(iterator);
          }
          
          state.stack.pop();
          context.offset = stackFrame.offset;
          context.text = context.text.concat(stackFrame.text);
          
          const endOffset = context.offset;
          annoData.start = {
            path: context.path.slice(0),
            offset: startOffset
          };
          annoData.end = {
            offset: endOffset
          };
          
          let parentFrame = last$1$2(state.stack);
          parentFrame.annos = parentFrame.annos.concat(stackFrame.annos, annoData);
        } else {
          console.warn('Unknown element type. Taking plain text.', el.outerHTML);
          text = this._prepareText(el.textContent);
          context.text = context.text.concat(text);
          context.offset += text.length;
        }
      }
      
      return context.text
    }

    _getConverterForElement (el, mode) {
      var converters;
      if (mode === 'block') {
        if (!el.tagName) return null
        converters = this._blockConverters;
      } else if (mode === 'inline') {
        converters = this._propertyAnnotationConverters;
      } else {
        converters = this._allConverters;
      }
      var converter = null;
      for (var i = 0; i < converters.length; i++) {
        if (this._converterCanBeApplied(converters[i], el)) {
          converter = converters[i];
          break
        }
      }
      return converter
    }

    _converterCanBeApplied (converter, el) {
      return converter.matchElement(el, this)
    }

    
    _wrapInlineElementsIntoBlockElement (childIterator) {
      if (!childIterator.hasNext()) return

      const converter = this._defaultBlockConverter;
      if (!converter) {
        throw new Error('Wrapping inline elements automatically is not supported in this schema.')
      }

      let dom = childIterator.peek().getOwnerDocument();
      let wrapper = dom.createElement('wrapper');
      while (childIterator.hasNext()) {
        const el = childIterator.next();
        
        const blockTypeConverter = this._getConverterForElement(el, 'block');
        if (blockTypeConverter) {
          childIterator.back();
          break
        }
        wrapper.append(el.clone());
      }
      const type = this.schema.getDefaultTextType();
      const id = this._getNextId(dom, type);
      let nodeData = { type, id };
      this.state.pushContext('wrapper', converter);
      nodeData = converter.import(wrapper, nodeData, this) || nodeData;
      let context = this.state.popContext();
      let annos = context.annos;
      
      const node = this._createNode(nodeData);
      
      annos.forEach((a) => {
        this._createNode(a);
      });
      return node
    }

    
    
    
    
    _prepareText (text) {
      const state = this.state;
      if (state.preserveWhitespace) {
        return text
      }
      var repl = SPACE$2;
      
      text = text.replace(TABS_OR_NL$2, '');
      
      
      
      
      
      if (state.lastChar === SPACE$2) {
        
        text = text.replace(WS_LEFT$2, '');
      } else {
        text = text.replace(WS_LEFT$2, repl);
      }
      text = text.replace(WS_RIGHT$2, repl);
      
      
      
      if (this.config.REMOVE_INNER_WS || state.removeInnerWhitespace) {
        text = text.replace(WS_ALL$2, SPACE$2);
      }
      state.lastChar = text[text.length - 1] || state.lastChar;
      return text
    }

    
    _trimTextContent (el) {
      var nodes = el.getChildNodes();
      var firstNode = nodes[0];
      var lastNode = last$1$2(nodes);
      var text, trimmed;
      
      if (firstNode && firstNode.isTextNode()) {
        text = firstNode.textContent;
        trimmed = this._trimLeft(text);
        firstNode.textContent = trimmed;
      }
      if (lastNode && lastNode.isTextNode()) {
        text = lastNode.textContent;
        trimmed = this._trimRight(text);
        lastNode.textContent = trimmed;
      }
      return el
    }

    _trimLeft (text) {
      return text.replace(WS_LEFT$2, '')
    }

    _trimRight (text) {
      return text.replace(WS_RIGHT$2, '')
    }
  }

  class DOMImporterState$2 {
    constructor () {
      this.reset();
    }

    reset () {
      this.preserveWhitespace = false;
      this.nodes = [];
      this.annotations = [];
      this.containerId = null;
      this.container = [];
      this.ids = {};
      
      this.contexts = [];
      
      this.stack = [];
      this.lastChar = '';
      this.skipTypes = {};
      this.ignoreAnnotations = false;
      this.isConverting = false;

      
      
      this.uuid = createCountingIdGenerator$2();
    }

    pushContext (tagName, converter) {
      this.contexts.push({ tagName: tagName, converter: converter, annos: [] });
    }

    popContext () {
      return this.contexts.pop()
    }

    getCurrentContext () {
      return last$1$2(this.contexts)
    }
  }

  DOMImporter$2.State = DOMImporterState$2;

  DOMImporter$2.INVISIBLE_CHARACTER = INVISIBLE_CHARACTER$2;

  const START$4 = 'START';
  const END$4 = 'END';
  const EPSILON$4 = 'EPSILON';
  const TEXT$1$2 = 'TEXT';

  class DFA$2 {
    constructor (transitions) {
      if (!transitions || Object.keys(transitions).length === 0) {
        transitions = { START: { EPSILON: END$4 } };
      }
      this.transitions = transitions;
    }

    consume (state, id) {
      const T = this.transitions;
      
      
      if (!T[state]) return -1
      let nextState = T[state][id];
      if (nextState !== undefined) {
        return nextState
      }
      while (T[state][EPSILON$4] !== undefined) {
        state = T[state][EPSILON$4];
        if (state === END$4) {
          return -1
        }
        nextState = T[state][id];
        if (nextState !== undefined) {
          return nextState
        }
      }
      return -1
    }

    canConsume (state, id) {
      let nextState = this.consume(state, id);
      return (nextState !== -1)
    }

    isFinished (state) {
      const T = this.transitions;
      if (state === 'END') return true
      
      if (!T[state]) return false
      while (T[state][EPSILON$4] !== undefined) {
        state = T[state][EPSILON$4];
        if (state === 'END') return true
      }
      return false
    }

    

    
    _tokensByPath () {
      const result = [];
      const transitions = this.transitions;
      if (!transitions) return []

      
      let first = {};
      forEach$2(transitions[START$4], (to, token) => {
        if (!first[to]) first[to] = [];
        first[to].push(token);
      });

      let visited = {START: true, END: true};
      forEach$2(first, (tokens, state) => {
        
        
        
        
        let _siblings = {};
        tokens.forEach((t) => {
          if (t !== EPSILON$4) {
            _siblings[t] = true;
          }
        });
        let stack = [state];
        while (stack.length > 0) {
          let from = stack.pop();
          if (state === END$4) continue
          visited[from] = true;
          let T = transitions[from];
          if (!T) throw new Error(`Internal Error: no transition from state ${from}`)
          let tokens = Object.keys(T);
          for (let i = 0; i < tokens.length; i++) {
            const token = tokens[i];
            const to = T[token];
            if (!visited[to]) stack.push(to);
            if (token !== EPSILON$4) {
              _siblings[token] = true;
            }
          }
        }
        let _siblingTokens = Object.keys(_siblings);
        if (_siblingTokens.length > 0) {
          result.push(_siblingTokens);
        }
      });
      return result
    }
  }

  DFA$2.START = START$4;
  DFA$2.END = END$4;
  DFA$2.EPSILON = EPSILON$4;
  DFA$2.TEXT = TEXT$1$2;

  function _isTextNodeEmpty$2 (el) {
    return Boolean(/^\s*$/.exec(el.textContent))
  }

  const START$1$2 = DFA$2.START;
  const END$1$2 = DFA$2.END;
  const EPSILON$1$2 = DFA$2.EPSILON;


  class DFABuilder$2 {
    constructor (transitions) {
      this.transitions = transitions;
    }

    addTransition (from, to, tokens) {
      if (!this.transitions) this.transitions = {};
      if (!isArray$3(tokens)) tokens = [tokens];
      tokens.forEach(token => _addTransition$2(this.transitions, from, to, token));
      return this
    }

    
    append (other) {
      if (this.transitions && other.transitions) {
        let t1 = cloneDeep$2(this.transitions);
        let t2 = cloneDeep$2(other.transitions);
        
        
        let firstIsOptional = Boolean(t1[START$1$2][EPSILON$1$2]);
        let secondIsOptional = Boolean(t2[START$1$2][EPSILON$1$2]);

        if (firstIsOptional) {
          
          
          delete t1[START$1$2][EPSILON$1$2];
        }
        
        
        let newState = uuid$2();
        
        
        forEach$2(t1, (T) => {
          forEach$2(T, (to, token) => {
            if (to === END$1$2) {
              T[token] = newState;
            }
          });
        });
        
        
        if (firstIsOptional) {
          forEach$2(t2[START$1$2], (to, token) => {
            _addTransition$2(t1, START$1$2, to, token);
          });
        }
        
        
        t2[newState] = t2[START$1$2];
        forEach$2(t2, (T) => {
          forEach$2(T, (to, token) => {
            if (to === START$1$2) {
              T[token] = newState;
            }
          });
        });
        delete t2[START$1$2];
        
        forEach$2(t2, (T, from) => {
          forEach$2(T, (to, token) => {
            _addTransition$2(t1, from, to, token);
          });
        });
        
        
        if (firstIsOptional && secondIsOptional) {
          _addTransition$2(t1, START$1$2, END$1$2, EPSILON$1$2);
        }
        this.transitions = t1;
      } else if (other.transitions) {
        this.transitions = cloneDeep$2(other.transitions);
      }
      return this
    }

    
    merge (other) {
      if (this.transitions && other.transitions) {
        let t1 = this.transitions;
        let t2 = other.transitions;
        forEach$2(t2, (T, from) => {
          forEach$2(T, (to, token) => {
            _addTransition$2(t1, from, to, token);
          });
        });
      } else if (other.transitions) {
        this.transitions = cloneDeep$2(other.transitions);
      }
      return this
    }

    
    optional () {
      let dfa = new DFABuilder$2(cloneDeep$2(this.transitions));
      if (this.transitions) {
        dfa.addTransition(START$1$2, END$1$2, EPSILON$1$2);
      }
      return dfa
    }

    
    kleene () {
      let dfa = this.plus();
      return dfa.optional()
    }

    
    plus () {
      let dfa;
      if (this.transitions) {
        let t1 = cloneDeep$2(this.transitions);
        
        
        
        
        const isOptional = Boolean(t1[START$1$2][EPSILON$1$2]);
        delete t1[START$1$2][EPSILON$1$2];
        
        
        let newState = uuid$2();
        forEach$2(t1, (T) => {
          forEach$2(T, (to, token) => {
            if (to === END$1$2) {
              T[token] = newState;
            }
          });
        });
        
        _addTransition$2(t1, newState, END$1$2, EPSILON$1$2);
        
        forEach$2(t1[START$1$2], (to, token) => {
          _addTransition$2(t1, newState, to, token);
        });
        
        if (isOptional) {
          _addTransition$2(t1, START$1$2, END$1$2, EPSILON$1$2);
        }
        dfa = new DFABuilder$2(t1);
      } else {
        dfa = new DFABuilder$2(cloneDeep$2(this.transitions));
      }
      return dfa
    }

    toJSON () {
      return cloneDeep$2(this.transitions)
    }

    
    copy () {
      let t = cloneDeep$2(this.transitions);
      if (this.transitions) {
        let states = Object.keys(t);
        let map = { START: START$1$2, END: END$1$2 };
        states.forEach((id) => {
          if (id === START$1$2 || id === END$1$2) return
          map[id] = uuid$2();
        });
        forEach$2(t, (T, from) => {
          if (from !== START$1$2 && from !== END$1$2) {
            t[map[from]] = T;
            delete t[from];
          }
          forEach$2(T, (to, token) => {
            if (to !== START$1$2 && to !== END$1$2) {
              T[token] = map[to];
            }
          });
        });
      }
      return new DFABuilder$2(t)
    }
  }

  DFABuilder$2.singleToken = function (token) {
    let dfa = new DFABuilder$2();
    dfa.addTransition(START$1$2, END$1$2, token);
    return dfa
  };

  function _addTransition$2 (transitions, from, to, token) {
    let T = transitions[from];
    if (!T) {
      transitions[from] = T = {};
    }
    if (token === EPSILON$1$2 && from === START$1$2 && to !== END$1$2) {
      throw new Error('The only EPSILON transition from START must be START->END')
    }
    if (T[token] && T[token] !== to) {
      console.error('Token %s already used. Ignoring this transition.', token);
      return
      
    }
    T[token] = to;
  }

  const { START: START$2$2, END: END$2$2, TEXT: TEXT$3$2, EPSILON: EPSILON$2$2 } = DFA$2;



  class Expression$2 {
    constructor (name, root) {
      this.name = name;
      this.root = root;

      this._initialize();
    }

    _initialize () {
      this._compile();
    }

    toString () {
      return this.root.toString()
    }

    copy () {
      return this.root.copy()
    }

    toJSON () {
      return {
        name: this.name,
        content: this.root.toJSON()
      }
    }

    isAllowed (tagName) {
      return Boolean(this._allowedChildren[tagName])
    }

    
    _normalize () {
      this.root._normalize();
    }

    
    _compile () {
      
      this.root._compile();
    }

    _describeError (state, token) {
      let msg = [];
      if (token !== TEXT$3$2) {
        if (!this.isAllowed(token)) {
          msg.push(`<${token}> is not valid in <${this.name}>\nSchema: ${this.toString()}`);
        } else {
          
          msg.push(`<${token}> is not allowed at the current position in <${this.name}>.\n${this.toString()}`);
          
          
          
        }
      } else {
        msg.push(`TEXT is not allowed at the current position: ${state.trace.join(',')}\n${this.toString()}`);
      }
      return msg.join('')
    }
  }

  Expression$2.fromJSON = function (data) {
    const name = data.name;
    const root = _fromJSON$2(data.content);
    return createExpression$2(name, root)
  };

  function createExpression$2 (name, root) {
    if (root instanceof Interleave$2) {
      return new InterleaveExpr$2(name, root)
    } else {
      return new DFAExpr$2(name, root)
    }
  }

  class DFAExpr$2 extends Expression$2 {
    
    getInitialState () {
      return {
        dfaState: START$2$2,
        errors: [],
        trace: []
      }
    }

    canConsume (state, token) {
      return this.dfa.canConsume(state.dfaState, token)
    }

    consume (state, token) {
      const dfa = this.dfa;
      let oldState = state.dfaState;
      let newState = dfa.consume(oldState, token);
      state.dfaState = newState;
      if (newState === -1) {
        state.errors.push({
          msg: this._describeError(state, token),
          
          
          el: state.el
        });
        return false
      } else {
        state.trace.push(token);
        return true
      }
    }

    isFinished (state) {
      return this.dfa.isFinished(state.dfaState)
    }

    _initialize () {
      super._initialize();

      this._computeAllowedChildren();
    }

    _compile () {
      super._compile();
      this.dfa = new DFA$2(this.root.dfa.transitions);
    }

    _computeAllowedChildren () {
      this._allowedChildren = _collectAllTokensFromDFA$2(this.dfa);
    }

    _findInsertPos (el, newTag, mode) {
      const root = this.root;
      if (root instanceof Sequence$2) {
        return this._findInsertPosInSequence(el, newTag, mode)
      } else if (root instanceof Plus$2 || root instanceof Kleene$2) {
        if (mode === 'first') {
          return 0
        } else {
          return el.childNodes.length
        }
      }
    }

    _isValid (_tokens) {
      let state = this.getInitialState();
      for (let i = 0; i < _tokens.length; i++) {
        const token = _tokens[i];
        
        
        
        if (!token) continue
        if (!this.consume(state, token)) {
          return false
        }
      }
      return this.isFinished(state)
    }

    _findInsertPosInSequence (el, newTag, mode) {
      const childNodes = el.getChildNodes();
      
      
      
      const tokens = [];
      childNodes.forEach((child) => {
        
        
        
        const tagName = child.tagName;
        if (!tagName) {
          if (child._isDOMElement && child.isTextNode() && !_isTextNodeEmpty$2(child)) {
            tokens.push(TEXT$3$2);
          } else {
            tokens.push(null);
          }
        } else {
          tokens.push(tagName);
        }
      });
      const L = tokens.length;
      const self = this;
      function _isValid (pos) {
        let _tokens = tokens.slice(0);
        _tokens.splice(pos, 0, newTag);
        return self._isValid(_tokens)
      }
      if (mode === 'first') {
        for (let pos = 0; pos <= L; pos++) {
          if (_isValid(pos)) {
            return pos
          }
        }
      } else {
        for (let pos = L; pos >= 0; pos--) {
          if (_isValid(pos)) {
            return pos
          }
        }
      }
      return -1
    }
  }

  function _collectAllTokensFromDFA$2 (dfa) {
    
    const children = {};
    if (dfa.transitions) {
      forEach$2(dfa.transitions, (T) => {
        Object.keys(T).forEach((tagName) => {
          if (tagName === EPSILON$2$2) return
          children[tagName] = true;
        });
      });
    }
    return children
  }

  class InterleaveExpr$2 extends Expression$2 {
    getInitialState () {
      const dfas = this.dfas;
      const dfaStates = new Array(dfas.length);
      dfaStates.fill(START$2$2);
      return {
        dfaStates,
        errors: [],
        trace: [],
        
        lastDFA: 0
      }
    }

    canConsume (state, token) {
      return (this._findNextDFA(state, token) >= 0)
    }

    consume (state, token) {
      const idx = this._findNextDFA(state, token);
      if (idx < 0) {
        state.errors.push({
          msg: this._describeError(state, token)
        });
        return false
      } else {
        const dfa = this.dfas[idx];
        const oldState = state.dfaStates[idx];
        const newState = dfa.consume(oldState, token);
        state.dfaStates[idx] = newState;
        state.trace.push(token);
        return true
      }
    }

    isFinished (state) {
      const dfas = this.dfas;
      for (let i = 0; i < dfas.length; i++) {
        const dfa = dfas[i];
        const dfaState = state.dfaStates[i];
        if (!dfa.isFinished(dfaState)) {
          return false
        }
      }
      return true
    }

    _initialize () {
      super._initialize();

      this._computeAllowedChildren();
    }

    _compile () {
      super._compile();

      this.blocks = this.root.blocks;
      this.dfas = this.blocks.map(b => new DFA$2(b.dfa.transitions));
    }

    _computeAllowedChildren () {
      this._allowedChildren = Object.assign(...this.blocks.map((block) => {
        return _collectAllTokensFromDFA$2(block.dfa)
      }));
    }

    _findNextDFA (state, token) {
      console.assert(state.dfaStates.length === this.dfas.length);
      const dfas = this.dfas;
      for (let i = 0; i < state.dfaStates.length; i++) {
        const dfa = dfas[i];
        const dfaState = state.dfaStates[i];
        if (dfa.canConsume(dfaState, token)) {
          return i
        }
      }
      return -1
    }

    _findInsertPos(el, newTag, mode) { 
      
      return el.childNodes.length
    }
  }

  class Token$2 {
    constructor (name) {
      this.name = name;
    }

    toString () {
      return this.name
    }

    toJSON () {
      return this.name
    }

    copy () {
      return new Token$2(this.name)
    }

    _normalize () {}

    _compile () {
      this.dfa = DFABuilder$2.singleToken(this.name);
    }
  }

  Token$2.fromJSON = function (data) {
    return new Token$2(data)
  };


  class Choice$2 {
    constructor (blocks) {
      this.blocks = blocks;
    }

    copy () {
      return new Choice$2(this.blocks.map(b => b.copy()))
    }

    toJSON () {
      return {
        type: '|',
        blocks: this.blocks.map(b => b.toJSON())
      }
    }

    _normalize () {
      const blocks = this.blocks;
      for (let i = blocks.length - 1; i >= 0; i--) {
        let block = blocks[i];
        block._normalize();
        
        if (block instanceof Choice$2) {
          blocks.splice(i, 1, ...(block.blocks));
        }
      }
    }

    _compile () {
      let dfa = new DFABuilder$2();
      this.blocks.forEach((block) => {
        if (block instanceof Token$2) {
          dfa.addTransition(START$2$2, END$2$2, block.name);
        } else if (block instanceof Interleave$2) {
          throw new Error('Nested interleave blocks are not supported.')
        } else {
          if (!block.dfa) {
            block._compile();
          }
          dfa.merge(block.dfa);
        }
      });
      this.dfa = dfa;
      return dfa
    }

    toString () {
      return '(' + this.blocks.map(b => b.toString()).join('|') + ')'
    }
  }

  Choice$2.fromJSON = function (data) {
    return new Choice$2(data.blocks.map((block) => {
      return _fromJSON$2(block)
    }))
  };


  class Sequence$2 {
    constructor (blocks) {
      this.blocks = blocks;
    }

    copy () {
      return new Sequence$2(this.blocks.map(b => b.copy()))
    }

    toJSON () {
      return {
        type: ',',
        blocks: this.blocks.map(b => b.toJSON())
      }
    }

    _compile () {
      let dfa = new DFABuilder$2();
      this.blocks.forEach((block) => {
        if (block instanceof Token$2) {
          dfa.append(DFABuilder$2.singleToken(block.name));
        } else if (block instanceof Interleave$2) {
          throw new Error('Nested interleave blocks are not supported.')
        } else {
          if (!block.dfa) {
            block._compile();
          }
          dfa.append(block.dfa);
        }
      });
      this.dfa = dfa;
      return dfa
    }

    _normalize () {
      const blocks = this.blocks;
      for (let i = blocks.length - 1; i >= 0; i--) {
        let block = blocks[i];
        block._normalize();
        
        if (block instanceof Sequence$2) {
          blocks.splice(i, 1, ...(block.blocks));
        }
      }
    }

    toString () {
      return '(' + this.blocks.map(b => b.toString()).join(',') + ')'
    }
  }

  Sequence$2.fromJSON = function (data) {
    return new Sequence$2(data.blocks.map((block) => {
      return _fromJSON$2(block)
    }))
  };


  class Interleave$2 {
    constructor (blocks) {
      this.blocks = blocks;
    }

    copy () {
      return new Interleave$2(this.blocks.map(b => b.copy()))
    }

    toString () {
      return '(' + this.blocks.map(b => b.toString()).join('~') + ')'
    }

    toJSON () {
      return {
        type: '~',
        blocks: this.blocks.map(b => b.toJSON())
      }
    }

    _normalize () {
      
    }

    _compile () {
      this.blocks.forEach(block => block._compile());
    }
  }

  Interleave$2.fromJSON = function (data) {
    return new Interleave$2(data.blocks.map((block) => {
      return _fromJSON$2(block)
    }))
  };


  class Optional$2 {
    constructor (block) {
      this.block = block;
    }

    copy () {
      return new Optional$2(this.block.copy())
    }

    toJSON () {
      return {
        type: '?',
        block: this.block.toJSON()
      }
    }

    _compile () {
      const block = this.block;
      if (block instanceof Interleave$2) {
        throw new Error('Nested interleave blocks are not supported.')
      }
      if (!block.dfa) {
        block._compile();
      }
      this.dfa = block.dfa.optional();
      return this.dfa
    }

    _normalize () {
      const block = this.block;
      block._normalize();
      if (block instanceof Optional$2) {
        this.block = block.block;
      } else if (block instanceof Kleene$2) {
        console.error('FIXME -  <optional> is useless here', this.toString());
      }
    }

    toString () {
      return this.block.toString() + '?'
    }
  }

  Optional$2.fromJSON = function (data) {
    return new Optional$2(_fromJSON$2(data.block))
  };


  class Kleene$2 {
    constructor (block) {
      this.block = block;
    }

    copy () {
      return new Kleene$2(this.block.copy())
    }

    toJSON () {
      return {
        type: '*',
        block: this.block.toJSON()
      }
    }

    _compile () {
      const block = this.block;
      if (block instanceof Interleave$2) {
        throw new Error('Nested interleave blocks are not supported.')
      }
      if (!block.dfa) {
        block._compile();
      }
      this.dfa = block.dfa.kleene();
      return this.dfa
    }

    _normalize () {
      const block = this.block;
      block._normalize();
      if (block instanceof Optional$2 || block instanceof Kleene$2) {
        this.block = block.block;
      } else if (block instanceof Plus$2) {
        throw new Error('This does not make sense:' + this.toString())
      }
    }

    toString () {
      return this.block.toString() + '*'
    }
  }

  Kleene$2.fromJSON = function (data) {
    return new Kleene$2(_fromJSON$2(data.block))
  };


  class Plus$2 {
    constructor (block) {
      this.block = block;
    }

    copy () {
      return new Plus$2(this.block.copy())
    }

    toJSON () {
      return {
        type: '+',
        block: this.block.toJSON()
      }
    }

    _compile () {
      const block = this.block;
      if (block instanceof Interleave$2) {
        throw new Error('Nested interleave blocks are not supported.')
      }
      if (!block.dfa) {
        block._compile();
      }
      this.dfa = block.dfa.plus();
      return this.dfa
    }

    _normalize () {
      const block = this.block;
      block._normalize();
      if (block instanceof Optional$2 || block instanceof Kleene$2) {
        throw new Error('This does not make sense:' + this.toString())
      } else if (block instanceof Plus$2) {
        this.block = block.block;
      }
    }

    toString () {
      return this.block.toString() + '+'
    }
  }

  Plus$2.fromJSON = function (data) {
    return new Plus$2(_fromJSON$2(data.block))
  };

  function _fromJSON$2 (data) {
    switch (data.type) {
      case ',':
        return Sequence$2.fromJSON(data)
      case '~':
        return Interleave$2.fromJSON(data)
      case '|':
        return Choice$2.fromJSON(data)
      case '?':
        return Optional$2.fromJSON(data)
      case '+':
        return Plus$2.fromJSON(data)
      case '*':
        return Kleene$2.fromJSON(data)
      default:
        if (isString$2(data)) {
          return new Token$2(data)
        }
        throw new Error('Unsupported data.')
    }
  }

  const { TEXT: TEXT$4$2 } = DFA$2;

  class XMLSchema$2 {
    constructor (elementSchemas, startElement) {
      if (!elementSchemas[startElement]) {
        throw new Error('startElement must be a valid element.')
      }
      this._elementSchemas = {};
      this.startElement = startElement;
      
      forEach$2(elementSchemas, (spec, name) => {
        this._elementSchemas[name] = new ElementSchema$2(spec.name, spec.type, spec.attributes, spec.expr);
      });
    }

    getIdAttribute () {
      return 'id'
    }

    getTagNames () {
      return Object.keys(this._elementSchemas)
    }

    getElementSchema (name) {
      return this._elementSchemas[name]
    }

    getStartElement () {
      return this.startElement
    }

    toJSON () {
      let result = {
        start: this.getStartElement(),
        elements: {}
      };
      forEach$2(this._elementSchemas, (schema, name) => {
        result.elements[name] = schema.toJSON();
      });
      return result
    }

    
    toMD () {
      let result = [];
      let elementNames = Object.keys(this._elementSchemas);
      elementNames.sort();
      elementNames.forEach((name) => {
        let elementSchema = this._elementSchemas[name];
        result.push(`# <${elementSchema.name}>`);
        result.push('');
        result.push(`type: ${elementSchema.type}`);
        result.push('attributes: ' + map$3(elementSchema.attributes, (_, name) => { return name }).join(', '));
        result.push('children:');
        result.push('  ' + elementSchema.expr.toString());
        result.push('');
      });
      return result.join('\n')
    }

    validateElement (el) {
      let tagName = el.tagName;
      let elementSchema = this.getElementSchema(tagName);
      return _validateElement$2(elementSchema, el)
    }
  }

  XMLSchema$2.fromJSON = function (data) {
    let elementSchemas = {};
    forEach$2(data.elements, (elData) => {
      let elSchema = ElementSchema$2.fromJSON(elData);
      elementSchemas[elSchema.name] = elSchema;
    });
    return new XMLSchema$2(elementSchemas, data.start)
  };

  class ElementSchema$2 {
    constructor (name, type, attributes, expr) {
      this.name = name;
      this.type = type;
      this.attributes = attributes;
      this.expr = expr;

      if (!name) {
        throw new Error("'name' is mandatory")
      }
      if (!type) {
        throw new Error("'type' is mandatory")
      }
      if (!attributes) {
        throw new Error("'attributes' is mandatory")
      }
      if (!expr) {
        throw new Error("'expr' is mandatory")
      }
    }

    toJSON () {
      return {
        name: this.name,
        type: this.type,
        attributes: this.attributes,
        elements: this.expr.toJSON()
      }
    }

    isAllowed (tagName) {
      return this.expr.isAllowed(tagName)
    }

    isTextAllowed () {
      return this.expr.isAllowed(TEXT$4$2)
    }

    printStructure () {
      return `${this.name} ::= ${this.expr.toString()}`
    }

    findFirstValidPos (el, newTag) {
      return this.expr._findInsertPos(el, newTag, 'first')
    }

    findLastValidPos (el, newTag) {
      return this.expr._findInsertPos(el, newTag, 'last')
    }
  }

  ElementSchema$2.fromJSON = function (data) {
    return new ElementSchema$2(
      data.name,
      data.type,
      data.attributes,
      Expression$2.fromJSON(data.elements)
    )
  };

  function _validateElement$2 (elementSchema, el) {
    let errors = [];
    let valid = true;
    { 
      const res = _checkAttributes$2(elementSchema, el);
      if (!res.ok) {
        errors = errors.concat(res.errors);
        valid = false;
      }
    }
    
    if (elementSchema.type === 'external') ; else {
      
      let res;
      if (el._isXMLTextElement) {
        res = _checkChildren$2(elementSchema, el.toXML());
      } else {
        res = _checkChildren$2(elementSchema, el);
      }
      if (!res.ok) {
        errors = errors.concat(res.errors);
        valid = false;
      }
    }
    return {
      errors,
      ok: valid
    }
  }

  function _checkAttributes$2(elementSchema, el) { 
    return { ok: true }
  }

  function _checkChildren$2 (elementSchema, el) {
    
    
    if (elementSchema.type === 'external') {
      return true
    }
    const expr = elementSchema.expr;
    const state = expr.getInitialState();
    const iterator = el.getChildNodeIterator();
    let valid = true;
    while (valid && iterator.hasNext()) {
      const childEl = iterator.next();
      let token;
      if (childEl.isTextNode()) {
        
        if (elementSchema.type !== 'text' && _isTextNodeEmpty$2(childEl)) {
          continue
        }
        token = TEXT$4$2;
      } else if (childEl.isElementNode()) {
        token = childEl.tagName;
      } else {
        continue
      }
      if (!expr.consume(state, token)) {
        valid = false;
      }
    }
    
    if (state.errors.length > 0) {
      state.errors.forEach((err) => {
        err.el = el;
      });
    }
    if (valid && !expr.isFinished(state)) {
      state.errors.push({
        msg: `<${el.tagName}> is incomplete.\nSchema: ${expr.toString()}`,
        el
      });
      valid = false;
    }
    if (valid) {
      state.ok = true;
    }
    return state
  }

  var ManifestSchemaData$2 = {"start":"dar","elements":{"dar":{"name":"dar","type":"element","attributes":{},"elements":{"name":"dar","content":{"type":",","blocks":["documents","assets"]}}},"documents":{"name":"documents","type":"element","attributes":{},"elements":{"name":"documents","content":{"type":"*","block":"document"}}},"assets":{"name":"assets","type":"element","attributes":{},"elements":{"name":"assets","content":{"type":"*","block":"asset"}}},"document":{"name":"document","type":"element","attributes":{"id":{"name":"id"},"name":{"name":"name"},"type":{"name":"type"},"path":{"name":"path"}},"elements":{"name":"document","content":{"type":",","blocks":[]}}},"asset":{"name":"asset","type":"element","attributes":{"id":{"name":"id"},"name":{"name":"name"},"type":{"name":"type"},"path":{"name":"path"}},"elements":{"name":"asset","content":{"type":",","blocks":[]}}}}};

  const ManifestSchema$2 = XMLSchema$2.fromJSON(ManifestSchemaData$2);


  ManifestSchema$2.getName = function() {
    return 'RDC-Manifest'
  };

  ManifestSchema$2.getVersion = function() {
    return '1.0'
  };

  ManifestSchema$2.getDocTypeParams = function() {
    return ['manifest', 'RDC-Manifest 1.0', ManifestSchema$2.uri]
  };


  ManifestSchema$2.getDefaultTextType = function () {
    return 'text'
  };

  ManifestSchema$2.uri = '//Manifest-1.0.dtd';

  class Data$2 extends EventEmitter$2 {
    
    constructor (schema, nodeFactory) {
      super();

      
      if (!schema) {
        throw new Error('schema is mandatory')
      }
      if (!nodeFactory) {
        throw new Error('nodeFactory is mandatory')
      }
      

      this.schema = schema;
      this.nodeFactory = nodeFactory;
      this.nodes = {};
      this.indexes = {};

      
      
      this.__QUEUE_INDEXING__ = false;
      this.queue = [];
    }

    
    contains (id) {
      return Boolean(this.nodes[id])
    }

    
    get (path, strict) {
      let result = this._get(path);
      if (strict && result === undefined) {
        if (isString$2(path)) {
          throw new Error("Could not find node with id '" + path + "'.")
        } else if (!this.contains(path[0])) {
          throw new Error("Could not find node with id '" + path[0] + "'.")
        } else {
          throw new Error("Property for path '" + path + "' us undefined.")
        }
      }
      return result
    }

    _get (path) {
      if (!path) return undefined
      let result;
      if (isString$2(path)) {
        result = this.nodes[path];
      } else if (path.length === 1) {
        result = this.nodes[path[0]];
      } else if (path.length > 1) {
        let context = this.nodes[path[0]];
        for (let i = 1; i < path.length - 1; i++) {
          if (!context) return undefined
          context = context[path[i]];
        }
        if (!context) return undefined
        result = context[path[path.length - 1]];
      }
      return result
    }

    
    getNodes () {
      return this.nodes
    }

    
    create (nodeData) {
      var node = this.nodeFactory.create(nodeData.type, nodeData);
      if (!node) {
        throw new Error('Illegal argument: could not create node for data:', nodeData)
      }
      if (this.contains(node.id)) {
        throw new Error('Node already exists: ' + node.id)
      }
      if (!node.id || !node.type) {
        throw new Error('Node id and type are mandatory.')
      }
      this.nodes[node.id] = node;

      var change = {
        type: 'create',
        node: node
      };

      if (this.__QUEUE_INDEXING__) {
        this.queue.push(change);
      } else {
        this._updateIndexes(change);
      }

      return node
    }

    
    delete (nodeId) {
      var node = this.nodes[nodeId];
      if (!node) return
      node.dispose();
      delete this.nodes[nodeId];

      var change = {
        type: 'delete',
        node: node
      };

      if (this.__QUEUE_INDEXING__) {
        this.queue.push(change);
      } else {
        this._updateIndexes(change);
      }

      return node
    }

    
    set (path, newValue) {
      let node = this.get(path[0]);
      let oldValue = this._set(path, newValue);
      var change = {
        type: 'set',
        node: node,
        path: path,
        newValue: newValue,
        oldValue: oldValue
      };
      if (this.__QUEUE_INDEXING__) {
        this.queue.push(change);
      } else {
        this._updateIndexes(change);
      }
      return oldValue
    }

    _set (path, newValue) {
      let oldValue = _setValue$2(this.nodes, path, newValue);
      return oldValue
    }

    
    update (path, diff) {
      var realPath = this.getRealPath(path);
      if (!realPath) {
        console.error('Could not resolve path', path);
        return
      }
      let node = this.get(realPath[0]);
      let oldValue = this._get(realPath);
      let newValue;
      if (diff.isOperation) {
        newValue = diff.apply(oldValue);
      } else {
        diff = this._normalizeDiff(oldValue, diff);
        if (isString$2(oldValue)) {
          switch (diff.type) {
            case 'delete': {
              newValue = oldValue.split('').splice(diff.start, diff.end - diff.start).join('');
              break
            }
            case 'insert': {
              newValue = [oldValue.substring(0, diff.start), diff.text, oldValue.substring(diff.start)].join('');
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else if (isArray$3(oldValue)) {
          newValue = oldValue.slice(0);
          switch (diff.type) {
            case 'delete': {
              newValue.splice(diff.pos, 1);
              break
            }
            case 'insert': {
              newValue.splice(diff.pos, 0, diff.value);
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else if (oldValue._isCoordinate) {
          switch (diff.type) {
            case 'shift': {
              
              oldValue = { path: oldValue.path, offset: oldValue.offset };
              newValue = oldValue;
              newValue.offset += diff.value;
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else {
          throw new Error('Diff is not supported:', JSON.stringify(diff))
        }
      }
      this._set(realPath, newValue);

      var change = {
        type: 'update',
        node: node,
        path: realPath,
        newValue: newValue,
        oldValue: oldValue
      };

      if (this.__QUEUE_INDEXING__) {
        this.queue.push(change);
      } else {
        this._updateIndexes(change);
      }

      return oldValue
    }

    
    _normalizeDiff (value, diff) {
      if (isString$2(value)) {
        
        if (diff['delete']) {
          console.warn('DEPRECATED: use doc.update(path, {type:"delete", start:s, end: e}) instead');
          diff = {
            type: 'delete',
            start: diff['delete'].start,
            end: diff['delete'].end
          };
        } else if (diff['insert']) {
          console.warn('DEPRECATED: use doc.update(path, {type:"insert", start:s, text: t}) instead');
          diff = {
            type: 'insert',
            start: diff['insert'].offset,
            text: diff['insert'].value
          };
        }
      } else if (isArray$3(value)) {
        
        if (diff['delete']) {
          console.warn('DEPRECATED: use doc.update(path, {type:"delete", pos:1}) instead');
          diff = {
            type: 'delete',
            pos: diff['delete'].offset
          };
        } else if (diff['insert']) {
          console.warn('DEPRECATED: use doc.update(path, {type:"insert", pos:1, value: "foo"}) instead');
          diff = {
            type: 'insert',
            pos: diff['insert'].offset,
            value: diff['insert'].value
          };
        }
      } else if (value._isCoordinate) {
        if (diff.hasOwnProperty('shift')) {
          console.warn('DEPRECATED: use doc.update(path, {type:"shift", value:2}) instead');
          diff = {
            type: 'shift',
            value: diff['shift']
          };
        }
      }
      return diff
    }

    
    toJSON () {
      let nodes = {};
      forEach$2(this.nodes, (node) => {
        nodes[node.id] = node.toJSON();
      });
      return {
        schema: [this.schema.id, this.schema.version],
        nodes: nodes
      }
    }

    reset () {
      this.clear();
    }

    
    clear () {
      this.nodes = {};
      forEach$2(this.indexes, index => index.clear());
    }

    
    addIndex (name, index) {
      if (this.indexes[name]) {
        console.error('Index with name %s already exists.', name);
      }
      index.reset(this);
      this.indexes[name] = index;
      return index
    }

    
    getIndex (name) {
      return this.indexes[name]
    }

    
    _updateIndexes (change) {
      if (!change || this.__QUEUE_INDEXING__) return
      forEach$2(this.indexes, function (index) {
        if (index.select(change.node)) {
          if (!index[change.type]) {
            console.error('Contract: every NodeIndex must implement ' + change.type);
          }
          index[change.type](change.node, change.path, change.newValue, change.oldValue);
        }
      });
    }

    
    _stopIndexing () {
      this.__QUEUE_INDEXING__ = true;
    }

    
    _startIndexing () {
      this.__QUEUE_INDEXING__ = false;
      while (this.queue.length > 0) {
        var change = this.queue.shift();
        this._updateIndexes(change);
      }
    }
  }

  function _setValue$2 (root, path, newValue) {
    let ctx = root;
    let L = path.length;
    for (let i = 0; i < L - 1; i++) {
      ctx = ctx[path[i]];
      if (!ctx) throw new Error('Can not set value.')
    }
    let oldValue = ctx[path[L - 1]];
    ctx[path[L - 1]] = newValue;
    return oldValue
  }

  class IncrementalData$2 extends Data$2 {
    
    create (nodeData) {
      if (nodeData._isNode) {
        nodeData = nodeData.toJSON();
      }
      let op = ObjectOperation$2.Create([nodeData.id], nodeData);
      this.apply(op);
      return op
    }

    
    delete (nodeId) {
      var op = null;
      var node = this.get(nodeId);
      if (node) {
        var nodeData = node.toJSON();
        op = ObjectOperation$2.Delete([nodeId], nodeData);
        this.apply(op);
      }
      return op
    }

    
    update (path, diff) {
      var diffOp = this._getDiffOp(path, diff);
      var op = ObjectOperation$2.Update(path, diffOp);
      this.apply(op);
      return op
    }

    
    set (path, newValue) {
      var oldValue = this.get(path);
      var op = ObjectOperation$2.Set(path, oldValue, newValue);
      this.apply(op);
      return op
    }

    
    apply (op) {
      if (op.type === ObjectOperation$2.NOP) return
      else if (op.type === ObjectOperation$2.CREATE) {
        
        super.create(cloneDeep$2(op.val));
      } else if (op.type === ObjectOperation$2.DELETE) {
        super.delete(op.val.id);
      } else if (op.type === ObjectOperation$2.UPDATE) {
        var oldVal = this.get(op.path);
        var diff = op.diff;
        if (op.propertyType === 'array') {
          if (!(diff._isArrayOperation)) {
            diff = ArrayOperation$2.fromJSON(diff);
          }
          
          diff.apply(oldVal);
        } else if (op.propertyType === 'string') {
          if (!(diff._isTextOperation)) {
            diff = TextOperation$2.fromJSON(diff);
          }
          var newVal = diff.apply(oldVal);
          super.set(op.path, newVal);
        } else if (op.propertyType === 'coordinate') {
          if (!(diff._isCoordinateOperation)) {
            diff = CoordinateOperation$2.fromJSON(diff);
          }
          diff.apply(oldVal);
        } else {
          throw new Error('Unsupported type for operational update.')
        }
      } else if (op.type === ObjectOperation$2.SET) {
        super.set(op.path, op.val);
      } else {
        throw new Error('Illegal state.')
      }
      this.emit('operation:applied', op, this);
    }

    
    _getDiffOp (path, diff) {
      var diffOp = null;
      if (diff.isOperation) {
        diffOp = diff;
      } else {
        var value = this.get(path);
        diff = this._normalizeDiff(value, diff);
        if (value === null || value === undefined) {
          throw new Error('Property has not been initialized: ' + JSON.stringify(path))
        } else if (isString$2(value)) {
          switch (diff.type) {
            case 'delete': {
              diffOp = TextOperation$2.Delete(diff.start, value.substring(diff.start, diff.end));
              break
            }
            case 'insert': {
              diffOp = TextOperation$2.Insert(diff.start, diff.text);
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else if (isArray$3(value)) {
          switch (diff.type) {
            case 'delete': {
              diffOp = ArrayOperation$2.Delete(diff.pos, value[diff.pos]);
              break
            }
            case 'insert': {
              diffOp = ArrayOperation$2.Insert(diff.pos, diff.value);
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        } else if (value._isCoordinate) {
          switch (diff.type) {
            case 'shift': {
              diffOp = CoordinateOperation$2.Shift(diff.value);
              break
            }
            default:
              throw new Error('Unknown diff type')
          }
        }
      }
      if (!diffOp) {
        throw new Error('Unsupported diff: ' + JSON.stringify(diff))
      }
      return diffOp
    }
  }

  class PropertyIndex$2 extends NodeIndex$2 {
    constructor (property) {
      super();

      this._property = property || 'id';
      this.index = new TreeIndex$2();
    }

    
    get (path) {
      return this.index.get(path) || {}
    }

    
    getAll (path) {
      return this.index.getAll(path)
    }

    clear () {
      this.index.clear();
    }

    
    select(node) { 
      return true
    }

    
    create (node) {
      var values = node[this._property];
      if (!isArray$3(values)) {
        values = [values];
      }
      forEach$2(values, function (value) {
        this.index.set([value, node.id], node);
      }.bind(this));
    }

    
    delete (node) {
      var values = node[this._property];
      if (!isArray$3(values)) {
        values = [values];
      }
      forEach$2(values, function (value) {
        this.index.delete([value, node.id]);
      }.bind(this));
    }

    
    update (node, path, newValue, oldValue) {
      if (!this.select(node) || path[1] !== this._property) return
      var values = oldValue;
      if (!isArray$3(values)) {
        values = [values];
      }
      forEach$2(values, function (value) {
        this.index.delete([value, node.id]);
      }.bind(this));
      values = newValue;
      if (!isArray$3(values)) {
        values = [values];
      }
      forEach$2(values, function (value) {
        this.index.set([value, node.id], node);
      }.bind(this));
    }

    set (node, path, newValue, oldValue) {
      this.update(node, path, newValue, oldValue);
    }

    _initialize (data) {
      forEach$2(data.getNodes(), function (node) {
        if (this.select(node)) {
          this.create(node);
        }
      }.bind(this));
    }
  }

  class AnnotationIndex$2 extends DocumentIndex$2 {
    constructor () {
      super();

      this.byPath = new TreeIndex$2();
      this.byType = new TreeIndex$2();
    }

    select (node) {
      return node.isPropertyAnnotation()
    }

    clear () {
      this.byPath.clear();
      this.byType.clear();
    }

    
    get (path, start, end, type) {
      var annotations;
      if (isString$2(path) || path.length === 1) {
        annotations = this.byPath.getAll(path) || {};
      } else {
        annotations = this.byPath.get(path);
      }
      annotations = map$3(annotations);
      if (isNumber$2(start)) {
        annotations = filter$2(annotations, AnnotationIndex$2.filterByRange(start, end));
      }
      if (type) {
        annotations = filter$2(annotations, DocumentIndex$2.filterByType(type));
      }
      return annotations
    }

    create (anno) {
      const path = anno.start.path;
      this.byType.set([anno.type, anno.id], anno);
      if (path && path.length > 0) {
        this.byPath.set(anno.start.path.concat([anno.id]), anno);
      }
    }

    delete (anno) {
      this._delete(anno.type, anno.id, anno.start.path);
    }

    _delete (type, id, path) {
      this.byType.delete([type, id]);
      if (path && path.length > 0) {
        this.byPath.delete(path.concat([id]));
      }
    }

    update (node, path, newValue, oldValue) {
      
      if (this.select(node) && path[1] === 'start' && path[2] === 'path') {
        this._delete(node.type, node.id, oldValue);
        this.create(node);
      }
    }
  }

  AnnotationIndex$2.filterByRange = function (start, end) {
    return function (anno) {
      var aStart = anno.start.offset;
      var aEnd = anno.end.offset;
      var overlap = (aEnd >= start);
      
      if (isNumber$2(end)) {
        overlap = overlap && (aStart <= end);
      }
      return overlap
    }
  };

  class ContainerAnnotationIndex$2 extends DocumentIndex$2 {
    constructor () {
      super();
      this.byId = new TreeIndex$2();
    }

    select (node) {
      return node.isContainerAnnotation()
    }

    clear () {
      this.byId.clear();
    }

    get (containerId, type) {
      var annotations = map$3(this.byId.get(containerId));
      if (isString$2(type)) {
        annotations = filter$2(annotations, DocumentIndex$2.filterByType);
      }
      return annotations
    }

    create (anno) {
      this.byId.set([anno.containerId, anno.id], anno);
    }

    delete (anno) {
      this.byId.delete([anno.containerId, anno.id]);
    }

    update(node, path, newValue, oldValue) { 
      
    }
  }

  class DocumentNodeFactory$2 {
    constructor (doc) {
      this.doc = doc;
    }

    create (nodeType, nodeData) {
      var NodeClass = this.doc.schema.getNodeClass(nodeType);
      if (!NodeClass) {
        throw new Error('No node registered by that name: ' + nodeType)
      }
      return new NodeClass(this.doc, nodeData)
    }
  }

  function copySelection$2 (doc, selection) {
    if (!selection) throw new Error("'selection' is mandatory.")
    let copy = null;
    if (!selection.isNull() && !selection.isCollapsed()) {
      
      if (selection.isPropertySelection()) {
        copy = _copyPropertySelection$2(doc, selection);
      } else if (selection.isContainerSelection()) {
        copy = _copyContainerSelection$2(doc, selection);
      } else if (selection.isNodeSelection()) {
        copy = _copyNodeSelection$2(doc, selection);
      } else {
        console.error('Copy is not yet supported for selection type.');
      }
    }
    return copy
  }

  function _copyPropertySelection$2 (doc, selection) {
    let path = selection.start.path;
    let offset = selection.start.offset;
    let endOffset = selection.end.offset;
    let text = doc.get(path);
    let snippet = doc.createSnippet();
    let containerNode = snippet.getContainer();
    snippet.create({
      type: doc.schema.getDefaultTextType(),
      id: TEXT_SNIPPET_ID$2,
      content: text.substring(offset, endOffset)
    });
    containerNode.show(TEXT_SNIPPET_ID$2);
    let annotations = doc.getIndex('annotations').get(path, offset, endOffset);
    forEach$2(annotations, function (anno) {
      let data = cloneDeep$2(anno.toJSON());
      let path = [TEXT_SNIPPET_ID$2, 'content'];
      data.start = {
        path: path,
        offset: Math.max(offset, anno.start.offset) - offset
      };
      data.end = {
        path: path,
        offset: Math.min(endOffset, anno.end.offset) - offset
      };
      snippet.create(data);
    });
    return snippet
  }

  function _copyContainerSelection$2 (tx, sel) {
    let snippet = tx.createSnippet();
    let container = snippet.getContainer();

    let nodeIds = sel.getNodeIds();
    let L = nodeIds.length;
    if (L === 0) return snippet

    let start = sel.start;
    let end = sel.end;

    let skippedFirst = false;
    let skippedLast = false;

    
    let created = {};
    for (let i = 0; i < L; i++) {
      let id = nodeIds[i];
      let node = tx.get(id);
      
      if (i === 0 && isLast$2(tx, start)) {
        skippedFirst = true;
        continue
      }
      if (i === L - 1 && isFirst$2(tx, end)) {
        skippedLast = true;
        continue
      }
      if (!created[id]) {
        copyNode$2(node).forEach((nodeData) => {
          let copy = snippet.create(nodeData);
          created[copy.id] = true;
        });
        container.show(id);
      }
    }
    if (!skippedFirst) {
      
      let startNode = snippet.get(start.getNodeId()).getContainerRoot();
      if (startNode.isText()) {
        deleteTextRange$2(snippet, null, start);
      } else if (startNode.isList()) {
        deleteListRange$2(snippet, startNode, null, start);
      }
    }
    if (!skippedLast) {
      
      let endNode = snippet.get(end.getNodeId()).getContainerRoot();
      if (endNode.isText()) {
        deleteTextRange$2(snippet, end, null);
      } else if (endNode.isList()) {
        deleteListRange$2(snippet, endNode, end, null);
      }
    }
    return snippet
  }

  function _copyNodeSelection$2 (doc, selection) {
    let snippet = doc.createSnippet();
    let containerNode = snippet.getContainer();
    let nodeId = selection.getNodeId();
    let node = doc.get(nodeId);
    copyNode$2(node).forEach((nodeData) => {
      snippet.create(nodeData);
    });
    containerNode.show(node.id);
    return snippet
  }

  function paste$2 (tx, args) {
    let sel = tx.selection;
    if (!sel || sel.isNull() || sel.isCustomSelection()) {
      throw new Error('Can not paste, without selection or a custom selection.')
    }
    args = args || {};
    args.text = args.text || '';
    let pasteDoc = args.doc;
    
    
    let inContainer = Boolean(sel.containerId);

    
    
    
    if (!pasteDoc && !inContainer) {
      tx.insertText(args.text);
      return
    }
    if (!pasteDoc) {
      pasteDoc = _convertPlainTextToDocument$2(tx, args);
    }
    if (pasteDoc && !inContainer) {
      
      let nodes = [];
      let container = pasteDoc.get('snippet');
      let content = container.getContent();
      content.forEach(nodeId => {
        let text = pasteDoc.get(nodeId).getText();
        nodes.push(text);
      });
      tx.insertText(nodes.join('\r\n'));
      return
    }
    if (!sel.isCollapsed()) {
      tx.deleteSelection();
    }
    let snippet = pasteDoc.get(SNIPPET_ID$2);
    if (snippet.getLength() > 0) {
      let first = snippet.getChildAt(0);
      if (first.isText()) {
        _pasteAnnotatedText$2(tx, pasteDoc);
        
        
        
        snippet.hideAt(0);
      }
      
      if (snippet.getLength() > 0) {
        _pasteDocument$2(tx, pasteDoc);
      }
    }
    return args
  }


  function _convertPlainTextToDocument$2 (tx, args) {
    let lines = args.text.split(/\s*\n\s*\n/);
    let pasteDoc = tx.getDocument().newInstance();
    let defaultTextType = pasteDoc.getSchema().getDefaultTextType();
    let container = pasteDoc.create({
      type: '@container',
      id: SNIPPET_ID$2,
      nodes: []
    });
    let node;
    if (lines.length === 1) {
      node = pasteDoc.create({
        id: TEXT_SNIPPET_ID$2,
        type: defaultTextType,
        content: lines[0]
      });
      container.show(node.id);
    } else {
      for (let i = 0; i < lines.length; i++) {
        node = pasteDoc.create({
          id: uuid$2(defaultTextType),
          type: defaultTextType,
          content: lines[i]
        });
        container.show(node.id);
      }
    }
    return pasteDoc
  }

  function _pasteAnnotatedText$2 (tx, copy) {
    let sel = tx.selection;
    const nodes = copy.get(SNIPPET_ID$2).nodes;
    const firstId = nodes[0];
    const first = copy.get(firstId);
    const textPath = first.getPath();
    const text = copy.get(textPath);
    const annotations = copy.getIndex('annotations').get(textPath);
    
    let path = sel.start.path;
    let offset = sel.start.offset;
    tx.insertText(text);
    
    forEach$2(annotations, function (anno) {
      let data = anno.toJSON();
      data.start.path = path.slice(0);
      data.start.offset += offset;
      data.end.offset += offset;
      
      if (tx.get(data.id)) data.id = uuid$2(data.type);
      tx.create(data);
    });
  }

  function _pasteDocument$2 (tx, pasteDoc) {
    let sel = tx.selection;
    let containerId = sel.containerId;
    let container = tx.get(containerId);
    let insertPos;
    if (sel.isPropertySelection()) {
      let startPath = sel.start.path;
      let nodeId = sel.start.getNodeId();
      let startPos = container.getPosition(nodeId, 'strict');
      let text = tx.get(startPath);
      
      
      if (text.length === 0) {
        insertPos = startPos;
        container.hide(nodeId);
        deleteNode$2(tx, tx.get(nodeId));
      } else if (text.length === sel.start.offset) {
        insertPos = startPos + 1;
      } else {
        tx.break();
        insertPos = startPos + 1;
      }
    } else if (sel.isNodeSelection()) {
      let nodePos = container.getPosition(sel.getNodeId(), 'strict');
      if (sel.isBefore()) {
        insertPos = nodePos;
      } else if (sel.isAfter()) {
        insertPos = nodePos + 1;
      } else {
        throw new Error('Illegal state: the selection should be collapsed.')
      }
    }
    
    let nodeIds = pasteDoc.get(SNIPPET_ID$2).nodes;
    let insertedNodes = [];
    let visited = {};
    for (let i = 0; i < nodeIds.length; i++) {
      let node = pasteDoc.get(nodeIds[i]);
      
      
      
      
      
      let newId = _transferWithDisambiguatedIds$2(node.getDocument(), tx, node.id, visited);
      
      node = tx.get(newId);
      container.showAt(insertPos++, newId);
      insertedNodes.push(node);
    }

    if (insertedNodes.length > 0) {
      let lastNode = last$1$2(insertedNodes);
      setCursor$2(tx, lastNode, containerId, 'after');
    }
  }






  function _transferWithDisambiguatedIds$2 (sourceDoc, targetDoc, id, visited) {
    if (visited[id]) throw new Error('FIXME: dont call me twice')
    const node = sourceDoc.get(id, 'strict');
    let oldId = node.id;
    let newId;
    if (targetDoc.contains(node.id)) {
      
      newId = uuid$2(node.type);
      node.id = newId;
    }
    visited[id] = node.id;
    const annotationIndex = sourceDoc.getIndex('annotations');
    const nodeSchema = node.getSchema();
    
    let annos = [];
    
    
    
    for (let prop of nodeSchema) {
      const name = prop.name;
      if (name === 'id' || name === 'type') continue
      
      if ((prop.isReference() && prop.isOwned()) || (prop.type === 'file')) {
        
        
        let val = node[prop.name];
        if (prop.isArray()) {
          _transferArrayOfReferences$2(sourceDoc, targetDoc, val, visited);
        } else {
          let id = val;
          if (!visited[id]) {
            node[name] = _transferWithDisambiguatedIds$2(sourceDoc, targetDoc, id, visited);
          }
        }
      
      } else if (prop.isText()) {
        
        
        
        
        
        let _annos = annotationIndex.get([oldId, prop.name]);
        for (let i = 0; i < _annos.length; i++) {
          let anno = _annos[i];
          if (anno.start.path[0] === oldId && newId) {
            anno.start.path[0] = newId;
          }
          if (anno.end.path[0] === oldId && newId) {
            anno.end.path[0] = newId;
          }
          annos.push(anno);
        }
      }
    }
    targetDoc.create(node);
    for (let i = 0; i < annos.length; i++) {
      _transferWithDisambiguatedIds$2(sourceDoc, targetDoc, annos[i].id, visited);
    }
    return node.id
  }

  function _transferArrayOfReferences$2 (sourceDoc, targetDoc, arr, visited) {
    for (let i = 0; i < arr.length; i++) {
      let val = arr[i];
      
      if (isArray$3(val)) {
        _transferArrayOfReferences$2(sourceDoc, targetDoc, val, visited);
      } else {
        let id = val;
        if (id && !visited[id]) {
          arr[i] = _transferWithDisambiguatedIds$2(sourceDoc, targetDoc, id, visited);
        }
      }
    }
  }

  class Editing$2 {
    
    annotate (tx, annotation) {
      let sel = tx.selection;
      let schema = tx.getSchema();
      let AnnotationClass = schema.getNodeClass(annotation.type);
      if (!AnnotationClass) throw new Error('Unknown annotation type', annotation)
      let start = sel.start;
      let end = sel.end;
      let containerId = sel.containerId;
      let nodeData = { start, end, containerId };
      
      
      if (sel.isPropertySelection()) {
        if (!AnnotationClass.prototype._isAnnotation) {
          throw new Error('Annotation can not be created for a selection.')
        }
      } else if (sel.isContainerSelection()) {
        if (AnnotationClass.prototype._isPropertyAnnotation) {
          console.warn('NOT SUPPORTED YET: creating property annotations for a non collapsed container selection.');
        }
      }
      Object.assign(nodeData, annotation);
      return tx.create(nodeData)
    }

    break (tx) {
      let sel = tx.selection;
      if (sel.isNodeSelection()) {
        let containerId = sel.containerId;
        let container = tx.get(containerId);
        let nodeId = sel.getNodeId();
        let nodePos = container.getPosition(nodeId, 'strict');
        let textNode = this.createTextNode(tx, container);
        if (sel.isBefore()) {
          container.showAt(nodePos, textNode.id);
          
        } else {
          container.showAt(nodePos + 1, textNode.id);
          setCursor$2(tx, textNode, containerId, 'before');
        }
      } else if (sel.isCustomSelection()) ; else if (sel.isCollapsed() || sel.isPropertySelection()) {
        let containerId = sel.containerId;
        if (!sel.isCollapsed()) {
          
          this._deletePropertySelection(tx, sel);
          tx.setSelection(sel.collapse('left'));
        }
        
        if (containerId) {
          let container = tx.get(containerId);
          let nodeId = sel.start.path[0];
          let node = tx.get(nodeId);
          this._breakNode(tx, node, sel.start, container);
        }
      } else if (sel.isContainerSelection()) {
        let start = sel.start;
        let containerId = sel.containerId;
        let container = tx.get(containerId);
        let startNodeId = start.path[0];
        let nodePos = container.getPosition(startNodeId, 'strict');
        this._deleteContainerSelection(tx, sel, { noMerge: true });
        setCursor$2(tx, container.getNodeAt(nodePos + 1), containerId, 'before');
      }
    }

    createTextNode (tx, container, text) { 
      
      
      return tx.createDefaultTextNode(text)
    }

    createListNode (tx, container, params = {}) { 
      
      
      return tx.create({ type: 'list', items: [], listType: params.listType || 'bullet' })
    }

    delete (tx, direction) {
      let sel = tx.selection;
      
      
      
      
      if (sel.isNodeSelection()) {
        this._deleteNodeSelection(tx, sel, direction);
      
      } else if (sel.isCustomSelection()) ; else if (sel.isCollapsed()) {
        
        
        
        let path = sel.start.path;
        let node = tx.get(path[0]);
        let text = tx.get(path);
        let offset = sel.start.offset;
        let needsMerge = (sel.containerId && (
          (offset === 0 && direction === 'left') ||
          (offset === text.length && direction === 'right')
        ));
        if (needsMerge) {
          
          
          
          
          
          
          
          let root = node.getContainerRoot();
          if (root.isList() && offset === 0 && direction === 'left') {
            return this.toggleList(tx)
          } else {
            let container = tx.get(sel.containerId);
            this._merge(tx, root, sel.start, direction, container);
          }
        } else {
          
          if ((offset === 0 && direction === 'left') ||
            (offset === text.length && direction === 'right')) {
            return
          }
          let startOffset = (direction === 'left') ? offset - 1 : offset;
          let endOffset = startOffset + 1;
          let start = { path: path, offset: startOffset };
          let end = { path: path, offset: endOffset };
          deleteTextRange$2(tx, start, end);
          tx.setSelection({
            type: 'property',
            path: path,
            startOffset: startOffset,
            containerId: sel.containerId
          });
        }
      
      } else if (sel.isPropertySelection()) {
        deleteTextRange$2(tx, sel.start, sel.end);
        tx.setSelection(sel.collapse('left'));
      
      } else if (sel.isContainerSelection()) {
        this._deleteContainerSelection(tx, sel);
      } else {
        console.warn('Unsupported case: tx.delete(%)', direction, sel);
      }
    }

    _deleteNodeSelection (tx, sel, direction) {
      let nodeId = sel.getNodeId();
      let container = tx.get(sel.containerId);
      let nodePos = container.getPosition(nodeId, 'strict');
      if (sel.isFull() ||
          (sel.isBefore() && direction === 'right') ||
          (sel.isAfter() && direction === 'left')) {
        
        container.hideAt(nodePos);
        deleteNode$2(tx, tx.get(nodeId));
        let newNode = this.createTextNode(tx, container);
        container.showAt(nodePos, newNode.id);
        tx.setSelection({
          type: 'property',
          path: newNode.getPath(),
          startOffset: 0,
          containerId: container.id
        });
      } else {
        
        if (sel.isBefore() && direction === 'left') {
          if (nodePos > 0) {
            let previous = container.getNodeAt(nodePos - 1);
            if (previous.isText()) {
              tx.setSelection({
                type: 'property',
                path: previous.getPath(),
                startOffset: previous.getLength()
              });
              this.delete(tx, direction);
            } else {
              tx.setSelection({
                type: 'node',
                nodeId: previous.id,
                containerId: container.id
              });
            }
          }
        } else if (sel.isAfter() && direction === 'right') {
          if (nodePos < container.getLength() - 1) {
            let next = container.getNodeAt(nodePos + 1);
            if (next.isText()) {
              tx.setSelection({
                type: 'property',
                path: next.getPath(),
                startOffset: 0
              });
              this.delete(tx, direction);
            } else {
              tx.setSelection({
                type: 'node',
                nodeId: next.id,
                containerId: container.id
              });
            }
          }
        } else {
          console.warn('Unsupported case: delete(%s)', direction, sel);
        }
      }
    }

    _deletePropertySelection (tx, sel) {
      let path = sel.start.path;
      let start = sel.start.offset;
      let end = sel.end.offset;
      tx.update(path, { type: 'delete', start: start, end: end });
      annotationHelpers$2.deletedText(tx, path, start, end);
    }

    
    _deleteContainerSelection (tx, sel, options = {}) {
      let containerId = sel.containerId;
      let container = tx.get(containerId);
      let start = sel.start;
      let end = sel.end;
      let startId = start.getNodeId();
      let endId = end.getNodeId();
      let startPos = container.getPosition(startId, 'strict');
      let endPos = container.getPosition(endId, 'strict');

      
      if (startPos === endPos) {
        
        let node = tx.get(startId).getContainerRoot();
        
        if (node.isText()) {
          deleteTextRange$2(tx, start, end);
        } else if (node.isList()) {
          deleteListRange$2(tx, node, start, end);
        } else {
          throw new Error('Not supported yet.')
        }
        tx.setSelection(sel.collapse('left'));
        return
      }

      

      let firstNode = tx.get(start.getNodeId());
      let lastNode = tx.get(end.getNodeId());
      let firstEntirelySelected = isEntirelySelected$2(tx, firstNode, start, null);
      let lastEntirelySelected = isEntirelySelected$2(tx, lastNode, null, end);

      
      if (lastEntirelySelected) {
        container.hideAt(endPos);
        deleteNode$2(tx, lastNode);
      } else {
        
        let node = lastNode.getContainerRoot();
        
        if (node.isText()) {
          deleteTextRange$2(tx, null, end);
        } else if (node.isList()) {
          deleteListRange$2(tx, node, null, end);
        }
      }

      
      for (let i = endPos - 1; i > startPos; i--) {
        let nodeId = container.getNodeIdAt(i);
        container.hideAt(i);
        deleteNode$2(tx, tx.get(nodeId));
      }

      
      if (firstEntirelySelected) {
        container.hideAt(startPos);
        deleteNode$2(tx, firstNode);
      } else {
        
        let node = firstNode.getContainerRoot();
        
        if (node.isText()) {
          deleteTextRange$2(tx, start, null);
        } else if (node.isList()) {
          deleteListRange$2(tx, node, start, null);
        }
      }

      
      if (firstEntirelySelected && lastEntirelySelected) {
        
        let textNode = this.createTextNode(tx, container);
        container.showAt(startPos, textNode.id);
        tx.setSelection({
          type: 'property',
          path: textNode.getPath(),
          startOffset: 0,
          containerId: containerId
        });
      } else if (!firstEntirelySelected && !lastEntirelySelected) {
        if (!options.noMerge) {
          this._merge(tx, firstNode, sel.start, 'right', container);
        }
        tx.setSelection(sel.collapse('left'));
      } else if (firstEntirelySelected) {
        setCursor$2(tx, lastNode, container.id, 'before');
      } else {
        setCursor$2(tx, firstNode, container.id, 'after');
      }
    }

    insertInlineNode (tx, nodeData) {
      let sel = tx.selection;
      let text = '\uFEFF';
      this.insertText(tx, text);
      sel = tx.selection;
      let endOffset = tx.selection.end.offset;
      let startOffset = endOffset - text.length;
      nodeData = Object.assign({}, nodeData, {
        start: {
          path: sel.path,
          offset: startOffset
        },
        end: {
          path: sel.path,
          offset: endOffset
        }
      });
      return tx.create(nodeData)
    }

    insertBlockNode (tx, nodeData) {
      let sel = tx.selection;
      
      let blockNode;
      if (!nodeData._isNode || !tx.get(nodeData.id)) {
        blockNode = tx.create(nodeData);
      } else {
        blockNode = tx.get(nodeData.id);
      }
      
      if (sel.isNodeSelection()) {
        let containerId = sel.containerId;
        let container = tx.get(containerId);
        let nodeId = sel.getNodeId();
        let nodePos = container.getPosition(nodeId, 'strict');
        
        if (sel.isBefore()) {
          container.showAt(nodePos, blockNode.id);
        
        } else if (sel.isAfter()) {
          container.showAt(nodePos + 1, blockNode.id);
          tx.setSelection({
            type: 'node',
            containerId: containerId,
            nodeId: blockNode.id,
            mode: 'after'
          });
        } else {
          container.hideAt(nodePos);
          deleteNode$2(tx, tx.get(nodeId));
          container.showAt(nodePos, blockNode.id);
          tx.setSelection({
            type: 'node',
            containerId: containerId,
            nodeId: blockNode.id,
            mode: 'after'
          });
        }
      } else if (sel.isPropertySelection()) {
        
        if (!sel.containerId) throw new Error('insertBlockNode can only be used within a container.')
        let container = tx.get(sel.containerId);
        if (!sel.isCollapsed()) {
          this._deletePropertySelection(tx, sel);
          tx.setSelection(sel.collapse('left'));
        }
        let node = tx.get(sel.path[0]);
        
        if (!node) throw new Error('Invalid selection.')
        let nodePos = container.getPosition(node.id, 'strict');
        
        if (node.isText()) {
          let text = node.getText();
          
          if (text.length === 0) {
            container.hideAt(nodePos);
            deleteNode$2(tx, node);
            container.showAt(nodePos, blockNode.id);
            setCursor$2(tx, blockNode, container.id, 'after');
          
          } else if (sel.start.offset === 0) {
            container.showAt(nodePos, blockNode.id);
          
          } else if (sel.start.offset === text.length) {
            container.showAt(nodePos + 1, blockNode.id);
            setCursor$2(tx, blockNode, container.id, 'before');
          
          } else {
            this.break(tx);
            container.showAt(nodePos + 1, blockNode.id);
            setCursor$2(tx, blockNode, container.id, 'after');
          }
        } else {
          console.error('Not supported: insertBlockNode() on a custom node');
        }
      } else if (sel.isContainerSelection()) {
        if (sel.isCollapsed()) {
          let start = sel.start;
          
          if (start.isPropertyCoordinate()) {
            tx.setSelection({
              type: 'property',
              path: start.path,
              startOffset: start.offset,
              containerId: sel.containerId
            });
          } else if (start.isNodeCoordinate()) {
            tx.setSelection({
              type: 'node',
              containerId: sel.containerId,
              nodeId: start.path[0],
              mode: start.offset === 0 ? 'before' : 'after'
            });
          } else {
            throw new Error('Unsupported selection for insertBlockNode')
          }
          return this.insertBlockNode(tx, blockNode)
        } else {
          this.break(tx);
          return this.insertBlockNode(tx, blockNode)
        }
      }
      return blockNode
    }

    insertText (tx, text) {
      let sel = tx.selection;
      
      
      
      if (sel.isNodeSelection()) {
        let containerId = sel.containerId;
        let container = tx.get(containerId);
        let nodeId = sel.getNodeId();
        let nodePos = container.getPosition(nodeId, 'strict');
        let textNode = this.createTextNode(tx, container, text);
        if (sel.isBefore()) {
          container.showAt(nodePos, textNode);
        } else if (sel.isAfter()) {
          container.showAt(nodePos + 1, textNode);
        } else {
          container.hide(nodeId);
          deleteNode$2(tx, tx.get(nodeId));
          container.showAt(nodePos, textNode);
        }
        setCursor$2(tx, textNode, sel.containerId, 'after');
      } else if (sel.isCustomSelection()) ; else if (sel.isCollapsed() || sel.isPropertySelection()) {
        
        this._insertText(tx, sel, text);
        
      } else if (sel.isContainerSelection()) {
        this._deleteContainerSelection(tx, sel);
        this.insertText(tx, text);
      }
    }

    paste (tx, content) {
      if (!content) return
      
      if (isString$2(content)) {
        paste$2(tx, {text: content});
      } else if (content._isDocument) {
        paste$2(tx, {doc: content});
      } else {
        throw new Error('Illegal content for paste.')
      }
    }

    
    switchTextType (tx, data) {
      let sel = tx.selection;
      
      if (!sel.isPropertySelection()) {
        throw new Error('Selection must be a PropertySelection.')
      }
      let containerId = sel.containerId;
      
      if (!containerId) {
        throw new Error('Selection must be within a container.')
      }
      let path = sel.path;
      let nodeId = path[0];
      let node = tx.get(nodeId);
      
      if (!(node.isInstanceOf('text'))) {
        throw new Error('Trying to use switchTextType on a non text node.')
      }
      const newId = uuid$2(data.type);
      
      const oldPath = node.getPath();
      console.assert(oldPath.length === 2, 'Currently we assume that TextNodes store the plain-text on the first level');
      const textProp = oldPath[1];
      let newPath = [newId, textProp];
      
      let newNodeData = Object.assign({
        id: newId,
        type: data.type,
        direction: node.direction
      }, data);
      newNodeData[textProp] = node.getText();

      let newNode = tx.create(newNodeData);
      annotationHelpers$2.transferAnnotations(tx, path, 0, newPath, 0);

      
      let container = tx.get(sel.containerId);
      let pos = container.getPosition(nodeId, 'strict');
      container.hide(nodeId);
      deleteNode$2(tx, node);
      container.showAt(pos, newNode.id);

      tx.setSelection({
        type: 'property',
        path: newPath,
        startOffset: sel.start.offset,
        endOffset: sel.end.offset,
        containerId: containerId
      });

      return newNode
    }

    toggleList (tx, params) {
      let sel = tx.selection;
      let container = tx.get(sel.containerId);
      
      if (!container) {
        throw new Error('Selection must be within a container.')
      }
      if (sel.isPropertySelection()) {
        let nodeId = sel.start.path[0];
        
        let node = tx.get(nodeId).getContainerRoot();
        let nodePos = container.getPosition(node.id, 'strict');
        
        if (node.isText()) {
          container.hideAt(nodePos);
          let newList = this.createListNode(tx, container, params);
          let newItem = newList.createListItem(node.getText());
          annotationHelpers$2.transferAnnotations(tx, node.getPath(), 0, newItem.getPath(), 0);
          newList.appendItem(newItem);
          deleteNode$2(tx, node);
          container.showAt(nodePos, newList.id);
          tx.setSelection({
            type: 'property',
            path: newItem.getPath(),
            startOffset: sel.start.offset,
            containerId: sel.containerId
          });
        } else if (node.isList()) {
          let itemId = sel.start.path[0];
          let item = tx.get(itemId);
          let itemPos = node.getItemPosition(item);
          let newTextNode = this.createTextNode(tx, container, item.getText());
          annotationHelpers$2.transferAnnotations(tx, item.getPath(), 0, newTextNode.getPath(), 0);
          
          node.removeItemAt(itemPos);
          if (node.isEmpty()) {
            container.hideAt(nodePos);
            deleteNode$2(tx, node);
            container.showAt(nodePos, newTextNode.id);
          } else if (itemPos === 0) {
            container.showAt(nodePos, newTextNode.id);
          } else if (node.getLength() <= itemPos) {
            container.showAt(nodePos + 1, newTextNode.id);
          } else {
            
            let tail = [];
            const items = node.getItems();
            const L = items.length;
            for (let i = L - 1; i >= itemPos; i--) {
              tail.unshift(items[i]);
              node.removeItemAt(i);
            }
            let newList = this.createListNode(tx, container, node);
            for (let i = 0; i < tail.length; i++) {
              newList.appendItem(tail[i]);
            }
            container.showAt(nodePos + 1, newTextNode.id);
            container.showAt(nodePos + 2, newList.id);
          }
          tx.setSelection({
            type: 'property',
            path: newTextNode.getPath(),
            startOffset: sel.start.offset,
            containerId: sel.containerId
          });
        }
      } else if (sel.isContainerSelection()) {
        console.error('TODO: support toggleList with ContainerSelection');
      }
    }

    indent (tx) {
      let sel = tx.selection;
      if (sel.isPropertySelection()) {
        let nodeId = sel.start.getNodeId();
        
        let node = tx.get(nodeId).getContainerRoot();
        if (node.isList()) {
          let itemId = sel.start.path[0];
          let item = tx.get(itemId);
          let level = item.getLevel();
          
          if (item && level < 3) {
            item.setLevel(item.level + 1);
            
            tx.set([node.id, '_itemsChanged'], true);
          }
        }
      } else if (sel.isContainerSelection()) {
        console.error('TODO: support toggleList with ContainerSelection');
      }
    }

    dedent (tx) {
      let sel = tx.selection;
      if (sel.isPropertySelection()) {
        let nodeId = sel.start.getNodeId();
        
        let node = tx.get(nodeId).getContainerRoot();
        if (node.isList()) {
          let itemId = sel.start.path[0];
          let item = tx.get(itemId);
          let level = item.getLevel();
          if (item) {
            if (level > 1) {
              item.setLevel(item.level - 1);
              
              tx.set([node.id, '_itemsChanged'], true);
            }
            
            
            
            
            
          }
        }
      } else if (sel.isContainerSelection()) {
        console.error('TODO: support toggleList with ContainerSelection');
      }
    }

    
    _insertText (tx, sel, text) {
      let start = sel.start;
      let end = sel.end;
      
      if (!isArrayEqual$2(start.path, end.path)) {
        throw new Error('Unsupported state: range should be on one property')
      }
      let path = start.path;
      let startOffset = start.offset;
      let endOffset = end.offset;
      let typeover = !sel.isCollapsed();
      let L = text.length;
      
      if (typeover) {
        tx.update(path, { type: 'delete', start: startOffset, end: endOffset });
      }
      
      tx.update(path, { type: 'insert', start: startOffset, text: text });
      
      let annos = tx.getAnnotations(path);
      annos.forEach(function (anno) {
        let annoStart = anno.start.offset;
        let annoEnd = anno.end.offset;

        
        
        if (annoEnd < startOffset) ; else if (annoStart >= endOffset) {
          tx.update([anno.id, 'start'], { type: 'shift', value: startOffset - endOffset + L });
          tx.update([anno.id, 'end'], { type: 'shift', value: startOffset - endOffset + L });
        
        
        
        
        } else if (
          (annoStart >= startOffset && annoEnd < endOffset) ||
          (anno._isInlineNode && annoStart >= startOffset && annoEnd <= endOffset)
        ) {
          tx.delete(anno.id);
        
        } else if (annoStart >= startOffset && annoEnd >= endOffset) {
          
          if (annoStart > startOffset || !typeover) {
            tx.update([anno.id, 'start'], { type: 'shift', value: startOffset - annoStart + L });
          }
          tx.update([anno.id, 'end'], { type: 'shift', value: startOffset - endOffset + L });
        
        } else if (annoStart < startOffset && annoEnd < endOffset) {
          
          tx.update([anno.id, 'end'], { type: 'shift', value: startOffset - annoEnd + L });
        
        } else if (annoEnd === startOffset && !anno.constructor.autoExpandRight) ; else if (annoStart < startOffset && annoEnd >= endOffset) {
          if (anno._isInlineNode) ; else {
            tx.update([anno.id, 'end'], { type: 'shift', value: startOffset - endOffset + L });
          }
        } else {
          console.warn('TODO: handle annotation update case.');
        }
      });
      let offset = startOffset + text.length;
      tx.setSelection({
        type: 'property',
        path: start.path,
        startOffset: offset,
        containerId: sel.containerId,
        surfaceId: sel.surfaceId
      });
    }

    _breakNode (tx, node, coor, container) {
      
      node = node.getContainerRoot();
      
      if (node.isText()) {
        this._breakTextNode(tx, node, coor, container);
      } else if (node.isList()) {
        this._breakListNode(tx, node, coor, container);
      } else {
        throw new Error('Not supported')
      }
    }

    _breakTextNode (tx, node, coor, container) {
      let path = coor.path;
      let offset = coor.offset;
      let nodePos = container.getPosition(node.id, 'strict');
      let text = node.getText();

      
      
      if (offset === 0) {
        let newNode = tx.create({
          type: node.type,
          content: ''
        });
        
        container.showAt(nodePos, newNode.id);
        tx.setSelection({
          type: 'property',
          path: path,
          startOffset: 0,
          containerId: container.id
        });
      
      } else {
        const textPath = node.getPath();
        const textProp = textPath[1];
        const newId = uuid$2(node.type);
        let newNodeData = node.toJSON();
        newNodeData.id = newId;
        newNodeData[textProp] = text.substring(offset);
        
        if (offset === text.length) {
          newNodeData.type = tx.getSchema().getDefaultTextType();
        }
        let newNode = tx.create(newNodeData);
        
        if (offset < text.length) {
          
          annotationHelpers$2.transferAnnotations(tx, path, offset, newNode.getPath(), 0);
          
          tx.update(path, { type: 'delete', start: offset, end: text.length });
        }
        
        container.showAt(nodePos + 1, newNode.id);
        
        tx.setSelection({
          type: 'property',
          path: newNode.getPath(),
          startOffset: 0,
          containerId: container.id
        });
      }
    }

    _breakListNode (tx, node, coor, container) {
      let path = coor.path;
      let offset = coor.offset;
      let listItem = tx.get(path[0]);

      let L = node.length;
      let itemPos = node.getItemPosition(listItem);
      let text = listItem.getText();
      let textProp = listItem.getPath()[1];
      let newItemData = listItem.toJSON();
      delete newItemData.id;
      if (offset === 0) {
        
        if (!text) {
          
          
          let nodePos = container.getPosition(node.id, 'strict');
          let newTextNode = this.createTextNode(tx, container);
          
          if (L < 2) {
            container.hide(node.id);
            deleteNode$2(tx, node);
            container.showAt(nodePos, newTextNode.id);
          
          } else if (itemPos === 0) {
            node.removeItem(listItem);
            deleteNode$2(tx, listItem);
            container.showAt(nodePos, newTextNode.id);
          
          } else if (itemPos >= L - 1) {
            node.removeItem(listItem);
            deleteNode$2(tx, listItem);
            container.showAt(nodePos + 1, newTextNode.id);
          
          } else {
            let tail = [];
            const items = node.getItems().slice();
            for (let i = L - 1; i > itemPos; i--) {
              tail.unshift(items[i]);
              node.removeItem(items[i]);
            }
            node.removeItem(items[itemPos]);
            let newList = this.createListNode(tx, container, node);
            for (let i = 0; i < tail.length; i++) {
              newList.appendItem(tail[i]);
            }
            container.showAt(nodePos + 1, newTextNode.id);
            container.showAt(nodePos + 2, newList.id);
          }
          tx.setSelection({
            type: 'property',
            path: newTextNode.getPath(),
            startOffset: 0
          });
        
        } else {
          newItemData[textProp] = '';
          let newItem = tx.create(newItemData);
          node.insertItemAt(itemPos, newItem);
          tx.setSelection({
            type: 'property',
            path: listItem.getPath(),
            startOffset: 0
          });
        }
      
      } else {
        newItemData[textProp] = text.substring(offset);
        let newItem = tx.create(newItemData);
        
        if (offset < text.length) {
          
          annotationHelpers$2.transferAnnotations(tx, path, offset, newItem.getPath(), 0);
          
          tx.update(path, { type: 'delete', start: offset, end: text.length });
        }
        node.insertItemAt(itemPos + 1, newItem);
        tx.setSelection({
          type: 'property',
          path: newItem.getPath(),
          startOffset: 0
        });
      }
    }

    _merge (tx, node, coor, direction, container) {
      
      
      if (node.isList()) {
        let list = node;
        let itemId = coor.path[0];
        let item = tx.get(itemId);
        let itemPos = list.getItemPosition(item);
        let withinListNode = (
          (direction === 'left' && itemPos > 0) ||
          (direction === 'right' && itemPos < list.items.length - 1)
        );
        if (withinListNode) {
          itemPos = (direction === 'left') ? itemPos - 1 : itemPos;
          let target = list.getItemAt(itemPos);
          let targetLength = target.getLength();
          mergeListItems$2(tx, list.id, itemPos);
          tx.setSelection({
            type: 'property',
            path: target.getPath(),
            startOffset: targetLength,
            containerId: container.id
          });
          return
        }
      }
      
      let nodePos = container.getPosition(node, 'strict');
      if (direction === 'left' && nodePos > 0) {
        this._mergeNodes(tx, container, nodePos - 1, direction);
      } else if (direction === 'right' && nodePos < container.getLength() - 1) {
        this._mergeNodes(tx, container, nodePos, direction);
      }
    }

    _mergeNodes (tx, container, pos, direction) {
      let first = container.getChildAt(pos);
      let second = container.getChildAt(pos + 1);
      if (first.isText()) {
        
        if (first.isEmpty()) {
          container.hide(first.id);
          deleteNode$2(tx, first);
          
          
          setCursor$2(tx, second, container.id, 'before');
          return
        }
        let target = first;
        let targetPath = target.getPath();
        let targetLength = target.getLength();
        if (second.isText()) {
          let source = second;
          let sourcePath = source.getPath();
          container.hide(source.id);
          
          tx.update(targetPath, { type: 'insert', start: targetLength, text: source.getText() });
          
          annotationHelpers$2.transferAnnotations(tx, sourcePath, 0, targetPath, targetLength);
          deleteNode$2(tx, source);
          tx.setSelection({
            type: 'property',
            path: targetPath,
            startOffset: targetLength,
            containerId: container.id
          });
        } else if (second.isList()) {
          let list = second;
          if (!second.isEmpty()) {
            let source = list.getFirstItem();
            let sourcePath = source.getPath();
            
            list.removeItemAt(0);
            
            tx.update(targetPath, { type: 'insert', start: targetLength, text: source.getText() });
            
            annotationHelpers$2.transferAnnotations(tx, sourcePath, 0, targetPath, targetLength);
            
            deleteNode$2(tx, source);
          }
          if (list.isEmpty()) {
            container.hide(list.id);
            deleteNode$2(tx, list);
          }
          tx.setSelection({
            type: 'property',
            path: targetPath,
            startOffset: targetLength,
            containerId: container.id
          });
        } else {
          selectNode$2(tx, direction === 'left' ? first.id : second.id, container.id);
        }
      } else if (first.isList()) {
        if (second.isText()) {
          let target = first.getLastItem();
          let targetPath = target.getPath();
          let targetLength = target.getLength();
          let third = (container.length > pos + 2) ? container.getChildAt(pos + 2) : null;
          if (second.getLength() === 0) {
            container.hide(second.id);
            deleteNode$2(tx, second);
          } else {
            let source = second;
            let sourcePath = source.getPath();
            container.hide(source.id);
            tx.update(targetPath, { type: 'insert', start: targetLength, text: source.getText() });
            annotationHelpers$2.transferAnnotations(tx, sourcePath, 0, targetPath, targetLength);
            deleteNode$2(tx, source);
          }
          
          if (third && third.type === first.type) {
            this._mergeTwoLists(tx, container, first, third);
          }
          tx.setSelection({
            type: 'property',
            path: target.getPath(),
            startOffset: targetLength,
            containerId: container.id
          });
        } else if (second.isList()) {
          
          if (direction !== 'right') {
            
            
            throw new Error('Illegal state')
          }
          let item = first.getLastItem();
          this._mergeTwoLists(tx, container, first, second);
          tx.setSelection({
            type: 'property',
            path: item.getPath(),
            startOffset: item.getLength(),
            containerId: container.id
          });
        } else {
          selectNode$2(tx, direction === 'left' ? first.id : second.id, container.id);
        }
      } else {
        if (second.isText() && second.isEmpty()) {
          container.hide(second.id);
          deleteNode$2(tx, second);
          setCursor$2(tx, first, container.id, 'after');
        } else {
          selectNode$2(tx, direction === 'left' ? first.id : second.id, container.id);
        }
      }
    }

    _mergeTwoLists (tx, container, first, second) {
      container.hide(second.id);
      let secondItems = second.getItems().slice();
      for (let i = 0; i < secondItems.length; i++) {
        second.removeItemAt(0);
        first.appendItem(secondItems[i]);
      }
      deleteNode$2(tx, second);
    }
  }

  class EditingInterface$2 {
    constructor (doc, options = {}) {
      this._document = doc;
      this._selection = null;
      
      this._impl = options.editing || new Editing$2();
      this._direction = null;
    }

    dispose () {}

    getDocument () {
      return this._document
    }

    

    get (...args) {
      return this._document.get(...args)
    }

    contains (id) {
      return this._document.contains(id)
    }

    create (nodeData) {
      return this._document.create(nodeData)
    }

    createDefaultTextNode (content) {
      return this._document.createDefaultTextNode(content, this._direction)
    }

    delete (nodeId) {
      return this._document.delete(nodeId)
    }

    set (path, value) {
      return this._document.set(path, value)
    }

    update (path, diffOp) {
      return this._document.update(path, diffOp)
    }

    updateNode (id, newProps) {
      return this._document.updateNode(id, newProps)
    }

    

    createSelection (selData) {
      
      
      
      
      
      
      
      selData = augmentSelection$2(selData, this._selection);
      return this._document.createSelection(selData)
    }

    setSelection (sel) {
      if (!sel) {
        sel = Selection$2.nullSelection;
      } else if (isPlainObject$3(sel)) {
        sel = this.createSelection(sel);
      } else {
        sel = augmentSelection$2(sel, this._selection);
      }
      this._selection = sel;
      return sel
    }

    getSelection () {
      return this._selection
    }

    get selection () {
      return this._selection
    }

    set selection (sel) {
      this.setSelection(sel);
    }

    
    get textDirection () {
      return this._direction
    }

    set textDirection (dir) {
      this._direction = dir;
    }

    

    annotate (annotationData) {
      const sel = this._selection;
      if (sel && (sel.isPropertySelection() || sel.isContainerSelection())) {
        return this._impl.annotate(this, annotationData)
      }
    }

    break () {
      if (this._selection && !this._selection.isNull()) {
        this._impl.break(this);
      }
    }

    copySelection () {
      const sel = this._selection;
      if (sel && !sel.isNull() && !sel.isCollapsed()) {
        return copySelection$2(this.getDocument(), this._selection)
      }
    }

    deleteSelection (options) {
      const sel = this._selection;
      if (sel && !sel.isNull() && !sel.isCollapsed()) {
        this._impl.delete(this, 'right', options);
      }
    }

    deleteCharacter (direction) {
      const sel = this._selection;
      if (!sel || sel.isNull()) ; else if (!sel.isCollapsed()) {
        this.deleteSelection();
      } else {
        this._impl.delete(this, direction);
      }
    }

    insertText (text) {
      const sel = this._selection;
      if (sel && !sel.isNull()) {
        this._impl.insertText(this, text);
      }
    }

    
    insertInlineNode (inlineNode) {
      const sel = this._selection;
      if (sel && !sel.isNull() && sel.isPropertySelection()) {
        return this._impl.insertInlineNode(this, inlineNode)
      }
    }

    insertBlockNode (blockNode) {
      const sel = this._selection;
      if (sel && !sel.isNull()) {
        return this._impl.insertBlockNode(this, blockNode)
      }
    }

    paste (content) {
      const sel = this._selection;
      if (sel && !sel.isNull() && !sel.isCustomSelection()) {
        return this._impl.paste(this, content)
      }
    }

    switchTextType (nodeData) {
      const sel = this._selection;
      if (sel && !sel.isNull()) {
        return this._impl.switchTextType(this, nodeData)
      }
    }

    toggleList (params) {
      const sel = this._selection;
      if (sel && !sel.isNull()) {
        return this._impl.toggleList(this, params)
      }
    }

    indent () {
      const sel = this._selection;
      if (sel && !sel.isNull()) {
        return this._impl.indent(this)
      }
    }

    dedent () {
      const sel = this._selection;
      if (sel && !sel.isNull()) {
        return this._impl.dedent(this)
      }
    }

    

    getIndex (...args) {
      return this._document.getIndex(...args)
    }

    getAnnotations (...args) {
      return this._document.getAnnotations(...args)
    }

    getSchema () {
      return this._document.getSchema()
    }

    createSnippet () {
      return this._document.createSnippet()
    }
  }

  class JSONConverter$2 {
    importDocument (doc, json) {
      if (!json.nodes) {
        throw new Error('Invalid JSON format.')
      }
      var schema = doc.getSchema();
      if (json.schema && schema.name !== json.schema.name) {
        throw new Error('Incompatible schema.')
      }
      
      var nodes = json.nodes;
      
      
      
      doc.import(function (tx) {
        forEach$2(nodes, function (node) {
          
          if (tx.get(node.id)) {
            tx.delete(node.id);
          }
          tx.create(node);
        });
      });
      return doc
    }

    exportDocument (doc) {
      var schema = doc.getSchema();
      var json = {
        schema: {
          name: schema.name
        },
        nodes: {}
      };
      forEach$2(doc.getNodes(), function (node) {
        if (node._isDocumentNode) {
          json.nodes[node.id] = node.toJSON();
        }
      });
      return json
    }
  }

  class ParentNodeHook$2 {
    constructor (doc) {
      this.doc = doc;
      
      this.parents = {};
      doc.data.on('operation:applied', this._onOperationApplied, this);
    }

    _onOperationApplied (op) {
      const doc = this.doc;
      let node = doc.get(op.path[0]);
      let nodeSchema, hasOwnedProperties;
      if (node) {
        nodeSchema = node.getSchema();
        hasOwnedProperties = nodeSchema.hasOwnedProperties();
      }
      
      
      switch (op.type) {
        case 'create': {
          if (hasOwnedProperties) {
            nodeSchema.getOwnedProperties().forEach(p => {
              let refs = node[p.name];
              if (refs) {
                this._setParent(node, refs);
              }
              this._setRegisteredParent(node);
            });
          }
          break
        }
        case 'update': {
          if (hasOwnedProperties) {
            let propName = op.path[1];
            if (nodeSchema.isOwned(propName)) {
              let update = op.diff;
              if (update.isDelete()) {
                this._setParent(null, update.getValue());
              } else {
                this._setParent(node, update.getValue());
              }
            }
          }
          break
        }
        case 'set': {
          if (hasOwnedProperties) {
            let propName = op.path[1];
            if (nodeSchema.isOwned(propName)) {
              let oldValue = op.getOldValue();
              let newValue = op.getValue();
              
              this._setParent(null, oldValue);
              this._setParent(node, newValue);
            }
          }
          break
        }
        default:
          
      }
    }

    _setParent (parent, ids) {
      if (ids) {
        if (isArray$3(ids)) {
          ids.forEach(id => this.__setParent(parent, id));
        } else {
          this.__setParent(parent, ids);
        }
      }
    }

    __setParent (parent, id) {
      
      
      
      
      
      
      this.parents[id] = parent;
      let child = this.doc.get(id);
      if (child) {
        child.setParent(parent);
      }
    }

    _setRegisteredParent (child) {
      let parent = this.parents[child.id];
      if (parent) {
        child.setParent(parent);
      }
    }
  }

  ParentNodeHook$2.register = function (doc) {
    return new ParentNodeHook$2(doc)
  };

  const converter$2 = new JSONConverter$2();



  class Document$3 extends EventEmitter$2 {
    
    constructor (schema, ...args) {
      super();

      this.schema = schema;
      
      if (!schema) {
        throw new Error('A document needs a schema for reflection.')
      }

      
      this._ops = [];

      this._initialize(...args);
    }

    _initialize () {
      this.__id__ = uuid$2();
      this.nodeFactory = new DocumentNodeFactory$2(this);
      this.data = new IncrementalData$2(this.schema, this.nodeFactory);
      
      this.addIndex('type', new PropertyIndex$2('type'));
      
      this.addIndex('annotations', new AnnotationIndex$2());
      
      
      
      this.addIndex('container-annotations', new ContainerAnnotationIndex$2());
      
      
      ParentNodeHook$2.register(this);
    }

    dispose () {
      this.off();
      this.data.off();
    }

    get id () {
      return this.__id__
    }

    
    getSchema () {
      return this.schema
    }

    
    contains (id) {
      return this.data.contains(id)
    }

    
    get (path, strict) {
      return this.data.get(path, strict)
    }

    
    getNodes () {
      return this.data.getNodes()
    }

    getAnnotations (path) {
      return this.getIndex('annotations').get(path)
    }

    
    import (importer) {
      try {
        this.data._stopIndexing();
        importer(this);
        this.data._startIndexing();
      } finally {
        this.data.queue = [];
        this.data._startIndexing();
      }
    }

    
    create (nodeData) {
      if (!nodeData.id) {
        nodeData.id = uuid$2(nodeData.type);
      }
      if (!nodeData.type) {
        throw new Error('No node type provided')
      }
      const op = this._create(nodeData);
      if (op) {
        this._ops.push(op);
        if (!this._isTransactionDocument) {
          this._emitChange(op);
        }
        return this.get(nodeData.id)
      }
    }

    createDefaultTextNode (text, dir) {
      return this.create({
        type: this.getSchema().getDefaultTextType(),
        content: text || '',
        direction: dir
      })
    }

    
    delete (nodeId) {
      const node = this.get(nodeId);
      const op = this._delete(nodeId);
      if (op) {
        this._ops.push(op);
        if (!this._isTransactionDocument) {
          this._emitChange(op);
        }
      }
      return node
    }

    
    set (path, value) {
      const oldValue = this.get(path);
      const op = this._set(path, value);
      if (op) {
        this._ops.push(op);
        if (!this._isTransactionDocument) {
          this._emitChange(op);
        }
      }
      return oldValue
    }

    
    update (path, diff) {
      const op = this._update(path, diff);
      if (op) {
        this._ops.push(op);
        if (!this._isTransactionDocument) {
          this._emitChange(op);
        }
      }
      return op
    }

    
    updateNode (id, newProps) {
      let node = this.get(id);
      forEach$2(newProps, (value, key) => {
        if (!isEqual$2(node[key], newProps[key])) {
          this.set([id, key], value);
        }
      });
    }

    
    addIndex (name, index) {
      return this.data.addIndex(name, index)
    }

    
    getIndex (name) {
      return this.data.getIndex(name)
    }

    
    createSelection (data) {
      let sel;
      if (isNil$2(data)) return Selection$2.nullSelection
      if (arguments.length !== 1 || !isPlainObject$3(data)) {
        throw new Error('Illegal argument: call createSelection({ type: ... }')
      } else {
        switch (data.type) {
          case 'property': {
            if (isNil$2(data.endOffset)) {
              data.endOffset = data.startOffset;
            }
            if (!data.hasOwnProperty('reverse')) {
              if (data.startOffset > data.endOffset) {
                [data.startOffset, data.endOffset] = [data.endOffset, data.startOffset];
                data.reverse = !data.reverse;
              }
            }
            
            let text = this.get(data.path, 'strict');
            if (data.startOffset < 0 || data.startOffset > text.length) {
              throw new Error('Invalid startOffset: target property has length ' + text.length + ', given startOffset is ' + data.startOffset)
            }
            if (data.endOffset < 0 || data.endOffset > text.length) {
              throw new Error('Invalid startOffset: target property has length ' + text.length + ', given endOffset is ' + data.endOffset)
            }
            sel = new PropertySelection$2(data);
            break
          }
          case 'container': {
            let container = this.get(data.containerId, 'strict');
            if (!container) throw new Error('Can not create ContainerSelection: container "' + data.containerId + '" does not exist.')
            let start = this._normalizeCoor({ path: data.startPath, offset: data.startOffset });
            let end = this._normalizeCoor({ path: data.endPath, offset: data.endOffset });
            let startAddress = container.getAddress(start);
            let endAddress = container.getAddress(end);
            if (!startAddress) {
              throw new Error('Invalid arguments for ContainerSelection: ', start.toString())
            }
            if (!endAddress) {
              throw new Error('Invalid arguments for ContainerSelection: ', end.toString())
            }
            if (!data.hasOwnProperty('reverse')) {
              if (endAddress.isBefore(startAddress, 'strict')) {
                [start, end] = [end, start];
                data.reverse = true;
              }
            }
            sel = new ContainerSelection$2(container.id, start.path, start.offset, end.path, end.offset, data.reverse, data.surfaceId);
            break
          }
          case 'node': {
            sel = createNodeSelection$2({
              doc: this,
              nodeId: data.nodeId,
              mode: data.mode,
              containerId: data.containerId,
              reverse: data.reverse,
              surfaceId: data.surfaceId
            });
            break
          }
          case 'custom': {
            sel = CustomSelection$2.fromJSON(data);
            break
          }
          default:
            throw new Error('Illegal selection type', data)
        }
      }
      if (!sel.isNull()) {
        sel.attach(this);
      }
      return sel
    }

    newInstance () {
      var DocumentClass = this.constructor;
      return new DocumentClass(this.schema)
    }

    
    createSnippet () {
      var snippet = this.newInstance();
      var snippetContainer = snippet.create({
        type: '@container',
        id: SNIPPET_ID$2
      });
      snippet.getContainer = function () {
        return snippetContainer
      };
      return snippet
    }

    createFromDocument (doc) {
      
      this.clear();

      
      
      
      
      
      
      
      let nodes = Object.values(doc.getNodes());
      let levels = {};
      let visited = new Set();
      nodes.forEach(n => {
        if (!visited.has(n)) this._computeDependencyLevel(n, levels, visited);
      });
      
      nodes.sort((a, b) => {
        return levels[b.id] - levels[a.id]
      });
      nodes.forEach(n => this.create(n));
      return this
    }

    _computeDependencyLevel (node, levels, visited) {
      if (!node) throw new Error('node was nil')
      if (visited.has(node)) throw new Error('Cyclic node dependency')
      visited.add(node);
      
      
      
      let level = 0;
      if (node.isAnnotation() || node.isInline()) {
        level = -1;
      } else {
        let parent = node.getParent();
        if (parent) {
          let parentLevel;
          if (levels.hasOwnProperty(parent.id)) {
            parentLevel = levels[parent.id];
          } else {
            parentLevel = this._computeDependencyLevel(parent, levels, visited);
          }
          level = parentLevel + 1;
        }
      }
      levels[node.id] = level;
      return level
    }

    
    toJSON () {
      return converter$2.exportDocument(this)
    }

    clone () {
      let copy = this.newInstance();
      copy.createFromDocument(this);
      return copy
    }

    clear () {
      this.data.clear();
      this._ops.length = 0;
    }

    
    createEditingInterface () {
      return new EditingInterface$2(this)
    }

    invert (change) {
      return change.invert()
    }

    _apply (documentChange) {
      forEach$2(documentChange.ops, (op) => {
        this._applyOp(op);
      });
      
      documentChange._extractInformation(this);
    }

    _applyOp (op) {
      this.data.apply(op);
      this.emit('operation:applied', op);
    }

    _create (nodeData) {
      return this.data.create(nodeData)
    }

    _delete (nodeId) {
      return this.data.delete(nodeId)
    }

    _set (path, value) {
      return this.data.set(path, value)
    }

    _update (path, diff) {
      return this.data.update(path, diff)
    }

    _emitChange (op) {
      const change = new DocumentChange$2([op], {}, {});
      change._extractInformation(this);
      this._notifyChangeListeners(change, { hidden: true });
    }

    _notifyChangeListeners (change, info) {
      info = info || {};
      this.emit('document:changed', change, info, this);
    }

    
    _createSelectionFromRange (range) {
      if (!range) return Selection$2.nullSelection
      let inOneNode = isEqual$2(range.start.path, range.end.path);
      if (inOneNode) {
        if (range.start.isNodeCoordinate()) {
          
          
          return new NodeSelection$2(range.containerId, range.start.getNodeId(), 'full', range.reverse, range.surfaceId)
        } else {
          return this.createSelection({
            type: 'property',
            path: range.start.path,
            startOffset: range.start.offset,
            endOffset: range.end.offset,
            reverse: range.reverse,
            containerId: range.containerId,
            surfaceId: range.surfaceId
          })
        }
      } else {
        return this.createSelection({
          type: 'container',
          startPath: range.start.path,
          startOffset: range.start.offset,
          endPath: range.end.path,
          endOffset: range.end.offset,
          reverse: range.reverse,
          containerId: range.containerId,
          surfaceId: range.surfaceId
        })
      }
    }

    _normalizeCoor ({ path, offset }) {
      
      if (path.length === 1) {
        let node = this.get(path[0]).getContainerRoot();
        if (node.isText()) {
          
          return new Coordinate$2(node.getPath(), offset === 0 ? 0 : node.getLength())
        } else if (node.isList()) {
          
          if (offset === 0) {
            let item = node.getItemAt(0);
            return new Coordinate$2(item.getPath(), 0)
          } else {
            let item = this.get(last$1$2(node.items));
            return new Coordinate$2(item.getPath(), item.getLength())
          }
        }
      }
      return new Coordinate$2(path, offset)
    }
  }

  Document$3.prototype._isDocument = true;

  class XMLParentNodeHook$2 {
    constructor (doc) {
      this.doc = doc;
      
      this.parents = {};
      doc.data.on('operation:applied', this._onOperationApplied, this);
    }

    _onOperationApplied (op) {
      const doc = this.doc;
      const parents = this.parents;
      let node = doc.get(op.path[0]);
      switch (op.type) {
        case 'create': {
          if (node._childNodes) {
            _setParent(node, node._childNodes);
          }
          _setRegisteredParent(node);
          break
        }
        case 'update': {
          
          
          let update = op.diff;
          if (op.path[1] === '_childNodes') {
            if (update.isInsert()) {
              _setParent(node, update.getValue());
            } else if (update.isDelete()) {
              _setParent(null, update.getValue());
            }
          }
          break
        }
        case 'set': {
          if (op.path[1] === '_childNodes') {
            _setParent(null, op.getOldValue());
            _setParent(node, op.getValue());
          }
          break
        }
        default:
          
      }

      function _setParent (parent, ids) {
        if (ids) {
          if (isArray$3(ids)) {
            ids.forEach(_set);
          } else {
            _set(ids);
          }
        }
        function _set (id) {
          
          
          
          parents[id] = parent;
          let child = doc.get(id);
          if (child) {
            child.parentNode = parent;
          }
        }
      }
      function _setRegisteredParent (child) {
        let parent = parents[child.id];
        if (parent) {
          child.parentNode = parent;
        }
      }
    }

    static register (doc) {
      return new XMLParentNodeHook$2(doc)
    }
  }

  class XMLEditingInterface$2 extends EditingInterface$2 {
    find (cssSelector) {
      return this.getDocument().find(cssSelector)
    }

    findAll (cssSelector) {
      return this.getDocument().findAll(cssSelector)
    }

    createElement (...args) {
      return this.getDocument().createElement(...args)
    }
  }

  class XMLDocument$2 extends Document$3 {
    _initialize () {
      this.nodeFactory = new DocumentNodeFactory$2(this);
      this.data = new IncrementalData$2(this.schema, this.nodeFactory);
      
      this.addIndex('type', new PropertyIndex$2('type'));
      
      this.addIndex('annotations', new AnnotationIndex$2());
      XMLParentNodeHook$2.register(this);
    }

    toXML () {
      let dom = DefaultDOMElement$2.createDocument('xml');
      dom.setDoctype(...this.getDocTypeParams());
      let xml = dom.createProcessingInstruction('xml', 'version="1.0" encoding="UTF-8"');
      dom.insertAt(0, xml);
      let rootElement = this.getRootNode().toXML();
      dom.append(rootElement);
      return dom
    }

    getDocTypeParams () {
      
      throw new Error('This method is abstract')
    }

    getXMLSchema () {
      
      throw new Error('This method is abstract')
    }

    getRootNode () {
      
      throw new Error('This method is abstract')
    }

    
    getDocTypeAsString () {
      return new Error('This method is abstract')
    }

    createEditingInterface () {
      return new XMLEditingInterface$2(this)
    }

    find (cssSelector) {
      return this.getRootNode().find(cssSelector)
    }

    findAll (cssSelector) {
      return this.getRootNode().findAll(cssSelector)
    }

    createElement (tagName, data) {
      let node = this.create(Object.assign({
        id: uuid$2(tagName),
        type: tagName
      }, data));
      return node
    }

    getElementSchema (type) {
      return this.getXMLSchema().getElementSchema(type)
    }

    _validateChange (change) {
      let changed = {};
      let deleted = [];
      change.ops.forEach((op) => {
        switch (op.type) {
          case 'delete': {
            deleted.push(op.val.id);
            break
          }
          case 'create': {
            changed[op.val.id] = true;
            break
          }
          default: {
            if (op.path) {
              changed[op.path[0]] = true;
            }
          }
        }
      });
      
      deleted.forEach(id => delete changed[id]);

      const xmlSchema = this.getXMLSchema();
      let errors = [];
      Object.keys(changed).forEach((id) => {
        let node = this.get(id);
        let res = xmlSchema.validateElement(node);
        if (!res.ok) {
          errors = errors.concat(res.errors);
        }
      });
      return {
        ok: errors.length === 0,
        errors
      }
    }

    
    _apply (documentChange) {
      
      if (!documentChange.ops) return

      let created = {};
      let deleted = {};
      let updated = {};

      const doc = this;

      function _recordUpdate (id, op) {
        let record = updated[id];
        if (!record) {
          record = updated[id] = { ops: [] };
        }
        if (op) record.ops.push(op);
        return record
      }

      
      function _checkAnnotation (op) {
        
        switch (op.type) {
          case 'create':
          case 'delete': {
            const annoData = op.val;
            if (annoData.hasOwnProperty('start')) {
              updated[annoData.start.path] = true;
              let node = doc.get(annoData.start.path[0]);
              if (node) {
                _recordUpdate(node.id, op);
              }
            }
            if (annoData.hasOwnProperty('end')) {
              updated[annoData.end.path] = true;
              let node = doc.get(annoData.start.path[0]);
              if (node) {
                _recordUpdate(node.id, op);
              }
            }
            break
          }
          case 'update':
          case 'set': {
            let anno = doc.get(op.path[0]);
            if (anno) {
              if (anno.isPropertyAnnotation()) {
                updated[anno.start.path] = true;
                let node = doc.get(anno.start.path[0]);
                if (node) {
                  _recordUpdate(node.id, op);
                }
              } else if (anno.isContainerAnnotation()) ;
            }
            break
          }
          default:
            
        }
      }

      documentChange.ops.forEach((op) => {
        
        
        
        switch (op.type) {
          case 'create': {
            created[op.path[0]] = op.val;
            break
          }
          case 'delete': {
            let node = this.get(op.path[0]);
            deleted[node.id] = node;
            break
          }
          case 'update':
          case 'set': {
            updated[op.path] = true;
            let node = this.get(op.path[0]);
            _recordUpdate(node.id, op);
            break
          }
          default:
            
        }
        
        _checkAnnotation(op);

        this._applyOp(op);
      });
      
      Object.keys(deleted).forEach((id) => {
        delete updated[id];
      });
      documentChange.created = created;
      documentChange.updated = updated;
      documentChange.deleted = deleted;
      documentChange._extracted = true;
    }
  }

  XMLDocument$2.prototype._isXMLDocument = true;

  function findParentDOMElement$2 (nativeEl) {
    while (nativeEl) {
      let el = DefaultDOMElement$2.unwrap(nativeEl);
      if (el) return el
      nativeEl = nativeEl.parentNode;
    }
  }

  function findParent$2 (el, selector) {
    while (el) {
      if (el.is(selector)) return el
      el = el.getParent();
    }
  }

  function stop$2 (event) {
    event.stopPropagation();
  }

  function stopAndPrevent$2 (event) {
    event.stopPropagation();
    event.preventDefault();
  }

  function walk$2 (el, cb) {
    _walk$2(el, cb, 0);
  }

  function _walk$2 (el, cb, level) {
    cb(el, level);
    if (el.getChildCount() > 0) {
      let it = el.getChildNodeIterator();
      while (it.hasNext()) {
        _walk$2(it.next(), cb, level + 1);
      }
    }
  }

  function isRightButton$2 (event) {
    let isRightButton = false;
    if ('which' in event) {
      isRightButton = (event.which === 3);
    } else if ('button' in event) {
      isRightButton = (event.button === 2);
    }
    return isRightButton
  }

  function getBoundingRect$2 (el) {
    let _rect = el.getNativeElement().getBoundingClientRect();
    return {
      top: _rect.top,
      left: _rect.left,
      height: _rect.height,
      width: _rect.width
    }
  }

  function getBoundingRectForRects$2 (...rects) {
    let top, left, bottom, right;
    if (rects.length > 0) {
      let first = rects[0];
      top = first.top;
      left = first.left;
      bottom = top + first.height;
      right = left + first.width;
      for (let i = 1; i < rects.length; i++) {
        let r = rects[i];
        top = Math.min(top, r.top);
        left = Math.min(left, r.left);
        bottom = Math.max(bottom, r.top + r.height);
        right = Math.max(right, r.left + r.width);
      }
    }
    return {
      top,
      left,
      right,
      bottom,
      height: bottom - top,
      width: right - left
    }
  }

  function isXInside$2 (x, rect) {
    return x >= rect.left && x <= rect.left + rect.width
  }

  function isYInside$2 (y, rect) {
    return y >= rect.top && y <= rect.top + rect.height
  }

  var domHelpers$2 = /*#__PURE__*/Object.freeze({
    findParentDOMElement: findParentDOMElement$2,
    findParent: findParent$2,
    stop: stop$2,
    stopAndPrevent: stopAndPrevent$2,
    walk: walk$2,
    isRightButton: isRightButton$2,
    getBoundingRect: getBoundingRect$2,
    getBoundingRectForRects: getBoundingRectForRects$2,
    isXInside: isXInside$2,
    isYInside: isYInside$2
  });

  function encodeXMLEntities$2 (str) {
    return String(str)
      .replace(/&/g, '&amp;')
      .replace(/</g, '&lt;')
      .replace(/>/g, '&gt;')
      .replace(/"/g, '&quot;')
  }

  class DOMExporter$2 {
    constructor (config, context) {
      this.context = context || {};
      if (!config.converters) {
        throw new Error('config.converters is mandatory')
      }
      if (!config.converters._isRegistry) {
        this.converters = new Registry$2();
        config.converters.forEach(function (Converter) {
          let converter = isFunction$1$2(Converter) ? new Converter() : Converter;
          if (!converter.type) {
            console.error('Converter must provide the type of the associated node.', converter);
            return
          }
          this.converters.add(converter.type, converter);
        }.bind(this));
      } else {
        this.converters = config.converters;
      }

      this.state = {
        doc: null
      };
      this.config = config;
      
      
      this._elementFactory = config.elementFactory;
      if (!this._elementFactory) {
        throw new Error("'elementFactory' is mandatory")
      }
      this.$$ = this.createElement.bind(this);
    }

    exportDocument (doc) {
      
      
      
      
      return this.convertDocument(doc)
    }

    
    convertDocument(doc) { 
      throw new Error('This method is abstract')
    }

    convertContainer (container) {
      if (!container) {
        throw new Error('Illegal arguments: container is mandatory.')
      }
      const doc = container.getDocument();
      this.state.doc = doc;
      let elements = [];
      container.getContent().forEach((id) => {
        const node = doc.get(id);
        const nodeEl = this.convertNode(node);
        elements.push(nodeEl);
      });
      return elements
    }

    convertNode (node) {
      if (isString$2(node)) {
        
        node = this.state.doc.get(node);
      } else {
        this.state.doc = node.getDocument();
      }
      var converter = this.getNodeConverter(node);
      
      
      if (node.isPropertyAnnotation() && (!converter || !converter.export)) {
        return this._convertPropertyAnnotation(node)
      }
      if (!converter) {
        converter = this.getDefaultBlockConverter();
      }
      var el;
      if (converter.tagName) {
        el = this.$$(converter.tagName);
      } else {
        el = this.$$('div');
      }
      el.attr(this.config.idAttribute, node.id);
      if (converter.export) {
        el = converter.export(node, el, this) || el;
      } else {
        el = this.getDefaultBlockConverter().export(node, el, this) || el;
      }
      return el
    }

    convertProperty (doc, path, options) {
      this.initialize(doc, options);
      var wrapper = this.$$('div')
        .append(this.annotatedText(path));
      return wrapper.innerHTML
    }

    annotatedText (path) {
      var doc = this.state.doc;
      var text = doc.get(path);
      var annotations = doc.getIndex('annotations').get(path);
      return this._annotatedText(text, annotations)
    }

    getNodeConverter (node) {
      return this.converters.get(node.type)
    }

    getDefaultBlockConverter () {
      throw new Error('This method is abstract.')
    }

    getDefaultPropertyAnnotationConverter () {
      throw new Error('This method is abstract.')
    }

    getDocument () {
      return this.state.doc
    }

    createElement (str) {
      return this._elementFactory.createElement(str)
    }

    _annotatedText (text, annotations) {
      var self = this;

      var annotator = new Fragmenter$2();
      annotator.onText = function (context, text) {
        context.children.push(encodeXMLEntities$2(text));
      };
      annotator.onEnter = function (fragment) {
        var anno = fragment.node;
        return {
          annotation: anno,
          children: []
        }
      };
      annotator.onExit = function (fragment, context, parentContext) {
        var anno = context.annotation;
        var converter = self.getNodeConverter(anno);
        if (!converter) {
          converter = self.getDefaultPropertyAnnotationConverter();
        }
        var el;
        if (converter.tagName) {
          el = this.$$(converter.tagName);
        } else {
          el = this.$$('span');
        }
        el.attr(this.config.idAttribute, anno.id);
        el.append(context.children);
        if (converter.export) {
          el = converter.export(anno, el, self) || el;
        }
        parentContext.children.push(el);
      }.bind(this);
      var wrapper = { children: [] };
      annotator.start(wrapper, text, annotations);
      return wrapper.children
    }

    
    _convertPropertyAnnotation (anno) {
      
      var wrapper = this.$$('div').append(this.annotatedText(anno.path));
      var el = wrapper.find('[' + this.config.idAttribute + '="' + anno.id + '"]');
      return el
    }
  }

  class HTMLExporter$2 extends DOMExporter$2 {
    constructor (config, context) {
      super(_defaultConfig$3(config), context);
    }

    exportDocument (doc) {
      let htmlEl = DefaultDOMElement$2.parseHTML('<html><head></head><body></body></html>');
      return this.convertDocument(doc, htmlEl)
    }

    getDefaultBlockConverter () {
      return defaultBlockConverter$3 
    }

    getDefaultPropertyAnnotationConverter () {
      return defaultAnnotationConverter$3 
    }
  }

  function _defaultConfig$3 (config) {
    config = Object.assign({
      idAttribute: 'data-id'
    }, config);
    if (!config.elementFactory) {
      config.elementFactory = DefaultDOMElement$2.createDocument('html');
    }
    return config
  }

  const defaultAnnotationConverter$3 = {
    tagName: 'span',
    export: function (node, el) {
      el.tagName = 'span';
      el.attr('data-type', node.type);
      var properties = node.toJSON();
      forEach$2(properties, function (value, name) {
        if (name === 'id' || name === 'type') return
        if (isString$2(value) || isNumber$2(value) || isBoolean$2(value)) {
          el.attr('data-' + name, value);
        }
      });
    }
  };

  const defaultBlockConverter$3 = {
    export: function (node, el, converter) {
      el.attr('data-type', node.type);
      var properties = node.toJSON();
      forEach$2(properties, function (value, name) {
        if (name === 'id' || name === 'type') {
          return
        }
        var prop = converter.$$('div').attr('property', name);
        if (node.getPropertyType(name) === 'string' && value) {
          prop.append(converter.annotatedText([node.id, name]));
        } else {
          prop.text(value);
        }
        el.append(prop);
      });
    }
  };

  class ClipboardExporter$2 extends HTMLExporter$2 {
    
    exportDocument (doc) {
      this.state.doc = doc;
      let html;
      let elements = this.convertDocument(doc);
      
      if (elements.length === 1 && elements[0].attr('data-id') === TEXT_SNIPPET_ID$2) {
        html = elements[0].innerHTML;
      } else {
        html = elements.map(function (el) {
          return el.outerHTML
        }).join('');
      }
      let jsonConverter = new JSONConverter$2();
      let jsonStr = JSON.stringify(jsonConverter.exportDocument(doc));
      let substanceContent = `<script id="substance-clipboard" type="application/json">${jsonStr}</script>`;
      return '<html><head>' + substanceContent + '</head><body>' + html + '</body></html>'
    }

    
    convertDocument (doc) {
      let content = doc.get(SNIPPET_ID$2);
      if (!content) {
        throw new Error('Illegal clipboard document: could not find container "' + SNIPPET_ID$2 + '"')
      }
      return this.convertContainer(content)
    }
  }

  class HTMLImporter$2 extends DOMImporter$2 {
    constructor (config) {
      super(Object.assign({ idAttribute: 'data-id' }, config));

      
      this._el = DefaultDOMElement$2.parseHTML('<html></html>');
    }

    importDocument (html) {
      this.reset();
      var parsed = DefaultDOMElement$2.parseHTML(html);
      this.convertDocument(parsed);
      return this.state.doc
    }

    
    convertDocument(documentEl) { 
      throw new Error('This method is abstract')
    }
  }

  const INLINENODES$2 = ['a', 'b', 'big', 'i', 'small', 'tt', 'abbr', 'acronym', 'cite', 'code', 'dfn', 'em', 'kbd', 'strong', 'samp', 'time', 'var', 'bdo', 'br', 'img', 'map', 'object', 'q', 'script', 'span', 'sub', 'sup', 'button', 'input', 'label', 'select', 'textarea'].reduce((m, n) => { m[n] = true; return m }, {});


  class ClipboardImporter$2 extends HTMLImporter$2 {
    constructor (config) {
      super(_withCatchAllConverter$2(config));
      
      this.IGNORE_DEFAULT_WARNINGS = true;

      Object.assign(config, {
        trimWhitespaces: true,
        REMOVE_INNER_WS: true
      });

      this.editorOptions = config.editorOptions;
    }

    
    importDocument (html) {
      if (platform$2.isWindows) {
        
        
        let match = /<!--StartFragment-->(.*)<!--EndFragment-->/.exec(html);
        if (match) {
          html = match[1];
        }
      }

      
      
      
      if (html.search(/script id=.substance-clipboard./) >= 0) {
        let htmlDoc = DefaultDOMElement$2.parseHTML(html);
        let substanceData = htmlDoc.find('#substance-clipboard');
        if (substanceData) {
          let jsonStr = substanceData.textContent;
          try {
            return this.importFromJSON(jsonStr)
          } finally {
            
          }
        }
      }

      if (this.editorOptions && this.editorOptions['forcePlainTextPaste']) {
        return null
      }

      let htmlDoc = DefaultDOMElement$2.parseHTML(html);
      let generatorMeta = htmlDoc.find('meta[name="generator"]');
      let xmnlsw = htmlDoc.find('html').getAttribute('xmlns:w');
      if (generatorMeta) {
        let generator = generatorMeta.getAttribute('content');
        if (generator.indexOf('LibreOffice') > -1) this._isLibreOffice = true;
      } else if (xmnlsw) {
        if (xmnlsw.indexOf('office:word') > -1) {
          this._isMicrosoftWord = true;
          
          
          
          let match = /<!--StartFragment-->([\s\S]*)<!--EndFragment-->/.exec(html);
          if (match) {
            htmlDoc = DefaultDOMElement$2.parseHTML(match[1]);
          }
        }
      } else if (html.indexOf('docs-internal-guid') > -1) {
        this._isGoogleDoc = true;
      }

      let body = htmlDoc.find('body');
      body = this._sanitizeBody(body);
      if (!body) {
        console.warn('Invalid HTML.');
        return null
      }
      this._wrapIntoParagraph(body);
      this.reset();
      this.convertBody(body);
      const doc = this.state.doc;
      return doc
    }

    _sanitizeBody (body) {
      
      body.findAll('meta').forEach(el => el.remove());

      
      
      if (this._isLibreOffice || this._isMicrosoftWord) {
        let bodyHtml = body.getInnerHTML();
        body.setInnerHTML(bodyHtml.replace(/\r\n|\r|\n/g, ' '));
      } else if (this._isGoogleDoc) {
        body = this._fixupGoogleDocsBody(body);
      }

      return body
    }

    _fixupGoogleDocsBody (body) {
      if (!body) return
      
      
      
      
      let bold = body.find('b');
      if (bold && /^docs-internal/.exec(bold.id)) {
        body = bold;
      }

      body.findAll('span').forEach(span => {
        
        
        
        
        
        
        
        
        let nodeTypes = [];
        if (span.getStyle('font-weight') === '700') nodeTypes.push('b');
        if (span.getStyle('font-style') === 'italic') nodeTypes.push('i');
        if (span.getStyle('vertical-align') === 'super') nodeTypes.push('sup');
        if (span.getStyle('vertical-align') === 'sub') nodeTypes.push('sub');
        
        span.removeAttribute('style');

        function createInlineNodes (parentEl, isRoot) {
          if (nodeTypes.length > 0) {
            let el = parentEl.createElement(nodeTypes[0]);
            if (nodeTypes.length === 1) el.append(span.textContent);
            if (isRoot) {
              parentEl.replaceChild(span, el);
            } else {
              parentEl.appendChild(el);
            }
            nodeTypes.shift();
            createInlineNodes(el);
          }
        }
        createInlineNodes(span.getParent(), true);
      });

      let tags = ['b', 'i', 'sup', 'sub'];

      tags.forEach(tag => {
        body.findAll(tag).forEach(el => {
          
          
          let previousSiblingEl = el.getPreviousSibling();
          if (previousSiblingEl && el.tagName === previousSiblingEl.tagName) {
            let parentEl = el.getParent();
            let newEl = parentEl.createElement(tag);
            newEl.setInnerHTML(previousSiblingEl.getInnerHTML() + el.getInnerHTML());
            parentEl.replaceChild(el, newEl);
            parentEl.removeChild(previousSiblingEl);
          }

          
          
          
          
          if (previousSiblingEl && previousSiblingEl.tagName && el.getChildCount() > 0 && el.getChildAt(0).tagName === previousSiblingEl.tagName) {
            let parentEl = el.getParent();
            let childEl = el.getChildAt(0);
            let newEl = parentEl.createElement(previousSiblingEl.tagName);
            let newChildEl = newEl.createElement(tag);
            newChildEl.setTextContent(childEl.textContent);
            newEl.appendChild(newChildEl);
            parentEl.replaceChild(el, newEl);
          }
        });
      });

      return body
    }

    _wrapIntoParagraph (body) {
      let childNodes = body.getChildNodes();
      let shouldWrap = false;
      for (let i = 0; i < childNodes.length; i++) {
        const c = childNodes[i];
        if (c.isTextNode()) {
          if (!(/^\s+$/.exec(c.textContent))) {
            shouldWrap = true;
            break
          }
        } else if (INLINENODES$2[c.tagName]) {
          shouldWrap = true;
          break
        }
      }
      if (shouldWrap) {
        let p = body.createElement('p');
        p.append(childNodes);
        body.append(p);
      }
    }

    importFromJSON (jsonStr) {
      this.reset();
      let doc = this.getDocument();
      let jsonData = JSON.parse(jsonStr);
      let converter = new JSONConverter$2();
      converter.importDocument(doc, jsonData);
      return doc
    }

    
    convertBody (body) {
      this.convertContainer(body.childNodes, SNIPPET_ID$2);
    }

    
    _createDocument () {
      let emptyDoc = super._createDocument();
      return emptyDoc.createSnippet()
    }
  }

  function _withCatchAllConverter$2 (config) {
    config = Object.assign({}, config);
    let defaultTextType = config.schema.getDefaultTextType();
    if (defaultTextType) {
      config.converters = config.converters.concat([{
        type: defaultTextType,
        matchElement: function (el) { return el.is('div') },
        import: function (el, node, converter) {
          node.content = converter.annotatedText(el, [node.id, 'content']);
        }
      }]);
    }
    return config
  }

  class ContainerAdapter$2 extends Container$2 {
    constructor (doc, path) {
      super(doc, { id: String(path) });
      this.document = doc;
      this.path = path;

      
      
      doc.data.nodes[this.id] = this;
    }

    getContentPath () {
      return this.path
    }

    get nodes () {
      return this.document.get(this.path)
    }
  }

  ContainerAdapter$2.prototype._isDocumentNode = false;
  ContainerAdapter$2.prototype._isContainer = false;

  class FileNode$2 extends DocumentNode$2 {
    getUrl () {
      if (this.proxy) {
        return this.proxy.getUrl()
      } else {
        
        console.warn('No file proxy attached to ', this.id);
        return ''
      }
    }

    setProxy (proxy) {
      this.proxy = proxy;
    }
  }

  FileNode$2.type = 'file';

  FileNode$2.schema = {
    url: { type: 'string', optional: true },
    fileType: { type: 'string', optional: true },
    mimeType: { type: 'string', optional: true },
    sourceFile: { type: 'object', optional: true }
  };

  FileNode$2.prototype._isFileNode = true;
  FileNode$2._isFileNode = true;

  class InlineNode$2 extends PropertyAnnotation$2 {}

  InlineNode$2.prototype._isInlineNode = true;

  InlineNode$2.isInline = true;

  class Range$2 {
    constructor (start, end, reverse, containerId, surfaceId) {
      
      if (arguments[0] === 'SKIP') return
      if (arguments.length === 1 && isPlainObject$3(arguments[0])) {
        let data = arguments[0];
        this.start = data.start;
        this.end = data.end;
        this.reverse = Boolean(data.reverse);
        this.containerId = data.containerId;
        this.surfaceId = data.surfaceId;
      } else {
        this.start = start;
        this.end = end;
        this.reverse = Boolean(reverse);
        this.containerId = containerId;
        this.surfaceId = surfaceId;
      }
    }

    isCollapsed () {
      return this.start.equals(this.end)
    }

    equals (other) {
      if (this === other) return true
      else {
        return (
          this.containerId === other.containerId &&
          this.start.equals(other.start) &&
          this.end.equals(other.end)
        )
      }
    }

    isReverse () {
      return this.reverse
    }

    toString () {
      let str = [this.start.toString(), '->', this.end.toString()];
      if (this.isReverse()) {
        str.push('[reverse]');
      }
      if (this.containerId) {
        str.push('[container=' + this.containerId + ']');
      }
      if (this.surfaceId) {
        str.push('[surface=' + this.surfaceId + ']');
      }
      return str.join('')
    }
  }

  Range$2.prototype._isRange = true;

  class AbstractClipboard$2 {
    constructor (configurator) {
      let schema = configurator.getSchema();
      let converterRegistry = configurator.converterRegistry;
      let htmlConverters = [];
      if (converterRegistry && converterRegistry.contains('html')) {
        htmlConverters = converterRegistry.get('html').values() || [];
      }
      
      
      this._config = {
        schema: schema,
        DocumentClass: schema.getDocumentClass(),
        converters: htmlConverters,
        editorOptions: configurator.getEditorOptions()
      };
    }

    
    onCopy (event) {
      
      let clipboardData = this._copy();
      substanceGlobals$2._clipboardData = event.clipboardData;

      if (event.clipboardData && clipboardData.doc) {
        event.preventDefault();
        
        event.clipboardData.setData('text/plain', clipboardData.text);
        event.clipboardData.setData('text/html', clipboardData.html);
      }
    }

    
    onCut (event) {
      
      event.preventDefault();
      
      this.onCopy(event);
      this._cut();
    }

    
    
    onPaste (event) {
      let clipboardData = event.clipboardData;

      let types = {};
      for (let i = 0; i < clipboardData.types.length; i++) {
        types[clipboardData.types[i]] = true;
      }
      

      event.preventDefault();
      event.stopPropagation();

      let plainText;
      let html;
      if (types['text/plain']) {
        plainText = clipboardData.getData('text/plain');
      }
      if (types['text/html']) {
        html = clipboardData.getData('text/html');
      }

      substanceGlobals$2.clipboardData = {
        text: plainText,
        html: html
      };

      

      
      
      if (html) {
        if (!this._pasteHtml(html, plainText)) {
          this._pastePlainText(plainText);
        }
      } else {
        this._pastePlainText(plainText);
      }
    }

    _getImporter () {
      return new ClipboardImporter$2(this._config)
    }

    _getExporter () {
      return new ClipboardExporter$2(this._config)
    }

    
    _copy () {
      throw new Error('This method is abstract.')
    }

    
    _cut () {
      throw new Error('This method is abstract.')
    }

    
    _pastePlainText(plainText) { 
      throw new Error('This method is abstract.')
    }

    
    _pasteHtml(html, text) { 
      throw new Error('This method is abstract.')
    }
  }

  function levenshtein$2 (a, b) {
    let m = [];
    for (let i = 0; i <= b.length; i++) {
      m[i] = [i];
      if (i === 0) continue
      let ib = i - 1;
      for (let j = 0; j <= a.length; j++) {
        m[0][j] = j;
        if (j === 0) continue
        let jb = j - 1;
        m[i][j] = b.charAt(ib) === a.charAt(jb) ? m[ib][jb] : Math.min(
          m[ib][jb] + 1,
          m[i][jb] + 1,
          m[ib][j] + 1
        );
      }
    }
    return m
  }

  function diff$2 (a, b, offset) {
    if (!isString$2(a) || !isString$2(b)) {
      throw new Error('Illegal arguments.')
    }
    offset = offset || 0;
    let changes = [];
    if (a || b) {
      if (!a && b) {
        changes.push({ type: 'insert', start: offset, text: b });
      } else if (a && !b) {
        changes.push({ type: 'delete', start: offset, end: offset + a.length });
      } else {
        let m = levenshtein$2(a, b);
        changes = _diff$2(a, b, m, offset);
      }
    }
    return changes
  }

  function _diff$2 (a, b, m, offset) {
    let i = b.length;
    let j = a.length;
    let changes = [];
    let current;
    while (i > 0 && j > 0) {
      _next();
    }
    _commit();
    return changes

    function _next () {
      let d = m[i][j];
      let ib = i - 1;
      let jb = j - 1;
      
      if (m[ib][jb] < d) {
        if (current && current.type === 'replace') {
          current.start--;
          current.text.unshift(b.charAt(ib));
        } else {
          _commit();
          current = { type: 'replace', start: jb, end: j, text: [b.charAt(ib)] };
        }
        i--;
        j--;
      
      } else if (m[ib][j] < d) {
        if (current && current.type === 'insert') {
          current.start--;
          current.text.unshift(b.charAt(ib));
        } else {
          _commit();
          current = { type: 'insert', start: jb, text: [b.charAt(ib)] };
        }
        i--;
      
      } else if (m[i][jb] < d) {
        if (current && current.type === 'delete') {
          current.start--;
        } else {
          _commit();
          current = { type: 'delete', start: jb, end: j };
        }
        j--;
      
      } else {
        _commit();
        i--;
        j--;
      }
    }

    function _commit () {
      if (current) {
        switch (current.type) {
          case 'insert':
            current.start += offset;
            current.text = current.text.join('');
            break
          case 'delete':
            current.start += offset;
            current.end += offset;
            break
          case 'replace':
            current.start += offset;
            current.end += offset;
            current.text = current.text.join('');
            break
          default:
            throw new Error('Invalid state')
        }
        changes.push(current);
        current = null;
      }
    }
  }

  class AnnotationComponent$2 extends Component$2 {
    
    didMount () {
      let node = this.props.node;
      node.on('highlighted', this.onHighlightedChanged, this);
    }

    
    dispose () {
      let node = this.props.node;
      node.off(this);
    }

    render ($$) {
      let el = $$(this.getTagName())
        .attr('data-id', this.props.node.id)
        .addClass(this.getClassNames());
      if (this.props.node.highlighted) {
        el.addClass('sm-highlighted');
      }
      el.append(this.props.children);
      return el
    }

    getClassNames () {
      return 'sc-' + this.props.node.type
    }

    onHighlightedChanged () {
      if (this.props.node.highlighted) {
        this.el.addClass('sm-highlighted');
      } else {
        this.el.removeClass('sm-highlighted');
      }
    }

    getTagName () {
      return 'span'
    }
  }

  class AbstractIsolatedNodeComponent$2 extends Component$2 {
    constructor (...args) {
      super(...args);

      this.name = this.props.node.id;
      this._state = { selectionFragment: null };

      this.handleAction('escape', this.escape);
      this.ContentClass = this._getContentClass(this.props.node);

      
      let useBlocker = platform$2.isFF || !this.ContentClass.noBlocker;
      this.blockingMode = useBlocker ? 'closed' : 'open';
    }

    getInitialState () {
      let selState = this.context.editorSession.getSelectionState();
      return this._deriveStateFromSelectionState(selState)
    }

    getChildContext () {
      return {
        parentSurfaceId: this.getId(),
        isolatedNodeComponent: this,
        
        
        surface: undefined
      }
    }

    didMount () {
      super.didMount();

      let editorSession = this.context.editorSession;
      editorSession.onRender('selection', this._onSelectionChanged, this);
    }

    dispose () {
      super.dispose.call(this);

      let editorSession = this.context.editorSession;
      editorSession.off(this);
    }

    renderContent ($$, node, options = {}) {
      let ComponentClass = this.ContentClass;
      if (!ComponentClass) {
        console.error('Could not resolve a component for type: ' + node.type);
        return $$(this.__elementTag)
      } else {
        let props = Object.assign({
          disabled: this.props.disabled,
          node: node,
          isolatedNodeState: this.state.mode,
          focused: (this.state.mode === 'focused')
        }, options);
        return $$(ComponentClass, props)
      }
    }

    getId () {
      
      
      if (!this._id) {
        this._id = this.context.parentSurfaceId + '/' + this.name;
      }
      return this._id
    }

    get id () { return this.getId() }

    getMode () {
      return this.state.mode
    }

    isOpen () {
      return this.blockingMode === 'open'
    }

    isClosed () {
      return this.blockingMode === 'closed'
    }

    isNotSelected () {
      return !this.state.mode
    }

    isSelected () {
      return this.state.mode === 'selected'
    }

    isCoSelected () {
      return this.state.mode === 'co-selected'
    }

    isFocused () {
      return this.state.mode === 'focused'
    }

    isCoFocused () {
      return this.state.mode === 'co-focused'
    }

    getParentSurface () {
      return this.context.surface
    }

    escape () {
      
      this.selectNode();
    }

    _onSelectionChanged () {
      let editorSession = this.context.editorSession;
      let newState = this._deriveStateFromSelectionState(editorSession.getSelectionState());
      if (!newState && this.state.mode) {
        this.extendState({ mode: null });
      } else if (newState && newState.mode !== this.state.mode) {
        this.extendState(newState);
      }
    }

    onKeydown (event) {
      
      
      
      
      if (event.keyCode === keys$1$2.ESCAPE && this.state.mode === 'focused') {
        event.stopPropagation();
        event.preventDefault();
        this.escape();
      }
    }

    _getContentClass (node) {
      let ComponentClass;
      
      ComponentClass = this.getComponent(node.type, true);
      
      if (!ComponentClass) {
        ComponentClass = Component$2;
      }
      return ComponentClass
    }

    _getSurface (selState) {
      let surface = selState.get('surface');
      if (surface === undefined) {
        let sel = selState.getSelection();
        if (sel && sel.surfaceId) {
          let surfaceManager = this.context.surfaceManager;
          surface = surfaceManager.getSurface(sel.surfaceId);
        } else {
          surface = null;
        }
        selState.set('surface', surface);
      }
      return surface
    }

    
    
    _getIsolatedNodes (selState) {
      let isolatedNodes = selState.get('isolatedNodes');
      if (!isolatedNodes) {
        let sel = selState.getSelection();
        isolatedNodes = [];
        if (sel && sel.surfaceId) {
          let surfaceManager = this.context.surfaceManager;
          let surface = surfaceManager.getSurface(sel.surfaceId);
          if (surface) {
            isolatedNodes = surface.getComponentPath().filter(comp => comp._isAbstractIsolatedNodeComponent);
          }
        }
        selState.set('isolatedNodes', isolatedNodes);
      }
      return isolatedNodes
    }

    _shouldConsumeEvent (event) {
      let comp = Component$2.unwrap(event.target);
      let isolatedNodeComponent = this._getIsolatedNode(comp);
      return (isolatedNodeComponent === this)
    }

    _getIsolatedNode (comp) {
      if (comp._isAbstractIsolatedNodeComponent) {
        return this
      } else if (comp.context.isolatedNodeComponent) {
        return comp.context.isolatedNodeComponent
      } else if (comp.context.surface) {
        return comp.context.surface.context.isolatedNodeComponent
      }
    }
  }

  AbstractIsolatedNodeComponent$2.prototype._isAbstractIsolatedNodeComponent = true;

  class InlineNodeComponent$2 extends AbstractIsolatedNodeComponent$2 {
    render ($$) {
      const node = this.props.node;
      const ContentClass = this.ContentClass;
      const state = this.state;

      let el = $$('span');
      el.addClass(this.getClassNames())
        .addClass('sc-inline-node')
        .addClass('sm-' + this.props.node.type)
        .attr('data-id', node.id)
        .attr('data-inline', '1');

      let disabled = this.isDisabled();

      if (state.mode) {
        el.addClass('sm-' + state.mode);
      } else {
        el.addClass('sm-not-selected');
      }

      if (!ContentClass.noStyle) {
        el.addClass('sm-default-style');
      }

      
      
      
      el.on('keydown', this.onKeydown);

      el.append(
        this.renderContent($$, node)
          .ref('content')
          .addClass('se-content')
      );

      if (disabled) {
        el.addClass('sm-disabled')
          .attr('contenteditable', false)
          .on('click', this.onClick);
      }

      
      
      
      
      
      
      
      el.attr('draggable', true);

      return el
    }

    isDisabled () {
      return !this.state.mode || ['co-selected', 'cursor'].indexOf(this.state.mode) > -1
    }

    getClassNames () {
      return ''
    }

    onClick (event) {
      if (!this._shouldConsumeEvent(event)) {
        return
      }
      this.selectNode();
    }

    selectNode () {
      
      let editorSession = this.context.editorSession;
      let surface = this.context.surface;
      let node = this.props.node;
      editorSession.setSelection({
        type: 'property',
        path: node.start.path,
        startOffset: node.start.offset,
        endOffset: node.end.offset,
        containerId: surface.getContainerId(),
        surfaceId: surface.id
      });
    }

    _getContentClass (node) {
      let ComponentClass;
      
      ComponentClass = this.getComponent(node.type, true);
      
      
      if (!ComponentClass) {
        ComponentClass = this.getComponent('unsupported-inline-node', true);
      }
      if (!ComponentClass) {
        console.error(`No component registered for inline node '${node.type}'.`);
        ComponentClass = StubInlineNodeComponent$2;
      }
      return ComponentClass
    }

    
    
    _deriveStateFromSelectionState (selState) {
      let surface = this._getSurface(selState);
      if (!surface) return null
      
      if (surface === this.context.surface) {
        let sel = selState.getSelection();
        let node = this.props.node;
        if (sel.isPropertySelection() && !sel.isCollapsed() && isEqual$2(sel.start.path, node.start.path)) {
          let nodeSel = node.getSelection();
          if (nodeSel.equals(sel)) {
            return { mode: 'selected' }
          }
          if (sel.contains(nodeSel)) {
            return { mode: 'co-selected' }
          }
        }
      }
      let isolatedNodeComponent = surface.context.isolatedNodeComponent;
      if (!isolatedNodeComponent) return null
      if (isolatedNodeComponent === this) {
        return { mode: 'focused' }
      }
      let isolatedNodes = this._getIsolatedNodes(selState);
      if (isolatedNodes.indexOf(this) > -1) {
        return { mode: 'co-focused' }
      }
      return null
    }
  }

  InlineNodeComponent$2.prototype._isInlineNodeComponent = true;

  class StubInlineNodeComponent$2 extends Component$2 {
    render ($$) {
      const node = this.props.node;
      return $$('span').text('???').attr('data-id', node.id).attr('data-type', node.type)
    }
  }

  class AnnotatedTextComponent$2 extends Component$2 {
    render ($$) {
      let el = this._renderContent($$)
        .addClass('sc-annotated-text')
        .css({ whiteSpace: 'pre-wrap' });
      return el
    }

    getPath () {
      return this.props.path
    }

    getText () {
      return this.getDocument().get(this.props.path) || ''
    }

    isEmpty () {
      return !(this.getText())
    }

    getAnnotations () {
      return this.getDocument().getIndex('annotations').get(this.props.path)
    }

    getDocument () {
      return this.props.doc || this.context.doc
    }

    _getTagName () {
      return this.props.tagName
    }

    _onDocumentChange (update) {
      if (update.change && update.change.updated[this.getPath()]) {
        this.rerender();
      }
    }

    _renderContent ($$) {
      let text = this.getText();
      let annotations = this.getAnnotations();
      let el = $$(this._getTagName() || 'span');
      if (annotations && annotations.length > 0) {
        let fragmenter = new Fragmenter$2({
          onText: this._renderTextNode.bind(this),
          onEnter: this._renderFragment.bind(this, $$),
          onExit: this._finishFragment.bind(this)
        });
        fragmenter.start(el, text, annotations);
      } else {
        el.append(text);
      }
      return el
    }

    _renderTextNode (context, text) {
      if (text && text.length > 0) {
        context.append(text);
      }
    }

    _renderFragment ($$, fragment) {
      let doc = this.getDocument();
      let componentRegistry = this.getComponentRegistry();
      let node = fragment.node;
      
      if (node.type === 'container-annotation-fragment') ; else if (node.type === 'container-annotation-anchor') ; else {
        let ComponentClass = componentRegistry.get(node.type) || AnnotationComponent$2;
        if (node.constructor.isInline &&
            
            !ComponentClass.prototype._isInlineNodeComponent &&
            
            !ComponentClass.isCustom) {
          ComponentClass = InlineNodeComponent$2;
        }
        let el = $$(ComponentClass, { doc: doc, node: node });
        return el
      }
    }

    _finishFragment (fragment, context, parentContext) {
      parentContext.append(context);
    }
  }

  class CursorComponent$2 extends Component$2 {
    render ($$) {
      
      let el = $$('span').addClass('se-cursor');
      
      
      
      el.append('\uFEFF');
      el.append($$('div').addClass('se-cursor-inner'));

      if (this.props.collaborator) {
        let collaboratorIndex = this.props.collaborator.colorIndex;
        el.addClass('sm-collaborator-' + collaboratorIndex);
      } else {
        el.addClass('sm-local-user');
      }

      return el
    }
  }

  class SelectionFragmentComponent$2 extends Component$2 {
    render ($$) {
      
      let el = $$('span').addClass('se-selection-fragment');
      if (this.props.collaborator) {
        let collaboratorIndex = this.props.collaborator.colorIndex;
        el.addClass('sm-collaborator-' + collaboratorIndex);
      } else {
        el.addClass('sm-local-user');
      }
      el.append(this.props.children);
      return el
    }
  }

  class TextPropertyComponent$2 extends AnnotatedTextComponent$2 {
    getInitialState () {
      const markersManager = this.context.markersManager;
      let path = this.getPath();
      let markers;
      if (markersManager) {
        
        markersManager.register(this);
        markers = markersManager.getMarkers(path, {
          surfaceId: this.getSurfaceId(),
          containerId: this.getContainerId()
        });
      } else {
        const doc = this.getDocument();
        markers = doc.getAnnotations(path);
      }
      return {
        markers: markers
      }
    }

    didMount () {
      
      
      
      
      
    }

    dispose () {
      if (this.context.markersManager) {
        this.context.markersManager.deregister(this);
      }
      if (this.domObserver) {
        this.domObserver.disconnect();
      }
    }

    render ($$) {
      let path = this.getPath();

      let el = this._renderContent($$)
        .addClass('sc-text-property')
        .attr({
          'data-path': path.join('.')
        })
        .css({
          'white-space': 'pre-wrap'
        });

      if (this.isEmpty()) {
        el.addClass('sm-empty');
        if (this.props.placeholder) {
          el.append(
            $$('span').addClass('se-placeholder').append(
              this.props.placeholder
            )
          );
        }
      }

      if (!this.props.withoutBreak) {
        el.append($$('br'));
      }

      return el
    }

    getAnnotations () {
      if (this.props.markers) {
        return this.state.markers.concat(this.props.markers)
      } else {
        return this.state.markers
      }
    }

    _renderFragment ($$, fragment) {
      let node = fragment.node;
      let id = node.id;
      let el;
      if (node.type === 'cursor') {
        el = $$(CursorComponent$2, { collaborator: node.collaborator });
      } else if (node.type === 'selection-fragment') {
        el = $$(SelectionFragmentComponent$2, { collaborator: node.collaborator });
      } else {
        el = super._renderFragment.apply(this, arguments);
        if (id) {
          el.ref(id + '@' + fragment.counter);
        }
      }
      el.attr('data-offset', fragment.pos);
      return el
    }

    _onDomMutations (mutations) {
      
      
      if (mutations.length === 2 && mutations[0].target === mutations[1].target) {
        let textEl = DefaultDOMElement$2.unwrap(mutations[0].target);
        if (textEl) {
          this._applyTextMutation(textEl, mutations[0].oldValue);
          return
        }
      }
      
      this.rerender();
    }

    _applyTextMutation (textEl, oldText) {
      
      let offset = _getCharPos$2(textEl, 0);
      let newText = textEl.textContent;
      let changes = diff$2(oldText, newText, offset);

      let editorSession = this.context.editorSession;
      let path = this.getPath();
      editorSession.transaction(function (tx) {
        changes.forEach(function (change) {
          
          if (change.type === 'replace') {
            tx.update(path, { type: 'delete', start: change.start, end: change.end });
            tx.update(path, { type: 'insert', start: change.start, text: change.text });
          } else {
            tx.update(path, change);
          }
        });
      });
    }

    getSurface () {
      return this.props.surface || this.context.surface
    }

    getSurfaceId () {
      let surface = this.getSurface();
      return surface ? surface.id : null
    }

    getContainerId () {
      let surface = this.getSurface();
      return surface ? surface.getContainerId() : null
    }

    isEditable () {
      const surface = this.getSurface();
      return surface ? surface.isEditable() : false
    }

    isReadonly () {
      const surface = this.getSurface();
      return surface ? surface.isReadonly() : true
    }

    getDOMCoordinate (charPos) {
      return this._getDOMCoordinate(this.el, charPos)
    }

    _finishFragment (fragment, context, parentContext) {
      context.attr('data-length', fragment.length);
      parentContext.append(context);
    }

    _getDOMCoordinate (el, charPos) {
      let l;
      let idx = 0;
      if (charPos === 0) {
        return {
          container: el.getNativeElement(),
          offset: 0
        }
      }
      for (let child = el.getFirstChild(); child; child = child.getNextSibling(), idx++) {
        if (child.isTextNode()) {
          l = child.textContent.length;
          if (l >= charPos) {
            return {
              container: child.getNativeElement(),
              offset: charPos
            }
          } else {
            charPos -= l;
          }
        } else if (child.isElementNode()) {
          let length = child.getAttribute('data-length');
          if (length) {
            l = parseInt(length, 10);
            if (l >= charPos) {
              
              if (child.attr('data-inline')) {
                let nextSibling = child.getNextSibling();
                if (nextSibling && nextSibling.isTextNode()) {
                  return {
                    container: nextSibling.getNativeElement(),
                    offset: 0
                  }
                } else {
                  return {
                    container: el.getNativeElement(),
                    offset: el.getChildIndex(child) + 1
                  }
                }
              }
              return this._getDOMCoordinate(child, charPos, idx)
            } else {
              charPos -= l;
            }
          } else {
            console.error('FIXME: Can not map to DOM coordinates.');
            return null
          }
        }
      }
    }
  }

  TextPropertyComponent$2.prototype._isTextPropertyComponent = true;




  TextPropertyComponent$2.getCoordinate = function (root, el, offset) {
    let context = _getPropertyContext$2(root, el, offset);
    if (!context) {
      return null
    }
    
    
    let charPos;
    if (el.parentNode && el.parentNode.is('.se-placeholder')) {
      charPos = 0;
    } else {
      
      
      
      charPos = _getCharPos$2(context.node, context.offset);
    }
    if (isNumber$2(charPos)) {
      let coor = new Coordinate$2(context.path, charPos);
      coor._comp = context.comp;
      return coor
    } else {
      return null
    }
  };

  function _getPropertyContext$2 (root, node, offset) {
    let result = {
      comp: null,
      el: null,
      path: null,
      node: node,
      offset: offset
    };
    while (node && node !== root) {
      if (node.isElementNode()) {
        let comp = Component$2.unwrap(node);
        if (comp && comp._isTextPropertyComponent) {
          result.comp = comp;
          result.el = node;
          result.path = comp.getPath();
          return result
        }
        
        
        
        if (node.getAttribute('data-inline')) {
          result.node = node;
          if (offset > 0) {
            result.offset = 1;
          }
        }
      }
      node = node.getParent();
    }
    return null
  }

  function _getCharPos$2 (node, offset) {
    let charPos = offset;
    let parent, childIdx;

    

    parent = node.getParent();
    if (node.isTextNode()) {
      
      if (node === parent.firstChild) {
        
        let parentPath = parent.getAttribute('data-path');
        let parentOffset = parent.getAttribute('data-offset');
        if (parentPath) {
          charPos = offset;
        
        } else if (parentOffset) {
          charPos = parseInt(parentOffset, 10) + offset;
        
        } else {
          charPos = _getCharPos$2(parent, 0) + offset;
        }
      } else {
        
        childIdx = parent.getChildIndex(node);
        charPos = _getCharPos$2(parent, childIdx) + offset;
      }
    } else if (node.isElementNode()) {
      let pathStr = node.getAttribute('data-path');
      let offsetStr = node.getAttribute('data-offset');
      
      
      if (pathStr) {
        charPos = _countCharacters$2(node, offset);
      
      
      } else if (offsetStr) {
        childIdx = parent.getChildIndex(node);
        charPos = parseInt(offsetStr, 10) + _countCharacters$2(node, offset);
      
      
      } else {
        childIdx = parent.getChildIndex(node);
        charPos = _getCharPos$2(parent, childIdx) + _countCharacters$2(node, offset);
      }
    } else {
      
      return null
    }
    return charPos
  }

  function _countCharacters$2 (el, maxIdx) {
    let charPos = 0;
    
    if (el.getAttribute('data-inline')) {
      return maxIdx === 0 ? 0 : 1
    }
    let l = el.getChildCount();
    if (arguments.length === 1) {
      maxIdx = l;
    }
    maxIdx = Math.min(l, maxIdx);
    for (let i = 0, child = el.getFirstChild(); i < maxIdx; child = child.getNextSibling(), i++) {
      if (child.isTextNode()) {
        charPos += child.getTextContent().length;
      } else if (child.isElementNode()) {
        let length = child.getAttribute('data-length');
        if (child.getAttribute('data-inline')) {
          charPos += 1;
        } else if (length) {
          charPos += parseInt(length, 10);
        } else {
          charPos += _countCharacters$2(child);
        }
      }
    }
    return charPos
  }

  const BRACKET$2 = 'X';
  const SLASH$2 = '/'.charCodeAt(0);


  class IsolatedNodeComponent$2 extends AbstractIsolatedNodeComponent$2 {
    render ($$) {
      let node = this.props.node;
      let ContentClass = this.ContentClass;
      let disabled = this.props.disabled;

      
      let el = $$('div');
      el.addClass(this.getClassNames())
        .addClass('sc-isolated-node')
        .addClass('sm-' + this.props.node.type)
        .attr('data-id', node.id);
      if (disabled) {
        el.addClass('sm-disabled');
      }
      if (this.state.mode) {
        el.addClass('sm-' + this.state.mode);
      }
      if (!ContentClass.noStyle) {
        el.addClass('sm-default-style');
      }
      
      el.on('keydown', this.onKeydown);

      
      let shouldRenderBlocker = (
        this.blockingMode === 'closed' &&
        !this.state.unblocked
      );

      
      
      el.append(
        $$('div').addClass('se-bracket sm-left').ref('left')
          .append(BRACKET$2)
      );

      let content = this.renderContent($$, node, {
        disabled: this.props.disabled || shouldRenderBlocker
      }).ref('content');
      content.attr('contenteditable', false);

      el.append(content);
      el.append($$(Blocker$2).ref('blocker'));
      el.append(
        $$('div').addClass('se-bracket sm-right').ref('right')
          .append(BRACKET$2)
      );

      if (!shouldRenderBlocker) {
        el.addClass('sm-no-blocker');
        el.on('click', this.onClick)
          .on('dblclick', this.onDblClick);
      }
      el.on('mousedown', this._reserveMousedown, this);

      return el
    }

    getClassNames () {
      return ''
    }

    getContent () {
      return this.refs.content
    }

    selectNode () {
      
      let editorSession = this.context.editorSession;
      let surface = this.context.surface;
      let nodeId = this.props.node.id;
      editorSession.setSelection({
        type: 'node',
        nodeId: nodeId,
        containerId: surface.getContainerId(),
        surfaceId: surface.id
      });
    }

    
    
    onClick (event) {
      
      event.stopPropagation();
    }

    onDblClick (event) {
      
      event.stopPropagation();
    }

    grabFocus (event) {
      let content = this.refs.content;
      if (content.grabFocus) {
        content.grabFocus(event);
        return true
      }
    }

    
    
    _reserveMousedown (event) {
      if (event.__reserved__) ; else {
        
        event.__reserved__ = this;
      }
    }

    _deriveStateFromSelectionState (selState) {
      let surface = this._getSurface(selState);
      let newState = { mode: null, unblocked: null };
      if (!surface) return newState
      
      if (surface === this.context.surface) {
        let sel = selState.getSelection();
        let nodeId = this.props.node.id;
        if (sel.isNodeSelection() && sel.getNodeId() === nodeId) {
          if (sel.isFull()) {
            newState.mode = 'selected';
            newState.unblocked = true;
          } else if (sel.isBefore()) {
            newState.mode = 'cursor';
            newState.position = 'before';
          } else if (sel.isAfter()) {
            newState.mode = 'cursor';
            newState.position = 'after';
          }
        }
        if (sel.isContainerSelection() && sel.containsNode(nodeId)) {
          newState.mode = 'co-selected';
        }
      } else {
        let sel = selState.getSelection();
        let surfaceId = sel.surfaceId;
        let id = this.getId();
        if (id.length < surfaceId.length && surfaceId.startsWith(id) && surfaceId.charCodeAt(id.length) === SLASH$2) {
          let tail = surfaceId.slice(id.length + 1);
          if (tail.indexOf('/') > 0) {
            newState.mode = 'co-focused';
            newState.unblocked = true;
          } else {
            newState.mode = 'focused';
            newState.unblocked = true;
          }
        }
      }
      return newState
    }
  }

  IsolatedNodeComponent$2.prototype._isIsolatedNodeComponent = true;

  IsolatedNodeComponent$2.prototype._isDisabled = IsolatedNodeComponent$2.prototype.isDisabled;

  IsolatedNodeComponent$2.getDOMCoordinate = function (comp, coor) {
    let { start, end } = IsolatedNodeComponent$2.getDOMCoordinates(comp);
    if (coor.offset === 0) return start
    else return end
  };

  IsolatedNodeComponent$2.getDOMCoordinates = function (comp) {
    const left = comp.refs.left;
    const right = comp.refs.right;
    return {
      start: {
        container: left.getNativeElement(),
        offset: 0
      },
      end: {
        container: right.getNativeElement(),
        offset: right.getChildCount()
      }
    }
  };

  IsolatedNodeComponent$2.getCoordinate = function (nodeEl, options) {
    let comp = Component$2.unwrap(nodeEl, 'strict').context.isolatedNodeComponent;
    let offset = null;
    if (options.direction === 'left' || nodeEl === comp.refs.left.el) {
      offset = 0;
    } else if (options.direction === 'right' || nodeEl === comp.refs.right.el) {
      offset = 1;
    }
    let coor;
    if (offset !== null) {
      coor = new Coordinate$2([comp.props.node.id], offset);
      coor._comp = comp;
    }
    return coor
  };

  class Blocker$2 extends Component$2 {
    render ($$) {
      return $$('div').addClass('sc-isolated-node-blocker')
        .attr('draggable', true)
        .attr('contenteditable', false)
        .on('click', this.onClick)
        .on('dblclick', this.onDblClick)
    }

    onClick (event) {
      if (event.target !== this.getNativeElement()) return
      
      event.stopPropagation();
      const comp = this._getIsolatedNodeComponent();
      comp.extendState({ mode: 'selected', unblocked: true });
      comp.selectNode();
    }

    onDblClick (event) {
      
      event.stopPropagation();
    }

    _getIsolatedNodeComponent () {
      return this.context.isolatedNodeComponent
    }
  }

  const DEBUG$1$2 = false;


  class DOMSelection$2 {
    constructor (editor) {
      this.editor = editor;
      if (platform$2.inBrowser) {
        this.wRange = window.document.createRange();
      }
      
      
      this.state = { dom: null, model: null };
    }

    
    getSelection (options) {
      
      if (!platform$2.inBrowser) return
      let range = this.mapDOMSelection(options);
      let doc = this.editor.getDocument();
      
      return doc._createSelectionFromRange(range)
    }

    getSelectionForDOMRange (wrange) {
      let range = this.mapDOMRange(wrange);
      let doc = this.editor.getDocument();
      return doc._createSelectionFromRange(range)
    }

    
    mapDOMSelection (options) {
      let wSel = window.getSelection();
      let state = this.state;
      let range;
      
      
      if (DEBUG$1$2) console.info('DOM->Model: ', wSel.anchorNode, wSel.anchorOffset, wSel.focusNode, wSel.focusOffset);
      if (wSel.rangeCount === 0) return _null()
      let anchorNode = DefaultDOMElement$2.wrapNativeElement(wSel.anchorNode);
      if (wSel.isCollapsed) {
        let coor = this._getCoordinate(anchorNode, wSel.anchorOffset, options);
        if (!coor) return _null()
        range = _createRange$2({
          start: coor,
          end: coor
        });
      } else {
        let focusNode = DefaultDOMElement$2.wrapNativeElement(wSel.focusNode);
        range = this._getRange(anchorNode, wSel.anchorOffset, focusNode, wSel.focusOffset, options);
      }
      if (DEBUG$1$2) console.info('DOM->Model: range ', range ? range.toString() : null);
      state.model = range;
      return range

      function _null () {
        state.dom = null;
        state.model = null;
        return null
      }
    }

    
    setSelection (sel) {
      
      if (!platform$2.inBrowser) return
      let state = this.state;
      let wSel = window.getSelection();
      let wRange = this.wRange;
      if (!sel || sel.isNull()) return this.clear()
      
      let {start, end} = this.mapModelToDOMCoordinates(sel);
      if (!start) return this.clear()
      if (sel.isReverse()) {
        [start, end] = [end, start];
      }
      state.dom = {
        anchorNode: start.container,
        anchorOffset: start.offset,
        focusNode: end.container,
        focusOffset: end.offset
      };
      _set(state.dom);

      function _set ({anchorNode, anchorOffset, focusNode, focusOffset}) {
        wSel.removeAllRanges();
        wRange.setStart(anchorNode, anchorOffset);
        wRange.setEnd(anchorNode, anchorOffset);
        wSel.addRange(wRange);
        if (focusNode !== anchorOffset || focusOffset !== anchorOffset) {
          wSel.extend(focusNode, focusOffset);
        }
      }
    }

    mapModelToDOMCoordinates (sel) {
      if (DEBUG$1$2) console.info('Model->DOM: sel =', sel.toString());
      let rootEl;
      let surface = this.editor.getSurfaceManager().getSurface(sel.surfaceId);
      if (!surface) {
        console.warn('No surface:', sel.surfaceId);
        rootEl = this.editor.getElement();
      } else {
        rootEl = surface.el;
      }
      if (sel.isNull() || sel.isCustomSelection()) {
        return {}
      }

      let start, end;
      if (sel.isPropertySelection() || sel.isContainerSelection()) {
        start = this._getDOMCoordinate(rootEl, sel.start);
        if (!start) {
          console.warn('FIXME: selection seems to be invalid.');
          return {}
        }
        if (sel.isCollapsed()) {
          end = start;
        } else {
          end = this._getDOMCoordinate(rootEl, sel.end);
          if (!end) {
            console.warn('FIXME: selection seems to be invalid.');
            return {}
          }
        }
      } else if (sel.isNodeSelection()) {
        let comp = Component$2.unwrap(rootEl.find('*[data-id="' + sel.getNodeId() + '"]'));
        if (!comp) {
          console.error('Could not find component with id', sel.getNodeId());
          return {}
        }
        if (comp._isIsolatedNodeComponent) {
          let coors = IsolatedNodeComponent$2.getDOMCoordinates(comp, sel);
          start = coors.start;
          end = coors.end;
          
          
          
        } else {
          let _nodeEl = comp.el;
          start = {
            container: _nodeEl.getNativeElement(),
            offset: 0
          };
          end = {
            container: _nodeEl.getNativeElement(),
            offset: _nodeEl.getChildCount()
          };
        }
      }
      if (DEBUG$1$2) console.info('Model->DOM:', start.container, start.offset, end.container, end.offset, 'isReverse?', sel.isReverse());
      return {start, end}
    }

    _getDOMCoordinate (rootEl, coor) {
      let domCoor = null;
      let comp;
      if (coor.isNodeCoordinate()) {
        comp = Component$2.unwrap(rootEl.find('*[data-id="' + coor.getNodeId() + '"]'));
        if (comp) {
          if (comp._isIsolatedNodeComponent) {
            domCoor = IsolatedNodeComponent$2.getDOMCoordinate(comp, coor);
          } else {
            let domOffset = 0;
            if (coor.offset > 0) {
              domOffset = comp.getChildCount();
            }
            domCoor = {
              container: comp.getNativeElement(),
              offset: domOffset
            };
          }
        }
      } else {
        comp = Component$2.unwrap(rootEl.find('.sc-text-property[data-path="' + coor.path.join('.') + '"]'));
        if (comp) {
          domCoor = comp.getDOMCoordinate(coor.offset);
        }
      }
      return domCoor
    }

    
    mapDOMRange (wRange, options) {
      return this._getRange(
        DefaultDOMElement$2.wrapNativeElement(wRange.startContainer),
        wRange.startOffset,
        DefaultDOMElement$2.wrapNativeElement(wRange.endContainer),
        wRange.endOffset, options)
    }

    
    clear () {
      window.getSelection().removeAllRanges();
      this.state.dom = null;
      this.state.model = null;
    }

    collapse (dir) {
      let wSel = window.getSelection();
      let wRange;
      if (wSel.rangeCount > 0) {
        wRange = wSel.getRangeAt(0);
        wRange.collapse(dir === 'left');
        wSel.removeAllRanges();
        wSel.addRange(wRange);
      }
    }

    select (el) {
      let wSel = window.getSelection();
      let wRange = window.document.createRange();
      wRange.selectNode(el.getNativeElement());
      wSel.removeAllRanges();
      wSel.addRange(wRange);
    }

    extend (el, offset) {
      let wSel = window.getSelection();
      wSel.extend(el.getNativeElement(), offset);
    }

    setCursor (el, offset) {
      let wSel = window.getSelection();
      let wRange = window.document.createRange();
      wRange.setStart(el.getNativeElement(), offset);
      wSel.removeAllRanges();
      wSel.addRange(wRange);
    }

    
    _getRange (anchorNode, anchorOffset, focusNode, focusOffset, options = {}) {
      let isReverse = DefaultDOMElement$2.isReverse(anchorNode, anchorOffset, focusNode, focusOffset);
      let isCollapsed = (anchorNode === focusNode && anchorOffset === focusOffset);
      let start, end;
      if (isCollapsed) {
        start = end = this._getCoordinate(anchorNode, anchorOffset, options);
      } else {
        start = this._getCoordinate(anchorNode, anchorOffset, { direction: isReverse ? 'right' : 'left' });
        end = this._getCoordinate(focusNode, focusOffset, options);
      }
      if (start && end) {
        return _createRange$2({ start, end, isReverse })
      } else {
        return null
      }
    }

    
    _getCoordinate (nodeEl, offset, options = {}) {
      let coor = null;
      
      if (!coor) {
        coor = TextPropertyComponent$2.getCoordinate(this.editor.getElement(), nodeEl, offset);
      }
      let comp = Component$2.unwrap(nodeEl);
      if (!coor && comp) {
        
        if (comp.context.isolatedNodeComponent) {
          coor = IsolatedNodeComponent$2.getCoordinate(nodeEl, options);
        }
      }
      
      
      if (!coor) {
        
        if (comp && comp._isContainerEditor) {
          let childIdx = (offset === 0) ? 0 : offset - 1;
          let isBefore = (offset === 0);
          let container = comp.getContainer();
          let childNode = container.getNodeAt(childIdx);
          let childComp = comp.getChildAt(childIdx);
          coor = new Coordinate$2([childNode.id], isBefore ? 0 : 1);
          coor._comp = childComp;
        
        } else if (nodeEl.isElementNode() && nodeEl.getChildCount() > 0) {
          let child = (offset > 0) ? nodeEl.getChildAt(offset - 1) : nodeEl.firstChild;
          let prop;
          let childComp = Component$2.unwrap(child);
          if (childComp && childComp._isTextPropertyComponent) {
            prop = child;
          }
          
          if (prop) {
            coor = TextPropertyComponent$2.getCoordinate(nodeEl, prop, (offset > 0) ? prop.getChildCount() : 0);
          }
        }
      }
      return coor
    }
  }


  function _createRange$2 ({start, end, isReverse}) {
    if (isReverse) {
      [start, end] = [end, start];
    }
    if (!start._comp || !end._comp) {
      console.error('FIXME: getCoordinate() should provide a component instance');
      return null
    }
    let surface = start._comp.context.surface;
    if (!surface) {
      console.error('FIXME: Editable components should have their surface in the context');
      return null
    }
    if (surface !== end._comp.context.surface) {
      console.error('Coordinates are within two different surfaces. Can not create a selection.');
      return null
    }
    return new Range$2(start, end, isReverse, surface.getContainerId(), surface.id)
  }

  class Command$2 {
    
    constructor (config) {
      this.config = config || {};
      this.name = this.config.name;
      if (!this.name) {
        throw new Error("'name' is required")
      }
    }

    get isAsync () {
      return false
    }

    
    getName () {
      return this.name
    }

    
    getCommandState(params, context) { 
      throw new Error('Command.getCommandState() is abstract.')
    }

    
    execute(params, context) { 
      throw new Error('Command.execute() is abstract.')
    }

    
    isAnnotationCommand () {
      return false
    }

    
    isInsertCommand () {
      return false
    }

    
    isSwitchTypeCommand () {
      return false
    }

    _getEditorSession (params, context) {
      let editorSession = params.editorSession || context.editorSession;
      if (!editorSession) {
        throw new Error("'editorSession' is required.")
      }
      return editorSession
    }

    _getSelection (params) {
      let sel = params.selection || params.selectionState.getSelection();
      if (!sel) {
        throw new Error("'selection' is required.")
      }
      return sel
    }
  }

  Command$2.prototype._isCommand = true;

  class NodeComponent$2 extends Component$2 {
    didMount () {
      this.context.editorSession.onRender('document', this.rerender, this, { path: [this.props.node.id] });
    }

    dispose () {
      this.context.editorSession.off(this);
    }

    render ($$) {
      let tagName = this.getTagName();
      let el = $$(tagName)
        .attr('data-id', this.props.node.id)
        .addClass(this.getClassNames());
      return el
    }

    getTagName () {
      return 'div'
    }

    getClassNames () {
      return ''
    }

    rerender (...args) {
      
      if (this.props.node.isDisposed()) return
      super.rerender(...args);
    }
  }

  class Clipboard$2 extends AbstractClipboard$2 {
    constructor (editorSession) {
      super(editorSession.getConfigurator());

      this.editorSession = editorSession;
    }

    
    _copy () {
      let editorSession = this.getEditorSession();
      let sel = editorSession.getSelection();
      let doc = editorSession.getDocument();
      let clipboardDoc = null;
      let clipboardText = '';
      let clipboardHtml = '';
      let htmlExporter = this._getExporter();
      if (!sel.isCollapsed()) {
        clipboardText = getTextForSelection$2(doc, sel) || '';
        clipboardDoc = copySelection$2(doc, sel);
        clipboardHtml = htmlExporter.exportDocument(clipboardDoc);
      }
      return {
        doc: clipboardDoc,
        html: clipboardHtml,
        text: clipboardText
      }
    }

    _cut () {
      let editorSession = this.getEditorSession();
      editorSession.transaction((tx) => {
        tx.deleteSelection();
      });
    }

    
    _pastePlainText (plainText) {
      let editorSession = this.getEditorSession();
      editorSession.transaction(function (tx) {
        tx.paste(plainText);
      }, { action: 'paste' });
    }

    
    _pasteHtml (html, text) {
      let htmlImporter = this._getImporter();
      let content = htmlImporter.importDocument(html) || text;
      if (content) {
        let editorSession = this.getEditorSession();
        editorSession.transaction((tx) => {
          tx.paste(content);
        }, { action: 'paste' });
      }
      return true
    }

    getEditorSession () {
      return this.editorSession
    }
  }

  function startsWith$2 (str, prefix) {
    if (!isString$2(str)) return false
    if (str.startsWith) return str.startsWith(prefix)
    if (!isString$2(prefix)) prefix = String(prefix);
    return str.slice(0, prefix.length) === prefix
  }

  class UnsupportedNodeComponent$2 extends Component$2 {
    render ($$) {
      return $$('pre')
        .addClass('content-node unsupported')
        .attr({
          'data-id': this.props.node.id,
          contentEditable: false
        })
        .append(
          JSON.stringify(this.props.node.properties, null, 2)
        )
    }
  }

  const BROWSER_DELAY$2 = platform$2.isFF ? 1 : 0;


  class Surface$2 extends Component$2 {
    constructor (...args) {
      super(...args);

      this._initialize();
    }

    _initialize () {
      
      
      this.editorSession = this.props.editorSession || this.context.editorSession;
      if (!this.editorSession) {
        throw new Error('No EditorSession provided')
      }
      this.name = this.props.name;
      if (!this.name) {
        throw new Error('Surface must have a name.')
      }
      if (this.name.indexOf('/') > -1) {
        
        throw new Error("Surface.name must not contain '/'")
      }
      
      
      this._surfaceId = Surface$2.createSurfaceId(this);

      this.clipboard = new Clipboard$2(this.editorSession);

      this.domSelection = this.context.domSelection || new DOMSelection$2(this);

      this._state = {
        
        skipNextFocusEvent: false
      };
    }

    getChildContext () {
      return {
        surface: this,
        parentSurfaceId: this.getId(),
        doc: this.getDocument(),
        
        
        isolatedNodeComponent: null
      }
    }

    didMount () {
      const editorSession = this.getEditorSession();
      const surfaceManager = this.getSurfaceManager();
      if (surfaceManager) {
        surfaceManager.registerSurface(this);
      }
      editorSession.onRender('selection', this._onSelectionChanged, this);

      const globalEventHandler = editorSession.globalEventHandler;
      globalEventHandler.addEventListener('keydown', this._muteNativeHandlers, this);
    }

    dispose () {
      const editorSession = this.getEditorSession();
      const surfaceManager = this.getSurfaceManager();
      editorSession.off(this);
      if (surfaceManager) {
        surfaceManager.unregisterSurface(this);
      }

      const globalEventHandler = editorSession.globalEventHandler;
      globalEventHandler.removeEventListener('keydown', this._muteNativeHandlers);
    }

    didUpdate () {
      this._updateContentEditableState();
    }

    render ($$) {
      let tagName = this.props.tagName || 'div';
      let el = $$(tagName)
        .addClass('sc-surface')
        .attr('tabindex', 2)
        .attr('data-surface-id', this.id);

      if (!this.isDisabled()) {
        if (this.isEditable()) {
          
          el.on('keydown', this.onKeyDown);
          
          if (!platform$2.isIE) {
            el.on('compositionstart', this.onCompositionStart);
            el.on('compositionend', this.onCompositionEnd);
          }
          
          
          
          
          if (platform$2.inBrowser && window.TextEvent && !platform$2.isIE) {
            el.on('textInput', this.onTextInput);
          } else {
            el.on('keypress', this.onTextInputShim);
          }
          el.on('paste', this._onPaste);
          el.on('cut', this._onCut);
        }
        if (!this.isReadonly()) {
          
          el.on('mousedown', this.onMouseDown);
          el.on('contextmenu', this.onContextMenu);
          
          
          el.on('focus', this.onNativeFocus);
          el.on('blur', this.onNativeBlur);
        }
        el.on('copy', this._onCopy);
      }
      return el
    }

    renderNode ($$, node) {
      let componentRegistry = this.getComponentRegistry();
      let ComponentClass = componentRegistry.get(node.type);
      if (!ComponentClass) {
        console.error('Could not resolve a component for type: ' + node.type);
        ComponentClass = UnsupportedNodeComponent$2;
      }
      return $$(ComponentClass, this._extractNodeProps(node)).ref(node.id)
    }

    getComponentRegistry () {
      return this.context.componentRegistry || this.props.componentRegistry
    }

    getName () {
      return this.name
    }

    getId () {
      return this._surfaceId
    }

    getSurfaceId () {
      return this._surfaceId
    }

    isDisabled () {
      return this.props.disabled
    }

    isEditable () {
      return (this.props.editing === 'full' || this.props.editing === undefined)
    }

    isSelectable () {
      return (this.props.editing === 'selection' || this.props.editing === 'full')
    }

    isReadonly () {
      return this.props.editing === 'readonly'
    }

    getElement () {
      return this.el
    }

    getDocument () {
      return this.getEditorSession().getDocument()
    }

    getEditorSession () {
      return this.editorSession
    }

    getSurfaceManager () {
      return this.context.surfaceManager
    }

    isEnabled () {
      return !this.state.disabled
    }

    isContainerEditor () {
      return false
    }

    isCustomEditor () {
      return false
    }

    hasNativeSpellcheck () {
      return this.props.spellcheck === 'native'
    }

    getContainerId () {
      return null
    }

    focus () {
      if (this.getEditorSession().getFocusedSurface() !== this) {
        this.selectFirst();
      }
    }

    blur () {
      if (this.getEditorSession().getFocusedSurface() === this) {
        this.getEditorSession().setSelection(null);
      }
    }

    selectFirst () {
      throw new Error('This method is abstract.')
    }

    
    rerenderDOMSelection () {
      if (this.isDisabled()) return
      if (platform$2.inBrowser) {
        
        let sel = this.getEditorSession().getSelection();
        if (sel.surfaceId === this.getId()) {
          this.domSelection.setSelection(sel);
          
          
          
          const scrollPane = this.context.scrollPane;
          if (scrollPane) {
            this.context.scrollPane.onSelectionPositioned();
          }
        }
      }
    }

    getDomNodeForId (nodeId) {
      return this.el.getNativeElement().querySelector('*[data-id="' + nodeId + '"]')
    }

    

    
    onKeyDown (event) {
      if (!this._shouldConsumeEvent(event)) return
      

      
      if (event.key === 'Dead') return

      
      let custom = this.getEditorSession().keyboardManager.onKeydown(event);
      if (!custom) {
        
        switch (event.keyCode) {
          
          case keys$1$2.LEFT:
          case keys$1$2.RIGHT:
            return this._handleLeftOrRightArrowKey(event)
          case keys$1$2.UP:
          case keys$1$2.DOWN:
            return this._handleUpOrDownArrowKey(event)
          case keys$1$2.HOME:
          case keys$1$2.END:
            return this._handleHomeOrEndKey(event)
          case keys$1$2.PAGEUP:
          case keys$1$2.PAGEDOWN:
            return this._handlePageUpOrDownKey(event)
          
          case keys$1$2.ENTER:
            return this._handleEnterKey(event)
          case keys$1$2.TAB:
            return this._handleTabKey(event)
          case keys$1$2.BACKSPACE:
          case keys$1$2.DELETE:
            return this._handleDeleteKey(event)
          case keys$1$2.ESCAPE:
            return this._handleEscapeKey(event)
          case keys$1$2.SPACE:
            return this._handleSpaceKey(event)
          default:
            break
        }
      }
    }

    onTextInput (event) {
      if (!this._shouldConsumeEvent(event)) return
      
      event.preventDefault();
      event.stopPropagation();
      if (!event.data) return
      const editorSession = this.getEditorSession();
      let text = event.data;
      if (!editorSession.keyboardManager.onTextInput(text)) {
        editorSession.transaction((tx) => {
          tx.insertText(text);
        }, { action: 'type' });
      }
    }

    
    onCompositionStart (event) {
      if (!this._shouldConsumeEvent(event)) return
      
      
      
      
      
      
      
      if (event.data) {
        let l = event.data.length;
        let sel = this.getEditorSession().getSelection();
        if (sel.isPropertySelection() && sel.isCollapsed()) {
          
          let offset = sel.start.offset;
          this.getEditorSession().setSelection(sel.createWithNewRange(offset - l, offset));
        }
      }
    }

    onCompositionEnd (event) {
      if (!this._shouldConsumeEvent(event)) return
      
      
      
      if (platform$2.isFF) {
        event.preventDefault();
        event.stopPropagation();
        if (!event.data) return
        this._delayed(() => {
          let text = event.data;
          if (!this.getEditorSession().keyboardManager.onTextInput(text)) {
            this.getEditorSession().transaction((tx) => {
              tx.insertText(text);
            }, { action: 'type' });
          }
        });
      }
    }

    
    onTextInputShim (event) {
      if (!this._shouldConsumeEvent(event)) return
      
      if (
        
        event.which === 0 || event.charCode === 0 ||
        
        event.keyCode === keys$1$2.TAB || event.keyCode === keys$1$2.ESCAPE ||
        
        Boolean(event.metaKey) || (Boolean(event.ctrlKey) ^ Boolean(event.altKey))
      ) {
        return
      }
      let character = String.fromCharCode(event.which);
      if (!event.shiftKey) {
        character = character.toLowerCase();
      }
      event.preventDefault();
      event.stopPropagation();
      if (!this.getEditorSession().keyboardManager.onTextInput(character)) {
        if (character.length > 0) {
          this.getEditorSession().transaction((tx) => {
            tx.insertText(character);
          }, { action: 'type' });
        }
      }
    }

    
    
    
    
    
    onMouseDown (event) {
      if (!this._shouldConsumeEvent(event)) {
        
        return
      }

      
      
      
      
      
      if (event.__reserved__) {
        
        return
      } else {
        
        event.__reserved__ = this;
      }

      
      
      
      if (this.isEditable()) {
        this.el.setAttribute('contenteditable', true);
      }

      
      if (event.button !== 0) {
        return
      }

      
      if (!(platform$2.isIE && platform$2.version < 12) && event.detail >= 3) {
        let sel = this.getEditorSession().getSelection();
        if (sel.isPropertySelection()) {
          this._selectProperty(sel.path);
          event.preventDefault();
          event.stopPropagation();
          return
        } else if (sel.isContainerSelection()) {
          this._selectProperty(sel.startPath);
          event.preventDefault();
          event.stopPropagation();
          return
        }
      }
      
      
      
      
      this._state.skipNextFocusEvent = true;

      
      if (platform$2.inBrowser) {
        let documentEl = DefaultDOMElement$2.wrapNativeElement(window.document);
        documentEl.on('mouseup', this.onMouseUp, this, { once: true });
      }
    }

    onMouseUp (e) {
      
      
      
      
      
      e.stopPropagation();
      
      
      
      
      this._delayed(() => {
        let sel = this.domSelection.getSelection();
        this._setSelection(sel);
      });
    }

    
    
    
    onContextMenu (event) {
      if (!this._shouldConsumeEvent(event)) return
      let sel = this.domSelection.getSelection();
      this._setSelection(sel);
    }

    onNativeBlur () {
      
      let _state = this._state;
      _state.hasNativeFocus = false;
    }

    onNativeFocus () {
      
      let _state = this._state;
      _state.hasNativeFocus = true;
    }

    _onCopy (e) {
      this.clipboard.onCopy(e);
    }

    _onCut (e) {
      this.clipboard.onCut(e);
    }

    _onPaste (e) {
      this.clipboard.onPaste(e);
    }

    

    _onSelectionChanged (selection) {
      let newMode = this._deriveModeFromSelection(selection);
      if (this.state.mode !== newMode) {
        this.extendState({
          mode: newMode
        });
      }
    }

    
    _deriveModeFromSelection (sel) {
      if (!sel) return null
      let surfaceId = sel.surfaceId;
      let id = this.getId();
      let mode;
      if (startsWith$2(surfaceId, id)) {
        if (surfaceId.length === id.length) {
          mode = 'focused';
        } else {
          mode = 'co-focused';
        }
      }
      return mode
    }

    _updateContentEditableState () {
      
      
      
      
      
      let enableContenteditable = false;
      if (this.isEditable() && !this.props.disabled) {
        enableContenteditable = true;
        if (this.state.mode === 'co-focused') {
          let selState = this.context.editorSession.getSelectionState();
          let sel = selState.getSelection();
          let surface = this.getSurfaceManager().getSurface(sel.surfaceId);
          if (surface) {
            let isolatedNodeComponent = surface.context.isolatedNodeComponent;
            if (isolatedNodeComponent) {
              enableContenteditable = isolatedNodeComponent.isOpen();
            }
          }
        }
      }
      if (enableContenteditable) {
        this.el.setAttribute('contenteditable', true);
      } else {
        
        this.el.removeAttribute('contenteditable');
      }
    }

    _blur () {
      if (this.el) {
        this.el.blur();
      }
    }

    _focus () {
      if (this.isDisabled()) return
      
      
      
      if (platform$2.isFF) {
        this.domSelection.clear();
        this.el.getNativeElement().blur();
      }
      this._focusElement();
    }

    _focusElement () {
      this._state.hasNativeFocus = true;
      
      
      
      
      if (this.el && !platform$2.isWebkit) {
        this._state.skipNextFocusEvent = true;
        
        
        this.el.focus();
        this._state.skipNextFocusEvent = false;
      }
    }

    _handleLeftOrRightArrowKey (event) {
      event.stopPropagation();
      let direction = (event.keyCode === keys$1$2.LEFT) ? 'left' : 'right';
      
      
      this._delayed(() => {
        this._updateModelSelection({direction});
      });
    }

    _handleUpOrDownArrowKey (event) {
      event.stopPropagation();
      
      
      this._delayed(() => {
        let options = {
          direction: (event.keyCode === keys$1$2.UP) ? 'left' : 'right'
        };
        this._updateModelSelection(options);
      });
    }

    _handleHomeOrEndKey (event) {
      event.stopPropagation();
      
      
      this._delayed(() => {
        let options = {
          direction: (event.keyCode === keys$1$2.HOME) ? 'left' : 'right'
        };
        this._updateModelSelection(options);
      });
    }

    _handlePageUpOrDownKey (event) {
      event.stopPropagation();
      
      
      this._delayed(() => {
        let options = {
          direction: (event.keyCode === keys$1$2.PAGEUP) ? 'left' : 'right'
        };
        this._updateModelSelection(options);
      });
    }

    _handleSpaceKey (event) {
      event.stopPropagation();
      event.preventDefault();
      const text = ' ';
      if (!this.getEditorSession().keyboardManager.onTextInput(text)) {
        this.getEditorSession().transaction((tx) => {
          tx.insertText(text);
        }, { action: 'type' });
      }
    }

    _handleTabKey (event) {
      event.stopPropagation();
      this.el.emit('tab', {
        altKey: event.altKey,
        ctrlKey: event.ctrlKey,
        metaKey: event.metaKey,
        shiftKey: event.shiftKey,
        code: event.code
      });
      if (this.props.handleTab === false) {
        event.preventDefault();
      } else {
        this.__handleTab(event);
      }
    }

    __handleTab () {
      
      
      this._delayed(() => {
        this._updateModelSelection();
      });
    }

    _handleEnterKey (event) {
      event.preventDefault();
      event.stopPropagation();
      this.getEditorSession().transaction((tx) => {
        tx.break();
      }, { action: 'break' });
    }

    _handleEscapeKey () {}

    _handleDeleteKey (event) {
      event.preventDefault();
      event.stopPropagation();
      let direction = (event.keyCode === keys$1$2.BACKSPACE) ? 'left' : 'right';
      this.getEditorSession().transaction((tx) => {
        tx.deleteCharacter(direction);
      }, { action: 'delete' });
    }

    _hasNativeFocus () {
      return Boolean(this._state.hasNativeFocus)
    }

    _setSelection (sel) {
      
      
      
      
      
      
      
      
      if (!sel.isNull() && sel.surfaceId === this.id && platform$2.isFF) {
        this._focusElement();
      }
      this.getEditorSession().setSelection(sel);
    }

    _updateModelSelection (options) {
      let sel = this.domSelection.getSelection(options);
      
      
      
      this._setSelection(sel);
    }

    _selectProperty (path) {
      let doc = this.getDocument();
      let text = doc.get(path);
      this._setSelection(doc.createSelection({
        type: 'property',
        path: path,
        startOffset: 0,
        endOffset: text.length
      }));
    }

    
    
    _renderNode ($$, nodeId) {
      let doc = this.getDocument();
      let node = doc.get(nodeId);
      let componentRegistry = this.context.componentRegistry || this.props.componentRegistry;
      let ComponentClass = componentRegistry.get(node.type);
      if (!ComponentClass) {
        console.error('Could not resolve a component for type: ' + node.type);
        ComponentClass = UnsupportedNodeComponent$2;
      }
      return $$(ComponentClass, {
        doc: doc,
        node: node
      })
    }

    _extractNodeProps (node) {
      let doc = this.getDocument();
      return {
        placeholder: this.props.placeholder,
        doc: doc,
        node: node
      }
    }

    
    _shouldConsumeEvent (event) {
      
      let comp = Component$2.unwrap(event.target);
      return (comp && (comp === this || comp.context.surface === this))
    }

    
    getSelectionFromEvent (event) {
      let domRange = getDOMRangeFromEvent$2(event);
      let sel = this.domSelection.getSelectionForDOMRange(domRange);
      sel.surfaceId = this.getId();
      return sel
    }

    setSelectionFromEvent (event) {
      let sel = this.getSelectionFromEvent(event);
      if (sel) {
        this._state.skipNextFocusEvent = true;
        this._setSelection(sel);
      } else {
        console.error('Could not create a selection from event.');
      }
    }

    get id () {
      return this._surfaceId
    }

    _delayed (fn) {
      window.setTimeout(fn, BROWSER_DELAY$2);
    }

    
    _muteNativeHandlers (event) {
      let contentEditableShortcuts;

      if (platform$2.isMac) {
        contentEditableShortcuts = [
          'META+66', 
          'META+73', 
          'META+85' 
        ];
      } else {
        contentEditableShortcuts = [
          'CTRL+66', 
          'CTRL+73', 
          'CTRL+85' 
        ];
      }

      const key = parseKeyEvent$2(event);
      if (contentEditableShortcuts.indexOf(key) > -1) {
        event.preventDefault();
      }
    }
  }

  Surface$2.prototype._isSurface = true;


  Surface$2.createSurfaceId = function (surface) {
    let parentSurfaceId = surface.context.parentSurfaceId;
    if (parentSurfaceId) {
      return parentSurfaceId + '/' + surface.name
    } else {
      return surface.name
    }
  };

  class ContainerEditor$2 extends Surface$2 {
    constructor (parent, props, el) {
      
      props.containerId = props.containerId || props.node.id;
      props.name = props.name || props.containerId || props.node.id;

      super(parent, props, el);

      this.containerId = this.props.containerId;
      if (!isString$2(this.containerId)) {
        throw new Error("Property 'containerId' is mandatory.")
      }
      let doc = this.getDocument();
      this.container = doc.get(this.containerId);
      if (!this.container) {
        throw new Error('Container with id ' + this.containerId + ' does not exist.')
      }

      this.editingBehavior = this.context.editingBehavior || new EditingBehavior$2();

      this._deriveInternalState(this.props);
    }

    
    shouldRerender (newProps) {
      if (newProps.disabled !== this.props.disabled) return true
      
      
      return false
    }

    willReceiveProps (newProps) {
      super.willReceiveProps.apply(this, arguments);
      this._deriveInternalState(newProps);
    }

    didMount () {
      super.didMount.apply(this, arguments);
      let editorSession = this.getEditorSession();
      editorSession.onUpdate('document', this._onContainerChanged, this, {
        path: this.container.getContentPath()
      });
      this._attachPlaceholder();
    }

    dispose () {
      super.dispose.apply(this, arguments);
      let editorSession = this.getEditorSession();
      editorSession.off(this);
    }

    render ($$) {
      let el = super.render($$);

      let doc = this.getDocument();
      let containerId = this.getContainerId();
      let containerNode = doc.get(containerId);
      if (!containerNode) {
        console.warn('No container node found for ', containerId);
      }
      el.addClass('sc-container-editor container-node ' + containerId)
        .attr('data-id', containerId);

      
      el.attr('spellcheck', this.props.spellcheck === 'native');

      containerNode.getNodes().forEach(function (node, index) {
        el.append(this._renderNode($$, node, index));
      }.bind(this));

      
      if (!this.props.disabled && !this.isEmpty()) {
        el.addClass('sm-enabled');
        el.setAttribute('contenteditable', true);
      }

      return el
    }

    selectFirst () {
      const container = this.getContainer();
      if (container.getLength() > 0) {
        const editorSession = this.getEditorSession();
        const first = container.getChildAt(0);
        setCursor$2(editorSession, first, container.id, 'before');
      }
    }

    _renderNode ($$, node, nodeIndex) {
      let props = { node };
      if (!node) throw new Error('Illegal argument')
      if (node.isText()) {
        return this.renderNode($$, node, nodeIndex)
      } else {
        let componentRegistry = this.context.componentRegistry;
        let ComponentClass = componentRegistry.get(node.type);
        if (ComponentClass.prototype._isCustomNodeComponent || ComponentClass.prototype._isIsolatedNodeComponent) {
          return $$(ComponentClass, props).ref(node.id)
        } else {
          return $$(IsolatedNodeComponent$2, props).ref(node.id)
        }
      }
    }

    _extractNodeProps (node) {
      let doc = this.getDocument();
      return {
        doc: doc,
        node: node
      }
    }

    _deriveInternalState (props) {
      let _state = this._state;
      if (!props.hasOwnProperty('enabled') || props.enabled) {
        _state.enabled = true;
      } else {
        _state.enabled = false;
      }
    }

    _selectNextIsolatedNode (direction) {
      let selState = this.getEditorSession().getSelectionState();
      let node = (direction === 'left') ? selState.getPreviousNode() : selState.getNextNode();
      let isIsolatedNode = !node.isText() && !node.isList();
      if (!node || !isIsolatedNode) return false
      if (
        (direction === 'left' && selState.isFirst()) ||
        (direction === 'right' && selState.isLast())
      ) {
        this.getEditorSession().setSelection({
          type: 'node',
          nodeId: node.id,
          containerId: selState.getContainer().id,
          surfaceId: this.id
        });
        return true
      }
      return false
    }

    _handleLeftOrRightArrowKey (event) {
      event.stopPropagation();
      const doc = this.getDocument();
      const sel = this.getEditorSession().getSelection();
      const left = (event.keyCode === keys$1$2.LEFT);
      const right = !left;
      const direction = left ? 'left' : 'right';

      if (sel && !sel.isNull()) {
        const container = doc.get(sel.containerId, 'strict');

        
        if (sel.isNodeSelection()) {
          let nodePos = container.getPosition(doc.get(sel.getNodeId()));
          if ((left && nodePos === 0) || (right && nodePos === container.length - 1)) {
            event.preventDefault();
            return
          }
        }

        if (sel.isNodeSelection() && !event.shiftKey) {
          this.domSelection.collapse(direction);
        }
      }

      this._delayed(() => {
        this._updateModelSelection({ direction });
      });
    }

    _handleUpOrDownArrowKey (event) {
      event.stopPropagation();
      const doc = this.getDocument();
      const sel = this.getEditorSession().getSelection();
      const up = (event.keyCode === keys$1$2.UP);
      const down = !up;
      const direction = up ? 'left' : 'right';

      if (sel && !sel.isNull()) {
        const container = doc.get(sel.containerId, 'strict');
        
        if (sel.isNodeSelection()) {
          let nodePos = container.getPosition(doc.get(sel.getNodeId()));
          if ((up && nodePos === 0) || (down && nodePos === container.length - 1)) {
            event.preventDefault();
            return
          }
          
          
          let editorSession = this.getEditorSession();
          
          
          
          if (!event.shiftKey) {
            event.preventDefault();
            if (up) {
              let prev = container.getChildAt(nodePos - 1);
              setCursor$2(editorSession, prev, sel.containerId, 'after');
              return
            } else {
              let next = container.getChildAt(nodePos + 1);
              setCursor$2(editorSession, next, sel.containerId, 'before');
              return
            }
          }
        }
      }

      this._delayed(() => {
        this._updateModelSelection({ direction });
      });
    }

    _handleTabKey (event) {
      const editorSession = this.getEditorSession();
      const sel = editorSession.getSelection();
      
      if (sel.isNodeSelection() && sel.isFull()) {
        const comp = this.refs[sel.getNodeId()];
        if (comp && stepIntoIsolatedNode$2(editorSession, comp)) {
          event.preventDefault();
          event.stopPropagation();
          return
        }
      }
      super._handleTabKey(event);
    }

    __handleTab (e) {
      e.preventDefault();
      if (e.shiftKey) {
        this.getEditorSession().transaction((tx) => {
          tx.dedent();
        }, { action: 'dedent' });
      } else {
        this.getEditorSession().transaction((tx) => {
          tx.indent();
        }, { action: 'indent' });
      }
    }

    
    isContainerEditor () {
      return true
    }

    
    getContainerId () {
      return this.containerId
    }

    getContainer () {
      return this.getDocument().get(this.getContainerId())
    }

    isEmpty () {
      let containerNode = this.getContainer();
      return (containerNode && containerNode.length === 0)
    }

    
    _attachPlaceholder () {
      let firstNode = this.childNodes[0];
      
      if (this.placeholderNode) {
        this.placeholderNode.extendProps({
          placeholder: undefined
        });
      }

      if (this.childNodes.length === 1 && this.props.placeholder) {
        firstNode.extendProps({
          placeholder: this.props.placeholder
        });
        this.placeholderNode = firstNode;
      }
    }

    isEditable () {
      return super.isEditable.call(this) && !this.isEmpty()
    }

    
    _onContainerChanged (change) {
      let doc = this.getDocument();
      
      let renderContext = RenderingEngine$2.createContext(this);
      let $$ = renderContext.$$;
      let container = this.getContainer();
      let path = container.getContentPath();
      for (let i = 0; i < change.ops.length; i++) {
        let op = change.ops[i];
        if (op.type === 'update' && op.path[0] === path[0]) {
          let diff = op.diff;
          if (diff.type === 'insert') {
            let nodeId = diff.getValue();
            let node = doc.get(nodeId);
            let nodeEl;
            if (node) {
              nodeEl = this._renderNode($$, node);
            } else {
              
              
              
              nodeEl = $$('div');
            }
            this.insertAt(diff.getOffset(), nodeEl);
          } else if (diff.type === 'delete') {
            this.removeAt(diff.getOffset());
          }
        }
      }
      this._attachPlaceholder();
    }
  }

  ContainerEditor$2.prototype._isContainerEditor = true;

  class TextPropertyEditor$2 extends Surface$2 {
    constructor (parent, props) {
      
      props.name = props.name || props.path.join('.');
      super(parent, props);

      if (!props.path) {
        throw new Error("Property 'path' is mandatory.")
      }
    }

    render ($$) {
      let el = super.render.apply(this, arguments);
      el.addClass('sc-text-property-editor');

      if (!this.props.disabled) {
        el.addClass('sm-enabled');
        el.attr('contenteditable', true);
        
        el.attr('spellcheck', this.props.spellcheck === 'native');
      }

      el.append(
        $$(TextPropertyComponent$2, {
          placeholder: this.props.placeholder,
          tagName: this.props.tagName || 'div',
          path: this.props.path,
          markers: this.props.markers,
          withoutBreak: this.props.withoutBreak
        }).ref('property')
      );

      return el
    }

    selectFirst () {
      this.editorSession.setSelection({
        type: 'property',
        path: this.getPath(),
        startOffset: 0,
        surfaceId: this.id
      });
    }

    getPath () {
      return this.props.path
    }

    _handleEnterKey (event) {
      event.stopPropagation();
      if (this.props.handleEnter === false || !this.props.multiLine) {
        event.preventDefault();
      }
      this.el.emit('enter', {
        altKey: event.altKey,
        ctrlKey: event.ctrlKey,
        metaKey: event.metaKey,
        shiftKey: event.shiftKey,
        code: event.code
      });
    }

    

    _handleEscapeKey (event) {
      this.el.emit('escape', {
        altKey: event.altKey,
        ctrlKey: event.ctrlKey,
        metaKey: event.metaKey,
        shiftKey: event.shiftKey,
        code: event.code
      });
    }
  }

  TextPropertyEditor$2.prototype._isTextPropertyEditor = true;

  function series$2 (tasks, cb, i) {
    i = i || 0;
    tasks[i](function (err, ...args) {
      
      if (err) return cb(err)
      if (i === tasks.length - 1) {
        cb(err, ...args); 
      } else {
        series$2(tasks, cb, i + 1);
      }
    });
  }

  var async$2 = /*#__PURE__*/Object.freeze({
    series: series$2
  });

  const ALPHABET$2 = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';

  function getColumnLabel$3 (colIdx) {
    if (!isNumber$2(colIdx)) {
      throw new Error('Illegal argument.')
    }
    let label = '';
    while(true) { 
      let mod = colIdx % ALPHABET$2.length;
      colIdx = Math.floor(colIdx / ALPHABET$2.length);
      label = ALPHABET$2[mod] + label;
      if (colIdx > 0) colIdx--;
      else if (colIdx === 0) break
    }
    return label
  }

  function getRowCol$2 (cellLabel) {
    var match = /^([A-Z]+)([1-9][0-9]*)$/.exec(cellLabel);
    return [
      parseInt(match[2], 10) - 1,
      getColumnIndex$2(match[1])
    ]
  }

  const A$2 = 'A'.charCodeAt(0);
  const ALPHABET_LENGTH$2 = ALPHABET$2.length;

  function getColumnIndex$2 (colStr) {
    let index = 0;
    let rank = 0;
    for (let i = colStr.length - 1; i >= 0; i--) {
      let idx = colStr.charCodeAt(i) - A$2;
      if (idx < 0 || idx >= ALPHABET_LENGTH$2) throw new Error('Illegal column label: ' + colStr)
      
      index += Math.pow(ALPHABET_LENGTH$2, rank) * (idx + 1);
      rank++;
    }
    
    return index - 1
  }

  function getCellLabel$3 (rowIdx, colIdx) {
    let colLabel = getColumnLabel$3(colIdx);
    let rowLabel = rowIdx + 1;
    return colLabel + rowLabel
  }

  function getIndexesFromRange$4 (start, end) {
    let [startRow, startCol] = getRowCol$2(start);
    let endRow, endCol;
    if (end) {
      ([endRow, endCol] = getRowCol$2(end));
      if (startRow > endRow) ([startRow, endRow] = [endRow, startRow]);
      if (startCol > endCol) ([startCol, endCol] = [endCol, startCol]);
    } else {
      ([endRow, endCol] = [startRow, startCol]);
    }
    return { startRow, startCol, endRow, endCol }
  }

  function getRangeFromMatrix$3 (cells, startRow, startCol, endRow, endCol, force2D) {
    if (!force2D) {
      if (startRow === endRow && startCol === endCol) {
        let row = cells[startRow];
        if (row) return row[startCol]
        else return undefined
      }
      if (startRow === endRow) {
        let row = cells[startRow];
        if (row) return row.slice(startCol, endCol + 1)
        else return []
      }
      if (startCol === endCol) {
        let res = [];
        for (let i = startRow; i <= endRow; i++) {
          let row = cells[i];
          if (row) res.push(row[startCol]);
        }
        return res
      }
    }
    let res = [];
    for (var i = startRow; i < endRow + 1; i++) {
      let row = cells[i];
      if (row) res.push(row.slice(startCol, endCol + 1));
    }
    return res
  }

  var tableHelpers$2 = /*#__PURE__*/Object.freeze({
    ALPHABET: ALPHABET$2,
    getColumnLabel: getColumnLabel$3,
    getRowCol: getRowCol$2,
    getColumnIndex: getColumnIndex$2,
    getCellLabel: getCellLabel$3,
    getIndexesFromRange: getIndexesFromRange$4,
    getRangeFromMatrix: getRangeFromMatrix$3
  });

  function renderListNode$2 (listNode, $$) {
    let items = listNode.getItems();
    let stack = [$$(_getTagName(1))];
    for (let i = 0; i < items.length; i++) {
      const item = items[i];
      
      const level = item.getLevel();
      if (level < stack.length) {
        for (let j = stack.length; j > level; j--) {
          stack.pop();
        }
      } else if (level > stack.length) {
        for (let j = stack.length; j < level; j++) {
          let list = stack[j - 1];
          let childCount = list.getChildCount();
          let item;
          if (childCount === 0) {
            item = $$('li');
            list.append(item);
          } else {
            item = list.getChildAt(childCount - 1);
          }
          let sublist = $$(_getTagName(j + 1));
          item.append(sublist);
          stack.push(sublist);
        }
      }
      console.assert(level === stack.length, 'item.level should now be the same as stack.length');
      
      last$1$2(stack).append(
        $$(item)
      );
    }
    for (let j = stack.length; j > 1; j--) {
      stack.pop();
    }

    return stack[0]

    function _getTagName (level) {
      let listType = listNode.getListType(level);
      return listType === 'order' ? 'ol' : 'ul'
    }
  }

  const DISABLED$2 = Object.freeze({
    disabled: true
  });

  class ListItemComponent$2 extends Component$2 {
    render ($$) {
      const node = this.props.node;
      const path = node.getPath();

      let el = $$('li').addClass('sc-list-item');
      el.append(
        $$(TextPropertyComponent$2, { path }).ref('text')
      );
      
      if (this.props.children) {
        el.append(this.props.children);
      }
      return el
    }
  }

  class ListComponent$2 extends NodeComponent$2 {
    render ($$) {
      let node = this.props.node;
      let el = renderListNode$2(node, (item) => {
        
        if (isString$2(item)) {
          return $$(item)
        } else if (item.type === 'list-item') {
          return $$(ListItemComponent$2, {
            node: item
          }).ref(item.id)
        }
      });
      el.addClass('sc-list').attr('data-id', node.id);
      return el
    }
  }


  ListComponent$2.prototype._isCustomNodeComponent = true;

  class TableCellComponent$2 extends Component$2 {
    render ($$) {
      let node = this.props.node;
      let el = $$('td').addClass('sc-table-cell');
      el.append(
        $$(TextPropertyEditor$2, {
          path: node.getPath(),
          disabled: this.props.disabled
        }).ref('editor')
      );
      if (node.rowspan > 0) {
        el.attr('rowspan', node.rowspan);
      }
      if (node.colspan > 0) {
        el.attr('colspan', node.colspan);
      }
      return el
    }

    grabFocus () {
      let node = this.props.node;
      this.context.editorSession.setSelection({
        type: 'property',
        path: node.getPath(),
        startOffset: node.getLength(),
        surfaceId: this.refs.editor.id
      });
    }
  }

  TableCellComponent$2.prototype._isTableCellComponent = true;

  // TODO: we need to think about this more
  // we allow to pack primitive values by the calling environment
  // e.g. to add type information or apply some encoding
  // With nested types e.g. single values of an array or object
  // the packing should not be done, instead the packer would need to do something
  // on the higher level
  const UNPACKED = true;

  class ExprNode {
    constructor (id, start, end) {
      this.id = id;
      if (!isNumber$2(start) || !isNumber$2(end)) {
        throw new Error("'start' and 'end' are mandatory")
      }
      this.start = start;
      this.end = end;
    }

    async evaluate (context, unpacked) {
      throw new Error('This is abstract')
    }
  }

  class Definition extends ExprNode {
    constructor (id, start, end, name, rhs) {
      super(id, start, end);
      this.name = name;
      this.rhs = rhs;
      rhs.parent = this;
    }

    get type () { return 'definition' }

    async evaluate (context, unpacked) {
      return this.rhs.evaluate(context, unpacked)
    }
  }

  class ArrayNode extends ExprNode {
    constructor (id, start, end, items) {
      super(id, start, end);
      this.items = items;
      items.forEach(item => {
        item.parent = this;
      });
    }

    get type () { return 'array' }

    async evaluate (context, unpacked) {
      let vals = await Promise.all(this.items.map(i => i.evaluate(context, UNPACKED)));
      if (unpacked) {
        return vals
      } else {
        return context.pack(vals, this.type)
      }
    }
  }

  class ObjectNode extends ExprNode {
    constructor (id, start, end, entries) {
      super(id, start, end);
      this.entries = entries;
      entries.forEach(entry => {
        entry.parent = this;
      });
    }

    get type () { return 'object' }

    async evaluate (context, unpacked) {
      let vals = await Promise.all(this.entries.map(e => e.val.evaluate(context, UNPACKED)));
      let obj = {};
      for (let i = 0; i < this.entries.length; i++) {
        let entry = this.entries[i];
        obj[entry.key] = vals[i];
      }
      if (unpacked) {
        return obj
      } else {
        return context.pack(obj)
      }
    }
  }

  class ConstantNode extends ExprNode {
    constructor (id, start, end, val) {
      super(id, start, end);
      this.val = val;
    }

    async evaluate (context, unpacked) {
      if (unpacked) {
        return this.val
      } else {
        return context.pack(this.val, this.type)
      }
    }
  }

  /*
    A constant value.
  */
  class NumberNode extends ConstantNode {
    get type () { return 'number' }
  }

  class BooleanNode extends ConstantNode {
    get type () { return 'boolean' }
  }

  class StringNode extends ConstantNode {
    get type () { return 'string' }
  }

  class Var extends ExprNode {
    constructor (id, start, end, name) {
      super(id, start, end);
      this.name = name;
    }

    get type () { return 'var' }

    async evaluate (context, unpacked) {
      let val = await context.resolve(this.name);
      if (unpacked) {
        return context.unpack(val)
      } else {
        return val
      }
    }
  }

  class EmptyArgument extends ExprNode {
    get type () { return 'empty-arg' }

    async evaluate () {}
  }

  class FunctionCall extends ExprNode {
    constructor (id, start, end, name, args = [], namedArgs = [], modifiers = []) {
      super(id, start, end);
      this.name = name;
      this.args = args;
      this.modifiers = modifiers;
      this.namedArgs = namedArgs;
      args.forEach((arg) => {
        arg.parent = this;
      });
      namedArgs.forEach((arg) => {
        arg.parent = this;
      });
    }

    get type () { return 'call' }

    async evaluate (context, unpacked) {
      // TODO: what do we expect here? do we want to
      // pack the args, or leave this to context.callFunction()
      let args = await Promise.all(this.args.map(a => a.evaluate(context)));
      let namedArgs = await Promise.all(this.namedArgs.map(a => {
        return {
          name: a.name,
          value: a.evaluate(context)
        }
      }));
      let result = await context.callFunction(this.name, args, namedArgs);
      if (unpacked) {
        return context.unpack(result)
      } else {
        return result
      }
    }
  }

  class NamedArgument extends ExprNode {
    constructor (id, start, end, name, rhs) {
      super(id, start, end);
      this.name = name;
      this.rhs = rhs;
      rhs.parent = this;
    }

    get type () { return 'named-argument' }

    async evaluate (context, unpacked) {
      return this.rhs.evaluate(context, unpacked)
    }
  }

  class PipeOp extends ExprNode {
    constructor (id, start, end, left, right) {
      super(id, start, end);
      this.left = left;
      this.right = right;
      this.left.parent = this;
      this.right.parent = this;
    }

    get type () { return 'pipe' }

    async evaluate (context, unpacked) {
      // first call the left one
      let pipeArg = await this.left.evaluate(context);

      // HACK: adding an additional argument to the RHS
      const right = this.right;
      const args = [{
        name: '_pipe',
        async evaluate (context) {
          return pipeArg
        }
      }].concat(right.args);
      let rightProxy = new FunctionCall(right.id, right.start, right.end,
        right.name, args, right.namedArgs, right.modifiers);

      return rightProxy.evaluate(context, unpacked)
    }
  }

  class ErrorNode extends ExprNode {
    constructor (id, start, end, exception) {
      super(id, start, end);

      this.exception = exception;
    }

    get type () { return 'error' }

    async evaluate () {}
  }

  /**
  * Maps the ANTLR4 AST to a our custom tree model.
  */
  function createNodeTree (ast) {
    let state = {
      // generating ids by counting created nodes
      nodeId: 0,
      nodes: [],
      // extra list to all variables, cells, ranges
      // to be able to compute dependencies
      inputs: [],
      // tokens for code highlighting
      tokens: []
    };
    let root = createFromAST(state, ast);
    return { root, state }
  }

  function createFromAST (state, ast) {
    let node;
    let [start, end] = _getStartStop(ast);
    switch (ast.type) {
      case 'evaluation':
        return createFromAST(state, ast.children[0])
      case 'definition': {
        let lhs = ast.children[0];
        state.tokens.push(new Token$3('output-name', lhs.symbol));
        node = new Definition(state.nodeId++, start, end, lhs.getText(), createFromAST(state, ast.children[2]));
        break
      }
      case 'simple':
        return createFromAST(state, ast.children[0])
      // Member selection operator `.`
      case 'select_id': {
        const value = createFromAST(state, ast.children[0]);
        // Create a new string node from the member ID
        const id = new StringNode(state.nodeId++, start, end, ast.children[2].getText());
        state.nodes.push(id);
        node = new FunctionCall(state.nodeId++, start, end, 'select', [value, id]);
        break
      }
      // Member selection operator `[]`
      case 'select_expr': {
        const args = exprSequence(state, [ast.children[0], ast.children[2]]);
        node = new FunctionCall(state.nodeId++, start, end, 'select', args);
        break
      }
      // Unary operators
      case 'not':
      case 'positive':
      case 'negative': {
        const name = ast.type;
        const args = exprSequence(state, [ast.children[1]]);
        node = new FunctionCall(state.nodeId++, start, end, name, args);
        break
      }
      // Binary operators (in order of precedence)
      case 'pow':
      case 'multiply':
      case 'divide':
      case 'remainder':
      case 'add':
      case 'subtract':
      case 'less':
      case 'less_or_equal':
      case 'greater':
      case 'greater_or_equal':
      case 'equal':
      case 'not_equal':
      case 'and':
      case 'or': {
        const name = ast.type;
        const args = exprSequence(state, [ast.children[0], ast.children[2]]);
        node = new FunctionCall(state.nodeId++, start, end, name, args);
        break
      }
      // Pipe operator
      case 'pipe': {
        node = new PipeOp(state.nodeId++, start, end,
          createFromAST(state, ast.children[0]),
          createFromAST(state, ast.children[2])
        );
        break
      }
      case 'int':
      case 'float':
      case 'number': {
        if (ast.children.length !== 1) {
          node = new ErrorNode(state.nodeId++, start, end, 'Invalid number.');
        } else {
          let token = ast.children[0].children[0];
          state.tokens.push(new Token$3('number-literal', token.symbol));
          node = new NumberNode(state.nodeId++, start, end, Number(token.getText()));
        }
        break
      }
      case 'boolean': {
        let token = ast.children[0];
        state.tokens.push(new Token$3('boolean-literal', token.symbol));
        let bool = (token.getText() === 'true');
        node = new BooleanNode(state.nodeId++, start, end, bool);
        break
      }
      case 'string': {
        let ctx = ast.children[0];
        state.tokens.push(new Token$3('string-literal', ctx.symbol));
        let str = ctx.getText().slice(1, -1);
        node = new StringNode(state.nodeId++, start, end, str);
        break
      }
      case 'array': {
        const ctx = ast.children[0];
        const seq = ctx.children[1];
        let vals = [];
        if (seq && seq.items) {
          vals = exprSequence(state, seq.items);
        }
        node = new ArrayNode(state.nodeId++, start, end, vals);
        break
      }
      case 'object': {
        const ctx = ast.children[0];
        const keys = ctx.keys;
        const vals = ctx.vals;
        const entries = [];
        for (let i = 0; i < keys.length; i++) {
          state.tokens.push(new Token$3('key', keys[i]));
          if (keys[i] && vals[i]) {
            let key = keys[i].text;
            let val = createFromAST(state, vals[i]);
            entries.push({ key, val });
          }
        }
        node = new ObjectNode(state.nodeId++, start, end, entries);
        break
      }
      case 'var': {
        node = new Var(state.nodeId++, start, end, ast.getText());
        state.tokens.push(new Token$3('input-variable-name', {
          start: ast.start.start,
          stop: ast.stop.stop
        }));
        state.inputs.push(node);
        break
      }
      case 'group': {
        // No need to create an extra node for a group expression '(..)'
        return createFromAST(state, ast.children[1])
      }
      case '_call':
        // HACK: sometimes we need to unwrap
        ast = ast.children[0]; // eslint-disable-line no-fallthrough
      case 'call': {
        // ATTENTION we need to be robust regarding partial expressions
        let name = ast.name ? ast.name.text : '';
        let args, namedArgs;
        let argsCtx = ast.args;
        if (argsCtx) {
          args = argSequence(state, argsCtx.args);
          namedArgs = argSequence(state, argsCtx.namedArgs);
        }
        if (ast.name) {
          state.tokens.push(
            new Token$3('function-name', ast.name)
          );
        }
        node = new FunctionCall(state.nodeId++, start, end, name, args, namedArgs);
        break
      }
      case 'named-argument': {
        let name = ast.children[0];
        state.tokens.push(new Token$3('key', name.symbol));
        node = new NamedArgument(state.nodeId++, start, end, name.toString(),
          createFromAST(state, ast.children[2])
        );
        break
      }
      default: {
        if (ast.exception) {
          // console.log('Creating ErrorNode with exception', ast)
          node = new ErrorNode(state.nodeId++, start, end, ast.exception);
        } else {
          // console.log('Creating ErrorNode', ast)
          node = new ErrorNode(state.nodeId++, start, end, 'Parser error.');
        }
      }
    }
    state.nodes.push(node);
    return node
  }

  class Token$3 {
    constructor (type, symbol) {
      /* istanbul ignore next */
      const start = symbol.start;
      const stop = symbol.stop;
      if (!isString$2(type)) {
        throw new Error('Illegal argument: "type" must be a string')
      }
      /* istanbul ignore next */
      if (!isNumber$2(start)) {
        throw new Error('Illegal argument: "start" must be a number')
      }
      /* istanbul ignore next */
      if (!isNumber$2(stop)) {
        throw new Error('Illegal argument: "stop" must be a number')
      }
      this.type = type;
      this.start = start;
      // ATTENTION: seems that symbol.end is inclusive
      this.end = stop + 1;
      this.text = symbol.text;
    }
  }

  function exprSequence (state, items) {
    if (!items) return []
    return items.map(c => createFromAST(state, c))
  }

  function argSequence (state, args) {
    if (!args) return []
    // HACK: we need to allow empty arguments to support incomplete expressions such as in `sum(x,,y)`
    // Still we want to treat it as an error
    let result = [];
    args = args.children;
    let last;
    for (let i = 0; i < args.length; i++) {
      let n = args[i];
      if (n.getText() === ',') {
        if (!last) {
          let [start, end] = _getStartStop(n);
          result.push(new EmptyArgument(state.nodeId++, start, end));
        }
        last = false;
      } else {
        result.push(createFromAST(state, n));
        last = n;
      }
    }
    if (!last) {
      last = args[args.length - 1];
      if (last.getText() === ',') {
        let [start, end] = _getStartStop(last);
        result.push(new EmptyArgument(state.nodeId++, start, end));
      }
    }
    return result
  }

  function _getStartStop (n) {
    if (n.start) {
      if (n.stop) {
        return [n.start.start, n.stop.stop + 1]
      } else {
        return [n.start.start, n.start.stop + 1]
      }
    } else if (n.symbol) {
      return [n.symbol.start, n.symbol.stop + 1]
    } else {
      return [undefined, undefined]
    }
  }

  class Expression$3 extends EventEmitter$2 {
    constructor (source, root, syntaxError, nodes, tokens) {
      super();

      this.source = source;
      this.root = root;

      this.syntaxError = syntaxError;
      this.nodes = nodes;
      this.tokens = tokens;
    }

    get name () {
      if (this.isDefinition()) {
        return this.root.name
      }
    }

    isDefinition () {
      return (this.root && this.root.type === 'definition')
    }

    async evaluate (context) {
      return this.root.evaluate(context)
    }
  }

  Expression$3.create = function (source, ast, syntaxError) {
    let { root, state } = createNodeTree(ast);
    return new Expression$3(source, root, syntaxError, state.nodes, state.tokens)
  };

  let _exports = (function e(t,n,r){function s(o,u){if(!n[o]){if(!t[o]){var a=typeof require=="function"&&require;if(!u&&a)return a(o,!0);if(i)return i(o,!0);var f=new Error("Cannot find module '"+o+"'");throw f.code="MODULE_NOT_FOUND",f}var l=n[o]={exports:{}};t[o][0].call(l.exports,function(e){var n=t[o][1][e];return s(n?n:e)},l,l.exports,e,t,n,r);}return n[o].exports}var i=typeof require=="function"&&require;for(var o=0;o<r.length;o++)s(r[o]);return s})({1:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */

  // This implementation of {@link TokenStream} loads tokens from a
  // {@link TokenSource} on-demand, and places the tokens in a buffer to provide
  // access to any previous token by index.
  //
  // <p>
  // This token stream ignores the value of {@link Token//getChannel}. If your
  // parser requires the token stream filter tokens to only those on a particular
  // channel, such as {@link Token//DEFAULT_CHANNEL} or
  // {@link Token//HIDDEN_CHANNEL}, use a filtering token stream such a
  // {@link CommonTokenStream}.</p>

  var Token = require('./Token').Token;
  var Lexer = require('./Lexer').Lexer;
  var Interval = require('./IntervalSet').Interval;

  // this is just to keep meaningful parameter types to Parser
  function TokenStream() {
  	return this;
  }

  function BufferedTokenStream(tokenSource) {

  	TokenStream.call(this);
  	// The {@link TokenSource} from which tokens for this stream are fetched.
  	this.tokenSource = tokenSource;

  	// A collection of all tokens fetched from the token source. The list is
  	// considered a complete view of the input once {@link //fetchedEOF} is set
  	// to {@code true}.
  	this.tokens = [];

  	// The index into {@link //tokens} of the current token (next token to
  	// {@link //consume}). {@link //tokens}{@code [}{@link //p}{@code ]} should
  	// be
  	// {@link //LT LT(1)}.
  	//
  	// <p>This field is set to -1 when the stream is first constructed or when
  	// {@link //setTokenSource} is called, indicating that the first token has
  	// not yet been fetched from the token source. For additional information,
  	// see the documentation of {@link IntStream} for a description of
  	// Initializing Methods.</p>
  	this.index = -1;

  	// Indicates whether the {@link Token//EOF} token has been fetched from
  	// {@link //tokenSource} and added to {@link //tokens}. This field improves
  	// performance for the following cases:
  	//
  	// <ul>
  	// <li>{@link //consume}: The lookahead check in {@link //consume} to
  	// prevent
  	// consuming the EOF symbol is optimized by checking the values of
  	// {@link //fetchedEOF} and {@link //p} instead of calling {@link
  	// //LA}.</li>
  	// <li>{@link //fetch}: The check to prevent adding multiple EOF symbols
  	// into
  	// {@link //tokens} is trivial with this field.</li>
  	// <ul>
  	this.fetchedEOF = false;
  	return this;
  }

  BufferedTokenStream.prototype = Object.create(TokenStream.prototype);
  BufferedTokenStream.prototype.constructor = BufferedTokenStream;

  BufferedTokenStream.prototype.mark = function() {
  	return 0;
  };

  BufferedTokenStream.prototype.release = function(marker) {
  	// no resources to release
  };

  BufferedTokenStream.prototype.reset = function() {
  	this.seek(0);
  };

  BufferedTokenStream.prototype.seek = function(index) {
  	this.lazyInit();
  	this.index = this.adjustSeekIndex(index);
  };

  BufferedTokenStream.prototype.get = function(index) {
  	this.lazyInit();
  	return this.tokens[index];
  };

  BufferedTokenStream.prototype.consume = function() {
  	var skipEofCheck = false;
  	if (this.index >= 0) {
  		if (this.fetchedEOF) {
  			// the last token in tokens is EOF. skip check if p indexes any
  			// fetched token except the last.
  			skipEofCheck = this.index < this.tokens.length - 1;
  		} else {
  			// no EOF token in tokens. skip check if p indexes a fetched token.
  			skipEofCheck = this.index < this.tokens.length;
  		}
  	} else {
  		// not yet initialized
  		skipEofCheck = false;
  	}
  	if (!skipEofCheck && this.LA(1) === Token.EOF) {
  		throw "cannot consume EOF";
  	}
  	if (this.sync(this.index + 1)) {
  		this.index = this.adjustSeekIndex(this.index + 1);
  	}
  };

  // Make sure index {@code i} in tokens has a token.
  //
  // @return {@code true} if a token is located at index {@code i}, otherwise
  // {@code false}.
  // @see //get(int i)
  // /
  BufferedTokenStream.prototype.sync = function(i) {
  	var n = i - this.tokens.length + 1; // how many more elements we need?
  	if (n > 0) {
  		var fetched = this.fetch(n);
  		return fetched >= n;
  	}
  	return true;
  };

  // Add {@code n} elements to buffer.
  //
  // @return The actual number of elements added to the buffer.
  // /
  BufferedTokenStream.prototype.fetch = function(n) {
  	if (this.fetchedEOF) {
  		return 0;
  	}
  	for (var i = 0; i < n; i++) {
  		var t = this.tokenSource.nextToken();
  		t.tokenIndex = this.tokens.length;
  		this.tokens.push(t);
  		if (t.type === Token.EOF) {
  			this.fetchedEOF = true;
  			return i + 1;
  		}
  	}
  	return n;
  };

  // Get all tokens from start..stop inclusively///
  BufferedTokenStream.prototype.getTokens = function(start, stop, types) {
  	if (types === undefined) {
  		types = null;
  	}
  	if (start < 0 || stop < 0) {
  		return null;
  	}
  	this.lazyInit();
  	var subset = [];
  	if (stop >= this.tokens.length) {
  		stop = this.tokens.length - 1;
  	}
  	for (var i = start; i < stop; i++) {
  		var t = this.tokens[i];
  		if (t.type === Token.EOF) {
  			break;
  		}
  		if (types === null || types.contains(t.type)) {
  			subset.push(t);
  		}
  	}
  	return subset;
  };

  BufferedTokenStream.prototype.LA = function(i) {
  	return this.LT(i).type;
  };

  BufferedTokenStream.prototype.LB = function(k) {
  	if (this.index - k < 0) {
  		return null;
  	}
  	return this.tokens[this.index - k];
  };

  BufferedTokenStream.prototype.LT = function(k) {
  	this.lazyInit();
  	if (k === 0) {
  		return null;
  	}
  	if (k < 0) {
  		return this.LB(-k);
  	}
  	var i = this.index + k - 1;
  	this.sync(i);
  	if (i >= this.tokens.length) { // return EOF token
  		// EOF must be last token
  		return this.tokens[this.tokens.length - 1];
  	}
  	return this.tokens[i];
  };

  // Allowed derived classes to modify the behavior of operations which change
  // the current stream position by adjusting the target token index of a seek
  // operation. The default implementation simply returns {@code i}. If an
  // exception is thrown in this method, the current stream index should not be
  // changed.
  //
  // <p>For example, {@link CommonTokenStream} overrides this method to ensure
  // that
  // the seek target is always an on-channel token.</p>
  //
  // @param i The target token index.
  // @return The adjusted target token index.

  BufferedTokenStream.prototype.adjustSeekIndex = function(i) {
  	return i;
  };

  BufferedTokenStream.prototype.lazyInit = function() {
  	if (this.index === -1) {
  		this.setup();
  	}
  };

  BufferedTokenStream.prototype.setup = function() {
  	this.sync(0);
  	this.index = this.adjustSeekIndex(0);
  };

  // Reset this token stream by setting its token source.///
  BufferedTokenStream.prototype.setTokenSource = function(tokenSource) {
  	this.tokenSource = tokenSource;
  	this.tokens = [];
  	this.index = -1;
  	this.fetchedEOF = false;
  };


  // Given a starting index, return the index of the next token on channel.
  // Return i if tokens[i] is on channel. Return -1 if there are no tokens
  // on channel between i and EOF.
  // /
  BufferedTokenStream.prototype.nextTokenOnChannel = function(i, channel) {
  	this.sync(i);
  	if (i >= this.tokens.length) {
  		return -1;
  	}
  	var token = this.tokens[i];
  	while (token.channel !== this.channel) {
  		if (token.type === Token.EOF) {
  			return -1;
  		}
  		i += 1;
  		this.sync(i);
  		token = this.tokens[i];
  	}
  	return i;
  };

  // Given a starting index, return the index of the previous token on channel.
  // Return i if tokens[i] is on channel. Return -1 if there are no tokens
  // on channel between i and 0.
  BufferedTokenStream.prototype.previousTokenOnChannel = function(i, channel) {
  	while (i >= 0 && this.tokens[i].channel !== channel) {
  		i -= 1;
  	}
  	return i;
  };

  // Collect all tokens on specified channel to the right of
  // the current token up until we see a token on DEFAULT_TOKEN_CHANNEL or
  // EOF. If channel is -1, find any non default channel token.
  BufferedTokenStream.prototype.getHiddenTokensToRight = function(tokenIndex,
  		channel) {
  	if (channel === undefined) {
  		channel = -1;
  	}
  	this.lazyInit();
  	if (tokenIndex < 0 || tokenIndex >= this.tokens.length) {
  		throw "" + tokenIndex + " not in 0.." + this.tokens.length - 1;
  	}
  	var nextOnChannel = this.nextTokenOnChannel(tokenIndex + 1, Lexer.DEFAULT_TOKEN_CHANNEL);
  	var from_ = tokenIndex + 1;
  	// if none onchannel to right, nextOnChannel=-1 so set to = last token
  	var to = nextOnChannel === -1 ? this.tokens.length - 1 : nextOnChannel;
  	return this.filterForChannel(from_, to, channel);
  };

  // Collect all tokens on specified channel to the left of
  // the current token up until we see a token on DEFAULT_TOKEN_CHANNEL.
  // If channel is -1, find any non default channel token.
  BufferedTokenStream.prototype.getHiddenTokensToLeft = function(tokenIndex,
  		channel) {
  	if (channel === undefined) {
  		channel = -1;
  	}
  	this.lazyInit();
  	if (tokenIndex < 0 || tokenIndex >= this.tokens.length) {
  		throw "" + tokenIndex + " not in 0.." + this.tokens.length - 1;
  	}
  	var prevOnChannel = this.previousTokenOnChannel(tokenIndex - 1, Lexer.DEFAULT_TOKEN_CHANNEL);
  	if (prevOnChannel === tokenIndex - 1) {
  		return null;
  	}
  	// if none on channel to left, prevOnChannel=-1 then from=0
  	var from_ = prevOnChannel + 1;
  	var to = tokenIndex - 1;
  	return this.filterForChannel(from_, to, channel);
  };

  BufferedTokenStream.prototype.filterForChannel = function(left, right, channel) {
  	var hidden = [];
  	for (var i = left; i < right + 1; i++) {
  		var t = this.tokens[i];
  		if (channel === -1) {
  			if (t.channel !== Lexer.DEFAULT_TOKEN_CHANNEL) {
  				hidden.push(t);
  			}
  		} else if (t.channel === channel) {
  			hidden.push(t);
  		}
  	}
  	if (hidden.length === 0) {
  		return null;
  	}
  	return hidden;
  };

  BufferedTokenStream.prototype.getSourceName = function() {
  	return this.tokenSource.getSourceName();
  };

  // Get the text of all tokens in this buffer.///
  BufferedTokenStream.prototype.getText = function(interval) {
  	this.lazyInit();
  	this.fill();
  	if (interval === undefined || interval === null) {
  		interval = new Interval(0, this.tokens.length - 1);
  	}
  	var start = interval.start;
  	if (start instanceof Token) {
  		start = start.tokenIndex;
  	}
  	var stop = interval.stop;
  	if (stop instanceof Token) {
  		stop = stop.tokenIndex;
  	}
  	if (start === null || stop === null || start < 0 || stop < 0) {
  		return "";
  	}
  	if (stop >= this.tokens.length) {
  		stop = this.tokens.length - 1;
  	}
  	var s = "";
  	for (var i = start; i < stop + 1; i++) {
  		var t = this.tokens[i];
  		if (t.type === Token.EOF) {
  			break;
  		}
  		s = s + t.text;
  	}
  	return s;
  };

  // Get all tokens from lexer until EOF///
  BufferedTokenStream.prototype.fill = function() {
  	this.lazyInit();
  	while (this.fetch(1000) === 1000) {
  		continue;
  	}
  };

  exports.BufferedTokenStream = BufferedTokenStream;

  },{"./IntervalSet":6,"./Lexer":8,"./Token":14}],2:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  //

  //
  // This default implementation of {@link TokenFactory} creates
  // {@link CommonToken} objects.
  //

  var CommonToken = require('./Token').CommonToken;

  function TokenFactory() {
  	return this;
  }

  function CommonTokenFactory(copyText) {
  	TokenFactory.call(this);
      // Indicates whether {@link CommonToken//setText} should be called after
      // constructing tokens to explicitly set the text. This is useful for cases
      // where the input stream might not be able to provide arbitrary substrings
      // of text from the input after the lexer creates a token (e.g. the
      // implementation of {@link CharStream//getText} in
      // {@link UnbufferedCharStream} throws an
      // {@link UnsupportedOperationException}). Explicitly setting the token text
      // allows {@link Token//getText} to be called at any time regardless of the
      // input stream implementation.
      //
      // <p>
      // The default value is {@code false} to avoid the performance and memory
      // overhead of copying text for every token unless explicitly requested.</p>
      //
      this.copyText = copyText===undefined ? false : copyText;
  	return this;
  }

  CommonTokenFactory.prototype = Object.create(TokenFactory.prototype);
  CommonTokenFactory.prototype.constructor = CommonTokenFactory;

  //
  // The default {@link CommonTokenFactory} instance.
  //
  // <p>
  // This token factory does not explicitly copy token text when constructing
  // tokens.</p>
  //
  CommonTokenFactory.DEFAULT = new CommonTokenFactory();

  CommonTokenFactory.prototype.create = function(source, type, text, channel, start, stop, line, column) {
      var t = new CommonToken(source, type, channel, start, stop);
      t.line = line;
      t.column = column;
      if (text !==null) {
          t.text = text;
      } else if (this.copyText && source[1] !==null) {
          t.text = source[1].getText(start,stop);
      }
      return t;
  };

  CommonTokenFactory.prototype.createThin = function(type, text) {
      var t = new CommonToken(null, type);
      t.text = text;
      return t;
  };

  exports.CommonTokenFactory = CommonTokenFactory;

  },{"./Token":14}],3:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  ///

  //
  // This class extends {@link BufferedTokenStream} with functionality to filter
  // token streams to tokens on a particular channel (tokens where
  // {@link Token//getChannel} returns a particular value).
  //
  // <p>
  // This token stream provides access to all tokens by index or when calling
  // methods like {@link //getText}. The channel filtering is only used for code
  // accessing tokens via the lookahead methods {@link //LA}, {@link //LT}, and
  // {@link //LB}.</p>
  //
  // <p>
  // By default, tokens are placed on the default channel
  // ({@link Token//DEFAULT_CHANNEL}), but may be reassigned by using the
  // {@code ->channel(HIDDEN)} lexer command, or by using an embedded action to
  // call {@link Lexer//setChannel}.
  // </p>
  //
  // <p>
  // Note: lexer rules which use the {@code ->skip} lexer command or call
  // {@link Lexer//skip} do not produce tokens at all, so input text matched by
  // such a rule will not be available as part of the token stream, regardless of
  // channel.</p>
  ///

  var Token = require('./Token').Token;
  var BufferedTokenStream = require('./BufferedTokenStream').BufferedTokenStream;

  function CommonTokenStream(lexer, channel) {
  	BufferedTokenStream.call(this, lexer);
      this.channel = channel===undefined ? Token.DEFAULT_CHANNEL : channel;
      return this;
  }

  CommonTokenStream.prototype = Object.create(BufferedTokenStream.prototype);
  CommonTokenStream.prototype.constructor = CommonTokenStream;

  CommonTokenStream.prototype.adjustSeekIndex = function(i) {
      return this.nextTokenOnChannel(i, this.channel);
  };

  CommonTokenStream.prototype.LB = function(k) {
      if (k===0 || this.index-k<0) {
          return null;
      }
      var i = this.index;
      var n = 1;
      // find k good tokens looking backwards
      while (n <= k) {
          // skip off-channel tokens
          i = this.previousTokenOnChannel(i - 1, this.channel);
          n += 1;
      }
      if (i < 0) {
          return null;
      }
      return this.tokens[i];
  };

  CommonTokenStream.prototype.LT = function(k) {
      this.lazyInit();
      if (k === 0) {
          return null;
      }
      if (k < 0) {
          return this.LB(-k);
      }
      var i = this.index;
      var n = 1; // we know tokens[pos] is a good one
      // find k good tokens
      while (n < k) {
          // skip off-channel tokens, but make sure to not look past EOF
          if (this.sync(i + 1)) {
              i = this.nextTokenOnChannel(i + 1, this.channel);
          }
          n += 1;
      }
      return this.tokens[i];
  };

  // Count EOF just once.///
  CommonTokenStream.prototype.getNumberOfOnChannelTokens = function() {
      var n = 0;
      this.fill();
      for (var i =0; i< this.tokens.length;i++) {
          var t = this.tokens[i];
          if( t.channel===this.channel) {
              n += 1;
          }
          if( t.type===Token.EOF) {
              break;
          }
      }
      return n;
  };

  exports.CommonTokenStream = CommonTokenStream;
  },{"./BufferedTokenStream":1,"./Token":14}],4:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  //

  //
  //  This is an InputStream that is loaded from a file all at once
  //  when you construct the object.
  //
  var InputStream = require('./InputStream').InputStream;
  var isNodeJs = typeof window === 'undefined' && typeof importScripts === 'undefined';
  var fs = isNodeJs ? require("fs") : null;

  function FileStream(fileName, decodeToUnicodeCodePoints) {
  	var data = fs.readFileSync(fileName, "utf8");
  	InputStream.call(this, data, decodeToUnicodeCodePoints);
  	this.fileName = fileName;
  	return this;
  }

  FileStream.prototype = Object.create(InputStream.prototype);
  FileStream.prototype.constructor = FileStream;

  exports.FileStream = FileStream;

  },{"./InputStream":5,"fs":47}],5:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  //

  var Token = require('./Token').Token;
  require('./polyfills/codepointat');
  require('./polyfills/fromcodepoint');

  // Vacuum all input from a string and then treat it like a buffer.

  function _loadString(stream, decodeToUnicodeCodePoints) {
  	stream._index = 0;
  	stream.data = [];
  	if (stream.decodeToUnicodeCodePoints) {
  		for (var i = 0; i < stream.strdata.length; ) {
  			var codePoint = stream.strdata.codePointAt(i);
  			stream.data.push(codePoint);
  			i += codePoint <= 0xFFFF ? 1 : 2;
  		}
  	} else {
  		for (var i = 0; i < stream.strdata.length; i++) {
  			var codeUnit = stream.strdata.charCodeAt(i);
  			stream.data.push(codeUnit);
  		}
  	}
  	stream._size = stream.data.length;
  }

  // If decodeToUnicodeCodePoints is true, the input is treated
  // as a series of Unicode code points.
  //
  // Otherwise, the input is treated as a series of 16-bit UTF-16 code
  // units.
  function InputStream(data, decodeToUnicodeCodePoints) {
  	this.name = "<empty>";
  	this.strdata = data;
  	this.decodeToUnicodeCodePoints = decodeToUnicodeCodePoints || false;
  	_loadString(this);
  	return this;
  }

  Object.defineProperty(InputStream.prototype, "index", {
  	get : function() {
  		return this._index;
  	}
  });

  Object.defineProperty(InputStream.prototype, "size", {
  	get : function() {
  		return this._size;
  	}
  });

  // Reset the stream so that it's in the same state it was
  // when the object was created *except* the data array is not
  // touched.
  //
  InputStream.prototype.reset = function() {
  	this._index = 0;
  };

  InputStream.prototype.consume = function() {
  	if (this._index >= this._size) {
  		// assert this.LA(1) == Token.EOF
  		throw ("cannot consume EOF");
  	}
  	this._index += 1;
  };

  InputStream.prototype.LA = function(offset) {
  	if (offset === 0) {
  		return 0; // undefined
  	}
  	if (offset < 0) {
  		offset += 1; // e.g., translate LA(-1) to use offset=0
  	}
  	var pos = this._index + offset - 1;
  	if (pos < 0 || pos >= this._size) { // invalid
  		return Token.EOF;
  	}
  	return this.data[pos];
  };

  InputStream.prototype.LT = function(offset) {
  	return this.LA(offset);
  };

  // mark/release do nothing; we have entire buffer
  InputStream.prototype.mark = function() {
  	return -1;
  };

  InputStream.prototype.release = function(marker) {
  };

  // consume() ahead until p==_index; can't just set p=_index as we must
  // update line and column. If we seek backwards, just set p
  //
  InputStream.prototype.seek = function(_index) {
  	if (_index <= this._index) {
  		this._index = _index; // just jump; don't update stream state (line,
  								// ...)
  		return;
  	}
  	// seek forward
  	this._index = Math.min(_index, this._size);
  };

  InputStream.prototype.getText = function(start, stop) {
  	if (stop >= this._size) {
  		stop = this._size - 1;
  	}
  	if (start >= this._size) {
  		return "";
  	} else {
  		if (this.decodeToUnicodeCodePoints) {
  			var result = "";
  			for (var i = start; i <= stop; i++) {
  				result += String.fromCodePoint(this.data[i]);
  			}
  			return result;
  		} else {
  			return this.strdata.slice(start, stop + 1);
  		}
  	}
  };

  InputStream.prototype.toString = function() {
  	return this.strdata;
  };

  exports.InputStream = InputStream;

  },{"./Token":14,"./polyfills/codepointat":42,"./polyfills/fromcodepoint":43}],6:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */

  /*jslint smarttabs:true */

  var Token = require('./Token').Token;

  /* stop is not included! */
  function Interval(start, stop) {
  	this.start = start;
  	this.stop = stop;
  	return this;
  }

  Interval.prototype.contains = function(item) {
  	return item >= this.start && item < this.stop;
  };

  Interval.prototype.toString = function() {
  	if(this.start===this.stop-1) {
  		return this.start.toString();
  	} else {
  		return this.start.toString() + ".." + (this.stop-1).toString();
  	}
  };


  Object.defineProperty(Interval.prototype, "length", {
  	get : function() {
  		return this.stop - this.start;
  	}
  });

  function IntervalSet() {
  	this.intervals = null;
  	this.readOnly = false;
  }

  IntervalSet.prototype.first = function(v) {
  	if (this.intervals === null || this.intervals.length===0) {
  		return Token.INVALID_TYPE;
  	} else {
  		return this.intervals[0].start;
  	}
  };

  IntervalSet.prototype.addOne = function(v) {
  	this.addInterval(new Interval(v, v + 1));
  };

  IntervalSet.prototype.addRange = function(l, h) {
  	this.addInterval(new Interval(l, h + 1));
  };

  IntervalSet.prototype.addInterval = function(v) {
  	if (this.intervals === null) {
  		this.intervals = [];
  		this.intervals.push(v);
  	} else {
  		// find insert pos
  		for (var k = 0; k < this.intervals.length; k++) {
  			var i = this.intervals[k];
  			// distinct range -> insert
  			if (v.stop < i.start) {
  				this.intervals.splice(k, 0, v);
  				return;
  			}
  			// contiguous range -> adjust
  			else if (v.stop === i.start) {
  				this.intervals[k].start = v.start;
  				return;
  			}
  			// overlapping range -> adjust and reduce
  			else if (v.start <= i.stop) {
  				this.intervals[k] = new Interval(Math.min(i.start, v.start), Math.max(i.stop, v.stop));
  				this.reduce(k);
  				return;
  			}
  		}
  		// greater than any existing
  		this.intervals.push(v);
  	}
  };

  IntervalSet.prototype.addSet = function(other) {
  	if (other.intervals !== null) {
  		for (var k = 0; k < other.intervals.length; k++) {
  			var i = other.intervals[k];
  			this.addInterval(new Interval(i.start, i.stop));
  		}
  	}
  	return this;
  };

  IntervalSet.prototype.reduce = function(k) {
  	// only need to reduce if k is not the last
  	if (k < this.intervalslength - 1) {
  		var l = this.intervals[k];
  		var r = this.intervals[k + 1];
  		// if r contained in l
  		if (l.stop >= r.stop) {
  			this.intervals.pop(k + 1);
  			this.reduce(k);
  		} else if (l.stop >= r.start) {
  			this.intervals[k] = new Interval(l.start, r.stop);
  			this.intervals.pop(k + 1);
  		}
  	}
  };

  IntervalSet.prototype.complement = function(start, stop) {
      var result = new IntervalSet();
      result.addInterval(new Interval(start,stop+1));
      for(var i=0; i<this.intervals.length; i++) {
          result.removeRange(this.intervals[i]);
      }
      return result;
  };

  IntervalSet.prototype.contains = function(item) {
  	if (this.intervals === null) {
  		return false;
  	} else {
  		for (var k = 0; k < this.intervals.length; k++) {
  			if(this.intervals[k].contains(item)) {
  				return true;
  			}
  		}
  		return false;
  	}
  };

  Object.defineProperty(IntervalSet.prototype, "length", {
  	get : function() {
  		var len = 0;
  		this.intervals.map(function(i) {len += i.length;});
  		return len;
  	}
  });

  IntervalSet.prototype.removeRange = function(v) {
      if(v.start===v.stop-1) {
          this.removeOne(v.start);
      } else if (this.intervals!==null) {
          var k = 0;
          for(var n=0; n<this.intervals.length; n++) {
              var i = this.intervals[k];
              // intervals are ordered
              if (v.stop<=i.start) {
                  return;
              }
              // check for including range, split it
              else if(v.start>i.start && v.stop<i.stop) {
                  this.intervals[k] = new Interval(i.start, v.start);
                  var x = new Interval(v.stop, i.stop);
                  this.intervals.splice(k, 0, x);
                  return;
              }
              // check for included range, remove it
              else if(v.start<=i.start && v.stop>=i.stop) {
                  this.intervals.splice(k, 1);
                  k = k - 1; // need another pass
              }
              // check for lower boundary
              else if(v.start<i.stop) {
                  this.intervals[k] = new Interval(i.start, v.start);
              }
              // check for upper boundary
              else if(v.stop<i.stop) {
                  this.intervals[k] = new Interval(v.stop, i.stop);
              }
              k += 1;
          }
      }
  };

  IntervalSet.prototype.removeOne = function(v) {
  	if (this.intervals !== null) {
  		for (var k = 0; k < this.intervals.length; k++) {
  			var i = this.intervals[k];
  			// intervals is ordered
  			if (v < i.start) {
  				return;
  			}
  			// check for single value range
  			else if (v === i.start && v === i.stop - 1) {
  				this.intervals.splice(k, 1);
  				return;
  			}
  			// check for lower boundary
  			else if (v === i.start) {
  				this.intervals[k] = new Interval(i.start + 1, i.stop);
  				return;
  			}
  			// check for upper boundary
  			else if (v === i.stop - 1) {
  				this.intervals[k] = new Interval(i.start, i.stop - 1);
  				return;
  			}
  			// split existing range
  			else if (v < i.stop - 1) {
  				var x = new Interval(i.start, v);
  				i.start = v + 1;
  				this.intervals.splice(k, 0, x);
  				return;
  			}
  		}
  	}
  };

  IntervalSet.prototype.toString = function(literalNames, symbolicNames, elemsAreChar) {
  	literalNames = literalNames || null;
  	symbolicNames = symbolicNames || null;
  	elemsAreChar = elemsAreChar || false;
  	if (this.intervals === null) {
  		return "{}";
  	} else if(literalNames!==null || symbolicNames!==null) {
  		return this.toTokenString(literalNames, symbolicNames);
  	} else if(elemsAreChar) {
  		return this.toCharString();
  	} else {
  		return this.toIndexString();
  	}
  };

  IntervalSet.prototype.toCharString = function() {
  	var names = [];
  	for (var i = 0; i < this.intervals.length; i++) {
  		var v = this.intervals[i];
  		if(v.stop===v.start+1) {
  			if ( v.start===Token.EOF ) {
  				names.push("<EOF>");
  			} else {
  				names.push("'" + String.fromCharCode(v.start) + "'");
  			}
  		} else {
  			names.push("'" + String.fromCharCode(v.start) + "'..'" + String.fromCharCode(v.stop-1) + "'");
  		}
  	}
  	if (names.length > 1) {
  		return "{" + names.join(", ") + "}";
  	} else {
  		return names[0];
  	}
  };


  IntervalSet.prototype.toIndexString = function() {
  	var names = [];
  	for (var i = 0; i < this.intervals.length; i++) {
  		var v = this.intervals[i];
  		if(v.stop===v.start+1) {
  			if ( v.start===Token.EOF ) {
  				names.push("<EOF>");
  			} else {
  				names.push(v.start.toString());
  			}
  		} else {
  			names.push(v.start.toString() + ".." + (v.stop-1).toString());
  		}
  	}
  	if (names.length > 1) {
  		return "{" + names.join(", ") + "}";
  	} else {
  		return names[0];
  	}
  };


  IntervalSet.prototype.toTokenString = function(literalNames, symbolicNames) {
  	var names = [];
  	for (var i = 0; i < this.intervals.length; i++) {
  		var v = this.intervals[i];
  		for (var j = v.start; j < v.stop; j++) {
  			names.push(this.elementName(literalNames, symbolicNames, j));
  		}
  	}
  	if (names.length > 1) {
  		return "{" + names.join(", ") + "}";
  	} else {
  		return names[0];
  	}
  };

  IntervalSet.prototype.elementName = function(literalNames, symbolicNames, a) {
  	if (a === Token.EOF) {
  		return "<EOF>";
  	} else if (a === Token.EPSILON) {
  		return "<EPSILON>";
  	} else {
  		return literalNames[a] || symbolicNames[a];
  	}
  };

  exports.Interval = Interval;
  exports.IntervalSet = IntervalSet;

  },{"./Token":14}],7:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  ///

  var Set = require('./Utils').Set;
  var BitSet = require('./Utils').BitSet;
  var Token = require('./Token').Token;
  var ATNConfig = require('./atn/ATNConfig').ATNConfig;
  var Interval = require('./IntervalSet').Interval;
  var IntervalSet = require('./IntervalSet').IntervalSet;
  var RuleStopState = require('./atn/ATNState').RuleStopState;
  var RuleTransition = require('./atn/Transition').RuleTransition;
  var NotSetTransition = require('./atn/Transition').NotSetTransition;
  var WildcardTransition = require('./atn/Transition').WildcardTransition;
  var AbstractPredicateTransition = require('./atn/Transition').AbstractPredicateTransition;

  var pc = require('./PredictionContext');
  var predictionContextFromRuleContext = pc.predictionContextFromRuleContext;
  var PredictionContext = pc.PredictionContext;
  var SingletonPredictionContext = pc.SingletonPredictionContext;

  function LL1Analyzer (atn) {
      this.atn = atn;
  }

  //* Special value added to the lookahead sets to indicate that we hit
  //  a predicate during analysis if {@code seeThruPreds==false}.
  ///
  LL1Analyzer.HIT_PRED = Token.INVALID_TYPE;


  //*
  // Calculates the SLL(1) expected lookahead set for each outgoing transition
  // of an {@link ATNState}. The returned array has one element for each
  // outgoing transition in {@code s}. If the closure from transition
  // <em>i</em> leads to a semantic predicate before matching a symbol, the
  // element at index <em>i</em> of the result will be {@code null}.
  //
  // @param s the ATN state
  // @return the expected symbols for each outgoing transition of {@code s}.
  ///
  LL1Analyzer.prototype.getDecisionLookahead = function(s) {
      if (s === null) {
          return null;
      }
      var count = s.transitions.length;
      var look = [];
      for(var alt=0; alt< count; alt++) {
          look[alt] = new IntervalSet();
          var lookBusy = new Set();
          var seeThruPreds = false; // fail to get lookahead upon pred
          this._LOOK(s.transition(alt).target, null, PredictionContext.EMPTY,
                look[alt], lookBusy, new BitSet(), seeThruPreds, false);
          // Wipe out lookahead for this alternative if we found nothing
          // or we had a predicate when we !seeThruPreds
          if (look[alt].length===0 || look[alt].contains(LL1Analyzer.HIT_PRED)) {
              look[alt] = null;
          }
      }
      return look;
  };

  //*
  // Compute set of tokens that can follow {@code s} in the ATN in the
  // specified {@code ctx}.
  //
  // <p>If {@code ctx} is {@code null} and the end of the rule containing
  // {@code s} is reached, {@link Token//EPSILON} is added to the result set.
  // If {@code ctx} is not {@code null} and the end of the outermost rule is
  // reached, {@link Token//EOF} is added to the result set.</p>
  //
  // @param s the ATN state
  // @param stopState the ATN state to stop at. This can be a
  // {@link BlockEndState} to detect epsilon paths through a closure.
  // @param ctx the complete parser context, or {@code null} if the context
  // should be ignored
  //
  // @return The set of tokens that can follow {@code s} in the ATN in the
  // specified {@code ctx}.
  ///
  LL1Analyzer.prototype.LOOK = function(s, stopState, ctx) {
      var r = new IntervalSet();
      var seeThruPreds = true; // ignore preds; get all lookahead
  	ctx = ctx || null;
      var lookContext = ctx!==null ? predictionContextFromRuleContext(s.atn, ctx) : null;
      this._LOOK(s, stopState, lookContext, r, new Set(), new BitSet(), seeThruPreds, true);
      return r;
  };

  //*
  // Compute set of tokens that can follow {@code s} in the ATN in the
  // specified {@code ctx}.
  //
  // <p>If {@code ctx} is {@code null} and {@code stopState} or the end of the
  // rule containing {@code s} is reached, {@link Token//EPSILON} is added to
  // the result set. If {@code ctx} is not {@code null} and {@code addEOF} is
  // {@code true} and {@code stopState} or the end of the outermost rule is
  // reached, {@link Token//EOF} is added to the result set.</p>
  //
  // @param s the ATN state.
  // @param stopState the ATN state to stop at. This can be a
  // {@link BlockEndState} to detect epsilon paths through a closure.
  // @param ctx The outer context, or {@code null} if the outer context should
  // not be used.
  // @param look The result lookahead set.
  // @param lookBusy A set used for preventing epsilon closures in the ATN
  // from causing a stack overflow. Outside code should pass
  // {@code new Set<ATNConfig>} for this argument.
  // @param calledRuleStack A set used for preventing left recursion in the
  // ATN from causing a stack overflow. Outside code should pass
  // {@code new BitSet()} for this argument.
  // @param seeThruPreds {@code true} to true semantic predicates as
  // implicitly {@code true} and "see through them", otherwise {@code false}
  // to treat semantic predicates as opaque and add {@link //HIT_PRED} to the
  // result if one is encountered.
  // @param addEOF Add {@link Token//EOF} to the result if the end of the
  // outermost context is reached. This parameter has no effect if {@code ctx}
  // is {@code null}.
  ///
  LL1Analyzer.prototype._LOOK = function(s, stopState , ctx, look, lookBusy, calledRuleStack, seeThruPreds, addEOF) {
      var c = new ATNConfig({state:s, alt:0, context: ctx}, null);
      if (lookBusy.contains(c)) {
          return;
      }
      lookBusy.add(c);
      if (s === stopState) {
          if (ctx ===null) {
              look.addOne(Token.EPSILON);
              return;
          } else if (ctx.isEmpty() && addEOF) {
              look.addOne(Token.EOF);
              return;
          }
      }
      if (s instanceof RuleStopState ) {
          if (ctx ===null) {
              look.addOne(Token.EPSILON);
              return;
          } else if (ctx.isEmpty() && addEOF) {
              look.addOne(Token.EOF);
              return;
          }
          if (ctx !== PredictionContext.EMPTY) {
              // run thru all possible stack tops in ctx
              for(var i=0; i<ctx.length; i++) {
                  var returnState = this.atn.states[ctx.getReturnState(i)];
                  var removed = calledRuleStack.contains(returnState.ruleIndex);
                  try {
                      calledRuleStack.remove(returnState.ruleIndex);
                      this._LOOK(returnState, stopState, ctx.getParent(i), look, lookBusy, calledRuleStack, seeThruPreds, addEOF);
                  } finally {
                      if (removed) {
                          calledRuleStack.add(returnState.ruleIndex);
                      }
                  }
              }
              return;
          }
      }
      for(var j=0; j<s.transitions.length; j++) {
          var t = s.transitions[j];
          if (t.constructor === RuleTransition) {
              if (calledRuleStack.contains(t.target.ruleIndex)) {
                  continue;
              }
              var newContext = SingletonPredictionContext.create(ctx, t.followState.stateNumber);
              try {
                  calledRuleStack.add(t.target.ruleIndex);
                  this._LOOK(t.target, stopState, newContext, look, lookBusy, calledRuleStack, seeThruPreds, addEOF);
              } finally {
                  calledRuleStack.remove(t.target.ruleIndex);
              }
          } else if (t instanceof AbstractPredicateTransition ) {
              if (seeThruPreds) {
                  this._LOOK(t.target, stopState, ctx, look, lookBusy, calledRuleStack, seeThruPreds, addEOF);
              } else {
                  look.addOne(LL1Analyzer.HIT_PRED);
              }
          } else if( t.isEpsilon) {
              this._LOOK(t.target, stopState, ctx, look, lookBusy, calledRuleStack, seeThruPreds, addEOF);
          } else if (t.constructor === WildcardTransition) {
              look.addRange( Token.MIN_USER_TOKEN_TYPE, this.atn.maxTokenType );
          } else {
              var set = t.label;
              if (set !== null) {
                  if (t instanceof NotSetTransition) {
                      set = set.complement(Token.MIN_USER_TOKEN_TYPE, this.atn.maxTokenType);
                  }
                  look.addSet(set);
              }
          }
      }
  };

  exports.LL1Analyzer = LL1Analyzer;


  },{"./IntervalSet":6,"./PredictionContext":11,"./Token":14,"./Utils":15,"./atn/ATNConfig":17,"./atn/ATNState":22,"./atn/Transition":30}],8:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  ///

  // A lexer is recognizer that draws input symbols from a character stream.
  //  lexer grammars result in a subclass of this object. A Lexer object
  //  uses simplified match() and error recovery mechanisms in the interest of speed.

  var Token = require('./Token').Token;
  var Recognizer = require('./Recognizer').Recognizer;
  var CommonTokenFactory = require('./CommonTokenFactory').CommonTokenFactory;
  var RecognitionException  = require('./error/Errors').RecognitionException;
  var LexerNoViableAltException = require('./error/Errors').LexerNoViableAltException;

  function Lexer(input) {
  	Recognizer.call(this);
  	this._input = input;
  	this._factory = CommonTokenFactory.DEFAULT;
  	this._tokenFactorySourcePair = [ this, input ];

  	this._interp = null; // child classes must populate this

  	// The goal of all lexer rules/methods is to create a token object.
  	// this is an instance variable as multiple rules may collaborate to
  	// create a single token. nextToken will return this object after
  	// matching lexer rule(s). If you subclass to allow multiple token
  	// emissions, then set this to the last token to be matched or
  	// something nonnull so that the auto token emit mechanism will not
  	// emit another token.
  	this._token = null;

  	// What character index in the stream did the current token start at?
  	// Needed, for example, to get the text for current token. Set at
  	// the start of nextToken.
  	this._tokenStartCharIndex = -1;

  	// The line on which the first character of the token resides///
  	this._tokenStartLine = -1;

  	// The character position of first character within the line///
  	this._tokenStartColumn = -1;

  	// Once we see EOF on char stream, next token will be EOF.
  	// If you have DONE : EOF ; then you see DONE EOF.
  	this._hitEOF = false;

  	// The channel number for the current token///
  	this._channel = Token.DEFAULT_CHANNEL;

  	// The token type for the current token///
  	this._type = Token.INVALID_TYPE;

  	this._modeStack = [];
  	this._mode = Lexer.DEFAULT_MODE;

  	// You can set the text for the current token to override what is in
  	// the input char buffer. Use setText() or can set this instance var.
  	// /
  	this._text = null;

  	return this;
  }

  Lexer.prototype = Object.create(Recognizer.prototype);
  Lexer.prototype.constructor = Lexer;

  Lexer.DEFAULT_MODE = 0;
  Lexer.MORE = -2;
  Lexer.SKIP = -3;

  Lexer.DEFAULT_TOKEN_CHANNEL = Token.DEFAULT_CHANNEL;
  Lexer.HIDDEN = Token.HIDDEN_CHANNEL;
  Lexer.MIN_CHAR_VALUE = 0x0000;
  Lexer.MAX_CHAR_VALUE = 0x10FFFF;

  Lexer.prototype.reset = function() {
  	// wack Lexer state variables
  	if (this._input !== null) {
  		this._input.seek(0); // rewind the input
  	}
  	this._token = null;
  	this._type = Token.INVALID_TYPE;
  	this._channel = Token.DEFAULT_CHANNEL;
  	this._tokenStartCharIndex = -1;
  	this._tokenStartColumn = -1;
  	this._tokenStartLine = -1;
  	this._text = null;

  	this._hitEOF = false;
  	this._mode = Lexer.DEFAULT_MODE;
  	this._modeStack = [];

  	this._interp.reset();
  };

  // Return a token from this source; i.e., match a token on the char stream.
  Lexer.prototype.nextToken = function() {
  	if (this._input === null) {
  		throw "nextToken requires a non-null input stream.";
  	}

  	// Mark start location in char stream so unbuffered streams are
  	// guaranteed at least have text of current token
  	var tokenStartMarker = this._input.mark();
  	try {
  		while (true) {
  			if (this._hitEOF) {
  				this.emitEOF();
  				return this._token;
  			}
  			this._token = null;
  			this._channel = Token.DEFAULT_CHANNEL;
  			this._tokenStartCharIndex = this._input.index;
  			this._tokenStartColumn = this._interp.column;
  			this._tokenStartLine = this._interp.line;
  			this._text = null;
  			var continueOuter = false;
  			while (true) {
  				this._type = Token.INVALID_TYPE;
  				var ttype = Lexer.SKIP;
  				try {
  					ttype = this._interp.match(this._input, this._mode);
  				} catch (e) {
  				    if(e instanceof RecognitionException) {
                          this.notifyListeners(e); // report error
                          this.recover(e);
                      } else {
                          console.log(e.stack);
                          throw e;
                      }
  				}
  				if (this._input.LA(1) === Token.EOF) {
  					this._hitEOF = true;
  				}
  				if (this._type === Token.INVALID_TYPE) {
  					this._type = ttype;
  				}
  				if (this._type === Lexer.SKIP) {
  					continueOuter = true;
  					break;
  				}
  				if (this._type !== Lexer.MORE) {
  					break;
  				}
  			}
  			if (continueOuter) {
  				continue;
  			}
  			if (this._token === null) {
  				this.emit();
  			}
  			return this._token;
  		}
  	} finally {
  		// make sure we release marker after match or
  		// unbuffered char stream will keep buffering
  		this._input.release(tokenStartMarker);
  	}
  };

  // Instruct the lexer to skip creating a token for current lexer rule
  // and look for another token. nextToken() knows to keep looking when
  // a lexer rule finishes with token set to SKIP_TOKEN. Recall that
  // if token==null at end of any token rule, it creates one for you
  // and emits it.
  // /
  Lexer.prototype.skip = function() {
  	this._type = Lexer.SKIP;
  };

  Lexer.prototype.more = function() {
  	this._type = Lexer.MORE;
  };

  Lexer.prototype.mode = function(m) {
  	this._mode = m;
  };

  Lexer.prototype.pushMode = function(m) {
  	if (this._interp.debug) {
  		console.log("pushMode " + m);
  	}
  	this._modeStack.push(this._mode);
  	this.mode(m);
  };

  Lexer.prototype.popMode = function() {
  	if (this._modeStack.length === 0) {
  		throw "Empty Stack";
  	}
  	if (this._interp.debug) {
  		console.log("popMode back to " + this._modeStack.slice(0, -1));
  	}
  	this.mode(this._modeStack.pop());
  	return this._mode;
  };

  // Set the char stream and reset the lexer
  Object.defineProperty(Lexer.prototype, "inputStream", {
  	get : function() {
  		return this._input;
  	},
  	set : function(input) {
  		this._input = null;
  		this._tokenFactorySourcePair = [ this, this._input ];
  		this.reset();
  		this._input = input;
  		this._tokenFactorySourcePair = [ this, this._input ];
  	}
  });

  Object.defineProperty(Lexer.prototype, "sourceName", {
  	get : function sourceName() {
  		return this._input.sourceName;
  	}
  });

  // By default does not support multiple emits per nextToken invocation
  // for efficiency reasons. Subclass and override this method, nextToken,
  // and getToken (to push tokens into a list and pull from that list
  // rather than a single variable as this implementation does).
  // /
  Lexer.prototype.emitToken = function(token) {
  	this._token = token;
  };

  // The standard method called to automatically emit a token at the
  // outermost lexical rule. The token object should point into the
  // char buffer start..stop. If there is a text override in 'text',
  // use that to set the token's text. Override this method to emit
  // custom Token objects or provide a new factory.
  // /
  Lexer.prototype.emit = function() {
  	var t = this._factory.create(this._tokenFactorySourcePair, this._type,
  			this._text, this._channel, this._tokenStartCharIndex, this
  					.getCharIndex() - 1, this._tokenStartLine,
  			this._tokenStartColumn);
  	this.emitToken(t);
  	return t;
  };

  Lexer.prototype.emitEOF = function() {
  	var cpos = this.column;
  	var lpos = this.line;
  	var eof = this._factory.create(this._tokenFactorySourcePair, Token.EOF,
  			null, Token.DEFAULT_CHANNEL, this._input.index,
  			this._input.index - 1, lpos, cpos);
  	this.emitToken(eof);
  	return eof;
  };

  Object.defineProperty(Lexer.prototype, "type", {
  	get : function() {
  		return this.type;
  	},
  	set : function(type) {
  		this._type = type;
  	}
  });

  Object.defineProperty(Lexer.prototype, "line", {
  	get : function() {
  		return this._interp.line;
  	},
  	set : function(line) {
  		this._interp.line = line;
  	}
  });

  Object.defineProperty(Lexer.prototype, "column", {
  	get : function() {
  		return this._interp.column;
  	},
  	set : function(column) {
  		this._interp.column = column;
  	}
  });


  // What is the index of the current character of lookahead?///
  Lexer.prototype.getCharIndex = function() {
  	return this._input.index;
  };

  // Return the text matched so far for the current token or any text override.
  //Set the complete text of this token; it wipes any previous changes to the text.
  Object.defineProperty(Lexer.prototype, "text", {
  	get : function() {
  		if (this._text !== null) {
  			return this._text;
  		} else {
  			return this._interp.getText(this._input);
  		}
  	},
  	set : function(text) {
  		this._text = text;
  	}
  });
  // Return a list of all Token objects in input char stream.
  // Forces load of all tokens. Does not include EOF token.
  // /
  Lexer.prototype.getAllTokens = function() {
  	var tokens = [];
  	var t = this.nextToken();
  	while (t.type !== Token.EOF) {
  		tokens.push(t);
  		t = this.nextToken();
  	}
  	return tokens;
  };

  Lexer.prototype.notifyListeners = function(e) {
  	var start = this._tokenStartCharIndex;
  	var stop = this._input.index;
  	var text = this._input.getText(start, stop);
  	var msg = "token recognition error at: '" + this.getErrorDisplay(text) + "'";
  	var listener = this.getErrorListenerDispatch();
  	listener.syntaxError(this, null, this._tokenStartLine,
  			this._tokenStartColumn, msg, e);
  };

  Lexer.prototype.getErrorDisplay = function(s) {
  	var d = [];
  	for (var i = 0; i < s.length; i++) {
  		d.push(s[i]);
  	}
  	return d.join('');
  };

  Lexer.prototype.getErrorDisplayForChar = function(c) {
  	if (c.charCodeAt(0) === Token.EOF) {
  		return "<EOF>";
  	} else if (c === '\n') {
  		return "\\n";
  	} else if (c === '\t') {
  		return "\\t";
  	} else if (c === '\r') {
  		return "\\r";
  	} else {
  		return c;
  	}
  };

  Lexer.prototype.getCharErrorDisplay = function(c) {
  	return "'" + this.getErrorDisplayForChar(c) + "'";
  };

  // Lexers can normally match any char in it's vocabulary after matching
  // a token, so do the easy thing and just kill a character and hope
  // it all works out. You can instead use the rule invocation stack
  // to do sophisticated error recovery if you are in a fragment rule.
  // /
  Lexer.prototype.recover = function(re) {
  	if (this._input.LA(1) !== Token.EOF) {
  		if (re instanceof LexerNoViableAltException) {
  			// skip a char and try again
  			this._interp.consume(this._input);
  		} else {
  			// TODO: Do we lose character or line position information?
  			this._input.consume();
  		}
  	}
  };

  exports.Lexer = Lexer;

  },{"./CommonTokenFactory":2,"./Recognizer":12,"./Token":14,"./error/Errors":39}],9:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */

  var Token = require('./Token').Token;
  var ParseTreeListener = require('./tree/Tree').ParseTreeListener;
  var Recognizer = require('./Recognizer').Recognizer;
  var DefaultErrorStrategy = require('./error/ErrorStrategy').DefaultErrorStrategy;
  var ATNDeserializer = require('./atn/ATNDeserializer').ATNDeserializer;
  var ATNDeserializationOptions = require('./atn/ATNDeserializationOptions').ATNDeserializationOptions;
  var TerminalNode = require('./tree/Tree').TerminalNode;
  var ErrorNode = require('./tree/Tree').ErrorNode;

  function TraceListener(parser) {
  	ParseTreeListener.call(this);
      this.parser = parser;
  	return this;
  }

  TraceListener.prototype = Object.create(ParseTreeListener);
  TraceListener.prototype.constructor = TraceListener;

  TraceListener.prototype.enterEveryRule = function(ctx) {
  	console.log("enter   " + this.parser.ruleNames[ctx.ruleIndex] + ", LT(1)=" + this.parser._input.LT(1).text);
  };

  TraceListener.prototype.visitTerminal = function( node) {
  	console.log("consume " + node.symbol + " rule " + this.parser.ruleNames[this.parser._ctx.ruleIndex]);
  };

  TraceListener.prototype.exitEveryRule = function(ctx) {
  	console.log("exit    " + this.parser.ruleNames[ctx.ruleIndex] + ", LT(1)=" + this.parser._input.LT(1).text);
  };

  // this is all the parsing support code essentially; most of it is error
  // recovery stuff.//
  function Parser(input) {
  	Recognizer.call(this);
  	// The input stream.
  	this._input = null;
  	// The error handling strategy for the parser. The default value is a new
  	// instance of {@link DefaultErrorStrategy}.
  	this._errHandler = new DefaultErrorStrategy();
  	this._precedenceStack = [];
  	this._precedenceStack.push(0);
  	// The {@link ParserRuleContext} object for the currently executing rule.
  	// this is always non-null during the parsing process.
  	this._ctx = null;
  	// Specifies whether or not the parser should construct a parse tree during
  	// the parsing process. The default value is {@code true}.
  	this.buildParseTrees = true;
  	// When {@link //setTrace}{@code (true)} is called, a reference to the
  	// {@link TraceListener} is stored here so it can be easily removed in a
  	// later call to {@link //setTrace}{@code (false)}. The listener itself is
  	// implemented as a parser listener so this field is not directly used by
  	// other parser methods.
  	this._tracer = null;
  	// The list of {@link ParseTreeListener} listeners registered to receive
  	// events during the parse.
  	this._parseListeners = null;
  	// The number of syntax errors reported during parsing. this value is
  	// incremented each time {@link //notifyErrorListeners} is called.
  	this._syntaxErrors = 0;
  	this.setInputStream(input);
  	return this;
  }

  Parser.prototype = Object.create(Recognizer.prototype);
  Parser.prototype.contructor = Parser;

  // this field maps from the serialized ATN string to the deserialized {@link
  // ATN} with
  // bypass alternatives.
  //
  // @see ATNDeserializationOptions//isGenerateRuleBypassTransitions()
  //
  Parser.bypassAltsAtnCache = {};

  // reset the parser's state//
  Parser.prototype.reset = function() {
  	if (this._input !== null) {
  		this._input.seek(0);
  	}
  	this._errHandler.reset(this);
  	this._ctx = null;
  	this._syntaxErrors = 0;
  	this.setTrace(false);
  	this._precedenceStack = [];
  	this._precedenceStack.push(0);
  	if (this._interp !== null) {
  		this._interp.reset();
  	}
  };

  // Match current input symbol against {@code ttype}. If the symbol type
  // matches, {@link ANTLRErrorStrategy//reportMatch} and {@link //consume} are
  // called to complete the match process.
  //
  // <p>If the symbol type does not match,
  // {@link ANTLRErrorStrategy//recoverInline} is called on the current error
  // strategy to attempt recovery. If {@link //getBuildParseTree} is
  // {@code true} and the token index of the symbol returned by
  // {@link ANTLRErrorStrategy//recoverInline} is -1, the symbol is added to
  // the parse tree by calling {@link ParserRuleContext//addErrorNode}.</p>
  //
  // @param ttype the token type to match
  // @return the matched symbol
  // @throws RecognitionException if the current input symbol did not match
  // {@code ttype} and the error strategy could not recover from the
  // mismatched symbol

  Parser.prototype.match = function(ttype) {
  	var t = this.getCurrentToken();
  	if (t.type === ttype) {
  		this._errHandler.reportMatch(this);
  		this.consume();
  	} else {
  		t = this._errHandler.recoverInline(this);
  		if (this.buildParseTrees && t.tokenIndex === -1) {
  			// we must have conjured up a new token during single token
  			// insertion
  			// if it's not the current symbol
  			this._ctx.addErrorNode(t);
  		}
  	}
  	return t;
  };
  // Match current input symbol as a wildcard. If the symbol type matches
  // (i.e. has a value greater than 0), {@link ANTLRErrorStrategy//reportMatch}
  // and {@link //consume} are called to complete the match process.
  //
  // <p>If the symbol type does not match,
  // {@link ANTLRErrorStrategy//recoverInline} is called on the current error
  // strategy to attempt recovery. If {@link //getBuildParseTree} is
  // {@code true} and the token index of the symbol returned by
  // {@link ANTLRErrorStrategy//recoverInline} is -1, the symbol is added to
  // the parse tree by calling {@link ParserRuleContext//addErrorNode}.</p>
  //
  // @return the matched symbol
  // @throws RecognitionException if the current input symbol did not match
  // a wildcard and the error strategy could not recover from the mismatched
  // symbol

  Parser.prototype.matchWildcard = function() {
  	var t = this.getCurrentToken();
  	if (t.type > 0) {
  		this._errHandler.reportMatch(this);
  		this.consume();
  	} else {
  		t = this._errHandler.recoverInline(this);
  		if (this._buildParseTrees && t.tokenIndex === -1) {
  			// we must have conjured up a new token during single token
  			// insertion
  			// if it's not the current symbol
  			this._ctx.addErrorNode(t);
  		}
  	}
  	return t;
  };

  Parser.prototype.getParseListeners = function() {
  	return this._parseListeners || [];
  };

  // Registers {@code listener} to receive events during the parsing process.
  //
  // <p>To support output-preserving grammar transformations (including but not
  // limited to left-recursion removal, automated left-factoring, and
  // optimized code generation), calls to listener methods during the parse
  // may differ substantially from calls made by
  // {@link ParseTreeWalker//DEFAULT} used after the parse is complete. In
  // particular, rule entry and exit events may occur in a different order
  // during the parse than after the parser. In addition, calls to certain
  // rule entry methods may be omitted.</p>
  //
  // <p>With the following specific exceptions, calls to listener events are
  // <em>deterministic</em>, i.e. for identical input the calls to listener
  // methods will be the same.</p>
  //
  // <ul>
  // <li>Alterations to the grammar used to generate code may change the
  // behavior of the listener calls.</li>
  // <li>Alterations to the command line options passed to ANTLR 4 when
  // generating the parser may change the behavior of the listener calls.</li>
  // <li>Changing the version of the ANTLR Tool used to generate the parser
  // may change the behavior of the listener calls.</li>
  // </ul>
  //
  // @param listener the listener to add
  //
  // @throws NullPointerException if {@code} listener is {@code null}
  //
  Parser.prototype.addParseListener = function(listener) {
  	if (listener === null) {
  		throw "listener";
  	}
  	if (this._parseListeners === null) {
  		this._parseListeners = [];
  	}
  	this._parseListeners.push(listener);
  };

  //
  // Remove {@code listener} from the list of parse listeners.
  //
  // <p>If {@code listener} is {@code null} or has not been added as a parse
  // listener, this method does nothing.</p>
  // @param listener the listener to remove
  //
  Parser.prototype.removeParseListener = function(listener) {
  	if (this._parseListeners !== null) {
  		var idx = this._parseListeners.indexOf(listener);
  		if (idx >= 0) {
  			this._parseListeners.splice(idx, 1);
  		}
  		if (this._parseListeners.length === 0) {
  			this._parseListeners = null;
  		}
  	}
  };

  // Remove all parse listeners.
  Parser.prototype.removeParseListeners = function() {
  	this._parseListeners = null;
  };

  // Notify any parse listeners of an enter rule event.
  Parser.prototype.triggerEnterRuleEvent = function() {
  	if (this._parseListeners !== null) {
          var ctx = this._ctx;
  		this._parseListeners.map(function(listener) {
  			listener.enterEveryRule(ctx);
  			ctx.enterRule(listener);
  		});
  	}
  };

  //
  // Notify any parse listeners of an exit rule event.
  //
  // @see //addParseListener
  //
  Parser.prototype.triggerExitRuleEvent = function() {
  	if (this._parseListeners !== null) {
  		// reverse order walk of listeners
          var ctx = this._ctx;
  		this._parseListeners.slice(0).reverse().map(function(listener) {
  			ctx.exitRule(listener);
  			listener.exitEveryRule(ctx);
  		});
  	}
  };

  Parser.prototype.getTokenFactory = function() {
  	return this._input.tokenSource._factory;
  };

  // Tell our token source and error strategy about a new way to create tokens.//
  Parser.prototype.setTokenFactory = function(factory) {
  	this._input.tokenSource._factory = factory;
  };

  // The ATN with bypass alternatives is expensive to create so we create it
  // lazily.
  //
  // @throws UnsupportedOperationException if the current parser does not
  // implement the {@link //getSerializedATN()} method.
  //
  Parser.prototype.getATNWithBypassAlts = function() {
  	var serializedAtn = this.getSerializedATN();
  	if (serializedAtn === null) {
  		throw "The current parser does not support an ATN with bypass alternatives.";
  	}
  	var result = this.bypassAltsAtnCache[serializedAtn];
  	if (result === null) {
  		var deserializationOptions = new ATNDeserializationOptions();
  		deserializationOptions.generateRuleBypassTransitions = true;
  		result = new ATNDeserializer(deserializationOptions)
  				.deserialize(serializedAtn);
  		this.bypassAltsAtnCache[serializedAtn] = result;
  	}
  	return result;
  };

  // The preferred method of getting a tree pattern. For example, here's a
  // sample use:
  //
  // <pre>
  // ParseTree t = parser.expr();
  // ParseTreePattern p = parser.compileParseTreePattern("&lt;ID&gt;+0",
  // MyParser.RULE_expr);
  // ParseTreeMatch m = p.match(t);
  // String id = m.get("ID");
  // </pre>

  var Lexer = require('./Lexer').Lexer;

  Parser.prototype.compileParseTreePattern = function(pattern, patternRuleIndex, lexer) {
  	lexer = lexer || null;
  	if (lexer === null) {
  		if (this.getTokenStream() !== null) {
  			var tokenSource = this.getTokenStream().tokenSource;
  			if (tokenSource instanceof Lexer) {
  				lexer = tokenSource;
  			}
  		}
  	}
  	if (lexer === null) {
  		throw "Parser can't discover a lexer to use";
  	}
  	var m = new ParseTreePatternMatcher(lexer, this);
  	return m.compile(pattern, patternRuleIndex);
  };

  Parser.prototype.getInputStream = function() {
  	return this.getTokenStream();
  };

  Parser.prototype.setInputStream = function(input) {
  	this.setTokenStream(input);
  };

  Parser.prototype.getTokenStream = function() {
  	return this._input;
  };

  // Set the token stream and reset the parser.//
  Parser.prototype.setTokenStream = function(input) {
  	this._input = null;
  	this.reset();
  	this._input = input;
  };

  // Match needs to return the current input symbol, which gets put
  // into the label for the associated token ref; e.g., x=ID.
  //
  Parser.prototype.getCurrentToken = function() {
  	return this._input.LT(1);
  };

  Parser.prototype.notifyErrorListeners = function(msg, offendingToken, err) {
  	offendingToken = offendingToken || null;
  	err = err || null;
  	if (offendingToken === null) {
  		offendingToken = this.getCurrentToken();
  	}
  	this._syntaxErrors += 1;
  	var line = offendingToken.line;
  	var column = offendingToken.column;
  	var listener = this.getErrorListenerDispatch();
  	listener.syntaxError(this, offendingToken, line, column, msg, err);
  };

  //
  // Consume and return the {@linkplain //getCurrentToken current symbol}.
  //
  // <p>E.g., given the following input with {@code A} being the current
  // lookahead symbol, this function moves the cursor to {@code B} and returns
  // {@code A}.</p>
  //
  // <pre>
  // A B
  // ^
  // </pre>
  //
  // If the parser is not in error recovery mode, the consumed symbol is added
  // to the parse tree using {@link ParserRuleContext//addChild(Token)}, and
  // {@link ParseTreeListener//visitTerminal} is called on any parse listeners.
  // If the parser <em>is</em> in error recovery mode, the consumed symbol is
  // added to the parse tree using
  // {@link ParserRuleContext//addErrorNode(Token)}, and
  // {@link ParseTreeListener//visitErrorNode} is called on any parse
  // listeners.
  //
  Parser.prototype.consume = function() {
  	var o = this.getCurrentToken();
  	if (o.type !== Token.EOF) {
  		this.getInputStream().consume();
  	}
  	var hasListener = this._parseListeners !== null && this._parseListeners.length > 0;
  	if (this.buildParseTrees || hasListener) {
  		var node;
  		if (this._errHandler.inErrorRecoveryMode(this)) {
  			node = this._ctx.addErrorNode(o);
  		} else {
  			node = this._ctx.addTokenNode(o);
  		}
          node.invokingState = this.state;
  		if (hasListener) {
  			this._parseListeners.map(function(listener) {
  				if (node instanceof ErrorNode || (node.isErrorNode !== undefined && node.isErrorNode())) {
  					listener.visitErrorNode(node);
  				} else if (node instanceof TerminalNode) {
  					listener.visitTerminal(node);
  				}
  			});
  		}
  	}
  	return o;
  };

  Parser.prototype.addContextToParseTree = function() {
  	// add current context to parent if we have a parent
  	if (this._ctx.parentCtx !== null) {
  		this._ctx.parentCtx.addChild(this._ctx);
  	}
  };

  // Always called by generated parsers upon entry to a rule. Access field
  // {@link //_ctx} get the current context.

  Parser.prototype.enterRule = function(localctx, state, ruleIndex) {
  	this.state = state;
  	this._ctx = localctx;
  	this._ctx.start = this._input.LT(1);
  	if (this.buildParseTrees) {
  		this.addContextToParseTree();
  	}
  	if (this._parseListeners !== null) {
  		this.triggerEnterRuleEvent();
  	}
  };

  Parser.prototype.exitRule = function() {
  	this._ctx.stop = this._input.LT(-1);
  	// trigger event on _ctx, before it reverts to parent
  	if (this._parseListeners !== null) {
  		this.triggerExitRuleEvent();
  	}
  	this.state = this._ctx.invokingState;
  	this._ctx = this._ctx.parentCtx;
  };

  Parser.prototype.enterOuterAlt = function(localctx, altNum) {
     	localctx.setAltNumber(altNum);
  	// if we have new localctx, make sure we replace existing ctx
  	// that is previous child of parse tree
  	if (this.buildParseTrees && this._ctx !== localctx) {
  		if (this._ctx.parentCtx !== null) {
  			this._ctx.parentCtx.removeLastChild();
  			this._ctx.parentCtx.addChild(localctx);
  		}
  	}
  	this._ctx = localctx;
  };

  // Get the precedence level for the top-most precedence rule.
  //
  // @return The precedence level for the top-most precedence rule, or -1 if
  // the parser context is not nested within a precedence rule.

  Parser.prototype.getPrecedence = function() {
  	if (this._precedenceStack.length === 0) {
  		return -1;
  	} else {
  		return this._precedenceStack[this._precedenceStack.length-1];
  	}
  };

  Parser.prototype.enterRecursionRule = function(localctx, state, ruleIndex,
  		precedence) {
  	this.state = state;
  	this._precedenceStack.push(precedence);
  	this._ctx = localctx;
  	this._ctx.start = this._input.LT(1);
  	if (this._parseListeners !== null) {
  		this.triggerEnterRuleEvent(); // simulates rule entry for
  										// left-recursive rules
  	}
  };

  //
  // Like {@link //enterRule} but for recursive rules.

  Parser.prototype.pushNewRecursionContext = function(localctx, state, ruleIndex) {
  	var previous = this._ctx;
  	previous.parentCtx = localctx;
  	previous.invokingState = state;
  	previous.stop = this._input.LT(-1);

  	this._ctx = localctx;
  	this._ctx.start = previous.start;
  	if (this.buildParseTrees) {
  		this._ctx.addChild(previous);
  	}
  	if (this._parseListeners !== null) {
  		this.triggerEnterRuleEvent(); // simulates rule entry for
  										// left-recursive rules
  	}
  };

  Parser.prototype.unrollRecursionContexts = function(parentCtx) {
  	this._precedenceStack.pop();
  	this._ctx.stop = this._input.LT(-1);
  	var retCtx = this._ctx; // save current ctx (return value)
  	// unroll so _ctx is as it was before call to recursive method
  	if (this._parseListeners !== null) {
  		while (this._ctx !== parentCtx) {
  			this.triggerExitRuleEvent();
  			this._ctx = this._ctx.parentCtx;
  		}
  	} else {
  		this._ctx = parentCtx;
  	}
  	// hook into tree
  	retCtx.parentCtx = parentCtx;
  	if (this.buildParseTrees && parentCtx !== null) {
  		// add return ctx into invoking rule's tree
  		parentCtx.addChild(retCtx);
  	}
  };

  Parser.prototype.getInvokingContext = function(ruleIndex) {
  	var ctx = this._ctx;
  	while (ctx !== null) {
  		if (ctx.ruleIndex === ruleIndex) {
  			return ctx;
  		}
  		ctx = ctx.parentCtx;
  	}
  	return null;
  };

  Parser.prototype.precpred = function(localctx, precedence) {
  	return precedence >= this._precedenceStack[this._precedenceStack.length-1];
  };

  Parser.prototype.inContext = function(context) {
  	// TODO: useful in parser?
  	return false;
  };

  //
  // Checks whether or not {@code symbol} can follow the current state in the
  // ATN. The behavior of this method is equivalent to the following, but is
  // implemented such that the complete context-sensitive follow set does not
  // need to be explicitly constructed.
  //
  // <pre>
  // return getExpectedTokens().contains(symbol);
  // </pre>
  //
  // @param symbol the symbol type to check
  // @return {@code true} if {@code symbol} can follow the current state in
  // the ATN, otherwise {@code false}.

  Parser.prototype.isExpectedToken = function(symbol) {
  	var atn = this._interp.atn;
  	var ctx = this._ctx;
  	var s = atn.states[this.state];
  	var following = atn.nextTokens(s);
  	if (following.contains(symbol)) {
  		return true;
  	}
  	if (!following.contains(Token.EPSILON)) {
  		return false;
  	}
  	while (ctx !== null && ctx.invokingState >= 0 && following.contains(Token.EPSILON)) {
  		var invokingState = atn.states[ctx.invokingState];
  		var rt = invokingState.transitions[0];
  		following = atn.nextTokens(rt.followState);
  		if (following.contains(symbol)) {
  			return true;
  		}
  		ctx = ctx.parentCtx;
  	}
  	if (following.contains(Token.EPSILON) && symbol === Token.EOF) {
  		return true;
  	} else {
  		return false;
  	}
  };

  // Computes the set of input symbols which could follow the current parser
  // state and context, as given by {@link //getState} and {@link //getContext},
  // respectively.
  //
  // @see ATN//getExpectedTokens(int, RuleContext)
  //
  Parser.prototype.getExpectedTokens = function() {
  	return this._interp.atn.getExpectedTokens(this.state, this._ctx);
  };

  Parser.prototype.getExpectedTokensWithinCurrentRule = function() {
  	var atn = this._interp.atn;
  	var s = atn.states[this.state];
  	return atn.nextTokens(s);
  };

  // Get a rule's index (i.e., {@code RULE_ruleName} field) or -1 if not found.//
  Parser.prototype.getRuleIndex = function(ruleName) {
  	var ruleIndex = this.getRuleIndexMap()[ruleName];
  	if (ruleIndex !== null) {
  		return ruleIndex;
  	} else {
  		return -1;
  	}
  };

  // Return List&lt;String&gt; of the rule names in your parser instance
  // leading up to a call to the current rule. You could override if
  // you want more details such as the file/line info of where
  // in the ATN a rule is invoked.
  //
  // this is very useful for error messages.
  //
  Parser.prototype.getRuleInvocationStack = function(p) {
  	p = p || null;
  	if (p === null) {
  		p = this._ctx;
  	}
  	var stack = [];
  	while (p !== null) {
  		// compute what follows who invoked us
  		var ruleIndex = p.ruleIndex;
  		if (ruleIndex < 0) {
  			stack.push("n/a");
  		} else {
  			stack.push(this.ruleNames[ruleIndex]);
  		}
  		p = p.parentCtx;
  	}
  	return stack;
  };

  // For debugging and other purposes.//
  Parser.prototype.getDFAStrings = function() {
  	return this._interp.decisionToDFA.toString();
  };
  // For debugging and other purposes.//
  Parser.prototype.dumpDFA = function() {
  	var seenOne = false;
  	for (var i = 0; i < this._interp.decisionToDFA.length; i++) {
  		var dfa = this._interp.decisionToDFA[i];
  		if (dfa.states.length > 0) {
  			if (seenOne) {
  				console.log();
  			}
  			this.printer.println("Decision " + dfa.decision + ":");
  			this.printer.print(dfa.toString(this.literalNames, this.symbolicNames));
  			seenOne = true;
  		}
  	}
  };

  /*
  "			printer = function() {\r\n" +
  "				this.println = function(s) { document.getElementById('output') += s + '\\n'; }\r\n" +
  "				this.print = function(s) { document.getElementById('output') += s; }\r\n" +
  "			};\r\n" +
  */

  Parser.prototype.getSourceName = function() {
  	return this._input.sourceName;
  };

  // During a parse is sometimes useful to listen in on the rule entry and exit
  // events as well as token matches. this is for quick and dirty debugging.
  //
  Parser.prototype.setTrace = function(trace) {
  	if (!trace) {
  		this.removeParseListener(this._tracer);
  		this._tracer = null;
  	} else {
  		if (this._tracer !== null) {
  			this.removeParseListener(this._tracer);
  		}
  		this._tracer = new TraceListener(this);
  		this.addParseListener(this._tracer);
  	}
  };

  exports.Parser = Parser;
  },{"./Lexer":8,"./Recognizer":12,"./Token":14,"./atn/ATNDeserializationOptions":19,"./atn/ATNDeserializer":20,"./error/ErrorStrategy":38,"./tree/Tree":44}],10:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */

  //* A rule invocation record for parsing.
  //
  //  Contains all of the information about the current rule not stored in the
  //  RuleContext. It handles parse tree children list, Any ATN state
  //  tracing, and the default values available for rule indications:
  //  start, stop, rule index, current alt number, current
  //  ATN state.
  //
  //  Subclasses made for each rule and grammar track the parameters,
  //  return values, locals, and labels specific to that rule. These
  //  are the objects that are returned from rules.
  //
  //  Note text is not an actual field of a rule return value; it is computed
  //  from start and stop using the input stream's toString() method.  I
  //  could add a ctor to this so that we can pass in and store the input
  //  stream, but I'm not sure we want to do that.  It would seem to be undefined
  //  to get the .text property anyway if the rule matches tokens from multiple
  //  input streams.
  //
  //  I do not use getters for fields of objects that are used simply to
  //  group values such as this aggregate.  The getters/setters are there to
  //  satisfy the superclass interface.

  var RuleContext = require('./RuleContext').RuleContext;
  var Tree = require('./tree/Tree');
  var INVALID_INTERVAL = Tree.INVALID_INTERVAL;
  var TerminalNode = Tree.TerminalNode;
  var TerminalNodeImpl = Tree.TerminalNodeImpl;
  var ErrorNodeImpl = Tree.ErrorNodeImpl;
  var Interval = require("./IntervalSet").Interval;

  function ParserRuleContext(parent, invokingStateNumber) {
  	parent = parent || null;
  	invokingStateNumber = invokingStateNumber || null;
  	RuleContext.call(this, parent, invokingStateNumber);
  	this.ruleIndex = -1;
      // * If we are debugging or building a parse tree for a visitor,
      // we need to track all of the tokens and rule invocations associated
      // with this rule's context. This is empty for parsing w/o tree constr.
      // operation because we don't the need to track the details about
      // how we parse this rule.
      // /
      this.children = null;
      this.start = null;
      this.stop = null;
      // The exception that forced this rule to return. If the rule successfully
      // completed, this is {@code null}.
      this.exception = null;
  }

  ParserRuleContext.prototype = Object.create(RuleContext.prototype);
  ParserRuleContext.prototype.constructor = ParserRuleContext;

  // * COPY a ctx (I'm deliberately not using copy constructor)///
  ParserRuleContext.prototype.copyFrom = function(ctx) {
      // from RuleContext
      this.parentCtx = ctx.parentCtx;
      this.invokingState = ctx.invokingState;
      this.children = null;
      this.start = ctx.start;
      this.stop = ctx.stop;
      // copy any error nodes to alt label node
      if(ctx.children) {
          this.children = [];
          // reset parent pointer for any error nodes
      	ctx.children.map(function(child) {
      		if (child instanceof ErrorNodeImpl) {
                  this.children.push(child);
                  child.parentCtx = this;
              }
  		}, this);
  	}
  };

  // Double dispatch methods for listeners
  ParserRuleContext.prototype.enterRule = function(listener) {
  };

  ParserRuleContext.prototype.exitRule = function(listener) {
  };

  // * Does not set parent link; other add methods do that///
  ParserRuleContext.prototype.addChild = function(child) {
      if (this.children === null) {
          this.children = [];
      }
      this.children.push(child);
      return child;
  };

  // * Used by enterOuterAlt to toss out a RuleContext previously added as
  // we entered a rule. If we have // label, we will need to remove
  // generic ruleContext object.
  // /
  ParserRuleContext.prototype.removeLastChild = function() {
      if (this.children !== null) {
          this.children.pop();
      }
  };

  ParserRuleContext.prototype.addTokenNode = function(token) {
      var node = new TerminalNodeImpl(token);
      this.addChild(node);
      node.parentCtx = this;
      return node;
  };

  ParserRuleContext.prototype.addErrorNode = function(badToken) {
      var node = new ErrorNodeImpl(badToken);
      this.addChild(node);
      node.parentCtx = this;
      return node;
  };

  ParserRuleContext.prototype.getChild = function(i, type) {
  	type = type || null;
  	if (type === null) {
  		return this.children.length>=i ? this.children[i] : null;
  	} else {
  		for(var j=0; j<this.children.length; j++) {
  			var child = this.children[j];
  			if(child instanceof type) {
  				if(i===0) {
  					return child;
  				} else {
  					i -= 1;
  				}
  			}
  		}
  		return null;
      }
  };


  ParserRuleContext.prototype.getToken = function(ttype, i) {
  	for(var j=0; j<this.children.length; j++) {
  		var child = this.children[j];
  		if (child instanceof TerminalNode) {
  			if (child.symbol.type === ttype) {
  				if(i===0) {
  					return child;
  				} else {
  					i -= 1;
  				}
  			}
          }
  	}
      return null;
  };

  ParserRuleContext.prototype.getTokens = function(ttype ) {
      if (this.children=== null) {
          return [];
      } else {
  		var tokens = [];
  		for(var j=0; j<this.children.length; j++) {
  			var child = this.children[j];
  			if (child instanceof TerminalNode) {
  				if (child.symbol.type === ttype) {
  					tokens.push(child);
  				}
  			}
  		}
  		return tokens;
      }
  };

  ParserRuleContext.prototype.getTypedRuleContext = function(ctxType, i) {
      return this.getChild(i, ctxType);
  };

  ParserRuleContext.prototype.getTypedRuleContexts = function(ctxType) {
      if (this.children=== null) {
          return [];
      } else {
  		var contexts = [];
  		for(var j=0; j<this.children.length; j++) {
  			var child = this.children[j];
  			if (child instanceof ctxType) {
  				contexts.push(child);
  			}
  		}
  		return contexts;
  	}
  };

  ParserRuleContext.prototype.getChildCount = function() {
  	if (this.children=== null) {
  		return 0;
  	} else {
  		return this.children.length;
  	}
  };

  ParserRuleContext.prototype.getSourceInterval = function() {
      if( this.start === null || this.stop === null) {
          return INVALID_INTERVAL;
      } else {
          return new Interval(this.start.tokenIndex, this.stop.tokenIndex);
      }
  };

  RuleContext.EMPTY = new ParserRuleContext();

  function InterpreterRuleContext(parent, invokingStateNumber, ruleIndex) {
  	ParserRuleContext.call(parent, invokingStateNumber);
      this.ruleIndex = ruleIndex;
      return this;
  }

  InterpreterRuleContext.prototype = Object.create(ParserRuleContext.prototype);
  InterpreterRuleContext.prototype.constructor = InterpreterRuleContext;

  exports.ParserRuleContext = ParserRuleContext;
  },{"./IntervalSet":6,"./RuleContext":13,"./tree/Tree":44}],11:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  ///

  var RuleContext = require('./RuleContext').RuleContext;
  var Hash = require('./Utils').Hash;

  function PredictionContext(cachedHashCode) {
  	this.cachedHashCode = cachedHashCode;
  }

  // Represents {@code $} in local context prediction, which means wildcard.
  // {@code//+x =//}.
  // /
  PredictionContext.EMPTY = null;

  // Represents {@code $} in an array in full context mode, when {@code $}
  // doesn't mean wildcard: {@code $ + x = [$,x]}. Here,
  // {@code $} = {@link //EMPTY_RETURN_STATE}.
  // /
  PredictionContext.EMPTY_RETURN_STATE = 0x7FFFFFFF;

  PredictionContext.globalNodeCount = 1;
  PredictionContext.id = PredictionContext.globalNodeCount;

  // Stores the computed hash code of this {@link PredictionContext}. The hash
  // code is computed in parts to match the following reference algorithm.
  //
  // <pre>
  // private int referenceHashCode() {
  // int hash = {@link MurmurHash//initialize MurmurHash.initialize}({@link
  // //INITIAL_HASH});
  //
  // for (int i = 0; i &lt; {@link //size()}; i++) {
  // hash = {@link MurmurHash//update MurmurHash.update}(hash, {@link //getParent
  // getParent}(i));
  // }
  //
  // for (int i = 0; i &lt; {@link //size()}; i++) {
  // hash = {@link MurmurHash//update MurmurHash.update}(hash, {@link
  // //getReturnState getReturnState}(i));
  // }
  //
  // hash = {@link MurmurHash//finish MurmurHash.finish}(hash, 2// {@link
  // //size()});
  // return hash;
  // }
  // </pre>
  // /

  // This means only the {@link //EMPTY} context is in set.
  PredictionContext.prototype.isEmpty = function() {
  	return this === PredictionContext.EMPTY;
  };

  PredictionContext.prototype.hasEmptyPath = function() {
  	return this.getReturnState(this.length - 1) === PredictionContext.EMPTY_RETURN_STATE;
  };

  PredictionContext.prototype.hashCode = function() {
  	return this.cachedHashCode;
  };


  PredictionContext.prototype.updateHashCode = function(hash) {
      hash.update(this.cachedHashCode);
  };
  /*
  function calculateHashString(parent, returnState) {
  	return "" + parent + returnState;
  }
  */

  // Used to cache {@link PredictionContext} objects. Its used for the shared
  // context cash associated with contexts in DFA states. This cache
  // can be used for both lexers and parsers.

  function PredictionContextCache() {
  	this.cache = {};
  	return this;
  }

  // Add a context to the cache and return it. If the context already exists,
  // return that one instead and do not add a new context to the cache.
  // Protect shared cache from unsafe thread access.
  //
  PredictionContextCache.prototype.add = function(ctx) {
  	if (ctx === PredictionContext.EMPTY) {
  		return PredictionContext.EMPTY;
  	}
  	var existing = this.cache[ctx] || null;
  	if (existing !== null) {
  		return existing;
  	}
  	this.cache[ctx] = ctx;
  	return ctx;
  };

  PredictionContextCache.prototype.get = function(ctx) {
  	return this.cache[ctx] || null;
  };

  Object.defineProperty(PredictionContextCache.prototype, "length", {
  	get : function() {
  		return this.cache.length;
  	}
  });

  function SingletonPredictionContext(parent, returnState) {
  	var hashCode = 0;
  	if(parent !== null) {
  		var hash = new Hash();
  		hash.update(parent, returnState);
          hashCode = hash.finish();
  	}
  	PredictionContext.call(this, hashCode);
  	this.parentCtx = parent;
  	this.returnState = returnState;
  }

  SingletonPredictionContext.prototype = Object.create(PredictionContext.prototype);
  SingletonPredictionContext.prototype.contructor = SingletonPredictionContext;

  SingletonPredictionContext.create = function(parent, returnState) {
  	if (returnState === PredictionContext.EMPTY_RETURN_STATE && parent === null) {
  		// someone can pass in the bits of an array ctx that mean $
  		return PredictionContext.EMPTY;
  	} else {
  		return new SingletonPredictionContext(parent, returnState);
  	}
  };

  Object.defineProperty(SingletonPredictionContext.prototype, "length", {
  	get : function() {
  		return 1;
  	}
  });

  SingletonPredictionContext.prototype.getParent = function(index) {
  	return this.parentCtx;
  };

  SingletonPredictionContext.prototype.getReturnState = function(index) {
  	return this.returnState;
  };

  SingletonPredictionContext.prototype.equals = function(other) {
  	if (this === other) {
  		return true;
  	} else if (!(other instanceof SingletonPredictionContext)) {
  		return false;
  	} else if (this.hashCode() !== other.hashCode()) {
  		return false; // can't be same if hash is different
  	} else {
  		if(this.returnState !== other.returnState)
              return false;
          else if(this.parentCtx==null)
              return other.parentCtx==null
  		else
              return this.parentCtx.equals(other.parentCtx);
  	}
  };

  SingletonPredictionContext.prototype.toString = function() {
  	var up = this.parentCtx === null ? "" : this.parentCtx.toString();
  	if (up.length === 0) {
  		if (this.returnState === PredictionContext.EMPTY_RETURN_STATE) {
  			return "$";
  		} else {
  			return "" + this.returnState;
  		}
  	} else {
  		return "" + this.returnState + " " + up;
  	}
  };

  function EmptyPredictionContext() {
  	SingletonPredictionContext.call(this, null, PredictionContext.EMPTY_RETURN_STATE);
  	return this;
  }

  EmptyPredictionContext.prototype = Object.create(SingletonPredictionContext.prototype);
  EmptyPredictionContext.prototype.constructor = EmptyPredictionContext;

  EmptyPredictionContext.prototype.isEmpty = function() {
  	return true;
  };

  EmptyPredictionContext.prototype.getParent = function(index) {
  	return null;
  };

  EmptyPredictionContext.prototype.getReturnState = function(index) {
  	return this.returnState;
  };

  EmptyPredictionContext.prototype.equals = function(other) {
  	return this === other;
  };

  EmptyPredictionContext.prototype.toString = function() {
  	return "$";
  };

  PredictionContext.EMPTY = new EmptyPredictionContext();

  function ArrayPredictionContext(parents, returnStates) {
  	// Parent can be null only if full ctx mode and we make an array
  	// from {@link //EMPTY} and non-empty. We merge {@link //EMPTY} by using
  	// null parent and
  	// returnState == {@link //EMPTY_RETURN_STATE}.
  	var h = new Hash();
  	h.update(parents, returnStates);
  	var hashCode = h.finish();
  	PredictionContext.call(this, hashCode);
  	this.parents = parents;
  	this.returnStates = returnStates;
  	return this;
  }

  ArrayPredictionContext.prototype = Object.create(PredictionContext.prototype);
  ArrayPredictionContext.prototype.constructor = ArrayPredictionContext;

  ArrayPredictionContext.prototype.isEmpty = function() {
  	// since EMPTY_RETURN_STATE can only appear in the last position, we
  	// don't need to verify that size==1
  	return this.returnStates[0] === PredictionContext.EMPTY_RETURN_STATE;
  };

  Object.defineProperty(ArrayPredictionContext.prototype, "length", {
  	get : function() {
  		return this.returnStates.length;
  	}
  });

  ArrayPredictionContext.prototype.getParent = function(index) {
  	return this.parents[index];
  };

  ArrayPredictionContext.prototype.getReturnState = function(index) {
  	return this.returnStates[index];
  };

  ArrayPredictionContext.prototype.equals = function(other) {
  	if (this === other) {
  		return true;
  	} else if (!(other instanceof ArrayPredictionContext)) {
  		return false;
  	} else if (this.hashCode() !== other.hashCode()) {
  		return false; // can't be same if hash is different
  	} else {
  		return this.returnStates === other.returnStates &&
  				this.parents === other.parents;
  	}
  };

  ArrayPredictionContext.prototype.toString = function() {
  	if (this.isEmpty()) {
  		return "[]";
  	} else {
  		var s = "[";
  		for (var i = 0; i < this.returnStates.length; i++) {
  			if (i > 0) {
  				s = s + ", ";
  			}
  			if (this.returnStates[i] === PredictionContext.EMPTY_RETURN_STATE) {
  				s = s + "$";
  				continue;
  			}
  			s = s + this.returnStates[i];
  			if (this.parents[i] !== null) {
  				s = s + " " + this.parents[i];
  			} else {
  				s = s + "null";
  			}
  		}
  		return s + "]";
  	}
  };

  // Convert a {@link RuleContext} tree to a {@link PredictionContext} graph.
  // Return {@link //EMPTY} if {@code outerContext} is empty or null.
  // /
  function predictionContextFromRuleContext(atn, outerContext) {
  	if (outerContext === undefined || outerContext === null) {
  		outerContext = RuleContext.EMPTY;
  	}
  	// if we are in RuleContext of start rule, s, then PredictionContext
  	// is EMPTY. Nobody called us. (if we are empty, return empty)
  	if (outerContext.parentCtx === null || outerContext === RuleContext.EMPTY) {
  		return PredictionContext.EMPTY;
  	}
  	// If we have a parent, convert it to a PredictionContext graph
  	var parent = predictionContextFromRuleContext(atn, outerContext.parentCtx);
  	var state = atn.states[outerContext.invokingState];
  	var transition = state.transitions[0];
  	return SingletonPredictionContext.create(parent, transition.followState.stateNumber);
  }
  /*
  function calculateListsHashString(parents, returnStates) {
  	var s = "";
  	parents.map(function(p) {
  		s = s + p;
  	});
  	returnStates.map(function(r) {
  		s = s + r;
  	});
  	return s;
  }
  */
  function merge(a, b, rootIsWildcard, mergeCache) {
  	// share same graph if both same
  	if (a === b) {
  		return a;
  	}
  	if (a instanceof SingletonPredictionContext && b instanceof SingletonPredictionContext) {
  		return mergeSingletons(a, b, rootIsWildcard, mergeCache);
  	}
  	// At least one of a or b is array
  	// If one is $ and rootIsWildcard, return $ as// wildcard
  	if (rootIsWildcard) {
  		if (a instanceof EmptyPredictionContext) {
  			return a;
  		}
  		if (b instanceof EmptyPredictionContext) {
  			return b;
  		}
  	}
  	// convert singleton so both are arrays to normalize
  	if (a instanceof SingletonPredictionContext) {
  		a = new ArrayPredictionContext([a.getParent()], [a.returnState]);
  	}
  	if (b instanceof SingletonPredictionContext) {
  		b = new ArrayPredictionContext([b.getParent()], [b.returnState]);
  	}
  	return mergeArrays(a, b, rootIsWildcard, mergeCache);
  }

  //
  // Merge two {@link SingletonPredictionContext} instances.
  //
  // <p>Stack tops equal, parents merge is same; return left graph.<br>
  // <embed src="images/SingletonMerge_SameRootSamePar.svg"
  // type="image/svg+xml"/></p>
  //
  // <p>Same stack top, parents differ; merge parents giving array node, then
  // remainders of those graphs. A new root node is created to point to the
  // merged parents.<br>
  // <embed src="images/SingletonMerge_SameRootDiffPar.svg"
  // type="image/svg+xml"/></p>
  //
  // <p>Different stack tops pointing to same parent. Make array node for the
  // root where both element in the root point to the same (original)
  // parent.<br>
  // <embed src="images/SingletonMerge_DiffRootSamePar.svg"
  // type="image/svg+xml"/></p>
  //
  // <p>Different stack tops pointing to different parents. Make array node for
  // the root where each element points to the corresponding original
  // parent.<br>
  // <embed src="images/SingletonMerge_DiffRootDiffPar.svg"
  // type="image/svg+xml"/></p>
  //
  // @param a the first {@link SingletonPredictionContext}
  // @param b the second {@link SingletonPredictionContext}
  // @param rootIsWildcard {@code true} if this is a local-context merge,
  // otherwise false to indicate a full-context merge
  // @param mergeCache
  // /
  function mergeSingletons(a, b, rootIsWildcard, mergeCache) {
  	if (mergeCache !== null) {
  		var previous = mergeCache.get(a, b);
  		if (previous !== null) {
  			return previous;
  		}
  		previous = mergeCache.get(b, a);
  		if (previous !== null) {
  			return previous;
  		}
  	}

  	var rootMerge = mergeRoot(a, b, rootIsWildcard);
  	if (rootMerge !== null) {
  		if (mergeCache !== null) {
  			mergeCache.set(a, b, rootMerge);
  		}
  		return rootMerge;
  	}
  	if (a.returnState === b.returnState) {
  		var parent = merge(a.parentCtx, b.parentCtx, rootIsWildcard, mergeCache);
  		// if parent is same as existing a or b parent or reduced to a parent,
  		// return it
  		if (parent === a.parentCtx) {
  			return a; // ax + bx = ax, if a=b
  		}
  		if (parent === b.parentCtx) {
  			return b; // ax + bx = bx, if a=b
  		}
  		// else: ax + ay = a'[x,y]
  		// merge parents x and y, giving array node with x,y then remainders
  		// of those graphs. dup a, a' points at merged array
  		// new joined parent so create new singleton pointing to it, a'
  		var spc = SingletonPredictionContext.create(parent, a.returnState);
  		if (mergeCache !== null) {
  			mergeCache.set(a, b, spc);
  		}
  		return spc;
  	} else { // a != b payloads differ
  		// see if we can collapse parents due to $+x parents if local ctx
  		var singleParent = null;
  		if (a === b || (a.parentCtx !== null && a.parentCtx === b.parentCtx)) { // ax +
  																				// bx =
  																				// [a,b]x
  			singleParent = a.parentCtx;
  		}
  		if (singleParent !== null) { // parents are same
  			// sort payloads and use same parent
  			var payloads = [ a.returnState, b.returnState ];
  			if (a.returnState > b.returnState) {
  				payloads[0] = b.returnState;
  				payloads[1] = a.returnState;
  			}
  			var parents = [ singleParent, singleParent ];
  			var apc = new ArrayPredictionContext(parents, payloads);
  			if (mergeCache !== null) {
  				mergeCache.set(a, b, apc);
  			}
  			return apc;
  		}
  		// parents differ and can't merge them. Just pack together
  		// into array; can't merge.
  		// ax + by = [ax,by]
  		var payloads = [ a.returnState, b.returnState ];
  		var parents = [ a.parentCtx, b.parentCtx ];
  		if (a.returnState > b.returnState) { // sort by payload
  			payloads[0] = b.returnState;
  			payloads[1] = a.returnState;
  			parents = [ b.parentCtx, a.parentCtx ];
  		}
  		var a_ = new ArrayPredictionContext(parents, payloads);
  		if (mergeCache !== null) {
  			mergeCache.set(a, b, a_);
  		}
  		return a_;
  	}
  }

  //
  // Handle case where at least one of {@code a} or {@code b} is
  // {@link //EMPTY}. In the following diagrams, the symbol {@code $} is used
  // to represent {@link //EMPTY}.
  //
  // <h2>Local-Context Merges</h2>
  //
  // <p>These local-context merge operations are used when {@code rootIsWildcard}
  // is true.</p>
  //
  // <p>{@link //EMPTY} is superset of any graph; return {@link //EMPTY}.<br>
  // <embed src="images/LocalMerge_EmptyRoot.svg" type="image/svg+xml"/></p>
  //
  // <p>{@link //EMPTY} and anything is {@code //EMPTY}, so merged parent is
  // {@code //EMPTY}; return left graph.<br>
  // <embed src="images/LocalMerge_EmptyParent.svg" type="image/svg+xml"/></p>
  //
  // <p>Special case of last merge if local context.<br>
  // <embed src="images/LocalMerge_DiffRoots.svg" type="image/svg+xml"/></p>
  //
  // <h2>Full-Context Merges</h2>
  //
  // <p>These full-context merge operations are used when {@code rootIsWildcard}
  // is false.</p>
  //
  // <p><embed src="images/FullMerge_EmptyRoots.svg" type="image/svg+xml"/></p>
  //
  // <p>Must keep all contexts; {@link //EMPTY} in array is a special value (and
  // null parent).<br>
  // <embed src="images/FullMerge_EmptyRoot.svg" type="image/svg+xml"/></p>
  //
  // <p><embed src="images/FullMerge_SameRoot.svg" type="image/svg+xml"/></p>
  //
  // @param a the first {@link SingletonPredictionContext}
  // @param b the second {@link SingletonPredictionContext}
  // @param rootIsWildcard {@code true} if this is a local-context merge,
  // otherwise false to indicate a full-context merge
  // /
  function mergeRoot(a, b, rootIsWildcard) {
  	if (rootIsWildcard) {
  		if (a === PredictionContext.EMPTY) {
  			return PredictionContext.EMPTY; // // + b =//
  		}
  		if (b === PredictionContext.EMPTY) {
  			return PredictionContext.EMPTY; // a +// =//
  		}
  	} else {
  		if (a === PredictionContext.EMPTY && b === PredictionContext.EMPTY) {
  			return PredictionContext.EMPTY; // $ + $ = $
  		} else if (a === PredictionContext.EMPTY) { // $ + x = [$,x]
  			var payloads = [ b.returnState,
  					PredictionContext.EMPTY_RETURN_STATE ];
  			var parents = [ b.parentCtx, null ];
  			return new ArrayPredictionContext(parents, payloads);
  		} else if (b === PredictionContext.EMPTY) { // x + $ = [$,x] ($ is always first if present)
  			var payloads = [ a.returnState, PredictionContext.EMPTY_RETURN_STATE ];
  			var parents = [ a.parentCtx, null ];
  			return new ArrayPredictionContext(parents, payloads);
  		}
  	}
  	return null;
  }

  //
  // Merge two {@link ArrayPredictionContext} instances.
  //
  // <p>Different tops, different parents.<br>
  // <embed src="images/ArrayMerge_DiffTopDiffPar.svg" type="image/svg+xml"/></p>
  //
  // <p>Shared top, same parents.<br>
  // <embed src="images/ArrayMerge_ShareTopSamePar.svg" type="image/svg+xml"/></p>
  //
  // <p>Shared top, different parents.<br>
  // <embed src="images/ArrayMerge_ShareTopDiffPar.svg" type="image/svg+xml"/></p>
  //
  // <p>Shared top, all shared parents.<br>
  // <embed src="images/ArrayMerge_ShareTopSharePar.svg"
  // type="image/svg+xml"/></p>
  //
  // <p>Equal tops, merge parents and reduce top to
  // {@link SingletonPredictionContext}.<br>
  // <embed src="images/ArrayMerge_EqualTop.svg" type="image/svg+xml"/></p>
  // /
  function mergeArrays(a, b, rootIsWildcard, mergeCache) {
  	if (mergeCache !== null) {
  		var previous = mergeCache.get(a, b);
  		if (previous !== null) {
  			return previous;
  		}
  		previous = mergeCache.get(b, a);
  		if (previous !== null) {
  			return previous;
  		}
  	}
  	// merge sorted payloads a + b => M
  	var i = 0; // walks a
  	var j = 0; // walks b
  	var k = 0; // walks target M array

  	var mergedReturnStates = [];
  	var mergedParents = [];
  	// walk and merge to yield mergedParents, mergedReturnStates
  	while (i < a.returnStates.length && j < b.returnStates.length) {
  		var a_parent = a.parents[i];
  		var b_parent = b.parents[j];
  		if (a.returnStates[i] === b.returnStates[j]) {
  			// same payload (stack tops are equal), must yield merged singleton
  			var payload = a.returnStates[i];
  			// $+$ = $
  			var bothDollars = payload === PredictionContext.EMPTY_RETURN_STATE &&
  					a_parent === null && b_parent === null;
  			var ax_ax = (a_parent !== null && b_parent !== null && a_parent === b_parent); // ax+ax
  																							// ->
  																							// ax
  			if (bothDollars || ax_ax) {
  				mergedParents[k] = a_parent; // choose left
  				mergedReturnStates[k] = payload;
  			} else { // ax+ay -> a'[x,y]
  				var mergedParent = merge(a_parent, b_parent, rootIsWildcard, mergeCache);
  				mergedParents[k] = mergedParent;
  				mergedReturnStates[k] = payload;
  			}
  			i += 1; // hop over left one as usual
  			j += 1; // but also skip one in right side since we merge
  		} else if (a.returnStates[i] < b.returnStates[j]) { // copy a[i] to M
  			mergedParents[k] = a_parent;
  			mergedReturnStates[k] = a.returnStates[i];
  			i += 1;
  		} else { // b > a, copy b[j] to M
  			mergedParents[k] = b_parent;
  			mergedReturnStates[k] = b.returnStates[j];
  			j += 1;
  		}
  		k += 1;
  	}
  	// copy over any payloads remaining in either array
  	if (i < a.returnStates.length) {
  		for (var p = i; p < a.returnStates.length; p++) {
  			mergedParents[k] = a.parents[p];
  			mergedReturnStates[k] = a.returnStates[p];
  			k += 1;
  		}
  	} else {
  		for (var p = j; p < b.returnStates.length; p++) {
  			mergedParents[k] = b.parents[p];
  			mergedReturnStates[k] = b.returnStates[p];
  			k += 1;
  		}
  	}
  	// trim merged if we combined a few that had same stack tops
  	if (k < mergedParents.length) { // write index < last position; trim
  		if (k === 1) { // for just one merged element, return singleton top
  			var a_ = SingletonPredictionContext.create(mergedParents[0],
  					mergedReturnStates[0]);
  			if (mergeCache !== null) {
  				mergeCache.set(a, b, a_);
  			}
  			return a_;
  		}
  		mergedParents = mergedParents.slice(0, k);
  		mergedReturnStates = mergedReturnStates.slice(0, k);
  	}

  	var M = new ArrayPredictionContext(mergedParents, mergedReturnStates);

  	// if we created same array as a or b, return that instead
  	// TODO: track whether this is possible above during merge sort for speed
  	if (M === a) {
  		if (mergeCache !== null) {
  			mergeCache.set(a, b, a);
  		}
  		return a;
  	}
  	if (M === b) {
  		if (mergeCache !== null) {
  			mergeCache.set(a, b, b);
  		}
  		return b;
  	}
  	combineCommonParents(mergedParents);

  	if (mergeCache !== null) {
  		mergeCache.set(a, b, M);
  	}
  	return M;
  }

  //
  // Make pass over all <em>M</em> {@code parents}; merge any {@code equals()}
  // ones.
  // /
  function combineCommonParents(parents) {
  	var uniqueParents = {};

  	for (var p = 0; p < parents.length; p++) {
  		var parent = parents[p];
  		if (!(parent in uniqueParents)) {
  			uniqueParents[parent] = parent;
  		}
  	}
  	for (var q = 0; q < parents.length; q++) {
  		parents[q] = uniqueParents[parents[q]];
  	}
  }

  function getCachedPredictionContext(context, contextCache, visited) {
  	if (context.isEmpty()) {
  		return context;
  	}
  	var existing = visited[context] || null;
  	if (existing !== null) {
  		return existing;
  	}
  	existing = contextCache.get(context);
  	if (existing !== null) {
  		visited[context] = existing;
  		return existing;
  	}
  	var changed = false;
  	var parents = [];
  	for (var i = 0; i < parents.length; i++) {
  		var parent = getCachedPredictionContext(context.getParent(i), contextCache, visited);
  		if (changed || parent !== context.getParent(i)) {
  			if (!changed) {
  				parents = [];
  				for (var j = 0; j < context.length; j++) {
  					parents[j] = context.getParent(j);
  				}
  				changed = true;
  			}
  			parents[i] = parent;
  		}
  	}
  	if (!changed) {
  		contextCache.add(context);
  		visited[context] = context;
  		return context;
  	}
  	var updated = null;
  	if (parents.length === 0) {
  		updated = PredictionContext.EMPTY;
  	} else if (parents.length === 1) {
  		updated = SingletonPredictionContext.create(parents[0], context
  				.getReturnState(0));
  	} else {
  		updated = new ArrayPredictionContext(parents, context.returnStates);
  	}
  	contextCache.add(updated);
  	visited[updated] = updated;
  	visited[context] = updated;

  	return updated;
  }

  exports.merge = merge;
  exports.PredictionContext = PredictionContext;
  exports.PredictionContextCache = PredictionContextCache;
  exports.SingletonPredictionContext = SingletonPredictionContext;
  exports.predictionContextFromRuleContext = predictionContextFromRuleContext;
  exports.getCachedPredictionContext = getCachedPredictionContext;

  },{"./RuleContext":13,"./Utils":15}],12:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  //

  var Token = require('./Token').Token;
  var ConsoleErrorListener = require('./error/ErrorListener').ConsoleErrorListener;
  var ProxyErrorListener = require('./error/ErrorListener').ProxyErrorListener;

  function Recognizer() {
      this._listeners = [ ConsoleErrorListener.INSTANCE ];
      this._interp = null;
      this._stateNumber = -1;
      return this;
  }

  Recognizer.tokenTypeMapCache = {};
  Recognizer.ruleIndexMapCache = {};


  Recognizer.prototype.checkVersion = function(toolVersion) {
      var runtimeVersion = "4.7";
      if (runtimeVersion!==toolVersion) {
          console.log("ANTLR runtime and generated code versions disagree: "+runtimeVersion+"!="+toolVersion);
      }
  };

  Recognizer.prototype.addErrorListener = function(listener) {
      this._listeners.push(listener);
  };

  Recognizer.prototype.removeErrorListeners = function() {
      this._listeners = [];
  };

  Recognizer.prototype.getTokenTypeMap = function() {
      var tokenNames = this.getTokenNames();
      if (tokenNames===null) {
          throw("The current recognizer does not provide a list of token names.");
      }
      var result = this.tokenTypeMapCache[tokenNames];
      if(result===undefined) {
          result = tokenNames.reduce(function(o, k, i) { o[k] = i; });
          result.EOF = Token.EOF;
          this.tokenTypeMapCache[tokenNames] = result;
      }
      return result;
  };

  // Get a map from rule names to rule indexes.
  //
  // <p>Used for XPath and tree pattern compilation.</p>
  //
  Recognizer.prototype.getRuleIndexMap = function() {
      var ruleNames = this.ruleNames;
      if (ruleNames===null) {
          throw("The current recognizer does not provide a list of rule names.");
      }
      var result = this.ruleIndexMapCache[ruleNames];
      if(result===undefined) {
          result = ruleNames.reduce(function(o, k, i) { o[k] = i; });
          this.ruleIndexMapCache[ruleNames] = result;
      }
      return result;
  };

  Recognizer.prototype.getTokenType = function(tokenName) {
      var ttype = this.getTokenTypeMap()[tokenName];
      if (ttype !==undefined) {
          return ttype;
      } else {
          return Token.INVALID_TYPE;
      }
  };


  // What is the error header, normally line/character position information?//
  Recognizer.prototype.getErrorHeader = function(e) {
      var line = e.getOffendingToken().line;
      var column = e.getOffendingToken().column;
      return "line " + line + ":" + column;
  };


  // How should a token be displayed in an error message? The default
  //  is to display just the text, but during development you might
  //  want to have a lot of information spit out.  Override in that case
  //  to use t.toString() (which, for CommonToken, dumps everything about
  //  the token). This is better than forcing you to override a method in
  //  your token objects because you don't have to go modify your lexer
  //  so that it creates a new Java type.
  //
  // @deprecated This method is not called by the ANTLR 4 Runtime. Specific
  // implementations of {@link ANTLRErrorStrategy} may provide a similar
  // feature when necessary. For example, see
  // {@link DefaultErrorStrategy//getTokenErrorDisplay}.
  //
  Recognizer.prototype.getTokenErrorDisplay = function(t) {
      if (t===null) {
          return "<no token>";
      }
      var s = t.text;
      if (s===null) {
          if (t.type===Token.EOF) {
              s = "<EOF>";
          } else {
              s = "<" + t.type + ">";
          }
      }
      s = s.replace("\n","\\n").replace("\r","\\r").replace("\t","\\t");
      return "'" + s + "'";
  };

  Recognizer.prototype.getErrorListenerDispatch = function() {
      return new ProxyErrorListener(this._listeners);
  };

  // subclass needs to override these if there are sempreds or actions
  // that the ATN interp needs to execute
  Recognizer.prototype.sempred = function(localctx, ruleIndex, actionIndex) {
      return true;
  };

  Recognizer.prototype.precpred = function(localctx , precedence) {
      return true;
  };

  //Indicate that the recognizer has changed internal state that is
  //consistent with the ATN state passed in.  This way we always know
  //where we are in the ATN as the parser goes along. The rule
  //context objects form a stack that lets us see the stack of
  //invoking rules. Combine this and we have complete ATN
  //configuration information.

  Object.defineProperty(Recognizer.prototype, "state", {
  	get : function() {
  		return this._stateNumber;
  	},
  	set : function(state) {
  		this._stateNumber = state;
  	}
  });


  exports.Recognizer = Recognizer;

  },{"./Token":14,"./error/ErrorListener":37}],13:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  ///

  //  A rule context is a record of a single rule invocation. It knows
  //  which context invoked it, if any. If there is no parent context, then
  //  naturally the invoking state is not valid.  The parent link
  //  provides a chain upwards from the current rule invocation to the root
  //  of the invocation tree, forming a stack. We actually carry no
  //  information about the rule associated with this context (except
  //  when parsing). We keep only the state number of the invoking state from
  //  the ATN submachine that invoked this. Contrast this with the s
  //  pointer inside ParserRuleContext that tracks the current state
  //  being "executed" for the current rule.
  //
  //  The parent contexts are useful for computing lookahead sets and
  //  getting error information.
  //
  //  These objects are used during parsing and prediction.
  //  For the special case of parsers, we use the subclass
  //  ParserRuleContext.
  //
  //  @see ParserRuleContext
  ///

  var RuleNode = require('./tree/Tree').RuleNode;
  var INVALID_INTERVAL = require('./tree/Tree').INVALID_INTERVAL;
  var INVALID_ALT_NUMBER = require('./atn/ATN').INVALID_ALT_NUMBER;

  function RuleContext(parent, invokingState) {
  	RuleNode.call(this);
  	// What context invoked this rule?
  	this.parentCtx = parent || null;
  	// What state invoked the rule associated with this context?
  	// The "return address" is the followState of invokingState
  	// If parent is null, this should be -1.
  	this.invokingState = invokingState || -1;
  	return this;
  }

  RuleContext.prototype = Object.create(RuleNode.prototype);
  RuleContext.prototype.constructor = RuleContext;

  RuleContext.prototype.depth = function() {
  	var n = 0;
  	var p = this;
  	while (p !== null) {
  		p = p.parentCtx;
  		n += 1;
  	}
  	return n;
  };

  // A context is empty if there is no invoking state; meaning nobody call
  // current context.
  RuleContext.prototype.isEmpty = function() {
  	return this.invokingState === -1;
  };

  // satisfy the ParseTree / SyntaxTree interface

  RuleContext.prototype.getSourceInterval = function() {
  	return INVALID_INTERVAL;
  };

  RuleContext.prototype.getRuleContext = function() {
  	return this;
  };

  RuleContext.prototype.getPayload = function() {
  	return this;
  };

  // Return the combined text of all child nodes. This method only considers
  // tokens which have been added to the parse tree.
  // <p>
  // Since tokens on hidden channels (e.g. whitespace or comments) are not
  // added to the parse trees, they will not appear in the output of this
  // method.
  // /
  RuleContext.prototype.getText = function() {
  	if (this.getChildCount() === 0) {
  		return "";
  	} else {
  		return this.children.map(function(child) {
  			return child.getText();
  		}).join("");
  	}
  };

  // For rule associated with this parse tree internal node, return
  // the outer alternative number used to match the input. Default
  // implementation does not compute nor store this alt num. Create
  // a subclass of ParserRuleContext with backing field and set
  // option contextSuperClass.
  // to set it.
  RuleContext.prototype.getAltNumber = function() { return INVALID_ALT_NUMBER; };

  // Set the outer alternative number for this context node. Default
  // implementation does nothing to avoid backing field overhead for
  // trees that don't need it.  Create
  // a subclass of ParserRuleContext with backing field and set
  // option contextSuperClass.
  RuleContext.prototype.setAltNumber = function(altNumber) { };

  RuleContext.prototype.getChild = function(i) {
  	return null;
  };

  RuleContext.prototype.getChildCount = function() {
  	return 0;
  };

  RuleContext.prototype.accept = function(visitor) {
  	return visitor.visitChildren(this);
  };

  //need to manage circular dependencies, so export now
  exports.RuleContext = RuleContext;
  var Trees = require('./tree/Trees').Trees;


  // Print out a whole tree, not just a node, in LISP format
  // (root child1 .. childN). Print just a node if this is a leaf.
  //

  RuleContext.prototype.toStringTree = function(ruleNames, recog) {
  	return Trees.toStringTree(this, ruleNames, recog);
  };

  RuleContext.prototype.toString = function(ruleNames, stop) {
  	ruleNames = ruleNames || null;
  	stop = stop || null;
  	var p = this;
  	var s = "[";
  	while (p !== null && p !== stop) {
  		if (ruleNames === null) {
  			if (!p.isEmpty()) {
  				s += p.invokingState;
  			}
  		} else {
  			var ri = p.ruleIndex;
  			var ruleName = (ri >= 0 && ri < ruleNames.length) ? ruleNames[ri]
  					: "" + ri;
  			s += ruleName;
  		}
  		if (p.parentCtx !== null && (ruleNames !== null || !p.parentCtx.isEmpty())) {
  			s += " ";
  		}
  		p = p.parentCtx;
  	}
  	s += "]";
  	return s;
  };


  },{"./atn/ATN":16,"./tree/Tree":44,"./tree/Trees":45}],14:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  //

  // A token has properties: text, type, line, character position in the line
  // (so we can ignore tabs), token channel, index, and source from which
  // we obtained this token.

  function Token() {
  	this.source = null;
  	this.type = null; // token type of the token
  	this.channel = null; // The parser ignores everything not on DEFAULT_CHANNEL
  	this.start = null; // optional; return -1 if not implemented.
  	this.stop = null; // optional; return -1 if not implemented.
  	this.tokenIndex = null; // from 0..n-1 of the token object in the input stream
  	this.line = null; // line=1..n of the 1st character
  	this.column = null; // beginning of the line at which it occurs, 0..n-1
  	this._text = null; // text of the token.
  	return this;
  }

  Token.INVALID_TYPE = 0;

  // During lookahead operations, this "token" signifies we hit rule end ATN state
  // and did not follow it despite needing to.
  Token.EPSILON = -2;

  Token.MIN_USER_TOKEN_TYPE = 1;

  Token.EOF = -1;

  // All tokens go to the parser (unless skip() is called in that rule)
  // on a particular "channel". The parser tunes to a particular channel
  // so that whitespace etc... can go to the parser on a "hidden" channel.

  Token.DEFAULT_CHANNEL = 0;

  // Anything on different channel than DEFAULT_CHANNEL is not parsed
  // by parser.

  Token.HIDDEN_CHANNEL = 1;

  // Explicitly set the text for this token. If {code text} is not
  // {@code null}, then {@link //getText} will return this value rather than
  // extracting the text from the input.
  //
  // @param text The explicit text of the token, or {@code null} if the text
  // should be obtained from the input along with the start and stop indexes
  // of the token.

  Object.defineProperty(Token.prototype, "text", {
  	get : function() {
  		return this._text;
  	},
  	set : function(text) {
  		this._text = text;
  	}
  });

  Token.prototype.getTokenSource = function() {
  	return this.source[0];
  };

  Token.prototype.getInputStream = function() {
  	return this.source[1];
  };

  function CommonToken(source, type, channel, start, stop) {
  	Token.call(this);
  	this.source = source !== undefined ? source : CommonToken.EMPTY_SOURCE;
  	this.type = type !== undefined ? type : null;
  	this.channel = channel !== undefined ? channel : Token.DEFAULT_CHANNEL;
  	this.start = start !== undefined ? start : -1;
  	this.stop = stop !== undefined ? stop : -1;
  	this.tokenIndex = -1;
  	if (this.source[0] !== null) {
  		this.line = source[0].line;
  		this.column = source[0].column;
  	} else {
  		this.column = -1;
  	}
  	return this;
  }

  CommonToken.prototype = Object.create(Token.prototype);
  CommonToken.prototype.constructor = CommonToken;

  // An empty {@link Pair} which is used as the default value of
  // {@link //source} for tokens that do not have a source.
  CommonToken.EMPTY_SOURCE = [ null, null ];

  // Constructs a new {@link CommonToken} as a copy of another {@link Token}.
  //
  // <p>
  // If {@code oldToken} is also a {@link CommonToken} instance, the newly
  // constructed token will share a reference to the {@link //text} field and
  // the {@link Pair} stored in {@link //source}. Otherwise, {@link //text} will
  // be assigned the result of calling {@link //getText}, and {@link //source}
  // will be constructed from the result of {@link Token//getTokenSource} and
  // {@link Token//getInputStream}.</p>
  //
  // @param oldToken The token to copy.
  //
  CommonToken.prototype.clone = function() {
  	var t = new CommonToken(this.source, this.type, this.channel, this.start,
  			this.stop);
  	t.tokenIndex = this.tokenIndex;
  	t.line = this.line;
  	t.column = this.column;
  	t.text = this.text;
  	return t;
  };

  Object.defineProperty(CommonToken.prototype, "text", {
  	get : function() {
  		if (this._text !== null) {
  			return this._text;
  		}
  		var input = this.getInputStream();
  		if (input === null) {
  			return null;
  		}
  		var n = input.size;
  		if (this.start < n && this.stop < n) {
  			return input.getText(this.start, this.stop);
  		} else {
  			return "<EOF>";
  		}
  	},
  	set : function(text) {
  		this._text = text;
  	}
  });

  CommonToken.prototype.toString = function() {
  	var txt = this.text;
  	if (txt !== null) {
  		txt = txt.replace(/\n/g, "\\n").replace(/\r/g, "\\r").replace(/\t/g, "\\t");
  	} else {
  		txt = "<no text>";
  	}
  	return "[@" + this.tokenIndex + "," + this.start + ":" + this.stop + "='" +
  			txt + "',<" + this.type + ">" +
  			(this.channel > 0 ? ",channel=" + this.channel : "") + "," +
  			this.line + ":" + this.column + "]";
  };

  exports.Token = Token;
  exports.CommonToken = CommonToken;

  },{}],15:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */

  function arrayToString(a) {
      return "[" + a.join(", ") + "]";
  }

  String.prototype.seed = String.prototype.seed || Math.round(Math.random() * Math.pow(2, 32));

  String.prototype.hashCode = function () {
      var remainder, bytes, h1, h1b, c1, c2, k1, i,
          key = this.toString();

      remainder = key.length & 3; // key.length % 4
      bytes = key.length - remainder;
      h1 = String.prototype.seed;
      c1 = 0xcc9e2d51;
      c2 = 0x1b873593;
      i = 0;

      while (i < bytes) {
          k1 =
              ((key.charCodeAt(i) & 0xff)) |
              ((key.charCodeAt(++i) & 0xff) << 8) |
              ((key.charCodeAt(++i) & 0xff) << 16) |
              ((key.charCodeAt(++i) & 0xff) << 24);
          ++i;

          k1 = ((((k1 & 0xffff) * c1) + ((((k1 >>> 16) * c1) & 0xffff) << 16))) & 0xffffffff;
          k1 = (k1 << 15) | (k1 >>> 17);
          k1 = ((((k1 & 0xffff) * c2) + ((((k1 >>> 16) * c2) & 0xffff) << 16))) & 0xffffffff;

          h1 ^= k1;
          h1 = (h1 << 13) | (h1 >>> 19);
          h1b = ((((h1 & 0xffff) * 5) + ((((h1 >>> 16) * 5) & 0xffff) << 16))) & 0xffffffff;
          h1 = (((h1b & 0xffff) + 0x6b64) + ((((h1b >>> 16) + 0xe654) & 0xffff) << 16));
      }

      k1 = 0;

      switch (remainder) {
          case 3:
              k1 ^= (key.charCodeAt(i + 2) & 0xff) << 16;
          case 2:
              k1 ^= (key.charCodeAt(i + 1) & 0xff) << 8;
          case 1:
              k1 ^= (key.charCodeAt(i) & 0xff);

              k1 = (((k1 & 0xffff) * c1) + ((((k1 >>> 16) * c1) & 0xffff) << 16)) & 0xffffffff;
              k1 = (k1 << 15) | (k1 >>> 17);
              k1 = (((k1 & 0xffff) * c2) + ((((k1 >>> 16) * c2) & 0xffff) << 16)) & 0xffffffff;
              h1 ^= k1;
      }

      h1 ^= key.length;

      h1 ^= h1 >>> 16;
      h1 = (((h1 & 0xffff) * 0x85ebca6b) + ((((h1 >>> 16) * 0x85ebca6b) & 0xffff) << 16)) & 0xffffffff;
      h1 ^= h1 >>> 13;
      h1 = ((((h1 & 0xffff) * 0xc2b2ae35) + ((((h1 >>> 16) * 0xc2b2ae35) & 0xffff) << 16))) & 0xffffffff;
      h1 ^= h1 >>> 16;

      return h1 >>> 0;
  };

  function standardEqualsFunction(a, b) {
      return a.equals(b);
  }

  function standardHashCodeFunction(a) {
      return a.hashCode();
  }

  function Set(hashFunction, equalsFunction) {
      this.data = {};
      this.hashFunction = hashFunction || standardHashCodeFunction;
      this.equalsFunction = equalsFunction || standardEqualsFunction;
      return this;
  }

  Object.defineProperty(Set.prototype, "length", {
      get: function () {
          var l = 0;
          for (var key in this.data) {
              if (key.indexOf("hash_") === 0) {
                  l = l + this.data[key].length;
              }
          }
          return l;
      }
  });

  Set.prototype.add = function (value) {
      var hash = this.hashFunction(value);
      var key = "hash_" + hash;
      if (key in this.data) {
          var values = this.data[key];
          for (var i = 0; i < values.length; i++) {
              if (this.equalsFunction(value, values[i])) {
                  return values[i];
              }
          }
          values.push(value);
          return value;
      } else {
          this.data[key] = [value];
          return value;
      }
  };

  Set.prototype.contains = function (value) {
      return this.get(value) != null;
  };

  Set.prototype.get = function (value) {
      var hash = this.hashFunction(value);
      var key = "hash_" + hash;
      if (key in this.data) {
          var values = this.data[key];
          for (var i = 0; i < values.length; i++) {
              if (this.equalsFunction(value, values[i])) {
                  return values[i];
              }
          }
      }
      return null;
  };

  Set.prototype.values = function () {
      var l = [];
      for (var key in this.data) {
          if (key.indexOf("hash_") === 0) {
              l = l.concat(this.data[key]);
          }
      }
      return l;
  };

  Set.prototype.toString = function () {
      return arrayToString(this.values());
  };

  function BitSet() {
      this.data = [];
      return this;
  }

  BitSet.prototype.add = function (value) {
      this.data[value] = true;
  };

  BitSet.prototype.or = function (set) {
      var bits = this;
      Object.keys(set.data).map(function (alt) {
          bits.add(alt);
      });
  };

  BitSet.prototype.remove = function (value) {
      delete this.data[value];
  };

  BitSet.prototype.contains = function (value) {
      return this.data[value] === true;
  };

  BitSet.prototype.values = function () {
      return Object.keys(this.data);
  };

  BitSet.prototype.minValue = function () {
      return Math.min.apply(null, this.values());
  };

  BitSet.prototype.hashCode = function () {
      var hash = new Hash();
      hash.update(this.values());
      return hash.finish();
  };

  BitSet.prototype.equals = function (other) {
      if (!(other instanceof BitSet)) {
          return false;
      }
      return this.hashCode() === other.hashCode();
  };

  Object.defineProperty(BitSet.prototype, "length", {
      get: function () {
          return this.values().length;
      }
  });

  BitSet.prototype.toString = function () {
      return "{" + this.values().join(", ") + "}";
  };

  function Map(hashFunction, equalsFunction) {
      this.data = {};
      this.hashFunction = hashFunction || standardHashCodeFunction;
      this.equalsFunction = equalsFunction || standardEqualsFunction;
      return this;
  }

  Object.defineProperty(Map.prototype, "length", {
      get: function () {
          var l = 0;
          for (var hashKey in this.data) {
              if (hashKey.indexOf("hash_") === 0) {
                  l = l + this.data[hashKey].length;
              }
          }
          return l;
      }
  });

  Map.prototype.put = function (key, value) {
      var hashKey = "hash_" + this.hashFunction(key);
      if (hashKey in this.data) {
          var entries = this.data[hashKey];
          for (var i = 0; i < entries.length; i++) {
              var entry = entries[i];
              if (this.equalsFunction(key, entry.key)) {
                  var oldValue = entry.value;
                  entry.value = value;
                  return oldValue;
              }
          }
          entries.push({key:key, value:value});
          return value;
      } else {
          this.data[hashKey] = [{key:key, value:value}];
          return value;
      }
  };

  Map.prototype.containsKey = function (key) {
      var hashKey = "hash_" + this.hashFunction(key);
      if(hashKey in this.data) {
          var entries = this.data[hashKey];
          for (var i = 0; i < entries.length; i++) {
              var entry = entries[i];
              if (this.equalsFunction(key, entry.key))
                  return true;
          }
      }
      return false;
  };

  Map.prototype.get = function (key) {
      var hashKey = "hash_" + this.hashFunction(key);
      if(hashKey in this.data) {
          var entries = this.data[hashKey];
          for (var i = 0; i < entries.length; i++) {
              var entry = entries[i];
              if (this.equalsFunction(key, entry.key))
                  return entry.value;
          }
      }
      return null;
  };

  Map.prototype.entries = function () {
      var l = [];
      for (var key in this.data) {
          if (key.indexOf("hash_") === 0) {
              l = l.concat(this.data[key]);
          }
      }
      return l;
  };


  Map.prototype.getKeys = function () {
      return this.entries().map(function(e) {
          return e.key;
      });
  };


  Map.prototype.getValues = function () {
      return this.entries().map(function(e) {
              return e.value;
      });
  };


  Map.prototype.toString = function () {
      var ss = this.entries().map(function(entry) {
          return '{' + entry.key + ':' + entry.value + '}';
      });
      return '[' + ss.join(", ") + ']';
  };


  function AltDict() {
      this.data = {};
      return this;
  }


  AltDict.prototype.get = function (key) {
      key = "k-" + key;
      if (key in this.data) {
          return this.data[key];
      } else {
          return null;
      }
  };

  AltDict.prototype.put = function (key, value) {
      key = "k-" + key;
      this.data[key] = value;
  };

  AltDict.prototype.values = function () {
      var data = this.data;
      var keys = Object.keys(this.data);
      return keys.map(function (key) {
          return data[key];
      });
  };

  function DoubleDict() {
      return this;
  }

  function Hash() {
      this.count = 0;
      this.hash = 0;
      return this;
  }

  Hash.prototype.update = function () {
      for(var i=0;i<arguments.length;i++) {
          var value = arguments[i];
          if (value == null)
              continue;
          if(Array.isArray(value))
              this.update.apply(value);
          else {
              var k = 0;
              switch (typeof(value)) {
                  case 'undefined':
                  case 'function':
                      continue;
                  case 'number':
                  case 'boolean':
                      k = value;
                      break;
                  case 'string':
                      k = value.hashCode();
                      break;
                  default:
                      value.updateHashCode(this);
                      continue;
              }
              k = k * 0xCC9E2D51;
              k = (k << 15) | (k >>> (32 - 15));
              k = k * 0x1B873593;
              this.count = this.count + 1;
              var hash = this.hash ^ k;
              hash = (hash << 13) | (hash >>> (32 - 13));
              hash = hash * 5 + 0xE6546B64;
              this.hash = hash;
          }
      }
  };

  Hash.prototype.finish = function () {
      var hash = this.hash ^ (this.count * 4);
      hash = hash ^ (hash >>> 16);
      hash = hash * 0x85EBCA6B;
      hash = hash ^ (hash >>> 13);
      hash = hash * 0xC2B2AE35;
      hash = hash ^ (hash >>> 16);
      return hash;
  };

  function hashStuff() {
      var hash = new Hash();
      hash.update.apply(arguments);
      return hash.finish();
  }

  DoubleDict.prototype.get = function (a, b) {
      var d = this[a] || null;
      return d === null ? null : (d[b] || null);
  };

  DoubleDict.prototype.set = function (a, b, o) {
      var d = this[a] || null;
      if (d === null) {
          d = {};
          this[a] = d;
      }
      d[b] = o;
  };


  function escapeWhitespace(s, escapeSpaces) {
      s = s.replace("\t", "\\t");
      s = s.replace("\n", "\\n");
      s = s.replace("\r", "\\r");
      if (escapeSpaces) {
          s = s.replace(" ", "\u00B7");
      }
      return s;
  }

  function titleCase(str) {
      return str.replace(/\w\S*/g, function (txt) {
          return txt.charAt(0).toUpperCase() + txt.substr(1);
      });
  }
  function equalArrays(a, b)
  {
      if (!Array.isArray(a) || !Array.isArray(b))
          return false;
      if (a == b)
          return true;
      if (a.length != b.length)
          return false;
      for (var i = 0; i < a.length; i++) {
          if (a[i] == b[i])
              continue;
          if (!a[i].equals(b[i]))
              return false;
      }
      return true;
  }
  exports.Hash = Hash;
  exports.Set = Set;
  exports.Map = Map;
  exports.BitSet = BitSet;
  exports.AltDict = AltDict;
  exports.DoubleDict = DoubleDict;
  exports.hashStuff = hashStuff;
  exports.escapeWhitespace = escapeWhitespace;
  exports.arrayToString = arrayToString;
  exports.titleCase = titleCase;
  exports.equalArrays = equalArrays;
  },{}],16:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */

  var LL1Analyzer = require('./../LL1Analyzer').LL1Analyzer;
  var IntervalSet = require('./../IntervalSet').IntervalSet;

  function ATN(grammarType , maxTokenType) {

      // Used for runtime deserialization of ATNs from strings///
      // The type of the ATN.
      this.grammarType = grammarType;
      // The maximum value for any symbol recognized by a transition in the ATN.
      this.maxTokenType = maxTokenType;
      this.states = [];
      // Each subrule/rule is a decision point and we must track them so we
      //  can go back later and build DFA predictors for them.  This includes
      //  all the rules, subrules, optional blocks, ()+, ()* etc...
      this.decisionToState = [];
      // Maps from rule index to starting state number.
      this.ruleToStartState = [];
      // Maps from rule index to stop state number.
      this.ruleToStopState = null;
      this.modeNameToStartState = {};
      // For lexer ATNs, this maps the rule index to the resulting token type.
      // For parser ATNs, this maps the rule index to the generated bypass token
      // type if the
      // {@link ATNDeserializationOptions//isGenerateRuleBypassTransitions}
      // deserialization option was specified; otherwise, this is {@code null}.
      this.ruleToTokenType = null;
      // For lexer ATNs, this is an array of {@link LexerAction} objects which may
      // be referenced by action transitions in the ATN.
      this.lexerActions = null;
      this.modeToStartState = [];

      return this;
  }

  // Compute the set of valid tokens that can occur starting in state {@code s}.
  //  If {@code ctx} is null, the set of tokens will not include what can follow
  //  the rule surrounding {@code s}. In other words, the set will be
  //  restricted to tokens reachable staying within {@code s}'s rule.
  ATN.prototype.nextTokensInContext = function(s, ctx) {
      var anal = new LL1Analyzer(this);
      return anal.LOOK(s, null, ctx);
  };

  // Compute the set of valid tokens that can occur starting in {@code s} and
  // staying in same rule. {@link Token//EPSILON} is in set if we reach end of
  // rule.
  ATN.prototype.nextTokensNoContext = function(s) {
      if (s.nextTokenWithinRule !== null ) {
          return s.nextTokenWithinRule;
      }
      s.nextTokenWithinRule = this.nextTokensInContext(s, null);
      s.nextTokenWithinRule.readOnly = true;
      return s.nextTokenWithinRule;
  };

  ATN.prototype.nextTokens = function(s, ctx) {
      if ( ctx===undefined ) {
          return this.nextTokensNoContext(s);
      } else {
          return this.nextTokensInContext(s, ctx);
      }
  };

  ATN.prototype.addState = function( state) {
      if ( state !== null ) {
          state.atn = this;
          state.stateNumber = this.states.length;
      }
      this.states.push(state);
  };

  ATN.prototype.removeState = function( state) {
      this.states[state.stateNumber] = null; // just free mem, don't shift states in list
  };

  ATN.prototype.defineDecisionState = function( s) {
      this.decisionToState.push(s);
      s.decision = this.decisionToState.length-1;
      return s.decision;
  };

  ATN.prototype.getDecisionState = function( decision) {
      if (this.decisionToState.length===0) {
          return null;
      } else {
          return this.decisionToState[decision];
      }
  };

  // Computes the set of input symbols which could follow ATN state number
  // {@code stateNumber} in the specified full {@code context}. This method
  // considers the complete parser context, but does not evaluate semantic
  // predicates (i.e. all predicates encountered during the calculation are
  // assumed true). If a path in the ATN exists from the starting state to the
  // {@link RuleStopState} of the outermost context without matching any
  // symbols, {@link Token//EOF} is added to the returned set.
  //
  // <p>If {@code context} is {@code null}, it is treated as
  // {@link ParserRuleContext//EMPTY}.</p>
  //
  // @param stateNumber the ATN state number
  // @param context the full parse context
  // @return The set of potentially valid input symbols which could follow the
  // specified state in the specified context.
  // @throws IllegalArgumentException if the ATN does not contain a state with
  // number {@code stateNumber}
  var Token = require('./../Token').Token;

  ATN.prototype.getExpectedTokens = function( stateNumber, ctx ) {
      if ( stateNumber < 0 || stateNumber >= this.states.length ) {
          throw("Invalid state number.");
      }
      var s = this.states[stateNumber];
      var following = this.nextTokens(s);
      if (!following.contains(Token.EPSILON)) {
          return following;
      }
      var expected = new IntervalSet();
      expected.addSet(following);
      expected.removeOne(Token.EPSILON);
      while (ctx !== null && ctx.invokingState >= 0 && following.contains(Token.EPSILON)) {
          var invokingState = this.states[ctx.invokingState];
          var rt = invokingState.transitions[0];
          following = this.nextTokens(rt.followState);
          expected.addSet(following);
          expected.removeOne(Token.EPSILON);
          ctx = ctx.parentCtx;
      }
      if (following.contains(Token.EPSILON)) {
          expected.addOne(Token.EOF);
      }
      return expected;
  };

  ATN.INVALID_ALT_NUMBER = 0;

  exports.ATN = ATN;
  },{"./../IntervalSet":6,"./../LL1Analyzer":7,"./../Token":14}],17:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  ///

  // A tuple: (ATN state, predicted alt, syntactic, semantic context).
  //  The syntactic context is a graph-structured stack node whose
  //  path(s) to the root is the rule invocation(s)
  //  chain used to arrive at the state.  The semantic context is
  //  the tree of semantic predicates encountered before reaching
  //  an ATN state.
  ///

  var DecisionState = require('./ATNState').DecisionState;
  var SemanticContext = require('./SemanticContext').SemanticContext;
  var Hash = require("../Utils").Hash;


  function checkParams(params, isCfg) {
  	if(params===null) {
  		var result = { state:null, alt:null, context:null, semanticContext:null };
  		if(isCfg) {
  			result.reachesIntoOuterContext = 0;
  		}
  		return result;
  	} else {
  		var props = {};
  		props.state = params.state || null;
  		props.alt = (params.alt === undefined) ? null : params.alt;
  		props.context = params.context || null;
  		props.semanticContext = params.semanticContext || null;
  		if(isCfg) {
  			props.reachesIntoOuterContext = params.reachesIntoOuterContext || 0;
  			props.precedenceFilterSuppressed = params.precedenceFilterSuppressed || false;
  		}
  		return props;
  	}
  }

  function ATNConfig(params, config) {
  	this.checkContext(params, config);
  	params = checkParams(params);
  	config = checkParams(config, true);
      // The ATN state associated with this configuration///
      this.state = params.state!==null ? params.state : config.state;
      // What alt (or lexer rule) is predicted by this configuration///
      this.alt = params.alt!==null ? params.alt : config.alt;
      // The stack of invoking states leading to the rule/states associated
      //  with this config.  We track only those contexts pushed during
      //  execution of the ATN simulator.
      this.context = params.context!==null ? params.context : config.context;
      this.semanticContext = params.semanticContext!==null ? params.semanticContext :
          (config.semanticContext!==null ? config.semanticContext : SemanticContext.NONE);
      // We cannot execute predicates dependent upon local context unless
      // we know for sure we are in the correct context. Because there is
      // no way to do this efficiently, we simply cannot evaluate
      // dependent predicates unless we are in the rule that initially
      // invokes the ATN simulator.
      //
      // closure() tracks the depth of how far we dip into the
      // outer context: depth &gt; 0.  Note that it may not be totally
      // accurate depth since I don't ever decrement. TODO: make it a boolean then
      this.reachesIntoOuterContext = config.reachesIntoOuterContext;
      this.precedenceFilterSuppressed = config.precedenceFilterSuppressed;
      return this;
  }

  ATNConfig.prototype.checkContext = function(params, config) {
  	if((params.context===null || params.context===undefined) &&
  			(config===null || config.context===null || config.context===undefined)) {
  		this.context = null;
  	}
  };


  ATNConfig.prototype.hashCode = function() {
      var hash = new Hash();
      this.updateHashCode(hash);
      return hash.finish();
  };


  ATNConfig.prototype.updateHashCode = function(hash) {
      hash.update(this.state.stateNumber, this.alt, this.context, this.semanticContext);
  };

  // An ATN configuration is equal to another if both have
  //  the same state, they predict the same alternative, and
  //  syntactic/semantic contexts are the same.

  ATNConfig.prototype.equals = function(other) {
      if (this === other) {
          return true;
      } else if (! (other instanceof ATNConfig)) {
          return false;
      } else {
          return this.state.stateNumber===other.state.stateNumber &&
              this.alt===other.alt &&
              (this.context===null ? other.context===null : this.context.equals(other.context)) &&
              this.semanticContext.equals(other.semanticContext) &&
              this.precedenceFilterSuppressed===other.precedenceFilterSuppressed;
      }
  };


  ATNConfig.prototype.hashCodeForConfigSet = function() {
      var hash = new Hash();
      hash.update(this.state.stateNumber, this.alt, this.semanticContext);
      return hash.finish();
  };


  ATNConfig.prototype.equalsForConfigSet = function(other) {
      if (this === other) {
          return true;
      } else if (! (other instanceof ATNConfig)) {
          return false;
      } else {
          return this.state.stateNumber===other.state.stateNumber &&
              this.alt===other.alt &&
              this.semanticContext.equals(other.semanticContext);
      }
  };


  ATNConfig.prototype.toString = function() {
      return "(" + this.state + "," + this.alt +
          (this.context!==null ? ",[" + this.context.toString() + "]" : "") +
          (this.semanticContext !== SemanticContext.NONE ?
                  ("," + this.semanticContext.toString())
                  : "") +
          (this.reachesIntoOuterContext>0 ?
                  (",up=" + this.reachesIntoOuterContext)
                  : "") + ")";
  };


  function LexerATNConfig(params, config) {
  	ATNConfig.call(this, params, config);

      // This is the backing field for {@link //getLexerActionExecutor}.
  	var lexerActionExecutor = params.lexerActionExecutor || null;
      this.lexerActionExecutor = lexerActionExecutor || (config!==null ? config.lexerActionExecutor : null);
      this.passedThroughNonGreedyDecision = config!==null ? this.checkNonGreedyDecision(config, this.state) : false;
      return this;
  }

  LexerATNConfig.prototype = Object.create(ATNConfig.prototype);
  LexerATNConfig.prototype.constructor = LexerATNConfig;

  LexerATNConfig.prototype.updateHashCode = function(hash) {
      hash.update(this.state.stateNumber, this.alt, this.context, this.semanticContext, this.passedThroughNonGreedyDecision, this.lexerActionExecutor);
  };

  LexerATNConfig.prototype.equals = function(other) {
      return this === other ||
              (other instanceof LexerATNConfig &&
              this.passedThroughNonGreedyDecision == other.passedThroughNonGreedyDecision &&
              (this.lexerActionExecutor ? this.lexerActionExecutor.equals(other.lexerActionExecutor) : !other.lexerActionExecutor) &&
              ATNConfig.prototype.equals.call(this, other));
  };

  LexerATNConfig.prototype.hashCodeForConfigSet = LexerATNConfig.prototype.hashCode;

  LexerATNConfig.prototype.equalsForConfigSet = LexerATNConfig.prototype.equals;


  LexerATNConfig.prototype.checkNonGreedyDecision = function(source, target) {
      return source.passedThroughNonGreedyDecision ||
          (target instanceof DecisionState) && target.nonGreedy;
  };

  exports.ATNConfig = ATNConfig;
  exports.LexerATNConfig = LexerATNConfig;
  },{"../Utils":15,"./ATNState":22,"./SemanticContext":29}],18:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */

  //
  // Specialized {@link Set}{@code <}{@link ATNConfig}{@code >} that can track
  // info about the set, with support for combining similar configurations using a
  // graph-structured stack.
  ///

  var ATN = require('./ATN').ATN;
  var Utils = require('./../Utils');
  var Hash = Utils.Hash;
  var Set = Utils.Set;
  var SemanticContext = require('./SemanticContext').SemanticContext;
  var merge = require('./../PredictionContext').merge;

  function hashATNConfig(c) {
  	return c.hashCodeForConfigSet();
  }

  function equalATNConfigs(a, b) {
  	if ( a===b ) {
  		return true;
  	} else if ( a===null || b===null ) {
  		return false;
  	} else
         return a.equalsForConfigSet(b);
   }


  function ATNConfigSet(fullCtx) {
  	//
  	// The reason that we need this is because we don't want the hash map to use
  	// the standard hash code and equals. We need all configurations with the
  	// same
  	// {@code (s,i,_,semctx)} to be equal. Unfortunately, this key effectively
  	// doubles
  	// the number of objects associated with ATNConfigs. The other solution is
  	// to
  	// use a hash table that lets us specify the equals/hashcode operation.
  	// All configs but hashed by (s, i, _, pi) not including context. Wiped out
  	// when we go readonly as this set becomes a DFA state.
  	this.configLookup = new Set(hashATNConfig, equalATNConfigs);
  	// Indicates that this configuration set is part of a full context
  	// LL prediction. It will be used to determine how to merge $. With SLL
  	// it's a wildcard whereas it is not for LL context merge.
  	this.fullCtx = fullCtx === undefined ? true : fullCtx;
  	// Indicates that the set of configurations is read-only. Do not
  	// allow any code to manipulate the set; DFA states will point at
  	// the sets and they must not change. This does not protect the other
  	// fields; in particular, conflictingAlts is set after
  	// we've made this readonly.
  	this.readOnly = false;
  	// Track the elements as they are added to the set; supports get(i)///
  	this.configs = [];

  	// TODO: these fields make me pretty uncomfortable but nice to pack up info
  	// together, saves recomputation
  	// TODO: can we track conflicts as they are added to save scanning configs
  	// later?
  	this.uniqueAlt = 0;
  	this.conflictingAlts = null;

  	// Used in parser and lexer. In lexer, it indicates we hit a pred
  	// while computing a closure operation. Don't make a DFA state from this.
  	this.hasSemanticContext = false;
  	this.dipsIntoOuterContext = false;

  	this.cachedHashCode = -1;

  	return this;
  }

  // Adding a new config means merging contexts with existing configs for
  // {@code (s, i, pi, _)}, where {@code s} is the
  // {@link ATNConfig//state}, {@code i} is the {@link ATNConfig//alt}, and
  // {@code pi} is the {@link ATNConfig//semanticContext}. We use
  // {@code (s,i,pi)} as key.
  //
  // <p>This method updates {@link //dipsIntoOuterContext} and
  // {@link //hasSemanticContext} when necessary.</p>
  // /
  ATNConfigSet.prototype.add = function(config, mergeCache) {
  	if (mergeCache === undefined) {
  		mergeCache = null;
  	}
  	if (this.readOnly) {
  		throw "This set is readonly";
  	}
  	if (config.semanticContext !== SemanticContext.NONE) {
  		this.hasSemanticContext = true;
  	}
  	if (config.reachesIntoOuterContext > 0) {
  		this.dipsIntoOuterContext = true;
  	}
  	var existing = this.configLookup.add(config);
  	if (existing === config) {
  		this.cachedHashCode = -1;
  		this.configs.push(config); // track order here
  		return true;
  	}
  	// a previous (s,i,pi,_), merge with it and save result
  	var rootIsWildcard = !this.fullCtx;
  	var merged = merge(existing.context, config.context, rootIsWildcard, mergeCache);
  	// no need to check for existing.context, config.context in cache
  	// since only way to create new graphs is "call rule" and here. We
  	// cache at both places.
  	existing.reachesIntoOuterContext = Math.max( existing.reachesIntoOuterContext, config.reachesIntoOuterContext);
  	// make sure to preserve the precedence filter suppression during the merge
  	if (config.precedenceFilterSuppressed) {
  		existing.precedenceFilterSuppressed = true;
  	}
  	existing.context = merged; // replace context; no need to alt mapping
  	return true;
  };

  ATNConfigSet.prototype.getStates = function() {
  	var states = new Set();
  	for (var i = 0; i < this.configs.length; i++) {
  		states.add(this.configs[i].state);
  	}
  	return states;
  };

  ATNConfigSet.prototype.getPredicates = function() {
  	var preds = [];
  	for (var i = 0; i < this.configs.length; i++) {
  		var c = this.configs[i].semanticContext;
  		if (c !== SemanticContext.NONE) {
  			preds.push(c.semanticContext);
  		}
  	}
  	return preds;
  };

  Object.defineProperty(ATNConfigSet.prototype, "items", {
  	get : function() {
  		return this.configs;
  	}
  });

  ATNConfigSet.prototype.optimizeConfigs = function(interpreter) {
  	if (this.readOnly) {
  		throw "This set is readonly";
  	}
  	if (this.configLookup.length === 0) {
  		return;
  	}
  	for (var i = 0; i < this.configs.length; i++) {
  		var config = this.configs[i];
  		config.context = interpreter.getCachedContext(config.context);
  	}
  };

  ATNConfigSet.prototype.addAll = function(coll) {
  	for (var i = 0; i < coll.length; i++) {
  		this.add(coll[i]);
  	}
  	return false;
  };

  ATNConfigSet.prototype.equals = function(other) {
  	return this === other ||
  		(other instanceof ATNConfigSet &&
  		Utils.equalArrays(this.configs, other.configs) &&
  		this.fullCtx === other.fullCtx &&
  		this.uniqueAlt === other.uniqueAlt &&
  		this.conflictingAlts === other.conflictingAlts &&
  		this.hasSemanticContext === other.hasSemanticContext &&
  		this.dipsIntoOuterContext === other.dipsIntoOuterContext);
  };

  ATNConfigSet.prototype.hashCode = function() {
      var hash = new Hash();
      this.updateHashCode(hash);
      return hash.finish();
  };


  ATNConfigSet.prototype.updateHashCode = function(hash) {
  	if (this.readOnly) {
  		if (this.cachedHashCode === -1) {
              var hash = new Hash();
              hash.update(this.configs);
  			this.cachedHashCode = hash.finish();
  		}
          hash.update(this.cachedHashCode);
  	} else {
          hash.update(this.configs);
  	}
  };


  Object.defineProperty(ATNConfigSet.prototype, "length", {
  	get : function() {
  		return this.configs.length;
  	}
  });

  ATNConfigSet.prototype.isEmpty = function() {
  	return this.configs.length === 0;
  };

  ATNConfigSet.prototype.contains = function(item) {
  	if (this.configLookup === null) {
  		throw "This method is not implemented for readonly sets.";
  	}
  	return this.configLookup.contains(item);
  };

  ATNConfigSet.prototype.containsFast = function(item) {
  	if (this.configLookup === null) {
  		throw "This method is not implemented for readonly sets.";
  	}
  	return this.configLookup.containsFast(item);
  };

  ATNConfigSet.prototype.clear = function() {
  	if (this.readOnly) {
  		throw "This set is readonly";
  	}
  	this.configs = [];
  	this.cachedHashCode = -1;
  	this.configLookup = new Set();
  };

  ATNConfigSet.prototype.setReadonly = function(readOnly) {
  	this.readOnly = readOnly;
  	if (readOnly) {
  		this.configLookup = null; // can't mod, no need for lookup cache
  	}
  };

  ATNConfigSet.prototype.toString = function() {
  	return Utils.arrayToString(this.configs) +
  		(this.hasSemanticContext ? ",hasSemanticContext=" + this.hasSemanticContext : "") +
  		(this.uniqueAlt !== ATN.INVALID_ALT_NUMBER ? ",uniqueAlt=" + this.uniqueAlt : "") +
  		(this.conflictingAlts !== null ? ",conflictingAlts=" + this.conflictingAlts : "") +
  		(this.dipsIntoOuterContext ? ",dipsIntoOuterContext" : "");
  };

  function OrderedATNConfigSet() {
  	ATNConfigSet.call(this);
  	this.configLookup = new Set();
  	return this;
  }

  OrderedATNConfigSet.prototype = Object.create(ATNConfigSet.prototype);
  OrderedATNConfigSet.prototype.constructor = OrderedATNConfigSet;

  exports.ATNConfigSet = ATNConfigSet;
  exports.OrderedATNConfigSet = OrderedATNConfigSet;

  },{"./../PredictionContext":11,"./../Utils":15,"./ATN":16,"./SemanticContext":29}],19:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */

  function ATNDeserializationOptions(copyFrom) {
  	if(copyFrom===undefined) {
  		copyFrom = null;
  	}
  	this.readOnly = false;
      this.verifyATN = copyFrom===null ? true : copyFrom.verifyATN;
      this.generateRuleBypassTransitions = copyFrom===null ? false : copyFrom.generateRuleBypassTransitions;

      return this;
  }

  ATNDeserializationOptions.defaultOptions = new ATNDeserializationOptions();
  ATNDeserializationOptions.defaultOptions.readOnly = true;

  //    def __setattr__(self, key, value):
  //        if key!="readOnly" and self.readOnly:
  //            raise Exception("The object is read only.")
  //        super(type(self), self).__setattr__(key,value)

  exports.ATNDeserializationOptions = ATNDeserializationOptions;

  },{}],20:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */

  var Token = require('./../Token').Token;
  var ATN = require('./ATN').ATN;
  var ATNType = require('./ATNType').ATNType;
  var ATNStates = require('./ATNState');
  var ATNState = ATNStates.ATNState;
  var BasicState = ATNStates.BasicState;
  var DecisionState = ATNStates.DecisionState;
  var BlockStartState = ATNStates.BlockStartState;
  var BlockEndState = ATNStates.BlockEndState;
  var LoopEndState = ATNStates.LoopEndState;
  var RuleStartState = ATNStates.RuleStartState;
  var RuleStopState = ATNStates.RuleStopState;
  var TokensStartState = ATNStates.TokensStartState;
  var PlusLoopbackState = ATNStates.PlusLoopbackState;
  var StarLoopbackState = ATNStates.StarLoopbackState;
  var StarLoopEntryState = ATNStates.StarLoopEntryState;
  var PlusBlockStartState = ATNStates.PlusBlockStartState;
  var StarBlockStartState = ATNStates.StarBlockStartState;
  var BasicBlockStartState = ATNStates.BasicBlockStartState;
  var Transitions = require('./Transition');
  var Transition = Transitions.Transition;
  var AtomTransition = Transitions.AtomTransition;
  var SetTransition = Transitions.SetTransition;
  var NotSetTransition = Transitions.NotSetTransition;
  var RuleTransition = Transitions.RuleTransition;
  var RangeTransition = Transitions.RangeTransition;
  var ActionTransition = Transitions.ActionTransition;
  var EpsilonTransition = Transitions.EpsilonTransition;
  var WildcardTransition = Transitions.WildcardTransition;
  var PredicateTransition = Transitions.PredicateTransition;
  var PrecedencePredicateTransition = Transitions.PrecedencePredicateTransition;
  var IntervalSet = require('./../IntervalSet').IntervalSet;
  var Interval = require('./../IntervalSet').Interval;
  var ATNDeserializationOptions = require('./ATNDeserializationOptions').ATNDeserializationOptions;
  var LexerActions = require('./LexerAction');
  var LexerActionType = LexerActions.LexerActionType;
  var LexerSkipAction = LexerActions.LexerSkipAction;
  var LexerChannelAction = LexerActions.LexerChannelAction;
  var LexerCustomAction = LexerActions.LexerCustomAction;
  var LexerMoreAction = LexerActions.LexerMoreAction;
  var LexerTypeAction = LexerActions.LexerTypeAction;
  var LexerPushModeAction = LexerActions.LexerPushModeAction;
  var LexerPopModeAction = LexerActions.LexerPopModeAction;
  var LexerModeAction = LexerActions.LexerModeAction;
  // This is the earliest supported serialized UUID.
  // stick to serialized version for now, we don't need a UUID instance
  var BASE_SERIALIZED_UUID = "AADB8D7E-AEEF-4415-AD2B-8204D6CF042E";

  //
  // This UUID indicates the serialized ATN contains two sets of
  // IntervalSets, where the second set's values are encoded as
  // 32-bit integers to support the full Unicode SMP range up to U+10FFFF.
  //
  var ADDED_UNICODE_SMP = "59627784-3BE5-417A-B9EB-8131A7286089";

  // This list contains all of the currently supported UUIDs, ordered by when
  // the feature first appeared in this branch.
  var SUPPORTED_UUIDS = [ BASE_SERIALIZED_UUID, ADDED_UNICODE_SMP ];

  var SERIALIZED_VERSION = 3;

  // This is the current serialized UUID.
  var SERIALIZED_UUID = ADDED_UNICODE_SMP;

  function initArray( length, value) {
  	var tmp = [];
  	tmp[length-1] = value;
  	return tmp.map(function(i) {return value;});
  }

  function ATNDeserializer (options) {

      if ( options=== undefined || options === null ) {
          options = ATNDeserializationOptions.defaultOptions;
      }
      this.deserializationOptions = options;
      this.stateFactories = null;
      this.actionFactories = null;

      return this;
  }

  // Determines if a particular serialized representation of an ATN supports
  // a particular feature, identified by the {@link UUID} used for serializing
  // the ATN at the time the feature was first introduced.
  //
  // @param feature The {@link UUID} marking the first time the feature was
  // supported in the serialized ATN.
  // @param actualUuid The {@link UUID} of the actual serialized ATN which is
  // currently being deserialized.
  // @return {@code true} if the {@code actualUuid} value represents a
  // serialized ATN at or after the feature identified by {@code feature} was
  // introduced; otherwise, {@code false}.

  ATNDeserializer.prototype.isFeatureSupported = function(feature, actualUuid) {
      var idx1 = SUPPORTED_UUIDS.indexOf(feature);
      if (idx1<0) {
          return false;
      }
      var idx2 = SUPPORTED_UUIDS.indexOf(actualUuid);
      return idx2 >= idx1;
  };

  ATNDeserializer.prototype.deserialize = function(data) {
      this.reset(data);
      this.checkVersion();
      this.checkUUID();
      var atn = this.readATN();
      this.readStates(atn);
      this.readRules(atn);
      this.readModes(atn);
      var sets = [];
      // First, deserialize sets with 16-bit arguments <= U+FFFF.
      this.readSets(atn, sets, this.readInt.bind(this));
      // Next, if the ATN was serialized with the Unicode SMP feature,
      // deserialize sets with 32-bit arguments <= U+10FFFF.
      if (this.isFeatureSupported(ADDED_UNICODE_SMP, this.uuid)) {
          this.readSets(atn, sets, this.readInt32.bind(this));
      }
      this.readEdges(atn, sets);
      this.readDecisions(atn);
      this.readLexerActions(atn);
      this.markPrecedenceDecisions(atn);
      this.verifyATN(atn);
      if (this.deserializationOptions.generateRuleBypassTransitions && atn.grammarType === ATNType.PARSER ) {
          this.generateRuleBypassTransitions(atn);
          // re-verify after modification
          this.verifyATN(atn);
      }
      return atn;
  };

  ATNDeserializer.prototype.reset = function(data) {
  	var adjust = function(c) {
          var v = c.charCodeAt(0);
          return v>1  ? v-2 : -1;
  	};
      var temp = data.split("").map(adjust);
      // don't adjust the first value since that's the version number
      temp[0] = data.charCodeAt(0);
      this.data = temp;
      this.pos = 0;
  };

  ATNDeserializer.prototype.checkVersion = function() {
      var version = this.readInt();
      if ( version !== SERIALIZED_VERSION ) {
          throw ("Could not deserialize ATN with version " + version + " (expected " + SERIALIZED_VERSION + ").");
      }
  };

  ATNDeserializer.prototype.checkUUID = function() {
      var uuid = this.readUUID();
      if (SUPPORTED_UUIDS.indexOf(uuid)<0) {
          throw (SERIALIZED_UUID);
      }
      this.uuid = uuid;
  };

  ATNDeserializer.prototype.readATN = function() {
      var grammarType = this.readInt();
      var maxTokenType = this.readInt();
      return new ATN(grammarType, maxTokenType);
  };

  ATNDeserializer.prototype.readStates = function(atn) {
  	var j, pair, stateNumber;
      var loopBackStateNumbers = [];
      var endStateNumbers = [];
      var nstates = this.readInt();
      for(var i=0; i<nstates; i++) {
          var stype = this.readInt();
          // ignore bad type of states
          if (stype===ATNState.INVALID_TYPE) {
              atn.addState(null);
              continue;
          }
          var ruleIndex = this.readInt();
          if (ruleIndex === 0xFFFF) {
              ruleIndex = -1;
          }
          var s = this.stateFactory(stype, ruleIndex);
          if (stype === ATNState.LOOP_END) { // special case
              var loopBackStateNumber = this.readInt();
              loopBackStateNumbers.push([s, loopBackStateNumber]);
          } else if(s instanceof BlockStartState) {
              var endStateNumber = this.readInt();
              endStateNumbers.push([s, endStateNumber]);
          }
          atn.addState(s);
      }
      // delay the assignment of loop back and end states until we know all the
  	// state instances have been initialized
      for (j=0; j<loopBackStateNumbers.length; j++) {
          pair = loopBackStateNumbers[j];
          pair[0].loopBackState = atn.states[pair[1]];
      }

      for (j=0; j<endStateNumbers.length; j++) {
          pair = endStateNumbers[j];
          pair[0].endState = atn.states[pair[1]];
      }

      var numNonGreedyStates = this.readInt();
      for (j=0; j<numNonGreedyStates; j++) {
          stateNumber = this.readInt();
          atn.states[stateNumber].nonGreedy = true;
      }

      var numPrecedenceStates = this.readInt();
      for (j=0; j<numPrecedenceStates; j++) {
          stateNumber = this.readInt();
          atn.states[stateNumber].isPrecedenceRule = true;
      }
  };

  ATNDeserializer.prototype.readRules = function(atn) {
      var i;
      var nrules = this.readInt();
      if (atn.grammarType === ATNType.LEXER ) {
          atn.ruleToTokenType = initArray(nrules, 0);
      }
      atn.ruleToStartState = initArray(nrules, 0);
      for (i=0; i<nrules; i++) {
          var s = this.readInt();
          var startState = atn.states[s];
          atn.ruleToStartState[i] = startState;
          if ( atn.grammarType === ATNType.LEXER ) {
              var tokenType = this.readInt();
              if (tokenType === 0xFFFF) {
                  tokenType = Token.EOF;
              }
              atn.ruleToTokenType[i] = tokenType;
          }
      }
      atn.ruleToStopState = initArray(nrules, 0);
      for (i=0; i<atn.states.length; i++) {
          var state = atn.states[i];
          if (!(state instanceof RuleStopState)) {
              continue;
          }
          atn.ruleToStopState[state.ruleIndex] = state;
          atn.ruleToStartState[state.ruleIndex].stopState = state;
      }
  };

  ATNDeserializer.prototype.readModes = function(atn) {
      var nmodes = this.readInt();
      for (var i=0; i<nmodes; i++) {
          var s = this.readInt();
          atn.modeToStartState.push(atn.states[s]);
      }
  };

  ATNDeserializer.prototype.readSets = function(atn, sets, readUnicode) {
      var m = this.readInt();
      for (var i=0; i<m; i++) {
          var iset = new IntervalSet();
          sets.push(iset);
          var n = this.readInt();
          var containsEof = this.readInt();
          if (containsEof!==0) {
              iset.addOne(-1);
          }
          for (var j=0; j<n; j++) {
              var i1 = readUnicode();
              var i2 = readUnicode();
              iset.addRange(i1, i2);
          }
      }
  };

  ATNDeserializer.prototype.readEdges = function(atn, sets) {
  	var i, j, state, trans, target;
      var nedges = this.readInt();
      for (i=0; i<nedges; i++) {
          var src = this.readInt();
          var trg = this.readInt();
          var ttype = this.readInt();
          var arg1 = this.readInt();
          var arg2 = this.readInt();
          var arg3 = this.readInt();
          trans = this.edgeFactory(atn, ttype, src, trg, arg1, arg2, arg3, sets);
          var srcState = atn.states[src];
          srcState.addTransition(trans);
      }
      // edges for rule stop states can be derived, so they aren't serialized
      for (i=0; i<atn.states.length; i++) {
          state = atn.states[i];
          for (j=0; j<state.transitions.length; j++) {
              var t = state.transitions[j];
              if (!(t instanceof RuleTransition)) {
                  continue;
              }
  			var outermostPrecedenceReturn = -1;
  			if (atn.ruleToStartState[t.target.ruleIndex].isPrecedenceRule) {
  				if (t.precedence === 0) {
  					outermostPrecedenceReturn = t.target.ruleIndex;
  				}
  			}

  			trans = new EpsilonTransition(t.followState, outermostPrecedenceReturn);
              atn.ruleToStopState[t.target.ruleIndex].addTransition(trans);
          }
      }

      for (i=0; i<atn.states.length; i++) {
          state = atn.states[i];
          if (state instanceof BlockStartState) {
              // we need to know the end state to set its start state
              if (state.endState === null) {
                  throw ("IllegalState");
              }
              // block end states can only be associated to a single block start
  			// state
              if ( state.endState.startState !== null) {
                  throw ("IllegalState");
              }
              state.endState.startState = state;
          }
          if (state instanceof PlusLoopbackState) {
              for (j=0; j<state.transitions.length; j++) {
                  target = state.transitions[j].target;
                  if (target instanceof PlusBlockStartState) {
                      target.loopBackState = state;
                  }
              }
          } else if (state instanceof StarLoopbackState) {
              for (j=0; j<state.transitions.length; j++) {
                  target = state.transitions[j].target;
                  if (target instanceof StarLoopEntryState) {
                      target.loopBackState = state;
                  }
              }
          }
      }
  };

  ATNDeserializer.prototype.readDecisions = function(atn) {
      var ndecisions = this.readInt();
      for (var i=0; i<ndecisions; i++) {
          var s = this.readInt();
          var decState = atn.states[s];
          atn.decisionToState.push(decState);
          decState.decision = i;
      }
  };

  ATNDeserializer.prototype.readLexerActions = function(atn) {
      if (atn.grammarType === ATNType.LEXER) {
          var count = this.readInt();
          atn.lexerActions = initArray(count, null);
          for (var i=0; i<count; i++) {
              var actionType = this.readInt();
              var data1 = this.readInt();
              if (data1 === 0xFFFF) {
                  data1 = -1;
              }
              var data2 = this.readInt();
              if (data2 === 0xFFFF) {
                  data2 = -1;
              }
              var lexerAction = this.lexerActionFactory(actionType, data1, data2);
              atn.lexerActions[i] = lexerAction;
          }
      }
  };

  ATNDeserializer.prototype.generateRuleBypassTransitions = function(atn) {
  	var i;
      var count = atn.ruleToStartState.length;
      for(i=0; i<count; i++) {
          atn.ruleToTokenType[i] = atn.maxTokenType + i + 1;
      }
      for(i=0; i<count; i++) {
          this.generateRuleBypassTransition(atn, i);
      }
  };

  ATNDeserializer.prototype.generateRuleBypassTransition = function(atn, idx) {
  	var i, state;
      var bypassStart = new BasicBlockStartState();
      bypassStart.ruleIndex = idx;
      atn.addState(bypassStart);

      var bypassStop = new BlockEndState();
      bypassStop.ruleIndex = idx;
      atn.addState(bypassStop);

      bypassStart.endState = bypassStop;
      atn.defineDecisionState(bypassStart);

      bypassStop.startState = bypassStart;

      var excludeTransition = null;
      var endState = null;

      if (atn.ruleToStartState[idx].isPrecedenceRule) {
          // wrap from the beginning of the rule to the StarLoopEntryState
          endState = null;
          for(i=0; i<atn.states.length; i++) {
              state = atn.states[i];
              if (this.stateIsEndStateFor(state, idx)) {
                  endState = state;
                  excludeTransition = state.loopBackState.transitions[0];
                  break;
              }
          }
          if (excludeTransition === null) {
              throw ("Couldn't identify final state of the precedence rule prefix section.");
          }
      } else {
          endState = atn.ruleToStopState[idx];
      }

      // all non-excluded transitions that currently target end state need to
  	// target blockEnd instead
      for(i=0; i<atn.states.length; i++) {
          state = atn.states[i];
          for(var j=0; j<state.transitions.length; j++) {
              var transition = state.transitions[j];
              if (transition === excludeTransition) {
                  continue;
              }
              if (transition.target === endState) {
                  transition.target = bypassStop;
              }
          }
      }

      // all transitions leaving the rule start state need to leave blockStart
  	// instead
      var ruleToStartState = atn.ruleToStartState[idx];
      var count = ruleToStartState.transitions.length;
      while ( count > 0) {
          bypassStart.addTransition(ruleToStartState.transitions[count-1]);
          ruleToStartState.transitions = ruleToStartState.transitions.slice(-1);
      }
      // link the new states
      atn.ruleToStartState[idx].addTransition(new EpsilonTransition(bypassStart));
      bypassStop.addTransition(new EpsilonTransition(endState));

      var matchState = new BasicState();
      atn.addState(matchState);
      matchState.addTransition(new AtomTransition(bypassStop, atn.ruleToTokenType[idx]));
      bypassStart.addTransition(new EpsilonTransition(matchState));
  };

  ATNDeserializer.prototype.stateIsEndStateFor = function(state, idx) {
      if ( state.ruleIndex !== idx) {
          return null;
      }
      if (!( state instanceof StarLoopEntryState)) {
          return null;
      }
      var maybeLoopEndState = state.transitions[state.transitions.length - 1].target;
      if (!( maybeLoopEndState instanceof LoopEndState)) {
          return null;
      }
      if (maybeLoopEndState.epsilonOnlyTransitions &&
          (maybeLoopEndState.transitions[0].target instanceof RuleStopState)) {
          return state;
      } else {
          return null;
      }
  };

  //
  // Analyze the {@link StarLoopEntryState} states in the specified ATN to set
  // the {@link StarLoopEntryState//isPrecedenceDecision} field to the
  // correct value.
  //
  // @param atn The ATN.
  //
  ATNDeserializer.prototype.markPrecedenceDecisions = function(atn) {
  	for(var i=0; i<atn.states.length; i++) {
  		var state = atn.states[i];
  		if (!( state instanceof StarLoopEntryState)) {
              continue;
          }
          // We analyze the ATN to determine if this ATN decision state is the
          // decision for the closure block that determines whether a
          // precedence rule should continue or complete.
          //
          if ( atn.ruleToStartState[state.ruleIndex].isPrecedenceRule) {
              var maybeLoopEndState = state.transitions[state.transitions.length - 1].target;
              if (maybeLoopEndState instanceof LoopEndState) {
                  if ( maybeLoopEndState.epsilonOnlyTransitions &&
                          (maybeLoopEndState.transitions[0].target instanceof RuleStopState)) {
                      state.isPrecedenceDecision = true;
                  }
              }
          }
  	}
  };

  ATNDeserializer.prototype.verifyATN = function(atn) {
      if (!this.deserializationOptions.verifyATN) {
          return;
      }
      // verify assumptions
  	for(var i=0; i<atn.states.length; i++) {
          var state = atn.states[i];
          if (state === null) {
              continue;
          }
          this.checkCondition(state.epsilonOnlyTransitions || state.transitions.length <= 1);
          if (state instanceof PlusBlockStartState) {
              this.checkCondition(state.loopBackState !== null);
          } else  if (state instanceof StarLoopEntryState) {
              this.checkCondition(state.loopBackState !== null);
              this.checkCondition(state.transitions.length === 2);
              if (state.transitions[0].target instanceof StarBlockStartState) {
                  this.checkCondition(state.transitions[1].target instanceof LoopEndState);
                  this.checkCondition(!state.nonGreedy);
              } else if (state.transitions[0].target instanceof LoopEndState) {
                  this.checkCondition(state.transitions[1].target instanceof StarBlockStartState);
                  this.checkCondition(state.nonGreedy);
              } else {
                  throw("IllegalState");
              }
          } else if (state instanceof StarLoopbackState) {
              this.checkCondition(state.transitions.length === 1);
              this.checkCondition(state.transitions[0].target instanceof StarLoopEntryState);
          } else if (state instanceof LoopEndState) {
              this.checkCondition(state.loopBackState !== null);
          } else if (state instanceof RuleStartState) {
              this.checkCondition(state.stopState !== null);
          } else if (state instanceof BlockStartState) {
              this.checkCondition(state.endState !== null);
          } else if (state instanceof BlockEndState) {
              this.checkCondition(state.startState !== null);
          } else if (state instanceof DecisionState) {
              this.checkCondition(state.transitions.length <= 1 || state.decision >= 0);
          } else {
              this.checkCondition(state.transitions.length <= 1 || (state instanceof RuleStopState));
          }
  	}
  };

  ATNDeserializer.prototype.checkCondition = function(condition, message) {
      if (!condition) {
          if (message === undefined || message===null) {
              message = "IllegalState";
          }
          throw (message);
      }
  };

  ATNDeserializer.prototype.readInt = function() {
      return this.data[this.pos++];
  };

  ATNDeserializer.prototype.readInt32 = function() {
      var low = this.readInt();
      var high = this.readInt();
      return low | (high << 16);
  };

  ATNDeserializer.prototype.readLong = function() {
      var low = this.readInt32();
      var high = this.readInt32();
      return (low & 0x00000000FFFFFFFF) | (high << 32);
  };

  function createByteToHex() {
  	var bth = [];
  	for (var i = 0; i < 256; i++) {
  		bth[i] = (i + 0x100).toString(16).substr(1).toUpperCase();
  	}
  	return bth;
  }

  var byteToHex = createByteToHex();

  ATNDeserializer.prototype.readUUID = function() {
  	var bb = [];
  	for(var i=7;i>=0;i--) {
  		var int = this.readInt();
  		/* jshint bitwise: false */
  		bb[(2*i)+1] = int & 0xFF;
  		bb[2*i] = (int >> 8) & 0xFF;
  	}
      return byteToHex[bb[0]] + byteToHex[bb[1]] +
      byteToHex[bb[2]] + byteToHex[bb[3]] + '-' +
      byteToHex[bb[4]] + byteToHex[bb[5]] + '-' +
      byteToHex[bb[6]] + byteToHex[bb[7]] + '-' +
      byteToHex[bb[8]] + byteToHex[bb[9]] + '-' +
      byteToHex[bb[10]] + byteToHex[bb[11]] +
      byteToHex[bb[12]] + byteToHex[bb[13]] +
      byteToHex[bb[14]] + byteToHex[bb[15]];
  };

  ATNDeserializer.prototype.edgeFactory = function(atn, type, src, trg, arg1, arg2, arg3, sets) {
      var target = atn.states[trg];
      switch(type) {
      case Transition.EPSILON:
          return new EpsilonTransition(target);
      case Transition.RANGE:
          return arg3 !== 0 ? new RangeTransition(target, Token.EOF, arg2) : new RangeTransition(target, arg1, arg2);
      case Transition.RULE:
          return new RuleTransition(atn.states[arg1], arg2, arg3, target);
      case Transition.PREDICATE:
          return new PredicateTransition(target, arg1, arg2, arg3 !== 0);
      case Transition.PRECEDENCE:
          return new PrecedencePredicateTransition(target, arg1);
      case Transition.ATOM:
          return arg3 !== 0 ? new AtomTransition(target, Token.EOF) : new AtomTransition(target, arg1);
      case Transition.ACTION:
          return new ActionTransition(target, arg1, arg2, arg3 !== 0);
      case Transition.SET:
          return new SetTransition(target, sets[arg1]);
      case Transition.NOT_SET:
          return new NotSetTransition(target, sets[arg1]);
      case Transition.WILDCARD:
          return new WildcardTransition(target);
      default:
          throw "The specified transition type: " + type + " is not valid.";
      }
  };

  ATNDeserializer.prototype.stateFactory = function(type, ruleIndex) {
      if (this.stateFactories === null) {
          var sf = [];
          sf[ATNState.INVALID_TYPE] = null;
          sf[ATNState.BASIC] = function() { return new BasicState(); };
          sf[ATNState.RULE_START] = function() { return new RuleStartState(); };
          sf[ATNState.BLOCK_START] = function() { return new BasicBlockStartState(); };
          sf[ATNState.PLUS_BLOCK_START] = function() { return new PlusBlockStartState(); };
          sf[ATNState.STAR_BLOCK_START] = function() { return new StarBlockStartState(); };
          sf[ATNState.TOKEN_START] = function() { return new TokensStartState(); };
          sf[ATNState.RULE_STOP] = function() { return new RuleStopState(); };
          sf[ATNState.BLOCK_END] = function() { return new BlockEndState(); };
          sf[ATNState.STAR_LOOP_BACK] = function() { return new StarLoopbackState(); };
          sf[ATNState.STAR_LOOP_ENTRY] = function() { return new StarLoopEntryState(); };
          sf[ATNState.PLUS_LOOP_BACK] = function() { return new PlusLoopbackState(); };
          sf[ATNState.LOOP_END] = function() { return new LoopEndState(); };
          this.stateFactories = sf;
      }
      if (type>this.stateFactories.length || this.stateFactories[type] === null) {
          throw("The specified state type " + type + " is not valid.");
      } else {
          var s = this.stateFactories[type]();
          if (s!==null) {
              s.ruleIndex = ruleIndex;
              return s;
          }
      }
  };

  ATNDeserializer.prototype.lexerActionFactory = function(type, data1, data2) {
      if (this.actionFactories === null) {
          var af = [];
          af[LexerActionType.CHANNEL] = function(data1, data2) { return new LexerChannelAction(data1); };
          af[LexerActionType.CUSTOM] = function(data1, data2) { return new LexerCustomAction(data1, data2); };
          af[LexerActionType.MODE] = function(data1, data2) { return new LexerModeAction(data1); };
          af[LexerActionType.MORE] = function(data1, data2) { return LexerMoreAction.INSTANCE; };
          af[LexerActionType.POP_MODE] = function(data1, data2) { return LexerPopModeAction.INSTANCE; };
          af[LexerActionType.PUSH_MODE] = function(data1, data2) { return new LexerPushModeAction(data1); };
          af[LexerActionType.SKIP] = function(data1, data2) { return LexerSkipAction.INSTANCE; };
          af[LexerActionType.TYPE] = function(data1, data2) { return new LexerTypeAction(data1); };
          this.actionFactories = af;
      }
      if (type>this.actionFactories.length || this.actionFactories[type] === null) {
          throw("The specified lexer action type " + type + " is not valid.");
      } else {
          return this.actionFactories[type](data1, data2);
      }
  };


  exports.ATNDeserializer = ATNDeserializer;
  },{"./../IntervalSet":6,"./../Token":14,"./ATN":16,"./ATNDeserializationOptions":19,"./ATNState":22,"./ATNType":23,"./LexerAction":25,"./Transition":30}],21:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  ///

  var DFAState = require('./../dfa/DFAState').DFAState;
  var ATNConfigSet = require('./ATNConfigSet').ATNConfigSet;
  var getCachedPredictionContext = require('./../PredictionContext').getCachedPredictionContext;

  function ATNSimulator(atn, sharedContextCache) {

      // The context cache maps all PredictionContext objects that are ==
      //  to a single cached copy. This cache is shared across all contexts
      //  in all ATNConfigs in all DFA states.  We rebuild each ATNConfigSet
      //  to use only cached nodes/graphs in addDFAState(). We don't want to
      //  fill this during closure() since there are lots of contexts that
      //  pop up but are not used ever again. It also greatly slows down closure().
      //
      //  <p>This cache makes a huge difference in memory and a little bit in speed.
      //  For the Java grammar on java.*, it dropped the memory requirements
      //  at the end from 25M to 16M. We don't store any of the full context
      //  graphs in the DFA because they are limited to local context only,
      //  but apparently there's a lot of repetition there as well. We optimize
      //  the config contexts before storing the config set in the DFA states
      //  by literally rebuilding them with cached subgraphs only.</p>
      //
      //  <p>I tried a cache for use during closure operations, that was
      //  whacked after each adaptivePredict(). It cost a little bit
      //  more time I think and doesn't save on the overall footprint
      //  so it's not worth the complexity.</p>
      ///
      this.atn = atn;
      this.sharedContextCache = sharedContextCache;
      return this;
  }

  // Must distinguish between missing edge and edge we know leads nowhere///
  ATNSimulator.ERROR = new DFAState(0x7FFFFFFF, new ATNConfigSet());


  ATNSimulator.prototype.getCachedContext = function(context) {
      if (this.sharedContextCache ===null) {
          return context;
      }
      var visited = {};
      return getCachedPredictionContext(context, this.sharedContextCache, visited);
  };

  exports.ATNSimulator = ATNSimulator;

  },{"./../PredictionContext":11,"./../dfa/DFAState":34,"./ATNConfigSet":18}],22:[function(require,module,exports){

  function ATNState() {
      // Which ATN are we in?
      this.atn = null;
      this.stateNumber = ATNState.INVALID_STATE_NUMBER;
      this.stateType = null;
      this.ruleIndex = 0; // at runtime, we don't have Rule objects
      this.epsilonOnlyTransitions = false;
      // Track the transitions emanating from this ATN state.
      this.transitions = [];
      // Used to cache lookahead during parsing, not used during construction
      this.nextTokenWithinRule = null;
      return this;
  }

  // constants for serialization
  ATNState.INVALID_TYPE = 0;
  ATNState.BASIC = 1;
  ATNState.RULE_START = 2;
  ATNState.BLOCK_START = 3;
  ATNState.PLUS_BLOCK_START = 4;
  ATNState.STAR_BLOCK_START = 5;
  ATNState.TOKEN_START = 6;
  ATNState.RULE_STOP = 7;
  ATNState.BLOCK_END = 8;
  ATNState.STAR_LOOP_BACK = 9;
  ATNState.STAR_LOOP_ENTRY = 10;
  ATNState.PLUS_LOOP_BACK = 11;
  ATNState.LOOP_END = 12;

  ATNState.serializationNames = [
              "INVALID",
              "BASIC",
              "RULE_START",
              "BLOCK_START",
              "PLUS_BLOCK_START",
              "STAR_BLOCK_START",
              "TOKEN_START",
              "RULE_STOP",
              "BLOCK_END",
              "STAR_LOOP_BACK",
              "STAR_LOOP_ENTRY",
              "PLUS_LOOP_BACK",
              "LOOP_END" ];

  ATNState.INVALID_STATE_NUMBER = -1;

  ATNState.prototype.toString = function() {
  	return this.stateNumber;
  };

  ATNState.prototype.equals = function(other) {
      if (other instanceof ATNState) {
          return this.stateNumber===other.stateNumber;
      } else {
          return false;
      }
  };

  ATNState.prototype.isNonGreedyExitState = function() {
      return false;
  };


  ATNState.prototype.addTransition = function(trans, index) {
  	if(index===undefined) {
  		index = -1;
  	}
      if (this.transitions.length===0) {
          this.epsilonOnlyTransitions = trans.isEpsilon;
      } else if(this.epsilonOnlyTransitions !== trans.isEpsilon) {
          this.epsilonOnlyTransitions = false;
      }
      if (index===-1) {
          this.transitions.push(trans);
      } else {
          this.transitions.splice(index, 1, trans);
      }
  };

  function BasicState() {
  	ATNState.call(this);
      this.stateType = ATNState.BASIC;
      return this;
  }

  BasicState.prototype = Object.create(ATNState.prototype);
  BasicState.prototype.constructor = BasicState;


  function DecisionState() {
  	ATNState.call(this);
      this.decision = -1;
      this.nonGreedy = false;
      return this;
  }

  DecisionState.prototype = Object.create(ATNState.prototype);
  DecisionState.prototype.constructor = DecisionState;


  //  The start of a regular {@code (...)} block.
  function BlockStartState() {
  	DecisionState.call(this);
  	this.endState = null;
  	return this;
  }

  BlockStartState.prototype = Object.create(DecisionState.prototype);
  BlockStartState.prototype.constructor = BlockStartState;


  function BasicBlockStartState() {
  	BlockStartState.call(this);
  	this.stateType = ATNState.BLOCK_START;
  	return this;
  }

  BasicBlockStartState.prototype = Object.create(BlockStartState.prototype);
  BasicBlockStartState.prototype.constructor = BasicBlockStartState;


  // Terminal node of a simple {@code (a|b|c)} block.
  function BlockEndState() {
  	ATNState.call(this);
  	this.stateType = ATNState.BLOCK_END;
      this.startState = null;
      return this;
  }

  BlockEndState.prototype = Object.create(ATNState.prototype);
  BlockEndState.prototype.constructor = BlockEndState;


  // The last node in the ATN for a rule, unless that rule is the start symbol.
  //  In that case, there is one transition to EOF. Later, we might encode
  //  references to all calls to this rule to compute FOLLOW sets for
  //  error handling.
  //
  function RuleStopState() {
  	ATNState.call(this);
      this.stateType = ATNState.RULE_STOP;
      return this;
  }

  RuleStopState.prototype = Object.create(ATNState.prototype);
  RuleStopState.prototype.constructor = RuleStopState;

  function RuleStartState() {
  	ATNState.call(this);
  	this.stateType = ATNState.RULE_START;
  	this.stopState = null;
  	this.isPrecedenceRule = false;
  	return this;
  }

  RuleStartState.prototype = Object.create(ATNState.prototype);
  RuleStartState.prototype.constructor = RuleStartState;

  // Decision state for {@code A+} and {@code (A|B)+}.  It has two transitions:
  //  one to the loop back to start of the block and one to exit.
  //
  function PlusLoopbackState() {
  	DecisionState.call(this);
  	this.stateType = ATNState.PLUS_LOOP_BACK;
  	return this;
  }

  PlusLoopbackState.prototype = Object.create(DecisionState.prototype);
  PlusLoopbackState.prototype.constructor = PlusLoopbackState;


  // Start of {@code (A|B|...)+} loop. Technically a decision state, but
  //  we don't use for code generation; somebody might need it, so I'm defining
  //  it for completeness. In reality, the {@link PlusLoopbackState} node is the
  //  real decision-making note for {@code A+}.
  //
  function PlusBlockStartState() {
  	BlockStartState.call(this);
  	this.stateType = ATNState.PLUS_BLOCK_START;
      this.loopBackState = null;
      return this;
  }

  PlusBlockStartState.prototype = Object.create(BlockStartState.prototype);
  PlusBlockStartState.prototype.constructor = PlusBlockStartState;

  // The block that begins a closure loop.
  function StarBlockStartState() {
  	BlockStartState.call(this);
  	this.stateType = ATNState.STAR_BLOCK_START;
  	return this;
  }

  StarBlockStartState.prototype = Object.create(BlockStartState.prototype);
  StarBlockStartState.prototype.constructor = StarBlockStartState;


  function StarLoopbackState() {
  	ATNState.call(this);
  	this.stateType = ATNState.STAR_LOOP_BACK;
  	return this;
  }

  StarLoopbackState.prototype = Object.create(ATNState.prototype);
  StarLoopbackState.prototype.constructor = StarLoopbackState;


  function StarLoopEntryState() {
  	DecisionState.call(this);
  	this.stateType = ATNState.STAR_LOOP_ENTRY;
      this.loopBackState = null;
      // Indicates whether this state can benefit from a precedence DFA during SLL decision making.
      this.isPrecedenceDecision = null;
      return this;
  }

  StarLoopEntryState.prototype = Object.create(DecisionState.prototype);
  StarLoopEntryState.prototype.constructor = StarLoopEntryState;


  // Mark the end of a * or + loop.
  function LoopEndState() {
  	ATNState.call(this);
  	this.stateType = ATNState.LOOP_END;
  	this.loopBackState = null;
  	return this;
  }

  LoopEndState.prototype = Object.create(ATNState.prototype);
  LoopEndState.prototype.constructor = LoopEndState;


  // The Tokens rule start state linking to each lexer rule start state */
  function TokensStartState() {
  	DecisionState.call(this);
  	this.stateType = ATNState.TOKEN_START;
  	return this;
  }

  TokensStartState.prototype = Object.create(DecisionState.prototype);
  TokensStartState.prototype.constructor = TokensStartState;

  exports.ATNState = ATNState;
  exports.BasicState = BasicState;
  exports.DecisionState = DecisionState;
  exports.BlockStartState = BlockStartState;
  exports.BlockEndState = BlockEndState;
  exports.LoopEndState = LoopEndState;
  exports.RuleStartState = RuleStartState;
  exports.RuleStopState = RuleStopState;
  exports.TokensStartState = TokensStartState;
  exports.PlusLoopbackState = PlusLoopbackState;
  exports.StarLoopbackState = StarLoopbackState;
  exports.StarLoopEntryState = StarLoopEntryState;
  exports.PlusBlockStartState = PlusBlockStartState;
  exports.StarBlockStartState = StarBlockStartState;
  exports.BasicBlockStartState = BasicBlockStartState;

  },{}],23:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  ///

  // Represents the type of recognizer an ATN applies to.

  function ATNType() {

  }

  ATNType.LEXER = 0;
  ATNType.PARSER = 1;

  exports.ATNType = ATNType;


  },{}],24:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  ///

  // When we hit an accept state in either the DFA or the ATN, we
  //  have to notify the character stream to start buffering characters
  //  via {@link IntStream//mark} and record the current state. The current sim state
  //  includes the current index into the input, the current line,
  //  and current character position in that line. Note that the Lexer is
  //  tracking the starting line and characterization of the token. These
  //  variables track the "state" of the simulator when it hits an accept state.
  //
  //  <p>We track these variables separately for the DFA and ATN simulation
  //  because the DFA simulation often has to fail over to the ATN
  //  simulation. If the ATN simulation fails, we need the DFA to fall
  //  back to its previously accepted state, if any. If the ATN succeeds,
  //  then the ATN does the accept and the DFA simulator that invoked it
  //  can simply return the predicted token type.</p>
  ///

  var Token = require('./../Token').Token;
  var Lexer = require('./../Lexer').Lexer;
  var ATN = require('./ATN').ATN;
  var ATNSimulator = require('./ATNSimulator').ATNSimulator;
  var DFAState = require('./../dfa/DFAState').DFAState;
  var ATNConfigSet = require('./ATNConfigSet').ATNConfigSet;
  var OrderedATNConfigSet = require('./ATNConfigSet').OrderedATNConfigSet;
  var PredictionContext = require('./../PredictionContext').PredictionContext;
  var SingletonPredictionContext = require('./../PredictionContext').SingletonPredictionContext;
  var RuleStopState = require('./ATNState').RuleStopState;
  var LexerATNConfig = require('./ATNConfig').LexerATNConfig;
  var Transition = require('./Transition').Transition;
  var LexerActionExecutor = require('./LexerActionExecutor').LexerActionExecutor;
  var LexerNoViableAltException = require('./../error/Errors').LexerNoViableAltException;

  function resetSimState(sim) {
  	sim.index = -1;
  	sim.line = 0;
  	sim.column = -1;
  	sim.dfaState = null;
  }

  function SimState() {
  	resetSimState(this);
  	return this;
  }

  SimState.prototype.reset = function() {
  	resetSimState(this);
  };

  function LexerATNSimulator(recog, atn, decisionToDFA, sharedContextCache) {
  	ATNSimulator.call(this, atn, sharedContextCache);
  	this.decisionToDFA = decisionToDFA;
  	this.recog = recog;
  	// The current token's starting index into the character stream.
  	// Shared across DFA to ATN simulation in case the ATN fails and the
  	// DFA did not have a previous accept state. In this case, we use the
  	// ATN-generated exception object.
  	this.startIndex = -1;
  	// line number 1..n within the input///
  	this.line = 1;
  	// The index of the character relative to the beginning of the line
  	// 0..n-1///
  	this.column = 0;
  	this.mode = Lexer.DEFAULT_MODE;
  	// Used during DFA/ATN exec to record the most recent accept configuration
  	// info
  	this.prevAccept = new SimState();
  	// done
  	return this;
  }

  LexerATNSimulator.prototype = Object.create(ATNSimulator.prototype);
  LexerATNSimulator.prototype.constructor = LexerATNSimulator;

  LexerATNSimulator.debug = false;
  LexerATNSimulator.dfa_debug = false;

  LexerATNSimulator.MIN_DFA_EDGE = 0;
  LexerATNSimulator.MAX_DFA_EDGE = 127; // forces unicode to stay in ATN

  LexerATNSimulator.match_calls = 0;

  LexerATNSimulator.prototype.copyState = function(simulator) {
  	this.column = simulator.column;
  	this.line = simulator.line;
  	this.mode = simulator.mode;
  	this.startIndex = simulator.startIndex;
  };

  LexerATNSimulator.prototype.match = function(input, mode) {
  	this.match_calls += 1;
  	this.mode = mode;
  	var mark = input.mark();
  	try {
  		this.startIndex = input.index;
  		this.prevAccept.reset();
  		var dfa = this.decisionToDFA[mode];
  		if (dfa.s0 === null) {
  			return this.matchATN(input);
  		} else {
  			return this.execATN(input, dfa.s0);
  		}
  	} finally {
  		input.release(mark);
  	}
  };

  LexerATNSimulator.prototype.reset = function() {
  	this.prevAccept.reset();
  	this.startIndex = -1;
  	this.line = 1;
  	this.column = 0;
  	this.mode = Lexer.DEFAULT_MODE;
  };

  LexerATNSimulator.prototype.matchATN = function(input) {
  	var startState = this.atn.modeToStartState[this.mode];

  	if (LexerATNSimulator.debug) {
  		console.log("matchATN mode " + this.mode + " start: " + startState);
  	}
  	var old_mode = this.mode;
  	var s0_closure = this.computeStartState(input, startState);
  	var suppressEdge = s0_closure.hasSemanticContext;
  	s0_closure.hasSemanticContext = false;

  	var next = this.addDFAState(s0_closure);
  	if (!suppressEdge) {
  		this.decisionToDFA[this.mode].s0 = next;
  	}

  	var predict = this.execATN(input, next);

  	if (LexerATNSimulator.debug) {
  		console.log("DFA after matchATN: " + this.decisionToDFA[old_mode].toLexerString());
  	}
  	return predict;
  };

  LexerATNSimulator.prototype.execATN = function(input, ds0) {
  	if (LexerATNSimulator.debug) {
  		console.log("start state closure=" + ds0.configs);
  	}
  	if (ds0.isAcceptState) {
  		// allow zero-length tokens
  		this.captureSimState(this.prevAccept, input, ds0);
  	}
  	var t = input.LA(1);
  	var s = ds0; // s is current/from DFA state

  	while (true) { // while more work
  		if (LexerATNSimulator.debug) {
  			console.log("execATN loop starting closure: " + s.configs);
  		}

  		// As we move src->trg, src->trg, we keep track of the previous trg to
  		// avoid looking up the DFA state again, which is expensive.
  		// If the previous target was already part of the DFA, we might
  		// be able to avoid doing a reach operation upon t. If s!=null,
  		// it means that semantic predicates didn't prevent us from
  		// creating a DFA state. Once we know s!=null, we check to see if
  		// the DFA state has an edge already for t. If so, we can just reuse
  		// it's configuration set; there's no point in re-computing it.
  		// This is kind of like doing DFA simulation within the ATN
  		// simulation because DFA simulation is really just a way to avoid
  		// computing reach/closure sets. Technically, once we know that
  		// we have a previously added DFA state, we could jump over to
  		// the DFA simulator. But, that would mean popping back and forth
  		// a lot and making things more complicated algorithmically.
  		// This optimization makes a lot of sense for loops within DFA.
  		// A character will take us back to an existing DFA state
  		// that already has lots of edges out of it. e.g., .* in comments.
  		// print("Target for:" + str(s) + " and:" + str(t))
  		var target = this.getExistingTargetState(s, t);
  		// print("Existing:" + str(target))
  		if (target === null) {
  			target = this.computeTargetState(input, s, t);
  			// print("Computed:" + str(target))
  		}
  		if (target === ATNSimulator.ERROR) {
  			break;
  		}
  		// If this is a consumable input element, make sure to consume before
  		// capturing the accept state so the input index, line, and char
  		// position accurately reflect the state of the interpreter at the
  		// end of the token.
  		if (t !== Token.EOF) {
  			this.consume(input);
  		}
  		if (target.isAcceptState) {
  			this.captureSimState(this.prevAccept, input, target);
  			if (t === Token.EOF) {
  				break;
  			}
  		}
  		t = input.LA(1);
  		s = target; // flip; current DFA target becomes new src/from state
  	}
  	return this.failOrAccept(this.prevAccept, input, s.configs, t);
  };

  // Get an existing target state for an edge in the DFA. If the target state
  // for the edge has not yet been computed or is otherwise not available,
  // this method returns {@code null}.
  //
  // @param s The current DFA state
  // @param t The next input symbol
  // @return The existing target DFA state for the given input symbol
  // {@code t}, or {@code null} if the target state for this edge is not
  // already cached
  LexerATNSimulator.prototype.getExistingTargetState = function(s, t) {
  	if (s.edges === null || t < LexerATNSimulator.MIN_DFA_EDGE || t > LexerATNSimulator.MAX_DFA_EDGE) {
  		return null;
  	}

  	var target = s.edges[t - LexerATNSimulator.MIN_DFA_EDGE];
  	if(target===undefined) {
  		target = null;
  	}
  	if (LexerATNSimulator.debug && target !== null) {
  		console.log("reuse state " + s.stateNumber + " edge to " + target.stateNumber);
  	}
  	return target;
  };

  // Compute a target state for an edge in the DFA, and attempt to add the
  // computed state and corresponding edge to the DFA.
  //
  // @param input The input stream
  // @param s The current DFA state
  // @param t The next input symbol
  //
  // @return The computed target DFA state for the given input symbol
  // {@code t}. If {@code t} does not lead to a valid DFA state, this method
  // returns {@link //ERROR}.
  LexerATNSimulator.prototype.computeTargetState = function(input, s, t) {
  	var reach = new OrderedATNConfigSet();
  	// if we don't find an existing DFA state
  	// Fill reach starting from closure, following t transitions
  	this.getReachableConfigSet(input, s.configs, reach, t);

  	if (reach.items.length === 0) { // we got nowhere on t from s
  		if (!reach.hasSemanticContext) {
  			// we got nowhere on t, don't throw out this knowledge; it'd
  			// cause a failover from DFA later.
  			this.addDFAEdge(s, t, ATNSimulator.ERROR);
  		}
  		// stop when we can't match any more char
  		return ATNSimulator.ERROR;
  	}
  	// Add an edge from s to target DFA found/created for reach
  	return this.addDFAEdge(s, t, null, reach);
  };

  LexerATNSimulator.prototype.failOrAccept = function(prevAccept, input, reach, t) {
  	if (this.prevAccept.dfaState !== null) {
  		var lexerActionExecutor = prevAccept.dfaState.lexerActionExecutor;
  		this.accept(input, lexerActionExecutor, this.startIndex,
  				prevAccept.index, prevAccept.line, prevAccept.column);
  		return prevAccept.dfaState.prediction;
  	} else {
  		// if no accept and EOF is first char, return EOF
  		if (t === Token.EOF && input.index === this.startIndex) {
  			return Token.EOF;
  		}
  		throw new LexerNoViableAltException(this.recog, input, this.startIndex, reach);
  	}
  };

  // Given a starting configuration set, figure out all ATN configurations
  // we can reach upon input {@code t}. Parameter {@code reach} is a return
  // parameter.
  LexerATNSimulator.prototype.getReachableConfigSet = function(input, closure,
  		reach, t) {
  	// this is used to skip processing for configs which have a lower priority
  	// than a config that already reached an accept state for the same rule
  	var skipAlt = ATN.INVALID_ALT_NUMBER;
  	for (var i = 0; i < closure.items.length; i++) {
  		var cfg = closure.items[i];
  		var currentAltReachedAcceptState = (cfg.alt === skipAlt);
  		if (currentAltReachedAcceptState && cfg.passedThroughNonGreedyDecision) {
  			continue;
  		}
  		if (LexerATNSimulator.debug) {
  			console.log("testing %s at %s\n", this.getTokenName(t), cfg
  					.toString(this.recog, true));
  		}
  		for (var j = 0; j < cfg.state.transitions.length; j++) {
  			var trans = cfg.state.transitions[j]; // for each transition
  			var target = this.getReachableTarget(trans, t);
  			if (target !== null) {
  				var lexerActionExecutor = cfg.lexerActionExecutor;
  				if (lexerActionExecutor !== null) {
  					lexerActionExecutor = lexerActionExecutor.fixOffsetBeforeMatch(input.index - this.startIndex);
  				}
  				var treatEofAsEpsilon = (t === Token.EOF);
  				var config = new LexerATNConfig({state:target, lexerActionExecutor:lexerActionExecutor}, cfg);
  				if (this.closure(input, config, reach,
  						currentAltReachedAcceptState, true, treatEofAsEpsilon)) {
  					// any remaining configs for this alt have a lower priority
  					// than the one that just reached an accept state.
  					skipAlt = cfg.alt;
  				}
  			}
  		}
  	}
  };

  LexerATNSimulator.prototype.accept = function(input, lexerActionExecutor,
  		startIndex, index, line, charPos) {
  	if (LexerATNSimulator.debug) {
  		console.log("ACTION %s\n", lexerActionExecutor);
  	}
  	// seek to after last char in token
  	input.seek(index);
  	this.line = line;
  	this.column = charPos;
  	if (lexerActionExecutor !== null && this.recog !== null) {
  		lexerActionExecutor.execute(this.recog, input, startIndex);
  	}
  };

  LexerATNSimulator.prototype.getReachableTarget = function(trans, t) {
  	if (trans.matches(t, 0, Lexer.MAX_CHAR_VALUE)) {
  		return trans.target;
  	} else {
  		return null;
  	}
  };

  LexerATNSimulator.prototype.computeStartState = function(input, p) {
  	var initialContext = PredictionContext.EMPTY;
  	var configs = new OrderedATNConfigSet();
  	for (var i = 0; i < p.transitions.length; i++) {
  		var target = p.transitions[i].target;
          var cfg = new LexerATNConfig({state:target, alt:i+1, context:initialContext}, null);
  		this.closure(input, cfg, configs, false, false, false);
  	}
  	return configs;
  };

  // Since the alternatives within any lexer decision are ordered by
  // preference, this method stops pursuing the closure as soon as an accept
  // state is reached. After the first accept state is reached by depth-first
  // search from {@code config}, all other (potentially reachable) states for
  // this rule would have a lower priority.
  //
  // @return {@code true} if an accept state is reached, otherwise
  // {@code false}.
  LexerATNSimulator.prototype.closure = function(input, config, configs,
  		currentAltReachedAcceptState, speculative, treatEofAsEpsilon) {
  	var cfg = null;
  	if (LexerATNSimulator.debug) {
  		console.log("closure(" + config.toString(this.recog, true) + ")");
  	}
  	if (config.state instanceof RuleStopState) {
  		if (LexerATNSimulator.debug) {
  			if (this.recog !== null) {
  				console.log("closure at %s rule stop %s\n", this.recog.ruleNames[config.state.ruleIndex], config);
  			} else {
  				console.log("closure at rule stop %s\n", config);
  			}
  		}
  		if (config.context === null || config.context.hasEmptyPath()) {
  			if (config.context === null || config.context.isEmpty()) {
  				configs.add(config);
  				return true;
  			} else {
  				configs.add(new LexerATNConfig({ state:config.state, context:PredictionContext.EMPTY}, config));
  				currentAltReachedAcceptState = true;
  			}
  		}
  		if (config.context !== null && !config.context.isEmpty()) {
  			for (var i = 0; i < config.context.length; i++) {
  				if (config.context.getReturnState(i) !== PredictionContext.EMPTY_RETURN_STATE) {
  					var newContext = config.context.getParent(i); // "pop" return state
  					var returnState = this.atn.states[config.context.getReturnState(i)];
  					cfg = new LexerATNConfig({ state:returnState, context:newContext }, config);
  					currentAltReachedAcceptState = this.closure(input, cfg,
  							configs, currentAltReachedAcceptState, speculative,
  							treatEofAsEpsilon);
  				}
  			}
  		}
  		return currentAltReachedAcceptState;
  	}
  	// optimization
  	if (!config.state.epsilonOnlyTransitions) {
  		if (!currentAltReachedAcceptState || !config.passedThroughNonGreedyDecision) {
  			configs.add(config);
  		}
  	}
  	for (var j = 0; j < config.state.transitions.length; j++) {
  		var trans = config.state.transitions[j];
  		cfg = this.getEpsilonTarget(input, config, trans, configs, speculative, treatEofAsEpsilon);
  		if (cfg !== null) {
  			currentAltReachedAcceptState = this.closure(input, cfg, configs,
  					currentAltReachedAcceptState, speculative, treatEofAsEpsilon);
  		}
  	}
  	return currentAltReachedAcceptState;
  };

  // side-effect: can alter configs.hasSemanticContext
  LexerATNSimulator.prototype.getEpsilonTarget = function(input, config, trans,
  		configs, speculative, treatEofAsEpsilon) {
  	var cfg = null;
  	if (trans.serializationType === Transition.RULE) {
  		var newContext = SingletonPredictionContext.create(config.context, trans.followState.stateNumber);
  		cfg = new LexerATNConfig( { state:trans.target, context:newContext}, config);
  	} else if (trans.serializationType === Transition.PRECEDENCE) {
  		throw "Precedence predicates are not supported in lexers.";
  	} else if (trans.serializationType === Transition.PREDICATE) {
  		// Track traversing semantic predicates. If we traverse,
  		// we cannot add a DFA state for this "reach" computation
  		// because the DFA would not test the predicate again in the
  		// future. Rather than creating collections of semantic predicates
  		// like v3 and testing them on prediction, v4 will test them on the
  		// fly all the time using the ATN not the DFA. This is slower but
  		// semantically it's not used that often. One of the key elements to
  		// this predicate mechanism is not adding DFA states that see
  		// predicates immediately afterwards in the ATN. For example,

  		// a : ID {p1}? | ID {p2}? ;

  		// should create the start state for rule 'a' (to save start state
  		// competition), but should not create target of ID state. The
  		// collection of ATN states the following ID references includes
  		// states reached by traversing predicates. Since this is when we
  		// test them, we cannot cash the DFA state target of ID.

  		if (LexerATNSimulator.debug) {
  			console.log("EVAL rule " + trans.ruleIndex + ":" + trans.predIndex);
  		}
  		configs.hasSemanticContext = true;
  		if (this.evaluatePredicate(input, trans.ruleIndex, trans.predIndex, speculative)) {
  			cfg = new LexerATNConfig({ state:trans.target}, config);
  		}
  	} else if (trans.serializationType === Transition.ACTION) {
  		if (config.context === null || config.context.hasEmptyPath()) {
  			// execute actions anywhere in the start rule for a token.
  			//
  			// TODO: if the entry rule is invoked recursively, some
  			// actions may be executed during the recursive call. The
  			// problem can appear when hasEmptyPath() is true but
  			// isEmpty() is false. In this case, the config needs to be
  			// split into two contexts - one with just the empty path
  			// and another with everything but the empty path.
  			// Unfortunately, the current algorithm does not allow
  			// getEpsilonTarget to return two configurations, so
  			// additional modifications are needed before we can support
  			// the split operation.
  			var lexerActionExecutor = LexerActionExecutor.append(config.lexerActionExecutor,
  					this.atn.lexerActions[trans.actionIndex]);
  			cfg = new LexerATNConfig({ state:trans.target, lexerActionExecutor:lexerActionExecutor }, config);
  		} else {
  			// ignore actions in referenced rules
  			cfg = new LexerATNConfig( { state:trans.target}, config);
  		}
  	} else if (trans.serializationType === Transition.EPSILON) {
  		cfg = new LexerATNConfig({ state:trans.target}, config);
  	} else if (trans.serializationType === Transition.ATOM ||
  				trans.serializationType === Transition.RANGE ||
  				trans.serializationType === Transition.SET) {
  		if (treatEofAsEpsilon) {
  			if (trans.matches(Token.EOF, 0, Lexer.MAX_CHAR_VALUE)) {
  				cfg = new LexerATNConfig( { state:trans.target }, config);
  			}
  		}
  	}
  	return cfg;
  };

  // Evaluate a predicate specified in the lexer.
  //
  // <p>If {@code speculative} is {@code true}, this method was called before
  // {@link //consume} for the matched character. This method should call
  // {@link //consume} before evaluating the predicate to ensure position
  // sensitive values, including {@link Lexer//getText}, {@link Lexer//getLine},
  // and {@link Lexer//getcolumn}, properly reflect the current
  // lexer state. This method should restore {@code input} and the simulator
  // to the original state before returning (i.e. undo the actions made by the
  // call to {@link //consume}.</p>
  //
  // @param input The input stream.
  // @param ruleIndex The rule containing the predicate.
  // @param predIndex The index of the predicate within the rule.
  // @param speculative {@code true} if the current index in {@code input} is
  // one character before the predicate's location.
  //
  // @return {@code true} if the specified predicate evaluates to
  // {@code true}.
  // /
  LexerATNSimulator.prototype.evaluatePredicate = function(input, ruleIndex,
  		predIndex, speculative) {
  	// assume true if no recognizer was provided
  	if (this.recog === null) {
  		return true;
  	}
  	if (!speculative) {
  		return this.recog.sempred(null, ruleIndex, predIndex);
  	}
  	var savedcolumn = this.column;
  	var savedLine = this.line;
  	var index = input.index;
  	var marker = input.mark();
  	try {
  		this.consume(input);
  		return this.recog.sempred(null, ruleIndex, predIndex);
  	} finally {
  		this.column = savedcolumn;
  		this.line = savedLine;
  		input.seek(index);
  		input.release(marker);
  	}
  };

  LexerATNSimulator.prototype.captureSimState = function(settings, input, dfaState) {
  	settings.index = input.index;
  	settings.line = this.line;
  	settings.column = this.column;
  	settings.dfaState = dfaState;
  };

  LexerATNSimulator.prototype.addDFAEdge = function(from_, tk, to, cfgs) {
  	if (to === undefined) {
  		to = null;
  	}
  	if (cfgs === undefined) {
  		cfgs = null;
  	}
  	if (to === null && cfgs !== null) {
  		// leading to this call, ATNConfigSet.hasSemanticContext is used as a
  		// marker indicating dynamic predicate evaluation makes this edge
  		// dependent on the specific input sequence, so the static edge in the
  		// DFA should be omitted. The target DFAState is still created since
  		// execATN has the ability to resynchronize with the DFA state cache
  		// following the predicate evaluation step.
  		//
  		// TJP notes: next time through the DFA, we see a pred again and eval.
  		// If that gets us to a previously created (but dangling) DFA
  		// state, we can continue in pure DFA mode from there.
  		// /
  		var suppressEdge = cfgs.hasSemanticContext;
  		cfgs.hasSemanticContext = false;

  		to = this.addDFAState(cfgs);

  		if (suppressEdge) {
  			return to;
  		}
  	}
  	// add the edge
  	if (tk < LexerATNSimulator.MIN_DFA_EDGE || tk > LexerATNSimulator.MAX_DFA_EDGE) {
  		// Only track edges within the DFA bounds
  		return to;
  	}
  	if (LexerATNSimulator.debug) {
  		console.log("EDGE " + from_ + " -> " + to + " upon " + tk);
  	}
  	if (from_.edges === null) {
  		// make room for tokens 1..n and -1 masquerading as index 0
  		from_.edges = [];
  	}
  	from_.edges[tk - LexerATNSimulator.MIN_DFA_EDGE] = to; // connect

  	return to;
  };

  // Add a new DFA state if there isn't one with this set of
  // configurations already. This method also detects the first
  // configuration containing an ATN rule stop state. Later, when
  // traversing the DFA, we will know which rule to accept.
  LexerATNSimulator.prototype.addDFAState = function(configs) {
  	var proposed = new DFAState(null, configs);
  	var firstConfigWithRuleStopState = null;
  	for (var i = 0; i < configs.items.length; i++) {
  		var cfg = configs.items[i];
  		if (cfg.state instanceof RuleStopState) {
  			firstConfigWithRuleStopState = cfg;
  			break;
  		}
  	}
  	if (firstConfigWithRuleStopState !== null) {
  		proposed.isAcceptState = true;
  		proposed.lexerActionExecutor = firstConfigWithRuleStopState.lexerActionExecutor;
  		proposed.prediction = this.atn.ruleToTokenType[firstConfigWithRuleStopState.state.ruleIndex];
  	}
  	var dfa = this.decisionToDFA[this.mode];
  	var existing = dfa.states.get(proposed);
  	if (existing!==null) {
  		return existing;
  	}
  	var newState = proposed;
  	newState.stateNumber = dfa.states.length;
  	configs.setReadonly(true);
  	newState.configs = configs;
  	dfa.states.add(newState);
  	return newState;
  };

  LexerATNSimulator.prototype.getDFA = function(mode) {
  	return this.decisionToDFA[mode];
  };

  // Get the text matched so far for the current token.
  LexerATNSimulator.prototype.getText = function(input) {
  	// index is first lookahead char, don't include.
  	return input.getText(this.startIndex, input.index - 1);
  };

  LexerATNSimulator.prototype.consume = function(input) {
  	var curChar = input.LA(1);
  	if (curChar === "\n".charCodeAt(0)) {
  		this.line += 1;
  		this.column = 0;
  	} else {
  		this.column += 1;
  	}
  	input.consume();
  };

  LexerATNSimulator.prototype.getTokenName = function(tt) {
  	if (tt === -1) {
  		return "EOF";
  	} else {
  		return "'" + String.fromCharCode(tt) + "'";
  	}
  };

  exports.LexerATNSimulator = LexerATNSimulator;

  },{"./../Lexer":8,"./../PredictionContext":11,"./../Token":14,"./../dfa/DFAState":34,"./../error/Errors":39,"./ATN":16,"./ATNConfig":17,"./ATNConfigSet":18,"./ATNSimulator":21,"./ATNState":22,"./LexerActionExecutor":26,"./Transition":30}],25:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
   //

  function LexerActionType() {
  }

  LexerActionType.CHANNEL = 0;     //The type of a {@link LexerChannelAction} action.
  LexerActionType.CUSTOM = 1;      //The type of a {@link LexerCustomAction} action.
  LexerActionType.MODE = 2;        //The type of a {@link LexerModeAction} action.
  LexerActionType.MORE = 3;        //The type of a {@link LexerMoreAction} action.
  LexerActionType.POP_MODE = 4;    //The type of a {@link LexerPopModeAction} action.
  LexerActionType.PUSH_MODE = 5;   //The type of a {@link LexerPushModeAction} action.
  LexerActionType.SKIP = 6;        //The type of a {@link LexerSkipAction} action.
  LexerActionType.TYPE = 7;        //The type of a {@link LexerTypeAction} action.

  function LexerAction(action) {
      this.actionType = action;
      this.isPositionDependent = false;
      return this;
  }

  LexerAction.prototype.hashCode = function() {
      var hash = new Hash();
      this.updateHashCode(hash);
      return hash.finish()
  };

  LexerAction.prototype.updateHashCode = function(hash) {
      hash.update(this.actionType);
  };

  LexerAction.prototype.equals = function(other) {
      return this === other;
  };



  //
  // Implements the {@code skip} lexer action by calling {@link Lexer//skip}.
  //
  // <p>The {@code skip} command does not have any parameters, so this action is
  // implemented as a singleton instance exposed by {@link //INSTANCE}.</p>
  function LexerSkipAction() {
  	LexerAction.call(this, LexerActionType.SKIP);
  	return this;
  }

  LexerSkipAction.prototype = Object.create(LexerAction.prototype);
  LexerSkipAction.prototype.constructor = LexerSkipAction;

  // Provides a singleton instance of this parameterless lexer action.
  LexerSkipAction.INSTANCE = new LexerSkipAction();

  LexerSkipAction.prototype.execute = function(lexer) {
      lexer.skip();
  };

  LexerSkipAction.prototype.toString = function() {
  	return "skip";
  };

  //  Implements the {@code type} lexer action by calling {@link Lexer//setType}
  // with the assigned type.
  function LexerTypeAction(type) {
  	LexerAction.call(this, LexerActionType.TYPE);
  	this.type = type;
  	return this;
  }

  LexerTypeAction.prototype = Object.create(LexerAction.prototype);
  LexerTypeAction.prototype.constructor = LexerTypeAction;

  LexerTypeAction.prototype.execute = function(lexer) {
      lexer.type = this.type;
  };

  LexerTypeAction.prototype.updateHashCode = function(hash) {
      hash.update(this.actionType, this.type);
  };


  LexerTypeAction.prototype.equals = function(other) {
      if(this === other) {
          return true;
      } else if (! (other instanceof LexerTypeAction)) {
          return false;
      } else {
          return this.type === other.type;
      }
  };

  LexerTypeAction.prototype.toString = function() {
      return "type(" + this.type + ")";
  };

  // Implements the {@code pushMode} lexer action by calling
  // {@link Lexer//pushMode} with the assigned mode.
  function LexerPushModeAction(mode) {
  	LexerAction.call(this, LexerActionType.PUSH_MODE);
      this.mode = mode;
      return this;
  }

  LexerPushModeAction.prototype = Object.create(LexerAction.prototype);
  LexerPushModeAction.prototype.constructor = LexerPushModeAction;

  // <p>This action is implemented by calling {@link Lexer//pushMode} with the
  // value provided by {@link //getMode}.</p>
  LexerPushModeAction.prototype.execute = function(lexer) {
      lexer.pushMode(this.mode);
  };

  LexerPushModeAction.prototype.updateHashCode = function(hash) {
      hash.update(this.actionType, this.mode);
  };

  LexerPushModeAction.prototype.equals = function(other) {
      if (this === other) {
          return true;
      } else if (! (other instanceof LexerPushModeAction)) {
          return false;
      } else {
          return this.mode === other.mode;
      }
  };

  LexerPushModeAction.prototype.toString = function() {
  	return "pushMode(" + this.mode + ")";
  };


  // Implements the {@code popMode} lexer action by calling {@link Lexer//popMode}.
  //
  // <p>The {@code popMode} command does not have any parameters, so this action is
  // implemented as a singleton instance exposed by {@link //INSTANCE}.</p>
  function LexerPopModeAction() {
  	LexerAction.call(this,LexerActionType.POP_MODE);
  	return this;
  }

  LexerPopModeAction.prototype = Object.create(LexerAction.prototype);
  LexerPopModeAction.prototype.constructor = LexerPopModeAction;

  LexerPopModeAction.INSTANCE = new LexerPopModeAction();

  // <p>This action is implemented by calling {@link Lexer//popMode}.</p>
  LexerPopModeAction.prototype.execute = function(lexer) {
      lexer.popMode();
  };

  LexerPopModeAction.prototype.toString = function() {
  	return "popMode";
  };

  // Implements the {@code more} lexer action by calling {@link Lexer//more}.
  //
  // <p>The {@code more} command does not have any parameters, so this action is
  // implemented as a singleton instance exposed by {@link //INSTANCE}.</p>
  function LexerMoreAction() {
  	LexerAction.call(this, LexerActionType.MORE);
  	return this;
  }

  LexerMoreAction.prototype = Object.create(LexerAction.prototype);
  LexerMoreAction.prototype.constructor = LexerMoreAction;

  LexerMoreAction.INSTANCE = new LexerMoreAction();

  // <p>This action is implemented by calling {@link Lexer//popMode}.</p>
  LexerMoreAction.prototype.execute = function(lexer) {
      lexer.more();
  };

  LexerMoreAction.prototype.toString = function() {
      return "more";
  };


  // Implements the {@code mode} lexer action by calling {@link Lexer//mode} with
  // the assigned mode.
  function LexerModeAction(mode) {
  	LexerAction.call(this, LexerActionType.MODE);
      this.mode = mode;
      return this;
  }

  LexerModeAction.prototype = Object.create(LexerAction.prototype);
  LexerModeAction.prototype.constructor = LexerModeAction;

  // <p>This action is implemented by calling {@link Lexer//mode} with the
  // value provided by {@link //getMode}.</p>
  LexerModeAction.prototype.execute = function(lexer) {
      lexer.mode(this.mode);
  };

  LexerModeAction.prototype.updateHashCode = function(hash) {
      hash.update(this.actionType, this.mode);
  };

  LexerModeAction.prototype.equals = function(other) {
      if (this === other) {
          return true;
      } else if (! (other instanceof LexerModeAction)) {
          return false;
      } else {
          return this.mode === other.mode;
      }
  };

  LexerModeAction.prototype.toString = function() {
      return "mode(" + this.mode + ")";
  };

  // Executes a custom lexer action by calling {@link Recognizer//action} with the
  // rule and action indexes assigned to the custom action. The implementation of
  // a custom action is added to the generated code for the lexer in an override
  // of {@link Recognizer//action} when the grammar is compiled.
  //
  // <p>This class may represent embedded actions created with the <code>{...}</code>
  // syntax in ANTLR 4, as well as actions created for lexer commands where the
  // command argument could not be evaluated when the grammar was compiled.</p>


      // Constructs a custom lexer action with the specified rule and action
      // indexes.
      //
      // @param ruleIndex The rule index to use for calls to
      // {@link Recognizer//action}.
      // @param actionIndex The action index to use for calls to
      // {@link Recognizer//action}.

  function LexerCustomAction(ruleIndex, actionIndex) {
  	LexerAction.call(this, LexerActionType.CUSTOM);
      this.ruleIndex = ruleIndex;
      this.actionIndex = actionIndex;
      this.isPositionDependent = true;
      return this;
  }

  LexerCustomAction.prototype = Object.create(LexerAction.prototype);
  LexerCustomAction.prototype.constructor = LexerCustomAction;

  // <p>Custom actions are implemented by calling {@link Lexer//action} with the
  // appropriate rule and action indexes.</p>
  LexerCustomAction.prototype.execute = function(lexer) {
      lexer.action(null, this.ruleIndex, this.actionIndex);
  };

  LexerCustomAction.prototype.updateHashCode = function(hash) {
      hash.update(this.actionType, this.ruleIndex, this.actionIndex);
  };

  LexerCustomAction.prototype.equals = function(other) {
      if (this === other) {
          return true;
      } else if (! (other instanceof LexerCustomAction)) {
          return false;
      } else {
          return this.ruleIndex === other.ruleIndex && this.actionIndex === other.actionIndex;
      }
  };

  // Implements the {@code channel} lexer action by calling
  // {@link Lexer//setChannel} with the assigned channel.
  // Constructs a new {@code channel} action with the specified channel value.
  // @param channel The channel value to pass to {@link Lexer//setChannel}.
  function LexerChannelAction(channel) {
  	LexerAction.call(this, LexerActionType.CHANNEL);
      this.channel = channel;
      return this;
  }

  LexerChannelAction.prototype = Object.create(LexerAction.prototype);
  LexerChannelAction.prototype.constructor = LexerChannelAction;

  // <p>This action is implemented by calling {@link Lexer//setChannel} with the
  // value provided by {@link //getChannel}.</p>
  LexerChannelAction.prototype.execute = function(lexer) {
      lexer._channel = this.channel;
  };

  LexerChannelAction.prototype.updateHashCode = function(hash) {
      hash.update(this.actionType, this.channel);
  };

  LexerChannelAction.prototype.equals = function(other) {
      if (this === other) {
          return true;
      } else if (! (other instanceof LexerChannelAction)) {
          return false;
      } else {
          return this.channel === other.channel;
      }
  };

  LexerChannelAction.prototype.toString = function() {
      return "channel(" + this.channel + ")";
  };

  // This implementation of {@link LexerAction} is used for tracking input offsets
  // for position-dependent actions within a {@link LexerActionExecutor}.
  //
  // <p>This action is not serialized as part of the ATN, and is only required for
  // position-dependent lexer actions which appear at a location other than the
  // end of a rule. For more information about DFA optimizations employed for
  // lexer actions, see {@link LexerActionExecutor//append} and
  // {@link LexerActionExecutor//fixOffsetBeforeMatch}.</p>

  // Constructs a new indexed custom action by associating a character offset
  // with a {@link LexerAction}.
  //
  // <p>Note: This class is only required for lexer actions for which
  // {@link LexerAction//isPositionDependent} returns {@code true}.</p>
  //
  // @param offset The offset into the input {@link CharStream}, relative to
  // the token start index, at which the specified lexer action should be
  // executed.
  // @param action The lexer action to execute at a particular offset in the
  // input {@link CharStream}.
  function LexerIndexedCustomAction(offset, action) {
  	LexerAction.call(this, action.actionType);
      this.offset = offset;
      this.action = action;
      this.isPositionDependent = true;
      return this;
  }

  LexerIndexedCustomAction.prototype = Object.create(LexerAction.prototype);
  LexerIndexedCustomAction.prototype.constructor = LexerIndexedCustomAction;

  // <p>This method calls {@link //execute} on the result of {@link //getAction}
  // using the provided {@code lexer}.</p>
  LexerIndexedCustomAction.prototype.execute = function(lexer) {
      // assume the input stream position was properly set by the calling code
      this.action.execute(lexer);
  };

  LexerIndexedCustomAction.prototype.updateHashCode = function(hash) {
      hash.update(this.actionType, this.offset, this.action);
  };

  LexerIndexedCustomAction.prototype.equals = function(other) {
      if (this === other) {
          return true;
      } else if (! (other instanceof LexerIndexedCustomAction)) {
          return false;
      } else {
          return this.offset === other.offset && this.action === other.action;
      }
  };


  exports.LexerActionType = LexerActionType;
  exports.LexerSkipAction = LexerSkipAction;
  exports.LexerChannelAction = LexerChannelAction;
  exports.LexerCustomAction = LexerCustomAction;
  exports.LexerIndexedCustomAction = LexerIndexedCustomAction;
  exports.LexerMoreAction = LexerMoreAction;
  exports.LexerTypeAction = LexerTypeAction;
  exports.LexerPushModeAction = LexerPushModeAction;
  exports.LexerPopModeAction = LexerPopModeAction;
  exports.LexerModeAction = LexerModeAction;
  },{}],26:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  ///

  // Represents an executor for a sequence of lexer actions which traversed during
  // the matching operation of a lexer rule (token).
  //
  // <p>The executor tracks position information for position-dependent lexer actions
  // efficiently, ensuring that actions appearing only at the end of the rule do
  // not cause bloating of the {@link DFA} created for the lexer.</p>

  var hashStuff = require("../Utils").hashStuff;
  var LexerIndexedCustomAction = require('./LexerAction').LexerIndexedCustomAction;

  function LexerActionExecutor(lexerActions) {
  	this.lexerActions = lexerActions === null ? [] : lexerActions;
  	// Caches the result of {@link //hashCode} since the hash code is an element
  	// of the performance-critical {@link LexerATNConfig//hashCode} operation.
  	this.cachedHashCode = hashStuff(lexerActions); // "".join([str(la) for la in
  	// lexerActions]))
  	return this;
  }

  // Creates a {@link LexerActionExecutor} which executes the actions for
  // the input {@code lexerActionExecutor} followed by a specified
  // {@code lexerAction}.
  //
  // @param lexerActionExecutor The executor for actions already traversed by
  // the lexer while matching a token within a particular
  // {@link LexerATNConfig}. If this is {@code null}, the method behaves as
  // though it were an empty executor.
  // @param lexerAction The lexer action to execute after the actions
  // specified in {@code lexerActionExecutor}.
  //
  // @return A {@link LexerActionExecutor} for executing the combine actions
  // of {@code lexerActionExecutor} and {@code lexerAction}.
  LexerActionExecutor.append = function(lexerActionExecutor, lexerAction) {
  	if (lexerActionExecutor === null) {
  		return new LexerActionExecutor([ lexerAction ]);
  	}
  	var lexerActions = lexerActionExecutor.lexerActions.concat([ lexerAction ]);
  	return new LexerActionExecutor(lexerActions);
  };

  // Creates a {@link LexerActionExecutor} which encodes the current offset
  // for position-dependent lexer actions.
  //
  // <p>Normally, when the executor encounters lexer actions where
  // {@link LexerAction//isPositionDependent} returns {@code true}, it calls
  // {@link IntStream//seek} on the input {@link CharStream} to set the input
  // position to the <em>end</em> of the current token. This behavior provides
  // for efficient DFA representation of lexer actions which appear at the end
  // of a lexer rule, even when the lexer rule matches a variable number of
  // characters.</p>
  //
  // <p>Prior to traversing a match transition in the ATN, the current offset
  // from the token start index is assigned to all position-dependent lexer
  // actions which have not already been assigned a fixed offset. By storing
  // the offsets relative to the token start index, the DFA representation of
  // lexer actions which appear in the middle of tokens remains efficient due
  // to sharing among tokens of the same length, regardless of their absolute
  // position in the input stream.</p>
  //
  // <p>If the current executor already has offsets assigned to all
  // position-dependent lexer actions, the method returns {@code this}.</p>
  //
  // @param offset The current offset to assign to all position-dependent
  // lexer actions which do not already have offsets assigned.
  //
  // @return A {@link LexerActionExecutor} which stores input stream offsets
  // for all position-dependent lexer actions.
  // /
  LexerActionExecutor.prototype.fixOffsetBeforeMatch = function(offset) {
  	var updatedLexerActions = null;
  	for (var i = 0; i < this.lexerActions.length; i++) {
  		if (this.lexerActions[i].isPositionDependent &&
  				!(this.lexerActions[i] instanceof LexerIndexedCustomAction)) {
  			if (updatedLexerActions === null) {
  				updatedLexerActions = this.lexerActions.concat([]);
  			}
  			updatedLexerActions[i] = new LexerIndexedCustomAction(offset,
  					this.lexerActions[i]);
  		}
  	}
  	if (updatedLexerActions === null) {
  		return this;
  	} else {
  		return new LexerActionExecutor(updatedLexerActions);
  	}
  };

  // Execute the actions encapsulated by this executor within the context of a
  // particular {@link Lexer}.
  //
  // <p>This method calls {@link IntStream//seek} to set the position of the
  // {@code input} {@link CharStream} prior to calling
  // {@link LexerAction//execute} on a position-dependent action. Before the
  // method returns, the input position will be restored to the same position
  // it was in when the method was invoked.</p>
  //
  // @param lexer The lexer instance.
  // @param input The input stream which is the source for the current token.
  // When this method is called, the current {@link IntStream//index} for
  // {@code input} should be the start of the following token, i.e. 1
  // character past the end of the current token.
  // @param startIndex The token start index. This value may be passed to
  // {@link IntStream//seek} to set the {@code input} position to the beginning
  // of the token.
  // /
  LexerActionExecutor.prototype.execute = function(lexer, input, startIndex) {
  	var requiresSeek = false;
  	var stopIndex = input.index;
  	try {
  		for (var i = 0; i < this.lexerActions.length; i++) {
  			var lexerAction = this.lexerActions[i];
  			if (lexerAction instanceof LexerIndexedCustomAction) {
  				var offset = lexerAction.offset;
  				input.seek(startIndex + offset);
  				lexerAction = lexerAction.action;
  				requiresSeek = (startIndex + offset) !== stopIndex;
  			} else if (lexerAction.isPositionDependent) {
  				input.seek(stopIndex);
  				requiresSeek = false;
  			}
  			lexerAction.execute(lexer);
  		}
  	} finally {
  		if (requiresSeek) {
  			input.seek(stopIndex);
  		}
  	}
  };

  LexerActionExecutor.prototype.hashCode = function() {
  	return this.cachedHashCode;
  };

  LexerActionExecutor.prototype.updateHashCode = function(hash) {
      hash.update(this.cachedHashCode);
  };


  LexerActionExecutor.prototype.equals = function(other) {
  	if (this === other) {
  		return true;
  	} else if (!(other instanceof LexerActionExecutor)) {
  		return false;
  	} else if (this.cachedHashCode != other.cachedHashCode) {
  		return false;
  	} else if (this.lexerActions.length != other.lexerActions.length) {
  		return false;
  	} else {
  		var numActions = this.lexerActions.length;
  		for (var idx = 0; idx < numActions; ++idx) {
  			if (!this.lexerActions[idx].equals(other.lexerActions[idx])) {
  				return false;
  			}
  		}
  		return true;
  	}
  };

  exports.LexerActionExecutor = LexerActionExecutor;

  },{"../Utils":15,"./LexerAction":25}],27:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  //

  //
  // The embodiment of the adaptive LL(*), ALL(*), parsing strategy.
  //
  // <p>
  // The basic complexity of the adaptive strategy makes it harder to understand.
  // We begin with ATN simulation to build paths in a DFA. Subsequent prediction
  // requests go through the DFA first. If they reach a state without an edge for
  // the current symbol, the algorithm fails over to the ATN simulation to
  // complete the DFA path for the current input (until it finds a conflict state
  // or uniquely predicting state).</p>
  //
  // <p>
  // All of that is done without using the outer context because we want to create
  // a DFA that is not dependent upon the rule invocation stack when we do a
  // prediction. One DFA works in all contexts. We avoid using context not
  // necessarily because it's slower, although it can be, but because of the DFA
  // caching problem. The closure routine only considers the rule invocation stack
  // created during prediction beginning in the decision rule. For example, if
  // prediction occurs without invoking another rule's ATN, there are no context
  // stacks in the configurations. When lack of context leads to a conflict, we
  // don't know if it's an ambiguity or a weakness in the strong LL(*) parsing
  // strategy (versus full LL(*)).</p>
  //
  // <p>
  // When SLL yields a configuration set with conflict, we rewind the input and
  // retry the ATN simulation, this time using full outer context without adding
  // to the DFA. Configuration context stacks will be the full invocation stacks
  // from the start rule. If we get a conflict using full context, then we can
  // definitively say we have a true ambiguity for that input sequence. If we
  // don't get a conflict, it implies that the decision is sensitive to the outer
  // context. (It is not context-sensitive in the sense of context-sensitive
  // grammars.)</p>
  //
  // <p>
  // The next time we reach this DFA state with an SLL conflict, through DFA
  // simulation, we will again retry the ATN simulation using full context mode.
  // This is slow because we can't save the results and have to "interpret" the
  // ATN each time we get that input.</p>
  //
  // <p>
  // <strong>CACHING FULL CONTEXT PREDICTIONS</strong></p>
  //
  // <p>
  // We could cache results from full context to predicted alternative easily and
  // that saves a lot of time but doesn't work in presence of predicates. The set
  // of visible predicates from the ATN start state changes depending on the
  // context, because closure can fall off the end of a rule. I tried to cache
  // tuples (stack context, semantic context, predicted alt) but it was slower
  // than interpreting and much more complicated. Also required a huge amount of
  // memory. The goal is not to create the world's fastest parser anyway. I'd like
  // to keep this algorithm simple. By launching multiple threads, we can improve
  // the speed of parsing across a large number of files.</p>
  //
  // <p>
  // There is no strict ordering between the amount of input used by SLL vs LL,
  // which makes it really hard to build a cache for full context. Let's say that
  // we have input A B C that leads to an SLL conflict with full context X. That
  // implies that using X we might only use A B but we could also use A B C D to
  // resolve conflict. Input A B C D could predict alternative 1 in one position
  // in the input and A B C E could predict alternative 2 in another position in
  // input. The conflicting SLL configurations could still be non-unique in the
  // full context prediction, which would lead us to requiring more input than the
  // original A B C.	To make a	prediction cache work, we have to track	the exact
  // input	used during the previous prediction. That amounts to a cache that maps
  // X to a specific DFA for that context.</p>
  //
  // <p>
  // Something should be done for left-recursive expression predictions. They are
  // likely LL(1) + pred eval. Easier to do the whole SLL unless error and retry
  // with full LL thing Sam does.</p>
  //
  // <p>
  // <strong>AVOIDING FULL CONTEXT PREDICTION</strong></p>
  //
  // <p>
  // We avoid doing full context retry when the outer context is empty, we did not
  // dip into the outer context by falling off the end of the decision state rule,
  // or when we force SLL mode.</p>
  //
  // <p>
  // As an example of the not dip into outer context case, consider as super
  // constructor calls versus function calls. One grammar might look like
  // this:</p>
  //
  // <pre>
  // ctorBody
  //   : '{' superCall? stat* '}'
  //   ;
  // </pre>
  //
  // <p>
  // Or, you might see something like</p>
  //
  // <pre>
  // stat
  //   : superCall ';'
  //   | expression ';'
  //   | ...
  //   ;
  // </pre>
  //
  // <p>
  // In both cases I believe that no closure operations will dip into the outer
  // context. In the first case ctorBody in the worst case will stop at the '}'.
  // In the 2nd case it should stop at the ';'. Both cases should stay within the
  // entry rule and not dip into the outer context.</p>
  //
  // <p>
  // <strong>PREDICATES</strong></p>
  //
  // <p>
  // Predicates are always evaluated if present in either SLL or LL both. SLL and
  // LL simulation deals with predicates differently. SLL collects predicates as
  // it performs closure operations like ANTLR v3 did. It delays predicate
  // evaluation until it reaches and accept state. This allows us to cache the SLL
  // ATN simulation whereas, if we had evaluated predicates on-the-fly during
  // closure, the DFA state configuration sets would be different and we couldn't
  // build up a suitable DFA.</p>
  //
  // <p>
  // When building a DFA accept state during ATN simulation, we evaluate any
  // predicates and return the sole semantically valid alternative. If there is
  // more than 1 alternative, we report an ambiguity. If there are 0 alternatives,
  // we throw an exception. Alternatives without predicates act like they have
  // true predicates. The simple way to think about it is to strip away all
  // alternatives with false predicates and choose the minimum alternative that
  // remains.</p>
  //
  // <p>
  // When we start in the DFA and reach an accept state that's predicated, we test
  // those and return the minimum semantically viable alternative. If no
  // alternatives are viable, we throw an exception.</p>
  //
  // <p>
  // During full LL ATN simulation, closure always evaluates predicates and
  // on-the-fly. This is crucial to reducing the configuration set size during
  // closure. It hits a landmine when parsing with the Java grammar, for example,
  // without this on-the-fly evaluation.</p>
  //
  // <p>
  // <strong>SHARING DFA</strong></p>
  //
  // <p>
  // All instances of the same parser share the same decision DFAs through a
  // static field. Each instance gets its own ATN simulator but they share the
  // same {@link //decisionToDFA} field. They also share a
  // {@link PredictionContextCache} object that makes sure that all
  // {@link PredictionContext} objects are shared among the DFA states. This makes
  // a big size difference.</p>
  //
  // <p>
  // <strong>THREAD SAFETY</strong></p>
  //
  // <p>
  // The {@link ParserATNSimulator} locks on the {@link //decisionToDFA} field when
  // it adds a new DFA object to that array. {@link //addDFAEdge}
  // locks on the DFA for the current decision when setting the
  // {@link DFAState//edges} field. {@link //addDFAState} locks on
  // the DFA for the current decision when looking up a DFA state to see if it
  // already exists. We must make sure that all requests to add DFA states that
  // are equivalent result in the same shared DFA object. This is because lots of
  // threads will be trying to update the DFA at once. The
  // {@link //addDFAState} method also locks inside the DFA lock
  // but this time on the shared context cache when it rebuilds the
  // configurations' {@link PredictionContext} objects using cached
  // subgraphs/nodes. No other locking occurs, even during DFA simulation. This is
  // safe as long as we can guarantee that all threads referencing
  // {@code s.edge[t]} get the same physical target {@link DFAState}, or
  // {@code null}. Once into the DFA, the DFA simulation does not reference the
  // {@link DFA//states} map. It follows the {@link DFAState//edges} field to new
  // targets. The DFA simulator will either find {@link DFAState//edges} to be
  // {@code null}, to be non-{@code null} and {@code dfa.edges[t]} null, or
  // {@code dfa.edges[t]} to be non-null. The
  // {@link //addDFAEdge} method could be racing to set the field
  // but in either case the DFA simulator works; if {@code null}, and requests ATN
  // simulation. It could also race trying to get {@code dfa.edges[t]}, but either
  // way it will work because it's not doing a test and set operation.</p>
  //
  // <p>
  // <strong>Starting with SLL then failing to combined SLL/LL (Two-Stage
  // Parsing)</strong></p>
  //
  // <p>
  // Sam pointed out that if SLL does not give a syntax error, then there is no
  // point in doing full LL, which is slower. We only have to try LL if we get a
  // syntax error. For maximum speed, Sam starts the parser set to pure SLL
  // mode with the {@link BailErrorStrategy}:</p>
  //
  // <pre>
  // parser.{@link Parser//getInterpreter() getInterpreter()}.{@link //setPredictionMode setPredictionMode}{@code (}{@link PredictionMode//SLL}{@code )};
  // parser.{@link Parser//setErrorHandler setErrorHandler}(new {@link BailErrorStrategy}());
  // </pre>
  //
  // <p>
  // If it does not get a syntax error, then we're done. If it does get a syntax
  // error, we need to retry with the combined SLL/LL strategy.</p>
  //
  // <p>
  // The reason this works is as follows. If there are no SLL conflicts, then the
  // grammar is SLL (at least for that input set). If there is an SLL conflict,
  // the full LL analysis must yield a set of viable alternatives which is a
  // subset of the alternatives reported by SLL. If the LL set is a singleton,
  // then the grammar is LL but not SLL. If the LL set is the same size as the SLL
  // set, the decision is SLL. If the LL set has size &gt; 1, then that decision
  // is truly ambiguous on the current input. If the LL set is smaller, then the
  // SLL conflict resolution might choose an alternative that the full LL would
  // rule out as a possibility based upon better context information. If that's
  // the case, then the SLL parse will definitely get an error because the full LL
  // analysis says it's not viable. If SLL conflict resolution chooses an
  // alternative within the LL set, them both SLL and LL would choose the same
  // alternative because they both choose the minimum of multiple conflicting
  // alternatives.</p>
  //
  // <p>
  // Let's say we have a set of SLL conflicting alternatives {@code {1, 2, 3}} and
  // a smaller LL set called <em>s</em>. If <em>s</em> is {@code {2, 3}}, then SLL
  // parsing will get an error because SLL will pursue alternative 1. If
  // <em>s</em> is {@code {1, 2}} or {@code {1, 3}} then both SLL and LL will
  // choose the same alternative because alternative one is the minimum of either
  // set. If <em>s</em> is {@code {2}} or {@code {3}} then SLL will get a syntax
  // error. If <em>s</em> is {@code {1}} then SLL will succeed.</p>
  //
  // <p>
  // Of course, if the input is invalid, then we will get an error for sure in
  // both SLL and LL parsing. Erroneous input will therefore require 2 passes over
  // the input.</p>
  //

  var Utils = require('./../Utils');
  var Set = Utils.Set;
  var BitSet = Utils.BitSet;
  var DoubleDict = Utils.DoubleDict;
  var ATN = require('./ATN').ATN;
  var ATNState = require('./ATNState').ATNState;
  var ATNConfig = require('./ATNConfig').ATNConfig;
  var ATNConfigSet = require('./ATNConfigSet').ATNConfigSet;
  var Token = require('./../Token').Token;
  var DFAState = require('./../dfa/DFAState').DFAState;
  var PredPrediction = require('./../dfa/DFAState').PredPrediction;
  var ATNSimulator = require('./ATNSimulator').ATNSimulator;
  var PredictionMode = require('./PredictionMode').PredictionMode;
  var RuleContext = require('./../RuleContext').RuleContext;
  var ParserRuleContext = require('./../ParserRuleContext').ParserRuleContext;
  var SemanticContext = require('./SemanticContext').SemanticContext;
  var StarLoopEntryState = require('./ATNState').StarLoopEntryState;
  var RuleStopState = require('./ATNState').RuleStopState;
  var PredictionContext = require('./../PredictionContext').PredictionContext;
  var Interval = require('./../IntervalSet').Interval;
  var Transitions = require('./Transition');
  var Transition = Transitions.Transition;
  var SetTransition = Transitions.SetTransition;
  var NotSetTransition = Transitions.NotSetTransition;
  var RuleTransition = Transitions.RuleTransition;
  var ActionTransition = Transitions.ActionTransition;
  var NoViableAltException = require('./../error/Errors').NoViableAltException;

  var SingletonPredictionContext = require('./../PredictionContext').SingletonPredictionContext;
  var predictionContextFromRuleContext = require('./../PredictionContext').predictionContextFromRuleContext;

  function ParserATNSimulator(parser, atn, decisionToDFA, sharedContextCache) {
  	ATNSimulator.call(this, atn, sharedContextCache);
      this.parser = parser;
      this.decisionToDFA = decisionToDFA;
      // SLL, LL, or LL + exact ambig detection?//
      this.predictionMode = PredictionMode.LL;
      // LAME globals to avoid parameters!!!!! I need these down deep in predTransition
      this._input = null;
      this._startIndex = 0;
      this._outerContext = null;
      this._dfa = null;
      // Each prediction operation uses a cache for merge of prediction contexts.
      //  Don't keep around as it wastes huge amounts of memory. DoubleKeyMap
      //  isn't synchronized but we're ok since two threads shouldn't reuse same
      //  parser/atnsim object because it can only handle one input at a time.
      //  This maps graphs a and b to merged result c. (a,b)&rarr;c. We can avoid
      //  the merge if we ever see a and b again.  Note that (b,a)&rarr;c should
      //  also be examined during cache lookup.
      //
      this.mergeCache = null;
      return this;
  }

  ParserATNSimulator.prototype = Object.create(ATNSimulator.prototype);
  ParserATNSimulator.prototype.constructor = ParserATNSimulator;

  ParserATNSimulator.prototype.debug = false;
  ParserATNSimulator.prototype.debug_closure = false;
  ParserATNSimulator.prototype.debug_add = false;
  ParserATNSimulator.prototype.debug_list_atn_decisions = false;
  ParserATNSimulator.prototype.dfa_debug = false;
  ParserATNSimulator.prototype.retry_debug = false;


  ParserATNSimulator.prototype.reset = function() {
  };

  ParserATNSimulator.prototype.adaptivePredict = function(input, decision, outerContext) {
      if (this.debug || this.debug_list_atn_decisions) {
          console.log("adaptivePredict decision " + decision +
                                 " exec LA(1)==" + this.getLookaheadName(input) +
                                 " line " + input.LT(1).line + ":" +
                                 input.LT(1).column);
      }
      this._input = input;
      this._startIndex = input.index;
      this._outerContext = outerContext;

      var dfa = this.decisionToDFA[decision];
      this._dfa = dfa;
      var m = input.mark();
      var index = input.index;

      // Now we are certain to have a specific decision's DFA
      // But, do we still need an initial state?
      try {
          var s0;
          if (dfa.precedenceDfa) {
              // the start state for a precedence DFA depends on the current
              // parser precedence, and is provided by a DFA method.
              s0 = dfa.getPrecedenceStartState(this.parser.getPrecedence());
          } else {
              // the start state for a "regular" DFA is just s0
              s0 = dfa.s0;
          }
          if (s0===null) {
              if (outerContext===null) {
                  outerContext = RuleContext.EMPTY;
              }
              if (this.debug || this.debug_list_atn_decisions) {
                  console.log("predictATN decision " + dfa.decision +
                                     " exec LA(1)==" + this.getLookaheadName(input) +
                                     ", outerContext=" + outerContext.toString(this.parser.ruleNames));
              }

              var fullCtx = false;
              var s0_closure = this.computeStartState(dfa.atnStartState, RuleContext.EMPTY, fullCtx);

              if( dfa.precedenceDfa) {
                  // If this is a precedence DFA, we use applyPrecedenceFilter
                  // to convert the computed start state to a precedence start
                  // state. We then use DFA.setPrecedenceStartState to set the
                  // appropriate start state for the precedence level rather
                  // than simply setting DFA.s0.
                  //
                  dfa.s0.configs = s0_closure; // not used for prediction but useful to know start configs anyway
                  s0_closure = this.applyPrecedenceFilter(s0_closure);
                  s0 = this.addDFAState(dfa, new DFAState(null, s0_closure));
                  dfa.setPrecedenceStartState(this.parser.getPrecedence(), s0);
              } else {
                  s0 = this.addDFAState(dfa, new DFAState(null, s0_closure));
                  dfa.s0 = s0;
              }
          }
          var alt = this.execATN(dfa, s0, input, index, outerContext);
          if (this.debug) {
              console.log("DFA after predictATN: " + dfa.toString(this.parser.literalNames));
          }
          return alt;
      } finally {
          this._dfa = null;
          this.mergeCache = null; // wack cache after each prediction
          input.seek(index);
          input.release(m);
      }
  };
  // Performs ATN simulation to compute a predicted alternative based
  //  upon the remaining input, but also updates the DFA cache to avoid
  //  having to traverse the ATN again for the same input sequence.

  // There are some key conditions we're looking for after computing a new
  // set of ATN configs (proposed DFA state):
        // if the set is empty, there is no viable alternative for current symbol
        // does the state uniquely predict an alternative?
        // does the state have a conflict that would prevent us from
        //   putting it on the work list?

  // We also have some key operations to do:
        // add an edge from previous DFA state to potentially new DFA state, D,
        //   upon current symbol but only if adding to work list, which means in all
        //   cases except no viable alternative (and possibly non-greedy decisions?)
        // collecting predicates and adding semantic context to DFA accept states
        // adding rule context to context-sensitive DFA accept states
        // consuming an input symbol
        // reporting a conflict
        // reporting an ambiguity
        // reporting a context sensitivity
        // reporting insufficient predicates

  // cover these cases:
  //    dead end
  //    single alt
  //    single alt + preds
  //    conflict
  //    conflict + preds
  //
  ParserATNSimulator.prototype.execATN = function(dfa, s0, input, startIndex, outerContext ) {
      if (this.debug || this.debug_list_atn_decisions) {
          console.log("execATN decision " + dfa.decision +
                  " exec LA(1)==" + this.getLookaheadName(input) +
                  " line " + input.LT(1).line + ":" + input.LT(1).column);
      }
      var alt;
      var previousD = s0;

      if (this.debug) {
          console.log("s0 = " + s0);
      }
      var t = input.LA(1);
      while(true) { // while more work
          var D = this.getExistingTargetState(previousD, t);
          if(D===null) {
              D = this.computeTargetState(dfa, previousD, t);
          }
          if(D===ATNSimulator.ERROR) {
              // if any configs in previous dipped into outer context, that
              // means that input up to t actually finished entry rule
              // at least for SLL decision. Full LL doesn't dip into outer
              // so don't need special case.
              // We will get an error no matter what so delay until after
              // decision; better error message. Also, no reachable target
              // ATN states in SLL implies LL will also get nowhere.
              // If conflict in states that dip out, choose min since we
              // will get error no matter what.
              var e = this.noViableAlt(input, outerContext, previousD.configs, startIndex);
              input.seek(startIndex);
              alt = this.getSynValidOrSemInvalidAltThatFinishedDecisionEntryRule(previousD.configs, outerContext);
              if(alt!==ATN.INVALID_ALT_NUMBER) {
                  return alt;
              } else {
                  throw e;
              }
          }
          if(D.requiresFullContext && this.predictionMode !== PredictionMode.SLL) {
              // IF PREDS, MIGHT RESOLVE TO SINGLE ALT => SLL (or syntax error)
              var conflictingAlts = null;
              if (D.predicates!==null) {
                  if (this.debug) {
                      console.log("DFA state has preds in DFA sim LL failover");
                  }
                  var conflictIndex = input.index;
                  if(conflictIndex !== startIndex) {
                      input.seek(startIndex);
                  }
                  conflictingAlts = this.evalSemanticContext(D.predicates, outerContext, true);
                  if (conflictingAlts.length===1) {
                      if(this.debug) {
                          console.log("Full LL avoided");
                      }
                      return conflictingAlts.minValue();
                  }
                  if (conflictIndex !== startIndex) {
                      // restore the index so reporting the fallback to full
                      // context occurs with the index at the correct spot
                      input.seek(conflictIndex);
                  }
              }
              if (this.dfa_debug) {
                  console.log("ctx sensitive state " + outerContext +" in " + D);
              }
              var fullCtx = true;
              var s0_closure = this.computeStartState(dfa.atnStartState, outerContext, fullCtx);
              this.reportAttemptingFullContext(dfa, conflictingAlts, D.configs, startIndex, input.index);
              alt = this.execATNWithFullContext(dfa, D, s0_closure, input, startIndex, outerContext);
              return alt;
          }
          if (D.isAcceptState) {
              if (D.predicates===null) {
                  return D.prediction;
              }
              var stopIndex = input.index;
              input.seek(startIndex);
              var alts = this.evalSemanticContext(D.predicates, outerContext, true);
              if (alts.length===0) {
                  throw this.noViableAlt(input, outerContext, D.configs, startIndex);
              } else if (alts.length===1) {
                  return alts.minValue();
              } else {
                  // report ambiguity after predicate evaluation to make sure the correct set of ambig alts is reported.
                  this.reportAmbiguity(dfa, D, startIndex, stopIndex, false, alts, D.configs);
                  return alts.minValue();
              }
          }
          previousD = D;

          if (t !== Token.EOF) {
              input.consume();
              t = input.LA(1);
          }
      }
  };
  //
  // Get an existing target state for an edge in the DFA. If the target state
  // for the edge has not yet been computed or is otherwise not available,
  // this method returns {@code null}.
  //
  // @param previousD The current DFA state
  // @param t The next input symbol
  // @return The existing target DFA state for the given input symbol
  // {@code t}, or {@code null} if the target state for this edge is not
  // already cached
  //
  ParserATNSimulator.prototype.getExistingTargetState = function(previousD, t) {
      var edges = previousD.edges;
      if (edges===null) {
          return null;
      } else {
          return edges[t + 1] || null;
      }
  };
  //
  // Compute a target state for an edge in the DFA, and attempt to add the
  // computed state and corresponding edge to the DFA.
  //
  // @param dfa The DFA
  // @param previousD The current DFA state
  // @param t The next input symbol
  //
  // @return The computed target DFA state for the given input symbol
  // {@code t}. If {@code t} does not lead to a valid DFA state, this method
  // returns {@link //ERROR}.
  //
  ParserATNSimulator.prototype.computeTargetState = function(dfa, previousD, t) {
     var reach = this.computeReachSet(previousD.configs, t, false);
      if(reach===null) {
          this.addDFAEdge(dfa, previousD, t, ATNSimulator.ERROR);
          return ATNSimulator.ERROR;
      }
      // create new target state; we'll add to DFA after it's complete
      var D = new DFAState(null, reach);

      var predictedAlt = this.getUniqueAlt(reach);

      if (this.debug) {
          var altSubSets = PredictionMode.getConflictingAltSubsets(reach);
          console.log("SLL altSubSets=" + Utils.arrayToString(altSubSets) +
                      ", previous=" + previousD.configs +
                      ", configs=" + reach +
                      ", predict=" + predictedAlt +
                      ", allSubsetsConflict=" +
                      PredictionMode.allSubsetsConflict(altSubSets) + ", conflictingAlts=" +
                      this.getConflictingAlts(reach));
      }
      if (predictedAlt!==ATN.INVALID_ALT_NUMBER) {
          // NO CONFLICT, UNIQUELY PREDICTED ALT
          D.isAcceptState = true;
          D.configs.uniqueAlt = predictedAlt;
          D.prediction = predictedAlt;
      } else if (PredictionMode.hasSLLConflictTerminatingPrediction(this.predictionMode, reach)) {
          // MORE THAN ONE VIABLE ALTERNATIVE
          D.configs.conflictingAlts = this.getConflictingAlts(reach);
          D.requiresFullContext = true;
          // in SLL-only mode, we will stop at this state and return the minimum alt
          D.isAcceptState = true;
          D.prediction = D.configs.conflictingAlts.minValue();
      }
      if (D.isAcceptState && D.configs.hasSemanticContext) {
          this.predicateDFAState(D, this.atn.getDecisionState(dfa.decision));
          if( D.predicates!==null) {
              D.prediction = ATN.INVALID_ALT_NUMBER;
          }
      }
      // all adds to dfa are done after we've created full D state
      D = this.addDFAEdge(dfa, previousD, t, D);
      return D;
  };

  ParserATNSimulator.prototype.predicateDFAState = function(dfaState, decisionState) {
      // We need to test all predicates, even in DFA states that
      // uniquely predict alternative.
      var nalts = decisionState.transitions.length;
      // Update DFA so reach becomes accept state with (predicate,alt)
      // pairs if preds found for conflicting alts
      var altsToCollectPredsFrom = this.getConflictingAltsOrUniqueAlt(dfaState.configs);
      var altToPred = this.getPredsForAmbigAlts(altsToCollectPredsFrom, dfaState.configs, nalts);
      if (altToPred!==null) {
          dfaState.predicates = this.getPredicatePredictions(altsToCollectPredsFrom, altToPred);
          dfaState.prediction = ATN.INVALID_ALT_NUMBER; // make sure we use preds
      } else {
          // There are preds in configs but they might go away
          // when OR'd together like {p}? || NONE == NONE. If neither
          // alt has preds, resolve to min alt
          dfaState.prediction = altsToCollectPredsFrom.minValue();
      }
  };

  // comes back with reach.uniqueAlt set to a valid alt
  ParserATNSimulator.prototype.execATNWithFullContext = function(dfa, D, // how far we got before failing over
                                       s0,
                                       input,
                                       startIndex,
                                       outerContext) {
      if (this.debug || this.debug_list_atn_decisions) {
          console.log("execATNWithFullContext "+s0);
      }
      var fullCtx = true;
      var foundExactAmbig = false;
      var reach = null;
      var previous = s0;
      input.seek(startIndex);
      var t = input.LA(1);
      var predictedAlt = -1;
      while (true) { // while more work
          reach = this.computeReachSet(previous, t, fullCtx);
          if (reach===null) {
              // if any configs in previous dipped into outer context, that
              // means that input up to t actually finished entry rule
              // at least for LL decision. Full LL doesn't dip into outer
              // so don't need special case.
              // We will get an error no matter what so delay until after
              // decision; better error message. Also, no reachable target
              // ATN states in SLL implies LL will also get nowhere.
              // If conflict in states that dip out, choose min since we
              // will get error no matter what.
              var e = this.noViableAlt(input, outerContext, previous, startIndex);
              input.seek(startIndex);
              var alt = this.getSynValidOrSemInvalidAltThatFinishedDecisionEntryRule(previous, outerContext);
              if(alt!==ATN.INVALID_ALT_NUMBER) {
                  return alt;
              } else {
                  throw e;
              }
          }
          var altSubSets = PredictionMode.getConflictingAltSubsets(reach);
          if(this.debug) {
              console.log("LL altSubSets=" + altSubSets + ", predict=" +
                    PredictionMode.getUniqueAlt(altSubSets) + ", resolvesToJustOneViableAlt=" +
                    PredictionMode.resolvesToJustOneViableAlt(altSubSets));
          }
          reach.uniqueAlt = this.getUniqueAlt(reach);
          // unique prediction?
          if(reach.uniqueAlt!==ATN.INVALID_ALT_NUMBER) {
              predictedAlt = reach.uniqueAlt;
              break;
          } else if (this.predictionMode !== PredictionMode.LL_EXACT_AMBIG_DETECTION) {
              predictedAlt = PredictionMode.resolvesToJustOneViableAlt(altSubSets);
              if(predictedAlt !== ATN.INVALID_ALT_NUMBER) {
                  break;
              }
          } else {
              // In exact ambiguity mode, we never try to terminate early.
              // Just keeps scarfing until we know what the conflict is
              if (PredictionMode.allSubsetsConflict(altSubSets) && PredictionMode.allSubsetsEqual(altSubSets)) {
                  foundExactAmbig = true;
                  predictedAlt = PredictionMode.getSingleViableAlt(altSubSets);
                  break;
              }
              // else there are multiple non-conflicting subsets or
              // we're not sure what the ambiguity is yet.
              // So, keep going.
          }
          previous = reach;
          if( t !== Token.EOF) {
              input.consume();
              t = input.LA(1);
          }
      }
      // If the configuration set uniquely predicts an alternative,
      // without conflict, then we know that it's a full LL decision
      // not SLL.
      if (reach.uniqueAlt !== ATN.INVALID_ALT_NUMBER ) {
          this.reportContextSensitivity(dfa, predictedAlt, reach, startIndex, input.index);
          return predictedAlt;
      }
      // We do not check predicates here because we have checked them
      // on-the-fly when doing full context prediction.

      //
      // In non-exact ambiguity detection mode, we might	actually be able to
      // detect an exact ambiguity, but I'm not going to spend the cycles
      // needed to check. We only emit ambiguity warnings in exact ambiguity
      // mode.
      //
      // For example, we might know that we have conflicting configurations.
      // But, that does not mean that there is no way forward without a
      // conflict. It's possible to have nonconflicting alt subsets as in:

      // altSubSets=[{1, 2}, {1, 2}, {1}, {1, 2}]

      // from
      //
      //    [(17,1,[5 $]), (13,1,[5 10 $]), (21,1,[5 10 $]), (11,1,[$]),
      //     (13,2,[5 10 $]), (21,2,[5 10 $]), (11,2,[$])]
      //
      // In this case, (17,1,[5 $]) indicates there is some next sequence that
      // would resolve this without conflict to alternative 1. Any other viable
      // next sequence, however, is associated with a conflict.  We stop
      // looking for input because no amount of further lookahead will alter
      // the fact that we should predict alternative 1.  We just can't say for
      // sure that there is an ambiguity without looking further.

      this.reportAmbiguity(dfa, D, startIndex, input.index, foundExactAmbig, null, reach);

      return predictedAlt;
  };

  ParserATNSimulator.prototype.computeReachSet = function(closure, t, fullCtx) {
      if (this.debug) {
          console.log("in computeReachSet, starting closure: " + closure);
      }
      if( this.mergeCache===null) {
          this.mergeCache = new DoubleDict();
      }
      var intermediate = new ATNConfigSet(fullCtx);

      // Configurations already in a rule stop state indicate reaching the end
      // of the decision rule (local context) or end of the start rule (full
      // context). Once reached, these configurations are never updated by a
      // closure operation, so they are handled separately for the performance
      // advantage of having a smaller intermediate set when calling closure.
      //
      // For full-context reach operations, separate handling is required to
      // ensure that the alternative matching the longest overall sequence is
      // chosen when multiple such configurations can match the input.

      var skippedStopStates = null;

      // First figure out where we can reach on input t
      for (var i=0; i<closure.items.length;i++) {
          var c = closure.items[i];
          if(this.debug_add) {
              console.log("testing " + this.getTokenName(t) + " at " + c);
          }
          if (c.state instanceof RuleStopState) {
              if (fullCtx || t === Token.EOF) {
                  if (skippedStopStates===null) {
                      skippedStopStates = [];
                  }
                  skippedStopStates.push(c);
                  if(this.debug_add) {
                      console.log("added " + c + " to skippedStopStates");
                  }
              }
              continue;
          }
          for(var j=0;j<c.state.transitions.length;j++) {
              var trans = c.state.transitions[j];
              var target = this.getReachableTarget(trans, t);
              if (target!==null) {
                  var cfg = new ATNConfig({state:target}, c);
                  intermediate.add(cfg, this.mergeCache);
                  if(this.debug_add) {
                      console.log("added " + cfg + " to intermediate");
                  }
              }
          }
      }
      // Now figure out where the reach operation can take us...
      var reach = null;

      // This block optimizes the reach operation for intermediate sets which
      // trivially indicate a termination state for the overall
      // adaptivePredict operation.
      //
      // The conditions assume that intermediate
      // contains all configurations relevant to the reach set, but this
      // condition is not true when one or more configurations have been
      // withheld in skippedStopStates, or when the current symbol is EOF.
      //
      if (skippedStopStates===null && t!==Token.EOF) {
          if (intermediate.items.length===1) {
              // Don't pursue the closure if there is just one state.
              // It can only have one alternative; just add to result
              // Also don't pursue the closure if there is unique alternative
              // among the configurations.
              reach = intermediate;
          } else if (this.getUniqueAlt(intermediate)!==ATN.INVALID_ALT_NUMBER) {
              // Also don't pursue the closure if there is unique alternative
              // among the configurations.
              reach = intermediate;
          }
      }
      // If the reach set could not be trivially determined, perform a closure
      // operation on the intermediate set to compute its initial value.
      //
      if (reach===null) {
          reach = new ATNConfigSet(fullCtx);
          var closureBusy = new Set();
          var treatEofAsEpsilon = t === Token.EOF;
          for (var k=0; k<intermediate.items.length;k++) {
              this.closure(intermediate.items[k], reach, closureBusy, false, fullCtx, treatEofAsEpsilon);
          }
      }
      if (t === Token.EOF) {
          // After consuming EOF no additional input is possible, so we are
          // only interested in configurations which reached the end of the
          // decision rule (local context) or end of the start rule (full
          // context). Update reach to contain only these configurations. This
          // handles both explicit EOF transitions in the grammar and implicit
          // EOF transitions following the end of the decision or start rule.
          //
          // When reach==intermediate, no closure operation was performed. In
          // this case, removeAllConfigsNotInRuleStopState needs to check for
          // reachable rule stop states as well as configurations already in
          // a rule stop state.
          //
          // This is handled before the configurations in skippedStopStates,
          // because any configurations potentially added from that list are
          // already guaranteed to meet this condition whether or not it's
          // required.
          //
          reach = this.removeAllConfigsNotInRuleStopState(reach, reach === intermediate);
      }
      // If skippedStopStates!==null, then it contains at least one
      // configuration. For full-context reach operations, these
      // configurations reached the end of the start rule, in which case we
      // only add them back to reach if no configuration during the current
      // closure operation reached such a state. This ensures adaptivePredict
      // chooses an alternative matching the longest overall sequence when
      // multiple alternatives are viable.
      //
      if (skippedStopStates!==null && ( (! fullCtx) || (! PredictionMode.hasConfigInRuleStopState(reach)))) {
          for (var l=0; l<skippedStopStates.length;l++) {
              reach.add(skippedStopStates[l], this.mergeCache);
          }
      }
      if (reach.items.length===0) {
          return null;
      } else {
          return reach;
      }
  };
  //
  // Return a configuration set containing only the configurations from
  // {@code configs} which are in a {@link RuleStopState}. If all
  // configurations in {@code configs} are already in a rule stop state, this
  // method simply returns {@code configs}.
  //
  // <p>When {@code lookToEndOfRule} is true, this method uses
  // {@link ATN//nextTokens} for each configuration in {@code configs} which is
  // not already in a rule stop state to see if a rule stop state is reachable
  // from the configuration via epsilon-only transitions.</p>
  //
  // @param configs the configuration set to update
  // @param lookToEndOfRule when true, this method checks for rule stop states
  // reachable by epsilon-only transitions from each configuration in
  // {@code configs}.
  //
  // @return {@code configs} if all configurations in {@code configs} are in a
  // rule stop state, otherwise return a new configuration set containing only
  // the configurations from {@code configs} which are in a rule stop state
  //
  ParserATNSimulator.prototype.removeAllConfigsNotInRuleStopState = function(configs, lookToEndOfRule) {
      if (PredictionMode.allConfigsInRuleStopStates(configs)) {
          return configs;
      }
      var result = new ATNConfigSet(configs.fullCtx);
      for(var i=0; i<configs.items.length;i++) {
          var config = configs.items[i];
          if (config.state instanceof RuleStopState) {
              result.add(config, this.mergeCache);
              continue;
          }
          if (lookToEndOfRule && config.state.epsilonOnlyTransitions) {
              var nextTokens = this.atn.nextTokens(config.state);
              if (nextTokens.contains(Token.EPSILON)) {
                  var endOfRuleState = this.atn.ruleToStopState[config.state.ruleIndex];
                  result.add(new ATNConfig({state:endOfRuleState}, config), this.mergeCache);
              }
          }
      }
      return result;
  };

  ParserATNSimulator.prototype.computeStartState = function(p, ctx, fullCtx) {
      // always at least the implicit call to start rule
      var initialContext = predictionContextFromRuleContext(this.atn, ctx);
      var configs = new ATNConfigSet(fullCtx);
      for(var i=0;i<p.transitions.length;i++) {
          var target = p.transitions[i].target;
          var c = new ATNConfig({ state:target, alt:i+1, context:initialContext }, null);
          var closureBusy = new Set();
          this.closure(c, configs, closureBusy, true, fullCtx, false);
      }
      return configs;
  };

  //
  // This method transforms the start state computed by
  // {@link //computeStartState} to the special start state used by a
  // precedence DFA for a particular precedence value. The transformation
  // process applies the following changes to the start state's configuration
  // set.
  //
  // <ol>
  // <li>Evaluate the precedence predicates for each configuration using
  // {@link SemanticContext//evalPrecedence}.</li>
  // <li>Remove all configurations which predict an alternative greater than
  // 1, for which another configuration that predicts alternative 1 is in the
  // same ATN state with the same prediction context. This transformation is
  // valid for the following reasons:
  // <ul>
  // <li>The closure block cannot contain any epsilon transitions which bypass
  // the body of the closure, so all states reachable via alternative 1 are
  // part of the precedence alternatives of the transformed left-recursive
  // rule.</li>
  // <li>The "primary" portion of a left recursive rule cannot contain an
  // epsilon transition, so the only way an alternative other than 1 can exist
  // in a state that is also reachable via alternative 1 is by nesting calls
  // to the left-recursive rule, with the outer calls not being at the
  // preferred precedence level.</li>
  // </ul>
  // </li>
  // </ol>
  //
  // <p>
  // The prediction context must be considered by this filter to address
  // situations like the following.
  // </p>
  // <code>
  // <pre>
  // grammar TA;
  // prog: statement* EOF;
  // statement: letterA | statement letterA 'b' ;
  // letterA: 'a';
  // </pre>
  // </code>
  // <p>
  // If the above grammar, the ATN state immediately before the token
  // reference {@code 'a'} in {@code letterA} is reachable from the left edge
  // of both the primary and closure blocks of the left-recursive rule
  // {@code statement}. The prediction context associated with each of these
  // configurations distinguishes between them, and prevents the alternative
  // which stepped out to {@code prog} (and then back in to {@code statement}
  // from being eliminated by the filter.
  // </p>
  //
  // @param configs The configuration set computed by
  // {@link //computeStartState} as the start state for the DFA.
  // @return The transformed configuration set representing the start state
  // for a precedence DFA at a particular precedence level (determined by
  // calling {@link Parser//getPrecedence}).
  //
  ParserATNSimulator.prototype.applyPrecedenceFilter = function(configs) {
  	var config;
  	var statesFromAlt1 = [];
      var configSet = new ATNConfigSet(configs.fullCtx);
      for(var i=0; i<configs.items.length; i++) {
          config = configs.items[i];
          // handle alt 1 first
          if (config.alt !== 1) {
              continue;
          }
          var updatedContext = config.semanticContext.evalPrecedence(this.parser, this._outerContext);
          if (updatedContext===null) {
              // the configuration was eliminated
              continue;
          }
          statesFromAlt1[config.state.stateNumber] = config.context;
          if (updatedContext !== config.semanticContext) {
              configSet.add(new ATNConfig({semanticContext:updatedContext}, config), this.mergeCache);
          } else {
              configSet.add(config, this.mergeCache);
          }
      }
      for(i=0; i<configs.items.length; i++) {
          config = configs.items[i];
          if (config.alt === 1) {
              // already handled
              continue;
          }
          // In the future, this elimination step could be updated to also
          // filter the prediction context for alternatives predicting alt>1
          // (basically a graph subtraction algorithm).
  		if (!config.precedenceFilterSuppressed) {
              var context = statesFromAlt1[config.state.stateNumber] || null;
              if (context!==null && context.equals(config.context)) {
                  // eliminated
                  continue;
              }
  		}
          configSet.add(config, this.mergeCache);
      }
      return configSet;
  };

  ParserATNSimulator.prototype.getReachableTarget = function(trans, ttype) {
      if (trans.matches(ttype, 0, this.atn.maxTokenType)) {
          return trans.target;
      } else {
          return null;
      }
  };

  ParserATNSimulator.prototype.getPredsForAmbigAlts = function(ambigAlts, configs, nalts) {
      // REACH=[1|1|[]|0:0, 1|2|[]|0:1]
      // altToPred starts as an array of all null contexts. The entry at index i
      // corresponds to alternative i. altToPred[i] may have one of three values:
      //   1. null: no ATNConfig c is found such that c.alt==i
      //   2. SemanticContext.NONE: At least one ATNConfig c exists such that
      //      c.alt==i and c.semanticContext==SemanticContext.NONE. In other words,
      //      alt i has at least one unpredicated config.
      //   3. Non-NONE Semantic Context: There exists at least one, and for all
      //      ATNConfig c such that c.alt==i, c.semanticContext!=SemanticContext.NONE.
      //
      // From this, it is clear that NONE||anything==NONE.
      //
      var altToPred = [];
      for(var i=0;i<configs.items.length;i++) {
          var c = configs.items[i];
          if(ambigAlts.contains( c.alt )) {
              altToPred[c.alt] = SemanticContext.orContext(altToPred[c.alt] || null, c.semanticContext);
          }
      }
      var nPredAlts = 0;
      for (i =1;i< nalts+1;i++) {
          var pred = altToPred[i] || null;
          if (pred===null) {
              altToPred[i] = SemanticContext.NONE;
          } else if (pred !== SemanticContext.NONE) {
              nPredAlts += 1;
          }
      }
      // nonambig alts are null in altToPred
      if (nPredAlts===0) {
          altToPred = null;
      }
      if (this.debug) {
          console.log("getPredsForAmbigAlts result " + Utils.arrayToString(altToPred));
      }
      return altToPred;
  };

  ParserATNSimulator.prototype.getPredicatePredictions = function(ambigAlts, altToPred) {
      var pairs = [];
      var containsPredicate = false;
      for (var i=1; i<altToPred.length;i++) {
          var pred = altToPred[i];
          // unpredicated is indicated by SemanticContext.NONE
          if( ambigAlts!==null && ambigAlts.contains( i )) {
              pairs.push(new PredPrediction(pred, i));
          }
          if (pred !== SemanticContext.NONE) {
              containsPredicate = true;
          }
      }
      if (! containsPredicate) {
          return null;
      }
      return pairs;
  };

  //
  // This method is used to improve the localization of error messages by
  // choosing an alternative rather than throwing a
  // {@link NoViableAltException} in particular prediction scenarios where the
  // {@link //ERROR} state was reached during ATN simulation.
  //
  // <p>
  // The default implementation of this method uses the following
  // algorithm to identify an ATN configuration which successfully parsed the
  // decision entry rule. Choosing such an alternative ensures that the
  // {@link ParserRuleContext} returned by the calling rule will be complete
  // and valid, and the syntax error will be reported later at a more
  // localized location.</p>
  //
  // <ul>
  // <li>If a syntactically valid path or paths reach the end of the decision rule and
  // they are semantically valid if predicated, return the min associated alt.</li>
  // <li>Else, if a semantically invalid but syntactically valid path exist
  // or paths exist, return the minimum associated alt.
  // </li>
  // <li>Otherwise, return {@link ATN//INVALID_ALT_NUMBER}.</li>
  // </ul>
  //
  // <p>
  // In some scenarios, the algorithm described above could predict an
  // alternative which will result in a {@link FailedPredicateException} in
  // the parser. Specifically, this could occur if the <em>only</em> configuration
  // capable of successfully parsing to the end of the decision rule is
  // blocked by a semantic predicate. By choosing this alternative within
  // {@link //adaptivePredict} instead of throwing a
  // {@link NoViableAltException}, the resulting
  // {@link FailedPredicateException} in the parser will identify the specific
  // predicate which is preventing the parser from successfully parsing the
  // decision rule, which helps developers identify and correct logic errors
  // in semantic predicates.
  // </p>
  //
  // @param configs The ATN configurations which were valid immediately before
  // the {@link //ERROR} state was reached
  // @param outerContext The is the \gamma_0 initial parser context from the paper
  // or the parser stack at the instant before prediction commences.
  //
  // @return The value to return from {@link //adaptivePredict}, or
  // {@link ATN//INVALID_ALT_NUMBER} if a suitable alternative was not
  // identified and {@link //adaptivePredict} should report an error instead.
  //
  ParserATNSimulator.prototype.getSynValidOrSemInvalidAltThatFinishedDecisionEntryRule = function(configs, outerContext) {
      var cfgs = this.splitAccordingToSemanticValidity(configs, outerContext);
      var semValidConfigs = cfgs[0];
      var semInvalidConfigs = cfgs[1];
      var alt = this.getAltThatFinishedDecisionEntryRule(semValidConfigs);
      if (alt!==ATN.INVALID_ALT_NUMBER) { // semantically/syntactically viable path exists
          return alt;
      }
      // Is there a syntactically valid path with a failed pred?
      if (semInvalidConfigs.items.length>0) {
          alt = this.getAltThatFinishedDecisionEntryRule(semInvalidConfigs);
          if (alt!==ATN.INVALID_ALT_NUMBER) { // syntactically viable path exists
              return alt;
          }
      }
      return ATN.INVALID_ALT_NUMBER;
  };

  ParserATNSimulator.prototype.getAltThatFinishedDecisionEntryRule = function(configs) {
      var alts = [];
      for(var i=0;i<configs.items.length; i++) {
          var c = configs.items[i];
          if (c.reachesIntoOuterContext>0 || ((c.state instanceof RuleStopState) && c.context.hasEmptyPath())) {
              if(alts.indexOf(c.alt)<0) {
                  alts.push(c.alt);
              }
          }
      }
      if (alts.length===0) {
          return ATN.INVALID_ALT_NUMBER;
      } else {
          return Math.min.apply(null, alts);
      }
  };
  // Walk the list of configurations and split them according to
  //  those that have preds evaluating to true/false.  If no pred, assume
  //  true pred and include in succeeded set.  Returns Pair of sets.
  //
  //  Create a new set so as not to alter the incoming parameter.
  //
  //  Assumption: the input stream has been restored to the starting point
  //  prediction, which is where predicates need to evaluate.
  //
  ParserATNSimulator.prototype.splitAccordingToSemanticValidity = function( configs, outerContext) {
      var succeeded = new ATNConfigSet(configs.fullCtx);
      var failed = new ATNConfigSet(configs.fullCtx);
      for(var i=0;i<configs.items.length; i++) {
          var c = configs.items[i];
          if (c.semanticContext !== SemanticContext.NONE) {
              var predicateEvaluationResult = c.semanticContext.evaluate(this.parser, outerContext);
              if (predicateEvaluationResult) {
                  succeeded.add(c);
              } else {
                  failed.add(c);
              }
          } else {
              succeeded.add(c);
          }
      }
      return [succeeded, failed];
  };

  // Look through a list of predicate/alt pairs, returning alts for the
  //  pairs that win. A {@code NONE} predicate indicates an alt containing an
  //  unpredicated config which behaves as "always true." If !complete
  //  then we stop at the first predicate that evaluates to true. This
  //  includes pairs with null predicates.
  //
  ParserATNSimulator.prototype.evalSemanticContext = function(predPredictions, outerContext, complete) {
      var predictions = new BitSet();
      for(var i=0;i<predPredictions.length;i++) {
      	var pair = predPredictions[i];
          if (pair.pred === SemanticContext.NONE) {
              predictions.add(pair.alt);
              if (! complete) {
                  break;
              }
              continue;
          }
          var predicateEvaluationResult = pair.pred.evaluate(this.parser, outerContext);
          if (this.debug || this.dfa_debug) {
              console.log("eval pred " + pair + "=" + predicateEvaluationResult);
          }
          if (predicateEvaluationResult) {
              if (this.debug || this.dfa_debug) {
                  console.log("PREDICT " + pair.alt);
              }
              predictions.add(pair.alt);
              if (! complete) {
                  break;
              }
          }
      }
      return predictions;
  };

  // TODO: If we are doing predicates, there is no point in pursuing
  //     closure operations if we reach a DFA state that uniquely predicts
  //     alternative. We will not be caching that DFA state and it is a
  //     waste to pursue the closure. Might have to advance when we do
  //     ambig detection thought :(
  //

  ParserATNSimulator.prototype.closure = function(config, configs, closureBusy, collectPredicates, fullCtx, treatEofAsEpsilon) {
      var initialDepth = 0;
      this.closureCheckingStopState(config, configs, closureBusy, collectPredicates,
                               fullCtx, initialDepth, treatEofAsEpsilon);
  };


  ParserATNSimulator.prototype.closureCheckingStopState = function(config, configs, closureBusy, collectPredicates, fullCtx, depth, treatEofAsEpsilon) {
      if (this.debug || this.debug_closure) {
          console.log("closure(" + config.toString(this.parser,true) + ")");
          // console.log("configs(" + configs.toString() + ")");
          if(config.reachesIntoOuterContext>50) {
              throw "problem";
          }
      }
      if (config.state instanceof RuleStopState) {
          // We hit rule end. If we have context info, use it
          // run thru all possible stack tops in ctx
          if (! config.context.isEmpty()) {
              for ( var i =0; i<config.context.length; i++) {
                  if (config.context.getReturnState(i) === PredictionContext.EMPTY_RETURN_STATE) {
                      if (fullCtx) {
                          configs.add(new ATNConfig({state:config.state, context:PredictionContext.EMPTY}, config), this.mergeCache);
                          continue;
                      } else {
                          // we have no context info, just chase follow links (if greedy)
                          if (this.debug) {
                              console.log("FALLING off rule " + this.getRuleName(config.state.ruleIndex));
                          }
                          this.closure_(config, configs, closureBusy, collectPredicates,
                                   fullCtx, depth, treatEofAsEpsilon);
                      }
                      continue;
                  }
                  var returnState = this.atn.states[config.context.getReturnState(i)];
                  var newContext = config.context.getParent(i); // "pop" return state
                  var parms = {state:returnState, alt:config.alt, context:newContext, semanticContext:config.semanticContext};
                  var c = new ATNConfig(parms, null);
                  // While we have context to pop back from, we may have
                  // gotten that context AFTER having falling off a rule.
                  // Make sure we track that we are now out of context.
                  c.reachesIntoOuterContext = config.reachesIntoOuterContext;
                  this.closureCheckingStopState(c, configs, closureBusy, collectPredicates, fullCtx, depth - 1, treatEofAsEpsilon);
              }
              return;
          } else if( fullCtx) {
              // reached end of start rule
              configs.add(config, this.mergeCache);
              return;
          } else {
              // else if we have no context info, just chase follow links (if greedy)
              if (this.debug) {
                  console.log("FALLING off rule " + this.getRuleName(config.state.ruleIndex));
              }
          }
      }
      this.closure_(config, configs, closureBusy, collectPredicates, fullCtx, depth, treatEofAsEpsilon);
  };


  // Do the actual work of walking epsilon edges//
  ParserATNSimulator.prototype.closure_ = function(config, configs, closureBusy, collectPredicates, fullCtx, depth, treatEofAsEpsilon) {
      var p = config.state;
      // optimization
      if (! p.epsilonOnlyTransitions) {
          configs.add(config, this.mergeCache);
          // make sure to not return here, because EOF transitions can act as
          // both epsilon transitions and non-epsilon transitions.
      }
      for(var i = 0;i<p.transitions.length; i++) {
          if(i==0 && this.canDropLoopEntryEdgeInLeftRecursiveRule(config))
              continue;

          var t = p.transitions[i];
          var continueCollecting = collectPredicates && !(t instanceof ActionTransition);
          var c = this.getEpsilonTarget(config, t, continueCollecting, depth === 0, fullCtx, treatEofAsEpsilon);
          if (c!==null) {
  			if (!t.isEpsilon && closureBusy.add(c)!==c){
  				// avoid infinite recursion for EOF* and EOF+
  				continue;
  			}
              var newDepth = depth;
              if ( config.state instanceof RuleStopState) {
                  // target fell off end of rule; mark resulting c as having dipped into outer context
                  // We can't get here if incoming config was rule stop and we had context
                  // track how far we dip into outer context.  Might
                  // come in handy and we avoid evaluating context dependent
                  // preds if this is > 0.

                  if (closureBusy.add(c)!==c) {
                      // avoid infinite recursion for right-recursive rules
                      continue;
                  }

  				if (this._dfa !== null && this._dfa.precedenceDfa) {
  					if (t.outermostPrecedenceReturn === this._dfa.atnStartState.ruleIndex) {
  						c.precedenceFilterSuppressed = true;
  					}
  				}

                  c.reachesIntoOuterContext += 1;
                  configs.dipsIntoOuterContext = true; // TODO: can remove? only care when we add to set per middle of this method
                  newDepth -= 1;
                  if (this.debug) {
                      console.log("dips into outer ctx: " + c);
                  }
              } else if (t instanceof RuleTransition) {
                  // latch when newDepth goes negative - once we step out of the entry context we can't return
                  if (newDepth >= 0) {
                      newDepth += 1;
                  }
              }
              this.closureCheckingStopState(c, configs, closureBusy, continueCollecting, fullCtx, newDepth, treatEofAsEpsilon);
          }
      }
  };


  ParserATNSimulator.prototype.canDropLoopEntryEdgeInLeftRecursiveRule = function(config) {
      // return False
      var p = config.state;
      // First check to see if we are in StarLoopEntryState generated during
      // left-recursion elimination. For efficiency, also check if
      // the context has an empty stack case. If so, it would mean
      // global FOLLOW so we can't perform optimization
      // Are we the special loop entry/exit state? or SLL wildcard
      if(p.stateType != ATNState.STAR_LOOP_ENTRY)
          return false;
      if(p.stateType != ATNState.STAR_LOOP_ENTRY || !p.isPrecedenceDecision ||
             config.context.isEmpty() || config.context.hasEmptyPath())
          return false;

      // Require all return states to return back to the same rule that p is in.
      var numCtxs = config.context.length;
      for(var i=0; i<numCtxs; i++) { // for each stack context
          var returnState = this.atn.states[config.context.getReturnState(i)];
          if (returnState.ruleIndex != p.ruleIndex)
              return false;
      }

      var decisionStartState = p.transitions[0].target;
      var blockEndStateNum = decisionStartState.endState.stateNumber;
      var blockEndState = this.atn.states[blockEndStateNum];

      // Verify that the top of each stack context leads to loop entry/exit
      // state through epsilon edges and w/o leaving rule.
      for(var i=0; i<numCtxs; i++) { // for each stack context
          var returnStateNumber = config.context.getReturnState(i);
          var returnState = this.atn.states[returnStateNumber];
          // all states must have single outgoing epsilon edge
          if (returnState.transitions.length != 1 || !returnState.transitions[0].isEpsilon)
              return false;

          // Look for prefix op case like 'not expr', (' type ')' expr
          var returnStateTarget = returnState.transitions[0].target;
          if ( returnState.stateType == ATNState.BLOCK_END && returnStateTarget == p )
              continue;

          // Look for 'expr op expr' or case where expr's return state is block end
          // of (...)* internal block; the block end points to loop back
          // which points to p but we don't need to check that
          if ( returnState == blockEndState )
              continue;

          // Look for ternary expr ? expr : expr. The return state points at block end,
          // which points at loop entry state
          if ( returnStateTarget == blockEndState )
              continue;

          // Look for complex prefix 'between expr and expr' case where 2nd expr's
          // return state points at block end state of (...)* internal block
          if (returnStateTarget.stateType == ATNState.BLOCK_END && returnStateTarget.transitions.length == 1
                  && returnStateTarget.transitions[0].isEpsilon && returnStateTarget.transitions[0].target == p)
              continue;

          // anything else ain't conforming
          return false;
      }
      return true;
  };


  ParserATNSimulator.prototype.getRuleName = function( index) {
      if (this.parser!==null && index>=0) {
          return this.parser.ruleNames[index];
      } else {
          return "<rule " + index + ">";
      }
  };

  ParserATNSimulator.prototype.getEpsilonTarget = function(config, t, collectPredicates, inContext, fullCtx, treatEofAsEpsilon) {
      switch(t.serializationType) {
      case Transition.RULE:
          return this.ruleTransition(config, t);
      case Transition.PRECEDENCE:
          return this.precedenceTransition(config, t, collectPredicates, inContext, fullCtx);
      case Transition.PREDICATE:
          return this.predTransition(config, t, collectPredicates, inContext, fullCtx);
      case Transition.ACTION:
          return this.actionTransition(config, t);
      case Transition.EPSILON:
          return new ATNConfig({state:t.target}, config);
      case Transition.ATOM:
      case Transition.RANGE:
      case Transition.SET:
          // EOF transitions act like epsilon transitions after the first EOF
          // transition is traversed
          if (treatEofAsEpsilon) {
              if (t.matches(Token.EOF, 0, 1)) {
                  return new ATNConfig({state: t.target}, config);
              }
          }
          return null;
      default:
      	return null;
      }
  };

  ParserATNSimulator.prototype.actionTransition = function(config, t) {
      if (this.debug) {
          var index = t.actionIndex==-1 ? 65535 : t.actionIndex;
          console.log("ACTION edge " + t.ruleIndex + ":" + index);
      }
      return new ATNConfig({state:t.target}, config);
  };

  ParserATNSimulator.prototype.precedenceTransition = function(config, pt,  collectPredicates, inContext, fullCtx) {
      if (this.debug) {
          console.log("PRED (collectPredicates=" + collectPredicates + ") " +
                  pt.precedence + ">=_p, ctx dependent=true");
          if (this.parser!==null) {
          	console.log("context surrounding pred is " + Utils.arrayToString(this.parser.getRuleInvocationStack()));
          }
      }
      var c = null;
      if (collectPredicates && inContext) {
          if (fullCtx) {
              // In full context mode, we can evaluate predicates on-the-fly
              // during closure, which dramatically reduces the size of
              // the config sets. It also obviates the need to test predicates
              // later during conflict resolution.
              var currentPosition = this._input.index;
              this._input.seek(this._startIndex);
              var predSucceeds = pt.getPredicate().evaluate(this.parser, this._outerContext);
              this._input.seek(currentPosition);
              if (predSucceeds) {
                  c = new ATNConfig({state:pt.target}, config); // no pred context
              }
          } else {
              var newSemCtx = SemanticContext.andContext(config.semanticContext, pt.getPredicate());
              c = new ATNConfig({state:pt.target, semanticContext:newSemCtx}, config);
          }
      } else {
          c = new ATNConfig({state:pt.target}, config);
      }
      if (this.debug) {
          console.log("config from pred transition=" + c);
      }
      return c;
  };

  ParserATNSimulator.prototype.predTransition = function(config, pt, collectPredicates, inContext, fullCtx) {
      if (this.debug) {
          console.log("PRED (collectPredicates=" + collectPredicates + ") " + pt.ruleIndex +
                  ":" + pt.predIndex + ", ctx dependent=" + pt.isCtxDependent);
          if (this.parser!==null) {
              console.log("context surrounding pred is " + Utils.arrayToString(this.parser.getRuleInvocationStack()));
          }
      }
      var c = null;
      if (collectPredicates && ((pt.isCtxDependent && inContext) || ! pt.isCtxDependent)) {
          if (fullCtx) {
              // In full context mode, we can evaluate predicates on-the-fly
              // during closure, which dramatically reduces the size of
              // the config sets. It also obviates the need to test predicates
              // later during conflict resolution.
              var currentPosition = this._input.index;
              this._input.seek(this._startIndex);
              var predSucceeds = pt.getPredicate().evaluate(this.parser, this._outerContext);
              this._input.seek(currentPosition);
              if (predSucceeds) {
                  c = new ATNConfig({state:pt.target}, config); // no pred context
              }
          } else {
              var newSemCtx = SemanticContext.andContext(config.semanticContext, pt.getPredicate());
              c = new ATNConfig({state:pt.target, semanticContext:newSemCtx}, config);
          }
      } else {
          c = new ATNConfig({state:pt.target}, config);
      }
      if (this.debug) {
          console.log("config from pred transition=" + c);
      }
      return c;
  };

  ParserATNSimulator.prototype.ruleTransition = function(config, t) {
      if (this.debug) {
          console.log("CALL rule " + this.getRuleName(t.target.ruleIndex) + ", ctx=" + config.context);
      }
      var returnState = t.followState;
      var newContext = SingletonPredictionContext.create(config.context, returnState.stateNumber);
      return new ATNConfig({state:t.target, context:newContext}, config );
  };

  ParserATNSimulator.prototype.getConflictingAlts = function(configs) {
      var altsets = PredictionMode.getConflictingAltSubsets(configs);
      return PredictionMode.getAlts(altsets);
  };

   // Sam pointed out a problem with the previous definition, v3, of
   // ambiguous states. If we have another state associated with conflicting
   // alternatives, we should keep going. For example, the following grammar
   //
   // s : (ID | ID ID?) ';' ;
   //
   // When the ATN simulation reaches the state before ';', it has a DFA
   // state that looks like: [12|1|[], 6|2|[], 12|2|[]]. Naturally
   // 12|1|[] and 12|2|[] conflict, but we cannot stop processing this node
   // because alternative to has another way to continue, via [6|2|[]].
   // The key is that we have a single state that has config's only associated
   // with a single alternative, 2, and crucially the state transitions
   // among the configurations are all non-epsilon transitions. That means
   // we don't consider any conflicts that include alternative 2. So, we
   // ignore the conflict between alts 1 and 2. We ignore a set of
   // conflicting alts when there is an intersection with an alternative
   // associated with a single alt state in the state&rarr;config-list map.
   //
   // It's also the case that we might have two conflicting configurations but
   // also a 3rd nonconflicting configuration for a different alternative:
   // [1|1|[], 1|2|[], 8|3|[]]. This can come about from grammar:
   //
   // a : A | A | A B ;
   //
   // After matching input A, we reach the stop state for rule A, state 1.
   // State 8 is the state right before B. Clearly alternatives 1 and 2
   // conflict and no amount of further lookahead will separate the two.
   // However, alternative 3 will be able to continue and so we do not
   // stop working on this state. In the previous example, we're concerned
   // with states associated with the conflicting alternatives. Here alt
   // 3 is not associated with the conflicting configs, but since we can continue
   // looking for input reasonably, I don't declare the state done. We
   // ignore a set of conflicting alts when we have an alternative
   // that we still need to pursue.
  //

  ParserATNSimulator.prototype.getConflictingAltsOrUniqueAlt = function(configs) {
      var conflictingAlts = null;
      if (configs.uniqueAlt!== ATN.INVALID_ALT_NUMBER) {
          conflictingAlts = new BitSet();
          conflictingAlts.add(configs.uniqueAlt);
      } else {
          conflictingAlts = configs.conflictingAlts;
      }
      return conflictingAlts;
  };

  ParserATNSimulator.prototype.getTokenName = function( t) {
      if (t===Token.EOF) {
          return "EOF";
      }
      if( this.parser!==null && this.parser.literalNames!==null) {
          if (t >= this.parser.literalNames.length && t >= this.parser.symbolicNames.length) {
              console.log("" + t + " ttype out of range: " + this.parser.literalNames);
              console.log("" + this.parser.getInputStream().getTokens());
          } else {
              var name = this.parser.literalNames[t] || this.parser.symbolicNames[t];
              return name + "<" + t + ">";
          }
      }
      return "" + t;
  };

  ParserATNSimulator.prototype.getLookaheadName = function(input) {
      return this.getTokenName(input.LA(1));
  };

  // Used for debugging in adaptivePredict around execATN but I cut
  //  it out for clarity now that alg. works well. We can leave this
  //  "dead" code for a bit.
  //
  ParserATNSimulator.prototype.dumpDeadEndConfigs = function(nvae) {
      console.log("dead end configs: ");
      var decs = nvae.getDeadEndConfigs();
      for(var i=0; i<decs.length; i++) {
      	var c = decs[i];
          var trans = "no edges";
          if (c.state.transitions.length>0) {
              var t = c.state.transitions[0];
              if (t instanceof AtomTransition) {
                  trans = "Atom "+ this.getTokenName(t.label);
              } else if (t instanceof SetTransition) {
                  var neg = (t instanceof NotSetTransition);
                  trans = (neg ? "~" : "") + "Set " + t.set;
              }
          }
          console.error(c.toString(this.parser, true) + ":" + trans);
      }
  };

  ParserATNSimulator.prototype.noViableAlt = function(input, outerContext, configs, startIndex) {
      return new NoViableAltException(this.parser, input, input.get(startIndex), input.LT(1), configs, outerContext);
  };

  ParserATNSimulator.prototype.getUniqueAlt = function(configs) {
      var alt = ATN.INVALID_ALT_NUMBER;
      for(var i=0;i<configs.items.length;i++) {
      	var c = configs.items[i];
          if (alt === ATN.INVALID_ALT_NUMBER) {
              alt = c.alt; // found first alt
          } else if( c.alt!==alt) {
              return ATN.INVALID_ALT_NUMBER;
          }
      }
      return alt;
  };

  //
  // Add an edge to the DFA, if possible. This method calls
  // {@link //addDFAState} to ensure the {@code to} state is present in the
  // DFA. If {@code from} is {@code null}, or if {@code t} is outside the
  // range of edges that can be represented in the DFA tables, this method
  // returns without adding the edge to the DFA.
  //
  // <p>If {@code to} is {@code null}, this method returns {@code null}.
  // Otherwise, this method returns the {@link DFAState} returned by calling
  // {@link //addDFAState} for the {@code to} state.</p>
  //
  // @param dfa The DFA
  // @param from The source state for the edge
  // @param t The input symbol
  // @param to The target state for the edge
  //
  // @return If {@code to} is {@code null}, this method returns {@code null};
  // otherwise this method returns the result of calling {@link //addDFAState}
  // on {@code to}
  //
  ParserATNSimulator.prototype.addDFAEdge = function(dfa, from_, t, to) {
      if( this.debug) {
          console.log("EDGE " + from_ + " -> " + to + " upon " + this.getTokenName(t));
      }
      if (to===null) {
          return null;
      }
      to = this.addDFAState(dfa, to); // used existing if possible not incoming
      if (from_===null || t < -1 || t > this.atn.maxTokenType) {
          return to;
      }
      if (from_.edges===null) {
          from_.edges = [];
      }
      from_.edges[t+1] = to; // connect

      if (this.debug) {
          var literalNames = this.parser===null ? null : this.parser.literalNames;
          var symbolicNames = this.parser===null ? null : this.parser.symbolicNames;
          console.log("DFA=\n" + dfa.toString(literalNames, symbolicNames));
      }
      return to;
  };
  //
  // Add state {@code D} to the DFA if it is not already present, and return
  // the actual instance stored in the DFA. If a state equivalent to {@code D}
  // is already in the DFA, the existing state is returned. Otherwise this
  // method returns {@code D} after adding it to the DFA.
  //
  // <p>If {@code D} is {@link //ERROR}, this method returns {@link //ERROR} and
  // does not change the DFA.</p>
  //
  // @param dfa The dfa
  // @param D The DFA state to add
  // @return The state stored in the DFA. This will be either the existing
  // state if {@code D} is already in the DFA, or {@code D} itself if the
  // state was not already present.
  //
  ParserATNSimulator.prototype.addDFAState = function(dfa, D) {
      if (D == ATNSimulator.ERROR) {
          return D;
      }
      var existing = dfa.states.get(D);
      if(existing!==null) {
          return existing;
      }
      D.stateNumber = dfa.states.length;
      if (! D.configs.readOnly) {
          D.configs.optimizeConfigs(this);
          D.configs.setReadonly(true);
      }
      dfa.states.add(D);
      if (this.debug) {
          console.log("adding new DFA state: " + D);
      }
      return D;
  };

  ParserATNSimulator.prototype.reportAttemptingFullContext = function(dfa, conflictingAlts, configs, startIndex, stopIndex) {
      if (this.debug || this.retry_debug) {
          var interval = new Interval(startIndex, stopIndex + 1);
          console.log("reportAttemptingFullContext decision=" + dfa.decision + ":" + configs +
                             ", input=" + this.parser.getTokenStream().getText(interval));
      }
      if (this.parser!==null) {
          this.parser.getErrorListenerDispatch().reportAttemptingFullContext(this.parser, dfa, startIndex, stopIndex, conflictingAlts, configs);
      }
  };

  ParserATNSimulator.prototype.reportContextSensitivity = function(dfa, prediction, configs, startIndex, stopIndex) {
      if (this.debug || this.retry_debug) {
          var interval = new Interval(startIndex, stopIndex + 1);
          console.log("reportContextSensitivity decision=" + dfa.decision + ":" + configs +
                             ", input=" + this.parser.getTokenStream().getText(interval));
      }
      if (this.parser!==null) {
          this.parser.getErrorListenerDispatch().reportContextSensitivity(this.parser, dfa, startIndex, stopIndex, prediction, configs);
      }
  };

  // If context sensitive parsing, we know it's ambiguity not conflict//
  ParserATNSimulator.prototype.reportAmbiguity = function(dfa, D, startIndex, stopIndex,
                                 exact, ambigAlts, configs ) {
      if (this.debug || this.retry_debug) {
          var interval = new Interval(startIndex, stopIndex + 1);
          console.log("reportAmbiguity " + ambigAlts + ":" + configs +
                             ", input=" + this.parser.getTokenStream().getText(interval));
      }
      if (this.parser!==null) {
          this.parser.getErrorListenerDispatch().reportAmbiguity(this.parser, dfa, startIndex, stopIndex, exact, ambigAlts, configs);
      }
  };

  exports.ParserATNSimulator = ParserATNSimulator;
  },{"./../IntervalSet":6,"./../ParserRuleContext":10,"./../PredictionContext":11,"./../RuleContext":13,"./../Token":14,"./../Utils":15,"./../dfa/DFAState":34,"./../error/Errors":39,"./ATN":16,"./ATNConfig":17,"./ATNConfigSet":18,"./ATNSimulator":21,"./ATNState":22,"./PredictionMode":28,"./SemanticContext":29,"./Transition":30}],28:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  //
  //
  // This enumeration defines the prediction modes available in ANTLR 4 along with
  // utility methods for analyzing configuration sets for conflicts and/or
  // ambiguities.

  var Set = require('./../Utils').Set;
  var Map = require('./../Utils').Map;
  var BitSet = require('./../Utils').BitSet;
  var AltDict = require('./../Utils').AltDict;
  var ATN = require('./ATN').ATN;
  var RuleStopState = require('./ATNState').RuleStopState;
  var ATNConfigSet = require('./ATNConfigSet').ATNConfigSet;
  var ATNConfig = require('./ATNConfig').ATNConfig;
  var SemanticContext = require('./SemanticContext').SemanticContext;
  var Hash = require("../Utils").Hash;
  var hashStuff = require('./../Utils').hashStuff;
  var equalArrays = require('./../Utils').equalArrays;

  function PredictionMode() {
  	return this;
  }

  //
  // The SLL(*) prediction mode. This prediction mode ignores the current
  // parser context when making predictions. This is the fastest prediction
  // mode, and provides correct results for many grammars. This prediction
  // mode is more powerful than the prediction mode provided by ANTLR 3, but
  // may result in syntax errors for grammar and input combinations which are
  // not SLL.
  //
  // <p>
  // When using this prediction mode, the parser will either return a correct
  // parse tree (i.e. the same parse tree that would be returned with the
  // {@link //LL} prediction mode), or it will report a syntax error. If a
  // syntax error is encountered when using the {@link //SLL} prediction mode,
  // it may be due to either an actual syntax error in the input or indicate
  // that the particular combination of grammar and input requires the more
  // powerful {@link //LL} prediction abilities to complete successfully.</p>
  //
  // <p>
  // This prediction mode does not provide any guarantees for prediction
  // behavior for syntactically-incorrect inputs.</p>
  //
  PredictionMode.SLL = 0;
  //
  // The LL(*) prediction mode. This prediction mode allows the current parser
  // context to be used for resolving SLL conflicts that occur during
  // prediction. This is the fastest prediction mode that guarantees correct
  // parse results for all combinations of grammars with syntactically correct
  // inputs.
  //
  // <p>
  // When using this prediction mode, the parser will make correct decisions
  // for all syntactically-correct grammar and input combinations. However, in
  // cases where the grammar is truly ambiguous this prediction mode might not
  // report a precise answer for <em>exactly which</em> alternatives are
  // ambiguous.</p>
  //
  // <p>
  // This prediction mode does not provide any guarantees for prediction
  // behavior for syntactically-incorrect inputs.</p>
  //
  PredictionMode.LL = 1;
  //
  // The LL(*) prediction mode with exact ambiguity detection. In addition to
  // the correctness guarantees provided by the {@link //LL} prediction mode,
  // this prediction mode instructs the prediction algorithm to determine the
  // complete and exact set of ambiguous alternatives for every ambiguous
  // decision encountered while parsing.
  //
  // <p>
  // This prediction mode may be used for diagnosing ambiguities during
  // grammar development. Due to the performance overhead of calculating sets
  // of ambiguous alternatives, this prediction mode should be avoided when
  // the exact results are not necessary.</p>
  //
  // <p>
  // This prediction mode does not provide any guarantees for prediction
  // behavior for syntactically-incorrect inputs.</p>
  //
  PredictionMode.LL_EXACT_AMBIG_DETECTION = 2;


  //
  // Computes the SLL prediction termination condition.
  //
  // <p>
  // This method computes the SLL prediction termination condition for both of
  // the following cases.</p>
  //
  // <ul>
  // <li>The usual SLL+LL fallback upon SLL conflict</li>
  // <li>Pure SLL without LL fallback</li>
  // </ul>
  //
  // <p><strong>COMBINED SLL+LL PARSING</strong></p>
  //
  // <p>When LL-fallback is enabled upon SLL conflict, correct predictions are
  // ensured regardless of how the termination condition is computed by this
  // method. Due to the substantially higher cost of LL prediction, the
  // prediction should only fall back to LL when the additional lookahead
  // cannot lead to a unique SLL prediction.</p>
  //
  // <p>Assuming combined SLL+LL parsing, an SLL configuration set with only
  // conflicting subsets should fall back to full LL, even if the
  // configuration sets don't resolve to the same alternative (e.g.
  // {@code {1,2}} and {@code {3,4}}. If there is at least one non-conflicting
  // configuration, SLL could continue with the hopes that more lookahead will
  // resolve via one of those non-conflicting configurations.</p>
  //
  // <p>Here's the prediction termination rule them: SLL (for SLL+LL parsing)
  // stops when it sees only conflicting configuration subsets. In contrast,
  // full LL keeps going when there is uncertainty.</p>
  //
  // <p><strong>HEURISTIC</strong></p>
  //
  // <p>As a heuristic, we stop prediction when we see any conflicting subset
  // unless we see a state that only has one alternative associated with it.
  // The single-alt-state thing lets prediction continue upon rules like
  // (otherwise, it would admit defeat too soon):</p>
  //
  // <p>{@code [12|1|[], 6|2|[], 12|2|[]]. s : (ID | ID ID?) ';' ;}</p>
  //
  // <p>When the ATN simulation reaches the state before {@code ';'}, it has a
  // DFA state that looks like: {@code [12|1|[], 6|2|[], 12|2|[]]}. Naturally
  // {@code 12|1|[]} and {@code 12|2|[]} conflict, but we cannot stop
  // processing this node because alternative to has another way to continue,
  // via {@code [6|2|[]]}.</p>
  //
  // <p>It also let's us continue for this rule:</p>
  //
  // <p>{@code [1|1|[], 1|2|[], 8|3|[]] a : A | A | A B ;}</p>
  //
  // <p>After matching input A, we reach the stop state for rule A, state 1.
  // State 8 is the state right before B. Clearly alternatives 1 and 2
  // conflict and no amount of further lookahead will separate the two.
  // However, alternative 3 will be able to continue and so we do not stop
  // working on this state. In the previous example, we're concerned with
  // states associated with the conflicting alternatives. Here alt 3 is not
  // associated with the conflicting configs, but since we can continue
  // looking for input reasonably, don't declare the state done.</p>
  //
  // <p><strong>PURE SLL PARSING</strong></p>
  //
  // <p>To handle pure SLL parsing, all we have to do is make sure that we
  // combine stack contexts for configurations that differ only by semantic
  // predicate. From there, we can do the usual SLL termination heuristic.</p>
  //
  // <p><strong>PREDICATES IN SLL+LL PARSING</strong></p>
  //
  // <p>SLL decisions don't evaluate predicates until after they reach DFA stop
  // states because they need to create the DFA cache that works in all
  // semantic situations. In contrast, full LL evaluates predicates collected
  // during start state computation so it can ignore predicates thereafter.
  // This means that SLL termination detection can totally ignore semantic
  // predicates.</p>
  //
  // <p>Implementation-wise, {@link ATNConfigSet} combines stack contexts but not
  // semantic predicate contexts so we might see two configurations like the
  // following.</p>
  //
  // <p>{@code (s, 1, x, {}), (s, 1, x', {p})}</p>
  //
  // <p>Before testing these configurations against others, we have to merge
  // {@code x} and {@code x'} (without modifying the existing configurations).
  // For example, we test {@code (x+x')==x''} when looking for conflicts in
  // the following configurations.</p>
  //
  // <p>{@code (s, 1, x, {}), (s, 1, x', {p}), (s, 2, x'', {})}</p>
  //
  // <p>If the configuration set has predicates (as indicated by
  // {@link ATNConfigSet//hasSemanticContext}), this algorithm makes a copy of
  // the configurations to strip out all of the predicates so that a standard
  // {@link ATNConfigSet} will merge everything ignoring predicates.</p>
  //
  PredictionMode.hasSLLConflictTerminatingPrediction = function( mode, configs) {
      // Configs in rule stop states indicate reaching the end of the decision
      // rule (local context) or end of start rule (full context). If all
      // configs meet this condition, then none of the configurations is able
      // to match additional input so we terminate prediction.
      //
      if (PredictionMode.allConfigsInRuleStopStates(configs)) {
          return true;
      }
      // pure SLL mode parsing
      if (mode === PredictionMode.SLL) {
          // Don't bother with combining configs from different semantic
          // contexts if we can fail over to full LL; costs more time
          // since we'll often fail over anyway.
          if (configs.hasSemanticContext) {
              // dup configs, tossing out semantic predicates
              var dup = new ATNConfigSet();
              for(var i=0;i<configs.items.length;i++) {
              	var c = configs.items[i];
                  c = new ATNConfig({semanticContext:SemanticContext.NONE}, c);
                  dup.add(c);
              }
              configs = dup;
          }
          // now we have combined contexts for configs with dissimilar preds
      }
      // pure SLL or combined SLL+LL mode parsing
      var altsets = PredictionMode.getConflictingAltSubsets(configs);
      return PredictionMode.hasConflictingAltSet(altsets) && !PredictionMode.hasStateAssociatedWithOneAlt(configs);
  };

  // Checks if any configuration in {@code configs} is in a
  // {@link RuleStopState}. Configurations meeting this condition have reached
  // the end of the decision rule (local context) or end of start rule (full
  // context).
  //
  // @param configs the configuration set to test
  // @return {@code true} if any configuration in {@code configs} is in a
  // {@link RuleStopState}, otherwise {@code false}
  PredictionMode.hasConfigInRuleStopState = function(configs) {
  	for(var i=0;i<configs.items.length;i++) {
  		var c = configs.items[i];
          if (c.state instanceof RuleStopState) {
              return true;
          }
  	}
      return false;
  };

  // Checks if all configurations in {@code configs} are in a
  // {@link RuleStopState}. Configurations meeting this condition have reached
  // the end of the decision rule (local context) or end of start rule (full
  // context).
  //
  // @param configs the configuration set to test
  // @return {@code true} if all configurations in {@code configs} are in a
  // {@link RuleStopState}, otherwise {@code false}
  PredictionMode.allConfigsInRuleStopStates = function(configs) {
  	for(var i=0;i<configs.items.length;i++) {
  		var c = configs.items[i];
          if (!(c.state instanceof RuleStopState)) {
              return false;
          }
  	}
      return true;
  };

  //
  // Full LL prediction termination.
  //
  // <p>Can we stop looking ahead during ATN simulation or is there some
  // uncertainty as to which alternative we will ultimately pick, after
  // consuming more input? Even if there are partial conflicts, we might know
  // that everything is going to resolve to the same minimum alternative. That
  // means we can stop since no more lookahead will change that fact. On the
  // other hand, there might be multiple conflicts that resolve to different
  // minimums. That means we need more look ahead to decide which of those
  // alternatives we should predict.</p>
  //
  // <p>The basic idea is to split the set of configurations {@code C}, into
  // conflicting subsets {@code (s, _, ctx, _)} and singleton subsets with
  // non-conflicting configurations. Two configurations conflict if they have
  // identical {@link ATNConfig//state} and {@link ATNConfig//context} values
  // but different {@link ATNConfig//alt} value, e.g. {@code (s, i, ctx, _)}
  // and {@code (s, j, ctx, _)} for {@code i!=j}.</p>
  //
  // <p>Reduce these configuration subsets to the set of possible alternatives.
  // You can compute the alternative subsets in one pass as follows:</p>
  //
  // <p>{@code A_s,ctx = {i | (s, i, ctx, _)}} for each configuration in
  // {@code C} holding {@code s} and {@code ctx} fixed.</p>
  //
  // <p>Or in pseudo-code, for each configuration {@code c} in {@code C}:</p>
  //
  // <pre>
  // map[c] U= c.{@link ATNConfig//alt alt} // map hash/equals uses s and x, not
  // alt and not pred
  // </pre>
  //
  // <p>The values in {@code map} are the set of {@code A_s,ctx} sets.</p>
  //
  // <p>If {@code |A_s,ctx|=1} then there is no conflict associated with
  // {@code s} and {@code ctx}.</p>
  //
  // <p>Reduce the subsets to singletons by choosing a minimum of each subset. If
  // the union of these alternative subsets is a singleton, then no amount of
  // more lookahead will help us. We will always pick that alternative. If,
  // however, there is more than one alternative, then we are uncertain which
  // alternative to predict and must continue looking for resolution. We may
  // or may not discover an ambiguity in the future, even if there are no
  // conflicting subsets this round.</p>
  //
  // <p>The biggest sin is to terminate early because it means we've made a
  // decision but were uncertain as to the eventual outcome. We haven't used
  // enough lookahead. On the other hand, announcing a conflict too late is no
  // big deal; you will still have the conflict. It's just inefficient. It
  // might even look until the end of file.</p>
  //
  // <p>No special consideration for semantic predicates is required because
  // predicates are evaluated on-the-fly for full LL prediction, ensuring that
  // no configuration contains a semantic context during the termination
  // check.</p>
  //
  // <p><strong>CONFLICTING CONFIGS</strong></p>
  //
  // <p>Two configurations {@code (s, i, x)} and {@code (s, j, x')}, conflict
  // when {@code i!=j} but {@code x=x'}. Because we merge all
  // {@code (s, i, _)} configurations together, that means that there are at
  // most {@code n} configurations associated with state {@code s} for
  // {@code n} possible alternatives in the decision. The merged stacks
  // complicate the comparison of configuration contexts {@code x} and
  // {@code x'}. Sam checks to see if one is a subset of the other by calling
  // merge and checking to see if the merged result is either {@code x} or
  // {@code x'}. If the {@code x} associated with lowest alternative {@code i}
  // is the superset, then {@code i} is the only possible prediction since the
  // others resolve to {@code min(i)} as well. However, if {@code x} is
  // associated with {@code j>i} then at least one stack configuration for
  // {@code j} is not in conflict with alternative {@code i}. The algorithm
  // should keep going, looking for more lookahead due to the uncertainty.</p>
  //
  // <p>For simplicity, I'm doing a equality check between {@code x} and
  // {@code x'} that lets the algorithm continue to consume lookahead longer
  // than necessary. The reason I like the equality is of course the
  // simplicity but also because that is the test you need to detect the
  // alternatives that are actually in conflict.</p>
  //
  // <p><strong>CONTINUE/STOP RULE</strong></p>
  //
  // <p>Continue if union of resolved alternative sets from non-conflicting and
  // conflicting alternative subsets has more than one alternative. We are
  // uncertain about which alternative to predict.</p>
  //
  // <p>The complete set of alternatives, {@code [i for (_,i,_)]}, tells us which
  // alternatives are still in the running for the amount of input we've
  // consumed at this point. The conflicting sets let us to strip away
  // configurations that won't lead to more states because we resolve
  // conflicts to the configuration with a minimum alternate for the
  // conflicting set.</p>
  //
  // <p><strong>CASES</strong></p>
  //
  // <ul>
  //
  // <li>no conflicts and more than 1 alternative in set =&gt; continue</li>
  //
  // <li> {@code (s, 1, x)}, {@code (s, 2, x)}, {@code (s, 3, z)},
  // {@code (s', 1, y)}, {@code (s', 2, y)} yields non-conflicting set
  // {@code {3}} U conflicting sets {@code min({1,2})} U {@code min({1,2})} =
  // {@code {1,3}} =&gt; continue
  // </li>
  //
  // <li>{@code (s, 1, x)}, {@code (s, 2, x)}, {@code (s', 1, y)},
  // {@code (s', 2, y)}, {@code (s'', 1, z)} yields non-conflicting set
  // {@code {1}} U conflicting sets {@code min({1,2})} U {@code min({1,2})} =
  // {@code {1}} =&gt; stop and predict 1</li>
  //
  // <li>{@code (s, 1, x)}, {@code (s, 2, x)}, {@code (s', 1, y)},
  // {@code (s', 2, y)} yields conflicting, reduced sets {@code {1}} U
  // {@code {1}} = {@code {1}} =&gt; stop and predict 1, can announce
  // ambiguity {@code {1,2}}</li>
  //
  // <li>{@code (s, 1, x)}, {@code (s, 2, x)}, {@code (s', 2, y)},
  // {@code (s', 3, y)} yields conflicting, reduced sets {@code {1}} U
  // {@code {2}} = {@code {1,2}} =&gt; continue</li>
  //
  // <li>{@code (s, 1, x)}, {@code (s, 2, x)}, {@code (s', 3, y)},
  // {@code (s', 4, y)} yields conflicting, reduced sets {@code {1}} U
  // {@code {3}} = {@code {1,3}} =&gt; continue</li>
  //
  // </ul>
  //
  // <p><strong>EXACT AMBIGUITY DETECTION</strong></p>
  //
  // <p>If all states report the same conflicting set of alternatives, then we
  // know we have the exact ambiguity set.</p>
  //
  // <p><code>|A_<em>i</em>|&gt;1</code> and
  // <code>A_<em>i</em> = A_<em>j</em></code> for all <em>i</em>, <em>j</em>.</p>
  //
  // <p>In other words, we continue examining lookahead until all {@code A_i}
  // have more than one alternative and all {@code A_i} are the same. If
  // {@code A={{1,2}, {1,3}}}, then regular LL prediction would terminate
  // because the resolved set is {@code {1}}. To determine what the real
  // ambiguity is, we have to know whether the ambiguity is between one and
  // two or one and three so we keep going. We can only stop prediction when
  // we need exact ambiguity detection when the sets look like
  // {@code A={{1,2}}} or {@code {{1,2},{1,2}}}, etc...</p>
  //
  PredictionMode.resolvesToJustOneViableAlt = function(altsets) {
      return PredictionMode.getSingleViableAlt(altsets);
  };

  //
  // Determines if every alternative subset in {@code altsets} contains more
  // than one alternative.
  //
  // @param altsets a collection of alternative subsets
  // @return {@code true} if every {@link BitSet} in {@code altsets} has
  // {@link BitSet//cardinality cardinality} &gt; 1, otherwise {@code false}
  //
  PredictionMode.allSubsetsConflict = function(altsets) {
      return ! PredictionMode.hasNonConflictingAltSet(altsets);
  };
  //
  // Determines if any single alternative subset in {@code altsets} contains
  // exactly one alternative.
  //
  // @param altsets a collection of alternative subsets
  // @return {@code true} if {@code altsets} contains a {@link BitSet} with
  // {@link BitSet//cardinality cardinality} 1, otherwise {@code false}
  //
  PredictionMode.hasNonConflictingAltSet = function(altsets) {
  	for(var i=0;i<altsets.length;i++) {
  		var alts = altsets[i];
          if (alts.length===1) {
              return true;
          }
  	}
      return false;
  };

  //
  // Determines if any single alternative subset in {@code altsets} contains
  // more than one alternative.
  //
  // @param altsets a collection of alternative subsets
  // @return {@code true} if {@code altsets} contains a {@link BitSet} with
  // {@link BitSet//cardinality cardinality} &gt; 1, otherwise {@code false}
  //
  PredictionMode.hasConflictingAltSet = function(altsets) {
  	for(var i=0;i<altsets.length;i++) {
  		var alts = altsets[i];
          if (alts.length>1) {
              return true;
          }
  	}
      return false;
  };

  //
  // Determines if every alternative subset in {@code altsets} is equivalent.
  //
  // @param altsets a collection of alternative subsets
  // @return {@code true} if every member of {@code altsets} is equal to the
  // others, otherwise {@code false}
  //
  PredictionMode.allSubsetsEqual = function(altsets) {
      var first = null;
  	for(var i=0;i<altsets.length;i++) {
  		var alts = altsets[i];
          if (first === null) {
              first = alts;
          } else if (alts!==first) {
              return false;
          }
  	}
      return true;
  };

  //
  // Returns the unique alternative predicted by all alternative subsets in
  // {@code altsets}. If no such alternative exists, this method returns
  // {@link ATN//INVALID_ALT_NUMBER}.
  //
  // @param altsets a collection of alternative subsets
  //
  PredictionMode.getUniqueAlt = function(altsets) {
      var all = PredictionMode.getAlts(altsets);
      if (all.length===1) {
          return all.minValue();
      } else {
          return ATN.INVALID_ALT_NUMBER;
      }
  };

  // Gets the complete set of represented alternatives for a collection of
  // alternative subsets. This method returns the union of each {@link BitSet}
  // in {@code altsets}.
  //
  // @param altsets a collection of alternative subsets
  // @return the set of represented alternatives in {@code altsets}
  //
  PredictionMode.getAlts = function(altsets) {
      var all = new BitSet();
      altsets.map( function(alts) { all.or(alts); });
      return all;
  };

  //
  // This function gets the conflicting alt subsets from a configuration set.
  // For each configuration {@code c} in {@code configs}:
  //
  // <pre>
  // map[c] U= c.{@link ATNConfig//alt alt} // map hash/equals uses s and x, not
  // alt and not pred
  // </pre>

  PredictionMode.getConflictingAltSubsets = function(configs) {
      var configToAlts = new Map();
      configToAlts.hashFunction = function(cfg) { hashStuff(cfg.state.stateNumber, cfg.context); };
      configToAlts.equalsFunction = function(c1, c2) { return c1.state.stateNumber==c2.state.stateNumber && c1.context.equals(c2.context);};
      configs.items.map(function(cfg) {
          var alts = configToAlts.get(cfg);
          if (alts === null) {
              alts = new BitSet();
              configToAlts.put(cfg, alts);
          }
          alts.add(cfg.alt);
  	});
      return configToAlts.getValues();
  };

  //
  // Get a map from state to alt subset from a configuration set. For each
  // configuration {@code c} in {@code configs}:
  //
  // <pre>
  // map[c.{@link ATNConfig//state state}] U= c.{@link ATNConfig//alt alt}
  // </pre>
  //
  PredictionMode.getStateToAltMap = function(configs) {
      var m = new AltDict();
      configs.items.map(function(c) {
          var alts = m.get(c.state);
          if (alts === null) {
              alts = new BitSet();
              m.put(c.state, alts);
          }
          alts.add(c.alt);
      });
      return m;
  };

  PredictionMode.hasStateAssociatedWithOneAlt = function(configs) {
      var values = PredictionMode.getStateToAltMap(configs).values();
      for(var i=0;i<values.length;i++) {
          if (values[i].length===1) {
              return true;
          }
      }
      return false;
  };

  PredictionMode.getSingleViableAlt = function(altsets) {
      var result = null;
  	for(var i=0;i<altsets.length;i++) {
  		var alts = altsets[i];
          var minAlt = alts.minValue();
          if(result===null) {
              result = minAlt;
          } else if(result!==minAlt) { // more than 1 viable alt
              return ATN.INVALID_ALT_NUMBER;
          }
  	}
      return result;
  };

  exports.PredictionMode = PredictionMode;

  },{"../Utils":15,"./../Utils":15,"./ATN":16,"./ATNConfig":17,"./ATNConfigSet":18,"./ATNState":22,"./SemanticContext":29}],29:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  //

  // A tree structure used to record the semantic context in which
  //  an ATN configuration is valid.  It's either a single predicate,
  //  a conjunction {@code p1&&p2}, or a sum of products {@code p1||p2}.
  //
  //  <p>I have scoped the {@link AND}, {@link OR}, and {@link Predicate} subclasses of
  //  {@link SemanticContext} within the scope of this outer class.</p>
  //

  var Set = require('./../Utils').Set;
  var Hash = require('./../Utils').Hash;

  function SemanticContext() {
  	return this;
  }

  SemanticContext.prototype.hashCode = function() {
      var hash = new Hash();
      this.updateHashCode(hash);
      return hash.finish();
  };

  // For context independent predicates, we evaluate them without a local
  // context (i.e., null context). That way, we can evaluate them without
  // having to create proper rule-specific context during prediction (as
  // opposed to the parser, which creates them naturally). In a practical
  // sense, this avoids a cast exception from RuleContext to myruleContext.
  //
  // <p>For context dependent predicates, we must pass in a local context so that
  // references such as $arg evaluate properly as _localctx.arg. We only
  // capture context dependent predicates in the context in which we begin
  // prediction, so we passed in the outer context here in case of context
  // dependent predicate evaluation.</p>
  //
  SemanticContext.prototype.evaluate = function(parser, outerContext) {
  };

  //
  // Evaluate the precedence predicates for the context and reduce the result.
  //
  // @param parser The parser instance.
  // @param outerContext The current parser context object.
  // @return The simplified semantic context after precedence predicates are
  // evaluated, which will be one of the following values.
  // <ul>
  // <li>{@link //NONE}: if the predicate simplifies to {@code true} after
  // precedence predicates are evaluated.</li>
  // <li>{@code null}: if the predicate simplifies to {@code false} after
  // precedence predicates are evaluated.</li>
  // <li>{@code this}: if the semantic context is not changed as a result of
  // precedence predicate evaluation.</li>
  // <li>A non-{@code null} {@link SemanticContext}: the new simplified
  // semantic context after precedence predicates are evaluated.</li>
  // </ul>
  //
  SemanticContext.prototype.evalPrecedence = function(parser, outerContext) {
  	return this;
  };

  SemanticContext.andContext = function(a, b) {
  	if (a === null || a === SemanticContext.NONE) {
  		return b;
  	}
  	if (b === null || b === SemanticContext.NONE) {
  		return a;
  	}
  	var result = new AND(a, b);
  	if (result.opnds.length === 1) {
  		return result.opnds[0];
  	} else {
  		return result;
  	}
  };

  SemanticContext.orContext = function(a, b) {
  	if (a === null) {
  		return b;
  	}
  	if (b === null) {
  		return a;
  	}
  	if (a === SemanticContext.NONE || b === SemanticContext.NONE) {
  		return SemanticContext.NONE;
  	}
  	var result = new OR(a, b);
  	if (result.opnds.length === 1) {
  		return result.opnds[0];
  	} else {
  		return result;
  	}
  };

  function Predicate(ruleIndex, predIndex, isCtxDependent) {
  	SemanticContext.call(this);
  	this.ruleIndex = ruleIndex === undefined ? -1 : ruleIndex;
  	this.predIndex = predIndex === undefined ? -1 : predIndex;
  	this.isCtxDependent = isCtxDependent === undefined ? false : isCtxDependent; // e.g., $i ref in pred
  	return this;
  }

  Predicate.prototype = Object.create(SemanticContext.prototype);
  Predicate.prototype.constructor = Predicate;

  //The default {@link SemanticContext}, which is semantically equivalent to
  //a predicate of the form {@code {true}?}.
  //
  SemanticContext.NONE = new Predicate();


  Predicate.prototype.evaluate = function(parser, outerContext) {
  	var localctx = this.isCtxDependent ? outerContext : null;
  	return parser.sempred(localctx, this.ruleIndex, this.predIndex);
  };

  Predicate.prototype.updateHashCode = function(hash) {
  	hash.update(this.ruleIndex, this.predIndex, this.isCtxDependent);
  };

  Predicate.prototype.equals = function(other) {
  	if (this === other) {
  		return true;
  	} else if (!(other instanceof Predicate)) {
  		return false;
  	} else {
  		return this.ruleIndex === other.ruleIndex &&
  				this.predIndex === other.predIndex &&
  				this.isCtxDependent === other.isCtxDependent;
  	}
  };

  Predicate.prototype.toString = function() {
  	return "{" + this.ruleIndex + ":" + this.predIndex + "}?";
  };

  function PrecedencePredicate(precedence) {
  	SemanticContext.call(this);
  	this.precedence = precedence === undefined ? 0 : precedence;
  }

  PrecedencePredicate.prototype = Object.create(SemanticContext.prototype);
  PrecedencePredicate.prototype.constructor = PrecedencePredicate;

  PrecedencePredicate.prototype.evaluate = function(parser, outerContext) {
  	return parser.precpred(outerContext, this.precedence);
  };

  PrecedencePredicate.prototype.evalPrecedence = function(parser, outerContext) {
  	if (parser.precpred(outerContext, this.precedence)) {
  		return SemanticContext.NONE;
  	} else {
  		return null;
  	}
  };

  PrecedencePredicate.prototype.compareTo = function(other) {
  	return this.precedence - other.precedence;
  };

  PrecedencePredicate.prototype.updateHashCode = function(hash) {
      hash.update(31);
  };

  PrecedencePredicate.prototype.equals = function(other) {
  	if (this === other) {
  		return true;
  	} else if (!(other instanceof PrecedencePredicate)) {
  		return false;
  	} else {
  		return this.precedence === other.precedence;
  	}
  };

  PrecedencePredicate.prototype.toString = function() {
  	return "{"+this.precedence+">=prec}?";
  };



  PrecedencePredicate.filterPrecedencePredicates = function(set) {
  	var result = [];
  	set.values().map( function(context) {
  		if (context instanceof PrecedencePredicate) {
  			result.push(context);
  		}
  	});
  	return result;
  };


  // A semantic context which is true whenever none of the contained contexts
  // is false.
  //
  function AND(a, b) {
  	SemanticContext.call(this);
  	var operands = new Set();
  	if (a instanceof AND) {
  		a.opnds.map(function(o) {
  			operands.add(o);
  		});
  	} else {
  		operands.add(a);
  	}
  	if (b instanceof AND) {
  		b.opnds.map(function(o) {
  			operands.add(o);
  		});
  	} else {
  		operands.add(b);
  	}
  	var precedencePredicates = PrecedencePredicate.filterPrecedencePredicates(operands);
  	if (precedencePredicates.length > 0) {
  		// interested in the transition with the lowest precedence
  		var reduced = null;
  		precedencePredicates.map( function(p) {
  			if(reduced===null || p.precedence<reduced.precedence) {
  				reduced = p;
  			}
  		});
  		operands.add(reduced);
  	}
  	this.opnds = operands.values();
  	return this;
  }

  AND.prototype = Object.create(SemanticContext.prototype);
  AND.prototype.constructor = AND;

  AND.prototype.equals = function(other) {
  	if (this === other) {
  		return true;
  	} else if (!(other instanceof AND)) {
  		return false;
  	} else {
  		return this.opnds === other.opnds;
  	}
  };

  AND.prototype.updateHashCode = function(hash) {
      hash.update(this.opnds, "AND");
  };
  //
  // {@inheritDoc}
  //
  // <p>
  // The evaluation of predicates by this context is short-circuiting, but
  // unordered.</p>
  //
  AND.prototype.evaluate = function(parser, outerContext) {
  	for (var i = 0; i < this.opnds.length; i++) {
  		if (!this.opnds[i].evaluate(parser, outerContext)) {
  			return false;
  		}
  	}
  	return true;
  };

  AND.prototype.evalPrecedence = function(parser, outerContext) {
  	var differs = false;
  	var operands = [];
  	for (var i = 0; i < this.opnds.length; i++) {
  		var context = this.opnds[i];
  		var evaluated = context.evalPrecedence(parser, outerContext);
  		differs |= (evaluated !== context);
  		if (evaluated === null) {
  			// The AND context is false if any element is false
  			return null;
  		} else if (evaluated !== SemanticContext.NONE) {
  			// Reduce the result by skipping true elements
  			operands.push(evaluated);
  		}
  	}
  	if (!differs) {
  		return this;
  	}
  	if (operands.length === 0) {
  		// all elements were true, so the AND context is true
  		return SemanticContext.NONE;
  	}
  	var result = null;
  	operands.map(function(o) {
  		result = result === null ? o : SemanticContext.andContext(result, o);
  	});
  	return result;
  };

  AND.prototype.toString = function() {
  	var s = "";
  	this.opnds.map(function(o) {
  		s += "&& " + o.toString();
  	});
  	return s.length > 3 ? s.slice(3) : s;
  };

  //
  // A semantic context which is true whenever at least one of the contained
  // contexts is true.
  //
  function OR(a, b) {
  	SemanticContext.call(this);
  	var operands = new Set();
  	if (a instanceof OR) {
  		a.opnds.map(function(o) {
  			operands.add(o);
  		});
  	} else {
  		operands.add(a);
  	}
  	if (b instanceof OR) {
  		b.opnds.map(function(o) {
  			operands.add(o);
  		});
  	} else {
  		operands.add(b);
  	}

  	var precedencePredicates = PrecedencePredicate.filterPrecedencePredicates(operands);
  	if (precedencePredicates.length > 0) {
  		// interested in the transition with the highest precedence
  		var s = precedencePredicates.sort(function(a, b) {
  			return a.compareTo(b);
  		});
  		var reduced = s[s.length-1];
  		operands.add(reduced);
  	}
  	this.opnds = operands.values();
  	return this;
  }

  OR.prototype = Object.create(SemanticContext.prototype);
  OR.prototype.constructor = OR;

  OR.prototype.constructor = function(other) {
  	if (this === other) {
  		return true;
  	} else if (!(other instanceof OR)) {
  		return false;
  	} else {
  		return this.opnds === other.opnds;
  	}
  };

  OR.prototype.updateHashCode = function(hash) {
      hash.update(this.opnds, "OR");
  };

  // <p>
  // The evaluation of predicates by this context is short-circuiting, but
  // unordered.</p>
  //
  OR.prototype.evaluate = function(parser, outerContext) {
  	for (var i = 0; i < this.opnds.length; i++) {
  		if (this.opnds[i].evaluate(parser, outerContext)) {
  			return true;
  		}
  	}
  	return false;
  };

  OR.prototype.evalPrecedence = function(parser, outerContext) {
  	var differs = false;
  	var operands = [];
  	for (var i = 0; i < this.opnds.length; i++) {
  		var context = this.opnds[i];
  		var evaluated = context.evalPrecedence(parser, outerContext);
  		differs |= (evaluated !== context);
  		if (evaluated === SemanticContext.NONE) {
  			// The OR context is true if any element is true
  			return SemanticContext.NONE;
  		} else if (evaluated !== null) {
  			// Reduce the result by skipping false elements
  			operands.push(evaluated);
  		}
  	}
  	if (!differs) {
  		return this;
  	}
  	if (operands.length === 0) {
  		// all elements were false, so the OR context is false
  		return null;
  	}
  	var result = null;
  	operands.map(function(o) {
  		return result === null ? o : SemanticContext.orContext(result, o);
  	});
  	return result;
  };

  OR.prototype.toString = function() {
  	var s = "";
  	this.opnds.map(function(o) {
  		s += "|| " + o.toString();
  	});
  	return s.length > 3 ? s.slice(3) : s;
  };

  exports.SemanticContext = SemanticContext;
  exports.PrecedencePredicate = PrecedencePredicate;
  exports.Predicate = Predicate;

  },{"./../Utils":15}],30:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  //

  //  An ATN transition between any two ATN states.  Subclasses define
  //  atom, set, epsilon, action, predicate, rule transitions.
  //
  //  <p>This is a one way link.  It emanates from a state (usually via a list of
  //  transitions) and has a target state.</p>
  //
  //  <p>Since we never have to change the ATN transitions once we construct it,
  //  we can fix these transitions as specific classes. The DFA transitions
  //  on the other hand need to update the labels as it adds transitions to
  //  the states. We'll use the term Edge for the DFA to distinguish them from
  //  ATN transitions.</p>

  var Token = require('./../Token').Token;
  var Interval = require('./../IntervalSet').Interval;
  var IntervalSet = require('./../IntervalSet').IntervalSet;
  var Predicate = require('./SemanticContext').Predicate;
  var PrecedencePredicate = require('./SemanticContext').PrecedencePredicate;

  function Transition (target) {
      // The target of this transition.
      if (target===undefined || target===null) {
          throw "target cannot be null.";
      }
      this.target = target;
      // Are we epsilon, action, sempred?
      this.isEpsilon = false;
      this.label = null;
      return this;
  }
      // constants for serialization
  Transition.EPSILON = 1;
  Transition.RANGE = 2;
  Transition.RULE = 3;
  Transition.PREDICATE = 4; // e.g., {isType(input.LT(1))}?
  Transition.ATOM = 5;
  Transition.ACTION = 6;
  Transition.SET = 7; // ~(A|B) or ~atom, wildcard, which convert to next 2
  Transition.NOT_SET = 8;
  Transition.WILDCARD = 9;
  Transition.PRECEDENCE = 10;

  Transition.serializationNames = [
              "INVALID",
              "EPSILON",
              "RANGE",
              "RULE",
              "PREDICATE",
              "ATOM",
              "ACTION",
              "SET",
              "NOT_SET",
              "WILDCARD",
              "PRECEDENCE"
          ];

  Transition.serializationTypes = {
          EpsilonTransition: Transition.EPSILON,
          RangeTransition: Transition.RANGE,
          RuleTransition: Transition.RULE,
          PredicateTransition: Transition.PREDICATE,
          AtomTransition: Transition.ATOM,
          ActionTransition: Transition.ACTION,
          SetTransition: Transition.SET,
          NotSetTransition: Transition.NOT_SET,
          WildcardTransition: Transition.WILDCARD,
          PrecedencePredicateTransition: Transition.PRECEDENCE
      };


  // TODO: make all transitions sets? no, should remove set edges
  function AtomTransition(target, label) {
  	Transition.call(this, target);
  	this.label_ = label; // The token type or character value; or, signifies special label.
      this.label = this.makeLabel();
      this.serializationType = Transition.ATOM;
      return this;
  }

  AtomTransition.prototype = Object.create(Transition.prototype);
  AtomTransition.prototype.constructor = AtomTransition;

  AtomTransition.prototype.makeLabel = function() {
  	var s = new IntervalSet();
      s.addOne(this.label_);
      return s;
  };

  AtomTransition.prototype.matches = function( symbol, minVocabSymbol,  maxVocabSymbol) {
      return this.label_ === symbol;
  };

  AtomTransition.prototype.toString = function() {
  	return this.label_;
  };

  function RuleTransition(ruleStart, ruleIndex, precedence, followState) {
  	Transition.call(this, ruleStart);
      this.ruleIndex = ruleIndex; // ptr to the rule definition object for this rule ref
      this.precedence = precedence;
      this.followState = followState; // what node to begin computations following ref to rule
      this.serializationType = Transition.RULE;
      this.isEpsilon = true;
      return this;
  }

  RuleTransition.prototype = Object.create(Transition.prototype);
  RuleTransition.prototype.constructor = RuleTransition;

  RuleTransition.prototype.matches = function(symbol, minVocabSymbol,  maxVocabSymbol) {
  	return false;
  };


  function EpsilonTransition(target, outermostPrecedenceReturn) {
  	Transition.call(this, target);
      this.serializationType = Transition.EPSILON;
      this.isEpsilon = true;
      this.outermostPrecedenceReturn = outermostPrecedenceReturn;
      return this;
  }

  EpsilonTransition.prototype = Object.create(Transition.prototype);
  EpsilonTransition.prototype.constructor = EpsilonTransition;

  EpsilonTransition.prototype.matches = function( symbol, minVocabSymbol,  maxVocabSymbol) {
  	return false;
  };

  EpsilonTransition.prototype.toString = function() {
  	return "epsilon";
  };

  function RangeTransition(target, start, stop) {
  	Transition.call(this, target);
  	this.serializationType = Transition.RANGE;
      this.start = start;
      this.stop = stop;
      this.label = this.makeLabel();
      return this;
  }

  RangeTransition.prototype = Object.create(Transition.prototype);
  RangeTransition.prototype.constructor = RangeTransition;

  RangeTransition.prototype.makeLabel = function() {
      var s = new IntervalSet();
      s.addRange(this.start, this.stop);
      return s;
  };

  RangeTransition.prototype.matches = function(symbol, minVocabSymbol,  maxVocabSymbol) {
  	return symbol >= this.start && symbol <= this.stop;
  };

  RangeTransition.prototype.toString = function() {
  	return "'" + String.fromCharCode(this.start) + "'..'" + String.fromCharCode(this.stop) + "'";
  };

  function AbstractPredicateTransition(target) {
  	Transition.call(this, target);
  	return this;
  }

  AbstractPredicateTransition.prototype = Object.create(Transition.prototype);
  AbstractPredicateTransition.prototype.constructor = AbstractPredicateTransition;

  function PredicateTransition(target, ruleIndex, predIndex, isCtxDependent) {
  	AbstractPredicateTransition.call(this, target);
      this.serializationType = Transition.PREDICATE;
      this.ruleIndex = ruleIndex;
      this.predIndex = predIndex;
      this.isCtxDependent = isCtxDependent; // e.g., $i ref in pred
      this.isEpsilon = true;
      return this;
  }

  PredicateTransition.prototype = Object.create(AbstractPredicateTransition.prototype);
  PredicateTransition.prototype.constructor = PredicateTransition;

  PredicateTransition.prototype.matches = function(symbol, minVocabSymbol,  maxVocabSymbol) {
  	return false;
  };

  PredicateTransition.prototype.getPredicate = function() {
  	return new Predicate(this.ruleIndex, this.predIndex, this.isCtxDependent);
  };

  PredicateTransition.prototype.toString = function() {
  	return "pred_" + this.ruleIndex + ":" + this.predIndex;
  };

  function ActionTransition(target, ruleIndex, actionIndex, isCtxDependent) {
  	Transition.call(this, target);
      this.serializationType = Transition.ACTION;
      this.ruleIndex = ruleIndex;
      this.actionIndex = actionIndex===undefined ? -1 : actionIndex;
      this.isCtxDependent = isCtxDependent===undefined ? false : isCtxDependent; // e.g., $i ref in pred
      this.isEpsilon = true;
      return this;
  }

  ActionTransition.prototype = Object.create(Transition.prototype);
  ActionTransition.prototype.constructor = ActionTransition;


  ActionTransition.prototype.matches = function(symbol, minVocabSymbol,  maxVocabSymbol) {
  	return false;
  };

  ActionTransition.prototype.toString = function() {
  	return "action_" + this.ruleIndex + ":" + this.actionIndex;
  };


  // A transition containing a set of values.
  function SetTransition(target, set) {
  	Transition.call(this, target);
  	this.serializationType = Transition.SET;
      if (set !==undefined && set !==null) {
          this.label = set;
      } else {
          this.label = new IntervalSet();
          this.label.addOne(Token.INVALID_TYPE);
      }
      return this;
  }

  SetTransition.prototype = Object.create(Transition.prototype);
  SetTransition.prototype.constructor = SetTransition;

  SetTransition.prototype.matches = function(symbol, minVocabSymbol,  maxVocabSymbol) {
  	return this.label.contains(symbol);
  };


  SetTransition.prototype.toString = function() {
  	return this.label.toString();
  };

  function NotSetTransition(target, set) {
  	SetTransition.call(this, target, set);
  	this.serializationType = Transition.NOT_SET;
  	return this;
  }

  NotSetTransition.prototype = Object.create(SetTransition.prototype);
  NotSetTransition.prototype.constructor = NotSetTransition;

  NotSetTransition.prototype.matches = function(symbol, minVocabSymbol,  maxVocabSymbol) {
  	return symbol >= minVocabSymbol && symbol <= maxVocabSymbol &&
  			!SetTransition.prototype.matches.call(this, symbol, minVocabSymbol, maxVocabSymbol);
  };

  NotSetTransition.prototype.toString = function() {
  	return '~' + SetTransition.prototype.toString.call(this);
  };

  function WildcardTransition(target) {
  	Transition.call(this, target);
  	this.serializationType = Transition.WILDCARD;
  	return this;
  }

  WildcardTransition.prototype = Object.create(Transition.prototype);
  WildcardTransition.prototype.constructor = WildcardTransition;


  WildcardTransition.prototype.matches = function(symbol, minVocabSymbol,  maxVocabSymbol) {
  	return symbol >= minVocabSymbol && symbol <= maxVocabSymbol;
  };

  WildcardTransition.prototype.toString = function() {
  	return ".";
  };

  function PrecedencePredicateTransition(target, precedence) {
  	AbstractPredicateTransition.call(this, target);
      this.serializationType = Transition.PRECEDENCE;
      this.precedence = precedence;
      this.isEpsilon = true;
      return this;
  }

  PrecedencePredicateTransition.prototype = Object.create(AbstractPredicateTransition.prototype);
  PrecedencePredicateTransition.prototype.constructor = PrecedencePredicateTransition;

  PrecedencePredicateTransition.prototype.matches = function(symbol, minVocabSymbol,  maxVocabSymbol) {
  	return false;
  };

  PrecedencePredicateTransition.prototype.getPredicate = function() {
  	return new PrecedencePredicate(this.precedence);
  };

  PrecedencePredicateTransition.prototype.toString = function() {
  	return this.precedence + " >= _p";
  };

  exports.Transition = Transition;
  exports.AtomTransition = AtomTransition;
  exports.SetTransition = SetTransition;
  exports.NotSetTransition = NotSetTransition;
  exports.RuleTransition = RuleTransition;
  exports.ActionTransition = ActionTransition;
  exports.EpsilonTransition = EpsilonTransition;
  exports.RangeTransition = RangeTransition;
  exports.WildcardTransition = WildcardTransition;
  exports.PredicateTransition = PredicateTransition;
  exports.PrecedencePredicateTransition = PrecedencePredicateTransition;
  exports.AbstractPredicateTransition = AbstractPredicateTransition;
  },{"./../IntervalSet":6,"./../Token":14,"./SemanticContext":29}],31:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */

  exports.ATN = require('./ATN').ATN;
  exports.ATNDeserializer = require('./ATNDeserializer').ATNDeserializer;
  exports.LexerATNSimulator = require('./LexerATNSimulator').LexerATNSimulator;
  exports.ParserATNSimulator = require('./ParserATNSimulator').ParserATNSimulator;
  exports.PredictionMode = require('./PredictionMode').PredictionMode;

  },{"./ATN":16,"./ATNDeserializer":20,"./LexerATNSimulator":24,"./ParserATNSimulator":27,"./PredictionMode":28}],32:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */

  var Set = require("../Utils").Set;
  var DFAState = require('./DFAState').DFAState;
  var StarLoopEntryState = require('../atn/ATNState').StarLoopEntryState;
  var ATNConfigSet = require('./../atn/ATNConfigSet').ATNConfigSet;
  var DFASerializer = require('./DFASerializer').DFASerializer;
  var LexerDFASerializer = require('./DFASerializer').LexerDFASerializer;



  function DFA(atnStartState, decision) {
  	if (decision === undefined) {
  		decision = 0;
  	}
  	// From which ATN state did we create this DFA?
  	this.atnStartState = atnStartState;
  	this.decision = decision;
  	// A set of all DFA states. Use {@link Map} so we can get old state back
  	// ({@link Set} only allows you to see if it's there).
  	this._states = new Set();
  	this.s0 = null;
  	// {@code true} if this DFA is for a precedence decision; otherwise,
  	// {@code false}. This is the backing field for {@link //isPrecedenceDfa},
  	// {@link //setPrecedenceDfa}.
  	this.precedenceDfa = false;
      if (atnStartState instanceof StarLoopEntryState)
      {
          if (atnStartState.isPrecedenceDecision) {
              this.precedenceDfa = true;
              var precedenceState = new DFAState(null, new ATNConfigSet());
              precedenceState.edges = [];
              precedenceState.isAcceptState = false;
              precedenceState.requiresFullContext = false;
              this.s0 = precedenceState;
          }
      }
  	return this;
  }

  // Get the start state for a specific precedence value.
  //
  // @param precedence The current precedence.
  // @return The start state corresponding to the specified precedence, or
  // {@code null} if no start state exists for the specified precedence.
  //
  // @throws IllegalStateException if this is not a precedence DFA.
  // @see //isPrecedenceDfa()

  DFA.prototype.getPrecedenceStartState = function(precedence) {
  	if (!(this.precedenceDfa)) {
  		throw ("Only precedence DFAs may contain a precedence start state.");
  	}
  	// s0.edges is never null for a precedence DFA
  	if (precedence < 0 || precedence >= this.s0.edges.length) {
  		return null;
  	}
  	return this.s0.edges[precedence] || null;
  };

  // Set the start state for a specific precedence value.
  //
  // @param precedence The current precedence.
  // @param startState The start state corresponding to the specified
  // precedence.
  //
  // @throws IllegalStateException if this is not a precedence DFA.
  // @see //isPrecedenceDfa()
  //
  DFA.prototype.setPrecedenceStartState = function(precedence, startState) {
  	if (!(this.precedenceDfa)) {
  		throw ("Only precedence DFAs may contain a precedence start state.");
  	}
  	if (precedence < 0) {
  		return;
  	}

  	// synchronization on s0 here is ok. when the DFA is turned into a
  	// precedence DFA, s0 will be initialized once and not updated again
  	// s0.edges is never null for a precedence DFA
  	this.s0.edges[precedence] = startState;
  };

  //
  // Sets whether this is a precedence DFA. If the specified value differs
  // from the current DFA configuration, the following actions are taken;
  // otherwise no changes are made to the current DFA.
  //
  // <ul>
  // <li>The {@link //states} map is cleared</li>
  // <li>If {@code precedenceDfa} is {@code false}, the initial state
  // {@link //s0} is set to {@code null}; otherwise, it is initialized to a new
  // {@link DFAState} with an empty outgoing {@link DFAState//edges} array to
  // store the start states for individual precedence values.</li>
  // <li>The {@link //precedenceDfa} field is updated</li>
  // </ul>
  //
  // @param precedenceDfa {@code true} if this is a precedence DFA; otherwise,
  // {@code false}

  DFA.prototype.setPrecedenceDfa = function(precedenceDfa) {
  	if (this.precedenceDfa!==precedenceDfa) {
  		this._states = new DFAStatesSet();
  		if (precedenceDfa) {
  			var precedenceState = new DFAState(null, new ATNConfigSet());
  			precedenceState.edges = [];
  			precedenceState.isAcceptState = false;
  			precedenceState.requiresFullContext = false;
  			this.s0 = precedenceState;
  		} else {
  			this.s0 = null;
  		}
  		this.precedenceDfa = precedenceDfa;
  	}
  };

  Object.defineProperty(DFA.prototype, "states", {
  	get : function() {
  		return this._states;
  	}
  });

  // Return a list of all states in this DFA, ordered by state number.
  DFA.prototype.sortedStates = function() {
  	var list = this._states.values();
  	return list.sort(function(a, b) {
  		return a.stateNumber - b.stateNumber;
  	});
  };

  DFA.prototype.toString = function(literalNames, symbolicNames) {
  	literalNames = literalNames || null;
  	symbolicNames = symbolicNames || null;
  	if (this.s0 === null) {
  		return "";
  	}
  	var serializer = new DFASerializer(this, literalNames, symbolicNames);
  	return serializer.toString();
  };

  DFA.prototype.toLexerString = function() {
  	if (this.s0 === null) {
  		return "";
  	}
  	var serializer = new LexerDFASerializer(this);
  	return serializer.toString();
  };

  exports.DFA = DFA;

  },{"../Utils":15,"../atn/ATNState":22,"./../atn/ATNConfigSet":18,"./DFASerializer":33,"./DFAState":34}],33:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */

  // A DFA walker that knows how to dump them to serialized strings.#/


  function DFASerializer(dfa, literalNames, symbolicNames) {
  	this.dfa = dfa;
  	this.literalNames = literalNames || [];
  	this.symbolicNames = symbolicNames || [];
  	return this;
  }

  DFASerializer.prototype.toString = function() {
     if(this.dfa.s0 === null) {
         return null;
     }
     var buf = "";
     var states = this.dfa.sortedStates();
     for(var i=0;i<states.length;i++) {
         var s = states[i];
         if(s.edges!==null) {
              var n = s.edges.length;
              for(var j=0;j<n;j++) {
                  var t = s.edges[j] || null;
                  if(t!==null && t.stateNumber !== 0x7FFFFFFF) {
                      buf = buf.concat(this.getStateString(s));
                      buf = buf.concat("-");
                      buf = buf.concat(this.getEdgeLabel(j));
                      buf = buf.concat("->");
                      buf = buf.concat(this.getStateString(t));
                      buf = buf.concat('\n');
                  }
              }
         }
     }
     return buf.length===0 ? null : buf;
  };

  DFASerializer.prototype.getEdgeLabel = function(i) {
      if (i===0) {
          return "EOF";
      } else if(this.literalNames !==null || this.symbolicNames!==null) {
          return this.literalNames[i-1] || this.symbolicNames[i-1];
      } else {
          return String.fromCharCode(i-1);
      }
  };

  DFASerializer.prototype.getStateString = function(s) {
      var baseStateStr = ( s.isAcceptState ? ":" : "") + "s" + s.stateNumber + ( s.requiresFullContext ? "^" : "");
      if(s.isAcceptState) {
          if (s.predicates !== null) {
              return baseStateStr + "=>" + s.predicates.toString();
          } else {
              return baseStateStr + "=>" + s.prediction.toString();
          }
      } else {
          return baseStateStr;
      }
  };

  function LexerDFASerializer(dfa) {
  	DFASerializer.call(this, dfa, null);
  	return this;
  }

  LexerDFASerializer.prototype = Object.create(DFASerializer.prototype);
  LexerDFASerializer.prototype.constructor = LexerDFASerializer;

  LexerDFASerializer.prototype.getEdgeLabel = function(i) {
  	return "'" + String.fromCharCode(i) + "'";
  };

  exports.DFASerializer = DFASerializer;
  exports.LexerDFASerializer = LexerDFASerializer;


  },{}],34:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  ///

  var ATNConfigSet = require('./../atn/ATNConfigSet').ATNConfigSet;
  var Utils = require('./../Utils');
  var Hash = Utils.Hash;
  var Set = Utils.Set;

  // Map a predicate to a predicted alternative.///

  function PredPrediction(pred, alt) {
  	this.alt = alt;
  	this.pred = pred;
  	return this;
  }

  PredPrediction.prototype.toString = function() {
  	return "(" + this.pred + ", " + this.alt + ")";
  };

  // A DFA state represents a set of possible ATN configurations.
  // As Aho, Sethi, Ullman p. 117 says "The DFA uses its state
  // to keep track of all possible states the ATN can be in after
  // reading each input symbol. That is to say, after reading
  // input a1a2..an, the DFA is in a state that represents the
  // subset T of the states of the ATN that are reachable from the
  // ATN's start state along some path labeled a1a2..an."
  // In conventional NFA&rarr;DFA conversion, therefore, the subset T
  // would be a bitset representing the set of states the
  // ATN could be in. We need to track the alt predicted by each
  // state as well, however. More importantly, we need to maintain
  // a stack of states, tracking the closure operations as they
  // jump from rule to rule, emulating rule invocations (method calls).
  // I have to add a stack to simulate the proper lookahead sequences for
  // the underlying LL grammar from which the ATN was derived.
  //
  // <p>I use a set of ATNConfig objects not simple states. An ATNConfig
  // is both a state (ala normal conversion) and a RuleContext describing
  // the chain of rules (if any) followed to arrive at that state.</p>
  //
  // <p>A DFA state may have multiple references to a particular state,
  // but with different ATN contexts (with same or different alts)
  // meaning that state was reached via a different set of rule invocations.</p>
  // /

  function DFAState(stateNumber, configs) {
  	if (stateNumber === null) {
  		stateNumber = -1;
  	}
  	if (configs === null) {
  		configs = new ATNConfigSet();
  	}
  	this.stateNumber = stateNumber;
  	this.configs = configs;
  	// {@code edges[symbol]} points to target of symbol. Shift up by 1 so (-1)
  	// {@link Token//EOF} maps to {@code edges[0]}.
  	this.edges = null;
  	this.isAcceptState = false;
  	// if accept state, what ttype do we match or alt do we predict?
  	// This is set to {@link ATN//INVALID_ALT_NUMBER} when {@link
  	// //predicates}{@code !=null} or
  	// {@link //requiresFullContext}.
  	this.prediction = 0;
  	this.lexerActionExecutor = null;
  	// Indicates that this state was created during SLL prediction that
  	// discovered a conflict between the configurations in the state. Future
  	// {@link ParserATNSimulator//execATN} invocations immediately jumped doing
  	// full context prediction if this field is true.
  	this.requiresFullContext = false;
  	// During SLL parsing, this is a list of predicates associated with the
  	// ATN configurations of the DFA state. When we have predicates,
  	// {@link //requiresFullContext} is {@code false} since full context
  	// prediction evaluates predicates
  	// on-the-fly. If this is not null, then {@link //prediction} is
  	// {@link ATN//INVALID_ALT_NUMBER}.
  	//
  	// <p>We only use these for non-{@link //requiresFullContext} but
  	// conflicting states. That
  	// means we know from the context (it's $ or we don't dip into outer
  	// context) that it's an ambiguity not a conflict.</p>
  	//
  	// <p>This list is computed by {@link
  	// ParserATNSimulator//predicateDFAState}.</p>
  	this.predicates = null;
  	return this;
  }

  // Get the set of all alts mentioned by all ATN configurations in this
  // DFA state.
  DFAState.prototype.getAltSet = function() {
  	var alts = new Set();
  	if (this.configs !== null) {
  		for (var i = 0; i < this.configs.length; i++) {
  			var c = this.configs[i];
  			alts.add(c.alt);
  		}
  	}
  	if (alts.length === 0) {
  		return null;
  	} else {
  		return alts;
  	}
  };

  // Two {@link DFAState} instances are equal if their ATN configuration sets
  // are the same. This method is used to see if a state already exists.
  //
  // <p>Because the number of alternatives and number of ATN configurations are
  // finite, there is a finite number of DFA states that can be processed.
  // This is necessary to show that the algorithm terminates.</p>
  //
  // <p>Cannot test the DFA state numbers here because in
  // {@link ParserATNSimulator//addDFAState} we need to know if any other state
  // exists that has this exact set of ATN configurations. The
  // {@link //stateNumber} is irrelevant.</p>
  DFAState.prototype.equals = function(other) {
  	// compare set of ATN configurations in this set with other
  	return this === other ||
  			(other instanceof DFAState &&
  				this.configs.equals(other.configs));
  };

  DFAState.prototype.toString = function() {
  	var s = "" + this.stateNumber + ":" + this.configs;
  	if(this.isAcceptState) {
          s = s + "=>";
          if (this.predicates !== null)
              s = s + this.predicates;
          else
              s = s + this.prediction;
      }
  	return s;
  };

  DFAState.prototype.hashCode = function() {
  	var hash = new Hash();
  	hash.update(this.configs);
  	if(this.isAcceptState) {
          if (this.predicates !== null)
              hash.update(this.predicates);
          else
              hash.update(this.prediction);
      }
      return hash.finish();
  };

  exports.DFAState = DFAState;
  exports.PredPrediction = PredPrediction;

  },{"./../Utils":15,"./../atn/ATNConfigSet":18}],35:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */

  exports.DFA = require('./DFA').DFA;
  exports.DFASerializer = require('./DFASerializer').DFASerializer;
  exports.LexerDFASerializer = require('./DFASerializer').LexerDFASerializer;
  exports.PredPrediction = require('./DFAState').PredPrediction;

  },{"./DFA":32,"./DFASerializer":33,"./DFAState":34}],36:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  //

  //
  // This implementation of {@link ANTLRErrorListener} can be used to identify
  // certain potential correctness and performance problems in grammars. "Reports"
  // are made by calling {@link Parser//notifyErrorListeners} with the appropriate
  // message.
  //
  // <ul>
  // <li><b>Ambiguities</b>: These are cases where more than one path through the
  // grammar can match the input.</li>
  // <li><b>Weak context sensitivity</b>: These are cases where full-context
  // prediction resolved an SLL conflict to a unique alternative which equaled the
  // minimum alternative of the SLL conflict.</li>
  // <li><b>Strong (forced) context sensitivity</b>: These are cases where the
  // full-context prediction resolved an SLL conflict to a unique alternative,
  // <em>and</em> the minimum alternative of the SLL conflict was found to not be
  // a truly viable alternative. Two-stage parsing cannot be used for inputs where
  // this situation occurs.</li>
  // </ul>

  var BitSet = require('./../Utils').BitSet;
  var ErrorListener = require('./ErrorListener').ErrorListener;
  var Interval = require('./../IntervalSet').Interval;

  function DiagnosticErrorListener(exactOnly) {
  	ErrorListener.call(this);
  	exactOnly = exactOnly || true;
  	// whether all ambiguities or only exact ambiguities are reported.
  	this.exactOnly = exactOnly;
  	return this;
  }

  DiagnosticErrorListener.prototype = Object.create(ErrorListener.prototype);
  DiagnosticErrorListener.prototype.constructor = DiagnosticErrorListener;

  DiagnosticErrorListener.prototype.reportAmbiguity = function(recognizer, dfa,
  		startIndex, stopIndex, exact, ambigAlts, configs) {
  	if (this.exactOnly && !exact) {
  		return;
  	}
  	var msg = "reportAmbiguity d=" +
  			this.getDecisionDescription(recognizer, dfa) +
  			": ambigAlts=" +
  			this.getConflictingAlts(ambigAlts, configs) +
  			", input='" +
  			recognizer.getTokenStream().getText(new Interval(startIndex, stopIndex)) + "'";
  	recognizer.notifyErrorListeners(msg);
  };

  DiagnosticErrorListener.prototype.reportAttemptingFullContext = function(
  		recognizer, dfa, startIndex, stopIndex, conflictingAlts, configs) {
  	var msg = "reportAttemptingFullContext d=" +
  			this.getDecisionDescription(recognizer, dfa) +
  			", input='" +
  			recognizer.getTokenStream().getText(new Interval(startIndex, stopIndex)) + "'";
  	recognizer.notifyErrorListeners(msg);
  };

  DiagnosticErrorListener.prototype.reportContextSensitivity = function(
  		recognizer, dfa, startIndex, stopIndex, prediction, configs) {
  	var msg = "reportContextSensitivity d=" +
  			this.getDecisionDescription(recognizer, dfa) +
  			", input='" +
  			recognizer.getTokenStream().getText(new Interval(startIndex, stopIndex)) + "'";
  	recognizer.notifyErrorListeners(msg);
  };

  DiagnosticErrorListener.prototype.getDecisionDescription = function(recognizer, dfa) {
  	var decision = dfa.decision;
  	var ruleIndex = dfa.atnStartState.ruleIndex;

  	var ruleNames = recognizer.ruleNames;
  	if (ruleIndex < 0 || ruleIndex >= ruleNames.length) {
  		return "" + decision;
  	}
  	var ruleName = ruleNames[ruleIndex] || null;
  	if (ruleName === null || ruleName.length === 0) {
  		return "" + decision;
  	}
  	return "" + decision + " (" + ruleName + ")";
  };

  //
  // Computes the set of conflicting or ambiguous alternatives from a
  // configuration set, if that information was not already provided by the
  // parser.
  //
  // @param reportedAlts The set of conflicting or ambiguous alternatives, as
  // reported by the parser.
  // @param configs The conflicting or ambiguous configuration set.
  // @return Returns {@code reportedAlts} if it is not {@code null}, otherwise
  // returns the set of alternatives represented in {@code configs}.
  //
  DiagnosticErrorListener.prototype.getConflictingAlts = function(reportedAlts, configs) {
  	if (reportedAlts !== null) {
  		return reportedAlts;
  	}
  	var result = new BitSet();
  	for (var i = 0; i < configs.items.length; i++) {
  		result.add(configs.items[i].alt);
  	}
  	return "{" + result.values().join(", ") + "}";
  };

  exports.DiagnosticErrorListener = DiagnosticErrorListener;
  },{"./../IntervalSet":6,"./../Utils":15,"./ErrorListener":37}],37:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */

  // Provides an empty default implementation of {@link ANTLRErrorListener}. The
  // default implementation of each method does nothing, but can be overridden as
  // necessary.

  function ErrorListener() {
  	return this;
  }

  ErrorListener.prototype.syntaxError = function(recognizer, offendingSymbol, line, column, msg, e) {
  };

  ErrorListener.prototype.reportAmbiguity = function(recognizer, dfa, startIndex, stopIndex, exact, ambigAlts, configs) {
  };

  ErrorListener.prototype.reportAttemptingFullContext = function(recognizer, dfa, startIndex, stopIndex, conflictingAlts, configs) {
  };

  ErrorListener.prototype.reportContextSensitivity = function(recognizer, dfa, startIndex, stopIndex, prediction, configs) {
  };

  function ConsoleErrorListener() {
  	ErrorListener.call(this);
  	return this;
  }

  ConsoleErrorListener.prototype = Object.create(ErrorListener.prototype);
  ConsoleErrorListener.prototype.constructor = ConsoleErrorListener;

  //
  // Provides a default instance of {@link ConsoleErrorListener}.
  //
  ConsoleErrorListener.INSTANCE = new ConsoleErrorListener();

  //
  // {@inheritDoc}
  //
  // <p>
  // This implementation prints messages to {@link System//err} containing the
  // values of {@code line}, {@code charPositionInLine}, and {@code msg} using
  // the following format.</p>
  //
  // <pre>
  // line <em>line</em>:<em>charPositionInLine</em> <em>msg</em>
  // </pre>
  //
  ConsoleErrorListener.prototype.syntaxError = function(recognizer, offendingSymbol, line, column, msg, e) {
      console.error("line " + line + ":" + column + " " + msg);
  };

  function ProxyErrorListener(delegates) {
  	ErrorListener.call(this);
      if (delegates===null) {
          throw "delegates";
      }
      this.delegates = delegates;
  	return this;
  }

  ProxyErrorListener.prototype = Object.create(ErrorListener.prototype);
  ProxyErrorListener.prototype.constructor = ProxyErrorListener;

  ProxyErrorListener.prototype.syntaxError = function(recognizer, offendingSymbol, line, column, msg, e) {
      this.delegates.map(function(d) { d.syntaxError(recognizer, offendingSymbol, line, column, msg, e); });
  };

  ProxyErrorListener.prototype.reportAmbiguity = function(recognizer, dfa, startIndex, stopIndex, exact, ambigAlts, configs) {
      this.delegates.map(function(d) { d.reportAmbiguity(recognizer, dfa, startIndex, stopIndex, exact, ambigAlts, configs); });
  };

  ProxyErrorListener.prototype.reportAttemptingFullContext = function(recognizer, dfa, startIndex, stopIndex, conflictingAlts, configs) {
  	this.delegates.map(function(d) { d.reportAttemptingFullContext(recognizer, dfa, startIndex, stopIndex, conflictingAlts, configs); });
  };

  ProxyErrorListener.prototype.reportContextSensitivity = function(recognizer, dfa, startIndex, stopIndex, prediction, configs) {
  	this.delegates.map(function(d) { d.reportContextSensitivity(recognizer, dfa, startIndex, stopIndex, prediction, configs); });
  };

  exports.ErrorListener = ErrorListener;
  exports.ConsoleErrorListener = ConsoleErrorListener;
  exports.ProxyErrorListener = ProxyErrorListener;


  },{}],38:[function(require,module,exports){
  //
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  //

  var Token = require('./../Token').Token;
  var Errors = require('./Errors');
  var NoViableAltException = Errors.NoViableAltException;
  var InputMismatchException = Errors.InputMismatchException;
  var FailedPredicateException = Errors.FailedPredicateException;
  var ParseCancellationException = Errors.ParseCancellationException;
  var ATNState = require('./../atn/ATNState').ATNState;
  var Interval = require('./../IntervalSet').Interval;
  var IntervalSet = require('./../IntervalSet').IntervalSet;

  function ErrorStrategy() {

  }

  ErrorStrategy.prototype.reset = function(recognizer){
  };

  ErrorStrategy.prototype.recoverInline = function(recognizer){
  };

  ErrorStrategy.prototype.recover = function(recognizer, e){
  };

  ErrorStrategy.prototype.sync = function(recognizer){
  };

  ErrorStrategy.prototype.inErrorRecoveryMode = function(recognizer){
  };

  ErrorStrategy.prototype.reportError = function(recognizer){
  };



  // This is the default implementation of {@link ANTLRErrorStrategy} used for
  // error reporting and recovery in ANTLR parsers.
  //
  function DefaultErrorStrategy() {
  	ErrorStrategy.call(this);
      // Indicates whether the error strategy is currently "recovering from an
      // error". This is used to suppress reporting multiple error messages while
      // attempting to recover from a detected syntax error.
      //
      // @see //inErrorRecoveryMode
      //
      this.errorRecoveryMode = false;

      // The index into the input stream where the last error occurred.
      // This is used to prevent infinite loops where an error is found
      // but no token is consumed during recovery...another error is found,
      // ad nauseum. This is a failsafe mechanism to guarantee that at least
      // one token/tree node is consumed for two errors.
      //
      this.lastErrorIndex = -1;
      this.lastErrorStates = null;
      return this;
  }

  DefaultErrorStrategy.prototype = Object.create(ErrorStrategy.prototype);
  DefaultErrorStrategy.prototype.constructor = DefaultErrorStrategy;

  // <p>The default implementation simply calls {@link //endErrorCondition} to
  // ensure that the handler is not in error recovery mode.</p>
  DefaultErrorStrategy.prototype.reset = function(recognizer) {
      this.endErrorCondition(recognizer);
  };

  //
  // This method is called to enter error recovery mode when a recognition
  // exception is reported.
  //
  // @param recognizer the parser instance
  //
  DefaultErrorStrategy.prototype.beginErrorCondition = function(recognizer) {
      this.errorRecoveryMode = true;
  };

  DefaultErrorStrategy.prototype.inErrorRecoveryMode = function(recognizer) {
      return this.errorRecoveryMode;
  };

  //
  // This method is called to leave error recovery mode after recovering from
  // a recognition exception.
  //
  // @param recognizer
  //
  DefaultErrorStrategy.prototype.endErrorCondition = function(recognizer) {
      this.errorRecoveryMode = false;
      this.lastErrorStates = null;
      this.lastErrorIndex = -1;
  };

  //
  // {@inheritDoc}
  //
  // <p>The default implementation simply calls {@link //endErrorCondition}.</p>
  //
  DefaultErrorStrategy.prototype.reportMatch = function(recognizer) {
      this.endErrorCondition(recognizer);
  };

  //
  // {@inheritDoc}
  //
  // <p>The default implementation returns immediately if the handler is already
  // in error recovery mode. Otherwise, it calls {@link //beginErrorCondition}
  // and dispatches the reporting task based on the runtime type of {@code e}
  // according to the following table.</p>
  //
  // <ul>
  // <li>{@link NoViableAltException}: Dispatches the call to
  // {@link //reportNoViableAlternative}</li>
  // <li>{@link InputMismatchException}: Dispatches the call to
  // {@link //reportInputMismatch}</li>
  // <li>{@link FailedPredicateException}: Dispatches the call to
  // {@link //reportFailedPredicate}</li>
  // <li>All other types: calls {@link Parser//notifyErrorListeners} to report
  // the exception</li>
  // </ul>
  //
  DefaultErrorStrategy.prototype.reportError = function(recognizer, e) {
     // if we've already reported an error and have not matched a token
     // yet successfully, don't report any errors.
      if(this.inErrorRecoveryMode(recognizer)) {
          return; // don't report spurious errors
      }
      this.beginErrorCondition(recognizer);
      if ( e instanceof NoViableAltException ) {
          this.reportNoViableAlternative(recognizer, e);
      } else if ( e instanceof InputMismatchException ) {
          this.reportInputMismatch(recognizer, e);
      } else if ( e instanceof FailedPredicateException ) {
          this.reportFailedPredicate(recognizer, e);
      } else {
          console.log("unknown recognition error type: " + e.constructor.name);
          console.log(e.stack);
          recognizer.notifyErrorListeners(e.getOffendingToken(), e.getMessage(), e);
      }
  };
  //
  // {@inheritDoc}
  //
  // <p>The default implementation resynchronizes the parser by consuming tokens
  // until we find one in the resynchronization set--loosely the set of tokens
  // that can follow the current rule.</p>
  //
  DefaultErrorStrategy.prototype.recover = function(recognizer, e) {
      if (this.lastErrorIndex===recognizer.getInputStream().index &&
          this.lastErrorStates !== null && this.lastErrorStates.indexOf(recognizer.state)>=0) {
  		// uh oh, another error at same token index and previously-visited
  		// state in ATN; must be a case where LT(1) is in the recovery
  		// token set so nothing got consumed. Consume a single token
  		// at least to prevent an infinite loop; this is a failsafe.
  		recognizer.consume();
      }
      this.lastErrorIndex = recognizer._input.index;
      if (this.lastErrorStates === null) {
          this.lastErrorStates = [];
      }
      this.lastErrorStates.push(recognizer.state);
      var followSet = this.getErrorRecoverySet(recognizer);
      this.consumeUntil(recognizer, followSet);
  };

  // The default implementation of {@link ANTLRErrorStrategy//sync} makes sure
  // that the current lookahead symbol is consistent with what were expecting
  // at this point in the ATN. You can call this anytime but ANTLR only
  // generates code to check before subrules/loops and each iteration.
  //
  // <p>Implements Jim Idle's magic sync mechanism in closures and optional
  // subrules. E.g.,</p>
  //
  // <pre>
  // a : sync ( stuff sync )* ;
  // sync : {consume to what can follow sync} ;
  // </pre>
  //
  // At the start of a sub rule upon error, {@link //sync} performs single
  // token deletion, if possible. If it can't do that, it bails on the current
  // rule and uses the default error recovery, which consumes until the
  // resynchronization set of the current rule.
  //
  // <p>If the sub rule is optional ({@code (...)?}, {@code (...)*}, or block
  // with an empty alternative), then the expected set includes what follows
  // the subrule.</p>
  //
  // <p>During loop iteration, it consumes until it sees a token that can start a
  // sub rule or what follows loop. Yes, that is pretty aggressive. We opt to
  // stay in the loop as long as possible.</p>
  //
  // <p><strong>ORIGINS</strong></p>
  //
  // <p>Previous versions of ANTLR did a poor job of their recovery within loops.
  // A single mismatch token or missing token would force the parser to bail
  // out of the entire rules surrounding the loop. So, for rule</p>
  //
  // <pre>
  // classDef : 'class' ID '{' member* '}'
  // </pre>
  //
  // input with an extra token between members would force the parser to
  // consume until it found the next class definition rather than the next
  // member definition of the current class.
  //
  // <p>This functionality cost a little bit of effort because the parser has to
  // compare token set at the start of the loop and at each iteration. If for
  // some reason speed is suffering for you, you can turn off this
  // functionality by simply overriding this method as a blank { }.</p>
  //
  DefaultErrorStrategy.prototype.sync = function(recognizer) {
      // If already recovering, don't try to sync
      if (this.inErrorRecoveryMode(recognizer)) {
          return;
      }
      var s = recognizer._interp.atn.states[recognizer.state];
      var la = recognizer.getTokenStream().LA(1);
      // try cheaper subset first; might get lucky. seems to shave a wee bit off
      var nextTokens = recognizer.atn.nextTokens(s);
      if (nextTokens.contains(Token.EPSILON) || nextTokens.contains(la)) {
          return;
      }
      switch (s.stateType) {
      case ATNState.BLOCK_START:
      case ATNState.STAR_BLOCK_START:
      case ATNState.PLUS_BLOCK_START:
      case ATNState.STAR_LOOP_ENTRY:
         // report error and recover if possible
          if( this.singleTokenDeletion(recognizer) !== null) {
              return;
          } else {
              throw new InputMismatchException(recognizer);
          }
          break;
      case ATNState.PLUS_LOOP_BACK:
      case ATNState.STAR_LOOP_BACK:
          this.reportUnwantedToken(recognizer);
          var expecting = new IntervalSet();
          expecting.addSet(recognizer.getExpectedTokens());
          var whatFollowsLoopIterationOrRule = expecting.addSet(this.getErrorRecoverySet(recognizer));
          this.consumeUntil(recognizer, whatFollowsLoopIterationOrRule);
          break;
      default:
          // do nothing if we can't identify the exact kind of ATN state
      }
  };

  // This is called by {@link //reportError} when the exception is a
  // {@link NoViableAltException}.
  //
  // @see //reportError
  //
  // @param recognizer the parser instance
  // @param e the recognition exception
  //
  DefaultErrorStrategy.prototype.reportNoViableAlternative = function(recognizer, e) {
      var tokens = recognizer.getTokenStream();
      var input;
      if(tokens !== null) {
          if (e.startToken.type===Token.EOF) {
              input = "<EOF>";
          } else {
              input = tokens.getText(new Interval(e.startToken, e.offendingToken));
          }
      } else {
          input = "<unknown input>";
      }
      var msg = "no viable alternative at input " + this.escapeWSAndQuote(input);
      recognizer.notifyErrorListeners(msg, e.offendingToken, e);
  };

  //
  // This is called by {@link //reportError} when the exception is an
  // {@link InputMismatchException}.
  //
  // @see //reportError
  //
  // @param recognizer the parser instance
  // @param e the recognition exception
  //
  DefaultErrorStrategy.prototype.reportInputMismatch = function(recognizer, e) {
      var msg = "mismatched input " + this.getTokenErrorDisplay(e.offendingToken) +
            " expecting " + e.getExpectedTokens().toString(recognizer.literalNames, recognizer.symbolicNames);
      recognizer.notifyErrorListeners(msg, e.offendingToken, e);
  };

  //
  // This is called by {@link //reportError} when the exception is a
  // {@link FailedPredicateException}.
  //
  // @see //reportError
  //
  // @param recognizer the parser instance
  // @param e the recognition exception
  //
  DefaultErrorStrategy.prototype.reportFailedPredicate = function(recognizer, e) {
      var ruleName = recognizer.ruleNames[recognizer._ctx.ruleIndex];
      var msg = "rule " + ruleName + " " + e.message;
      recognizer.notifyErrorListeners(msg, e.offendingToken, e);
  };

  // This method is called to report a syntax error which requires the removal
  // of a token from the input stream. At the time this method is called, the
  // erroneous symbol is current {@code LT(1)} symbol and has not yet been
  // removed from the input stream. When this method returns,
  // {@code recognizer} is in error recovery mode.
  //
  // <p>This method is called when {@link //singleTokenDeletion} identifies
  // single-token deletion as a viable recovery strategy for a mismatched
  // input error.</p>
  //
  // <p>The default implementation simply returns if the handler is already in
  // error recovery mode. Otherwise, it calls {@link //beginErrorCondition} to
  // enter error recovery mode, followed by calling
  // {@link Parser//notifyErrorListeners}.</p>
  //
  // @param recognizer the parser instance
  //
  DefaultErrorStrategy.prototype.reportUnwantedToken = function(recognizer) {
      if (this.inErrorRecoveryMode(recognizer)) {
          return;
      }
      this.beginErrorCondition(recognizer);
      var t = recognizer.getCurrentToken();
      var tokenName = this.getTokenErrorDisplay(t);
      var expecting = this.getExpectedTokens(recognizer);
      var msg = "extraneous input " + tokenName + " expecting " +
          expecting.toString(recognizer.literalNames, recognizer.symbolicNames);
      recognizer.notifyErrorListeners(msg, t, null);
  };
  // This method is called to report a syntax error which requires the
  // insertion of a missing token into the input stream. At the time this
  // method is called, the missing token has not yet been inserted. When this
  // method returns, {@code recognizer} is in error recovery mode.
  //
  // <p>This method is called when {@link //singleTokenInsertion} identifies
  // single-token insertion as a viable recovery strategy for a mismatched
  // input error.</p>
  //
  // <p>The default implementation simply returns if the handler is already in
  // error recovery mode. Otherwise, it calls {@link //beginErrorCondition} to
  // enter error recovery mode, followed by calling
  // {@link Parser//notifyErrorListeners}.</p>
  //
  // @param recognizer the parser instance
  //
  DefaultErrorStrategy.prototype.reportMissingToken = function(recognizer) {
      if ( this.inErrorRecoveryMode(recognizer)) {
          return;
      }
      this.beginErrorCondition(recognizer);
      var t = recognizer.getCurrentToken();
      var expecting = this.getExpectedTokens(recognizer);
      var msg = "missing " + expecting.toString(recognizer.literalNames, recognizer.symbolicNames) +
            " at " + this.getTokenErrorDisplay(t);
      recognizer.notifyErrorListeners(msg, t, null);
  };

  // <p>The default implementation attempts to recover from the mismatched input
  // by using single token insertion and deletion as described below. If the
  // recovery attempt fails, this method throws an
  // {@link InputMismatchException}.</p>
  //
  // <p><strong>EXTRA TOKEN</strong> (single token deletion)</p>
  //
  // <p>{@code LA(1)} is not what we are looking for. If {@code LA(2)} has the
  // right token, however, then assume {@code LA(1)} is some extra spurious
  // token and delete it. Then consume and return the next token (which was
  // the {@code LA(2)} token) as the successful result of the match operation.</p>
  //
  // <p>This recovery strategy is implemented by {@link
  // //singleTokenDeletion}.</p>
  //
  // <p><strong>MISSING TOKEN</strong> (single token insertion)</p>
  //
  // <p>If current token (at {@code LA(1)}) is consistent with what could come
  // after the expected {@code LA(1)} token, then assume the token is missing
  // and use the parser's {@link TokenFactory} to create it on the fly. The
  // "insertion" is performed by returning the created token as the successful
  // result of the match operation.</p>
  //
  // <p>This recovery strategy is implemented by {@link
  // //singleTokenInsertion}.</p>
  //
  // <p><strong>EXAMPLE</strong></p>
  //
  // <p>For example, Input {@code i=(3;} is clearly missing the {@code ')'}. When
  // the parser returns from the nested call to {@code expr}, it will have
  // call chain:</p>
  //
  // <pre>
  // stat &rarr; expr &rarr; atom
  // </pre>
  //
  // and it will be trying to match the {@code ')'} at this point in the
  // derivation:
  //
  // <pre>
  // =&gt; ID '=' '(' INT ')' ('+' atom)* ';'
  // ^
  // </pre>
  //
  // The attempt to match {@code ')'} will fail when it sees {@code ';'} and
  // call {@link //recoverInline}. To recover, it sees that {@code LA(1)==';'}
  // is in the set of tokens that can follow the {@code ')'} token reference
  // in rule {@code atom}. It can assume that you forgot the {@code ')'}.
  //
  DefaultErrorStrategy.prototype.recoverInline = function(recognizer) {
      // SINGLE TOKEN DELETION
      var matchedSymbol = this.singleTokenDeletion(recognizer);
      if (matchedSymbol !== null) {
          // we have deleted the extra token.
          // now, move past ttype token as if all were ok
          recognizer.consume();
          return matchedSymbol;
      }
      // SINGLE TOKEN INSERTION
      if (this.singleTokenInsertion(recognizer)) {
          return this.getMissingSymbol(recognizer);
      }
      // even that didn't work; must throw the exception
      throw new InputMismatchException(recognizer);
  };

  //
  // This method implements the single-token insertion inline error recovery
  // strategy. It is called by {@link //recoverInline} if the single-token
  // deletion strategy fails to recover from the mismatched input. If this
  // method returns {@code true}, {@code recognizer} will be in error recovery
  // mode.
  //
  // <p>This method determines whether or not single-token insertion is viable by
  // checking if the {@code LA(1)} input symbol could be successfully matched
  // if it were instead the {@code LA(2)} symbol. If this method returns
  // {@code true}, the caller is responsible for creating and inserting a
  // token with the correct type to produce this behavior.</p>
  //
  // @param recognizer the parser instance
  // @return {@code true} if single-token insertion is a viable recovery
  // strategy for the current mismatched input, otherwise {@code false}
  //
  DefaultErrorStrategy.prototype.singleTokenInsertion = function(recognizer) {
      var currentSymbolType = recognizer.getTokenStream().LA(1);
      // if current token is consistent with what could come after current
      // ATN state, then we know we're missing a token; error recovery
      // is free to conjure up and insert the missing token
      var atn = recognizer._interp.atn;
      var currentState = atn.states[recognizer.state];
      var next = currentState.transitions[0].target;
      var expectingAtLL2 = atn.nextTokens(next, recognizer._ctx);
      if (expectingAtLL2.contains(currentSymbolType) ){
          this.reportMissingToken(recognizer);
          return true;
      } else {
          return false;
      }
  };

  // This method implements the single-token deletion inline error recovery
  // strategy. It is called by {@link //recoverInline} to attempt to recover
  // from mismatched input. If this method returns null, the parser and error
  // handler state will not have changed. If this method returns non-null,
  // {@code recognizer} will <em>not</em> be in error recovery mode since the
  // returned token was a successful match.
  //
  // <p>If the single-token deletion is successful, this method calls
  // {@link //reportUnwantedToken} to report the error, followed by
  // {@link Parser//consume} to actually "delete" the extraneous token. Then,
  // before returning {@link //reportMatch} is called to signal a successful
  // match.</p>
  //
  // @param recognizer the parser instance
  // @return the successfully matched {@link Token} instance if single-token
  // deletion successfully recovers from the mismatched input, otherwise
  // {@code null}
  //
  DefaultErrorStrategy.prototype.singleTokenDeletion = function(recognizer) {
      var nextTokenType = recognizer.getTokenStream().LA(2);
      var expecting = this.getExpectedTokens(recognizer);
      if (expecting.contains(nextTokenType)) {
          this.reportUnwantedToken(recognizer);
          // print("recoverFromMismatchedToken deleting " \
          // + str(recognizer.getTokenStream().LT(1)) \
          // + " since " + str(recognizer.getTokenStream().LT(2)) \
          // + " is what we want", file=sys.stderr)
          recognizer.consume(); // simply delete extra token
          // we want to return the token we're actually matching
          var matchedSymbol = recognizer.getCurrentToken();
          this.reportMatch(recognizer); // we know current token is correct
          return matchedSymbol;
      } else {
          return null;
      }
  };

  // Conjure up a missing token during error recovery.
  //
  // The recognizer attempts to recover from single missing
  // symbols. But, actions might refer to that missing symbol.
  // For example, x=ID {f($x);}. The action clearly assumes
  // that there has been an identifier matched previously and that
  // $x points at that token. If that token is missing, but
  // the next token in the stream is what we want we assume that
  // this token is missing and we keep going. Because we
  // have to return some token to replace the missing token,
  // we have to conjure one up. This method gives the user control
  // over the tokens returned for missing tokens. Mostly,
  // you will want to create something special for identifier
  // tokens. For literals such as '{' and ',', the default
  // action in the parser or tree parser works. It simply creates
  // a CommonToken of the appropriate type. The text will be the token.
  // If you change what tokens must be created by the lexer,
  // override this method to create the appropriate tokens.
  //
  DefaultErrorStrategy.prototype.getMissingSymbol = function(recognizer) {
      var currentSymbol = recognizer.getCurrentToken();
      var expecting = this.getExpectedTokens(recognizer);
      var expectedTokenType = expecting.first(); // get any element
      var tokenText;
      if (expectedTokenType===Token.EOF) {
          tokenText = "<missing EOF>";
      } else {
          tokenText = "<missing " + recognizer.literalNames[expectedTokenType] + ">";
      }
      var current = currentSymbol;
      var lookback = recognizer.getTokenStream().LT(-1);
      if (current.type===Token.EOF && lookback !== null) {
          current = lookback;
      }
      return recognizer.getTokenFactory().create(current.source,
          expectedTokenType, tokenText, Token.DEFAULT_CHANNEL,
          -1, -1, current.line, current.column);
  };

  DefaultErrorStrategy.prototype.getExpectedTokens = function(recognizer) {
      return recognizer.getExpectedTokens();
  };

  // How should a token be displayed in an error message? The default
  // is to display just the text, but during development you might
  // want to have a lot of information spit out. Override in that case
  // to use t.toString() (which, for CommonToken, dumps everything about
  // the token). This is better than forcing you to override a method in
  // your token objects because you don't have to go modify your lexer
  // so that it creates a new Java type.
  //
  DefaultErrorStrategy.prototype.getTokenErrorDisplay = function(t) {
      if (t === null) {
          return "<no token>";
      }
      var s = t.text;
      if (s === null) {
          if (t.type===Token.EOF) {
              s = "<EOF>";
          } else {
              s = "<" + t.type + ">";
          }
      }
      return this.escapeWSAndQuote(s);
  };

  DefaultErrorStrategy.prototype.escapeWSAndQuote = function(s) {
      s = s.replace(/\n/g,"\\n");
      s = s.replace(/\r/g,"\\r");
      s = s.replace(/\t/g,"\\t");
      return "'" + s + "'";
  };

  // Compute the error recovery set for the current rule. During
  // rule invocation, the parser pushes the set of tokens that can
  // follow that rule reference on the stack; this amounts to
  // computing FIRST of what follows the rule reference in the
  // enclosing rule. See LinearApproximator.FIRST().
  // This local follow set only includes tokens
  // from within the rule; i.e., the FIRST computation done by
  // ANTLR stops at the end of a rule.
  //
  // EXAMPLE
  //
  // When you find a "no viable alt exception", the input is not
  // consistent with any of the alternatives for rule r. The best
  // thing to do is to consume tokens until you see something that
  // can legally follow a call to r//or* any rule that called r.
  // You don't want the exact set of viable next tokens because the
  // input might just be missing a token--you might consume the
  // rest of the input looking for one of the missing tokens.
  //
  // Consider grammar:
  //
  // a : '[' b ']'
  // | '(' b ')'
  // ;
  // b : c '^' INT ;
  // c : ID
  // | INT
  // ;
  //
  // At each rule invocation, the set of tokens that could follow
  // that rule is pushed on a stack. Here are the various
  // context-sensitive follow sets:
  //
  // FOLLOW(b1_in_a) = FIRST(']') = ']'
  // FOLLOW(b2_in_a) = FIRST(')') = ')'
  // FOLLOW(c_in_b) = FIRST('^') = '^'
  //
  // Upon erroneous input "[]", the call chain is
  //
  // a -> b -> c
  //
  // and, hence, the follow context stack is:
  //
  // depth follow set start of rule execution
  // 0 <EOF> a (from main())
  // 1 ']' b
  // 2 '^' c
  //
  // Notice that ')' is not included, because b would have to have
  // been called from a different context in rule a for ')' to be
  // included.
  //
  // For error recovery, we cannot consider FOLLOW(c)
  // (context-sensitive or otherwise). We need the combined set of
  // all context-sensitive FOLLOW sets--the set of all tokens that
  // could follow any reference in the call chain. We need to
  // resync to one of those tokens. Note that FOLLOW(c)='^' and if
  // we resync'd to that token, we'd consume until EOF. We need to
  // sync to context-sensitive FOLLOWs for a, b, and c: {']','^'}.
  // In this case, for input "[]", LA(1) is ']' and in the set, so we would
  // not consume anything. After printing an error, rule c would
  // return normally. Rule b would not find the required '^' though.
  // At this point, it gets a mismatched token error and throws an
  // exception (since LA(1) is not in the viable following token
  // set). The rule exception handler tries to recover, but finds
  // the same recovery set and doesn't consume anything. Rule b
  // exits normally returning to rule a. Now it finds the ']' (and
  // with the successful match exits errorRecovery mode).
  //
  // So, you can see that the parser walks up the call chain looking
  // for the token that was a member of the recovery set.
  //
  // Errors are not generated in errorRecovery mode.
  //
  // ANTLR's error recovery mechanism is based upon original ideas:
  //
  // "Algorithms + Data Structures = Programs" by Niklaus Wirth
  //
  // and
  //
  // "A note on error recovery in recursive descent parsers":
  // http://portal.acm.org/citation.cfm?id=947902.947905
  //
  // Later, Josef Grosch had some good ideas:
  //
  // "Efficient and Comfortable Error Recovery in Recursive Descent
  // Parsers":
  // ftp://www.cocolab.com/products/cocktail/doca4.ps/ell.ps.zip
  //
  // Like Grosch I implement context-sensitive FOLLOW sets that are combined
  // at run-time upon error to avoid overhead during parsing.
  //
  DefaultErrorStrategy.prototype.getErrorRecoverySet = function(recognizer) {
      var atn = recognizer._interp.atn;
      var ctx = recognizer._ctx;
      var recoverSet = new IntervalSet();
      while (ctx !== null && ctx.invokingState>=0) {
          // compute what follows who invoked us
          var invokingState = atn.states[ctx.invokingState];
          var rt = invokingState.transitions[0];
          var follow = atn.nextTokens(rt.followState);
          recoverSet.addSet(follow);
          ctx = ctx.parentCtx;
      }
      recoverSet.removeOne(Token.EPSILON);
      return recoverSet;
  };

  // Consume tokens until one matches the given token set.//
  DefaultErrorStrategy.prototype.consumeUntil = function(recognizer, set) {
      var ttype = recognizer.getTokenStream().LA(1);
      while( ttype !== Token.EOF && !set.contains(ttype)) {
          recognizer.consume();
          ttype = recognizer.getTokenStream().LA(1);
      }
  };

  //
  // This implementation of {@link ANTLRErrorStrategy} responds to syntax errors
  // by immediately canceling the parse operation with a
  // {@link ParseCancellationException}. The implementation ensures that the
  // {@link ParserRuleContext//exception} field is set for all parse tree nodes
  // that were not completed prior to encountering the error.
  //
  // <p>
  // This error strategy is useful in the following scenarios.</p>
  //
  // <ul>
  // <li><strong>Two-stage parsing:</strong> This error strategy allows the first
  // stage of two-stage parsing to immediately terminate if an error is
  // encountered, and immediately fall back to the second stage. In addition to
  // avoiding wasted work by attempting to recover from errors here, the empty
  // implementation of {@link BailErrorStrategy//sync} improves the performance of
  // the first stage.</li>
  // <li><strong>Silent validation:</strong> When syntax errors are not being
  // reported or logged, and the parse result is simply ignored if errors occur,
  // the {@link BailErrorStrategy} avoids wasting work on recovering from errors
  // when the result will be ignored either way.</li>
  // </ul>
  //
  // <p>
  // {@code myparser.setErrorHandler(new BailErrorStrategy());}</p>
  //
  // @see Parser//setErrorHandler(ANTLRErrorStrategy)
  //
  function BailErrorStrategy() {
  	DefaultErrorStrategy.call(this);
  	return this;
  }

  BailErrorStrategy.prototype = Object.create(DefaultErrorStrategy.prototype);
  BailErrorStrategy.prototype.constructor = BailErrorStrategy;

  // Instead of recovering from exception {@code e}, re-throw it wrapped
  // in a {@link ParseCancellationException} so it is not caught by the
  // rule function catches. Use {@link Exception//getCause()} to get the
  // original {@link RecognitionException}.
  //
  BailErrorStrategy.prototype.recover = function(recognizer, e) {
      var context = recognizer._ctx;
      while (context !== null) {
          context.exception = e;
          context = context.parentCtx;
      }
      throw new ParseCancellationException(e);
  };

  // Make sure we don't attempt to recover inline; if the parser
  // successfully recovers, it won't throw an exception.
  //
  BailErrorStrategy.prototype.recoverInline = function(recognizer) {
      this.recover(recognizer, new InputMismatchException(recognizer));
  };

  // Make sure we don't attempt to recover from problems in subrules.//
  BailErrorStrategy.prototype.sync = function(recognizer) {
      // pass
  };

  exports.BailErrorStrategy = BailErrorStrategy;
  exports.DefaultErrorStrategy = DefaultErrorStrategy;
  },{"./../IntervalSet":6,"./../Token":14,"./../atn/ATNState":22,"./Errors":39}],39:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */

  // The root of the ANTLR exception hierarchy. In general, ANTLR tracks just
  //  3 kinds of errors: prediction errors, failed predicate errors, and
  //  mismatched input errors. In each case, the parser knows where it is
  //  in the input, where it is in the ATN, the rule invocation stack,
  //  and what kind of problem occurred.

  var PredicateTransition = require('./../atn/Transition').PredicateTransition;

  function RecognitionException(params) {
  	Error.call(this);
  	if (!!Error.captureStackTrace) {
          Error.captureStackTrace(this, RecognitionException);
  	}
  	this.message = params.message;
      this.recognizer = params.recognizer;
      this.input = params.input;
      this.ctx = params.ctx;
      // The current {@link Token} when an error occurred. Since not all streams
      // support accessing symbols by index, we have to track the {@link Token}
      // instance itself.
      this.offendingToken = null;
      // Get the ATN state number the parser was in at the time the error
      // occurred. For {@link NoViableAltException} and
      // {@link LexerNoViableAltException} exceptions, this is the
      // {@link DecisionState} number. For others, it is the state whose outgoing
      // edge we couldn't match.
      this.offendingState = -1;
      if (this.recognizer!==null) {
          this.offendingState = this.recognizer.state;
      }
      return this;
  }

  RecognitionException.prototype = Object.create(Error.prototype);
  RecognitionException.prototype.constructor = RecognitionException;

  // <p>If the state number is not known, this method returns -1.</p>

  //
  // Gets the set of input symbols which could potentially follow the
  // previously matched symbol at the time this exception was thrown.
  //
  // <p>If the set of expected tokens is not known and could not be computed,
  // this method returns {@code null}.</p>
  //
  // @return The set of token types that could potentially follow the current
  // state in the ATN, or {@code null} if the information is not available.
  // /
  RecognitionException.prototype.getExpectedTokens = function() {
      if (this.recognizer!==null) {
          return this.recognizer.atn.getExpectedTokens(this.offendingState, this.ctx);
      } else {
          return null;
      }
  };

  RecognitionException.prototype.toString = function() {
      return this.message;
  };

  function LexerNoViableAltException(lexer, input, startIndex, deadEndConfigs) {
  	RecognitionException.call(this, {message:"", recognizer:lexer, input:input, ctx:null});
      this.startIndex = startIndex;
      this.deadEndConfigs = deadEndConfigs;
      return this;
  }

  LexerNoViableAltException.prototype = Object.create(RecognitionException.prototype);
  LexerNoViableAltException.prototype.constructor = LexerNoViableAltException;

  LexerNoViableAltException.prototype.toString = function() {
      var symbol = "";
      if (this.startIndex >= 0 && this.startIndex < this.input.size) {
          symbol = this.input.getText((this.startIndex,this.startIndex));
      }
      return "LexerNoViableAltException" + symbol;
  };

  // Indicates that the parser could not decide which of two or more paths
  // to take based upon the remaining input. It tracks the starting token
  // of the offending input and also knows where the parser was
  // in the various paths when the error. Reported by reportNoViableAlternative()
  //
  function NoViableAltException(recognizer, input, startToken, offendingToken, deadEndConfigs, ctx) {
  	ctx = ctx || recognizer._ctx;
  	offendingToken = offendingToken || recognizer.getCurrentToken();
  	startToken = startToken || recognizer.getCurrentToken();
  	input = input || recognizer.getInputStream();
  	RecognitionException.call(this, {message:"", recognizer:recognizer, input:input, ctx:ctx});
      // Which configurations did we try at input.index() that couldn't match
  	// input.LT(1)?//
      this.deadEndConfigs = deadEndConfigs;
      // The token object at the start index; the input stream might
      // not be buffering tokens so get a reference to it. (At the
      // time the error occurred, of course the stream needs to keep a
      // buffer all of the tokens but later we might not have access to those.)
      this.startToken = startToken;
      this.offendingToken = offendingToken;
  }

  NoViableAltException.prototype = Object.create(RecognitionException.prototype);
  NoViableAltException.prototype.constructor = NoViableAltException;

  // This signifies any kind of mismatched input exceptions such as
  // when the current input does not match the expected token.
  //
  function InputMismatchException(recognizer) {
  	RecognitionException.call(this, {message:"", recognizer:recognizer, input:recognizer.getInputStream(), ctx:recognizer._ctx});
      this.offendingToken = recognizer.getCurrentToken();
  }

  InputMismatchException.prototype = Object.create(RecognitionException.prototype);
  InputMismatchException.prototype.constructor = InputMismatchException;

  // A semantic predicate failed during validation. Validation of predicates
  // occurs when normally parsing the alternative just like matching a token.
  // Disambiguating predicate evaluation occurs when we test a predicate during
  // prediction.

  function FailedPredicateException(recognizer, predicate, message) {
  	RecognitionException.call(this, {message:this.formatMessage(predicate,message || null), recognizer:recognizer,
                           input:recognizer.getInputStream(), ctx:recognizer._ctx});
      var s = recognizer._interp.atn.states[recognizer.state];
      var trans = s.transitions[0];
      if (trans instanceof PredicateTransition) {
          this.ruleIndex = trans.ruleIndex;
          this.predicateIndex = trans.predIndex;
      } else {
          this.ruleIndex = 0;
          this.predicateIndex = 0;
      }
      this.predicate = predicate;
      this.offendingToken = recognizer.getCurrentToken();
      return this;
  }

  FailedPredicateException.prototype = Object.create(RecognitionException.prototype);
  FailedPredicateException.prototype.constructor = FailedPredicateException;

  FailedPredicateException.prototype.formatMessage = function(predicate, message) {
      if (message !==null) {
          return message;
      } else {
          return "failed predicate: {" + predicate + "}?";
      }
  };

  function ParseCancellationException() {
  	Error.call(this);
  	Error.captureStackTrace(this, ParseCancellationException);
  	return this;
  }

  ParseCancellationException.prototype = Object.create(Error.prototype);
  ParseCancellationException.prototype.constructor = ParseCancellationException;

  exports.RecognitionException = RecognitionException;
  exports.NoViableAltException = NoViableAltException;
  exports.LexerNoViableAltException = LexerNoViableAltException;
  exports.InputMismatchException = InputMismatchException;
  exports.FailedPredicateException = FailedPredicateException;
  exports.ParseCancellationException = ParseCancellationException;

  },{"./../atn/Transition":30}],40:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */

  exports.RecognitionException = require('./Errors').RecognitionException;
  exports.NoViableAltException = require('./Errors').NoViableAltException;
  exports.LexerNoViableAltException = require('./Errors').LexerNoViableAltException;
  exports.InputMismatchException = require('./Errors').InputMismatchException;
  exports.FailedPredicateException = require('./Errors').FailedPredicateException;
  exports.DiagnosticErrorListener = require('./DiagnosticErrorListener').DiagnosticErrorListener;
  exports.BailErrorStrategy = require('./ErrorStrategy').BailErrorStrategy;
  exports.ErrorListener = require('./ErrorListener').ErrorListener;

  },{"./DiagnosticErrorListener":36,"./ErrorListener":37,"./ErrorStrategy":38,"./Errors":39}],41:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  exports.atn = require('./atn/index');
  exports.codepointat = require('./polyfills/codepointat');
  exports.dfa = require('./dfa/index');
  exports.fromcodepoint = require('./polyfills/fromcodepoint');
  exports.tree = require('./tree/index');
  exports.error = require('./error/index');
  exports.Token = require('./Token').Token;
  exports.CommonToken = require('./Token').CommonToken;
  exports.InputStream = require('./InputStream').InputStream;
  exports.FileStream = require('./FileStream').FileStream;
  exports.CommonTokenStream = require('./CommonTokenStream').CommonTokenStream;
  exports.Lexer = require('./Lexer').Lexer;
  exports.Parser = require('./Parser').Parser;
  var pc = require('./PredictionContext');
  exports.PredictionContextCache = pc.PredictionContextCache;
  exports.ParserRuleContext = require('./ParserRuleContext').ParserRuleContext;
  exports.Interval = require('./IntervalSet').Interval;
  exports.Utils = require('./Utils');

  },{"./CommonTokenStream":3,"./FileStream":4,"./InputStream":5,"./IntervalSet":6,"./Lexer":8,"./Parser":9,"./ParserRuleContext":10,"./PredictionContext":11,"./Token":14,"./Utils":15,"./atn/index":31,"./dfa/index":35,"./error/index":40,"./polyfills/codepointat":42,"./polyfills/fromcodepoint":43,"./tree/index":46}],42:[function(require,module,exports){
  /*! https://mths.be/codepointat v0.2.0 by @mathias */
  if (!String.prototype.codePointAt) {
  	(function() {
  		var defineProperty = (function() {
  			// IE 8 only supports `Object.defineProperty` on DOM elements
  			try {
  				var object = {};
  				var $defineProperty = Object.defineProperty;
  				var result = $defineProperty(object, object, object) && $defineProperty;
  			} catch(error) {}
  			return result;
  		}());
  		var codePointAt = function(position) {
  			if (this == null) {
  				throw TypeError();
  			}
  			var string = String(this);
  			var size = string.length;
  			// `ToInteger`
  			var index = position ? Number(position) : 0;
  			if (index != index) { // better `isNaN`
  				index = 0;
  			}
  			// Account for out-of-bounds indices:
  			if (index < 0 || index >= size) {
  				return undefined;
  			}
  			// Get the first code unit
  			var first = string.charCodeAt(index);
  			var second;
  			if ( // check if it’s the start of a surrogate pair
  				first >= 0xD800 && first <= 0xDBFF && // high surrogate
  				size > index + 1 // there is a next code unit
  			) {
  				second = string.charCodeAt(index + 1);
  				if (second >= 0xDC00 && second <= 0xDFFF) { // low surrogate
  					// https://mathiasbynens.be/notes/javascript-encoding#surrogate-formulae
  					return (first - 0xD800) * 0x400 + second - 0xDC00 + 0x10000;
  				}
  			}
  			return first;
  		};
  		if (defineProperty) {
  			defineProperty(String.prototype, 'codePointAt', {
  				'value': codePointAt,
  				'configurable': true,
  				'writable': true
  			});
  		} else {
  			String.prototype.codePointAt = codePointAt;
  		}
  	}());
  }

  },{}],43:[function(require,module,exports){
  /*! https://mths.be/fromcodepoint v0.2.1 by @mathias */
  if (!String.fromCodePoint) {
  	(function() {
  		var defineProperty = (function() {
  			// IE 8 only supports `Object.defineProperty` on DOM elements
  			try {
  				var object = {};
  				var $defineProperty = Object.defineProperty;
  				var result = $defineProperty(object, object, object) && $defineProperty;
  			} catch(error) {}
  			return result;
  		}());
  		var stringFromCharCode = String.fromCharCode;
  		var floor = Math.floor;
  		var fromCodePoint = function(_) {
  			var MAX_SIZE = 0x4000;
  			var codeUnits = [];
  			var highSurrogate;
  			var lowSurrogate;
  			var index = -1;
  			var length = arguments.length;
  			if (!length) {
  				return '';
  			}
  			var result = '';
  			while (++index < length) {
  				var codePoint = Number(arguments[index]);
  				if (
  					!isFinite(codePoint) || // `NaN`, `+Infinity`, or `-Infinity`
  					codePoint < 0 || // not a valid Unicode code point
  					codePoint > 0x10FFFF || // not a valid Unicode code point
  					floor(codePoint) != codePoint // not an integer
  				) {
  					throw RangeError('Invalid code point: ' + codePoint);
  				}
  				if (codePoint <= 0xFFFF) { // BMP code point
  					codeUnits.push(codePoint);
  				} else { // Astral code point; split in surrogate halves
  					// https://mathiasbynens.be/notes/javascript-encoding#surrogate-formulae
  					codePoint -= 0x10000;
  					highSurrogate = (codePoint >> 10) + 0xD800;
  					lowSurrogate = (codePoint % 0x400) + 0xDC00;
  					codeUnits.push(highSurrogate, lowSurrogate);
  				}
  				if (index + 1 == length || codeUnits.length > MAX_SIZE) {
  					result += stringFromCharCode.apply(null, codeUnits);
  					codeUnits.length = 0;
  				}
  			}
  			return result;
  		};
  		if (defineProperty) {
  			defineProperty(String, 'fromCodePoint', {
  				'value': fromCodePoint,
  				'configurable': true,
  				'writable': true
  			});
  		} else {
  			String.fromCodePoint = fromCodePoint;
  		}
  	}());
  }

  },{}],44:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */
  ///

  // The basic notion of a tree has a parent, a payload, and a list of children.
  //  It is the most abstract interface for all the trees used by ANTLR.
  ///

  var Token = require('./../Token').Token;
  var Interval = require('./../IntervalSet').Interval;
  var INVALID_INTERVAL = new Interval(-1, -2);
  var Utils = require('../Utils.js');


  function Tree() {
  	return this;
  }

  function SyntaxTree() {
  	Tree.call(this);
  	return this;
  }

  SyntaxTree.prototype = Object.create(Tree.prototype);
  SyntaxTree.prototype.constructor = SyntaxTree;

  function ParseTree() {
  	SyntaxTree.call(this);
  	return this;
  }

  ParseTree.prototype = Object.create(SyntaxTree.prototype);
  ParseTree.prototype.constructor = ParseTree;

  function RuleNode() {
  	ParseTree.call(this);
  	return this;
  }

  RuleNode.prototype = Object.create(ParseTree.prototype);
  RuleNode.prototype.constructor = RuleNode;

  function TerminalNode() {
  	ParseTree.call(this);
  	return this;
  }

  TerminalNode.prototype = Object.create(ParseTree.prototype);
  TerminalNode.prototype.constructor = TerminalNode;

  function ErrorNode() {
  	TerminalNode.call(this);
  	return this;
  }

  ErrorNode.prototype = Object.create(TerminalNode.prototype);
  ErrorNode.prototype.constructor = ErrorNode;

  function ParseTreeVisitor() {
  	return this;
  }

  ParseTreeVisitor.prototype.visit = function(ctx) {
   	if (Array.isArray(ctx)) {
  		return ctx.map(function(child) {
              return child.accept(this);
          }, this);
  	} else {
  		return ctx.accept(this);
  	}
  };

  ParseTreeVisitor.prototype.visitChildren = function(ctx) {
    return this.visit(ctx.children);
  };

  ParseTreeVisitor.prototype.visitTerminal = function(node) {
  };

  ParseTreeVisitor.prototype.visitErrorNode = function(node) {
  };


  function ParseTreeListener() {
  	return this;
  }

  ParseTreeListener.prototype.visitTerminal = function(node) {
  };

  ParseTreeListener.prototype.visitErrorNode = function(node) {
  };

  ParseTreeListener.prototype.enterEveryRule = function(node) {
  };

  ParseTreeListener.prototype.exitEveryRule = function(node) {
  };

  function TerminalNodeImpl(symbol) {
  	TerminalNode.call(this);
  	this.parentCtx = null;
  	this.symbol = symbol;
  	return this;
  }

  TerminalNodeImpl.prototype = Object.create(TerminalNode.prototype);
  TerminalNodeImpl.prototype.constructor = TerminalNodeImpl;

  TerminalNodeImpl.prototype.getChild = function(i) {
  	return null;
  };

  TerminalNodeImpl.prototype.getSymbol = function() {
  	return this.symbol;
  };

  TerminalNodeImpl.prototype.getParent = function() {
  	return this.parentCtx;
  };

  TerminalNodeImpl.prototype.getPayload = function() {
  	return this.symbol;
  };

  TerminalNodeImpl.prototype.getSourceInterval = function() {
  	if (this.symbol === null) {
  		return INVALID_INTERVAL;
  	}
  	var tokenIndex = this.symbol.tokenIndex;
  	return new Interval(tokenIndex, tokenIndex);
  };

  TerminalNodeImpl.prototype.getChildCount = function() {
  	return 0;
  };

  TerminalNodeImpl.prototype.accept = function(visitor) {
  	return visitor.visitTerminal(this);
  };

  TerminalNodeImpl.prototype.getText = function() {
  	return this.symbol.text;
  };

  TerminalNodeImpl.prototype.toString = function() {
  	if (this.symbol.type === Token.EOF) {
  		return "<EOF>";
  	} else {
  		return this.symbol.text;
  	}
  };

  // Represents a token that was consumed during resynchronization
  // rather than during a valid match operation. For example,
  // we will create this kind of a node during single token insertion
  // and deletion as well as during "consume until error recovery set"
  // upon no viable alternative exceptions.

  function ErrorNodeImpl(token) {
  	TerminalNodeImpl.call(this, token);
  	return this;
  }

  ErrorNodeImpl.prototype = Object.create(TerminalNodeImpl.prototype);
  ErrorNodeImpl.prototype.constructor = ErrorNodeImpl;

  ErrorNodeImpl.prototype.isErrorNode = function() {
  	return true;
  };

  ErrorNodeImpl.prototype.accept = function(visitor) {
  	return visitor.visitErrorNode(this);
  };

  function ParseTreeWalker() {
  	return this;
  }

  ParseTreeWalker.prototype.walk = function(listener, t) {
  	var errorNode = t instanceof ErrorNode ||
  			(t.isErrorNode !== undefined && t.isErrorNode());
  	if (errorNode) {
  		listener.visitErrorNode(t);
  	} else if (t instanceof TerminalNode) {
  		listener.visitTerminal(t);
  	} else {
  		this.enterRule(listener, t);
  		for (var i = 0; i < t.getChildCount(); i++) {
  			var child = t.getChild(i);
  			this.walk(listener, child);
  		}
  		this.exitRule(listener, t);
  	}
  };
  //
  // The discovery of a rule node, involves sending two events: the generic
  // {@link ParseTreeListener//enterEveryRule} and a
  // {@link RuleContext}-specific event. First we trigger the generic and then
  // the rule specific. We to them in reverse order upon finishing the node.
  //
  ParseTreeWalker.prototype.enterRule = function(listener, r) {
  	var ctx = r.getRuleContext();
  	listener.enterEveryRule(ctx);
  	ctx.enterRule(listener);
  };

  ParseTreeWalker.prototype.exitRule = function(listener, r) {
  	var ctx = r.getRuleContext();
  	ctx.exitRule(listener);
  	listener.exitEveryRule(ctx);
  };

  ParseTreeWalker.DEFAULT = new ParseTreeWalker();

  exports.RuleNode = RuleNode;
  exports.ErrorNode = ErrorNode;
  exports.TerminalNode = TerminalNode;
  exports.ErrorNodeImpl = ErrorNodeImpl;
  exports.TerminalNodeImpl = TerminalNodeImpl;
  exports.ParseTreeListener = ParseTreeListener;
  exports.ParseTreeVisitor = ParseTreeVisitor;
  exports.ParseTreeWalker = ParseTreeWalker;
  exports.INVALID_INTERVAL = INVALID_INTERVAL;

  },{"../Utils.js":15,"./../IntervalSet":6,"./../Token":14}],45:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */

  var Utils = require('./../Utils');
  var Token = require('./../Token').Token;
  var RuleNode = require('./Tree').RuleNode;
  var ErrorNode = require('./Tree').ErrorNode;
  var TerminalNode = require('./Tree').TerminalNode;
  var ParserRuleContext = require('./../ParserRuleContext').ParserRuleContext;
  var RuleContext = require('./../RuleContext').RuleContext;
  var INVALID_ALT_NUMBER = require('./../atn/ATN').INVALID_ALT_NUMBER;


  /** A set of utility routines useful for all kinds of ANTLR trees. */
  function Trees() {
  }

  // Print out a whole tree in LISP form. {@link //getNodeText} is used on the
  //  node payloads to get the text for the nodes.  Detect
  //  parse trees and extract data appropriately.
  Trees.toStringTree = function(tree, ruleNames, recog) {
  	ruleNames = ruleNames || null;
  	recog = recog || null;
      if(recog!==null) {
         ruleNames = recog.ruleNames;
      }
      var s = Trees.getNodeText(tree, ruleNames);
      s = Utils.escapeWhitespace(s, false);
      var c = tree.getChildCount();
      if(c===0) {
          return s;
      }
      var res = "(" + s + ' ';
      if(c>0) {
          s = Trees.toStringTree(tree.getChild(0), ruleNames);
          res = res.concat(s);
      }
      for(var i=1;i<c;i++) {
          s = Trees.toStringTree(tree.getChild(i), ruleNames);
          res = res.concat(' ' + s);
      }
      res = res.concat(")");
      return res;
  };

  Trees.getNodeText = function(t, ruleNames, recog) {
  	ruleNames = ruleNames || null;
  	recog = recog || null;
      if(recog!==null) {
          ruleNames = recog.ruleNames;
      }
      if(ruleNames!==null) {
         if (t instanceof RuleContext) {
             var altNumber = t.getAltNumber();
             if ( altNumber!=INVALID_ALT_NUMBER ) {
                 return ruleNames[t.ruleIndex]+":"+altNumber;
             }
             return ruleNames[t.ruleIndex];
         } else if ( t instanceof ErrorNode) {
             return t.toString();
         } else if(t instanceof TerminalNode) {
             if(t.symbol!==null) {
                 return t.symbol.text;
             }
         }
      }
      // no recog for rule names
      var payload = t.getPayload();
      if (payload instanceof Token ) {
         return payload.text;
      }
      return t.getPayload().toString();
  };


  // Return ordered list of all children of this node
  Trees.getChildren = function(t) {
  	var list = [];
  	for(var i=0;i<t.getChildCount();i++) {
  		list.push(t.getChild(i));
  	}
  	return list;
  };

  // Return a list of all ancestors of this node.  The first node of
  //  list is the root and the last is the parent of this node.
  //
  Trees.getAncestors = function(t) {
      var ancestors = [];
      t = t.getParent();
      while(t!==null) {
          ancestors = [t].concat(ancestors);
          t = t.getParent();
      }
      return ancestors;
  };

  Trees.findAllTokenNodes = function(t, ttype) {
      return Trees.findAllNodes(t, ttype, true);
  };

  Trees.findAllRuleNodes = function(t, ruleIndex) {
  	return Trees.findAllNodes(t, ruleIndex, false);
  };

  Trees.findAllNodes = function(t, index, findTokens) {
  	var nodes = [];
  	Trees._findAllNodes(t, index, findTokens, nodes);
  	return nodes;
  };

  Trees._findAllNodes = function(t, index, findTokens, nodes) {
  	// check this node (the root) first
  	if(findTokens && (t instanceof TerminalNode)) {
  		if(t.symbol.type===index) {
  			nodes.push(t);
  		}
  	} else if(!findTokens && (t instanceof ParserRuleContext)) {
  		if(t.ruleIndex===index) {
  			nodes.push(t);
  		}
  	}
  	// check children
  	for(var i=0;i<t.getChildCount();i++) {
  		Trees._findAllNodes(t.getChild(i), index, findTokens, nodes);
  	}
  };

  Trees.descendants = function(t) {
  	var nodes = [t];
      for(var i=0;i<t.getChildCount();i++) {
          nodes = nodes.concat(Trees.descendants(t.getChild(i)));
      }
      return nodes;
  };


  exports.Trees = Trees;
  },{"./../ParserRuleContext":10,"./../RuleContext":13,"./../Token":14,"./../Utils":15,"./../atn/ATN":16,"./Tree":44}],46:[function(require,module,exports){
  /* Copyright (c) 2012-2016 The ANTLR Project. All rights reserved.
   * Use of this file is governed by the BSD 3-clause license that
   * can be found in the LICENSE.txt file in the project root.
   */

  var Tree = require('./Tree');
  exports.Trees = require('./Trees').Trees;
  exports.RuleNode = Tree.RuleNode;
  exports.ParseTreeListener = Tree.ParseTreeListener;
  exports.ParseTreeVisitor = Tree.ParseTreeVisitor;
  exports.ParseTreeWalker = Tree.ParseTreeWalker;

  },{"./Tree":44,"./Trees":45}],47:[function(require,module,exports){

  },{}]},{},[41])(41);

  var serializedATN = ["\u0003\u0430\ud6d1\u8206\uad2d\u4417\uaef1\u8d80\uaadd",
      "\u0002#\u00c1\b\u0001\u0004\u0002\t\u0002\u0004\u0003\t\u0003\u0004",
      "\u0004\t\u0004\u0004\u0005\t\u0005\u0004\u0006\t\u0006\u0004\u0007\t",
      "\u0007\u0004\b\t\b\u0004\t\t\t\u0004\n\t\n\u0004\u000b\t\u000b\u0004",
      "\f\t\f\u0004\r\t\r\u0004\u000e\t\u000e\u0004\u000f\t\u000f\u0004\u0010",
      "\t\u0010\u0004\u0011\t\u0011\u0004\u0012\t\u0012\u0004\u0013\t\u0013",
      "\u0004\u0014\t\u0014\u0004\u0015\t\u0015\u0004\u0016\t\u0016\u0004\u0017",
      "\t\u0017\u0004\u0018\t\u0018\u0004\u0019\t\u0019\u0004\u001a\t\u001a",
      "\u0004\u001b\t\u001b\u0004\u001c\t\u001c\u0004\u001d\t\u001d\u0004\u001e",
      "\t\u001e\u0004\u001f\t\u001f\u0004 \t \u0004!\t!\u0004\"\t\"\u0003\u0002",
      "\u0003\u0002\u0003\u0003\u0003\u0003\u0003\u0004\u0003\u0004\u0003\u0005",
      "\u0003\u0005\u0003\u0006\u0003\u0006\u0003\u0007\u0003\u0007\u0003\b",
      "\u0003\b\u0003\t\u0003\t\u0003\n\u0003\n\u0003\u000b\u0003\u000b\u0003",
      "\f\u0003\f\u0003\r\u0003\r\u0003\r\u0003\u000e\u0003\u000e\u0003\u000f",
      "\u0003\u000f\u0003\u000f\u0003\u0010\u0003\u0010\u0003\u0010\u0003\u0011",
      "\u0003\u0011\u0003\u0011\u0003\u0012\u0003\u0012\u0003\u0012\u0003\u0013",
      "\u0003\u0013\u0003\u0013\u0003\u0014\u0003\u0014\u0003\u0015\u0003\u0015",
      "\u0003\u0016\u0003\u0016\u0003\u0017\u0003\u0017\u0003\u0017\u0003\u0018",
      "\u0003\u0018\u0003\u0019\u0003\u0019\u0003\u001a\u0003\u001a\u0003\u001b",
      "\u0003\u001b\u0003\u001c\u0003\u001c\u0003\u001c\u0003\u001c\u0003\u001c",
      "\u0003\u001c\u0003\u001c\u0003\u001c\u0003\u001c\u0005\u001c\u008a\n",
      "\u001c\u0003\u001d\u0003\u001d\u0007\u001d\u008e\n\u001d\f\u001d\u000e",
      "\u001d\u0091\u000b\u001d\u0003\u001e\u0006\u001e\u0094\n\u001e\r\u001e",
      "\u000e\u001e\u0095\u0003\u001f\u0006\u001f\u0099\n\u001f\r\u001f\u000e",
      "\u001f\u009a\u0003\u001f\u0003\u001f\u0006\u001f\u009f\n\u001f\r\u001f",
      "\u000e\u001f\u00a0\u0003 \u0003 \u0003 \u0003 \u0007 \u00a7\n \f \u000e",
      " \u00aa\u000b \u0003 \u0003 \u0003 \u0003 \u0003 \u0007 \u00b1\n \f",
      " \u000e \u00b4\u000b \u0003 \u0005 \u00b7\n \u0003!\u0003!\u0003\"\u0006",
      "\"\u00bc\n\"\r\"\u000e\"\u00bd\u0003\"\u0003\"\u0004\u00a8\u00b2\u0002",
      "#\u0003\u0003\u0005\u0004\u0007\u0005\t\u0006\u000b\u0007\r\b\u000f",
      "\t\u0011\n\u0013\u000b\u0015\f\u0017\r\u0019\u000e\u001b\u000f\u001d",
      "\u0010\u001f\u0011!\u0012#\u0013%\u0014\'\u0015)\u0016+\u0017-\u0018",
      "/\u00191\u001a3\u001b5\u001c7\u001d9\u001e;\u001f= ?!A\"C#\u0003\u0002",
      "\u0007\u0005\u0002B\\aac|\u0006\u00022;B\\aac|\u0003\u00022;\u0003\u0002",
      "))\u0005\u0002\u000b\f\u000f\u000f\"\"\u00cb\u0002\u0003\u0003\u0002",
      "\u0002\u0002\u0002\u0005\u0003\u0002\u0002\u0002\u0002\u0007\u0003\u0002",
      "\u0002\u0002\u0002\t\u0003\u0002\u0002\u0002\u0002\u000b\u0003\u0002",
      "\u0002\u0002\u0002\r\u0003\u0002\u0002\u0002\u0002\u000f\u0003\u0002",
      "\u0002\u0002\u0002\u0011\u0003\u0002\u0002\u0002\u0002\u0013\u0003\u0002",
      "\u0002\u0002\u0002\u0015\u0003\u0002\u0002\u0002\u0002\u0017\u0003\u0002",
      "\u0002\u0002\u0002\u0019\u0003\u0002\u0002\u0002\u0002\u001b\u0003\u0002",
      "\u0002\u0002\u0002\u001d\u0003\u0002\u0002\u0002\u0002\u001f\u0003\u0002",
      "\u0002\u0002\u0002!\u0003\u0002\u0002\u0002\u0002#\u0003\u0002\u0002",
      "\u0002\u0002%\u0003\u0002\u0002\u0002\u0002\'\u0003\u0002\u0002\u0002",
      "\u0002)\u0003\u0002\u0002\u0002\u0002+\u0003\u0002\u0002\u0002\u0002",
      "-\u0003\u0002\u0002\u0002\u0002/\u0003\u0002\u0002\u0002\u00021\u0003",
      "\u0002\u0002\u0002\u00023\u0003\u0002\u0002\u0002\u00025\u0003\u0002",
      "\u0002\u0002\u00027\u0003\u0002\u0002\u0002\u00029\u0003\u0002\u0002",
      "\u0002\u0002;\u0003\u0002\u0002\u0002\u0002=\u0003\u0002\u0002\u0002",
      "\u0002?\u0003\u0002\u0002\u0002\u0002A\u0003\u0002\u0002\u0002\u0002",
      "C\u0003\u0002\u0002\u0002\u0003E\u0003\u0002\u0002\u0002\u0005G\u0003",
      "\u0002\u0002\u0002\u0007I\u0003\u0002\u0002\u0002\tK\u0003\u0002\u0002",
      "\u0002\u000bM\u0003\u0002\u0002\u0002\rO\u0003\u0002\u0002\u0002\u000f",
      "Q\u0003\u0002\u0002\u0002\u0011S\u0003\u0002\u0002\u0002\u0013U\u0003",
      "\u0002\u0002\u0002\u0015W\u0003\u0002\u0002\u0002\u0017Y\u0003\u0002",
      "\u0002\u0002\u0019[\u0003\u0002\u0002\u0002\u001b^\u0003\u0002\u0002",
      "\u0002\u001d`\u0003\u0002\u0002\u0002\u001fc\u0003\u0002\u0002\u0002",
      "!f\u0003\u0002\u0002\u0002#i\u0003\u0002\u0002\u0002%l\u0003\u0002\u0002",
      "\u0002\'o\u0003\u0002\u0002\u0002)q\u0003\u0002\u0002\u0002+s\u0003",
      "\u0002\u0002\u0002-u\u0003\u0002\u0002\u0002/x\u0003\u0002\u0002\u0002",
      "1z\u0003\u0002\u0002\u00023|\u0003\u0002\u0002\u00025~\u0003\u0002\u0002",
      "\u00027\u0089\u0003\u0002\u0002\u00029\u008b\u0003\u0002\u0002\u0002",
      ";\u0093\u0003\u0002\u0002\u0002=\u0098\u0003\u0002\u0002\u0002?\u00b6",
      "\u0003\u0002\u0002\u0002A\u00b8\u0003\u0002\u0002\u0002C\u00bb\u0003",
      "\u0002\u0002\u0002EF\u00070\u0002\u0002F\u0004\u0003\u0002\u0002\u0002",
      "GH\u0007]\u0002\u0002H\u0006\u0003\u0002\u0002\u0002IJ\u0007_\u0002",
      "\u0002J\b\u0003\u0002\u0002\u0002KL\u0007#\u0002\u0002L\n\u0003\u0002",
      "\u0002\u0002MN\u0007-\u0002\u0002N\f\u0003\u0002\u0002\u0002OP\u0007",
      "/\u0002\u0002P\u000e\u0003\u0002\u0002\u0002QR\u0007`\u0002\u0002R\u0010",
      "\u0003\u0002\u0002\u0002ST\u0007,\u0002\u0002T\u0012\u0003\u0002\u0002",
      "\u0002UV\u00071\u0002\u0002V\u0014\u0003\u0002\u0002\u0002WX\u0007\'",
      "\u0002\u0002X\u0016\u0003\u0002\u0002\u0002YZ\u0007>\u0002\u0002Z\u0018",
      "\u0003\u0002\u0002\u0002[\\\u0007>\u0002\u0002\\]\u0007?\u0002\u0002",
      "]\u001a\u0003\u0002\u0002\u0002^_\u0007@\u0002\u0002_\u001c\u0003\u0002",
      "\u0002\u0002`a\u0007@\u0002\u0002ab\u0007?\u0002\u0002b\u001e\u0003",
      "\u0002\u0002\u0002cd\u0007?\u0002\u0002de\u0007?\u0002\u0002e \u0003",
      "\u0002\u0002\u0002fg\u0007#\u0002\u0002gh\u0007?\u0002\u0002h\"\u0003",
      "\u0002\u0002\u0002ij\u0007(\u0002\u0002jk\u0007(\u0002\u0002k$\u0003",
      "\u0002\u0002\u0002lm\u0007~\u0002\u0002mn\u0007~\u0002\u0002n&\u0003",
      "\u0002\u0002\u0002op\u0007~\u0002\u0002p(\u0003\u0002\u0002\u0002qr",
      "\u0007*\u0002\u0002r*\u0003\u0002\u0002\u0002st\u0007+\u0002\u0002t",
      ",\u0003\u0002\u0002\u0002uv\u0007*\u0002\u0002vw\u0007+\u0002\u0002",
      "w.\u0003\u0002\u0002\u0002xy\u0007.\u0002\u0002y0\u0003\u0002\u0002",
      "\u0002z{\u0007}\u0002\u0002{2\u0003\u0002\u0002\u0002|}\u0007<\u0002",
      "\u0002}4\u0003\u0002\u0002\u0002~\u007f\u0007\u007f\u0002\u0002\u007f",
      "6\u0003\u0002\u0002\u0002\u0080\u0081\u0007v\u0002\u0002\u0081\u0082",
      "\u0007t\u0002\u0002\u0082\u0083\u0007w\u0002\u0002\u0083\u008a\u0007",
      "g\u0002\u0002\u0084\u0085\u0007h\u0002\u0002\u0085\u0086\u0007c\u0002",
      "\u0002\u0086\u0087\u0007n\u0002\u0002\u0087\u0088\u0007u\u0002\u0002",
      "\u0088\u008a\u0007g\u0002\u0002\u0089\u0080\u0003\u0002\u0002\u0002",
      "\u0089\u0084\u0003\u0002\u0002\u0002\u008a8\u0003\u0002\u0002\u0002",
      "\u008b\u008f\t\u0002\u0002\u0002\u008c\u008e\t\u0003\u0002\u0002\u008d",
      "\u008c\u0003\u0002\u0002\u0002\u008e\u0091\u0003\u0002\u0002\u0002\u008f",
      "\u008d\u0003\u0002\u0002\u0002\u008f\u0090\u0003\u0002\u0002\u0002\u0090",
      ":\u0003\u0002\u0002\u0002\u0091\u008f\u0003\u0002\u0002\u0002\u0092",
      "\u0094\t\u0004\u0002\u0002\u0093\u0092\u0003\u0002\u0002\u0002\u0094",
      "\u0095\u0003\u0002\u0002\u0002\u0095\u0093\u0003\u0002\u0002\u0002\u0095",
      "\u0096\u0003\u0002\u0002\u0002\u0096<\u0003\u0002\u0002\u0002\u0097",
      "\u0099\t\u0004\u0002\u0002\u0098\u0097\u0003\u0002\u0002\u0002\u0099",
      "\u009a\u0003\u0002\u0002\u0002\u009a\u0098\u0003\u0002\u0002\u0002\u009a",
      "\u009b\u0003\u0002\u0002\u0002\u009b\u009c\u0003\u0002\u0002\u0002\u009c",
      "\u009e\u00070\u0002\u0002\u009d\u009f\t\u0004\u0002\u0002\u009e\u009d",
      "\u0003\u0002\u0002\u0002\u009f\u00a0\u0003\u0002\u0002\u0002\u00a0\u009e",
      "\u0003\u0002\u0002\u0002\u00a0\u00a1\u0003\u0002\u0002\u0002\u00a1>",
      "\u0003\u0002\u0002\u0002\u00a2\u00a8\u0007$\u0002\u0002\u00a3\u00a4",
      "\u0007^\u0002\u0002\u00a4\u00a7\u0007$\u0002\u0002\u00a5\u00a7\u000b",
      "\u0002\u0002\u0002\u00a6\u00a3\u0003\u0002\u0002\u0002\u00a6\u00a5\u0003",
      "\u0002\u0002\u0002\u00a7\u00aa\u0003\u0002\u0002\u0002\u00a8\u00a9\u0003",
      "\u0002\u0002\u0002\u00a8\u00a6\u0003\u0002\u0002\u0002\u00a9\u00ab\u0003",
      "\u0002\u0002\u0002\u00aa\u00a8\u0003\u0002\u0002\u0002\u00ab\u00b7\u0007",
      "$\u0002\u0002\u00ac\u00b2\t\u0005\u0002\u0002\u00ad\u00ae\u0007^\u0002",
      "\u0002\u00ae\u00b1\t\u0005\u0002\u0002\u00af\u00b1\u000b\u0002\u0002",
      "\u0002\u00b0\u00ad\u0003\u0002\u0002\u0002\u00b0\u00af\u0003\u0002\u0002",
      "\u0002\u00b1\u00b4\u0003\u0002\u0002\u0002\u00b2\u00b3\u0003\u0002\u0002",
      "\u0002\u00b2\u00b0\u0003\u0002\u0002\u0002\u00b3\u00b5\u0003\u0002\u0002",
      "\u0002\u00b4\u00b2\u0003\u0002\u0002\u0002\u00b5\u00b7\t\u0005\u0002",
      "\u0002\u00b6\u00a2\u0003\u0002\u0002\u0002\u00b6\u00ac\u0003\u0002\u0002",
      "\u0002\u00b7@\u0003\u0002\u0002\u0002\u00b8\u00b9\u0007?\u0002\u0002",
      "\u00b9B\u0003\u0002\u0002\u0002\u00ba\u00bc\t\u0006\u0002\u0002\u00bb",
      "\u00ba\u0003\u0002\u0002\u0002\u00bc\u00bd\u0003\u0002\u0002\u0002\u00bd",
      "\u00bb\u0003\u0002\u0002\u0002\u00bd\u00be\u0003\u0002\u0002\u0002\u00be",
      "\u00bf\u0003\u0002\u0002\u0002\u00bf\u00c0\b\"\u0002\u0002\u00c0D\u0003",
      "\u0002\u0002\u0002\u000e\u0002\u0089\u008f\u0095\u009a\u00a0\u00a6\u00a8",
      "\u00b0\u00b2\u00b6\u00bd\u0003\b\u0002\u0002"].join("");


  var atn = new _exports.atn.ATNDeserializer().deserialize(serializedATN);

  var decisionsToDFA = atn.decisionToState.map( function(ds, index) { return new _exports.dfa.DFA(ds, index); });

  function MiniLexer(input) {
  	_exports.Lexer.call(this, input);
      this._interp = new _exports.atn.LexerATNSimulator(this, atn, decisionsToDFA, new _exports.PredictionContextCache());
      return this;
  }

  MiniLexer.prototype = Object.create(_exports.Lexer.prototype);
  MiniLexer.prototype.constructor = MiniLexer;

  MiniLexer.EOF = _exports.Token.EOF;
  MiniLexer.T__0 = 1;
  MiniLexer.T__1 = 2;
  MiniLexer.T__2 = 3;
  MiniLexer.T__3 = 4;
  MiniLexer.T__4 = 5;
  MiniLexer.T__5 = 6;
  MiniLexer.T__6 = 7;
  MiniLexer.T__7 = 8;
  MiniLexer.T__8 = 9;
  MiniLexer.T__9 = 10;
  MiniLexer.T__10 = 11;
  MiniLexer.T__11 = 12;
  MiniLexer.T__12 = 13;
  MiniLexer.T__13 = 14;
  MiniLexer.T__14 = 15;
  MiniLexer.T__15 = 16;
  MiniLexer.T__16 = 17;
  MiniLexer.T__17 = 18;
  MiniLexer.T__18 = 19;
  MiniLexer.T__19 = 20;
  MiniLexer.T__20 = 21;
  MiniLexer.T__21 = 22;
  MiniLexer.T__22 = 23;
  MiniLexer.T__23 = 24;
  MiniLexer.T__24 = 25;
  MiniLexer.T__25 = 26;
  MiniLexer.BOOLEAN = 27;
  MiniLexer.ID = 28;
  MiniLexer.INT = 29;
  MiniLexer.FLOAT = 30;
  MiniLexer.STRING = 31;
  MiniLexer.EQ = 32;
  MiniLexer.WS = 33;


  MiniLexer.prototype.modeNames = [ "DEFAULT_MODE" ];

  MiniLexer.prototype.literalNames = [ null, "'.'", "'['", "']'", "'!'", "'+'", 
                                       "'-'", "'^'", "'*'", "'/'", "'%'", 
                                       "'<'", "'<='", "'>'", "'>='", "'=='", 
                                       "'!='", "'&&'", "'||'", "'|'", "'('", 
                                       "')'", "'()'", "','", "'{'", "':'", 
                                       "'}'", null, null, null, null, null, 
                                       "'='" ];

  MiniLexer.prototype.symbolicNames = [ null, null, null, null, null, null, 
                                        null, null, null, null, null, null, 
                                        null, null, null, null, null, null, 
                                        null, null, null, null, null, null, 
                                        null, null, null, "BOOLEAN", "ID", 
                                        "INT", "FLOAT", "STRING", "EQ", "WS" ];

  MiniLexer.prototype.ruleNames = [ "T__0", "T__1", "T__2", "T__3", "T__4", 
                                    "T__5", "T__6", "T__7", "T__8", "T__9", 
                                    "T__10", "T__11", "T__12", "T__13", "T__14", 
                                    "T__15", "T__16", "T__17", "T__18", "T__19", 
                                    "T__20", "T__21", "T__22", "T__23", "T__24", 
                                    "T__25", "BOOLEAN", "ID", "INT", "FLOAT", 
                                    "STRING", "EQ", "WS" ];

  MiniLexer.prototype.grammarFileName = "Mini.g4";



  var MiniLexer_2 = MiniLexer;

  var MiniLexer_1 = {
  	MiniLexer: MiniLexer_2
  };

  function MiniListener() {
  	_exports.tree.ParseTreeListener.call(this);
  	return this;
  }

  MiniListener.prototype = Object.create(_exports.tree.ParseTreeListener.prototype);
  MiniListener.prototype.constructor = MiniListener;


  MiniListener.prototype.enterMini = function(ctx) {
  };


  MiniListener.prototype.exitMini = function(ctx) {
  };



  MiniListener.prototype.enterMainExpr = function(ctx) {
  };


  MiniListener.prototype.exitMainExpr = function(ctx) {
  };



  MiniListener.prototype.enterExpr = function(ctx) {
  };


  MiniListener.prototype.exitExpr = function(ctx) {
  };



  MiniListener.prototype.enterFunction_call = function(ctx) {
  };


  MiniListener.prototype.exitFunction_call = function(ctx) {
  };



  MiniListener.prototype.enterNumber = function(ctx) {
  };


  MiniListener.prototype.exitNumber = function(ctx) {
  };



  MiniListener.prototype.enterSeq = function(ctx) {
  };


  MiniListener.prototype.exitSeq = function(ctx) {
  };



  MiniListener.prototype.enterId_seq = function(ctx) {
  };


  MiniListener.prototype.exitId_seq = function(ctx) {
  };



  MiniListener.prototype.enterCall_arguments = function(ctx) {
  };


  MiniListener.prototype.exitCall_arguments = function(ctx) {
  };



  MiniListener.prototype.enterPositional_arguments = function(ctx) {
  };


  MiniListener.prototype.exitPositional_arguments = function(ctx) {
  };



  MiniListener.prototype.enterNamed_arguments = function(ctx) {
  };


  MiniListener.prototype.exitNamed_arguments = function(ctx) {
  };



  MiniListener.prototype.enterNamed_argument = function(ctx) {
  };


  MiniListener.prototype.exitNamed_argument = function(ctx) {
  };



  MiniListener.prototype.enterArray = function(ctx) {
  };


  MiniListener.prototype.exitArray = function(ctx) {
  };



  MiniListener.prototype.enterObject = function(ctx) {
  };


  MiniListener.prototype.exitObject = function(ctx) {
  };



  var MiniListener_2 = MiniListener;

  var MiniListener_1 = {
  	MiniListener: MiniListener_2
  };

  var MiniListener$1 = MiniListener_1.MiniListener;

  var serializedATN$1 = ["\u0003\u0430\ud6d1\u8206\uad2d\u4417\uaef1\u8d80\uaadd",
      "\u0003#\u00de\u0004\u0002\t\u0002\u0004\u0003\t\u0003\u0004\u0004\t",
      "\u0004\u0004\u0005\t\u0005\u0004\u0006\t\u0006\u0004\u0007\t\u0007\u0004",
      "\b\t\b\u0004\t\t\t\u0004\n\t\n\u0004\u000b\t\u000b\u0004\f\t\f\u0004",
      "\r\t\r\u0004\u000e\t\u000e\u0003\u0002\u0003\u0002\u0003\u0002\u0003",
      "\u0002\u0003\u0002\u0003\u0002\u0003\u0002\u0003\u0002\u0003\u0002\u0003",
      "\u0002\u0005\u0002\'\n\u0002\u0003\u0003\u0003\u0003\u0003\u0003\u0003",
      "\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003",
      "\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003",
      "\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003",
      "\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003",
      "\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0005\u0004I",
      "\n\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004",
      "\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004",
      "\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004",
      "\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004",
      "\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004",
      "\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004",
      "\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004",
      "\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004\u0003\u0004",
      "\u0003\u0004\u0003\u0004\u0003\u0004\u0007\u0004}\n\u0004\f\u0004\u000e",
      "\u0004\u0080\u000b\u0004\u0003\u0005\u0003\u0005\u0003\u0005\u0003\u0005",
      "\u0003\u0005\u0003\u0005\u0003\u0005\u0003\u0005\u0003\u0005\u0005\u0005",
      "\u008b\n\u0005\u0003\u0006\u0003\u0006\u0003\u0006\u0003\u0006\u0005",
      "\u0006\u0091\n\u0006\u0003\u0007\u0003\u0007\u0003\u0007\u0007\u0007",
      "\u0096\n\u0007\f\u0007\u000e\u0007\u0099\u000b\u0007\u0003\b\u0003\b",
      "\u0003\b\u0007\b\u009e\n\b\f\b\u000e\b\u00a1\u000b\b\u0003\t\u0003\t",
      "\u0003\t\u0003\t\u0003\t\u0003\t\u0003\t\u0005\t\u00aa\n\t\u0003\n\u0003",
      "\n\u0003\n\u0005\n\u00af\n\n\u0007\n\u00b1\n\n\f\n\u000e\n\u00b4\u000b",
      "\n\u0003\u000b\u0003\u000b\u0003\u000b\u0005\u000b\u00b9\n\u000b\u0007",
      "\u000b\u00bb\n\u000b\f\u000b\u000e\u000b\u00be\u000b\u000b\u0003\f\u0003",
      "\f\u0003\f\u0003\f\u0003\f\u0003\r\u0003\r\u0005\r\u00c7\n\r\u0003\r",
      "\u0003\r\u0003\r\u0003\u000e\u0003\u000e\u0003\u000e\u0003\u000e\u0003",
      "\u000e\u0003\u000e\u0003\u000e\u0003\u000e\u0007\u000e\u00d4\n\u000e",
      "\f\u000e\u000e\u000e\u00d7\u000b\u000e\u0005\u000e\u00d9\n\u000e\u0003",
      "\u000e\u0003\u000e\u0003\u000e\u0003\u000e\u0002\u0003\u0006\u000f\u0002",
      "\u0004\u0006\b\n\f\u000e\u0010\u0012\u0014\u0016\u0018\u001a\u0002\u0007",
      "\u0003\u0002\u0006\b\u0003\u0002\n\f\u0003\u0002\u0007\b\u0003\u0002",
      "\r\u0010\u0003\u0002\u0011\u0012\u00f1\u0002&\u0003\u0002\u0002\u0002",
      "\u0004(\u0003\u0002\u0002\u0002\u0006H\u0003\u0002\u0002\u0002\b\u008a",
      "\u0003\u0002\u0002\u0002\n\u0090\u0003\u0002\u0002\u0002\f\u0092\u0003",
      "\u0002\u0002\u0002\u000e\u009a\u0003\u0002\u0002\u0002\u0010\u00a9\u0003",
      "\u0002\u0002\u0002\u0012\u00ab\u0003\u0002\u0002\u0002\u0014\u00b5\u0003",
      "\u0002\u0002\u0002\u0016\u00bf\u0003\u0002\u0002\u0002\u0018\u00c4\u0003",
      "\u0002\u0002\u0002\u001a\u00cb\u0003\u0002\u0002\u0002\u001c\u001d\u0005",
      "\u0004\u0003\u0002\u001d\u001e\u0007\u0002\u0002\u0003\u001e\u001f\b",
      "\u0002\u0001\u0002\u001f\'\u0003\u0002\u0002\u0002 !\u0007\u001e\u0002",
      "\u0002!\"\u0007\"\u0002\u0002\"#\u0005\u0004\u0003\u0002#$\u0007\u0002",
      "\u0002\u0003$%\b\u0002\u0001\u0002%\'\u0003\u0002\u0002\u0002&\u001c",
      "\u0003\u0002\u0002\u0002& \u0003\u0002\u0002\u0002\'\u0003\u0003\u0002",
      "\u0002\u0002()\u0005\u0006\u0004\u0002)*\b\u0003\u0001\u0002*\u0005",
      "\u0003\u0002\u0002\u0002+,\b\u0004\u0001\u0002,-\t\u0002\u0002\u0002",
      "-.\u0005\u0006\u0004\u0013./\b\u0004\u0001\u0002/I\u0003\u0002\u0002",
      "\u000201\u0007\u001d\u0002\u00021I\b\u0004\u0001\u000223\u0005\n\u0006",
      "\u000234\b\u0004\u0001\u00024I\u0003\u0002\u0002\u000256\u0007\u001e",
      "\u0002\u000267\u0006\u0004\u0002\u00027I\b\u0004\u0001\u000289\u0005",
      "\b\u0005\u00029:\b\u0004\u0001\u0002:I\u0003\u0002\u0002\u0002;<\u0007",
      "\u0016\u0002\u0002<=\u0005\u0006\u0004\u0002=>\u0007\u0017\u0002\u0002",
      ">?\b\u0004\u0001\u0002?I\u0003\u0002\u0002\u0002@A\u0005\u0018\r\u0002",
      "AB\b\u0004\u0001\u0002BI\u0003\u0002\u0002\u0002CD\u0005\u001a\u000e",
      "\u0002DE\b\u0004\u0001\u0002EI\u0003\u0002\u0002\u0002FG\u0007!\u0002",
      "\u0002GI\b\u0004\u0001\u0002H+\u0003\u0002\u0002\u0002H0\u0003\u0002",
      "\u0002\u0002H2\u0003\u0002\u0002\u0002H5\u0003\u0002\u0002\u0002H8\u0003",
      "\u0002\u0002\u0002H;\u0003\u0002\u0002\u0002H@\u0003\u0002\u0002\u0002",
      "HC\u0003\u0002\u0002\u0002HF\u0003\u0002\u0002\u0002I~\u0003\u0002\u0002",
      "\u0002JK\f\u0012\u0002\u0002KL\u0007\t\u0002\u0002LM\u0005\u0006\u0004",
      "\u0013MN\b\u0004\u0001\u0002N}\u0003\u0002\u0002\u0002OP\f\u0011\u0002",
      "\u0002PQ\t\u0003\u0002\u0002QR\u0005\u0006\u0004\u0012RS\b\u0004\u0001",
      "\u0002S}\u0003\u0002\u0002\u0002TU\f\u0010\u0002\u0002UV\t\u0004\u0002",
      "\u0002VW\u0005\u0006\u0004\u0011WX\b\u0004\u0001\u0002X}\u0003\u0002",
      "\u0002\u0002YZ\f\u000f\u0002\u0002Z[\t\u0005\u0002\u0002[\\\u0005\u0006",
      "\u0004\u0010\\]\b\u0004\u0001\u0002]}\u0003\u0002\u0002\u0002^_\f\u000e",
      "\u0002\u0002_`\t\u0006\u0002\u0002`a\u0005\u0006\u0004\u000fab\b\u0004",
      "\u0001\u0002b}\u0003\u0002\u0002\u0002cd\f\r\u0002\u0002de\u0007\u0013",
      "\u0002\u0002ef\u0005\u0006\u0004\u000efg\b\u0004\u0001\u0002g}\u0003",
      "\u0002\u0002\u0002hi\f\f\u0002\u0002ij\u0007\u0014\u0002\u0002jk\u0005",
      "\u0006\u0004\rkl\b\u0004\u0001\u0002l}\u0003\u0002\u0002\u0002mn\f\u0015",
      "\u0002\u0002no\u0007\u0003\u0002\u0002op\u0007\u001e\u0002\u0002p}\b",
      "\u0004\u0001\u0002qr\f\u0014\u0002\u0002rs\u0007\u0004\u0002\u0002s",
      "t\u0005\u0006\u0004\u0002tu\u0007\u0005\u0002\u0002uv\b\u0004\u0001",
      "\u0002v}\u0003\u0002\u0002\u0002wx\f\u000b\u0002\u0002xy\u0007\u0015",
      "\u0002\u0002yz\u0005\b\u0005\u0002z{\b\u0004\u0001\u0002{}\u0003\u0002",
      "\u0002\u0002|J\u0003\u0002\u0002\u0002|O\u0003\u0002\u0002\u0002|T\u0003",
      "\u0002\u0002\u0002|Y\u0003\u0002\u0002\u0002|^\u0003\u0002\u0002\u0002",
      "|c\u0003\u0002\u0002\u0002|h\u0003\u0002\u0002\u0002|m\u0003\u0002\u0002",
      "\u0002|q\u0003\u0002\u0002\u0002|w\u0003\u0002\u0002\u0002}\u0080\u0003",
      "\u0002\u0002\u0002~|\u0003\u0002\u0002\u0002~\u007f\u0003\u0002\u0002",
      "\u0002\u007f\u0007\u0003\u0002\u0002\u0002\u0080~\u0003\u0002\u0002",
      "\u0002\u0081\u0082\u0007\u001e\u0002\u0002\u0082\u0083\u0007\u0018\u0002",
      "\u0002\u0083\u008b\b\u0005\u0001\u0002\u0084\u0085\u0007\u001e\u0002",
      "\u0002\u0085\u0086\u0007\u0016\u0002\u0002\u0086\u0087\u0005\u0010\t",
      "\u0002\u0087\u0088\u0007\u0017\u0002\u0002\u0088\u0089\b\u0005\u0001",
      "\u0002\u0089\u008b\u0003\u0002\u0002\u0002\u008a\u0081\u0003\u0002\u0002",
      "\u0002\u008a\u0084\u0003\u0002\u0002\u0002\u008b\t\u0003\u0002\u0002",
      "\u0002\u008c\u008d\u0007\u001f\u0002\u0002\u008d\u0091\b\u0006\u0001",
      "\u0002\u008e\u008f\u0007 \u0002\u0002\u008f\u0091\b\u0006\u0001\u0002",
      "\u0090\u008c\u0003\u0002\u0002\u0002\u0090\u008e\u0003\u0002\u0002\u0002",
      "\u0091\u000b\u0003\u0002\u0002\u0002\u0092\u0097\u0005\u0006\u0004\u0002",
      "\u0093\u0094\u0007\u0019\u0002\u0002\u0094\u0096\u0005\u0006\u0004\u0002",
      "\u0095\u0093\u0003\u0002\u0002\u0002\u0096\u0099\u0003\u0002\u0002\u0002",
      "\u0097\u0095\u0003\u0002\u0002\u0002\u0097\u0098\u0003\u0002\u0002\u0002",
      "\u0098\r\u0003\u0002\u0002\u0002\u0099\u0097\u0003\u0002\u0002\u0002",
      "\u009a\u009f\u0007\u001e\u0002\u0002\u009b\u009c\u0007\u0019\u0002\u0002",
      "\u009c\u009e\u0007\u001e\u0002\u0002\u009d\u009b\u0003\u0002\u0002\u0002",
      "\u009e\u00a1\u0003\u0002\u0002\u0002\u009f\u009d\u0003\u0002\u0002\u0002",
      "\u009f\u00a0\u0003\u0002\u0002\u0002\u00a0\u000f\u0003\u0002\u0002\u0002",
      "\u00a1\u009f\u0003\u0002\u0002\u0002\u00a2\u00aa\u0003\u0002\u0002\u0002",
      "\u00a3\u00aa\u0005\u0012\n\u0002\u00a4\u00aa\u0005\u0014\u000b\u0002",
      "\u00a5\u00a6\u0005\u0012\n\u0002\u00a6\u00a7\u0007\u0019\u0002\u0002",
      "\u00a7\u00a8\u0005\u0014\u000b\u0002\u00a8\u00aa\u0003\u0002\u0002\u0002",
      "\u00a9\u00a2\u0003\u0002\u0002\u0002\u00a9\u00a3\u0003\u0002\u0002\u0002",
      "\u00a9\u00a4\u0003\u0002\u0002\u0002\u00a9\u00a5\u0003\u0002\u0002\u0002",
      "\u00aa\u0011\u0003\u0002\u0002\u0002\u00ab\u00b2\u0005\u0006\u0004\u0002",
      "\u00ac\u00ae\u0007\u0019\u0002\u0002\u00ad\u00af\u0005\u0006\u0004\u0002",
      "\u00ae\u00ad\u0003\u0002\u0002\u0002\u00ae\u00af\u0003\u0002\u0002\u0002",
      "\u00af\u00b1\u0003\u0002\u0002\u0002\u00b0\u00ac\u0003\u0002\u0002\u0002",
      "\u00b1\u00b4\u0003\u0002\u0002\u0002\u00b2\u00b0\u0003\u0002\u0002\u0002",
      "\u00b2\u00b3\u0003\u0002\u0002\u0002\u00b3\u0013\u0003\u0002\u0002\u0002",
      "\u00b4\u00b2\u0003\u0002\u0002\u0002\u00b5\u00bc\u0005\u0016\f\u0002",
      "\u00b6\u00b8\u0007\u0019\u0002\u0002\u00b7\u00b9\u0005\u0016\f\u0002",
      "\u00b8\u00b7\u0003\u0002\u0002\u0002\u00b8\u00b9\u0003\u0002\u0002\u0002",
      "\u00b9\u00bb\u0003\u0002\u0002\u0002\u00ba\u00b6\u0003\u0002\u0002\u0002",
      "\u00bb\u00be\u0003\u0002\u0002\u0002\u00bc\u00ba\u0003\u0002\u0002\u0002",
      "\u00bc\u00bd\u0003\u0002\u0002\u0002\u00bd\u0015\u0003\u0002\u0002\u0002",
      "\u00be\u00bc\u0003\u0002\u0002\u0002\u00bf\u00c0\u0007\u001e\u0002\u0002",
      "\u00c0\u00c1\u0007\"\u0002\u0002\u00c1\u00c2\u0005\u0006\u0004\u0002",
      "\u00c2\u00c3\b\f\u0001\u0002\u00c3\u0017\u0003\u0002\u0002\u0002\u00c4",
      "\u00c6\u0007\u0004\u0002\u0002\u00c5\u00c7\u0005\f\u0007\u0002\u00c6",
      "\u00c5\u0003\u0002\u0002\u0002\u00c6\u00c7\u0003\u0002\u0002\u0002\u00c7",
      "\u00c8\u0003\u0002\u0002\u0002\u00c8\u00c9\u0007\u0005\u0002\u0002\u00c9",
      "\u00ca\b\r\u0001\u0002\u00ca\u0019\u0003\u0002\u0002\u0002\u00cb\u00d8",
      "\u0007\u001a\u0002\u0002\u00cc\u00cd\u0007\u001e\u0002\u0002\u00cd\u00ce",
      "\u0007\u001b\u0002\u0002\u00ce\u00d5\u0005\u0006\u0004\u0002\u00cf\u00d0",
      "\u0007\u0019\u0002\u0002\u00d0\u00d1\u0007\u001e\u0002\u0002\u00d1\u00d2",
      "\u0007\u001b\u0002\u0002\u00d2\u00d4\u0005\u0006\u0004\u0002\u00d3\u00cf",
      "\u0003\u0002\u0002\u0002\u00d4\u00d7\u0003\u0002\u0002\u0002\u00d5\u00d3",
      "\u0003\u0002\u0002\u0002\u00d5\u00d6\u0003\u0002\u0002\u0002\u00d6\u00d9",
      "\u0003\u0002\u0002\u0002\u00d7\u00d5\u0003\u0002\u0002\u0002\u00d8\u00cc",
      "\u0003\u0002\u0002\u0002\u00d8\u00d9\u0003\u0002\u0002\u0002\u00d9\u00da",
      "\u0003\u0002\u0002\u0002\u00da\u00db\u0007\u001c\u0002\u0002\u00db\u00dc",
      "\b\u000e\u0001\u0002\u00dc\u001b\u0003\u0002\u0002\u0002\u0012&H|~\u008a",
      "\u0090\u0097\u009f\u00a9\u00ae\u00b2\u00b8\u00bc\u00c6\u00d5\u00d8"].join("");


  var atn$1 = new _exports.atn.ATNDeserializer().deserialize(serializedATN$1);

  var decisionsToDFA$1 = atn$1.decisionToState.map( function(ds, index) { return new _exports.dfa.DFA(ds, index); });

  var sharedContextCache = new _exports.PredictionContextCache();

  var literalNames = [ null, "'.'", "'['", "']'", "'!'", "'+'", "'-'", "'^'", 
                       "'*'", "'/'", "'%'", "'<'", "'<='", "'>'", "'>='", 
                       "'=='", "'!='", "'&&'", "'||'", "'|'", "'('", "')'", 
                       "'()'", "','", "'{'", "':'", "'}'", null, null, null, 
                       null, null, "'='" ];

  var symbolicNames = [ null, null, null, null, null, null, null, null, null, 
                        null, null, null, null, null, null, null, null, null, 
                        null, null, null, null, null, null, null, null, null, 
                        "BOOLEAN", "ID", "INT", "FLOAT", "STRING", "EQ", "WS" ];

  var ruleNames =  [ "mini", "mainExpr", "expr", "function_call", "number", 
                     "seq", "id_seq", "call_arguments", "positional_arguments", 
                     "named_arguments", "named_argument", "array", "object" ];

  function MiniParser (input) {
  	_exports.Parser.call(this, input);
      this._interp = new _exports.atn.ParserATNSimulator(this, atn$1, decisionsToDFA$1, sharedContextCache);
      this.ruleNames = ruleNames;
      this.literalNames = literalNames;
      this.symbolicNames = symbolicNames;
      return this;
  }

  MiniParser.prototype = Object.create(_exports.Parser.prototype);
  MiniParser.prototype.constructor = MiniParser;

  Object.defineProperty(MiniParser.prototype, "atn", {
  	get : function() {
  		return atn$1;
  	}
  });

  MiniParser.EOF = _exports.Token.EOF;
  MiniParser.T__0 = 1;
  MiniParser.T__1 = 2;
  MiniParser.T__2 = 3;
  MiniParser.T__3 = 4;
  MiniParser.T__4 = 5;
  MiniParser.T__5 = 6;
  MiniParser.T__6 = 7;
  MiniParser.T__7 = 8;
  MiniParser.T__8 = 9;
  MiniParser.T__9 = 10;
  MiniParser.T__10 = 11;
  MiniParser.T__11 = 12;
  MiniParser.T__12 = 13;
  MiniParser.T__13 = 14;
  MiniParser.T__14 = 15;
  MiniParser.T__15 = 16;
  MiniParser.T__16 = 17;
  MiniParser.T__17 = 18;
  MiniParser.T__18 = 19;
  MiniParser.T__19 = 20;
  MiniParser.T__20 = 21;
  MiniParser.T__21 = 22;
  MiniParser.T__22 = 23;
  MiniParser.T__23 = 24;
  MiniParser.T__24 = 25;
  MiniParser.T__25 = 26;
  MiniParser.BOOLEAN = 27;
  MiniParser.ID = 28;
  MiniParser.INT = 29;
  MiniParser.FLOAT = 30;
  MiniParser.STRING = 31;
  MiniParser.EQ = 32;
  MiniParser.WS = 33;

  MiniParser.RULE_mini = 0;
  MiniParser.RULE_mainExpr = 1;
  MiniParser.RULE_expr = 2;
  MiniParser.RULE_function_call = 3;
  MiniParser.RULE_number = 4;
  MiniParser.RULE_seq = 5;
  MiniParser.RULE_id_seq = 6;
  MiniParser.RULE_call_arguments = 7;
  MiniParser.RULE_positional_arguments = 8;
  MiniParser.RULE_named_arguments = 9;
  MiniParser.RULE_named_argument = 10;
  MiniParser.RULE_array = 11;
  MiniParser.RULE_object = 12;

  function MiniContext(parser, parent, invokingState) {
  	if(parent===undefined) {
  	    parent = null;
  	}
  	if(invokingState===undefined || invokingState===null) {
  		invokingState = -1;
  	}
  	_exports.ParserRuleContext.call(this, parent, invokingState);
      this.parser = parser;
      this.ruleIndex = MiniParser.RULE_mini;
      return this;
  }

  MiniContext.prototype = Object.create(_exports.ParserRuleContext.prototype);
  MiniContext.prototype.constructor = MiniContext;

  MiniContext.prototype.mainExpr = function() {
      return this.getTypedRuleContext(MainExprContext,0);
  };

  MiniContext.prototype.EOF = function() {
      return this.getToken(MiniParser.EOF, 0);
  };

  MiniContext.prototype.ID = function() {
      return this.getToken(MiniParser.ID, 0);
  };

  MiniContext.prototype.EQ = function() {
      return this.getToken(MiniParser.EQ, 0);
  };

  MiniContext.prototype.enterRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.enterMini(this);
  	}
  };

  MiniContext.prototype.exitRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.exitMini(this);
  	}
  };




  MiniParser.MiniContext = MiniContext;

  MiniParser.prototype.mini = function() {

      var localctx = new MiniContext(this, this._ctx, this.state);
      this.enterRule(localctx, 0, MiniParser.RULE_mini);
      try {
          this.state = 36;
          this._errHandler.sync(this);
          var la_ = this._interp.adaptivePredict(this._input,0,this._ctx);
          switch(la_) {
          case 1:
              this.enterOuterAlt(localctx, 1);
              this.state = 26;
              this.mainExpr();
              this.state = 27;
              this.match(MiniParser.EOF);
               localctx.type = 'evaluation'; 
              break;

          case 2:
              this.enterOuterAlt(localctx, 2);
              this.state = 30;
              this.match(MiniParser.ID);
              this.state = 31;
              this.match(MiniParser.EQ);
              this.state = 32;
              this.mainExpr();
              this.state = 33;
              this.match(MiniParser.EOF);
               localctx.type = 'definition';
              break;

          }
      } catch (re) {
      	if(re instanceof _exports.error.RecognitionException) {
  	        localctx.exception = re;
  	        this._errHandler.reportError(this, re);
  	        this._errHandler.recover(this, re);
  	    } else {
  	    	throw re;
  	    }
      } finally {
          this.exitRule();
      }
      return localctx;
  };

  function MainExprContext(parser, parent, invokingState) {
  	if(parent===undefined) {
  	    parent = null;
  	}
  	if(invokingState===undefined || invokingState===null) {
  		invokingState = -1;
  	}
  	_exports.ParserRuleContext.call(this, parent, invokingState);
      this.parser = parser;
      this.ruleIndex = MiniParser.RULE_mainExpr;
      return this;
  }

  MainExprContext.prototype = Object.create(_exports.ParserRuleContext.prototype);
  MainExprContext.prototype.constructor = MainExprContext;

  MainExprContext.prototype.expr = function() {
      return this.getTypedRuleContext(ExprContext,0);
  };

  MainExprContext.prototype.enterRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.enterMainExpr(this);
  	}
  };

  MainExprContext.prototype.exitRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.exitMainExpr(this);
  	}
  };




  MiniParser.MainExprContext = MainExprContext;

  MiniParser.prototype.mainExpr = function() {

      var localctx = new MainExprContext(this, this._ctx, this.state);
      this.enterRule(localctx, 2, MiniParser.RULE_mainExpr);
      try {
          this.enterOuterAlt(localctx, 1);
          this.state = 38;
          this.expr(0);
           localctx.type = 'simple'; 
      } catch (re) {
      	if(re instanceof _exports.error.RecognitionException) {
  	        localctx.exception = re;
  	        this._errHandler.reportError(this, re);
  	        this._errHandler.recover(this, re);
  	    } else {
  	    	throw re;
  	    }
      } finally {
          this.exitRule();
      }
      return localctx;
  };

  function ExprContext(parser, parent, invokingState) {
  	if(parent===undefined) {
  	    parent = null;
  	}
  	if(invokingState===undefined || invokingState===null) {
  		invokingState = -1;
  	}
  	_exports.ParserRuleContext.call(this, parent, invokingState);
      this.parser = parser;
      this.ruleIndex = MiniParser.RULE_expr;
      this.op = null; 
      return this;
  }

  ExprContext.prototype = Object.create(_exports.ParserRuleContext.prototype);
  ExprContext.prototype.constructor = ExprContext;

  ExprContext.prototype.expr = function(i) {
      if(i===undefined) {
          i = null;
      }
      if(i===null) {
          return this.getTypedRuleContexts(ExprContext);
      } else {
          return this.getTypedRuleContext(ExprContext,i);
      }
  };

  ExprContext.prototype.BOOLEAN = function() {
      return this.getToken(MiniParser.BOOLEAN, 0);
  };

  ExprContext.prototype.number = function() {
      return this.getTypedRuleContext(NumberContext,0);
  };

  ExprContext.prototype.ID = function() {
      return this.getToken(MiniParser.ID, 0);
  };

  ExprContext.prototype.function_call = function() {
      return this.getTypedRuleContext(Function_callContext,0);
  };

  ExprContext.prototype.array = function() {
      return this.getTypedRuleContext(ArrayContext,0);
  };

  ExprContext.prototype.object = function() {
      return this.getTypedRuleContext(ObjectContext,0);
  };

  ExprContext.prototype.STRING = function() {
      return this.getToken(MiniParser.STRING, 0);
  };

  ExprContext.prototype.enterRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.enterExpr(this);
  	}
  };

  ExprContext.prototype.exitRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.exitExpr(this);
  	}
  };



  MiniParser.prototype.expr = function(_p) {
  	if(_p===undefined) {
  	    _p = 0;
  	}
      var _parentctx = this._ctx;
      var _parentState = this.state;
      var localctx = new ExprContext(this, this._ctx, _parentState);
      var _startState = 4;
      this.enterRecursionRule(localctx, 4, MiniParser.RULE_expr, _p);
      var _la = 0; 
      try {
          this.enterOuterAlt(localctx, 1);
          this.state = 70;
          this._errHandler.sync(this);
          var la_ = this._interp.adaptivePredict(this._input,1,this._ctx);
          switch(la_) {
          case 1:
              this.state = 42;
              localctx.op = this._input.LT(1);
              _la = this._input.LA(1);
              if(!((((_la) & ~0x1f) == 0 && ((1 << _la) & ((1 << MiniParser.T__3) | (1 << MiniParser.T__4) | (1 << MiniParser.T__5))) !== 0))) {
                  localctx.op = this._errHandler.recoverInline(this);
              }
              else {
              	this._errHandler.reportMatch(this);
                  this.consume();
              }
              this.state = 43;
              this.expr(17);
               switch (localctx.op.text) {
                          case '!': localctx.type = 'not';               break
                          case '+': localctx.type = 'positive';          break
                          case '-': localctx.type = 'negative';          break
                      }
              break;

          case 2:
              this.state = 46;
              this.match(MiniParser.BOOLEAN);
               localctx.type = 'boolean'; 
              break;

          case 3:
              this.state = 48;
              this.number();
               localctx.type = 'number'; 
              break;

          case 4:
              this.state = 51;
              this.match(MiniParser.ID);
              this.state = 52;
              if (!( this._input.LA(1) !== MiniParser.EQ)) {
                  throw new _exports.error.FailedPredicateException(this, "this._input.LA(1) !== MiniParser.EQ");
              }
               localctx.type = 'var'; 
              break;

          case 5:
              this.state = 54;
              this.function_call();
               localctx.type = '_call'; 
              break;

          case 6:
              this.state = 57;
              this.match(MiniParser.T__19);
              this.state = 58;
              this.expr(0);
              this.state = 59;
              this.match(MiniParser.T__20);
               localctx.type = 'group'; 
              break;

          case 7:
              this.state = 62;
              this.array();
               localctx.type = 'array'; 
              break;

          case 8:
              this.state = 65;
              this.object();
               localctx.type = 'object'; 
              break;

          case 9:
              this.state = 68;
              this.match(MiniParser.STRING);
               localctx.type = 'string'; 
              break;

          }
          this._ctx.stop = this._input.LT(-1);
          this.state = 124;
          this._errHandler.sync(this);
          var _alt = this._interp.adaptivePredict(this._input,3,this._ctx);
          while(_alt!=2 && _alt!=_exports.atn.ATN.INVALID_ALT_NUMBER) {
              if(_alt===1) {
                  if(this._parseListeners!==null) {
                      this.triggerExitRuleEvent();
                  }
                  this.state = 122;
                  this._errHandler.sync(this);
                  var la_ = this._interp.adaptivePredict(this._input,2,this._ctx);
                  switch(la_) {
                  case 1:
                      localctx = new ExprContext(this, _parentctx, _parentState);
                      this.pushNewRecursionContext(localctx, _startState, MiniParser.RULE_expr);
                      this.state = 72;
                      if (!( this.precpred(this._ctx, 16))) {
                          throw new _exports.error.FailedPredicateException(this, "this.precpred(this._ctx, 16)");
                      }
                      this.state = 73;
                      this.match(MiniParser.T__6);
                      this.state = 74;
                      this.expr(17);

                                            localctx.type = 'pow';
                                        
                      break;

                  case 2:
                      localctx = new ExprContext(this, _parentctx, _parentState);
                      this.pushNewRecursionContext(localctx, _startState, MiniParser.RULE_expr);
                      this.state = 77;
                      if (!( this.precpred(this._ctx, 15))) {
                          throw new _exports.error.FailedPredicateException(this, "this.precpred(this._ctx, 15)");
                      }
                      this.state = 78;
                      localctx.op = this._input.LT(1);
                      _la = this._input.LA(1);
                      if(!((((_la) & ~0x1f) == 0 && ((1 << _la) & ((1 << MiniParser.T__7) | (1 << MiniParser.T__8) | (1 << MiniParser.T__9))) !== 0))) {
                          localctx.op = this._errHandler.recoverInline(this);
                      }
                      else {
                      	this._errHandler.reportMatch(this);
                          this.consume();
                      }
                      this.state = 79;
                      this.expr(16);
                       switch (localctx.op.text) {
                                            case '*': localctx.type = 'multiply';          break
                                            case '/': localctx.type = 'divide';            break
                                            case '%': localctx.type = 'remainder';         break
                                        }
                      break;

                  case 3:
                      localctx = new ExprContext(this, _parentctx, _parentState);
                      this.pushNewRecursionContext(localctx, _startState, MiniParser.RULE_expr);
                      this.state = 82;
                      if (!( this.precpred(this._ctx, 14))) {
                          throw new _exports.error.FailedPredicateException(this, "this.precpred(this._ctx, 14)");
                      }
                      this.state = 83;
                      localctx.op = this._input.LT(1);
                      _la = this._input.LA(1);
                      if(!(_la===MiniParser.T__4 || _la===MiniParser.T__5)) {
                          localctx.op = this._errHandler.recoverInline(this);
                      }
                      else {
                      	this._errHandler.reportMatch(this);
                          this.consume();
                      }
                      this.state = 84;
                      this.expr(15);

                                            localctx.type = (localctx.op.text === '+') ? 'add' : 'subtract';
                                        
                      break;

                  case 4:
                      localctx = new ExprContext(this, _parentctx, _parentState);
                      this.pushNewRecursionContext(localctx, _startState, MiniParser.RULE_expr);
                      this.state = 87;
                      if (!( this.precpred(this._ctx, 13))) {
                          throw new _exports.error.FailedPredicateException(this, "this.precpred(this._ctx, 13)");
                      }
                      this.state = 88;
                      localctx.op = this._input.LT(1);
                      _la = this._input.LA(1);
                      if(!((((_la) & ~0x1f) == 0 && ((1 << _la) & ((1 << MiniParser.T__10) | (1 << MiniParser.T__11) | (1 << MiniParser.T__12) | (1 << MiniParser.T__13))) !== 0))) {
                          localctx.op = this._errHandler.recoverInline(this);
                      }
                      else {
                      	this._errHandler.reportMatch(this);
                          this.consume();
                      }
                      this.state = 89;
                      this.expr(14);
                       switch (localctx.op.text) {
                                            case '<':  localctx.type = 'less';             break
                                            case '<=': localctx.type = 'less_or_equal';    break
                                            case '>':  localctx.type = 'greater';          break
                                            case '>=': localctx.type = 'greater_or_equal'; break
                                        }
                      break;

                  case 5:
                      localctx = new ExprContext(this, _parentctx, _parentState);
                      this.pushNewRecursionContext(localctx, _startState, MiniParser.RULE_expr);
                      this.state = 92;
                      if (!( this.precpred(this._ctx, 12))) {
                          throw new _exports.error.FailedPredicateException(this, "this.precpred(this._ctx, 12)");
                      }
                      this.state = 93;
                      localctx.op = this._input.LT(1);
                      _la = this._input.LA(1);
                      if(!(_la===MiniParser.T__14 || _la===MiniParser.T__15)) {
                          localctx.op = this._errHandler.recoverInline(this);
                      }
                      else {
                      	this._errHandler.reportMatch(this);
                          this.consume();
                      }
                      this.state = 94;
                      this.expr(13);

                                            localctx.type = (localctx.op.text === '==') ? 'equal' : 'not_equal';
                                        
                      break;

                  case 6:
                      localctx = new ExprContext(this, _parentctx, _parentState);
                      this.pushNewRecursionContext(localctx, _startState, MiniParser.RULE_expr);
                      this.state = 97;
                      if (!( this.precpred(this._ctx, 11))) {
                          throw new _exports.error.FailedPredicateException(this, "this.precpred(this._ctx, 11)");
                      }
                      this.state = 98;
                      this.match(MiniParser.T__16);
                      this.state = 99;
                      this.expr(12);
                       localctx.type = 'and'; 
                      break;

                  case 7:
                      localctx = new ExprContext(this, _parentctx, _parentState);
                      this.pushNewRecursionContext(localctx, _startState, MiniParser.RULE_expr);
                      this.state = 102;
                      if (!( this.precpred(this._ctx, 10))) {
                          throw new _exports.error.FailedPredicateException(this, "this.precpred(this._ctx, 10)");
                      }
                      this.state = 103;
                      this.match(MiniParser.T__17);
                      this.state = 104;
                      this.expr(11);
                       localctx.type = 'or'; 
                      break;

                  case 8:
                      localctx = new ExprContext(this, _parentctx, _parentState);
                      this.pushNewRecursionContext(localctx, _startState, MiniParser.RULE_expr);
                      this.state = 107;
                      if (!( this.precpred(this._ctx, 19))) {
                          throw new _exports.error.FailedPredicateException(this, "this.precpred(this._ctx, 19)");
                      }
                      this.state = 108;
                      this.match(MiniParser.T__0);
                      this.state = 109;
                      this.match(MiniParser.ID);

                                            localctx.type = 'select_id';
                                        
                      break;

                  case 9:
                      localctx = new ExprContext(this, _parentctx, _parentState);
                      this.pushNewRecursionContext(localctx, _startState, MiniParser.RULE_expr);
                      this.state = 111;
                      if (!( this.precpred(this._ctx, 18))) {
                          throw new _exports.error.FailedPredicateException(this, "this.precpred(this._ctx, 18)");
                      }
                      this.state = 112;
                      this.match(MiniParser.T__1);
                      this.state = 113;
                      this.expr(0);
                      this.state = 114;
                      this.match(MiniParser.T__2);

                                            localctx.type = 'select_expr';
                                        
                      break;

                  case 10:
                      localctx = new ExprContext(this, _parentctx, _parentState);
                      this.pushNewRecursionContext(localctx, _startState, MiniParser.RULE_expr);
                      this.state = 117;
                      if (!( this.precpred(this._ctx, 9))) {
                          throw new _exports.error.FailedPredicateException(this, "this.precpred(this._ctx, 9)");
                      }
                      this.state = 118;
                      this.match(MiniParser.T__18);
                      this.state = 119;
                      this.function_call();
                       localctx.type = 'pipe'; 
                      break;

                  } 
              }
              this.state = 126;
              this._errHandler.sync(this);
              _alt = this._interp.adaptivePredict(this._input,3,this._ctx);
          }

      } catch( error) {
          if(error instanceof _exports.error.RecognitionException) {
  	        localctx.exception = error;
  	        this._errHandler.reportError(this, error);
  	        this._errHandler.recover(this, error);
  	    } else {
  	    	throw error;
  	    }
      } finally {
          this.unrollRecursionContexts(_parentctx);
      }
      return localctx;
  };

  function Function_callContext(parser, parent, invokingState) {
  	if(parent===undefined) {
  	    parent = null;
  	}
  	if(invokingState===undefined || invokingState===null) {
  		invokingState = -1;
  	}
  	_exports.ParserRuleContext.call(this, parent, invokingState);
      this.parser = parser;
      this.ruleIndex = MiniParser.RULE_function_call;
      this.name = null; 
      this.args = null; 
      return this;
  }

  Function_callContext.prototype = Object.create(_exports.ParserRuleContext.prototype);
  Function_callContext.prototype.constructor = Function_callContext;

  Function_callContext.prototype.ID = function() {
      return this.getToken(MiniParser.ID, 0);
  };

  Function_callContext.prototype.call_arguments = function() {
      return this.getTypedRuleContext(Call_argumentsContext,0);
  };

  Function_callContext.prototype.enterRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.enterFunction_call(this);
  	}
  };

  Function_callContext.prototype.exitRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.exitFunction_call(this);
  	}
  };




  MiniParser.Function_callContext = Function_callContext;

  MiniParser.prototype.function_call = function() {

      var localctx = new Function_callContext(this, this._ctx, this.state);
      this.enterRule(localctx, 6, MiniParser.RULE_function_call);
      try {
          this.state = 136;
          this._errHandler.sync(this);
          var la_ = this._interp.adaptivePredict(this._input,4,this._ctx);
          switch(la_) {
          case 1:
              this.enterOuterAlt(localctx, 1);
              this.state = 127;
              localctx.name = this.match(MiniParser.ID);
              this.state = 128;
              this.match(MiniParser.T__21);
               localctx.type = 'call'; 
              break;

          case 2:
              this.enterOuterAlt(localctx, 2);
              this.state = 130;
              localctx.name = this.match(MiniParser.ID);
              this.state = 131;
              this.match(MiniParser.T__19);
              this.state = 132;
              localctx.args = this.call_arguments();
              this.state = 133;
              this.match(MiniParser.T__20);
               localctx.type = 'call'; 
              break;

          }
      } catch (re) {
      	if(re instanceof _exports.error.RecognitionException) {
  	        localctx.exception = re;
  	        this._errHandler.reportError(this, re);
  	        this._errHandler.recover(this, re);
  	    } else {
  	    	throw re;
  	    }
      } finally {
          this.exitRule();
      }
      return localctx;
  };

  function NumberContext(parser, parent, invokingState) {
  	if(parent===undefined) {
  	    parent = null;
  	}
  	if(invokingState===undefined || invokingState===null) {
  		invokingState = -1;
  	}
  	_exports.ParserRuleContext.call(this, parent, invokingState);
      this.parser = parser;
      this.ruleIndex = MiniParser.RULE_number;
      return this;
  }

  NumberContext.prototype = Object.create(_exports.ParserRuleContext.prototype);
  NumberContext.prototype.constructor = NumberContext;

  NumberContext.prototype.INT = function() {
      return this.getToken(MiniParser.INT, 0);
  };

  NumberContext.prototype.FLOAT = function() {
      return this.getToken(MiniParser.FLOAT, 0);
  };

  NumberContext.prototype.enterRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.enterNumber(this);
  	}
  };

  NumberContext.prototype.exitRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.exitNumber(this);
  	}
  };




  MiniParser.NumberContext = NumberContext;

  MiniParser.prototype.number = function() {

      var localctx = new NumberContext(this, this._ctx, this.state);
      this.enterRule(localctx, 8, MiniParser.RULE_number);
      try {
          this.state = 142;
          this._errHandler.sync(this);
          switch(this._input.LA(1)) {
          case MiniParser.INT:
              this.enterOuterAlt(localctx, 1);
              this.state = 138;
              this.match(MiniParser.INT);
               localctx.type = 'int'; 
              break;
          case MiniParser.FLOAT:
              this.enterOuterAlt(localctx, 2);
              this.state = 140;
              this.match(MiniParser.FLOAT);
               localctx.type = 'float'; 
              break;
          default:
              throw new _exports.error.NoViableAltException(this);
          }
      } catch (re) {
      	if(re instanceof _exports.error.RecognitionException) {
  	        localctx.exception = re;
  	        this._errHandler.reportError(this, re);
  	        this._errHandler.recover(this, re);
  	    } else {
  	    	throw re;
  	    }
      } finally {
          this.exitRule();
      }
      return localctx;
  };

  function SeqContext(parser, parent, invokingState) {
  	if(parent===undefined) {
  	    parent = null;
  	}
  	if(invokingState===undefined || invokingState===null) {
  		invokingState = -1;
  	}
  	_exports.ParserRuleContext.call(this, parent, invokingState);
      this.parser = parser;
      this.ruleIndex = MiniParser.RULE_seq;
      this._expr = null; 
      this.items = []; 
      return this;
  }

  SeqContext.prototype = Object.create(_exports.ParserRuleContext.prototype);
  SeqContext.prototype.constructor = SeqContext;

  SeqContext.prototype.expr = function(i) {
      if(i===undefined) {
          i = null;
      }
      if(i===null) {
          return this.getTypedRuleContexts(ExprContext);
      } else {
          return this.getTypedRuleContext(ExprContext,i);
      }
  };

  SeqContext.prototype.enterRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.enterSeq(this);
  	}
  };

  SeqContext.prototype.exitRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.exitSeq(this);
  	}
  };




  MiniParser.SeqContext = SeqContext;

  MiniParser.prototype.seq = function() {

      var localctx = new SeqContext(this, this._ctx, this.state);
      this.enterRule(localctx, 10, MiniParser.RULE_seq);
      var _la = 0; 
      try {
          this.enterOuterAlt(localctx, 1);
          this.state = 144;
          localctx._expr = this.expr(0);
          localctx.items.push(localctx._expr);
          this.state = 149;
          this._errHandler.sync(this);
          _la = this._input.LA(1);
          while(_la===MiniParser.T__22) {
              this.state = 145;
              this.match(MiniParser.T__22);
              this.state = 146;
              localctx._expr = this.expr(0);
              localctx.items.push(localctx._expr);
              this.state = 151;
              this._errHandler.sync(this);
              _la = this._input.LA(1);
          }
      } catch (re) {
      	if(re instanceof _exports.error.RecognitionException) {
  	        localctx.exception = re;
  	        this._errHandler.reportError(this, re);
  	        this._errHandler.recover(this, re);
  	    } else {
  	    	throw re;
  	    }
      } finally {
          this.exitRule();
      }
      return localctx;
  };

  function Id_seqContext(parser, parent, invokingState) {
  	if(parent===undefined) {
  	    parent = null;
  	}
  	if(invokingState===undefined || invokingState===null) {
  		invokingState = -1;
  	}
  	_exports.ParserRuleContext.call(this, parent, invokingState);
      this.parser = parser;
      this.ruleIndex = MiniParser.RULE_id_seq;
      this._ID = null; 
      this.items = []; 
      return this;
  }

  Id_seqContext.prototype = Object.create(_exports.ParserRuleContext.prototype);
  Id_seqContext.prototype.constructor = Id_seqContext;

  Id_seqContext.prototype.ID = function(i) {
  	if(i===undefined) {
  		i = null;
  	}
      if(i===null) {
          return this.getTokens(MiniParser.ID);
      } else {
          return this.getToken(MiniParser.ID, i);
      }
  };


  Id_seqContext.prototype.enterRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.enterId_seq(this);
  	}
  };

  Id_seqContext.prototype.exitRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.exitId_seq(this);
  	}
  };




  MiniParser.Id_seqContext = Id_seqContext;

  MiniParser.prototype.id_seq = function() {

      var localctx = new Id_seqContext(this, this._ctx, this.state);
      this.enterRule(localctx, 12, MiniParser.RULE_id_seq);
      var _la = 0; 
      try {
          this.enterOuterAlt(localctx, 1);
          this.state = 152;
          localctx._ID = this.match(MiniParser.ID);
          localctx.items.push(localctx._ID);
          this.state = 157;
          this._errHandler.sync(this);
          _la = this._input.LA(1);
          while(_la===MiniParser.T__22) {
              this.state = 153;
              this.match(MiniParser.T__22);
              this.state = 154;
              localctx._ID = this.match(MiniParser.ID);
              localctx.items.push(localctx._ID);
              this.state = 159;
              this._errHandler.sync(this);
              _la = this._input.LA(1);
          }
      } catch (re) {
      	if(re instanceof _exports.error.RecognitionException) {
  	        localctx.exception = re;
  	        this._errHandler.reportError(this, re);
  	        this._errHandler.recover(this, re);
  	    } else {
  	    	throw re;
  	    }
      } finally {
          this.exitRule();
      }
      return localctx;
  };

  function Call_argumentsContext(parser, parent, invokingState) {
  	if(parent===undefined) {
  	    parent = null;
  	}
  	if(invokingState===undefined || invokingState===null) {
  		invokingState = -1;
  	}
  	_exports.ParserRuleContext.call(this, parent, invokingState);
      this.parser = parser;
      this.ruleIndex = MiniParser.RULE_call_arguments;
      this.args = null; 
      this.namedArgs = null; 
      return this;
  }

  Call_argumentsContext.prototype = Object.create(_exports.ParserRuleContext.prototype);
  Call_argumentsContext.prototype.constructor = Call_argumentsContext;

  Call_argumentsContext.prototype.positional_arguments = function() {
      return this.getTypedRuleContext(Positional_argumentsContext,0);
  };

  Call_argumentsContext.prototype.named_arguments = function() {
      return this.getTypedRuleContext(Named_argumentsContext,0);
  };

  Call_argumentsContext.prototype.enterRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.enterCall_arguments(this);
  	}
  };

  Call_argumentsContext.prototype.exitRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.exitCall_arguments(this);
  	}
  };




  MiniParser.Call_argumentsContext = Call_argumentsContext;

  MiniParser.prototype.call_arguments = function() {

      var localctx = new Call_argumentsContext(this, this._ctx, this.state);
      this.enterRule(localctx, 14, MiniParser.RULE_call_arguments);
      try {
          this.state = 167;
          this._errHandler.sync(this);
          var la_ = this._interp.adaptivePredict(this._input,8,this._ctx);
          switch(la_) {
          case 1:
              this.enterOuterAlt(localctx, 1);

              break;

          case 2:
              this.enterOuterAlt(localctx, 2);
              this.state = 161;
              localctx.args = this.positional_arguments();
              break;

          case 3:
              this.enterOuterAlt(localctx, 3);
              this.state = 162;
              localctx.namedArgs = this.named_arguments();
              break;

          case 4:
              this.enterOuterAlt(localctx, 4);
              this.state = 163;
              localctx.args = this.positional_arguments();
              this.state = 164;
              this.match(MiniParser.T__22);
              this.state = 165;
              localctx.namedArgs = this.named_arguments();
              break;

          }
      } catch (re) {
      	if(re instanceof _exports.error.RecognitionException) {
  	        localctx.exception = re;
  	        this._errHandler.reportError(this, re);
  	        this._errHandler.recover(this, re);
  	    } else {
  	    	throw re;
  	    }
      } finally {
          this.exitRule();
      }
      return localctx;
  };

  function Positional_argumentsContext(parser, parent, invokingState) {
  	if(parent===undefined) {
  	    parent = null;
  	}
  	if(invokingState===undefined || invokingState===null) {
  		invokingState = -1;
  	}
  	_exports.ParserRuleContext.call(this, parent, invokingState);
      this.parser = parser;
      this.ruleIndex = MiniParser.RULE_positional_arguments;
      return this;
  }

  Positional_argumentsContext.prototype = Object.create(_exports.ParserRuleContext.prototype);
  Positional_argumentsContext.prototype.constructor = Positional_argumentsContext;

  Positional_argumentsContext.prototype.expr = function(i) {
      if(i===undefined) {
          i = null;
      }
      if(i===null) {
          return this.getTypedRuleContexts(ExprContext);
      } else {
          return this.getTypedRuleContext(ExprContext,i);
      }
  };

  Positional_argumentsContext.prototype.enterRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.enterPositional_arguments(this);
  	}
  };

  Positional_argumentsContext.prototype.exitRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.exitPositional_arguments(this);
  	}
  };




  MiniParser.Positional_argumentsContext = Positional_argumentsContext;

  MiniParser.prototype.positional_arguments = function() {

      var localctx = new Positional_argumentsContext(this, this._ctx, this.state);
      this.enterRule(localctx, 16, MiniParser.RULE_positional_arguments);
      var _la = 0; 
      try {
          this.enterOuterAlt(localctx, 1);
          this.state = 169;
          this.expr(0);
          this.state = 176;
          this._errHandler.sync(this);
          var _alt = this._interp.adaptivePredict(this._input,10,this._ctx);
          while(_alt!=2 && _alt!=_exports.atn.ATN.INVALID_ALT_NUMBER) {
              if(_alt===1) {
                  this.state = 170;
                  this.match(MiniParser.T__22);
                  this.state = 172;
                  this._errHandler.sync(this);
                  _la = this._input.LA(1);
                  if((((_la) & ~0x1f) == 0 && ((1 << _la) & ((1 << MiniParser.T__1) | (1 << MiniParser.T__3) | (1 << MiniParser.T__4) | (1 << MiniParser.T__5) | (1 << MiniParser.T__19) | (1 << MiniParser.T__23) | (1 << MiniParser.BOOLEAN) | (1 << MiniParser.ID) | (1 << MiniParser.INT) | (1 << MiniParser.FLOAT) | (1 << MiniParser.STRING))) !== 0)) {
                      this.state = 171;
                      this.expr(0);
                  }
           
              }
              this.state = 178;
              this._errHandler.sync(this);
              _alt = this._interp.adaptivePredict(this._input,10,this._ctx);
          }

      } catch (re) {
      	if(re instanceof _exports.error.RecognitionException) {
  	        localctx.exception = re;
  	        this._errHandler.reportError(this, re);
  	        this._errHandler.recover(this, re);
  	    } else {
  	    	throw re;
  	    }
      } finally {
          this.exitRule();
      }
      return localctx;
  };

  function Named_argumentsContext(parser, parent, invokingState) {
  	if(parent===undefined) {
  	    parent = null;
  	}
  	if(invokingState===undefined || invokingState===null) {
  		invokingState = -1;
  	}
  	_exports.ParserRuleContext.call(this, parent, invokingState);
      this.parser = parser;
      this.ruleIndex = MiniParser.RULE_named_arguments;
      return this;
  }

  Named_argumentsContext.prototype = Object.create(_exports.ParserRuleContext.prototype);
  Named_argumentsContext.prototype.constructor = Named_argumentsContext;

  Named_argumentsContext.prototype.named_argument = function(i) {
      if(i===undefined) {
          i = null;
      }
      if(i===null) {
          return this.getTypedRuleContexts(Named_argumentContext);
      } else {
          return this.getTypedRuleContext(Named_argumentContext,i);
      }
  };

  Named_argumentsContext.prototype.enterRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.enterNamed_arguments(this);
  	}
  };

  Named_argumentsContext.prototype.exitRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.exitNamed_arguments(this);
  	}
  };




  MiniParser.Named_argumentsContext = Named_argumentsContext;

  MiniParser.prototype.named_arguments = function() {

      var localctx = new Named_argumentsContext(this, this._ctx, this.state);
      this.enterRule(localctx, 18, MiniParser.RULE_named_arguments);
      var _la = 0; 
      try {
          this.enterOuterAlt(localctx, 1);
          this.state = 179;
          this.named_argument();
          this.state = 186;
          this._errHandler.sync(this);
          _la = this._input.LA(1);
          while(_la===MiniParser.T__22) {
              this.state = 180;
              this.match(MiniParser.T__22);
              this.state = 182;
              this._errHandler.sync(this);
              _la = this._input.LA(1);
              if(_la===MiniParser.ID) {
                  this.state = 181;
                  this.named_argument();
              }

              this.state = 188;
              this._errHandler.sync(this);
              _la = this._input.LA(1);
          }
      } catch (re) {
      	if(re instanceof _exports.error.RecognitionException) {
  	        localctx.exception = re;
  	        this._errHandler.reportError(this, re);
  	        this._errHandler.recover(this, re);
  	    } else {
  	    	throw re;
  	    }
      } finally {
          this.exitRule();
      }
      return localctx;
  };

  function Named_argumentContext(parser, parent, invokingState) {
  	if(parent===undefined) {
  	    parent = null;
  	}
  	if(invokingState===undefined || invokingState===null) {
  		invokingState = -1;
  	}
  	_exports.ParserRuleContext.call(this, parent, invokingState);
      this.parser = parser;
      this.ruleIndex = MiniParser.RULE_named_argument;
      return this;
  }

  Named_argumentContext.prototype = Object.create(_exports.ParserRuleContext.prototype);
  Named_argumentContext.prototype.constructor = Named_argumentContext;

  Named_argumentContext.prototype.ID = function() {
      return this.getToken(MiniParser.ID, 0);
  };

  Named_argumentContext.prototype.EQ = function() {
      return this.getToken(MiniParser.EQ, 0);
  };

  Named_argumentContext.prototype.expr = function() {
      return this.getTypedRuleContext(ExprContext,0);
  };

  Named_argumentContext.prototype.enterRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.enterNamed_argument(this);
  	}
  };

  Named_argumentContext.prototype.exitRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.exitNamed_argument(this);
  	}
  };




  MiniParser.Named_argumentContext = Named_argumentContext;

  MiniParser.prototype.named_argument = function() {

      var localctx = new Named_argumentContext(this, this._ctx, this.state);
      this.enterRule(localctx, 20, MiniParser.RULE_named_argument);
      try {
          this.enterOuterAlt(localctx, 1);
          this.state = 189;
          this.match(MiniParser.ID);
          this.state = 190;
          this.match(MiniParser.EQ);
          this.state = 191;
          this.expr(0);
           localctx.type = 'named-argument'; 
      } catch (re) {
      	if(re instanceof _exports.error.RecognitionException) {
  	        localctx.exception = re;
  	        this._errHandler.reportError(this, re);
  	        this._errHandler.recover(this, re);
  	    } else {
  	    	throw re;
  	    }
      } finally {
          this.exitRule();
      }
      return localctx;
  };

  function ArrayContext(parser, parent, invokingState) {
  	if(parent===undefined) {
  	    parent = null;
  	}
  	if(invokingState===undefined || invokingState===null) {
  		invokingState = -1;
  	}
  	_exports.ParserRuleContext.call(this, parent, invokingState);
      this.parser = parser;
      this.ruleIndex = MiniParser.RULE_array;
      return this;
  }

  ArrayContext.prototype = Object.create(_exports.ParserRuleContext.prototype);
  ArrayContext.prototype.constructor = ArrayContext;

  ArrayContext.prototype.seq = function() {
      return this.getTypedRuleContext(SeqContext,0);
  };

  ArrayContext.prototype.enterRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.enterArray(this);
  	}
  };

  ArrayContext.prototype.exitRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.exitArray(this);
  	}
  };




  MiniParser.ArrayContext = ArrayContext;

  MiniParser.prototype.array = function() {

      var localctx = new ArrayContext(this, this._ctx, this.state);
      this.enterRule(localctx, 22, MiniParser.RULE_array);
      var _la = 0; 
      try {
          this.enterOuterAlt(localctx, 1);
          this.state = 194;
          this.match(MiniParser.T__1);
          this.state = 196;
          this._errHandler.sync(this);
          _la = this._input.LA(1);
          if((((_la) & ~0x1f) == 0 && ((1 << _la) & ((1 << MiniParser.T__1) | (1 << MiniParser.T__3) | (1 << MiniParser.T__4) | (1 << MiniParser.T__5) | (1 << MiniParser.T__19) | (1 << MiniParser.T__23) | (1 << MiniParser.BOOLEAN) | (1 << MiniParser.ID) | (1 << MiniParser.INT) | (1 << MiniParser.FLOAT) | (1 << MiniParser.STRING))) !== 0)) {
              this.state = 195;
              this.seq();
          }

          this.state = 198;
          this.match(MiniParser.T__2);
           localctx.type = 'array'; 
      } catch (re) {
      	if(re instanceof _exports.error.RecognitionException) {
  	        localctx.exception = re;
  	        this._errHandler.reportError(this, re);
  	        this._errHandler.recover(this, re);
  	    } else {
  	    	throw re;
  	    }
      } finally {
          this.exitRule();
      }
      return localctx;
  };

  function ObjectContext(parser, parent, invokingState) {
  	if(parent===undefined) {
  	    parent = null;
  	}
  	if(invokingState===undefined || invokingState===null) {
  		invokingState = -1;
  	}
  	_exports.ParserRuleContext.call(this, parent, invokingState);
      this.parser = parser;
      this.ruleIndex = MiniParser.RULE_object;
      this._ID = null; 
      this.keys = []; 
      this._expr = null; 
      this.vals = []; 
      return this;
  }

  ObjectContext.prototype = Object.create(_exports.ParserRuleContext.prototype);
  ObjectContext.prototype.constructor = ObjectContext;

  ObjectContext.prototype.ID = function(i) {
  	if(i===undefined) {
  		i = null;
  	}
      if(i===null) {
          return this.getTokens(MiniParser.ID);
      } else {
          return this.getToken(MiniParser.ID, i);
      }
  };


  ObjectContext.prototype.expr = function(i) {
      if(i===undefined) {
          i = null;
      }
      if(i===null) {
          return this.getTypedRuleContexts(ExprContext);
      } else {
          return this.getTypedRuleContext(ExprContext,i);
      }
  };

  ObjectContext.prototype.enterRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.enterObject(this);
  	}
  };

  ObjectContext.prototype.exitRule = function(listener) {
      if(listener instanceof MiniListener$1 ) {
          listener.exitObject(this);
  	}
  };




  MiniParser.ObjectContext = ObjectContext;

  MiniParser.prototype.object = function() {

      var localctx = new ObjectContext(this, this._ctx, this.state);
      this.enterRule(localctx, 24, MiniParser.RULE_object);
      var _la = 0; 
      try {
          this.enterOuterAlt(localctx, 1);
          this.state = 201;
          this.match(MiniParser.T__23);
          this.state = 214;
          this._errHandler.sync(this);
          _la = this._input.LA(1);
          if(_la===MiniParser.ID) {
              this.state = 202;
              localctx._ID = this.match(MiniParser.ID);
              localctx.keys.push(localctx._ID);
              this.state = 203;
              this.match(MiniParser.T__24);
              this.state = 204;
              localctx._expr = this.expr(0);
              localctx.vals.push(localctx._expr);
              this.state = 211;
              this._errHandler.sync(this);
              _la = this._input.LA(1);
              while(_la===MiniParser.T__22) {
                  this.state = 205;
                  this.match(MiniParser.T__22);
                  this.state = 206;
                  localctx._ID = this.match(MiniParser.ID);
                  localctx.keys.push(localctx._ID);
                  this.state = 207;
                  this.match(MiniParser.T__24);
                  this.state = 208;
                  localctx._expr = this.expr(0);
                  localctx.vals.push(localctx._expr);
                  this.state = 213;
                  this._errHandler.sync(this);
                  _la = this._input.LA(1);
              }
          }

          this.state = 216;
          this.match(MiniParser.T__25);
           localctx.type = 'object'; 
      } catch (re) {
      	if(re instanceof _exports.error.RecognitionException) {
  	        localctx.exception = re;
  	        this._errHandler.reportError(this, re);
  	        this._errHandler.recover(this, re);
  	    } else {
  	    	throw re;
  	    }
      } finally {
          this.exitRule();
      }
      return localctx;
  };


  MiniParser.prototype.sempred = function(localctx, ruleIndex, predIndex) {
  	switch(ruleIndex) {
  	case 2:
  			return this.expr_sempred(localctx, predIndex);
      default:
          throw "No predicate with index:" + ruleIndex;
     }
  };

  MiniParser.prototype.expr_sempred = function(localctx, predIndex) {
  	switch(predIndex) {
  		case 0:
  			return this._input.LA(1) !== MiniParser.EQ;
  		case 1:
  			return this.precpred(this._ctx, 16);
  		case 2:
  			return this.precpred(this._ctx, 15);
  		case 3:
  			return this.precpred(this._ctx, 14);
  		case 4:
  			return this.precpred(this._ctx, 13);
  		case 5:
  			return this.precpred(this._ctx, 12);
  		case 6:
  			return this.precpred(this._ctx, 11);
  		case 7:
  			return this.precpred(this._ctx, 10);
  		case 8:
  			return this.precpred(this._ctx, 19);
  		case 9:
  			return this.precpred(this._ctx, 18);
  		case 10:
  			return this.precpred(this._ctx, 9);
  		default:
  			throw "No predicate with index:" + predIndex;
  	}
  };


  var MiniParser_2 = MiniParser;

  var MiniParser_1 = {
  	MiniParser: MiniParser_2
  };

  const { InputStream, CommonTokenStream } = _exports;
  const treeWalker = _exports.tree.ParseTreeWalker.DEFAULT;
  const MiniLexer$1 = MiniLexer_1.MiniLexer;
  const MiniParser$1 = MiniParser_1.MiniParser;

  class ErrorListener {
    constructor () {
      this.syntaxErrors = [];
    }

    syntaxError (recognizer, offendingSymbol, line, column, msg, error) {
      let row = line - 1;
      let start, token;
      if (offendingSymbol) {
        start = offendingSymbol.start;
        token = offendingSymbol.text;
      }
      this.syntaxErrors.push({
        type: 'syntax-error',
        row,
        column,
        start,
        token,
        msg,
        error
      });
    }

    reportAttemptingFullContext () {
      console.error('Attempting Full Context: ', arguments);
    }

    reportAmbiguity () {
      console.error('Ambiguity:', arguments);
    }

    reportContextSensitivity () {
      console.error('ContextSensitivity:', arguments);
    }
  }

  function parse$4 (source = '', options = {}) {
    const errorListener = new ErrorListener();
    const lexer = new MiniLexer$1(new InputStream(source));
    const parser = new MiniParser$1(new CommonTokenStream(lexer));
    parser.buildParseTrees = true;
    if (!options.debug) {
      lexer.removeErrorListeners();
      parser.removeErrorListeners();
    }
    lexer.addErrorListener(errorListener);
    parser.addErrorListener(errorListener);
    // NOTE: 'mini' is the start rule as defined in the grammar file
    let ast = parser.mini();
    let syntaxError = errorListener.syntaxErrors[0];
    return Expression$3.create(source, ast, syntaxError)
  }

  // TODO: apart from the dependency to the type system this is pretty independent now

  const INPUT_TYPES = new Set(['var', 'call']);

  class MiniContext$1 {
    constructor (host) {
      this._host = host;
    }

    get id () {
      return 'mini'
    }

    async compile (cell) {
      return this._compile(cell)
    }

    async execute (cell) {
      return this._execute(cell)
    }

    // called during evaluation of mini expressions
    async _callFunction (funcNode, args, namedArgs) {
      return this._host.callFunction(funcNode, args, namedArgs)
    }

    _compile (cell) {
      const code = cell.code;
      if (!code) {
        Object.assign(cell, {
          inputs: [],
          outputs: [],
          messages: [],
          tokens: [],
          nodes: []
        });
        return cell
      }
      let expr = parse$4(code);
      let inputs = [];
      let outputs = [];
      let tokens, nodes;
      let messages = [];
      if (expr.syntaxError) {
        messages.push({
          type: 'error',
          message: expr.syntaxError.msg
        });
      }
      const _inputs = new Set();
      expr.nodes.forEach(n => {
        if (INPUT_TYPES.has(n.type)) {
          const name = n.name;
          // making sure that inputs are not added twice
          if (!_inputs.has(name)) {
            inputs.push({name});
            _inputs.add(name);
          }
        }
      });
      if (expr.name) {
        const name = expr.name;
        outputs.push({name});
      }
      if (expr.tokens) {
        // some tokens are used for code highlighting
        // some for function documentation
        tokens = expr.tokens;
      }

      nodes = [];
      expr.nodes.forEach((n) => {
        if (n.type === 'call') {
          let args = n.args.map((a) => {
            return {
              start: a.start,
              end: a.end
            }
          }).concat(n.namedArgs.map((a) => {
            return {
              start: a.start,
              end: a.end,
              name: a.name
            }
          }));
          let node = {
            type: 'function',
            name: n.name,
            start: n.start,
            end: n.end,
            args
          };
          nodes.push(node);
        }
      });

      Object.assign(cell, {
        type: 'cell',
        code,
        inputs,
        outputs,
        messages,
        tokens,
        nodes,
        _expr: expr
      });
      return cell
    }

    async _execute (cell) {
      let expr = cell._expr || parse$4(cell.code);
      // don't evaluate the expression in presence of an syntax error
      if (expr.syntaxError) {
        return Promise.resolve(cell)
      }
      const outputName = expr.name;
      let adapter = new _MiniContextAdapter(this, cell.inputs);
      try {
        let value = await expr.evaluate(adapter);
        if (adapter.messages.length > 0) {
          cell.messages = adapter.messages;
        }
        // HACK: Mini allows only one output
        let output = {value};
        if (outputName) {
          output.name = outputName;
        }
        cell.outputs = [output];
      } catch (err) {
        cell.messages = [err];
      }
      return cell
    }
  }

  // an adapter between mini and the context
  // used to provide data, and
  class _MiniContextAdapter {
    constructor (miniContext, inputs) {
      this.miniContext = miniContext;
      let _inputs = new Map();
      inputs.forEach(i => _inputs.set(i.name, i.value));
      this.inputs = _inputs;
      this.messages = [];
    }

    resolve (name) {
      return this.inputs.get(name)
    }

    // coerce and pack
    pack (value, ctx) {
      if (ctx === 'array') {
        value = value.map(v => pack(v));
        return coerceArray(value)
      }
      return pack(value)
    }

    unpack (val) {
      // TODO: better understand if it is ok to make this robust
      // by guarding undefined values, and not obfuscating an error occurring elsewhere
      // it happened whenever undefined is returned by a called function
      if (!val) return undefined
      return val.data
    }

    async callFunction (name, args, namedArgs) {
      let func = this.inputs.get(name);
      if (!func) {
        throw new Error('Function not provided.')
      }
      let res = await this.miniContext._callFunction(func, args, namedArgs);
      this.messages.concat(res.messages);
      return res.value
    }
  }

  function importFromJSON (engine, data) {
    data.resources.forEach(res => {
      if (!res.id) res.id = uuid();
      switch (res.type) {
        case 'document': {
          engine.addDocument(res);
          break
        }
        case 'sheet': {
          engine.addSheet(res);
          break
        }
        default:
          //
          console.error('Unsupported resource type');
      }
    });
  }

  // Note: async because we gonna have an asynchronous configuration with remote contexts
  async function setupContext (config, options = {}) {
    let ContextClass = options.ContextClass || CompositeContext;
    let context = new ContextClass();
    context.configure(config);
    return context
  }

  exports.Cell = Cell$1;
  exports.CellGraph = CellGraph;
  exports.CellSymbol = CellSymbol;
  exports.CompositeContext = CompositeContext;
  exports.Engine = Engine;
  exports.MiniContext = MiniContext$1;
  exports.importFromJson = importFromJSON;
  exports.setupContext = setupContext;
  exports.CellError = CellError;
  exports.ContextError = ContextError;
  exports.GraphError = GraphError;
  exports.SyntaxError = SyntaxError$1;
  exports.UnresolvedInputError = UnresolvedInputError;
  exports.CyclicDependencyError = CyclicDependencyError;
  exports.OutputCollisionError = OutputCollisionError;
  exports.RuntimeError = RuntimeError;
  exports.ValidationError = ValidationError;
  exports.UNKNOWN = UNKNOWN;
  exports.ANALYSED = ANALYSED;
  exports.BROKEN = BROKEN;
  exports.FAILED = FAILED;
  exports.BLOCKED = BLOCKED;
  exports.WAITING = WAITING;
  exports.READY = READY;
  exports.RUNNING = RUNNING;
  exports.OK = OK;
  exports.cellStateToInteger = cellStateToInteger;
  exports.cellStateToString = cellStateToString;
  exports.createSymbol = createSymbol;
  exports.extractSymbols = extractSymbols;
  exports.transpile = transpile;
  exports.parseSymbol = parseSymbol;
  exports.qualifiedId = qualifiedId;
  exports.toIdentifier = toIdentifier;
  exports.recordTransformations = recordTransformations;
  exports.applyCellTransformations = applyCellTransformations;
  exports.transformRange = transformRange;
  exports.isExpression = isExpression;
  exports.parseValue = parseValue;
  exports.BROKEN_REF = BROKEN_REF;
  exports.ID = ID;
  exports.NAME = NAME;
  exports.CELL_ID = CELL_ID;
  exports.EXPRESSION_CELL = EXPRESSION_CELL;
  exports.EXPRESSION_CELL_RE = EXPRESSION_CELL_RE;
  exports.REF = REF;
  exports.REF_RE = REF_RE;

  Object.defineProperty(exports, '__esModule', { value: true });

})));

//# sourceMappingURL=./stencila-engine.js.map